
#include "nir.h"
#include "nir_builder.h"
#include "nir_search.h"
#include "nir_search_helpers.h"

/* What follows is NIR algebraic transform code for the following 2061
 * transforms:
 *    ('imul', 'a', '#b(is_pos_power_of_two)') => ('ishl', 'a', ('find_lsb', 'b'))
 *    ('imul', 'a@8', 128) => ('ishl', 'a', 7)
 *    ('imul', 'a@16', 32768) => ('ishl', 'a', 15)
 *    ('imul', 'a@32', 2147483648) => ('ishl', 'a', 31)
 *    ('imul', 'a@64', 9223372036854775808) => ('ishl', 'a', 63)
 *    ('imul', 'a', '#b(is_neg_power_of_two)') => ('ineg', ('ishl', 'a', ('find_lsb', ('iabs', 'b'))))
 *    ('ishl', 'a', '#b') => ('imul', 'a', ('ishl', 1, 'b'))
 *    ('imul@64', 'a', '#b(is_bitcount2)') => ('iadd', ('ishl', 'a', ('ufind_msb', 'b')), ('ishl', 'a', ('find_lsb', 'b')))
 *    ('unpack_64_2x32_split_x', ('imul_2x32_64(is_used_once)', 'a', 'b')) => ('imul', 'a', 'b')
 *    ('unpack_64_2x32_split_x', ('umul_2x32_64(is_used_once)', 'a', 'b')) => ('imul', 'a', 'b')
 *    ('imul_2x32_64', 'a', 'b') => ('pack_64_2x32_split', ('imul', 'a', 'b'), ('imul_high', 'a', 'b'))
 *    ('umul_2x32_64', 'a', 'b') => ('pack_64_2x32_split', ('imul', 'a', 'b'), ('umul_high', 'a', 'b'))
 *    ('udiv', 'a', 1) => a
 *    ('idiv', 'a', 1) => a
 *    ('umod', 'a', 1) => 0
 *    ('imod', 'a', 1) => 0
 *    ('imod', 'a', -1) => 0
 *    ('irem', 'a', 1) => 0
 *    ('irem', 'a', -1) => 0
 *    ('udiv', 'a', '#b(is_pos_power_of_two)') => ('ushr', 'a', ('find_lsb', 'b'))
 *    ('idiv', 'a', '#b(is_pos_power_of_two)') => ('imul', ('isign', 'a'), ('ushr', ('iabs', 'a'), ('find_lsb', 'b')))
 *    ('idiv', 'a', '#b(is_neg_power_of_two)') => ('ineg', ('imul', ('isign', 'a'), ('ushr', ('iabs', 'a'), ('find_lsb', ('iabs', 'b')))))
 *    ('umod', 'a', '#b(is_pos_power_of_two)') => ('iand', 'a', ('isub', 'b', 1))
 *    ('imod', 'a', '#b(is_pos_power_of_two)') => ('iand', 'a', ('isub', 'b', 1))
 *    ('imod', 'a', '#b(is_neg_power_of_two)') => ('bcsel', ('ieq', ('ior', 'a', 'b'), 'b'), 0, ('ior', 'a', 'b'))
 *    ('irem', 'a', '#b(is_pos_power_of_two)') => ('isub', 'a', ('iand', ('bcsel', ('ilt', 'a', 0), ('iadd', 'a', ('isub', 'b', 1)), 'a'), ('ineg', 'b')))
 *    ('irem', 'a', '#b(is_neg_power_of_two)') => ('irem', 'a', ('iabs', 'b'))
 *    ('~fneg', ('fneg', 'a')) => a
 *    ('ineg', ('ineg', 'a')) => a
 *    ('fabs', ('fneg', 'a')) => ('fabs', 'a')
 *    ('fabs', ('u2f', 'a')) => ('u2f', 'a')
 *    ('iabs', ('iabs', 'a')) => ('iabs', 'a')
 *    ('iabs', ('ineg', 'a')) => ('iabs', 'a')
 *    ('f2b', ('fneg', 'a')) => ('f2b', 'a')
 *    ('~fadd', 'a', 0.0) => a
 *    ('fadd(is_only_used_as_float)', 'a@16', 0.0) => a
 *    ('fadd(is_only_used_as_float)', 'a@32', 0.0) => a
 *    ('iadd', 'a', 0) => a
 *    ('iadd_sat', 'a', 0) => a
 *    ('isub_sat', 'a', 0) => a
 *    ('uadd_sat', 'a', 0) => a
 *    ('usub_sat', 'a', 0) => a
 *    ('usadd_4x8_vc4', 'a', 0) => a
 *    ('usadd_4x8_vc4', 'a', -1) => -1
 *    ('~fadd', ('fmul', 'a', 'b'), ('fmul', 'a', 'c')) => ('fmul', 'a', ('fadd', 'b', 'c'))
 *    ('~fadd', ('fmulz', 'a', 'b'), ('fmulz', 'a', 'c')) => ('fmulz', 'a', ('fadd', 'b', 'c'))
 *    ('~ffma', 'a', 'b', ('ffma(is_used_once)', 'a', 'c', 'd')) => ('ffma', 'a', ('fadd', 'b', 'c'), 'd')
 *    ('~ffma', 'a', 'b', ('fmul(is_used_once)', 'a', 'c')) => ('fmul', 'a', ('fadd', 'b', 'c'))
 *    ('~fadd', ('fmul(is_used_once)', 'a', 'b'), ('ffma(is_used_once)', 'a', 'c', 'd')) => ('ffma', 'a', ('fadd', 'b', 'c'), 'd')
 *    ('~ffma', 'a', ('fmul(is_used_once)', 'b', 'c'), ('fmul(is_used_once)', 'b', 'd')) => ('fmul', 'b', ('ffma', 'a', 'c', 'd'))
 *    ('~ffmaz', 'a', 'b', ('ffmaz(is_used_once)', 'a', 'c', 'd')) => ('ffmaz', 'a', ('fadd', 'b', 'c'), 'd')
 *    ('~ffmaz', 'a', 'b', ('fmulz(is_used_once)', 'a', 'c')) => ('fmulz', 'a', ('fadd', 'b', 'c'))
 *    ('~fadd', ('fmulz(is_used_once)', 'a', 'b'), ('ffmaz(is_used_once)', 'a', 'c', 'd')) => ('ffmaz', 'a', ('fadd', 'b', 'c'), 'd')
 *    ('~ffmaz', 'a', ('fmulz(is_used_once)', 'b', 'c'), ('fmulz(is_used_once)', 'b', 'd')) => ('fmulz', 'b', ('ffmaz', 'a', 'c', 'd'))
 *    ('iadd', ('imul', 'a', 'b'), ('imul', 'a', 'c')) => ('imul', 'a', ('iadd', 'b', 'c'))
 *    ('iadd', ('ishl', 'b', 'a'), ('ishl', 'c', 'a')) => ('ishl', ('iadd', 'b', 'c'), 'a')
 *    ('iand', ('ior', 'a', 'b'), ('ior', 'a', 'c')) => ('ior', 'a', ('iand', 'b', 'c'))
 *    ('ior', ('iand', 'a', 'b'), ('iand', 'a', 'c')) => ('iand', 'a', ('ior', 'b', 'c'))
 *    ('ieq', ('iand', 'a', '#b(is_pos_power_of_two)'), 'b') => ('ine', ('iand', 'a', 'b'), 0)
 *    ('ine', ('iand', 'a', '#b(is_pos_power_of_two)'), 'b') => ('ieq', ('iand', 'a', 'b'), 0)
 *    ('~fadd', ('fneg', 'a'), 'a') => 0.0
 *    ('iadd', ('ineg', 'a'), 'a') => 0
 *    ('iadd', ('ineg', 'a'), ('iadd', 'a', 'b')) => b
 *    ('iadd', 'a', ('iadd', ('ineg', 'a'), 'b')) => b
 *    ('~fadd', ('fneg', 'a'), ('fadd', 'a', 'b')) => b
 *    ('~fadd', 'a', ('fadd', ('fneg', 'a'), 'b')) => b
 *    ('fadd', ('fsat', 'a'), ('fsat', ('fneg', 'a'))) => ('fsat', ('fabs', 'a'))
 *    ('~fmul', 'a', 0.0) => 0.0
 *    ('fmul', 'a@16', 0.0) => 0.0
 *    ('fmul', 'a@32', 0.0) => 0.0
 *    ('fmulz', 'a', 0.0) => 0.0
 *    ('fmulz', 'a', 'b(is_finite_not_zero)') => ('fmul', 'a', 'b')
 *    ('fmulz', 'a(is_finite)', 'b(is_finite)') => ('fmul', 'a', 'b')
 *    ('fmulz', 'a', 'a') => ('fmul', 'a', 'a')
 *    ('ffmaz', 'a', 'b(is_finite_not_zero)', 'c') => ('ffma', 'a', 'b', 'c')
 *    ('ffmaz', 'a(is_finite)', 'b(is_finite)', 'c') => ('ffma', 'a', 'b', 'c')
 *    ('ffmaz', 'a', 'a', 'b') => ('ffma', 'a', 'a', 'b')
 *    ('imul', 'a', 0) => 0
 *    ('umul_unorm_4x8_vc4', 'a', 0) => 0
 *    ('umul_unorm_4x8_vc4', 'a', -1) => a
 *    ('~fmul', 'a', 1.0) => a
 *    ('~fmulz', 'a', 1.0) => a
 *    ('fmul(is_only_used_as_float)', 'a', 1.0) => a
 *    ('imul', 'a', 1) => a
 *    ('fmul', 'a', -1.0) => ('fneg', 'a')
 *    ('imul', 'a', -1) => ('ineg', 'a')
 *    ('fmul', ('fsign', 'a'), ('fmul', 'a', 'a')) => ('fmul', ('fabs', 'a'), 'a')
 *    ('fmul', ('fmul', ('fsign', 'a'), 'a'), 'a') => ('fmul', ('fabs', 'a'), 'a')
 *    ('~ffma', 0.0, 'a', 'b') => b
 *    ('ffma@16(is_only_used_as_float)', 0.0, 'a', 'b') => b
 *    ('ffma@32(is_only_used_as_float)', 0.0, 'a', 'b') => b
 *    ('ffmaz', 0.0, 'a', 'b') => ('fadd', 0.0, 'b')
 *    ('~ffma', 'a', 'b', 0.0) => ('fmul', 'a', 'b')
 *    ('ffma@16', 'a', 'b', 0.0) => ('fmul', 'a', 'b')
 *    ('ffma@32', 'a', 'b', 0.0) => ('fmul', 'a', 'b')
 *    ('ffmaz', 'a', 'b', 0.0) => ('fmulz', 'a', 'b')
 *    ('ffma', 1.0, 'a', 'b') => ('fadd', 'a', 'b')
 *    ('ffmaz', 1.0, 'a', 'b') => ('fadd', 'a', 'b')
 *    ('ffma', -1.0, 'a', 'b') => ('fadd', ('fneg', 'a'), 'b')
 *    ('ffmaz', -1.0, 'a', 'b') => ('fadd', ('fneg', 'a'), 'b')
 *    ('~ffma', '#a', '#b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c')
 *    ('~ffmaz', '#a', '#b', 'c') => ('fadd', ('fmulz', 'a', 'b'), 'c')
 *    ('~flrp', 'a', 'b', 0.0) => a
 *    ('~flrp', 'a', 'b', 1.0) => b
 *    ('~flrp', 'a', 'a', 'b') => a
 *    ('~flrp', 0.0, 'a', 'b') => ('fmul', 'a', 'b')
 *    ('~flrp', 'a', ('fadd(is_used_once)', 'a', 'b'), 'c') => ('fadd', ('fmul', 'b', 'c'), 'a')
 *    ('sdot_4x8_iadd', 'a', 0, 'b') => b
 *    ('udot_4x8_uadd', 'a', 0, 'b') => b
 *    ('sdot_4x8_iadd_sat', 'a', 0, 'b') => b
 *    ('udot_4x8_uadd_sat', 'a', 0, 'b') => b
 *    ('sdot_2x16_iadd', 'a', 0, 'b') => b
 *    ('udot_2x16_uadd', 'a', 0, 'b') => b
 *    ('sdot_2x16_iadd_sat', 'a', 0, 'b') => b
 *    ('udot_2x16_uadd_sat', 'a', 0, 'b') => b
 *    ('sudot_4x8_iadd', 'a', 0, 'b') => b
 *    ('sudot_4x8_iadd', 0, 'a', 'b') => b
 *    ('sudot_4x8_iadd_sat', 'a', 0, 'b') => b
 *    ('sudot_4x8_iadd_sat', 0, 'a', 'b') => b
 *    ('iadd', ('sdot_4x8_iadd(is_used_once)', 'a', 'b', '#c'), '#d') => ('sdot_4x8_iadd', 'a', 'b', ('iadd', 'c', 'd'))
 *    ('iadd', ('udot_4x8_uadd(is_used_once)', 'a', 'b', '#c'), '#d') => ('udot_4x8_uadd', 'a', 'b', ('iadd', 'c', 'd'))
 *    ('iadd', ('sudot_4x8_iadd(is_used_once)', 'a', 'b', '#c'), '#d') => ('sudot_4x8_iadd', 'a', 'b', ('iadd', 'c', 'd'))
 *    ('iadd', ('sdot_2x16_iadd(is_used_once)', 'a', 'b', '#c'), '#d') => ('sdot_2x16_iadd', 'a', 'b', ('iadd', 'c', 'd'))
 *    ('iadd', ('udot_2x16_uadd(is_used_once)', 'a', 'b', '#c'), '#d') => ('udot_2x16_uadd', 'a', 'b', ('iadd', 'c', 'd'))
 *    ('iadd', ('sdot_4x8_iadd', 'a(is_not_const)', 'b', 0), 'c') => ('sdot_4x8_iadd', 'a', 'b', 'c')
 *    ('iadd', ('udot_4x8_uadd', 'a(is_not_const)', 'b', 0), 'c') => ('udot_4x8_uadd', 'a', 'b', 'c')
 *    ('iadd', ('sudot_4x8_iadd', 'a(is_not_const)', 'b', 0), 'c') => ('sudot_4x8_iadd', 'a', 'b', 'c')
 *    ('iadd', ('sudot_4x8_iadd', 'a', 'b(is_not_const)', 0), 'c') => ('sudot_4x8_iadd', 'a', 'b', 'c')
 *    ('iadd', ('sdot_2x16_iadd', 'a(is_not_const)', 'b', 0), 'c') => ('sdot_2x16_iadd', 'a', 'b', 'c')
 *    ('iadd', ('udot_2x16_uadd', 'a(is_not_const)', 'b', 0), 'c') => ('udot_2x16_uadd', 'a', 'b', 'c')
 *    ('sdot_4x8_iadd', '#a', '#b', 'c(is_not_const)') => ('iadd', ('sdot_4x8_iadd', 'a', 'b', 0), 'c')
 *    ('udot_4x8_uadd', '#a', '#b', 'c(is_not_const)') => ('iadd', ('udot_4x8_uadd', 'a', 'b', 0), 'c')
 *    ('sudot_4x8_iadd', '#a', '#b', 'c(is_not_const)') => ('iadd', ('sudot_4x8_iadd', 'a', 'b', 0), 'c')
 *    ('sdot_2x16_iadd', '#a', '#b', 'c(is_not_const)') => ('iadd', ('sdot_2x16_iadd', 'a', 'b', 0), 'c')
 *    ('udot_2x16_uadd', '#a', '#b', 'c(is_not_const)') => ('iadd', ('udot_2x16_uadd', 'a', 'b', 0), 'c')
 *    ('sdot_4x8_iadd_sat', '#a', '#b', 'c(is_not_const)') => ('iadd_sat', ('sdot_4x8_iadd', 'a', 'b', 0), 'c')
 *    ('udot_4x8_uadd_sat', '#a', '#b', 'c(is_not_const)') => ('uadd_sat', ('udot_4x8_uadd', 'a', 'b', 0), 'c')
 *    ('sudot_4x8_iadd_sat', '#a', '#b', 'c(is_not_const)') => ('iadd_sat', ('sudot_4x8_iadd', 'a', 'b', 0), 'c')
 *    ('sdot_2x16_iadd_sat', '#a', '#b', 'c(is_not_const)') => ('iadd_sat', ('sdot_2x16_iadd', 'a', 'b', 0), 'c')
 *    ('udot_2x16_uadd_sat', '#a', '#b', 'c(is_not_const)') => ('uadd_sat', ('udot_2x16_uadd', 'a', 'b', 0), 'c')
 *    ('fmul@32', ('bcsel', ('feq', 'b', 0.0, 'ignore_exact'), 0.0, 'a'), ('bcsel', ('feq', 'a', 0.0, 'ignore_exact'), 0.0, 'b')) => ('fmulz', 'a', 'b')
 *    ('fmul@32', 'a', ('bcsel', ('feq', 'a', 0.0, 'ignore_exact'), 0.0, '#b(is_not_const_zero)')) => ('fmulz', 'a', 'b')
 *    ('ffma@32', ('bcsel', ('feq', 'b', 0.0, 'ignore_exact'), 0.0, 'a'), ('bcsel', ('feq', 'a', 0.0, 'ignore_exact'), 0.0, 'b'), 'c') => ('ffmaz', 'a', 'b', 'c')
 *    ('ffma@32', 'a', ('bcsel', ('feq', 'a', 0.0, 'ignore_exact'), 0.0, '#b(is_not_const_zero)'), 'c') => ('ffmaz', 'a', 'b', 'c')
 *    ('bcsel', ('feq', 'b', 0.0, 'ignore_exact'), 1.0, ('fexp2', ('fmul@32', 'a', 'b'))) => ('fexp2', ('fmulz', 'a', 'b'))
 *    ('sdot_4x8_iadd', 'a', 'b', 'c') => ('iadd', ('iadd', ('iadd', ('imul', ('extract_i8', 'a', 0), ('extract_i8', 'b', 0)), ('imul', ('extract_i8', 'a', 1), ('extract_i8', 'b', 1))), ('iadd', ('imul', ('extract_i8', 'a', 2), ('extract_i8', 'b', 2)), ('imul', ('extract_i8', 'a', 3), ('extract_i8', 'b', 3)))), 'c')
 *    ('udot_4x8_uadd', 'a', 'b', 'c') => ('iadd', ('iadd', ('iadd', ('imul', ('extract_u8', 'a', 0), ('extract_u8', 'b', 0)), ('imul', ('extract_u8', 'a', 1), ('extract_u8', 'b', 1))), ('iadd', ('imul', ('extract_u8', 'a', 2), ('extract_u8', 'b', 2)), ('imul', ('extract_u8', 'a', 3), ('extract_u8', 'b', 3)))), 'c')
 *    ('sudot_4x8_iadd', 'a', 'b', 'c') => ('iadd', ('iadd', ('iadd', ('imul', ('extract_i8', 'a', 0), ('extract_u8', 'b', 0)), ('imul', ('extract_i8', 'a', 1), ('extract_u8', 'b', 1))), ('iadd', ('imul', ('extract_i8', 'a', 2), ('extract_u8', 'b', 2)), ('imul', ('extract_i8', 'a', 3), ('extract_u8', 'b', 3)))), 'c')
 *    ('sdot_2x16_iadd', 'a', 'b', 'c') => ('iadd', ('iadd', ('imul', ('extract_i16', 'a', 0), ('extract_i16', 'b', 0)), ('imul', ('extract_i16', 'a', 1), ('extract_i16', 'b', 1))), 'c')
 *    ('udot_2x16_uadd', 'a', 'b', 'c') => ('iadd', ('iadd', ('imul', ('extract_u16', 'a', 0), ('extract_u16', 'b', 0)), ('imul', ('extract_u16', 'a', 1), ('extract_u16', 'b', 1))), 'c')
 *    ('udot_4x8_uadd_sat', 'a', 'b', '#c(is_ult_0xfffc07fc)') => ('udot_4x8_uadd', 'a', 'b', 'c')
 *    ('udot_4x8_uadd_sat', 'a', 'b', 'c') => ('uadd_sat', ('iadd', ('iadd', ('imul', ('extract_u8', 'a', 0), ('extract_u8', 'b', 0)), ('imul', ('extract_u8', 'a', 1), ('extract_u8', 'b', 1))), ('iadd', ('imul', ('extract_u8', 'a', 2), ('extract_u8', 'b', 2)), ('imul', ('extract_u8', 'a', 3), ('extract_u8', 'b', 3)))), 'c')
 *    ('sdot_4x8_iadd_sat', 'a', 'b', 'c') => ('iadd_sat', ('iadd', ('iadd', ('imul', ('extract_i8', 'a', 0), ('extract_i8', 'b', 0)), ('imul', ('extract_i8', 'a', 1), ('extract_i8', 'b', 1))), ('iadd', ('imul', ('extract_i8', 'a', 2), ('extract_i8', 'b', 2)), ('imul', ('extract_i8', 'a', 3), ('extract_i8', 'b', 3)))), 'c')
 *    ('sudot_4x8_iadd_sat', 'a', 'b', 'c') => ('iadd_sat', ('iadd', ('iadd', ('imul', ('extract_i8', 'a', 0), ('extract_u8', 'b', 0)), ('imul', ('extract_i8', 'a', 1), ('extract_u8', 'b', 1))), ('iadd', ('imul', ('extract_i8', 'a', 2), ('extract_u8', 'b', 2)), ('imul', ('extract_i8', 'a', 3), ('extract_u8', 'b', 3)))), 'c')
 *    ('udot_2x16_uadd_sat', 'a', 'b', 'c') => ('uadd_sat', ('iadd', ('imul', ('extract_u16', 'a', 0), ('extract_u16', 'b', 0)), ('imul', ('extract_u16', 'a', 1), ('extract_u16', 'b', 1))), 'c')
 *    ('sdot_2x16_iadd_sat', 'a', 'b', 'c') => ('iadd_sat', ('iadd', ('imul', ('extract_i16', 'a', 0), ('extract_i16', 'b', 0)), ('imul', ('extract_i16', 'a', 1), ('extract_i16', 'b', 1))), 'c')
 *    ('~flrp@16', 'a', 'b', ('b2f', 'c@1')) => ('bcsel', 'c', 'b', 'a')
 *    ('~flrp@16', 'a', ('fadd', 'a', 'b'), 'c') => ('fadd', ('fmul', 'b', 'c'), 'a')
 *    ('~flrp@16', ('fadd(is_used_once)', 'a', 'b'), ('fadd(is_used_once)', 'a', 'c'), 'd') => ('fadd', ('flrp', 'b', 'c', 'd'), 'a')
 *    ('~flrp@16', 'a', ('fmul(is_used_once)', 'a', 'b'), 'c') => ('fmul', ('flrp', 1.0, 'b', 'c'), 'a')
 *    ('~fadd@16', ('fmul', 'a', ('fadd', 1.0, ('fneg', 'c'))), ('fmul', 'b', 'c')) => ('flrp', 'a', 'b', 'c')
 *    ('~fadd@16', ('fmul', 'a', ('fsat', ('fadd', 1.0, ('fneg', 'c')))), ('fmul', 'b', ('fsat', 'c'))) => ('flrp', 'a', 'b', ('fsat', 'c'))
 *    ('~fadd@16', 'a', ('fmul', 'c', ('fadd', 'b', ('fneg', 'a')))) => ('flrp', 'a', 'b', 'c')
 *    ('~fadd@16', ('fmul', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1')))), ('fmul', 'b', ('b2f', 'c'))) => ('bcsel', 'c', 'b', 'a')
 *    ('~fadd@16', 'a', ('fmul', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')))) => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@16', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1'))), ('fmul', 'b', ('b2f', 'c@1'))) => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@16', 'b', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), ('b2f', 'c@1'), 'a')) => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@16', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')), 'a') => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@16', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), 'b', 'd'), ('fmul', 'a', 'b')) => ('bcsel', 'c', 'd', ('fmul', 'a', 'b'))
 *    ('~fadd@16', 1.0, ('fneg', ('fmul', ('fadd', 1.0, ('fneg', 'a')), ('fadd', 1.0, ('fneg', 'b'))))) => ('flrp', 'b', 1.0, 'a')
 *    ('~flrp@32', 'a', 'b', ('b2f', 'c@1')) => ('bcsel', 'c', 'b', 'a')
 *    ('~flrp@32', 'a', ('fadd', 'a', 'b'), 'c') => ('fadd', ('fmul', 'b', 'c'), 'a')
 *    ('~flrp@32', ('fadd(is_used_once)', 'a', 'b'), ('fadd(is_used_once)', 'a', 'c'), 'd') => ('fadd', ('flrp', 'b', 'c', 'd'), 'a')
 *    ('~flrp@32', 'a', ('fmul(is_used_once)', 'a', 'b'), 'c') => ('fmul', ('flrp', 1.0, 'b', 'c'), 'a')
 *    ('~fadd@32', ('fmul', 'a', ('fadd', 1.0, ('fneg', 'c'))), ('fmul', 'b', 'c')) => ('flrp', 'a', 'b', 'c')
 *    ('~fadd@32', ('fmul', 'a', ('fsat', ('fadd', 1.0, ('fneg', 'c')))), ('fmul', 'b', ('fsat', 'c'))) => ('flrp', 'a', 'b', ('fsat', 'c'))
 *    ('~fadd@32', 'a', ('fmul', 'c', ('fadd', 'b', ('fneg', 'a')))) => ('flrp', 'a', 'b', 'c')
 *    ('~fadd@32', ('fmul', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1')))), ('fmul', 'b', ('b2f', 'c'))) => ('bcsel', 'c', 'b', 'a')
 *    ('~fadd@32', 'a', ('fmul', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')))) => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@32', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1'))), ('fmul', 'b', ('b2f', 'c@1'))) => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@32', 'b', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), ('b2f', 'c@1'), 'a')) => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@32', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')), 'a') => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@32', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), 'b', 'd'), ('fmul', 'a', 'b')) => ('bcsel', 'c', 'd', ('fmul', 'a', 'b'))
 *    ('~fadd@32', 1.0, ('fneg', ('fmul', ('fadd', 1.0, ('fneg', 'a')), ('fadd', 1.0, ('fneg', 'b'))))) => ('flrp', 'b', 1.0, 'a')
 *    ('~flrp@64', 'a', 'b', ('b2f', 'c@1')) => ('bcsel', 'c', 'b', 'a')
 *    ('~flrp@64', 'a', ('fadd', 'a', 'b'), 'c') => ('fadd', ('fmul', 'b', 'c'), 'a')
 *    ('~flrp@64', ('fadd(is_used_once)', 'a', 'b'), ('fadd(is_used_once)', 'a', 'c'), 'd') => ('fadd', ('flrp', 'b', 'c', 'd'), 'a')
 *    ('~flrp@64', 'a', ('fmul(is_used_once)', 'a', 'b'), 'c') => ('fmul', ('flrp', 1.0, 'b', 'c'), 'a')
 *    ('~fadd@64', ('fmul', 'a', ('fadd', 1.0, ('fneg', 'c'))), ('fmul', 'b', 'c')) => ('flrp', 'a', 'b', 'c')
 *    ('~fadd@64', ('fmul', 'a', ('fsat', ('fadd', 1.0, ('fneg', 'c')))), ('fmul', 'b', ('fsat', 'c'))) => ('flrp', 'a', 'b', ('fsat', 'c'))
 *    ('~fadd@64', 'a', ('fmul', 'c', ('fadd', 'b', ('fneg', 'a')))) => ('flrp', 'a', 'b', 'c')
 *    ('~fadd@64', ('fmul', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1')))), ('fmul', 'b', ('b2f', 'c'))) => ('bcsel', 'c', 'b', 'a')
 *    ('~fadd@64', 'a', ('fmul', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')))) => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@64', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1'))), ('fmul', 'b', ('b2f', 'c@1'))) => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@64', 'b', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), ('b2f', 'c@1'), 'a')) => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@64', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')), 'a') => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@64', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), 'b', 'd'), ('fmul', 'a', 'b')) => ('bcsel', 'c', 'd', ('fmul', 'a', 'b'))
 *    ('~fadd@64', 1.0, ('fneg', ('fmul', ('fadd', 1.0, ('fneg', 'a')), ('fadd', 1.0, ('fneg', 'b'))))) => ('flrp', 'b', 1.0, 'a')
 *    ('~flrp', ('fmul(is_used_once)', 'a', 'b'), ('fmul(is_used_once)', 'a', 'c'), 'd') => ('fmul', ('flrp', 'b', 'c', 'd'), 'a')
 *    ('~flrp', 'a', 0.0, 'c') => ('fadd', ('fmul', ('fneg', 'a'), 'c'), 'a')
 *    ('ftrunc@16', 'a') => ('bcsel', ('flt', 'a', 0.0), ('fneg', ('ffloor', ('fabs', 'a'))), ('ffloor', ('fabs', 'a')))
 *    ('ftrunc@32', 'a') => ('bcsel', ('flt', 'a', 0.0), ('fneg', ('ffloor', ('fabs', 'a'))), ('ffloor', ('fabs', 'a')))
 *    ('ftrunc@64', 'a') => ('bcsel', ('flt', 'a', 0.0), ('fneg', ('ffloor', ('fabs', 'a'))), ('ffloor', ('fabs', 'a')))
 *    ('ffloor@16', 'a') => ('fsub', 'a', ('ffract', 'a'))
 *    ('ffloor@32', 'a') => ('fsub', 'a', ('ffract', 'a'))
 *    ('ffloor@64', 'a') => ('fsub', 'a', ('ffract', 'a'))
 *    ('fadd@16', 'a', ('fadd@16', 'b', ('fneg', ('ffract', 'a')))) => ('fadd@16', 'b', ('ffloor', 'a'))
 *    ('fadd@32', 'a', ('fadd@32', 'b', ('fneg', ('ffract', 'a')))) => ('fadd@32', 'b', ('ffloor', 'a'))
 *    ('fadd@64', 'a', ('fadd@64', 'b', ('fneg', ('ffract', 'a')))) => ('fadd@64', 'b', ('ffloor', 'a'))
 *    ('fadd@16', 'a', ('fneg', ('ffract', 'a'))) => ('ffloor', 'a')
 *    ('fadd@32', 'a', ('fneg', ('ffract', 'a'))) => ('ffloor', 'a')
 *    ('fadd@64', 'a', ('fneg', ('ffract', 'a'))) => ('ffloor', 'a')
 *    ('ffract@16', 'a') => ('fsub', 'a', ('ffloor', 'a'))
 *    ('ffract@32', 'a') => ('fsub', 'a', ('ffloor', 'a'))
 *    ('ffract@64', 'a') => ('fsub', 'a', ('ffloor', 'a'))
 *    ('fceil', 'a') => ('fneg', ('ffloor', ('fneg', 'a')))
 *    ('ffma@16', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c')
 *    ('ffma@32', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c')
 *    ('ffma@64', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c')
 *    ('ffmaz', 'a', 'b', 'c') => ('fadd', ('fmulz', 'a', 'b'), 'c')
 *    ('~ffma@16', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c')
 *    ('~ffma@32', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c')
 *    ('~ffma@64', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c')
 *    ('~ffmaz', 'a', 'b', 'c') => ('fadd', ('fmulz', 'a', 'b'), 'c')
 *    ('~fmul', ('fadd', ('iand', ('ineg', ('b2i', 'a@bool')), ('fmul', 'b', 'c')), '#d'), '#e') => ('bcsel', 'a', ('fmul', ('fadd', ('fmul', 'b', 'c'), 'd'), 'e'), ('fmul', 'd', 'e'))
 *    ('fdph', 'a', 'b') => ('fdot4', ('vec4', 'a.x', 'a.y', 'a.z', 1.0), 'b')
 *    ('fdot4', ('vec4', 'a', 'b', 'c', 1.0), 'd') => ('fdph', ('vec3', 'a', 'b', 'c'), 'd')
 *    ('fdot4', ('vec4', 'a', 0.0, 0.0, 0.0), 'b') => ('fmul', 'a', 'b')
 *    ('fdot4', ('vec4', 'a', 'b', 0.0, 0.0), 'c') => ('fdot2', ('vec2', 'a', 'b'), 'c')
 *    ('fdot4', ('vec4', 'a', 'b', 'c', 0.0), 'd') => ('fdot3', ('vec3', 'a', 'b', 'c'), 'd')
 *    ('fdot3', ('vec3', 'a', 0.0, 0.0), 'b') => ('fmul', 'a', 'b')
 *    ('fdot3', ('vec3', 'a', 'b', 0.0), 'c') => ('fdot2', ('vec2', 'a', 'b'), 'c')
 *    ('fdot2', ('vec2', 'a', 0.0), 'b') => ('fmul', 'a', 'b')
 *    ('fdot2', 'a', 1.0) => ('fadd', 'a.x', 'a.y')
 *    ('fdot2', 'a', 'b') => ('fsum2', ('fmul', 'a', 'b'))
 *    ('fdot3', 'a', 'b') => ('fsum3', ('fmul', 'a', 'b'))
 *    ('fdot4', 'a', 'b') => ('fsum4', ('fmul', 'a', 'b'))
 *    ('fsum2', 'a') => ('fadd', 'a.x', 'a.y')
 *    ('~fadd', ('fneg(is_used_once)', ('fsat(is_used_once)', 'a(is_not_fmul)')), 1.0) => ('fsat', ('fadd', 1.0, ('fneg', 'a')))
 *    ('ishl', ('iadd', ('imul', 'a', '#b'), '#c'), '#d') => ('iadd', ('imul', 'a', ('ishl', 'b', 'd')), ('ishl', 'c', 'd'))
 *    ('ishl', ('imul', 'a', '#b'), '#c') => ('imul', 'a', ('ishl', 'b', 'c'))
 *    ('ishl@8', ('ishl@8', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 7), ('iand', 'c', 7)), 8), ('ishl@8', 'a', ('iadd', 'b', 'c')), 0)
 *    ('ushr@8', ('ushr@8', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 7), ('iand', 'c', 7)), 8), ('ushr@8', 'a', ('iadd', 'b', 'c')), 0)
 *    ('ishr@8', ('ishr@8', 'a', '#b'), '#c') => ('ishr@8', 'a', ('imin', ('iadd', ('iand', 'b', 7), ('iand', 'c', 7)), 7))
 *    ('ishl@16', ('ishl@16', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 15), ('iand', 'c', 15)), 16), ('ishl@16', 'a', ('iadd', 'b', 'c')), 0)
 *    ('ushr@16', ('ushr@16', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 15), ('iand', 'c', 15)), 16), ('ushr@16', 'a', ('iadd', 'b', 'c')), 0)
 *    ('ishr@16', ('ishr@16', 'a', '#b'), '#c') => ('ishr@16', 'a', ('imin', ('iadd', ('iand', 'b', 15), ('iand', 'c', 15)), 15))
 *    ('ishl@32', ('ishl@32', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 31), ('iand', 'c', 31)), 32), ('ishl@32', 'a', ('iadd', 'b', 'c')), 0)
 *    ('ushr@32', ('ushr@32', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 31), ('iand', 'c', 31)), 32), ('ushr@32', 'a', ('iadd', 'b', 'c')), 0)
 *    ('ishr@32', ('ishr@32', 'a', '#b'), '#c') => ('ishr@32', 'a', ('imin', ('iadd', ('iand', 'b', 31), ('iand', 'c', 31)), 31))
 *    ('ishl@64', ('ishl@64', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 63), ('iand', 'c', 63)), 64), ('ishl@64', 'a', ('iadd', 'b', 'c')), 0)
 *    ('ushr@64', ('ushr@64', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 63), ('iand', 'c', 63)), 64), ('ushr@64', 'a', ('iadd', 'b', 'c')), 0)
 *    ('ishr@64', ('ishr@64', 'a', '#b'), '#c') => ('ishr@64', 'a', ('imin', ('iadd', ('iand', 'b', 63), ('iand', 'c', 63)), 63))
 *    ('ishl', ('ushr', 'a@8', '#b'), 'b') => ('iand', 'a', ('ishl', 255, 'b'))
 *    ('ishl', ('ishr', 'a@8', '#b'), 'b') => ('iand', 'a', ('ishl', 255, 'b'))
 *    ('ushr', ('ishl', 'a@8', '#b'), 'b') => ('iand', 'a', ('ushr', 255, 'b'))
 *    ('ishl', ('ushr', 'a@16', '#b'), 'b') => ('iand', 'a', ('ishl', 65535, 'b'))
 *    ('ishl', ('ishr', 'a@16', '#b'), 'b') => ('iand', 'a', ('ishl', 65535, 'b'))
 *    ('ushr', ('ishl', 'a@16', '#b'), 'b') => ('iand', 'a', ('ushr', 65535, 'b'))
 *    ('ishl', ('ushr', 'a@32', '#b'), 'b') => ('iand', 'a', ('ishl', 4294967295, 'b'))
 *    ('ishl', ('ishr', 'a@32', '#b'), 'b') => ('iand', 'a', ('ishl', 4294967295, 'b'))
 *    ('ushr', ('ishl', 'a@32', '#b'), 'b') => ('iand', 'a', ('ushr', 4294967295, 'b'))
 *    ('ishl', ('ushr', 'a@64', '#b'), 'b') => ('iand', 'a', ('ishl', 18446744073709551615, 'b'))
 *    ('ishl', ('ishr', 'a@64', '#b'), 'b') => ('iand', 'a', ('ishl', 18446744073709551615, 'b'))
 *    ('ushr', ('ishl', 'a@64', '#b'), 'b') => ('iand', 'a', ('ushr', 18446744073709551615, 'b'))
 *    ('iand', ('ishl', 'a@32', '#b(is_first_5_bits_uge_2)'), -4) => ('ishl', 'a', 'b')
 *    ('iand', ('imul', 'a', '#b(is_unsigned_multiple_of_4)'), -4) => ('imul', 'a', 'b')
 *    ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_2)'), 4294967294) => ('iadd', ('iand', 'a', 4294967294), 'b')
 *    ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_4)'), 4294967292) => ('iadd', ('iand', 'a', 4294967292), 'b')
 *    ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_8)'), 4294967288) => ('iadd', ('iand', 'a', 4294967288), 'b')
 *    ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_16)'), 4294967280) => ('iadd', ('iand', 'a', 4294967280), 'b')
 *    ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_32)'), 4294967264) => ('iadd', ('iand', 'a', 4294967264), 'b')
 *    ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_64)'), 4294967232) => ('iadd', ('iand', 'a', 4294967232), 'b')
 *    ('ishl@32', ('iand', 'a@32', 2147483647), 1) => ('ishl', 'a', 1)
 *    ('iand', ('ishl', 'a@32', 1), 4294967294) => ('ishl', 'a', 1)
 *    ('iand', ('ushr', 'a@32', 1), 2147483647) => ('ushr', 'a', 1)
 *    ('ishl@32', ('iand', 'a@32', 1073741823), 2) => ('ishl', 'a', 2)
 *    ('iand', ('ishl', 'a@32', 2), 4294967292) => ('ishl', 'a', 2)
 *    ('iand', ('ushr', 'a@32', 2), 1073741823) => ('ushr', 'a', 2)
 *    ('ishl@32', ('iand', 'a@32', 65535), 16) => ('ishl', 'a', 16)
 *    ('iand', ('ishl', 'a@32', 16), 4294901760) => ('ishl', 'a', 16)
 *    ('iand', ('ushr', 'a@32', 16), 65535) => ('ushr', 'a', 16)
 *    ('ishl@32', ('iand', 'a@32', 255), 24) => ('ishl', 'a', 24)
 *    ('iand', ('ishl', 'a@32', 24), 4278190080) => ('ishl', 'a', 24)
 *    ('iand', ('ushr', 'a@32', 24), 255) => ('ushr', 'a', 24)
 *    ('ishl', ('iadd', 'a', '#b'), '#c') => ('iadd', ('ishl', 'a', 'c'), ('ishl', 'b', 'c'))
 *    ('imul', ('iadd(is_used_once)', 'a', '#b'), '#c') => ('iadd', ('imul', 'a', 'c'), ('imul', 'b', 'c'))
 *    ('imul', ('iadd(is_used_once)', ('iadd(is_used_once)', 'a', '#b'), 'c'), '#d') => ('iadd', ('imul', ('iadd', 'a', 'c'), 'd'), ('imul', 'b', 'd'))
 *    ('ishl', ('iadd(is_used_once)', ('iadd(is_used_once)', 'a', '#b'), 'c'), '#d') => ('iadd', ('ishl', ('iadd', 'a', 'c'), 'd'), ('ishl', 'b', 'd'))
 *    ('inot', ('flt(is_used_once)', 'a(is_a_number)', 'b(is_a_number)')) => ('fge', 'a', 'b')
 *    ('inot', ('fge(is_used_once)', 'a(is_a_number)', 'b(is_a_number)')) => ('flt', 'a', 'b')
 *    ('inot', ('feq(is_used_once)', 'a', 'b')) => ('fneu', 'a', 'b')
 *    ('inot', ('fneu(is_used_once)', 'a', 'b')) => ('feq', 'a', 'b')
 *    ('inot', ('ilt(is_used_once)', 'a', 'b')) => ('ige', 'a', 'b')
 *    ('inot', ('ult(is_used_once)', 'a', 'b')) => ('uge', 'a', 'b')
 *    ('inot', ('ige(is_used_once)', 'a', 'b')) => ('ilt', 'a', 'b')
 *    ('inot', ('uge(is_used_once)', 'a', 'b')) => ('ult', 'a', 'b')
 *    ('inot', ('ieq(is_used_once)', 'a', 'b')) => ('ine', 'a', 'b')
 *    ('inot', ('ine(is_used_once)', 'a', 'b')) => ('ieq', 'a', 'b')
 *    ('iand', ('feq', 'a', 'b'), ('fneu', 'a', 'b')) => False
 *    ('iand', ('flt', 'a', 'b'), ('flt', 'b', 'a')) => False
 *    ('iand', ('ieq', 'a', 'b'), ('ine', 'a', 'b')) => False
 *    ('iand', ('ilt', 'a', 'b'), ('ilt', 'b', 'a')) => False
 *    ('iand', ('ult', 'a', 'b'), ('ult', 'b', 'a')) => False
 *    ('flt', ('fneg', 'a'), ('fneg', 'b')) => ('flt', 'b', 'a')
 *    ('fge', ('fneg', 'a'), ('fneg', 'b')) => ('fge', 'b', 'a')
 *    ('feq', ('fneg', 'a'), ('fneg', 'b')) => ('feq', 'b', 'a')
 *    ('fneu', ('fneg', 'a'), ('fneg', 'b')) => ('fneu', 'b', 'a')
 *    ('flt', ('fneg', 'a'), -1.0) => ('flt', 1.0, 'a')
 *    ('flt', -1.0, ('fneg', 'a')) => ('flt', 'a', 1.0)
 *    ('fge', ('fneg', 'a'), -1.0) => ('fge', 1.0, 'a')
 *    ('fge', -1.0, ('fneg', 'a')) => ('fge', 'a', 1.0)
 *    ('fneu', ('fneg', 'a'), -1.0) => ('fneu', 1.0, 'a')
 *    ('feq', -1.0, ('fneg', 'a')) => ('feq', 'a', 1.0)
 *    ('ieq', ('ineg', 'a'), 0) => ('ieq', 'a', 0)
 *    ('ine', ('ineg', 'a'), 0) => ('ine', 'a', 0)
 *    ('ieq', ('iabs', 'a'), 0) => ('ieq', 'a', 0)
 *    ('ine', ('iabs', 'a'), 0) => ('ine', 'a', 0)
 *    ('flt', '#b(is_gt_0_and_lt_1)', ('fsat(is_used_once)', 'a')) => ('flt', 'b', 'a')
 *    ('fge', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('fge', 'a', 'b')
 *    ('feq', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('feq', 'a', 'b')
 *    ('fneu', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('fneu', 'a', 'b')
 *    ('fge', ('fsat(is_used_once)', 'a'), 1.0) => ('fge', 'a', 1.0)
 *    ('flt', 0.0, ('fsat(is_used_once)', 'a')) => ('flt', 0.0, 'a')
 *    ('fge', 0.0, ('b2f', 'a@1')) => ('inot', 'a')
 *    ('fge', ('fneg', ('b2f', 'a@1')), 0.0) => ('inot', 'a')
 *    ('fneu', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1')), 0.0) => ('ior', 'a', 'b')
 *    ('fneu', ('bcsel', 'a', 1.0, ('b2f', 'b@1')), 0.0) => ('ior', 'a', 'b')
 *    ('fneu', ('b2f', 'a@1'), ('fneg', ('b2f', 'b@1'))) => ('ior', 'a', 'b')
 *    ('fneu', ('fmul', ('b2f', 'a@1'), ('b2f', 'b@1')), 0.0) => ('iand', 'a', 'b')
 *    ('fneu', ('bcsel', 'a', ('b2f', 'b@1'), 0.0), 0.0) => ('iand', 'a', 'b')
 *    ('fneu', ('fadd', ('b2f', 'a@1'), ('fneg', ('b2f', 'b@1'))), 0.0) => ('ixor', 'a', 'b')
 *    ('fneu', ('b2f', 'a@1'), ('b2f', 'b@1')) => ('ixor', 'a', 'b')
 *    ('fneu', ('fneg', ('b2f', 'a@1')), ('fneg', ('b2f', 'b@1'))) => ('ixor', 'a', 'b')
 *    ('feq', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1')), 0.0) => ('inot', ('ior', 'a', 'b'))
 *    ('feq', ('bcsel', 'a', 1.0, ('b2f', 'b@1')), 0.0) => ('inot', ('ior', 'a', 'b'))
 *    ('feq', ('b2f', 'a@1'), ('fneg', ('b2f', 'b@1'))) => ('inot', ('ior', 'a', 'b'))
 *    ('feq', ('fmul', ('b2f', 'a@1'), ('b2f', 'b@1')), 0.0) => ('inot', ('iand', 'a', 'b'))
 *    ('feq', ('bcsel', 'a', ('b2f', 'b@1'), 0.0), 0.0) => ('inot', ('iand', 'a', 'b'))
 *    ('feq', ('fadd', ('b2f', 'a@1'), ('fneg', ('b2f', 'b@1'))), 0.0) => ('ieq', 'a', 'b')
 *    ('feq', ('b2f', 'a@1'), ('b2f', 'b@1')) => ('ieq', 'a', 'b')
 *    ('feq', ('fneg', ('b2f', 'a@1')), ('fneg', ('b2f', 'b@1'))) => ('ieq', 'a', 'b')
 *    ('flt', ('fneg', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1'))), 0.0) => ('ior', 'a', 'b')
 *    ('flt', 0.0, ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1'))) => ('ior', 'a', 'b')
 *    ('fge', ('fneg', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1'))), 0.0) => ('inot', ('ior', 'a', 'b'))
 *    ('fge', 0.0, ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1'))) => ('inot', ('ior', 'a', 'b'))
 *    ('flt', 'a', ('fneg', 'a')) => ('flt', 'a', 0.0)
 *    ('fge', 'a', ('fneg', 'a')) => ('fge', 'a', 0.0)
 *    ('flt', ('fmin', 'c', ('fneg', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1')))), 0.0) => ('ior', ('flt', 'c', 0.0), ('ior', 'a', 'b'))
 *    ('~flt', ('fadd', 'a', 'b'), 'a') => ('flt', 'b', 0.0)
 *    ('~fge', ('fadd', 'a', 'b'), 'a') => ('fge', 'b', 0.0)
 *    ('~feq', ('fadd', 'a', 'b'), 'a') => ('feq', 'b', 0.0)
 *    ('~fneu', ('fadd', 'a', 'b'), 'a') => ('fneu', 'b', 0.0)
 *    ('~flt', ('fadd(is_used_once)', 'a', '#b'), '#c') => ('flt', 'a', ('fadd', 'c', ('fneg', 'b')))
 *    ('~flt', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', '#b')), '#c') => ('flt', ('fneg', ('fadd', 'c', 'b')), 'a')
 *    ('~fge', ('fadd(is_used_once)', 'a', '#b'), '#c') => ('fge', 'a', ('fadd', 'c', ('fneg', 'b')))
 *    ('~fge', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', '#b')), '#c') => ('fge', ('fneg', ('fadd', 'c', 'b')), 'a')
 *    ('~feq', ('fadd(is_used_once)', 'a', '#b'), '#c') => ('feq', 'a', ('fadd', 'c', ('fneg', 'b')))
 *    ('~feq', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', '#b')), '#c') => ('feq', ('fneg', ('fadd', 'c', 'b')), 'a')
 *    ('~fneu', ('fadd(is_used_once)', 'a', '#b'), '#c') => ('fneu', 'a', ('fadd', 'c', ('fneg', 'b')))
 *    ('~fneu', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', '#b')), '#c') => ('fneu', ('fneg', ('fadd', 'c', 'b')), 'a')
 *    ('ieq', ('iadd', 'a', 'b'), 'a') => ('ieq', 'b', 0)
 *    ('ine', ('iadd', 'a', 'b'), 'a') => ('ine', 'b', 0)
 *    ('feq', ('b2f', 'a@1'), 0.0) => ('inot', 'a')
 *    ('fneu', ('b2f', 'a@1'), 0.0) => a
 *    ('ieq', ('b2i', 'a@1'), 0) => ('inot', 'a')
 *    ('ine', ('b2i', 'a@1'), 0) => a
 *    ('fneu', ('u2f', 'a'), 0.0) => ('ine', 'a', 0)
 *    ('feq', ('u2f', 'a'), 0.0) => ('ieq', 'a', 0)
 *    ('fge', ('u2f', 'a'), 0.0) => True
 *    ('fge', 0.0, ('u2f', 'a')) => ('uge', 0, 'a')
 *    ('flt', ('u2f', 'a'), 0.0) => False
 *    ('flt', 0.0, ('u2f', 'a')) => ('ult', 0, 'a')
 *    ('fneu', ('i2f', 'a'), 0.0) => ('ine', 'a', 0)
 *    ('feq', ('i2f', 'a'), 0.0) => ('ieq', 'a', 0)
 *    ('fge', ('i2f', 'a'), 0.0) => ('ige', 'a', 0)
 *    ('fge', 0.0, ('i2f', 'a')) => ('ige', 0, 'a')
 *    ('flt', ('i2f', 'a'), 0.0) => ('ilt', 'a', 0)
 *    ('flt', 0.0, ('i2f', 'a')) => ('ilt', 0, 'a')
 *    ('~flt', 0.0, ('fabs', 'a')) => ('fneu', 'a', 0.0)
 *    ('~flt', ('fneg', ('fabs', 'a')), 0.0) => ('fneu', 'a', 0.0)
 *    ('fge', 0.0, ('fabs', 'a')) => ('feq', 'a', 0.0)
 *    ('fge', ('fneg', ('fabs', 'a')), 0.0) => ('feq', 'a', 0.0)
 *    ('iand', ('fge', 'a', 0.0), ('fge', 1.0, 'a')) => ('feq', 'a', ('fsat', 'a'))
 *    ('fmax', ('b2f(is_used_once)', 'a@1'), ('b2f', 'b@1')) => ('b2f', ('ior', 'a', 'b'))
 *    ('fmax', ('fneg(is_used_once)', ('b2f(is_used_once)', 'a@1')), ('fneg', ('b2f', 'b@1'))) => ('fneg', ('b2f', ('iand', 'a', 'b')))
 *    ('fmin', ('b2f(is_used_once)', 'a@1'), ('b2f', 'b@1')) => ('b2f', ('iand', 'a', 'b'))
 *    ('fmin', ('fneg(is_used_once)', ('b2f(is_used_once)', 'a@1')), ('fneg', ('b2f', 'b@1'))) => ('fneg', ('b2f', ('ior', 'a', 'b')))
 *    ('fmin', ('b2f', 'a@1'), '#b') => ('bcsel', 'a', ('fmin', 'b', 1.0), ('fmin', 'b', 0.0))
 *    ('flt', ('fadd(is_used_once)', 'a', ('fneg', 'b')), 0.0) => ('flt', 'a', 'b')
 *    ('fge', ('fneg', ('fabs', 'a')), 0.0) => ('feq', 'a', 0.0)
 *    ('~bcsel', ('flt', 'b', 'a'), 'b', 'a') => ('fmin', 'a', 'b')
 *    ('~bcsel', ('flt', 'a', 'b'), 'b', 'a') => ('fmax', 'a', 'b')
 *    ('~bcsel', ('fge', 'a', 'b'), 'b', 'a') => ('fmin', 'a', 'b')
 *    ('~bcsel', ('fge', 'b', 'a'), 'b', 'a') => ('fmax', 'a', 'b')
 *    ('bcsel', ('ult', 'b', 'a'), 'b', 'a') => ('umin', 'a', 'b')
 *    ('bcsel', ('ult', 'a', 'b'), 'b', 'a') => ('umax', 'a', 'b')
 *    ('bcsel', ('uge', 'a', 'b'), 'b', 'a') => ('umin', 'a', 'b')
 *    ('bcsel', ('uge', 'b', 'a'), 'b', 'a') => ('umax', 'a', 'b')
 *    ('bcsel', ('ilt', 'b', 'a'), 'b', 'a') => ('imin', 'a', 'b')
 *    ('bcsel', ('ilt', 'a', 'b'), 'b', 'a') => ('imax', 'a', 'b')
 *    ('bcsel', ('ige', 'a', 'b'), 'b', 'a') => ('imin', 'a', 'b')
 *    ('bcsel', ('ige', 'b', 'a'), 'b', 'a') => ('imax', 'a', 'b')
 *    ('bcsel', ('inot', 'a'), 'b', 'c') => ('bcsel', 'a', 'c', 'b')
 *    ('bcsel', 'a', ('bcsel', 'a', 'b', 'c'), 'd') => ('bcsel', 'a', 'b', 'd')
 *    ('bcsel', 'a', 'b', ('bcsel', 'a', 'c', 'd')) => ('bcsel', 'a', 'b', 'd')
 *    ('bcsel', 'a', ('bcsel', 'b', 'c', 'd'), ('bcsel(is_used_once)', 'b', 'c', 'e')) => ('bcsel', 'b', 'c', ('bcsel', 'a', 'd', 'e'))
 *    ('bcsel', 'a', ('bcsel(is_used_once)', 'b', 'c', 'd'), ('bcsel', 'b', 'c', 'e')) => ('bcsel', 'b', 'c', ('bcsel', 'a', 'd', 'e'))
 *    ('bcsel', 'a', ('bcsel', 'b', 'c', 'd'), ('bcsel(is_used_once)', 'b', 'e', 'd')) => ('bcsel', 'b', ('bcsel', 'a', 'c', 'e'), 'd')
 *    ('bcsel', 'a', ('bcsel(is_used_once)', 'b', 'c', 'd'), ('bcsel', 'b', 'e', 'd')) => ('bcsel', 'b', ('bcsel', 'a', 'c', 'e'), 'd')
 *    ('bcsel', 'a', True, 'b') => ('ior', 'a', 'b')
 *    ('bcsel', 'a', 'a', 'b') => ('ior', 'a', 'b')
 *    ('bcsel', 'a', 'b', False) => ('iand', 'a', 'b')
 *    ('bcsel', 'a', 'b', 'a') => ('iand', 'a', 'b')
 *    ('~fmin', 'a', 'a') => a
 *    ('~fmax', 'a', 'a') => a
 *    ('imin', 'a', 'a') => a
 *    ('imax', 'a', 'a') => a
 *    ('umin', 'a', 'a') => a
 *    ('umin', 'a', 0) => 0
 *    ('umin', 'a', -1) => a
 *    ('umax', 'a', 'a') => a
 *    ('umax', 'a', 0) => a
 *    ('umax', 'a', -1) => -1
 *    ('fmax', ('fmax', 'a', 'b'), 'b') => ('fmax', 'a', 'b')
 *    ('umax', ('umax', 'a', 'b'), 'b') => ('umax', 'a', 'b')
 *    ('imax', ('imax', 'a', 'b'), 'b') => ('imax', 'a', 'b')
 *    ('fmin', ('fmin', 'a', 'b'), 'b') => ('fmin', 'a', 'b')
 *    ('umin', ('umin', 'a', 'b'), 'b') => ('umin', 'a', 'b')
 *    ('imin', ('imin', 'a', 'b'), 'b') => ('imin', 'a', 'b')
 *    ('fmax', ('fmax', ('fmax', 'a', 'b'), 'c'), 'a') => ('fmax', ('fmax', 'a', 'b'), 'c')
 *    ('umax', ('umax', ('umax', 'a', 'b'), 'c'), 'a') => ('umax', ('umax', 'a', 'b'), 'c')
 *    ('imax', ('imax', ('imax', 'a', 'b'), 'c'), 'a') => ('imax', ('imax', 'a', 'b'), 'c')
 *    ('fmin', ('fmin', ('fmin', 'a', 'b'), 'c'), 'a') => ('fmin', ('fmin', 'a', 'b'), 'c')
 *    ('umin', ('umin', ('umin', 'a', 'b'), 'c'), 'a') => ('umin', ('umin', 'a', 'b'), 'c')
 *    ('imin', ('imin', ('imin', 'a', 'b'), 'c'), 'a') => ('imin', ('imin', 'a', 'b'), 'c')
 *    ('ieq', ('b2i8', 'a@1'), ('b2i8', 'b@1')) => ('ieq', 'a', 'b')
 *    ('ine', ('b2i8', 'a@1'), ('b2i8', 'b@1')) => ('ine', 'a', 'b')
 *    ('ieq', ('b2i16', 'a@1'), ('b2i16', 'b@1')) => ('ieq', 'a', 'b')
 *    ('ine', ('b2i16', 'a@1'), ('b2i16', 'b@1')) => ('ine', 'a', 'b')
 *    ('ieq', ('b2i32', 'a@1'), ('b2i32', 'b@1')) => ('ieq', 'a', 'b')
 *    ('ine', ('b2i32', 'a@1'), ('b2i32', 'b@1')) => ('ine', 'a', 'b')
 *    ('ieq', ('b2i64', 'a@1'), ('b2i64', 'b@1')) => ('ieq', 'a', 'b')
 *    ('ine', ('b2i64', 'a@1'), ('b2i64', 'b@1')) => ('ine', 'a', 'b')
 *    ('feq', ('b2f16', 'a@1'), ('b2f16', 'b@1')) => ('ieq', 'a', 'b')
 *    ('fneu', ('b2f16', 'a@1'), ('b2f16', 'b@1')) => ('ine', 'a', 'b')
 *    ('feq', ('b2f32', 'a@1'), ('b2f32', 'b@1')) => ('ieq', 'a', 'b')
 *    ('fneu', ('b2f32', 'a@1'), ('b2f32', 'b@1')) => ('ine', 'a', 'b')
 *    ('feq', ('b2f64', 'a@1'), ('b2f64', 'b@1')) => ('ieq', 'a', 'b')
 *    ('fneu', ('b2f64', 'a@1'), ('b2f64', 'b@1')) => ('ine', 'a', 'b')
 *    ('iand@8', 'a', ('inot', ('ishr', 'a', 7))) => ('imax', 'a', 0)
 *    ('ieq', ('iand', 'a@8', 128), 0) => ('ige', 'a', 0)
 *    ('ine', ('iand', 'a@8', 128), 128) => ('ige', 'a', 0)
 *    ('ine', ('iand', 'a@8', 128), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('iand', 'a@8', 128), 128) => ('ilt', 'a', 0)
 *    ('ine', ('ushr', 'a@8', 7), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('ushr', 'a@8', 7), 0) => ('ige', 'a', 0)
 *    ('ieq', ('ushr', 'a@8', 7), 1) => ('ilt', 'a', 0)
 *    ('ine', ('ushr', 'a@8', 7), 1) => ('ige', 'a', 0)
 *    ('ine', ('ishr', 'a@8', 7), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('ishr', 'a@8', 7), 0) => ('ige', 'a', 0)
 *    ('ieq', ('ishr', 'a@8', 7), -1) => ('ilt', 'a', 0)
 *    ('ine', ('ishr', 'a@8', 7), -1) => ('ige', 'a', 0)
 *    ('iand@16', 'a', ('inot', ('ishr', 'a', 15))) => ('imax', 'a', 0)
 *    ('ieq', ('iand', 'a@16', 32768), 0) => ('ige', 'a', 0)
 *    ('ine', ('iand', 'a@16', 32768), 32768) => ('ige', 'a', 0)
 *    ('ine', ('iand', 'a@16', 32768), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('iand', 'a@16', 32768), 32768) => ('ilt', 'a', 0)
 *    ('ine', ('ushr', 'a@16', 15), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('ushr', 'a@16', 15), 0) => ('ige', 'a', 0)
 *    ('ieq', ('ushr', 'a@16', 15), 1) => ('ilt', 'a', 0)
 *    ('ine', ('ushr', 'a@16', 15), 1) => ('ige', 'a', 0)
 *    ('ine', ('ishr', 'a@16', 15), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('ishr', 'a@16', 15), 0) => ('ige', 'a', 0)
 *    ('ieq', ('ishr', 'a@16', 15), -1) => ('ilt', 'a', 0)
 *    ('ine', ('ishr', 'a@16', 15), -1) => ('ige', 'a', 0)
 *    ('iand@32', 'a', ('inot', ('ishr', 'a', 31))) => ('imax', 'a', 0)
 *    ('ieq', ('iand', 'a@32', 2147483648), 0) => ('ige', 'a', 0)
 *    ('ine', ('iand', 'a@32', 2147483648), 2147483648) => ('ige', 'a', 0)
 *    ('ine', ('iand', 'a@32', 2147483648), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('iand', 'a@32', 2147483648), 2147483648) => ('ilt', 'a', 0)
 *    ('ine', ('ushr', 'a@32', 31), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('ushr', 'a@32', 31), 0) => ('ige', 'a', 0)
 *    ('ieq', ('ushr', 'a@32', 31), 1) => ('ilt', 'a', 0)
 *    ('ine', ('ushr', 'a@32', 31), 1) => ('ige', 'a', 0)
 *    ('ine', ('ishr', 'a@32', 31), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('ishr', 'a@32', 31), 0) => ('ige', 'a', 0)
 *    ('ieq', ('ishr', 'a@32', 31), -1) => ('ilt', 'a', 0)
 *    ('ine', ('ishr', 'a@32', 31), -1) => ('ige', 'a', 0)
 *    ('iand@64', 'a', ('inot', ('ishr', 'a', 63))) => ('imax', 'a', 0)
 *    ('ieq', ('iand', 'a@64', 9223372036854775808), 0) => ('ige', 'a', 0)
 *    ('ine', ('iand', 'a@64', 9223372036854775808), 9223372036854775808) => ('ige', 'a', 0)
 *    ('ine', ('iand', 'a@64', 9223372036854775808), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('iand', 'a@64', 9223372036854775808), 9223372036854775808) => ('ilt', 'a', 0)
 *    ('ine', ('ushr', 'a@64', 63), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('ushr', 'a@64', 63), 0) => ('ige', 'a', 0)
 *    ('ieq', ('ushr', 'a@64', 63), 1) => ('ilt', 'a', 0)
 *    ('ine', ('ushr', 'a@64', 63), 1) => ('ige', 'a', 0)
 *    ('ine', ('ishr', 'a@64', 63), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('ishr', 'a@64', 63), 0) => ('ige', 'a', 0)
 *    ('ieq', ('ishr', 'a@64', 63), -1) => ('ilt', 'a', 0)
 *    ('ine', ('ishr', 'a@64', 63), -1) => ('ige', 'a', 0)
 *    ('fmin', 'a', ('fneg', 'a')) => ('fneg', ('fabs', 'a'))
 *    ('imin', 'a', ('ineg', 'a')) => ('ineg', ('iabs', 'a'))
 *    ('fmin', 'a', ('fneg', ('fabs', 'a'))) => ('fneg', ('fabs', 'a'))
 *    ('imin', 'a', ('ineg', ('iabs', 'a'))) => ('ineg', ('iabs', 'a'))
 *    ('~fmin', 'a', ('fabs', 'a')) => a
 *    ('imin', 'a', ('iabs', 'a')) => a
 *    ('~fmax', 'a', ('fneg', ('fabs', 'a'))) => a
 *    ('imax', 'a', ('ineg', ('iabs', 'a'))) => a
 *    ('fmax', 'a', ('fabs', 'a')) => ('fabs', 'a')
 *    ('imax', 'a', ('iabs', 'a')) => ('iabs', 'a')
 *    ('fmax', 'a', ('fneg', 'a')) => ('fabs', 'a')
 *    ('imax', 'a', ('ineg', 'a')) => ('iabs', 'a')
 *    ('~fmax', ('fabs', 'a'), 0.0) => ('fabs', 'a')
 *    ('fmin', ('fmax', 'a', 0.0), 1.0) => ('fsat', 'a')
 *    ('~fmax', ('fmin', 'a', 1.0), 0.0) => ('fsat', 'a')
 *    ('~fmin', ('fmax', 'a', -1.0), 0.0) => ('fneg', ('fsat', ('fneg', 'a')))
 *    ('~fmax', ('fmin', 'a', 0.0), -1.0) => ('fneg', ('fsat', ('fneg', 'a')))
 *    ('fsat', ('fsign', 'a')) => ('b2f', ('!flt', 0.0, 'a'))
 *    ('fsat', ('b2f', 'a')) => ('b2f', 'a')
 *    ('fsat', 'a') => ('fmin', ('fmax', 'a', 0.0), 1.0)
 *    ('fsat', ('fsat', 'a')) => ('fsat', 'a')
 *    ('fsat', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', 'b'))) => ('fsat', ('fadd', ('fneg', 'a'), ('fneg', 'b')))
 *    ('fsat', ('fneg(is_used_once)', ('fmul(is_used_once)', 'a', 'b'))) => ('fsat', ('fmul', ('fneg', 'a'), 'b'))
 *    ('fsat', ('fneg(is_used_once)', ('fmulz(is_used_once)', 'a', 'b'))) => ('fsat', ('fmulz', ('fneg', 'a'), 'b'))
 *    ('fsat', ('fabs(is_used_once)', ('fmul(is_used_once)', 'a', 'b'))) => ('fsat', ('fmul', ('fabs', 'a'), ('fabs', 'b')))
 *    ('fmin', ('fmax', ('fmin', ('fmax', 'a', 'b'), 'c'), 'b'), 'c') => ('fmin', ('fmax', 'a', 'b'), 'c')
 *    ('imin', ('imax', ('imin', ('imax', 'a', 'b'), 'c'), 'b'), 'c') => ('imin', ('imax', 'a', 'b'), 'c')
 *    ('umin', ('umax', ('umin', ('umax', 'a', 'b'), 'c'), 'b'), 'c') => ('umin', ('umax', 'a', 'b'), 'c')
 *    ('fmax', ('fsat', 'a'), '#b(is_zero_to_one)') => ('fsat', ('fmax', 'a', 'b'))
 *    ('~fmin', ('fsat', 'a'), '#b(is_zero_to_one)') => ('fsat', ('fmin', 'a', 'b'))
 *    ('fmax', ('fneg', ('fmin', 'b', 'a')), 'b') => ('fmax', ('fabs', 'b'), ('fneg', 'a'))
 *    ('fmin', ('fneg', ('fmax', 'b', 'a')), 'b') => ('fmin', ('fneg', ('fabs', 'b')), ('fneg', 'a'))
 *    ('fmax', ('fadd(is_used_once)', ('fneg', 'a(is_not_negative)'), '#b(is_zero_to_one)'), 0.0) => ('fsat', ('fadd', ('fneg', 'a'), 'b'))
 *    ('extract_u8', ('imin', ('imax', 'a', 0), 255), 0) => ('imin', ('imax', 'a', 0), 255)
 *    ('ior', ('flt(is_used_once)', 'a', 'b'), ('flt', 'a', 'c')) => ('flt', 'a', ('!fmax', 'b', 'c'))
 *    ('ior', ('flt(is_used_once)', 'a', 'c'), ('flt', 'b', 'c')) => ('flt', ('!fmin', 'a', 'b'), 'c')
 *    ('ior', ('fge(is_used_once)', 'a', 'b'), ('fge', 'a', 'c')) => ('fge', 'a', ('!fmin', 'b', 'c'))
 *    ('ior', ('fge(is_used_once)', 'a', 'c'), ('fge', 'b', 'c')) => ('fge', ('!fmax', 'a', 'b'), 'c')
 *    ('ior', ('flt', 'a', '#b'), ('flt', 'a', '#c')) => ('flt', 'a', ('!fmax', 'b', 'c'))
 *    ('ior', ('flt', '#a', 'c'), ('flt', '#b', 'c')) => ('flt', ('!fmin', 'a', 'b'), 'c')
 *    ('ior', ('fge', 'a', '#b'), ('fge', 'a', '#c')) => ('fge', 'a', ('!fmin', 'b', 'c'))
 *    ('ior', ('fge', '#a', 'c'), ('fge', '#b', 'c')) => ('fge', ('!fmax', 'a', 'b'), 'c')
 *    ('~iand', ('flt(is_used_once)', 'a', 'b'), ('flt', 'a', 'c')) => ('flt', 'a', ('fmin', 'b', 'c'))
 *    ('~iand', ('flt(is_used_once)', 'a', 'c'), ('flt', 'b', 'c')) => ('flt', ('fmax', 'a', 'b'), 'c')
 *    ('~iand', ('fge(is_used_once)', 'a', 'b'), ('fge', 'a', 'c')) => ('fge', 'a', ('fmax', 'b', 'c'))
 *    ('~iand', ('fge(is_used_once)', 'a', 'c'), ('fge', 'b', 'c')) => ('fge', ('fmin', 'a', 'b'), 'c')
 *    ('iand', ('flt', 'a', '#b(is_a_number)'), ('flt', 'a', '#c(is_a_number)')) => ('flt', 'a', ('fmin', 'b', 'c'))
 *    ('iand', ('flt', '#a(is_a_number)', 'c'), ('flt', '#b(is_a_number)', 'c')) => ('flt', ('fmax', 'a', 'b'), 'c')
 *    ('iand', ('fge', 'a', '#b(is_a_number)'), ('fge', 'a', '#c(is_a_number)')) => ('fge', 'a', ('fmax', 'b', 'c'))
 *    ('iand', ('fge', '#a(is_a_number)', 'c'), ('fge', '#b(is_a_number)', 'c')) => ('fge', ('fmin', 'a', 'b'), 'c')
 *    ('ior', ('ilt(is_used_once)', 'a', 'b'), ('ilt', 'a', 'c')) => ('ilt', 'a', ('imax', 'b', 'c'))
 *    ('ior', ('ilt(is_used_once)', 'a', 'c'), ('ilt', 'b', 'c')) => ('ilt', ('imin', 'a', 'b'), 'c')
 *    ('ior', ('ige(is_used_once)', 'a', 'b'), ('ige', 'a', 'c')) => ('ige', 'a', ('imin', 'b', 'c'))
 *    ('ior', ('ige(is_used_once)', 'a', 'c'), ('ige', 'b', 'c')) => ('ige', ('imax', 'a', 'b'), 'c')
 *    ('ior', ('ult(is_used_once)', 'a', 'b'), ('ult', 'a', 'c')) => ('ult', 'a', ('umax', 'b', 'c'))
 *    ('ior', ('ult(is_used_once)', 'a', 'c'), ('ult', 'b', 'c')) => ('ult', ('umin', 'a', 'b'), 'c')
 *    ('ior', ('uge(is_used_once)', 'a', 'b'), ('uge', 'a', 'c')) => ('uge', 'a', ('umin', 'b', 'c'))
 *    ('ior', ('uge(is_used_once)', 'a', 'c'), ('uge', 'b', 'c')) => ('uge', ('umax', 'a', 'b'), 'c')
 *    ('iand', ('ilt(is_used_once)', 'a', 'b'), ('ilt', 'a', 'c')) => ('ilt', 'a', ('imin', 'b', 'c'))
 *    ('iand', ('ilt(is_used_once)', 'a', 'c'), ('ilt', 'b', 'c')) => ('ilt', ('imax', 'a', 'b'), 'c')
 *    ('iand', ('ige(is_used_once)', 'a', 'b'), ('ige', 'a', 'c')) => ('ige', 'a', ('imax', 'b', 'c'))
 *    ('iand', ('ige(is_used_once)', 'a', 'c'), ('ige', 'b', 'c')) => ('ige', ('imin', 'a', 'b'), 'c')
 *    ('iand', ('ult(is_used_once)', 'a', 'b'), ('ult', 'a', 'c')) => ('ult', 'a', ('umin', 'b', 'c'))
 *    ('iand', ('ult(is_used_once)', 'a', 'c'), ('ult', 'b', 'c')) => ('ult', ('umax', 'a', 'b'), 'c')
 *    ('iand', ('uge(is_used_once)', 'a', 'b'), ('uge', 'a', 'c')) => ('uge', 'a', ('umax', 'b', 'c'))
 *    ('iand', ('uge(is_used_once)', 'a', 'c'), ('uge', 'b', 'c')) => ('uge', ('umin', 'a', 'b'), 'c')
 *    ('ior', ('ior(is_used_once)', ('flt(is_used_once)', 'a', 'c'), 'd'), ('flt', 'b', 'c')) => ('ior', ('flt', ('!fmin', 'a', 'b'), 'c'), 'd')
 *    ('ior', ('ior(is_used_once)', ('flt', 'a', 'c'), 'd'), ('flt(is_used_once)', 'b', 'c')) => ('ior', ('flt', ('!fmin', 'a', 'b'), 'c'), 'd')
 *    ('ior', ('ior(is_used_once)', ('flt(is_used_once)', 'a', 'b'), 'd'), ('flt', 'a', 'c')) => ('ior', ('flt', 'a', ('!fmax', 'b', 'c')), 'd')
 *    ('ior', ('ior(is_used_once)', ('flt', 'a', 'b'), 'd'), ('flt(is_used_once)', 'a', 'c')) => ('ior', ('flt', 'a', ('!fmax', 'b', 'c')), 'd')
 *    ('ior', ('flt', 'a(is_a_number)', 'b(is_a_number)'), ('flt', 'b', 'a')) => ('fneu', 'a', 'b')
 *    ('umin', ('iand', 'a', '#b(is_pos_power_of_two)'), ('iand', 'c', 'b')) => ('iand', ('iand', 'a', 'b'), ('iand', 'c', 'b'))
 *    ('ior', ('flt(is_used_once)', 0.0, 'a@16'), ('flt', 'b@16', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('ior', ('flt', 0.0, 'a@16'), ('flt(is_used_once)', 'b@16', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('ior', ('fge(is_used_once)', 0.0, 'a@16'), ('fge', 'b@16', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('ior', ('fge', 0.0, 'a@16'), ('fge(is_used_once)', 'b@16', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('~iand', ('flt(is_used_once)', 0.0, 'a@16'), ('flt', 'b@16', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('~iand', ('flt', 0.0, 'a@16'), ('flt(is_used_once)', 'b@16', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('~iand', ('fge(is_used_once)', 0.0, 'a@16'), ('fge', 'b@16', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('~iand', ('fge', 0.0, 'a@16'), ('fge(is_used_once)', 'b@16', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('bcsel@16', ('feq', 'a', 0.0), 1.0, ('i2f16', ('iadd', ('b2i16', ('flt', 0.0, 'a@16')), ('ineg', ('b2i16', ('flt', 'a@16', 0.0)))))) => ('i2f16', ('iadd', ('b2i32', ('!fge', 'a', 0.0)), ('ineg', ('b2i32', ('!flt', 'a', 0.0)))))
 *    ('bcsel', 'a', ('b2f(is_used_once)', 'b@16'), ('b2f', 'c@16')) => ('b2f', ('bcsel', 'a', 'b', 'c'))
 *    ('~i2f16', ('f2i', 'a@16')) => ('ftrunc', 'a')
 *    ('~f2i16', ('i2f', 'a@16')) => a
 *    ('~f2i16', ('u2f', 'a@16')) => a
 *    ('~f2u16', ('i2f', 'a@16')) => a
 *    ('~f2u16', ('u2f', 'a@16')) => a
 *    ('fadd', ('b2f16', ('flt', 0.0, 'a@16')), ('fneg', ('b2f16', ('flt', 'a@16', 0.0)))) => ('fsign', 'a')
 *    ('iadd', ('b2i16', ('flt', 0, 'a@16')), ('ineg', ('b2i16', ('flt', 'a@16', 0)))) => ('f2i16', ('fsign', 'a'))
 *    ('f2f16', ('f2f32', 'a@16')) => a
 *    ('i2i16', ('i2i32', 'a@16')) => a
 *    ('u2u16', ('u2u32', 'a@16')) => a
 *    ('f2f16', ('b2f32', 'a@1')) => ('b2f16', 'a')
 *    ('i2i16', ('b2i32', 'a@1')) => ('b2i16', 'a')
 *    ('u2u16', ('b2i32', 'a@1')) => ('b2i16', 'a')
 *    ('f2u32', ('f2f32', 'a@16')) => ('f2u32', 'a')
 *    ('f2i32', ('f2f32', 'a@16')) => ('f2i32', 'a')
 *    ('f2f16', ('u2f32', 'a')) => ('u2f16', 'a')
 *    ('f2f16', ('i2f32', 'a')) => ('i2f16', 'a')
 *    ('u2f32', ('u2u32', 'a@16')) => ('u2f32', 'a')
 *    ('i2f32', ('i2i32', 'a@16')) => ('i2f32', 'a')
 *    ('f2f16', ('f2f64', 'a@16')) => a
 *    ('i2i16', ('i2i64', 'a@16')) => a
 *    ('u2u16', ('u2u64', 'a@16')) => a
 *    ('f2f16', ('b2f64', 'a@1')) => ('b2f16', 'a')
 *    ('i2i16', ('b2i64', 'a@1')) => ('b2i16', 'a')
 *    ('u2u16', ('b2i64', 'a@1')) => ('b2i16', 'a')
 *    ('f2u64', ('f2f64', 'a@16')) => ('f2u64', 'a')
 *    ('f2i64', ('f2f64', 'a@16')) => ('f2i64', 'a')
 *    ('f2f16', ('u2f64', 'a')) => ('u2f16', 'a')
 *    ('f2f16', ('i2f64', 'a')) => ('i2f16', 'a')
 *    ('u2f64', ('u2u64', 'a@16')) => ('u2f64', 'a')
 *    ('i2f64', ('i2i64', 'a@16')) => ('i2f64', 'a')
 *    ('ior', ('flt(is_used_once)', 0.0, 'a@32'), ('flt', 'b@32', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('ior', ('flt', 0.0, 'a@32'), ('flt(is_used_once)', 'b@32', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('ior', ('fge(is_used_once)', 0.0, 'a@32'), ('fge', 'b@32', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('ior', ('fge', 0.0, 'a@32'), ('fge(is_used_once)', 'b@32', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('~iand', ('flt(is_used_once)', 0.0, 'a@32'), ('flt', 'b@32', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('~iand', ('flt', 0.0, 'a@32'), ('flt(is_used_once)', 'b@32', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('~iand', ('fge(is_used_once)', 0.0, 'a@32'), ('fge', 'b@32', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('~iand', ('fge', 0.0, 'a@32'), ('fge(is_used_once)', 'b@32', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('bcsel@32', ('feq', 'a', 0.0), 1.0, ('i2f32', ('iadd', ('b2i32', ('flt', 0.0, 'a@32')), ('ineg', ('b2i32', ('flt', 'a@32', 0.0)))))) => ('i2f32', ('iadd', ('b2i32', ('!fge', 'a', 0.0)), ('ineg', ('b2i32', ('!flt', 'a', 0.0)))))
 *    ('bcsel', 'a', ('b2f(is_used_once)', 'b@32'), ('b2f', 'c@32')) => ('b2f', ('bcsel', 'a', 'b', 'c'))
 *    ('~i2f32', ('f2i', 'a@32')) => ('ftrunc', 'a')
 *    ('~f2i32', ('i2f', 'a@32')) => a
 *    ('~f2i32', ('u2f', 'a@32')) => a
 *    ('~f2u32', ('i2f', 'a@32')) => a
 *    ('~f2u32', ('u2f', 'a@32')) => a
 *    ('fadd', ('b2f32', ('flt', 0.0, 'a@32')), ('fneg', ('b2f32', ('flt', 'a@32', 0.0)))) => ('fsign', 'a')
 *    ('iadd', ('b2i32', ('flt', 0, 'a@32')), ('ineg', ('b2i32', ('flt', 'a@32', 0)))) => ('f2i32', ('fsign', 'a'))
 *    ('f2f32', ('f2f64', 'a@32')) => a
 *    ('i2i32', ('i2i64', 'a@32')) => a
 *    ('u2u32', ('u2u64', 'a@32')) => a
 *    ('f2f32', ('b2f64', 'a@1')) => ('b2f32', 'a')
 *    ('i2i32', ('b2i64', 'a@1')) => ('b2i32', 'a')
 *    ('u2u32', ('b2i64', 'a@1')) => ('b2i32', 'a')
 *    ('f2u64', ('f2f64', 'a@32')) => ('f2u64', 'a')
 *    ('f2i64', ('f2f64', 'a@32')) => ('f2i64', 'a')
 *    ('f2f32', ('u2f64', 'a')) => ('u2f32', 'a')
 *    ('f2f32', ('i2f64', 'a')) => ('i2f32', 'a')
 *    ('u2f64', ('u2u64', 'a@32')) => ('u2f64', 'a')
 *    ('i2f64', ('i2i64', 'a@32')) => ('i2f64', 'a')
 *    ('ior', ('flt(is_used_once)', 0.0, 'a@64'), ('flt', 'b@64', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('ior', ('flt', 0.0, 'a@64'), ('flt(is_used_once)', 'b@64', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('ior', ('fge(is_used_once)', 0.0, 'a@64'), ('fge', 'b@64', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('ior', ('fge', 0.0, 'a@64'), ('fge(is_used_once)', 'b@64', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('~iand', ('flt(is_used_once)', 0.0, 'a@64'), ('flt', 'b@64', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('~iand', ('flt', 0.0, 'a@64'), ('flt(is_used_once)', 'b@64', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('~iand', ('fge(is_used_once)', 0.0, 'a@64'), ('fge', 'b@64', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('~iand', ('fge', 0.0, 'a@64'), ('fge(is_used_once)', 'b@64', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('bcsel@64', ('feq', 'a', 0.0), 1.0, ('i2f64', ('iadd', ('b2i64', ('flt', 0.0, 'a@64')), ('ineg', ('b2i64', ('flt', 'a@64', 0.0)))))) => ('i2f64', ('iadd', ('b2i32', ('!fge', 'a', 0.0)), ('ineg', ('b2i32', ('!flt', 'a', 0.0)))))
 *    ('bcsel', 'a', ('b2f(is_used_once)', 'b@64'), ('b2f', 'c@64')) => ('b2f', ('bcsel', 'a', 'b', 'c'))
 *    ('~i2f64', ('f2i', 'a@64')) => ('ftrunc', 'a')
 *    ('~f2i64', ('i2f', 'a@64')) => a
 *    ('~f2i64', ('u2f', 'a@64')) => a
 *    ('~f2u64', ('i2f', 'a@64')) => a
 *    ('~f2u64', ('u2f', 'a@64')) => a
 *    ('fadd', ('b2f64', ('flt', 0.0, 'a@64')), ('fneg', ('b2f64', ('flt', 'a@64', 0.0)))) => ('fsign', 'a')
 *    ('iadd', ('b2i64', ('flt', 0, 'a@64')), ('ineg', ('b2i64', ('flt', 'a@64', 0)))) => ('f2i64', ('fsign', 'a'))
 *    ('f2fmp', ('u2f32', 'a@32')) => ('u2fmp', 'a')
 *    ('f2fmp', ('i2f32', 'a@32')) => ('i2fmp', 'a')
 *    ('f2u16', ('f2fmp', 'a@32')) => ('f2u16', 'a')
 *    ('f2i16', ('f2fmp', 'a@32')) => ('f2i16', 'a')
 *    ('i2imp', ('f2u32', 'a@32')) => ('f2ump', 'a')
 *    ('i2imp', ('f2i32', 'a@32')) => ('f2imp', 'a')
 *    ('u2f16', ('i2imp', 'a@32')) => ('u2f16', 'a')
 *    ('i2f16', ('i2imp', 'a@32')) => ('i2f16', 'a')
 *    ('iand', ('u2u16', ('u2u8', 'a@16')), '#b') => ('iand', 'a', ('iand', 'b', 255))
 *    ('u2u16', ('u2u8(is_used_once)', ('iand', 'a@16', '#b'))) => ('iand', 'a', ('iand', 'b', 255))
 *    ('u2u8', ('iand', ('u2u16', ('u2u8', 'a@16')), ('u2u16', ('u2u8', 'b@16')))) => ('u2u8', ('iand', 'a', 'b'))
 *    ('u2u8', ('iand', ('u2u16', ('u2u8', 'a@32')), ('u2u16', ('u2u8', 'b@32')))) => ('u2u8', ('iand', 'a', 'b'))
 *    ('iand', ('extract_i8', 'a', '#b'), ('extract_i8', 'c', 'b')) => ('extract_i8', ('iand', 'a', 'c'), 'b')
 *    ('iand', ('extract_u8', 'a', '#b'), ('extract_u8', 'c', 'b')) => ('extract_u8', ('iand', 'a', 'c'), 'b')
 *    ('iand', ('extract_i16', 'a', '#b'), ('extract_i16', 'c', 'b')) => ('extract_i16', ('iand', 'a', 'c'), 'b')
 *    ('iand', ('extract_u16', 'a', '#b'), ('extract_u16', 'c', 'b')) => ('extract_u16', ('iand', 'a', 'c'), 'b')
 *    ('iand', ('ushr(is_used_once)', 'a', '#b'), ('ushr', 'c', 'b')) => ('ushr', ('iand', 'a', 'c'), 'b')
 *    ('iand', ('ishr(is_used_once)', 'a', '#b'), ('ishr', 'c', 'b')) => ('ishr', ('iand', 'a', 'c'), 'b')
 *    ('iand', ('ishl(is_used_once)', 'a', '#b'), ('ishl', 'c', 'b')) => ('ishl', ('iand', 'a', 'c'), 'b')
 *    ('u2u8', ('ior', ('u2u16', ('u2u8', 'a@16')), ('u2u16', ('u2u8', 'b@16')))) => ('u2u8', ('ior', 'a', 'b'))
 *    ('u2u8', ('ior', ('u2u16', ('u2u8', 'a@32')), ('u2u16', ('u2u8', 'b@32')))) => ('u2u8', ('ior', 'a', 'b'))
 *    ('ior', ('extract_i8', 'a', '#b'), ('extract_i8', 'c', 'b')) => ('extract_i8', ('ior', 'a', 'c'), 'b')
 *    ('ior', ('extract_u8', 'a', '#b'), ('extract_u8', 'c', 'b')) => ('extract_u8', ('ior', 'a', 'c'), 'b')
 *    ('ior', ('extract_i16', 'a', '#b'), ('extract_i16', 'c', 'b')) => ('extract_i16', ('ior', 'a', 'c'), 'b')
 *    ('ior', ('extract_u16', 'a', '#b'), ('extract_u16', 'c', 'b')) => ('extract_u16', ('ior', 'a', 'c'), 'b')
 *    ('ior', ('ushr(is_used_once)', 'a', '#b'), ('ushr', 'c', 'b')) => ('ushr', ('ior', 'a', 'c'), 'b')
 *    ('ior', ('ishr(is_used_once)', 'a', '#b'), ('ishr', 'c', 'b')) => ('ishr', ('ior', 'a', 'c'), 'b')
 *    ('ior', ('ishl(is_used_once)', 'a', '#b'), ('ishl', 'c', 'b')) => ('ishl', ('ior', 'a', 'c'), 'b')
 *    ('u2u8', ('ixor', ('u2u16', ('u2u8', 'a@16')), ('u2u16', ('u2u8', 'b@16')))) => ('u2u8', ('ixor', 'a', 'b'))
 *    ('u2u8', ('ixor', ('u2u16', ('u2u8', 'a@32')), ('u2u16', ('u2u8', 'b@32')))) => ('u2u8', ('ixor', 'a', 'b'))
 *    ('ixor', ('extract_i8', 'a', '#b'), ('extract_i8', 'c', 'b')) => ('extract_i8', ('ixor', 'a', 'c'), 'b')
 *    ('ixor', ('extract_u8', 'a', '#b'), ('extract_u8', 'c', 'b')) => ('extract_u8', ('ixor', 'a', 'c'), 'b')
 *    ('ixor', ('extract_i16', 'a', '#b'), ('extract_i16', 'c', 'b')) => ('extract_i16', ('ixor', 'a', 'c'), 'b')
 *    ('ixor', ('extract_u16', 'a', '#b'), ('extract_u16', 'c', 'b')) => ('extract_u16', ('ixor', 'a', 'c'), 'b')
 *    ('ixor', ('ushr(is_used_once)', 'a', '#b'), ('ushr', 'c', 'b')) => ('ushr', ('ixor', 'a', 'c'), 'b')
 *    ('ixor', ('ishr(is_used_once)', 'a', '#b'), ('ishr', 'c', 'b')) => ('ishr', ('ixor', 'a', 'c'), 'b')
 *    ('ixor', ('ishl(is_used_once)', 'a', '#b'), ('ishl', 'c', 'b')) => ('ishl', ('ixor', 'a', 'c'), 'b')
 *    ('iand', ('ieq', 'a@8', 0), ('ieq', 'b@8', 0)) => ('ieq', ('ior', 'a', 'b'), 0)
 *    ('ior', ('ine', 'a@8', 0), ('ine', 'b@8', 0)) => ('ine', ('ior', 'a', 'b'), 0)
 *    ('iand', ('ieq', 'a@8', 0), ('ieq', 'b@8', 0)) => ('ieq', ('umax', 'a', 'b'), 0)
 *    ('ior', ('ieq', 'a@8', 0), ('ieq', 'b@8', 0)) => ('ieq', ('umin', 'a', 'b'), 0)
 *    ('iand', ('ine', 'a@8', 0), ('ine', 'b@8', 0)) => ('ine', ('umin', 'a', 'b'), 0)
 *    ('ior', ('ine', 'a@8', 0), ('ine', 'b@8', 0)) => ('ine', ('umax', 'a', 'b'), 0)
 *    ('ineg', ('b2i8', 'a@8')) => a
 *    ('ishl', 'a@8', ('iand', 7, 'b')) => ('ishl', 'a', 'b')
 *    ('ishr', 'a@8', ('iand', 7, 'b')) => ('ishr', 'a', 'b')
 *    ('ushr', 'a@8', ('iand', 7, 'b')) => ('ushr', 'a', 'b')
 *    ('ushr', 'a@8', ('ishl(is_used_once)', ('iand', 'b', 1), 2)) => ('ushr', 'a', ('ishl', 'b', 2))
 *    ('iand', ('ieq', 'a@16', 0), ('ieq', 'b@16', 0)) => ('ieq', ('ior', 'a', 'b'), 0)
 *    ('ior', ('ine', 'a@16', 0), ('ine', 'b@16', 0)) => ('ine', ('ior', 'a', 'b'), 0)
 *    ('iand', ('ieq', 'a@16', 0), ('ieq', 'b@16', 0)) => ('ieq', ('umax', 'a', 'b'), 0)
 *    ('ior', ('ieq', 'a@16', 0), ('ieq', 'b@16', 0)) => ('ieq', ('umin', 'a', 'b'), 0)
 *    ('iand', ('ine', 'a@16', 0), ('ine', 'b@16', 0)) => ('ine', ('umin', 'a', 'b'), 0)
 *    ('ior', ('ine', 'a@16', 0), ('ine', 'b@16', 0)) => ('ine', ('umax', 'a', 'b'), 0)
 *    ('ineg', ('b2i16', 'a@16')) => a
 *    ('ishl', 'a@16', ('iand', 15, 'b')) => ('ishl', 'a', 'b')
 *    ('ishr', 'a@16', ('iand', 15, 'b')) => ('ishr', 'a', 'b')
 *    ('ushr', 'a@16', ('iand', 15, 'b')) => ('ushr', 'a', 'b')
 *    ('ushr', 'a@16', ('ishl(is_used_once)', ('iand', 'b', 1), 3)) => ('ushr', 'a', ('ishl', 'b', 3))
 *    ('iand', ('ieq', 'a@32', 0), ('ieq', 'b@32', 0)) => ('ieq', ('ior', 'a', 'b'), 0)
 *    ('ior', ('ine', 'a@32', 0), ('ine', 'b@32', 0)) => ('ine', ('ior', 'a', 'b'), 0)
 *    ('iand', ('ieq', 'a@32', 0), ('ieq', 'b@32', 0)) => ('ieq', ('umax', 'a', 'b'), 0)
 *    ('ior', ('ieq', 'a@32', 0), ('ieq', 'b@32', 0)) => ('ieq', ('umin', 'a', 'b'), 0)
 *    ('iand', ('ine', 'a@32', 0), ('ine', 'b@32', 0)) => ('ine', ('umin', 'a', 'b'), 0)
 *    ('ior', ('ine', 'a@32', 0), ('ine', 'b@32', 0)) => ('ine', ('umax', 'a', 'b'), 0)
 *    ('ineg', ('b2i32', 'a@32')) => a
 *    ('ishl', 'a@32', ('iand', 31, 'b')) => ('ishl', 'a', 'b')
 *    ('ishr', 'a@32', ('iand', 31, 'b')) => ('ishr', 'a', 'b')
 *    ('ushr', 'a@32', ('iand', 31, 'b')) => ('ushr', 'a', 'b')
 *    ('ushr', 'a@32', ('ishl(is_used_once)', ('iand', 'b', 1), 4)) => ('ushr', 'a', ('ishl', 'b', 4))
 *    ('iand', ('ieq', 'a@64', 0), ('ieq', 'b@64', 0)) => ('ieq', ('ior', 'a', 'b'), 0)
 *    ('ior', ('ine', 'a@64', 0), ('ine', 'b@64', 0)) => ('ine', ('ior', 'a', 'b'), 0)
 *    ('iand', ('ieq', 'a@64', 0), ('ieq', 'b@64', 0)) => ('ieq', ('umax', 'a', 'b'), 0)
 *    ('ior', ('ieq', 'a@64', 0), ('ieq', 'b@64', 0)) => ('ieq', ('umin', 'a', 'b'), 0)
 *    ('iand', ('ine', 'a@64', 0), ('ine', 'b@64', 0)) => ('ine', ('umin', 'a', 'b'), 0)
 *    ('ior', ('ine', 'a@64', 0), ('ine', 'b@64', 0)) => ('ine', ('umax', 'a', 'b'), 0)
 *    ('ineg', ('b2i64', 'a@64')) => a
 *    ('ishl', 'a@64', ('iand', 63, 'b')) => ('ishl', 'a', 'b')
 *    ('ishr', 'a@64', ('iand', 63, 'b')) => ('ishr', 'a', 'b')
 *    ('ushr', 'a@64', ('iand', 63, 'b')) => ('ushr', 'a', 'b')
 *    ('ushr', 'a@64', ('ishl(is_used_once)', ('iand', 'b', 1), 5)) => ('ushr', 'a', ('ishl', 'b', 5))
 *    ('ior', ('ieq', 'a', 0), ('ieq', 'a', 1)) => ('uge', 1, 'a')
 *    ('ior', ('uge', 1, 'a'), ('ieq', 'a', 2)) => ('uge', 2, 'a')
 *    ('ior', ('uge', 2, 'a'), ('ieq', 'a', 3)) => ('uge', 3, 'a')
 *    ('ior', 'a', ('ieq', 'a', False)) => True
 *    ('ine', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ine', 'a', 'b')
 *    ('b2i', ('ine', 'a@1', 'b@1')) => ('b2i', ('ixor', 'a', 'b'))
 *    ('ishl', ('b2i32', ('ine', ('iand', 'a@32', '#b(is_pos_power_of_two)'), 0)), '#c') => ('bcsel', ('ige', ('iand', 'c', 31), ('find_lsb', 'b')), ('ishl', ('iand', 'a', 'b'), ('iadd', ('iand', 'c', 31), ('ineg', ('find_lsb', 'b')))), ('ushr', ('iand', 'a', 'b'), ('iadd', ('ineg', ('iand', 'c', 31)), ('find_lsb', 'b'))))
 *    ('b2i32', ('ine', ('iand', 'a@32', '#b(is_pos_power_of_two)'), 0)) => ('ushr', ('iand', 'a', 'b'), ('find_lsb', 'b'))
 *    ('ior', ('b2i', 'a'), ('iand', 'b', 1)) => ('iand', ('ior', ('b2i', 'a'), 'b'), 1)
 *    ('iand', ('b2i', 'a'), ('iand', 'b', 1)) => ('iand', ('b2i', 'a'), 'b')
 *    ('iand', ('inot', ('iand', ('ior', ('ieq', 'a', 0), 'b'), 'c')), ('ilt', 'a', 0)) => ('iand', ('inot', ('iand', 'b', 'c')), ('ilt', 'a', 0))
 *    ('iand', ('inot', ('iand', ('ieq', ('umin', 'a', 'b'), 0), 'c')), ('ilt', 'a', 0)) => ('iand', ('inot', ('iand', ('ieq', 'b', 0), 'c')), ('ilt', 'a', 0))
 *    ('flt', 'a', ('fmax', 'b', 'a')) => ('flt', 'a', 'b')
 *    ('flt', ('fmin', 'a', 'b'), 'a') => ('flt', 'b', 'a')
 *    ('~fge', 'a', ('fmin', 'b', 'a')) => True
 *    ('~fge', ('fmax', 'a', 'b'), 'a') => True
 *    ('flt', 'a', ('fmin', 'b', 'a')) => False
 *    ('flt', ('fmax', 'a', 'b'), 'a') => False
 *    ('~fge', 'a', ('fmax', 'b', 'a')) => ('fge', 'a', 'b')
 *    ('~fge', ('fmin', 'a', 'b'), 'a') => ('fge', 'b', 'a')
 *    ('ilt', 'a', ('imax', 'b', 'a')) => ('ilt', 'a', 'b')
 *    ('ilt', ('imin', 'a', 'b'), 'a') => ('ilt', 'b', 'a')
 *    ('ige', 'a', ('imin', 'b', 'a')) => True
 *    ('ige', ('imax', 'a', 'b'), 'a') => True
 *    ('ult', 'a', ('umax', 'b', 'a')) => ('ult', 'a', 'b')
 *    ('ult', ('umin', 'a', 'b'), 'a') => ('ult', 'b', 'a')
 *    ('uge', 'a', ('umin', 'b', 'a')) => True
 *    ('uge', ('umax', 'a', 'b'), 'a') => True
 *    ('ilt', 'a', ('imin', 'b', 'a')) => False
 *    ('ilt', ('imax', 'a', 'b'), 'a') => False
 *    ('ige', 'a', ('imax', 'b', 'a')) => ('ige', 'a', 'b')
 *    ('ige', ('imin', 'a', 'b'), 'a') => ('ige', 'b', 'a')
 *    ('ult', 'a', ('umin', 'b', 'a')) => False
 *    ('ult', ('umax', 'a', 'b'), 'a') => False
 *    ('uge', 'a', ('umax', 'b', 'a')) => ('uge', 'a', 'b')
 *    ('uge', ('umin', 'a', 'b'), 'a') => ('uge', 'b', 'a')
 *    ('ult', 'a', ('iand', 'b', 'a')) => False
 *    ('ult', ('ior', 'a', 'b'), 'a') => False
 *    ('uge', 'a', ('iand', 'b', 'a')) => True
 *    ('uge', ('ior', 'a', 'b'), 'a') => True
 *    ('ilt', '#a', ('imax', '#b', 'c')) => ('ior', ('ilt', 'a', 'b'), ('ilt', 'a', 'c'))
 *    ('ilt', ('imin', '#a', 'b'), '#c') => ('ior', ('ilt', 'a', 'c'), ('ilt', 'b', 'c'))
 *    ('ige', '#a', ('imin', '#b', 'c')) => ('ior', ('ige', 'a', 'b'), ('ige', 'a', 'c'))
 *    ('ige', ('imax', '#a', 'b'), '#c') => ('ior', ('ige', 'a', 'c'), ('ige', 'b', 'c'))
 *    ('ult', '#a', ('umax', '#b', 'c')) => ('ior', ('ult', 'a', 'b'), ('ult', 'a', 'c'))
 *    ('ult', ('umin', '#a', 'b'), '#c') => ('ior', ('ult', 'a', 'c'), ('ult', 'b', 'c'))
 *    ('uge', '#a', ('umin', '#b', 'c')) => ('ior', ('uge', 'a', 'b'), ('uge', 'a', 'c'))
 *    ('uge', ('umax', '#a', 'b'), '#c') => ('ior', ('uge', 'a', 'c'), ('uge', 'b', 'c'))
 *    ('ilt', '#a', ('imin', '#b', 'c')) => ('iand', ('ilt', 'a', 'b'), ('ilt', 'a', 'c'))
 *    ('ilt', ('imax', '#a', 'b'), '#c') => ('iand', ('ilt', 'a', 'c'), ('ilt', 'b', 'c'))
 *    ('ige', '#a', ('imax', '#b', 'c')) => ('iand', ('ige', 'a', 'b'), ('ige', 'a', 'c'))
 *    ('ige', ('imin', '#a', 'b'), '#c') => ('iand', ('ige', 'a', 'c'), ('ige', 'b', 'c'))
 *    ('ult', '#a', ('umin', '#b', 'c')) => ('iand', ('ult', 'a', 'b'), ('ult', 'a', 'c'))
 *    ('ult', ('umax', '#a', 'b'), '#c') => ('iand', ('ult', 'a', 'c'), ('ult', 'b', 'c'))
 *    ('uge', '#a', ('umax', '#b', 'c')) => ('iand', ('uge', 'a', 'b'), ('uge', 'a', 'c'))
 *    ('uge', ('umin', '#a', 'b'), '#c') => ('iand', ('uge', 'a', 'c'), ('uge', 'b', 'c'))
 *    ('bcsel', ('ilt', 'a', 0), ('ineg', ('ishr', 'a', 'b')), ('ishr', 'a', 'b')) => ('iabs', ('ishr', 'a', 'b'))
 *    ('iabs', ('ishr', ('iabs', 'a'), 'b')) => ('ishr', ('iabs', 'a'), 'b')
 *    ('fabs', ('slt', 'a', 'b')) => ('slt', 'a', 'b')
 *    ('fabs', ('sge', 'a', 'b')) => ('sge', 'a', 'b')
 *    ('fabs', ('seq', 'a', 'b')) => ('seq', 'a', 'b')
 *    ('fabs', ('sne', 'a', 'b')) => ('sne', 'a', 'b')
 *    ('slt', 'a', 'b') => ('b2f', ('flt', 'a', 'b'))
 *    ('sge', 'a', 'b') => ('b2f', ('fge', 'a', 'b'))
 *    ('seq', 'a', 'b') => ('b2f', ('feq', 'a', 'b'))
 *    ('sne', 'a', 'b') => ('b2f', ('fneu', 'a', 'b'))
 *    ('seq', ('seq', 'a', 'b'), 1.0) => ('seq', 'a', 'b')
 *    ('seq', ('sne', 'a', 'b'), 1.0) => ('sne', 'a', 'b')
 *    ('seq', ('slt', 'a', 'b'), 1.0) => ('slt', 'a', 'b')
 *    ('seq', ('sge', 'a', 'b'), 1.0) => ('sge', 'a', 'b')
 *    ('sne', ('seq', 'a', 'b'), 0.0) => ('seq', 'a', 'b')
 *    ('sne', ('sne', 'a', 'b'), 0.0) => ('sne', 'a', 'b')
 *    ('sne', ('slt', 'a', 'b'), 0.0) => ('slt', 'a', 'b')
 *    ('sne', ('sge', 'a', 'b'), 0.0) => ('sge', 'a', 'b')
 *    ('seq', ('seq', 'a', 'b'), 0.0) => ('sne', 'a', 'b')
 *    ('seq', ('sne', 'a', 'b'), 0.0) => ('seq', 'a', 'b')
 *    ('seq', ('slt', 'a', 'b'), 0.0) => ('sge', 'a', 'b')
 *    ('seq', ('sge', 'a', 'b'), 0.0) => ('slt', 'a', 'b')
 *    ('sne', ('seq', 'a', 'b'), 1.0) => ('sne', 'a', 'b')
 *    ('sne', ('sne', 'a', 'b'), 1.0) => ('seq', 'a', 'b')
 *    ('sne', ('slt', 'a', 'b'), 1.0) => ('sge', 'a', 'b')
 *    ('sne', ('sge', 'a', 'b'), 1.0) => ('slt', 'a', 'b')
 *    ('fall_equal2', 'a', 'b') => ('fmin', ('seq', 'a.x', 'b.x'), ('seq', 'a.y', 'b.y'))
 *    ('fall_equal3', 'a', 'b') => ('seq', ('fany_nequal3', 'a', 'b'), 0.0)
 *    ('fall_equal4', 'a', 'b') => ('seq', ('fany_nequal4', 'a', 'b'), 0.0)
 *    ('fall_equal8', 'a', 'b') => ('seq', ('fany_nequal8', 'a', 'b'), 0.0)
 *    ('fall_equal16', 'a', 'b') => ('seq', ('fany_nequal16', 'a', 'b'), 0.0)
 *    ('fany_nequal2', 'a', 'b') => ('fmax', ('sne', 'a.x', 'b.x'), ('sne', 'a.y', 'b.y'))
 *    ('fany_nequal3', 'a', 'b') => ('fsat', ('fdot3', ('sne', 'a', 'b'), ('sne', 'a', 'b')))
 *    ('fany_nequal4', 'a', 'b') => ('fsat', ('fdot4', ('sne', 'a', 'b'), ('sne', 'a', 'b')))
 *    ('fany_nequal8', 'a', 'b') => ('fsat', ('fdot8', ('sne', 'a', 'b'), ('sne', 'a', 'b')))
 *    ('fany_nequal16', 'a', 'b') => ('fsat', ('fdot16', ('sne', 'a', 'b'), ('sne', 'a', 'b')))
 *    ('ball_iequal2', 'a', 'b') => ('iand', ('ieq', 'a.x', 'b.x'), ('ieq', 'a.y', 'b.y'))
 *    ('ball_iequal3', 'a', 'b') => ('iand', ('ieq', 'a.x', 'b.x'), ('iand', ('ieq', 'a.y', 'b.y'), ('ieq', 'a.z', 'b.z')))
 *    ('ball_iequal4', 'a', 'b') => ('iand', ('iand', ('ieq', 'a.x', 'b.x'), ('ieq', 'a.y', 'b.y')), ('iand', ('ieq', 'a.z', 'b.z'), ('ieq', 'a.w', 'b.w')))
 *    ('ball_iequal8', 'a', 'b') => ('iand', ('iand', ('iand', ('ieq', 'a.a', 'b.a'), ('ieq', 'a.b', 'b.b')), ('iand', ('ieq', 'a.c', 'b.c'), ('ieq', 'a.d', 'b.d'))), ('iand', ('iand', ('ieq', 'a.e', 'b.e'), ('ieq', 'a.f', 'b.f')), ('iand', ('ieq', 'a.g', 'b.g'), ('ieq', 'a.h', 'b.h'))))
 *    ('ball_iequal16', 'a', 'b') => ('iand', ('iand', ('iand', ('iand', ('ieq', 'a.a', 'b.a'), ('ieq', 'a.b', 'b.b')), ('iand', ('ieq', 'a.c', 'b.c'), ('ieq', 'a.d', 'b.d'))), ('iand', ('iand', ('ieq', 'a.e', 'b.e'), ('ieq', 'a.f', 'b.f')), ('iand', ('ieq', 'a.g', 'b.g'), ('ieq', 'a.h', 'b.h')))), ('iand', ('iand', ('iand', ('ieq', 'a.i', 'b.i'), ('ieq', 'a.j', 'b.j')), ('iand', ('ieq', 'a.k', 'b.k'), ('ieq', 'a.l', 'b.l'))), ('iand', ('iand', ('ieq', 'a.m', 'b.m'), ('ieq', 'a.n', 'b.n')), ('iand', ('ieq', 'a.o', 'b.o'), ('ieq', 'a.p', 'b.p')))))
 *    ('ball_fequal2', 'a', 'b') => ('iand', ('feq', 'a.x', 'b.x'), ('feq', 'a.y', 'b.y'))
 *    ('ball_fequal3', 'a', 'b') => ('iand', ('feq', 'a.x', 'b.x'), ('iand', ('feq', 'a.y', 'b.y'), ('feq', 'a.z', 'b.z')))
 *    ('ball_fequal4', 'a', 'b') => ('iand', ('iand', ('feq', 'a.x', 'b.x'), ('feq', 'a.y', 'b.y')), ('iand', ('feq', 'a.z', 'b.z'), ('feq', 'a.w', 'b.w')))
 *    ('ball_fequal8', 'a', 'b') => ('iand', ('iand', ('iand', ('feq', 'a.a', 'b.a'), ('feq', 'a.b', 'b.b')), ('iand', ('feq', 'a.c', 'b.c'), ('feq', 'a.d', 'b.d'))), ('iand', ('iand', ('feq', 'a.e', 'b.e'), ('feq', 'a.f', 'b.f')), ('iand', ('feq', 'a.g', 'b.g'), ('feq', 'a.h', 'b.h'))))
 *    ('ball_fequal16', 'a', 'b') => ('iand', ('iand', ('iand', ('iand', ('feq', 'a.a', 'b.a'), ('feq', 'a.b', 'b.b')), ('iand', ('feq', 'a.c', 'b.c'), ('feq', 'a.d', 'b.d'))), ('iand', ('iand', ('feq', 'a.e', 'b.e'), ('feq', 'a.f', 'b.f')), ('iand', ('feq', 'a.g', 'b.g'), ('feq', 'a.h', 'b.h')))), ('iand', ('iand', ('iand', ('feq', 'a.i', 'b.i'), ('feq', 'a.j', 'b.j')), ('iand', ('feq', 'a.k', 'b.k'), ('feq', 'a.l', 'b.l'))), ('iand', ('iand', ('feq', 'a.m', 'b.m'), ('feq', 'a.n', 'b.n')), ('iand', ('feq', 'a.o', 'b.o'), ('feq', 'a.p', 'b.p')))))
 *    ('bany_inequal2', 'a', 'b') => ('ior', ('ine', 'a.x', 'b.x'), ('ine', 'a.y', 'b.y'))
 *    ('bany_inequal3', 'a', 'b') => ('ior', ('ine', 'a.x', 'b.x'), ('ior', ('ine', 'a.y', 'b.y'), ('ine', 'a.z', 'b.z')))
 *    ('bany_inequal4', 'a', 'b') => ('ior', ('ior', ('ine', 'a.x', 'b.x'), ('ine', 'a.y', 'b.y')), ('ior', ('ine', 'a.z', 'b.z'), ('ine', 'a.w', 'b.w')))
 *    ('bany_inequal8', 'a', 'b') => ('ior', ('ior', ('ior', ('ine', 'a.a', 'b.a'), ('ine', 'a.b', 'b.b')), ('ior', ('ine', 'a.c', 'b.c'), ('ine', 'a.d', 'b.d'))), ('ior', ('ior', ('ine', 'a.e', 'b.e'), ('ine', 'a.f', 'b.f')), ('ior', ('ine', 'a.g', 'b.g'), ('ine', 'a.h', 'b.h'))))
 *    ('bany_inequal16', 'a', 'b') => ('ior', ('ior', ('ior', ('ior', ('ine', 'a.a', 'b.a'), ('ine', 'a.b', 'b.b')), ('ior', ('ine', 'a.c', 'b.c'), ('ine', 'a.d', 'b.d'))), ('ior', ('ior', ('ine', 'a.e', 'b.e'), ('ine', 'a.f', 'b.f')), ('ior', ('ine', 'a.g', 'b.g'), ('ine', 'a.h', 'b.h')))), ('ior', ('ior', ('ior', ('ine', 'a.i', 'b.i'), ('ine', 'a.j', 'b.j')), ('ior', ('ine', 'a.k', 'b.k'), ('ine', 'a.l', 'b.l'))), ('ior', ('ior', ('ine', 'a.m', 'b.m'), ('ine', 'a.n', 'b.n')), ('ior', ('ine', 'a.o', 'b.o'), ('ine', 'a.p', 'b.p')))))
 *    ('bany_fnequal2', 'a', 'b') => ('ior', ('fneu', 'a.x', 'b.x'), ('fneu', 'a.y', 'b.y'))
 *    ('bany_fnequal3', 'a', 'b') => ('ior', ('fneu', 'a.x', 'b.x'), ('ior', ('fneu', 'a.y', 'b.y'), ('fneu', 'a.z', 'b.z')))
 *    ('bany_fnequal4', 'a', 'b') => ('ior', ('ior', ('fneu', 'a.x', 'b.x'), ('fneu', 'a.y', 'b.y')), ('ior', ('fneu', 'a.z', 'b.z'), ('fneu', 'a.w', 'b.w')))
 *    ('bany_fnequal8', 'a', 'b') => ('ior', ('ior', ('ior', ('fneu', 'a.a', 'b.a'), ('fneu', 'a.b', 'b.b')), ('ior', ('fneu', 'a.c', 'b.c'), ('fneu', 'a.d', 'b.d'))), ('ior', ('ior', ('fneu', 'a.e', 'b.e'), ('fneu', 'a.f', 'b.f')), ('ior', ('fneu', 'a.g', 'b.g'), ('fneu', 'a.h', 'b.h'))))
 *    ('bany_fnequal16', 'a', 'b') => ('ior', ('ior', ('ior', ('ior', ('fneu', 'a.a', 'b.a'), ('fneu', 'a.b', 'b.b')), ('ior', ('fneu', 'a.c', 'b.c'), ('fneu', 'a.d', 'b.d'))), ('ior', ('ior', ('fneu', 'a.e', 'b.e'), ('fneu', 'a.f', 'b.f')), ('ior', ('fneu', 'a.g', 'b.g'), ('fneu', 'a.h', 'b.h')))), ('ior', ('ior', ('ior', ('fneu', 'a.i', 'b.i'), ('fneu', 'a.j', 'b.j')), ('ior', ('fneu', 'a.k', 'b.k'), ('fneu', 'a.l', 'b.l'))), ('ior', ('ior', ('fneu', 'a.m', 'b.m'), ('fneu', 'a.n', 'b.n')), ('ior', ('fneu', 'a.o', 'b.o'), ('fneu', 'a.p', 'b.p')))))
 *    ('feq', ('seq', 'a', 'b'), 1.0) => ('feq', 'a', 'b')
 *    ('feq', ('sne', 'a', 'b'), 1.0) => ('fneu', 'a', 'b')
 *    ('feq', ('slt', 'a', 'b'), 1.0) => ('flt', 'a', 'b')
 *    ('feq', ('sge', 'a', 'b'), 1.0) => ('fge', 'a', 'b')
 *    ('fneu', ('seq', 'a', 'b'), 0.0) => ('feq', 'a', 'b')
 *    ('fneu', ('sne', 'a', 'b'), 0.0) => ('fneu', 'a', 'b')
 *    ('fneu', ('slt', 'a', 'b'), 0.0) => ('flt', 'a', 'b')
 *    ('fneu', ('sge', 'a', 'b'), 0.0) => ('fge', 'a', 'b')
 *    ('feq', ('seq', 'a', 'b'), 0.0) => ('fneu', 'a', 'b')
 *    ('feq', ('sne', 'a', 'b'), 0.0) => ('feq', 'a', 'b')
 *    ('feq', ('slt', 'a', 'b'), 0.0) => ('fge', 'a', 'b')
 *    ('feq', ('sge', 'a', 'b'), 0.0) => ('flt', 'a', 'b')
 *    ('fneu', ('seq', 'a', 'b'), 1.0) => ('fneu', 'a', 'b')
 *    ('fneu', ('sne', 'a', 'b'), 1.0) => ('feq', 'a', 'b')
 *    ('fneu', ('slt', 'a', 'b'), 1.0) => ('fge', 'a', 'b')
 *    ('fneu', ('sge', 'a', 'b'), 1.0) => ('flt', 'a', 'b')
 *    ('fneu', ('fneg', 'a'), 'a') => ('fneu', 'a', 0.0)
 *    ('feq', ('fneg', 'a'), 'a') => ('feq', 'a', 0.0)
 *    ('imul', ('b2i', 'a@1'), ('b2i', 'b@1')) => ('b2i', ('iand', 'a', 'b'))
 *    ('iand', ('b2i', 'a@1'), ('b2i', 'b@1')) => ('b2i', ('iand', 'a', 'b'))
 *    ('ior', ('b2i', 'a@1'), ('b2i', 'b@1')) => ('b2i', ('ior', 'a', 'b'))
 *    ('fmul', ('b2f', 'a@1'), ('b2f', 'b@1')) => ('b2f', ('iand', 'a', 'b'))
 *    ('fsat', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1'))) => ('b2f', ('ior', 'a', 'b'))
 *    ('iand', 'a@bool16', 1.0) => ('b2f', 'a')
 *    ('iand', 'a@bool32', 1.0) => ('b2f', 'a')
 *    ('flt', ('fneg', ('b2f', 'a@1')), 0) => a
 *    ('ilt', 'a', 'a') => False
 *    ('ige', 'a', 'a') => True
 *    ('ieq', 'a', 'a') => True
 *    ('ine', 'a', 'a') => False
 *    ('ult', 'a', 'a') => False
 *    ('uge', 'a', 'a') => True
 *    ('flt', 'a', 'a') => False
 *    ('fge', 'a(is_a_number)', 'a') => True
 *    ('feq', 'a(is_a_number)', 'a') => True
 *    ('fneu', 'a(is_a_number)', 'a') => False
 *    ('iand', 'a', 'a') => a
 *    ('iand', 'a', 0) => 0
 *    ('iand', 'a', -1) => a
 *    ('iand', 'a', ('inot', 'a')) => 0
 *    ('ior', 'a', 'a') => a
 *    ('ior', 'a', 0) => a
 *    ('ior', 'a', -1) => -1
 *    ('ior', 'a', ('inot', 'a')) => -1
 *    ('ixor', 'a', 'a') => 0
 *    ('ixor', 'a', 0) => a
 *    ('ixor', 'a', ('ixor', 'a', 'b')) => b
 *    ('ixor', 'a', -1) => ('inot', 'a')
 *    ('inot', ('inot', 'a')) => a
 *    ('ior', ('iand', 'a', 'b'), 'b') => b
 *    ('ior', ('ior', 'a', 'b'), 'b') => ('ior', 'a', 'b')
 *    ('iand', ('ior', 'a', 'b'), 'b') => b
 *    ('iand', ('iand', 'a', 'b'), 'b') => ('iand', 'a', 'b')
 *    ('iand', ('iand(is_used_once)', 'a', 'b'), ('iand(is_used_once)', 'a', 'c')) => ('iand', 'a', ('iand', 'b', 'c'))
 *    ('iand@64', 'a', '#b(is_lower_half_zero)') => ('pack_64_2x32_split', 0, ('iand', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b')))
 *    ('iand@64', 'a', '#b(is_upper_half_zero)') => ('pack_64_2x32_split', ('iand', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_x', 'b')), 0)
 *    ('iand@64', 'a', '#b(is_lower_half_negative_one)') => ('pack_64_2x32_split', ('unpack_64_2x32_split_x', 'a'), ('iand', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b')))
 *    ('iand@64', 'a', '#b(is_upper_half_negative_one)') => ('pack_64_2x32_split', ('iand', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_x', 'b')), ('unpack_64_2x32_split_y', 'a'))
 *    ('ior@64', 'a', '#b(is_lower_half_zero)') => ('pack_64_2x32_split', ('unpack_64_2x32_split_x', 'a'), ('ior', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b')))
 *    ('ior@64', 'a', '#b(is_upper_half_zero)') => ('pack_64_2x32_split', ('ior', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_x', 'b')), ('unpack_64_2x32_split_y', 'a'))
 *    ('ior@64', 'a', '#b(is_lower_half_negative_one)') => ('pack_64_2x32_split', -1, ('ior', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b')))
 *    ('ior@64', 'a', '#b(is_upper_half_negative_one)') => ('pack_64_2x32_split', ('ior', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_x', 'b')), -1)
 *    ('ixor@64', 'a', '#b(is_lower_half_zero)') => ('pack_64_2x32_split', ('unpack_64_2x32_split_x', 'a'), ('ixor', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b')))
 *    ('ixor@64', 'a', '#b(is_upper_half_zero)') => ('pack_64_2x32_split', ('ixor', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_x', 'b')), ('unpack_64_2x32_split_y', 'a'))
 *    ('iand', ('inot', 'a'), ('inot', 'b')) => ('inot', ('ior', 'a', 'b'))
 *    ('ior', ('inot', 'a'), ('inot', 'b')) => ('inot', ('iand', 'a', 'b'))
 *    ('ishl', 0, 'a') => 0
 *    ('ishl', 'a', 0) => a
 *    ('ishr', 0, 'a') => 0
 *    ('ishr', -1, 'a') => -1
 *    ('ishr', 'a', 0) => a
 *    ('ushr', 0, 'a') => 0
 *    ('ushr', 'a', 0) => a
 *    ('ior', ('ishl@16', 'a', 'b'), ('ushr@16', 'a', ('iadd', 16, ('ineg', 'b')))) => ('urol', 'a', 'b')
 *    ('ior', ('ishl@16', 'a', 'b'), ('ushr@16', 'a', ('isub', 16, 'b'))) => ('urol', 'a', 'b')
 *    ('ior', ('ishl@32', 'a', 'b'), ('ushr@32', 'a', ('iadd', 32, ('ineg', 'b')))) => ('urol', 'a', 'b')
 *    ('ior', ('ishl@32', 'a', 'b'), ('ushr@32', 'a', ('isub', 32, 'b'))) => ('urol', 'a', 'b')
 *    ('ior', ('ushr@16', 'a', 'b'), ('ishl@16', 'a', ('iadd', 16, ('ineg', 'b')))) => ('uror', 'a', 'b')
 *    ('ior', ('ushr@16', 'a', 'b'), ('ishl@16', 'a', ('isub', 16, 'b'))) => ('uror', 'a', 'b')
 *    ('ior', ('ushr@32', 'a', 'b'), ('ishl@32', 'a', ('iadd', 32, ('ineg', 'b')))) => ('uror', 'a', 'b')
 *    ('ior', ('ushr@32', 'a', 'b'), ('ishl@32', 'a', ('isub', 32, 'b'))) => ('uror', 'a', 'b')
 *    ('urol@8', 'a', 'b') => ('ior', ('ishl', 'a', 'b'), ('ushr', 'a', ('isub', 8, 'b')))
 *    ('urol@16', 'a', 'b') => ('ior', ('ishl', 'a', 'b'), ('ushr', 'a', ('isub', 16, 'b')))
 *    ('urol@32', 'a', 'b') => ('ior', ('ishl', 'a', 'b'), ('ushr', 'a', ('isub', 32, 'b')))
 *    ('urol@64', 'a', 'b') => ('ior', ('ishl', 'a', 'b'), ('ushr', 'a', ('isub', 64, 'b')))
 *    ('uror@8', 'a', 'b') => ('ior', ('ushr', 'a', 'b'), ('ishl', 'a', ('isub', 8, 'b')))
 *    ('uror@16', 'a', 'b') => ('ior', ('ushr', 'a', 'b'), ('ishl', 'a', ('isub', 16, 'b')))
 *    ('uror@32', 'a', 'b') => ('ior', ('ushr', 'a', 'b'), ('ishl', 'a', ('isub', 32, 'b')))
 *    ('uror@64', 'a', 'b') => ('ior', ('ushr', 'a', 'b'), ('ishl', 'a', ('isub', 64, 'b')))
 *    ('~fexp2', ('flog2', 'a')) => a
 *    ('~flog2', ('fexp2', 'a')) => a
 *    ('fpow', 'a', 'b') => ('fexp2', ('fmul', ('flog2', 'a'), 'b'))
 *    ('~fexp2', ('fmul', ('flog2', 'a'), 'b')) => ('fpow', 'a', 'b')
 *    ('~fexp2', ('fadd', ('fmul', ('flog2', 'a'), 'b'), ('fmul', ('flog2', 'c'), 'd'))) => ('~fmul', ('fpow', 'a', 'b'), ('fpow', 'c', 'd'))
 *    ('~fexp2', ('fmul', ('flog2', 'a'), 0.5)) => ('fsqrt', 'a')
 *    ('~fexp2', ('fmul', ('flog2', 'a'), 2.0)) => ('fmul', 'a', 'a')
 *    ('~fexp2', ('fmul', ('flog2', 'a'), 4.0)) => ('fmul', ('fmul', 'a', 'a'), ('fmul', 'a', 'a'))
 *    ('~fpow', 'a', 1.0) => a
 *    ('~fpow', 'a', 2.0) => ('fmul', 'a', 'a')
 *    ('~fpow', 'a', 4.0) => ('fmul', ('fmul', 'a', 'a'), ('fmul', 'a', 'a'))
 *    ('~fpow', 2.0, 'a') => ('fexp2', 'a')
 *    ('~fpow', ('fpow', 'a', 2.2), 0.454545) => a
 *    ('~fpow', ('fabs', ('fpow', 'a', 2.2)), 0.454545) => ('fabs', 'a')
 *    ('~fsqrt', ('fexp2', 'a')) => ('fexp2', ('fmul', 0.5, 'a'))
 *    ('~frcp', ('fexp2', 'a')) => ('fexp2', ('fneg', 'a'))
 *    ('~frsq', ('fexp2', 'a')) => ('fexp2', ('fmul', -0.5, 'a'))
 *    ('~flog2', ('fsqrt', 'a')) => ('fmul', 0.5, ('flog2', 'a'))
 *    ('~flog2', ('frcp', 'a')) => ('fneg', ('flog2', 'a'))
 *    ('~flog2', ('frsq', 'a')) => ('fmul', -0.5, ('flog2', 'a'))
 *    ('~flog2', ('fpow', 'a', 'b')) => ('fmul', 'b', ('flog2', 'a'))
 *    ('~fmul', ('fexp2(is_used_once)', 'a'), ('fexp2(is_used_once)', 'b')) => ('fexp2', ('fadd', 'a', 'b'))
 *    ('bcsel', ('flt', 'a', 0.0), 0.0, ('fsqrt', 'a')) => ('fsqrt', ('fmax', 'a', 0.0))
 *    ('~fmul', ('fsqrt', 'a'), ('fsqrt', 'a')) => ('fabs', 'a')
 *    ('~fmulz', ('fsqrt', 'a'), ('fsqrt', 'a')) => ('fabs', 'a')
 *    ('~fdiv', 1.0, 'a') => ('frcp', 'a')
 *    ('fdiv', 'a', 'b') => ('fmul', 'a', ('frcp', 'b'))
 *    ('~frcp', ('frcp', 'a')) => a
 *    ('~frcp', ('fsqrt', 'a')) => ('frsq', 'a')
 *    ('fsqrt', 'a') => ('frcp', ('frsq', 'a'))
 *    ('~frcp', ('frsq', 'a')) => ('fsqrt', 'a')
 *    ('fsin', 'a') => ('ffma', ('ffma', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0)))), 4.0), ('fabs', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0)))), 4.0)), ('fneg', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0)))), 4.0))), 0.225, ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0)))), 4.0))
 *    ('fcos', 'a') => ('ffma', ('ffma', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0)))), 4.0), ('fabs', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0)))), 4.0)), ('fneg', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0)))), 4.0))), 0.225, ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0)))), 4.0))
 *    ('ieq', 'a', True) => a
 *    ('ine(is_not_used_by_if)', 'a', True) => ('inot', 'a')
 *    ('ine', 'a', False) => a
 *    ('ieq(is_not_used_by_if)', 'a', False) => ('inot', 'a')
 *    ('bcsel', 'a', True, False) => a
 *    ('bcsel', 'a', False, True) => ('inot', 'a')
 *    ('bcsel', True, 'b', 'c') => b
 *    ('bcsel', False, 'b', 'c') => c
 *    ('bcsel@16', 'a', 1.0, 0.0) => ('b2f', 'a')
 *    ('bcsel@16', 'a', 0.0, 1.0) => ('b2f', ('inot', 'a'))
 *    ('bcsel@16', 'a', -1.0, -0.0) => ('fneg', ('b2f', 'a'))
 *    ('bcsel@16', 'a', -0.0, -1.0) => ('fneg', ('b2f', ('inot', 'a')))
 *    ('bcsel@32', 'a', 1.0, 0.0) => ('b2f', 'a')
 *    ('bcsel@32', 'a', 0.0, 1.0) => ('b2f', ('inot', 'a'))
 *    ('bcsel@32', 'a', -1.0, -0.0) => ('fneg', ('b2f', 'a'))
 *    ('bcsel@32', 'a', -0.0, -1.0) => ('fneg', ('b2f', ('inot', 'a')))
 *    ('bcsel@64', 'a', 1.0, 0.0) => ('b2f', 'a')
 *    ('bcsel@64', 'a', 0.0, 1.0) => ('b2f', ('inot', 'a'))
 *    ('bcsel@64', 'a', -1.0, -0.0) => ('fneg', ('b2f', 'a'))
 *    ('bcsel@64', 'a', -0.0, -1.0) => ('fneg', ('b2f', ('inot', 'a')))
 *    ('bcsel', 'a', 'b', 'b') => b
 *    ('~fcsel', 'a', 'b', 'b') => b
 *    ('bcsel', 'a', -1, 0) => ('ineg', ('b2i', 'a@1'))
 *    ('bcsel', 'a', 0, -1) => ('ineg', ('b2i', ('inot', 'a')))
 *    ('bcsel', 'a', 1, 0) => ('b2i', 'a@1')
 *    ('bcsel', 'a', 0, 1) => ('b2i', ('inot', 'a'))
 *    ('iand', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('iand', 'a', 'b')))
 *    ('ior', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('ior', 'a', 'b')))
 *    ('ieq', ('ineg', ('b2i', 'a@1')), -1) => a
 *    ('ine', ('ineg', ('b2i', 'a@1')), -1) => ('inot', 'a')
 *    ('ige', ('ineg', ('b2i', 'a@1')), 0) => ('inot', 'a')
 *    ('ilt', ('ineg', ('b2i', 'a@1')), 0) => a
 *    ('ult', 0, ('ineg', ('b2i', 'a@1'))) => a
 *    ('iand', ('ineg', ('b2i', 'a')), 1.0) => ('b2f', 'a')
 *    ('iand', ('ineg', ('b2i', 'a')), 1) => ('b2i', 'a')
 *    ('imax', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('iand', 'a', 'b')))
 *    ('imin', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('ior', 'a', 'b')))
 *    ('umax', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('ior', 'a', 'b')))
 *    ('umin', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('iand', 'a', 'b')))
 *    ('umax', ('b2i', 'a@1'), ('b2i', 'b@1')) => ('b2i', ('ior', 'a', 'b'))
 *    ('umin', ('b2i', 'a@1'), ('b2i', 'b@1')) => ('b2i', ('iand', 'a', 'b'))
 *    ('ine', ('umin', ('ineg', ('b2i', 'a@1')), 'b'), 0) => ('iand', 'a', ('ine', 'b', 0))
 *    ('ine', ('umax', ('ineg', ('b2i', 'a@1')), 'b'), 0) => ('ior', 'a', ('ine', 'b', 0))
 *    ('f2i', ('ftrunc', 'a')) => ('f2i', 'a')
 *    ('f2u', ('ftrunc', 'a')) => ('f2u', 'a')
 *    ('inot', ('f2b1', 'a')) => ('feq', 'a', 0.0)
 *    ('f2fmp', ('f2f32', 'a@16')) => a
 *    ('i2imp', ('i2i32', 'a@16')) => a
 *    ('i2imp', ('u2u32', 'a@16')) => a
 *    ('f2imp', ('f2f32', 'a@16')) => ('f2i16', 'a')
 *    ('f2ump', ('f2f32', 'a@16')) => ('f2u16', 'a')
 *    ('i2fmp', ('i2i32', 'a@16')) => ('i2f16', 'a')
 *    ('u2fmp', ('u2u32', 'a@16')) => ('u2f16', 'a')
 *    ('f2fmp', ('b2f32', 'a@1')) => ('b2f16', 'a')
 *    ('i2imp', ('b2i32', 'a@1')) => ('b2i16', 'a')
 *    ('i2imp', ('b2i32', 'a@1')) => ('b2i16', 'a')
 *    ('f2imp', ('b2f32', 'a@1')) => ('b2i16', 'a')
 *    ('f2ump', ('b2f32', 'a@1')) => ('b2i16', 'a')
 *    ('i2fmp', ('b2i32', 'a@1')) => ('b2f16', 'a')
 *    ('u2fmp', ('b2i32', 'a@1')) => ('b2f16', 'a')
 *    ('f2f32', ('f2fmp', 'a@32')) => a
 *    ('i2i32', ('i2imp', 'a@32')) => a
 *    ('u2u32', ('i2imp', 'a@32')) => a
 *    ('i2i32', ('f2imp', 'a@32')) => ('f2i32', 'a')
 *    ('u2u32', ('f2ump', 'a@32')) => ('f2u32', 'a')
 *    ('f2f32', ('i2fmp', 'a@32')) => ('i2f32', 'a')
 *    ('f2f32', ('u2fmp', 'a@32')) => ('u2f32', 'a')
 *    ('~f2f32', ('f2f64', 'a@32')) => a
 *    ('ffloor', 'a(is_integral)') => a
 *    ('fceil', 'a(is_integral)') => a
 *    ('ftrunc', 'a(is_integral)') => a
 *    ('fround_even', 'a(is_integral)') => a
 *    ('~ffract', 'a(is_integral)') => 0.0
 *    ('fabs', 'a(is_not_negative)') => a
 *    ('iabs', 'a(is_not_negative)') => a
 *    ('fsat', 'a(is_not_positive)') => 0.0
 *    ('~fmin', 'a(is_not_negative)', 1.0) => ('fsat', 'a')
 *    ('flt', ('fadd', ('fmul', ('fsat', 'a'), ('fneg', ('fsat', 'a'))), 1.0), 0.0) => False
 *    ('flt', ('fadd', ('fneg', ('fmul', ('fsat', 'a'), ('fsat', 'a'))), 1.0), 0.0) => False
 *    ('fmax', ('fadd', ('fmul', ('fsat', 'a'), ('fneg', ('fsat', 'a'))), 1.0), 0.0) => ('fadd', ('fmul', ('fsat', 'a'), ('fneg', ('fsat', 'a'))), 1.0)
 *    ('fmax', ('fadd', ('fneg', ('fmul', ('fsat', 'a'), ('fsat', 'a'))), 1.0), 0.0) => ('fadd', ('fneg', ('fmul', ('fsat', 'a'), ('fsat', 'a'))), 1.0)
 *    ('fneu', 'a(is_not_zero)', 0.0) => True
 *    ('feq', 'a(is_not_zero)', 0.0) => False
 *    ('fge', 'a(is_a_number_not_negative)', 'b(is_a_number_not_positive)') => True
 *    ('fge', 'a(is_not_positive)', 'b(is_gt_zero)') => False
 *    ('fge', 'a(is_lt_zero)', 'b(is_not_negative)') => False
 *    ('flt', 'a(is_not_negative)', 'b(is_not_positive)') => False
 *    ('flt', 'a(is_a_number_not_positive)', 'b(is_a_number_gt_zero)') => True
 *    ('flt', 'a(is_a_number_lt_zero)', 'b(is_a_number_not_negative)') => True
 *    ('ine', 'a(is_not_zero)', 0) => True
 *    ('ieq', 'a(is_not_zero)', 0) => False
 *    ('ige', 'a(is_not_negative)', 'b(is_not_positive)') => True
 *    ('ige', 'a(is_not_positive)', 'b(is_gt_zero)') => False
 *    ('ige', 'a(is_lt_zero)', 'b(is_not_negative)') => False
 *    ('ilt', 'a(is_not_negative)', 'b(is_not_positive)') => False
 *    ('ilt', 'a(is_not_positive)', 'b(is_gt_zero)') => True
 *    ('ilt', 'a(is_lt_zero)', 'b(is_not_negative)') => True
 *    ('ult', 0, 'a(is_gt_zero)') => True
 *    ('ult', 'a', 0) => False
 *    ('unpack_64_2x32_split_x', ('pack_64_2x32_split', 'a', 'b')) => a
 *    ('unpack_64_2x32_split_y', ('pack_64_2x32_split', 'a', 'b')) => b
 *    ('unpack_64_2x32_split_x', ('pack_64_2x32', 'a')) => a.x
 *    ('unpack_64_2x32_split_y', ('pack_64_2x32', 'a')) => a.y
 *    ('unpack_64_2x32_split_x', ('u2u64', 'a@32')) => a
 *    ('unpack_64_2x32_split_y', ('u2u64', 'a')) => 0
 *    ('unpack_64_2x32_split_x', ('i2i64', 'a@32')) => a
 *    ('unpack_64_2x32_split_y', ('i2i64(is_used_once)', 'a@32')) => ('ishr', 'a', 31)
 *    ('unpack_64_2x32', ('pack_64_2x32_split', 'a', 'b')) => ('vec2', 'a', 'b')
 *    ('unpack_64_2x32', ('pack_64_2x32', 'a')) => a
 *    ('unpack_double_2x32_dxil', ('pack_double_2x32_dxil', 'a')) => a
 *    ('pack_64_2x32_split', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_y', 'a')) => a
 *    ('pack_64_2x32', ('vec2', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_y', 'a'))) => a
 *    ('pack_64_2x32', ('unpack_64_2x32', 'a')) => a
 *    ('pack_double_2x32_dxil', ('unpack_double_2x32_dxil', 'a')) => a
 *    ('iand', ('ieq', ('unpack_32_2x16_split_x', 'a'), '#b'), ('ieq', ('unpack_32_2x16_split_y', 'a'), '#c')) => ('ieq', 'a', ('pack_32_2x16_split', 'b', 'c'))
 *    ('ushr', 'a@16', 8) => ('extract_u8', 'a', 1)
 *    ('ushr', 'a@32', 24) => ('extract_u8', 'a', 3)
 *    ('ushr', 'a@64', 56) => ('extract_u8', 'a', 7)
 *    ('ishr', 'a@16', 8) => ('extract_i8', 'a', 1)
 *    ('ishr', 'a@32', 24) => ('extract_i8', 'a', 3)
 *    ('ishr', 'a@64', 56) => ('extract_i8', 'a', 7)
 *    ('iand', 255, 'a') => ('extract_u8', 'a', 0)
 *    ('u2u8', ('extract_u16', 'a', 1)) => ('u2u8', ('extract_u8', 'a', 2))
 *    ('u2u8', ('ushr', 'a', 8)) => ('u2u8', ('extract_u8', 'a', 1))
 *    ('i2i16', ('u2u8', ('extract_u8', 'a', 'b'))) => ('i2i16', ('extract_i8', 'a', 'b'))
 *    ('u2u16', ('u2u8', ('extract_u8', 'a', 'b'))) => ('u2u16', ('extract_u8', 'a', 'b'))
 *    ('ubfe', 'a', 0, 8) => ('extract_u8', 'a', 0)
 *    ('ubfe', 'a', 8, 8) => ('extract_u8', 'a', 1)
 *    ('ubfe', 'a', 16, 8) => ('extract_u8', 'a', 2)
 *    ('ubfe', 'a', 24, 8) => ('extract_u8', 'a', 3)
 *    ('ibfe', 'a', 0, 8) => ('extract_i8', 'a', 0)
 *    ('ibfe', 'a', 8, 8) => ('extract_i8', 'a', 1)
 *    ('ibfe', 'a', 16, 8) => ('extract_i8', 'a', 2)
 *    ('ibfe', 'a', 24, 8) => ('extract_i8', 'a', 3)
 *    ('extract_u8', ('extract_i8', 'a', 'b'), 0) => ('extract_u8', 'a', 'b')
 *    ('extract_u8', ('extract_u8', 'a', 'b'), 0) => ('extract_u8', 'a', 'b')
 *    ('ushr', ('ishl', 'a@32', 16), 16) => ('extract_u16', 'a', 0)
 *    ('ushr', 'a@32', 16) => ('extract_u16', 'a', 1)
 *    ('ishr', ('ishl', 'a@32', 16), 16) => ('extract_i16', 'a', 0)
 *    ('ishr', 'a@32', 16) => ('extract_i16', 'a', 1)
 *    ('iand', 65535, 'a') => ('extract_u16', 'a', 0)
 *    ('ubfe', 'a', 0, 16) => ('extract_u16', 'a', 0)
 *    ('ubfe', 'a', 16, 16) => ('extract_u16', 'a', 1)
 *    ('ibfe', 'a', 0, 16) => ('extract_i16', 'a', 0)
 *    ('ibfe', 'a', 16, 16) => ('extract_i16', 'a', 1)
 *    ('ior', ('ishl', ('u2u32', 'a@8'), 24), ('ior', ('ishl', ('u2u32', 'b@8'), 16), ('ior', ('ishl', ('u2u32', 'c@8'), 8), ('u2u32', 'd@8')))) => ('pack_32_4x8', ('vec4', 'd', 'c', 'b', 'a'))
 *    ('extract_u16', ('extract_i16', 'a', 'b'), 0) => ('extract_u16', 'a', 'b')
 *    ('extract_u16', ('extract_u16', 'a', 'b'), 0) => ('extract_u16', 'a', 'b')
 *    ('pack_64_2x32_split', 'a', 'b') => ('ior', ('u2u64', 'a'), ('ishl', ('u2u64', 'b'), 32))
 *    ('pack_32_2x16_split', 'a', 'b') => ('ior', ('u2u32', 'a'), ('ishl', ('u2u32', 'b'), 16))
 *    ('unpack_64_2x32_split_x', 'a') => ('u2u32', 'a')
 *    ('unpack_64_2x32_split_y', 'a') => ('u2u32', ('ushr', 'a', 32))
 *    ('unpack_32_2x16_split_x', 'a') => ('u2u16', 'a')
 *    ('unpack_32_2x16_split_y', 'a') => ('u2u16', ('ushr', 'a', 16))
 *    ('unpack_half_2x16_split_x', ('iand', 'a', 65535)) => ('unpack_half_2x16_split_x', 'a')
 *    ('unpack_32_2x16_split_x', ('iand', 'a', 65535)) => ('unpack_32_2x16_split_x', 'a')
 *    ('unpack_64_2x32_split_x', ('iand', 'a', 4294967295)) => ('unpack_64_2x32_split_x', 'a')
 *    ('unpack_half_2x16_split_y', ('iand', 'a', 4294901760)) => ('unpack_half_2x16_split_y', 'a')
 *    ('unpack_32_2x16_split_y', ('iand', 'a', 4294901760)) => ('unpack_32_2x16_split_y', 'a')
 *    ('unpack_64_2x32_split_y', ('iand', 'a', 18446744069414584320)) => ('unpack_64_2x32_split_y', 'a')
 *    ('unpack_half_2x16_split_x', ('extract_u16', 'a', 0)) => ('unpack_half_2x16_split_x', 'a')
 *    ('unpack_half_2x16_split_x', ('extract_u16', 'a', 1)) => ('unpack_half_2x16_split_y', 'a')
 *    ('unpack_half_2x16_split_x', ('ushr', 'a', 16)) => ('unpack_half_2x16_split_y', 'a')
 *    ('unpack_32_2x16_split_x', ('extract_u16', 'a', 0)) => ('unpack_32_2x16_split_x', 'a')
 *    ('unpack_32_2x16_split_x', ('extract_u16', 'a', 1)) => ('unpack_32_2x16_split_y', 'a')
 *    ('ishl', ('pack_half_2x16', ('vec2', 'a', 0)), 16) => ('pack_half_2x16', ('vec2', 0, 'a'))
 *    ('ushr', ('pack_half_2x16', ('vec2', 0, 'a')), 16) => ('pack_half_2x16', ('vec2', 'a', 0))
 *    ('iadd', ('pack_half_2x16', ('vec2', 'a', 0)), ('pack_half_2x16', ('vec2', 0, 'b'))) => ('pack_half_2x16', ('vec2', 'a', 'b'))
 *    ('ior', ('pack_half_2x16', ('vec2', 'a', 0)), ('pack_half_2x16', ('vec2', 0, 'b'))) => ('pack_half_2x16', ('vec2', 'a', 'b'))
 *    ('ishl', ('pack_half_2x16_split', 'a', 0), 16) => ('pack_half_2x16_split', 0, 'a')
 *    ('ushr', ('pack_half_2x16_split', 0, 'a'), 16) => ('pack_half_2x16_split', 'a', 0)
 *    ('extract_u16', ('pack_half_2x16_split', 0, 'a'), 1) => ('pack_half_2x16_split', 'a', 0)
 *    ('iadd', ('pack_half_2x16_split', 'a', 0), ('pack_half_2x16_split', 0, 'b')) => ('pack_half_2x16_split', 'a', 'b')
 *    ('ior', ('pack_half_2x16_split', 'a', 0), ('pack_half_2x16_split', 0, 'b')) => ('pack_half_2x16_split', 'a', 'b')
 *    ('extract_i8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 0) => ('i2i', 'a')
 *    ('extract_i8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 1) => ('i2i', 'b')
 *    ('extract_i8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 2) => ('i2i', 'c')
 *    ('extract_i8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 3) => ('i2i', 'd')
 *    ('extract_u8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 0) => ('u2u', 'a')
 *    ('extract_u8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 1) => ('u2u', 'b')
 *    ('extract_u8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 2) => ('u2u', 'c')
 *    ('extract_u8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 3) => ('u2u', 'd')
 *    ('extract_u8', ('ushr', 'a@16', 8), 0) => ('extract_u8', 'a', 1)
 *    ('extract_u8', ('ushr', 'a@32', 8), 0) => ('extract_u8', 'a', 1)
 *    ('extract_u8', ('ushr', 'a@32', 16), 0) => ('extract_u8', 'a', 2)
 *    ('extract_u8', ('ushr', 'a@32', 24), 0) => ('extract_u8', 'a', 3)
 *    ('extract_u8', ('ushr', 'a@64', 8), 0) => ('extract_u8', 'a', 1)
 *    ('extract_u8', ('ushr', 'a@64', 16), 0) => ('extract_u8', 'a', 2)
 *    ('extract_u8', ('ushr', 'a@64', 24), 0) => ('extract_u8', 'a', 3)
 *    ('extract_u8', ('ushr', 'a@64', 32), 0) => ('extract_u8', 'a', 4)
 *    ('extract_u8', ('ushr', 'a@64', 40), 0) => ('extract_u8', 'a', 5)
 *    ('extract_u8', ('ushr', 'a@64', 48), 0) => ('extract_u8', 'a', 6)
 *    ('extract_u8', ('ushr', 'a@64', 56), 0) => ('extract_u8', 'a', 7)
 *    ('extract_u8', ('ishr', 'a@16', 8), 0) => ('extract_u8', 'a', 1)
 *    ('extract_u8', ('ishr', 'a@32', 8), 0) => ('extract_u8', 'a', 1)
 *    ('extract_u8', ('ishr', 'a@32', 16), 0) => ('extract_u8', 'a', 2)
 *    ('extract_u8', ('ishr', 'a@32', 24), 0) => ('extract_u8', 'a', 3)
 *    ('extract_u8', ('ishr', 'a@64', 8), 0) => ('extract_u8', 'a', 1)
 *    ('extract_u8', ('ishr', 'a@64', 16), 0) => ('extract_u8', 'a', 2)
 *    ('extract_u8', ('ishr', 'a@64', 24), 0) => ('extract_u8', 'a', 3)
 *    ('extract_u8', ('ishr', 'a@64', 32), 0) => ('extract_u8', 'a', 4)
 *    ('extract_u8', ('ishr', 'a@64', 40), 0) => ('extract_u8', 'a', 5)
 *    ('extract_u8', ('ishr', 'a@64', 48), 0) => ('extract_u8', 'a', 6)
 *    ('extract_u8', ('ishr', 'a@64', 56), 0) => ('extract_u8', 'a', 7)
 *    ('extract_u8', ('extract_u16', 'a', 1), 0) => ('extract_u8', 'a', 2)
 *    ('extract_u8', ('ishl', 'a@16', 8), 1) => ('extract_u8', 'a', 0)
 *    ('extract_u8', ('ishl', 'a@32', 8), 3) => ('extract_u8', 'a', 2)
 *    ('extract_u8', ('ishl', 'a@32', 16), 3) => ('extract_u8', 'a', 1)
 *    ('extract_u8', ('ishl', 'a@32', 24), 3) => ('extract_u8', 'a', 0)
 *    ('extract_u8', ('ishl', 'a@64', 8), 7) => ('extract_u8', 'a', 6)
 *    ('extract_u8', ('ishl', 'a@64', 16), 7) => ('extract_u8', 'a', 5)
 *    ('extract_u8', ('ishl', 'a@64', 24), 7) => ('extract_u8', 'a', 4)
 *    ('extract_u8', ('ishl', 'a@64', 32), 7) => ('extract_u8', 'a', 3)
 *    ('extract_u8', ('ishl', 'a@64', 40), 7) => ('extract_u8', 'a', 2)
 *    ('extract_u8', ('ishl', 'a@64', 48), 7) => ('extract_u8', 'a', 1)
 *    ('extract_u8', ('ishl', 'a@64', 56), 7) => ('extract_u8', 'a', 0)
 *    ('extract_i8', ('ishl', 'a@16', 8), 1) => ('extract_i8', 'a', 0)
 *    ('extract_i8', ('ishl', 'a@32', 8), 3) => ('extract_i8', 'a', 2)
 *    ('extract_i8', ('ishl', 'a@32', 16), 3) => ('extract_i8', 'a', 1)
 *    ('extract_i8', ('ishl', 'a@32', 24), 3) => ('extract_i8', 'a', 0)
 *    ('extract_i8', ('ishl', 'a@64', 8), 7) => ('extract_i8', 'a', 6)
 *    ('extract_i8', ('ishl', 'a@64', 16), 7) => ('extract_i8', 'a', 5)
 *    ('extract_i8', ('ishl', 'a@64', 24), 7) => ('extract_i8', 'a', 4)
 *    ('extract_i8', ('ishl', 'a@64', 32), 7) => ('extract_i8', 'a', 3)
 *    ('extract_i8', ('ishl', 'a@64', 40), 7) => ('extract_i8', 'a', 2)
 *    ('extract_i8', ('ishl', 'a@64', 48), 7) => ('extract_i8', 'a', 1)
 *    ('extract_i8', ('ishl', 'a@64', 56), 7) => ('extract_i8', 'a', 0)
 *    ('ussub_4x8_vc4', 'a', 0) => a
 *    ('ussub_4x8_vc4', 'a', -1) => 0
 *    ('fsub', 'a', 'b') => ('fadd', 'a', ('fneg', 'b'))
 *    ('isub', 'a', 'b') => ('iadd', 'a', ('ineg', 'b'))
 *    ('uabs_usub', 'a', 'b') => ('bcsel', ('ult', 'a', 'b'), ('ineg', ('isub', 'a', 'b')), ('isub', 'a', 'b'))
 *    ('uabs_isub', 'a', 'b') => ('bcsel', ('ilt', 'a', 'b'), ('ineg', ('isub', 'a', 'b')), ('isub', 'a', 'b'))
 *    ('fmul(is_used_by_non_fsat)', ('fneg', 'a'), 'b') => ('fneg', ('fmul', 'a', 'b'))
 *    ('fmulz(is_used_by_non_fsat)', ('fneg', 'a'), 'b') => ('fneg', ('fmulz', 'a', 'b'))
 *    ('ffma', ('fneg', 'a'), ('fneg', 'b'), 'c') => ('ffma', 'a', 'b', 'c')
 *    ('ffmaz', ('fneg', 'a'), ('fneg', 'b'), 'c') => ('ffmaz', 'a', 'b', 'c')
 *    ('imul', ('ineg', 'a'), 'b') => ('ineg', ('imul', 'a', 'b'))
 *    ('~fmul(is_used_once)', ('fmul(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c') => ('fmul', ('fmul', 'a', 'c'), 'b')
 *    ('~fmulz(is_used_once)', ('fmulz(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c') => ('fmulz', ('fmulz', 'a', 'c'), 'b')
 *    ('~fmul(is_used_once)', ('fmulz(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c(is_finite_not_zero)') => ('fmulz', ('fmul', 'a', 'c'), 'b')
 *    ('imul(is_used_once)', ('imul(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c') => ('imul', ('imul', 'a', 'c'), 'b')
 *    ('~ffma', ('fmul(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c', 'd') => ('ffma', ('fmul', 'a', 'c'), 'b', 'd')
 *    ('~ffmaz', ('fmulz(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c', 'd') => ('ffmaz', ('fmulz', 'a', 'c'), 'b', 'd')
 *    ('~ffma', ('fmulz(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c(is_finite_not_zero)', 'd') => ('ffmaz', ('fmul', 'a', 'c'), 'b', 'd')
 *    ('~fadd(is_used_once)', ('fadd(is_used_once)', 'a(is_not_const)', 'b(is_fmul)'), '#c') => ('fadd', ('fadd', 'a', 'c'), 'b')
 *    ('~fadd(is_used_once)', ('fadd(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c') => ('fadd', ('fadd', 'a', 'c'), 'b')
 *    ('~fadd(is_used_once)', ('ffma(is_used_once)', 'a(is_not_const)', 'b', 'c(is_not_const)'), '#d') => ('fadd', ('ffma', 'a', 'b', 'd'), 'c')
 *    ('~fadd(is_used_once)', ('ffmaz(is_used_once)', 'a(is_not_const)', 'b', 'c(is_not_const)'), '#d') => ('fadd', ('ffmaz', 'a', 'b', 'd'), 'c')
 *    ('iadd(is_used_once)', ('iadd(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c') => ('iadd', ('iadd', 'a', 'c'), 'b')
 *    ('~fmul', '#a', ('fmul', 'b(is_not_const)', '#c')) => ('fmul', ('fmul', 'a', 'c'), 'b')
 *    ('~fmulz', '#a', ('fmulz', 'b(is_not_const)', '#c')) => ('fmulz', ('fmulz', 'a', 'c'), 'b')
 *    ('~fmul', '#a(is_finite_not_zero)', ('fmulz', 'b(is_not_const)', '#c')) => ('fmulz', ('fmul', 'a', 'c'), 'b')
 *    ('~ffma', '#a', ('fmul', 'b(is_not_const)', '#c'), 'd') => ('ffma', ('fmul', 'a', 'c'), 'b', 'd')
 *    ('~ffmaz', '#a', ('fmulz', 'b(is_not_const)', '#c'), 'd') => ('ffmaz', ('fmulz', 'a', 'c'), 'b', 'd')
 *    ('~ffmaz', '#a(is_finite_not_zero)', ('fmulz', 'b(is_not_const)', '#c'), 'd') => ('ffmaz', ('fmul', 'a', 'c'), 'b', 'd')
 *    ('imul', '#a', ('imul', 'b(is_not_const)', '#c')) => ('imul', ('imul', 'a', 'c'), 'b')
 *    ('~fadd', '#a', ('fadd', 'b(is_not_const)', '#c')) => ('fadd', ('fadd', 'a', 'c'), 'b')
 *    ('~fadd', '#a', ('fneg', ('fadd', 'b(is_not_const)', '#c'))) => ('fadd', ('fadd', 'a', ('fneg', 'c')), ('fneg', 'b'))
 *    ('~fadd', '#a', ('ffma', 'b(is_not_const)', 'c(is_not_const)', '#d')) => ('ffma', 'b', 'c', ('fadd', 'a', 'd'))
 *    ('~fadd', '#a', ('fneg', ('ffma', 'b(is_not_const)', 'c(is_not_const)', '#d'))) => ('ffma', ('fneg', 'b'), 'c', ('fadd', 'a', ('fneg', 'd')))
 *    ('~fadd', '#a', ('ffmaz', 'b(is_not_const)', 'c(is_not_const)', '#d')) => ('ffmaz', 'b', 'c', ('fadd', 'a', 'd'))
 *    ('~fadd', '#a', ('fneg', ('ffmaz', 'b(is_not_const)', 'c(is_not_const)', '#d'))) => ('ffmaz', ('fneg', 'b'), 'c', ('fadd', 'a', ('fneg', 'd')))
 *    ('iadd', '#a', ('iadd', 'b(is_not_const)', '#c')) => ('iadd', ('iadd', 'a', 'c'), 'b')
 *    ('iand', '#a', ('iand', 'b(is_not_const)', '#c')) => ('iand', ('iand', 'a', 'c'), 'b')
 *    ('ior', '#a', ('ior', 'b(is_not_const)', '#c')) => ('ior', ('ior', 'a', 'c'), 'b')
 *    ('ixor', '#a', ('ixor', 'b(is_not_const)', '#c')) => ('ixor', ('ixor', 'a', 'c'), 'b')
 *    ('~fadd', ('fadd(is_used_once)', 'a(is_fmul)', 'b(is_fmul)'), 'c(is_not_fmul)') => ('fadd', ('fadd', 'a', 'c'), 'b')
 *    ('idiv', ('imul(no_signed_wrap)', 'a', 'b'), 'b') => a
 *    ('bcsel', ('ige', ('find_lsb', 'a'), 0), ('find_lsb', 'a'), -1) => ('find_lsb', 'a')
 *    ('bcsel', ('ige', ('ifind_msb', 'a'), 0), ('ifind_msb', 'a'), -1) => ('ifind_msb', 'a')
 *    ('bcsel', ('ige', ('ufind_msb', 'a'), 0), ('ufind_msb', 'a'), -1) => ('ufind_msb', 'a')
 *    ('bcsel', ('ige', ('ifind_msb_rev', 'a'), 0), ('ifind_msb_rev', 'a'), -1) => ('ifind_msb_rev', 'a')
 *    ('bcsel', ('ige', ('ufind_msb_rev', 'a'), 0), ('ufind_msb_rev', 'a'), -1) => ('ufind_msb_rev', 'a')
 *    ('bcsel', ('ine', 'a', 0), ('find_lsb', 'a'), -1) => ('find_lsb', 'a')
 *    ('bcsel', ('ine', 'a', 0), ('ifind_msb', 'a'), -1) => ('ifind_msb', 'a')
 *    ('bcsel', ('ine', 'a', 0), ('ufind_msb', 'a'), -1) => ('ufind_msb', 'a')
 *    ('bcsel', ('ine', 'a', 0), ('ifind_msb_rev', 'a'), -1) => ('ifind_msb_rev', 'a')
 *    ('bcsel', ('ine', 'a', 0), ('ufind_msb_rev', 'a'), -1) => ('ufind_msb_rev', 'a')
 *    ('bcsel', ('ine', 'a', -1), ('ifind_msb', 'a'), -1) => ('ifind_msb', 'a')
 *    ('bcsel', ('ine', 'a', -1), ('ifind_msb_rev', 'a'), -1) => ('ifind_msb_rev', 'a')
 *    ('bcsel', ('ine', ('ifind_msb', 'a@32'), -1), ('iadd', 31, ('ineg', ('ifind_msb', 'a'))), -1) => ('ifind_msb_rev', 'a')
 *    ('bcsel', ('ine', ('ufind_msb', 'a@32'), -1), ('iadd', 31, ('ineg', ('ufind_msb', 'a'))), -1) => ('ufind_msb_rev', 'a')
 *    ('bcsel', ('ieq', ('ifind_msb', 'a@32'), -1), -1, ('iadd', 31, ('ineg', ('ifind_msb', 'a')))) => ('ifind_msb_rev', 'a')
 *    ('bcsel', ('ieq', ('ufind_msb', 'a@32'), -1), -1, ('iadd', 31, ('ineg', ('ufind_msb', 'a')))) => ('ufind_msb_rev', 'a')
 *    ('bcsel', ('ine', ('ifind_msb', 'a@32'), -1), ('iadd', 31, ('ineg', ('ifind_msb', 'a'))), ('ifind_msb', 'a')) => ('ifind_msb_rev', 'a')
 *    ('bcsel', ('ine', ('ufind_msb', 'a@32'), -1), ('iadd', 31, ('ineg', ('ufind_msb', 'a'))), ('ufind_msb', 'a')) => ('ufind_msb_rev', 'a')
 *    ('bcsel', ('ieq', ('ifind_msb', 'a@32'), -1), ('ifind_msb', 'a'), ('iadd', 31, ('ineg', ('ifind_msb', 'a')))) => ('ifind_msb_rev', 'a')
 *    ('bcsel', ('ieq', ('ufind_msb', 'a@32'), -1), ('ufind_msb', 'a'), ('iadd', 31, ('ineg', ('ufind_msb', 'a')))) => ('ufind_msb_rev', 'a')
 *    ('bcsel', ('ine', 'a@32', 0), ('iadd', 31, ('ineg', ('ufind_msb', 'a'))), -1) => ('ufind_msb_rev', 'a')
 *    ('bcsel', ('ieq', 'a@32', 0), -1, ('iadd', 31, ('ineg', ('ufind_msb', 'a')))) => ('ufind_msb_rev', 'a')
 *    ('bcsel', ('ine', 'a@32', 0), ('iadd', 31, ('ineg', ('ufind_msb', 'a'))), ('ufind_msb', 'a')) => ('ufind_msb_rev', 'a')
 *    ('bcsel', ('ieq', 'a@32', 0), ('ufind_msb', 'a'), ('iadd', 31, ('ineg', ('ufind_msb', 'a')))) => ('ufind_msb_rev', 'a')
 *    ('bcsel', ('ine', ('ifind_msb_rev', 'a@32'), -1), ('iadd', 31, ('ineg', ('ifind_msb_rev', 'a'))), -1) => ('ifind_msb', 'a')
 *    ('bcsel', ('ine', ('ufind_msb_rev', 'a@32'), -1), ('iadd', 31, ('ineg', ('ufind_msb_rev', 'a'))), -1) => ('ufind_msb', 'a')
 *    ('bcsel', ('ieq', ('ifind_msb_rev', 'a@32'), -1), -1, ('iadd', 31, ('ineg', ('ifind_msb_rev', 'a')))) => ('ifind_msb', 'a')
 *    ('bcsel', ('ieq', ('ufind_msb_rev', 'a@32'), -1), -1, ('iadd', 31, ('ineg', ('ufind_msb_rev', 'a')))) => ('ufind_msb', 'a')
 *    ('bcsel', ('ine', ('ifind_msb_rev', 'a@32'), -1), ('iadd', 31, ('ineg', ('ifind_msb_rev', 'a'))), ('ifind_msb_rev', 'a')) => ('ifind_msb', 'a')
 *    ('bcsel', ('ine', ('ufind_msb_rev', 'a@32'), -1), ('iadd', 31, ('ineg', ('ufind_msb_rev', 'a'))), ('ufind_msb_rev', 'a')) => ('ufind_msb', 'a')
 *    ('bcsel', ('ieq', ('ifind_msb_rev', 'a@32'), -1), ('ifind_msb_rev', 'a'), ('iadd', 31, ('ineg', ('ifind_msb_rev', 'a')))) => ('ifind_msb', 'a')
 *    ('bcsel', ('ieq', ('ufind_msb_rev', 'a@32'), -1), ('ufind_msb_rev', 'a'), ('iadd', 31, ('ineg', ('ufind_msb_rev', 'a')))) => ('ufind_msb', 'a')
 *    ('bcsel', ('ine', 'a@32', 0), ('iadd', 31, ('ineg', ('ufind_msb_rev', 'a'))), -1) => ('ufind_msb', 'a')
 *    ('bcsel', ('ieq', 'a@32', 0), -1, ('iadd', 31, ('ineg', ('ufind_msb_rev', 'a')))) => ('ufind_msb', 'a')
 *    ('bcsel', ('ine', 'a@32', 0), ('iadd', 31, ('ineg', ('ufind_msb_rev', 'a'))), ('ufind_msb_rev', 'a')) => ('ufind_msb', 'a')
 *    ('bcsel', ('ieq', 'a@32', 0), ('ufind_msb_rev', 'a'), ('iadd', 31, ('ineg', ('ufind_msb_rev', 'a')))) => ('ufind_msb', 'a')
 *    ('find_lsb', ('bitfield_reverse', 'a')) => ('ufind_msb_rev', 'a')
 *    ('ufind_msb_rev', ('bitfield_reverse', 'a')) => ('find_lsb', 'a')
 *    ('~fmul', ('bcsel(is_used_once)', 'c', -1.0, 1.0), 'b') => ('bcsel', 'c', ('fneg', 'b'), 'b')
 *    ('~fmul', ('bcsel(is_used_once)', 'c', 1.0, -1.0), 'b') => ('bcsel', 'c', 'b', ('fneg', 'b'))
 *    ('~fmulz', ('bcsel(is_used_once)', 'c', -1.0, 1.0), 'b') => ('bcsel', 'c', ('fneg', 'b'), 'b')
 *    ('~fmulz', ('bcsel(is_used_once)', 'c', 1.0, -1.0), 'b') => ('bcsel', 'c', 'b', ('fneg', 'b'))
 *    ('fabs', ('bcsel(is_used_once)', 'b', ('fneg', 'a'), 'a')) => ('fabs', 'a')
 *    ('fabs', ('bcsel(is_used_once)', 'b', 'a', ('fneg', 'a'))) => ('fabs', 'a')
 *    ('~bcsel', ('flt', 'a', 0.0), ('fneg', 'a'), 'a') => ('fabs', 'a')
 *    ('bcsel', 'a', ('bcsel', 'b', 'c', 'd'), 'd') => ('bcsel', ('iand', 'a', 'b'), 'c', 'd')
 *    ('bcsel', 'a', 'b', ('bcsel', 'c', 'b', 'd')) => ('bcsel', ('ior', 'a', 'c'), 'b', 'd')
 *    ('fmod', 'a', 'b') => ('fsub', 'a', ('fmul', 'b', ('ffloor', ('fdiv', 'a', 'b'))))
 *    ('frem', 'a', 'b') => ('fsub', 'a', ('fmul', 'b', ('ftrunc', ('fdiv', 'a', 'b'))))
 *    ('uadd_carry', 'a', 'b') => ('b2i', ('ult', ('iadd', 'a', 'b'), 'a'))
 *    ('usub_borrow', 'a', 'b') => ('b2i', ('ult', 'a', 'b'))
 *    ('bitfield_insert', 'base', 'insert', 'offset', 'bits') => ('bcsel', ('ult', 31, 'bits'), 'insert', ('bfi', ('bfm', 'bits', 'offset'), 'insert', 'base'))
 *    ('ihadd', 'a', 'b') => ('iadd', ('iand', 'a', 'b'), ('ishr', ('ixor', 'a', 'b'), 1))
 *    ('uhadd', 'a', 'b') => ('iadd', ('iand', 'a', 'b'), ('ushr', ('ixor', 'a', 'b'), 1))
 *    ('irhadd', 'a', 'b') => ('isub', ('ior', 'a', 'b'), ('ishr', ('ixor', 'a', 'b'), 1))
 *    ('urhadd', 'a', 'b') => ('isub', ('ior', 'a', 'b'), ('ushr', ('ixor', 'a', 'b'), 1))
 *    ('ihadd@64', 'a', 'b') => ('iadd', ('iand', 'a', 'b'), ('ishr', ('ixor', 'a', 'b'), 1))
 *    ('uhadd@64', 'a', 'b') => ('iadd', ('iand', 'a', 'b'), ('ushr', ('ixor', 'a', 'b'), 1))
 *    ('irhadd@64', 'a', 'b') => ('isub', ('ior', 'a', 'b'), ('ishr', ('ixor', 'a', 'b'), 1))
 *    ('urhadd@64', 'a', 'b') => ('isub', ('ior', 'a', 'b'), ('ushr', ('ixor', 'a', 'b'), 1))
 *    ('imul_32x16', 'a', 'b') => ('imul', 'a', ('extract_i16', 'b', 0))
 *    ('umul_32x16', 'a', 'b') => ('imul', 'a', ('extract_u16', 'b', 0))
 *    ('uadd_sat@64', 'a', 'b') => ('bcsel', ('ult', ('iadd', 'a', 'b'), 'a'), -1, ('iadd', 'a', 'b'))
 *    ('uadd_sat', 'a', 'b') => ('bcsel', ('ult', ('iadd', 'a', 'b'), 'a'), -1, ('iadd', 'a', 'b'))
 *    ('usub_sat', 'a', 'b') => ('bcsel', ('ult', 'a', 'b'), 0, ('isub', 'a', 'b'))
 *    ('usub_sat@64', 'a', 'b') => ('bcsel', ('ult', 'a', 'b'), 0, ('isub', 'a', 'b'))
 *    ('iadd_sat@64', 'a', 'b') => ('bcsel', ('iand', ('iand', ('ilt', 'a', 0), ('ilt', 'b', 0)), ('ige', ('iadd', 'a', 'b'), 0)), 9223372036854775808, ('bcsel', ('ior', ('ior', ('ilt', 'a', 0), ('ilt', 'b', 0)), ('ige', ('iadd', 'a', 'b'), 0)), ('iadd', 'a', 'b'), 9223372036854775807))
 *    ('isub_sat@64', 'a', 'b') => ('bcsel', ('iand', ('iand', ('ilt', 'a', 0), ('ige', 'b', 0)), ('ige', ('isub', 'a', 'b'), 0)), 9223372036854775808, ('bcsel', ('ior', ('ior', ('ilt', 'a', 0), ('ige', 'b', 0)), ('ige', ('isub', 'a', 'b'), 0)), ('isub', 'a', 'b'), 9223372036854775807))
 *    ('ilt', ('imax(is_used_once)', 'a@64', 'b@64'), 0) => ('ilt', ('imax', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b')), 0)
 *    ('ilt', ('imin(is_used_once)', 'a@64', 'b@64'), 0) => ('ilt', ('imin', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b')), 0)
 *    ('ige', ('imax(is_used_once)', 'a@64', 'b@64'), 0) => ('ige', ('imax', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b')), 0)
 *    ('ige', ('imin(is_used_once)', 'a@64', 'b@64'), 0) => ('ige', ('imin', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b')), 0)
 *    ('ilt', 'a@64', 0) => ('ilt', ('unpack_64_2x32_split_y', 'a'), 0)
 *    ('ige', 'a@64', 0) => ('ige', ('unpack_64_2x32_split_y', 'a'), 0)
 *    ('ine', 'a@64', 0) => ('ine', ('ior', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_y', 'a')), 0)
 *    ('ieq', 'a@64', 0) => ('ieq', ('ior', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_y', 'a')), 0)
 *    ('ult', 0, 'a@64') => ('ine', ('ior', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_y', 'a')), 0)
 *    ('bitfield_insert', 'base', 'insert', 'offset', 'bits') => ('bcsel', ('ult', 31, 'bits'), 'insert', ('ior', ('iand', 'base', ('inot', ('ishl', ('isub', ('ishl', 1, 'bits'), 1), 'offset'))), ('iand', ('ishl', 'insert', 'offset'), ('ishl', ('isub', ('ishl', 1, 'bits'), 1), 'offset'))))
 *    ('bitfield_insert', 'base', 'insert', 'offset', 'bits') => ('bcsel', ('ult', 31, 'bits'), 'insert', ('bitfield_select', ('bfm', 'bits', 'offset'), ('ishl', 'insert', 'offset'), 'base'))
 *    ('ibitfield_extract', 'value', 'offset', 'bits') => ('bcsel', ('ult', 31, 'bits'), 'value', ('ibfe', 'value', 'offset', 'bits'))
 *    ('ubitfield_extract', 'value', 'offset', 'bits') => ('bcsel', ('ult', 31, 'bits'), 'value', ('ubfe', 'value', 'offset', 'bits'))
 *    ('bitfield_select', 'a', 'b', 0) => ('iand', 'a', 'b')
 *    ('bitfield_select', 'a', ('iand', 'a', 'b'), 'c') => ('bitfield_select', 'a', 'b', 'c')
 *    ('ubfe', 'value', 'offset', ('iand', 31, 'bits')) => ('ubfe', 'value', 'offset', 'bits')
 *    ('ubfe', 'value', ('iand', 31, 'offset'), 'bits') => ('ubfe', 'value', 'offset', 'bits')
 *    ('ibfe', 'value', 'offset', ('iand', 31, 'bits')) => ('ibfe', 'value', 'offset', 'bits')
 *    ('ibfe', 'value', ('iand', 31, 'offset'), 'bits') => ('ibfe', 'value', 'offset', 'bits')
 *    ('bfm', 'bits', ('iand', 31, 'offset')) => ('bfm', 'bits', 'offset')
 *    ('bfm', ('iand', 31, 'bits'), 'offset') => ('bfm', 'bits', 'offset')
 *    ('ult', 'a', ('umin', ('iand', 'a', 'b'), 'c')) => False
 *    ('ult', 31, ('umin', '#bits(is_ult_32)', 'a')) => False
 *    ('ubfe', 'value', 'offset', ('umin', 'width', ('iadd', 32, ('ineg', ('iand', 31, 'offset'))))) => ('ubfe', 'value', 'offset', 'width')
 *    ('ibfe', 'value', 'offset', ('umin', 'width', ('iadd', 32, ('ineg', ('iand', 31, 'offset'))))) => ('ibfe', 'value', 'offset', 'width')
 *    ('bfm', ('umin', 'width', ('iadd', 32, ('ineg', ('iand', 31, 'offset')))), 'offset') => ('bfm', 'width', 'offset')
 *    ('iadd@32', ('ishl', 1, 'a'), -1) => ('bfm', 'a', 0)
 *    ('ishl', ('bfm', 'a', 0), 'b') => ('bfm', 'a', 'b')
 *    ('ubfe', 'a', 'b', 0) => 0
 *    ('ibfe', 'a', 'b', 0) => 0
 *    ('ubfe', 'a', 0, '#b') => ('iand', 'a', ('ushr', 4294967295, ('ineg', 'b')))
 *    ('b2i32', ('ine', ('ubfe', 'a', 'b', 1), 0)) => ('ubfe', 'a', 'b', 1)
 *    ('b2i32', ('ine', ('ibfe', 'a', 'b', 1), 0)) => ('ubfe', 'a', 'b', 1)
 *    ('ine', ('ibfe(is_used_once)', 'a', '#b', '#c'), 0) => ('ine', ('iand', 'a', ('ishl', ('ushr', 4294967295, ('ineg', 'c')), 'b')), 0)
 *    ('ieq', ('ibfe(is_used_once)', 'a', '#b', '#c'), 0) => ('ieq', ('iand', 'a', ('ishl', ('ushr', 4294967295, ('ineg', 'c')), 'b')), 0)
 *    ('ine', ('ubfe(is_used_once)', 'a', '#b', '#c'), 0) => ('ine', ('iand', 'a', ('ishl', ('ushr', 4294967295, ('ineg', 'c')), 'b')), 0)
 *    ('ieq', ('ubfe(is_used_once)', 'a', '#b', '#c'), 0) => ('ieq', ('iand', 'a', ('ishl', ('ushr', 4294967295, ('ineg', 'c')), 'b')), 0)
 *    ('ibitfield_extract', 'value', 'offset', 'bits') => ('bcsel', ('ieq', 0, 'bits'), 0, ('ishr', ('ishl', 'value', ('isub', ('isub', 32, 'bits'), 'offset')), ('isub', 32, 'bits')))
 *    ('ubitfield_extract', 'value', 'offset', 'bits') => ('iand', ('ushr', 'value', 'offset'), ('bcsel', ('ieq', 'bits', 32), 4294967295, ('isub', ('ishl', 1, 'bits'), 1)))
 *    ('ifind_msb', 'value') => ('ufind_msb', ('bcsel', ('ilt', 'value', 0), ('inot', 'value'), 'value'))
 *    ('ifind_msb', 'value') => ('bcsel', ('ige', ('ifind_msb_rev', 'value'), 0), ('isub', 31, ('ifind_msb_rev', 'value')), ('ifind_msb_rev', 'value'))
 *    ('ufind_msb', 'value') => ('bcsel', ('ige', ('ufind_msb_rev', 'value'), 0), ('isub', 31, ('ufind_msb_rev', 'value')), ('ufind_msb_rev', 'value'))
 *    ('uclz', 'a') => ('umin', 32, ('ufind_msb_rev', 'a'))
 *    ('find_lsb', 'value') => ('ufind_msb', ('iand', 'value', ('ineg', 'value')))
 *    ('extract_i8', 'a', 'b@32') => ('ishr', ('ishl', 'a', ('imul', ('isub', 3, 'b'), 8)), 24)
 *    ('extract_u8', 'a', 'b@32') => ('iand', ('ushr', 'a', ('imul', 'b', 8)), 255)
 *    ('extract_i16', 'a', 'b@32') => ('ishr', ('ishl', 'a', ('imul', ('isub', 1, 'b'), 16)), 16)
 *    ('extract_u16', 'a', 'b@32') => ('iand', ('ushr', 'a', ('imul', 'b', 16)), 65535)
 *    ('pack_unorm_2x16', 'v') => ('pack_uvec2_to_uint', ('f2u32', ('fround_even', ('fmul', ('fsat', 'v'), 65535.0))))
 *    ('pack_unorm_4x8', 'v') => ('pack_uvec4_to_uint', ('f2u32', ('fround_even', ('fmul', ('fsat', 'v'), 255.0))))
 *    ('pack_snorm_2x16', 'v') => ('pack_uvec2_to_uint', ('f2i32', ('fround_even', ('fmul', ('fmin', 1.0, ('fmax', -1.0, 'v')), 32767.0))))
 *    ('pack_snorm_4x8', 'v') => ('pack_uvec4_to_uint', ('f2i32', ('fround_even', ('fmul', ('fmin', 1.0, ('fmax', -1.0, 'v')), 127.0))))
 *    ('unpack_unorm_2x16', 'v') => ('fdiv', ('u2f32', ('vec2', ('extract_u16', 'v', 0), ('extract_u16', 'v', 1))), 65535.0)
 *    ('unpack_unorm_4x8', 'v') => ('fdiv', ('u2f32', ('vec4', ('extract_u8', 'v', 0), ('extract_u8', 'v', 1), ('extract_u8', 'v', 2), ('extract_u8', 'v', 3))), 255.0)
 *    ('unpack_snorm_2x16', 'v') => ('fmin', 1.0, ('fmax', -1.0, ('fdiv', ('i2f', ('vec2', ('extract_i16', 'v', 0), ('extract_i16', 'v', 1))), 32767.0)))
 *    ('unpack_snorm_4x8', 'v') => ('fmin', 1.0, ('fmax', -1.0, ('fdiv', ('i2f', ('vec4', ('extract_i8', 'v', 0), ('extract_i8', 'v', 1), ('extract_i8', 'v', 2), ('extract_i8', 'v', 3))), 127.0)))
 *    ('pack_half_2x16_split', 'a@32', 'b@32') => ('ior', ('ishl', ('u2u32', ('f2f16', 'b')), 16), ('u2u32', ('f2f16', 'a')))
 *    ('unpack_half_2x16_split_x', 'a@32') => ('f2f32', ('u2u16', 'a'))
 *    ('unpack_half_2x16_split_y', 'a@32') => ('f2f32', ('u2u16', ('ushr', 'a', 16)))
 *    ('pack_32_2x16_split', 'a@16', 'b@16') => ('ior', ('ishl', ('u2u32', 'b'), 16), ('u2u32', 'a'))
 *    ('unpack_32_2x16_split_x', 'a@32') => ('u2u16', 'a')
 *    ('unpack_32_2x16_split_y', 'a@32') => ('u2u16', ('ushr', 'a', 16))
 *    ('isign', 'a') => ('imin', ('imax', 'a', -1), 1)
 *    ('imin', ('imax', 'a', -1), 1) => ('isign', 'a')
 *    ('imax', ('imin', 'a', 1), -1) => ('isign', 'a')
 *    ('fsign', 'a') => ('fsub', ('b2f', ('!flt', 0.0, 'a')), ('b2f', ('!flt', 'a', 0.0)))
 *    ('amul', 'a', 'b') => ('imul', 'a', 'b')
 *    ('umul24', 'a', 'b') => ('imul', ('iand', 'a', 16777215), ('iand', 'b', 16777215))
 *    ('umad24', 'a', 'b', 'c') => ('iadd', ('imul', ('iand', 'a', 16777215), ('iand', 'b', 16777215)), 'c')
 *    ('imul24_relaxed', 'a', 'b') => ('imul24', 'a', 'b')
 *    ('imul24_relaxed', 'a', 'b') => ('imul', 'a', 'b')
 *    ('umad24_relaxed', 'a', 'b', 'c') => ('umad24', 'a', 'b', 'c')
 *    ('umad24_relaxed', 'a', 'b', 'c') => ('iadd', ('umul24_relaxed', 'a', 'b'), 'c')
 *    ('umul24_relaxed', 'a', 'b') => ('umul24', 'a', 'b')
 *    ('umul24_relaxed', 'a', 'b') => ('imul', 'a', 'b')
 *    ('imad24_ir3', 'a', 'b', 0) => ('imul24', 'a', 'b')
 *    ('imad24_ir3', 'a', 0, 'c') => c
 *    ('imad24_ir3', 'a', 1, 'c') => ('iadd', 'a', 'c')
 *    ('imad24_ir3', '#a', '#b', 'c') => ('iadd', ('imul', 'a', 'b'), 'c')
 *    ('imul24', 'a', '#b@32(is_pos_power_of_two)') => ('ishl', 'a', ('find_lsb', 'b'))
 *    ('imul24', 'a', '#b@32(is_neg_power_of_two)') => ('ineg', ('ishl', 'a', ('find_lsb', ('iabs', 'b'))))
 *    ('imul24', 'a', 0) => 0
 *    ('bcsel', ('ult', ('iadd', 'a', 'b'), 'a'), -1, ('iadd@8', 'a', 'b')) => ('uadd_sat', 'a', 'b')
 *    ('bcsel', ('uge', ('iadd', 'a', 'b'), 'a'), ('iadd@8', 'a', 'b'), -1) => ('uadd_sat', 'a', 'b')
 *    ('bcsel', ('ieq', ('uadd_carry', 'a', 'b'), 0), ('iadd@8', 'a', 'b'), -1) => ('uadd_sat', 'a', 'b')
 *    ('bcsel', ('ine', ('uadd_carry', 'a', 'b'), 0), -1, ('iadd@8', 'a', 'b')) => ('uadd_sat', 'a', 'b')
 *    ('bcsel', ('ult', ('iadd', 'a', 'b'), 'a'), -1, ('iadd@16', 'a', 'b')) => ('uadd_sat', 'a', 'b')
 *    ('bcsel', ('uge', ('iadd', 'a', 'b'), 'a'), ('iadd@16', 'a', 'b'), -1) => ('uadd_sat', 'a', 'b')
 *    ('bcsel', ('ieq', ('uadd_carry', 'a', 'b'), 0), ('iadd@16', 'a', 'b'), -1) => ('uadd_sat', 'a', 'b')
 *    ('bcsel', ('ine', ('uadd_carry', 'a', 'b'), 0), -1, ('iadd@16', 'a', 'b')) => ('uadd_sat', 'a', 'b')
 *    ('bcsel', ('ult', ('iadd', 'a', 'b'), 'a'), -1, ('iadd@32', 'a', 'b')) => ('uadd_sat', 'a', 'b')
 *    ('bcsel', ('uge', ('iadd', 'a', 'b'), 'a'), ('iadd@32', 'a', 'b'), -1) => ('uadd_sat', 'a', 'b')
 *    ('bcsel', ('ieq', ('uadd_carry', 'a', 'b'), 0), ('iadd@32', 'a', 'b'), -1) => ('uadd_sat', 'a', 'b')
 *    ('bcsel', ('ine', ('uadd_carry', 'a', 'b'), 0), -1, ('iadd@32', 'a', 'b')) => ('uadd_sat', 'a', 'b')
 *    ('bcsel', ('ult', ('iadd', 'a', 'b'), 'a'), -1, ('iadd@64', 'a', 'b')) => ('uadd_sat', 'a', 'b')
 *    ('bcsel', ('uge', ('iadd', 'a', 'b'), 'a'), ('iadd@64', 'a', 'b'), -1) => ('uadd_sat', 'a', 'b')
 *    ('bcsel', ('ieq', ('uadd_carry', 'a', 'b'), 0), ('iadd@64', 'a', 'b'), -1) => ('uadd_sat', 'a', 'b')
 *    ('bcsel', ('ine', ('uadd_carry', 'a', 'b'), 0), -1, ('iadd@64', 'a', 'b')) => ('uadd_sat', 'a', 'b')
 *    ('bcsel', ('ult', 'a', 'b'), 0, ('iadd@8', 'a', ('ineg', 'b'))) => ('usub_sat', 'a', 'b')
 *    ('bcsel', ('uge', 'a', 'b'), ('iadd@8', 'a', ('ineg', 'b')), 0) => ('usub_sat', 'a', 'b')
 *    ('bcsel', ('ieq', ('usub_borrow', 'a', 'b'), 0), ('iadd@8', 'a', ('ineg', 'b')), 0) => ('usub_sat', 'a', 'b')
 *    ('bcsel', ('ine', ('usub_borrow', 'a', 'b'), 0), 0, ('iadd@8', 'a', ('ineg', 'b'))) => ('usub_sat', 'a', 'b')
 *    ('bcsel', ('ult', 'a', 'b'), 0, ('iadd@16', 'a', ('ineg', 'b'))) => ('usub_sat', 'a', 'b')
 *    ('bcsel', ('uge', 'a', 'b'), ('iadd@16', 'a', ('ineg', 'b')), 0) => ('usub_sat', 'a', 'b')
 *    ('bcsel', ('ieq', ('usub_borrow', 'a', 'b'), 0), ('iadd@16', 'a', ('ineg', 'b')), 0) => ('usub_sat', 'a', 'b')
 *    ('bcsel', ('ine', ('usub_borrow', 'a', 'b'), 0), 0, ('iadd@16', 'a', ('ineg', 'b'))) => ('usub_sat', 'a', 'b')
 *    ('bcsel', ('ult', 'a', 'b'), 0, ('iadd@32', 'a', ('ineg', 'b'))) => ('usub_sat', 'a', 'b')
 *    ('bcsel', ('uge', 'a', 'b'), ('iadd@32', 'a', ('ineg', 'b')), 0) => ('usub_sat', 'a', 'b')
 *    ('bcsel', ('ieq', ('usub_borrow', 'a', 'b'), 0), ('iadd@32', 'a', ('ineg', 'b')), 0) => ('usub_sat', 'a', 'b')
 *    ('bcsel', ('ine', ('usub_borrow', 'a', 'b'), 0), 0, ('iadd@32', 'a', ('ineg', 'b'))) => ('usub_sat', 'a', 'b')
 *    ('bcsel', ('ult', 'a', 'b'), 0, ('iadd@64', 'a', ('ineg', 'b'))) => ('usub_sat', 'a', 'b')
 *    ('bcsel', ('uge', 'a', 'b'), ('iadd@64', 'a', ('ineg', 'b')), 0) => ('usub_sat', 'a', 'b')
 *    ('bcsel', ('ieq', ('usub_borrow', 'a', 'b'), 0), ('iadd@64', 'a', ('ineg', 'b')), 0) => ('usub_sat', 'a', 'b')
 *    ('bcsel', ('ine', ('usub_borrow', 'a', 'b'), 0), 0, ('iadd@64', 'a', ('ineg', 'b'))) => ('usub_sat', 'a', 'b')
 *    ('iadd_sat@8', 'a', 'b') => ('bcsel', ('ige', 'b', 1), ('bcsel', ('ilt', ('iadd', 'a', 'b'), 'a'), 127, ('iadd', 'a', 'b')), ('bcsel', ('ilt', 'a', ('iadd', 'a', 'b')), 128, ('iadd', 'a', 'b')))
 *    ('isub_sat@8', 'a', 'b') => ('bcsel', ('ilt', 'b', 0), ('bcsel', ('ilt', ('isub', 'a', 'b'), 'a'), 127, ('isub', 'a', 'b')), ('bcsel', ('ilt', 'a', ('isub', 'a', 'b')), 128, ('isub', 'a', 'b')))
 *    ('iadd_sat@16', 'a', 'b') => ('bcsel', ('ige', 'b', 1), ('bcsel', ('ilt', ('iadd', 'a', 'b'), 'a'), 32767, ('iadd', 'a', 'b')), ('bcsel', ('ilt', 'a', ('iadd', 'a', 'b')), 32768, ('iadd', 'a', 'b')))
 *    ('isub_sat@16', 'a', 'b') => ('bcsel', ('ilt', 'b', 0), ('bcsel', ('ilt', ('isub', 'a', 'b'), 'a'), 32767, ('isub', 'a', 'b')), ('bcsel', ('ilt', 'a', ('isub', 'a', 'b')), 32768, ('isub', 'a', 'b')))
 *    ('iadd_sat@32', 'a', 'b') => ('bcsel', ('ige', 'b', 1), ('bcsel', ('ilt', ('iadd', 'a', 'b'), 'a'), 2147483647, ('iadd', 'a', 'b')), ('bcsel', ('ilt', 'a', ('iadd', 'a', 'b')), 2147483648, ('iadd', 'a', 'b')))
 *    ('isub_sat@32', 'a', 'b') => ('bcsel', ('ilt', 'b', 0), ('bcsel', ('ilt', ('isub', 'a', 'b'), 'a'), 2147483647, ('isub', 'a', 'b')), ('bcsel', ('ilt', 'a', ('isub', 'a', 'b')), 2147483648, ('isub', 'a', 'b')))
 *    ('iadd_sat@64', 'a', 'b') => ('bcsel', ('ige', 'b', 1), ('bcsel', ('ilt', ('iadd', 'a', 'b'), 'a'), 9223372036854775807, ('iadd', 'a', 'b')), ('bcsel', ('ilt', 'a', ('iadd', 'a', 'b')), 9223372036854775808, ('iadd', 'a', 'b')))
 *    ('isub_sat@64', 'a', 'b') => ('bcsel', ('ilt', 'b', 0), ('bcsel', ('ilt', ('isub', 'a', 'b'), 'a'), 9223372036854775807, ('isub', 'a', 'b')), ('bcsel', ('ilt', 'a', ('isub', 'a', 'b')), 9223372036854775808, ('isub', 'a', 'b')))
 *    ('inot', ('ior(is_used_once)', ('feq', 'a', 'b'), ('feq', 'c', 'd'))) => ('iand', ('fneu', 'a', 'b'), ('fneu', 'c', 'd'))
 *    ('inot', ('iand(is_used_once)', ('feq', 'a', 'b'), ('feq', 'c', 'd'))) => ('ior', ('fneu', 'a', 'b'), ('fneu', 'c', 'd'))
 *    ('inot', ('ior(is_used_once)', ('feq', 'a', 'b'), ('fneu', 'c', 'd'))) => ('iand', ('fneu', 'a', 'b'), ('feq', 'c', 'd'))
 *    ('inot', ('iand(is_used_once)', ('feq', 'a', 'b'), ('fneu', 'c', 'd'))) => ('ior', ('fneu', 'a', 'b'), ('feq', 'c', 'd'))
 *    ('inot', ('ior(is_used_once)', ('fneu', 'a', 'b'), ('fneu', 'c', 'd'))) => ('iand', ('feq', 'a', 'b'), ('feq', 'c', 'd'))
 *    ('inot', ('iand(is_used_once)', ('fneu', 'a', 'b'), ('fneu', 'c', 'd'))) => ('ior', ('feq', 'a', 'b'), ('feq', 'c', 'd'))
 *    ('f2b1', ('b2f', 'a@1')) => a
 *    ('f2b8', ('b2f', 'a@8')) => a
 *    ('f2b16', ('b2f', 'a@16')) => a
 *    ('f2b32', ('b2f', 'a@32')) => a
 *    ('f2f', ('b2f', 'a')) => ('b2f', 'a')
 *    ('f2u', ('b2f', 'a')) => ('b2i', 'a')
 *    ('f2i', ('b2f', 'a')) => ('b2i', 'a')
 *    ('u2f', ('b2i', 'a')) => ('b2f', 'a')
 *    ('u2u', ('b2i', 'a')) => ('b2i', 'a')
 *    ('i2f', ('b2i', 'a')) => ('b2f', 'a')
 *    ('i2i', ('b2i', 'a')) => ('b2i', 'a')
 *    ('i2i1', 'a@1') => a
 *    ('i2i8', 'a@8') => a
 *    ('i2i16', 'a@16') => a
 *    ('i2i32', 'a@32') => a
 *    ('i2i64', 'a@64') => a
 *    ('u2u1', 'a@1') => a
 *    ('u2u8', 'a@8') => a
 *    ('u2u16', 'a@16') => a
 *    ('u2u32', 'a@32') => a
 *    ('u2u64', 'a@64') => a
 *    ('f2f16', 'a@16') => a
 *    ('f2f32', 'a@32') => a
 *    ('f2f64', 'a@64') => a
 *    ('b2b1', 'a@1') => a
 *    ('b2b8', 'a@8') => a
 *    ('b2b16', 'a@16') => a
 *    ('b2b32', 'a@32') => a
 *    ('i2i1', ('i2i8', 'a')) => ('i2i1', 'a')
 *    ('i2i1', ('u2u8', 'a')) => ('u2u1', 'a')
 *    ('u2u1', ('i2i8', 'a')) => ('i2i1', 'a')
 *    ('u2u1', ('u2u8', 'a')) => ('u2u1', 'a')
 *    ('i2i1', ('i2i16', 'a')) => ('i2i1', 'a')
 *    ('i2i1', ('u2u16', 'a')) => ('u2u1', 'a')
 *    ('u2u1', ('i2i16', 'a')) => ('i2i1', 'a')
 *    ('u2u1', ('u2u16', 'a')) => ('u2u1', 'a')
 *    ('i2i1', ('i2i32', 'a')) => ('i2i1', 'a')
 *    ('i2i1', ('u2u32', 'a')) => ('u2u1', 'a')
 *    ('u2u1', ('i2i32', 'a')) => ('i2i1', 'a')
 *    ('u2u1', ('u2u32', 'a')) => ('u2u1', 'a')
 *    ('i2i1', ('i2i64', 'a')) => ('i2i1', 'a')
 *    ('i2i1', ('u2u64', 'a')) => ('u2u1', 'a')
 *    ('u2u1', ('i2i64', 'a')) => ('i2i1', 'a')
 *    ('u2u1', ('u2u64', 'a')) => ('u2u1', 'a')
 *    ('i2i8', ('i2i1', 'a@1')) => ('i2i8', 'a')
 *    ('u2u8', ('u2u1', 'a@1')) => ('u2u8', 'a')
 *    ('i2i8', ('i2i16', 'a')) => ('i2i8', 'a')
 *    ('i2i8', ('u2u16', 'a')) => ('u2u8', 'a')
 *    ('u2u8', ('i2i16', 'a')) => ('i2i8', 'a')
 *    ('u2u8', ('u2u16', 'a')) => ('u2u8', 'a')
 *    ('i2i8', ('i2i32', 'a')) => ('i2i8', 'a')
 *    ('i2i8', ('u2u32', 'a')) => ('u2u8', 'a')
 *    ('u2u8', ('i2i32', 'a')) => ('i2i8', 'a')
 *    ('u2u8', ('u2u32', 'a')) => ('u2u8', 'a')
 *    ('i2i8', ('i2i64', 'a')) => ('i2i8', 'a')
 *    ('i2i8', ('u2u64', 'a')) => ('u2u8', 'a')
 *    ('u2u8', ('i2i64', 'a')) => ('i2i8', 'a')
 *    ('u2u8', ('u2u64', 'a')) => ('u2u8', 'a')
 *    ('i2i16', ('i2i1', 'a@1')) => ('i2i16', 'a')
 *    ('u2u16', ('u2u1', 'a@1')) => ('u2u16', 'a')
 *    ('i2i16', ('i2i8', 'a@1')) => ('i2i16', 'a')
 *    ('u2u16', ('u2u8', 'a@1')) => ('u2u16', 'a')
 *    ('i2i16', ('i2i8', 'a@8')) => ('i2i16', 'a')
 *    ('u2u16', ('u2u8', 'a@8')) => ('u2u16', 'a')
 *    ('i2i16', ('i2i32', 'a')) => ('i2i16', 'a')
 *    ('i2i16', ('u2u32', 'a')) => ('u2u16', 'a')
 *    ('u2u16', ('i2i32', 'a')) => ('i2i16', 'a')
 *    ('u2u16', ('u2u32', 'a')) => ('u2u16', 'a')
 *    ('i2i16', ('i2i64', 'a')) => ('i2i16', 'a')
 *    ('i2i16', ('u2u64', 'a')) => ('u2u16', 'a')
 *    ('u2u16', ('i2i64', 'a')) => ('i2i16', 'a')
 *    ('u2u16', ('u2u64', 'a')) => ('u2u16', 'a')
 *    ('i2i32', ('i2i1', 'a@1')) => ('i2i32', 'a')
 *    ('u2u32', ('u2u1', 'a@1')) => ('u2u32', 'a')
 *    ('i2i32', ('i2i8', 'a@1')) => ('i2i32', 'a')
 *    ('u2u32', ('u2u8', 'a@1')) => ('u2u32', 'a')
 *    ('i2i32', ('i2i8', 'a@8')) => ('i2i32', 'a')
 *    ('u2u32', ('u2u8', 'a@8')) => ('u2u32', 'a')
 *    ('i2i32', ('i2i16', 'a@1')) => ('i2i32', 'a')
 *    ('u2u32', ('u2u16', 'a@1')) => ('u2u32', 'a')
 *    ('i2i32', ('i2i16', 'a@8')) => ('i2i32', 'a')
 *    ('u2u32', ('u2u16', 'a@8')) => ('u2u32', 'a')
 *    ('i2i32', ('i2i16', 'a@16')) => ('i2i32', 'a')
 *    ('u2u32', ('u2u16', 'a@16')) => ('u2u32', 'a')
 *    ('i2i32', ('i2i64', 'a')) => ('i2i32', 'a')
 *    ('i2i32', ('u2u64', 'a')) => ('u2u32', 'a')
 *    ('u2u32', ('i2i64', 'a')) => ('i2i32', 'a')
 *    ('u2u32', ('u2u64', 'a')) => ('u2u32', 'a')
 *    ('i2i64', ('i2i1', 'a@1')) => ('i2i64', 'a')
 *    ('u2u64', ('u2u1', 'a@1')) => ('u2u64', 'a')
 *    ('i2i64', ('i2i8', 'a@1')) => ('i2i64', 'a')
 *    ('u2u64', ('u2u8', 'a@1')) => ('u2u64', 'a')
 *    ('i2i64', ('i2i8', 'a@8')) => ('i2i64', 'a')
 *    ('u2u64', ('u2u8', 'a@8')) => ('u2u64', 'a')
 *    ('i2i64', ('i2i16', 'a@1')) => ('i2i64', 'a')
 *    ('u2u64', ('u2u16', 'a@1')) => ('u2u64', 'a')
 *    ('i2i64', ('i2i16', 'a@8')) => ('i2i64', 'a')
 *    ('u2u64', ('u2u16', 'a@8')) => ('u2u64', 'a')
 *    ('i2i64', ('i2i16', 'a@16')) => ('i2i64', 'a')
 *    ('u2u64', ('u2u16', 'a@16')) => ('u2u64', 'a')
 *    ('i2i64', ('i2i32', 'a@1')) => ('i2i64', 'a')
 *    ('u2u64', ('u2u32', 'a@1')) => ('u2u64', 'a')
 *    ('i2i64', ('i2i32', 'a@8')) => ('i2i64', 'a')
 *    ('u2u64', ('u2u32', 'a@8')) => ('u2u64', 'a')
 *    ('i2i64', ('i2i32', 'a@16')) => ('i2i64', 'a')
 *    ('u2u64', ('u2u32', 'a@16')) => ('u2u64', 'a')
 *    ('i2i64', ('i2i32', 'a@32')) => ('i2i64', 'a')
 *    ('u2u64', ('u2u32', 'a@32')) => ('u2u64', 'a')
 *    ('i2i8', ('pack_64_2x32_split', 'a', 'b')) => ('i2i8', 'a')
 *    ('i2i8', ('pack_64_2x32_split', 'a', 'b')) => ('i2i8', 'a')
 *    ('i2i16', ('pack_64_2x32_split', 'a', 'b')) => ('i2i16', 'a')
 *    ('i2i16', ('pack_64_2x32_split', 'a', 'b')) => ('i2i16', 'a')
 *    ('i2i32', ('pack_64_2x32_split', 'a', 'b')) => ('i2i32', 'a')
 *    ('i2i32', ('pack_64_2x32_split', 'a', 'b')) => ('i2i32', 'a')
 *    ('u2u8', ('pack_64_2x32_split', 'a', 'b')) => ('u2u8', 'a')
 *    ('u2u8', ('pack_64_2x32_split', 'a', 'b')) => ('u2u8', 'a')
 *    ('u2u16', ('pack_64_2x32_split', 'a', 'b')) => ('u2u16', 'a')
 *    ('u2u16', ('pack_64_2x32_split', 'a', 'b')) => ('u2u16', 'a')
 *    ('u2u32', ('pack_64_2x32_split', 'a', 'b')) => ('u2u32', 'a')
 *    ('u2u32', ('pack_64_2x32_split', 'a', 'b')) => ('u2u32', 'a')
 *    ('ieq', ('i2i16', 'a@8'), ('i2i16', 'b@8')) => ('ieq', 'a', ('i2i8', 'b'))
 *    ('ine', ('i2i16', 'a@8'), ('i2i16', 'b@8')) => ('ine', 'a', ('i2i8', 'b'))
 *    ('ige', ('i2i16', 'a@8'), ('i2i16', 'b@8')) => ('ige', 'a', ('i2i8', 'b'))
 *    ('ilt', ('i2i16', 'a@8'), ('i2i16', 'b@8')) => ('ilt', 'a', ('i2i8', 'b'))
 *    ('ige', ('i2i16', 'b@8'), ('i2i16', 'a@8')) => ('ige', ('i2i8', 'b'), 'a')
 *    ('ilt', ('i2i16', 'b@8'), ('i2i16', 'a@8')) => ('ilt', ('i2i8', 'b'), 'a')
 *    ('ieq', ('i2i16', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('i2i8', 'b')), ('ieq', ('i2i16', ('i2i8', 'b')), 'b'))
 *    ('ine', ('i2i16', 'a@8'), '#b') => ('ior', ('ine', 'a', ('i2i8', 'b')), ('ine', ('i2i16', ('i2i8', 'b')), 'b'))
 *    ('ilt', ('i2i16', 'a@8'), '#b') => ('iand', ('ilt', -128, 'b'), ('ior', ('ilt', 127, 'b'), ('ilt', 'a', ('i2i8', 'b'))))
 *    ('ilt', '#a', ('i2i16', 'b@8')) => ('iand', ('ilt', 'a', 127), ('ior', ('ilt', 'a', -128), ('ilt', ('i2i8', 'a'), 'b')))
 *    ('ige', ('i2i16', 'a@8'), '#b') => ('iand', ('ige', 127, 'b'), ('ior', ('ige', -128, 'b'), ('ige', 'a', ('i2i8', 'b'))))
 *    ('ige', '#a', ('i2i16', 'b@8')) => ('iand', ('ige', 'a', -128), ('ior', ('ige', 'a', 127), ('ige', ('i2i8', 'a'), 'b')))
 *    ('ieq', ('i2i32', 'a@8'), ('i2i32', 'b@8')) => ('ieq', 'a', ('i2i8', 'b'))
 *    ('ine', ('i2i32', 'a@8'), ('i2i32', 'b@8')) => ('ine', 'a', ('i2i8', 'b'))
 *    ('ige', ('i2i32', 'a@8'), ('i2i32', 'b@8')) => ('ige', 'a', ('i2i8', 'b'))
 *    ('ilt', ('i2i32', 'a@8'), ('i2i32', 'b@8')) => ('ilt', 'a', ('i2i8', 'b'))
 *    ('ige', ('i2i32', 'b@8'), ('i2i32', 'a@8')) => ('ige', ('i2i8', 'b'), 'a')
 *    ('ilt', ('i2i32', 'b@8'), ('i2i32', 'a@8')) => ('ilt', ('i2i8', 'b'), 'a')
 *    ('ieq', ('i2i32', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('i2i8', 'b')), ('ieq', ('i2i32', ('i2i8', 'b')), 'b'))
 *    ('ine', ('i2i32', 'a@8'), '#b') => ('ior', ('ine', 'a', ('i2i8', 'b')), ('ine', ('i2i32', ('i2i8', 'b')), 'b'))
 *    ('ilt', ('i2i32', 'a@8'), '#b') => ('iand', ('ilt', -128, 'b'), ('ior', ('ilt', 127, 'b'), ('ilt', 'a', ('i2i8', 'b'))))
 *    ('ilt', '#a', ('i2i32', 'b@8')) => ('iand', ('ilt', 'a', 127), ('ior', ('ilt', 'a', -128), ('ilt', ('i2i8', 'a'), 'b')))
 *    ('ige', ('i2i32', 'a@8'), '#b') => ('iand', ('ige', 127, 'b'), ('ior', ('ige', -128, 'b'), ('ige', 'a', ('i2i8', 'b'))))
 *    ('ige', '#a', ('i2i32', 'b@8')) => ('iand', ('ige', 'a', -128), ('ior', ('ige', 'a', 127), ('ige', ('i2i8', 'a'), 'b')))
 *    ('ieq', ('i2i64', 'a@8'), ('i2i64', 'b@8')) => ('ieq', 'a', ('i2i8', 'b'))
 *    ('ine', ('i2i64', 'a@8'), ('i2i64', 'b@8')) => ('ine', 'a', ('i2i8', 'b'))
 *    ('ige', ('i2i64', 'a@8'), ('i2i64', 'b@8')) => ('ige', 'a', ('i2i8', 'b'))
 *    ('ilt', ('i2i64', 'a@8'), ('i2i64', 'b@8')) => ('ilt', 'a', ('i2i8', 'b'))
 *    ('ige', ('i2i64', 'b@8'), ('i2i64', 'a@8')) => ('ige', ('i2i8', 'b'), 'a')
 *    ('ilt', ('i2i64', 'b@8'), ('i2i64', 'a@8')) => ('ilt', ('i2i8', 'b'), 'a')
 *    ('ieq', ('i2i64', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('i2i8', 'b')), ('ieq', ('i2i64', ('i2i8', 'b')), 'b'))
 *    ('ine', ('i2i64', 'a@8'), '#b') => ('ior', ('ine', 'a', ('i2i8', 'b')), ('ine', ('i2i64', ('i2i8', 'b')), 'b'))
 *    ('ilt', ('i2i64', 'a@8'), '#b') => ('iand', ('ilt', -128, 'b'), ('ior', ('ilt', 127, 'b'), ('ilt', 'a', ('i2i8', 'b'))))
 *    ('ilt', '#a', ('i2i64', 'b@8')) => ('iand', ('ilt', 'a', 127), ('ior', ('ilt', 'a', -128), ('ilt', ('i2i8', 'a'), 'b')))
 *    ('ige', ('i2i64', 'a@8'), '#b') => ('iand', ('ige', 127, 'b'), ('ior', ('ige', -128, 'b'), ('ige', 'a', ('i2i8', 'b'))))
 *    ('ige', '#a', ('i2i64', 'b@8')) => ('iand', ('ige', 'a', -128), ('ior', ('ige', 'a', 127), ('ige', ('i2i8', 'a'), 'b')))
 *    ('ieq', ('i2i32', 'a@16'), ('i2i32', 'b@8')) => ('ieq', 'a', ('i2i16', 'b'))
 *    ('ine', ('i2i32', 'a@16'), ('i2i32', 'b@8')) => ('ine', 'a', ('i2i16', 'b'))
 *    ('ige', ('i2i32', 'a@16'), ('i2i32', 'b@8')) => ('ige', 'a', ('i2i16', 'b'))
 *    ('ilt', ('i2i32', 'a@16'), ('i2i32', 'b@8')) => ('ilt', 'a', ('i2i16', 'b'))
 *    ('ige', ('i2i32', 'b@8'), ('i2i32', 'a@16')) => ('ige', ('i2i16', 'b'), 'a')
 *    ('ilt', ('i2i32', 'b@8'), ('i2i32', 'a@16')) => ('ilt', ('i2i16', 'b'), 'a')
 *    ('ieq', ('i2i32', 'a@16'), ('i2i32', 'b@16')) => ('ieq', 'a', ('i2i16', 'b'))
 *    ('ine', ('i2i32', 'a@16'), ('i2i32', 'b@16')) => ('ine', 'a', ('i2i16', 'b'))
 *    ('ige', ('i2i32', 'a@16'), ('i2i32', 'b@16')) => ('ige', 'a', ('i2i16', 'b'))
 *    ('ilt', ('i2i32', 'a@16'), ('i2i32', 'b@16')) => ('ilt', 'a', ('i2i16', 'b'))
 *    ('ige', ('i2i32', 'b@16'), ('i2i32', 'a@16')) => ('ige', ('i2i16', 'b'), 'a')
 *    ('ilt', ('i2i32', 'b@16'), ('i2i32', 'a@16')) => ('ilt', ('i2i16', 'b'), 'a')
 *    ('ieq', ('i2i32', 'a@16'), '#b') => ('iand', ('ieq', 'a', ('i2i16', 'b')), ('ieq', ('i2i32', ('i2i16', 'b')), 'b'))
 *    ('ine', ('i2i32', 'a@16'), '#b') => ('ior', ('ine', 'a', ('i2i16', 'b')), ('ine', ('i2i32', ('i2i16', 'b')), 'b'))
 *    ('ilt', ('i2i32', 'a@16'), '#b') => ('iand', ('ilt', -32768, 'b'), ('ior', ('ilt', 32767, 'b'), ('ilt', 'a', ('i2i16', 'b'))))
 *    ('ilt', '#a', ('i2i32', 'b@16')) => ('iand', ('ilt', 'a', 32767), ('ior', ('ilt', 'a', -32768), ('ilt', ('i2i16', 'a'), 'b')))
 *    ('ige', ('i2i32', 'a@16'), '#b') => ('iand', ('ige', 32767, 'b'), ('ior', ('ige', -32768, 'b'), ('ige', 'a', ('i2i16', 'b'))))
 *    ('ige', '#a', ('i2i32', 'b@16')) => ('iand', ('ige', 'a', -32768), ('ior', ('ige', 'a', 32767), ('ige', ('i2i16', 'a'), 'b')))
 *    ('ieq', ('i2i64', 'a@16'), ('i2i64', 'b@8')) => ('ieq', 'a', ('i2i16', 'b'))
 *    ('ine', ('i2i64', 'a@16'), ('i2i64', 'b@8')) => ('ine', 'a', ('i2i16', 'b'))
 *    ('ige', ('i2i64', 'a@16'), ('i2i64', 'b@8')) => ('ige', 'a', ('i2i16', 'b'))
 *    ('ilt', ('i2i64', 'a@16'), ('i2i64', 'b@8')) => ('ilt', 'a', ('i2i16', 'b'))
 *    ('ige', ('i2i64', 'b@8'), ('i2i64', 'a@16')) => ('ige', ('i2i16', 'b'), 'a')
 *    ('ilt', ('i2i64', 'b@8'), ('i2i64', 'a@16')) => ('ilt', ('i2i16', 'b'), 'a')
 *    ('ieq', ('i2i64', 'a@16'), ('i2i64', 'b@16')) => ('ieq', 'a', ('i2i16', 'b'))
 *    ('ine', ('i2i64', 'a@16'), ('i2i64', 'b@16')) => ('ine', 'a', ('i2i16', 'b'))
 *    ('ige', ('i2i64', 'a@16'), ('i2i64', 'b@16')) => ('ige', 'a', ('i2i16', 'b'))
 *    ('ilt', ('i2i64', 'a@16'), ('i2i64', 'b@16')) => ('ilt', 'a', ('i2i16', 'b'))
 *    ('ige', ('i2i64', 'b@16'), ('i2i64', 'a@16')) => ('ige', ('i2i16', 'b'), 'a')
 *    ('ilt', ('i2i64', 'b@16'), ('i2i64', 'a@16')) => ('ilt', ('i2i16', 'b'), 'a')
 *    ('ieq', ('i2i64', 'a@16'), '#b') => ('iand', ('ieq', 'a', ('i2i16', 'b')), ('ieq', ('i2i64', ('i2i16', 'b')), 'b'))
 *    ('ine', ('i2i64', 'a@16'), '#b') => ('ior', ('ine', 'a', ('i2i16', 'b')), ('ine', ('i2i64', ('i2i16', 'b')), 'b'))
 *    ('ilt', ('i2i64', 'a@16'), '#b') => ('iand', ('ilt', -32768, 'b'), ('ior', ('ilt', 32767, 'b'), ('ilt', 'a', ('i2i16', 'b'))))
 *    ('ilt', '#a', ('i2i64', 'b@16')) => ('iand', ('ilt', 'a', 32767), ('ior', ('ilt', 'a', -32768), ('ilt', ('i2i16', 'a'), 'b')))
 *    ('ige', ('i2i64', 'a@16'), '#b') => ('iand', ('ige', 32767, 'b'), ('ior', ('ige', -32768, 'b'), ('ige', 'a', ('i2i16', 'b'))))
 *    ('ige', '#a', ('i2i64', 'b@16')) => ('iand', ('ige', 'a', -32768), ('ior', ('ige', 'a', 32767), ('ige', ('i2i16', 'a'), 'b')))
 *    ('ieq', ('i2i64', 'a@32'), ('i2i64', 'b@8')) => ('ieq', 'a', ('i2i32', 'b'))
 *    ('ine', ('i2i64', 'a@32'), ('i2i64', 'b@8')) => ('ine', 'a', ('i2i32', 'b'))
 *    ('ige', ('i2i64', 'a@32'), ('i2i64', 'b@8')) => ('ige', 'a', ('i2i32', 'b'))
 *    ('ilt', ('i2i64', 'a@32'), ('i2i64', 'b@8')) => ('ilt', 'a', ('i2i32', 'b'))
 *    ('ige', ('i2i64', 'b@8'), ('i2i64', 'a@32')) => ('ige', ('i2i32', 'b'), 'a')
 *    ('ilt', ('i2i64', 'b@8'), ('i2i64', 'a@32')) => ('ilt', ('i2i32', 'b'), 'a')
 *    ('ieq', ('i2i64', 'a@32'), ('i2i64', 'b@16')) => ('ieq', 'a', ('i2i32', 'b'))
 *    ('ine', ('i2i64', 'a@32'), ('i2i64', 'b@16')) => ('ine', 'a', ('i2i32', 'b'))
 *    ('ige', ('i2i64', 'a@32'), ('i2i64', 'b@16')) => ('ige', 'a', ('i2i32', 'b'))
 *    ('ilt', ('i2i64', 'a@32'), ('i2i64', 'b@16')) => ('ilt', 'a', ('i2i32', 'b'))
 *    ('ige', ('i2i64', 'b@16'), ('i2i64', 'a@32')) => ('ige', ('i2i32', 'b'), 'a')
 *    ('ilt', ('i2i64', 'b@16'), ('i2i64', 'a@32')) => ('ilt', ('i2i32', 'b'), 'a')
 *    ('ieq', ('i2i64', 'a@32'), ('i2i64', 'b@32')) => ('ieq', 'a', ('i2i32', 'b'))
 *    ('ine', ('i2i64', 'a@32'), ('i2i64', 'b@32')) => ('ine', 'a', ('i2i32', 'b'))
 *    ('ige', ('i2i64', 'a@32'), ('i2i64', 'b@32')) => ('ige', 'a', ('i2i32', 'b'))
 *    ('ilt', ('i2i64', 'a@32'), ('i2i64', 'b@32')) => ('ilt', 'a', ('i2i32', 'b'))
 *    ('ige', ('i2i64', 'b@32'), ('i2i64', 'a@32')) => ('ige', ('i2i32', 'b'), 'a')
 *    ('ilt', ('i2i64', 'b@32'), ('i2i64', 'a@32')) => ('ilt', ('i2i32', 'b'), 'a')
 *    ('ieq', ('i2i64', 'a@32'), '#b') => ('iand', ('ieq', 'a', ('i2i32', 'b')), ('ieq', ('i2i64', ('i2i32', 'b')), 'b'))
 *    ('ine', ('i2i64', 'a@32'), '#b') => ('ior', ('ine', 'a', ('i2i32', 'b')), ('ine', ('i2i64', ('i2i32', 'b')), 'b'))
 *    ('ilt', ('i2i64', 'a@32'), '#b') => ('iand', ('ilt', -2147483648, 'b'), ('ior', ('ilt', 2147483647, 'b'), ('ilt', 'a', ('i2i32', 'b'))))
 *    ('ilt', '#a', ('i2i64', 'b@32')) => ('iand', ('ilt', 'a', 2147483647), ('ior', ('ilt', 'a', -2147483648), ('ilt', ('i2i32', 'a'), 'b')))
 *    ('ige', ('i2i64', 'a@32'), '#b') => ('iand', ('ige', 2147483647, 'b'), ('ior', ('ige', -2147483648, 'b'), ('ige', 'a', ('i2i32', 'b'))))
 *    ('ige', '#a', ('i2i64', 'b@32')) => ('iand', ('ige', 'a', -2147483648), ('ior', ('ige', 'a', 2147483647), ('ige', ('i2i32', 'a'), 'b')))
 *    ('ieq', ('u2u16', 'a@8'), ('u2u16', 'b@8')) => ('ieq', 'a', ('u2u8', 'b'))
 *    ('ine', ('u2u16', 'a@8'), ('u2u16', 'b@8')) => ('ine', 'a', ('u2u8', 'b'))
 *    ('uge', ('u2u16', 'a@8'), ('u2u16', 'b@8')) => ('uge', 'a', ('u2u8', 'b'))
 *    ('ult', ('u2u16', 'a@8'), ('u2u16', 'b@8')) => ('ult', 'a', ('u2u8', 'b'))
 *    ('uge', ('u2u16', 'b@8'), ('u2u16', 'a@8')) => ('uge', ('u2u8', 'b'), 'a')
 *    ('ult', ('u2u16', 'b@8'), ('u2u16', 'a@8')) => ('ult', ('u2u8', 'b'), 'a')
 *    ('ieq', ('u2u16', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('u2u8', 'b')), ('ieq', ('u2u16', ('u2u8', 'b')), 'b'))
 *    ('ine', ('u2u16', 'a@8'), '#b') => ('ior', ('ine', 'a', ('u2u8', 'b')), ('ine', ('u2u16', ('u2u8', 'b')), 'b'))
 *    ('ult', ('u2u16', 'a@8'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 255, 'b'), ('ult', 'a', ('u2u8', 'b'))))
 *    ('ult', '#a', ('u2u16', 'b@8')) => ('iand', ('ult', 'a', 255), ('ior', ('ult', 'a', 0), ('ult', ('u2u8', 'a'), 'b')))
 *    ('uge', ('u2u16', 'a@8'), '#b') => ('iand', ('uge', 255, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u8', 'b'))))
 *    ('uge', '#a', ('u2u16', 'b@8')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 255), ('uge', ('u2u8', 'a'), 'b')))
 *    ('ieq', ('u2u32', 'a@8'), ('u2u32', 'b@8')) => ('ieq', 'a', ('u2u8', 'b'))
 *    ('ine', ('u2u32', 'a@8'), ('u2u32', 'b@8')) => ('ine', 'a', ('u2u8', 'b'))
 *    ('uge', ('u2u32', 'a@8'), ('u2u32', 'b@8')) => ('uge', 'a', ('u2u8', 'b'))
 *    ('ult', ('u2u32', 'a@8'), ('u2u32', 'b@8')) => ('ult', 'a', ('u2u8', 'b'))
 *    ('uge', ('u2u32', 'b@8'), ('u2u32', 'a@8')) => ('uge', ('u2u8', 'b'), 'a')
 *    ('ult', ('u2u32', 'b@8'), ('u2u32', 'a@8')) => ('ult', ('u2u8', 'b'), 'a')
 *    ('ieq', ('u2u32', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('u2u8', 'b')), ('ieq', ('u2u32', ('u2u8', 'b')), 'b'))
 *    ('ine', ('u2u32', 'a@8'), '#b') => ('ior', ('ine', 'a', ('u2u8', 'b')), ('ine', ('u2u32', ('u2u8', 'b')), 'b'))
 *    ('ult', ('u2u32', 'a@8'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 255, 'b'), ('ult', 'a', ('u2u8', 'b'))))
 *    ('ult', '#a', ('u2u32', 'b@8')) => ('iand', ('ult', 'a', 255), ('ior', ('ult', 'a', 0), ('ult', ('u2u8', 'a'), 'b')))
 *    ('uge', ('u2u32', 'a@8'), '#b') => ('iand', ('uge', 255, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u8', 'b'))))
 *    ('uge', '#a', ('u2u32', 'b@8')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 255), ('uge', ('u2u8', 'a'), 'b')))
 *    ('ieq', ('u2u64', 'a@8'), ('u2u64', 'b@8')) => ('ieq', 'a', ('u2u8', 'b'))
 *    ('ine', ('u2u64', 'a@8'), ('u2u64', 'b@8')) => ('ine', 'a', ('u2u8', 'b'))
 *    ('uge', ('u2u64', 'a@8'), ('u2u64', 'b@8')) => ('uge', 'a', ('u2u8', 'b'))
 *    ('ult', ('u2u64', 'a@8'), ('u2u64', 'b@8')) => ('ult', 'a', ('u2u8', 'b'))
 *    ('uge', ('u2u64', 'b@8'), ('u2u64', 'a@8')) => ('uge', ('u2u8', 'b'), 'a')
 *    ('ult', ('u2u64', 'b@8'), ('u2u64', 'a@8')) => ('ult', ('u2u8', 'b'), 'a')
 *    ('ieq', ('u2u64', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('u2u8', 'b')), ('ieq', ('u2u64', ('u2u8', 'b')), 'b'))
 *    ('ine', ('u2u64', 'a@8'), '#b') => ('ior', ('ine', 'a', ('u2u8', 'b')), ('ine', ('u2u64', ('u2u8', 'b')), 'b'))
 *    ('ult', ('u2u64', 'a@8'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 255, 'b'), ('ult', 'a', ('u2u8', 'b'))))
 *    ('ult', '#a', ('u2u64', 'b@8')) => ('iand', ('ult', 'a', 255), ('ior', ('ult', 'a', 0), ('ult', ('u2u8', 'a'), 'b')))
 *    ('uge', ('u2u64', 'a@8'), '#b') => ('iand', ('uge', 255, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u8', 'b'))))
 *    ('uge', '#a', ('u2u64', 'b@8')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 255), ('uge', ('u2u8', 'a'), 'b')))
 *    ('ieq', ('u2u32', 'a@16'), ('u2u32', 'b@8')) => ('ieq', 'a', ('u2u16', 'b'))
 *    ('ine', ('u2u32', 'a@16'), ('u2u32', 'b@8')) => ('ine', 'a', ('u2u16', 'b'))
 *    ('uge', ('u2u32', 'a@16'), ('u2u32', 'b@8')) => ('uge', 'a', ('u2u16', 'b'))
 *    ('ult', ('u2u32', 'a@16'), ('u2u32', 'b@8')) => ('ult', 'a', ('u2u16', 'b'))
 *    ('uge', ('u2u32', 'b@8'), ('u2u32', 'a@16')) => ('uge', ('u2u16', 'b'), 'a')
 *    ('ult', ('u2u32', 'b@8'), ('u2u32', 'a@16')) => ('ult', ('u2u16', 'b'), 'a')
 *    ('ieq', ('u2u32', 'a@16'), ('u2u32', 'b@16')) => ('ieq', 'a', ('u2u16', 'b'))
 *    ('ine', ('u2u32', 'a@16'), ('u2u32', 'b@16')) => ('ine', 'a', ('u2u16', 'b'))
 *    ('uge', ('u2u32', 'a@16'), ('u2u32', 'b@16')) => ('uge', 'a', ('u2u16', 'b'))
 *    ('ult', ('u2u32', 'a@16'), ('u2u32', 'b@16')) => ('ult', 'a', ('u2u16', 'b'))
 *    ('uge', ('u2u32', 'b@16'), ('u2u32', 'a@16')) => ('uge', ('u2u16', 'b'), 'a')
 *    ('ult', ('u2u32', 'b@16'), ('u2u32', 'a@16')) => ('ult', ('u2u16', 'b'), 'a')
 *    ('ieq', ('u2u32', 'a@16'), '#b') => ('iand', ('ieq', 'a', ('u2u16', 'b')), ('ieq', ('u2u32', ('u2u16', 'b')), 'b'))
 *    ('ine', ('u2u32', 'a@16'), '#b') => ('ior', ('ine', 'a', ('u2u16', 'b')), ('ine', ('u2u32', ('u2u16', 'b')), 'b'))
 *    ('ult', ('u2u32', 'a@16'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 65535, 'b'), ('ult', 'a', ('u2u16', 'b'))))
 *    ('ult', '#a', ('u2u32', 'b@16')) => ('iand', ('ult', 'a', 65535), ('ior', ('ult', 'a', 0), ('ult', ('u2u16', 'a'), 'b')))
 *    ('uge', ('u2u32', 'a@16'), '#b') => ('iand', ('uge', 65535, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u16', 'b'))))
 *    ('uge', '#a', ('u2u32', 'b@16')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 65535), ('uge', ('u2u16', 'a'), 'b')))
 *    ('ieq', ('u2u64', 'a@16'), ('u2u64', 'b@8')) => ('ieq', 'a', ('u2u16', 'b'))
 *    ('ine', ('u2u64', 'a@16'), ('u2u64', 'b@8')) => ('ine', 'a', ('u2u16', 'b'))
 *    ('uge', ('u2u64', 'a@16'), ('u2u64', 'b@8')) => ('uge', 'a', ('u2u16', 'b'))
 *    ('ult', ('u2u64', 'a@16'), ('u2u64', 'b@8')) => ('ult', 'a', ('u2u16', 'b'))
 *    ('uge', ('u2u64', 'b@8'), ('u2u64', 'a@16')) => ('uge', ('u2u16', 'b'), 'a')
 *    ('ult', ('u2u64', 'b@8'), ('u2u64', 'a@16')) => ('ult', ('u2u16', 'b'), 'a')
 *    ('ieq', ('u2u64', 'a@16'), ('u2u64', 'b@16')) => ('ieq', 'a', ('u2u16', 'b'))
 *    ('ine', ('u2u64', 'a@16'), ('u2u64', 'b@16')) => ('ine', 'a', ('u2u16', 'b'))
 *    ('uge', ('u2u64', 'a@16'), ('u2u64', 'b@16')) => ('uge', 'a', ('u2u16', 'b'))
 *    ('ult', ('u2u64', 'a@16'), ('u2u64', 'b@16')) => ('ult', 'a', ('u2u16', 'b'))
 *    ('uge', ('u2u64', 'b@16'), ('u2u64', 'a@16')) => ('uge', ('u2u16', 'b'), 'a')
 *    ('ult', ('u2u64', 'b@16'), ('u2u64', 'a@16')) => ('ult', ('u2u16', 'b'), 'a')
 *    ('ieq', ('u2u64', 'a@16'), '#b') => ('iand', ('ieq', 'a', ('u2u16', 'b')), ('ieq', ('u2u64', ('u2u16', 'b')), 'b'))
 *    ('ine', ('u2u64', 'a@16'), '#b') => ('ior', ('ine', 'a', ('u2u16', 'b')), ('ine', ('u2u64', ('u2u16', 'b')), 'b'))
 *    ('ult', ('u2u64', 'a@16'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 65535, 'b'), ('ult', 'a', ('u2u16', 'b'))))
 *    ('ult', '#a', ('u2u64', 'b@16')) => ('iand', ('ult', 'a', 65535), ('ior', ('ult', 'a', 0), ('ult', ('u2u16', 'a'), 'b')))
 *    ('uge', ('u2u64', 'a@16'), '#b') => ('iand', ('uge', 65535, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u16', 'b'))))
 *    ('uge', '#a', ('u2u64', 'b@16')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 65535), ('uge', ('u2u16', 'a'), 'b')))
 *    ('ieq', ('u2u64', 'a@32'), ('u2u64', 'b@8')) => ('ieq', 'a', ('u2u32', 'b'))
 *    ('ine', ('u2u64', 'a@32'), ('u2u64', 'b@8')) => ('ine', 'a', ('u2u32', 'b'))
 *    ('uge', ('u2u64', 'a@32'), ('u2u64', 'b@8')) => ('uge', 'a', ('u2u32', 'b'))
 *    ('ult', ('u2u64', 'a@32'), ('u2u64', 'b@8')) => ('ult', 'a', ('u2u32', 'b'))
 *    ('uge', ('u2u64', 'b@8'), ('u2u64', 'a@32')) => ('uge', ('u2u32', 'b'), 'a')
 *    ('ult', ('u2u64', 'b@8'), ('u2u64', 'a@32')) => ('ult', ('u2u32', 'b'), 'a')
 *    ('ieq', ('u2u64', 'a@32'), ('u2u64', 'b@16')) => ('ieq', 'a', ('u2u32', 'b'))
 *    ('ine', ('u2u64', 'a@32'), ('u2u64', 'b@16')) => ('ine', 'a', ('u2u32', 'b'))
 *    ('uge', ('u2u64', 'a@32'), ('u2u64', 'b@16')) => ('uge', 'a', ('u2u32', 'b'))
 *    ('ult', ('u2u64', 'a@32'), ('u2u64', 'b@16')) => ('ult', 'a', ('u2u32', 'b'))
 *    ('uge', ('u2u64', 'b@16'), ('u2u64', 'a@32')) => ('uge', ('u2u32', 'b'), 'a')
 *    ('ult', ('u2u64', 'b@16'), ('u2u64', 'a@32')) => ('ult', ('u2u32', 'b'), 'a')
 *    ('ieq', ('u2u64', 'a@32'), ('u2u64', 'b@32')) => ('ieq', 'a', ('u2u32', 'b'))
 *    ('ine', ('u2u64', 'a@32'), ('u2u64', 'b@32')) => ('ine', 'a', ('u2u32', 'b'))
 *    ('uge', ('u2u64', 'a@32'), ('u2u64', 'b@32')) => ('uge', 'a', ('u2u32', 'b'))
 *    ('ult', ('u2u64', 'a@32'), ('u2u64', 'b@32')) => ('ult', 'a', ('u2u32', 'b'))
 *    ('uge', ('u2u64', 'b@32'), ('u2u64', 'a@32')) => ('uge', ('u2u32', 'b'), 'a')
 *    ('ult', ('u2u64', 'b@32'), ('u2u64', 'a@32')) => ('ult', ('u2u32', 'b'), 'a')
 *    ('ieq', ('u2u64', 'a@32'), '#b') => ('iand', ('ieq', 'a', ('u2u32', 'b')), ('ieq', ('u2u64', ('u2u32', 'b')), 'b'))
 *    ('ine', ('u2u64', 'a@32'), '#b') => ('ior', ('ine', 'a', ('u2u32', 'b')), ('ine', ('u2u64', ('u2u32', 'b')), 'b'))
 *    ('ult', ('u2u64', 'a@32'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 4294967295, 'b'), ('ult', 'a', ('u2u32', 'b'))))
 *    ('ult', '#a', ('u2u64', 'b@32')) => ('iand', ('ult', 'a', 4294967295), ('ior', ('ult', 'a', 0), ('ult', ('u2u32', 'a'), 'b')))
 *    ('uge', ('u2u64', 'a@32'), '#b') => ('iand', ('uge', 4294967295, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u32', 'b'))))
 *    ('uge', '#a', ('u2u64', 'b@32')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 4294967295), ('uge', ('u2u32', 'a'), 'b')))
 *    ('feq', ('f2f32', 'a@16'), ('f2f32', 'b@16')) => ('feq', 'a', ('f2f16', 'b'))
 *    ('fneu', ('f2f32', 'a@16'), ('f2f32', 'b@16')) => ('fneu', 'a', ('f2f16', 'b'))
 *    ('fge', ('f2f32', 'a@16'), ('f2f32', 'b@16')) => ('fge', 'a', ('f2f16', 'b'))
 *    ('flt', ('f2f32', 'a@16'), ('f2f32', 'b@16')) => ('flt', 'a', ('f2f16', 'b'))
 *    ('fge', ('f2f32', 'b@16'), ('f2f32', 'a@16')) => ('fge', ('f2f16', 'b'), 'a')
 *    ('flt', ('f2f32', 'b@16'), ('f2f32', 'a@16')) => ('flt', ('f2f16', 'b'), 'a')
 *    ('feq', ('f2f64', 'a@16'), ('f2f64', 'b@16')) => ('feq', 'a', ('f2f16', 'b'))
 *    ('fneu', ('f2f64', 'a@16'), ('f2f64', 'b@16')) => ('fneu', 'a', ('f2f16', 'b'))
 *    ('fge', ('f2f64', 'a@16'), ('f2f64', 'b@16')) => ('fge', 'a', ('f2f16', 'b'))
 *    ('flt', ('f2f64', 'a@16'), ('f2f64', 'b@16')) => ('flt', 'a', ('f2f16', 'b'))
 *    ('fge', ('f2f64', 'b@16'), ('f2f64', 'a@16')) => ('fge', ('f2f16', 'b'), 'a')
 *    ('flt', ('f2f64', 'b@16'), ('f2f64', 'a@16')) => ('flt', ('f2f16', 'b'), 'a')
 *    ('feq', ('f2f64', 'a@32'), ('f2f64', 'b@16')) => ('feq', 'a', ('f2f32', 'b'))
 *    ('fneu', ('f2f64', 'a@32'), ('f2f64', 'b@16')) => ('fneu', 'a', ('f2f32', 'b'))
 *    ('fge', ('f2f64', 'a@32'), ('f2f64', 'b@16')) => ('fge', 'a', ('f2f32', 'b'))
 *    ('flt', ('f2f64', 'a@32'), ('f2f64', 'b@16')) => ('flt', 'a', ('f2f32', 'b'))
 *    ('fge', ('f2f64', 'b@16'), ('f2f64', 'a@32')) => ('fge', ('f2f32', 'b'), 'a')
 *    ('flt', ('f2f64', 'b@16'), ('f2f64', 'a@32')) => ('flt', ('f2f32', 'b'), 'a')
 *    ('feq', ('f2f64', 'a@32'), ('f2f64', 'b@32')) => ('feq', 'a', ('f2f32', 'b'))
 *    ('fneu', ('f2f64', 'a@32'), ('f2f64', 'b@32')) => ('fneu', 'a', ('f2f32', 'b'))
 *    ('fge', ('f2f64', 'a@32'), ('f2f64', 'b@32')) => ('fge', 'a', ('f2f32', 'b'))
 *    ('flt', ('f2f64', 'a@32'), ('f2f64', 'b@32')) => ('flt', 'a', ('f2f32', 'b'))
 *    ('fge', ('f2f64', 'b@32'), ('f2f64', 'a@32')) => ('fge', ('f2f32', 'b'), 'a')
 *    ('flt', ('f2f64', 'b@32'), ('f2f64', 'a@32')) => ('flt', ('f2f32', 'b'), 'a')
 *    ('i2i32', ('iand', 'a@64', 4294967295)) => ('u2u32', 'a')
 *    ('i2i16', ('iand', 'a@32', 65535)) => ('u2u16', 'a')
 *    ('i2i16', ('iand', 'a@64', 65535)) => ('u2u16', 'a')
 *    ('i2i8', ('iand', 'a@16', 255)) => ('u2u8', 'a')
 *    ('i2i8', ('iand', 'a@32', 255)) => ('u2u8', 'a')
 *    ('i2i8', ('iand', 'a@64', 255)) => ('u2u8', 'a')
 *    ('u2u16(only_lower_8_bits_used)', ('u2u8', 'a@16')) => a
 *    ('extract_u8(only_lower_8_bits_used)', 'a@16', 0) => a
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('u2u16', ('u2u8', 'a@16')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('u2u16', ('i2i8', 'a@16')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('extract_u8', 'a@16', 0), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('u2u16', ('u2u8', 'a@16')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('u2u16', ('i2i8', 'a@16')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('extract_u8', 'a@16', 0), 'b') => ('iadd', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('u2u16', ('u2u8', 'a@16')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('u2u16', ('i2i8', 'a@16')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('extract_u8', 'a@16', 0), 'b') => ('imul', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('u2u16', ('u2u8', 'a@16')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('u2u16', ('i2i8', 'a@16')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('extract_u8', 'a@16', 0), 'b') => ('iand', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('u2u16', ('u2u8', 'a@16')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('u2u16', ('i2i8', 'a@16')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('extract_u8', 'a@16', 0), 'b') => ('ior', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('u2u16', ('u2u8', 'a@16')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('u2u16', ('i2i8', 'a@16')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('extract_u8', 'a@16', 0), 'b') => ('ixor', 'a', 'b')
 *    ('i2i16(only_lower_8_bits_used)', ('u2u8', 'a@16')) => a
 *    ('extract_i8(only_lower_8_bits_used)', 'a@16', 0) => a
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('i2i16', ('u2u8', 'a@16')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('i2i16', ('i2i8', 'a@16')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('extract_i8', 'a@16', 0), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('i2i16', ('u2u8', 'a@16')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('i2i16', ('i2i8', 'a@16')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('extract_i8', 'a@16', 0), 'b') => ('iadd', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('i2i16', ('u2u8', 'a@16')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('i2i16', ('i2i8', 'a@16')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('extract_i8', 'a@16', 0), 'b') => ('imul', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('i2i16', ('u2u8', 'a@16')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('i2i16', ('i2i8', 'a@16')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('extract_i8', 'a@16', 0), 'b') => ('iand', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('i2i16', ('u2u8', 'a@16')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('i2i16', ('i2i8', 'a@16')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('extract_i8', 'a@16', 0), 'b') => ('ior', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('i2i16', ('u2u8', 'a@16')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('i2i16', ('i2i8', 'a@16')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('extract_i8', 'a@16', 0), 'b') => ('ixor', 'a', 'b')
 *    ('u2u32(only_lower_8_bits_used)', ('u2u8', 'a@32')) => a
 *    ('extract_u8(only_lower_8_bits_used)', 'a@32', 0) => a
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('u2u32', ('u2u8', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('u2u32', ('i2i8', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('extract_u8', 'a@32', 0), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('u2u32', ('u2u8', 'a@32')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('u2u32', ('i2i8', 'a@32')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('extract_u8', 'a@32', 0), 'b') => ('iadd', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('u2u32', ('u2u8', 'a@32')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('u2u32', ('i2i8', 'a@32')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('extract_u8', 'a@32', 0), 'b') => ('imul', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('u2u32', ('u2u8', 'a@32')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('u2u32', ('i2i8', 'a@32')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('extract_u8', 'a@32', 0), 'b') => ('iand', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('u2u32', ('u2u8', 'a@32')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('u2u32', ('i2i8', 'a@32')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('extract_u8', 'a@32', 0), 'b') => ('ior', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('u2u32', ('u2u8', 'a@32')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('u2u32', ('i2i8', 'a@32')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('extract_u8', 'a@32', 0), 'b') => ('ixor', 'a', 'b')
 *    ('i2i32(only_lower_8_bits_used)', ('u2u8', 'a@32')) => a
 *    ('extract_i8(only_lower_8_bits_used)', 'a@32', 0) => a
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('i2i32', ('u2u8', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('i2i32', ('i2i8', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('extract_i8', 'a@32', 0), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('i2i32', ('u2u8', 'a@32')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('i2i32', ('i2i8', 'a@32')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('extract_i8', 'a@32', 0), 'b') => ('iadd', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('i2i32', ('u2u8', 'a@32')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('i2i32', ('i2i8', 'a@32')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('extract_i8', 'a@32', 0), 'b') => ('imul', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('i2i32', ('u2u8', 'a@32')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('i2i32', ('i2i8', 'a@32')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('extract_i8', 'a@32', 0), 'b') => ('iand', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('i2i32', ('u2u8', 'a@32')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('i2i32', ('i2i8', 'a@32')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('extract_i8', 'a@32', 0), 'b') => ('ior', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('i2i32', ('u2u8', 'a@32')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('i2i32', ('i2i8', 'a@32')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('extract_i8', 'a@32', 0), 'b') => ('ixor', 'a', 'b')
 *    ('u2u32(only_lower_16_bits_used)', ('u2u16', 'a@32')) => a
 *    ('extract_u16(only_lower_16_bits_used)', 'a@32', 0) => a
 *    ('bcsel(only_lower_16_bits_used)', 'c', ('u2u32', ('u2u16', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_16_bits_used)', 'c', ('u2u32', ('i2i16', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_16_bits_used)', 'c', ('extract_u16', 'a@32', 0), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('iadd(only_lower_16_bits_used)', ('u2u32', ('u2u16', 'a@32')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_16_bits_used)', ('u2u32', ('i2i16', 'a@32')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_16_bits_used)', ('extract_u16', 'a@32', 0), 'b') => ('iadd', 'a', 'b')
 *    ('imul(only_lower_16_bits_used)', ('u2u32', ('u2u16', 'a@32')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_16_bits_used)', ('u2u32', ('i2i16', 'a@32')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_16_bits_used)', ('extract_u16', 'a@32', 0), 'b') => ('imul', 'a', 'b')
 *    ('iand(only_lower_16_bits_used)', ('u2u32', ('u2u16', 'a@32')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_16_bits_used)', ('u2u32', ('i2i16', 'a@32')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_16_bits_used)', ('extract_u16', 'a@32', 0), 'b') => ('iand', 'a', 'b')
 *    ('ior(only_lower_16_bits_used)', ('u2u32', ('u2u16', 'a@32')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_16_bits_used)', ('u2u32', ('i2i16', 'a@32')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_16_bits_used)', ('extract_u16', 'a@32', 0), 'b') => ('ior', 'a', 'b')
 *    ('ixor(only_lower_16_bits_used)', ('u2u32', ('u2u16', 'a@32')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_16_bits_used)', ('u2u32', ('i2i16', 'a@32')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_16_bits_used)', ('extract_u16', 'a@32', 0), 'b') => ('ixor', 'a', 'b')
 *    ('i2i32(only_lower_16_bits_used)', ('u2u16', 'a@32')) => a
 *    ('extract_i16(only_lower_16_bits_used)', 'a@32', 0) => a
 *    ('bcsel(only_lower_16_bits_used)', 'c', ('i2i32', ('u2u16', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_16_bits_used)', 'c', ('i2i32', ('i2i16', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_16_bits_used)', 'c', ('extract_i16', 'a@32', 0), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('iadd(only_lower_16_bits_used)', ('i2i32', ('u2u16', 'a@32')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_16_bits_used)', ('i2i32', ('i2i16', 'a@32')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_16_bits_used)', ('extract_i16', 'a@32', 0), 'b') => ('iadd', 'a', 'b')
 *    ('imul(only_lower_16_bits_used)', ('i2i32', ('u2u16', 'a@32')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_16_bits_used)', ('i2i32', ('i2i16', 'a@32')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_16_bits_used)', ('extract_i16', 'a@32', 0), 'b') => ('imul', 'a', 'b')
 *    ('iand(only_lower_16_bits_used)', ('i2i32', ('u2u16', 'a@32')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_16_bits_used)', ('i2i32', ('i2i16', 'a@32')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_16_bits_used)', ('extract_i16', 'a@32', 0), 'b') => ('iand', 'a', 'b')
 *    ('ior(only_lower_16_bits_used)', ('i2i32', ('u2u16', 'a@32')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_16_bits_used)', ('i2i32', ('i2i16', 'a@32')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_16_bits_used)', ('extract_i16', 'a@32', 0), 'b') => ('ior', 'a', 'b')
 *    ('ixor(only_lower_16_bits_used)', ('i2i32', ('u2u16', 'a@32')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_16_bits_used)', ('i2i32', ('i2i16', 'a@32')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_16_bits_used)', ('extract_i16', 'a@32', 0), 'b') => ('ixor', 'a', 'b')
 *    ('ldexp@16', 'x', 'exp') => ('fmul', ('fmul', 'x', ('i2i16', ('ishl', ('iadd', ('ishr', ('imin', ('imax', 'exp', -30), 30), 1), 15), 10))), ('i2i16', ('ishl', ('iadd', ('isub', ('imin', ('imax', 'exp', -30), 30), ('ishr', ('imin', ('imax', 'exp', -30), 30), 1)), 15), 10)))
 *    ('ldexp@32', 'x', 'exp') => ('fmul', ('fmul', 'x', ('ishl', ('iadd', ('ishr', ('imin', ('imax', 'exp', -254), 254), 1), 127), 23)), ('ishl', ('iadd', ('isub', ('imin', ('imax', 'exp', -254), 254), ('ishr', ('imin', ('imax', 'exp', -254), 254), 1)), 127), 23))
 *    ('ldexp@64', 'x', 'exp') => ('fmul', ('fmul', 'x', ('pack_64_2x32_split', 0, ('ishl', ('iadd', ('ishr', ('imin', ('imax', 'exp', -2046), 2046), 1), 1023), 20))), ('pack_64_2x32_split', 0, ('ishl', ('iadd', ('isub', ('imin', ('imax', 'exp', -2046), 2046), ('ishr', ('imin', ('imax', 'exp', -2046), 2046), 1)), 1023), 20)))
 *    ('ior(many-comm-expr)', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 252645135), 4), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 4042322160), 4)), 858993459), 2), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 252645135), 4), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 4042322160), 4)), 3435973836), 2)), 1431655765), 1), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 252645135), 4), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 4042322160), 4)), 858993459), 2), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 252645135), 4), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 4042322160), 4)), 3435973836), 2)), 2863311530), 1)) => ('bitfield_reverse', 'x')
 *    ('ior(many-comm-expr)', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 3435973836), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 858993459)), 4), 4042322160), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 3435973836), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 858993459)), 4), 252645135)), 8), 4278255360), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 3435973836), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 858993459)), 4), 4042322160), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 3435973836), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 858993459)), 4), 252645135)), 8), 16711935)) => ('bitfield_reverse', 'x')
 *    ('ball_iequal2', ('ieq', 'a', 'b'), -1) => ('ball_iequal2', 'a', 'b')
 *    ('ball_iequal2', ('feq', 'a', 'b'), -1) => ('ball_fequal2', 'a', 'b')
 *    ('bany_inequal2', ('ine', 'a', 'b'), 0) => ('bany_inequal2', 'a', 'b')
 *    ('bany_inequal2', ('fneu', 'a', 'b'), 0) => ('bany_fnequal2', 'a', 'b')
 *    ('ball_iequal3', ('ieq', 'a', 'b'), -1) => ('ball_iequal3', 'a', 'b')
 *    ('ball_iequal3', ('feq', 'a', 'b'), -1) => ('ball_fequal3', 'a', 'b')
 *    ('bany_inequal3', ('ine', 'a', 'b'), 0) => ('bany_inequal3', 'a', 'b')
 *    ('bany_inequal3', ('fneu', 'a', 'b'), 0) => ('bany_fnequal3', 'a', 'b')
 *    ('ball_iequal4', ('ieq', 'a', 'b'), -1) => ('ball_iequal4', 'a', 'b')
 *    ('ball_iequal4', ('feq', 'a', 'b'), -1) => ('ball_fequal4', 'a', 'b')
 *    ('bany_inequal4', ('ine', 'a', 'b'), 0) => ('bany_inequal4', 'a', 'b')
 *    ('bany_inequal4', ('fneu', 'a', 'b'), 0) => ('bany_fnequal4', 'a', 'b')
 *    ('ball_iequal8', ('ieq', 'a', 'b'), -1) => ('ball_iequal8', 'a', 'b')
 *    ('ball_iequal8', ('feq', 'a', 'b'), -1) => ('ball_fequal8', 'a', 'b')
 *    ('bany_inequal8', ('ine', 'a', 'b'), 0) => ('bany_inequal8', 'a', 'b')
 *    ('bany_inequal8', ('fneu', 'a', 'b'), 0) => ('bany_fnequal8', 'a', 'b')
 *    ('ball_iequal16', ('ieq', 'a', 'b'), -1) => ('ball_iequal16', 'a', 'b')
 *    ('ball_iequal16', ('feq', 'a', 'b'), -1) => ('ball_fequal16', 'a', 'b')
 *    ('bany_inequal16', ('ine', 'a', 'b'), 0) => ('bany_inequal16', 'a', 'b')
 *    ('bany_inequal16', ('fneu', 'a', 'b'), 0) => ('bany_fnequal16', 'a', 'b')
 *    ('iand', ('feq', 'a', 'a'), ('flt', 'a', 'b')) => ('!flt', 'a', 'b')
 *    ('iand', ('feq', 'a', 'a'), ('flt', 'b', 'a')) => ('!flt', 'b', 'a')
 *    ('iand', ('feq', 'a', 'a'), ('fge', 'a', 'b')) => ('!fge', 'a', 'b')
 *    ('iand', ('feq', 'a', 'a'), ('fge', 'b', 'a')) => ('!fge', 'b', 'a')
 *    ('iand', ('feq', 'a', 'a'), ('feq', 'a', 'b')) => ('!feq', 'a', 'b')
 *    ('iand', ('feq', 'a', 'a'), ('feq', 'b', 'a')) => ('!feq', 'b', 'a')
 *    ('feq', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('feq', 'b', 'd'), ('feq', 'c', 'd'))
 *    ('fneu', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('fneu', 'b', 'd'), ('fneu', 'c', 'd'))
 *    ('ieq', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('ieq', 'b', 'd'), ('ieq', 'c', 'd'))
 *    ('ine', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('ine', 'b', 'd'), ('ine', 'c', 'd'))
 *    ('flt', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('flt', 'b', 'd'), ('flt', 'c', 'd'))
 *    ('flt', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('flt', 'd', 'b'), ('flt', 'd', 'c'))
 *    ('fge', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('fge', 'b', 'd'), ('fge', 'c', 'd'))
 *    ('fge', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('fge', 'd', 'b'), ('fge', 'd', 'c'))
 *    ('ilt', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('ilt', 'b', 'd'), ('ilt', 'c', 'd'))
 *    ('ilt', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('ilt', 'd', 'b'), ('ilt', 'd', 'c'))
 *    ('ige', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('ige', 'b', 'd'), ('ige', 'c', 'd'))
 *    ('ige', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('ige', 'd', 'b'), ('ige', 'd', 'c'))
 *    ('ult', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('ult', 'b', 'd'), ('ult', 'c', 'd'))
 *    ('ult', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('ult', 'd', 'b'), ('ult', 'd', 'c'))
 *    ('uge', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('uge', 'b', 'd'), ('uge', 'c', 'd'))
 *    ('uge', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('uge', 'd', 'b'), ('uge', 'd', 'c'))
 *    ('fadd', ('bcsel(is_used_once)', 'a', '#b', 'c'), '#d') => ('bcsel', 'a', ('fadd', 'b', 'd'), ('fadd', 'c', 'd'))
 *    ('fmul', ('bcsel(is_used_once)', 'a', '#b', 'c'), '#d') => ('bcsel', 'a', ('fmul', 'b', 'd'), ('fmul', 'c', 'd'))
 *    ('fmulz', ('bcsel(is_used_once)', 'a', '#b', 'c'), '#d') => ('bcsel', 'a', ('fmulz', 'b', 'd'), ('fmulz', 'c', 'd'))
 *    ('iadd', ('bcsel(is_used_once)', 'a', '#b', 'c'), '#d') => ('bcsel', 'a', ('iadd', 'b', 'd'), ('iadd', 'c', 'd'))
 *    ('imul', ('bcsel(is_used_once)', 'a', '#b', 'c'), '#d') => ('bcsel', 'a', ('imul', 'b', 'd'), ('imul', 'c', 'd'))
 *    ('fddx', 'a') => 0.0
 *    ('fddx_fine', 'a') => 0.0
 *    ('fddx_coarse', 'a') => 0.0
 *    ('fddy', 'a') => 0.0
 *    ('fddy_fine', 'a') => 0.0
 *    ('fddy_coarse', 'a') => 0.0
 *    ('umul_low', '#a(is_lower_half_zero)', 'b') => 0
 *    ('imadsh_mix16', '#a@32(is_lower_half_zero)', 'b@32', 'c@32') => c
 *    ('imadsh_mix16', 'a@32', '#b@32(is_upper_half_zero)', 'c@32') => c
 *    ('bcsel', 'a', ('flrp(is_used_once)', 'b', 'c', 'd'), ('flrp', 'b', 'c', 'e')) => ('flrp', 'b', 'c', ('bcsel', 'a', 'd', 'e'))
 *    ('bcsel', 'a', ('flrp', 'b', 'c', 'd'), ('flrp(is_used_once)', 'b', 'c', 'e')) => ('flrp', 'b', 'c', ('bcsel', 'a', 'd', 'e'))
 *    ('bcsel', 'a', ('flrp(is_used_once)', 'b', 'c', 'd'), ('flrp', 'b', 'e', 'd')) => ('flrp', 'b', ('bcsel', 'a', 'c', 'e'), 'd')
 *    ('bcsel', 'a', ('flrp', 'b', 'c', 'd'), ('flrp(is_used_once)', 'b', 'e', 'd')) => ('flrp', 'b', ('bcsel', 'a', 'c', 'e'), 'd')
 *    ('bcsel', 'a', ('flrp(is_used_once)', 'b', 'c', 'd'), ('flrp', 'e', 'c', 'd')) => ('flrp', ('bcsel', 'a', 'b', 'e'), 'c', 'd')
 *    ('bcsel', 'a', ('flrp', 'b', 'c', 'd'), ('flrp(is_used_once)', 'e', 'c', 'd')) => ('flrp', ('bcsel', 'a', 'b', 'e'), 'c', 'd')
 *    ('bcsel', 'a', ('fmulz(is_used_once)', 'b', 'c'), ('fmulz', 'b', 'd(is_not_const)')) => ('fmulz', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmulz(is_used_once)', 'b', 'c(is_not_const)'), ('fmulz', 'b', 'd')) => ('fmulz', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmulz', 'b', 'c(is_not_const)'), ('fmulz(is_used_once)', 'b', 'd')) => ('fmulz', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmulz', 'b', 'c'), ('fmulz(is_used_once)', 'b', 'd(is_not_const)')) => ('fmulz', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmul(is_used_once)', 'b', 'c'), ('fmul', 'b', 'd(is_not_const)')) => ('fmul', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmul(is_used_once)', 'b', 'c(is_not_const)'), ('fmul', 'b', 'd')) => ('fmul', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmul', 'b', 'c(is_not_const)'), ('fmul(is_used_once)', 'b', 'd')) => ('fmul', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmul', 'b', 'c'), ('fmul(is_used_once)', 'b', 'd(is_not_const)')) => ('fmul', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('iadd(is_used_once)', 'b', 'c'), ('iadd', 'b', 'd(is_not_const)')) => ('iadd', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('iadd(is_used_once)', 'b', 'c(is_not_const)'), ('iadd', 'b', 'd')) => ('iadd', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('iadd', 'b', 'c(is_not_const)'), ('iadd(is_used_once)', 'b', 'd')) => ('iadd', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('iadd', 'b', 'c'), ('iadd(is_used_once)', 'b', 'd(is_not_const)')) => ('iadd', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imul(is_used_once)', 'b', 'c'), ('imul', 'b', 'd(is_not_const)')) => ('imul', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imul(is_used_once)', 'b', 'c(is_not_const)'), ('imul', 'b', 'd')) => ('imul', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imul', 'b', 'c(is_not_const)'), ('imul(is_used_once)', 'b', 'd')) => ('imul', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imul', 'b', 'c'), ('imul(is_used_once)', 'b', 'd(is_not_const)')) => ('imul', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('iand(is_used_once)', 'b', 'c'), ('iand', 'b', 'd(is_not_const)')) => ('iand', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('iand(is_used_once)', 'b', 'c(is_not_const)'), ('iand', 'b', 'd')) => ('iand', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('iand', 'b', 'c(is_not_const)'), ('iand(is_used_once)', 'b', 'd')) => ('iand', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('iand', 'b', 'c'), ('iand(is_used_once)', 'b', 'd(is_not_const)')) => ('iand', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('ior(is_used_once)', 'b', 'c'), ('ior', 'b', 'd(is_not_const)')) => ('ior', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('ior(is_used_once)', 'b', 'c(is_not_const)'), ('ior', 'b', 'd')) => ('ior', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('ior', 'b', 'c(is_not_const)'), ('ior(is_used_once)', 'b', 'd')) => ('ior', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('ior', 'b', 'c'), ('ior(is_used_once)', 'b', 'd(is_not_const)')) => ('ior', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('ixor(is_used_once)', 'b', 'c'), ('ixor', 'b', 'd(is_not_const)')) => ('ixor', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('ixor(is_used_once)', 'b', 'c(is_not_const)'), ('ixor', 'b', 'd')) => ('ixor', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('ixor', 'b', 'c(is_not_const)'), ('ixor(is_used_once)', 'b', 'd')) => ('ixor', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('ixor', 'b', 'c'), ('ixor(is_used_once)', 'b', 'd(is_not_const)')) => ('ixor', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmin(is_used_once)', 'b', 'c'), ('fmin', 'b', 'd(is_not_const)')) => ('fmin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmin(is_used_once)', 'b', 'c(is_not_const)'), ('fmin', 'b', 'd')) => ('fmin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmin', 'b', 'c(is_not_const)'), ('fmin(is_used_once)', 'b', 'd')) => ('fmin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmin', 'b', 'c'), ('fmin(is_used_once)', 'b', 'd(is_not_const)')) => ('fmin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmax(is_used_once)', 'b', 'c'), ('fmax', 'b', 'd(is_not_const)')) => ('fmax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmax(is_used_once)', 'b', 'c(is_not_const)'), ('fmax', 'b', 'd')) => ('fmax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmax', 'b', 'c(is_not_const)'), ('fmax(is_used_once)', 'b', 'd')) => ('fmax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmax', 'b', 'c'), ('fmax(is_used_once)', 'b', 'd(is_not_const)')) => ('fmax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imin(is_used_once)', 'b', 'c'), ('imin', 'b', 'd(is_not_const)')) => ('imin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imin(is_used_once)', 'b', 'c(is_not_const)'), ('imin', 'b', 'd')) => ('imin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imin', 'b', 'c(is_not_const)'), ('imin(is_used_once)', 'b', 'd')) => ('imin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imin', 'b', 'c'), ('imin(is_used_once)', 'b', 'd(is_not_const)')) => ('imin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imax(is_used_once)', 'b', 'c'), ('imax', 'b', 'd(is_not_const)')) => ('imax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imax(is_used_once)', 'b', 'c(is_not_const)'), ('imax', 'b', 'd')) => ('imax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imax', 'b', 'c(is_not_const)'), ('imax(is_used_once)', 'b', 'd')) => ('imax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imax', 'b', 'c'), ('imax(is_used_once)', 'b', 'd(is_not_const)')) => ('imax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('umin(is_used_once)', 'b', 'c'), ('umin', 'b', 'd(is_not_const)')) => ('umin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('umin(is_used_once)', 'b', 'c(is_not_const)'), ('umin', 'b', 'd')) => ('umin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('umin', 'b', 'c(is_not_const)'), ('umin(is_used_once)', 'b', 'd')) => ('umin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('umin', 'b', 'c'), ('umin(is_used_once)', 'b', 'd(is_not_const)')) => ('umin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('umax(is_used_once)', 'b', 'c'), ('umax', 'b', 'd(is_not_const)')) => ('umax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('umax(is_used_once)', 'b', 'c(is_not_const)'), ('umax', 'b', 'd')) => ('umax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('umax', 'b', 'c(is_not_const)'), ('umax(is_used_once)', 'b', 'd')) => ('umax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('umax', 'b', 'c'), ('umax(is_used_once)', 'b', 'd(is_not_const)')) => ('umax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fpow(is_used_once)', 'b', 'c'), ('fpow', 'b', 'd')) => ('fpow', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fpow', 'b', 'c'), ('fpow(is_used_once)', 'b', 'd')) => ('fpow', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fpow(is_used_once)', 'b', 'c'), ('fpow', 'd', 'c')) => ('fpow', ('bcsel', 'a', 'b', 'd'), 'c')
 *    ('bcsel', 'a', ('fpow', 'b', 'c'), ('fpow(is_used_once)', 'd', 'c')) => ('fpow', ('bcsel', 'a', 'b', 'd'), 'c')
 *    ('bcsel', 'c', ('frcp(is_used_once)', 'a'), ('frcp(is_used_once)', 'b')) => ('frcp', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('frsq(is_used_once)', 'a'), ('frsq(is_used_once)', 'b')) => ('frsq', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('fsqrt(is_used_once)', 'a'), ('fsqrt(is_used_once)', 'b')) => ('fsqrt', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('fexp2(is_used_once)', 'a'), ('fexp2(is_used_once)', 'b')) => ('fexp2', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('flog2(is_used_once)', 'a'), ('flog2(is_used_once)', 'b')) => ('flog2', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('fsign(is_used_once)', 'a'), ('fsign(is_used_once)', 'b')) => ('fsign', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('fsin(is_used_once)', 'a'), ('fsin(is_used_once)', 'b')) => ('fsin', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('fcos(is_used_once)', 'a'), ('fcos(is_used_once)', 'b')) => ('fcos', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('fsin_amd(is_used_once)', 'a'), ('fsin_amd(is_used_once)', 'b')) => ('fsin_amd', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('fcos_amd(is_used_once)', 'a'), ('fcos_amd(is_used_once)', 'b')) => ('fcos_amd', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('fneg(is_used_once)', 'a'), ('fneg(is_used_once)', 'b')) => ('fneg', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('fabs(is_used_once)', 'a'), ('fabs(is_used_once)', 'b')) => ('fabs', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('fsign(is_used_once)', 'a'), ('fsign(is_used_once)', 'b')) => ('fsign', ('bcsel', 'c', 'a', 'b'))
 *    ('ineg', ('bcsel', 'c', '#a', '#b')) => ('bcsel', 'c', ('ineg', 'a'), ('ineg', 'b'))
 *    ('iabs', ('bcsel', 'c', '#a', '#b')) => ('bcsel', 'c', ('iabs', 'a'), ('iabs', 'b'))
 *    ('inot', ('bcsel', 'c', '#a', '#b')) => ('bcsel', 'c', ('inot', 'a'), ('inot', 'b'))
 *    ('isign', ('bcsel', 'c', '#a', '#b')) => ('bcsel', 'c', ('isign', 'a'), ('isign', 'b'))
 *    ('fisnormal', 'a@16') => ('ult', 4095, ('iadd', ('ishl', 'a', 1), 2048))
 *    ('fisnormal', 'a@32') => ('ult', 33554431, ('iadd', ('ishl', 'a', 1), 16777216))
 *    ('fisnormal', 'a@64') => ('ult', 18014398509481983, ('iadd', ('ishl', 'a', 1), 9007199254740992))
 *    ('f2f16', ('vec2', 'a@32', 'b@32')) => ('vec2', ('f2f16', 'a'), ('f2f16', 'b'))
 *    ('f2fmp', ('vec2', 'a@32', 'b@32')) => ('vec2', ('f2fmp', 'a'), ('f2fmp', 'b'))
 *    ('u2u16', ('vec2', 'a@32', 'b@32')) => ('vec2', ('u2u16', 'a'), ('u2u16', 'b'))
 *    ('i2i16', ('vec2', 'a@32', 'b@32')) => ('vec2', ('i2i16', 'a'), ('i2i16', 'b'))
 *    ('i2imp', ('vec2', 'a@32', 'b@32')) => ('vec2', ('i2imp', 'a'), ('i2imp', 'b'))
 *    ('f2f16', ('vec3', 'a@32', 'b@32', 'c@32')) => ('vec3', ('f2f16', 'a'), ('f2f16', 'b'), ('f2f16', 'c'))
 *    ('f2fmp', ('vec3', 'a@32', 'b@32', 'c@32')) => ('vec3', ('f2fmp', 'a'), ('f2fmp', 'b'), ('f2fmp', 'c'))
 *    ('u2u16', ('vec3', 'a@32', 'b@32', 'c@32')) => ('vec3', ('u2u16', 'a'), ('u2u16', 'b'), ('u2u16', 'c'))
 *    ('i2i16', ('vec3', 'a@32', 'b@32', 'c@32')) => ('vec3', ('i2i16', 'a'), ('i2i16', 'b'), ('i2i16', 'c'))
 *    ('i2imp', ('vec3', 'a@32', 'b@32', 'c@32')) => ('vec3', ('i2imp', 'a'), ('i2imp', 'b'), ('i2imp', 'c'))
 *    ('f2f16', ('vec4', 'a@32', 'b@32', 'c@32', 'd@32')) => ('vec4', ('f2f16', 'a'), ('f2f16', 'b'), ('f2f16', 'c'), ('f2f16', 'd'))
 *    ('f2fmp', ('vec4', 'a@32', 'b@32', 'c@32', 'd@32')) => ('vec4', ('f2fmp', 'a'), ('f2fmp', 'b'), ('f2fmp', 'c'), ('f2fmp', 'd'))
 *    ('u2u16', ('vec4', 'a@32', 'b@32', 'c@32', 'd@32')) => ('vec4', ('u2u16', 'a'), ('u2u16', 'b'), ('u2u16', 'c'), ('u2u16', 'd'))
 *    ('i2i16', ('vec4', 'a@32', 'b@32', 'c@32', 'd@32')) => ('vec4', ('i2i16', 'a'), ('i2i16', 'b'), ('i2i16', 'c'), ('i2i16', 'd'))
 *    ('i2imp', ('vec4', 'a@32', 'b@32', 'c@32', 'd@32')) => ('vec4', ('i2imp', 'a'), ('i2imp', 'b'), ('i2imp', 'c'), ('i2imp', 'd'))
 */


static const nir_search_value_union nir_opt_algebraic_values[] = {
   /* ('imul', 'a', '#b(is_pos_power_of_two)') => ('ishl', 'a', ('find_lsb', 'b')) */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      true,
      nir_type_invalid,
      0,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 0, 1 },
      -1,
   } },

   /* replace0_0 -> 0 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_find_lsb,
      -1, 0,
      { 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 0, 4 },
      -1,
   } },

   /* ('imul', 'a@8', 128) => ('ishl', 'a', 7) */
   { .variable = {
      { nir_search_value_variable, 8 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .constant = {
      { nir_search_value_constant, 8 },
      nir_type_int, { 0x80 /* 128 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 6, 7 },
      -1,
   } },

   /* replace1_0 -> 6 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x7 /* 7 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 6, 9 },
      -1,
   } },

   /* ('imul', 'a@16', 32768) => ('ishl', 'a', 15) */
   { .variable = {
      { nir_search_value_variable, 16 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { 0x8000 /* 32768 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 11, 12 },
      -1,
   } },

   /* replace2_0 -> 11 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xf /* 15 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 11, 14 },
      -1,
   } },

   /* ('imul', 'a@32', 2147483648) => ('ishl', 'a', 31) */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x80000000 /* 2147483648 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 16, 17 },
      -1,
   } },

   /* replace3_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x1f /* 31 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 19 },
      -1,
   } },

   /* ('imul', 'a@64', 9223372036854775808) => ('ishl', 'a', 63) */
   { .variable = {
      { nir_search_value_variable, 64 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x8000000000000000 /* 9223372036854775808 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 21, 22 },
      -1,
   } },

   /* replace4_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x3f /* 63 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 24 },
      -1,
   } },

   /* ('imul', 'a', '#b(is_neg_power_of_two)') => ('ineg', ('ishl', 'a', ('find_lsb', ('iabs', 'b')))) */
   /* search5_0 -> 0 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      true,
      nir_type_invalid,
      1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 0, 26 },
      -1,
   } },

   /* replace5_0_0 -> 0 in the cache */
   /* replace5_0_1_0_0 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_find_lsb,
      -1, 0,
      { 28 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 0, 29 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 30 },
      -1,
   } },

   /* ('ishl', 'a', '#b') => ('imul', 'a', ('ishl', 1, 'b')) */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 32, 33 },
      -1,
   } },

   /* replace6_0 -> 32 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_int, { 0x1 /* 1 */ },
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 35, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 32, 37 },
      -1,
   } },

   /* ('imul@64', 'a', '#b(is_bitcount2)') => ('iadd', ('ishl', 'a', ('ufind_msb', 'b')), ('ishl', 'a', ('find_lsb', 'b'))) */
   /* search7_0 -> 21 in the cache */
   { .variable = {
      { nir_search_value_variable, 64 },
      1, /* b */
      true,
      nir_type_invalid,
      2,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 21, 39 },
      -1,
   } },

   /* replace7_0_0 -> 21 in the cache */
   { .variable = {
      { nir_search_value_variable, 64 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ufind_msb,
      -1, 0,
      { 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 42 },
      -1,
   } },
   /* replace7_1_0 -> 21 in the cache */
   /* replace7_1_1_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_find_lsb,
      -1, 0,
      { 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 44 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 43, 45 },
      -1,
   } },

   /* ('unpack_64_2x32_split_x', ('imul_2x32_64(is_used_once)', 'a', 'b')) => ('imul', 'a', 'b') */
   /* search8_0_0 -> 16 in the cache */
   /* search8_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_imul_2x32_64,
      0, 1,
      { 16, 36 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_x,
      -1, 1,
      { 47 },
      -1,
   } },

   /* replace8_0 -> 16 in the cache */
   /* replace8_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* ('unpack_64_2x32_split_x', ('umul_2x32_64(is_used_once)', 'a', 'b')) => ('imul', 'a', 'b') */
   /* search9_0_0 -> 16 in the cache */
   /* search9_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_umul_2x32_64,
      0, 1,
      { 16, 36 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_x,
      -1, 1,
      { 50 },
      -1,
   } },

   /* replace9_0 -> 16 in the cache */
   /* replace9_1 -> 36 in the cache */
   /* replace9 -> 49 in the cache */

   /* ('imul_2x32_64', 'a', 'b') => ('pack_64_2x32_split', ('imul', 'a', 'b'), ('imul_high', 'a', 'b')) */
   /* search10_0 -> 16 in the cache */
   /* search10_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_imul_2x32_64,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace10_0_0 -> 16 in the cache */
   /* replace10_0_1 -> 36 in the cache */
   /* replace10_0 -> 49 in the cache */
   /* replace10_1_0 -> 16 in the cache */
   /* replace10_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul_high,
      1, 1,
      { 16, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 2,
      { 49, 53 },
      -1,
   } },

   /* ('umul_2x32_64', 'a', 'b') => ('pack_64_2x32_split', ('imul', 'a', 'b'), ('umul_high', 'a', 'b')) */
   /* search11_0 -> 16 in the cache */
   /* search11_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_umul_2x32_64,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace11_0_0 -> 16 in the cache */
   /* replace11_0_1 -> 36 in the cache */
   /* replace11_0 -> 49 in the cache */
   /* replace11_1_0 -> 16 in the cache */
   /* replace11_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umul_high,
      1, 1,
      { 16, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 2,
      { 49, 56 },
      -1,
   } },

   /* ('udiv', 'a', 1) => a */
   /* search12_0 -> 32 in the cache */
   /* search12_1 -> 35 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_udiv,
      -1, 0,
      { 32, 35 },
      -1,
   } },

   /* replace12 -> 32 in the cache */

   /* ('idiv', 'a', 1) => a */
   /* search13_0 -> 32 in the cache */
   /* search13_1 -> 35 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_idiv,
      -1, 0,
      { 32, 35 },
      -1,
   } },

   /* replace13 -> 32 in the cache */

   /* ('umod', 'a', 1) => 0 */
   /* search14_0 -> 32 in the cache */
   /* search14_1 -> 35 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_umod,
      -1, 0,
      { 32, 35 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_int, { 0x0 /* 0 */ },
   } },

   /* ('imod', 'a', 1) => 0 */
   /* search15_0 -> 32 in the cache */
   /* search15_1 -> 35 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imod,
      -1, 0,
      { 32, 35 },
      -1,
   } },

   /* replace15 -> 61 in the cache */

   /* ('imod', 'a', -1) => 0 */
   /* search16_0 -> 32 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_int, { -0x1 /* -1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imod,
      -1, 0,
      { 32, 63 },
      -1,
   } },

   /* replace16 -> 61 in the cache */

   /* ('irem', 'a', 1) => 0 */
   /* search17_0 -> 32 in the cache */
   /* search17_1 -> 35 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_irem,
      -1, 0,
      { 32, 35 },
      -1,
   } },

   /* replace17 -> 61 in the cache */

   /* ('irem', 'a', -1) => 0 */
   /* search18_0 -> 32 in the cache */
   /* search18_1 -> 63 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_irem,
      -1, 0,
      { 32, 63 },
      -1,
   } },

   /* replace18 -> 61 in the cache */

   /* ('udiv', 'a', '#b(is_pos_power_of_two)') => ('ushr', 'a', ('find_lsb', 'b')) */
   /* search19_0 -> 0 in the cache */
   /* search19_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_udiv,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* replace19_0 -> 0 in the cache */
   /* replace19_1_0 -> 3 in the cache */
   /* replace19_1 -> 4 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 0, 4 },
      -1,
   } },

   /* ('idiv', 'a', '#b(is_pos_power_of_two)') => ('imul', ('isign', 'a'), ('ushr', ('iabs', 'a'), ('find_lsb', 'b'))) */
   /* search20_0 -> 0 in the cache */
   /* search20_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_idiv,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* replace20_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_isign,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace20_1_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace20_1_1_0 -> 3 in the cache */
   /* replace20_1_1 -> 4 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 71, 4 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 70, 72 },
      -1,
   } },

   /* ('idiv', 'a', '#b(is_neg_power_of_two)') => ('ineg', ('imul', ('isign', 'a'), ('ushr', ('iabs', 'a'), ('find_lsb', ('iabs', 'b'))))) */
   /* search21_0 -> 0 in the cache */
   /* search21_1 -> 26 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_idiv,
      -1, 0,
      { 0, 26 },
      -1,
   } },

   /* replace21_0_0_0 -> 0 in the cache */
   /* replace21_0_0 -> 70 in the cache */
   /* replace21_0_1_0_0 -> 0 in the cache */
   /* replace21_0_1_0 -> 71 in the cache */
   /* replace21_0_1_1_0_0 -> 3 in the cache */
   /* replace21_0_1_1_0 -> 28 in the cache */
   /* replace21_0_1_1 -> 29 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 71, 29 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 70, 75 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 1,
      { 76 },
      -1,
   } },

   /* ('umod', 'a', '#b(is_pos_power_of_two)') => ('iand', 'a', ('isub', 'b', 1)) */
   /* search22_0 -> 0 in the cache */
   /* search22_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umod,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* replace22_0 -> 0 in the cache */
   /* replace22_1_0 -> 3 in the cache */
   { .constant = {
      { nir_search_value_constant, -2 },
      nir_type_int, { 0x1 /* 1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 3, 79 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 0, 80 },
      -1,
   } },

   /* ('imod', 'a', '#b(is_pos_power_of_two)') => ('iand', 'a', ('isub', 'b', 1)) */
   /* search23_0 -> 0 in the cache */
   /* search23_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imod,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* replace23_0 -> 0 in the cache */
   /* replace23_1_0 -> 3 in the cache */
   /* replace23_1_1 -> 79 in the cache */
   /* replace23_1 -> 80 in the cache */
   /* replace23 -> 81 in the cache */

   /* ('imod', 'a', '#b(is_neg_power_of_two)') => ('bcsel', ('ieq', ('ior', 'a', 'b'), 'b'), 0, ('ior', 'a', 'b')) */
   /* search24_0 -> 0 in the cache */
   /* search24_1 -> 26 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imod,
      -1, 0,
      { 0, 26 },
      -1,
   } },

   /* replace24_0_0_0 -> 0 in the cache */
   /* replace24_0_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* replace24_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 84, 3 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, -2 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   /* replace24_2_0 -> 0 in the cache */
   /* replace24_2_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      2, 1,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 3,
      { 85, 86, 87 },
      -1,
   } },

   /* ('irem', 'a', '#b(is_pos_power_of_two)') => ('isub', 'a', ('iand', ('bcsel', ('ilt', 'a', 0), ('iadd', 'a', ('isub', 'b', 1)), 'a'), ('ineg', 'b'))) */
   /* search25_0 -> 0 in the cache */
   /* search25_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_irem,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* replace25_0 -> 0 in the cache */
   /* replace25_1_0_0_0 -> 0 in the cache */
   /* replace25_1_0_0_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 0, 86 },
      -1,
   } },
   /* replace25_1_0_1_0 -> 0 in the cache */
   /* replace25_1_0_1_1_0 -> 3 in the cache */
   /* replace25_1_0_1_1_1 -> 79 in the cache */
   /* replace25_1_0_1_1 -> 80 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 0, 80 },
      -1,
   } },
   /* replace25_1_0_2 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 90, 91, 0 },
      -1,
   } },
   /* replace25_1_1_0 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 92, 93 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 2,
      { 0, 94 },
      -1,
   } },

   /* ('irem', 'a', '#b(is_neg_power_of_two)') => ('irem', 'a', ('iabs', 'b')) */
   /* search26_0 -> 0 in the cache */
   /* search26_1 -> 26 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_irem,
      -1, 0,
      { 0, 26 },
      -1,
   } },

   /* replace26_0 -> 0 in the cache */
   /* replace26_1_0 -> 3 in the cache */
   /* replace26_1 -> 28 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_irem,
      -1, 0,
      { 0, 28 },
      -1,
   } },

   /* ('~fneg', ('fneg', 'a')) => a */
   /* search27_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 98 },
      -1,
   } },

   /* replace27 -> 32 in the cache */

   /* ('ineg', ('ineg', 'a')) => a */
   /* search28_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 100 },
      -1,
   } },

   /* replace28 -> 32 in the cache */

   /* ('fabs', ('fneg', 'a')) => ('fabs', 'a') */
   /* search29_0_0 -> 32 in the cache */
   /* search29_0 -> 98 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 98 },
      -1,
   } },

   /* replace29_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 32 },
      -1,
   } },

   /* ('fabs', ('u2f', 'a')) => ('u2f', 'a') */
   /* search30_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_u2f,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 104 },
      -1,
   } },

   /* replace30_0 -> 32 in the cache */
   /* replace30 -> 104 in the cache */

   /* ('iabs', ('iabs', 'a')) => ('iabs', 'a') */
   /* search31_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 106 },
      -1,
   } },

   /* replace31_0 -> 32 in the cache */
   /* replace31 -> 106 in the cache */

   /* ('iabs', ('ineg', 'a')) => ('iabs', 'a') */
   /* search32_0_0 -> 32 in the cache */
   /* search32_0 -> 100 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 100 },
      -1,
   } },

   /* replace32_0 -> 32 in the cache */
   /* replace32 -> 106 in the cache */

   /* ('f2b', ('fneg', 'a')) => ('f2b', 'a') */
   /* search33_0_0 -> 32 in the cache */
   /* search33_0 -> 98 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2b,
      -1, 0,
      { 98 },
      -1,
   } },

   /* replace33_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2b,
      -1, 0,
      { 32 },
      -1,
   } },

   /* ('~fadd', 'a', 0.0) => a */
   /* search34_0 -> 32 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 32, 111 },
      -1,
   } },

   /* replace34 -> 32 in the cache */

   /* ('fadd(is_only_used_as_float)', 'a@16', 0.0) => a */
   /* search35_0 -> 11 in the cache */
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 11, 113 },
      1,
   } },

   /* replace35 -> 11 in the cache */

   /* ('fadd(is_only_used_as_float)', 'a@32', 0.0) => a */
   /* search36_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 16, 115 },
      1,
   } },

   /* replace36 -> 16 in the cache */

   /* ('iadd', 'a', 0) => a */
   /* search37_0 -> 32 in the cache */
   /* search37_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 32, 61 },
      -1,
   } },

   /* replace37 -> 32 in the cache */

   /* ('iadd_sat', 'a', 0) => a */
   /* search38_0 -> 32 in the cache */
   /* search38_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iadd_sat,
      0, 1,
      { 32, 61 },
      -1,
   } },

   /* replace38 -> 32 in the cache */

   /* ('isub_sat', 'a', 0) => a */
   /* search39_0 -> 32 in the cache */
   /* search39_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_isub_sat,
      -1, 0,
      { 32, 61 },
      -1,
   } },

   /* replace39 -> 32 in the cache */

   /* ('uadd_sat', 'a', 0) => a */
   /* search40_0 -> 32 in the cache */
   /* search40_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_uadd_sat,
      0, 1,
      { 32, 61 },
      -1,
   } },

   /* replace40 -> 32 in the cache */

   /* ('usub_sat', 'a', 0) => a */
   /* search41_0 -> 32 in the cache */
   /* search41_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_usub_sat,
      -1, 0,
      { 32, 61 },
      -1,
   } },

   /* replace41 -> 32 in the cache */

   /* ('usadd_4x8_vc4', 'a', 0) => a */
   /* search42_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_usadd_4x8_vc4,
      0, 1,
      { 16, 122 },
      -1,
   } },

   /* replace42 -> 16 in the cache */

   /* ('usadd_4x8_vc4', 'a', -1) => -1 */
   /* search43_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x1 /* -1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_usadd_4x8_vc4,
      0, 1,
      { 16, 124 },
      -1,
   } },

   /* replace43 -> 124 in the cache */

   /* ('~fadd', ('fmul', 'a', 'b'), ('fmul', 'a', 'c')) => ('fmul', 'a', ('fadd', 'b', 'c')) */
   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 126, 127 },
      -1,
   } },
   /* search44_1_0 -> 126 in the cache */
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 126, 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 128, 130 },
      -1,
   } },

   /* replace44_0 -> 126 in the cache */
   /* replace44_1_0 -> 127 in the cache */
   /* replace44_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 127, 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 126, 132 },
      -1,
   } },

   /* ('~fadd', ('fmulz', 'a', 'b'), ('fmulz', 'a', 'c')) => ('fmulz', 'a', ('fadd', 'b', 'c')) */
   /* search45_0_0 -> 16 in the cache */
   /* search45_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 16, 36 },
      -1,
   } },
   /* search45_1_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      2, 1,
      { 16, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 134, 136 },
      -1,
   } },

   /* replace45_0 -> 16 in the cache */
   /* replace45_1_0 -> 36 in the cache */
   /* replace45_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 36, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 2,
      { 16, 138 },
      -1,
   } },

   /* ('~ffma', 'a', 'b', ('ffma(is_used_once)', 'a', 'c', 'd')) => ('ffma', 'a', ('fadd', 'b', 'c'), 'd') */
   /* search46_0 -> 0 in the cache */
   /* search46_1 -> 3 in the cache */
   /* search46_2_0 -> 0 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 0, 140, 141 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 0, 3, 142 },
      -1,
   } },

   /* replace46_0 -> 0 in the cache */
   /* replace46_1_0 -> 3 in the cache */
   /* replace46_1_1 -> 140 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 3, 140 },
      -1,
   } },
   /* replace46_2 -> 141 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 0, 144, 141 },
      -1,
   } },

   /* ('~ffma', 'a', 'b', ('fmul(is_used_once)', 'a', 'c')) => ('fmul', 'a', ('fadd', 'b', 'c')) */
   /* search47_0 -> 0 in the cache */
   /* search47_1 -> 3 in the cache */
   /* search47_2_0 -> 0 in the cache */
   /* search47_2_1 -> 140 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 0, 140 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 0, 3, 146 },
      -1,
   } },

   /* replace47_0 -> 0 in the cache */
   /* replace47_1_0 -> 3 in the cache */
   /* replace47_1_1 -> 140 in the cache */
   /* replace47_1 -> 144 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 0, 144 },
      -1,
   } },

   /* ('~fadd', ('fmul(is_used_once)', 'a', 'b'), ('ffma(is_used_once)', 'a', 'c', 'd')) => ('ffma', 'a', ('fadd', 'b', 'c'), 'd') */
   { .variable = {
      { nir_search_value_variable, -4 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 149, 150 },
      0,
   } },
   /* search48_1_0 -> 149 in the cache */
   { .variable = {
      { nir_search_value_variable, -4 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ffma,
      2, 1,
      { 149, 152, 153 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 151, 154 },
      -1,
   } },

   /* replace48_0 -> 149 in the cache */
   /* replace48_1_0 -> 150 in the cache */
   /* replace48_1_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 150, 152 },
      -1,
   } },
   /* replace48_2 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 149, 156, 153 },
      -1,
   } },

   /* ('~ffma', 'a', ('fmul(is_used_once)', 'b', 'c'), ('fmul(is_used_once)', 'b', 'd')) => ('fmul', 'b', ('ffma', 'a', 'c', 'd')) */
   /* search49_0 -> 149 in the cache */
   /* search49_1_0 -> 150 in the cache */
   /* search49_1_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 150, 152 },
      0,
   } },
   /* search49_2_0 -> 150 in the cache */
   /* search49_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 150, 153 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 3,
      { 149, 158, 159 },
      -1,
   } },

   /* replace49_0 -> 150 in the cache */
   /* replace49_1_0 -> 149 in the cache */
   /* replace49_1_1 -> 152 in the cache */
   /* replace49_1_2 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 149, 152, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 150, 161 },
      -1,
   } },

   /* ('~ffmaz', 'a', 'b', ('ffmaz(is_used_once)', 'a', 'c', 'd')) => ('ffmaz', 'a', ('fadd', 'b', 'c'), 'd') */
   /* search50_0 -> 16 in the cache */
   /* search50_1 -> 36 in the cache */
   /* search50_2_0 -> 16 in the cache */
   /* search50_2_1 -> 135 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 1,
      { 16, 135, 163 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffmaz,
      0, 2,
      { 16, 36, 164 },
      -1,
   } },

   /* replace50_0 -> 16 in the cache */
   /* replace50_1_0 -> 36 in the cache */
   /* replace50_1_1 -> 135 in the cache */
   /* replace50_1 -> 138 in the cache */
   /* replace50_2 -> 163 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 2,
      { 16, 138, 163 },
      -1,
   } },

   /* ('~ffmaz', 'a', 'b', ('fmulz(is_used_once)', 'a', 'c')) => ('fmulz', 'a', ('fadd', 'b', 'c')) */
   /* search51_0 -> 16 in the cache */
   /* search51_1 -> 36 in the cache */
   /* search51_2_0 -> 16 in the cache */
   /* search51_2_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 16, 135 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffmaz,
      0, 2,
      { 16, 36, 167 },
      -1,
   } },

   /* replace51_0 -> 16 in the cache */
   /* replace51_1_0 -> 36 in the cache */
   /* replace51_1_1 -> 135 in the cache */
   /* replace51_1 -> 138 in the cache */
   /* replace51 -> 139 in the cache */

   /* ('~fadd', ('fmulz(is_used_once)', 'a', 'b'), ('ffmaz(is_used_once)', 'a', 'c', 'd')) => ('ffmaz', 'a', ('fadd', 'b', 'c'), 'd') */
   /* search52_0_0 -> 16 in the cache */
   /* search52_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 16, 36 },
      0,
   } },
   /* search52_1_0 -> 16 in the cache */
   /* search52_1_1 -> 135 in the cache */
   /* search52_1_2 -> 163 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      2, 1,
      { 16, 135, 163 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 169, 170 },
      -1,
   } },

   /* replace52_0 -> 16 in the cache */
   /* replace52_1_0 -> 36 in the cache */
   /* replace52_1_1 -> 135 in the cache */
   /* replace52_1 -> 138 in the cache */
   /* replace52_2 -> 163 in the cache */
   /* replace52 -> 166 in the cache */

   /* ('~ffmaz', 'a', ('fmulz(is_used_once)', 'b', 'c'), ('fmulz(is_used_once)', 'b', 'd')) => ('fmulz', 'b', ('ffmaz', 'a', 'c', 'd')) */
   /* search53_0 -> 16 in the cache */
   /* search53_1_0 -> 36 in the cache */
   /* search53_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 36, 135 },
      0,
   } },
   /* search53_2_0 -> 36 in the cache */
   /* search53_2_1 -> 163 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      2, 1,
      { 36, 163 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffmaz,
      0, 3,
      { 16, 172, 173 },
      -1,
   } },

   /* replace53_0 -> 36 in the cache */
   /* replace53_1_0 -> 16 in the cache */
   /* replace53_1_1 -> 135 in the cache */
   /* replace53_1_2 -> 163 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 1,
      { 16, 135, 163 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 2,
      { 36, 175 },
      -1,
   } },

   /* ('iadd', ('imul', 'a', 'b'), ('imul', 'a', 'c')) => ('imul', 'a', ('iadd', 'b', 'c')) */
   /* search54_0_0 -> 126 in the cache */
   /* search54_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 126, 127 },
      -1,
   } },
   /* search54_1_0 -> 126 in the cache */
   /* search54_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      2, 1,
      { 126, 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 3,
      { 177, 178 },
      -1,
   } },

   /* replace54_0 -> 126 in the cache */
   /* replace54_1_0 -> 127 in the cache */
   /* replace54_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 127, 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      0, 2,
      { 126, 180 },
      -1,
   } },

   /* ('iadd', ('ishl', 'b', 'a'), ('ishl', 'c', 'a')) => ('ishl', ('iadd', 'b', 'c'), 'a') */
   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 182, 183 },
      -1,
   } },
   /* search55_1_0 -> 129 in the cache */
   /* search55_1_1 -> 183 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 129, 183 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 184, 185 },
      -1,
   } },

   /* replace55_0_0 -> 182 in the cache */
   /* replace55_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 182, 129 },
      -1,
   } },
   /* replace55_1 -> 183 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 187, 183 },
      -1,
   } },

   /* ('iand', ('ior', 'a', 'b'), ('ior', 'a', 'c')) => ('ior', 'a', ('iand', 'b', 'c')) */
   /* search56_0_0 -> 126 in the cache */
   /* search56_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 126, 127 },
      -1,
   } },
   /* search56_1_0 -> 126 in the cache */
   /* search56_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      2, 1,
      { 126, 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 189, 190 },
      -1,
   } },

   /* replace56_0 -> 126 in the cache */
   /* replace56_1_0 -> 127 in the cache */
   /* replace56_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 127, 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 126, 192 },
      -1,
   } },

   /* ('ior', ('iand', 'a', 'b'), ('iand', 'a', 'c')) => ('iand', 'a', ('ior', 'b', 'c')) */
   /* search57_0_0 -> 126 in the cache */
   /* search57_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 126, 127 },
      -1,
   } },
   /* search57_1_0 -> 126 in the cache */
   /* search57_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 126, 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 194, 195 },
      -1,
   } },

   /* replace57_0 -> 126 in the cache */
   /* replace57_1_0 -> 127 in the cache */
   /* replace57_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 127, 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 126, 197 },
      -1,
   } },

   /* ('ieq', ('iand', 'a', '#b(is_pos_power_of_two)'), 'b') => ('ine', ('iand', 'a', 'b'), 0) */
   /* search58_0_0 -> 0 in the cache */
   /* search58_0_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 0, 1 },
      -1,
   } },
   /* search58_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 199, 3 },
      -1,
   } },

   /* replace58_0_0 -> 0 in the cache */
   /* replace58_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* replace58_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 201, 86 },
      -1,
   } },

   /* ('ine', ('iand', 'a', '#b(is_pos_power_of_two)'), 'b') => ('ieq', ('iand', 'a', 'b'), 0) */
   /* search59_0_0 -> 0 in the cache */
   /* search59_0_1 -> 1 in the cache */
   /* search59_0 -> 199 in the cache */
   /* search59_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 199, 3 },
      -1,
   } },

   /* replace59_0_0 -> 0 in the cache */
   /* replace59_0_1 -> 3 in the cache */
   /* replace59_0 -> 201 in the cache */
   /* replace59_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 201, 86 },
      -1,
   } },

   /* ('~fadd', ('fneg', 'a'), 'a') => 0.0 */
   /* search60_0_0 -> 32 in the cache */
   /* search60_0 -> 98 in the cache */
   /* search60_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 98, 32 },
      -1,
   } },

   /* replace60 -> 111 in the cache */

   /* ('iadd', ('ineg', 'a'), 'a') => 0 */
   /* search61_0_0 -> 32 in the cache */
   /* search61_0 -> 100 in the cache */
   /* search61_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 100, 32 },
      -1,
   } },

   /* replace61 -> 61 in the cache */

   /* ('iadd', ('ineg', 'a'), ('iadd', 'a', 'b')) => b */
   /* search62_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 0 },
      -1,
   } },
   /* search62_1_0 -> 0 in the cache */
   /* search62_1_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 207, 208 },
      -1,
   } },

   /* replace62 -> 3 in the cache */

   /* ('iadd', 'a', ('iadd', ('ineg', 'a'), 'b')) => b */
   /* search63_0 -> 0 in the cache */
   /* search63_1_0_0 -> 0 in the cache */
   /* search63_1_0 -> 207 in the cache */
   /* search63_1_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 207, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 0, 210 },
      -1,
   } },

   /* replace63 -> 3 in the cache */

   /* ('~fadd', ('fneg', 'a'), ('fadd', 'a', 'b')) => b */
   /* search64_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 0 },
      -1,
   } },
   /* search64_1_0 -> 0 in the cache */
   /* search64_1_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 212, 213 },
      -1,
   } },

   /* replace64 -> 3 in the cache */

   /* ('~fadd', 'a', ('fadd', ('fneg', 'a'), 'b')) => b */
   /* search65_0 -> 0 in the cache */
   /* search65_1_0_0 -> 0 in the cache */
   /* search65_1_0 -> 212 in the cache */
   /* search65_1_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 212, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 0, 215 },
      -1,
   } },

   /* replace65 -> 3 in the cache */

   /* ('fadd', ('fsat', 'a'), ('fsat', ('fneg', 'a'))) => ('fsat', ('fabs', 'a')) */
   /* search66_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 32 },
      -1,
   } },
   /* search66_1_0_0 -> 32 in the cache */
   /* search66_1_0 -> 98 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 98 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 217, 218 },
      -1,
   } },

   /* replace66_0_0 -> 32 in the cache */
   /* replace66_0 -> 103 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 103 },
      -1,
   } },

   /* ('~fmul', 'a', 0.0) => 0.0 */
   /* search67_0 -> 32 in the cache */
   /* search67_1 -> 111 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 32, 111 },
      -1,
   } },

   /* replace67 -> 111 in the cache */

   /* ('fmul', 'a@16', 0.0) => 0.0 */
   /* search68_0 -> 11 in the cache */
   /* search68_1 -> 113 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 11, 113 },
      -1,
   } },

   /* replace68 -> 113 in the cache */

   /* ('fmul', 'a@32', 0.0) => 0.0 */
   /* search69_0 -> 16 in the cache */
   /* search69_1 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 16, 115 },
      -1,
   } },

   /* replace69 -> 115 in the cache */

   /* ('fmulz', 'a', 0.0) => 0.0 */
   /* search70_0 -> 16 in the cache */
   /* search70_1 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 16, 115 },
      -1,
   } },

   /* replace70 -> 115 in the cache */

   /* ('fmulz', 'a', 'b(is_finite_not_zero)') => ('fmul', 'a', 'b') */
   /* search71_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      false,
      nir_type_invalid,
      3,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 16, 225 },
      -1,
   } },

   /* replace71_0 -> 16 in the cache */
   /* replace71_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* ('fmulz', 'a(is_finite)', 'b(is_finite)') => ('fmul', 'a', 'b') */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      false,
      nir_type_invalid,
      4,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      false,
      nir_type_invalid,
      4,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 228, 229 },
      -1,
   } },

   /* replace72_0 -> 16 in the cache */
   /* replace72_1 -> 36 in the cache */
   /* replace72 -> 227 in the cache */

   /* ('fmulz', 'a', 'a') => ('fmul', 'a', 'a') */
   /* search73_0 -> 16 in the cache */
   /* search73_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      -1, 0,
      { 16, 16 },
      -1,
   } },

   /* replace73_0 -> 16 in the cache */
   /* replace73_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      -1, 0,
      { 16, 16 },
      -1,
   } },

   /* ('ffmaz', 'a', 'b(is_finite_not_zero)', 'c') => ('ffma', 'a', 'b', 'c') */
   /* search74_0 -> 16 in the cache */
   /* search74_1 -> 225 in the cache */
   /* search74_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 16, 225, 135 },
      -1,
   } },

   /* replace74_0 -> 16 in the cache */
   /* replace74_1 -> 36 in the cache */
   /* replace74_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 16, 36, 135 },
      -1,
   } },

   /* ('ffmaz', 'a(is_finite)', 'b(is_finite)', 'c') => ('ffma', 'a', 'b', 'c') */
   /* search75_0 -> 228 in the cache */
   /* search75_1 -> 229 in the cache */
   /* search75_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 228, 229, 135 },
      -1,
   } },

   /* replace75_0 -> 16 in the cache */
   /* replace75_1 -> 36 in the cache */
   /* replace75_2 -> 135 in the cache */
   /* replace75 -> 234 in the cache */

   /* ('ffmaz', 'a', 'a', 'b') => ('ffma', 'a', 'a', 'b') */
   /* search76_0 -> 16 in the cache */
   /* search76_1 -> 16 in the cache */
   /* search76_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      -1, 0,
      { 16, 16, 36 },
      -1,
   } },

   /* replace76_0 -> 16 in the cache */
   /* replace76_1 -> 16 in the cache */
   /* replace76_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffma,
      -1, 0,
      { 16, 16, 36 },
      -1,
   } },

   /* ('imul', 'a', 0) => 0 */
   /* search77_0 -> 32 in the cache */
   /* search77_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 32, 61 },
      -1,
   } },

   /* replace77 -> 61 in the cache */

   /* ('umul_unorm_4x8_vc4', 'a', 0) => 0 */
   /* search78_0 -> 16 in the cache */
   /* search78_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umul_unorm_4x8_vc4,
      0, 1,
      { 16, 122 },
      -1,
   } },

   /* replace78 -> 122 in the cache */

   /* ('umul_unorm_4x8_vc4', 'a', -1) => a */
   /* search79_0 -> 16 in the cache */
   /* search79_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umul_unorm_4x8_vc4,
      0, 1,
      { 16, 124 },
      -1,
   } },

   /* replace79 -> 16 in the cache */

   /* ('~fmul', 'a', 1.0) => a */
   /* search80_0 -> 32 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 32, 241 },
      -1,
   } },

   /* replace80 -> 32 in the cache */

   /* ('~fmulz', 'a', 1.0) => a */
   /* search81_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 16, 243 },
      -1,
   } },

   /* replace81 -> 16 in the cache */

   /* ('fmul(is_only_used_as_float)', 'a', 1.0) => a */
   /* search82_0 -> 32 in the cache */
   /* search82_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 32, 241 },
      1,
   } },

   /* replace82 -> 32 in the cache */

   /* ('imul', 'a', 1) => a */
   /* search83_0 -> 32 in the cache */
   /* search83_1 -> 35 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 32, 35 },
      -1,
   } },

   /* replace83 -> 32 in the cache */

   /* ('fmul', 'a', -1.0) => ('fneg', 'a') */
   /* search84_0 -> 32 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0xbff0000000000000 /* -1.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 32, 247 },
      -1,
   } },

   /* replace84_0 -> 32 in the cache */
   /* replace84 -> 98 in the cache */

   /* ('imul', 'a', -1) => ('ineg', 'a') */
   /* search85_0 -> 32 in the cache */
   /* search85_1 -> 63 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 32, 63 },
      -1,
   } },

   /* replace85_0 -> 32 in the cache */
   /* replace85 -> 100 in the cache */

   /* ('fmul', ('fsign', 'a'), ('fmul', 'a', 'a')) => ('fmul', ('fabs', 'a'), 'a') */
   /* search86_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 32 },
      -1,
   } },
   /* search86_1_0 -> 32 in the cache */
   /* search86_1_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      -1, 0,
      { 32, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 250, 251 },
      -1,
   } },

   /* replace86_0_0 -> 32 in the cache */
   /* replace86_0 -> 103 in the cache */
   /* replace86_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 103, 32 },
      -1,
   } },

   /* ('fmul', ('fmul', ('fsign', 'a'), 'a'), 'a') => ('fmul', ('fabs', 'a'), 'a') */
   /* search87_0_0_0 -> 32 in the cache */
   /* search87_0_0 -> 250 in the cache */
   /* search87_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 250, 32 },
      -1,
   } },
   /* search87_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 254, 32 },
      -1,
   } },

   /* replace87_0_0 -> 32 in the cache */
   /* replace87_0 -> 103 in the cache */
   /* replace87_1 -> 32 in the cache */
   /* replace87 -> 253 in the cache */

   /* ('~ffma', 0.0, 'a', 'b') => b */
   { .constant = {
      { nir_search_value_constant, -2 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },
   /* search88_1 -> 0 in the cache */
   /* search88_2 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 256, 0, 3 },
      -1,
   } },

   /* replace88 -> 3 in the cache */

   /* ('ffma@16(is_only_used_as_float)', 0.0, 'a', 'b') => b */
   /* search89_0 -> 113 in the cache */
   /* search89_1 -> 11 in the cache */
   { .variable = {
      { nir_search_value_variable, 16 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 113, 11, 258 },
      1,
   } },

   /* replace89 -> 258 in the cache */

   /* ('ffma@32(is_only_used_as_float)', 0.0, 'a', 'b') => b */
   /* search90_0 -> 115 in the cache */
   /* search90_1 -> 16 in the cache */
   /* search90_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 115, 16, 36 },
      1,
   } },

   /* replace90 -> 36 in the cache */

   /* ('ffmaz', 0.0, 'a', 'b') => ('fadd', 0.0, 'b') */
   /* search91_0 -> 115 in the cache */
   /* search91_1 -> 16 in the cache */
   /* search91_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 115, 16, 36 },
      -1,
   } },

   /* replace91_0 -> 115 in the cache */
   /* replace91_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 115, 36 },
      -1,
   } },

   /* ('~ffma', 'a', 'b', 0.0) => ('fmul', 'a', 'b') */
   /* search92_0 -> 0 in the cache */
   /* search92_1 -> 3 in the cache */
   /* search92_2 -> 256 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 0, 3, 256 },
      -1,
   } },

   /* replace92_0 -> 0 in the cache */
   /* replace92_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('ffma@16', 'a', 'b', 0.0) => ('fmul', 'a', 'b') */
   /* search93_0 -> 11 in the cache */
   /* search93_1 -> 258 in the cache */
   /* search93_2 -> 113 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 11, 258, 113 },
      -1,
   } },

   /* replace93_0 -> 11 in the cache */
   /* replace93_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 11, 258 },
      -1,
   } },

   /* ('ffma@32', 'a', 'b', 0.0) => ('fmul', 'a', 'b') */
   /* search94_0 -> 16 in the cache */
   /* search94_1 -> 36 in the cache */
   /* search94_2 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 16, 36, 115 },
      -1,
   } },

   /* replace94_0 -> 16 in the cache */
   /* replace94_1 -> 36 in the cache */
   /* replace94 -> 227 in the cache */

   /* ('ffmaz', 'a', 'b', 0.0) => ('fmulz', 'a', 'b') */
   /* search95_0 -> 16 in the cache */
   /* search95_1 -> 36 in the cache */
   /* search95_2 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 16, 36, 115 },
      -1,
   } },

   /* replace95_0 -> 16 in the cache */
   /* replace95_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* ('ffma', 1.0, 'a', 'b') => ('fadd', 'a', 'b') */
   { .constant = {
      { nir_search_value_constant, -2 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   /* search96_1 -> 0 in the cache */
   /* search96_2 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 270, 0, 3 },
      -1,
   } },

   /* replace96_0 -> 0 in the cache */
   /* replace96_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('ffmaz', 1.0, 'a', 'b') => ('fadd', 'a', 'b') */
   /* search97_0 -> 243 in the cache */
   /* search97_1 -> 16 in the cache */
   /* search97_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 243, 16, 36 },
      -1,
   } },

   /* replace97_0 -> 16 in the cache */
   /* replace97_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* ('ffma', -1.0, 'a', 'b') => ('fadd', ('fneg', 'a'), 'b') */
   { .constant = {
      { nir_search_value_constant, -2 },
      nir_type_float, { 0xbff0000000000000 /* -1.0 */ },
   } },
   /* search98_1 -> 0 in the cache */
   /* search98_2 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 275, 0, 3 },
      -1,
   } },

   /* replace98_0_0 -> 0 in the cache */
   /* replace98_0 -> 212 in the cache */
   /* replace98_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 212, 3 },
      -1,
   } },

   /* ('ffmaz', -1.0, 'a', 'b') => ('fadd', ('fneg', 'a'), 'b') */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0xbff0000000000000 /* -1.0 */ },
   } },
   /* search99_1 -> 16 in the cache */
   /* search99_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 278, 16, 36 },
      -1,
   } },

   /* replace99_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 16 },
      -1,
   } },
   /* replace99_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 280, 36 },
      -1,
   } },

   /* ('~ffma', '#a', '#b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c') */
   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* a */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search100_2 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 282, 283, 129 },
      -1,
   } },

   /* replace100_0_0 -> 126 in the cache */
   /* replace100_0_1 -> 127 in the cache */
   /* replace100_0 -> 128 in the cache */
   /* replace100_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 128, 129 },
      -1,
   } },

   /* ('~ffmaz', '#a', '#b', 'c') => ('fadd', ('fmulz', 'a', 'b'), 'c') */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search101_1 -> 33 in the cache */
   /* search101_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 286, 33, 135 },
      -1,
   } },

   /* replace101_0_0 -> 16 in the cache */
   /* replace101_0_1 -> 36 in the cache */
   /* replace101_0 -> 134 in the cache */
   /* replace101_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 134, 135 },
      -1,
   } },

   /* ('~flrp', 'a', 'b', 0.0) => a */
   /* search102_0 -> 0 in the cache */
   /* search102_1 -> 3 in the cache */
   /* search102_2 -> 256 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 0, 3, 256 },
      -1,
   } },

   /* replace102 -> 0 in the cache */

   /* ('~flrp', 'a', 'b', 1.0) => b */
   /* search103_0 -> 0 in the cache */
   /* search103_1 -> 3 in the cache */
   /* search103_2 -> 270 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 0, 3, 270 },
      -1,
   } },

   /* replace103 -> 3 in the cache */

   /* ('~flrp', 'a', 'a', 'b') => a */
   /* search104_0 -> 0 in the cache */
   /* search104_1 -> 0 in the cache */
   /* search104_2 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 0, 0, 3 },
      -1,
   } },

   /* replace104 -> 0 in the cache */

   /* ('~flrp', 0.0, 'a', 'b') => ('fmul', 'a', 'b') */
   /* search105_0 -> 256 in the cache */
   /* search105_1 -> 0 in the cache */
   /* search105_2 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 256, 0, 3 },
      -1,
   } },

   /* replace105_0 -> 0 in the cache */
   /* replace105_1 -> 3 in the cache */
   /* replace105 -> 264 in the cache */

   /* ('~flrp', 'a', ('fadd(is_used_once)', 'a', 'b'), 'c') => ('fadd', ('fmul', 'b', 'c'), 'a') */
   /* search106_0 -> 126 in the cache */
   /* search106_1_0 -> 126 in the cache */
   /* search106_1_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 126, 127 },
      0,
   } },
   /* search106_2 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 1,
      { 126, 293, 129 },
      -1,
   } },

   /* replace106_0_0 -> 127 in the cache */
   /* replace106_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 127, 129 },
      -1,
   } },
   /* replace106_1 -> 126 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 295, 126 },
      -1,
   } },

   /* ('sdot_4x8_iadd', 'a', 0, 'b') => b */
   /* search107_0 -> 16 in the cache */
   /* search107_1 -> 122 in the cache */
   /* search107_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_4x8_iadd,
      0, 1,
      { 16, 122, 36 },
      -1,
   } },

   /* replace107 -> 36 in the cache */

   /* ('udot_4x8_uadd', 'a', 0, 'b') => b */
   /* search108_0 -> 16 in the cache */
   /* search108_1 -> 122 in the cache */
   /* search108_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_4x8_uadd,
      0, 1,
      { 16, 122, 36 },
      -1,
   } },

   /* replace108 -> 36 in the cache */

   /* ('sdot_4x8_iadd_sat', 'a', 0, 'b') => b */
   /* search109_0 -> 16 in the cache */
   /* search109_1 -> 122 in the cache */
   /* search109_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_4x8_iadd_sat,
      0, 1,
      { 16, 122, 36 },
      -1,
   } },

   /* replace109 -> 36 in the cache */

   /* ('udot_4x8_uadd_sat', 'a', 0, 'b') => b */
   /* search110_0 -> 16 in the cache */
   /* search110_1 -> 122 in the cache */
   /* search110_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_4x8_uadd_sat,
      0, 1,
      { 16, 122, 36 },
      -1,
   } },

   /* replace110 -> 36 in the cache */

   /* ('sdot_2x16_iadd', 'a', 0, 'b') => b */
   /* search111_0 -> 16 in the cache */
   /* search111_1 -> 122 in the cache */
   /* search111_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_2x16_iadd,
      0, 1,
      { 16, 122, 36 },
      -1,
   } },

   /* replace111 -> 36 in the cache */

   /* ('udot_2x16_uadd', 'a', 0, 'b') => b */
   /* search112_0 -> 16 in the cache */
   /* search112_1 -> 122 in the cache */
   /* search112_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_2x16_uadd,
      0, 1,
      { 16, 122, 36 },
      -1,
   } },

   /* replace112 -> 36 in the cache */

   /* ('sdot_2x16_iadd_sat', 'a', 0, 'b') => b */
   /* search113_0 -> 16 in the cache */
   /* search113_1 -> 122 in the cache */
   /* search113_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_2x16_iadd_sat,
      0, 1,
      { 16, 122, 36 },
      -1,
   } },

   /* replace113 -> 36 in the cache */

   /* ('udot_2x16_uadd_sat', 'a', 0, 'b') => b */
   /* search114_0 -> 16 in the cache */
   /* search114_1 -> 122 in the cache */
   /* search114_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_2x16_uadd_sat,
      0, 1,
      { 16, 122, 36 },
      -1,
   } },

   /* replace114 -> 36 in the cache */

   /* ('sudot_4x8_iadd', 'a', 0, 'b') => b */
   /* search115_0 -> 16 in the cache */
   /* search115_1 -> 122 in the cache */
   /* search115_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd,
      -1, 0,
      { 16, 122, 36 },
      -1,
   } },

   /* replace115 -> 36 in the cache */

   /* ('sudot_4x8_iadd', 0, 'a', 'b') => b */
   /* search116_0 -> 122 in the cache */
   /* search116_1 -> 16 in the cache */
   /* search116_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd,
      -1, 0,
      { 122, 16, 36 },
      -1,
   } },

   /* replace116 -> 36 in the cache */

   /* ('sudot_4x8_iadd_sat', 'a', 0, 'b') => b */
   /* search117_0 -> 16 in the cache */
   /* search117_1 -> 122 in the cache */
   /* search117_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd_sat,
      -1, 0,
      { 16, 122, 36 },
      -1,
   } },

   /* replace117 -> 36 in the cache */

   /* ('sudot_4x8_iadd_sat', 0, 'a', 'b') => b */
   /* search118_0 -> 122 in the cache */
   /* search118_1 -> 16 in the cache */
   /* search118_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd_sat,
      -1, 0,
      { 122, 16, 36 },
      -1,
   } },

   /* replace118 -> 36 in the cache */

   /* ('iadd', ('sdot_4x8_iadd(is_used_once)', 'a', 'b', '#c'), '#d') => ('sdot_4x8_iadd', 'a', 'b', ('iadd', 'c', 'd')) */
   /* search119_0_0 -> 16 in the cache */
   /* search119_0_1 -> 36 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* c */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_4x8_iadd,
      1, 1,
      { 16, 36, 309 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      3, /* d */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 310, 311 },
      -1,
   } },

   /* replace119_0 -> 16 in the cache */
   /* replace119_1 -> 36 in the cache */
   /* replace119_2_0 -> 135 in the cache */
   /* replace119_2_1 -> 163 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 135, 163 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_4x8_iadd,
      0, 2,
      { 16, 36, 313 },
      -1,
   } },

   /* ('iadd', ('udot_4x8_uadd(is_used_once)', 'a', 'b', '#c'), '#d') => ('udot_4x8_uadd', 'a', 'b', ('iadd', 'c', 'd')) */
   /* search120_0_0 -> 16 in the cache */
   /* search120_0_1 -> 36 in the cache */
   /* search120_0_2 -> 309 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_4x8_uadd,
      1, 1,
      { 16, 36, 309 },
      0,
   } },
   /* search120_1 -> 311 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 315, 311 },
      -1,
   } },

   /* replace120_0 -> 16 in the cache */
   /* replace120_1 -> 36 in the cache */
   /* replace120_2_0 -> 135 in the cache */
   /* replace120_2_1 -> 163 in the cache */
   /* replace120_2 -> 313 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_4x8_uadd,
      0, 2,
      { 16, 36, 313 },
      -1,
   } },

   /* ('iadd', ('sudot_4x8_iadd(is_used_once)', 'a', 'b', '#c'), '#d') => ('sudot_4x8_iadd', 'a', 'b', ('iadd', 'c', 'd')) */
   /* search121_0_0 -> 16 in the cache */
   /* search121_0_1 -> 36 in the cache */
   /* search121_0_2 -> 309 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd,
      -1, 0,
      { 16, 36, 309 },
      0,
   } },
   /* search121_1 -> 311 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 318, 311 },
      -1,
   } },

   /* replace121_0 -> 16 in the cache */
   /* replace121_1 -> 36 in the cache */
   /* replace121_2_0 -> 135 in the cache */
   /* replace121_2_1 -> 163 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 135, 163 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd,
      -1, 1,
      { 16, 36, 320 },
      -1,
   } },

   /* ('iadd', ('sdot_2x16_iadd(is_used_once)', 'a', 'b', '#c'), '#d') => ('sdot_2x16_iadd', 'a', 'b', ('iadd', 'c', 'd')) */
   /* search122_0_0 -> 16 in the cache */
   /* search122_0_1 -> 36 in the cache */
   /* search122_0_2 -> 309 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_2x16_iadd,
      1, 1,
      { 16, 36, 309 },
      0,
   } },
   /* search122_1 -> 311 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 322, 311 },
      -1,
   } },

   /* replace122_0 -> 16 in the cache */
   /* replace122_1 -> 36 in the cache */
   /* replace122_2_0 -> 135 in the cache */
   /* replace122_2_1 -> 163 in the cache */
   /* replace122_2 -> 313 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_2x16_iadd,
      0, 2,
      { 16, 36, 313 },
      -1,
   } },

   /* ('iadd', ('udot_2x16_uadd(is_used_once)', 'a', 'b', '#c'), '#d') => ('udot_2x16_uadd', 'a', 'b', ('iadd', 'c', 'd')) */
   /* search123_0_0 -> 16 in the cache */
   /* search123_0_1 -> 36 in the cache */
   /* search123_0_2 -> 309 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_2x16_uadd,
      1, 1,
      { 16, 36, 309 },
      0,
   } },
   /* search123_1 -> 311 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 325, 311 },
      -1,
   } },

   /* replace123_0 -> 16 in the cache */
   /* replace123_1 -> 36 in the cache */
   /* replace123_2_0 -> 135 in the cache */
   /* replace123_2_1 -> 163 in the cache */
   /* replace123_2 -> 313 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_2x16_uadd,
      0, 2,
      { 16, 36, 313 },
      -1,
   } },

   /* ('iadd', ('sdot_4x8_iadd', 'a(is_not_const)', 'b', 0), 'c') => ('sdot_4x8_iadd', 'a', 'b', 'c') */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search124_0_1 -> 36 in the cache */
   /* search124_0_2 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_4x8_iadd,
      1, 1,
      { 328, 36, 122 },
      -1,
   } },
   /* search124_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 329, 135 },
      -1,
   } },

   /* replace124_0 -> 16 in the cache */
   /* replace124_1 -> 36 in the cache */
   /* replace124_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_4x8_iadd,
      0, 1,
      { 16, 36, 135 },
      -1,
   } },

   /* ('iadd', ('udot_4x8_uadd', 'a(is_not_const)', 'b', 0), 'c') => ('udot_4x8_uadd', 'a', 'b', 'c') */
   /* search125_0_0 -> 328 in the cache */
   /* search125_0_1 -> 36 in the cache */
   /* search125_0_2 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_4x8_uadd,
      1, 1,
      { 328, 36, 122 },
      -1,
   } },
   /* search125_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 332, 135 },
      -1,
   } },

   /* replace125_0 -> 16 in the cache */
   /* replace125_1 -> 36 in the cache */
   /* replace125_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_4x8_uadd,
      0, 1,
      { 16, 36, 135 },
      -1,
   } },

   /* ('iadd', ('sudot_4x8_iadd', 'a(is_not_const)', 'b', 0), 'c') => ('sudot_4x8_iadd', 'a', 'b', 'c') */
   /* search126_0_0 -> 328 in the cache */
   /* search126_0_1 -> 36 in the cache */
   /* search126_0_2 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd,
      -1, 0,
      { 328, 36, 122 },
      -1,
   } },
   /* search126_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 335, 135 },
      -1,
   } },

   /* replace126_0 -> 16 in the cache */
   /* replace126_1 -> 36 in the cache */
   /* replace126_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd,
      -1, 0,
      { 16, 36, 135 },
      -1,
   } },

   /* ('iadd', ('sudot_4x8_iadd', 'a', 'b(is_not_const)', 0), 'c') => ('sudot_4x8_iadd', 'a', 'b', 'c') */
   /* search127_0_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search127_0_2 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd,
      -1, 0,
      { 16, 338, 122 },
      -1,
   } },
   /* search127_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 339, 135 },
      -1,
   } },

   /* replace127_0 -> 16 in the cache */
   /* replace127_1 -> 36 in the cache */
   /* replace127_2 -> 135 in the cache */
   /* replace127 -> 337 in the cache */

   /* ('iadd', ('sdot_2x16_iadd', 'a(is_not_const)', 'b', 0), 'c') => ('sdot_2x16_iadd', 'a', 'b', 'c') */
   /* search128_0_0 -> 328 in the cache */
   /* search128_0_1 -> 36 in the cache */
   /* search128_0_2 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_2x16_iadd,
      1, 1,
      { 328, 36, 122 },
      -1,
   } },
   /* search128_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 341, 135 },
      -1,
   } },

   /* replace128_0 -> 16 in the cache */
   /* replace128_1 -> 36 in the cache */
   /* replace128_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_2x16_iadd,
      0, 1,
      { 16, 36, 135 },
      -1,
   } },

   /* ('iadd', ('udot_2x16_uadd', 'a(is_not_const)', 'b', 0), 'c') => ('udot_2x16_uadd', 'a', 'b', 'c') */
   /* search129_0_0 -> 328 in the cache */
   /* search129_0_1 -> 36 in the cache */
   /* search129_0_2 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_2x16_uadd,
      1, 1,
      { 328, 36, 122 },
      -1,
   } },
   /* search129_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 344, 135 },
      -1,
   } },

   /* replace129_0 -> 16 in the cache */
   /* replace129_1 -> 36 in the cache */
   /* replace129_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_2x16_uadd,
      0, 1,
      { 16, 36, 135 },
      -1,
   } },

   /* ('sdot_4x8_iadd', '#a', '#b', 'c(is_not_const)') => ('iadd', ('sdot_4x8_iadd', 'a', 'b', 0), 'c') */
   /* search130_0 -> 286 in the cache */
   /* search130_1 -> 33 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* c */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_4x8_iadd,
      0, 1,
      { 286, 33, 347 },
      -1,
   } },

   /* replace130_0_0 -> 16 in the cache */
   /* replace130_0_1 -> 36 in the cache */
   /* replace130_0_2 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_4x8_iadd,
      1, 1,
      { 16, 36, 122 },
      -1,
   } },
   /* replace130_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 349, 135 },
      -1,
   } },

   /* ('udot_4x8_uadd', '#a', '#b', 'c(is_not_const)') => ('iadd', ('udot_4x8_uadd', 'a', 'b', 0), 'c') */
   /* search131_0 -> 286 in the cache */
   /* search131_1 -> 33 in the cache */
   /* search131_2 -> 347 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_4x8_uadd,
      0, 1,
      { 286, 33, 347 },
      -1,
   } },

   /* replace131_0_0 -> 16 in the cache */
   /* replace131_0_1 -> 36 in the cache */
   /* replace131_0_2 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_4x8_uadd,
      1, 1,
      { 16, 36, 122 },
      -1,
   } },
   /* replace131_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 352, 135 },
      -1,
   } },

   /* ('sudot_4x8_iadd', '#a', '#b', 'c(is_not_const)') => ('iadd', ('sudot_4x8_iadd', 'a', 'b', 0), 'c') */
   /* search132_0 -> 286 in the cache */
   /* search132_1 -> 33 in the cache */
   /* search132_2 -> 347 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd,
      -1, 0,
      { 286, 33, 347 },
      -1,
   } },

   /* replace132_0_0 -> 16 in the cache */
   /* replace132_0_1 -> 36 in the cache */
   /* replace132_0_2 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd,
      -1, 0,
      { 16, 36, 122 },
      -1,
   } },
   /* replace132_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 355, 135 },
      -1,
   } },

   /* ('sdot_2x16_iadd', '#a', '#b', 'c(is_not_const)') => ('iadd', ('sdot_2x16_iadd', 'a', 'b', 0), 'c') */
   /* search133_0 -> 286 in the cache */
   /* search133_1 -> 33 in the cache */
   /* search133_2 -> 347 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_2x16_iadd,
      0, 1,
      { 286, 33, 347 },
      -1,
   } },

   /* replace133_0_0 -> 16 in the cache */
   /* replace133_0_1 -> 36 in the cache */
   /* replace133_0_2 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_2x16_iadd,
      1, 1,
      { 16, 36, 122 },
      -1,
   } },
   /* replace133_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 358, 135 },
      -1,
   } },

   /* ('udot_2x16_uadd', '#a', '#b', 'c(is_not_const)') => ('iadd', ('udot_2x16_uadd', 'a', 'b', 0), 'c') */
   /* search134_0 -> 286 in the cache */
   /* search134_1 -> 33 in the cache */
   /* search134_2 -> 347 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_2x16_uadd,
      0, 1,
      { 286, 33, 347 },
      -1,
   } },

   /* replace134_0_0 -> 16 in the cache */
   /* replace134_0_1 -> 36 in the cache */
   /* replace134_0_2 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_2x16_uadd,
      1, 1,
      { 16, 36, 122 },
      -1,
   } },
   /* replace134_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 361, 135 },
      -1,
   } },

   /* ('sdot_4x8_iadd_sat', '#a', '#b', 'c(is_not_const)') => ('iadd_sat', ('sdot_4x8_iadd', 'a', 'b', 0), 'c') */
   /* search135_0 -> 286 in the cache */
   /* search135_1 -> 33 in the cache */
   /* search135_2 -> 347 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_4x8_iadd_sat,
      0, 1,
      { 286, 33, 347 },
      -1,
   } },

   /* replace135_0_0 -> 16 in the cache */
   /* replace135_0_1 -> 36 in the cache */
   /* replace135_0_2 -> 122 in the cache */
   /* replace135_0 -> 349 in the cache */
   /* replace135_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd_sat,
      0, 2,
      { 349, 135 },
      -1,
   } },

   /* ('udot_4x8_uadd_sat', '#a', '#b', 'c(is_not_const)') => ('uadd_sat', ('udot_4x8_uadd', 'a', 'b', 0), 'c') */
   /* search136_0 -> 286 in the cache */
   /* search136_1 -> 33 in the cache */
   /* search136_2 -> 347 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_4x8_uadd_sat,
      0, 1,
      { 286, 33, 347 },
      -1,
   } },

   /* replace136_0_0 -> 16 in the cache */
   /* replace136_0_1 -> 36 in the cache */
   /* replace136_0_2 -> 122 in the cache */
   /* replace136_0 -> 352 in the cache */
   /* replace136_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_uadd_sat,
      0, 2,
      { 352, 135 },
      -1,
   } },

   /* ('sudot_4x8_iadd_sat', '#a', '#b', 'c(is_not_const)') => ('iadd_sat', ('sudot_4x8_iadd', 'a', 'b', 0), 'c') */
   /* search137_0 -> 286 in the cache */
   /* search137_1 -> 33 in the cache */
   /* search137_2 -> 347 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd_sat,
      -1, 0,
      { 286, 33, 347 },
      -1,
   } },

   /* replace137_0_0 -> 16 in the cache */
   /* replace137_0_1 -> 36 in the cache */
   /* replace137_0_2 -> 122 in the cache */
   /* replace137_0 -> 355 in the cache */
   /* replace137_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd_sat,
      0, 1,
      { 355, 135 },
      -1,
   } },

   /* ('sdot_2x16_iadd_sat', '#a', '#b', 'c(is_not_const)') => ('iadd_sat', ('sdot_2x16_iadd', 'a', 'b', 0), 'c') */
   /* search138_0 -> 286 in the cache */
   /* search138_1 -> 33 in the cache */
   /* search138_2 -> 347 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_2x16_iadd_sat,
      0, 1,
      { 286, 33, 347 },
      -1,
   } },

   /* replace138_0_0 -> 16 in the cache */
   /* replace138_0_1 -> 36 in the cache */
   /* replace138_0_2 -> 122 in the cache */
   /* replace138_0 -> 358 in the cache */
   /* replace138_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd_sat,
      0, 2,
      { 358, 135 },
      -1,
   } },

   /* ('udot_2x16_uadd_sat', '#a', '#b', 'c(is_not_const)') => ('uadd_sat', ('udot_2x16_uadd', 'a', 'b', 0), 'c') */
   /* search139_0 -> 286 in the cache */
   /* search139_1 -> 33 in the cache */
   /* search139_2 -> 347 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_2x16_uadd_sat,
      0, 1,
      { 286, 33, 347 },
      -1,
   } },

   /* replace139_0_0 -> 16 in the cache */
   /* replace139_0_1 -> 36 in the cache */
   /* replace139_0_2 -> 122 in the cache */
   /* replace139_0 -> 361 in the cache */
   /* replace139_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_uadd_sat,
      0, 2,
      { 361, 135 },
      -1,
   } },

   /* ('fmul@32', ('bcsel', ('feq', 'b', 0.0, 'ignore_exact'), 0.0, 'a'), ('bcsel', ('feq', 'a', 0.0, 'ignore_exact'), 0.0, 'b')) => ('fmulz', 'a', 'b') */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search140_0_0_1 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      true,
      nir_op_feq,
      1, 1,
      { 373, 115 },
      -1,
   } },
   /* search140_0_1 -> 115 in the cache */
   /* search140_0_2 -> 183 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 374, 115, 183 },
      -1,
   } },
   /* search140_1_0_0 -> 183 in the cache */
   /* search140_1_0_1 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      true,
      nir_op_feq,
      2, 1,
      { 183, 115 },
      -1,
   } },
   /* search140_1_1 -> 115 in the cache */
   /* search140_1_2 -> 373 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 376, 115, 373 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 3,
      { 375, 377 },
      -1,
   } },

   /* replace140_0 -> 183 in the cache */
   /* replace140_1 -> 373 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 183, 373 },
      -1,
   } },

   /* ('fmul@32', 'a', ('bcsel', ('feq', 'a', 0.0, 'ignore_exact'), 0.0, '#b(is_not_const_zero)')) => ('fmulz', 'a', 'b') */
   /* search141_0 -> 16 in the cache */
   /* search141_1_0_0 -> 16 in the cache */
   /* search141_1_0_1 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      true,
      nir_op_feq,
      1, 1,
      { 16, 115 },
      -1,
   } },
   /* search141_1_1 -> 115 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      6,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 380, 115, 381 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 16, 382 },
      -1,
   } },

   /* replace141_0 -> 16 in the cache */
   /* replace141_1 -> 36 in the cache */
   /* replace141 -> 269 in the cache */

   /* ('ffma@32', ('bcsel', ('feq', 'b', 0.0, 'ignore_exact'), 0.0, 'a'), ('bcsel', ('feq', 'a', 0.0, 'ignore_exact'), 0.0, 'b'), 'c') => ('ffmaz', 'a', 'b', 'c') */
   /* search142_0_0_0 -> 373 in the cache */
   /* search142_0_0_1 -> 115 in the cache */
   /* search142_0_0 -> 374 in the cache */
   /* search142_0_1 -> 115 in the cache */
   /* search142_0_2 -> 183 in the cache */
   /* search142_0 -> 375 in the cache */
   /* search142_1_0_0 -> 183 in the cache */
   /* search142_1_0_1 -> 115 in the cache */
   /* search142_1_0 -> 376 in the cache */
   /* search142_1_1 -> 115 in the cache */
   /* search142_1_2 -> 373 in the cache */
   /* search142_1 -> 377 in the cache */
   /* search142_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 3,
      { 375, 377, 135 },
      -1,
   } },

   /* replace142_0 -> 183 in the cache */
   /* replace142_1 -> 373 in the cache */
   /* replace142_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 183, 373, 135 },
      -1,
   } },

   /* ('ffma@32', 'a', ('bcsel', ('feq', 'a', 0.0, 'ignore_exact'), 0.0, '#b(is_not_const_zero)'), 'c') => ('ffmaz', 'a', 'b', 'c') */
   /* search143_0 -> 16 in the cache */
   /* search143_1_0_0 -> 16 in the cache */
   /* search143_1_0_1 -> 115 in the cache */
   /* search143_1_0 -> 380 in the cache */
   /* search143_1_1 -> 115 in the cache */
   /* search143_1_2 -> 381 in the cache */
   /* search143_1 -> 382 in the cache */
   /* search143_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 16, 382, 135 },
      -1,
   } },

   /* replace143_0 -> 16 in the cache */
   /* replace143_1 -> 36 in the cache */
   /* replace143_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 16, 36, 135 },
      -1,
   } },

   /* ('bcsel', ('feq', 'b', 0.0, 'ignore_exact'), 1.0, ('fexp2', ('fmul@32', 'a', 'b'))) => ('fexp2', ('fmulz', 'a', 'b')) */
   /* search144_0_0 -> 373 in the cache */
   /* search144_0_1 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      true,
      nir_op_feq,
      0, 1,
      { 373, 115 },
      -1,
   } },
   /* search144_1 -> 243 in the cache */
   /* search144_2_0_0 -> 183 in the cache */
   /* search144_2_0_1 -> 373 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 183, 373 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 1,
      { 389 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 388, 243, 390 },
      -1,
   } },

   /* replace144_0_0 -> 183 in the cache */
   /* replace144_0_1 -> 373 in the cache */
   /* replace144_0 -> 379 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 1,
      { 379 },
      -1,
   } },

   /* ('sdot_4x8_iadd', 'a', 'b', 'c') => ('iadd', ('iadd', ('iadd', ('imul', ('extract_i8', 'a', 0), ('extract_i8', 'b', 0)), ('imul', ('extract_i8', 'a', 1), ('extract_i8', 'b', 1))), ('iadd', ('imul', ('extract_i8', 'a', 2), ('extract_i8', 'b', 2)), ('imul', ('extract_i8', 'a', 3), ('extract_i8', 'b', 3)))), 'c') */
   /* search145_0 -> 16 in the cache */
   /* search145_1 -> 36 in the cache */
   /* search145_2 -> 135 in the cache */
   /* search145 -> 331 in the cache */

   /* replace145_0_0_0_0_0 -> 16 in the cache */
   /* replace145_0_0_0_0_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 16, 122 },
      -1,
   } },
   /* replace145_0_0_0_1_0 -> 36 in the cache */
   /* replace145_0_0_0_1_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 36, 122 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      3, 1,
      { 393, 394 },
      -1,
   } },
   /* replace145_0_0_1_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x1 /* 1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 16, 396 },
      -1,
   } },
   /* replace145_0_0_1_1_0 -> 36 in the cache */
   /* replace145_0_0_1_1_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 36, 396 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      4, 1,
      { 397, 398 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 3,
      { 395, 399 },
      -1,
   } },
   /* replace145_0_1_0_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x2 /* 2 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 16, 401 },
      -1,
   } },
   /* replace145_0_1_0_1_0 -> 36 in the cache */
   /* replace145_0_1_0_1_1 -> 401 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 36, 401 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      6, 1,
      { 402, 403 },
      -1,
   } },
   /* replace145_0_1_1_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x3 /* 3 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 16, 405 },
      -1,
   } },
   /* replace145_0_1_1_1_0 -> 36 in the cache */
   /* replace145_0_1_1_1_1 -> 405 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 36, 405 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      7, 1,
      { 406, 407 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      5, 3,
      { 404, 408 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 7,
      { 400, 409 },
      -1,
   } },
   /* replace145_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 8,
      { 410, 135 },
      -1,
   } },

   /* ('udot_4x8_uadd', 'a', 'b', 'c') => ('iadd', ('iadd', ('iadd', ('imul', ('extract_u8', 'a', 0), ('extract_u8', 'b', 0)), ('imul', ('extract_u8', 'a', 1), ('extract_u8', 'b', 1))), ('iadd', ('imul', ('extract_u8', 'a', 2), ('extract_u8', 'b', 2)), ('imul', ('extract_u8', 'a', 3), ('extract_u8', 'b', 3)))), 'c') */
   /* search146_0 -> 16 in the cache */
   /* search146_1 -> 36 in the cache */
   /* search146_2 -> 135 in the cache */
   /* search146 -> 334 in the cache */

   /* replace146_0_0_0_0_0 -> 16 in the cache */
   /* replace146_0_0_0_0_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 16, 122 },
      -1,
   } },
   /* replace146_0_0_0_1_0 -> 36 in the cache */
   /* replace146_0_0_0_1_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 36, 122 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      3, 1,
      { 412, 413 },
      -1,
   } },
   /* replace146_0_0_1_0_0 -> 16 in the cache */
   /* replace146_0_0_1_0_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 16, 396 },
      -1,
   } },
   /* replace146_0_0_1_1_0 -> 36 in the cache */
   /* replace146_0_0_1_1_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 36, 396 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      4, 1,
      { 415, 416 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 3,
      { 414, 417 },
      -1,
   } },
   /* replace146_0_1_0_0_0 -> 16 in the cache */
   /* replace146_0_1_0_0_1 -> 401 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 16, 401 },
      -1,
   } },
   /* replace146_0_1_0_1_0 -> 36 in the cache */
   /* replace146_0_1_0_1_1 -> 401 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 36, 401 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      6, 1,
      { 419, 420 },
      -1,
   } },
   /* replace146_0_1_1_0_0 -> 16 in the cache */
   /* replace146_0_1_1_0_1 -> 405 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 16, 405 },
      -1,
   } },
   /* replace146_0_1_1_1_0 -> 36 in the cache */
   /* replace146_0_1_1_1_1 -> 405 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 36, 405 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      7, 1,
      { 422, 423 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      5, 3,
      { 421, 424 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 7,
      { 418, 425 },
      -1,
   } },
   /* replace146_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 8,
      { 426, 135 },
      -1,
   } },

   /* ('sudot_4x8_iadd', 'a', 'b', 'c') => ('iadd', ('iadd', ('iadd', ('imul', ('extract_i8', 'a', 0), ('extract_u8', 'b', 0)), ('imul', ('extract_i8', 'a', 1), ('extract_u8', 'b', 1))), ('iadd', ('imul', ('extract_i8', 'a', 2), ('extract_u8', 'b', 2)), ('imul', ('extract_i8', 'a', 3), ('extract_u8', 'b', 3)))), 'c') */
   /* search147_0 -> 16 in the cache */
   /* search147_1 -> 36 in the cache */
   /* search147_2 -> 135 in the cache */
   /* search147 -> 337 in the cache */

   /* replace147_0_0_0_0_0 -> 16 in the cache */
   /* replace147_0_0_0_0_1 -> 122 in the cache */
   /* replace147_0_0_0_0 -> 393 in the cache */
   /* replace147_0_0_0_1_0 -> 36 in the cache */
   /* replace147_0_0_0_1_1 -> 122 in the cache */
   /* replace147_0_0_0_1 -> 413 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      3, 1,
      { 393, 413 },
      -1,
   } },
   /* replace147_0_0_1_0_0 -> 16 in the cache */
   /* replace147_0_0_1_0_1 -> 396 in the cache */
   /* replace147_0_0_1_0 -> 397 in the cache */
   /* replace147_0_0_1_1_0 -> 36 in the cache */
   /* replace147_0_0_1_1_1 -> 396 in the cache */
   /* replace147_0_0_1_1 -> 416 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      4, 1,
      { 397, 416 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 3,
      { 428, 429 },
      -1,
   } },
   /* replace147_0_1_0_0_0 -> 16 in the cache */
   /* replace147_0_1_0_0_1 -> 401 in the cache */
   /* replace147_0_1_0_0 -> 402 in the cache */
   /* replace147_0_1_0_1_0 -> 36 in the cache */
   /* replace147_0_1_0_1_1 -> 401 in the cache */
   /* replace147_0_1_0_1 -> 420 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      6, 1,
      { 402, 420 },
      -1,
   } },
   /* replace147_0_1_1_0_0 -> 16 in the cache */
   /* replace147_0_1_1_0_1 -> 405 in the cache */
   /* replace147_0_1_1_0 -> 406 in the cache */
   /* replace147_0_1_1_1_0 -> 36 in the cache */
   /* replace147_0_1_1_1_1 -> 405 in the cache */
   /* replace147_0_1_1_1 -> 423 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      7, 1,
      { 406, 423 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      5, 3,
      { 431, 432 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 7,
      { 430, 433 },
      -1,
   } },
   /* replace147_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 8,
      { 434, 135 },
      -1,
   } },

   /* ('sdot_2x16_iadd', 'a', 'b', 'c') => ('iadd', ('iadd', ('imul', ('extract_i16', 'a', 0), ('extract_i16', 'b', 0)), ('imul', ('extract_i16', 'a', 1), ('extract_i16', 'b', 1))), 'c') */
   /* search148_0 -> 16 in the cache */
   /* search148_1 -> 36 in the cache */
   /* search148_2 -> 135 in the cache */
   /* search148 -> 343 in the cache */

   /* replace148_0_0_0_0 -> 16 in the cache */
   /* replace148_0_0_0_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 16, 122 },
      -1,
   } },
   /* replace148_0_0_1_0 -> 36 in the cache */
   /* replace148_0_0_1_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 36, 122 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      2, 1,
      { 436, 437 },
      -1,
   } },
   /* replace148_0_1_0_0 -> 16 in the cache */
   /* replace148_0_1_0_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 16, 396 },
      -1,
   } },
   /* replace148_0_1_1_0 -> 36 in the cache */
   /* replace148_0_1_1_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 36, 396 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      3, 1,
      { 439, 440 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 3,
      { 438, 441 },
      -1,
   } },
   /* replace148_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 4,
      { 442, 135 },
      -1,
   } },

   /* ('udot_2x16_uadd', 'a', 'b', 'c') => ('iadd', ('iadd', ('imul', ('extract_u16', 'a', 0), ('extract_u16', 'b', 0)), ('imul', ('extract_u16', 'a', 1), ('extract_u16', 'b', 1))), 'c') */
   /* search149_0 -> 16 in the cache */
   /* search149_1 -> 36 in the cache */
   /* search149_2 -> 135 in the cache */
   /* search149 -> 346 in the cache */

   /* replace149_0_0_0_0 -> 16 in the cache */
   /* replace149_0_0_0_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 16, 122 },
      -1,
   } },
   /* replace149_0_0_1_0 -> 36 in the cache */
   /* replace149_0_0_1_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 36, 122 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      2, 1,
      { 444, 445 },
      -1,
   } },
   /* replace149_0_1_0_0 -> 16 in the cache */
   /* replace149_0_1_0_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 16, 396 },
      -1,
   } },
   /* replace149_0_1_1_0 -> 36 in the cache */
   /* replace149_0_1_1_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 36, 396 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      3, 1,
      { 447, 448 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 3,
      { 446, 449 },
      -1,
   } },
   /* replace149_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 4,
      { 450, 135 },
      -1,
   } },

   /* ('udot_4x8_uadd_sat', 'a', 'b', '#c(is_ult_0xfffc07fc)') => ('udot_4x8_uadd', 'a', 'b', 'c') */
   /* search150_0 -> 16 in the cache */
   /* search150_1 -> 36 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* c */
      true,
      nir_type_invalid,
      7,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_4x8_uadd_sat,
      0, 1,
      { 16, 36, 452 },
      -1,
   } },

   /* replace150_0 -> 16 in the cache */
   /* replace150_1 -> 36 in the cache */
   /* replace150_2 -> 135 in the cache */
   /* replace150 -> 334 in the cache */

   /* ('udot_4x8_uadd_sat', 'a', 'b', 'c') => ('uadd_sat', ('iadd', ('iadd', ('imul', ('extract_u8', 'a', 0), ('extract_u8', 'b', 0)), ('imul', ('extract_u8', 'a', 1), ('extract_u8', 'b', 1))), ('iadd', ('imul', ('extract_u8', 'a', 2), ('extract_u8', 'b', 2)), ('imul', ('extract_u8', 'a', 3), ('extract_u8', 'b', 3)))), 'c') */
   /* search151_0 -> 16 in the cache */
   /* search151_1 -> 36 in the cache */
   /* search151_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_4x8_uadd_sat,
      0, 1,
      { 16, 36, 135 },
      -1,
   } },

   /* replace151_0_0_0_0_0 -> 16 in the cache */
   /* replace151_0_0_0_0_1 -> 122 in the cache */
   /* replace151_0_0_0_0 -> 412 in the cache */
   /* replace151_0_0_0_1_0 -> 36 in the cache */
   /* replace151_0_0_0_1_1 -> 122 in the cache */
   /* replace151_0_0_0_1 -> 413 in the cache */
   /* replace151_0_0_0 -> 414 in the cache */
   /* replace151_0_0_1_0_0 -> 16 in the cache */
   /* replace151_0_0_1_0_1 -> 396 in the cache */
   /* replace151_0_0_1_0 -> 415 in the cache */
   /* replace151_0_0_1_1_0 -> 36 in the cache */
   /* replace151_0_0_1_1_1 -> 396 in the cache */
   /* replace151_0_0_1_1 -> 416 in the cache */
   /* replace151_0_0_1 -> 417 in the cache */
   /* replace151_0_0 -> 418 in the cache */
   /* replace151_0_1_0_0_0 -> 16 in the cache */
   /* replace151_0_1_0_0_1 -> 401 in the cache */
   /* replace151_0_1_0_0 -> 419 in the cache */
   /* replace151_0_1_0_1_0 -> 36 in the cache */
   /* replace151_0_1_0_1_1 -> 401 in the cache */
   /* replace151_0_1_0_1 -> 420 in the cache */
   /* replace151_0_1_0 -> 421 in the cache */
   /* replace151_0_1_1_0_0 -> 16 in the cache */
   /* replace151_0_1_1_0_1 -> 405 in the cache */
   /* replace151_0_1_1_0 -> 422 in the cache */
   /* replace151_0_1_1_1_0 -> 36 in the cache */
   /* replace151_0_1_1_1_1 -> 405 in the cache */
   /* replace151_0_1_1_1 -> 423 in the cache */
   /* replace151_0_1_1 -> 424 in the cache */
   /* replace151_0_1 -> 425 in the cache */
   /* replace151_0 -> 426 in the cache */
   /* replace151_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_uadd_sat,
      0, 8,
      { 426, 135 },
      -1,
   } },

   /* ('sdot_4x8_iadd_sat', 'a', 'b', 'c') => ('iadd_sat', ('iadd', ('iadd', ('imul', ('extract_i8', 'a', 0), ('extract_i8', 'b', 0)), ('imul', ('extract_i8', 'a', 1), ('extract_i8', 'b', 1))), ('iadd', ('imul', ('extract_i8', 'a', 2), ('extract_i8', 'b', 2)), ('imul', ('extract_i8', 'a', 3), ('extract_i8', 'b', 3)))), 'c') */
   /* search152_0 -> 16 in the cache */
   /* search152_1 -> 36 in the cache */
   /* search152_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_4x8_iadd_sat,
      0, 1,
      { 16, 36, 135 },
      -1,
   } },

   /* replace152_0_0_0_0_0 -> 16 in the cache */
   /* replace152_0_0_0_0_1 -> 122 in the cache */
   /* replace152_0_0_0_0 -> 393 in the cache */
   /* replace152_0_0_0_1_0 -> 36 in the cache */
   /* replace152_0_0_0_1_1 -> 122 in the cache */
   /* replace152_0_0_0_1 -> 394 in the cache */
   /* replace152_0_0_0 -> 395 in the cache */
   /* replace152_0_0_1_0_0 -> 16 in the cache */
   /* replace152_0_0_1_0_1 -> 396 in the cache */
   /* replace152_0_0_1_0 -> 397 in the cache */
   /* replace152_0_0_1_1_0 -> 36 in the cache */
   /* replace152_0_0_1_1_1 -> 396 in the cache */
   /* replace152_0_0_1_1 -> 398 in the cache */
   /* replace152_0_0_1 -> 399 in the cache */
   /* replace152_0_0 -> 400 in the cache */
   /* replace152_0_1_0_0_0 -> 16 in the cache */
   /* replace152_0_1_0_0_1 -> 401 in the cache */
   /* replace152_0_1_0_0 -> 402 in the cache */
   /* replace152_0_1_0_1_0 -> 36 in the cache */
   /* replace152_0_1_0_1_1 -> 401 in the cache */
   /* replace152_0_1_0_1 -> 403 in the cache */
   /* replace152_0_1_0 -> 404 in the cache */
   /* replace152_0_1_1_0_0 -> 16 in the cache */
   /* replace152_0_1_1_0_1 -> 405 in the cache */
   /* replace152_0_1_1_0 -> 406 in the cache */
   /* replace152_0_1_1_1_0 -> 36 in the cache */
   /* replace152_0_1_1_1_1 -> 405 in the cache */
   /* replace152_0_1_1_1 -> 407 in the cache */
   /* replace152_0_1_1 -> 408 in the cache */
   /* replace152_0_1 -> 409 in the cache */
   /* replace152_0 -> 410 in the cache */
   /* replace152_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd_sat,
      0, 8,
      { 410, 135 },
      -1,
   } },

   /* ('sudot_4x8_iadd_sat', 'a', 'b', 'c') => ('iadd_sat', ('iadd', ('iadd', ('imul', ('extract_i8', 'a', 0), ('extract_u8', 'b', 0)), ('imul', ('extract_i8', 'a', 1), ('extract_u8', 'b', 1))), ('iadd', ('imul', ('extract_i8', 'a', 2), ('extract_u8', 'b', 2)), ('imul', ('extract_i8', 'a', 3), ('extract_u8', 'b', 3)))), 'c') */
   /* search153_0 -> 16 in the cache */
   /* search153_1 -> 36 in the cache */
   /* search153_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd_sat,
      -1, 0,
      { 16, 36, 135 },
      -1,
   } },

   /* replace153_0_0_0_0_0 -> 16 in the cache */
   /* replace153_0_0_0_0_1 -> 122 in the cache */
   /* replace153_0_0_0_0 -> 393 in the cache */
   /* replace153_0_0_0_1_0 -> 36 in the cache */
   /* replace153_0_0_0_1_1 -> 122 in the cache */
   /* replace153_0_0_0_1 -> 413 in the cache */
   /* replace153_0_0_0 -> 428 in the cache */
   /* replace153_0_0_1_0_0 -> 16 in the cache */
   /* replace153_0_0_1_0_1 -> 396 in the cache */
   /* replace153_0_0_1_0 -> 397 in the cache */
   /* replace153_0_0_1_1_0 -> 36 in the cache */
   /* replace153_0_0_1_1_1 -> 396 in the cache */
   /* replace153_0_0_1_1 -> 416 in the cache */
   /* replace153_0_0_1 -> 429 in the cache */
   /* replace153_0_0 -> 430 in the cache */
   /* replace153_0_1_0_0_0 -> 16 in the cache */
   /* replace153_0_1_0_0_1 -> 401 in the cache */
   /* replace153_0_1_0_0 -> 402 in the cache */
   /* replace153_0_1_0_1_0 -> 36 in the cache */
   /* replace153_0_1_0_1_1 -> 401 in the cache */
   /* replace153_0_1_0_1 -> 420 in the cache */
   /* replace153_0_1_0 -> 431 in the cache */
   /* replace153_0_1_1_0_0 -> 16 in the cache */
   /* replace153_0_1_1_0_1 -> 405 in the cache */
   /* replace153_0_1_1_0 -> 406 in the cache */
   /* replace153_0_1_1_1_0 -> 36 in the cache */
   /* replace153_0_1_1_1_1 -> 405 in the cache */
   /* replace153_0_1_1_1 -> 423 in the cache */
   /* replace153_0_1_1 -> 432 in the cache */
   /* replace153_0_1 -> 433 in the cache */
   /* replace153_0 -> 434 in the cache */
   /* replace153_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd_sat,
      0, 8,
      { 434, 135 },
      -1,
   } },

   /* ('udot_2x16_uadd_sat', 'a', 'b', 'c') => ('uadd_sat', ('iadd', ('imul', ('extract_u16', 'a', 0), ('extract_u16', 'b', 0)), ('imul', ('extract_u16', 'a', 1), ('extract_u16', 'b', 1))), 'c') */
   /* search154_0 -> 16 in the cache */
   /* search154_1 -> 36 in the cache */
   /* search154_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_2x16_uadd_sat,
      0, 1,
      { 16, 36, 135 },
      -1,
   } },

   /* replace154_0_0_0_0 -> 16 in the cache */
   /* replace154_0_0_0_1 -> 122 in the cache */
   /* replace154_0_0_0 -> 444 in the cache */
   /* replace154_0_0_1_0 -> 36 in the cache */
   /* replace154_0_0_1_1 -> 122 in the cache */
   /* replace154_0_0_1 -> 445 in the cache */
   /* replace154_0_0 -> 446 in the cache */
   /* replace154_0_1_0_0 -> 16 in the cache */
   /* replace154_0_1_0_1 -> 396 in the cache */
   /* replace154_0_1_0 -> 447 in the cache */
   /* replace154_0_1_1_0 -> 36 in the cache */
   /* replace154_0_1_1_1 -> 396 in the cache */
   /* replace154_0_1_1 -> 448 in the cache */
   /* replace154_0_1 -> 449 in the cache */
   /* replace154_0 -> 450 in the cache */
   /* replace154_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_uadd_sat,
      0, 4,
      { 450, 135 },
      -1,
   } },

   /* ('sdot_2x16_iadd_sat', 'a', 'b', 'c') => ('iadd_sat', ('iadd', ('imul', ('extract_i16', 'a', 0), ('extract_i16', 'b', 0)), ('imul', ('extract_i16', 'a', 1), ('extract_i16', 'b', 1))), 'c') */
   /* search155_0 -> 16 in the cache */
   /* search155_1 -> 36 in the cache */
   /* search155_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_2x16_iadd_sat,
      0, 1,
      { 16, 36, 135 },
      -1,
   } },

   /* replace155_0_0_0_0 -> 16 in the cache */
   /* replace155_0_0_0_1 -> 122 in the cache */
   /* replace155_0_0_0 -> 436 in the cache */
   /* replace155_0_0_1_0 -> 36 in the cache */
   /* replace155_0_0_1_1 -> 122 in the cache */
   /* replace155_0_0_1 -> 437 in the cache */
   /* replace155_0_0 -> 438 in the cache */
   /* replace155_0_1_0_0 -> 16 in the cache */
   /* replace155_0_1_0_1 -> 396 in the cache */
   /* replace155_0_1_0 -> 439 in the cache */
   /* replace155_0_1_1_0 -> 36 in the cache */
   /* replace155_0_1_1_1 -> 396 in the cache */
   /* replace155_0_1_1 -> 440 in the cache */
   /* replace155_0_1 -> 441 in the cache */
   /* replace155_0 -> 442 in the cache */
   /* replace155_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd_sat,
      0, 4,
      { 442, 135 },
      -1,
   } },

   /* ('~flrp@16', 'a', 'b', ('b2f', 'c@1')) => ('bcsel', 'c', 'b', 'a') */
   /* search156_0 -> 11 in the cache */
   /* search156_1 -> 258 in the cache */
   { .variable = {
      { nir_search_value_variable, 1 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 464 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 11, 258, 465 },
      -1,
   } },

   /* replace156_0 -> 464 in the cache */
   /* replace156_1 -> 258 in the cache */
   /* replace156_2 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 464, 258, 11 },
      -1,
   } },

   /* ('~flrp@16', 'a', ('fadd', 'a', 'b'), 'c') => ('fadd', ('fmul', 'b', 'c'), 'a') */
   /* search157_0 -> 11 in the cache */
   /* search157_1_0 -> 11 in the cache */
   /* search157_1_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 11, 258 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 16 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 1,
      { 11, 468, 469 },
      -1,
   } },

   /* replace157_0_0 -> 258 in the cache */
   /* replace157_0_1 -> 469 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 258, 469 },
      -1,
   } },
   /* replace157_1 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 471, 11 },
      -1,
   } },

   /* ('~flrp@16', ('fadd(is_used_once)', 'a', 'b'), ('fadd(is_used_once)', 'a', 'c'), 'd') => ('fadd', ('flrp', 'b', 'c', 'd'), 'a') */
   /* search158_0_0 -> 11 in the cache */
   /* search158_0_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 11, 258 },
      0,
   } },
   /* search158_1_0 -> 11 in the cache */
   /* search158_1_1 -> 469 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 11, 469 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, 16 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 2,
      { 473, 474, 475 },
      -1,
   } },

   /* replace158_0_0 -> 258 in the cache */
   /* replace158_0_1 -> 469 in the cache */
   /* replace158_0_2 -> 475 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 258, 469, 475 },
      -1,
   } },
   /* replace158_1 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 477, 11 },
      -1,
   } },

   /* ('~flrp@16', 'a', ('fmul(is_used_once)', 'a', 'b'), 'c') => ('fmul', ('flrp', 1.0, 'b', 'c'), 'a') */
   /* search159_0 -> 11 in the cache */
   /* search159_1_0 -> 11 in the cache */
   /* search159_1_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 11, 258 },
      0,
   } },
   /* search159_2 -> 469 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 1,
      { 11, 479, 469 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   /* replace159_0_1 -> 258 in the cache */
   /* replace159_0_2 -> 469 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 481, 258, 469 },
      -1,
   } },
   /* replace159_1 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 482, 11 },
      -1,
   } },

   /* ('~fadd@16', ('fmul', 'a', ('fadd', 1.0, ('fneg', 'c'))), ('fmul', 'b', 'c')) => ('flrp', 'a', 'b', 'c') */
   /* search160_0_0 -> 11 in the cache */
   /* search160_0_1_0 -> 481 in the cache */
   { .variable = {
      { nir_search_value_variable, 16 },
      1, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 484 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 481, 485 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 11, 486 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 16 },
      2, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search160_1_1 -> 484 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 488, 484 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 487, 489 },
      -1,
   } },

   /* replace160_0 -> 11 in the cache */
   /* replace160_1 -> 488 in the cache */
   /* replace160_2 -> 484 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 11, 488, 484 },
      -1,
   } },

   /* ('~fadd@16', ('fmul', 'a', ('fsat', ('fadd', 1.0, ('fneg', 'c')))), ('fmul', 'b', ('fsat', 'c'))) => ('flrp', 'a', 'b', ('fsat', 'c')) */
   /* search161_0_0 -> 11 in the cache */
   /* search161_0_1_0_0 -> 481 in the cache */
   /* search161_0_1_0_1_0 -> 484 in the cache */
   /* search161_0_1_0_1 -> 485 in the cache */
   /* search161_0_1_0 -> 486 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 486 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 11, 492 },
      -1,
   } },
   /* search161_1_0 -> 488 in the cache */
   /* search161_1_1_0 -> 484 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 484 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 488, 494 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 493, 495 },
      -1,
   } },

   /* replace161_0 -> 11 in the cache */
   /* replace161_1 -> 488 in the cache */
   /* replace161_2_0 -> 484 in the cache */
   /* replace161_2 -> 494 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 11, 488, 494 },
      -1,
   } },

   /* ('~fadd@16', 'a', ('fmul', 'c', ('fadd', 'b', ('fneg', 'a')))) => ('flrp', 'a', 'b', 'c') */
   /* search162_0 -> 11 in the cache */
   /* search162_1_0 -> 484 in the cache */
   /* search162_1_1_0 -> 488 in the cache */
   /* search162_1_1_1_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 488, 498 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 484, 499 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 11, 500 },
      -1,
   } },

   /* replace162_0 -> 11 in the cache */
   /* replace162_1 -> 488 in the cache */
   /* replace162_2 -> 484 in the cache */
   /* replace162 -> 491 in the cache */

   /* ('~fadd@16', ('fmul', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1')))), ('fmul', 'b', ('b2f', 'c'))) => ('bcsel', 'c', 'b', 'a') */
   /* search163_0_0 -> 11 in the cache */
   /* search163_0_1_0 -> 481 in the cache */
   { .variable = {
      { nir_search_value_variable, 1 },
      1, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 502 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 503 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 481, 504 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 11, 505 },
      -1,
   } },
   /* search163_1_0 -> 488 in the cache */
   /* search163_1_1_0 -> 502 in the cache */
   /* search163_1_1 -> 503 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 488, 503 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 506, 507 },
      -1,
   } },

   /* replace163_0 -> 502 in the cache */
   /* replace163_1 -> 488 in the cache */
   /* replace163_2 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 502, 488, 11 },
      -1,
   } },

   /* ('~fadd@16', 'a', ('fmul', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')))) => ('bcsel', 'c', 'b', 'a') */
   /* search164_0 -> 11 in the cache */
   /* search164_1_0_0 -> 502 in the cache */
   /* search164_1_0 -> 503 in the cache */
   /* search164_1_1_0 -> 488 in the cache */
   /* search164_1_1_1_0 -> 11 in the cache */
   /* search164_1_1_1 -> 498 in the cache */
   /* search164_1_1 -> 499 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 503, 499 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 11, 510 },
      -1,
   } },

   /* replace164_0 -> 502 in the cache */
   /* replace164_1 -> 488 in the cache */
   /* replace164_2 -> 11 in the cache */
   /* replace164 -> 509 in the cache */

   /* ('~ffma@16', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1'))), ('fmul', 'b', ('b2f', 'c@1'))) => ('bcsel', 'c', 'b', 'a') */
   /* search165_0 -> 11 in the cache */
   /* search165_1_0 -> 481 in the cache */
   /* search165_1_1_0_0 -> 502 in the cache */
   /* search165_1_1_0 -> 503 in the cache */
   /* search165_1_1 -> 504 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 481, 504 },
      -1,
   } },
   /* search165_2_0 -> 488 in the cache */
   /* search165_2_1_0 -> 502 in the cache */
   /* search165_2_1 -> 503 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 488, 503 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 3,
      { 11, 512, 513 },
      -1,
   } },

   /* replace165_0 -> 502 in the cache */
   /* replace165_1 -> 488 in the cache */
   /* replace165_2 -> 11 in the cache */
   /* replace165 -> 509 in the cache */

   /* ('~ffma@16', 'b', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), ('b2f', 'c@1'), 'a')) => ('bcsel', 'c', 'b', 'a') */
   { .variable = {
      { nir_search_value_variable, 16 },
      0, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search166_1_0 -> 502 in the cache */
   /* search166_1 -> 503 in the cache */
   { .variable = {
      { nir_search_value_variable, 16 },
      2, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 516 },
      -1,
   } },
   /* search166_2_1_0 -> 502 in the cache */
   /* search166_2_1 -> 503 in the cache */
   /* search166_2_2 -> 516 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 517, 503, 516 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 515, 503, 518 },
      -1,
   } },

   /* replace166_0 -> 502 in the cache */
   /* replace166_1 -> 515 in the cache */
   /* replace166_2 -> 516 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 502, 515, 516 },
      -1,
   } },

   /* ('~ffma@16', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')), 'a') => ('bcsel', 'c', 'b', 'a') */
   { .variable = {
      { nir_search_value_variable, 1 },
      0, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 521 },
      -1,
   } },
   /* search167_1_0 -> 258 in the cache */
   /* search167_1_1_0 -> 516 in the cache */
   /* search167_1_1 -> 517 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 258, 517 },
      -1,
   } },
   /* search167_2 -> 516 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 522, 523, 516 },
      -1,
   } },

   /* replace167_0 -> 521 in the cache */
   /* replace167_1 -> 258 in the cache */
   /* replace167_2 -> 516 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 258, 516 },
      -1,
   } },

   /* ('~ffma@16', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), 'b', 'd'), ('fmul', 'a', 'b')) => ('bcsel', 'c', 'd', ('fmul', 'a', 'b')) */
   /* search168_0_0 -> 521 in the cache */
   /* search168_0 -> 522 in the cache */
   { .variable = {
      { nir_search_value_variable, 16 },
      1, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 526 },
      -1,
   } },
   /* search168_1_1 -> 488 in the cache */
   /* search168_1_2 -> 475 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 527, 488, 475 },
      -1,
   } },
   /* search168_2_0 -> 526 in the cache */
   /* search168_2_1 -> 488 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 526, 488 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 3,
      { 522, 528, 529 },
      -1,
   } },

   /* replace168_0 -> 521 in the cache */
   /* replace168_1 -> 475 in the cache */
   /* replace168_2_0 -> 526 in the cache */
   /* replace168_2_1 -> 488 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 526, 488 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 521, 475, 531 },
      -1,
   } },

   /* ('~fadd@16', 1.0, ('fneg', ('fmul', ('fadd', 1.0, ('fneg', 'a')), ('fadd', 1.0, ('fneg', 'b'))))) => ('flrp', 'b', 1.0, 'a') */
   /* search169_0 -> 481 in the cache */
   /* search169_1_0_0_0 -> 481 in the cache */
   /* search169_1_0_0_1_0 -> 11 in the cache */
   /* search169_1_0_0_1 -> 498 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 481, 498 },
      -1,
   } },
   /* search169_1_0_1_0 -> 481 in the cache */
   /* search169_1_0_1_1_0 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 258 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      3, 1,
      { 481, 534 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 3,
      { 533, 535 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 3,
      { 536 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 481, 537 },
      -1,
   } },

   /* replace169_0 -> 258 in the cache */
   /* replace169_1 -> 481 in the cache */
   /* replace169_2 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 258, 481, 11 },
      -1,
   } },

   /* ('~flrp@32', 'a', 'b', ('b2f', 'c@1')) => ('bcsel', 'c', 'b', 'a') */
   /* search170_0 -> 16 in the cache */
   /* search170_1 -> 36 in the cache */
   /* search170_2_0 -> 464 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 464 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 16, 36, 540 },
      -1,
   } },

   /* replace170_0 -> 464 in the cache */
   /* replace170_1 -> 36 in the cache */
   /* replace170_2 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 464, 36, 16 },
      -1,
   } },

   /* ('~flrp@32', 'a', ('fadd', 'a', 'b'), 'c') => ('fadd', ('fmul', 'b', 'c'), 'a') */
   /* search171_0 -> 16 in the cache */
   /* search171_1_0 -> 16 in the cache */
   /* search171_1_1 -> 36 in the cache */
   /* search171_1 -> 274 in the cache */
   /* search171_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 1,
      { 16, 274, 135 },
      -1,
   } },

   /* replace171_0_0 -> 36 in the cache */
   /* replace171_0_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 36, 135 },
      -1,
   } },
   /* replace171_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 544, 16 },
      -1,
   } },

   /* ('~flrp@32', ('fadd(is_used_once)', 'a', 'b'), ('fadd(is_used_once)', 'a', 'c'), 'd') => ('fadd', ('flrp', 'b', 'c', 'd'), 'a') */
   /* search172_0_0 -> 16 in the cache */
   /* search172_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 16, 36 },
      0,
   } },
   /* search172_1_0 -> 16 in the cache */
   /* search172_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 16, 135 },
      0,
   } },
   /* search172_2 -> 163 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 2,
      { 546, 547, 163 },
      -1,
   } },

   /* replace172_0_0 -> 36 in the cache */
   /* replace172_0_1 -> 135 in the cache */
   /* replace172_0_2 -> 163 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 36, 135, 163 },
      -1,
   } },
   /* replace172_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 549, 16 },
      -1,
   } },

   /* ('~flrp@32', 'a', ('fmul(is_used_once)', 'a', 'b'), 'c') => ('fmul', ('flrp', 1.0, 'b', 'c'), 'a') */
   /* search173_0 -> 16 in the cache */
   /* search173_1_0 -> 16 in the cache */
   /* search173_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 16, 36 },
      0,
   } },
   /* search173_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 1,
      { 16, 551, 135 },
      -1,
   } },

   /* replace173_0_0 -> 243 in the cache */
   /* replace173_0_1 -> 36 in the cache */
   /* replace173_0_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 243, 36, 135 },
      -1,
   } },
   /* replace173_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 553, 16 },
      -1,
   } },

   /* ('~fadd@32', ('fmul', 'a', ('fadd', 1.0, ('fneg', 'c'))), ('fmul', 'b', 'c')) => ('flrp', 'a', 'b', 'c') */
   /* search174_0_0 -> 16 in the cache */
   /* search174_0_1_0 -> 243 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 555 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 243, 556 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 16, 557 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search174_1_1 -> 555 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 559, 555 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 558, 560 },
      -1,
   } },

   /* replace174_0 -> 16 in the cache */
   /* replace174_1 -> 559 in the cache */
   /* replace174_2 -> 555 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 16, 559, 555 },
      -1,
   } },

   /* ('~fadd@32', ('fmul', 'a', ('fsat', ('fadd', 1.0, ('fneg', 'c')))), ('fmul', 'b', ('fsat', 'c'))) => ('flrp', 'a', 'b', ('fsat', 'c')) */
   /* search175_0_0 -> 16 in the cache */
   /* search175_0_1_0_0 -> 243 in the cache */
   /* search175_0_1_0_1_0 -> 555 in the cache */
   /* search175_0_1_0_1 -> 556 in the cache */
   /* search175_0_1_0 -> 557 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 557 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 16, 563 },
      -1,
   } },
   /* search175_1_0 -> 559 in the cache */
   /* search175_1_1_0 -> 555 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 555 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 559, 565 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 564, 566 },
      -1,
   } },

   /* replace175_0 -> 16 in the cache */
   /* replace175_1 -> 559 in the cache */
   /* replace175_2_0 -> 555 in the cache */
   /* replace175_2 -> 565 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 16, 559, 565 },
      -1,
   } },

   /* ('~fadd@32', 'a', ('fmul', 'c', ('fadd', 'b', ('fneg', 'a')))) => ('flrp', 'a', 'b', 'c') */
   /* search176_0 -> 16 in the cache */
   /* search176_1_0 -> 555 in the cache */
   /* search176_1_1_0 -> 559 in the cache */
   /* search176_1_1_1_0 -> 16 in the cache */
   /* search176_1_1_1 -> 280 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 559, 280 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 555, 569 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 16, 570 },
      -1,
   } },

   /* replace176_0 -> 16 in the cache */
   /* replace176_1 -> 559 in the cache */
   /* replace176_2 -> 555 in the cache */
   /* replace176 -> 562 in the cache */

   /* ('~fadd@32', ('fmul', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1')))), ('fmul', 'b', ('b2f', 'c'))) => ('bcsel', 'c', 'b', 'a') */
   /* search177_0_0 -> 16 in the cache */
   /* search177_0_1_0 -> 243 in the cache */
   /* search177_0_1_1_0_0 -> 502 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 502 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 572 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 243, 573 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 16, 574 },
      -1,
   } },
   /* search177_1_0 -> 559 in the cache */
   /* search177_1_1_0 -> 502 in the cache */
   /* search177_1_1 -> 572 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 559, 572 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 575, 576 },
      -1,
   } },

   /* replace177_0 -> 502 in the cache */
   /* replace177_1 -> 559 in the cache */
   /* replace177_2 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 502, 559, 16 },
      -1,
   } },

   /* ('~fadd@32', 'a', ('fmul', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')))) => ('bcsel', 'c', 'b', 'a') */
   /* search178_0 -> 16 in the cache */
   /* search178_1_0_0 -> 502 in the cache */
   /* search178_1_0 -> 572 in the cache */
   /* search178_1_1_0 -> 559 in the cache */
   /* search178_1_1_1_0 -> 16 in the cache */
   /* search178_1_1_1 -> 280 in the cache */
   /* search178_1_1 -> 569 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 572, 569 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 16, 579 },
      -1,
   } },

   /* replace178_0 -> 502 in the cache */
   /* replace178_1 -> 559 in the cache */
   /* replace178_2 -> 16 in the cache */
   /* replace178 -> 578 in the cache */

   /* ('~ffma@32', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1'))), ('fmul', 'b', ('b2f', 'c@1'))) => ('bcsel', 'c', 'b', 'a') */
   /* search179_0 -> 16 in the cache */
   /* search179_1_0 -> 243 in the cache */
   /* search179_1_1_0_0 -> 502 in the cache */
   /* search179_1_1_0 -> 572 in the cache */
   /* search179_1_1 -> 573 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 243, 573 },
      -1,
   } },
   /* search179_2_0 -> 559 in the cache */
   /* search179_2_1_0 -> 502 in the cache */
   /* search179_2_1 -> 572 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 559, 572 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 3,
      { 16, 581, 582 },
      -1,
   } },

   /* replace179_0 -> 502 in the cache */
   /* replace179_1 -> 559 in the cache */
   /* replace179_2 -> 16 in the cache */
   /* replace179 -> 578 in the cache */

   /* ('~ffma@32', 'b', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), ('b2f', 'c@1'), 'a')) => ('bcsel', 'c', 'b', 'a') */
   /* search180_0 -> 373 in the cache */
   /* search180_1_0 -> 502 in the cache */
   /* search180_1 -> 572 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 584 },
      -1,
   } },
   /* search180_2_1_0 -> 502 in the cache */
   /* search180_2_1 -> 572 in the cache */
   /* search180_2_2 -> 584 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 585, 572, 584 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 373, 572, 586 },
      -1,
   } },

   /* replace180_0 -> 502 in the cache */
   /* replace180_1 -> 373 in the cache */
   /* replace180_2 -> 584 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 502, 373, 584 },
      -1,
   } },

   /* ('~ffma@32', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')), 'a') => ('bcsel', 'c', 'b', 'a') */
   /* search181_0_0 -> 521 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 521 },
      -1,
   } },
   /* search181_1_0 -> 36 in the cache */
   /* search181_1_1_0 -> 584 in the cache */
   /* search181_1_1 -> 585 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 36, 585 },
      -1,
   } },
   /* search181_2 -> 584 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 589, 590, 584 },
      -1,
   } },

   /* replace181_0 -> 521 in the cache */
   /* replace181_1 -> 36 in the cache */
   /* replace181_2 -> 584 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 36, 584 },
      -1,
   } },

   /* ('~ffma@32', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), 'b', 'd'), ('fmul', 'a', 'b')) => ('bcsel', 'c', 'd', ('fmul', 'a', 'b')) */
   /* search182_0_0 -> 521 in the cache */
   /* search182_0 -> 589 in the cache */
   /* search182_1_0_0 -> 183 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 183 },
      -1,
   } },
   /* search182_1_1 -> 559 in the cache */
   /* search182_1_2 -> 163 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 593, 559, 163 },
      -1,
   } },
   /* search182_2_0 -> 183 in the cache */
   /* search182_2_1 -> 559 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 183, 559 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 3,
      { 589, 594, 595 },
      -1,
   } },

   /* replace182_0 -> 521 in the cache */
   /* replace182_1 -> 163 in the cache */
   /* replace182_2_0 -> 183 in the cache */
   /* replace182_2_1 -> 559 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 183, 559 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 521, 163, 597 },
      -1,
   } },

   /* ('~fadd@32', 1.0, ('fneg', ('fmul', ('fadd', 1.0, ('fneg', 'a')), ('fadd', 1.0, ('fneg', 'b'))))) => ('flrp', 'b', 1.0, 'a') */
   /* search183_0 -> 243 in the cache */
   /* search183_1_0_0_0 -> 243 in the cache */
   /* search183_1_0_0_1_0 -> 16 in the cache */
   /* search183_1_0_0_1 -> 280 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 243, 280 },
      -1,
   } },
   /* search183_1_0_1_0 -> 243 in the cache */
   /* search183_1_0_1_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      3, 1,
      { 243, 600 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 3,
      { 599, 601 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 3,
      { 602 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 243, 603 },
      -1,
   } },

   /* replace183_0 -> 36 in the cache */
   /* replace183_1 -> 243 in the cache */
   /* replace183_2 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 36, 243, 16 },
      -1,
   } },

   /* ('~flrp@64', 'a', 'b', ('b2f', 'c@1')) => ('bcsel', 'c', 'b', 'a') */
   /* search184_0 -> 21 in the cache */
   /* search184_1 -> 41 in the cache */
   /* search184_2_0 -> 464 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 464 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 21, 41, 606 },
      -1,
   } },

   /* replace184_0 -> 464 in the cache */
   /* replace184_1 -> 41 in the cache */
   /* replace184_2 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 464, 41, 21 },
      -1,
   } },

   /* ('~flrp@64', 'a', ('fadd', 'a', 'b'), 'c') => ('fadd', ('fmul', 'b', 'c'), 'a') */
   /* search185_0 -> 21 in the cache */
   /* search185_1_0 -> 21 in the cache */
   /* search185_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 21, 41 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 64 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 1,
      { 21, 609, 610 },
      -1,
   } },

   /* replace185_0_0 -> 41 in the cache */
   /* replace185_0_1 -> 610 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 41, 610 },
      -1,
   } },
   /* replace185_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 612, 21 },
      -1,
   } },

   /* ('~flrp@64', ('fadd(is_used_once)', 'a', 'b'), ('fadd(is_used_once)', 'a', 'c'), 'd') => ('fadd', ('flrp', 'b', 'c', 'd'), 'a') */
   /* search186_0_0 -> 21 in the cache */
   /* search186_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 21, 41 },
      0,
   } },
   /* search186_1_0 -> 21 in the cache */
   /* search186_1_1 -> 610 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 21, 610 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, 64 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 2,
      { 614, 615, 616 },
      -1,
   } },

   /* replace186_0_0 -> 41 in the cache */
   /* replace186_0_1 -> 610 in the cache */
   /* replace186_0_2 -> 616 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 41, 610, 616 },
      -1,
   } },
   /* replace186_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 618, 21 },
      -1,
   } },

   /* ('~flrp@64', 'a', ('fmul(is_used_once)', 'a', 'b'), 'c') => ('fmul', ('flrp', 1.0, 'b', 'c'), 'a') */
   /* search187_0 -> 21 in the cache */
   /* search187_1_0 -> 21 in the cache */
   /* search187_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 21, 41 },
      0,
   } },
   /* search187_2 -> 610 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 1,
      { 21, 620, 610 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   /* replace187_0_1 -> 41 in the cache */
   /* replace187_0_2 -> 610 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 622, 41, 610 },
      -1,
   } },
   /* replace187_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 623, 21 },
      -1,
   } },

   /* ('~fadd@64', ('fmul', 'a', ('fadd', 1.0, ('fneg', 'c'))), ('fmul', 'b', 'c')) => ('flrp', 'a', 'b', 'c') */
   /* search188_0_0 -> 21 in the cache */
   /* search188_0_1_0 -> 622 in the cache */
   { .variable = {
      { nir_search_value_variable, 64 },
      1, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 625 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 622, 626 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 21, 627 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 64 },
      2, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search188_1_1 -> 625 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 629, 625 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 628, 630 },
      -1,
   } },

   /* replace188_0 -> 21 in the cache */
   /* replace188_1 -> 629 in the cache */
   /* replace188_2 -> 625 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 21, 629, 625 },
      -1,
   } },

   /* ('~fadd@64', ('fmul', 'a', ('fsat', ('fadd', 1.0, ('fneg', 'c')))), ('fmul', 'b', ('fsat', 'c'))) => ('flrp', 'a', 'b', ('fsat', 'c')) */
   /* search189_0_0 -> 21 in the cache */
   /* search189_0_1_0_0 -> 622 in the cache */
   /* search189_0_1_0_1_0 -> 625 in the cache */
   /* search189_0_1_0_1 -> 626 in the cache */
   /* search189_0_1_0 -> 627 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 627 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 21, 633 },
      -1,
   } },
   /* search189_1_0 -> 629 in the cache */
   /* search189_1_1_0 -> 625 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 625 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 629, 635 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 634, 636 },
      -1,
   } },

   /* replace189_0 -> 21 in the cache */
   /* replace189_1 -> 629 in the cache */
   /* replace189_2_0 -> 625 in the cache */
   /* replace189_2 -> 635 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 21, 629, 635 },
      -1,
   } },

   /* ('~fadd@64', 'a', ('fmul', 'c', ('fadd', 'b', ('fneg', 'a')))) => ('flrp', 'a', 'b', 'c') */
   /* search190_0 -> 21 in the cache */
   /* search190_1_0 -> 625 in the cache */
   /* search190_1_1_0 -> 629 in the cache */
   /* search190_1_1_1_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 21 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 629, 639 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 625, 640 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 21, 641 },
      -1,
   } },

   /* replace190_0 -> 21 in the cache */
   /* replace190_1 -> 629 in the cache */
   /* replace190_2 -> 625 in the cache */
   /* replace190 -> 632 in the cache */

   /* ('~fadd@64', ('fmul', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1')))), ('fmul', 'b', ('b2f', 'c'))) => ('bcsel', 'c', 'b', 'a') */
   /* search191_0_0 -> 21 in the cache */
   /* search191_0_1_0 -> 622 in the cache */
   /* search191_0_1_1_0_0 -> 502 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 502 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 643 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 622, 644 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 21, 645 },
      -1,
   } },
   /* search191_1_0 -> 629 in the cache */
   /* search191_1_1_0 -> 502 in the cache */
   /* search191_1_1 -> 643 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 629, 643 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 646, 647 },
      -1,
   } },

   /* replace191_0 -> 502 in the cache */
   /* replace191_1 -> 629 in the cache */
   /* replace191_2 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 502, 629, 21 },
      -1,
   } },

   /* ('~fadd@64', 'a', ('fmul', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')))) => ('bcsel', 'c', 'b', 'a') */
   /* search192_0 -> 21 in the cache */
   /* search192_1_0_0 -> 502 in the cache */
   /* search192_1_0 -> 643 in the cache */
   /* search192_1_1_0 -> 629 in the cache */
   /* search192_1_1_1_0 -> 21 in the cache */
   /* search192_1_1_1 -> 639 in the cache */
   /* search192_1_1 -> 640 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 643, 640 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 21, 650 },
      -1,
   } },

   /* replace192_0 -> 502 in the cache */
   /* replace192_1 -> 629 in the cache */
   /* replace192_2 -> 21 in the cache */
   /* replace192 -> 649 in the cache */

   /* ('~ffma@64', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1'))), ('fmul', 'b', ('b2f', 'c@1'))) => ('bcsel', 'c', 'b', 'a') */
   /* search193_0 -> 21 in the cache */
   /* search193_1_0 -> 622 in the cache */
   /* search193_1_1_0_0 -> 502 in the cache */
   /* search193_1_1_0 -> 643 in the cache */
   /* search193_1_1 -> 644 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 622, 644 },
      -1,
   } },
   /* search193_2_0 -> 629 in the cache */
   /* search193_2_1_0 -> 502 in the cache */
   /* search193_2_1 -> 643 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 629, 643 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 3,
      { 21, 652, 653 },
      -1,
   } },

   /* replace193_0 -> 502 in the cache */
   /* replace193_1 -> 629 in the cache */
   /* replace193_2 -> 21 in the cache */
   /* replace193 -> 649 in the cache */

   /* ('~ffma@64', 'b', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), ('b2f', 'c@1'), 'a')) => ('bcsel', 'c', 'b', 'a') */
   { .variable = {
      { nir_search_value_variable, 64 },
      0, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search194_1_0 -> 502 in the cache */
   /* search194_1 -> 643 in the cache */
   { .variable = {
      { nir_search_value_variable, 64 },
      2, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 656 },
      -1,
   } },
   /* search194_2_1_0 -> 502 in the cache */
   /* search194_2_1 -> 643 in the cache */
   /* search194_2_2 -> 656 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 657, 643, 656 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 655, 643, 658 },
      -1,
   } },

   /* replace194_0 -> 502 in the cache */
   /* replace194_1 -> 655 in the cache */
   /* replace194_2 -> 656 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 502, 655, 656 },
      -1,
   } },

   /* ('~ffma@64', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')), 'a') => ('bcsel', 'c', 'b', 'a') */
   /* search195_0_0 -> 521 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 521 },
      -1,
   } },
   /* search195_1_0 -> 41 in the cache */
   /* search195_1_1_0 -> 656 in the cache */
   /* search195_1_1 -> 657 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 41, 657 },
      -1,
   } },
   /* search195_2 -> 656 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 661, 662, 656 },
      -1,
   } },

   /* replace195_0 -> 521 in the cache */
   /* replace195_1 -> 41 in the cache */
   /* replace195_2 -> 656 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 41, 656 },
      -1,
   } },

   /* ('~ffma@64', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), 'b', 'd'), ('fmul', 'a', 'b')) => ('bcsel', 'c', 'd', ('fmul', 'a', 'b')) */
   /* search196_0_0 -> 521 in the cache */
   /* search196_0 -> 661 in the cache */
   { .variable = {
      { nir_search_value_variable, 64 },
      1, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 665 },
      -1,
   } },
   /* search196_1_1 -> 629 in the cache */
   /* search196_1_2 -> 616 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 666, 629, 616 },
      -1,
   } },
   /* search196_2_0 -> 665 in the cache */
   /* search196_2_1 -> 629 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 665, 629 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 3,
      { 661, 667, 668 },
      -1,
   } },

   /* replace196_0 -> 521 in the cache */
   /* replace196_1 -> 616 in the cache */
   /* replace196_2_0 -> 665 in the cache */
   /* replace196_2_1 -> 629 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 665, 629 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 521, 616, 670 },
      -1,
   } },

   /* ('~fadd@64', 1.0, ('fneg', ('fmul', ('fadd', 1.0, ('fneg', 'a')), ('fadd', 1.0, ('fneg', 'b'))))) => ('flrp', 'b', 1.0, 'a') */
   /* search197_0 -> 622 in the cache */
   /* search197_1_0_0_0 -> 622 in the cache */
   /* search197_1_0_0_1_0 -> 21 in the cache */
   /* search197_1_0_0_1 -> 639 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 622, 639 },
      -1,
   } },
   /* search197_1_0_1_0 -> 622 in the cache */
   /* search197_1_0_1_1_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      3, 1,
      { 622, 673 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 3,
      { 672, 674 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 3,
      { 675 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 622, 676 },
      -1,
   } },

   /* replace197_0 -> 41 in the cache */
   /* replace197_1 -> 622 in the cache */
   /* replace197_2 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 41, 622, 21 },
      -1,
   } },

   /* ('~flrp', ('fmul(is_used_once)', 'a', 'b'), ('fmul(is_used_once)', 'a', 'c'), 'd') => ('fmul', ('flrp', 'b', 'c', 'd'), 'a') */
   /* search198_0_0 -> 149 in the cache */
   /* search198_0_1 -> 150 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 149, 150 },
      0,
   } },
   /* search198_1_0 -> 149 in the cache */
   /* search198_1_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 149, 152 },
      0,
   } },
   /* search198_2 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 2,
      { 679, 680, 153 },
      -1,
   } },

   /* replace198_0_0 -> 150 in the cache */
   /* replace198_0_1 -> 152 in the cache */
   /* replace198_0_2 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 150, 152, 153 },
      -1,
   } },
   /* replace198_1 -> 149 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 682, 149 },
      -1,
   } },

   /* ('~flrp', 'a', 0.0, 'c') => ('fadd', ('fmul', ('fneg', 'a'), 'c'), 'a') */
   /* search199_0 -> 0 in the cache */
   /* search199_1 -> 256 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 0, 256, 684 },
      -1,
   } },

   /* replace199_0_0_0 -> 0 in the cache */
   /* replace199_0_0 -> 212 in the cache */
   /* replace199_0_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 212, 684 },
      -1,
   } },
   /* replace199_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 686, 0 },
      -1,
   } },

   /* ('ftrunc@16', 'a') => ('bcsel', ('flt', 'a', 0.0), ('fneg', ('ffloor', ('fabs', 'a'))), ('ffloor', ('fabs', 'a'))) */
   /* search200_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ftrunc,
      -1, 0,
      { 11 },
      -1,
   } },

   /* replace200_0_0 -> 11 in the cache */
   /* replace200_0_1 -> 113 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 11, 113 },
      -1,
   } },
   /* replace200_1_0_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 0,
      { 690 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 691 },
      -1,
   } },
   /* replace200_2_0_0 -> 11 in the cache */
   /* replace200_2_0 -> 690 in the cache */
   /* replace200_2 -> 691 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 689, 692, 691 },
      -1,
   } },

   /* ('ftrunc@32', 'a') => ('bcsel', ('flt', 'a', 0.0), ('fneg', ('ffloor', ('fabs', 'a'))), ('ffloor', ('fabs', 'a'))) */
   /* search201_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ftrunc,
      -1, 0,
      { 16 },
      -1,
   } },

   /* replace201_0_0 -> 16 in the cache */
   /* replace201_0_1 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 16, 115 },
      -1,
   } },
   /* replace201_1_0_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 0,
      { 696 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 697 },
      -1,
   } },
   /* replace201_2_0_0 -> 16 in the cache */
   /* replace201_2_0 -> 696 in the cache */
   /* replace201_2 -> 697 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 695, 698, 697 },
      -1,
   } },

   /* ('ftrunc@64', 'a') => ('bcsel', ('flt', 'a', 0.0), ('fneg', ('ffloor', ('fabs', 'a'))), ('ffloor', ('fabs', 'a'))) */
   /* search202_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ftrunc,
      -1, 0,
      { 21 },
      -1,
   } },

   /* replace202_0_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 21, 701 },
      -1,
   } },
   /* replace202_1_0_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 21 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 0,
      { 703 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 704 },
      -1,
   } },
   /* replace202_2_0_0 -> 21 in the cache */
   /* replace202_2_0 -> 703 in the cache */
   /* replace202_2 -> 704 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 702, 705, 704 },
      -1,
   } },

   /* ('ffloor@16', 'a') => ('fsub', 'a', ('ffract', 'a')) */
   /* search203_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 0,
      { 11 },
      -1,
   } },

   /* replace203_0 -> 11 in the cache */
   /* replace203_1_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 11, 708 },
      -1,
   } },

   /* ('ffloor@32', 'a') => ('fsub', 'a', ('ffract', 'a')) */
   /* search204_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 0,
      { 16 },
      -1,
   } },

   /* replace204_0 -> 16 in the cache */
   /* replace204_1_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 16, 711 },
      -1,
   } },

   /* ('ffloor@64', 'a') => ('fsub', 'a', ('ffract', 'a')) */
   /* search205_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 0,
      { 21 },
      -1,
   } },

   /* replace205_0 -> 21 in the cache */
   /* replace205_1_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 0,
      { 21 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 21, 714 },
      -1,
   } },

   /* ('fadd@16', 'a', ('fadd@16', 'b', ('fneg', ('ffract', 'a')))) => ('fadd@16', 'b', ('ffloor', 'a')) */
   /* search206_0 -> 11 in the cache */
   /* search206_1_0 -> 258 in the cache */
   /* search206_1_1_0_0 -> 11 in the cache */
   /* search206_1_1_0 -> 708 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 708 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 258, 716 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 11, 717 },
      -1,
   } },

   /* replace206_0 -> 258 in the cache */
   /* replace206_1_0 -> 11 in the cache */
   /* replace206_1 -> 707 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 258, 707 },
      -1,
   } },

   /* ('fadd@32', 'a', ('fadd@32', 'b', ('fneg', ('ffract', 'a')))) => ('fadd@32', 'b', ('ffloor', 'a')) */
   /* search207_0 -> 16 in the cache */
   /* search207_1_0 -> 36 in the cache */
   /* search207_1_1_0_0 -> 16 in the cache */
   /* search207_1_1_0 -> 711 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 711 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 36, 720 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 16, 721 },
      -1,
   } },

   /* replace207_0 -> 36 in the cache */
   /* replace207_1_0 -> 16 in the cache */
   /* replace207_1 -> 710 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 36, 710 },
      -1,
   } },

   /* ('fadd@64', 'a', ('fadd@64', 'b', ('fneg', ('ffract', 'a')))) => ('fadd@64', 'b', ('ffloor', 'a')) */
   /* search208_0 -> 21 in the cache */
   /* search208_1_0 -> 41 in the cache */
   /* search208_1_1_0_0 -> 21 in the cache */
   /* search208_1_1_0 -> 714 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 714 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 41, 724 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 21, 725 },
      -1,
   } },

   /* replace208_0 -> 41 in the cache */
   /* replace208_1_0 -> 21 in the cache */
   /* replace208_1 -> 713 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 41, 713 },
      -1,
   } },

   /* ('fadd@16', 'a', ('fneg', ('ffract', 'a'))) => ('ffloor', 'a') */
   /* search209_0 -> 11 in the cache */
   /* search209_1_0_0 -> 11 in the cache */
   /* search209_1_0 -> 708 in the cache */
   /* search209_1 -> 716 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 11, 716 },
      -1,
   } },

   /* replace209_0 -> 11 in the cache */
   /* replace209 -> 707 in the cache */

   /* ('fadd@32', 'a', ('fneg', ('ffract', 'a'))) => ('ffloor', 'a') */
   /* search210_0 -> 16 in the cache */
   /* search210_1_0_0 -> 16 in the cache */
   /* search210_1_0 -> 711 in the cache */
   /* search210_1 -> 720 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 16, 720 },
      -1,
   } },

   /* replace210_0 -> 16 in the cache */
   /* replace210 -> 710 in the cache */

   /* ('fadd@64', 'a', ('fneg', ('ffract', 'a'))) => ('ffloor', 'a') */
   /* search211_0 -> 21 in the cache */
   /* search211_1_0_0 -> 21 in the cache */
   /* search211_1_0 -> 714 in the cache */
   /* search211_1 -> 724 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 21, 724 },
      -1,
   } },

   /* replace211_0 -> 21 in the cache */
   /* replace211 -> 713 in the cache */

   /* ('ffract@16', 'a') => ('fsub', 'a', ('ffloor', 'a')) */
   /* search212_0 -> 11 in the cache */
   /* search212 -> 708 in the cache */

   /* replace212_0 -> 11 in the cache */
   /* replace212_1_0 -> 11 in the cache */
   /* replace212_1 -> 707 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 11, 707 },
      -1,
   } },

   /* ('ffract@32', 'a') => ('fsub', 'a', ('ffloor', 'a')) */
   /* search213_0 -> 16 in the cache */
   /* search213 -> 711 in the cache */

   /* replace213_0 -> 16 in the cache */
   /* replace213_1_0 -> 16 in the cache */
   /* replace213_1 -> 710 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 16, 710 },
      -1,
   } },

   /* ('ffract@64', 'a') => ('fsub', 'a', ('ffloor', 'a')) */
   /* search214_0 -> 21 in the cache */
   /* search214 -> 714 in the cache */

   /* replace214_0 -> 21 in the cache */
   /* replace214_1_0 -> 21 in the cache */
   /* replace214_1 -> 713 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 21, 713 },
      -1,
   } },

   /* ('fceil', 'a') => ('fneg', ('ffloor', ('fneg', 'a'))) */
   /* search215_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fceil,
      -1, 0,
      { 32 },
      -1,
   } },

   /* replace215_0_0_0 -> 32 in the cache */
   /* replace215_0_0 -> 98 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 0,
      { 98 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 735 },
      -1,
   } },

   /* ('ffma@16', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c') */
   /* search216_0 -> 11 in the cache */
   /* search216_1 -> 258 in the cache */
   /* search216_2 -> 469 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 11, 258, 469 },
      -1,
   } },

   /* replace216_0_0 -> 11 in the cache */
   /* replace216_0_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 11, 258 },
      -1,
   } },
   /* replace216_1 -> 469 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 738, 469 },
      -1,
   } },

   /* ('ffma@32', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c') */
   /* search217_0 -> 16 in the cache */
   /* search217_1 -> 36 in the cache */
   /* search217_2 -> 135 in the cache */
   /* search217 -> 234 in the cache */

   /* replace217_0_0 -> 16 in the cache */
   /* replace217_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 16, 36 },
      -1,
   } },
   /* replace217_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 740, 135 },
      -1,
   } },

   /* ('ffma@64', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c') */
   /* search218_0 -> 21 in the cache */
   /* search218_1 -> 41 in the cache */
   /* search218_2 -> 610 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 21, 41, 610 },
      -1,
   } },

   /* replace218_0_0 -> 21 in the cache */
   /* replace218_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 21, 41 },
      -1,
   } },
   /* replace218_1 -> 610 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 743, 610 },
      -1,
   } },

   /* ('ffmaz', 'a', 'b', 'c') => ('fadd', ('fmulz', 'a', 'b'), 'c') */
   /* search219_0 -> 16 in the cache */
   /* search219_1 -> 36 in the cache */
   /* search219_2 -> 135 in the cache */
   /* search219 -> 387 in the cache */

   /* replace219_0_0 -> 16 in the cache */
   /* replace219_0_1 -> 36 in the cache */
   /* replace219_0 -> 134 in the cache */
   /* replace219_1 -> 135 in the cache */
   /* replace219 -> 288 in the cache */

   /* ('~ffma@16', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c') */
   /* search220_0 -> 11 in the cache */
   /* search220_1 -> 258 in the cache */
   /* search220_2 -> 469 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 11, 258, 469 },
      -1,
   } },

   /* replace220_0_0 -> 11 in the cache */
   /* replace220_0_1 -> 258 in the cache */
   /* replace220_0 -> 738 in the cache */
   /* replace220_1 -> 469 in the cache */
   /* replace220 -> 739 in the cache */

   /* ('~ffma@32', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c') */
   /* search221_0 -> 16 in the cache */
   /* search221_1 -> 36 in the cache */
   /* search221_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 16, 36, 135 },
      -1,
   } },

   /* replace221_0_0 -> 16 in the cache */
   /* replace221_0_1 -> 36 in the cache */
   /* replace221_0 -> 740 in the cache */
   /* replace221_1 -> 135 in the cache */
   /* replace221 -> 741 in the cache */

   /* ('~ffma@64', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c') */
   /* search222_0 -> 21 in the cache */
   /* search222_1 -> 41 in the cache */
   /* search222_2 -> 610 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 21, 41, 610 },
      -1,
   } },

   /* replace222_0_0 -> 21 in the cache */
   /* replace222_0_1 -> 41 in the cache */
   /* replace222_0 -> 743 in the cache */
   /* replace222_1 -> 610 in the cache */
   /* replace222 -> 744 in the cache */

   /* ('~ffmaz', 'a', 'b', 'c') => ('fadd', ('fmulz', 'a', 'b'), 'c') */
   /* search223_0 -> 16 in the cache */
   /* search223_1 -> 36 in the cache */
   /* search223_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 16, 36, 135 },
      -1,
   } },

   /* replace223_0_0 -> 16 in the cache */
   /* replace223_0_1 -> 36 in the cache */
   /* replace223_0 -> 134 in the cache */
   /* replace223_1 -> 135 in the cache */
   /* replace223 -> 288 in the cache */

   /* ('~fmul', ('fadd', ('iand', ('ineg', ('b2i', 'a@bool')), ('fmul', 'b', 'c')), '#d'), '#e') => ('bcsel', 'a', ('fmul', ('fadd', ('fmul', 'b', 'c'), 'd'), 'e'), ('fmul', 'd', 'e')) */
   { .variable = {
      { nir_search_value_variable, 1 },
      0, /* a */
      false,
      nir_type_bool,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 0,
      { 749 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 750 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 752, 753 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_iand,
      2, 2,
      { 751, 754 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      3, /* d */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 3,
      { 755, 756 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      4, /* e */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 4,
      { 757, 758 },
      -1,
   } },

   { .variable = {
      { nir_search_value_variable, 1 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* replace224_1_0_0_0 -> 752 in the cache */
   /* replace224_1_0_0_1 -> 753 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 752, 753 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 2,
      { 761, 762 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      4, /* e */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 3,
      { 763, 764 },
      -1,
   } },
   /* replace224_2_0 -> 762 in the cache */
   /* replace224_2_1 -> 764 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 762, 764 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 760, 765, 766 },
      -1,
   } },

   /* ('fdph', 'a', 'b') => ('fdot4', ('vec4', 'a.x', 'a.y', 'a.z', 1.0), 'b') */
   /* search225_0 -> 0 in the cache */
   /* search225_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdph,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {1},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {2},
   } },
   /* replace225_0_3 -> 270 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 769, 770, 771, 270 },
      -1,
   } },
   /* replace225_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot4,
      0, 1,
      { 772, 3 },
      -1,
   } },

   /* ('fdot4', ('vec4', 'a', 'b', 'c', 1.0), 'd') => ('fdph', ('vec3', 'a', 'b', 'c'), 'd') */
   /* search226_0_0 -> 149 in the cache */
   /* search226_0_1 -> 150 in the cache */
   /* search226_0_2 -> 152 in the cache */
   { .constant = {
      { nir_search_value_constant, -4 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 149, 150, 152, 774 },
      -1,
   } },
   /* search226_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fdot4,
      0, 1,
      { 775, 153 },
      -1,
   } },

   /* replace226_0_0 -> 149 in the cache */
   /* replace226_0_1 -> 150 in the cache */
   /* replace226_0_2 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_vec3,
      -1, 0,
      { 149, 150, 152 },
      -1,
   } },
   /* replace226_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fdph,
      -1, 0,
      { 777, 153 },
      -1,
   } },

   /* ('fdot4', ('vec4', 'a', 0.0, 0.0, 0.0), 'b') => ('fmul', 'a', 'b') */
   /* search227_0_0 -> 0 in the cache */
   /* search227_0_1 -> 256 in the cache */
   /* search227_0_2 -> 256 in the cache */
   /* search227_0_3 -> 256 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 0, 256, 256, 256 },
      -1,
   } },
   /* search227_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot4,
      0, 1,
      { 779, 3 },
      -1,
   } },

   /* replace227_0 -> 0 in the cache */
   /* replace227_1 -> 3 in the cache */
   /* replace227 -> 264 in the cache */

   /* ('fdot4', ('vec4', 'a', 'b', 0.0, 0.0), 'c') => ('fdot2', ('vec2', 'a', 'b'), 'c') */
   /* search228_0_0 -> 126 in the cache */
   /* search228_0_1 -> 127 in the cache */
   { .constant = {
      { nir_search_value_constant, -3 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },
   /* search228_0_3 -> 781 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 126, 127, 781, 781 },
      -1,
   } },
   /* search228_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fdot4,
      0, 1,
      { 782, 129 },
      -1,
   } },

   /* replace228_0_0 -> 126 in the cache */
   /* replace228_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 126, 127 },
      -1,
   } },
   /* replace228_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fdot2,
      0, 1,
      { 784, 129 },
      -1,
   } },

   /* ('fdot4', ('vec4', 'a', 'b', 'c', 0.0), 'd') => ('fdot3', ('vec3', 'a', 'b', 'c'), 'd') */
   /* search229_0_0 -> 149 in the cache */
   /* search229_0_1 -> 150 in the cache */
   /* search229_0_2 -> 152 in the cache */
   { .constant = {
      { nir_search_value_constant, -4 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 149, 150, 152, 786 },
      -1,
   } },
   /* search229_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fdot4,
      0, 1,
      { 787, 153 },
      -1,
   } },

   /* replace229_0_0 -> 149 in the cache */
   /* replace229_0_1 -> 150 in the cache */
   /* replace229_0_2 -> 152 in the cache */
   /* replace229_0 -> 777 in the cache */
   /* replace229_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fdot3,
      0, 1,
      { 777, 153 },
      -1,
   } },

   /* ('fdot3', ('vec3', 'a', 0.0, 0.0), 'b') => ('fmul', 'a', 'b') */
   /* search230_0_0 -> 0 in the cache */
   /* search230_0_1 -> 256 in the cache */
   /* search230_0_2 -> 256 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_vec3,
      -1, 0,
      { 0, 256, 256 },
      -1,
   } },
   /* search230_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot3,
      0, 1,
      { 790, 3 },
      -1,
   } },

   /* replace230_0 -> 0 in the cache */
   /* replace230_1 -> 3 in the cache */
   /* replace230 -> 264 in the cache */

   /* ('fdot3', ('vec3', 'a', 'b', 0.0), 'c') => ('fdot2', ('vec2', 'a', 'b'), 'c') */
   /* search231_0_0 -> 126 in the cache */
   /* search231_0_1 -> 127 in the cache */
   /* search231_0_2 -> 781 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_vec3,
      -1, 0,
      { 126, 127, 781 },
      -1,
   } },
   /* search231_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fdot3,
      0, 1,
      { 792, 129 },
      -1,
   } },

   /* replace231_0_0 -> 126 in the cache */
   /* replace231_0_1 -> 127 in the cache */
   /* replace231_0 -> 784 in the cache */
   /* replace231_1 -> 129 in the cache */
   /* replace231 -> 785 in the cache */

   /* ('fdot2', ('vec2', 'a', 0.0), 'b') => ('fmul', 'a', 'b') */
   /* search232_0_0 -> 0 in the cache */
   /* search232_0_1 -> 256 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 0, 256 },
      -1,
   } },
   /* search232_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot2,
      0, 1,
      { 794, 3 },
      -1,
   } },

   /* replace232_0 -> 0 in the cache */
   /* replace232_1 -> 3 in the cache */
   /* replace232 -> 264 in the cache */

   /* ('fdot2', 'a', 1.0) => ('fadd', 'a.x', 'a.y') */
   /* search233_0 -> 32 in the cache */
   /* search233_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fdot2,
      0, 1,
      { 32, 241 },
      -1,
   } },

   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0},
   } },
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {1},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      -1, 0,
      { 797, 798 },
      -1,
   } },

   /* ('fdot2', 'a', 'b') => ('fsum2', ('fmul', 'a', 'b')) */
   /* search234_0 -> 0 in the cache */
   /* search234_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot2,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace234_0_0 -> 0 in the cache */
   /* replace234_0_1 -> 3 in the cache */
   /* replace234_0 -> 264 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsum2,
      -1, 1,
      { 264 },
      -1,
   } },

   /* ('fdot3', 'a', 'b') => ('fsum3', ('fmul', 'a', 'b')) */
   /* search235_0 -> 0 in the cache */
   /* search235_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot3,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace235_0_0 -> 0 in the cache */
   /* replace235_0_1 -> 3 in the cache */
   /* replace235_0 -> 264 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsum3,
      -1, 1,
      { 264 },
      -1,
   } },

   /* ('fdot4', 'a', 'b') => ('fsum4', ('fmul', 'a', 'b')) */
   /* search236_0 -> 0 in the cache */
   /* search236_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot4,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace236_0_0 -> 0 in the cache */
   /* replace236_0_1 -> 3 in the cache */
   /* replace236_0 -> 264 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsum4,
      -1, 1,
      { 264 },
      -1,
   } },

   /* ('fsum2', 'a') => ('fadd', 'a.x', 'a.y') */
   /* search237_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsum2,
      -1, 0,
      { 32 },
      -1,
   } },

   /* replace237_0 -> 797 in the cache */
   /* replace237_1 -> 798 in the cache */
   /* replace237 -> 799 in the cache */

   /* ('~fadd', ('fneg(is_used_once)', ('fsat(is_used_once)', 'a(is_not_fmul)')), 1.0) => ('fsat', ('fadd', 1.0, ('fneg', 'a'))) */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      8,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 807 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 808 },
      0,
   } },
   /* search238_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 809, 241 },
      -1,
   } },

   /* replace238_0_0 -> 241 in the cache */
   /* replace238_0_1_0 -> 32 in the cache */
   /* replace238_0_1 -> 98 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 241, 98 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 811 },
      -1,
   } },

   /* ('ishl', ('iadd', ('imul', 'a', '#b'), '#c'), '#d') => ('iadd', ('imul', 'a', ('ishl', 'b', 'd')), ('ishl', 'c', 'd')) */
   /* search239_0_0_0 -> 126 in the cache */
   /* search239_0_0_1 -> 283 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 126, 283 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* c */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 813, 814 },
      -1,
   } },
   /* search239_1 -> 311 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 2,
      { 815, 311 },
      -1,
   } },

   /* replace239_0_0 -> 126 in the cache */
   /* replace239_0_1_0 -> 127 in the cache */
   /* replace239_0_1_1 -> 163 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 127, 163 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 126, 817 },
      -1,
   } },
   /* replace239_1_0 -> 129 in the cache */
   /* replace239_1_1 -> 163 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 129, 163 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 818, 819 },
      -1,
   } },

   /* ('ishl', ('imul', 'a', '#b'), '#c') => ('imul', 'a', ('ishl', 'b', 'c')) */
   /* search240_0_0 -> 0 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 0, 821 },
      -1,
   } },
   /* search240_1 -> 309 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 822, 309 },
      -1,
   } },

   /* replace240_0 -> 0 in the cache */
   /* replace240_1_0 -> 3 in the cache */
   /* replace240_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 0, 824 },
      -1,
   } },

   /* ('ishl@8', ('ishl@8', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 7), ('iand', 'c', 7)), 8), ('ishl@8', 'a', ('iadd', 'b', 'c')), 0) */
   /* search241_0_0 -> 6 in the cache */
   /* search241_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 6, 33 },
      -1,
   } },
   /* search241_1 -> 309 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 826, 309 },
      -1,
   } },

   /* replace241_0_0_0_0 -> 36 in the cache */
   /* replace241_0_0_0_1 -> 9 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 9 },
      -1,
   } },
   /* replace241_0_0_1_0 -> 135 in the cache */
   /* replace241_0_0_1_1 -> 9 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 135, 9 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 3,
      { 828, 829 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x8 /* 8 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 3,
      { 830, 831 },
      -1,
   } },
   /* replace241_1_0 -> 6 in the cache */
   /* replace241_1_1_0 -> 36 in the cache */
   /* replace241_1_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      3, 1,
      { 36, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 6, 833 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 8 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 832, 834, 835 },
      -1,
   } },

   /* ('ushr@8', ('ushr@8', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 7), ('iand', 'c', 7)), 8), ('ushr@8', 'a', ('iadd', 'b', 'c')), 0) */
   /* search242_0_0 -> 6 in the cache */
   /* search242_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 6, 33 },
      -1,
   } },
   /* search242_1 -> 309 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 837, 309 },
      -1,
   } },

   /* replace242_0_0_0_0 -> 36 in the cache */
   /* replace242_0_0_0_1 -> 9 in the cache */
   /* replace242_0_0_0 -> 828 in the cache */
   /* replace242_0_0_1_0 -> 135 in the cache */
   /* replace242_0_0_1_1 -> 9 in the cache */
   /* replace242_0_0_1 -> 829 in the cache */
   /* replace242_0_0 -> 830 in the cache */
   /* replace242_0_1 -> 831 in the cache */
   /* replace242_0 -> 832 in the cache */
   /* replace242_1_0 -> 6 in the cache */
   /* replace242_1_1_0 -> 36 in the cache */
   /* replace242_1_1_1 -> 135 in the cache */
   /* replace242_1_1 -> 833 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 6, 833 },
      -1,
   } },
   /* replace242_2 -> 835 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 832, 839, 835 },
      -1,
   } },

   /* ('ishr@8', ('ishr@8', 'a', '#b'), '#c') => ('ishr@8', 'a', ('imin', ('iadd', ('iand', 'b', 7), ('iand', 'c', 7)), 7)) */
   /* search243_0_0 -> 6 in the cache */
   /* search243_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 6, 33 },
      -1,
   } },
   /* search243_1 -> 309 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 841, 309 },
      -1,
   } },

   /* replace243_0 -> 6 in the cache */
   /* replace243_1_0_0_0 -> 36 in the cache */
   /* replace243_1_0_0_1 -> 9 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 36, 9 },
      -1,
   } },
   /* replace243_1_0_1_0 -> 135 in the cache */
   /* replace243_1_0_1_1 -> 9 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      3, 1,
      { 135, 9 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 3,
      { 843, 844 },
      -1,
   } },
   /* replace243_1_1 -> 9 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      0, 4,
      { 845, 9 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 4,
      { 6, 846 },
      -1,
   } },

   /* ('ishl@16', ('ishl@16', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 15), ('iand', 'c', 15)), 16), ('ishl@16', 'a', ('iadd', 'b', 'c')), 0) */
   /* search244_0_0 -> 11 in the cache */
   /* search244_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 11, 33 },
      -1,
   } },
   /* search244_1 -> 309 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 848, 309 },
      -1,
   } },

   /* replace244_0_0_0_0 -> 36 in the cache */
   /* replace244_0_0_0_1 -> 14 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 14 },
      -1,
   } },
   /* replace244_0_0_1_0 -> 135 in the cache */
   /* replace244_0_0_1_1 -> 14 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 135, 14 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 3,
      { 850, 851 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x10 /* 16 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 3,
      { 852, 853 },
      -1,
   } },
   /* replace244_1_0 -> 11 in the cache */
   /* replace244_1_1_0 -> 36 in the cache */
   /* replace244_1_1_1 -> 135 in the cache */
   /* replace244_1_1 -> 833 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 11, 833 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 854, 855, 856 },
      -1,
   } },

   /* ('ushr@16', ('ushr@16', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 15), ('iand', 'c', 15)), 16), ('ushr@16', 'a', ('iadd', 'b', 'c')), 0) */
   /* search245_0_0 -> 11 in the cache */
   /* search245_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 11, 33 },
      -1,
   } },
   /* search245_1 -> 309 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 858, 309 },
      -1,
   } },

   /* replace245_0_0_0_0 -> 36 in the cache */
   /* replace245_0_0_0_1 -> 14 in the cache */
   /* replace245_0_0_0 -> 850 in the cache */
   /* replace245_0_0_1_0 -> 135 in the cache */
   /* replace245_0_0_1_1 -> 14 in the cache */
   /* replace245_0_0_1 -> 851 in the cache */
   /* replace245_0_0 -> 852 in the cache */
   /* replace245_0_1 -> 853 in the cache */
   /* replace245_0 -> 854 in the cache */
   /* replace245_1_0 -> 11 in the cache */
   /* replace245_1_1_0 -> 36 in the cache */
   /* replace245_1_1_1 -> 135 in the cache */
   /* replace245_1_1 -> 833 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 11, 833 },
      -1,
   } },
   /* replace245_2 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 854, 860, 856 },
      -1,
   } },

   /* ('ishr@16', ('ishr@16', 'a', '#b'), '#c') => ('ishr@16', 'a', ('imin', ('iadd', ('iand', 'b', 15), ('iand', 'c', 15)), 15)) */
   /* search246_0_0 -> 11 in the cache */
   /* search246_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 11, 33 },
      -1,
   } },
   /* search246_1 -> 309 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 862, 309 },
      -1,
   } },

   /* replace246_0 -> 11 in the cache */
   /* replace246_1_0_0_0 -> 36 in the cache */
   /* replace246_1_0_0_1 -> 14 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 36, 14 },
      -1,
   } },
   /* replace246_1_0_1_0 -> 135 in the cache */
   /* replace246_1_0_1_1 -> 14 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      3, 1,
      { 135, 14 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 3,
      { 864, 865 },
      -1,
   } },
   /* replace246_1_1 -> 14 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      0, 4,
      { 866, 14 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 4,
      { 11, 867 },
      -1,
   } },

   /* ('ishl@32', ('ishl@32', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 31), ('iand', 'c', 31)), 32), ('ishl@32', 'a', ('iadd', 'b', 'c')), 0) */
   /* search247_0_0 -> 16 in the cache */
   /* search247_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 33 },
      -1,
   } },
   /* search247_1 -> 309 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 869, 309 },
      -1,
   } },

   /* replace247_0_0_0_0 -> 36 in the cache */
   /* replace247_0_0_0_1 -> 19 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 19 },
      -1,
   } },
   /* replace247_0_0_1_0 -> 135 in the cache */
   /* replace247_0_0_1_1 -> 19 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 135, 19 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 3,
      { 871, 872 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x20 /* 32 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 3,
      { 873, 874 },
      -1,
   } },
   /* replace247_1_0 -> 16 in the cache */
   /* replace247_1_1_0 -> 36 in the cache */
   /* replace247_1_1_1 -> 135 in the cache */
   /* replace247_1_1 -> 833 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 16, 833 },
      -1,
   } },
   /* replace247_2 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 875, 876, 122 },
      -1,
   } },

   /* ('ushr@32', ('ushr@32', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 31), ('iand', 'c', 31)), 32), ('ushr@32', 'a', ('iadd', 'b', 'c')), 0) */
   /* search248_0_0 -> 16 in the cache */
   /* search248_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 16, 33 },
      -1,
   } },
   /* search248_1 -> 309 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 878, 309 },
      -1,
   } },

   /* replace248_0_0_0_0 -> 36 in the cache */
   /* replace248_0_0_0_1 -> 19 in the cache */
   /* replace248_0_0_0 -> 871 in the cache */
   /* replace248_0_0_1_0 -> 135 in the cache */
   /* replace248_0_0_1_1 -> 19 in the cache */
   /* replace248_0_0_1 -> 872 in the cache */
   /* replace248_0_0 -> 873 in the cache */
   /* replace248_0_1 -> 874 in the cache */
   /* replace248_0 -> 875 in the cache */
   /* replace248_1_0 -> 16 in the cache */
   /* replace248_1_1_0 -> 36 in the cache */
   /* replace248_1_1_1 -> 135 in the cache */
   /* replace248_1_1 -> 833 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 16, 833 },
      -1,
   } },
   /* replace248_2 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 875, 880, 122 },
      -1,
   } },

   /* ('ishr@32', ('ishr@32', 'a', '#b'), '#c') => ('ishr@32', 'a', ('imin', ('iadd', ('iand', 'b', 31), ('iand', 'c', 31)), 31)) */
   /* search249_0_0 -> 16 in the cache */
   /* search249_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 16, 33 },
      -1,
   } },
   /* search249_1 -> 309 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 882, 309 },
      -1,
   } },

   /* replace249_0 -> 16 in the cache */
   /* replace249_1_0_0_0 -> 36 in the cache */
   /* replace249_1_0_0_1 -> 19 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 36, 19 },
      -1,
   } },
   /* replace249_1_0_1_0 -> 135 in the cache */
   /* replace249_1_0_1_1 -> 19 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      3, 1,
      { 135, 19 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 3,
      { 884, 885 },
      -1,
   } },
   /* replace249_1_1 -> 19 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      0, 4,
      { 886, 19 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 4,
      { 16, 887 },
      -1,
   } },

   /* ('ishl@64', ('ishl@64', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 63), ('iand', 'c', 63)), 64), ('ishl@64', 'a', ('iadd', 'b', 'c')), 0) */
   /* search250_0_0 -> 21 in the cache */
   /* search250_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 33 },
      -1,
   } },
   /* search250_1 -> 309 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 889, 309 },
      -1,
   } },

   /* replace250_0_0_0_0 -> 36 in the cache */
   /* replace250_0_0_0_1 -> 24 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 24 },
      -1,
   } },
   /* replace250_0_0_1_0 -> 135 in the cache */
   /* replace250_0_0_1_1 -> 24 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 135, 24 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 3,
      { 891, 892 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x40 /* 64 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 3,
      { 893, 894 },
      -1,
   } },
   /* replace250_1_0 -> 21 in the cache */
   /* replace250_1_1_0 -> 36 in the cache */
   /* replace250_1_1_1 -> 135 in the cache */
   /* replace250_1_1 -> 833 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 21, 833 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 895, 896, 897 },
      -1,
   } },

   /* ('ushr@64', ('ushr@64', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 63), ('iand', 'c', 63)), 64), ('ushr@64', 'a', ('iadd', 'b', 'c')), 0) */
   /* search251_0_0 -> 21 in the cache */
   /* search251_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 21, 33 },
      -1,
   } },
   /* search251_1 -> 309 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 899, 309 },
      -1,
   } },

   /* replace251_0_0_0_0 -> 36 in the cache */
   /* replace251_0_0_0_1 -> 24 in the cache */
   /* replace251_0_0_0 -> 891 in the cache */
   /* replace251_0_0_1_0 -> 135 in the cache */
   /* replace251_0_0_1_1 -> 24 in the cache */
   /* replace251_0_0_1 -> 892 in the cache */
   /* replace251_0_0 -> 893 in the cache */
   /* replace251_0_1 -> 894 in the cache */
   /* replace251_0 -> 895 in the cache */
   /* replace251_1_0 -> 21 in the cache */
   /* replace251_1_1_0 -> 36 in the cache */
   /* replace251_1_1_1 -> 135 in the cache */
   /* replace251_1_1 -> 833 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 21, 833 },
      -1,
   } },
   /* replace251_2 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 895, 901, 897 },
      -1,
   } },

   /* ('ishr@64', ('ishr@64', 'a', '#b'), '#c') => ('ishr@64', 'a', ('imin', ('iadd', ('iand', 'b', 63), ('iand', 'c', 63)), 63)) */
   /* search252_0_0 -> 21 in the cache */
   /* search252_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 21, 33 },
      -1,
   } },
   /* search252_1 -> 309 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 903, 309 },
      -1,
   } },

   /* replace252_0 -> 21 in the cache */
   /* replace252_1_0_0_0 -> 36 in the cache */
   /* replace252_1_0_0_1 -> 24 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 36, 24 },
      -1,
   } },
   /* replace252_1_0_1_0 -> 135 in the cache */
   /* replace252_1_0_1_1 -> 24 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      3, 1,
      { 135, 24 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 3,
      { 905, 906 },
      -1,
   } },
   /* replace252_1_1 -> 24 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      0, 4,
      { 907, 24 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 4,
      { 21, 908 },
      -1,
   } },

   /* ('ishl', ('ushr', 'a@8', '#b'), 'b') => ('iand', 'a', ('ishl', 255, 'b')) */
   /* search253_0_0 -> 6 in the cache */
   /* search253_0_1 -> 33 in the cache */
   /* search253_0 -> 837 in the cache */
   /* search253_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 837, 36 },
      -1,
   } },

   /* replace253_0 -> 6 in the cache */
   { .constant = {
      { nir_search_value_constant, 8 },
      nir_type_int, { 0xff /* 255 */ },
   } },
   /* replace253_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 911, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 6, 912 },
      -1,
   } },

   /* ('ishl', ('ishr', 'a@8', '#b'), 'b') => ('iand', 'a', ('ishl', 255, 'b')) */
   /* search254_0_0 -> 6 in the cache */
   /* search254_0_1 -> 33 in the cache */
   /* search254_0 -> 841 in the cache */
   /* search254_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 841, 36 },
      -1,
   } },

   /* replace254_0 -> 6 in the cache */
   /* replace254_1_0 -> 911 in the cache */
   /* replace254_1_1 -> 36 in the cache */
   /* replace254_1 -> 912 in the cache */
   /* replace254 -> 913 in the cache */

   /* ('ushr', ('ishl', 'a@8', '#b'), 'b') => ('iand', 'a', ('ushr', 255, 'b')) */
   /* search255_0_0 -> 6 in the cache */
   /* search255_0_1 -> 33 in the cache */
   /* search255_0 -> 826 in the cache */
   /* search255_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 826, 36 },
      -1,
   } },

   /* replace255_0 -> 6 in the cache */
   /* replace255_1_0 -> 911 in the cache */
   /* replace255_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 911, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 6, 916 },
      -1,
   } },

   /* ('ishl', ('ushr', 'a@16', '#b'), 'b') => ('iand', 'a', ('ishl', 65535, 'b')) */
   /* search256_0_0 -> 11 in the cache */
   /* search256_0_1 -> 33 in the cache */
   /* search256_0 -> 858 in the cache */
   /* search256_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 858, 36 },
      -1,
   } },

   /* replace256_0 -> 11 in the cache */
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { 0xffff /* 65535 */ },
   } },
   /* replace256_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 919, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 11, 920 },
      -1,
   } },

   /* ('ishl', ('ishr', 'a@16', '#b'), 'b') => ('iand', 'a', ('ishl', 65535, 'b')) */
   /* search257_0_0 -> 11 in the cache */
   /* search257_0_1 -> 33 in the cache */
   /* search257_0 -> 862 in the cache */
   /* search257_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 862, 36 },
      -1,
   } },

   /* replace257_0 -> 11 in the cache */
   /* replace257_1_0 -> 919 in the cache */
   /* replace257_1_1 -> 36 in the cache */
   /* replace257_1 -> 920 in the cache */
   /* replace257 -> 921 in the cache */

   /* ('ushr', ('ishl', 'a@16', '#b'), 'b') => ('iand', 'a', ('ushr', 65535, 'b')) */
   /* search258_0_0 -> 11 in the cache */
   /* search258_0_1 -> 33 in the cache */
   /* search258_0 -> 848 in the cache */
   /* search258_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 848, 36 },
      -1,
   } },

   /* replace258_0 -> 11 in the cache */
   /* replace258_1_0 -> 919 in the cache */
   /* replace258_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 919, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 11, 924 },
      -1,
   } },

   /* ('ishl', ('ushr', 'a@32', '#b'), 'b') => ('iand', 'a', ('ishl', 4294967295, 'b')) */
   /* search259_0_0 -> 16 in the cache */
   /* search259_0_1 -> 33 in the cache */
   /* search259_0 -> 878 in the cache */
   /* search259_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 878, 36 },
      -1,
   } },

   /* replace259_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xffffffff /* 4294967295 */ },
   } },
   /* replace259_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 927, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 16, 928 },
      -1,
   } },

   /* ('ishl', ('ishr', 'a@32', '#b'), 'b') => ('iand', 'a', ('ishl', 4294967295, 'b')) */
   /* search260_0_0 -> 16 in the cache */
   /* search260_0_1 -> 33 in the cache */
   /* search260_0 -> 882 in the cache */
   /* search260_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 882, 36 },
      -1,
   } },

   /* replace260_0 -> 16 in the cache */
   /* replace260_1_0 -> 927 in the cache */
   /* replace260_1_1 -> 36 in the cache */
   /* replace260_1 -> 928 in the cache */
   /* replace260 -> 929 in the cache */

   /* ('ushr', ('ishl', 'a@32', '#b'), 'b') => ('iand', 'a', ('ushr', 4294967295, 'b')) */
   /* search261_0_0 -> 16 in the cache */
   /* search261_0_1 -> 33 in the cache */
   /* search261_0 -> 869 in the cache */
   /* search261_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 869, 36 },
      -1,
   } },

   /* replace261_0 -> 16 in the cache */
   /* replace261_1_0 -> 927 in the cache */
   /* replace261_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 927, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 16, 932 },
      -1,
   } },

   /* ('ishl', ('ushr', 'a@64', '#b'), 'b') => ('iand', 'a', ('ishl', 18446744073709551615, 'b')) */
   /* search262_0_0 -> 21 in the cache */
   /* search262_0_1 -> 33 in the cache */
   /* search262_0 -> 899 in the cache */
   /* search262_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 899, 36 },
      -1,
   } },

   /* replace262_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0xffffffffffffffff /* 18446744073709551615 */ },
   } },
   /* replace262_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 935, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 21, 936 },
      -1,
   } },

   /* ('ishl', ('ishr', 'a@64', '#b'), 'b') => ('iand', 'a', ('ishl', 18446744073709551615, 'b')) */
   /* search263_0_0 -> 21 in the cache */
   /* search263_0_1 -> 33 in the cache */
   /* search263_0 -> 903 in the cache */
   /* search263_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 903, 36 },
      -1,
   } },

   /* replace263_0 -> 21 in the cache */
   /* replace263_1_0 -> 935 in the cache */
   /* replace263_1_1 -> 36 in the cache */
   /* replace263_1 -> 936 in the cache */
   /* replace263 -> 937 in the cache */

   /* ('ushr', ('ishl', 'a@64', '#b'), 'b') => ('iand', 'a', ('ushr', 18446744073709551615, 'b')) */
   /* search264_0_0 -> 21 in the cache */
   /* search264_0_1 -> 33 in the cache */
   /* search264_0 -> 889 in the cache */
   /* search264_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 889, 36 },
      -1,
   } },

   /* replace264_0 -> 21 in the cache */
   /* replace264_1_0 -> 935 in the cache */
   /* replace264_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 935, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 21, 940 },
      -1,
   } },

   /* ('iand', ('ishl', 'a@32', '#b(is_first_5_bits_uge_2)'), -4) => ('ishl', 'a', 'b') */
   /* search265_0_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      9,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 942 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x4 /* -4 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 943, 944 },
      -1,
   } },

   /* replace265_0 -> 16 in the cache */
   /* replace265_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* ('iand', ('imul', 'a', '#b(is_unsigned_multiple_of_4)'), -4) => ('imul', 'a', 'b') */
   /* search266_0_0 -> 0 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      true,
      nir_type_invalid,
      10,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 0, 947 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, -2 },
      nir_type_int, { -0x4 /* -4 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 948, 949 },
      -1,
   } },

   /* replace266_0 -> 0 in the cache */
   /* replace266_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_2)'), 4294967294) => ('iadd', ('iand', 'a', 4294967294), 'b') */
   /* search267_0_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      11,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 16, 952 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xfffffffe /* 4294967294 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 953, 954 },
      -1,
   } },

   /* replace267_0_0 -> 16 in the cache */
   /* replace267_0_1 -> 954 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 16, 954 },
      -1,
   } },
   /* replace267_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 956, 36 },
      -1,
   } },

   /* ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_4)'), 4294967292) => ('iadd', ('iand', 'a', 4294967292), 'b') */
   /* search268_0_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      10,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 16, 958 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xfffffffc /* 4294967292 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 959, 960 },
      -1,
   } },

   /* replace268_0_0 -> 16 in the cache */
   /* replace268_0_1 -> 960 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 16, 960 },
      -1,
   } },
   /* replace268_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 962, 36 },
      -1,
   } },

   /* ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_8)'), 4294967288) => ('iadd', ('iand', 'a', 4294967288), 'b') */
   /* search269_0_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      12,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 16, 964 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xfffffff8 /* 4294967288 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 965, 966 },
      -1,
   } },

   /* replace269_0_0 -> 16 in the cache */
   /* replace269_0_1 -> 966 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 16, 966 },
      -1,
   } },
   /* replace269_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 968, 36 },
      -1,
   } },

   /* ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_16)'), 4294967280) => ('iadd', ('iand', 'a', 4294967280), 'b') */
   /* search270_0_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      13,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 16, 970 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xfffffff0 /* 4294967280 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 971, 972 },
      -1,
   } },

   /* replace270_0_0 -> 16 in the cache */
   /* replace270_0_1 -> 972 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 16, 972 },
      -1,
   } },
   /* replace270_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 974, 36 },
      -1,
   } },

   /* ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_32)'), 4294967264) => ('iadd', ('iand', 'a', 4294967264), 'b') */
   /* search271_0_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      14,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 16, 976 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xffffffe0 /* 4294967264 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 977, 978 },
      -1,
   } },

   /* replace271_0_0 -> 16 in the cache */
   /* replace271_0_1 -> 978 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 16, 978 },
      -1,
   } },
   /* replace271_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 980, 36 },
      -1,
   } },

   /* ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_64)'), 4294967232) => ('iadd', ('iand', 'a', 4294967232), 'b') */
   /* search272_0_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      15,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 16, 982 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xffffffc0 /* 4294967232 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 983, 984 },
      -1,
   } },

   /* replace272_0_0 -> 16 in the cache */
   /* replace272_0_1 -> 984 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 16, 984 },
      -1,
   } },
   /* replace272_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 986, 36 },
      -1,
   } },

   /* ('ishl@32', ('iand', 'a@32', 2147483647), 1) => ('ishl', 'a', 1) */
   /* search273_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x7fffffff /* 2147483647 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 16, 988 },
      -1,
   } },
   /* search273_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 989, 396 },
      -1,
   } },

   /* replace273_0 -> 16 in the cache */
   /* replace273_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 396 },
      -1,
   } },

   /* ('iand', ('ishl', 'a@32', 1), 4294967294) => ('ishl', 'a', 1) */
   /* search274_0_0 -> 16 in the cache */
   /* search274_0_1 -> 396 in the cache */
   /* search274_0 -> 991 in the cache */
   /* search274_1 -> 954 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 991, 954 },
      -1,
   } },

   /* replace274_0 -> 16 in the cache */
   /* replace274_1 -> 396 in the cache */
   /* replace274 -> 991 in the cache */

   /* ('iand', ('ushr', 'a@32', 1), 2147483647) => ('ushr', 'a', 1) */
   /* search275_0_0 -> 16 in the cache */
   /* search275_0_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 16, 396 },
      -1,
   } },
   /* search275_1 -> 988 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 993, 988 },
      -1,
   } },

   /* replace275_0 -> 16 in the cache */
   /* replace275_1 -> 396 in the cache */
   /* replace275 -> 993 in the cache */

   /* ('ishl@32', ('iand', 'a@32', 1073741823), 2) => ('ishl', 'a', 2) */
   /* search276_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x3fffffff /* 1073741823 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 16, 995 },
      -1,
   } },
   /* search276_1 -> 401 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 996, 401 },
      -1,
   } },

   /* replace276_0 -> 16 in the cache */
   /* replace276_1 -> 401 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 401 },
      -1,
   } },

   /* ('iand', ('ishl', 'a@32', 2), 4294967292) => ('ishl', 'a', 2) */
   /* search277_0_0 -> 16 in the cache */
   /* search277_0_1 -> 401 in the cache */
   /* search277_0 -> 998 in the cache */
   /* search277_1 -> 960 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 998, 960 },
      -1,
   } },

   /* replace277_0 -> 16 in the cache */
   /* replace277_1 -> 401 in the cache */
   /* replace277 -> 998 in the cache */

   /* ('iand', ('ushr', 'a@32', 2), 1073741823) => ('ushr', 'a', 2) */
   /* search278_0_0 -> 16 in the cache */
   /* search278_0_1 -> 401 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 16, 401 },
      -1,
   } },
   /* search278_1 -> 995 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1000, 995 },
      -1,
   } },

   /* replace278_0 -> 16 in the cache */
   /* replace278_1 -> 401 in the cache */
   /* replace278 -> 1000 in the cache */

   /* ('ishl@32', ('iand', 'a@32', 65535), 16) => ('ishl', 'a', 16) */
   /* search279_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xffff /* 65535 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 16, 1002 },
      -1,
   } },
   /* search279_1 -> 853 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 1003, 853 },
      -1,
   } },

   /* replace279_0 -> 16 in the cache */
   /* replace279_1 -> 853 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 853 },
      -1,
   } },

   /* ('iand', ('ishl', 'a@32', 16), 4294901760) => ('ishl', 'a', 16) */
   /* search280_0_0 -> 16 in the cache */
   /* search280_0_1 -> 853 in the cache */
   /* search280_0 -> 1005 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xffff0000 /* 4294901760 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1005, 1006 },
      -1,
   } },

   /* replace280_0 -> 16 in the cache */
   /* replace280_1 -> 853 in the cache */
   /* replace280 -> 1005 in the cache */

   /* ('iand', ('ushr', 'a@32', 16), 65535) => ('ushr', 'a', 16) */
   /* search281_0_0 -> 16 in the cache */
   /* search281_0_1 -> 853 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 16, 853 },
      -1,
   } },
   /* search281_1 -> 1002 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1008, 1002 },
      -1,
   } },

   /* replace281_0 -> 16 in the cache */
   /* replace281_1 -> 853 in the cache */
   /* replace281 -> 1008 in the cache */

   /* ('ishl@32', ('iand', 'a@32', 255), 24) => ('ishl', 'a', 24) */
   /* search282_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xff /* 255 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 16, 1010 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x18 /* 24 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 1011, 1012 },
      -1,
   } },

   /* replace282_0 -> 16 in the cache */
   /* replace282_1 -> 1012 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 1012 },
      -1,
   } },

   /* ('iand', ('ishl', 'a@32', 24), 4278190080) => ('ishl', 'a', 24) */
   /* search283_0_0 -> 16 in the cache */
   /* search283_0_1 -> 1012 in the cache */
   /* search283_0 -> 1014 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xff000000 /* 4278190080 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1014, 1015 },
      -1,
   } },

   /* replace283_0 -> 16 in the cache */
   /* replace283_1 -> 1012 in the cache */
   /* replace283 -> 1014 in the cache */

   /* ('iand', ('ushr', 'a@32', 24), 255) => ('ushr', 'a', 24) */
   /* search284_0_0 -> 16 in the cache */
   /* search284_0_1 -> 1012 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 16, 1012 },
      -1,
   } },
   /* search284_1 -> 1010 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1017, 1010 },
      -1,
   } },

   /* replace284_0 -> 16 in the cache */
   /* replace284_1 -> 1012 in the cache */
   /* replace284 -> 1017 in the cache */

   /* ('ishl', ('iadd', 'a', '#b'), '#c') => ('iadd', ('ishl', 'a', 'c'), ('ishl', 'b', 'c')) */
   /* search285_0_0 -> 0 in the cache */
   /* search285_0_1 -> 821 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 0, 821 },
      -1,
   } },
   /* search285_1 -> 309 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 1019, 309 },
      -1,
   } },

   /* replace285_0_0 -> 0 in the cache */
   /* replace285_0_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 0, 135 },
      -1,
   } },
   /* replace285_1_0 -> 3 in the cache */
   /* replace285_1_1 -> 135 in the cache */
   /* replace285_1 -> 824 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 1021, 824 },
      -1,
   } },

   /* ('imul', ('iadd(is_used_once)', 'a', '#b'), '#c') => ('iadd', ('imul', 'a', 'c'), ('imul', 'b', 'c')) */
   /* search286_0_0 -> 126 in the cache */
   /* search286_0_1 -> 283 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 126, 283 },
      0,
   } },
   /* search286_1 -> 814 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      0, 2,
      { 1023, 814 },
      -1,
   } },

   /* replace286_0_0 -> 126 in the cache */
   /* replace286_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 126, 129 },
      -1,
   } },
   /* replace286_1_0 -> 127 in the cache */
   /* replace286_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      2, 1,
      { 127, 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 3,
      { 1025, 1026 },
      -1,
   } },

   /* ('imul', ('iadd(is_used_once)', ('iadd(is_used_once)', 'a', '#b'), 'c'), '#d') => ('iadd', ('imul', ('iadd', 'a', 'c'), 'd'), ('imul', 'b', 'd')) */
   /* search287_0_0_0 -> 149 in the cache */
   { .variable = {
      { nir_search_value_variable, -4 },
      1, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 1,
      { 149, 1028 },
      0,
   } },
   /* search287_0_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 2,
      { 1029, 152 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      3, /* d */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      0, 3,
      { 1030, 1031 },
      -1,
   } },

   /* replace287_0_0_0 -> 149 in the cache */
   /* replace287_0_0_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 1,
      { 149, 152 },
      -1,
   } },
   /* replace287_0_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      1, 2,
      { 1033, 153 },
      -1,
   } },
   /* replace287_1_0 -> 150 in the cache */
   /* replace287_1_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      3, 1,
      { 150, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 4,
      { 1034, 1035 },
      -1,
   } },

   /* ('ishl', ('iadd(is_used_once)', ('iadd(is_used_once)', 'a', '#b'), 'c'), '#d') => ('iadd', ('ishl', ('iadd', 'a', 'c'), 'd'), ('ishl', 'b', 'd')) */
   /* search288_0_0_0 -> 126 in the cache */
   /* search288_0_0_1 -> 283 in the cache */
   /* search288_0_0 -> 1023 in the cache */
   /* search288_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 1023, 129 },
      0,
   } },
   /* search288_1 -> 311 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 2,
      { 1037, 311 },
      -1,
   } },

   /* replace288_0_0_0 -> 126 in the cache */
   /* replace288_0_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 126, 129 },
      -1,
   } },
   /* replace288_0_1 -> 163 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 1039, 163 },
      -1,
   } },
   /* replace288_1_0 -> 127 in the cache */
   /* replace288_1_1 -> 163 in the cache */
   /* replace288_1 -> 817 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 1040, 817 },
      -1,
   } },

   /* ('inot', ('flt(is_used_once)', 'a(is_a_number)', 'b(is_a_number)')) => ('fge', 'a', 'b') */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      16,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      16,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1042, 1043 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 1044 },
      -1,
   } },

   /* replace289_0 -> 0 in the cache */
   /* replace289_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* ('inot', ('fge(is_used_once)', 'a(is_a_number)', 'b(is_a_number)')) => ('flt', 'a', 'b') */
   /* search290_0_0 -> 1042 in the cache */
   /* search290_0_1 -> 1043 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1042, 1043 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 1047 },
      -1,
   } },

   /* replace290_0 -> 0 in the cache */
   /* replace290_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* ('inot', ('feq(is_used_once)', 'a', 'b')) => ('fneu', 'a', 'b') */
   /* search291_0_0 -> 0 in the cache */
   /* search291_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 0, 3 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 1,
      { 1050 },
      -1,
   } },

   /* replace291_0 -> 0 in the cache */
   /* replace291_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('inot', ('fneu(is_used_once)', 'a', 'b')) => ('feq', 'a', 'b') */
   /* search292_0_0 -> 0 in the cache */
   /* search292_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 0, 3 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 1,
      { 1053 },
      -1,
   } },

   /* replace292_0 -> 0 in the cache */
   /* replace292_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('inot', ('ilt(is_used_once)', 'a', 'b')) => ('ige', 'a', 'b') */
   /* search293_0_0 -> 0 in the cache */
   /* search293_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 0, 3 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 1056 },
      -1,
   } },

   /* replace293_0 -> 0 in the cache */
   /* replace293_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* ('inot', ('ult(is_used_once)', 'a', 'b')) => ('uge', 'a', 'b') */
   /* search294_0_0 -> 0 in the cache */
   /* search294_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 0, 3 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 1059 },
      -1,
   } },

   /* replace294_0 -> 0 in the cache */
   /* replace294_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* ('inot', ('ige(is_used_once)', 'a', 'b')) => ('ilt', 'a', 'b') */
   /* search295_0_0 -> 0 in the cache */
   /* search295_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 0, 3 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 1062 },
      -1,
   } },

   /* replace295_0 -> 0 in the cache */
   /* replace295_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* ('inot', ('uge(is_used_once)', 'a', 'b')) => ('ult', 'a', 'b') */
   /* search296_0_0 -> 0 in the cache */
   /* search296_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 0, 3 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 1065 },
      -1,
   } },

   /* replace296_0 -> 0 in the cache */
   /* replace296_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* ('inot', ('ieq(is_used_once)', 'a', 'b')) => ('ine', 'a', 'b') */
   /* search297_0_0 -> 0 in the cache */
   /* search297_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 0, 3 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 1,
      { 1068 },
      -1,
   } },

   /* replace297_0 -> 0 in the cache */
   /* replace297_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('inot', ('ine(is_used_once)', 'a', 'b')) => ('ieq', 'a', 'b') */
   /* search298_0_0 -> 0 in the cache */
   /* search298_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 0, 3 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 1,
      { 1071 },
      -1,
   } },

   /* replace298_0 -> 0 in the cache */
   /* replace298_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('iand', ('feq', 'a', 'b'), ('fneu', 'a', 'b')) => False */
   /* search299_0_0 -> 0 in the cache */
   /* search299_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* search299_1_0 -> 0 in the cache */
   /* search299_1_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      2, 1,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 1074, 1075 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 1 },
      nir_type_bool, { NIR_FALSE /* False */ },
   } },

   /* ('iand', ('flt', 'a', 'b'), ('flt', 'b', 'a')) => False */
   /* search300_0_0 -> 0 in the cache */
   /* search300_0_1 -> 3 in the cache */
   /* search300_0 -> 1049 in the cache */
   /* search300_1_0 -> 3 in the cache */
   /* search300_1_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 3, 0 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1049, 1078 },
      -1,
   } },

   /* replace300 -> 1077 in the cache */

   /* ('iand', ('ieq', 'a', 'b'), ('ine', 'a', 'b')) => False */
   /* search301_0_0 -> 0 in the cache */
   /* search301_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* search301_1_0 -> 0 in the cache */
   /* search301_1_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 1080, 1081 },
      -1,
   } },

   /* replace301 -> 1077 in the cache */

   /* ('iand', ('ilt', 'a', 'b'), ('ilt', 'b', 'a')) => False */
   /* search302_0_0 -> 0 in the cache */
   /* search302_0_1 -> 3 in the cache */
   /* search302_0 -> 1064 in the cache */
   /* search302_1_0 -> 3 in the cache */
   /* search302_1_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 3, 0 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1064, 1083 },
      -1,
   } },

   /* replace302 -> 1077 in the cache */

   /* ('iand', ('ult', 'a', 'b'), ('ult', 'b', 'a')) => False */
   /* search303_0_0 -> 0 in the cache */
   /* search303_0_1 -> 3 in the cache */
   /* search303_0 -> 1067 in the cache */
   /* search303_1_0 -> 3 in the cache */
   /* search303_1_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 3, 0 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1067, 1085 },
      -1,
   } },

   /* replace303 -> 1077 in the cache */

   /* ('flt', ('fneg', 'a'), ('fneg', 'b')) => ('flt', 'b', 'a') */
   /* search304_0_0 -> 0 in the cache */
   /* search304_0 -> 212 in the cache */
   /* search304_1_0 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 212, 1087 },
      -1,
   } },

   /* replace304_0 -> 3 in the cache */
   /* replace304_1 -> 0 in the cache */
   /* replace304 -> 1078 in the cache */

   /* ('fge', ('fneg', 'a'), ('fneg', 'b')) => ('fge', 'b', 'a') */
   /* search305_0_0 -> 0 in the cache */
   /* search305_0 -> 212 in the cache */
   /* search305_1_0 -> 3 in the cache */
   /* search305_1 -> 1087 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 212, 1087 },
      -1,
   } },

   /* replace305_0 -> 3 in the cache */
   /* replace305_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 3, 0 },
      -1,
   } },

   /* ('feq', ('fneg', 'a'), ('fneg', 'b')) => ('feq', 'b', 'a') */
   /* search306_0_0 -> 0 in the cache */
   /* search306_0 -> 212 in the cache */
   /* search306_1_0 -> 3 in the cache */
   /* search306_1 -> 1087 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 212, 1087 },
      -1,
   } },

   /* replace306_0 -> 3 in the cache */
   /* replace306_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 3, 0 },
      -1,
   } },

   /* ('fneu', ('fneg', 'a'), ('fneg', 'b')) => ('fneu', 'b', 'a') */
   /* search307_0_0 -> 0 in the cache */
   /* search307_0 -> 212 in the cache */
   /* search307_1_0 -> 3 in the cache */
   /* search307_1 -> 1087 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 212, 1087 },
      -1,
   } },

   /* replace307_0 -> 3 in the cache */
   /* replace307_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 3, 0 },
      -1,
   } },

   /* ('flt', ('fneg', 'a'), -1.0) => ('flt', 1.0, 'a') */
   /* search308_0_0 -> 32 in the cache */
   /* search308_0 -> 98 in the cache */
   /* search308_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 98, 247 },
      -1,
   } },

   /* replace308_0 -> 241 in the cache */
   /* replace308_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 241, 32 },
      -1,
   } },

   /* ('flt', -1.0, ('fneg', 'a')) => ('flt', 'a', 1.0) */
   /* search309_0 -> 247 in the cache */
   /* search309_1_0 -> 32 in the cache */
   /* search309_1 -> 98 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 247, 98 },
      -1,
   } },

   /* replace309_0 -> 32 in the cache */
   /* replace309_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 32, 241 },
      -1,
   } },

   /* ('fge', ('fneg', 'a'), -1.0) => ('fge', 1.0, 'a') */
   /* search310_0_0 -> 32 in the cache */
   /* search310_0 -> 98 in the cache */
   /* search310_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 98, 247 },
      -1,
   } },

   /* replace310_0 -> 241 in the cache */
   /* replace310_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 241, 32 },
      -1,
   } },

   /* ('fge', -1.0, ('fneg', 'a')) => ('fge', 'a', 1.0) */
   /* search311_0 -> 247 in the cache */
   /* search311_1_0 -> 32 in the cache */
   /* search311_1 -> 98 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 247, 98 },
      -1,
   } },

   /* replace311_0 -> 32 in the cache */
   /* replace311_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 32, 241 },
      -1,
   } },

   /* ('fneu', ('fneg', 'a'), -1.0) => ('fneu', 1.0, 'a') */
   /* search312_0_0 -> 32 in the cache */
   /* search312_0 -> 98 in the cache */
   /* search312_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 98, 247 },
      -1,
   } },

   /* replace312_0 -> 241 in the cache */
   /* replace312_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 241, 32 },
      -1,
   } },

   /* ('feq', -1.0, ('fneg', 'a')) => ('feq', 'a', 1.0) */
   /* search313_0 -> 247 in the cache */
   /* search313_1_0 -> 32 in the cache */
   /* search313_1 -> 98 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 247, 98 },
      -1,
   } },

   /* replace313_0 -> 32 in the cache */
   /* replace313_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 32, 241 },
      -1,
   } },

   /* ('ieq', ('ineg', 'a'), 0) => ('ieq', 'a', 0) */
   /* search314_0_0 -> 32 in the cache */
   /* search314_0 -> 100 in the cache */
   /* search314_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 100, 61 },
      -1,
   } },

   /* replace314_0 -> 32 in the cache */
   /* replace314_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 32, 61 },
      -1,
   } },

   /* ('ine', ('ineg', 'a'), 0) => ('ine', 'a', 0) */
   /* search315_0_0 -> 32 in the cache */
   /* search315_0 -> 100 in the cache */
   /* search315_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 100, 61 },
      -1,
   } },

   /* replace315_0 -> 32 in the cache */
   /* replace315_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 32, 61 },
      -1,
   } },

   /* ('ieq', ('iabs', 'a'), 0) => ('ieq', 'a', 0) */
   /* search316_0_0 -> 32 in the cache */
   /* search316_0 -> 106 in the cache */
   /* search316_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 106, 61 },
      -1,
   } },

   /* replace316_0 -> 32 in the cache */
   /* replace316_1 -> 61 in the cache */
   /* replace316 -> 1108 in the cache */

   /* ('ine', ('iabs', 'a'), 0) => ('ine', 'a', 0) */
   /* search317_0_0 -> 32 in the cache */
   /* search317_0 -> 106 in the cache */
   /* search317_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 106, 61 },
      -1,
   } },

   /* replace317_0 -> 32 in the cache */
   /* replace317_1 -> 61 in the cache */
   /* replace317 -> 1110 in the cache */

   /* ('flt', '#b(is_gt_0_and_lt_1)', ('fsat(is_used_once)', 'a')) => ('flt', 'b', 'a') */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* b */
      true,
      nir_type_invalid,
      17,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 1114 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1113, 1115 },
      -1,
   } },

   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* replace318_1 -> 1114 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1117, 1114 },
      -1,
   } },

   /* ('fge', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('fge', 'a', 'b') */
   /* search319_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 0 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      true,
      nir_type_invalid,
      17,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1119, 1120 },
      -1,
   } },

   /* replace319_0 -> 0 in the cache */
   /* replace319_1 -> 3 in the cache */
   /* replace319 -> 1046 in the cache */

   /* ('feq', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('feq', 'a', 'b') */
   /* search320_0_0 -> 0 in the cache */
   /* search320_0 -> 1119 in the cache */
   /* search320_1 -> 1120 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1119, 1120 },
      -1,
   } },

   /* replace320_0 -> 0 in the cache */
   /* replace320_1 -> 3 in the cache */
   /* replace320 -> 1055 in the cache */

   /* ('fneu', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('fneu', 'a', 'b') */
   /* search321_0_0 -> 0 in the cache */
   /* search321_0 -> 1119 in the cache */
   /* search321_1 -> 1120 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1119, 1120 },
      -1,
   } },

   /* replace321_0 -> 0 in the cache */
   /* replace321_1 -> 3 in the cache */
   /* replace321 -> 1052 in the cache */

   /* ('fge', ('fsat(is_used_once)', 'a'), 1.0) => ('fge', 'a', 1.0) */
   /* search322_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 32 },
      0,
   } },
   /* search322_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1124, 241 },
      -1,
   } },

   /* replace322_0 -> 32 in the cache */
   /* replace322_1 -> 241 in the cache */
   /* replace322 -> 1102 in the cache */

   /* ('flt', 0.0, ('fsat(is_used_once)', 'a')) => ('flt', 0.0, 'a') */
   /* search323_0 -> 111 in the cache */
   /* search323_1_0 -> 32 in the cache */
   /* search323_1 -> 1124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 111, 1124 },
      -1,
   } },

   /* replace323_0 -> 111 in the cache */
   /* replace323_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 111, 32 },
      -1,
   } },

   /* ('fge', 0.0, ('b2f', 'a@1')) => ('inot', 'a') */
   { .constant = {
      { nir_search_value_constant, 0 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },
   /* search324_1_0 -> 760 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 760 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1128, 1129 },
      -1,
   } },

   /* replace324_0 -> 760 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 760 },
      -1,
   } },

   /* ('fge', ('fneg', ('b2f', 'a@1')), 0.0) => ('inot', 'a') */
   /* search325_0_0_0 -> 760 in the cache */
   /* search325_0_0 -> 1129 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 1129 },
      -1,
   } },
   /* search325_1 -> 1128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1132, 1128 },
      -1,
   } },

   /* replace325_0 -> 760 in the cache */
   /* replace325 -> 1131 in the cache */

   /* ('fneu', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1')), 0.0) => ('ior', 'a', 'b') */
   /* search326_0_0_0 -> 760 in the cache */
   /* search326_0_0 -> 1129 in the cache */
   { .variable = {
      { nir_search_value_variable, 1 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 1134 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 1129, 1135 },
      -1,
   } },
   /* search326_1 -> 1128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 1136, 1128 },
      -1,
   } },

   /* replace326_0 -> 760 in the cache */
   /* replace326_1 -> 1134 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 760, 1134 },
      -1,
   } },

   /* ('fneu', ('bcsel', 'a', 1.0, ('b2f', 'b@1')), 0.0) => ('ior', 'a', 'b') */
   /* search327_0_0 -> 760 in the cache */
   { .constant = {
      { nir_search_value_constant, 0 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   /* search327_0_2_0 -> 1134 in the cache */
   /* search327_0_2 -> 1135 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 1139, 1135 },
      -1,
   } },
   /* search327_1 -> 1128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1140, 1128 },
      -1,
   } },

   /* replace327_0 -> 760 in the cache */
   /* replace327_1 -> 1134 in the cache */
   /* replace327 -> 1138 in the cache */

   /* ('fneu', ('b2f', 'a@1'), ('fneg', ('b2f', 'b@1'))) => ('ior', 'a', 'b') */
   /* search328_0_0 -> 760 in the cache */
   /* search328_0 -> 1129 in the cache */
   /* search328_1_0_0 -> 1134 in the cache */
   /* search328_1_0 -> 1135 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 1135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1129, 1142 },
      -1,
   } },

   /* replace328_0 -> 760 in the cache */
   /* replace328_1 -> 1134 in the cache */
   /* replace328 -> 1138 in the cache */

   /* ('fneu', ('fmul', ('b2f', 'a@1'), ('b2f', 'b@1')), 0.0) => ('iand', 'a', 'b') */
   /* search329_0_0_0 -> 760 in the cache */
   /* search329_0_0 -> 1129 in the cache */
   /* search329_0_1_0 -> 1134 in the cache */
   /* search329_0_1 -> 1135 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 1129, 1135 },
      -1,
   } },
   /* search329_1 -> 1128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 1144, 1128 },
      -1,
   } },

   /* replace329_0 -> 760 in the cache */
   /* replace329_1 -> 1134 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 760, 1134 },
      -1,
   } },

   /* ('fneu', ('bcsel', 'a', ('b2f', 'b@1'), 0.0), 0.0) => ('iand', 'a', 'b') */
   /* search330_0_0 -> 760 in the cache */
   /* search330_0_1_0 -> 1134 in the cache */
   /* search330_0_1 -> 1135 in the cache */
   /* search330_0_2 -> 1128 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 1135, 1128 },
      -1,
   } },
   /* search330_1 -> 1128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1147, 1128 },
      -1,
   } },

   /* replace330_0 -> 760 in the cache */
   /* replace330_1 -> 1134 in the cache */
   /* replace330 -> 1146 in the cache */

   /* ('fneu', ('fadd', ('b2f', 'a@1'), ('fneg', ('b2f', 'b@1'))), 0.0) => ('ixor', 'a', 'b') */
   /* search331_0_0_0 -> 760 in the cache */
   /* search331_0_0 -> 1129 in the cache */
   /* search331_0_1_0_0 -> 1134 in the cache */
   /* search331_0_1_0 -> 1135 in the cache */
   /* search331_0_1 -> 1142 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 1129, 1142 },
      -1,
   } },
   /* search331_1 -> 1128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 1149, 1128 },
      -1,
   } },

   /* replace331_0 -> 760 in the cache */
   /* replace331_1 -> 1134 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 760, 1134 },
      -1,
   } },

   /* ('fneu', ('b2f', 'a@1'), ('b2f', 'b@1')) => ('ixor', 'a', 'b') */
   /* search332_0_0 -> 760 in the cache */
   /* search332_0 -> 1129 in the cache */
   /* search332_1_0 -> 1134 in the cache */
   /* search332_1 -> 1135 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1129, 1135 },
      -1,
   } },

   /* replace332_0 -> 760 in the cache */
   /* replace332_1 -> 1134 in the cache */
   /* replace332 -> 1151 in the cache */

   /* ('fneu', ('fneg', ('b2f', 'a@1')), ('fneg', ('b2f', 'b@1'))) => ('ixor', 'a', 'b') */
   /* search333_0_0_0 -> 760 in the cache */
   /* search333_0_0 -> 1129 in the cache */
   /* search333_0 -> 1132 in the cache */
   /* search333_1_0_0 -> 1134 in the cache */
   /* search333_1_0 -> 1135 in the cache */
   /* search333_1 -> 1142 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1132, 1142 },
      -1,
   } },

   /* replace333_0 -> 760 in the cache */
   /* replace333_1 -> 1134 in the cache */
   /* replace333 -> 1151 in the cache */

   /* ('feq', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1')), 0.0) => ('inot', ('ior', 'a', 'b')) */
   /* search334_0_0_0 -> 760 in the cache */
   /* search334_0_0 -> 1129 in the cache */
   /* search334_0_1_0 -> 1134 in the cache */
   /* search334_0_1 -> 1135 in the cache */
   /* search334_0 -> 1136 in the cache */
   /* search334_1 -> 1128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 1136, 1128 },
      -1,
   } },

   /* replace334_0_0 -> 760 in the cache */
   /* replace334_0_1 -> 1134 in the cache */
   /* replace334_0 -> 1138 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 1,
      { 1138 },
      -1,
   } },

   /* ('feq', ('bcsel', 'a', 1.0, ('b2f', 'b@1')), 0.0) => ('inot', ('ior', 'a', 'b')) */
   /* search335_0_0 -> 760 in the cache */
   /* search335_0_1 -> 1139 in the cache */
   /* search335_0_2_0 -> 1134 in the cache */
   /* search335_0_2 -> 1135 in the cache */
   /* search335_0 -> 1140 in the cache */
   /* search335_1 -> 1128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1140, 1128 },
      -1,
   } },

   /* replace335_0_0 -> 760 in the cache */
   /* replace335_0_1 -> 1134 in the cache */
   /* replace335_0 -> 1138 in the cache */
   /* replace335 -> 1155 in the cache */

   /* ('feq', ('b2f', 'a@1'), ('fneg', ('b2f', 'b@1'))) => ('inot', ('ior', 'a', 'b')) */
   /* search336_0_0 -> 760 in the cache */
   /* search336_0 -> 1129 in the cache */
   /* search336_1_0_0 -> 1134 in the cache */
   /* search336_1_0 -> 1135 in the cache */
   /* search336_1 -> 1142 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1129, 1142 },
      -1,
   } },

   /* replace336_0_0 -> 760 in the cache */
   /* replace336_0_1 -> 1134 in the cache */
   /* replace336_0 -> 1138 in the cache */
   /* replace336 -> 1155 in the cache */

   /* ('feq', ('fmul', ('b2f', 'a@1'), ('b2f', 'b@1')), 0.0) => ('inot', ('iand', 'a', 'b')) */
   /* search337_0_0_0 -> 760 in the cache */
   /* search337_0_0 -> 1129 in the cache */
   /* search337_0_1_0 -> 1134 in the cache */
   /* search337_0_1 -> 1135 in the cache */
   /* search337_0 -> 1144 in the cache */
   /* search337_1 -> 1128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 1144, 1128 },
      -1,
   } },

   /* replace337_0_0 -> 760 in the cache */
   /* replace337_0_1 -> 1134 in the cache */
   /* replace337_0 -> 1146 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 1,
      { 1146 },
      -1,
   } },

   /* ('feq', ('bcsel', 'a', ('b2f', 'b@1'), 0.0), 0.0) => ('inot', ('iand', 'a', 'b')) */
   /* search338_0_0 -> 760 in the cache */
   /* search338_0_1_0 -> 1134 in the cache */
   /* search338_0_1 -> 1135 in the cache */
   /* search338_0_2 -> 1128 in the cache */
   /* search338_0 -> 1147 in the cache */
   /* search338_1 -> 1128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1147, 1128 },
      -1,
   } },

   /* replace338_0_0 -> 760 in the cache */
   /* replace338_0_1 -> 1134 in the cache */
   /* replace338_0 -> 1146 in the cache */
   /* replace338 -> 1159 in the cache */

   /* ('feq', ('fadd', ('b2f', 'a@1'), ('fneg', ('b2f', 'b@1'))), 0.0) => ('ieq', 'a', 'b') */
   /* search339_0_0_0 -> 760 in the cache */
   /* search339_0_0 -> 1129 in the cache */
   /* search339_0_1_0_0 -> 1134 in the cache */
   /* search339_0_1_0 -> 1135 in the cache */
   /* search339_0_1 -> 1142 in the cache */
   /* search339_0 -> 1149 in the cache */
   /* search339_1 -> 1128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 1149, 1128 },
      -1,
   } },

   /* replace339_0 -> 760 in the cache */
   /* replace339_1 -> 1134 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 760, 1134 },
      -1,
   } },

   /* ('feq', ('b2f', 'a@1'), ('b2f', 'b@1')) => ('ieq', 'a', 'b') */
   /* search340_0_0 -> 760 in the cache */
   /* search340_0 -> 1129 in the cache */
   /* search340_1_0 -> 1134 in the cache */
   /* search340_1 -> 1135 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1129, 1135 },
      -1,
   } },

   /* replace340_0 -> 760 in the cache */
   /* replace340_1 -> 1134 in the cache */
   /* replace340 -> 1162 in the cache */

   /* ('feq', ('fneg', ('b2f', 'a@1')), ('fneg', ('b2f', 'b@1'))) => ('ieq', 'a', 'b') */
   /* search341_0_0_0 -> 760 in the cache */
   /* search341_0_0 -> 1129 in the cache */
   /* search341_0 -> 1132 in the cache */
   /* search341_1_0_0 -> 1134 in the cache */
   /* search341_1_0 -> 1135 in the cache */
   /* search341_1 -> 1142 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1132, 1142 },
      -1,
   } },

   /* replace341_0 -> 760 in the cache */
   /* replace341_1 -> 1134 in the cache */
   /* replace341 -> 1162 in the cache */

   /* ('flt', ('fneg', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1'))), 0.0) => ('ior', 'a', 'b') */
   /* search342_0_0_0_0 -> 760 in the cache */
   /* search342_0_0_0 -> 1129 in the cache */
   /* search342_0_0_1_0 -> 1134 in the cache */
   /* search342_0_0_1 -> 1135 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 1129, 1135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1165 },
      -1,
   } },
   /* search342_1 -> 1128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 1166, 1128 },
      -1,
   } },

   /* replace342_0 -> 760 in the cache */
   /* replace342_1 -> 1134 in the cache */
   /* replace342 -> 1138 in the cache */

   /* ('flt', 0.0, ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1'))) => ('ior', 'a', 'b') */
   /* search343_0 -> 1128 in the cache */
   /* search343_1_0_0 -> 760 in the cache */
   /* search343_1_0 -> 1129 in the cache */
   /* search343_1_1_0 -> 1134 in the cache */
   /* search343_1_1 -> 1135 in the cache */
   /* search343_1 -> 1165 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 1128, 1165 },
      -1,
   } },

   /* replace343_0 -> 760 in the cache */
   /* replace343_1 -> 1134 in the cache */
   /* replace343 -> 1138 in the cache */

   /* ('fge', ('fneg', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1'))), 0.0) => ('inot', ('ior', 'a', 'b')) */
   /* search344_0_0_0_0 -> 760 in the cache */
   /* search344_0_0_0 -> 1129 in the cache */
   /* search344_0_0_1_0 -> 1134 in the cache */
   /* search344_0_0_1 -> 1135 in the cache */
   /* search344_0_0 -> 1165 in the cache */
   /* search344_0 -> 1166 in the cache */
   /* search344_1 -> 1128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 1166, 1128 },
      -1,
   } },

   /* replace344_0_0 -> 760 in the cache */
   /* replace344_0_1 -> 1134 in the cache */
   /* replace344_0 -> 1138 in the cache */
   /* replace344 -> 1155 in the cache */

   /* ('fge', 0.0, ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1'))) => ('inot', ('ior', 'a', 'b')) */
   /* search345_0 -> 1128 in the cache */
   /* search345_1_0_0 -> 760 in the cache */
   /* search345_1_0 -> 1129 in the cache */
   /* search345_1_1_0 -> 1134 in the cache */
   /* search345_1_1 -> 1135 in the cache */
   /* search345_1 -> 1165 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 1128, 1165 },
      -1,
   } },

   /* replace345_0_0 -> 760 in the cache */
   /* replace345_0_1 -> 1134 in the cache */
   /* replace345_0 -> 1138 in the cache */
   /* replace345 -> 1155 in the cache */

   /* ('flt', 'a', ('fneg', 'a')) => ('flt', 'a', 0.0) */
   /* search346_0 -> 32 in the cache */
   /* search346_1_0 -> 32 in the cache */
   /* search346_1 -> 98 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 32, 98 },
      -1,
   } },

   /* replace346_0 -> 32 in the cache */
   /* replace346_1 -> 111 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 32, 111 },
      -1,
   } },

   /* ('fge', 'a', ('fneg', 'a')) => ('fge', 'a', 0.0) */
   /* search347_0 -> 32 in the cache */
   /* search347_1_0 -> 32 in the cache */
   /* search347_1 -> 98 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 32, 98 },
      -1,
   } },

   /* replace347_0 -> 32 in the cache */
   /* replace347_1 -> 111 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 32, 111 },
      -1,
   } },

   /* ('flt', ('fmin', 'c', ('fneg', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1')))), 0.0) => ('ior', ('flt', 'c', 0.0), ('ior', 'a', 'b')) */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 1 },
      1, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 1176 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 1 },
      2, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 1178 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 1177, 1179 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1180 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 2,
      { 1175, 1181 },
      -1,
   } },
   /* search348_1 -> 111 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 2,
      { 1182, 111 },
      -1,
   } },

   /* replace348_0_0 -> 1175 in the cache */
   /* replace348_0_1 -> 111 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1175, 111 },
      -1,
   } },
   /* replace348_1_0 -> 1176 in the cache */
   /* replace348_1_1 -> 1178 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 1176, 1178 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 1184, 1185 },
      -1,
   } },

   /* ('~flt', ('fadd', 'a', 'b'), 'a') => ('flt', 'b', 0.0) */
   /* search349_0_0 -> 0 in the cache */
   /* search349_0_1 -> 3 in the cache */
   /* search349_0 -> 272 in the cache */
   /* search349_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 272, 0 },
      -1,
   } },

   /* replace349_0 -> 3 in the cache */
   /* replace349_1 -> 256 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 3, 256 },
      -1,
   } },

   /* ('~fge', ('fadd', 'a', 'b'), 'a') => ('fge', 'b', 0.0) */
   /* search350_0_0 -> 0 in the cache */
   /* search350_0_1 -> 3 in the cache */
   /* search350_0 -> 272 in the cache */
   /* search350_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 272, 0 },
      -1,
   } },

   /* replace350_0 -> 3 in the cache */
   /* replace350_1 -> 256 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 3, 256 },
      -1,
   } },

   /* ('~feq', ('fadd', 'a', 'b'), 'a') => ('feq', 'b', 0.0) */
   /* search351_0_0 -> 0 in the cache */
   /* search351_0_1 -> 3 in the cache */
   /* search351_0 -> 213 in the cache */
   /* search351_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 213, 0 },
      -1,
   } },

   /* replace351_0 -> 3 in the cache */
   /* replace351_1 -> 256 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 3, 256 },
      -1,
   } },

   /* ('~fneu', ('fadd', 'a', 'b'), 'a') => ('fneu', 'b', 0.0) */
   /* search352_0_0 -> 0 in the cache */
   /* search352_0_1 -> 3 in the cache */
   /* search352_0 -> 213 in the cache */
   /* search352_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 213, 0 },
      -1,
   } },

   /* replace352_0 -> 3 in the cache */
   /* replace352_1 -> 256 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 3, 256 },
      -1,
   } },

   /* ('~flt', ('fadd(is_used_once)', 'a', '#b'), '#c') => ('flt', 'a', ('fadd', 'c', ('fneg', 'b'))) */
   /* search353_0_0 -> 126 in the cache */
   /* search353_0_1 -> 283 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 126, 283 },
      0,
   } },
   /* search353_1 -> 814 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 1195, 814 },
      -1,
   } },

   /* replace353_0 -> 126 in the cache */
   /* replace353_1_0 -> 129 in the cache */
   /* replace353_1_1_0 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 127 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 129, 1197 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 126, 1198 },
      -1,
   } },

   /* ('~flt', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', '#b')), '#c') => ('flt', ('fneg', ('fadd', 'c', 'b')), 'a') */
   /* search354_0_0_0 -> 126 in the cache */
   /* search354_0_0_1 -> 283 in the cache */
   /* search354_0_0 -> 1195 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1195 },
      0,
   } },
   /* search354_1 -> 814 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 1200, 814 },
      -1,
   } },

   /* replace354_0_0_0 -> 129 in the cache */
   /* replace354_0_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 129, 127 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1202 },
      -1,
   } },
   /* replace354_1 -> 126 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 1203, 126 },
      -1,
   } },

   /* ('~fge', ('fadd(is_used_once)', 'a', '#b'), '#c') => ('fge', 'a', ('fadd', 'c', ('fneg', 'b'))) */
   /* search355_0_0 -> 126 in the cache */
   /* search355_0_1 -> 283 in the cache */
   /* search355_0 -> 1195 in the cache */
   /* search355_1 -> 814 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 1195, 814 },
      -1,
   } },

   /* replace355_0 -> 126 in the cache */
   /* replace355_1_0 -> 129 in the cache */
   /* replace355_1_1_0 -> 127 in the cache */
   /* replace355_1_1 -> 1197 in the cache */
   /* replace355_1 -> 1198 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 126, 1198 },
      -1,
   } },

   /* ('~fge', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', '#b')), '#c') => ('fge', ('fneg', ('fadd', 'c', 'b')), 'a') */
   /* search356_0_0_0 -> 126 in the cache */
   /* search356_0_0_1 -> 283 in the cache */
   /* search356_0_0 -> 1195 in the cache */
   /* search356_0 -> 1200 in the cache */
   /* search356_1 -> 814 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 1200, 814 },
      -1,
   } },

   /* replace356_0_0_0 -> 129 in the cache */
   /* replace356_0_0_1 -> 127 in the cache */
   /* replace356_0_0 -> 1202 in the cache */
   /* replace356_0 -> 1203 in the cache */
   /* replace356_1 -> 126 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 1203, 126 },
      -1,
   } },

   /* ('~feq', ('fadd(is_used_once)', 'a', '#b'), '#c') => ('feq', 'a', ('fadd', 'c', ('fneg', 'b'))) */
   /* search357_0_0 -> 126 in the cache */
   /* search357_0_1 -> 283 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 126, 283 },
      0,
   } },
   /* search357_1 -> 814 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 1209, 814 },
      -1,
   } },

   /* replace357_0 -> 126 in the cache */
   /* replace357_1_0 -> 129 in the cache */
   /* replace357_1_1_0 -> 127 in the cache */
   /* replace357_1_1 -> 1197 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 129, 1197 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 126, 1211 },
      -1,
   } },

   /* ('~feq', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', '#b')), '#c') => ('feq', ('fneg', ('fadd', 'c', 'b')), 'a') */
   /* search358_0_0_0 -> 126 in the cache */
   /* search358_0_0_1 -> 283 in the cache */
   /* search358_0_0 -> 1209 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1209 },
      0,
   } },
   /* search358_1 -> 814 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 1213, 814 },
      -1,
   } },

   /* replace358_0_0_0 -> 129 in the cache */
   /* replace358_0_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 129, 127 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1215 },
      -1,
   } },
   /* replace358_1 -> 126 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 1216, 126 },
      -1,
   } },

   /* ('~fneu', ('fadd(is_used_once)', 'a', '#b'), '#c') => ('fneu', 'a', ('fadd', 'c', ('fneg', 'b'))) */
   /* search359_0_0 -> 126 in the cache */
   /* search359_0_1 -> 283 in the cache */
   /* search359_0 -> 1209 in the cache */
   /* search359_1 -> 814 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 1209, 814 },
      -1,
   } },

   /* replace359_0 -> 126 in the cache */
   /* replace359_1_0 -> 129 in the cache */
   /* replace359_1_1_0 -> 127 in the cache */
   /* replace359_1_1 -> 1197 in the cache */
   /* replace359_1 -> 1211 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 126, 1211 },
      -1,
   } },

   /* ('~fneu', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', '#b')), '#c') => ('fneu', ('fneg', ('fadd', 'c', 'b')), 'a') */
   /* search360_0_0_0 -> 126 in the cache */
   /* search360_0_0_1 -> 283 in the cache */
   /* search360_0_0 -> 1209 in the cache */
   /* search360_0 -> 1213 in the cache */
   /* search360_1 -> 814 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 1213, 814 },
      -1,
   } },

   /* replace360_0_0_0 -> 129 in the cache */
   /* replace360_0_0_1 -> 127 in the cache */
   /* replace360_0_0 -> 1215 in the cache */
   /* replace360_0 -> 1216 in the cache */
   /* replace360_1 -> 126 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 1216, 126 },
      -1,
   } },

   /* ('ieq', ('iadd', 'a', 'b'), 'a') => ('ieq', 'b', 0) */
   /* search361_0_0 -> 0 in the cache */
   /* search361_0_1 -> 3 in the cache */
   /* search361_0 -> 208 in the cache */
   /* search361_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 208, 0 },
      -1,
   } },

   /* replace361_0 -> 3 in the cache */
   /* replace361_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 3, 86 },
      -1,
   } },

   /* ('ine', ('iadd', 'a', 'b'), 'a') => ('ine', 'b', 0) */
   /* search362_0_0 -> 0 in the cache */
   /* search362_0_1 -> 3 in the cache */
   /* search362_0 -> 208 in the cache */
   /* search362_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 208, 0 },
      -1,
   } },

   /* replace362_0 -> 3 in the cache */
   /* replace362_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 3, 86 },
      -1,
   } },

   /* ('feq', ('b2f', 'a@1'), 0.0) => ('inot', 'a') */
   /* search363_0_0 -> 760 in the cache */
   /* search363_0 -> 1129 in the cache */
   /* search363_1 -> 1128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1129, 1128 },
      -1,
   } },

   /* replace363_0 -> 760 in the cache */
   /* replace363 -> 1131 in the cache */

   /* ('fneu', ('b2f', 'a@1'), 0.0) => a */
   /* search364_0_0 -> 760 in the cache */
   /* search364_0 -> 1129 in the cache */
   /* search364_1 -> 1128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1129, 1128 },
      -1,
   } },

   /* replace364 -> 760 in the cache */

   /* ('ieq', ('b2i', 'a@1'), 0) => ('inot', 'a') */
   /* search365_0_0 -> 760 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 0,
      { 760 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 0 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1228, 1229 },
      -1,
   } },

   /* replace365_0 -> 760 in the cache */
   /* replace365 -> 1131 in the cache */

   /* ('ine', ('b2i', 'a@1'), 0) => a */
   /* search366_0_0 -> 760 in the cache */
   /* search366_0 -> 1228 in the cache */
   /* search366_1 -> 1229 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1228, 1229 },
      -1,
   } },

   /* replace366 -> 760 in the cache */

   /* ('fneu', ('u2f', 'a'), 0.0) => ('ine', 'a', 0) */
   /* search367_0_0 -> 32 in the cache */
   /* search367_0 -> 104 in the cache */
   /* search367_1 -> 1128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 104, 1128 },
      -1,
   } },

   /* replace367_0 -> 32 in the cache */
   /* replace367_1 -> 61 in the cache */
   /* replace367 -> 1110 in the cache */

   /* ('feq', ('u2f', 'a'), 0.0) => ('ieq', 'a', 0) */
   /* search368_0_0 -> 32 in the cache */
   /* search368_0 -> 104 in the cache */
   /* search368_1 -> 1128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 104, 1128 },
      -1,
   } },

   /* replace368_0 -> 32 in the cache */
   /* replace368_1 -> 61 in the cache */
   /* replace368 -> 1108 in the cache */

   /* ('fge', ('u2f', 'a'), 0.0) => True */
   /* search369_0_0 -> 32 in the cache */
   /* search369_0 -> 104 in the cache */
   /* search369_1 -> 1128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 104, 1128 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 1 },
      nir_type_bool, { NIR_TRUE /* True */ },
   } },

   /* ('fge', 0.0, ('u2f', 'a')) => ('uge', 0, 'a') */
   /* search370_0 -> 1128 in the cache */
   /* search370_1_0 -> 32 in the cache */
   /* search370_1 -> 104 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1128, 104 },
      -1,
   } },

   /* replace370_0 -> 61 in the cache */
   /* replace370_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 61, 32 },
      -1,
   } },

   /* ('flt', ('u2f', 'a'), 0.0) => False */
   /* search371_0_0 -> 32 in the cache */
   /* search371_0 -> 104 in the cache */
   /* search371_1 -> 1128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 104, 1128 },
      -1,
   } },

   /* replace371 -> 1077 in the cache */

   /* ('flt', 0.0, ('u2f', 'a')) => ('ult', 0, 'a') */
   /* search372_0 -> 1128 in the cache */
   /* search372_1_0 -> 32 in the cache */
   /* search372_1 -> 104 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1128, 104 },
      -1,
   } },

   /* replace372_0 -> 61 in the cache */
   /* replace372_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 61, 32 },
      -1,
   } },

   /* ('fneu', ('i2f', 'a'), 0.0) => ('ine', 'a', 0) */
   /* search373_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_i2f,
      -1, 0,
      { 32 },
      -1,
   } },
   /* search373_1 -> 1128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1241, 1128 },
      -1,
   } },

   /* replace373_0 -> 32 in the cache */
   /* replace373_1 -> 61 in the cache */
   /* replace373 -> 1110 in the cache */

   /* ('feq', ('i2f', 'a'), 0.0) => ('ieq', 'a', 0) */
   /* search374_0_0 -> 32 in the cache */
   /* search374_0 -> 1241 in the cache */
   /* search374_1 -> 1128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1241, 1128 },
      -1,
   } },

   /* replace374_0 -> 32 in the cache */
   /* replace374_1 -> 61 in the cache */
   /* replace374 -> 1108 in the cache */

   /* ('fge', ('i2f', 'a'), 0.0) => ('ige', 'a', 0) */
   /* search375_0_0 -> 32 in the cache */
   /* search375_0 -> 1241 in the cache */
   /* search375_1 -> 1128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1241, 1128 },
      -1,
   } },

   /* replace375_0 -> 32 in the cache */
   /* replace375_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 32, 61 },
      -1,
   } },

   /* ('fge', 0.0, ('i2f', 'a')) => ('ige', 0, 'a') */
   /* search376_0 -> 1128 in the cache */
   /* search376_1_0 -> 32 in the cache */
   /* search376_1 -> 1241 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1128, 1241 },
      -1,
   } },

   /* replace376_0 -> 61 in the cache */
   /* replace376_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 61, 32 },
      -1,
   } },

   /* ('flt', ('i2f', 'a'), 0.0) => ('ilt', 'a', 0) */
   /* search377_0_0 -> 32 in the cache */
   /* search377_0 -> 1241 in the cache */
   /* search377_1 -> 1128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1241, 1128 },
      -1,
   } },

   /* replace377_0 -> 32 in the cache */
   /* replace377_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 32, 61 },
      -1,
   } },

   /* ('flt', 0.0, ('i2f', 'a')) => ('ilt', 0, 'a') */
   /* search378_0 -> 1128 in the cache */
   /* search378_1_0 -> 32 in the cache */
   /* search378_1 -> 1241 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1128, 1241 },
      -1,
   } },

   /* replace378_0 -> 61 in the cache */
   /* replace378_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 61, 32 },
      -1,
   } },

   /* ('~flt', 0.0, ('fabs', 'a')) => ('fneu', 'a', 0.0) */
   /* search379_0 -> 111 in the cache */
   /* search379_1_0 -> 32 in the cache */
   /* search379_1 -> 103 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 111, 103 },
      -1,
   } },

   /* replace379_0 -> 32 in the cache */
   /* replace379_1 -> 111 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 32, 111 },
      -1,
   } },

   /* ('~flt', ('fneg', ('fabs', 'a')), 0.0) => ('fneu', 'a', 0.0) */
   /* search380_0_0_0 -> 32 in the cache */
   /* search380_0_0 -> 103 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 103 },
      -1,
   } },
   /* search380_1 -> 111 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1254, 111 },
      -1,
   } },

   /* replace380_0 -> 32 in the cache */
   /* replace380_1 -> 111 in the cache */
   /* replace380 -> 1253 in the cache */

   /* ('fge', 0.0, ('fabs', 'a')) => ('feq', 'a', 0.0) */
   /* search381_0 -> 111 in the cache */
   /* search381_1_0 -> 32 in the cache */
   /* search381_1 -> 103 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 111, 103 },
      -1,
   } },

   /* replace381_0 -> 32 in the cache */
   /* replace381_1 -> 111 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 32, 111 },
      -1,
   } },

   /* ('fge', ('fneg', ('fabs', 'a')), 0.0) => ('feq', 'a', 0.0) */
   /* search382_0_0_0 -> 32 in the cache */
   /* search382_0_0 -> 103 in the cache */
   /* search382_0 -> 1254 in the cache */
   /* search382_1 -> 111 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1254, 111 },
      -1,
   } },

   /* replace382_0 -> 32 in the cache */
   /* replace382_1 -> 111 in the cache */
   /* replace382 -> 1257 in the cache */

   /* ('iand', ('fge', 'a', 0.0), ('fge', 1.0, 'a')) => ('feq', 'a', ('fsat', 'a')) */
   /* search383_0_0 -> 32 in the cache */
   /* search383_0_1 -> 111 in the cache */
   /* search383_0 -> 1174 in the cache */
   /* search383_1_0 -> 241 in the cache */
   /* search383_1_1 -> 32 in the cache */
   /* search383_1 -> 1100 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1174, 1100 },
      -1,
   } },

   /* replace383_0 -> 32 in the cache */
   /* replace383_1_0 -> 32 in the cache */
   /* replace383_1 -> 217 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 32, 217 },
      -1,
   } },

   /* ('fmax', ('b2f(is_used_once)', 'a@1'), ('b2f', 'b@1')) => ('b2f', ('ior', 'a', 'b')) */
   /* search384_0_0 -> 760 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 760 },
      0,
   } },
   /* search384_1_0 -> 1134 in the cache */
   /* search384_1 -> 1135 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 1261, 1135 },
      -1,
   } },

   /* replace384_0_0 -> 760 in the cache */
   /* replace384_0_1 -> 1134 in the cache */
   /* replace384_0 -> 1138 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 1,
      { 1138 },
      -1,
   } },

   /* ('fmax', ('fneg(is_used_once)', ('b2f(is_used_once)', 'a@1')), ('fneg', ('b2f', 'b@1'))) => ('fneg', ('b2f', ('iand', 'a', 'b'))) */
   /* search385_0_0_0 -> 760 in the cache */
   /* search385_0_0 -> 1261 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 1261 },
      0,
   } },
   /* search385_1_0_0 -> 1134 in the cache */
   /* search385_1_0 -> 1135 in the cache */
   /* search385_1 -> 1142 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 1264, 1142 },
      -1,
   } },

   /* replace385_0_0_0 -> 760 in the cache */
   /* replace385_0_0_1 -> 1134 in the cache */
   /* replace385_0_0 -> 1146 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 1,
      { 1146 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1266 },
      -1,
   } },

   /* ('fmin', ('b2f(is_used_once)', 'a@1'), ('b2f', 'b@1')) => ('b2f', ('iand', 'a', 'b')) */
   /* search386_0_0 -> 760 in the cache */
   /* search386_0 -> 1261 in the cache */
   /* search386_1_0 -> 1134 in the cache */
   /* search386_1 -> 1135 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 1261, 1135 },
      -1,
   } },

   /* replace386_0_0 -> 760 in the cache */
   /* replace386_0_1 -> 1134 in the cache */
   /* replace386_0 -> 1146 in the cache */
   /* replace386 -> 1266 in the cache */

   /* ('fmin', ('fneg(is_used_once)', ('b2f(is_used_once)', 'a@1')), ('fneg', ('b2f', 'b@1'))) => ('fneg', ('b2f', ('ior', 'a', 'b'))) */
   /* search387_0_0_0 -> 760 in the cache */
   /* search387_0_0 -> 1261 in the cache */
   /* search387_0 -> 1264 in the cache */
   /* search387_1_0_0 -> 1134 in the cache */
   /* search387_1_0 -> 1135 in the cache */
   /* search387_1 -> 1142 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 1264, 1142 },
      -1,
   } },

   /* replace387_0_0_0 -> 760 in the cache */
   /* replace387_0_0_1 -> 1134 in the cache */
   /* replace387_0_0 -> 1138 in the cache */
   /* replace387_0 -> 1263 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1263 },
      -1,
   } },

   /* ('fmin', ('b2f', 'a@1'), '#b') => ('bcsel', 'a', ('fmin', 'b', 1.0), ('fmin', 'b', 0.0)) */
   /* search388_0_0 -> 760 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 760 },
      -1,
   } },
   /* search388_1 -> 821 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 1271, 821 },
      -1,
   } },

   /* replace388_0 -> 760 in the cache */
   /* replace388_1_0 -> 3 in the cache */
   /* replace388_1_1 -> 270 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 3, 270 },
      -1,
   } },
   /* replace388_2_0 -> 3 in the cache */
   /* replace388_2_1 -> 256 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 3, 256 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 1273, 1274 },
      -1,
   } },

   /* ('flt', ('fadd(is_used_once)', 'a', ('fneg', 'b')), 0.0) => ('flt', 'a', 'b') */
   /* search389_0_0 -> 0 in the cache */
   /* search389_0_1_0 -> 3 in the cache */
   /* search389_0_1 -> 1087 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 0, 1087 },
      0,
   } },
   /* search389_1 -> 256 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 1276, 256 },
      -1,
   } },

   /* replace389_0 -> 0 in the cache */
   /* replace389_1 -> 3 in the cache */
   /* replace389 -> 1049 in the cache */

   /* ('fge', ('fneg', ('fabs', 'a')), 0.0) => ('feq', 'a', 0.0) */
   /* search390_0_0_0 -> 32 in the cache */
   /* search390_0_0 -> 103 in the cache */
   /* search390_0 -> 1254 in the cache */
   /* search390_1 -> 111 in the cache */
   /* search390 -> 1258 in the cache */

   /* replace390_0 -> 32 in the cache */
   /* replace390_1 -> 111 in the cache */
   /* replace390 -> 1257 in the cache */

   /* ('~bcsel', ('flt', 'b', 'a'), 'b', 'a') => ('fmin', 'a', 'b') */
   /* search391_0_0 -> 1117 in the cache */
   /* search391_0_1 -> 1114 in the cache */
   /* search391_0 -> 1118 in the cache */
   /* search391_1 -> 1117 in the cache */
   /* search391_2 -> 1114 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1118, 1117, 1114 },
      -1,
   } },

   /* replace391_0 -> 1114 in the cache */
   /* replace391_1 -> 1117 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 1114, 1117 },
      -1,
   } },

   /* ('~bcsel', ('flt', 'a', 'b'), 'b', 'a') => ('fmax', 'a', 'b') */
   /* search392_0_0 -> 0 in the cache */
   /* search392_0_1 -> 3 in the cache */
   /* search392_0 -> 1049 in the cache */
   /* search392_1 -> 3 in the cache */
   /* search392_2 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1049, 3, 0 },
      -1,
   } },

   /* replace392_0 -> 0 in the cache */
   /* replace392_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('~bcsel', ('fge', 'a', 'b'), 'b', 'a') => ('fmin', 'a', 'b') */
   /* search393_0_0 -> 0 in the cache */
   /* search393_0_1 -> 3 in the cache */
   /* search393_0 -> 1046 in the cache */
   /* search393_1 -> 3 in the cache */
   /* search393_2 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1046, 3, 0 },
      -1,
   } },

   /* replace393_0 -> 0 in the cache */
   /* replace393_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('~bcsel', ('fge', 'b', 'a'), 'b', 'a') => ('fmax', 'a', 'b') */
   /* search394_0_0 -> 1117 in the cache */
   /* search394_0_1 -> 1114 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1117, 1114 },
      -1,
   } },
   /* search394_1 -> 1117 in the cache */
   /* search394_2 -> 1114 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1284, 1117, 1114 },
      -1,
   } },

   /* replace394_0 -> 1114 in the cache */
   /* replace394_1 -> 1117 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 1114, 1117 },
      -1,
   } },

   /* ('bcsel', ('ult', 'b', 'a'), 'b', 'a') => ('umin', 'a', 'b') */
   /* search395_0_0 -> 1117 in the cache */
   /* search395_0_1 -> 1114 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 1117, 1114 },
      -1,
   } },
   /* search395_1 -> 1117 in the cache */
   /* search395_2 -> 1114 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1287, 1117, 1114 },
      -1,
   } },

   /* replace395_0 -> 1114 in the cache */
   /* replace395_1 -> 1117 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 1114, 1117 },
      -1,
   } },

   /* ('bcsel', ('ult', 'a', 'b'), 'b', 'a') => ('umax', 'a', 'b') */
   /* search396_0_0 -> 0 in the cache */
   /* search396_0_1 -> 3 in the cache */
   /* search396_0 -> 1067 in the cache */
   /* search396_1 -> 3 in the cache */
   /* search396_2 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1067, 3, 0 },
      -1,
   } },

   /* replace396_0 -> 0 in the cache */
   /* replace396_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('bcsel', ('uge', 'a', 'b'), 'b', 'a') => ('umin', 'a', 'b') */
   /* search397_0_0 -> 0 in the cache */
   /* search397_0_1 -> 3 in the cache */
   /* search397_0 -> 1061 in the cache */
   /* search397_1 -> 3 in the cache */
   /* search397_2 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1061, 3, 0 },
      -1,
   } },

   /* replace397_0 -> 0 in the cache */
   /* replace397_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('bcsel', ('uge', 'b', 'a'), 'b', 'a') => ('umax', 'a', 'b') */
   /* search398_0_0 -> 1117 in the cache */
   /* search398_0_1 -> 1114 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 1117, 1114 },
      -1,
   } },
   /* search398_1 -> 1117 in the cache */
   /* search398_2 -> 1114 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1294, 1117, 1114 },
      -1,
   } },

   /* replace398_0 -> 1114 in the cache */
   /* replace398_1 -> 1117 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 1114, 1117 },
      -1,
   } },

   /* ('bcsel', ('ilt', 'b', 'a'), 'b', 'a') => ('imin', 'a', 'b') */
   /* search399_0_0 -> 1117 in the cache */
   /* search399_0_1 -> 1114 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 1117, 1114 },
      -1,
   } },
   /* search399_1 -> 1117 in the cache */
   /* search399_2 -> 1114 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1297, 1117, 1114 },
      -1,
   } },

   /* replace399_0 -> 1114 in the cache */
   /* replace399_1 -> 1117 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 1114, 1117 },
      -1,
   } },

   /* ('bcsel', ('ilt', 'a', 'b'), 'b', 'a') => ('imax', 'a', 'b') */
   /* search400_0_0 -> 0 in the cache */
   /* search400_0_1 -> 3 in the cache */
   /* search400_0 -> 1064 in the cache */
   /* search400_1 -> 3 in the cache */
   /* search400_2 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1064, 3, 0 },
      -1,
   } },

   /* replace400_0 -> 0 in the cache */
   /* replace400_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('bcsel', ('ige', 'a', 'b'), 'b', 'a') => ('imin', 'a', 'b') */
   /* search401_0_0 -> 0 in the cache */
   /* search401_0_1 -> 3 in the cache */
   /* search401_0 -> 1058 in the cache */
   /* search401_1 -> 3 in the cache */
   /* search401_2 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1058, 3, 0 },
      -1,
   } },

   /* replace401_0 -> 0 in the cache */
   /* replace401_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('bcsel', ('ige', 'b', 'a'), 'b', 'a') => ('imax', 'a', 'b') */
   /* search402_0_0 -> 1117 in the cache */
   /* search402_0_1 -> 1114 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 1117, 1114 },
      -1,
   } },
   /* search402_1 -> 1117 in the cache */
   /* search402_2 -> 1114 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1304, 1117, 1114 },
      -1,
   } },

   /* replace402_0 -> 1114 in the cache */
   /* replace402_1 -> 1117 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 1114, 1117 },
      -1,
   } },

   /* ('bcsel', ('inot', 'a'), 'b', 'c') => ('bcsel', 'a', 'c', 'b') */
   /* search403_0_0 -> 760 in the cache */
   /* search403_0 -> 1131 in the cache */
   /* search403_1 -> 127 in the cache */
   /* search403_2 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1131, 127, 129 },
      -1,
   } },

   /* replace403_0 -> 760 in the cache */
   /* replace403_1 -> 129 in the cache */
   /* replace403_2 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 129, 127 },
      -1,
   } },

   /* ('bcsel', 'a', ('bcsel', 'a', 'b', 'c'), 'd') => ('bcsel', 'a', 'b', 'd') */
   /* search404_0 -> 760 in the cache */
   /* search404_1_0 -> 760 in the cache */
   /* search404_1_1 -> 150 in the cache */
   /* search404_1_2 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 150, 152 },
      -1,
   } },
   /* search404_2 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 1309, 153 },
      -1,
   } },

   /* replace404_0 -> 760 in the cache */
   /* replace404_1 -> 150 in the cache */
   /* replace404_2 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 150, 153 },
      -1,
   } },

   /* ('bcsel', 'a', 'b', ('bcsel', 'a', 'c', 'd')) => ('bcsel', 'a', 'b', 'd') */
   /* search405_0 -> 760 in the cache */
   /* search405_1 -> 150 in the cache */
   /* search405_2_0 -> 760 in the cache */
   /* search405_2_1 -> 152 in the cache */
   /* search405_2_2 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 152, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 150, 1312 },
      -1,
   } },

   /* replace405_0 -> 760 in the cache */
   /* replace405_1 -> 150 in the cache */
   /* replace405_2 -> 153 in the cache */
   /* replace405 -> 1311 in the cache */

   /* ('bcsel', 'a', ('bcsel', 'b', 'c', 'd'), ('bcsel(is_used_once)', 'b', 'c', 'e')) => ('bcsel', 'b', 'c', ('bcsel', 'a', 'd', 'e')) */
   /* search406_0 -> 760 in the cache */
   /* search406_1_0 -> 1134 in the cache */
   /* search406_1_1 -> 753 in the cache */
   /* search406_1_2 -> 762 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1134, 753, 762 },
      -1,
   } },
   /* search406_2_0 -> 1134 in the cache */
   /* search406_2_1 -> 753 in the cache */
   /* search406_2_2 -> 764 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1134, 753, 764 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 1314, 1315 },
      -1,
   } },

   /* replace406_0 -> 1134 in the cache */
   /* replace406_1 -> 753 in the cache */
   /* replace406_2_0 -> 760 in the cache */
   /* replace406_2_1 -> 762 in the cache */
   /* replace406_2_2 -> 764 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 762, 764 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1134, 753, 1317 },
      -1,
   } },

   /* ('bcsel', 'a', ('bcsel(is_used_once)', 'b', 'c', 'd'), ('bcsel', 'b', 'c', 'e')) => ('bcsel', 'b', 'c', ('bcsel', 'a', 'd', 'e')) */
   /* search407_0 -> 760 in the cache */
   /* search407_1_0 -> 1134 in the cache */
   /* search407_1_1 -> 753 in the cache */
   /* search407_1_2 -> 762 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1134, 753, 762 },
      0,
   } },
   /* search407_2_0 -> 1134 in the cache */
   /* search407_2_1 -> 753 in the cache */
   /* search407_2_2 -> 764 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1134, 753, 764 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 1319, 1320 },
      -1,
   } },

   /* replace407_0 -> 1134 in the cache */
   /* replace407_1 -> 753 in the cache */
   /* replace407_2_0 -> 760 in the cache */
   /* replace407_2_1 -> 762 in the cache */
   /* replace407_2_2 -> 764 in the cache */
   /* replace407_2 -> 1317 in the cache */
   /* replace407 -> 1318 in the cache */

   /* ('bcsel', 'a', ('bcsel', 'b', 'c', 'd'), ('bcsel(is_used_once)', 'b', 'e', 'd')) => ('bcsel', 'b', ('bcsel', 'a', 'c', 'e'), 'd') */
   /* search408_0 -> 760 in the cache */
   /* search408_1_0 -> 1134 in the cache */
   /* search408_1_1 -> 152 in the cache */
   /* search408_1_2 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1134, 152, 153 },
      -1,
   } },
   /* search408_2_0 -> 1134 in the cache */
   { .variable = {
      { nir_search_value_variable, -4 },
      4, /* e */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search408_2_2 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1134, 1323, 153 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 1322, 1324 },
      -1,
   } },

   /* replace408_0 -> 1134 in the cache */
   /* replace408_1_0 -> 760 in the cache */
   /* replace408_1_1 -> 152 in the cache */
   /* replace408_1_2 -> 1323 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 152, 1323 },
      -1,
   } },
   /* replace408_2 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1134, 1326, 153 },
      -1,
   } },

   /* ('bcsel', 'a', ('bcsel(is_used_once)', 'b', 'c', 'd'), ('bcsel', 'b', 'e', 'd')) => ('bcsel', 'b', ('bcsel', 'a', 'c', 'e'), 'd') */
   /* search409_0 -> 760 in the cache */
   /* search409_1_0 -> 1134 in the cache */
   /* search409_1_1 -> 152 in the cache */
   /* search409_1_2 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1134, 152, 153 },
      0,
   } },
   /* search409_2_0 -> 1134 in the cache */
   /* search409_2_1 -> 1323 in the cache */
   /* search409_2_2 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1134, 1323, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 1328, 1329 },
      -1,
   } },

   /* replace409_0 -> 1134 in the cache */
   /* replace409_1_0 -> 760 in the cache */
   /* replace409_1_1 -> 152 in the cache */
   /* replace409_1_2 -> 1323 in the cache */
   /* replace409_1 -> 1326 in the cache */
   /* replace409_2 -> 153 in the cache */
   /* replace409 -> 1327 in the cache */

   /* ('bcsel', 'a', True, 'b') => ('ior', 'a', 'b') */
   /* search410_0 -> 760 in the cache */
   /* search410_1 -> 1235 in the cache */
   /* search410_2 -> 1134 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 1235, 1134 },
      -1,
   } },

   /* replace410_0 -> 760 in the cache */
   /* replace410_1 -> 1134 in the cache */
   /* replace410 -> 1138 in the cache */

   /* ('bcsel', 'a', 'a', 'b') => ('ior', 'a', 'b') */
   /* search411_0 -> 760 in the cache */
   /* search411_1 -> 760 in the cache */
   /* search411_2 -> 1134 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 760, 1134 },
      -1,
   } },

   /* replace411_0 -> 760 in the cache */
   /* replace411_1 -> 1134 in the cache */
   /* replace411 -> 1138 in the cache */

   /* ('bcsel', 'a', 'b', False) => ('iand', 'a', 'b') */
   /* search412_0 -> 760 in the cache */
   /* search412_1 -> 1134 in the cache */
   /* search412_2 -> 1077 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 1134, 1077 },
      -1,
   } },

   /* replace412_0 -> 760 in the cache */
   /* replace412_1 -> 1134 in the cache */
   /* replace412 -> 1146 in the cache */

   /* ('bcsel', 'a', 'b', 'a') => ('iand', 'a', 'b') */
   /* search413_0 -> 760 in the cache */
   /* search413_1 -> 1134 in the cache */
   /* search413_2 -> 760 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 1134, 760 },
      -1,
   } },

   /* replace413_0 -> 760 in the cache */
   /* replace413_1 -> 1134 in the cache */
   /* replace413 -> 1146 in the cache */

   /* ('~fmin', 'a', 'a') => a */
   /* search414_0 -> 32 in the cache */
   /* search414_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmin,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace414 -> 32 in the cache */

   /* ('~fmax', 'a', 'a') => a */
   /* search415_0 -> 32 in the cache */
   /* search415_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmax,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace415 -> 32 in the cache */

   /* ('imin', 'a', 'a') => a */
   /* search416_0 -> 32 in the cache */
   /* search416_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imin,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace416 -> 32 in the cache */

   /* ('imax', 'a', 'a') => a */
   /* search417_0 -> 32 in the cache */
   /* search417_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imax,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace417 -> 32 in the cache */

   /* ('umin', 'a', 'a') => a */
   /* search418_0 -> 32 in the cache */
   /* search418_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_umin,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace418 -> 32 in the cache */

   /* ('umin', 'a', 0) => 0 */
   /* search419_0 -> 32 in the cache */
   /* search419_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 32, 61 },
      -1,
   } },

   /* replace419 -> 61 in the cache */

   /* ('umin', 'a', -1) => a */
   /* search420_0 -> 32 in the cache */
   /* search420_1 -> 63 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 32, 63 },
      -1,
   } },

   /* replace420 -> 32 in the cache */

   /* ('umax', 'a', 'a') => a */
   /* search421_0 -> 32 in the cache */
   /* search421_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_umax,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace421 -> 32 in the cache */

   /* ('umax', 'a', 0) => a */
   /* search422_0 -> 32 in the cache */
   /* search422_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 32, 61 },
      -1,
   } },

   /* replace422 -> 32 in the cache */

   /* ('umax', 'a', -1) => -1 */
   /* search423_0 -> 32 in the cache */
   /* search423_1 -> 63 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 32, 63 },
      -1,
   } },

   /* replace423 -> 63 in the cache */

   /* ('fmax', ('fmax', 'a', 'b'), 'b') => ('fmax', 'a', 'b') */
   /* search424_0_0 -> 0 in the cache */
   /* search424_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* search424_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 2,
      { 1345, 3 },
      -1,
   } },

   /* replace424_0 -> 0 in the cache */
   /* replace424_1 -> 3 in the cache */
   /* replace424 -> 1281 in the cache */

   /* ('umax', ('umax', 'a', 'b'), 'b') => ('umax', 'a', 'b') */
   /* search425_0_0 -> 0 in the cache */
   /* search425_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umax,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* search425_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umax,
      0, 2,
      { 1347, 3 },
      -1,
   } },

   /* replace425_0 -> 0 in the cache */
   /* replace425_1 -> 3 in the cache */
   /* replace425 -> 1291 in the cache */

   /* ('imax', ('imax', 'a', 'b'), 'b') => ('imax', 'a', 'b') */
   /* search426_0_0 -> 0 in the cache */
   /* search426_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imax,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* search426_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imax,
      0, 2,
      { 1349, 3 },
      -1,
   } },

   /* replace426_0 -> 0 in the cache */
   /* replace426_1 -> 3 in the cache */
   /* replace426 -> 1301 in the cache */

   /* ('fmin', ('fmin', 'a', 'b'), 'b') => ('fmin', 'a', 'b') */
   /* search427_0_0 -> 0 in the cache */
   /* search427_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* search427_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 2,
      { 1351, 3 },
      -1,
   } },

   /* replace427_0 -> 0 in the cache */
   /* replace427_1 -> 3 in the cache */
   /* replace427 -> 1283 in the cache */

   /* ('umin', ('umin', 'a', 'b'), 'b') => ('umin', 'a', 'b') */
   /* search428_0_0 -> 0 in the cache */
   /* search428_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* search428_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umin,
      0, 2,
      { 1353, 3 },
      -1,
   } },

   /* replace428_0 -> 0 in the cache */
   /* replace428_1 -> 3 in the cache */
   /* replace428 -> 1293 in the cache */

   /* ('imin', ('imin', 'a', 'b'), 'b') => ('imin', 'a', 'b') */
   /* search429_0_0 -> 0 in the cache */
   /* search429_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imin,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* search429_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imin,
      0, 2,
      { 1355, 3 },
      -1,
   } },

   /* replace429_0 -> 0 in the cache */
   /* replace429_1 -> 3 in the cache */
   /* replace429 -> 1303 in the cache */

   /* ('fmax', ('fmax', ('fmax', 'a', 'b'), 'c'), 'a') => ('fmax', ('fmax', 'a', 'b'), 'c') */
   /* search430_0_0_0 -> 126 in the cache */
   /* search430_0_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmax,
      2, 1,
      { 126, 127 },
      -1,
   } },
   /* search430_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 2,
      { 1357, 129 },
      -1,
   } },
   /* search430_1 -> 126 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 3,
      { 1358, 126 },
      -1,
   } },

   /* replace430_0_0 -> 126 in the cache */
   /* replace430_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 126, 127 },
      -1,
   } },
   /* replace430_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 2,
      { 1360, 129 },
      -1,
   } },

   /* ('umax', ('umax', ('umax', 'a', 'b'), 'c'), 'a') => ('umax', ('umax', 'a', 'b'), 'c') */
   /* search431_0_0_0 -> 126 in the cache */
   /* search431_0_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umax,
      2, 1,
      { 126, 127 },
      -1,
   } },
   /* search431_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umax,
      1, 2,
      { 1362, 129 },
      -1,
   } },
   /* search431_1 -> 126 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umax,
      0, 3,
      { 1363, 126 },
      -1,
   } },

   /* replace431_0_0 -> 126 in the cache */
   /* replace431_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umax,
      1, 1,
      { 126, 127 },
      -1,
   } },
   /* replace431_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umax,
      0, 2,
      { 1365, 129 },
      -1,
   } },

   /* ('imax', ('imax', ('imax', 'a', 'b'), 'c'), 'a') => ('imax', ('imax', 'a', 'b'), 'c') */
   /* search432_0_0_0 -> 126 in the cache */
   /* search432_0_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imax,
      2, 1,
      { 126, 127 },
      -1,
   } },
   /* search432_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imax,
      1, 2,
      { 1367, 129 },
      -1,
   } },
   /* search432_1 -> 126 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imax,
      0, 3,
      { 1368, 126 },
      -1,
   } },

   /* replace432_0_0 -> 126 in the cache */
   /* replace432_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imax,
      1, 1,
      { 126, 127 },
      -1,
   } },
   /* replace432_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imax,
      0, 2,
      { 1370, 129 },
      -1,
   } },

   /* ('fmin', ('fmin', ('fmin', 'a', 'b'), 'c'), 'a') => ('fmin', ('fmin', 'a', 'b'), 'c') */
   /* search433_0_0_0 -> 126 in the cache */
   /* search433_0_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmin,
      2, 1,
      { 126, 127 },
      -1,
   } },
   /* search433_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 2,
      { 1372, 129 },
      -1,
   } },
   /* search433_1 -> 126 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 3,
      { 1373, 126 },
      -1,
   } },

   /* replace433_0_0 -> 126 in the cache */
   /* replace433_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 126, 127 },
      -1,
   } },
   /* replace433_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 2,
      { 1375, 129 },
      -1,
   } },

   /* ('umin', ('umin', ('umin', 'a', 'b'), 'c'), 'a') => ('umin', ('umin', 'a', 'b'), 'c') */
   /* search434_0_0_0 -> 126 in the cache */
   /* search434_0_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umin,
      2, 1,
      { 126, 127 },
      -1,
   } },
   /* search434_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umin,
      1, 2,
      { 1377, 129 },
      -1,
   } },
   /* search434_1 -> 126 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umin,
      0, 3,
      { 1378, 126 },
      -1,
   } },

   /* replace434_0_0 -> 126 in the cache */
   /* replace434_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 126, 127 },
      -1,
   } },
   /* replace434_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umin,
      0, 2,
      { 1380, 129 },
      -1,
   } },

   /* ('imin', ('imin', ('imin', 'a', 'b'), 'c'), 'a') => ('imin', ('imin', 'a', 'b'), 'c') */
   /* search435_0_0_0 -> 126 in the cache */
   /* search435_0_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imin,
      2, 1,
      { 126, 127 },
      -1,
   } },
   /* search435_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imin,
      1, 2,
      { 1382, 129 },
      -1,
   } },
   /* search435_1 -> 126 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imin,
      0, 3,
      { 1383, 126 },
      -1,
   } },

   /* replace435_0_0 -> 126 in the cache */
   /* replace435_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imin,
      1, 1,
      { 126, 127 },
      -1,
   } },
   /* replace435_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imin,
      0, 2,
      { 1385, 129 },
      -1,
   } },

   /* ('ieq', ('b2i8', 'a@1'), ('b2i8', 'b@1')) => ('ieq', 'a', 'b') */
   /* search436_0_0 -> 760 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_b2i8,
      -1, 0,
      { 760 },
      -1,
   } },
   /* search436_1_0 -> 1134 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_b2i8,
      -1, 0,
      { 1134 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1387, 1388 },
      -1,
   } },

   /* replace436_0 -> 760 in the cache */
   /* replace436_1 -> 1134 in the cache */
   /* replace436 -> 1162 in the cache */

   /* ('ine', ('b2i8', 'a@1'), ('b2i8', 'b@1')) => ('ine', 'a', 'b') */
   /* search437_0_0 -> 760 in the cache */
   /* search437_0 -> 1387 in the cache */
   /* search437_1_0 -> 1134 in the cache */
   /* search437_1 -> 1388 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1387, 1388 },
      -1,
   } },

   /* replace437_0 -> 760 in the cache */
   /* replace437_1 -> 1134 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 760, 1134 },
      -1,
   } },

   /* ('ieq', ('b2i16', 'a@1'), ('b2i16', 'b@1')) => ('ieq', 'a', 'b') */
   /* search438_0_0 -> 760 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2i16,
      -1, 0,
      { 760 },
      -1,
   } },
   /* search438_1_0 -> 1134 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2i16,
      -1, 0,
      { 1134 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1392, 1393 },
      -1,
   } },

   /* replace438_0 -> 760 in the cache */
   /* replace438_1 -> 1134 in the cache */
   /* replace438 -> 1162 in the cache */

   /* ('ine', ('b2i16', 'a@1'), ('b2i16', 'b@1')) => ('ine', 'a', 'b') */
   /* search439_0_0 -> 760 in the cache */
   /* search439_0 -> 1392 in the cache */
   /* search439_1_0 -> 1134 in the cache */
   /* search439_1 -> 1393 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1392, 1393 },
      -1,
   } },

   /* replace439_0 -> 760 in the cache */
   /* replace439_1 -> 1134 in the cache */
   /* replace439 -> 1391 in the cache */

   /* ('ieq', ('b2i32', 'a@1'), ('b2i32', 'b@1')) => ('ieq', 'a', 'b') */
   /* search440_0_0 -> 760 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 760 },
      -1,
   } },
   /* search440_1_0 -> 1134 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 1134 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1396, 1397 },
      -1,
   } },

   /* replace440_0 -> 760 in the cache */
   /* replace440_1 -> 1134 in the cache */
   /* replace440 -> 1162 in the cache */

   /* ('ine', ('b2i32', 'a@1'), ('b2i32', 'b@1')) => ('ine', 'a', 'b') */
   /* search441_0_0 -> 760 in the cache */
   /* search441_0 -> 1396 in the cache */
   /* search441_1_0 -> 1134 in the cache */
   /* search441_1 -> 1397 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1396, 1397 },
      -1,
   } },

   /* replace441_0 -> 760 in the cache */
   /* replace441_1 -> 1134 in the cache */
   /* replace441 -> 1391 in the cache */

   /* ('ieq', ('b2i64', 'a@1'), ('b2i64', 'b@1')) => ('ieq', 'a', 'b') */
   /* search442_0_0 -> 760 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2i64,
      -1, 0,
      { 760 },
      -1,
   } },
   /* search442_1_0 -> 1134 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2i64,
      -1, 0,
      { 1134 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1400, 1401 },
      -1,
   } },

   /* replace442_0 -> 760 in the cache */
   /* replace442_1 -> 1134 in the cache */
   /* replace442 -> 1162 in the cache */

   /* ('ine', ('b2i64', 'a@1'), ('b2i64', 'b@1')) => ('ine', 'a', 'b') */
   /* search443_0_0 -> 760 in the cache */
   /* search443_0 -> 1400 in the cache */
   /* search443_1_0 -> 1134 in the cache */
   /* search443_1 -> 1401 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1400, 1401 },
      -1,
   } },

   /* replace443_0 -> 760 in the cache */
   /* replace443_1 -> 1134 in the cache */
   /* replace443 -> 1391 in the cache */

   /* ('feq', ('b2f16', 'a@1'), ('b2f16', 'b@1')) => ('ieq', 'a', 'b') */
   /* search444_0_0 -> 760 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2f16,
      -1, 0,
      { 760 },
      -1,
   } },
   /* search444_1_0 -> 1134 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2f16,
      -1, 0,
      { 1134 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1404, 1405 },
      -1,
   } },

   /* replace444_0 -> 760 in the cache */
   /* replace444_1 -> 1134 in the cache */
   /* replace444 -> 1162 in the cache */

   /* ('fneu', ('b2f16', 'a@1'), ('b2f16', 'b@1')) => ('ine', 'a', 'b') */
   /* search445_0_0 -> 760 in the cache */
   /* search445_0 -> 1404 in the cache */
   /* search445_1_0 -> 1134 in the cache */
   /* search445_1 -> 1405 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1404, 1405 },
      -1,
   } },

   /* replace445_0 -> 760 in the cache */
   /* replace445_1 -> 1134 in the cache */
   /* replace445 -> 1391 in the cache */

   /* ('feq', ('b2f32', 'a@1'), ('b2f32', 'b@1')) => ('ieq', 'a', 'b') */
   /* search446_0_0 -> 760 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2f32,
      -1, 0,
      { 760 },
      -1,
   } },
   /* search446_1_0 -> 1134 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2f32,
      -1, 0,
      { 1134 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1408, 1409 },
      -1,
   } },

   /* replace446_0 -> 760 in the cache */
   /* replace446_1 -> 1134 in the cache */
   /* replace446 -> 1162 in the cache */

   /* ('fneu', ('b2f32', 'a@1'), ('b2f32', 'b@1')) => ('ine', 'a', 'b') */
   /* search447_0_0 -> 760 in the cache */
   /* search447_0 -> 1408 in the cache */
   /* search447_1_0 -> 1134 in the cache */
   /* search447_1 -> 1409 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1408, 1409 },
      -1,
   } },

   /* replace447_0 -> 760 in the cache */
   /* replace447_1 -> 1134 in the cache */
   /* replace447 -> 1391 in the cache */

   /* ('feq', ('b2f64', 'a@1'), ('b2f64', 'b@1')) => ('ieq', 'a', 'b') */
   /* search448_0_0 -> 760 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2f64,
      -1, 0,
      { 760 },
      -1,
   } },
   /* search448_1_0 -> 1134 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2f64,
      -1, 0,
      { 1134 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1412, 1413 },
      -1,
   } },

   /* replace448_0 -> 760 in the cache */
   /* replace448_1 -> 1134 in the cache */
   /* replace448 -> 1162 in the cache */

   /* ('fneu', ('b2f64', 'a@1'), ('b2f64', 'b@1')) => ('ine', 'a', 'b') */
   /* search449_0_0 -> 760 in the cache */
   /* search449_0 -> 1412 in the cache */
   /* search449_1_0 -> 1134 in the cache */
   /* search449_1 -> 1413 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1412, 1413 },
      -1,
   } },

   /* replace449_0 -> 760 in the cache */
   /* replace449_1 -> 1134 in the cache */
   /* replace449 -> 1391 in the cache */

   /* ('iand@8', 'a', ('inot', ('ishr', 'a', 7))) => ('imax', 'a', 0) */
   /* search450_0 -> 6 in the cache */
   /* search450_1_0_0 -> 6 in the cache */
   /* search450_1_0_1 -> 9 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 6, 9 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 1416 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 6, 1417 },
      -1,
   } },

   /* replace450_0 -> 6 in the cache */
   /* replace450_1 -> 835 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 6, 835 },
      -1,
   } },

   /* ('ieq', ('iand', 'a@8', 128), 0) => ('ige', 'a', 0) */
   /* search451_0_0 -> 6 in the cache */
   /* search451_0_1 -> 7 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 6, 7 },
      -1,
   } },
   /* search451_1 -> 835 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 1420, 835 },
      -1,
   } },

   /* replace451_0 -> 6 in the cache */
   /* replace451_1 -> 835 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 6, 835 },
      -1,
   } },

   /* ('ine', ('iand', 'a@8', 128), 128) => ('ige', 'a', 0) */
   /* search452_0_0 -> 6 in the cache */
   /* search452_0_1 -> 7 in the cache */
   /* search452_0 -> 1420 in the cache */
   /* search452_1 -> 7 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 1420, 7 },
      -1,
   } },

   /* replace452_0 -> 6 in the cache */
   /* replace452_1 -> 835 in the cache */
   /* replace452 -> 1422 in the cache */

   /* ('ine', ('iand', 'a@8', 128), 0) => ('ilt', 'a', 0) */
   /* search453_0_0 -> 6 in the cache */
   /* search453_0_1 -> 7 in the cache */
   /* search453_0 -> 1420 in the cache */
   /* search453_1 -> 835 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 1420, 835 },
      -1,
   } },

   /* replace453_0 -> 6 in the cache */
   /* replace453_1 -> 835 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 6, 835 },
      -1,
   } },

   /* ('ieq', ('iand', 'a@8', 128), 128) => ('ilt', 'a', 0) */
   /* search454_0_0 -> 6 in the cache */
   /* search454_0_1 -> 7 in the cache */
   /* search454_0 -> 1420 in the cache */
   /* search454_1 -> 7 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 1420, 7 },
      -1,
   } },

   /* replace454_0 -> 6 in the cache */
   /* replace454_1 -> 835 in the cache */
   /* replace454 -> 1425 in the cache */

   /* ('ine', ('ushr', 'a@8', 7), 0) => ('ilt', 'a', 0) */
   /* search455_0_0 -> 6 in the cache */
   /* search455_0_1 -> 9 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 6, 9 },
      -1,
   } },
   /* search455_1 -> 835 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1427, 835 },
      -1,
   } },

   /* replace455_0 -> 6 in the cache */
   /* replace455_1 -> 835 in the cache */
   /* replace455 -> 1425 in the cache */

   /* ('ieq', ('ushr', 'a@8', 7), 0) => ('ige', 'a', 0) */
   /* search456_0_0 -> 6 in the cache */
   /* search456_0_1 -> 9 in the cache */
   /* search456_0 -> 1427 in the cache */
   /* search456_1 -> 835 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1427, 835 },
      -1,
   } },

   /* replace456_0 -> 6 in the cache */
   /* replace456_1 -> 835 in the cache */
   /* replace456 -> 1422 in the cache */

   /* ('ieq', ('ushr', 'a@8', 7), 1) => ('ilt', 'a', 0) */
   /* search457_0_0 -> 6 in the cache */
   /* search457_0_1 -> 9 in the cache */
   /* search457_0 -> 1427 in the cache */
   { .constant = {
      { nir_search_value_constant, 8 },
      nir_type_int, { 0x1 /* 1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1427, 1430 },
      -1,
   } },

   /* replace457_0 -> 6 in the cache */
   /* replace457_1 -> 835 in the cache */
   /* replace457 -> 1425 in the cache */

   /* ('ine', ('ushr', 'a@8', 7), 1) => ('ige', 'a', 0) */
   /* search458_0_0 -> 6 in the cache */
   /* search458_0_1 -> 9 in the cache */
   /* search458_0 -> 1427 in the cache */
   /* search458_1 -> 1430 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1427, 1430 },
      -1,
   } },

   /* replace458_0 -> 6 in the cache */
   /* replace458_1 -> 835 in the cache */
   /* replace458 -> 1422 in the cache */

   /* ('ine', ('ishr', 'a@8', 7), 0) => ('ilt', 'a', 0) */
   /* search459_0_0 -> 6 in the cache */
   /* search459_0_1 -> 9 in the cache */
   /* search459_0 -> 1416 in the cache */
   /* search459_1 -> 835 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1416, 835 },
      -1,
   } },

   /* replace459_0 -> 6 in the cache */
   /* replace459_1 -> 835 in the cache */
   /* replace459 -> 1425 in the cache */

   /* ('ieq', ('ishr', 'a@8', 7), 0) => ('ige', 'a', 0) */
   /* search460_0_0 -> 6 in the cache */
   /* search460_0_1 -> 9 in the cache */
   /* search460_0 -> 1416 in the cache */
   /* search460_1 -> 835 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1416, 835 },
      -1,
   } },

   /* replace460_0 -> 6 in the cache */
   /* replace460_1 -> 835 in the cache */
   /* replace460 -> 1422 in the cache */

   /* ('ieq', ('ishr', 'a@8', 7), -1) => ('ilt', 'a', 0) */
   /* search461_0_0 -> 6 in the cache */
   /* search461_0_1 -> 9 in the cache */
   /* search461_0 -> 1416 in the cache */
   { .constant = {
      { nir_search_value_constant, 8 },
      nir_type_int, { -0x1 /* -1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1416, 1435 },
      -1,
   } },

   /* replace461_0 -> 6 in the cache */
   /* replace461_1 -> 835 in the cache */
   /* replace461 -> 1425 in the cache */

   /* ('ine', ('ishr', 'a@8', 7), -1) => ('ige', 'a', 0) */
   /* search462_0_0 -> 6 in the cache */
   /* search462_0_1 -> 9 in the cache */
   /* search462_0 -> 1416 in the cache */
   /* search462_1 -> 1435 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1416, 1435 },
      -1,
   } },

   /* replace462_0 -> 6 in the cache */
   /* replace462_1 -> 835 in the cache */
   /* replace462 -> 1422 in the cache */

   /* ('iand@16', 'a', ('inot', ('ishr', 'a', 15))) => ('imax', 'a', 0) */
   /* search463_0 -> 11 in the cache */
   /* search463_1_0_0 -> 11 in the cache */
   /* search463_1_0_1 -> 14 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 11, 14 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 1438 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 11, 1439 },
      -1,
   } },

   /* replace463_0 -> 11 in the cache */
   /* replace463_1 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 11, 856 },
      -1,
   } },

   /* ('ieq', ('iand', 'a@16', 32768), 0) => ('ige', 'a', 0) */
   /* search464_0_0 -> 11 in the cache */
   /* search464_0_1 -> 12 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 11, 12 },
      -1,
   } },
   /* search464_1 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 1442, 856 },
      -1,
   } },

   /* replace464_0 -> 11 in the cache */
   /* replace464_1 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 11, 856 },
      -1,
   } },

   /* ('ine', ('iand', 'a@16', 32768), 32768) => ('ige', 'a', 0) */
   /* search465_0_0 -> 11 in the cache */
   /* search465_0_1 -> 12 in the cache */
   /* search465_0 -> 1442 in the cache */
   /* search465_1 -> 12 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 1442, 12 },
      -1,
   } },

   /* replace465_0 -> 11 in the cache */
   /* replace465_1 -> 856 in the cache */
   /* replace465 -> 1444 in the cache */

   /* ('ine', ('iand', 'a@16', 32768), 0) => ('ilt', 'a', 0) */
   /* search466_0_0 -> 11 in the cache */
   /* search466_0_1 -> 12 in the cache */
   /* search466_0 -> 1442 in the cache */
   /* search466_1 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 1442, 856 },
      -1,
   } },

   /* replace466_0 -> 11 in the cache */
   /* replace466_1 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 11, 856 },
      -1,
   } },

   /* ('ieq', ('iand', 'a@16', 32768), 32768) => ('ilt', 'a', 0) */
   /* search467_0_0 -> 11 in the cache */
   /* search467_0_1 -> 12 in the cache */
   /* search467_0 -> 1442 in the cache */
   /* search467_1 -> 12 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 1442, 12 },
      -1,
   } },

   /* replace467_0 -> 11 in the cache */
   /* replace467_1 -> 856 in the cache */
   /* replace467 -> 1447 in the cache */

   /* ('ine', ('ushr', 'a@16', 15), 0) => ('ilt', 'a', 0) */
   /* search468_0_0 -> 11 in the cache */
   /* search468_0_1 -> 14 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 11, 14 },
      -1,
   } },
   /* search468_1 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1449, 856 },
      -1,
   } },

   /* replace468_0 -> 11 in the cache */
   /* replace468_1 -> 856 in the cache */
   /* replace468 -> 1447 in the cache */

   /* ('ieq', ('ushr', 'a@16', 15), 0) => ('ige', 'a', 0) */
   /* search469_0_0 -> 11 in the cache */
   /* search469_0_1 -> 14 in the cache */
   /* search469_0 -> 1449 in the cache */
   /* search469_1 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1449, 856 },
      -1,
   } },

   /* replace469_0 -> 11 in the cache */
   /* replace469_1 -> 856 in the cache */
   /* replace469 -> 1444 in the cache */

   /* ('ieq', ('ushr', 'a@16', 15), 1) => ('ilt', 'a', 0) */
   /* search470_0_0 -> 11 in the cache */
   /* search470_0_1 -> 14 in the cache */
   /* search470_0 -> 1449 in the cache */
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { 0x1 /* 1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1449, 1452 },
      -1,
   } },

   /* replace470_0 -> 11 in the cache */
   /* replace470_1 -> 856 in the cache */
   /* replace470 -> 1447 in the cache */

   /* ('ine', ('ushr', 'a@16', 15), 1) => ('ige', 'a', 0) */
   /* search471_0_0 -> 11 in the cache */
   /* search471_0_1 -> 14 in the cache */
   /* search471_0 -> 1449 in the cache */
   /* search471_1 -> 1452 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1449, 1452 },
      -1,
   } },

   /* replace471_0 -> 11 in the cache */
   /* replace471_1 -> 856 in the cache */
   /* replace471 -> 1444 in the cache */

   /* ('ine', ('ishr', 'a@16', 15), 0) => ('ilt', 'a', 0) */
   /* search472_0_0 -> 11 in the cache */
   /* search472_0_1 -> 14 in the cache */
   /* search472_0 -> 1438 in the cache */
   /* search472_1 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1438, 856 },
      -1,
   } },

   /* replace472_0 -> 11 in the cache */
   /* replace472_1 -> 856 in the cache */
   /* replace472 -> 1447 in the cache */

   /* ('ieq', ('ishr', 'a@16', 15), 0) => ('ige', 'a', 0) */
   /* search473_0_0 -> 11 in the cache */
   /* search473_0_1 -> 14 in the cache */
   /* search473_0 -> 1438 in the cache */
   /* search473_1 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1438, 856 },
      -1,
   } },

   /* replace473_0 -> 11 in the cache */
   /* replace473_1 -> 856 in the cache */
   /* replace473 -> 1444 in the cache */

   /* ('ieq', ('ishr', 'a@16', 15), -1) => ('ilt', 'a', 0) */
   /* search474_0_0 -> 11 in the cache */
   /* search474_0_1 -> 14 in the cache */
   /* search474_0 -> 1438 in the cache */
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { -0x1 /* -1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1438, 1457 },
      -1,
   } },

   /* replace474_0 -> 11 in the cache */
   /* replace474_1 -> 856 in the cache */
   /* replace474 -> 1447 in the cache */

   /* ('ine', ('ishr', 'a@16', 15), -1) => ('ige', 'a', 0) */
   /* search475_0_0 -> 11 in the cache */
   /* search475_0_1 -> 14 in the cache */
   /* search475_0 -> 1438 in the cache */
   /* search475_1 -> 1457 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1438, 1457 },
      -1,
   } },

   /* replace475_0 -> 11 in the cache */
   /* replace475_1 -> 856 in the cache */
   /* replace475 -> 1444 in the cache */

   /* ('iand@32', 'a', ('inot', ('ishr', 'a', 31))) => ('imax', 'a', 0) */
   /* search476_0 -> 16 in the cache */
   /* search476_1_0_0 -> 16 in the cache */
   /* search476_1_0_1 -> 19 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 16, 19 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 1460 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 16, 1461 },
      -1,
   } },

   /* replace476_0 -> 16 in the cache */
   /* replace476_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 16, 122 },
      -1,
   } },

   /* ('ieq', ('iand', 'a@32', 2147483648), 0) => ('ige', 'a', 0) */
   /* search477_0_0 -> 16 in the cache */
   /* search477_0_1 -> 17 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 16, 17 },
      -1,
   } },
   /* search477_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 1464, 122 },
      -1,
   } },

   /* replace477_0 -> 16 in the cache */
   /* replace477_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 16, 122 },
      -1,
   } },

   /* ('ine', ('iand', 'a@32', 2147483648), 2147483648) => ('ige', 'a', 0) */
   /* search478_0_0 -> 16 in the cache */
   /* search478_0_1 -> 17 in the cache */
   /* search478_0 -> 1464 in the cache */
   /* search478_1 -> 17 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 1464, 17 },
      -1,
   } },

   /* replace478_0 -> 16 in the cache */
   /* replace478_1 -> 122 in the cache */
   /* replace478 -> 1466 in the cache */

   /* ('ine', ('iand', 'a@32', 2147483648), 0) => ('ilt', 'a', 0) */
   /* search479_0_0 -> 16 in the cache */
   /* search479_0_1 -> 17 in the cache */
   /* search479_0 -> 1464 in the cache */
   /* search479_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 1464, 122 },
      -1,
   } },

   /* replace479_0 -> 16 in the cache */
   /* replace479_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 16, 122 },
      -1,
   } },

   /* ('ieq', ('iand', 'a@32', 2147483648), 2147483648) => ('ilt', 'a', 0) */
   /* search480_0_0 -> 16 in the cache */
   /* search480_0_1 -> 17 in the cache */
   /* search480_0 -> 1464 in the cache */
   /* search480_1 -> 17 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 1464, 17 },
      -1,
   } },

   /* replace480_0 -> 16 in the cache */
   /* replace480_1 -> 122 in the cache */
   /* replace480 -> 1469 in the cache */

   /* ('ine', ('ushr', 'a@32', 31), 0) => ('ilt', 'a', 0) */
   /* search481_0_0 -> 16 in the cache */
   /* search481_0_1 -> 19 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 16, 19 },
      -1,
   } },
   /* search481_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1471, 122 },
      -1,
   } },

   /* replace481_0 -> 16 in the cache */
   /* replace481_1 -> 122 in the cache */
   /* replace481 -> 1469 in the cache */

   /* ('ieq', ('ushr', 'a@32', 31), 0) => ('ige', 'a', 0) */
   /* search482_0_0 -> 16 in the cache */
   /* search482_0_1 -> 19 in the cache */
   /* search482_0 -> 1471 in the cache */
   /* search482_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1471, 122 },
      -1,
   } },

   /* replace482_0 -> 16 in the cache */
   /* replace482_1 -> 122 in the cache */
   /* replace482 -> 1466 in the cache */

   /* ('ieq', ('ushr', 'a@32', 31), 1) => ('ilt', 'a', 0) */
   /* search483_0_0 -> 16 in the cache */
   /* search483_0_1 -> 19 in the cache */
   /* search483_0 -> 1471 in the cache */
   /* search483_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1471, 396 },
      -1,
   } },

   /* replace483_0 -> 16 in the cache */
   /* replace483_1 -> 122 in the cache */
   /* replace483 -> 1469 in the cache */

   /* ('ine', ('ushr', 'a@32', 31), 1) => ('ige', 'a', 0) */
   /* search484_0_0 -> 16 in the cache */
   /* search484_0_1 -> 19 in the cache */
   /* search484_0 -> 1471 in the cache */
   /* search484_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1471, 396 },
      -1,
   } },

   /* replace484_0 -> 16 in the cache */
   /* replace484_1 -> 122 in the cache */
   /* replace484 -> 1466 in the cache */

   /* ('ine', ('ishr', 'a@32', 31), 0) => ('ilt', 'a', 0) */
   /* search485_0_0 -> 16 in the cache */
   /* search485_0_1 -> 19 in the cache */
   /* search485_0 -> 1460 in the cache */
   /* search485_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1460, 122 },
      -1,
   } },

   /* replace485_0 -> 16 in the cache */
   /* replace485_1 -> 122 in the cache */
   /* replace485 -> 1469 in the cache */

   /* ('ieq', ('ishr', 'a@32', 31), 0) => ('ige', 'a', 0) */
   /* search486_0_0 -> 16 in the cache */
   /* search486_0_1 -> 19 in the cache */
   /* search486_0 -> 1460 in the cache */
   /* search486_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1460, 122 },
      -1,
   } },

   /* replace486_0 -> 16 in the cache */
   /* replace486_1 -> 122 in the cache */
   /* replace486 -> 1466 in the cache */

   /* ('ieq', ('ishr', 'a@32', 31), -1) => ('ilt', 'a', 0) */
   /* search487_0_0 -> 16 in the cache */
   /* search487_0_1 -> 19 in the cache */
   /* search487_0 -> 1460 in the cache */
   /* search487_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1460, 124 },
      -1,
   } },

   /* replace487_0 -> 16 in the cache */
   /* replace487_1 -> 122 in the cache */
   /* replace487 -> 1469 in the cache */

   /* ('ine', ('ishr', 'a@32', 31), -1) => ('ige', 'a', 0) */
   /* search488_0_0 -> 16 in the cache */
   /* search488_0_1 -> 19 in the cache */
   /* search488_0 -> 1460 in the cache */
   /* search488_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1460, 124 },
      -1,
   } },

   /* replace488_0 -> 16 in the cache */
   /* replace488_1 -> 122 in the cache */
   /* replace488 -> 1466 in the cache */

   /* ('iand@64', 'a', ('inot', ('ishr', 'a', 63))) => ('imax', 'a', 0) */
   /* search489_0 -> 21 in the cache */
   /* search489_1_0_0 -> 21 in the cache */
   /* search489_1_0_1 -> 24 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 21, 24 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 1480 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 21, 1481 },
      -1,
   } },

   /* replace489_0 -> 21 in the cache */
   /* replace489_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 21, 897 },
      -1,
   } },

   /* ('ieq', ('iand', 'a@64', 9223372036854775808), 0) => ('ige', 'a', 0) */
   /* search490_0_0 -> 21 in the cache */
   /* search490_0_1 -> 22 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 21, 22 },
      -1,
   } },
   /* search490_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 1484, 897 },
      -1,
   } },

   /* replace490_0 -> 21 in the cache */
   /* replace490_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 21, 897 },
      -1,
   } },

   /* ('ine', ('iand', 'a@64', 9223372036854775808), 9223372036854775808) => ('ige', 'a', 0) */
   /* search491_0_0 -> 21 in the cache */
   /* search491_0_1 -> 22 in the cache */
   /* search491_0 -> 1484 in the cache */
   /* search491_1 -> 22 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 1484, 22 },
      -1,
   } },

   /* replace491_0 -> 21 in the cache */
   /* replace491_1 -> 897 in the cache */
   /* replace491 -> 1486 in the cache */

   /* ('ine', ('iand', 'a@64', 9223372036854775808), 0) => ('ilt', 'a', 0) */
   /* search492_0_0 -> 21 in the cache */
   /* search492_0_1 -> 22 in the cache */
   /* search492_0 -> 1484 in the cache */
   /* search492_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 1484, 897 },
      -1,
   } },

   /* replace492_0 -> 21 in the cache */
   /* replace492_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 21, 897 },
      -1,
   } },

   /* ('ieq', ('iand', 'a@64', 9223372036854775808), 9223372036854775808) => ('ilt', 'a', 0) */
   /* search493_0_0 -> 21 in the cache */
   /* search493_0_1 -> 22 in the cache */
   /* search493_0 -> 1484 in the cache */
   /* search493_1 -> 22 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 1484, 22 },
      -1,
   } },

   /* replace493_0 -> 21 in the cache */
   /* replace493_1 -> 897 in the cache */
   /* replace493 -> 1489 in the cache */

   /* ('ine', ('ushr', 'a@64', 63), 0) => ('ilt', 'a', 0) */
   /* search494_0_0 -> 21 in the cache */
   /* search494_0_1 -> 24 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 21, 24 },
      -1,
   } },
   /* search494_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1491, 897 },
      -1,
   } },

   /* replace494_0 -> 21 in the cache */
   /* replace494_1 -> 897 in the cache */
   /* replace494 -> 1489 in the cache */

   /* ('ieq', ('ushr', 'a@64', 63), 0) => ('ige', 'a', 0) */
   /* search495_0_0 -> 21 in the cache */
   /* search495_0_1 -> 24 in the cache */
   /* search495_0 -> 1491 in the cache */
   /* search495_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1491, 897 },
      -1,
   } },

   /* replace495_0 -> 21 in the cache */
   /* replace495_1 -> 897 in the cache */
   /* replace495 -> 1486 in the cache */

   /* ('ieq', ('ushr', 'a@64', 63), 1) => ('ilt', 'a', 0) */
   /* search496_0_0 -> 21 in the cache */
   /* search496_0_1 -> 24 in the cache */
   /* search496_0 -> 1491 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x1 /* 1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1491, 1494 },
      -1,
   } },

   /* replace496_0 -> 21 in the cache */
   /* replace496_1 -> 897 in the cache */
   /* replace496 -> 1489 in the cache */

   /* ('ine', ('ushr', 'a@64', 63), 1) => ('ige', 'a', 0) */
   /* search497_0_0 -> 21 in the cache */
   /* search497_0_1 -> 24 in the cache */
   /* search497_0 -> 1491 in the cache */
   /* search497_1 -> 1494 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1491, 1494 },
      -1,
   } },

   /* replace497_0 -> 21 in the cache */
   /* replace497_1 -> 897 in the cache */
   /* replace497 -> 1486 in the cache */

   /* ('ine', ('ishr', 'a@64', 63), 0) => ('ilt', 'a', 0) */
   /* search498_0_0 -> 21 in the cache */
   /* search498_0_1 -> 24 in the cache */
   /* search498_0 -> 1480 in the cache */
   /* search498_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1480, 897 },
      -1,
   } },

   /* replace498_0 -> 21 in the cache */
   /* replace498_1 -> 897 in the cache */
   /* replace498 -> 1489 in the cache */

   /* ('ieq', ('ishr', 'a@64', 63), 0) => ('ige', 'a', 0) */
   /* search499_0_0 -> 21 in the cache */
   /* search499_0_1 -> 24 in the cache */
   /* search499_0 -> 1480 in the cache */
   /* search499_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1480, 897 },
      -1,
   } },

   /* replace499_0 -> 21 in the cache */
   /* replace499_1 -> 897 in the cache */
   /* replace499 -> 1486 in the cache */

   /* ('ieq', ('ishr', 'a@64', 63), -1) => ('ilt', 'a', 0) */
   /* search500_0_0 -> 21 in the cache */
   /* search500_0_1 -> 24 in the cache */
   /* search500_0 -> 1480 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { -0x1 /* -1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1480, 1499 },
      -1,
   } },

   /* replace500_0 -> 21 in the cache */
   /* replace500_1 -> 897 in the cache */
   /* replace500 -> 1489 in the cache */

   /* ('ine', ('ishr', 'a@64', 63), -1) => ('ige', 'a', 0) */
   /* search501_0_0 -> 21 in the cache */
   /* search501_0_1 -> 24 in the cache */
   /* search501_0 -> 1480 in the cache */
   /* search501_1 -> 1499 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1480, 1499 },
      -1,
   } },

   /* replace501_0 -> 21 in the cache */
   /* replace501_1 -> 897 in the cache */
   /* replace501 -> 1486 in the cache */

   /* ('fmin', 'a', ('fneg', 'a')) => ('fneg', ('fabs', 'a')) */
   /* search502_0 -> 32 in the cache */
   /* search502_1_0 -> 32 in the cache */
   /* search502_1 -> 98 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 32, 98 },
      -1,
   } },

   /* replace502_0_0 -> 32 in the cache */
   /* replace502_0 -> 103 in the cache */
   /* replace502 -> 1254 in the cache */

   /* ('imin', 'a', ('ineg', 'a')) => ('ineg', ('iabs', 'a')) */
   /* search503_0 -> 32 in the cache */
   /* search503_1_0 -> 32 in the cache */
   /* search503_1 -> 100 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 32, 100 },
      -1,
   } },

   /* replace503_0_0 -> 32 in the cache */
   /* replace503_0 -> 106 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 106 },
      -1,
   } },

   /* ('fmin', 'a', ('fneg', ('fabs', 'a'))) => ('fneg', ('fabs', 'a')) */
   /* search504_0 -> 32 in the cache */
   /* search504_1_0_0 -> 32 in the cache */
   /* search504_1_0 -> 103 in the cache */
   /* search504_1 -> 1254 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 32, 1254 },
      -1,
   } },

   /* replace504_0_0 -> 32 in the cache */
   /* replace504_0 -> 103 in the cache */
   /* replace504 -> 1254 in the cache */

   /* ('imin', 'a', ('ineg', ('iabs', 'a'))) => ('ineg', ('iabs', 'a')) */
   /* search505_0 -> 32 in the cache */
   /* search505_1_0_0 -> 32 in the cache */
   /* search505_1_0 -> 106 in the cache */
   /* search505_1 -> 1504 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 32, 1504 },
      -1,
   } },

   /* replace505_0_0 -> 32 in the cache */
   /* replace505_0 -> 106 in the cache */
   /* replace505 -> 1504 in the cache */

   /* ('~fmin', 'a', ('fabs', 'a')) => a */
   /* search506_0 -> 32 in the cache */
   /* search506_1_0 -> 32 in the cache */
   /* search506_1 -> 103 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 32, 103 },
      -1,
   } },

   /* replace506 -> 32 in the cache */

   /* ('imin', 'a', ('iabs', 'a')) => a */
   /* search507_0 -> 32 in the cache */
   /* search507_1_0 -> 32 in the cache */
   /* search507_1 -> 106 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 32, 106 },
      -1,
   } },

   /* replace507 -> 32 in the cache */

   /* ('~fmax', 'a', ('fneg', ('fabs', 'a'))) => a */
   /* search508_0 -> 32 in the cache */
   /* search508_1_0_0 -> 32 in the cache */
   /* search508_1_0 -> 103 in the cache */
   /* search508_1 -> 1254 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 32, 1254 },
      -1,
   } },

   /* replace508 -> 32 in the cache */

   /* ('imax', 'a', ('ineg', ('iabs', 'a'))) => a */
   /* search509_0 -> 32 in the cache */
   /* search509_1_0_0 -> 32 in the cache */
   /* search509_1_0 -> 106 in the cache */
   /* search509_1 -> 1504 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 32, 1504 },
      -1,
   } },

   /* replace509 -> 32 in the cache */

   /* ('fmax', 'a', ('fabs', 'a')) => ('fabs', 'a') */
   /* search510_0 -> 32 in the cache */
   /* search510_1_0 -> 32 in the cache */
   /* search510_1 -> 103 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 32, 103 },
      -1,
   } },

   /* replace510_0 -> 32 in the cache */
   /* replace510 -> 103 in the cache */

   /* ('imax', 'a', ('iabs', 'a')) => ('iabs', 'a') */
   /* search511_0 -> 32 in the cache */
   /* search511_1_0 -> 32 in the cache */
   /* search511_1 -> 106 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 32, 106 },
      -1,
   } },

   /* replace511_0 -> 32 in the cache */
   /* replace511 -> 106 in the cache */

   /* ('fmax', 'a', ('fneg', 'a')) => ('fabs', 'a') */
   /* search512_0 -> 32 in the cache */
   /* search512_1_0 -> 32 in the cache */
   /* search512_1 -> 98 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 32, 98 },
      -1,
   } },

   /* replace512_0 -> 32 in the cache */
   /* replace512 -> 103 in the cache */

   /* ('imax', 'a', ('ineg', 'a')) => ('iabs', 'a') */
   /* search513_0 -> 32 in the cache */
   /* search513_1_0 -> 32 in the cache */
   /* search513_1 -> 100 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 32, 100 },
      -1,
   } },

   /* replace513_0 -> 32 in the cache */
   /* replace513 -> 106 in the cache */

   /* ('~fmax', ('fabs', 'a'), 0.0) => ('fabs', 'a') */
   /* search514_0_0 -> 32 in the cache */
   /* search514_0 -> 103 in the cache */
   /* search514_1 -> 111 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 103, 111 },
      -1,
   } },

   /* replace514_0 -> 32 in the cache */
   /* replace514 -> 103 in the cache */

   /* ('fmin', ('fmax', 'a', 0.0), 1.0) => ('fsat', 'a') */
   /* search515_0_0 -> 32 in the cache */
   /* search515_0_1 -> 111 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 32, 111 },
      -1,
   } },
   /* search515_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 2,
      { 1516, 241 },
      -1,
   } },

   /* replace515_0 -> 32 in the cache */
   /* replace515 -> 217 in the cache */

   /* ('~fmax', ('fmin', 'a', 1.0), 0.0) => ('fsat', 'a') */
   /* search516_0_0 -> 32 in the cache */
   /* search516_0_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 32, 241 },
      -1,
   } },
   /* search516_1 -> 111 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmax,
      0, 2,
      { 1518, 111 },
      -1,
   } },

   /* replace516_0 -> 32 in the cache */
   /* replace516 -> 217 in the cache */

   /* ('~fmin', ('fmax', 'a', -1.0), 0.0) => ('fneg', ('fsat', ('fneg', 'a'))) */
   /* search517_0_0 -> 32 in the cache */
   /* search517_0_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 32, 247 },
      -1,
   } },
   /* search517_1 -> 111 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmin,
      0, 2,
      { 1520, 111 },
      -1,
   } },

   /* replace517_0_0_0 -> 32 in the cache */
   /* replace517_0_0 -> 98 in the cache */
   /* replace517_0 -> 218 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 218 },
      -1,
   } },

   /* ('~fmax', ('fmin', 'a', 0.0), -1.0) => ('fneg', ('fsat', ('fneg', 'a'))) */
   /* search518_0_0 -> 32 in the cache */
   /* search518_0_1 -> 111 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 32, 111 },
      -1,
   } },
   /* search518_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmax,
      0, 2,
      { 1523, 247 },
      -1,
   } },

   /* replace518_0_0_0 -> 32 in the cache */
   /* replace518_0_0 -> 98 in the cache */
   /* replace518_0 -> 218 in the cache */
   /* replace518 -> 1522 in the cache */

   /* ('fsat', ('fsign', 'a')) => ('b2f', ('!flt', 0.0, 'a')) */
   /* search519_0_0 -> 32 in the cache */
   /* search519_0 -> 250 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 250 },
      -1,
   } },

   /* replace519_0_0 -> 111 in the cache */
   /* replace519_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_flt,
      -1, 0,
      { 111, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 1526 },
      -1,
   } },

   /* ('fsat', ('b2f', 'a')) => ('b2f', 'a') */
   /* search520_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 1528 },
      -1,
   } },

   /* replace520_0 -> 32 in the cache */
   /* replace520 -> 1528 in the cache */

   /* ('fsat', 'a') => ('fmin', ('fmax', 'a', 0.0), 1.0) */
   /* search521_0 -> 32 in the cache */
   /* search521 -> 217 in the cache */

   /* replace521_0_0 -> 32 in the cache */
   /* replace521_0_1 -> 111 in the cache */
   /* replace521_0 -> 1516 in the cache */
   /* replace521_1 -> 241 in the cache */
   /* replace521 -> 1517 in the cache */

   /* ('fsat', ('fsat', 'a')) => ('fsat', 'a') */
   /* search522_0_0 -> 32 in the cache */
   /* search522_0 -> 217 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 217 },
      -1,
   } },

   /* replace522_0 -> 32 in the cache */
   /* replace522 -> 217 in the cache */

   /* ('fsat', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', 'b'))) => ('fsat', ('fadd', ('fneg', 'a'), ('fneg', 'b'))) */
   /* search523_0_0_0 -> 0 in the cache */
   /* search523_0_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 0, 3 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1531 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 1532 },
      -1,
   } },

   /* replace523_0_0_0 -> 0 in the cache */
   /* replace523_0_0 -> 212 in the cache */
   /* replace523_0_1_0 -> 3 in the cache */
   /* replace523_0_1 -> 1087 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 212, 1087 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 1534 },
      -1,
   } },

   /* ('fsat', ('fneg(is_used_once)', ('fmul(is_used_once)', 'a', 'b'))) => ('fsat', ('fmul', ('fneg', 'a'), 'b')) */
   /* search524_0_0_0 -> 0 in the cache */
   /* search524_0_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 0, 3 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1536 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 1537 },
      -1,
   } },

   /* replace524_0_0_0 -> 0 in the cache */
   /* replace524_0_0 -> 212 in the cache */
   /* replace524_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 212, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 1539 },
      -1,
   } },

   /* ('fsat', ('fneg(is_used_once)', ('fmulz(is_used_once)', 'a', 'b'))) => ('fsat', ('fmulz', ('fneg', 'a'), 'b')) */
   /* search525_0_0_0 -> 16 in the cache */
   /* search525_0_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 16, 36 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1541 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 1542 },
      -1,
   } },

   /* replace525_0_0_0 -> 16 in the cache */
   /* replace525_0_0 -> 280 in the cache */
   /* replace525_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 280, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 1544 },
      -1,
   } },

   /* ('fsat', ('fabs(is_used_once)', ('fmul(is_used_once)', 'a', 'b'))) => ('fsat', ('fmul', ('fabs', 'a'), ('fabs', 'b'))) */
   /* search526_0_0_0 -> 0 in the cache */
   /* search526_0_0_1 -> 3 in the cache */
   /* search526_0_0 -> 1536 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 1,
      { 1536 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 1546 },
      -1,
   } },

   /* replace526_0_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace526_0_1_0 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 1548, 1549 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 1550 },
      -1,
   } },

   /* ('fmin', ('fmax', ('fmin', ('fmax', 'a', 'b'), 'c'), 'b'), 'c') => ('fmin', ('fmax', 'a', 'b'), 'c') */
   /* search527_0_0_0_0 -> 126 in the cache */
   /* search527_0_0_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmax,
      3, 1,
      { 126, 127 },
      -1,
   } },
   /* search527_0_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmin,
      2, 2,
      { 1552, 129 },
      -1,
   } },
   /* search527_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 3,
      { 1553, 127 },
      -1,
   } },
   /* search527_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 4,
      { 1554, 129 },
      -1,
   } },

   /* replace527_0_0 -> 126 in the cache */
   /* replace527_0_1 -> 127 in the cache */
   /* replace527_0 -> 1360 in the cache */
   /* replace527_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 2,
      { 1360, 129 },
      -1,
   } },

   /* ('imin', ('imax', ('imin', ('imax', 'a', 'b'), 'c'), 'b'), 'c') => ('imin', ('imax', 'a', 'b'), 'c') */
   /* search528_0_0_0_0 -> 126 in the cache */
   /* search528_0_0_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imax,
      3, 1,
      { 126, 127 },
      -1,
   } },
   /* search528_0_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imin,
      2, 2,
      { 1557, 129 },
      -1,
   } },
   /* search528_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imax,
      1, 3,
      { 1558, 127 },
      -1,
   } },
   /* search528_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imin,
      0, 4,
      { 1559, 129 },
      -1,
   } },

   /* replace528_0_0 -> 126 in the cache */
   /* replace528_0_1 -> 127 in the cache */
   /* replace528_0 -> 1370 in the cache */
   /* replace528_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imin,
      0, 2,
      { 1370, 129 },
      -1,
   } },

   /* ('umin', ('umax', ('umin', ('umax', 'a', 'b'), 'c'), 'b'), 'c') => ('umin', ('umax', 'a', 'b'), 'c') */
   /* search529_0_0_0_0 -> 126 in the cache */
   /* search529_0_0_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umax,
      3, 1,
      { 126, 127 },
      -1,
   } },
   /* search529_0_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umin,
      2, 2,
      { 1562, 129 },
      -1,
   } },
   /* search529_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umax,
      1, 3,
      { 1563, 127 },
      -1,
   } },
   /* search529_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umin,
      0, 4,
      { 1564, 129 },
      -1,
   } },

   /* replace529_0_0 -> 126 in the cache */
   /* replace529_0_1 -> 127 in the cache */
   /* replace529_0 -> 1365 in the cache */
   /* replace529_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umin,
      0, 2,
      { 1365, 129 },
      -1,
   } },

   /* ('fmax', ('fsat', 'a'), '#b(is_zero_to_one)') => ('fsat', ('fmax', 'a', 'b')) */
   /* search530_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 0 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      true,
      nir_type_invalid,
      18,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 1567, 1568 },
      -1,
   } },

   /* replace530_0_0 -> 0 in the cache */
   /* replace530_0_1 -> 3 in the cache */
   /* replace530_0 -> 1281 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 1281 },
      -1,
   } },

   /* ('~fmin', ('fsat', 'a'), '#b(is_zero_to_one)') => ('fsat', ('fmin', 'a', 'b')) */
   /* search531_0_0 -> 0 in the cache */
   /* search531_0 -> 1567 in the cache */
   /* search531_1 -> 1568 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 1567, 1568 },
      -1,
   } },

   /* replace531_0_0 -> 0 in the cache */
   /* replace531_0_1 -> 3 in the cache */
   /* replace531_0 -> 1283 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 1283 },
      -1,
   } },

   /* ('fmax', ('fneg', ('fmin', 'b', 'a')), 'b') => ('fmax', ('fabs', 'b'), ('fneg', 'a')) */
   /* search532_0_0_0 -> 1117 in the cache */
   /* search532_0_0_1 -> 1114 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 1117, 1114 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1573 },
      -1,
   } },
   /* search532_1 -> 1117 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 2,
      { 1574, 1117 },
      -1,
   } },

   /* replace532_0_0 -> 1117 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 1117 },
      -1,
   } },
   /* replace532_1_0 -> 1114 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 1114 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 1576, 1577 },
      -1,
   } },

   /* ('fmin', ('fneg', ('fmax', 'b', 'a')), 'b') => ('fmin', ('fneg', ('fabs', 'b')), ('fneg', 'a')) */
   /* search533_0_0_0 -> 1117 in the cache */
   /* search533_0_0_1 -> 1114 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 1117, 1114 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1579 },
      -1,
   } },
   /* search533_1 -> 1117 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 2,
      { 1580, 1117 },
      -1,
   } },

   /* replace533_0_0_0 -> 1117 in the cache */
   /* replace533_0_0 -> 1576 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 1576 },
      -1,
   } },
   /* replace533_1_0 -> 1114 in the cache */
   /* replace533_1 -> 1577 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 1582, 1577 },
      -1,
   } },

   /* ('fmax', ('fadd(is_used_once)', ('fneg', 'a(is_not_negative)'), '#b(is_zero_to_one)'), 0.0) => ('fsat', ('fadd', ('fneg', 'a'), 'b')) */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      19,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 1584 },
      -1,
   } },
   /* search534_0_1 -> 1568 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 1585, 1568 },
      0,
   } },
   /* search534_1 -> 256 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 2,
      { 1586, 256 },
      -1,
   } },

   /* replace534_0_0_0 -> 0 in the cache */
   /* replace534_0_0 -> 212 in the cache */
   /* replace534_0_1 -> 3 in the cache */
   /* replace534_0 -> 277 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 277 },
      -1,
   } },

   /* ('extract_u8', ('imin', ('imax', 'a', 0), 255), 0) => ('imin', ('imax', 'a', 0), 255) */
   /* search535_0_0_0 -> 32 in the cache */
   /* search535_0_0_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imax,
      1, 1,
      { 32, 61 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_int, { 0xff /* 255 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imin,
      0, 2,
      { 1589, 1590 },
      -1,
   } },
   /* search535_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 2,
      { 1591, 61 },
      -1,
   } },

   /* replace535_0_0 -> 32 in the cache */
   /* replace535_0_1 -> 61 in the cache */
   /* replace535_0 -> 1589 in the cache */
   /* replace535_1 -> 1590 in the cache */
   /* replace535 -> 1591 in the cache */

   /* ('ior', ('flt(is_used_once)', 'a', 'b'), ('flt', 'a', 'c')) => ('flt', 'a', ('!fmax', 'b', 'c')) */
   /* search536_0_0 -> 126 in the cache */
   /* search536_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 126, 127 },
      0,
   } },
   /* search536_1_0 -> 126 in the cache */
   /* search536_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 126, 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1593, 1594 },
      -1,
   } },

   /* replace536_0 -> 126 in the cache */
   /* replace536_1_0 -> 127 in the cache */
   /* replace536_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      true,
      false,
      nir_op_fmax,
      0, 1,
      { 127, 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 126, 1596 },
      -1,
   } },

   /* ('ior', ('flt(is_used_once)', 'a', 'c'), ('flt', 'b', 'c')) => ('flt', ('!fmin', 'a', 'b'), 'c') */
   /* search537_0_0 -> 0 in the cache */
   /* search537_0_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 0, 684 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      2, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search537_1_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1599, 684 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1598, 1600 },
      -1,
   } },

   /* replace537_0_0 -> 0 in the cache */
   /* replace537_0_1 -> 1599 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      true,
      false,
      nir_op_fmin,
      0, 1,
      { 0, 1599 },
      -1,
   } },
   /* replace537_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 1602, 684 },
      -1,
   } },

   /* ('ior', ('fge(is_used_once)', 'a', 'b'), ('fge', 'a', 'c')) => ('fge', 'a', ('!fmin', 'b', 'c')) */
   /* search538_0_0 -> 126 in the cache */
   /* search538_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 126, 127 },
      0,
   } },
   /* search538_1_0 -> 126 in the cache */
   /* search538_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 126, 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1604, 1605 },
      -1,
   } },

   /* replace538_0 -> 126 in the cache */
   /* replace538_1_0 -> 127 in the cache */
   /* replace538_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      true,
      false,
      nir_op_fmin,
      0, 1,
      { 127, 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 126, 1607 },
      -1,
   } },

   /* ('ior', ('fge(is_used_once)', 'a', 'c'), ('fge', 'b', 'c')) => ('fge', ('!fmax', 'a', 'b'), 'c') */
   /* search539_0_0 -> 0 in the cache */
   /* search539_0_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 0, 684 },
      0,
   } },
   /* search539_1_0 -> 1599 in the cache */
   /* search539_1_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1599, 684 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1609, 1610 },
      -1,
   } },

   /* replace539_0_0 -> 0 in the cache */
   /* replace539_0_1 -> 1599 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      true,
      false,
      nir_op_fmax,
      0, 1,
      { 0, 1599 },
      -1,
   } },
   /* replace539_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 1612, 684 },
      -1,
   } },

   /* ('ior', ('flt', 'a', '#b'), ('flt', 'a', '#c')) => ('flt', 'a', ('!fmax', 'b', 'c')) */
   /* search540_0_0 -> 126 in the cache */
   /* search540_0_1 -> 283 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 126, 283 },
      -1,
   } },
   /* search540_1_0 -> 126 in the cache */
   /* search540_1_1 -> 814 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 126, 814 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1614, 1615 },
      -1,
   } },

   /* replace540_0 -> 126 in the cache */
   /* replace540_1_0 -> 127 in the cache */
   /* replace540_1_1 -> 129 in the cache */
   /* replace540_1 -> 1596 in the cache */
   /* replace540 -> 1597 in the cache */

   /* ('ior', ('flt', '#a', 'c'), ('flt', '#b', 'c')) => ('flt', ('!fmin', 'a', 'b'), 'c') */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search541_0_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1617, 684 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      2, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search541_1_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1619, 684 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1618, 1620 },
      -1,
   } },

   /* replace541_0_0 -> 0 in the cache */
   /* replace541_0_1 -> 1599 in the cache */
   /* replace541_0 -> 1602 in the cache */
   /* replace541_1 -> 684 in the cache */
   /* replace541 -> 1603 in the cache */

   /* ('ior', ('fge', 'a', '#b'), ('fge', 'a', '#c')) => ('fge', 'a', ('!fmin', 'b', 'c')) */
   /* search542_0_0 -> 126 in the cache */
   /* search542_0_1 -> 283 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 126, 283 },
      -1,
   } },
   /* search542_1_0 -> 126 in the cache */
   /* search542_1_1 -> 814 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 126, 814 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1622, 1623 },
      -1,
   } },

   /* replace542_0 -> 126 in the cache */
   /* replace542_1_0 -> 127 in the cache */
   /* replace542_1_1 -> 129 in the cache */
   /* replace542_1 -> 1607 in the cache */
   /* replace542 -> 1608 in the cache */

   /* ('ior', ('fge', '#a', 'c'), ('fge', '#b', 'c')) => ('fge', ('!fmax', 'a', 'b'), 'c') */
   /* search543_0_0 -> 1617 in the cache */
   /* search543_0_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1617, 684 },
      -1,
   } },
   /* search543_1_0 -> 1619 in the cache */
   /* search543_1_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1619, 684 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1625, 1626 },
      -1,
   } },

   /* replace543_0_0 -> 0 in the cache */
   /* replace543_0_1 -> 1599 in the cache */
   /* replace543_0 -> 1612 in the cache */
   /* replace543_1 -> 684 in the cache */
   /* replace543 -> 1613 in the cache */

   /* ('~iand', ('flt(is_used_once)', 'a', 'b'), ('flt', 'a', 'c')) => ('flt', 'a', ('fmin', 'b', 'c')) */
   /* search544_0_0 -> 126 in the cache */
   /* search544_0_1 -> 127 in the cache */
   /* search544_0 -> 1593 in the cache */
   /* search544_1_0 -> 126 in the cache */
   /* search544_1_1 -> 129 in the cache */
   /* search544_1 -> 1594 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1593, 1594 },
      -1,
   } },

   /* replace544_0 -> 126 in the cache */
   /* replace544_1_0 -> 127 in the cache */
   /* replace544_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 127, 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 126, 1629 },
      -1,
   } },

   /* ('~iand', ('flt(is_used_once)', 'a', 'c'), ('flt', 'b', 'c')) => ('flt', ('fmax', 'a', 'b'), 'c') */
   /* search545_0_0 -> 0 in the cache */
   /* search545_0_1 -> 684 in the cache */
   /* search545_0 -> 1598 in the cache */
   /* search545_1_0 -> 1599 in the cache */
   /* search545_1_1 -> 684 in the cache */
   /* search545_1 -> 1600 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1598, 1600 },
      -1,
   } },

   /* replace545_0_0 -> 0 in the cache */
   /* replace545_0_1 -> 1599 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 0, 1599 },
      -1,
   } },
   /* replace545_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 1632, 684 },
      -1,
   } },

   /* ('~iand', ('fge(is_used_once)', 'a', 'b'), ('fge', 'a', 'c')) => ('fge', 'a', ('fmax', 'b', 'c')) */
   /* search546_0_0 -> 126 in the cache */
   /* search546_0_1 -> 127 in the cache */
   /* search546_0 -> 1604 in the cache */
   /* search546_1_0 -> 126 in the cache */
   /* search546_1_1 -> 129 in the cache */
   /* search546_1 -> 1605 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1604, 1605 },
      -1,
   } },

   /* replace546_0 -> 126 in the cache */
   /* replace546_1_0 -> 127 in the cache */
   /* replace546_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 127, 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 126, 1635 },
      -1,
   } },

   /* ('~iand', ('fge(is_used_once)', 'a', 'c'), ('fge', 'b', 'c')) => ('fge', ('fmin', 'a', 'b'), 'c') */
   /* search547_0_0 -> 0 in the cache */
   /* search547_0_1 -> 684 in the cache */
   /* search547_0 -> 1609 in the cache */
   /* search547_1_0 -> 1599 in the cache */
   /* search547_1_1 -> 684 in the cache */
   /* search547_1 -> 1610 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1609, 1610 },
      -1,
   } },

   /* replace547_0_0 -> 0 in the cache */
   /* replace547_0_1 -> 1599 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 0, 1599 },
      -1,
   } },
   /* replace547_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 1638, 684 },
      -1,
   } },

   /* ('iand', ('flt', 'a', '#b(is_a_number)'), ('flt', 'a', '#c(is_a_number)')) => ('flt', 'a', ('fmin', 'b', 'c')) */
   /* search548_0_0 -> 126 in the cache */
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* b */
      true,
      nir_type_invalid,
      16,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 126, 1640 },
      -1,
   } },
   /* search548_1_0 -> 126 in the cache */
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* c */
      true,
      nir_type_invalid,
      16,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 126, 1642 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1641, 1643 },
      -1,
   } },

   /* replace548_0 -> 126 in the cache */
   /* replace548_1_0 -> 127 in the cache */
   /* replace548_1_1 -> 129 in the cache */
   /* replace548_1 -> 1629 in the cache */
   /* replace548 -> 1630 in the cache */

   /* ('iand', ('flt', '#a(is_a_number)', 'c'), ('flt', '#b(is_a_number)', 'c')) => ('flt', ('fmax', 'a', 'b'), 'c') */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      true,
      nir_type_invalid,
      16,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search549_0_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1645, 684 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      2, /* b */
      true,
      nir_type_invalid,
      16,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search549_1_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1647, 684 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1646, 1648 },
      -1,
   } },

   /* replace549_0_0 -> 0 in the cache */
   /* replace549_0_1 -> 1599 in the cache */
   /* replace549_0 -> 1632 in the cache */
   /* replace549_1 -> 684 in the cache */
   /* replace549 -> 1633 in the cache */

   /* ('iand', ('fge', 'a', '#b(is_a_number)'), ('fge', 'a', '#c(is_a_number)')) => ('fge', 'a', ('fmax', 'b', 'c')) */
   /* search550_0_0 -> 126 in the cache */
   /* search550_0_1 -> 1640 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 126, 1640 },
      -1,
   } },
   /* search550_1_0 -> 126 in the cache */
   /* search550_1_1 -> 1642 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 126, 1642 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1650, 1651 },
      -1,
   } },

   /* replace550_0 -> 126 in the cache */
   /* replace550_1_0 -> 127 in the cache */
   /* replace550_1_1 -> 129 in the cache */
   /* replace550_1 -> 1635 in the cache */
   /* replace550 -> 1636 in the cache */

   /* ('iand', ('fge', '#a(is_a_number)', 'c'), ('fge', '#b(is_a_number)', 'c')) => ('fge', ('fmin', 'a', 'b'), 'c') */
   /* search551_0_0 -> 1645 in the cache */
   /* search551_0_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1645, 684 },
      -1,
   } },
   /* search551_1_0 -> 1647 in the cache */
   /* search551_1_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1647, 684 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1653, 1654 },
      -1,
   } },

   /* replace551_0_0 -> 0 in the cache */
   /* replace551_0_1 -> 1599 in the cache */
   /* replace551_0 -> 1638 in the cache */
   /* replace551_1 -> 684 in the cache */
   /* replace551 -> 1639 in the cache */

   /* ('ior', ('ilt(is_used_once)', 'a', 'b'), ('ilt', 'a', 'c')) => ('ilt', 'a', ('imax', 'b', 'c')) */
   /* search552_0_0 -> 126 in the cache */
   /* search552_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 126, 127 },
      0,
   } },
   /* search552_1_0 -> 126 in the cache */
   /* search552_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 126, 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1656, 1657 },
      -1,
   } },

   /* replace552_0 -> 126 in the cache */
   /* replace552_1_0 -> 127 in the cache */
   /* replace552_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 127, 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 126, 1659 },
      -1,
   } },

   /* ('ior', ('ilt(is_used_once)', 'a', 'c'), ('ilt', 'b', 'c')) => ('ilt', ('imin', 'a', 'b'), 'c') */
   /* search553_0_0 -> 0 in the cache */
   /* search553_0_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 0, 684 },
      0,
   } },
   /* search553_1_0 -> 1599 in the cache */
   /* search553_1_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 1599, 684 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1661, 1662 },
      -1,
   } },

   /* replace553_0_0 -> 0 in the cache */
   /* replace553_0_1 -> 1599 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 0, 1599 },
      -1,
   } },
   /* replace553_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 1664, 684 },
      -1,
   } },

   /* ('ior', ('ige(is_used_once)', 'a', 'b'), ('ige', 'a', 'c')) => ('ige', 'a', ('imin', 'b', 'c')) */
   /* search554_0_0 -> 126 in the cache */
   /* search554_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 126, 127 },
      0,
   } },
   /* search554_1_0 -> 126 in the cache */
   /* search554_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 126, 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1666, 1667 },
      -1,
   } },

   /* replace554_0 -> 126 in the cache */
   /* replace554_1_0 -> 127 in the cache */
   /* replace554_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 127, 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 126, 1669 },
      -1,
   } },

   /* ('ior', ('ige(is_used_once)', 'a', 'c'), ('ige', 'b', 'c')) => ('ige', ('imax', 'a', 'b'), 'c') */
   /* search555_0_0 -> 0 in the cache */
   /* search555_0_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 0, 684 },
      0,
   } },
   /* search555_1_0 -> 1599 in the cache */
   /* search555_1_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 1599, 684 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1671, 1672 },
      -1,
   } },

   /* replace555_0_0 -> 0 in the cache */
   /* replace555_0_1 -> 1599 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 0, 1599 },
      -1,
   } },
   /* replace555_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 1674, 684 },
      -1,
   } },

   /* ('ior', ('ult(is_used_once)', 'a', 'b'), ('ult', 'a', 'c')) => ('ult', 'a', ('umax', 'b', 'c')) */
   /* search556_0_0 -> 126 in the cache */
   /* search556_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 126, 127 },
      0,
   } },
   /* search556_1_0 -> 126 in the cache */
   /* search556_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 126, 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1676, 1677 },
      -1,
   } },

   /* replace556_0 -> 126 in the cache */
   /* replace556_1_0 -> 127 in the cache */
   /* replace556_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 127, 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 126, 1679 },
      -1,
   } },

   /* ('ior', ('ult(is_used_once)', 'a', 'c'), ('ult', 'b', 'c')) => ('ult', ('umin', 'a', 'b'), 'c') */
   /* search557_0_0 -> 0 in the cache */
   /* search557_0_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 0, 684 },
      0,
   } },
   /* search557_1_0 -> 1599 in the cache */
   /* search557_1_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 1599, 684 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1681, 1682 },
      -1,
   } },

   /* replace557_0_0 -> 0 in the cache */
   /* replace557_0_1 -> 1599 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 0, 1599 },
      -1,
   } },
   /* replace557_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 1684, 684 },
      -1,
   } },

   /* ('ior', ('uge(is_used_once)', 'a', 'b'), ('uge', 'a', 'c')) => ('uge', 'a', ('umin', 'b', 'c')) */
   /* search558_0_0 -> 126 in the cache */
   /* search558_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 126, 127 },
      0,
   } },
   /* search558_1_0 -> 126 in the cache */
   /* search558_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 126, 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1686, 1687 },
      -1,
   } },

   /* replace558_0 -> 126 in the cache */
   /* replace558_1_0 -> 127 in the cache */
   /* replace558_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 127, 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 126, 1689 },
      -1,
   } },

   /* ('ior', ('uge(is_used_once)', 'a', 'c'), ('uge', 'b', 'c')) => ('uge', ('umax', 'a', 'b'), 'c') */
   /* search559_0_0 -> 0 in the cache */
   /* search559_0_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 0, 684 },
      0,
   } },
   /* search559_1_0 -> 1599 in the cache */
   /* search559_1_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 1599, 684 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1691, 1692 },
      -1,
   } },

   /* replace559_0_0 -> 0 in the cache */
   /* replace559_0_1 -> 1599 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 0, 1599 },
      -1,
   } },
   /* replace559_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 1694, 684 },
      -1,
   } },

   /* ('iand', ('ilt(is_used_once)', 'a', 'b'), ('ilt', 'a', 'c')) => ('ilt', 'a', ('imin', 'b', 'c')) */
   /* search560_0_0 -> 126 in the cache */
   /* search560_0_1 -> 127 in the cache */
   /* search560_0 -> 1656 in the cache */
   /* search560_1_0 -> 126 in the cache */
   /* search560_1_1 -> 129 in the cache */
   /* search560_1 -> 1657 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1656, 1657 },
      -1,
   } },

   /* replace560_0 -> 126 in the cache */
   /* replace560_1_0 -> 127 in the cache */
   /* replace560_1_1 -> 129 in the cache */
   /* replace560_1 -> 1669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 126, 1669 },
      -1,
   } },

   /* ('iand', ('ilt(is_used_once)', 'a', 'c'), ('ilt', 'b', 'c')) => ('ilt', ('imax', 'a', 'b'), 'c') */
   /* search561_0_0 -> 0 in the cache */
   /* search561_0_1 -> 684 in the cache */
   /* search561_0 -> 1661 in the cache */
   /* search561_1_0 -> 1599 in the cache */
   /* search561_1_1 -> 684 in the cache */
   /* search561_1 -> 1662 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1661, 1662 },
      -1,
   } },

   /* replace561_0_0 -> 0 in the cache */
   /* replace561_0_1 -> 1599 in the cache */
   /* replace561_0 -> 1674 in the cache */
   /* replace561_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 1674, 684 },
      -1,
   } },

   /* ('iand', ('ige(is_used_once)', 'a', 'b'), ('ige', 'a', 'c')) => ('ige', 'a', ('imax', 'b', 'c')) */
   /* search562_0_0 -> 126 in the cache */
   /* search562_0_1 -> 127 in the cache */
   /* search562_0 -> 1666 in the cache */
   /* search562_1_0 -> 126 in the cache */
   /* search562_1_1 -> 129 in the cache */
   /* search562_1 -> 1667 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1666, 1667 },
      -1,
   } },

   /* replace562_0 -> 126 in the cache */
   /* replace562_1_0 -> 127 in the cache */
   /* replace562_1_1 -> 129 in the cache */
   /* replace562_1 -> 1659 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 126, 1659 },
      -1,
   } },

   /* ('iand', ('ige(is_used_once)', 'a', 'c'), ('ige', 'b', 'c')) => ('ige', ('imin', 'a', 'b'), 'c') */
   /* search563_0_0 -> 0 in the cache */
   /* search563_0_1 -> 684 in the cache */
   /* search563_0 -> 1671 in the cache */
   /* search563_1_0 -> 1599 in the cache */
   /* search563_1_1 -> 684 in the cache */
   /* search563_1 -> 1672 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1671, 1672 },
      -1,
   } },

   /* replace563_0_0 -> 0 in the cache */
   /* replace563_0_1 -> 1599 in the cache */
   /* replace563_0 -> 1664 in the cache */
   /* replace563_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 1664, 684 },
      -1,
   } },

   /* ('iand', ('ult(is_used_once)', 'a', 'b'), ('ult', 'a', 'c')) => ('ult', 'a', ('umin', 'b', 'c')) */
   /* search564_0_0 -> 126 in the cache */
   /* search564_0_1 -> 127 in the cache */
   /* search564_0 -> 1676 in the cache */
   /* search564_1_0 -> 126 in the cache */
   /* search564_1_1 -> 129 in the cache */
   /* search564_1 -> 1677 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1676, 1677 },
      -1,
   } },

   /* replace564_0 -> 126 in the cache */
   /* replace564_1_0 -> 127 in the cache */
   /* replace564_1_1 -> 129 in the cache */
   /* replace564_1 -> 1689 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 126, 1689 },
      -1,
   } },

   /* ('iand', ('ult(is_used_once)', 'a', 'c'), ('ult', 'b', 'c')) => ('ult', ('umax', 'a', 'b'), 'c') */
   /* search565_0_0 -> 0 in the cache */
   /* search565_0_1 -> 684 in the cache */
   /* search565_0 -> 1681 in the cache */
   /* search565_1_0 -> 1599 in the cache */
   /* search565_1_1 -> 684 in the cache */
   /* search565_1 -> 1682 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1681, 1682 },
      -1,
   } },

   /* replace565_0_0 -> 0 in the cache */
   /* replace565_0_1 -> 1599 in the cache */
   /* replace565_0 -> 1694 in the cache */
   /* replace565_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 1694, 684 },
      -1,
   } },

   /* ('iand', ('uge(is_used_once)', 'a', 'b'), ('uge', 'a', 'c')) => ('uge', 'a', ('umax', 'b', 'c')) */
   /* search566_0_0 -> 126 in the cache */
   /* search566_0_1 -> 127 in the cache */
   /* search566_0 -> 1686 in the cache */
   /* search566_1_0 -> 126 in the cache */
   /* search566_1_1 -> 129 in the cache */
   /* search566_1 -> 1687 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1686, 1687 },
      -1,
   } },

   /* replace566_0 -> 126 in the cache */
   /* replace566_1_0 -> 127 in the cache */
   /* replace566_1_1 -> 129 in the cache */
   /* replace566_1 -> 1679 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 126, 1679 },
      -1,
   } },

   /* ('iand', ('uge(is_used_once)', 'a', 'c'), ('uge', 'b', 'c')) => ('uge', ('umin', 'a', 'b'), 'c') */
   /* search567_0_0 -> 0 in the cache */
   /* search567_0_1 -> 684 in the cache */
   /* search567_0 -> 1691 in the cache */
   /* search567_1_0 -> 1599 in the cache */
   /* search567_1_1 -> 684 in the cache */
   /* search567_1 -> 1692 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1691, 1692 },
      -1,
   } },

   /* replace567_0_0 -> 0 in the cache */
   /* replace567_0_1 -> 1599 in the cache */
   /* replace567_0 -> 1684 in the cache */
   /* replace567_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 1684, 684 },
      -1,
   } },

   /* ('ior', ('ior(is_used_once)', ('flt(is_used_once)', 'a', 'c'), 'd'), ('flt', 'b', 'c')) => ('ior', ('flt', ('!fmin', 'a', 'b'), 'c'), 'd') */
   /* search568_0_0_0 -> 0 in the cache */
   /* search568_0_0_1 -> 684 in the cache */
   /* search568_0_0 -> 1598 in the cache */
   { .variable = {
      { nir_search_value_variable, 1 },
      2, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 1598, 1712 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      3, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search568_1_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1714, 684 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 1713, 1715 },
      -1,
   } },

   /* replace568_0_0_0 -> 0 in the cache */
   /* replace568_0_0_1 -> 1714 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      true,
      false,
      nir_op_fmin,
      1, 1,
      { 0, 1714 },
      -1,
   } },
   /* replace568_0_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 1717, 684 },
      -1,
   } },
   /* replace568_1 -> 1712 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 1718, 1712 },
      -1,
   } },

   /* ('ior', ('ior(is_used_once)', ('flt', 'a', 'c'), 'd'), ('flt(is_used_once)', 'b', 'c')) => ('ior', ('flt', ('!fmin', 'a', 'b'), 'c'), 'd') */
   /* search569_0_0_0 -> 0 in the cache */
   /* search569_0_0_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 0, 684 },
      -1,
   } },
   /* search569_0_1 -> 1712 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 1720, 1712 },
      0,
   } },
   /* search569_1_0 -> 1714 in the cache */
   /* search569_1_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1714, 684 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 1721, 1722 },
      -1,
   } },

   /* replace569_0_0_0 -> 0 in the cache */
   /* replace569_0_0_1 -> 1714 in the cache */
   /* replace569_0_0 -> 1717 in the cache */
   /* replace569_0_1 -> 684 in the cache */
   /* replace569_0 -> 1718 in the cache */
   /* replace569_1 -> 1712 in the cache */
   /* replace569 -> 1719 in the cache */

   /* ('ior', ('ior(is_used_once)', ('flt(is_used_once)', 'a', 'b'), 'd'), ('flt', 'a', 'c')) => ('ior', ('flt', 'a', ('!fmax', 'b', 'c')), 'd') */
   /* search570_0_0_0 -> 149 in the cache */
   /* search570_0_0_1 -> 150 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 149, 150 },
      0,
   } },
   /* search570_0_1 -> 1712 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 1724, 1712 },
      0,
   } },
   /* search570_1_0 -> 149 in the cache */
   { .variable = {
      { nir_search_value_variable, -4 },
      3, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 149, 1726 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 1725, 1727 },
      -1,
   } },

   /* replace570_0_0 -> 149 in the cache */
   /* replace570_0_1_0 -> 150 in the cache */
   /* replace570_0_1_1 -> 1726 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      true,
      false,
      nir_op_fmax,
      1, 1,
      { 150, 1726 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 149, 1729 },
      -1,
   } },
   /* replace570_1 -> 1712 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 1730, 1712 },
      -1,
   } },

   /* ('ior', ('ior(is_used_once)', ('flt', 'a', 'b'), 'd'), ('flt(is_used_once)', 'a', 'c')) => ('ior', ('flt', 'a', ('!fmax', 'b', 'c')), 'd') */
   /* search571_0_0_0 -> 149 in the cache */
   /* search571_0_0_1 -> 150 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 149, 150 },
      -1,
   } },
   /* search571_0_1 -> 1712 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 1732, 1712 },
      0,
   } },
   /* search571_1_0 -> 149 in the cache */
   /* search571_1_1 -> 1726 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 149, 1726 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 1733, 1734 },
      -1,
   } },

   /* replace571_0_0 -> 149 in the cache */
   /* replace571_0_1_0 -> 150 in the cache */
   /* replace571_0_1_1 -> 1726 in the cache */
   /* replace571_0_1 -> 1729 in the cache */
   /* replace571_0 -> 1730 in the cache */
   /* replace571_1 -> 1712 in the cache */
   /* replace571 -> 1731 in the cache */

   /* ('ior', ('flt', 'a(is_a_number)', 'b(is_a_number)'), ('flt', 'b', 'a')) => ('fneu', 'a', 'b') */
   /* search572_0_0 -> 1042 in the cache */
   /* search572_0_1 -> 1043 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1042, 1043 },
      -1,
   } },
   /* search572_1_0 -> 3 in the cache */
   /* search572_1_1 -> 0 in the cache */
   /* search572_1 -> 1078 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1736, 1078 },
      -1,
   } },

   /* replace572_0 -> 0 in the cache */
   /* replace572_1 -> 3 in the cache */
   /* replace572 -> 1052 in the cache */

   /* ('umin', ('iand', 'a', '#b(is_pos_power_of_two)'), ('iand', 'c', 'b')) => ('iand', ('iand', 'a', 'b'), ('iand', 'c', 'b')) */
   /* search573_0_0 -> 0 in the cache */
   /* search573_0_1 -> 1 in the cache */
   /* search573_0 -> 199 in the cache */
   /* search573_1_0 -> 140 in the cache */
   /* search573_1_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 140, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umin,
      0, 3,
      { 199, 1738 },
      -1,
   } },

   /* replace573_0_0 -> 0 in the cache */
   /* replace573_0_1 -> 3 in the cache */
   /* replace573_0 -> 201 in the cache */
   /* replace573_1_0 -> 140 in the cache */
   /* replace573_1_1 -> 3 in the cache */
   /* replace573_1 -> 1738 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 201, 1738 },
      -1,
   } },

   /* ('ior', ('flt(is_used_once)', 0.0, 'a@16'), ('flt', 'b@16', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   /* search574_0_0 -> 113 in the cache */
   /* search574_0_1 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 113, 11 },
      0,
   } },
   /* search574_1_0 -> 258 in the cache */
   /* search574_1_1 -> 113 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 258, 113 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1741, 1742 },
      -1,
   } },

   /* replace574_0 -> 113 in the cache */
   /* replace574_1_0 -> 11 in the cache */
   /* replace574_1_1_0 -> 258 in the cache */
   /* replace574_1_1 -> 534 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 11, 534 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 113, 1744 },
      -1,
   } },

   /* ('ior', ('flt', 0.0, 'a@16'), ('flt(is_used_once)', 'b@16', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   /* search575_0_0 -> 113 in the cache */
   /* search575_0_1 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 113, 11 },
      -1,
   } },
   /* search575_1_0 -> 258 in the cache */
   /* search575_1_1 -> 113 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 258, 113 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1746, 1747 },
      -1,
   } },

   /* replace575_0 -> 113 in the cache */
   /* replace575_1_0 -> 11 in the cache */
   /* replace575_1_1_0 -> 258 in the cache */
   /* replace575_1_1 -> 534 in the cache */
   /* replace575_1 -> 1744 in the cache */
   /* replace575 -> 1745 in the cache */

   /* ('ior', ('fge(is_used_once)', 0.0, 'a@16'), ('fge', 'b@16', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search576_0_0 -> 113 in the cache */
   /* search576_0_1 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 113, 11 },
      0,
   } },
   /* search576_1_0 -> 258 in the cache */
   /* search576_1_1 -> 113 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 258, 113 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1749, 1750 },
      -1,
   } },

   /* replace576_0 -> 113 in the cache */
   /* replace576_1_0 -> 11 in the cache */
   /* replace576_1_1_0 -> 258 in the cache */
   /* replace576_1_1 -> 534 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 11, 534 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 113, 1752 },
      -1,
   } },

   /* ('ior', ('fge', 0.0, 'a@16'), ('fge(is_used_once)', 'b@16', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search577_0_0 -> 113 in the cache */
   /* search577_0_1 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 113, 11 },
      -1,
   } },
   /* search577_1_0 -> 258 in the cache */
   /* search577_1_1 -> 113 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 258, 113 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1754, 1755 },
      -1,
   } },

   /* replace577_0 -> 113 in the cache */
   /* replace577_1_0 -> 11 in the cache */
   /* replace577_1_1_0 -> 258 in the cache */
   /* replace577_1_1 -> 534 in the cache */
   /* replace577_1 -> 1752 in the cache */
   /* replace577 -> 1753 in the cache */

   /* ('~iand', ('flt(is_used_once)', 0.0, 'a@16'), ('flt', 'b@16', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search578_0_0 -> 113 in the cache */
   /* search578_0_1 -> 11 in the cache */
   /* search578_0 -> 1741 in the cache */
   /* search578_1_0 -> 258 in the cache */
   /* search578_1_1 -> 113 in the cache */
   /* search578_1 -> 1742 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1741, 1742 },
      -1,
   } },

   /* replace578_0 -> 113 in the cache */
   /* replace578_1_0 -> 11 in the cache */
   /* replace578_1_1_0 -> 258 in the cache */
   /* replace578_1_1 -> 534 in the cache */
   /* replace578_1 -> 1752 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 113, 1752 },
      -1,
   } },

   /* ('~iand', ('flt', 0.0, 'a@16'), ('flt(is_used_once)', 'b@16', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search579_0_0 -> 113 in the cache */
   /* search579_0_1 -> 11 in the cache */
   /* search579_0 -> 1746 in the cache */
   /* search579_1_0 -> 258 in the cache */
   /* search579_1_1 -> 113 in the cache */
   /* search579_1 -> 1747 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1746, 1747 },
      -1,
   } },

   /* replace579_0 -> 113 in the cache */
   /* replace579_1_0 -> 11 in the cache */
   /* replace579_1_1_0 -> 258 in the cache */
   /* replace579_1_1 -> 534 in the cache */
   /* replace579_1 -> 1752 in the cache */
   /* replace579 -> 1758 in the cache */

   /* ('~iand', ('fge(is_used_once)', 0.0, 'a@16'), ('fge', 'b@16', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   /* search580_0_0 -> 113 in the cache */
   /* search580_0_1 -> 11 in the cache */
   /* search580_0 -> 1749 in the cache */
   /* search580_1_0 -> 258 in the cache */
   /* search580_1_1 -> 113 in the cache */
   /* search580_1 -> 1750 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1749, 1750 },
      -1,
   } },

   /* replace580_0 -> 113 in the cache */
   /* replace580_1_0 -> 11 in the cache */
   /* replace580_1_1_0 -> 258 in the cache */
   /* replace580_1_1 -> 534 in the cache */
   /* replace580_1 -> 1744 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 113, 1744 },
      -1,
   } },

   /* ('~iand', ('fge', 0.0, 'a@16'), ('fge(is_used_once)', 'b@16', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   /* search581_0_0 -> 113 in the cache */
   /* search581_0_1 -> 11 in the cache */
   /* search581_0 -> 1754 in the cache */
   /* search581_1_0 -> 258 in the cache */
   /* search581_1_1 -> 113 in the cache */
   /* search581_1 -> 1755 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1754, 1755 },
      -1,
   } },

   /* replace581_0 -> 113 in the cache */
   /* replace581_1_0 -> 11 in the cache */
   /* replace581_1_1_0 -> 258 in the cache */
   /* replace581_1_1 -> 534 in the cache */
   /* replace581_1 -> 1744 in the cache */
   /* replace581 -> 1761 in the cache */

   /* ('bcsel@16', ('feq', 'a', 0.0), 1.0, ('i2f16', ('iadd', ('b2i16', ('flt', 0.0, 'a@16')), ('ineg', ('b2i16', ('flt', 'a@16', 0.0)))))) => ('i2f16', ('iadd', ('b2i32', ('!fge', 'a', 0.0)), ('ineg', ('b2i32', ('!flt', 'a', 0.0))))) */
   /* search582_0_0 -> 11 in the cache */
   /* search582_0_1 -> 113 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 11, 113 },
      -1,
   } },
   /* search582_1 -> 481 in the cache */
   /* search582_2_0_0_0_0 -> 113 in the cache */
   /* search582_2_0_0_0_1 -> 11 in the cache */
   /* search582_2_0_0_0 -> 1746 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2i16,
      -1, 0,
      { 1746 },
      -1,
   } },
   /* search582_2_0_1_0_0_0 -> 11 in the cache */
   /* search582_2_0_1_0_0_1 -> 113 in the cache */
   /* search582_2_0_1_0_0 -> 689 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2i16,
      -1, 0,
      { 689 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 1765 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 1764, 1766 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2f16,
      -1, 1,
      { 1767 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 1763, 481, 1768 },
      -1,
   } },

   /* replace582_0_0_0_0 -> 11 in the cache */
   /* replace582_0_0_0_1 -> 113 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_fge,
      -1, 0,
      { 11, 113 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 1770 },
      -1,
   } },
   /* replace582_0_1_0_0_0 -> 11 in the cache */
   /* replace582_0_1_0_0_1 -> 113 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_flt,
      -1, 0,
      { 11, 113 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 1772 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 1773 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 1771, 1774 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2f16,
      -1, 1,
      { 1775 },
      -1,
   } },

   /* ('bcsel', 'a', ('b2f(is_used_once)', 'b@16'), ('b2f', 'c@16')) => ('b2f', ('bcsel', 'a', 'b', 'c')) */
   /* search583_0 -> 760 in the cache */
   /* search583_1_0 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 258 },
      0,
   } },
   /* search583_2_0 -> 469 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 469 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 1777, 1778 },
      -1,
   } },

   /* replace583_0_0 -> 760 in the cache */
   /* replace583_0_1 -> 258 in the cache */
   /* replace583_0_2 -> 469 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 258, 469 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 1780 },
      -1,
   } },

   /* ('~i2f16', ('f2i', 'a@16')) => ('ftrunc', 'a') */
   /* search584_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2i,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_i2f16,
      -1, 0,
      { 1782 },
      -1,
   } },

   /* replace584_0 -> 11 in the cache */
   /* replace584 -> 688 in the cache */

   /* ('~f2i16', ('i2f', 'a@16')) => a */
   /* search585_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_i2f,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_f2i16,
      -1, 0,
      { 1784 },
      -1,
   } },

   /* replace585 -> 11 in the cache */

   /* ('~f2i16', ('u2f', 'a@16')) => a */
   /* search586_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_u2f,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_f2i16,
      -1, 0,
      { 1786 },
      -1,
   } },

   /* replace586 -> 11 in the cache */

   /* ('~f2u16', ('i2f', 'a@16')) => a */
   /* search587_0_0 -> 11 in the cache */
   /* search587_0 -> 1784 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_f2u16,
      -1, 0,
      { 1784 },
      -1,
   } },

   /* replace587 -> 11 in the cache */

   /* ('~f2u16', ('u2f', 'a@16')) => a */
   /* search588_0_0 -> 11 in the cache */
   /* search588_0 -> 1786 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_f2u16,
      -1, 0,
      { 1786 },
      -1,
   } },

   /* replace588 -> 11 in the cache */

   /* ('fadd', ('b2f16', ('flt', 0.0, 'a@16')), ('fneg', ('b2f16', ('flt', 'a@16', 0.0)))) => ('fsign', 'a') */
   /* search589_0_0_0 -> 113 in the cache */
   /* search589_0_0_1 -> 11 in the cache */
   /* search589_0_0 -> 1746 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2f16,
      -1, 0,
      { 1746 },
      -1,
   } },
   /* search589_1_0_0_0 -> 11 in the cache */
   /* search589_1_0_0_1 -> 113 in the cache */
   /* search589_1_0_0 -> 689 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2f16,
      -1, 0,
      { 689 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 1791 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 1790, 1792 },
      -1,
   } },

   /* replace589_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('iadd', ('b2i16', ('flt', 0, 'a@16')), ('ineg', ('b2i16', ('flt', 'a@16', 0)))) => ('f2i16', ('fsign', 'a')) */
   /* search590_0_0_0 -> 856 in the cache */
   /* search590_0_0_1 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 856, 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2i16,
      -1, 0,
      { 1795 },
      -1,
   } },
   /* search590_1_0_0_0 -> 11 in the cache */
   /* search590_1_0_0_1 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 11, 856 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2i16,
      -1, 0,
      { 1797 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 1798 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 1796, 1799 },
      -1,
   } },

   /* replace590_0_0 -> 11 in the cache */
   /* replace590_0 -> 1794 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2i16,
      -1, 0,
      { 1794 },
      -1,
   } },

   /* ('f2f16', ('f2f32', 'a@16')) => a */
   /* search591_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 1802 },
      -1,
   } },

   /* replace591 -> 11 in the cache */

   /* ('i2i16', ('i2i32', 'a@16')) => a */
   /* search592_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 1804 },
      -1,
   } },

   /* replace592 -> 11 in the cache */

   /* ('u2u16', ('u2u32', 'a@16')) => a */
   /* search593_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 1806 },
      -1,
   } },

   /* replace593 -> 11 in the cache */

   /* ('f2f16', ('b2f32', 'a@1')) => ('b2f16', 'a') */
   /* search594_0_0 -> 760 in the cache */
   /* search594_0 -> 1408 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 1408 },
      -1,
   } },

   /* replace594_0 -> 760 in the cache */
   /* replace594 -> 1404 in the cache */

   /* ('i2i16', ('b2i32', 'a@1')) => ('b2i16', 'a') */
   /* search595_0_0 -> 760 in the cache */
   /* search595_0 -> 1396 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 1396 },
      -1,
   } },

   /* replace595_0 -> 760 in the cache */
   /* replace595 -> 1392 in the cache */

   /* ('u2u16', ('b2i32', 'a@1')) => ('b2i16', 'a') */
   /* search596_0_0 -> 760 in the cache */
   /* search596_0 -> 1396 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 1396 },
      -1,
   } },

   /* replace596_0 -> 760 in the cache */
   /* replace596 -> 1392 in the cache */

   /* ('f2u32', ('f2f32', 'a@16')) => ('f2u32', 'a') */
   /* search597_0_0 -> 11 in the cache */
   /* search597_0 -> 1802 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2u32,
      -1, 0,
      { 1802 },
      -1,
   } },

   /* replace597_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2u32,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('f2i32', ('f2f32', 'a@16')) => ('f2i32', 'a') */
   /* search598_0_0 -> 11 in the cache */
   /* search598_0 -> 1802 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2i32,
      -1, 0,
      { 1802 },
      -1,
   } },

   /* replace598_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2i32,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('f2f16', ('u2f32', 'a')) => ('u2f16', 'a') */
   /* search599_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2f32,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 1815 },
      -1,
   } },

   /* replace599_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2f16,
      -1, 0,
      { 32 },
      -1,
   } },

   /* ('f2f16', ('i2f32', 'a')) => ('i2f16', 'a') */
   /* search600_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2f32,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 1818 },
      -1,
   } },

   /* replace600_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2f16,
      -1, 0,
      { 32 },
      -1,
   } },

   /* ('u2f32', ('u2u32', 'a@16')) => ('u2f32', 'a') */
   /* search601_0_0 -> 11 in the cache */
   /* search601_0 -> 1806 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2f32,
      -1, 0,
      { 1806 },
      -1,
   } },

   /* replace601_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2f32,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('i2f32', ('i2i32', 'a@16')) => ('i2f32', 'a') */
   /* search602_0_0 -> 11 in the cache */
   /* search602_0 -> 1804 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2f32,
      -1, 0,
      { 1804 },
      -1,
   } },

   /* replace602_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2f32,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('f2f16', ('f2f64', 'a@16')) => a */
   /* search603_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2f64,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 1825 },
      -1,
   } },

   /* replace603 -> 11 in the cache */

   /* ('i2i16', ('i2i64', 'a@16')) => a */
   /* search604_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 1827 },
      -1,
   } },

   /* replace604 -> 11 in the cache */

   /* ('u2u16', ('u2u64', 'a@16')) => a */
   /* search605_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 1829 },
      -1,
   } },

   /* replace605 -> 11 in the cache */

   /* ('f2f16', ('b2f64', 'a@1')) => ('b2f16', 'a') */
   /* search606_0_0 -> 760 in the cache */
   /* search606_0 -> 1412 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 1412 },
      -1,
   } },

   /* replace606_0 -> 760 in the cache */
   /* replace606 -> 1404 in the cache */

   /* ('i2i16', ('b2i64', 'a@1')) => ('b2i16', 'a') */
   /* search607_0_0 -> 760 in the cache */
   /* search607_0 -> 1400 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 1400 },
      -1,
   } },

   /* replace607_0 -> 760 in the cache */
   /* replace607 -> 1392 in the cache */

   /* ('u2u16', ('b2i64', 'a@1')) => ('b2i16', 'a') */
   /* search608_0_0 -> 760 in the cache */
   /* search608_0 -> 1400 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 1400 },
      -1,
   } },

   /* replace608_0 -> 760 in the cache */
   /* replace608 -> 1392 in the cache */

   /* ('f2u64', ('f2f64', 'a@16')) => ('f2u64', 'a') */
   /* search609_0_0 -> 11 in the cache */
   /* search609_0 -> 1825 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2u64,
      -1, 0,
      { 1825 },
      -1,
   } },

   /* replace609_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2u64,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('f2i64', ('f2f64', 'a@16')) => ('f2i64', 'a') */
   /* search610_0_0 -> 11 in the cache */
   /* search610_0 -> 1825 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2i64,
      -1, 0,
      { 1825 },
      -1,
   } },

   /* replace610_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2i64,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('f2f16', ('u2f64', 'a')) => ('u2f16', 'a') */
   /* search611_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2f64,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 1838 },
      -1,
   } },

   /* replace611_0 -> 32 in the cache */
   /* replace611 -> 1817 in the cache */

   /* ('f2f16', ('i2f64', 'a')) => ('i2f16', 'a') */
   /* search612_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2f64,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 1840 },
      -1,
   } },

   /* replace612_0 -> 32 in the cache */
   /* replace612 -> 1820 in the cache */

   /* ('u2f64', ('u2u64', 'a@16')) => ('u2f64', 'a') */
   /* search613_0_0 -> 11 in the cache */
   /* search613_0 -> 1829 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2f64,
      -1, 0,
      { 1829 },
      -1,
   } },

   /* replace613_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2f64,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('i2f64', ('i2i64', 'a@16')) => ('i2f64', 'a') */
   /* search614_0_0 -> 11 in the cache */
   /* search614_0 -> 1827 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2f64,
      -1, 0,
      { 1827 },
      -1,
   } },

   /* replace614_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2f64,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('ior', ('flt(is_used_once)', 0.0, 'a@32'), ('flt', 'b@32', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   /* search615_0_0 -> 115 in the cache */
   /* search615_0_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 115, 16 },
      0,
   } },
   /* search615_1_0 -> 36 in the cache */
   /* search615_1_1 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 36, 115 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1846, 1847 },
      -1,
   } },

   /* replace615_0 -> 115 in the cache */
   /* replace615_1_0 -> 16 in the cache */
   /* replace615_1_1_0 -> 36 in the cache */
   /* replace615_1_1 -> 600 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 16, 600 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 115, 1849 },
      -1,
   } },

   /* ('ior', ('flt', 0.0, 'a@32'), ('flt(is_used_once)', 'b@32', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   /* search616_0_0 -> 115 in the cache */
   /* search616_0_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 115, 16 },
      -1,
   } },
   /* search616_1_0 -> 36 in the cache */
   /* search616_1_1 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 36, 115 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1851, 1852 },
      -1,
   } },

   /* replace616_0 -> 115 in the cache */
   /* replace616_1_0 -> 16 in the cache */
   /* replace616_1_1_0 -> 36 in the cache */
   /* replace616_1_1 -> 600 in the cache */
   /* replace616_1 -> 1849 in the cache */
   /* replace616 -> 1850 in the cache */

   /* ('ior', ('fge(is_used_once)', 0.0, 'a@32'), ('fge', 'b@32', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search617_0_0 -> 115 in the cache */
   /* search617_0_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 115, 16 },
      0,
   } },
   /* search617_1_0 -> 36 in the cache */
   /* search617_1_1 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 36, 115 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1854, 1855 },
      -1,
   } },

   /* replace617_0 -> 115 in the cache */
   /* replace617_1_0 -> 16 in the cache */
   /* replace617_1_1_0 -> 36 in the cache */
   /* replace617_1_1 -> 600 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 16, 600 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 115, 1857 },
      -1,
   } },

   /* ('ior', ('fge', 0.0, 'a@32'), ('fge(is_used_once)', 'b@32', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search618_0_0 -> 115 in the cache */
   /* search618_0_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 115, 16 },
      -1,
   } },
   /* search618_1_0 -> 36 in the cache */
   /* search618_1_1 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 36, 115 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1859, 1860 },
      -1,
   } },

   /* replace618_0 -> 115 in the cache */
   /* replace618_1_0 -> 16 in the cache */
   /* replace618_1_1_0 -> 36 in the cache */
   /* replace618_1_1 -> 600 in the cache */
   /* replace618_1 -> 1857 in the cache */
   /* replace618 -> 1858 in the cache */

   /* ('~iand', ('flt(is_used_once)', 0.0, 'a@32'), ('flt', 'b@32', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search619_0_0 -> 115 in the cache */
   /* search619_0_1 -> 16 in the cache */
   /* search619_0 -> 1846 in the cache */
   /* search619_1_0 -> 36 in the cache */
   /* search619_1_1 -> 115 in the cache */
   /* search619_1 -> 1847 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1846, 1847 },
      -1,
   } },

   /* replace619_0 -> 115 in the cache */
   /* replace619_1_0 -> 16 in the cache */
   /* replace619_1_1_0 -> 36 in the cache */
   /* replace619_1_1 -> 600 in the cache */
   /* replace619_1 -> 1857 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 115, 1857 },
      -1,
   } },

   /* ('~iand', ('flt', 0.0, 'a@32'), ('flt(is_used_once)', 'b@32', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search620_0_0 -> 115 in the cache */
   /* search620_0_1 -> 16 in the cache */
   /* search620_0 -> 1851 in the cache */
   /* search620_1_0 -> 36 in the cache */
   /* search620_1_1 -> 115 in the cache */
   /* search620_1 -> 1852 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1851, 1852 },
      -1,
   } },

   /* replace620_0 -> 115 in the cache */
   /* replace620_1_0 -> 16 in the cache */
   /* replace620_1_1_0 -> 36 in the cache */
   /* replace620_1_1 -> 600 in the cache */
   /* replace620_1 -> 1857 in the cache */
   /* replace620 -> 1863 in the cache */

   /* ('~iand', ('fge(is_used_once)', 0.0, 'a@32'), ('fge', 'b@32', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   /* search621_0_0 -> 115 in the cache */
   /* search621_0_1 -> 16 in the cache */
   /* search621_0 -> 1854 in the cache */
   /* search621_1_0 -> 36 in the cache */
   /* search621_1_1 -> 115 in the cache */
   /* search621_1 -> 1855 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1854, 1855 },
      -1,
   } },

   /* replace621_0 -> 115 in the cache */
   /* replace621_1_0 -> 16 in the cache */
   /* replace621_1_1_0 -> 36 in the cache */
   /* replace621_1_1 -> 600 in the cache */
   /* replace621_1 -> 1849 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 115, 1849 },
      -1,
   } },

   /* ('~iand', ('fge', 0.0, 'a@32'), ('fge(is_used_once)', 'b@32', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   /* search622_0_0 -> 115 in the cache */
   /* search622_0_1 -> 16 in the cache */
   /* search622_0 -> 1859 in the cache */
   /* search622_1_0 -> 36 in the cache */
   /* search622_1_1 -> 115 in the cache */
   /* search622_1 -> 1860 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1859, 1860 },
      -1,
   } },

   /* replace622_0 -> 115 in the cache */
   /* replace622_1_0 -> 16 in the cache */
   /* replace622_1_1_0 -> 36 in the cache */
   /* replace622_1_1 -> 600 in the cache */
   /* replace622_1 -> 1849 in the cache */
   /* replace622 -> 1866 in the cache */

   /* ('bcsel@32', ('feq', 'a', 0.0), 1.0, ('i2f32', ('iadd', ('b2i32', ('flt', 0.0, 'a@32')), ('ineg', ('b2i32', ('flt', 'a@32', 0.0)))))) => ('i2f32', ('iadd', ('b2i32', ('!fge', 'a', 0.0)), ('ineg', ('b2i32', ('!flt', 'a', 0.0))))) */
   /* search623_0_0 -> 16 in the cache */
   /* search623_0_1 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 16, 115 },
      -1,
   } },
   /* search623_1 -> 243 in the cache */
   /* search623_2_0_0_0_0 -> 115 in the cache */
   /* search623_2_0_0_0_1 -> 16 in the cache */
   /* search623_2_0_0_0 -> 1851 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 1851 },
      -1,
   } },
   /* search623_2_0_1_0_0_0 -> 16 in the cache */
   /* search623_2_0_1_0_0_1 -> 115 in the cache */
   /* search623_2_0_1_0_0 -> 695 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 695 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 1870 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 1869, 1871 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2f32,
      -1, 1,
      { 1872 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 1868, 243, 1873 },
      -1,
   } },

   /* replace623_0_0_0_0 -> 16 in the cache */
   /* replace623_0_0_0_1 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_fge,
      -1, 0,
      { 16, 115 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 1875 },
      -1,
   } },
   /* replace623_0_1_0_0_0 -> 16 in the cache */
   /* replace623_0_1_0_0_1 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_flt,
      -1, 0,
      { 16, 115 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 1877 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 1878 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 1876, 1879 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2f32,
      -1, 1,
      { 1880 },
      -1,
   } },

   /* ('bcsel', 'a', ('b2f(is_used_once)', 'b@32'), ('b2f', 'c@32')) => ('b2f', ('bcsel', 'a', 'b', 'c')) */
   /* search624_0 -> 760 in the cache */
   /* search624_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 36 },
      0,
   } },
   /* search624_2_0 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 1882, 1883 },
      -1,
   } },

   /* replace624_0_0 -> 760 in the cache */
   /* replace624_0_1 -> 36 in the cache */
   /* replace624_0_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 36, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 1885 },
      -1,
   } },

   /* ('~i2f32', ('f2i', 'a@32')) => ('ftrunc', 'a') */
   /* search625_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2i,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_i2f32,
      -1, 0,
      { 1887 },
      -1,
   } },

   /* replace625_0 -> 16 in the cache */
   /* replace625 -> 694 in the cache */

   /* ('~f2i32', ('i2f', 'a@32')) => a */
   /* search626_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_i2f,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2i32,
      -1, 0,
      { 1889 },
      -1,
   } },

   /* replace626 -> 16 in the cache */

   /* ('~f2i32', ('u2f', 'a@32')) => a */
   /* search627_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_u2f,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2i32,
      -1, 0,
      { 1891 },
      -1,
   } },

   /* replace627 -> 16 in the cache */

   /* ('~f2u32', ('i2f', 'a@32')) => a */
   /* search628_0_0 -> 16 in the cache */
   /* search628_0 -> 1889 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2u32,
      -1, 0,
      { 1889 },
      -1,
   } },

   /* replace628 -> 16 in the cache */

   /* ('~f2u32', ('u2f', 'a@32')) => a */
   /* search629_0_0 -> 16 in the cache */
   /* search629_0 -> 1891 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2u32,
      -1, 0,
      { 1891 },
      -1,
   } },

   /* replace629 -> 16 in the cache */

   /* ('fadd', ('b2f32', ('flt', 0.0, 'a@32')), ('fneg', ('b2f32', ('flt', 'a@32', 0.0)))) => ('fsign', 'a') */
   /* search630_0_0_0 -> 115 in the cache */
   /* search630_0_0_1 -> 16 in the cache */
   /* search630_0_0 -> 1851 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2f32,
      -1, 0,
      { 1851 },
      -1,
   } },
   /* search630_1_0_0_0 -> 16 in the cache */
   /* search630_1_0_0_1 -> 115 in the cache */
   /* search630_1_0_0 -> 695 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2f32,
      -1, 0,
      { 695 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 1896 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 1895, 1897 },
      -1,
   } },

   /* replace630_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('iadd', ('b2i32', ('flt', 0, 'a@32')), ('ineg', ('b2i32', ('flt', 'a@32', 0)))) => ('f2i32', ('fsign', 'a')) */
   /* search631_0_0_0 -> 122 in the cache */
   /* search631_0_0_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 122, 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 1900 },
      -1,
   } },
   /* search631_1_0_0_0 -> 16 in the cache */
   /* search631_1_0_0_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 16, 122 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 1902 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 1903 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 1901, 1904 },
      -1,
   } },

   /* replace631_0_0 -> 16 in the cache */
   /* replace631_0 -> 1899 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2i32,
      -1, 0,
      { 1899 },
      -1,
   } },

   /* ('f2f32', ('f2f64', 'a@32')) => a */
   /* search632_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2f64,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 1907 },
      -1,
   } },

   /* replace632 -> 16 in the cache */

   /* ('i2i32', ('i2i64', 'a@32')) => a */
   /* search633_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 1909 },
      -1,
   } },

   /* replace633 -> 16 in the cache */

   /* ('u2u32', ('u2u64', 'a@32')) => a */
   /* search634_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 1911 },
      -1,
   } },

   /* replace634 -> 16 in the cache */

   /* ('f2f32', ('b2f64', 'a@1')) => ('b2f32', 'a') */
   /* search635_0_0 -> 760 in the cache */
   /* search635_0 -> 1412 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 1412 },
      -1,
   } },

   /* replace635_0 -> 760 in the cache */
   /* replace635 -> 1408 in the cache */

   /* ('i2i32', ('b2i64', 'a@1')) => ('b2i32', 'a') */
   /* search636_0_0 -> 760 in the cache */
   /* search636_0 -> 1400 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 1400 },
      -1,
   } },

   /* replace636_0 -> 760 in the cache */
   /* replace636 -> 1396 in the cache */

   /* ('u2u32', ('b2i64', 'a@1')) => ('b2i32', 'a') */
   /* search637_0_0 -> 760 in the cache */
   /* search637_0 -> 1400 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 1400 },
      -1,
   } },

   /* replace637_0 -> 760 in the cache */
   /* replace637 -> 1396 in the cache */

   /* ('f2u64', ('f2f64', 'a@32')) => ('f2u64', 'a') */
   /* search638_0_0 -> 16 in the cache */
   /* search638_0 -> 1907 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2u64,
      -1, 0,
      { 1907 },
      -1,
   } },

   /* replace638_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2u64,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('f2i64', ('f2f64', 'a@32')) => ('f2i64', 'a') */
   /* search639_0_0 -> 16 in the cache */
   /* search639_0 -> 1907 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2i64,
      -1, 0,
      { 1907 },
      -1,
   } },

   /* replace639_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2i64,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('f2f32', ('u2f64', 'a')) => ('u2f32', 'a') */
   /* search640_0_0 -> 32 in the cache */
   /* search640_0 -> 1838 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 1838 },
      -1,
   } },

   /* replace640_0 -> 32 in the cache */
   /* replace640 -> 1815 in the cache */

   /* ('f2f32', ('i2f64', 'a')) => ('i2f32', 'a') */
   /* search641_0_0 -> 32 in the cache */
   /* search641_0 -> 1840 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 1840 },
      -1,
   } },

   /* replace641_0 -> 32 in the cache */
   /* replace641 -> 1818 in the cache */

   /* ('u2f64', ('u2u64', 'a@32')) => ('u2f64', 'a') */
   /* search642_0_0 -> 16 in the cache */
   /* search642_0 -> 1911 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2f64,
      -1, 0,
      { 1911 },
      -1,
   } },

   /* replace642_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2f64,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('i2f64', ('i2i64', 'a@32')) => ('i2f64', 'a') */
   /* search643_0_0 -> 16 in the cache */
   /* search643_0 -> 1909 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2f64,
      -1, 0,
      { 1909 },
      -1,
   } },

   /* replace643_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2f64,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('ior', ('flt(is_used_once)', 0.0, 'a@64'), ('flt', 'b@64', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   /* search644_0_0 -> 701 in the cache */
   /* search644_0_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 701, 21 },
      0,
   } },
   /* search644_1_0 -> 41 in the cache */
   /* search644_1_1 -> 701 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 41, 701 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1926, 1927 },
      -1,
   } },

   /* replace644_0 -> 701 in the cache */
   /* replace644_1_0 -> 21 in the cache */
   /* replace644_1_1_0 -> 41 in the cache */
   /* replace644_1_1 -> 673 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 21, 673 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 701, 1929 },
      -1,
   } },

   /* ('ior', ('flt', 0.0, 'a@64'), ('flt(is_used_once)', 'b@64', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   /* search645_0_0 -> 701 in the cache */
   /* search645_0_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 701, 21 },
      -1,
   } },
   /* search645_1_0 -> 41 in the cache */
   /* search645_1_1 -> 701 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 41, 701 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1931, 1932 },
      -1,
   } },

   /* replace645_0 -> 701 in the cache */
   /* replace645_1_0 -> 21 in the cache */
   /* replace645_1_1_0 -> 41 in the cache */
   /* replace645_1_1 -> 673 in the cache */
   /* replace645_1 -> 1929 in the cache */
   /* replace645 -> 1930 in the cache */

   /* ('ior', ('fge(is_used_once)', 0.0, 'a@64'), ('fge', 'b@64', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search646_0_0 -> 701 in the cache */
   /* search646_0_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 701, 21 },
      0,
   } },
   /* search646_1_0 -> 41 in the cache */
   /* search646_1_1 -> 701 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 41, 701 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1934, 1935 },
      -1,
   } },

   /* replace646_0 -> 701 in the cache */
   /* replace646_1_0 -> 21 in the cache */
   /* replace646_1_1_0 -> 41 in the cache */
   /* replace646_1_1 -> 673 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 21, 673 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 701, 1937 },
      -1,
   } },

   /* ('ior', ('fge', 0.0, 'a@64'), ('fge(is_used_once)', 'b@64', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search647_0_0 -> 701 in the cache */
   /* search647_0_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 701, 21 },
      -1,
   } },
   /* search647_1_0 -> 41 in the cache */
   /* search647_1_1 -> 701 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 41, 701 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1939, 1940 },
      -1,
   } },

   /* replace647_0 -> 701 in the cache */
   /* replace647_1_0 -> 21 in the cache */
   /* replace647_1_1_0 -> 41 in the cache */
   /* replace647_1_1 -> 673 in the cache */
   /* replace647_1 -> 1937 in the cache */
   /* replace647 -> 1938 in the cache */

   /* ('~iand', ('flt(is_used_once)', 0.0, 'a@64'), ('flt', 'b@64', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search648_0_0 -> 701 in the cache */
   /* search648_0_1 -> 21 in the cache */
   /* search648_0 -> 1926 in the cache */
   /* search648_1_0 -> 41 in the cache */
   /* search648_1_1 -> 701 in the cache */
   /* search648_1 -> 1927 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1926, 1927 },
      -1,
   } },

   /* replace648_0 -> 701 in the cache */
   /* replace648_1_0 -> 21 in the cache */
   /* replace648_1_1_0 -> 41 in the cache */
   /* replace648_1_1 -> 673 in the cache */
   /* replace648_1 -> 1937 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 701, 1937 },
      -1,
   } },

   /* ('~iand', ('flt', 0.0, 'a@64'), ('flt(is_used_once)', 'b@64', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search649_0_0 -> 701 in the cache */
   /* search649_0_1 -> 21 in the cache */
   /* search649_0 -> 1931 in the cache */
   /* search649_1_0 -> 41 in the cache */
   /* search649_1_1 -> 701 in the cache */
   /* search649_1 -> 1932 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1931, 1932 },
      -1,
   } },

   /* replace649_0 -> 701 in the cache */
   /* replace649_1_0 -> 21 in the cache */
   /* replace649_1_1_0 -> 41 in the cache */
   /* replace649_1_1 -> 673 in the cache */
   /* replace649_1 -> 1937 in the cache */
   /* replace649 -> 1943 in the cache */

   /* ('~iand', ('fge(is_used_once)', 0.0, 'a@64'), ('fge', 'b@64', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   /* search650_0_0 -> 701 in the cache */
   /* search650_0_1 -> 21 in the cache */
   /* search650_0 -> 1934 in the cache */
   /* search650_1_0 -> 41 in the cache */
   /* search650_1_1 -> 701 in the cache */
   /* search650_1 -> 1935 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1934, 1935 },
      -1,
   } },

   /* replace650_0 -> 701 in the cache */
   /* replace650_1_0 -> 21 in the cache */
   /* replace650_1_1_0 -> 41 in the cache */
   /* replace650_1_1 -> 673 in the cache */
   /* replace650_1 -> 1929 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 701, 1929 },
      -1,
   } },

   /* ('~iand', ('fge', 0.0, 'a@64'), ('fge(is_used_once)', 'b@64', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   /* search651_0_0 -> 701 in the cache */
   /* search651_0_1 -> 21 in the cache */
   /* search651_0 -> 1939 in the cache */
   /* search651_1_0 -> 41 in the cache */
   /* search651_1_1 -> 701 in the cache */
   /* search651_1 -> 1940 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1939, 1940 },
      -1,
   } },

   /* replace651_0 -> 701 in the cache */
   /* replace651_1_0 -> 21 in the cache */
   /* replace651_1_1_0 -> 41 in the cache */
   /* replace651_1_1 -> 673 in the cache */
   /* replace651_1 -> 1929 in the cache */
   /* replace651 -> 1946 in the cache */

   /* ('bcsel@64', ('feq', 'a', 0.0), 1.0, ('i2f64', ('iadd', ('b2i64', ('flt', 0.0, 'a@64')), ('ineg', ('b2i64', ('flt', 'a@64', 0.0)))))) => ('i2f64', ('iadd', ('b2i32', ('!fge', 'a', 0.0)), ('ineg', ('b2i32', ('!flt', 'a', 0.0))))) */
   /* search652_0_0 -> 21 in the cache */
   /* search652_0_1 -> 701 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 21, 701 },
      -1,
   } },
   /* search652_1 -> 622 in the cache */
   /* search652_2_0_0_0_0 -> 701 in the cache */
   /* search652_2_0_0_0_1 -> 21 in the cache */
   /* search652_2_0_0_0 -> 1931 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2i64,
      -1, 0,
      { 1931 },
      -1,
   } },
   /* search652_2_0_1_0_0_0 -> 21 in the cache */
   /* search652_2_0_1_0_0_1 -> 701 in the cache */
   /* search652_2_0_1_0_0 -> 702 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2i64,
      -1, 0,
      { 702 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 1950 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 1949, 1951 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2f64,
      -1, 1,
      { 1952 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 1948, 622, 1953 },
      -1,
   } },

   /* replace652_0_0_0_0 -> 21 in the cache */
   /* replace652_0_0_0_1 -> 701 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_fge,
      -1, 0,
      { 21, 701 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 1955 },
      -1,
   } },
   /* replace652_0_1_0_0_0 -> 21 in the cache */
   /* replace652_0_1_0_0_1 -> 701 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_flt,
      -1, 0,
      { 21, 701 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 1957 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 1958 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 1956, 1959 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2f64,
      -1, 1,
      { 1960 },
      -1,
   } },

   /* ('bcsel', 'a', ('b2f(is_used_once)', 'b@64'), ('b2f', 'c@64')) => ('b2f', ('bcsel', 'a', 'b', 'c')) */
   /* search653_0 -> 760 in the cache */
   /* search653_1_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 41 },
      0,
   } },
   /* search653_2_0 -> 610 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 610 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 1962, 1963 },
      -1,
   } },

   /* replace653_0_0 -> 760 in the cache */
   /* replace653_0_1 -> 41 in the cache */
   /* replace653_0_2 -> 610 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 41, 610 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 1965 },
      -1,
   } },

   /* ('~i2f64', ('f2i', 'a@64')) => ('ftrunc', 'a') */
   /* search654_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2i,
      -1, 0,
      { 21 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_i2f64,
      -1, 0,
      { 1967 },
      -1,
   } },

   /* replace654_0 -> 21 in the cache */
   /* replace654 -> 700 in the cache */

   /* ('~f2i64', ('i2f', 'a@64')) => a */
   /* search655_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_i2f,
      -1, 0,
      { 21 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_f2i64,
      -1, 0,
      { 1969 },
      -1,
   } },

   /* replace655 -> 21 in the cache */

   /* ('~f2i64', ('u2f', 'a@64')) => a */
   /* search656_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_u2f,
      -1, 0,
      { 21 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_f2i64,
      -1, 0,
      { 1971 },
      -1,
   } },

   /* replace656 -> 21 in the cache */

   /* ('~f2u64', ('i2f', 'a@64')) => a */
   /* search657_0_0 -> 21 in the cache */
   /* search657_0 -> 1969 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_f2u64,
      -1, 0,
      { 1969 },
      -1,
   } },

   /* replace657 -> 21 in the cache */

   /* ('~f2u64', ('u2f', 'a@64')) => a */
   /* search658_0_0 -> 21 in the cache */
   /* search658_0 -> 1971 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_f2u64,
      -1, 0,
      { 1971 },
      -1,
   } },

   /* replace658 -> 21 in the cache */

   /* ('fadd', ('b2f64', ('flt', 0.0, 'a@64')), ('fneg', ('b2f64', ('flt', 'a@64', 0.0)))) => ('fsign', 'a') */
   /* search659_0_0_0 -> 701 in the cache */
   /* search659_0_0_1 -> 21 in the cache */
   /* search659_0_0 -> 1931 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2f64,
      -1, 0,
      { 1931 },
      -1,
   } },
   /* search659_1_0_0_0 -> 21 in the cache */
   /* search659_1_0_0_1 -> 701 in the cache */
   /* search659_1_0_0 -> 702 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2f64,
      -1, 0,
      { 702 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 1976 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 1975, 1977 },
      -1,
   } },

   /* replace659_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 21 },
      -1,
   } },

   /* ('iadd', ('b2i64', ('flt', 0, 'a@64')), ('ineg', ('b2i64', ('flt', 'a@64', 0)))) => ('f2i64', ('fsign', 'a')) */
   /* search660_0_0_0 -> 897 in the cache */
   /* search660_0_0_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 897, 21 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2i64,
      -1, 0,
      { 1980 },
      -1,
   } },
   /* search660_1_0_0_0 -> 21 in the cache */
   /* search660_1_0_0_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 21, 897 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2i64,
      -1, 0,
      { 1982 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 1983 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 1981, 1984 },
      -1,
   } },

   /* replace660_0_0 -> 21 in the cache */
   /* replace660_0 -> 1979 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2i64,
      -1, 0,
      { 1979 },
      -1,
   } },

   /* ('f2fmp', ('u2f32', 'a@32')) => ('u2fmp', 'a') */
   /* search661_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2f32,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 1987 },
      -1,
   } },

   /* replace661_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2fmp,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('f2fmp', ('i2f32', 'a@32')) => ('i2fmp', 'a') */
   /* search662_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2f32,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 1990 },
      -1,
   } },

   /* replace662_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2fmp,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('f2u16', ('f2fmp', 'a@32')) => ('f2u16', 'a') */
   /* search663_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2u16,
      -1, 0,
      { 1993 },
      -1,
   } },

   /* replace663_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2u16,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('f2i16', ('f2fmp', 'a@32')) => ('f2i16', 'a') */
   /* search664_0_0 -> 16 in the cache */
   /* search664_0 -> 1993 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2i16,
      -1, 0,
      { 1993 },
      -1,
   } },

   /* replace664_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2i16,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('i2imp', ('f2u32', 'a@32')) => ('f2ump', 'a') */
   /* search665_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2u32,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 1998 },
      -1,
   } },

   /* replace665_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2ump,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('i2imp', ('f2i32', 'a@32')) => ('f2imp', 'a') */
   /* search666_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2i32,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 2001 },
      -1,
   } },

   /* replace666_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2imp,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('u2f16', ('i2imp', 'a@32')) => ('u2f16', 'a') */
   /* search667_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2f16,
      -1, 0,
      { 2004 },
      -1,
   } },

   /* replace667_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2f16,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('i2f16', ('i2imp', 'a@32')) => ('i2f16', 'a') */
   /* search668_0_0 -> 16 in the cache */
   /* search668_0 -> 2004 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2f16,
      -1, 0,
      { 2004 },
      -1,
   } },

   /* replace668_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2f16,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('iand', ('u2u16', ('u2u8', 'a@16')), '#b') => ('iand', 'a', ('iand', 'b', 255)) */
   /* search669_0_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 2009 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 16 },
      1, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2010, 2011 },
      -1,
   } },

   /* replace669_0 -> 11 in the cache */
   /* replace669_1_0 -> 258 in the cache */
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { 0xff /* 255 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 258, 2013 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 11, 2014 },
      -1,
   } },

   /* ('u2u16', ('u2u8(is_used_once)', ('iand', 'a@16', '#b'))) => ('iand', 'a', ('iand', 'b', 255)) */
   /* search670_0_0_0 -> 11 in the cache */
   /* search670_0_0_1 -> 2011 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 11, 2011 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 2016 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 1,
      { 2017 },
      -1,
   } },

   /* replace670_0 -> 11 in the cache */
   /* replace670_1_0 -> 258 in the cache */
   /* replace670_1_1 -> 2013 in the cache */
   /* replace670_1 -> 2014 in the cache */
   /* replace670 -> 2015 in the cache */

   /* ('u2u8', ('iand', ('u2u16', ('u2u8', 'a@16')), ('u2u16', ('u2u8', 'b@16')))) => ('u2u8', ('iand', 'a', 'b')) */
   /* search671_0_0_0_0 -> 11 in the cache */
   /* search671_0_0_0 -> 2009 in the cache */
   /* search671_0_0 -> 2010 in the cache */
   /* search671_0_1_0_0 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 258 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 2019 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2010, 2020 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 2021 },
      -1,
   } },

   /* replace671_0_0 -> 11 in the cache */
   /* replace671_0_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 11, 258 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 2023 },
      -1,
   } },

   /* ('u2u8', ('iand', ('u2u16', ('u2u8', 'a@32')), ('u2u16', ('u2u8', 'b@32')))) => ('u2u8', ('iand', 'a', 'b')) */
   /* search672_0_0_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 2025 },
      -1,
   } },
   /* search672_0_1_0_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 2027 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2026, 2028 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 2029 },
      -1,
   } },

   /* replace672_0_0 -> 16 in the cache */
   /* replace672_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 16, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 2031 },
      -1,
   } },

   /* ('iand', ('extract_i8', 'a', '#b'), ('extract_i8', 'c', 'b')) => ('extract_i8', ('iand', 'a', 'c'), 'b') */
   /* search673_0_0 -> 0 in the cache */
   /* search673_0_1 -> 821 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 0, 821 },
      -1,
   } },
   /* search673_1_0 -> 140 in the cache */
   /* search673_1_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 140, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2033, 2034 },
      -1,
   } },

   /* replace673_0_0 -> 0 in the cache */
   /* replace673_0_1 -> 140 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 0, 140 },
      -1,
   } },
   /* replace673_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 1,
      { 2036, 3 },
      -1,
   } },

   /* ('iand', ('extract_u8', 'a', '#b'), ('extract_u8', 'c', 'b')) => ('extract_u8', ('iand', 'a', 'c'), 'b') */
   /* search674_0_0 -> 0 in the cache */
   /* search674_0_1 -> 821 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 0, 821 },
      -1,
   } },
   /* search674_1_0 -> 140 in the cache */
   /* search674_1_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 140, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2038, 2039 },
      -1,
   } },

   /* replace674_0_0 -> 0 in the cache */
   /* replace674_0_1 -> 140 in the cache */
   /* replace674_0 -> 2036 in the cache */
   /* replace674_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 1,
      { 2036, 3 },
      -1,
   } },

   /* ('iand', ('extract_i16', 'a', '#b'), ('extract_i16', 'c', 'b')) => ('extract_i16', ('iand', 'a', 'c'), 'b') */
   /* search675_0_0 -> 0 in the cache */
   /* search675_0_1 -> 821 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 0, 821 },
      -1,
   } },
   /* search675_1_0 -> 140 in the cache */
   /* search675_1_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 140, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2042, 2043 },
      -1,
   } },

   /* replace675_0_0 -> 0 in the cache */
   /* replace675_0_1 -> 140 in the cache */
   /* replace675_0 -> 2036 in the cache */
   /* replace675_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 1,
      { 2036, 3 },
      -1,
   } },

   /* ('iand', ('extract_u16', 'a', '#b'), ('extract_u16', 'c', 'b')) => ('extract_u16', ('iand', 'a', 'c'), 'b') */
   /* search676_0_0 -> 0 in the cache */
   /* search676_0_1 -> 821 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 0, 821 },
      -1,
   } },
   /* search676_1_0 -> 140 in the cache */
   /* search676_1_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 140, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2046, 2047 },
      -1,
   } },

   /* replace676_0_0 -> 0 in the cache */
   /* replace676_0_1 -> 140 in the cache */
   /* replace676_0 -> 2036 in the cache */
   /* replace676_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 1,
      { 2036, 3 },
      -1,
   } },

   /* ('iand', ('ushr(is_used_once)', 'a', '#b'), ('ushr', 'c', 'b')) => ('ushr', ('iand', 'a', 'c'), 'b') */
   /* search677_0_0 -> 126 in the cache */
   /* search677_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 126, 33 },
      0,
   } },
   /* search677_1_0 -> 129 in the cache */
   /* search677_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 129, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2050, 2051 },
      -1,
   } },

   /* replace677_0_0 -> 126 in the cache */
   /* replace677_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 126, 129 },
      -1,
   } },
   /* replace677_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 2053, 36 },
      -1,
   } },

   /* ('iand', ('ishr(is_used_once)', 'a', '#b'), ('ishr', 'c', 'b')) => ('ishr', ('iand', 'a', 'c'), 'b') */
   /* search678_0_0 -> 126 in the cache */
   /* search678_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 126, 33 },
      0,
   } },
   /* search678_1_0 -> 129 in the cache */
   /* search678_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 129, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2055, 2056 },
      -1,
   } },

   /* replace678_0_0 -> 126 in the cache */
   /* replace678_0_1 -> 129 in the cache */
   /* replace678_0 -> 2053 in the cache */
   /* replace678_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 2053, 36 },
      -1,
   } },

   /* ('iand', ('ishl(is_used_once)', 'a', '#b'), ('ishl', 'c', 'b')) => ('ishl', ('iand', 'a', 'c'), 'b') */
   /* search679_0_0 -> 126 in the cache */
   /* search679_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 126, 33 },
      0,
   } },
   /* search679_1_0 -> 129 in the cache */
   /* search679_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 129, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2059, 2060 },
      -1,
   } },

   /* replace679_0_0 -> 126 in the cache */
   /* replace679_0_1 -> 129 in the cache */
   /* replace679_0 -> 2053 in the cache */
   /* replace679_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 2053, 36 },
      -1,
   } },

   /* ('u2u8', ('ior', ('u2u16', ('u2u8', 'a@16')), ('u2u16', ('u2u8', 'b@16')))) => ('u2u8', ('ior', 'a', 'b')) */
   /* search680_0_0_0_0 -> 11 in the cache */
   /* search680_0_0_0 -> 2009 in the cache */
   /* search680_0_0 -> 2010 in the cache */
   /* search680_0_1_0_0 -> 258 in the cache */
   /* search680_0_1_0 -> 2019 in the cache */
   /* search680_0_1 -> 2020 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2010, 2020 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 2063 },
      -1,
   } },

   /* replace680_0_0 -> 11 in the cache */
   /* replace680_0_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 11, 258 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 2065 },
      -1,
   } },

   /* ('u2u8', ('ior', ('u2u16', ('u2u8', 'a@32')), ('u2u16', ('u2u8', 'b@32')))) => ('u2u8', ('ior', 'a', 'b')) */
   /* search681_0_0_0_0 -> 16 in the cache */
   /* search681_0_0_0 -> 2025 in the cache */
   /* search681_0_0 -> 2026 in the cache */
   /* search681_0_1_0_0 -> 36 in the cache */
   /* search681_0_1_0 -> 2027 in the cache */
   /* search681_0_1 -> 2028 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2026, 2028 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 2067 },
      -1,
   } },

   /* replace681_0_0 -> 16 in the cache */
   /* replace681_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 16, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 2069 },
      -1,
   } },

   /* ('ior', ('extract_i8', 'a', '#b'), ('extract_i8', 'c', 'b')) => ('extract_i8', ('ior', 'a', 'c'), 'b') */
   /* search682_0_0 -> 0 in the cache */
   /* search682_0_1 -> 821 in the cache */
   /* search682_0 -> 2033 in the cache */
   /* search682_1_0 -> 140 in the cache */
   /* search682_1_1 -> 3 in the cache */
   /* search682_1 -> 2034 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2033, 2034 },
      -1,
   } },

   /* replace682_0_0 -> 0 in the cache */
   /* replace682_0_1 -> 140 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 0, 140 },
      -1,
   } },
   /* replace682_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 1,
      { 2072, 3 },
      -1,
   } },

   /* ('ior', ('extract_u8', 'a', '#b'), ('extract_u8', 'c', 'b')) => ('extract_u8', ('ior', 'a', 'c'), 'b') */
   /* search683_0_0 -> 0 in the cache */
   /* search683_0_1 -> 821 in the cache */
   /* search683_0 -> 2038 in the cache */
   /* search683_1_0 -> 140 in the cache */
   /* search683_1_1 -> 3 in the cache */
   /* search683_1 -> 2039 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2038, 2039 },
      -1,
   } },

   /* replace683_0_0 -> 0 in the cache */
   /* replace683_0_1 -> 140 in the cache */
   /* replace683_0 -> 2072 in the cache */
   /* replace683_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 1,
      { 2072, 3 },
      -1,
   } },

   /* ('ior', ('extract_i16', 'a', '#b'), ('extract_i16', 'c', 'b')) => ('extract_i16', ('ior', 'a', 'c'), 'b') */
   /* search684_0_0 -> 0 in the cache */
   /* search684_0_1 -> 821 in the cache */
   /* search684_0 -> 2042 in the cache */
   /* search684_1_0 -> 140 in the cache */
   /* search684_1_1 -> 3 in the cache */
   /* search684_1 -> 2043 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2042, 2043 },
      -1,
   } },

   /* replace684_0_0 -> 0 in the cache */
   /* replace684_0_1 -> 140 in the cache */
   /* replace684_0 -> 2072 in the cache */
   /* replace684_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 1,
      { 2072, 3 },
      -1,
   } },

   /* ('ior', ('extract_u16', 'a', '#b'), ('extract_u16', 'c', 'b')) => ('extract_u16', ('ior', 'a', 'c'), 'b') */
   /* search685_0_0 -> 0 in the cache */
   /* search685_0_1 -> 821 in the cache */
   /* search685_0 -> 2046 in the cache */
   /* search685_1_0 -> 140 in the cache */
   /* search685_1_1 -> 3 in the cache */
   /* search685_1 -> 2047 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2046, 2047 },
      -1,
   } },

   /* replace685_0_0 -> 0 in the cache */
   /* replace685_0_1 -> 140 in the cache */
   /* replace685_0 -> 2072 in the cache */
   /* replace685_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 1,
      { 2072, 3 },
      -1,
   } },

   /* ('ior', ('ushr(is_used_once)', 'a', '#b'), ('ushr', 'c', 'b')) => ('ushr', ('ior', 'a', 'c'), 'b') */
   /* search686_0_0 -> 126 in the cache */
   /* search686_0_1 -> 33 in the cache */
   /* search686_0 -> 2050 in the cache */
   /* search686_1_0 -> 129 in the cache */
   /* search686_1_1 -> 36 in the cache */
   /* search686_1 -> 2051 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2050, 2051 },
      -1,
   } },

   /* replace686_0_0 -> 126 in the cache */
   /* replace686_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 126, 129 },
      -1,
   } },
   /* replace686_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 2081, 36 },
      -1,
   } },

   /* ('ior', ('ishr(is_used_once)', 'a', '#b'), ('ishr', 'c', 'b')) => ('ishr', ('ior', 'a', 'c'), 'b') */
   /* search687_0_0 -> 126 in the cache */
   /* search687_0_1 -> 33 in the cache */
   /* search687_0 -> 2055 in the cache */
   /* search687_1_0 -> 129 in the cache */
   /* search687_1_1 -> 36 in the cache */
   /* search687_1 -> 2056 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2055, 2056 },
      -1,
   } },

   /* replace687_0_0 -> 126 in the cache */
   /* replace687_0_1 -> 129 in the cache */
   /* replace687_0 -> 2081 in the cache */
   /* replace687_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 2081, 36 },
      -1,
   } },

   /* ('ior', ('ishl(is_used_once)', 'a', '#b'), ('ishl', 'c', 'b')) => ('ishl', ('ior', 'a', 'c'), 'b') */
   /* search688_0_0 -> 126 in the cache */
   /* search688_0_1 -> 33 in the cache */
   /* search688_0 -> 2059 in the cache */
   /* search688_1_0 -> 129 in the cache */
   /* search688_1_1 -> 36 in the cache */
   /* search688_1 -> 2060 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2059, 2060 },
      -1,
   } },

   /* replace688_0_0 -> 126 in the cache */
   /* replace688_0_1 -> 129 in the cache */
   /* replace688_0 -> 2081 in the cache */
   /* replace688_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 2081, 36 },
      -1,
   } },

   /* ('u2u8', ('ixor', ('u2u16', ('u2u8', 'a@16')), ('u2u16', ('u2u8', 'b@16')))) => ('u2u8', ('ixor', 'a', 'b')) */
   /* search689_0_0_0_0 -> 11 in the cache */
   /* search689_0_0_0 -> 2009 in the cache */
   /* search689_0_0 -> 2010 in the cache */
   /* search689_0_1_0_0 -> 258 in the cache */
   /* search689_0_1_0 -> 2019 in the cache */
   /* search689_0_1 -> 2020 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 2010, 2020 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 2087 },
      -1,
   } },

   /* replace689_0_0 -> 11 in the cache */
   /* replace689_0_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 11, 258 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 2089 },
      -1,
   } },

   /* ('u2u8', ('ixor', ('u2u16', ('u2u8', 'a@32')), ('u2u16', ('u2u8', 'b@32')))) => ('u2u8', ('ixor', 'a', 'b')) */
   /* search690_0_0_0_0 -> 16 in the cache */
   /* search690_0_0_0 -> 2025 in the cache */
   /* search690_0_0 -> 2026 in the cache */
   /* search690_0_1_0_0 -> 36 in the cache */
   /* search690_0_1_0 -> 2027 in the cache */
   /* search690_0_1 -> 2028 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 2026, 2028 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 2091 },
      -1,
   } },

   /* replace690_0_0 -> 16 in the cache */
   /* replace690_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 16, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 2093 },
      -1,
   } },

   /* ('ixor', ('extract_i8', 'a', '#b'), ('extract_i8', 'c', 'b')) => ('extract_i8', ('ixor', 'a', 'c'), 'b') */
   /* search691_0_0 -> 0 in the cache */
   /* search691_0_1 -> 821 in the cache */
   /* search691_0 -> 2033 in the cache */
   /* search691_1_0 -> 140 in the cache */
   /* search691_1_1 -> 3 in the cache */
   /* search691_1 -> 2034 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 2033, 2034 },
      -1,
   } },

   /* replace691_0_0 -> 0 in the cache */
   /* replace691_0_1 -> 140 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 0, 140 },
      -1,
   } },
   /* replace691_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 1,
      { 2096, 3 },
      -1,
   } },

   /* ('ixor', ('extract_u8', 'a', '#b'), ('extract_u8', 'c', 'b')) => ('extract_u8', ('ixor', 'a', 'c'), 'b') */
   /* search692_0_0 -> 0 in the cache */
   /* search692_0_1 -> 821 in the cache */
   /* search692_0 -> 2038 in the cache */
   /* search692_1_0 -> 140 in the cache */
   /* search692_1_1 -> 3 in the cache */
   /* search692_1 -> 2039 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 2038, 2039 },
      -1,
   } },

   /* replace692_0_0 -> 0 in the cache */
   /* replace692_0_1 -> 140 in the cache */
   /* replace692_0 -> 2096 in the cache */
   /* replace692_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 1,
      { 2096, 3 },
      -1,
   } },

   /* ('ixor', ('extract_i16', 'a', '#b'), ('extract_i16', 'c', 'b')) => ('extract_i16', ('ixor', 'a', 'c'), 'b') */
   /* search693_0_0 -> 0 in the cache */
   /* search693_0_1 -> 821 in the cache */
   /* search693_0 -> 2042 in the cache */
   /* search693_1_0 -> 140 in the cache */
   /* search693_1_1 -> 3 in the cache */
   /* search693_1 -> 2043 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 2042, 2043 },
      -1,
   } },

   /* replace693_0_0 -> 0 in the cache */
   /* replace693_0_1 -> 140 in the cache */
   /* replace693_0 -> 2096 in the cache */
   /* replace693_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 1,
      { 2096, 3 },
      -1,
   } },

   /* ('ixor', ('extract_u16', 'a', '#b'), ('extract_u16', 'c', 'b')) => ('extract_u16', ('ixor', 'a', 'c'), 'b') */
   /* search694_0_0 -> 0 in the cache */
   /* search694_0_1 -> 821 in the cache */
   /* search694_0 -> 2046 in the cache */
   /* search694_1_0 -> 140 in the cache */
   /* search694_1_1 -> 3 in the cache */
   /* search694_1 -> 2047 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 2046, 2047 },
      -1,
   } },

   /* replace694_0_0 -> 0 in the cache */
   /* replace694_0_1 -> 140 in the cache */
   /* replace694_0 -> 2096 in the cache */
   /* replace694_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 1,
      { 2096, 3 },
      -1,
   } },

   /* ('ixor', ('ushr(is_used_once)', 'a', '#b'), ('ushr', 'c', 'b')) => ('ushr', ('ixor', 'a', 'c'), 'b') */
   /* search695_0_0 -> 126 in the cache */
   /* search695_0_1 -> 33 in the cache */
   /* search695_0 -> 2050 in the cache */
   /* search695_1_0 -> 129 in the cache */
   /* search695_1_1 -> 36 in the cache */
   /* search695_1 -> 2051 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 2050, 2051 },
      -1,
   } },

   /* replace695_0_0 -> 126 in the cache */
   /* replace695_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 126, 129 },
      -1,
   } },
   /* replace695_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 2105, 36 },
      -1,
   } },

   /* ('ixor', ('ishr(is_used_once)', 'a', '#b'), ('ishr', 'c', 'b')) => ('ishr', ('ixor', 'a', 'c'), 'b') */
   /* search696_0_0 -> 126 in the cache */
   /* search696_0_1 -> 33 in the cache */
   /* search696_0 -> 2055 in the cache */
   /* search696_1_0 -> 129 in the cache */
   /* search696_1_1 -> 36 in the cache */
   /* search696_1 -> 2056 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 2055, 2056 },
      -1,
   } },

   /* replace696_0_0 -> 126 in the cache */
   /* replace696_0_1 -> 129 in the cache */
   /* replace696_0 -> 2105 in the cache */
   /* replace696_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 2105, 36 },
      -1,
   } },

   /* ('ixor', ('ishl(is_used_once)', 'a', '#b'), ('ishl', 'c', 'b')) => ('ishl', ('ixor', 'a', 'c'), 'b') */
   /* search697_0_0 -> 126 in the cache */
   /* search697_0_1 -> 33 in the cache */
   /* search697_0 -> 2059 in the cache */
   /* search697_1_0 -> 129 in the cache */
   /* search697_1_1 -> 36 in the cache */
   /* search697_1 -> 2060 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 2059, 2060 },
      -1,
   } },

   /* replace697_0_0 -> 126 in the cache */
   /* replace697_0_1 -> 129 in the cache */
   /* replace697_0 -> 2105 in the cache */
   /* replace697_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 2105, 36 },
      -1,
   } },

   /* ('iand', ('ieq', 'a@8', 0), ('ieq', 'b@8', 0)) => ('ieq', ('ior', 'a', 'b'), 0) */
   /* search698_0_0 -> 6 in the cache */
   /* search698_0_1 -> 835 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 6, 835 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 8 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search698_1_1 -> 835 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 2112, 835 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 2111, 2113 },
      -1,
   } },

   /* replace698_0_0 -> 6 in the cache */
   /* replace698_0_1 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 6, 2112 },
      -1,
   } },
   /* replace698_1 -> 835 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2115, 835 },
      -1,
   } },

   /* ('ior', ('ine', 'a@8', 0), ('ine', 'b@8', 0)) => ('ine', ('ior', 'a', 'b'), 0) */
   /* search699_0_0 -> 6 in the cache */
   /* search699_0_1 -> 835 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 6, 835 },
      -1,
   } },
   /* search699_1_0 -> 2112 in the cache */
   /* search699_1_1 -> 835 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 2112, 835 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 2117, 2118 },
      -1,
   } },

   /* replace699_0_0 -> 6 in the cache */
   /* replace699_0_1 -> 2112 in the cache */
   /* replace699_0 -> 2115 in the cache */
   /* replace699_1 -> 835 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2115, 835 },
      -1,
   } },

   /* ('iand', ('ieq', 'a@8', 0), ('ieq', 'b@8', 0)) => ('ieq', ('umax', 'a', 'b'), 0) */
   /* search700_0_0 -> 6 in the cache */
   /* search700_0_1 -> 835 in the cache */
   /* search700_0 -> 2111 in the cache */
   /* search700_1_0 -> 2112 in the cache */
   /* search700_1_1 -> 835 in the cache */
   /* search700_1 -> 2113 in the cache */
   /* search700 -> 2114 in the cache */

   /* replace700_0_0 -> 6 in the cache */
   /* replace700_0_1 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_umax,
      1, 1,
      { 6, 2112 },
      -1,
   } },
   /* replace700_1 -> 835 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2121, 835 },
      -1,
   } },

   /* ('ior', ('ieq', 'a@8', 0), ('ieq', 'b@8', 0)) => ('ieq', ('umin', 'a', 'b'), 0) */
   /* search701_0_0 -> 6 in the cache */
   /* search701_0_1 -> 835 in the cache */
   /* search701_0 -> 2111 in the cache */
   /* search701_1_0 -> 2112 in the cache */
   /* search701_1_1 -> 835 in the cache */
   /* search701_1 -> 2113 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 2111, 2113 },
      -1,
   } },

   /* replace701_0_0 -> 6 in the cache */
   /* replace701_0_1 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 6, 2112 },
      -1,
   } },
   /* replace701_1 -> 835 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2124, 835 },
      -1,
   } },

   /* ('iand', ('ine', 'a@8', 0), ('ine', 'b@8', 0)) => ('ine', ('umin', 'a', 'b'), 0) */
   /* search702_0_0 -> 6 in the cache */
   /* search702_0_1 -> 835 in the cache */
   /* search702_0 -> 2117 in the cache */
   /* search702_1_0 -> 2112 in the cache */
   /* search702_1_1 -> 835 in the cache */
   /* search702_1 -> 2118 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 2117, 2118 },
      -1,
   } },

   /* replace702_0_0 -> 6 in the cache */
   /* replace702_0_1 -> 2112 in the cache */
   /* replace702_0 -> 2124 in the cache */
   /* replace702_1 -> 835 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2124, 835 },
      -1,
   } },

   /* ('ior', ('ine', 'a@8', 0), ('ine', 'b@8', 0)) => ('ine', ('umax', 'a', 'b'), 0) */
   /* search703_0_0 -> 6 in the cache */
   /* search703_0_1 -> 835 in the cache */
   /* search703_0 -> 2117 in the cache */
   /* search703_1_0 -> 2112 in the cache */
   /* search703_1_1 -> 835 in the cache */
   /* search703_1 -> 2118 in the cache */
   /* search703 -> 2119 in the cache */

   /* replace703_0_0 -> 6 in the cache */
   /* replace703_0_1 -> 2112 in the cache */
   /* replace703_0 -> 2121 in the cache */
   /* replace703_1 -> 835 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2121, 835 },
      -1,
   } },

   /* ('ineg', ('b2i8', 'a@8')) => a */
   /* search704_0_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_b2i8,
      -1, 0,
      { 6 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 2129 },
      -1,
   } },

   /* replace704 -> 6 in the cache */

   /* ('ishl', 'a@8', ('iand', 7, 'b')) => ('ishl', 'a', 'b') */
   /* search705_0 -> 6 in the cache */
   /* search705_1_0 -> 9 in the cache */
   /* search705_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 9, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 6, 2131 },
      -1,
   } },

   /* replace705_0 -> 6 in the cache */
   /* replace705_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 6, 36 },
      -1,
   } },

   /* ('ishr', 'a@8', ('iand', 7, 'b')) => ('ishr', 'a', 'b') */
   /* search706_0 -> 6 in the cache */
   /* search706_1_0 -> 9 in the cache */
   /* search706_1_1 -> 36 in the cache */
   /* search706_1 -> 2131 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 6, 2131 },
      -1,
   } },

   /* replace706_0 -> 6 in the cache */
   /* replace706_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 6, 36 },
      -1,
   } },

   /* ('ushr', 'a@8', ('iand', 7, 'b')) => ('ushr', 'a', 'b') */
   /* search707_0 -> 6 in the cache */
   /* search707_1_0 -> 9 in the cache */
   /* search707_1_1 -> 36 in the cache */
   /* search707_1 -> 2131 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 6, 2131 },
      -1,
   } },

   /* replace707_0 -> 6 in the cache */
   /* replace707_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 6, 36 },
      -1,
   } },

   /* ('ushr', 'a@8', ('ishl(is_used_once)', ('iand', 'b', 1), 2)) => ('ushr', 'a', ('ishl', 'b', 2)) */
   /* search708_0 -> 6 in the cache */
   /* search708_1_0_0 -> 36 in the cache */
   /* search708_1_0_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 36, 396 },
      -1,
   } },
   /* search708_1_1 -> 401 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 2138, 401 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 6, 2139 },
      -1,
   } },

   /* replace708_0 -> 6 in the cache */
   /* replace708_1_0 -> 36 in the cache */
   /* replace708_1_1 -> 401 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 36, 401 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 6, 2141 },
      -1,
   } },

   /* ('iand', ('ieq', 'a@16', 0), ('ieq', 'b@16', 0)) => ('ieq', ('ior', 'a', 'b'), 0) */
   /* search709_0_0 -> 11 in the cache */
   /* search709_0_1 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 11, 856 },
      -1,
   } },
   /* search709_1_0 -> 258 in the cache */
   /* search709_1_1 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 258, 856 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 2143, 2144 },
      -1,
   } },

   /* replace709_0_0 -> 11 in the cache */
   /* replace709_0_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 11, 258 },
      -1,
   } },
   /* replace709_1 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2146, 856 },
      -1,
   } },

   /* ('ior', ('ine', 'a@16', 0), ('ine', 'b@16', 0)) => ('ine', ('ior', 'a', 'b'), 0) */
   /* search710_0_0 -> 11 in the cache */
   /* search710_0_1 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 11, 856 },
      -1,
   } },
   /* search710_1_0 -> 258 in the cache */
   /* search710_1_1 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 258, 856 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 2148, 2149 },
      -1,
   } },

   /* replace710_0_0 -> 11 in the cache */
   /* replace710_0_1 -> 258 in the cache */
   /* replace710_0 -> 2146 in the cache */
   /* replace710_1 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2146, 856 },
      -1,
   } },

   /* ('iand', ('ieq', 'a@16', 0), ('ieq', 'b@16', 0)) => ('ieq', ('umax', 'a', 'b'), 0) */
   /* search711_0_0 -> 11 in the cache */
   /* search711_0_1 -> 856 in the cache */
   /* search711_0 -> 2143 in the cache */
   /* search711_1_0 -> 258 in the cache */
   /* search711_1_1 -> 856 in the cache */
   /* search711_1 -> 2144 in the cache */
   /* search711 -> 2145 in the cache */

   /* replace711_0_0 -> 11 in the cache */
   /* replace711_0_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_umax,
      1, 1,
      { 11, 258 },
      -1,
   } },
   /* replace711_1 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2152, 856 },
      -1,
   } },

   /* ('ior', ('ieq', 'a@16', 0), ('ieq', 'b@16', 0)) => ('ieq', ('umin', 'a', 'b'), 0) */
   /* search712_0_0 -> 11 in the cache */
   /* search712_0_1 -> 856 in the cache */
   /* search712_0 -> 2143 in the cache */
   /* search712_1_0 -> 258 in the cache */
   /* search712_1_1 -> 856 in the cache */
   /* search712_1 -> 2144 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 2143, 2144 },
      -1,
   } },

   /* replace712_0_0 -> 11 in the cache */
   /* replace712_0_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 11, 258 },
      -1,
   } },
   /* replace712_1 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2155, 856 },
      -1,
   } },

   /* ('iand', ('ine', 'a@16', 0), ('ine', 'b@16', 0)) => ('ine', ('umin', 'a', 'b'), 0) */
   /* search713_0_0 -> 11 in the cache */
   /* search713_0_1 -> 856 in the cache */
   /* search713_0 -> 2148 in the cache */
   /* search713_1_0 -> 258 in the cache */
   /* search713_1_1 -> 856 in the cache */
   /* search713_1 -> 2149 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 2148, 2149 },
      -1,
   } },

   /* replace713_0_0 -> 11 in the cache */
   /* replace713_0_1 -> 258 in the cache */
   /* replace713_0 -> 2155 in the cache */
   /* replace713_1 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2155, 856 },
      -1,
   } },

   /* ('ior', ('ine', 'a@16', 0), ('ine', 'b@16', 0)) => ('ine', ('umax', 'a', 'b'), 0) */
   /* search714_0_0 -> 11 in the cache */
   /* search714_0_1 -> 856 in the cache */
   /* search714_0 -> 2148 in the cache */
   /* search714_1_0 -> 258 in the cache */
   /* search714_1_1 -> 856 in the cache */
   /* search714_1 -> 2149 in the cache */
   /* search714 -> 2150 in the cache */

   /* replace714_0_0 -> 11 in the cache */
   /* replace714_0_1 -> 258 in the cache */
   /* replace714_0 -> 2152 in the cache */
   /* replace714_1 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2152, 856 },
      -1,
   } },

   /* ('ineg', ('b2i16', 'a@16')) => a */
   /* search715_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2i16,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 2160 },
      -1,
   } },

   /* replace715 -> 11 in the cache */

   /* ('ishl', 'a@16', ('iand', 15, 'b')) => ('ishl', 'a', 'b') */
   /* search716_0 -> 11 in the cache */
   /* search716_1_0 -> 14 in the cache */
   /* search716_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 14, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 11, 2162 },
      -1,
   } },

   /* replace716_0 -> 11 in the cache */
   /* replace716_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 11, 36 },
      -1,
   } },

   /* ('ishr', 'a@16', ('iand', 15, 'b')) => ('ishr', 'a', 'b') */
   /* search717_0 -> 11 in the cache */
   /* search717_1_0 -> 14 in the cache */
   /* search717_1_1 -> 36 in the cache */
   /* search717_1 -> 2162 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 11, 2162 },
      -1,
   } },

   /* replace717_0 -> 11 in the cache */
   /* replace717_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 11, 36 },
      -1,
   } },

   /* ('ushr', 'a@16', ('iand', 15, 'b')) => ('ushr', 'a', 'b') */
   /* search718_0 -> 11 in the cache */
   /* search718_1_0 -> 14 in the cache */
   /* search718_1_1 -> 36 in the cache */
   /* search718_1 -> 2162 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 11, 2162 },
      -1,
   } },

   /* replace718_0 -> 11 in the cache */
   /* replace718_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 11, 36 },
      -1,
   } },

   /* ('ushr', 'a@16', ('ishl(is_used_once)', ('iand', 'b', 1), 3)) => ('ushr', 'a', ('ishl', 'b', 3)) */
   /* search719_0 -> 11 in the cache */
   /* search719_1_0_0 -> 36 in the cache */
   /* search719_1_0_1 -> 396 in the cache */
   /* search719_1_0 -> 2138 in the cache */
   /* search719_1_1 -> 405 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 2138, 405 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 11, 2169 },
      -1,
   } },

   /* replace719_0 -> 11 in the cache */
   /* replace719_1_0 -> 36 in the cache */
   /* replace719_1_1 -> 405 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 36, 405 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 11, 2171 },
      -1,
   } },

   /* ('iand', ('ieq', 'a@32', 0), ('ieq', 'b@32', 0)) => ('ieq', ('ior', 'a', 'b'), 0) */
   /* search720_0_0 -> 16 in the cache */
   /* search720_0_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 16, 122 },
      -1,
   } },
   /* search720_1_0 -> 36 in the cache */
   /* search720_1_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 36, 122 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 2173, 2174 },
      -1,
   } },

   /* replace720_0_0 -> 16 in the cache */
   /* replace720_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 16, 36 },
      -1,
   } },
   /* replace720_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2176, 122 },
      -1,
   } },

   /* ('ior', ('ine', 'a@32', 0), ('ine', 'b@32', 0)) => ('ine', ('ior', 'a', 'b'), 0) */
   /* search721_0_0 -> 16 in the cache */
   /* search721_0_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 16, 122 },
      -1,
   } },
   /* search721_1_0 -> 36 in the cache */
   /* search721_1_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 36, 122 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 2178, 2179 },
      -1,
   } },

   /* replace721_0_0 -> 16 in the cache */
   /* replace721_0_1 -> 36 in the cache */
   /* replace721_0 -> 2176 in the cache */
   /* replace721_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2176, 122 },
      -1,
   } },

   /* ('iand', ('ieq', 'a@32', 0), ('ieq', 'b@32', 0)) => ('ieq', ('umax', 'a', 'b'), 0) */
   /* search722_0_0 -> 16 in the cache */
   /* search722_0_1 -> 122 in the cache */
   /* search722_0 -> 2173 in the cache */
   /* search722_1_0 -> 36 in the cache */
   /* search722_1_1 -> 122 in the cache */
   /* search722_1 -> 2174 in the cache */
   /* search722 -> 2175 in the cache */

   /* replace722_0_0 -> 16 in the cache */
   /* replace722_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umax,
      1, 1,
      { 16, 36 },
      -1,
   } },
   /* replace722_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2182, 122 },
      -1,
   } },

   /* ('ior', ('ieq', 'a@32', 0), ('ieq', 'b@32', 0)) => ('ieq', ('umin', 'a', 'b'), 0) */
   /* search723_0_0 -> 16 in the cache */
   /* search723_0_1 -> 122 in the cache */
   /* search723_0 -> 2173 in the cache */
   /* search723_1_0 -> 36 in the cache */
   /* search723_1_1 -> 122 in the cache */
   /* search723_1 -> 2174 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 2173, 2174 },
      -1,
   } },

   /* replace723_0_0 -> 16 in the cache */
   /* replace723_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 16, 36 },
      -1,
   } },
   /* replace723_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2185, 122 },
      -1,
   } },

   /* ('iand', ('ine', 'a@32', 0), ('ine', 'b@32', 0)) => ('ine', ('umin', 'a', 'b'), 0) */
   /* search724_0_0 -> 16 in the cache */
   /* search724_0_1 -> 122 in the cache */
   /* search724_0 -> 2178 in the cache */
   /* search724_1_0 -> 36 in the cache */
   /* search724_1_1 -> 122 in the cache */
   /* search724_1 -> 2179 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 2178, 2179 },
      -1,
   } },

   /* replace724_0_0 -> 16 in the cache */
   /* replace724_0_1 -> 36 in the cache */
   /* replace724_0 -> 2185 in the cache */
   /* replace724_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2185, 122 },
      -1,
   } },

   /* ('ior', ('ine', 'a@32', 0), ('ine', 'b@32', 0)) => ('ine', ('umax', 'a', 'b'), 0) */
   /* search725_0_0 -> 16 in the cache */
   /* search725_0_1 -> 122 in the cache */
   /* search725_0 -> 2178 in the cache */
   /* search725_1_0 -> 36 in the cache */
   /* search725_1_1 -> 122 in the cache */
   /* search725_1 -> 2179 in the cache */
   /* search725 -> 2180 in the cache */

   /* replace725_0_0 -> 16 in the cache */
   /* replace725_0_1 -> 36 in the cache */
   /* replace725_0 -> 2182 in the cache */
   /* replace725_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2182, 122 },
      -1,
   } },

   /* ('ineg', ('b2i32', 'a@32')) => a */
   /* search726_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 2190 },
      -1,
   } },

   /* replace726 -> 16 in the cache */

   /* ('ishl', 'a@32', ('iand', 31, 'b')) => ('ishl', 'a', 'b') */
   /* search727_0 -> 16 in the cache */
   /* search727_1_0 -> 19 in the cache */
   /* search727_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 19, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 16, 2192 },
      -1,
   } },

   /* replace727_0 -> 16 in the cache */
   /* replace727_1 -> 36 in the cache */
   /* replace727 -> 946 in the cache */

   /* ('ishr', 'a@32', ('iand', 31, 'b')) => ('ishr', 'a', 'b') */
   /* search728_0 -> 16 in the cache */
   /* search728_1_0 -> 19 in the cache */
   /* search728_1_1 -> 36 in the cache */
   /* search728_1 -> 2192 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 16, 2192 },
      -1,
   } },

   /* replace728_0 -> 16 in the cache */
   /* replace728_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* ('ushr', 'a@32', ('iand', 31, 'b')) => ('ushr', 'a', 'b') */
   /* search729_0 -> 16 in the cache */
   /* search729_1_0 -> 19 in the cache */
   /* search729_1_1 -> 36 in the cache */
   /* search729_1 -> 2192 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 16, 2192 },
      -1,
   } },

   /* replace729_0 -> 16 in the cache */
   /* replace729_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* ('ushr', 'a@32', ('ishl(is_used_once)', ('iand', 'b', 1), 4)) => ('ushr', 'a', ('ishl', 'b', 4)) */
   /* search730_0 -> 16 in the cache */
   /* search730_1_0_0 -> 36 in the cache */
   /* search730_1_0_1 -> 396 in the cache */
   /* search730_1_0 -> 2138 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x4 /* 4 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 2138, 2198 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 16, 2199 },
      -1,
   } },

   /* replace730_0 -> 16 in the cache */
   /* replace730_1_0 -> 36 in the cache */
   /* replace730_1_1 -> 2198 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 36, 2198 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 16, 2201 },
      -1,
   } },

   /* ('iand', ('ieq', 'a@64', 0), ('ieq', 'b@64', 0)) => ('ieq', ('ior', 'a', 'b'), 0) */
   /* search731_0_0 -> 21 in the cache */
   /* search731_0_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 21, 897 },
      -1,
   } },
   /* search731_1_0 -> 41 in the cache */
   /* search731_1_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 41, 897 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 2203, 2204 },
      -1,
   } },

   /* replace731_0_0 -> 21 in the cache */
   /* replace731_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 21, 41 },
      -1,
   } },
   /* replace731_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2206, 897 },
      -1,
   } },

   /* ('ior', ('ine', 'a@64', 0), ('ine', 'b@64', 0)) => ('ine', ('ior', 'a', 'b'), 0) */
   /* search732_0_0 -> 21 in the cache */
   /* search732_0_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 21, 897 },
      -1,
   } },
   /* search732_1_0 -> 41 in the cache */
   /* search732_1_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 41, 897 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 2208, 2209 },
      -1,
   } },

   /* replace732_0_0 -> 21 in the cache */
   /* replace732_0_1 -> 41 in the cache */
   /* replace732_0 -> 2206 in the cache */
   /* replace732_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2206, 897 },
      -1,
   } },

   /* ('iand', ('ieq', 'a@64', 0), ('ieq', 'b@64', 0)) => ('ieq', ('umax', 'a', 'b'), 0) */
   /* search733_0_0 -> 21 in the cache */
   /* search733_0_1 -> 897 in the cache */
   /* search733_0 -> 2203 in the cache */
   /* search733_1_0 -> 41 in the cache */
   /* search733_1_1 -> 897 in the cache */
   /* search733_1 -> 2204 in the cache */
   /* search733 -> 2205 in the cache */

   /* replace733_0_0 -> 21 in the cache */
   /* replace733_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_umax,
      1, 1,
      { 21, 41 },
      -1,
   } },
   /* replace733_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2212, 897 },
      -1,
   } },

   /* ('ior', ('ieq', 'a@64', 0), ('ieq', 'b@64', 0)) => ('ieq', ('umin', 'a', 'b'), 0) */
   /* search734_0_0 -> 21 in the cache */
   /* search734_0_1 -> 897 in the cache */
   /* search734_0 -> 2203 in the cache */
   /* search734_1_0 -> 41 in the cache */
   /* search734_1_1 -> 897 in the cache */
   /* search734_1 -> 2204 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 2203, 2204 },
      -1,
   } },

   /* replace734_0_0 -> 21 in the cache */
   /* replace734_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 21, 41 },
      -1,
   } },
   /* replace734_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2215, 897 },
      -1,
   } },

   /* ('iand', ('ine', 'a@64', 0), ('ine', 'b@64', 0)) => ('ine', ('umin', 'a', 'b'), 0) */
   /* search735_0_0 -> 21 in the cache */
   /* search735_0_1 -> 897 in the cache */
   /* search735_0 -> 2208 in the cache */
   /* search735_1_0 -> 41 in the cache */
   /* search735_1_1 -> 897 in the cache */
   /* search735_1 -> 2209 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 2208, 2209 },
      -1,
   } },

   /* replace735_0_0 -> 21 in the cache */
   /* replace735_0_1 -> 41 in the cache */
   /* replace735_0 -> 2215 in the cache */
   /* replace735_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2215, 897 },
      -1,
   } },

   /* ('ior', ('ine', 'a@64', 0), ('ine', 'b@64', 0)) => ('ine', ('umax', 'a', 'b'), 0) */
   /* search736_0_0 -> 21 in the cache */
   /* search736_0_1 -> 897 in the cache */
   /* search736_0 -> 2208 in the cache */
   /* search736_1_0 -> 41 in the cache */
   /* search736_1_1 -> 897 in the cache */
   /* search736_1 -> 2209 in the cache */
   /* search736 -> 2210 in the cache */

   /* replace736_0_0 -> 21 in the cache */
   /* replace736_0_1 -> 41 in the cache */
   /* replace736_0 -> 2212 in the cache */
   /* replace736_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2212, 897 },
      -1,
   } },

   /* ('ineg', ('b2i64', 'a@64')) => a */
   /* search737_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2i64,
      -1, 0,
      { 21 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 2220 },
      -1,
   } },

   /* replace737 -> 21 in the cache */

   /* ('ishl', 'a@64', ('iand', 63, 'b')) => ('ishl', 'a', 'b') */
   /* search738_0 -> 21 in the cache */
   /* search738_1_0 -> 24 in the cache */
   /* search738_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 24, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 21, 2222 },
      -1,
   } },

   /* replace738_0 -> 21 in the cache */
   /* replace738_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 36 },
      -1,
   } },

   /* ('ishr', 'a@64', ('iand', 63, 'b')) => ('ishr', 'a', 'b') */
   /* search739_0 -> 21 in the cache */
   /* search739_1_0 -> 24 in the cache */
   /* search739_1_1 -> 36 in the cache */
   /* search739_1 -> 2222 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 21, 2222 },
      -1,
   } },

   /* replace739_0 -> 21 in the cache */
   /* replace739_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 21, 36 },
      -1,
   } },

   /* ('ushr', 'a@64', ('iand', 63, 'b')) => ('ushr', 'a', 'b') */
   /* search740_0 -> 21 in the cache */
   /* search740_1_0 -> 24 in the cache */
   /* search740_1_1 -> 36 in the cache */
   /* search740_1 -> 2222 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 21, 2222 },
      -1,
   } },

   /* replace740_0 -> 21 in the cache */
   /* replace740_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 21, 36 },
      -1,
   } },

   /* ('ushr', 'a@64', ('ishl(is_used_once)', ('iand', 'b', 1), 5)) => ('ushr', 'a', ('ishl', 'b', 5)) */
   /* search741_0 -> 21 in the cache */
   /* search741_1_0_0 -> 36 in the cache */
   /* search741_1_0_1 -> 396 in the cache */
   /* search741_1_0 -> 2138 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x5 /* 5 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 2138, 2229 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 21, 2230 },
      -1,
   } },

   /* replace741_0 -> 21 in the cache */
   /* replace741_1_0 -> 36 in the cache */
   /* replace741_1_1 -> 2229 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 36, 2229 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 21, 2232 },
      -1,
   } },

   /* ('ior', ('ieq', 'a', 0), ('ieq', 'a', 1)) => ('uge', 1, 'a') */
   /* search742_0_0 -> 32 in the cache */
   /* search742_0_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 32, 61 },
      -1,
   } },
   /* search742_1_0 -> 32 in the cache */
   /* search742_1_1 -> 35 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 32, 35 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 2234, 2235 },
      -1,
   } },

   /* replace742_0 -> 35 in the cache */
   /* replace742_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 35, 32 },
      -1,
   } },

   /* ('ior', ('uge', 1, 'a'), ('ieq', 'a', 2)) => ('uge', 2, 'a') */
   /* search743_0_0 -> 35 in the cache */
   /* search743_0_1 -> 32 in the cache */
   /* search743_0 -> 2237 in the cache */
   /* search743_1_0 -> 32 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_int, { 0x2 /* 2 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 32, 2238 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 2237, 2239 },
      -1,
   } },

   /* replace743_0 -> 2238 in the cache */
   /* replace743_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 2238, 32 },
      -1,
   } },

   /* ('ior', ('uge', 2, 'a'), ('ieq', 'a', 3)) => ('uge', 3, 'a') */
   /* search744_0_0 -> 2238 in the cache */
   /* search744_0_1 -> 32 in the cache */
   /* search744_0 -> 2241 in the cache */
   /* search744_1_0 -> 32 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_int, { 0x3 /* 3 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 32, 2242 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 2241, 2243 },
      -1,
   } },

   /* replace744_0 -> 2242 in the cache */
   /* replace744_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 2242, 32 },
      -1,
   } },

   /* ('ior', 'a', ('ieq', 'a', False)) => True */
   /* search745_0 -> 760 in the cache */
   /* search745_1_0 -> 760 in the cache */
   /* search745_1_1 -> 1077 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 760, 1077 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 760, 2246 },
      -1,
   } },

   /* replace745 -> 1235 in the cache */

   /* ('ine', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ine', 'a', 'b') */
   /* search746_0_0_0 -> 760 in the cache */
   /* search746_0_0 -> 1228 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 1228 },
      -1,
   } },
   /* search746_1_0_0 -> 1134 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 0,
      { 1134 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 2249 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2248, 2250 },
      -1,
   } },

   /* replace746_0 -> 760 in the cache */
   /* replace746_1 -> 1134 in the cache */
   /* replace746 -> 1391 in the cache */

   /* ('b2i', ('ine', 'a@1', 'b@1')) => ('b2i', ('ixor', 'a', 'b')) */
   /* search747_0_0 -> 760 in the cache */
   /* search747_0_1 -> 1134 in the cache */
   /* search747_0 -> 1391 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 1,
      { 1391 },
      -1,
   } },

   /* replace747_0_0 -> 760 in the cache */
   /* replace747_0_1 -> 1134 in the cache */
   /* replace747_0 -> 1151 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 1,
      { 1151 },
      -1,
   } },

   /* ('ishl', ('b2i32', ('ine', ('iand', 'a@32', '#b(is_pos_power_of_two)'), 0)), '#c') => ('bcsel', ('ige', ('iand', 'c', 31), ('find_lsb', 'b')), ('ishl', ('iand', 'a', 'b'), ('iadd', ('iand', 'c', 31), ('ineg', ('find_lsb', 'b')))), ('ushr', ('iand', 'a', 'b'), ('iadd', ('ineg', ('iand', 'c', 31)), ('find_lsb', 'b')))) */
   /* search748_0_0_0_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      0,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 16, 2254 },
      -1,
   } },
   /* search748_0_0_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2255, 122 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 2,
      { 2256 },
      -1,
   } },
   /* search748_1 -> 309 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 2,
      { 2257, 309 },
      -1,
   } },

   /* replace748_0_0_0 -> 135 in the cache */
   /* replace748_0_0_1 -> 19 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 135, 19 },
      -1,
   } },
   /* replace748_0_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_find_lsb,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 2259, 2260 },
      -1,
   } },
   /* replace748_1_0_0 -> 16 in the cache */
   /* replace748_1_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 16, 36 },
      -1,
   } },
   /* replace748_1_1_0_0 -> 135 in the cache */
   /* replace748_1_1_0_1 -> 19 in the cache */
   /* replace748_1_1_0 -> 885 in the cache */
   /* replace748_1_1_1_0_0 -> 36 in the cache */
   /* replace748_1_1_1_0 -> 2260 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 2260 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 2,
      { 885, 2263 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 3,
      { 2262, 2264 },
      -1,
   } },
   /* replace748_2_0_0 -> 16 in the cache */
   /* replace748_2_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      4, 1,
      { 16, 36 },
      -1,
   } },
   /* replace748_2_1_0_0_0 -> 135 in the cache */
   /* replace748_2_1_0_0_1 -> 19 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      6, 1,
      { 135, 19 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 1,
      { 2267 },
      -1,
   } },
   /* replace748_2_1_1_0 -> 36 in the cache */
   /* replace748_2_1_1 -> 2260 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      5, 2,
      { 2268, 2260 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 3,
      { 2266, 2269 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 7,
      { 2261, 2265, 2270 },
      -1,
   } },

   /* ('b2i32', ('ine', ('iand', 'a@32', '#b(is_pos_power_of_two)'), 0)) => ('ushr', ('iand', 'a', 'b'), ('find_lsb', 'b')) */
   /* search749_0_0_0 -> 16 in the cache */
   /* search749_0_0_1 -> 2254 in the cache */
   /* search749_0_0 -> 2255 in the cache */
   /* search749_0_1 -> 122 in the cache */
   /* search749_0 -> 2256 in the cache */
   /* search749 -> 2257 in the cache */

   /* replace749_0_0 -> 16 in the cache */
   /* replace749_0_1 -> 36 in the cache */
   /* replace749_0 -> 2031 in the cache */
   /* replace749_1_0 -> 36 in the cache */
   /* replace749_1 -> 2260 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 2031, 2260 },
      -1,
   } },

   /* ('ior', ('b2i', 'a'), ('iand', 'b', 1)) => ('iand', ('ior', ('b2i', 'a'), 'b'), 1) */
   /* search750_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 0,
      { 32 },
      -1,
   } },
   /* search750_1_0 -> 3 in the cache */
   /* search750_1_1 -> 79 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 3, 79 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 2273, 2274 },
      -1,
   } },

   /* replace750_0_0_0 -> 32 in the cache */
   /* replace750_0_0 -> 2273 in the cache */
   /* replace750_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 2273, 3 },
      -1,
   } },
   /* replace750_1 -> 79 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 2276, 79 },
      -1,
   } },

   /* ('iand', ('b2i', 'a'), ('iand', 'b', 1)) => ('iand', ('b2i', 'a'), 'b') */
   /* search751_0_0 -> 32 in the cache */
   /* search751_0 -> 2273 in the cache */
   /* search751_1_0 -> 3 in the cache */
   /* search751_1_1 -> 79 in the cache */
   /* search751_1 -> 2274 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 2273, 2274 },
      -1,
   } },

   /* replace751_0_0 -> 32 in the cache */
   /* replace751_0 -> 2273 in the cache */
   /* replace751_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2273, 3 },
      -1,
   } },

   /* ('iand', ('inot', ('iand', ('ior', ('ieq', 'a', 0), 'b'), 'c')), ('ilt', 'a', 0)) => ('iand', ('inot', ('iand', 'b', 'c')), ('ilt', 'a', 0)) */
   /* search752_0_0_0_0_0 -> 32 in the cache */
   /* search752_0_0_0_0_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      3, 1,
      { 32, 61 },
      -1,
   } },
   /* search752_0_0_0_1 -> 1134 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      2, 2,
      { 2280, 1134 },
      -1,
   } },
   /* search752_0_0_1 -> 464 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      1, 3,
      { 2281, 464 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 3,
      { 2282 },
      -1,
   } },
   /* search752_1_0 -> 32 in the cache */
   /* search752_1_1 -> 61 in the cache */
   /* search752_1 -> 1249 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 4,
      { 2283, 1249 },
      -1,
   } },

   /* replace752_0_0_0 -> 1134 in the cache */
   /* replace752_0_0_1 -> 464 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 1134, 464 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 1,
      { 2285 },
      -1,
   } },
   /* replace752_1_0 -> 32 in the cache */
   /* replace752_1_1 -> 61 in the cache */
   /* replace752_1 -> 1249 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 2286, 1249 },
      -1,
   } },

   /* ('iand', ('inot', ('iand', ('ieq', ('umin', 'a', 'b'), 0), 'c')), ('ilt', 'a', 0)) => ('iand', ('inot', ('iand', ('ieq', 'b', 0), 'c')), ('ilt', 'a', 0)) */
   /* search753_0_0_0_0_0 -> 0 in the cache */
   /* search753_0_0_0_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umin,
      3, 1,
      { 0, 3 },
      -1,
   } },
   /* search753_0_0_0_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 2,
      { 2288, 86 },
      -1,
   } },
   /* search753_0_0_1 -> 464 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      1, 3,
      { 2289, 464 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 3,
      { 2290 },
      -1,
   } },
   /* search753_1_0 -> 0 in the cache */
   /* search753_1_1 -> 86 in the cache */
   /* search753_1 -> 90 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 4,
      { 2291, 90 },
      -1,
   } },

   /* replace753_0_0_0_0 -> 3 in the cache */
   /* replace753_0_0_0_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 3, 86 },
      -1,
   } },
   /* replace753_0_0_1 -> 464 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      1, 2,
      { 2293, 464 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 2,
      { 2294 },
      -1,
   } },
   /* replace753_1_0 -> 0 in the cache */
   /* replace753_1_1 -> 86 in the cache */
   /* replace753_1 -> 90 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 2295, 90 },
      -1,
   } },

   /* ('flt', 'a', ('fmax', 'b', 'a')) => ('flt', 'a', 'b') */
   /* search754_0 -> 32 in the cache */
   { .variable = {
      { nir_search_value_variable, -1 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search754_1_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 2297, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 32, 2298 },
      -1,
   } },

   /* replace754_0 -> 32 in the cache */
   /* replace754_1 -> 2297 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 32, 2297 },
      -1,
   } },

   /* ('flt', ('fmin', 'a', 'b'), 'a') => ('flt', 'b', 'a') */
   /* search755_0_0 -> 0 in the cache */
   /* search755_0_1 -> 3 in the cache */
   /* search755_0 -> 1283 in the cache */
   /* search755_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 1283, 0 },
      -1,
   } },

   /* replace755_0 -> 3 in the cache */
   /* replace755_1 -> 0 in the cache */
   /* replace755 -> 1078 in the cache */

   /* ('~fge', 'a', ('fmin', 'b', 'a')) => True */
   /* search756_0 -> 32 in the cache */
   /* search756_1_0 -> 2297 in the cache */
   /* search756_1_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 2297, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 32, 2302 },
      -1,
   } },

   /* replace756 -> 1235 in the cache */

   /* ('~fge', ('fmax', 'a', 'b'), 'a') => True */
   /* search757_0_0 -> 0 in the cache */
   /* search757_0_1 -> 3 in the cache */
   /* search757_0 -> 1281 in the cache */
   /* search757_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 1281, 0 },
      -1,
   } },

   /* replace757 -> 1235 in the cache */

   /* ('flt', 'a', ('fmin', 'b', 'a')) => False */
   /* search758_0 -> 32 in the cache */
   /* search758_1_0 -> 2297 in the cache */
   /* search758_1_1 -> 32 in the cache */
   /* search758_1 -> 2302 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 32, 2302 },
      -1,
   } },

   /* replace758 -> 1077 in the cache */

   /* ('flt', ('fmax', 'a', 'b'), 'a') => False */
   /* search759_0_0 -> 0 in the cache */
   /* search759_0_1 -> 3 in the cache */
   /* search759_0 -> 1281 in the cache */
   /* search759_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 1281, 0 },
      -1,
   } },

   /* replace759 -> 1077 in the cache */

   /* ('~fge', 'a', ('fmax', 'b', 'a')) => ('fge', 'a', 'b') */
   /* search760_0 -> 32 in the cache */
   /* search760_1_0 -> 2297 in the cache */
   /* search760_1_1 -> 32 in the cache */
   /* search760_1 -> 2298 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 32, 2298 },
      -1,
   } },

   /* replace760_0 -> 32 in the cache */
   /* replace760_1 -> 2297 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 32, 2297 },
      -1,
   } },

   /* ('~fge', ('fmin', 'a', 'b'), 'a') => ('fge', 'b', 'a') */
   /* search761_0_0 -> 0 in the cache */
   /* search761_0_1 -> 3 in the cache */
   /* search761_0 -> 1283 in the cache */
   /* search761_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 1283, 0 },
      -1,
   } },

   /* replace761_0 -> 3 in the cache */
   /* replace761_1 -> 0 in the cache */
   /* replace761 -> 1090 in the cache */

   /* ('ilt', 'a', ('imax', 'b', 'a')) => ('ilt', 'a', 'b') */
   /* search762_0 -> 32 in the cache */
   /* search762_1_0 -> 2297 in the cache */
   /* search762_1_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 2297, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 32, 2310 },
      -1,
   } },

   /* replace762_0 -> 32 in the cache */
   /* replace762_1 -> 2297 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 32, 2297 },
      -1,
   } },

   /* ('ilt', ('imin', 'a', 'b'), 'a') => ('ilt', 'b', 'a') */
   /* search763_0_0 -> 0 in the cache */
   /* search763_0_1 -> 3 in the cache */
   /* search763_0 -> 1303 in the cache */
   /* search763_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 1303, 0 },
      -1,
   } },

   /* replace763_0 -> 3 in the cache */
   /* replace763_1 -> 0 in the cache */
   /* replace763 -> 1083 in the cache */

   /* ('ige', 'a', ('imin', 'b', 'a')) => True */
   /* search764_0 -> 32 in the cache */
   /* search764_1_0 -> 2297 in the cache */
   /* search764_1_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 2297, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 32, 2314 },
      -1,
   } },

   /* replace764 -> 1235 in the cache */

   /* ('ige', ('imax', 'a', 'b'), 'a') => True */
   /* search765_0_0 -> 0 in the cache */
   /* search765_0_1 -> 3 in the cache */
   /* search765_0 -> 1301 in the cache */
   /* search765_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 1301, 0 },
      -1,
   } },

   /* replace765 -> 1235 in the cache */

   /* ('ult', 'a', ('umax', 'b', 'a')) => ('ult', 'a', 'b') */
   /* search766_0 -> 32 in the cache */
   /* search766_1_0 -> 2297 in the cache */
   /* search766_1_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 2297, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 32, 2317 },
      -1,
   } },

   /* replace766_0 -> 32 in the cache */
   /* replace766_1 -> 2297 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 32, 2297 },
      -1,
   } },

   /* ('ult', ('umin', 'a', 'b'), 'a') => ('ult', 'b', 'a') */
   /* search767_0_0 -> 0 in the cache */
   /* search767_0_1 -> 3 in the cache */
   /* search767_0 -> 1293 in the cache */
   /* search767_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 1293, 0 },
      -1,
   } },

   /* replace767_0 -> 3 in the cache */
   /* replace767_1 -> 0 in the cache */
   /* replace767 -> 1085 in the cache */

   /* ('uge', 'a', ('umin', 'b', 'a')) => True */
   /* search768_0 -> 32 in the cache */
   /* search768_1_0 -> 2297 in the cache */
   /* search768_1_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 2297, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 32, 2321 },
      -1,
   } },

   /* replace768 -> 1235 in the cache */

   /* ('uge', ('umax', 'a', 'b'), 'a') => True */
   /* search769_0_0 -> 0 in the cache */
   /* search769_0_1 -> 3 in the cache */
   /* search769_0 -> 1291 in the cache */
   /* search769_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 1291, 0 },
      -1,
   } },

   /* replace769 -> 1235 in the cache */

   /* ('ilt', 'a', ('imin', 'b', 'a')) => False */
   /* search770_0 -> 32 in the cache */
   /* search770_1_0 -> 2297 in the cache */
   /* search770_1_1 -> 32 in the cache */
   /* search770_1 -> 2314 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 32, 2314 },
      -1,
   } },

   /* replace770 -> 1077 in the cache */

   /* ('ilt', ('imax', 'a', 'b'), 'a') => False */
   /* search771_0_0 -> 0 in the cache */
   /* search771_0_1 -> 3 in the cache */
   /* search771_0 -> 1301 in the cache */
   /* search771_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 1301, 0 },
      -1,
   } },

   /* replace771 -> 1077 in the cache */

   /* ('ige', 'a', ('imax', 'b', 'a')) => ('ige', 'a', 'b') */
   /* search772_0 -> 32 in the cache */
   /* search772_1_0 -> 2297 in the cache */
   /* search772_1_1 -> 32 in the cache */
   /* search772_1 -> 2310 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 32, 2310 },
      -1,
   } },

   /* replace772_0 -> 32 in the cache */
   /* replace772_1 -> 2297 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 32, 2297 },
      -1,
   } },

   /* ('ige', ('imin', 'a', 'b'), 'a') => ('ige', 'b', 'a') */
   /* search773_0_0 -> 0 in the cache */
   /* search773_0_1 -> 3 in the cache */
   /* search773_0 -> 1303 in the cache */
   /* search773_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 1303, 0 },
      -1,
   } },

   /* replace773_0 -> 3 in the cache */
   /* replace773_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3, 0 },
      -1,
   } },

   /* ('ult', 'a', ('umin', 'b', 'a')) => False */
   /* search774_0 -> 32 in the cache */
   /* search774_1_0 -> 2297 in the cache */
   /* search774_1_1 -> 32 in the cache */
   /* search774_1 -> 2321 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 32, 2321 },
      -1,
   } },

   /* replace774 -> 1077 in the cache */

   /* ('ult', ('umax', 'a', 'b'), 'a') => False */
   /* search775_0_0 -> 0 in the cache */
   /* search775_0_1 -> 3 in the cache */
   /* search775_0 -> 1291 in the cache */
   /* search775_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 1291, 0 },
      -1,
   } },

   /* replace775 -> 1077 in the cache */

   /* ('uge', 'a', ('umax', 'b', 'a')) => ('uge', 'a', 'b') */
   /* search776_0 -> 32 in the cache */
   /* search776_1_0 -> 2297 in the cache */
   /* search776_1_1 -> 32 in the cache */
   /* search776_1 -> 2317 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 32, 2317 },
      -1,
   } },

   /* replace776_0 -> 32 in the cache */
   /* replace776_1 -> 2297 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 32, 2297 },
      -1,
   } },

   /* ('uge', ('umin', 'a', 'b'), 'a') => ('uge', 'b', 'a') */
   /* search777_0_0 -> 0 in the cache */
   /* search777_0_1 -> 3 in the cache */
   /* search777_0 -> 1293 in the cache */
   /* search777_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 1293, 0 },
      -1,
   } },

   /* replace777_0 -> 3 in the cache */
   /* replace777_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 3, 0 },
      -1,
   } },

   /* ('ult', 'a', ('iand', 'b', 'a')) => False */
   /* search778_0 -> 32 in the cache */
   /* search778_1_0 -> 2297 in the cache */
   /* search778_1_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2297, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 32, 2336 },
      -1,
   } },

   /* replace778 -> 1077 in the cache */

   /* ('ult', ('ior', 'a', 'b'), 'a') => False */
   /* search779_0_0 -> 0 in the cache */
   /* search779_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 0, 3 },
      -1,
   } },
   /* search779_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 2338, 0 },
      -1,
   } },

   /* replace779 -> 1077 in the cache */

   /* ('uge', 'a', ('iand', 'b', 'a')) => True */
   /* search780_0 -> 32 in the cache */
   /* search780_1_0 -> 2297 in the cache */
   /* search780_1_1 -> 32 in the cache */
   /* search780_1 -> 2336 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 32, 2336 },
      -1,
   } },

   /* replace780 -> 1235 in the cache */

   /* ('uge', ('ior', 'a', 'b'), 'a') => True */
   /* search781_0_0 -> 0 in the cache */
   /* search781_0_1 -> 3 in the cache */
   /* search781_0 -> 2338 in the cache */
   /* search781_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 2338, 0 },
      -1,
   } },

   /* replace781 -> 1235 in the cache */

   /* ('ilt', '#a', ('imax', '#b', 'c')) => ('ior', ('ilt', 'a', 'b'), ('ilt', 'a', 'c')) */
   /* search782_0 -> 282 in the cache */
   /* search782_1_0 -> 283 in the cache */
   /* search782_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 283, 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 282, 2342 },
      -1,
   } },

   /* replace782_0_0 -> 126 in the cache */
   /* replace782_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 126, 127 },
      -1,
   } },
   /* replace782_1_0 -> 126 in the cache */
   /* replace782_1_1 -> 129 in the cache */
   /* replace782_1 -> 1657 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2344, 1657 },
      -1,
   } },

   /* ('ilt', ('imin', '#a', 'b'), '#c') => ('ior', ('ilt', 'a', 'c'), ('ilt', 'b', 'c')) */
   /* search783_0_0 -> 282 in the cache */
   /* search783_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 282, 127 },
      -1,
   } },
   /* search783_1 -> 814 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 2346, 814 },
      -1,
   } },

   /* replace783_0_0 -> 126 in the cache */
   /* replace783_0_1 -> 129 in the cache */
   /* replace783_0 -> 1657 in the cache */
   /* replace783_1_0 -> 127 in the cache */
   /* replace783_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 127, 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1657, 2348 },
      -1,
   } },

   /* ('ige', '#a', ('imin', '#b', 'c')) => ('ior', ('ige', 'a', 'b'), ('ige', 'a', 'c')) */
   /* search784_0 -> 282 in the cache */
   /* search784_1_0 -> 283 in the cache */
   /* search784_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 283, 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 282, 2350 },
      -1,
   } },

   /* replace784_0_0 -> 126 in the cache */
   /* replace784_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 126, 127 },
      -1,
   } },
   /* replace784_1_0 -> 126 in the cache */
   /* replace784_1_1 -> 129 in the cache */
   /* replace784_1 -> 1667 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2352, 1667 },
      -1,
   } },

   /* ('ige', ('imax', '#a', 'b'), '#c') => ('ior', ('ige', 'a', 'c'), ('ige', 'b', 'c')) */
   /* search785_0_0 -> 282 in the cache */
   /* search785_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 282, 127 },
      -1,
   } },
   /* search785_1 -> 814 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 2354, 814 },
      -1,
   } },

   /* replace785_0_0 -> 126 in the cache */
   /* replace785_0_1 -> 129 in the cache */
   /* replace785_0 -> 1667 in the cache */
   /* replace785_1_0 -> 127 in the cache */
   /* replace785_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 127, 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1667, 2356 },
      -1,
   } },

   /* ('ult', '#a', ('umax', '#b', 'c')) => ('ior', ('ult', 'a', 'b'), ('ult', 'a', 'c')) */
   /* search786_0 -> 282 in the cache */
   /* search786_1_0 -> 283 in the cache */
   /* search786_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 283, 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 282, 2358 },
      -1,
   } },

   /* replace786_0_0 -> 126 in the cache */
   /* replace786_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 126, 127 },
      -1,
   } },
   /* replace786_1_0 -> 126 in the cache */
   /* replace786_1_1 -> 129 in the cache */
   /* replace786_1 -> 1677 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2360, 1677 },
      -1,
   } },

   /* ('ult', ('umin', '#a', 'b'), '#c') => ('ior', ('ult', 'a', 'c'), ('ult', 'b', 'c')) */
   /* search787_0_0 -> 282 in the cache */
   /* search787_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 282, 127 },
      -1,
   } },
   /* search787_1 -> 814 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 2362, 814 },
      -1,
   } },

   /* replace787_0_0 -> 126 in the cache */
   /* replace787_0_1 -> 129 in the cache */
   /* replace787_0 -> 1677 in the cache */
   /* replace787_1_0 -> 127 in the cache */
   /* replace787_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 127, 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1677, 2364 },
      -1,
   } },

   /* ('uge', '#a', ('umin', '#b', 'c')) => ('ior', ('uge', 'a', 'b'), ('uge', 'a', 'c')) */
   /* search788_0 -> 282 in the cache */
   /* search788_1_0 -> 283 in the cache */
   /* search788_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 283, 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 282, 2366 },
      -1,
   } },

   /* replace788_0_0 -> 126 in the cache */
   /* replace788_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 126, 127 },
      -1,
   } },
   /* replace788_1_0 -> 126 in the cache */
   /* replace788_1_1 -> 129 in the cache */
   /* replace788_1 -> 1687 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2368, 1687 },
      -1,
   } },

   /* ('uge', ('umax', '#a', 'b'), '#c') => ('ior', ('uge', 'a', 'c'), ('uge', 'b', 'c')) */
   /* search789_0_0 -> 282 in the cache */
   /* search789_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 282, 127 },
      -1,
   } },
   /* search789_1 -> 814 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 2370, 814 },
      -1,
   } },

   /* replace789_0_0 -> 126 in the cache */
   /* replace789_0_1 -> 129 in the cache */
   /* replace789_0 -> 1687 in the cache */
   /* replace789_1_0 -> 127 in the cache */
   /* replace789_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 127, 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1687, 2372 },
      -1,
   } },

   /* ('ilt', '#a', ('imin', '#b', 'c')) => ('iand', ('ilt', 'a', 'b'), ('ilt', 'a', 'c')) */
   /* search790_0 -> 282 in the cache */
   /* search790_1_0 -> 283 in the cache */
   /* search790_1_1 -> 129 in the cache */
   /* search790_1 -> 2350 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 282, 2350 },
      -1,
   } },

   /* replace790_0_0 -> 126 in the cache */
   /* replace790_0_1 -> 127 in the cache */
   /* replace790_0 -> 2344 in the cache */
   /* replace790_1_0 -> 126 in the cache */
   /* replace790_1_1 -> 129 in the cache */
   /* replace790_1 -> 1657 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2344, 1657 },
      -1,
   } },

   /* ('ilt', ('imax', '#a', 'b'), '#c') => ('iand', ('ilt', 'a', 'c'), ('ilt', 'b', 'c')) */
   /* search791_0_0 -> 282 in the cache */
   /* search791_0_1 -> 127 in the cache */
   /* search791_0 -> 2354 in the cache */
   /* search791_1 -> 814 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 2354, 814 },
      -1,
   } },

   /* replace791_0_0 -> 126 in the cache */
   /* replace791_0_1 -> 129 in the cache */
   /* replace791_0 -> 1657 in the cache */
   /* replace791_1_0 -> 127 in the cache */
   /* replace791_1_1 -> 129 in the cache */
   /* replace791_1 -> 2348 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1657, 2348 },
      -1,
   } },

   /* ('ige', '#a', ('imax', '#b', 'c')) => ('iand', ('ige', 'a', 'b'), ('ige', 'a', 'c')) */
   /* search792_0 -> 282 in the cache */
   /* search792_1_0 -> 283 in the cache */
   /* search792_1_1 -> 129 in the cache */
   /* search792_1 -> 2342 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 282, 2342 },
      -1,
   } },

   /* replace792_0_0 -> 126 in the cache */
   /* replace792_0_1 -> 127 in the cache */
   /* replace792_0 -> 2352 in the cache */
   /* replace792_1_0 -> 126 in the cache */
   /* replace792_1_1 -> 129 in the cache */
   /* replace792_1 -> 1667 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2352, 1667 },
      -1,
   } },

   /* ('ige', ('imin', '#a', 'b'), '#c') => ('iand', ('ige', 'a', 'c'), ('ige', 'b', 'c')) */
   /* search793_0_0 -> 282 in the cache */
   /* search793_0_1 -> 127 in the cache */
   /* search793_0 -> 2346 in the cache */
   /* search793_1 -> 814 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 2346, 814 },
      -1,
   } },

   /* replace793_0_0 -> 126 in the cache */
   /* replace793_0_1 -> 129 in the cache */
   /* replace793_0 -> 1667 in the cache */
   /* replace793_1_0 -> 127 in the cache */
   /* replace793_1_1 -> 129 in the cache */
   /* replace793_1 -> 2356 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1667, 2356 },
      -1,
   } },

   /* ('ult', '#a', ('umin', '#b', 'c')) => ('iand', ('ult', 'a', 'b'), ('ult', 'a', 'c')) */
   /* search794_0 -> 282 in the cache */
   /* search794_1_0 -> 283 in the cache */
   /* search794_1_1 -> 129 in the cache */
   /* search794_1 -> 2366 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 282, 2366 },
      -1,
   } },

   /* replace794_0_0 -> 126 in the cache */
   /* replace794_0_1 -> 127 in the cache */
   /* replace794_0 -> 2360 in the cache */
   /* replace794_1_0 -> 126 in the cache */
   /* replace794_1_1 -> 129 in the cache */
   /* replace794_1 -> 1677 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2360, 1677 },
      -1,
   } },

   /* ('ult', ('umax', '#a', 'b'), '#c') => ('iand', ('ult', 'a', 'c'), ('ult', 'b', 'c')) */
   /* search795_0_0 -> 282 in the cache */
   /* search795_0_1 -> 127 in the cache */
   /* search795_0 -> 2370 in the cache */
   /* search795_1 -> 814 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 2370, 814 },
      -1,
   } },

   /* replace795_0_0 -> 126 in the cache */
   /* replace795_0_1 -> 129 in the cache */
   /* replace795_0 -> 1677 in the cache */
   /* replace795_1_0 -> 127 in the cache */
   /* replace795_1_1 -> 129 in the cache */
   /* replace795_1 -> 2364 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1677, 2364 },
      -1,
   } },

   /* ('uge', '#a', ('umax', '#b', 'c')) => ('iand', ('uge', 'a', 'b'), ('uge', 'a', 'c')) */
   /* search796_0 -> 282 in the cache */
   /* search796_1_0 -> 283 in the cache */
   /* search796_1_1 -> 129 in the cache */
   /* search796_1 -> 2358 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 282, 2358 },
      -1,
   } },

   /* replace796_0_0 -> 126 in the cache */
   /* replace796_0_1 -> 127 in the cache */
   /* replace796_0 -> 2368 in the cache */
   /* replace796_1_0 -> 126 in the cache */
   /* replace796_1_1 -> 129 in the cache */
   /* replace796_1 -> 1687 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2368, 1687 },
      -1,
   } },

   /* ('uge', ('umin', '#a', 'b'), '#c') => ('iand', ('uge', 'a', 'c'), ('uge', 'b', 'c')) */
   /* search797_0_0 -> 282 in the cache */
   /* search797_0_1 -> 127 in the cache */
   /* search797_0 -> 2362 in the cache */
   /* search797_1 -> 814 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 2362, 814 },
      -1,
   } },

   /* replace797_0_0 -> 126 in the cache */
   /* replace797_0_1 -> 129 in the cache */
   /* replace797_0 -> 1687 in the cache */
   /* replace797_1_0 -> 127 in the cache */
   /* replace797_1_1 -> 129 in the cache */
   /* replace797_1 -> 2372 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1687, 2372 },
      -1,
   } },

   /* ('bcsel', ('ilt', 'a', 0), ('ineg', ('ishr', 'a', 'b')), ('ishr', 'a', 'b')) => ('iabs', ('ishr', 'a', 'b')) */
   /* search798_0_0 -> 32 in the cache */
   /* search798_0_1 -> 61 in the cache */
   /* search798_0 -> 1249 in the cache */
   /* search798_1_0_0 -> 32 in the cache */
   /* search798_1_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 32, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 2390 },
      -1,
   } },
   /* search798_2_0 -> 32 in the cache */
   /* search798_2_1 -> 36 in the cache */
   /* search798_2 -> 2390 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1249, 2391, 2390 },
      -1,
   } },

   /* replace798_0_0 -> 32 in the cache */
   /* replace798_0_1 -> 36 in the cache */
   /* replace798_0 -> 2390 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 2390 },
      -1,
   } },

   /* ('iabs', ('ishr', ('iabs', 'a'), 'b')) => ('ishr', ('iabs', 'a'), 'b') */
   /* search799_0_0_0 -> 32 in the cache */
   /* search799_0_0 -> 106 in the cache */
   /* search799_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 106, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 2394 },
      -1,
   } },

   /* replace799_0_0 -> 32 in the cache */
   /* replace799_0 -> 106 in the cache */
   /* replace799_1 -> 36 in the cache */
   /* replace799 -> 2394 in the cache */

   /* ('fabs', ('slt', 'a', 'b')) => ('slt', 'a', 'b') */
   /* search800_0_0 -> 0 in the cache */
   /* search800_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_slt,
      -1, 0,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 2396 },
      -1,
   } },

   /* replace800_0 -> 0 in the cache */
   /* replace800_1 -> 3 in the cache */
   /* replace800 -> 2396 in the cache */

   /* ('fabs', ('sge', 'a', 'b')) => ('sge', 'a', 'b') */
   /* search801_0_0 -> 0 in the cache */
   /* search801_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_sge,
      -1, 0,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 2398 },
      -1,
   } },

   /* replace801_0 -> 0 in the cache */
   /* replace801_1 -> 3 in the cache */
   /* replace801 -> 2398 in the cache */

   /* ('fabs', ('seq', 'a', 'b')) => ('seq', 'a', 'b') */
   /* search802_0_0 -> 0 in the cache */
   /* search802_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_seq,
      0, 1,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 1,
      { 2400 },
      -1,
   } },

   /* replace802_0 -> 0 in the cache */
   /* replace802_1 -> 3 in the cache */
   /* replace802 -> 2400 in the cache */

   /* ('fabs', ('sne', 'a', 'b')) => ('sne', 'a', 'b') */
   /* search803_0_0 -> 0 in the cache */
   /* search803_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_sne,
      0, 1,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 1,
      { 2402 },
      -1,
   } },

   /* replace803_0 -> 0 in the cache */
   /* replace803_1 -> 3 in the cache */
   /* replace803 -> 2402 in the cache */

   /* ('slt', 'a', 'b') => ('b2f', ('flt', 'a', 'b')) */
   /* search804_0 -> 0 in the cache */
   /* search804_1 -> 3 in the cache */
   /* search804 -> 2396 in the cache */

   /* replace804_0_0 -> 0 in the cache */
   /* replace804_0_1 -> 3 in the cache */
   /* replace804_0 -> 1049 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 1049 },
      -1,
   } },

   /* ('sge', 'a', 'b') => ('b2f', ('fge', 'a', 'b')) */
   /* search805_0 -> 0 in the cache */
   /* search805_1 -> 3 in the cache */
   /* search805 -> 2398 in the cache */

   /* replace805_0_0 -> 0 in the cache */
   /* replace805_0_1 -> 3 in the cache */
   /* replace805_0 -> 1046 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 1046 },
      -1,
   } },

   /* ('seq', 'a', 'b') => ('b2f', ('feq', 'a', 'b')) */
   /* search806_0 -> 0 in the cache */
   /* search806_1 -> 3 in the cache */
   /* search806 -> 2400 in the cache */

   /* replace806_0_0 -> 0 in the cache */
   /* replace806_0_1 -> 3 in the cache */
   /* replace806_0 -> 1055 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 1,
      { 1055 },
      -1,
   } },

   /* ('sne', 'a', 'b') => ('b2f', ('fneu', 'a', 'b')) */
   /* search807_0 -> 0 in the cache */
   /* search807_1 -> 3 in the cache */
   /* search807 -> 2402 in the cache */

   /* replace807_0_0 -> 0 in the cache */
   /* replace807_0_1 -> 3 in the cache */
   /* replace807_0 -> 1052 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 1,
      { 1052 },
      -1,
   } },

   /* ('seq', ('seq', 'a', 'b'), 1.0) => ('seq', 'a', 'b') */
   /* search808_0_0 -> 0 in the cache */
   /* search808_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_seq,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* search808_1 -> 270 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_seq,
      0, 2,
      { 2408, 270 },
      -1,
   } },

   /* replace808_0 -> 0 in the cache */
   /* replace808_1 -> 3 in the cache */
   /* replace808 -> 2400 in the cache */

   /* ('seq', ('sne', 'a', 'b'), 1.0) => ('sne', 'a', 'b') */
   /* search809_0_0 -> 0 in the cache */
   /* search809_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_sne,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* search809_1 -> 270 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_seq,
      0, 2,
      { 2410, 270 },
      -1,
   } },

   /* replace809_0 -> 0 in the cache */
   /* replace809_1 -> 3 in the cache */
   /* replace809 -> 2402 in the cache */

   /* ('seq', ('slt', 'a', 'b'), 1.0) => ('slt', 'a', 'b') */
   /* search810_0_0 -> 0 in the cache */
   /* search810_0_1 -> 3 in the cache */
   /* search810_0 -> 2396 in the cache */
   /* search810_1 -> 270 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_seq,
      0, 1,
      { 2396, 270 },
      -1,
   } },

   /* replace810_0 -> 0 in the cache */
   /* replace810_1 -> 3 in the cache */
   /* replace810 -> 2396 in the cache */

   /* ('seq', ('sge', 'a', 'b'), 1.0) => ('sge', 'a', 'b') */
   /* search811_0_0 -> 0 in the cache */
   /* search811_0_1 -> 3 in the cache */
   /* search811_0 -> 2398 in the cache */
   /* search811_1 -> 270 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_seq,
      0, 1,
      { 2398, 270 },
      -1,
   } },

   /* replace811_0 -> 0 in the cache */
   /* replace811_1 -> 3 in the cache */
   /* replace811 -> 2398 in the cache */

   /* ('sne', ('seq', 'a', 'b'), 0.0) => ('seq', 'a', 'b') */
   /* search812_0_0 -> 0 in the cache */
   /* search812_0_1 -> 3 in the cache */
   /* search812_0 -> 2408 in the cache */
   /* search812_1 -> 256 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_sne,
      0, 2,
      { 2408, 256 },
      -1,
   } },

   /* replace812_0 -> 0 in the cache */
   /* replace812_1 -> 3 in the cache */
   /* replace812 -> 2400 in the cache */

   /* ('sne', ('sne', 'a', 'b'), 0.0) => ('sne', 'a', 'b') */
   /* search813_0_0 -> 0 in the cache */
   /* search813_0_1 -> 3 in the cache */
   /* search813_0 -> 2410 in the cache */
   /* search813_1 -> 256 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_sne,
      0, 2,
      { 2410, 256 },
      -1,
   } },

   /* replace813_0 -> 0 in the cache */
   /* replace813_1 -> 3 in the cache */
   /* replace813 -> 2402 in the cache */

   /* ('sne', ('slt', 'a', 'b'), 0.0) => ('slt', 'a', 'b') */
   /* search814_0_0 -> 0 in the cache */
   /* search814_0_1 -> 3 in the cache */
   /* search814_0 -> 2396 in the cache */
   /* search814_1 -> 256 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_sne,
      0, 1,
      { 2396, 256 },
      -1,
   } },

   /* replace814_0 -> 0 in the cache */
   /* replace814_1 -> 3 in the cache */
   /* replace814 -> 2396 in the cache */

   /* ('sne', ('sge', 'a', 'b'), 0.0) => ('sge', 'a', 'b') */
   /* search815_0_0 -> 0 in the cache */
   /* search815_0_1 -> 3 in the cache */
   /* search815_0 -> 2398 in the cache */
   /* search815_1 -> 256 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_sne,
      0, 1,
      { 2398, 256 },
      -1,
   } },

   /* replace815_0 -> 0 in the cache */
   /* replace815_1 -> 3 in the cache */
   /* replace815 -> 2398 in the cache */

   /* ('seq', ('seq', 'a', 'b'), 0.0) => ('sne', 'a', 'b') */
   /* search816_0_0 -> 0 in the cache */
   /* search816_0_1 -> 3 in the cache */
   /* search816_0 -> 2408 in the cache */
   /* search816_1 -> 256 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_seq,
      0, 2,
      { 2408, 256 },
      -1,
   } },

   /* replace816_0 -> 0 in the cache */
   /* replace816_1 -> 3 in the cache */
   /* replace816 -> 2402 in the cache */

   /* ('seq', ('sne', 'a', 'b'), 0.0) => ('seq', 'a', 'b') */
   /* search817_0_0 -> 0 in the cache */
   /* search817_0_1 -> 3 in the cache */
   /* search817_0 -> 2410 in the cache */
   /* search817_1 -> 256 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_seq,
      0, 2,
      { 2410, 256 },
      -1,
   } },

   /* replace817_0 -> 0 in the cache */
   /* replace817_1 -> 3 in the cache */
   /* replace817 -> 2400 in the cache */

   /* ('seq', ('slt', 'a', 'b'), 0.0) => ('sge', 'a', 'b') */
   /* search818_0_0 -> 0 in the cache */
   /* search818_0_1 -> 3 in the cache */
   /* search818_0 -> 2396 in the cache */
   /* search818_1 -> 256 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_seq,
      0, 1,
      { 2396, 256 },
      -1,
   } },

   /* replace818_0 -> 0 in the cache */
   /* replace818_1 -> 3 in the cache */
   /* replace818 -> 2398 in the cache */

   /* ('seq', ('sge', 'a', 'b'), 0.0) => ('slt', 'a', 'b') */
   /* search819_0_0 -> 0 in the cache */
   /* search819_0_1 -> 3 in the cache */
   /* search819_0 -> 2398 in the cache */
   /* search819_1 -> 256 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_seq,
      0, 1,
      { 2398, 256 },
      -1,
   } },

   /* replace819_0 -> 0 in the cache */
   /* replace819_1 -> 3 in the cache */
   /* replace819 -> 2396 in the cache */

   /* ('sne', ('seq', 'a', 'b'), 1.0) => ('sne', 'a', 'b') */
   /* search820_0_0 -> 0 in the cache */
   /* search820_0_1 -> 3 in the cache */
   /* search820_0 -> 2408 in the cache */
   /* search820_1 -> 270 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_sne,
      0, 2,
      { 2408, 270 },
      -1,
   } },

   /* replace820_0 -> 0 in the cache */
   /* replace820_1 -> 3 in the cache */
   /* replace820 -> 2402 in the cache */

   /* ('sne', ('sne', 'a', 'b'), 1.0) => ('seq', 'a', 'b') */
   /* search821_0_0 -> 0 in the cache */
   /* search821_0_1 -> 3 in the cache */
   /* search821_0 -> 2410 in the cache */
   /* search821_1 -> 270 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_sne,
      0, 2,
      { 2410, 270 },
      -1,
   } },

   /* replace821_0 -> 0 in the cache */
   /* replace821_1 -> 3 in the cache */
   /* replace821 -> 2400 in the cache */

   /* ('sne', ('slt', 'a', 'b'), 1.0) => ('sge', 'a', 'b') */
   /* search822_0_0 -> 0 in the cache */
   /* search822_0_1 -> 3 in the cache */
   /* search822_0 -> 2396 in the cache */
   /* search822_1 -> 270 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_sne,
      0, 1,
      { 2396, 270 },
      -1,
   } },

   /* replace822_0 -> 0 in the cache */
   /* replace822_1 -> 3 in the cache */
   /* replace822 -> 2398 in the cache */

   /* ('sne', ('sge', 'a', 'b'), 1.0) => ('slt', 'a', 'b') */
   /* search823_0_0 -> 0 in the cache */
   /* search823_0_1 -> 3 in the cache */
   /* search823_0 -> 2398 in the cache */
   /* search823_1 -> 270 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_sne,
      0, 1,
      { 2398, 270 },
      -1,
   } },

   /* replace823_0 -> 0 in the cache */
   /* replace823_1 -> 3 in the cache */
   /* replace823 -> 2396 in the cache */

   /* ('fall_equal2', 'a', 'b') => ('fmin', ('seq', 'a.x', 'b.x'), ('seq', 'a.y', 'b.y')) */
   /* search824_0 -> 16 in the cache */
   /* search824_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fall_equal2,
      0, 1,
      { 16, 36 },
      -1,
   } },

   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_seq,
      0, 1,
      { 2427, 2428 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {1},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {1},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_seq,
      1, 1,
      { 2430, 2431 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmin,
      -1, 2,
      { 2429, 2432 },
      -1,
   } },

   /* ('fall_equal3', 'a', 'b') => ('seq', ('fany_nequal3', 'a', 'b'), 0.0) */
   /* search825_0 -> 16 in the cache */
   /* search825_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fall_equal3,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace825_0_0 -> 16 in the cache */
   /* replace825_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fany_nequal3,
      1, 1,
      { 16, 36 },
      -1,
   } },
   /* replace825_1 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_seq,
      0, 2,
      { 2435, 115 },
      -1,
   } },

   /* ('fall_equal4', 'a', 'b') => ('seq', ('fany_nequal4', 'a', 'b'), 0.0) */
   /* search826_0 -> 16 in the cache */
   /* search826_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fall_equal4,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace826_0_0 -> 16 in the cache */
   /* replace826_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fany_nequal4,
      1, 1,
      { 16, 36 },
      -1,
   } },
   /* replace826_1 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_seq,
      0, 2,
      { 2438, 115 },
      -1,
   } },

   /* ('fall_equal8', 'a', 'b') => ('seq', ('fany_nequal8', 'a', 'b'), 0.0) */
   /* search827_0 -> 16 in the cache */
   /* search827_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fall_equal8,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace827_0_0 -> 16 in the cache */
   /* replace827_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fany_nequal8,
      1, 1,
      { 16, 36 },
      -1,
   } },
   /* replace827_1 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_seq,
      0, 2,
      { 2441, 115 },
      -1,
   } },

   /* ('fall_equal16', 'a', 'b') => ('seq', ('fany_nequal16', 'a', 'b'), 0.0) */
   /* search828_0 -> 16 in the cache */
   /* search828_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fall_equal16,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace828_0_0 -> 16 in the cache */
   /* replace828_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fany_nequal16,
      1, 1,
      { 16, 36 },
      -1,
   } },
   /* replace828_1 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_seq,
      0, 2,
      { 2444, 115 },
      -1,
   } },

   /* ('fany_nequal2', 'a', 'b') => ('fmax', ('sne', 'a.x', 'b.x'), ('sne', 'a.y', 'b.y')) */
   /* search829_0 -> 16 in the cache */
   /* search829_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fany_nequal2,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace829_0_0 -> 2427 in the cache */
   /* replace829_0_1 -> 2428 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sne,
      0, 1,
      { 2427, 2428 },
      -1,
   } },
   /* replace829_1_0 -> 2430 in the cache */
   /* replace829_1_1 -> 2431 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sne,
      1, 1,
      { 2430, 2431 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmax,
      -1, 2,
      { 2447, 2448 },
      -1,
   } },

   /* ('fany_nequal3', 'a', 'b') => ('fsat', ('fdot3', ('sne', 'a', 'b'), ('sne', 'a', 'b'))) */
   /* search830_0 -> 16 in the cache */
   /* search830_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fany_nequal3,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace830_0_0_0 -> 16 in the cache */
   /* replace830_0_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sne,
      0, 1,
      { 16, 36 },
      -1,
   } },
   /* replace830_0_1_0 -> 16 in the cache */
   /* replace830_0_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sne,
      1, 1,
      { 16, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fdot3,
      -1, 2,
      { 2451, 2452 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 2,
      { 2453 },
      -1,
   } },

   /* ('fany_nequal4', 'a', 'b') => ('fsat', ('fdot4', ('sne', 'a', 'b'), ('sne', 'a', 'b'))) */
   /* search831_0 -> 16 in the cache */
   /* search831_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fany_nequal4,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace831_0_0_0 -> 16 in the cache */
   /* replace831_0_0_1 -> 36 in the cache */
   /* replace831_0_0 -> 2451 in the cache */
   /* replace831_0_1_0 -> 16 in the cache */
   /* replace831_0_1_1 -> 36 in the cache */
   /* replace831_0_1 -> 2452 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fdot4,
      -1, 2,
      { 2451, 2452 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 2,
      { 2456 },
      -1,
   } },

   /* ('fany_nequal8', 'a', 'b') => ('fsat', ('fdot8', ('sne', 'a', 'b'), ('sne', 'a', 'b'))) */
   /* search832_0 -> 16 in the cache */
   /* search832_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fany_nequal8,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace832_0_0_0 -> 16 in the cache */
   /* replace832_0_0_1 -> 36 in the cache */
   /* replace832_0_0 -> 2451 in the cache */
   /* replace832_0_1_0 -> 16 in the cache */
   /* replace832_0_1_1 -> 36 in the cache */
   /* replace832_0_1 -> 2452 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fdot8,
      -1, 2,
      { 2451, 2452 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 2,
      { 2459 },
      -1,
   } },

   /* ('fany_nequal16', 'a', 'b') => ('fsat', ('fdot16', ('sne', 'a', 'b'), ('sne', 'a', 'b'))) */
   /* search833_0 -> 16 in the cache */
   /* search833_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fany_nequal16,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace833_0_0_0 -> 16 in the cache */
   /* replace833_0_0_1 -> 36 in the cache */
   /* replace833_0_0 -> 2451 in the cache */
   /* replace833_0_1_0 -> 16 in the cache */
   /* replace833_0_1_1 -> 36 in the cache */
   /* replace833_0_1 -> 2452 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fdot16,
      -1, 2,
      { 2451, 2452 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 2,
      { 2462 },
      -1,
   } },

   /* ('ball_iequal2', 'a', 'b') => ('iand', ('ieq', 'a.x', 'b.x'), ('ieq', 'a.y', 'b.y')) */
   /* search834_0 -> 0 in the cache */
   /* search834_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal2,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace834_0_0 -> 769 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 769, 2465 },
      -1,
   } },
   /* replace834_1_0 -> 770 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {1},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 770, 2467 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2466, 2468 },
      -1,
   } },

   /* ('ball_iequal3', 'a', 'b') => ('iand', ('ieq', 'a.x', 'b.x'), ('iand', ('ieq', 'a.y', 'b.y'), ('ieq', 'a.z', 'b.z'))) */
   /* search835_0 -> 0 in the cache */
   /* search835_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal3,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace835_0_0 -> 769 in the cache */
   /* replace835_0_1 -> 2465 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 769, 2465 },
      -1,
   } },
   /* replace835_1_0_0 -> 770 in the cache */
   /* replace835_1_0_1 -> 2467 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 770, 2467 },
      -1,
   } },
   /* replace835_1_1_0 -> 771 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {2},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      3, 1,
      { 771, 2473 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2472, 2474 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 4,
      { 2471, 2475 },
      -1,
   } },

   /* ('ball_iequal4', 'a', 'b') => ('iand', ('iand', ('ieq', 'a.x', 'b.x'), ('ieq', 'a.y', 'b.y')), ('iand', ('ieq', 'a.z', 'b.z'), ('ieq', 'a.w', 'b.w'))) */
   /* search836_0 -> 0 in the cache */
   /* search836_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal4,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace836_0_0_0 -> 769 in the cache */
   /* replace836_0_0_1 -> 2465 in the cache */
   /* replace836_0_0 -> 2466 in the cache */
   /* replace836_0_1_0 -> 770 in the cache */
   /* replace836_0_1_1 -> 2467 in the cache */
   /* replace836_0_1 -> 2468 in the cache */
   /* replace836_0 -> 2469 in the cache */
   /* replace836_1_0_0 -> 771 in the cache */
   /* replace836_1_0_1 -> 2473 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 771, 2473 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {3},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {3},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      3, 1,
      { 2479, 2480 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2478, 2481 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 4,
      { 2469, 2482 },
      -1,
   } },

   /* ('ball_iequal8', 'a', 'b') => ('iand', ('iand', ('iand', ('ieq', 'a.a', 'b.a'), ('ieq', 'a.b', 'b.b')), ('iand', ('ieq', 'a.c', 'b.c'), ('ieq', 'a.d', 'b.d'))), ('iand', ('iand', ('ieq', 'a.e', 'b.e'), ('ieq', 'a.f', 'b.f')), ('iand', ('ieq', 'a.g', 'b.g'), ('ieq', 'a.h', 'b.h')))) */
   /* search837_0 -> 0 in the cache */
   /* search837_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal8,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace837_0_0_0_0 -> 769 in the cache */
   /* replace837_0_0_0_1 -> 2465 in the cache */
   /* replace837_0_0_0 -> 2466 in the cache */
   /* replace837_0_0_1_0 -> 770 in the cache */
   /* replace837_0_0_1_1 -> 2467 in the cache */
   /* replace837_0_0_1 -> 2468 in the cache */
   /* replace837_0_0 -> 2469 in the cache */
   /* replace837_0_1_0_0 -> 771 in the cache */
   /* replace837_0_1_0_1 -> 2473 in the cache */
   /* replace837_0_1_0 -> 2478 in the cache */
   /* replace837_0_1_1_0 -> 2479 in the cache */
   /* replace837_0_1_1_1 -> 2480 in the cache */
   /* replace837_0_1_1 -> 2481 in the cache */
   /* replace837_0_1 -> 2482 in the cache */
   /* replace837_0 -> 2483 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {4},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {4},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      4, 1,
      { 2485, 2486 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {5},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {5},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      5, 1,
      { 2488, 2489 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2487, 2490 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {6},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {6},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      6, 1,
      { 2492, 2493 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {7},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {7},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      7, 1,
      { 2495, 2496 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2494, 2497 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 4,
      { 2491, 2498 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 8,
      { 2483, 2499 },
      -1,
   } },

   /* ('ball_iequal16', 'a', 'b') => ('iand', ('iand', ('iand', ('iand', ('ieq', 'a.a', 'b.a'), ('ieq', 'a.b', 'b.b')), ('iand', ('ieq', 'a.c', 'b.c'), ('ieq', 'a.d', 'b.d'))), ('iand', ('iand', ('ieq', 'a.e', 'b.e'), ('ieq', 'a.f', 'b.f')), ('iand', ('ieq', 'a.g', 'b.g'), ('ieq', 'a.h', 'b.h')))), ('iand', ('iand', ('iand', ('ieq', 'a.i', 'b.i'), ('ieq', 'a.j', 'b.j')), ('iand', ('ieq', 'a.k', 'b.k'), ('ieq', 'a.l', 'b.l'))), ('iand', ('iand', ('ieq', 'a.m', 'b.m'), ('ieq', 'a.n', 'b.n')), ('iand', ('ieq', 'a.o', 'b.o'), ('ieq', 'a.p', 'b.p'))))) */
   /* search838_0 -> 0 in the cache */
   /* search838_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal16,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace838_0_0_0_0_0 -> 769 in the cache */
   /* replace838_0_0_0_0_1 -> 2465 in the cache */
   /* replace838_0_0_0_0 -> 2466 in the cache */
   /* replace838_0_0_0_1_0 -> 770 in the cache */
   /* replace838_0_0_0_1_1 -> 2467 in the cache */
   /* replace838_0_0_0_1 -> 2468 in the cache */
   /* replace838_0_0_0 -> 2469 in the cache */
   /* replace838_0_0_1_0_0 -> 771 in the cache */
   /* replace838_0_0_1_0_1 -> 2473 in the cache */
   /* replace838_0_0_1_0 -> 2478 in the cache */
   /* replace838_0_0_1_1_0 -> 2479 in the cache */
   /* replace838_0_0_1_1_1 -> 2480 in the cache */
   /* replace838_0_0_1_1 -> 2481 in the cache */
   /* replace838_0_0_1 -> 2482 in the cache */
   /* replace838_0_0 -> 2483 in the cache */
   /* replace838_0_1_0_0_0 -> 2485 in the cache */
   /* replace838_0_1_0_0_1 -> 2486 in the cache */
   /* replace838_0_1_0_0 -> 2487 in the cache */
   /* replace838_0_1_0_1_0 -> 2488 in the cache */
   /* replace838_0_1_0_1_1 -> 2489 in the cache */
   /* replace838_0_1_0_1 -> 2490 in the cache */
   /* replace838_0_1_0 -> 2491 in the cache */
   /* replace838_0_1_1_0_0 -> 2492 in the cache */
   /* replace838_0_1_1_0_1 -> 2493 in the cache */
   /* replace838_0_1_1_0 -> 2494 in the cache */
   /* replace838_0_1_1_1_0 -> 2495 in the cache */
   /* replace838_0_1_1_1_1 -> 2496 in the cache */
   /* replace838_0_1_1_1 -> 2497 in the cache */
   /* replace838_0_1_1 -> 2498 in the cache */
   /* replace838_0_1 -> 2499 in the cache */
   /* replace838_0 -> 2500 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {8},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {8},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      8, 1,
      { 2502, 2503 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {9},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {9},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      9, 1,
      { 2505, 2506 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2504, 2507 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {10},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {10},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      10, 1,
      { 2509, 2510 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {11},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {11},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      11, 1,
      { 2512, 2513 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2511, 2514 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 4,
      { 2508, 2515 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {12},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {12},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      12, 1,
      { 2517, 2518 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {13},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {13},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      13, 1,
      { 2520, 2521 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2519, 2522 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {14},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {14},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      14, 1,
      { 2524, 2525 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {15},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      15, 1,
      { 2527, 2528 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2526, 2529 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 4,
      { 2523, 2530 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 8,
      { 2516, 2531 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 16,
      { 2500, 2532 },
      -1,
   } },

   /* ('ball_fequal2', 'a', 'b') => ('iand', ('feq', 'a.x', 'b.x'), ('feq', 'a.y', 'b.y')) */
   /* search839_0 -> 0 in the cache */
   /* search839_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_fequal2,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace839_0_0 -> 769 in the cache */
   /* replace839_0_1 -> 2465 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 769, 2465 },
      -1,
   } },
   /* replace839_1_0 -> 770 in the cache */
   /* replace839_1_1 -> 2467 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      1, 1,
      { 770, 2467 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2535, 2536 },
      -1,
   } },

   /* ('ball_fequal3', 'a', 'b') => ('iand', ('feq', 'a.x', 'b.x'), ('iand', ('feq', 'a.y', 'b.y'), ('feq', 'a.z', 'b.z'))) */
   /* search840_0 -> 0 in the cache */
   /* search840_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_fequal3,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace840_0_0 -> 769 in the cache */
   /* replace840_0_1 -> 2465 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      1, 1,
      { 769, 2465 },
      -1,
   } },
   /* replace840_1_0_0 -> 770 in the cache */
   /* replace840_1_0_1 -> 2467 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      2, 1,
      { 770, 2467 },
      -1,
   } },
   /* replace840_1_1_0 -> 771 in the cache */
   /* replace840_1_1_1 -> 2473 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      3, 1,
      { 771, 2473 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2540, 2541 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 4,
      { 2539, 2542 },
      -1,
   } },

   /* ('ball_fequal4', 'a', 'b') => ('iand', ('iand', ('feq', 'a.x', 'b.x'), ('feq', 'a.y', 'b.y')), ('iand', ('feq', 'a.z', 'b.z'), ('feq', 'a.w', 'b.w'))) */
   /* search841_0 -> 0 in the cache */
   /* search841_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_fequal4,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace841_0_0_0 -> 769 in the cache */
   /* replace841_0_0_1 -> 2465 in the cache */
   /* replace841_0_0 -> 2535 in the cache */
   /* replace841_0_1_0 -> 770 in the cache */
   /* replace841_0_1_1 -> 2467 in the cache */
   /* replace841_0_1 -> 2536 in the cache */
   /* replace841_0 -> 2537 in the cache */
   /* replace841_1_0_0 -> 771 in the cache */
   /* replace841_1_0_1 -> 2473 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      2, 1,
      { 771, 2473 },
      -1,
   } },
   /* replace841_1_1_0 -> 2479 in the cache */
   /* replace841_1_1_1 -> 2480 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      3, 1,
      { 2479, 2480 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2545, 2546 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 4,
      { 2537, 2547 },
      -1,
   } },

   /* ('ball_fequal8', 'a', 'b') => ('iand', ('iand', ('iand', ('feq', 'a.a', 'b.a'), ('feq', 'a.b', 'b.b')), ('iand', ('feq', 'a.c', 'b.c'), ('feq', 'a.d', 'b.d'))), ('iand', ('iand', ('feq', 'a.e', 'b.e'), ('feq', 'a.f', 'b.f')), ('iand', ('feq', 'a.g', 'b.g'), ('feq', 'a.h', 'b.h')))) */
   /* search842_0 -> 0 in the cache */
   /* search842_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_fequal8,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace842_0_0_0_0 -> 769 in the cache */
   /* replace842_0_0_0_1 -> 2465 in the cache */
   /* replace842_0_0_0 -> 2535 in the cache */
   /* replace842_0_0_1_0 -> 770 in the cache */
   /* replace842_0_0_1_1 -> 2467 in the cache */
   /* replace842_0_0_1 -> 2536 in the cache */
   /* replace842_0_0 -> 2537 in the cache */
   /* replace842_0_1_0_0 -> 771 in the cache */
   /* replace842_0_1_0_1 -> 2473 in the cache */
   /* replace842_0_1_0 -> 2545 in the cache */
   /* replace842_0_1_1_0 -> 2479 in the cache */
   /* replace842_0_1_1_1 -> 2480 in the cache */
   /* replace842_0_1_1 -> 2546 in the cache */
   /* replace842_0_1 -> 2547 in the cache */
   /* replace842_0 -> 2548 in the cache */
   /* replace842_1_0_0_0 -> 2485 in the cache */
   /* replace842_1_0_0_1 -> 2486 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      4, 1,
      { 2485, 2486 },
      -1,
   } },
   /* replace842_1_0_1_0 -> 2488 in the cache */
   /* replace842_1_0_1_1 -> 2489 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      5, 1,
      { 2488, 2489 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2550, 2551 },
      -1,
   } },
   /* replace842_1_1_0_0 -> 2492 in the cache */
   /* replace842_1_1_0_1 -> 2493 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      6, 1,
      { 2492, 2493 },
      -1,
   } },
   /* replace842_1_1_1_0 -> 2495 in the cache */
   /* replace842_1_1_1_1 -> 2496 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      7, 1,
      { 2495, 2496 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2553, 2554 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 4,
      { 2552, 2555 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 8,
      { 2548, 2556 },
      -1,
   } },

   /* ('ball_fequal16', 'a', 'b') => ('iand', ('iand', ('iand', ('iand', ('feq', 'a.a', 'b.a'), ('feq', 'a.b', 'b.b')), ('iand', ('feq', 'a.c', 'b.c'), ('feq', 'a.d', 'b.d'))), ('iand', ('iand', ('feq', 'a.e', 'b.e'), ('feq', 'a.f', 'b.f')), ('iand', ('feq', 'a.g', 'b.g'), ('feq', 'a.h', 'b.h')))), ('iand', ('iand', ('iand', ('feq', 'a.i', 'b.i'), ('feq', 'a.j', 'b.j')), ('iand', ('feq', 'a.k', 'b.k'), ('feq', 'a.l', 'b.l'))), ('iand', ('iand', ('feq', 'a.m', 'b.m'), ('feq', 'a.n', 'b.n')), ('iand', ('feq', 'a.o', 'b.o'), ('feq', 'a.p', 'b.p'))))) */
   /* search843_0 -> 0 in the cache */
   /* search843_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_fequal16,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace843_0_0_0_0_0 -> 769 in the cache */
   /* replace843_0_0_0_0_1 -> 2465 in the cache */
   /* replace843_0_0_0_0 -> 2535 in the cache */
   /* replace843_0_0_0_1_0 -> 770 in the cache */
   /* replace843_0_0_0_1_1 -> 2467 in the cache */
   /* replace843_0_0_0_1 -> 2536 in the cache */
   /* replace843_0_0_0 -> 2537 in the cache */
   /* replace843_0_0_1_0_0 -> 771 in the cache */
   /* replace843_0_0_1_0_1 -> 2473 in the cache */
   /* replace843_0_0_1_0 -> 2545 in the cache */
   /* replace843_0_0_1_1_0 -> 2479 in the cache */
   /* replace843_0_0_1_1_1 -> 2480 in the cache */
   /* replace843_0_0_1_1 -> 2546 in the cache */
   /* replace843_0_0_1 -> 2547 in the cache */
   /* replace843_0_0 -> 2548 in the cache */
   /* replace843_0_1_0_0_0 -> 2485 in the cache */
   /* replace843_0_1_0_0_1 -> 2486 in the cache */
   /* replace843_0_1_0_0 -> 2550 in the cache */
   /* replace843_0_1_0_1_0 -> 2488 in the cache */
   /* replace843_0_1_0_1_1 -> 2489 in the cache */
   /* replace843_0_1_0_1 -> 2551 in the cache */
   /* replace843_0_1_0 -> 2552 in the cache */
   /* replace843_0_1_1_0_0 -> 2492 in the cache */
   /* replace843_0_1_1_0_1 -> 2493 in the cache */
   /* replace843_0_1_1_0 -> 2553 in the cache */
   /* replace843_0_1_1_1_0 -> 2495 in the cache */
   /* replace843_0_1_1_1_1 -> 2496 in the cache */
   /* replace843_0_1_1_1 -> 2554 in the cache */
   /* replace843_0_1_1 -> 2555 in the cache */
   /* replace843_0_1 -> 2556 in the cache */
   /* replace843_0 -> 2557 in the cache */
   /* replace843_1_0_0_0_0 -> 2502 in the cache */
   /* replace843_1_0_0_0_1 -> 2503 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      8, 1,
      { 2502, 2503 },
      -1,
   } },
   /* replace843_1_0_0_1_0 -> 2505 in the cache */
   /* replace843_1_0_0_1_1 -> 2506 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      9, 1,
      { 2505, 2506 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2559, 2560 },
      -1,
   } },
   /* replace843_1_0_1_0_0 -> 2509 in the cache */
   /* replace843_1_0_1_0_1 -> 2510 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      10, 1,
      { 2509, 2510 },
      -1,
   } },
   /* replace843_1_0_1_1_0 -> 2512 in the cache */
   /* replace843_1_0_1_1_1 -> 2513 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      11, 1,
      { 2512, 2513 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2562, 2563 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 4,
      { 2561, 2564 },
      -1,
   } },
   /* replace843_1_1_0_0_0 -> 2517 in the cache */
   /* replace843_1_1_0_0_1 -> 2518 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      12, 1,
      { 2517, 2518 },
      -1,
   } },
   /* replace843_1_1_0_1_0 -> 2520 in the cache */
   /* replace843_1_1_0_1_1 -> 2521 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      13, 1,
      { 2520, 2521 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2566, 2567 },
      -1,
   } },
   /* replace843_1_1_1_0_0 -> 2524 in the cache */
   /* replace843_1_1_1_0_1 -> 2525 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      14, 1,
      { 2524, 2525 },
      -1,
   } },
   /* replace843_1_1_1_1_0 -> 2527 in the cache */
   /* replace843_1_1_1_1_1 -> 2528 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      15, 1,
      { 2527, 2528 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2569, 2570 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 4,
      { 2568, 2571 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 8,
      { 2565, 2572 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 16,
      { 2557, 2573 },
      -1,
   } },

   /* ('bany_inequal2', 'a', 'b') => ('ior', ('ine', 'a.x', 'b.x'), ('ine', 'a.y', 'b.y')) */
   /* search844_0 -> 0 in the cache */
   /* search844_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal2,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace844_0_0 -> 769 in the cache */
   /* replace844_0_1 -> 2465 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 769, 2465 },
      -1,
   } },
   /* replace844_1_0 -> 770 in the cache */
   /* replace844_1_1 -> 2467 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 770, 2467 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2576, 2577 },
      -1,
   } },

   /* ('bany_inequal3', 'a', 'b') => ('ior', ('ine', 'a.x', 'b.x'), ('ior', ('ine', 'a.y', 'b.y'), ('ine', 'a.z', 'b.z'))) */
   /* search845_0 -> 0 in the cache */
   /* search845_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal3,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace845_0_0 -> 769 in the cache */
   /* replace845_0_1 -> 2465 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 769, 2465 },
      -1,
   } },
   /* replace845_1_0_0 -> 770 in the cache */
   /* replace845_1_0_1 -> 2467 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 770, 2467 },
      -1,
   } },
   /* replace845_1_1_0 -> 771 in the cache */
   /* replace845_1_1_1 -> 2473 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      3, 1,
      { 771, 2473 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2581, 2582 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 4,
      { 2580, 2583 },
      -1,
   } },

   /* ('bany_inequal4', 'a', 'b') => ('ior', ('ior', ('ine', 'a.x', 'b.x'), ('ine', 'a.y', 'b.y')), ('ior', ('ine', 'a.z', 'b.z'), ('ine', 'a.w', 'b.w'))) */
   /* search846_0 -> 0 in the cache */
   /* search846_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal4,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace846_0_0_0 -> 769 in the cache */
   /* replace846_0_0_1 -> 2465 in the cache */
   /* replace846_0_0 -> 2576 in the cache */
   /* replace846_0_1_0 -> 770 in the cache */
   /* replace846_0_1_1 -> 2467 in the cache */
   /* replace846_0_1 -> 2577 in the cache */
   /* replace846_0 -> 2578 in the cache */
   /* replace846_1_0_0 -> 771 in the cache */
   /* replace846_1_0_1 -> 2473 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 771, 2473 },
      -1,
   } },
   /* replace846_1_1_0 -> 2479 in the cache */
   /* replace846_1_1_1 -> 2480 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      3, 1,
      { 2479, 2480 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2586, 2587 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 4,
      { 2578, 2588 },
      -1,
   } },

   /* ('bany_inequal8', 'a', 'b') => ('ior', ('ior', ('ior', ('ine', 'a.a', 'b.a'), ('ine', 'a.b', 'b.b')), ('ior', ('ine', 'a.c', 'b.c'), ('ine', 'a.d', 'b.d'))), ('ior', ('ior', ('ine', 'a.e', 'b.e'), ('ine', 'a.f', 'b.f')), ('ior', ('ine', 'a.g', 'b.g'), ('ine', 'a.h', 'b.h')))) */
   /* search847_0 -> 0 in the cache */
   /* search847_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal8,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace847_0_0_0_0 -> 769 in the cache */
   /* replace847_0_0_0_1 -> 2465 in the cache */
   /* replace847_0_0_0 -> 2576 in the cache */
   /* replace847_0_0_1_0 -> 770 in the cache */
   /* replace847_0_0_1_1 -> 2467 in the cache */
   /* replace847_0_0_1 -> 2577 in the cache */
   /* replace847_0_0 -> 2578 in the cache */
   /* replace847_0_1_0_0 -> 771 in the cache */
   /* replace847_0_1_0_1 -> 2473 in the cache */
   /* replace847_0_1_0 -> 2586 in the cache */
   /* replace847_0_1_1_0 -> 2479 in the cache */
   /* replace847_0_1_1_1 -> 2480 in the cache */
   /* replace847_0_1_1 -> 2587 in the cache */
   /* replace847_0_1 -> 2588 in the cache */
   /* replace847_0 -> 2589 in the cache */
   /* replace847_1_0_0_0 -> 2485 in the cache */
   /* replace847_1_0_0_1 -> 2486 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      4, 1,
      { 2485, 2486 },
      -1,
   } },
   /* replace847_1_0_1_0 -> 2488 in the cache */
   /* replace847_1_0_1_1 -> 2489 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      5, 1,
      { 2488, 2489 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2591, 2592 },
      -1,
   } },
   /* replace847_1_1_0_0 -> 2492 in the cache */
   /* replace847_1_1_0_1 -> 2493 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      6, 1,
      { 2492, 2493 },
      -1,
   } },
   /* replace847_1_1_1_0 -> 2495 in the cache */
   /* replace847_1_1_1_1 -> 2496 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      7, 1,
      { 2495, 2496 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2594, 2595 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 4,
      { 2593, 2596 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 8,
      { 2589, 2597 },
      -1,
   } },

   /* ('bany_inequal16', 'a', 'b') => ('ior', ('ior', ('ior', ('ior', ('ine', 'a.a', 'b.a'), ('ine', 'a.b', 'b.b')), ('ior', ('ine', 'a.c', 'b.c'), ('ine', 'a.d', 'b.d'))), ('ior', ('ior', ('ine', 'a.e', 'b.e'), ('ine', 'a.f', 'b.f')), ('ior', ('ine', 'a.g', 'b.g'), ('ine', 'a.h', 'b.h')))), ('ior', ('ior', ('ior', ('ine', 'a.i', 'b.i'), ('ine', 'a.j', 'b.j')), ('ior', ('ine', 'a.k', 'b.k'), ('ine', 'a.l', 'b.l'))), ('ior', ('ior', ('ine', 'a.m', 'b.m'), ('ine', 'a.n', 'b.n')), ('ior', ('ine', 'a.o', 'b.o'), ('ine', 'a.p', 'b.p'))))) */
   /* search848_0 -> 0 in the cache */
   /* search848_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal16,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace848_0_0_0_0_0 -> 769 in the cache */
   /* replace848_0_0_0_0_1 -> 2465 in the cache */
   /* replace848_0_0_0_0 -> 2576 in the cache */
   /* replace848_0_0_0_1_0 -> 770 in the cache */
   /* replace848_0_0_0_1_1 -> 2467 in the cache */
   /* replace848_0_0_0_1 -> 2577 in the cache */
   /* replace848_0_0_0 -> 2578 in the cache */
   /* replace848_0_0_1_0_0 -> 771 in the cache */
   /* replace848_0_0_1_0_1 -> 2473 in the cache */
   /* replace848_0_0_1_0 -> 2586 in the cache */
   /* replace848_0_0_1_1_0 -> 2479 in the cache */
   /* replace848_0_0_1_1_1 -> 2480 in the cache */
   /* replace848_0_0_1_1 -> 2587 in the cache */
   /* replace848_0_0_1 -> 2588 in the cache */
   /* replace848_0_0 -> 2589 in the cache */
   /* replace848_0_1_0_0_0 -> 2485 in the cache */
   /* replace848_0_1_0_0_1 -> 2486 in the cache */
   /* replace848_0_1_0_0 -> 2591 in the cache */
   /* replace848_0_1_0_1_0 -> 2488 in the cache */
   /* replace848_0_1_0_1_1 -> 2489 in the cache */
   /* replace848_0_1_0_1 -> 2592 in the cache */
   /* replace848_0_1_0 -> 2593 in the cache */
   /* replace848_0_1_1_0_0 -> 2492 in the cache */
   /* replace848_0_1_1_0_1 -> 2493 in the cache */
   /* replace848_0_1_1_0 -> 2594 in the cache */
   /* replace848_0_1_1_1_0 -> 2495 in the cache */
   /* replace848_0_1_1_1_1 -> 2496 in the cache */
   /* replace848_0_1_1_1 -> 2595 in the cache */
   /* replace848_0_1_1 -> 2596 in the cache */
   /* replace848_0_1 -> 2597 in the cache */
   /* replace848_0 -> 2598 in the cache */
   /* replace848_1_0_0_0_0 -> 2502 in the cache */
   /* replace848_1_0_0_0_1 -> 2503 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      8, 1,
      { 2502, 2503 },
      -1,
   } },
   /* replace848_1_0_0_1_0 -> 2505 in the cache */
   /* replace848_1_0_0_1_1 -> 2506 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      9, 1,
      { 2505, 2506 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2600, 2601 },
      -1,
   } },
   /* replace848_1_0_1_0_0 -> 2509 in the cache */
   /* replace848_1_0_1_0_1 -> 2510 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      10, 1,
      { 2509, 2510 },
      -1,
   } },
   /* replace848_1_0_1_1_0 -> 2512 in the cache */
   /* replace848_1_0_1_1_1 -> 2513 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      11, 1,
      { 2512, 2513 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2603, 2604 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 4,
      { 2602, 2605 },
      -1,
   } },
   /* replace848_1_1_0_0_0 -> 2517 in the cache */
   /* replace848_1_1_0_0_1 -> 2518 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      12, 1,
      { 2517, 2518 },
      -1,
   } },
   /* replace848_1_1_0_1_0 -> 2520 in the cache */
   /* replace848_1_1_0_1_1 -> 2521 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      13, 1,
      { 2520, 2521 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2607, 2608 },
      -1,
   } },
   /* replace848_1_1_1_0_0 -> 2524 in the cache */
   /* replace848_1_1_1_0_1 -> 2525 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      14, 1,
      { 2524, 2525 },
      -1,
   } },
   /* replace848_1_1_1_1_0 -> 2527 in the cache */
   /* replace848_1_1_1_1_1 -> 2528 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      15, 1,
      { 2527, 2528 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2610, 2611 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 4,
      { 2609, 2612 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 8,
      { 2606, 2613 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 16,
      { 2598, 2614 },
      -1,
   } },

   /* ('bany_fnequal2', 'a', 'b') => ('ior', ('fneu', 'a.x', 'b.x'), ('fneu', 'a.y', 'b.y')) */
   /* search849_0 -> 0 in the cache */
   /* search849_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_fnequal2,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace849_0_0 -> 769 in the cache */
   /* replace849_0_1 -> 2465 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 769, 2465 },
      -1,
   } },
   /* replace849_1_0 -> 770 in the cache */
   /* replace849_1_1 -> 2467 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      1, 1,
      { 770, 2467 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2617, 2618 },
      -1,
   } },

   /* ('bany_fnequal3', 'a', 'b') => ('ior', ('fneu', 'a.x', 'b.x'), ('ior', ('fneu', 'a.y', 'b.y'), ('fneu', 'a.z', 'b.z'))) */
   /* search850_0 -> 0 in the cache */
   /* search850_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_fnequal3,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace850_0_0 -> 769 in the cache */
   /* replace850_0_1 -> 2465 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      1, 1,
      { 769, 2465 },
      -1,
   } },
   /* replace850_1_0_0 -> 770 in the cache */
   /* replace850_1_0_1 -> 2467 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      2, 1,
      { 770, 2467 },
      -1,
   } },
   /* replace850_1_1_0 -> 771 in the cache */
   /* replace850_1_1_1 -> 2473 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      3, 1,
      { 771, 2473 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2622, 2623 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 4,
      { 2621, 2624 },
      -1,
   } },

   /* ('bany_fnequal4', 'a', 'b') => ('ior', ('ior', ('fneu', 'a.x', 'b.x'), ('fneu', 'a.y', 'b.y')), ('ior', ('fneu', 'a.z', 'b.z'), ('fneu', 'a.w', 'b.w'))) */
   /* search851_0 -> 0 in the cache */
   /* search851_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_fnequal4,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace851_0_0_0 -> 769 in the cache */
   /* replace851_0_0_1 -> 2465 in the cache */
   /* replace851_0_0 -> 2617 in the cache */
   /* replace851_0_1_0 -> 770 in the cache */
   /* replace851_0_1_1 -> 2467 in the cache */
   /* replace851_0_1 -> 2618 in the cache */
   /* replace851_0 -> 2619 in the cache */
   /* replace851_1_0_0 -> 771 in the cache */
   /* replace851_1_0_1 -> 2473 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      2, 1,
      { 771, 2473 },
      -1,
   } },
   /* replace851_1_1_0 -> 2479 in the cache */
   /* replace851_1_1_1 -> 2480 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      3, 1,
      { 2479, 2480 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2627, 2628 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 4,
      { 2619, 2629 },
      -1,
   } },

   /* ('bany_fnequal8', 'a', 'b') => ('ior', ('ior', ('ior', ('fneu', 'a.a', 'b.a'), ('fneu', 'a.b', 'b.b')), ('ior', ('fneu', 'a.c', 'b.c'), ('fneu', 'a.d', 'b.d'))), ('ior', ('ior', ('fneu', 'a.e', 'b.e'), ('fneu', 'a.f', 'b.f')), ('ior', ('fneu', 'a.g', 'b.g'), ('fneu', 'a.h', 'b.h')))) */
   /* search852_0 -> 0 in the cache */
   /* search852_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_fnequal8,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace852_0_0_0_0 -> 769 in the cache */
   /* replace852_0_0_0_1 -> 2465 in the cache */
   /* replace852_0_0_0 -> 2617 in the cache */
   /* replace852_0_0_1_0 -> 770 in the cache */
   /* replace852_0_0_1_1 -> 2467 in the cache */
   /* replace852_0_0_1 -> 2618 in the cache */
   /* replace852_0_0 -> 2619 in the cache */
   /* replace852_0_1_0_0 -> 771 in the cache */
   /* replace852_0_1_0_1 -> 2473 in the cache */
   /* replace852_0_1_0 -> 2627 in the cache */
   /* replace852_0_1_1_0 -> 2479 in the cache */
   /* replace852_0_1_1_1 -> 2480 in the cache */
   /* replace852_0_1_1 -> 2628 in the cache */
   /* replace852_0_1 -> 2629 in the cache */
   /* replace852_0 -> 2630 in the cache */
   /* replace852_1_0_0_0 -> 2485 in the cache */
   /* replace852_1_0_0_1 -> 2486 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      4, 1,
      { 2485, 2486 },
      -1,
   } },
   /* replace852_1_0_1_0 -> 2488 in the cache */
   /* replace852_1_0_1_1 -> 2489 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      5, 1,
      { 2488, 2489 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2632, 2633 },
      -1,
   } },
   /* replace852_1_1_0_0 -> 2492 in the cache */
   /* replace852_1_1_0_1 -> 2493 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      6, 1,
      { 2492, 2493 },
      -1,
   } },
   /* replace852_1_1_1_0 -> 2495 in the cache */
   /* replace852_1_1_1_1 -> 2496 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      7, 1,
      { 2495, 2496 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2635, 2636 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 4,
      { 2634, 2637 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 8,
      { 2630, 2638 },
      -1,
   } },

   /* ('bany_fnequal16', 'a', 'b') => ('ior', ('ior', ('ior', ('ior', ('fneu', 'a.a', 'b.a'), ('fneu', 'a.b', 'b.b')), ('ior', ('fneu', 'a.c', 'b.c'), ('fneu', 'a.d', 'b.d'))), ('ior', ('ior', ('fneu', 'a.e', 'b.e'), ('fneu', 'a.f', 'b.f')), ('ior', ('fneu', 'a.g', 'b.g'), ('fneu', 'a.h', 'b.h')))), ('ior', ('ior', ('ior', ('fneu', 'a.i', 'b.i'), ('fneu', 'a.j', 'b.j')), ('ior', ('fneu', 'a.k', 'b.k'), ('fneu', 'a.l', 'b.l'))), ('ior', ('ior', ('fneu', 'a.m', 'b.m'), ('fneu', 'a.n', 'b.n')), ('ior', ('fneu', 'a.o', 'b.o'), ('fneu', 'a.p', 'b.p'))))) */
   /* search853_0 -> 0 in the cache */
   /* search853_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_fnequal16,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace853_0_0_0_0_0 -> 769 in the cache */
   /* replace853_0_0_0_0_1 -> 2465 in the cache */
   /* replace853_0_0_0_0 -> 2617 in the cache */
   /* replace853_0_0_0_1_0 -> 770 in the cache */
   /* replace853_0_0_0_1_1 -> 2467 in the cache */
   /* replace853_0_0_0_1 -> 2618 in the cache */
   /* replace853_0_0_0 -> 2619 in the cache */
   /* replace853_0_0_1_0_0 -> 771 in the cache */
   /* replace853_0_0_1_0_1 -> 2473 in the cache */
   /* replace853_0_0_1_0 -> 2627 in the cache */
   /* replace853_0_0_1_1_0 -> 2479 in the cache */
   /* replace853_0_0_1_1_1 -> 2480 in the cache */
   /* replace853_0_0_1_1 -> 2628 in the cache */
   /* replace853_0_0_1 -> 2629 in the cache */
   /* replace853_0_0 -> 2630 in the cache */
   /* replace853_0_1_0_0_0 -> 2485 in the cache */
   /* replace853_0_1_0_0_1 -> 2486 in the cache */
   /* replace853_0_1_0_0 -> 2632 in the cache */
   /* replace853_0_1_0_1_0 -> 2488 in the cache */
   /* replace853_0_1_0_1_1 -> 2489 in the cache */
   /* replace853_0_1_0_1 -> 2633 in the cache */
   /* replace853_0_1_0 -> 2634 in the cache */
   /* replace853_0_1_1_0_0 -> 2492 in the cache */
   /* replace853_0_1_1_0_1 -> 2493 in the cache */
   /* replace853_0_1_1_0 -> 2635 in the cache */
   /* replace853_0_1_1_1_0 -> 2495 in the cache */
   /* replace853_0_1_1_1_1 -> 2496 in the cache */
   /* replace853_0_1_1_1 -> 2636 in the cache */
   /* replace853_0_1_1 -> 2637 in the cache */
   /* replace853_0_1 -> 2638 in the cache */
   /* replace853_0 -> 2639 in the cache */
   /* replace853_1_0_0_0_0 -> 2502 in the cache */
   /* replace853_1_0_0_0_1 -> 2503 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      8, 1,
      { 2502, 2503 },
      -1,
   } },
   /* replace853_1_0_0_1_0 -> 2505 in the cache */
   /* replace853_1_0_0_1_1 -> 2506 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      9, 1,
      { 2505, 2506 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2641, 2642 },
      -1,
   } },
   /* replace853_1_0_1_0_0 -> 2509 in the cache */
   /* replace853_1_0_1_0_1 -> 2510 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      10, 1,
      { 2509, 2510 },
      -1,
   } },
   /* replace853_1_0_1_1_0 -> 2512 in the cache */
   /* replace853_1_0_1_1_1 -> 2513 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      11, 1,
      { 2512, 2513 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2644, 2645 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 4,
      { 2643, 2646 },
      -1,
   } },
   /* replace853_1_1_0_0_0 -> 2517 in the cache */
   /* replace853_1_1_0_0_1 -> 2518 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      12, 1,
      { 2517, 2518 },
      -1,
   } },
   /* replace853_1_1_0_1_0 -> 2520 in the cache */
   /* replace853_1_1_0_1_1 -> 2521 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      13, 1,
      { 2520, 2521 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2648, 2649 },
      -1,
   } },
   /* replace853_1_1_1_0_0 -> 2524 in the cache */
   /* replace853_1_1_1_0_1 -> 2525 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      14, 1,
      { 2524, 2525 },
      -1,
   } },
   /* replace853_1_1_1_1_0 -> 2527 in the cache */
   /* replace853_1_1_1_1_1 -> 2528 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      15, 1,
      { 2527, 2528 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2651, 2652 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 4,
      { 2650, 2653 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 8,
      { 2647, 2654 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 16,
      { 2639, 2655 },
      -1,
   } },

   /* ('feq', ('seq', 'a', 'b'), 1.0) => ('feq', 'a', 'b') */
   /* search854_0_0 -> 0 in the cache */
   /* search854_0_1 -> 3 in the cache */
   /* search854_0 -> 2408 in the cache */
   /* search854_1 -> 270 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 2408, 270 },
      -1,
   } },

   /* replace854_0 -> 0 in the cache */
   /* replace854_1 -> 3 in the cache */
   /* replace854 -> 1055 in the cache */

   /* ('feq', ('sne', 'a', 'b'), 1.0) => ('fneu', 'a', 'b') */
   /* search855_0_0 -> 0 in the cache */
   /* search855_0_1 -> 3 in the cache */
   /* search855_0 -> 2410 in the cache */
   /* search855_1 -> 270 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 2410, 270 },
      -1,
   } },

   /* replace855_0 -> 0 in the cache */
   /* replace855_1 -> 3 in the cache */
   /* replace855 -> 1052 in the cache */

   /* ('feq', ('slt', 'a', 'b'), 1.0) => ('flt', 'a', 'b') */
   /* search856_0_0 -> 0 in the cache */
   /* search856_0_1 -> 3 in the cache */
   /* search856_0 -> 2396 in the cache */
   /* search856_1 -> 270 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 2396, 270 },
      -1,
   } },

   /* replace856_0 -> 0 in the cache */
   /* replace856_1 -> 3 in the cache */
   /* replace856 -> 1049 in the cache */

   /* ('feq', ('sge', 'a', 'b'), 1.0) => ('fge', 'a', 'b') */
   /* search857_0_0 -> 0 in the cache */
   /* search857_0_1 -> 3 in the cache */
   /* search857_0 -> 2398 in the cache */
   /* search857_1 -> 270 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 2398, 270 },
      -1,
   } },

   /* replace857_0 -> 0 in the cache */
   /* replace857_1 -> 3 in the cache */
   /* replace857 -> 1046 in the cache */

   /* ('fneu', ('seq', 'a', 'b'), 0.0) => ('feq', 'a', 'b') */
   /* search858_0_0 -> 0 in the cache */
   /* search858_0_1 -> 3 in the cache */
   /* search858_0 -> 2408 in the cache */
   /* search858_1 -> 256 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 2408, 256 },
      -1,
   } },

   /* replace858_0 -> 0 in the cache */
   /* replace858_1 -> 3 in the cache */
   /* replace858 -> 1055 in the cache */

   /* ('fneu', ('sne', 'a', 'b'), 0.0) => ('fneu', 'a', 'b') */
   /* search859_0_0 -> 0 in the cache */
   /* search859_0_1 -> 3 in the cache */
   /* search859_0 -> 2410 in the cache */
   /* search859_1 -> 256 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 2410, 256 },
      -1,
   } },

   /* replace859_0 -> 0 in the cache */
   /* replace859_1 -> 3 in the cache */
   /* replace859 -> 1052 in the cache */

   /* ('fneu', ('slt', 'a', 'b'), 0.0) => ('flt', 'a', 'b') */
   /* search860_0_0 -> 0 in the cache */
   /* search860_0_1 -> 3 in the cache */
   /* search860_0 -> 2396 in the cache */
   /* search860_1 -> 256 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 2396, 256 },
      -1,
   } },

   /* replace860_0 -> 0 in the cache */
   /* replace860_1 -> 3 in the cache */
   /* replace860 -> 1049 in the cache */

   /* ('fneu', ('sge', 'a', 'b'), 0.0) => ('fge', 'a', 'b') */
   /* search861_0_0 -> 0 in the cache */
   /* search861_0_1 -> 3 in the cache */
   /* search861_0 -> 2398 in the cache */
   /* search861_1 -> 256 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 2398, 256 },
      -1,
   } },

   /* replace861_0 -> 0 in the cache */
   /* replace861_1 -> 3 in the cache */
   /* replace861 -> 1046 in the cache */

   /* ('feq', ('seq', 'a', 'b'), 0.0) => ('fneu', 'a', 'b') */
   /* search862_0_0 -> 0 in the cache */
   /* search862_0_1 -> 3 in the cache */
   /* search862_0 -> 2408 in the cache */
   /* search862_1 -> 256 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 2408, 256 },
      -1,
   } },

   /* replace862_0 -> 0 in the cache */
   /* replace862_1 -> 3 in the cache */
   /* replace862 -> 1052 in the cache */

   /* ('feq', ('sne', 'a', 'b'), 0.0) => ('feq', 'a', 'b') */
   /* search863_0_0 -> 0 in the cache */
   /* search863_0_1 -> 3 in the cache */
   /* search863_0 -> 2410 in the cache */
   /* search863_1 -> 256 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 2410, 256 },
      -1,
   } },

   /* replace863_0 -> 0 in the cache */
   /* replace863_1 -> 3 in the cache */
   /* replace863 -> 1055 in the cache */

   /* ('feq', ('slt', 'a', 'b'), 0.0) => ('fge', 'a', 'b') */
   /* search864_0_0 -> 0 in the cache */
   /* search864_0_1 -> 3 in the cache */
   /* search864_0 -> 2396 in the cache */
   /* search864_1 -> 256 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 2396, 256 },
      -1,
   } },

   /* replace864_0 -> 0 in the cache */
   /* replace864_1 -> 3 in the cache */
   /* replace864 -> 1046 in the cache */

   /* ('feq', ('sge', 'a', 'b'), 0.0) => ('flt', 'a', 'b') */
   /* search865_0_0 -> 0 in the cache */
   /* search865_0_1 -> 3 in the cache */
   /* search865_0 -> 2398 in the cache */
   /* search865_1 -> 256 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 2398, 256 },
      -1,
   } },

   /* replace865_0 -> 0 in the cache */
   /* replace865_1 -> 3 in the cache */
   /* replace865 -> 1049 in the cache */

   /* ('fneu', ('seq', 'a', 'b'), 1.0) => ('fneu', 'a', 'b') */
   /* search866_0_0 -> 0 in the cache */
   /* search866_0_1 -> 3 in the cache */
   /* search866_0 -> 2408 in the cache */
   /* search866_1 -> 270 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 2408, 270 },
      -1,
   } },

   /* replace866_0 -> 0 in the cache */
   /* replace866_1 -> 3 in the cache */
   /* replace866 -> 1052 in the cache */

   /* ('fneu', ('sne', 'a', 'b'), 1.0) => ('feq', 'a', 'b') */
   /* search867_0_0 -> 0 in the cache */
   /* search867_0_1 -> 3 in the cache */
   /* search867_0 -> 2410 in the cache */
   /* search867_1 -> 270 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 2410, 270 },
      -1,
   } },

   /* replace867_0 -> 0 in the cache */
   /* replace867_1 -> 3 in the cache */
   /* replace867 -> 1055 in the cache */

   /* ('fneu', ('slt', 'a', 'b'), 1.0) => ('fge', 'a', 'b') */
   /* search868_0_0 -> 0 in the cache */
   /* search868_0_1 -> 3 in the cache */
   /* search868_0 -> 2396 in the cache */
   /* search868_1 -> 270 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 2396, 270 },
      -1,
   } },

   /* replace868_0 -> 0 in the cache */
   /* replace868_1 -> 3 in the cache */
   /* replace868 -> 1046 in the cache */

   /* ('fneu', ('sge', 'a', 'b'), 1.0) => ('flt', 'a', 'b') */
   /* search869_0_0 -> 0 in the cache */
   /* search869_0_1 -> 3 in the cache */
   /* search869_0 -> 2398 in the cache */
   /* search869_1 -> 270 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 2398, 270 },
      -1,
   } },

   /* replace869_0 -> 0 in the cache */
   /* replace869_1 -> 3 in the cache */
   /* replace869 -> 1049 in the cache */

   /* ('fneu', ('fneg', 'a'), 'a') => ('fneu', 'a', 0.0) */
   /* search870_0_0 -> 32 in the cache */
   /* search870_0 -> 98 in the cache */
   /* search870_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 98, 32 },
      -1,
   } },

   /* replace870_0 -> 32 in the cache */
   /* replace870_1 -> 111 in the cache */
   /* replace870 -> 1253 in the cache */

   /* ('feq', ('fneg', 'a'), 'a') => ('feq', 'a', 0.0) */
   /* search871_0_0 -> 32 in the cache */
   /* search871_0 -> 98 in the cache */
   /* search871_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 98, 32 },
      -1,
   } },

   /* replace871_0 -> 32 in the cache */
   /* replace871_1 -> 111 in the cache */
   /* replace871 -> 1257 in the cache */

   /* ('imul', ('b2i', 'a@1'), ('b2i', 'b@1')) => ('b2i', ('iand', 'a', 'b')) */
   /* search872_0_0 -> 760 in the cache */
   /* search872_0 -> 1228 in the cache */
   /* search872_1_0 -> 1134 in the cache */
   /* search872_1 -> 2249 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 1228, 2249 },
      -1,
   } },

   /* replace872_0_0 -> 760 in the cache */
   /* replace872_0_1 -> 1134 in the cache */
   /* replace872_0 -> 1146 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 1,
      { 1146 },
      -1,
   } },

   /* ('iand', ('b2i', 'a@1'), ('b2i', 'b@1')) => ('b2i', ('iand', 'a', 'b')) */
   /* search873_0_0 -> 760 in the cache */
   /* search873_0 -> 1228 in the cache */
   /* search873_1_0 -> 1134 in the cache */
   /* search873_1 -> 2249 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1228, 2249 },
      -1,
   } },

   /* replace873_0_0 -> 760 in the cache */
   /* replace873_0_1 -> 1134 in the cache */
   /* replace873_0 -> 1146 in the cache */
   /* replace873 -> 2676 in the cache */

   /* ('ior', ('b2i', 'a@1'), ('b2i', 'b@1')) => ('b2i', ('ior', 'a', 'b')) */
   /* search874_0_0 -> 760 in the cache */
   /* search874_0 -> 1228 in the cache */
   /* search874_1_0 -> 1134 in the cache */
   /* search874_1 -> 2249 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1228, 2249 },
      -1,
   } },

   /* replace874_0_0 -> 760 in the cache */
   /* replace874_0_1 -> 1134 in the cache */
   /* replace874_0 -> 1138 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 1,
      { 1138 },
      -1,
   } },

   /* ('fmul', ('b2f', 'a@1'), ('b2f', 'b@1')) => ('b2f', ('iand', 'a', 'b')) */
   /* search875_0_0 -> 760 in the cache */
   /* search875_0 -> 1129 in the cache */
   /* search875_1_0 -> 1134 in the cache */
   /* search875_1 -> 1135 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 1129, 1135 },
      -1,
   } },

   /* replace875_0_0 -> 760 in the cache */
   /* replace875_0_1 -> 1134 in the cache */
   /* replace875_0 -> 1146 in the cache */
   /* replace875 -> 1266 in the cache */

   /* ('fsat', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1'))) => ('b2f', ('ior', 'a', 'b')) */
   /* search876_0_0_0 -> 760 in the cache */
   /* search876_0_0 -> 1129 in the cache */
   /* search876_0_1_0 -> 1134 in the cache */
   /* search876_0_1 -> 1135 in the cache */
   /* search876_0 -> 1165 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 1165 },
      -1,
   } },

   /* replace876_0_0 -> 760 in the cache */
   /* replace876_0_1 -> 1134 in the cache */
   /* replace876_0 -> 1138 in the cache */
   /* replace876 -> 1263 in the cache */

   /* ('iand', 'a@bool16', 1.0) => ('b2f', 'a') */
   { .variable = {
      { nir_search_value_variable, 16 },
      0, /* a */
      false,
      nir_type_bool,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search877_1 -> 481 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2682, 481 },
      -1,
   } },

   /* replace877_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('iand', 'a@bool32', 1.0) => ('b2f', 'a') */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      false,
      nir_type_bool,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search878_1 -> 243 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2685, 243 },
      -1,
   } },

   /* replace878_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('flt', ('fneg', ('b2f', 'a@1')), 0) => a */
   /* search879_0_0_0 -> 760 in the cache */
   /* search879_0_0 -> 1129 in the cache */
   /* search879_0 -> 1132 in the cache */
   /* search879_1 -> 1229 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1132, 1229 },
      -1,
   } },

   /* replace879 -> 760 in the cache */

   /* ('ilt', 'a', 'a') => False */
   /* search880_0 -> 32 in the cache */
   /* search880_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace880 -> 1077 in the cache */

   /* ('ige', 'a', 'a') => True */
   /* search881_0 -> 32 in the cache */
   /* search881_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace881 -> 1235 in the cache */

   /* ('ieq', 'a', 'a') => True */
   /* search882_0 -> 32 in the cache */
   /* search882_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace882 -> 1235 in the cache */

   /* ('ine', 'a', 'a') => False */
   /* search883_0 -> 32 in the cache */
   /* search883_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace883 -> 1077 in the cache */

   /* ('ult', 'a', 'a') => False */
   /* search884_0 -> 32 in the cache */
   /* search884_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace884 -> 1077 in the cache */

   /* ('uge', 'a', 'a') => True */
   /* search885_0 -> 32 in the cache */
   /* search885_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace885 -> 1235 in the cache */

   /* ('flt', 'a', 'a') => False */
   /* search886_0 -> 32 in the cache */
   /* search886_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace886 -> 1077 in the cache */

   /* ('fge', 'a(is_a_number)', 'a') => True */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      16,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search887_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 2696, 32 },
      -1,
   } },

   /* replace887 -> 1235 in the cache */

   /* ('feq', 'a(is_a_number)', 'a') => True */
   /* search888_0 -> 2696 in the cache */
   /* search888_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      -1, 0,
      { 2696, 32 },
      -1,
   } },

   /* replace888 -> 1235 in the cache */

   /* ('fneu', 'a(is_a_number)', 'a') => False */
   /* search889_0 -> 2696 in the cache */
   /* search889_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      -1, 0,
      { 2696, 32 },
      -1,
   } },

   /* replace889 -> 1077 in the cache */

   /* ('iand', 'a', 'a') => a */
   /* search890_0 -> 32 in the cache */
   /* search890_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace890 -> 32 in the cache */

   /* ('iand', 'a', 0) => 0 */
   /* search891_0 -> 32 in the cache */
   /* search891_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 32, 61 },
      -1,
   } },

   /* replace891 -> 61 in the cache */

   /* ('iand', 'a', -1) => a */
   /* search892_0 -> 32 in the cache */
   /* search892_1 -> 63 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 32, 63 },
      -1,
   } },

   /* replace892 -> 32 in the cache */

   /* ('iand', 'a', ('inot', 'a')) => 0 */
   /* search893_0 -> 32 in the cache */
   /* search893_1_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 32, 2703 },
      -1,
   } },

   /* replace893 -> 61 in the cache */

   /* ('ior', 'a', 'a') => a */
   /* search894_0 -> 32 in the cache */
   /* search894_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace894 -> 32 in the cache */

   /* ('ior', 'a', 0) => a */
   /* search895_0 -> 32 in the cache */
   /* search895_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 32, 61 },
      -1,
   } },

   /* replace895 -> 32 in the cache */

   /* ('ior', 'a', -1) => -1 */
   /* search896_0 -> 32 in the cache */
   /* search896_1 -> 63 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 32, 63 },
      -1,
   } },

   /* replace896 -> 63 in the cache */

   /* ('ior', 'a', ('inot', 'a')) => -1 */
   /* search897_0 -> 32 in the cache */
   /* search897_1_0 -> 32 in the cache */
   /* search897_1 -> 2703 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 32, 2703 },
      -1,
   } },

   /* replace897 -> 63 in the cache */

   /* ('ixor', 'a', 'a') => 0 */
   /* search898_0 -> 32 in the cache */
   /* search898_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ixor,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace898 -> 61 in the cache */

   /* ('ixor', 'a', 0) => a */
   /* search899_0 -> 32 in the cache */
   /* search899_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 32, 61 },
      -1,
   } },

   /* replace899 -> 32 in the cache */

   /* ('ixor', 'a', ('ixor', 'a', 'b')) => b */
   /* search900_0 -> 0 in the cache */
   /* search900_1_0 -> 0 in the cache */
   /* search900_1_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ixor,
      1, 1,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 2,
      { 0, 2711 },
      -1,
   } },

   /* replace900 -> 3 in the cache */

   /* ('ixor', 'a', -1) => ('inot', 'a') */
   /* search901_0 -> 32 in the cache */
   /* search901_1 -> 63 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 32, 63 },
      -1,
   } },

   /* replace901_0 -> 32 in the cache */
   /* replace901 -> 2703 in the cache */

   /* ('inot', ('inot', 'a')) => a */
   /* search902_0_0 -> 32 in the cache */
   /* search902_0 -> 2703 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 2703 },
      -1,
   } },

   /* replace902 -> 32 in the cache */

   /* ('ior', ('iand', 'a', 'b'), 'b') => b */
   /* search903_0_0 -> 0 in the cache */
   /* search903_0_1 -> 3 in the cache */
   /* search903_0 -> 201 in the cache */
   /* search903_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 201, 3 },
      -1,
   } },

   /* replace903 -> 3 in the cache */

   /* ('ior', ('ior', 'a', 'b'), 'b') => ('ior', 'a', 'b') */
   /* search904_0_0 -> 0 in the cache */
   /* search904_0_1 -> 3 in the cache */
   /* search904_0 -> 84 in the cache */
   /* search904_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 84, 3 },
      -1,
   } },

   /* replace904_0 -> 0 in the cache */
   /* replace904_1 -> 3 in the cache */
   /* replace904 -> 2338 in the cache */

   /* ('iand', ('ior', 'a', 'b'), 'b') => b */
   /* search905_0_0 -> 0 in the cache */
   /* search905_0_1 -> 3 in the cache */
   /* search905_0 -> 84 in the cache */
   /* search905_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 84, 3 },
      -1,
   } },

   /* replace905 -> 3 in the cache */

   /* ('iand', ('iand', 'a', 'b'), 'b') => ('iand', 'a', 'b') */
   /* search906_0_0 -> 0 in the cache */
   /* search906_0_1 -> 3 in the cache */
   /* search906_0 -> 201 in the cache */
   /* search906_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 201, 3 },
      -1,
   } },

   /* replace906_0 -> 0 in the cache */
   /* replace906_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('iand', ('iand(is_used_once)', 'a', 'b'), ('iand(is_used_once)', 'a', 'c')) => ('iand', 'a', ('iand', 'b', 'c')) */
   /* search907_0_0 -> 126 in the cache */
   /* search907_0_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 126, 127 },
      0,
   } },
   /* search907_1_0 -> 126 in the cache */
   /* search907_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 126, 129 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 2720, 2721 },
      -1,
   } },

   /* replace907_0 -> 126 in the cache */
   /* replace907_1_0 -> 127 in the cache */
   /* replace907_1_1 -> 129 in the cache */
   /* replace907_1 -> 192 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 126, 192 },
      -1,
   } },

   /* ('iand@64', 'a', '#b(is_lower_half_zero)') => ('pack_64_2x32_split', 0, ('iand', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b'))) */
   /* search908_0 -> 21 in the cache */
   { .variable = {
      { nir_search_value_variable, 64 },
      1, /* b */
      true,
      nir_type_invalid,
      20,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 21, 2724 },
      -1,
   } },

   /* replace908_0 -> 122 in the cache */
   /* replace908_1_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_y,
      -1, 0,
      { 21 },
      -1,
   } },
   /* replace908_1_1_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_y,
      -1, 0,
      { 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2726, 2727 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 1,
      { 122, 2728 },
      -1,
   } },

   /* ('iand@64', 'a', '#b(is_upper_half_zero)') => ('pack_64_2x32_split', ('iand', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_x', 'b')), 0) */
   /* search909_0 -> 21 in the cache */
   { .variable = {
      { nir_search_value_variable, 64 },
      1, /* b */
      true,
      nir_type_invalid,
      21,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 21, 2730 },
      -1,
   } },

   /* replace909_0_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_x,
      -1, 0,
      { 21 },
      -1,
   } },
   /* replace909_0_1_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_x,
      -1, 0,
      { 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2732, 2733 },
      -1,
   } },
   /* replace909_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 1,
      { 2734, 122 },
      -1,
   } },

   /* ('iand@64', 'a', '#b(is_lower_half_negative_one)') => ('pack_64_2x32_split', ('unpack_64_2x32_split_x', 'a'), ('iand', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b'))) */
   /* search910_0 -> 21 in the cache */
   { .variable = {
      { nir_search_value_variable, 64 },
      1, /* b */
      true,
      nir_type_invalid,
      22,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 21, 2736 },
      -1,
   } },

   /* replace910_0_0 -> 21 in the cache */
   /* replace910_0 -> 2732 in the cache */
   /* replace910_1_0_0 -> 21 in the cache */
   /* replace910_1_0 -> 2726 in the cache */
   /* replace910_1_1_0 -> 41 in the cache */
   /* replace910_1_1 -> 2727 in the cache */
   /* replace910_1 -> 2728 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 1,
      { 2732, 2728 },
      -1,
   } },

   /* ('iand@64', 'a', '#b(is_upper_half_negative_one)') => ('pack_64_2x32_split', ('iand', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_x', 'b')), ('unpack_64_2x32_split_y', 'a')) */
   /* search911_0 -> 21 in the cache */
   { .variable = {
      { nir_search_value_variable, 64 },
      1, /* b */
      true,
      nir_type_invalid,
      23,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 21, 2739 },
      -1,
   } },

   /* replace911_0_0_0 -> 21 in the cache */
   /* replace911_0_0 -> 2732 in the cache */
   /* replace911_0_1_0 -> 41 in the cache */
   /* replace911_0_1 -> 2733 in the cache */
   /* replace911_0 -> 2734 in the cache */
   /* replace911_1_0 -> 21 in the cache */
   /* replace911_1 -> 2726 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 1,
      { 2734, 2726 },
      -1,
   } },

   /* ('ior@64', 'a', '#b(is_lower_half_zero)') => ('pack_64_2x32_split', ('unpack_64_2x32_split_x', 'a'), ('ior', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b'))) */
   /* search912_0 -> 21 in the cache */
   /* search912_1 -> 2724 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 21, 2724 },
      -1,
   } },

   /* replace912_0_0 -> 21 in the cache */
   /* replace912_0 -> 2732 in the cache */
   /* replace912_1_0_0 -> 21 in the cache */
   /* replace912_1_0 -> 2726 in the cache */
   /* replace912_1_1_0 -> 41 in the cache */
   /* replace912_1_1 -> 2727 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2726, 2727 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 1,
      { 2732, 2743 },
      -1,
   } },

   /* ('ior@64', 'a', '#b(is_upper_half_zero)') => ('pack_64_2x32_split', ('ior', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_x', 'b')), ('unpack_64_2x32_split_y', 'a')) */
   /* search913_0 -> 21 in the cache */
   /* search913_1 -> 2730 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 21, 2730 },
      -1,
   } },

   /* replace913_0_0_0 -> 21 in the cache */
   /* replace913_0_0 -> 2732 in the cache */
   /* replace913_0_1_0 -> 41 in the cache */
   /* replace913_0_1 -> 2733 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2732, 2733 },
      -1,
   } },
   /* replace913_1_0 -> 21 in the cache */
   /* replace913_1 -> 2726 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 1,
      { 2746, 2726 },
      -1,
   } },

   /* ('ior@64', 'a', '#b(is_lower_half_negative_one)') => ('pack_64_2x32_split', -1, ('ior', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b'))) */
   /* search914_0 -> 21 in the cache */
   /* search914_1 -> 2736 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 21, 2736 },
      -1,
   } },

   /* replace914_0 -> 124 in the cache */
   /* replace914_1_0_0 -> 21 in the cache */
   /* replace914_1_0 -> 2726 in the cache */
   /* replace914_1_1_0 -> 41 in the cache */
   /* replace914_1_1 -> 2727 in the cache */
   /* replace914_1 -> 2743 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 1,
      { 124, 2743 },
      -1,
   } },

   /* ('ior@64', 'a', '#b(is_upper_half_negative_one)') => ('pack_64_2x32_split', ('ior', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_x', 'b')), -1) */
   /* search915_0 -> 21 in the cache */
   /* search915_1 -> 2739 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 21, 2739 },
      -1,
   } },

   /* replace915_0_0_0 -> 21 in the cache */
   /* replace915_0_0 -> 2732 in the cache */
   /* replace915_0_1_0 -> 41 in the cache */
   /* replace915_0_1 -> 2733 in the cache */
   /* replace915_0 -> 2746 in the cache */
   /* replace915_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 1,
      { 2746, 124 },
      -1,
   } },

   /* ('ixor@64', 'a', '#b(is_lower_half_zero)') => ('pack_64_2x32_split', ('unpack_64_2x32_split_x', 'a'), ('ixor', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b'))) */
   /* search916_0 -> 21 in the cache */
   /* search916_1 -> 2724 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 21, 2724 },
      -1,
   } },

   /* replace916_0_0 -> 21 in the cache */
   /* replace916_0 -> 2732 in the cache */
   /* replace916_1_0_0 -> 21 in the cache */
   /* replace916_1_0 -> 2726 in the cache */
   /* replace916_1_1_0 -> 41 in the cache */
   /* replace916_1_1 -> 2727 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 2726, 2727 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 1,
      { 2732, 2753 },
      -1,
   } },

   /* ('ixor@64', 'a', '#b(is_upper_half_zero)') => ('pack_64_2x32_split', ('ixor', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_x', 'b')), ('unpack_64_2x32_split_y', 'a')) */
   /* search917_0 -> 21 in the cache */
   /* search917_1 -> 2730 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 21, 2730 },
      -1,
   } },

   /* replace917_0_0_0 -> 21 in the cache */
   /* replace917_0_0 -> 2732 in the cache */
   /* replace917_0_1_0 -> 41 in the cache */
   /* replace917_0_1 -> 2733 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 2732, 2733 },
      -1,
   } },
   /* replace917_1_0 -> 21 in the cache */
   /* replace917_1 -> 2726 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 1,
      { 2756, 2726 },
      -1,
   } },

   /* ('iand', ('inot', 'a'), ('inot', 'b')) => ('inot', ('ior', 'a', 'b')) */
   /* search918_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 0 },
      -1,
   } },
   /* search918_1_0 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2758, 2759 },
      -1,
   } },

   /* replace918_0_0 -> 0 in the cache */
   /* replace918_0_1 -> 3 in the cache */
   /* replace918_0 -> 2338 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 1,
      { 2338 },
      -1,
   } },

   /* ('ior', ('inot', 'a'), ('inot', 'b')) => ('inot', ('iand', 'a', 'b')) */
   /* search919_0_0 -> 0 in the cache */
   /* search919_0 -> 2758 in the cache */
   /* search919_1_0 -> 3 in the cache */
   /* search919_1 -> 2759 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2758, 2759 },
      -1,
   } },

   /* replace919_0_0 -> 0 in the cache */
   /* replace919_0_1 -> 3 in the cache */
   /* replace919_0 -> 2719 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 1,
      { 2719 },
      -1,
   } },

   /* ('ishl', 0, 'a') => 0 */
   /* search920_0 -> 1229 in the cache */
   /* search920_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 1229, 16 },
      -1,
   } },

   /* replace920 -> 1229 in the cache */

   /* ('ishl', 'a', 0) => a */
   /* search921_0 -> 32 in the cache */
   /* search921_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 32, 122 },
      -1,
   } },

   /* replace921 -> 32 in the cache */

   /* ('ishr', 0, 'a') => 0 */
   /* search922_0 -> 1229 in the cache */
   /* search922_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 1229, 16 },
      -1,
   } },

   /* replace922 -> 1229 in the cache */

   /* ('ishr', -1, 'a') => -1 */
   { .constant = {
      { nir_search_value_constant, 0 },
      nir_type_int, { -0x1 /* -1 */ },
   } },
   /* search923_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 2767, 16 },
      -1,
   } },

   /* replace923 -> 2767 in the cache */

   /* ('ishr', 'a', 0) => a */
   /* search924_0 -> 32 in the cache */
   /* search924_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 32, 122 },
      -1,
   } },

   /* replace924 -> 32 in the cache */

   /* ('ushr', 0, 'a') => 0 */
   /* search925_0 -> 1229 in the cache */
   /* search925_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 1229, 16 },
      -1,
   } },

   /* replace925 -> 1229 in the cache */

   /* ('ushr', 'a', 0) => a */
   /* search926_0 -> 32 in the cache */
   /* search926_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 32, 122 },
      -1,
   } },

   /* replace926 -> 32 in the cache */

   /* ('ior', ('ishl@16', 'a', 'b'), ('ushr@16', 'a', ('iadd', 16, ('ineg', 'b')))) => ('urol', 'a', 'b') */
   /* search927_0_0 -> 11 in the cache */
   /* search927_0_1 -> 36 in the cache */
   /* search927_0 -> 2164 in the cache */
   /* search927_1_0 -> 11 in the cache */
   /* search927_1_1_0 -> 853 in the cache */
   /* search927_1_1_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 853, 2772 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 11, 2773 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 2164, 2774 },
      -1,
   } },

   /* replace927_0 -> 11 in the cache */
   /* replace927_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_urol,
      -1, 0,
      { 11, 36 },
      -1,
   } },

   /* ('ior', ('ishl@16', 'a', 'b'), ('ushr@16', 'a', ('isub', 16, 'b'))) => ('urol', 'a', 'b') */
   /* search928_0_0 -> 11 in the cache */
   /* search928_0_1 -> 36 in the cache */
   /* search928_0 -> 2164 in the cache */
   /* search928_1_0 -> 11 in the cache */
   /* search928_1_1_0 -> 853 in the cache */
   /* search928_1_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 853, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 11, 2777 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2164, 2778 },
      -1,
   } },

   /* replace928_0 -> 11 in the cache */
   /* replace928_1 -> 36 in the cache */
   /* replace928 -> 2776 in the cache */

   /* ('ior', ('ishl@32', 'a', 'b'), ('ushr@32', 'a', ('iadd', 32, ('ineg', 'b')))) => ('urol', 'a', 'b') */
   /* search929_0_0 -> 16 in the cache */
   /* search929_0_1 -> 36 in the cache */
   /* search929_0 -> 946 in the cache */
   /* search929_1_0 -> 16 in the cache */
   /* search929_1_1_0 -> 874 in the cache */
   /* search929_1_1_1_0 -> 36 in the cache */
   /* search929_1_1_1 -> 2772 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 874, 2772 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 16, 2780 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 946, 2781 },
      -1,
   } },

   /* replace929_0 -> 16 in the cache */
   /* replace929_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_urol,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* ('ior', ('ishl@32', 'a', 'b'), ('ushr@32', 'a', ('isub', 32, 'b'))) => ('urol', 'a', 'b') */
   /* search930_0_0 -> 16 in the cache */
   /* search930_0_1 -> 36 in the cache */
   /* search930_0 -> 946 in the cache */
   /* search930_1_0 -> 16 in the cache */
   /* search930_1_1_0 -> 874 in the cache */
   /* search930_1_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 874, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 16, 2784 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 946, 2785 },
      -1,
   } },

   /* replace930_0 -> 16 in the cache */
   /* replace930_1 -> 36 in the cache */
   /* replace930 -> 2783 in the cache */

   /* ('ior', ('ushr@16', 'a', 'b'), ('ishl@16', 'a', ('iadd', 16, ('ineg', 'b')))) => ('uror', 'a', 'b') */
   /* search931_0_0 -> 11 in the cache */
   /* search931_0_1 -> 36 in the cache */
   /* search931_0 -> 2168 in the cache */
   /* search931_1_0 -> 11 in the cache */
   /* search931_1_1_0 -> 853 in the cache */
   /* search931_1_1_1_0 -> 36 in the cache */
   /* search931_1_1_1 -> 2772 in the cache */
   /* search931_1_1 -> 2773 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 11, 2773 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 2168, 2787 },
      -1,
   } },

   /* replace931_0 -> 11 in the cache */
   /* replace931_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_uror,
      -1, 0,
      { 11, 36 },
      -1,
   } },

   /* ('ior', ('ushr@16', 'a', 'b'), ('ishl@16', 'a', ('isub', 16, 'b'))) => ('uror', 'a', 'b') */
   /* search932_0_0 -> 11 in the cache */
   /* search932_0_1 -> 36 in the cache */
   /* search932_0 -> 2168 in the cache */
   /* search932_1_0 -> 11 in the cache */
   /* search932_1_1_0 -> 853 in the cache */
   /* search932_1_1_1 -> 36 in the cache */
   /* search932_1_1 -> 2777 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 11, 2777 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2168, 2790 },
      -1,
   } },

   /* replace932_0 -> 11 in the cache */
   /* replace932_1 -> 36 in the cache */
   /* replace932 -> 2789 in the cache */

   /* ('ior', ('ushr@32', 'a', 'b'), ('ishl@32', 'a', ('iadd', 32, ('ineg', 'b')))) => ('uror', 'a', 'b') */
   /* search933_0_0 -> 16 in the cache */
   /* search933_0_1 -> 36 in the cache */
   /* search933_0 -> 2197 in the cache */
   /* search933_1_0 -> 16 in the cache */
   /* search933_1_1_0 -> 874 in the cache */
   /* search933_1_1_1_0 -> 36 in the cache */
   /* search933_1_1_1 -> 2772 in the cache */
   /* search933_1_1 -> 2780 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 16, 2780 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 2197, 2792 },
      -1,
   } },

   /* replace933_0 -> 16 in the cache */
   /* replace933_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_uror,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* ('ior', ('ushr@32', 'a', 'b'), ('ishl@32', 'a', ('isub', 32, 'b'))) => ('uror', 'a', 'b') */
   /* search934_0_0 -> 16 in the cache */
   /* search934_0_1 -> 36 in the cache */
   /* search934_0 -> 2197 in the cache */
   /* search934_1_0 -> 16 in the cache */
   /* search934_1_1_0 -> 874 in the cache */
   /* search934_1_1_1 -> 36 in the cache */
   /* search934_1_1 -> 2784 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 2784 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2197, 2795 },
      -1,
   } },

   /* replace934_0 -> 16 in the cache */
   /* replace934_1 -> 36 in the cache */
   /* replace934 -> 2794 in the cache */

   /* ('urol@8', 'a', 'b') => ('ior', ('ishl', 'a', 'b'), ('ushr', 'a', ('isub', 8, 'b'))) */
   /* search935_0 -> 6 in the cache */
   /* search935_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_urol,
      -1, 0,
      { 6, 36 },
      -1,
   } },

   /* replace935_0_0 -> 6 in the cache */
   /* replace935_0_1 -> 36 in the cache */
   /* replace935_0 -> 2133 in the cache */
   /* replace935_1_0 -> 6 in the cache */
   /* replace935_1_1_0 -> 831 in the cache */
   /* replace935_1_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 831, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 6, 2798 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2133, 2799 },
      -1,
   } },

   /* ('urol@16', 'a', 'b') => ('ior', ('ishl', 'a', 'b'), ('ushr', 'a', ('isub', 16, 'b'))) */
   /* search936_0 -> 11 in the cache */
   /* search936_1 -> 36 in the cache */
   /* search936 -> 2776 in the cache */

   /* replace936_0_0 -> 11 in the cache */
   /* replace936_0_1 -> 36 in the cache */
   /* replace936_0 -> 2164 in the cache */
   /* replace936_1_0 -> 11 in the cache */
   /* replace936_1_1_0 -> 853 in the cache */
   /* replace936_1_1_1 -> 36 in the cache */
   /* replace936_1_1 -> 2777 in the cache */
   /* replace936_1 -> 2778 in the cache */
   /* replace936 -> 2779 in the cache */

   /* ('urol@32', 'a', 'b') => ('ior', ('ishl', 'a', 'b'), ('ushr', 'a', ('isub', 32, 'b'))) */
   /* search937_0 -> 16 in the cache */
   /* search937_1 -> 36 in the cache */
   /* search937 -> 2783 in the cache */

   /* replace937_0_0 -> 16 in the cache */
   /* replace937_0_1 -> 36 in the cache */
   /* replace937_0 -> 946 in the cache */
   /* replace937_1_0 -> 16 in the cache */
   /* replace937_1_1_0 -> 874 in the cache */
   /* replace937_1_1_1 -> 36 in the cache */
   /* replace937_1_1 -> 2784 in the cache */
   /* replace937_1 -> 2785 in the cache */
   /* replace937 -> 2786 in the cache */

   /* ('urol@64', 'a', 'b') => ('ior', ('ishl', 'a', 'b'), ('ushr', 'a', ('isub', 64, 'b'))) */
   /* search938_0 -> 21 in the cache */
   /* search938_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_urol,
      -1, 0,
      { 21, 36 },
      -1,
   } },

   /* replace938_0_0 -> 21 in the cache */
   /* replace938_0_1 -> 36 in the cache */
   /* replace938_0 -> 2224 in the cache */
   /* replace938_1_0 -> 21 in the cache */
   /* replace938_1_1_0 -> 894 in the cache */
   /* replace938_1_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 894, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 21, 2802 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2224, 2803 },
      -1,
   } },

   /* ('uror@8', 'a', 'b') => ('ior', ('ushr', 'a', 'b'), ('ishl', 'a', ('isub', 8, 'b'))) */
   /* search939_0 -> 6 in the cache */
   /* search939_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_uror,
      -1, 0,
      { 6, 36 },
      -1,
   } },

   /* replace939_0_0 -> 6 in the cache */
   /* replace939_0_1 -> 36 in the cache */
   /* replace939_0 -> 2137 in the cache */
   /* replace939_1_0 -> 6 in the cache */
   /* replace939_1_1_0 -> 831 in the cache */
   /* replace939_1_1_1 -> 36 in the cache */
   /* replace939_1_1 -> 2798 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 6, 2798 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2137, 2806 },
      -1,
   } },

   /* ('uror@16', 'a', 'b') => ('ior', ('ushr', 'a', 'b'), ('ishl', 'a', ('isub', 16, 'b'))) */
   /* search940_0 -> 11 in the cache */
   /* search940_1 -> 36 in the cache */
   /* search940 -> 2789 in the cache */

   /* replace940_0_0 -> 11 in the cache */
   /* replace940_0_1 -> 36 in the cache */
   /* replace940_0 -> 2168 in the cache */
   /* replace940_1_0 -> 11 in the cache */
   /* replace940_1_1_0 -> 853 in the cache */
   /* replace940_1_1_1 -> 36 in the cache */
   /* replace940_1_1 -> 2777 in the cache */
   /* replace940_1 -> 2790 in the cache */
   /* replace940 -> 2791 in the cache */

   /* ('uror@32', 'a', 'b') => ('ior', ('ushr', 'a', 'b'), ('ishl', 'a', ('isub', 32, 'b'))) */
   /* search941_0 -> 16 in the cache */
   /* search941_1 -> 36 in the cache */
   /* search941 -> 2794 in the cache */

   /* replace941_0_0 -> 16 in the cache */
   /* replace941_0_1 -> 36 in the cache */
   /* replace941_0 -> 2197 in the cache */
   /* replace941_1_0 -> 16 in the cache */
   /* replace941_1_1_0 -> 874 in the cache */
   /* replace941_1_1_1 -> 36 in the cache */
   /* replace941_1_1 -> 2784 in the cache */
   /* replace941_1 -> 2795 in the cache */
   /* replace941 -> 2796 in the cache */

   /* ('uror@64', 'a', 'b') => ('ior', ('ushr', 'a', 'b'), ('ishl', 'a', ('isub', 64, 'b'))) */
   /* search942_0 -> 21 in the cache */
   /* search942_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_uror,
      -1, 0,
      { 21, 36 },
      -1,
   } },

   /* replace942_0_0 -> 21 in the cache */
   /* replace942_0_1 -> 36 in the cache */
   /* replace942_0 -> 2228 in the cache */
   /* replace942_1_0 -> 21 in the cache */
   /* replace942_1_1_0 -> 894 in the cache */
   /* replace942_1_1_1 -> 36 in the cache */
   /* replace942_1_1 -> 2802 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 2802 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2228, 2809 },
      -1,
   } },

   /* ('~fexp2', ('flog2', 'a')) => a */
   /* search943_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fexp2,
      -1, 0,
      { 2811 },
      -1,
   } },

   /* replace943 -> 32 in the cache */

   /* ('~flog2', ('fexp2', 'a')) => a */
   /* search944_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 2813 },
      -1,
   } },

   /* replace944 -> 32 in the cache */

   /* ('fpow', 'a', 'b') => ('fexp2', ('fmul', ('flog2', 'a'), 'b')) */
   /* search945_0 -> 0 in the cache */
   /* search945_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* replace945_0_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace945_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 2816, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 1,
      { 2817 },
      -1,
   } },

   /* ('~fexp2', ('fmul', ('flog2', 'a'), 'b')) => ('fpow', 'a', 'b') */
   /* search946_0_0_0 -> 0 in the cache */
   /* search946_0_0 -> 2816 in the cache */
   /* search946_0_1 -> 3 in the cache */
   /* search946_0 -> 2817 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_fexp2,
      -1, 1,
      { 2817 },
      -1,
   } },

   /* replace946_0 -> 0 in the cache */
   /* replace946_1 -> 3 in the cache */
   /* replace946 -> 2815 in the cache */

   /* ('~fexp2', ('fadd', ('fmul', ('flog2', 'a'), 'b'), ('fmul', ('flog2', 'c'), 'd'))) => ('~fmul', ('fpow', 'a', 'b'), ('fpow', 'c', 'd')) */
   /* search947_0_0_0_0 -> 149 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 149 },
      -1,
   } },
   /* search947_0_0_1 -> 150 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 2820, 150 },
      -1,
   } },
   /* search947_0_1_0_0 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 152 },
      -1,
   } },
   /* search947_0_1_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 2822, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 2821, 2823 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_fexp2,
      -1, 3,
      { 2824 },
      -1,
   } },

   /* replace947_0_0 -> 149 in the cache */
   /* replace947_0_1 -> 150 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 149, 150 },
      -1,
   } },
   /* replace947_1_0 -> 152 in the cache */
   /* replace947_1_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 152, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 2826, 2827 },
      -1,
   } },

   /* ('~fexp2', ('fmul', ('flog2', 'a'), 0.5)) => ('fsqrt', 'a') */
   /* search948_0_0_0 -> 32 in the cache */
   /* search948_0_0 -> 2811 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x3fe0000000000000 /* 0.5 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 2811, 2829 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fexp2,
      -1, 1,
      { 2830 },
      -1,
   } },

   /* replace948_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsqrt,
      -1, 0,
      { 32 },
      -1,
   } },

   /* ('~fexp2', ('fmul', ('flog2', 'a'), 2.0)) => ('fmul', 'a', 'a') */
   /* search949_0_0_0 -> 32 in the cache */
   /* search949_0_0 -> 2811 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x4000000000000000 /* 2.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 2811, 2833 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fexp2,
      -1, 1,
      { 2834 },
      -1,
   } },

   /* replace949_0 -> 32 in the cache */
   /* replace949_1 -> 32 in the cache */
   /* replace949 -> 251 in the cache */

   /* ('~fexp2', ('fmul', ('flog2', 'a'), 4.0)) => ('fmul', ('fmul', 'a', 'a'), ('fmul', 'a', 'a')) */
   /* search950_0_0_0 -> 32 in the cache */
   /* search950_0_0 -> 2811 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x4010000000000000 /* 4.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 2811, 2836 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fexp2,
      -1, 1,
      { 2837 },
      -1,
   } },

   /* replace950_0_0 -> 32 in the cache */
   /* replace950_0_1 -> 32 in the cache */
   /* replace950_0 -> 251 in the cache */
   /* replace950_1_0 -> 32 in the cache */
   /* replace950_1_1 -> 32 in the cache */
   /* replace950_1 -> 251 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      -1, 0,
      { 251, 251 },
      -1,
   } },

   /* ('~fpow', 'a', 1.0) => a */
   /* search951_0 -> 32 in the cache */
   /* search951_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 32, 241 },
      -1,
   } },

   /* replace951 -> 32 in the cache */

   /* ('~fpow', 'a', 2.0) => ('fmul', 'a', 'a') */
   /* search952_0 -> 32 in the cache */
   /* search952_1 -> 2833 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 32, 2833 },
      -1,
   } },

   /* replace952_0 -> 32 in the cache */
   /* replace952_1 -> 32 in the cache */
   /* replace952 -> 251 in the cache */

   /* ('~fpow', 'a', 4.0) => ('fmul', ('fmul', 'a', 'a'), ('fmul', 'a', 'a')) */
   /* search953_0 -> 32 in the cache */
   /* search953_1 -> 2836 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 32, 2836 },
      -1,
   } },

   /* replace953_0_0 -> 32 in the cache */
   /* replace953_0_1 -> 32 in the cache */
   /* replace953_0 -> 251 in the cache */
   /* replace953_1_0 -> 32 in the cache */
   /* replace953_1_1 -> 32 in the cache */
   /* replace953_1 -> 251 in the cache */
   /* replace953 -> 2839 in the cache */

   /* ('~fpow', 2.0, 'a') => ('fexp2', 'a') */
   /* search954_0 -> 2833 in the cache */
   /* search954_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 2833, 32 },
      -1,
   } },

   /* replace954_0 -> 32 in the cache */
   /* replace954 -> 2813 in the cache */

   /* ('~fpow', ('fpow', 'a', 2.2), 0.454545) => a */
   /* search955_0_0 -> 32 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x400199999999999a /* 2.2 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 32, 2844 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x3fdd1743e963dc48 /* 0.454545 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 2845, 2846 },
      -1,
   } },

   /* replace955 -> 32 in the cache */

   /* ('~fpow', ('fabs', ('fpow', 'a', 2.2)), 0.454545) => ('fabs', 'a') */
   /* search956_0_0_0 -> 32 in the cache */
   /* search956_0_0_1 -> 2844 in the cache */
   /* search956_0_0 -> 2845 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 2845 },
      -1,
   } },
   /* search956_1 -> 2846 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 2848, 2846 },
      -1,
   } },

   /* replace956_0 -> 32 in the cache */
   /* replace956 -> 103 in the cache */

   /* ('~fsqrt', ('fexp2', 'a')) => ('fexp2', ('fmul', 0.5, 'a')) */
   /* search957_0_0 -> 32 in the cache */
   /* search957_0 -> 2813 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fsqrt,
      -1, 0,
      { 2813 },
      -1,
   } },

   /* replace957_0_0 -> 2829 in the cache */
   /* replace957_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 2829, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 1,
      { 2851 },
      -1,
   } },

   /* ('~frcp', ('fexp2', 'a')) => ('fexp2', ('fneg', 'a')) */
   /* search958_0_0 -> 32 in the cache */
   /* search958_0 -> 2813 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 2813 },
      -1,
   } },

   /* replace958_0_0 -> 32 in the cache */
   /* replace958_0 -> 98 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 0,
      { 98 },
      -1,
   } },

   /* ('~frsq', ('fexp2', 'a')) => ('fexp2', ('fmul', -0.5, 'a')) */
   /* search959_0_0 -> 32 in the cache */
   /* search959_0 -> 2813 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_frsq,
      -1, 0,
      { 2813 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0xbfe0000000000000 /* -0.5 */ },
   } },
   /* replace959_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 2856, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 1,
      { 2857 },
      -1,
   } },

   /* ('~flog2', ('fsqrt', 'a')) => ('fmul', 0.5, ('flog2', 'a')) */
   /* search960_0_0 -> 32 in the cache */
   /* search960_0 -> 2832 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 2832 },
      -1,
   } },

   /* replace960_0 -> 2829 in the cache */
   /* replace960_1_0 -> 32 in the cache */
   /* replace960_1 -> 2811 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 2829, 2811 },
      -1,
   } },

   /* ('~flog2', ('frcp', 'a')) => ('fneg', ('flog2', 'a')) */
   /* search961_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 2861 },
      -1,
   } },

   /* replace961_0_0 -> 32 in the cache */
   /* replace961_0 -> 2811 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 2811 },
      -1,
   } },

   /* ('~flog2', ('frsq', 'a')) => ('fmul', -0.5, ('flog2', 'a')) */
   /* search962_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_frsq,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 2864 },
      -1,
   } },

   /* replace962_0 -> 2856 in the cache */
   /* replace962_1_0 -> 32 in the cache */
   /* replace962_1 -> 2811 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 2856, 2811 },
      -1,
   } },

   /* ('~flog2', ('fpow', 'a', 'b')) => ('fmul', 'b', ('flog2', 'a')) */
   /* search963_0_0 -> 0 in the cache */
   /* search963_0_1 -> 3 in the cache */
   /* search963_0 -> 2815 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 2815 },
      -1,
   } },

   /* replace963_0 -> 3 in the cache */
   /* replace963_1_0 -> 0 in the cache */
   /* replace963_1 -> 2816 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 3, 2816 },
      -1,
   } },

   /* ('~fmul', ('fexp2(is_used_once)', 'a'), ('fexp2(is_used_once)', 'b')) => ('fexp2', ('fadd', 'a', 'b')) */
   /* search964_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 0,
      { 0 },
      0,
   } },
   /* search964_1_0 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 0,
      { 3 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 2869, 2870 },
      -1,
   } },

   /* replace964_0_0 -> 0 in the cache */
   /* replace964_0_1 -> 3 in the cache */
   /* replace964_0 -> 272 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 1,
      { 272 },
      -1,
   } },

   /* ('bcsel', ('flt', 'a', 0.0), 0.0, ('fsqrt', 'a')) => ('fsqrt', ('fmax', 'a', 0.0)) */
   /* search965_0_0 -> 32 in the cache */
   /* search965_0_1 -> 111 in the cache */
   /* search965_0 -> 1172 in the cache */
   /* search965_1 -> 111 in the cache */
   /* search965_2_0 -> 32 in the cache */
   /* search965_2 -> 2832 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1172, 111, 2832 },
      -1,
   } },

   /* replace965_0_0 -> 32 in the cache */
   /* replace965_0_1 -> 111 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 32, 111 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsqrt,
      -1, 1,
      { 2874 },
      -1,
   } },

   /* ('~fmul', ('fsqrt', 'a'), ('fsqrt', 'a')) => ('fabs', 'a') */
   /* search966_0_0 -> 32 in the cache */
   /* search966_0 -> 2832 in the cache */
   /* search966_1_0 -> 32 in the cache */
   /* search966_1 -> 2832 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmul,
      -1, 0,
      { 2832, 2832 },
      -1,
   } },

   /* replace966_0 -> 32 in the cache */
   /* replace966 -> 103 in the cache */

   /* ('~fmulz', ('fsqrt', 'a'), ('fsqrt', 'a')) => ('fabs', 'a') */
   /* search967_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsqrt,
      -1, 0,
      { 16 },
      -1,
   } },
   /* search967_1_0 -> 16 in the cache */
   /* search967_1 -> 2877 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fmulz,
      -1, 0,
      { 2877, 2877 },
      -1,
   } },

   /* replace967_0 -> 16 in the cache */
   /* replace967 -> 696 in the cache */

   /* ('~fdiv', 1.0, 'a') => ('frcp', 'a') */
   /* search968_0 -> 241 in the cache */
   /* search968_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fdiv,
      -1, 0,
      { 241, 32 },
      -1,
   } },

   /* replace968_0 -> 32 in the cache */
   /* replace968 -> 2861 in the cache */

   /* ('fdiv', 'a', 'b') => ('fmul', 'a', ('frcp', 'b')) */
   /* search969_0 -> 0 in the cache */
   /* search969_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdiv,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* replace969_0 -> 0 in the cache */
   /* replace969_1_0 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 0, 2881 },
      -1,
   } },

   /* ('~frcp', ('frcp', 'a')) => a */
   /* search970_0_0 -> 32 in the cache */
   /* search970_0 -> 2861 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 2861 },
      -1,
   } },

   /* replace970 -> 32 in the cache */

   /* ('~frcp', ('fsqrt', 'a')) => ('frsq', 'a') */
   /* search971_0_0 -> 32 in the cache */
   /* search971_0 -> 2832 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 2832 },
      -1,
   } },

   /* replace971_0 -> 32 in the cache */
   /* replace971 -> 2864 in the cache */

   /* ('fsqrt', 'a') => ('frcp', ('frsq', 'a')) */
   /* search972_0 -> 32 in the cache */
   /* search972 -> 2832 in the cache */

   /* replace972_0_0 -> 32 in the cache */
   /* replace972_0 -> 2864 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 2864 },
      -1,
   } },

   /* ('~frcp', ('frsq', 'a')) => ('fsqrt', 'a') */
   /* search973_0_0 -> 32 in the cache */
   /* search973_0 -> 2864 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 2864 },
      -1,
   } },

   /* replace973_0 -> 32 in the cache */
   /* replace973 -> 2832 in the cache */

   /* ('fsin', 'a') => ('ffma', ('ffma', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0)))), 4.0), ('fabs', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0)))), 4.0)), ('fneg', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0)))), 4.0))), 0.225, ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0)))), 4.0)) */
   /* search974_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsin,
      -1, 0,
      { 32 },
      -1,
   } },

   /* replace974_0_0_0_0_0_0 -> 2833 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x3fc45f306dc9c883 /* 0.15915494309189535 */ },
   } },
   /* replace974_0_0_0_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      5, 1,
      { 2888, 32 },
      -1,
   } },
   /* replace974_0_0_0_0_0_1_0_1 -> 2829 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      4, 2,
      { 2889, 2829 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2890 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 3,
      { 2833, 2891 },
      -1,
   } },
   /* replace974_0_0_0_0_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2892, 241 },
      -1,
   } },
   /* replace974_0_0_0_1_0_0_0 -> 2833 in the cache */
   /* replace974_0_0_0_1_0_0_1_0_0_0 -> 2888 in the cache */
   /* replace974_0_0_0_1_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      9, 1,
      { 2888, 32 },
      -1,
   } },
   /* replace974_0_0_0_1_0_0_1_0_1 -> 2829 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      8, 2,
      { 2894, 2829 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2895 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      7, 3,
      { 2833, 2896 },
      -1,
   } },
   /* replace974_0_0_0_1_0_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2897, 241 },
      -1,
   } },
   /* replace974_0_0_0_1_1_0_0_0 -> 2833 in the cache */
   /* replace974_0_0_0_1_1_0_0_1_0_0_0 -> 2888 in the cache */
   /* replace974_0_0_0_1_1_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      12, 1,
      { 2888, 32 },
      -1,
   } },
   /* replace974_0_0_0_1_1_0_0_1_0_1 -> 2829 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      11, 2,
      { 2899, 2829 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2900 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      10, 3,
      { 2833, 2901 },
      -1,
   } },
   /* replace974_0_0_0_1_1_0_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2902, 241 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 3,
      { 2903 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      6, 7,
      { 2898, 2904 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 10,
      { 2893, 2905 },
      -1,
   } },
   /* replace974_0_0_1 -> 2836 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 11,
      { 2906, 2836 },
      -1,
   } },
   /* replace974_0_1_0_0_0_0_0 -> 2833 in the cache */
   /* replace974_0_1_0_0_0_0_1_0_0_0 -> 2888 in the cache */
   /* replace974_0_1_0_0_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      16, 1,
      { 2888, 32 },
      -1,
   } },
   /* replace974_0_1_0_0_0_0_1_0_1 -> 2829 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      15, 2,
      { 2908, 2829 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2909 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      14, 3,
      { 2833, 2910 },
      -1,
   } },
   /* replace974_0_1_0_0_0_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2911, 241 },
      -1,
   } },
   /* replace974_0_1_0_0_1_0_0_0 -> 2833 in the cache */
   /* replace974_0_1_0_0_1_0_0_1_0_0_0 -> 2888 in the cache */
   /* replace974_0_1_0_0_1_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      20, 1,
      { 2888, 32 },
      -1,
   } },
   /* replace974_0_1_0_0_1_0_0_1_0_1 -> 2829 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      19, 2,
      { 2913, 2829 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2914 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      18, 3,
      { 2833, 2915 },
      -1,
   } },
   /* replace974_0_1_0_0_1_0_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2916, 241 },
      -1,
   } },
   /* replace974_0_1_0_0_1_1_0_0_0 -> 2833 in the cache */
   /* replace974_0_1_0_0_1_1_0_0_1_0_0_0 -> 2888 in the cache */
   /* replace974_0_1_0_0_1_1_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      23, 1,
      { 2888, 32 },
      -1,
   } },
   /* replace974_0_1_0_0_1_1_0_0_1_0_1 -> 2829 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      22, 2,
      { 2918, 2829 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2919 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      21, 3,
      { 2833, 2920 },
      -1,
   } },
   /* replace974_0_1_0_0_1_1_0_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2921, 241 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 3,
      { 2922 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      17, 7,
      { 2917, 2923 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 10,
      { 2912, 2924 },
      -1,
   } },
   /* replace974_0_1_0_1 -> 2836 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      13, 11,
      { 2925, 2836 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 11,
      { 2926 },
      -1,
   } },
   /* replace974_0_2_0_0_0_0_0 -> 2833 in the cache */
   /* replace974_0_2_0_0_0_0_1_0_0_0 -> 2888 in the cache */
   /* replace974_0_2_0_0_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      27, 1,
      { 2888, 32 },
      -1,
   } },
   /* replace974_0_2_0_0_0_0_1_0_1 -> 2829 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      26, 2,
      { 2928, 2829 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2929 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      25, 3,
      { 2833, 2930 },
      -1,
   } },
   /* replace974_0_2_0_0_0_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2931, 241 },
      -1,
   } },
   /* replace974_0_2_0_0_1_0_0_0 -> 2833 in the cache */
   /* replace974_0_2_0_0_1_0_0_1_0_0_0 -> 2888 in the cache */
   /* replace974_0_2_0_0_1_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      31, 1,
      { 2888, 32 },
      -1,
   } },
   /* replace974_0_2_0_0_1_0_0_1_0_1 -> 2829 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      30, 2,
      { 2933, 2829 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2934 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      29, 3,
      { 2833, 2935 },
      -1,
   } },
   /* replace974_0_2_0_0_1_0_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2936, 241 },
      -1,
   } },
   /* replace974_0_2_0_0_1_1_0_0_0 -> 2833 in the cache */
   /* replace974_0_2_0_0_1_1_0_0_1_0_0_0 -> 2888 in the cache */
   /* replace974_0_2_0_0_1_1_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      34, 1,
      { 2888, 32 },
      -1,
   } },
   /* replace974_0_2_0_0_1_1_0_0_1_0_1 -> 2829 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      33, 2,
      { 2938, 2829 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2939 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      32, 3,
      { 2833, 2940 },
      -1,
   } },
   /* replace974_0_2_0_0_1_1_0_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2941, 241 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 3,
      { 2942 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      28, 7,
      { 2937, 2943 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 10,
      { 2932, 2944 },
      -1,
   } },
   /* replace974_0_2_0_1 -> 2836 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      24, 11,
      { 2945, 2836 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 11,
      { 2946 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 34,
      { 2907, 2927, 2947 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x3fcccccccccccccd /* 0.225 */ },
   } },
   /* replace974_2_0_0_0_0 -> 2833 in the cache */
   /* replace974_2_0_0_0_1_0_0_0 -> 2888 in the cache */
   /* replace974_2_0_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      38, 1,
      { 2888, 32 },
      -1,
   } },
   /* replace974_2_0_0_0_1_0_1 -> 2829 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      37, 2,
      { 2950, 2829 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2951 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      36, 3,
      { 2833, 2952 },
      -1,
   } },
   /* replace974_2_0_0_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2953, 241 },
      -1,
   } },
   /* replace974_2_0_1_0_0_0 -> 2833 in the cache */
   /* replace974_2_0_1_0_0_1_0_0_0 -> 2888 in the cache */
   /* replace974_2_0_1_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      42, 1,
      { 2888, 32 },
      -1,
   } },
   /* replace974_2_0_1_0_0_1_0_1 -> 2829 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      41, 2,
      { 2955, 2829 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2956 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      40, 3,
      { 2833, 2957 },
      -1,
   } },
   /* replace974_2_0_1_0_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2958, 241 },
      -1,
   } },
   /* replace974_2_0_1_1_0_0_0 -> 2833 in the cache */
   /* replace974_2_0_1_1_0_0_1_0_0_0 -> 2888 in the cache */
   /* replace974_2_0_1_1_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      45, 1,
      { 2888, 32 },
      -1,
   } },
   /* replace974_2_0_1_1_0_0_1_0_1 -> 2829 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      44, 2,
      { 2960, 2829 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2961 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      43, 3,
      { 2833, 2962 },
      -1,
   } },
   /* replace974_2_0_1_1_0_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2963, 241 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 3,
      { 2964 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      39, 7,
      { 2959, 2965 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 10,
      { 2954, 2966 },
      -1,
   } },
   /* replace974_2_1 -> 2836 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      35, 11,
      { 2967, 2836 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 46,
      { 2948, 2949, 2968 },
      -1,
   } },

   /* ('fcos', 'a') => ('ffma', ('ffma', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0)))), 4.0), ('fabs', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0)))), 4.0)), ('fneg', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0)))), 4.0))), 0.225, ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0)))), 4.0)) */
   /* search975_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fcos,
      -1, 0,
      { 32 },
      -1,
   } },

   /* replace975_0_0_0_0_0_0 -> 2833 in the cache */
   /* replace975_0_0_0_0_0_1_0_0_0 -> 2888 in the cache */
   /* replace975_0_0_0_0_0_1_0_0_1 -> 32 in the cache */
   /* replace975_0_0_0_0_0_1_0_0 -> 2889 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x3fe8000000000000 /* 0.75 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      4, 2,
      { 2889, 2971 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2972 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 3,
      { 2833, 2973 },
      -1,
   } },
   /* replace975_0_0_0_0_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2974, 241 },
      -1,
   } },
   /* replace975_0_0_0_1_0_0_0 -> 2833 in the cache */
   /* replace975_0_0_0_1_0_0_1_0_0_0 -> 2888 in the cache */
   /* replace975_0_0_0_1_0_0_1_0_0_1 -> 32 in the cache */
   /* replace975_0_0_0_1_0_0_1_0_0 -> 2894 in the cache */
   /* replace975_0_0_0_1_0_0_1_0_1 -> 2971 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      8, 2,
      { 2894, 2971 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2976 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      7, 3,
      { 2833, 2977 },
      -1,
   } },
   /* replace975_0_0_0_1_0_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2978, 241 },
      -1,
   } },
   /* replace975_0_0_0_1_1_0_0_0 -> 2833 in the cache */
   /* replace975_0_0_0_1_1_0_0_1_0_0_0 -> 2888 in the cache */
   /* replace975_0_0_0_1_1_0_0_1_0_0_1 -> 32 in the cache */
   /* replace975_0_0_0_1_1_0_0_1_0_0 -> 2899 in the cache */
   /* replace975_0_0_0_1_1_0_0_1_0_1 -> 2971 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      11, 2,
      { 2899, 2971 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2980 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      10, 3,
      { 2833, 2981 },
      -1,
   } },
   /* replace975_0_0_0_1_1_0_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2982, 241 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 3,
      { 2983 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      6, 7,
      { 2979, 2984 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 10,
      { 2975, 2985 },
      -1,
   } },
   /* replace975_0_0_1 -> 2836 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 11,
      { 2986, 2836 },
      -1,
   } },
   /* replace975_0_1_0_0_0_0_0 -> 2833 in the cache */
   /* replace975_0_1_0_0_0_0_1_0_0_0 -> 2888 in the cache */
   /* replace975_0_1_0_0_0_0_1_0_0_1 -> 32 in the cache */
   /* replace975_0_1_0_0_0_0_1_0_0 -> 2908 in the cache */
   /* replace975_0_1_0_0_0_0_1_0_1 -> 2971 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      15, 2,
      { 2908, 2971 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2988 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      14, 3,
      { 2833, 2989 },
      -1,
   } },
   /* replace975_0_1_0_0_0_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2990, 241 },
      -1,
   } },
   /* replace975_0_1_0_0_1_0_0_0 -> 2833 in the cache */
   /* replace975_0_1_0_0_1_0_0_1_0_0_0 -> 2888 in the cache */
   /* replace975_0_1_0_0_1_0_0_1_0_0_1 -> 32 in the cache */
   /* replace975_0_1_0_0_1_0_0_1_0_0 -> 2913 in the cache */
   /* replace975_0_1_0_0_1_0_0_1_0_1 -> 2971 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      19, 2,
      { 2913, 2971 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2992 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      18, 3,
      { 2833, 2993 },
      -1,
   } },
   /* replace975_0_1_0_0_1_0_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2994, 241 },
      -1,
   } },
   /* replace975_0_1_0_0_1_1_0_0_0 -> 2833 in the cache */
   /* replace975_0_1_0_0_1_1_0_0_1_0_0_0 -> 2888 in the cache */
   /* replace975_0_1_0_0_1_1_0_0_1_0_0_1 -> 32 in the cache */
   /* replace975_0_1_0_0_1_1_0_0_1_0_0 -> 2918 in the cache */
   /* replace975_0_1_0_0_1_1_0_0_1_0_1 -> 2971 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      22, 2,
      { 2918, 2971 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2996 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      21, 3,
      { 2833, 2997 },
      -1,
   } },
   /* replace975_0_1_0_0_1_1_0_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2998, 241 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 3,
      { 2999 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      17, 7,
      { 2995, 3000 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 10,
      { 2991, 3001 },
      -1,
   } },
   /* replace975_0_1_0_1 -> 2836 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      13, 11,
      { 3002, 2836 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 11,
      { 3003 },
      -1,
   } },
   /* replace975_0_2_0_0_0_0_0 -> 2833 in the cache */
   /* replace975_0_2_0_0_0_0_1_0_0_0 -> 2888 in the cache */
   /* replace975_0_2_0_0_0_0_1_0_0_1 -> 32 in the cache */
   /* replace975_0_2_0_0_0_0_1_0_0 -> 2928 in the cache */
   /* replace975_0_2_0_0_0_0_1_0_1 -> 2971 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      26, 2,
      { 2928, 2971 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 3005 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      25, 3,
      { 2833, 3006 },
      -1,
   } },
   /* replace975_0_2_0_0_0_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 3007, 241 },
      -1,
   } },
   /* replace975_0_2_0_0_1_0_0_0 -> 2833 in the cache */
   /* replace975_0_2_0_0_1_0_0_1_0_0_0 -> 2888 in the cache */
   /* replace975_0_2_0_0_1_0_0_1_0_0_1 -> 32 in the cache */
   /* replace975_0_2_0_0_1_0_0_1_0_0 -> 2933 in the cache */
   /* replace975_0_2_0_0_1_0_0_1_0_1 -> 2971 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      30, 2,
      { 2933, 2971 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 3009 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      29, 3,
      { 2833, 3010 },
      -1,
   } },
   /* replace975_0_2_0_0_1_0_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 3011, 241 },
      -1,
   } },
   /* replace975_0_2_0_0_1_1_0_0_0 -> 2833 in the cache */
   /* replace975_0_2_0_0_1_1_0_0_1_0_0_0 -> 2888 in the cache */
   /* replace975_0_2_0_0_1_1_0_0_1_0_0_1 -> 32 in the cache */
   /* replace975_0_2_0_0_1_1_0_0_1_0_0 -> 2938 in the cache */
   /* replace975_0_2_0_0_1_1_0_0_1_0_1 -> 2971 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      33, 2,
      { 2938, 2971 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 3013 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      32, 3,
      { 2833, 3014 },
      -1,
   } },
   /* replace975_0_2_0_0_1_1_0_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 3015, 241 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 3,
      { 3016 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      28, 7,
      { 3012, 3017 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 10,
      { 3008, 3018 },
      -1,
   } },
   /* replace975_0_2_0_1 -> 2836 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      24, 11,
      { 3019, 2836 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 11,
      { 3020 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 34,
      { 2987, 3004, 3021 },
      -1,
   } },
   /* replace975_1 -> 2949 in the cache */
   /* replace975_2_0_0_0_0 -> 2833 in the cache */
   /* replace975_2_0_0_0_1_0_0_0 -> 2888 in the cache */
   /* replace975_2_0_0_0_1_0_0_1 -> 32 in the cache */
   /* replace975_2_0_0_0_1_0_0 -> 2950 in the cache */
   /* replace975_2_0_0_0_1_0_1 -> 2971 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      37, 2,
      { 2950, 2971 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 3023 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      36, 3,
      { 2833, 3024 },
      -1,
   } },
   /* replace975_2_0_0_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 3025, 241 },
      -1,
   } },
   /* replace975_2_0_1_0_0_0 -> 2833 in the cache */
   /* replace975_2_0_1_0_0_1_0_0_0 -> 2888 in the cache */
   /* replace975_2_0_1_0_0_1_0_0_1 -> 32 in the cache */
   /* replace975_2_0_1_0_0_1_0_0 -> 2955 in the cache */
   /* replace975_2_0_1_0_0_1_0_1 -> 2971 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      41, 2,
      { 2955, 2971 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 3027 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      40, 3,
      { 2833, 3028 },
      -1,
   } },
   /* replace975_2_0_1_0_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 3029, 241 },
      -1,
   } },
   /* replace975_2_0_1_1_0_0_0 -> 2833 in the cache */
   /* replace975_2_0_1_1_0_0_1_0_0_0 -> 2888 in the cache */
   /* replace975_2_0_1_1_0_0_1_0_0_1 -> 32 in the cache */
   /* replace975_2_0_1_1_0_0_1_0_0 -> 2960 in the cache */
   /* replace975_2_0_1_1_0_0_1_0_1 -> 2971 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      44, 2,
      { 2960, 2971 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 3031 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      43, 3,
      { 2833, 3032 },
      -1,
   } },
   /* replace975_2_0_1_1_0_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 3033, 241 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 3,
      { 3034 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      39, 7,
      { 3030, 3035 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 10,
      { 3026, 3036 },
      -1,
   } },
   /* replace975_2_1 -> 2836 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      35, 11,
      { 3037, 2836 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 46,
      { 3022, 2949, 3038 },
      -1,
   } },

   /* ('ieq', 'a', True) => a */
   /* search976_0 -> 760 in the cache */
   /* search976_1 -> 1235 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 760, 1235 },
      -1,
   } },

   /* replace976 -> 760 in the cache */

   /* ('ine(is_not_used_by_if)', 'a', True) => ('inot', 'a') */
   /* search977_0 -> 760 in the cache */
   /* search977_1 -> 1235 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 760, 1235 },
      2,
   } },

   /* replace977_0 -> 760 in the cache */
   /* replace977 -> 1131 in the cache */

   /* ('ine', 'a', False) => a */
   /* search978_0 -> 760 in the cache */
   /* search978_1 -> 1077 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 760, 1077 },
      -1,
   } },

   /* replace978 -> 760 in the cache */

   /* ('ieq(is_not_used_by_if)', 'a', False) => ('inot', 'a') */
   /* search979_0 -> 760 in the cache */
   /* search979_1 -> 1077 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 760, 1077 },
      2,
   } },

   /* replace979_0 -> 760 in the cache */
   /* replace979 -> 1131 in the cache */

   /* ('bcsel', 'a', True, False) => a */
   /* search980_0 -> 760 in the cache */
   /* search980_1 -> 1235 in the cache */
   /* search980_2 -> 1077 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 1235, 1077 },
      -1,
   } },

   /* replace980 -> 760 in the cache */

   /* ('bcsel', 'a', False, True) => ('inot', 'a') */
   /* search981_0 -> 760 in the cache */
   /* search981_1 -> 1077 in the cache */
   /* search981_2 -> 1235 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 1077, 1235 },
      -1,
   } },

   /* replace981_0 -> 760 in the cache */
   /* replace981 -> 1131 in the cache */

   /* ('bcsel', True, 'b', 'c') => b */
   /* search982_0 -> 1235 in the cache */
   /* search982_1 -> 1117 in the cache */
   /* search982_2 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1235, 1117, 684 },
      -1,
   } },

   /* replace982 -> 1117 in the cache */

   /* ('bcsel', False, 'b', 'c') => c */
   /* search983_0 -> 1077 in the cache */
   /* search983_1 -> 1117 in the cache */
   /* search983_2 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1077, 1117, 684 },
      -1,
   } },

   /* replace983 -> 684 in the cache */

   /* ('bcsel@16', 'a', 1.0, 0.0) => ('b2f', 'a') */
   /* search984_0 -> 760 in the cache */
   /* search984_1 -> 481 in the cache */
   /* search984_2 -> 113 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 481, 113 },
      -1,
   } },

   /* replace984_0 -> 760 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 760 },
      -1,
   } },

   /* ('bcsel@16', 'a', 0.0, 1.0) => ('b2f', ('inot', 'a')) */
   /* search985_0 -> 760 in the cache */
   /* search985_1 -> 113 in the cache */
   /* search985_2 -> 481 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 113, 481 },
      -1,
   } },

   /* replace985_0_0 -> 760 in the cache */
   /* replace985_0 -> 1131 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 1131 },
      -1,
   } },

   /* ('bcsel@16', 'a', -1.0, -0.0) => ('fneg', ('b2f', 'a')) */
   /* search986_0 -> 760 in the cache */
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_float, { 0xbff0000000000000 /* -1.0 */ },
   } },
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_float, { 0x8000000000000000 /* -0.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 3052, 3053 },
      -1,
   } },

   /* replace986_0_0 -> 760 in the cache */
   /* replace986_0 -> 3049 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 3049 },
      -1,
   } },

   /* ('bcsel@16', 'a', -0.0, -1.0) => ('fneg', ('b2f', ('inot', 'a'))) */
   /* search987_0 -> 760 in the cache */
   /* search987_1 -> 3053 in the cache */
   /* search987_2 -> 3052 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 3053, 3052 },
      -1,
   } },

   /* replace987_0_0_0 -> 760 in the cache */
   /* replace987_0_0 -> 1131 in the cache */
   /* replace987_0 -> 3051 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 3051 },
      -1,
   } },

   /* ('bcsel@32', 'a', 1.0, 0.0) => ('b2f', 'a') */
   /* search988_0 -> 760 in the cache */
   /* search988_1 -> 243 in the cache */
   /* search988_2 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 243, 115 },
      -1,
   } },

   /* replace988_0 -> 760 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 760 },
      -1,
   } },

   /* ('bcsel@32', 'a', 0.0, 1.0) => ('b2f', ('inot', 'a')) */
   /* search989_0 -> 760 in the cache */
   /* search989_1 -> 115 in the cache */
   /* search989_2 -> 243 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 115, 243 },
      -1,
   } },

   /* replace989_0_0 -> 760 in the cache */
   /* replace989_0 -> 1131 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 1131 },
      -1,
   } },

   /* ('bcsel@32', 'a', -1.0, -0.0) => ('fneg', ('b2f', 'a')) */
   /* search990_0 -> 760 in the cache */
   /* search990_1 -> 278 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0x8000000000000000 /* -0.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 278, 3062 },
      -1,
   } },

   /* replace990_0_0 -> 760 in the cache */
   /* replace990_0 -> 3059 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 3059 },
      -1,
   } },

   /* ('bcsel@32', 'a', -0.0, -1.0) => ('fneg', ('b2f', ('inot', 'a'))) */
   /* search991_0 -> 760 in the cache */
   /* search991_1 -> 3062 in the cache */
   /* search991_2 -> 278 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 3062, 278 },
      -1,
   } },

   /* replace991_0_0_0 -> 760 in the cache */
   /* replace991_0_0 -> 1131 in the cache */
   /* replace991_0 -> 3061 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 3061 },
      -1,
   } },

   /* ('bcsel@64', 'a', 1.0, 0.0) => ('b2f', 'a') */
   /* search992_0 -> 760 in the cache */
   /* search992_1 -> 622 in the cache */
   /* search992_2 -> 701 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 622, 701 },
      -1,
   } },

   /* replace992_0 -> 760 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 760 },
      -1,
   } },

   /* ('bcsel@64', 'a', 0.0, 1.0) => ('b2f', ('inot', 'a')) */
   /* search993_0 -> 760 in the cache */
   /* search993_1 -> 701 in the cache */
   /* search993_2 -> 622 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 701, 622 },
      -1,
   } },

   /* replace993_0_0 -> 760 in the cache */
   /* replace993_0 -> 1131 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 1131 },
      -1,
   } },

   /* ('bcsel@64', 'a', -1.0, -0.0) => ('fneg', ('b2f', 'a')) */
   /* search994_0 -> 760 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_float, { 0xbff0000000000000 /* -1.0 */ },
   } },
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_float, { 0x8000000000000000 /* -0.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 3071, 3072 },
      -1,
   } },

   /* replace994_0_0 -> 760 in the cache */
   /* replace994_0 -> 3068 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 3068 },
      -1,
   } },

   /* ('bcsel@64', 'a', -0.0, -1.0) => ('fneg', ('b2f', ('inot', 'a'))) */
   /* search995_0 -> 760 in the cache */
   /* search995_1 -> 3072 in the cache */
   /* search995_2 -> 3071 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 3072, 3071 },
      -1,
   } },

   /* replace995_0_0_0 -> 760 in the cache */
   /* replace995_0_0 -> 1131 in the cache */
   /* replace995_0 -> 3070 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 3070 },
      -1,
   } },

   /* ('bcsel', 'a', 'b', 'b') => b */
   /* search996_0 -> 760 in the cache */
   /* search996_1 -> 3 in the cache */
   /* search996_2 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 3, 3 },
      -1,
   } },

   /* replace996 -> 3 in the cache */

   /* ('~fcsel', 'a', 'b', 'b') => b */
   /* search997_0 -> 16 in the cache */
   /* search997_1 -> 36 in the cache */
   /* search997_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fcsel,
      -1, 0,
      { 16, 36, 36 },
      -1,
   } },

   /* replace997 -> 36 in the cache */

   /* ('bcsel', 'a', -1, 0) => ('ineg', ('b2i', 'a@1')) */
   /* search998_0 -> 760 in the cache */
   /* search998_1 -> 2767 in the cache */
   /* search998_2 -> 1229 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 2767, 1229 },
      -1,
   } },

   /* replace998_0_0 -> 760 in the cache */
   /* replace998_0 -> 1228 in the cache */
   /* replace998 -> 2248 in the cache */

   /* ('bcsel', 'a', 0, -1) => ('ineg', ('b2i', ('inot', 'a'))) */
   /* search999_0 -> 760 in the cache */
   /* search999_1 -> 1229 in the cache */
   /* search999_2 -> 2767 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 1229, 2767 },
      -1,
   } },

   /* replace999_0_0_0 -> 760 in the cache */
   /* replace999_0_0 -> 1131 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 0,
      { 1131 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 3081 },
      -1,
   } },

   /* ('bcsel', 'a', 1, 0) => ('b2i', 'a@1') */
   /* search1000_0 -> 760 in the cache */
   { .constant = {
      { nir_search_value_constant, 0 },
      nir_type_int, { 0x1 /* 1 */ },
   } },
   /* search1000_2 -> 1229 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 3083, 1229 },
      -1,
   } },

   /* replace1000_0 -> 760 in the cache */
   /* replace1000 -> 1228 in the cache */

   /* ('bcsel', 'a', 0, 1) => ('b2i', ('inot', 'a')) */
   /* search1001_0 -> 760 in the cache */
   /* search1001_1 -> 1229 in the cache */
   /* search1001_2 -> 3083 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 1229, 3083 },
      -1,
   } },

   /* replace1001_0_0 -> 760 in the cache */
   /* replace1001_0 -> 1131 in the cache */
   /* replace1001 -> 3081 in the cache */

   /* ('iand', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('iand', 'a', 'b'))) */
   /* search1002_0_0_0 -> 760 in the cache */
   /* search1002_0_0 -> 1228 in the cache */
   /* search1002_0 -> 2248 in the cache */
   /* search1002_1_0_0 -> 1134 in the cache */
   /* search1002_1_0 -> 2249 in the cache */
   /* search1002_1 -> 2250 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2248, 2250 },
      -1,
   } },

   /* replace1002_0_0_0 -> 760 in the cache */
   /* replace1002_0_0_1 -> 1134 in the cache */
   /* replace1002_0_0 -> 1146 in the cache */
   /* replace1002_0 -> 2676 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 1,
      { 2676 },
      -1,
   } },

   /* ('ior', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('ior', 'a', 'b'))) */
   /* search1003_0_0_0 -> 760 in the cache */
   /* search1003_0_0 -> 1228 in the cache */
   /* search1003_0 -> 2248 in the cache */
   /* search1003_1_0_0 -> 1134 in the cache */
   /* search1003_1_0 -> 2249 in the cache */
   /* search1003_1 -> 2250 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2248, 2250 },
      -1,
   } },

   /* replace1003_0_0_0 -> 760 in the cache */
   /* replace1003_0_0_1 -> 1134 in the cache */
   /* replace1003_0_0 -> 1138 in the cache */
   /* replace1003_0 -> 2679 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 1,
      { 2679 },
      -1,
   } },

   /* ('ieq', ('ineg', ('b2i', 'a@1')), -1) => a */
   /* search1004_0_0_0 -> 760 in the cache */
   /* search1004_0_0 -> 1228 in the cache */
   /* search1004_0 -> 2248 in the cache */
   /* search1004_1 -> 2767 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2248, 2767 },
      -1,
   } },

   /* replace1004 -> 760 in the cache */

   /* ('ine', ('ineg', ('b2i', 'a@1')), -1) => ('inot', 'a') */
   /* search1005_0_0_0 -> 760 in the cache */
   /* search1005_0_0 -> 1228 in the cache */
   /* search1005_0 -> 2248 in the cache */
   /* search1005_1 -> 2767 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2248, 2767 },
      -1,
   } },

   /* replace1005_0 -> 760 in the cache */
   /* replace1005 -> 1131 in the cache */

   /* ('ige', ('ineg', ('b2i', 'a@1')), 0) => ('inot', 'a') */
   /* search1006_0_0_0 -> 760 in the cache */
   /* search1006_0_0 -> 1228 in the cache */
   /* search1006_0 -> 2248 in the cache */
   /* search1006_1 -> 1229 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 2248, 1229 },
      -1,
   } },

   /* replace1006_0 -> 760 in the cache */
   /* replace1006 -> 1131 in the cache */

   /* ('ilt', ('ineg', ('b2i', 'a@1')), 0) => a */
   /* search1007_0_0_0 -> 760 in the cache */
   /* search1007_0_0 -> 1228 in the cache */
   /* search1007_0 -> 2248 in the cache */
   /* search1007_1 -> 1229 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 2248, 1229 },
      -1,
   } },

   /* replace1007 -> 760 in the cache */

   /* ('ult', 0, ('ineg', ('b2i', 'a@1'))) => a */
   /* search1008_0 -> 1229 in the cache */
   /* search1008_1_0_0 -> 760 in the cache */
   /* search1008_1_0 -> 1228 in the cache */
   /* search1008_1 -> 2248 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 1229, 2248 },
      -1,
   } },

   /* replace1008 -> 760 in the cache */

   /* ('iand', ('ineg', ('b2i', 'a')), 1.0) => ('b2f', 'a') */
   /* search1009_0_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 3095 },
      -1,
   } },
   /* search1009_1 -> 1139 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 3096, 1139 },
      -1,
   } },

   /* replace1009_0 -> 32 in the cache */
   /* replace1009 -> 1528 in the cache */

   /* ('iand', ('ineg', ('b2i', 'a')), 1) => ('b2i', 'a') */
   /* search1010_0_0_0 -> 32 in the cache */
   /* search1010_0_0 -> 3095 in the cache */
   /* search1010_0 -> 3096 in the cache */
   /* search1010_1 -> 3083 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 3096, 3083 },
      -1,
   } },

   /* replace1010_0 -> 32 in the cache */
   /* replace1010 -> 3095 in the cache */

   /* ('imax', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('iand', 'a', 'b'))) */
   /* search1011_0_0_0 -> 760 in the cache */
   /* search1011_0_0 -> 1228 in the cache */
   /* search1011_0 -> 2248 in the cache */
   /* search1011_1_0_0 -> 1134 in the cache */
   /* search1011_1_0 -> 2249 in the cache */
   /* search1011_1 -> 2250 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 2248, 2250 },
      -1,
   } },

   /* replace1011_0_0_0 -> 760 in the cache */
   /* replace1011_0_0_1 -> 1134 in the cache */
   /* replace1011_0_0 -> 1146 in the cache */
   /* replace1011_0 -> 2676 in the cache */
   /* replace1011 -> 3087 in the cache */

   /* ('imin', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('ior', 'a', 'b'))) */
   /* search1012_0_0_0 -> 760 in the cache */
   /* search1012_0_0 -> 1228 in the cache */
   /* search1012_0 -> 2248 in the cache */
   /* search1012_1_0_0 -> 1134 in the cache */
   /* search1012_1_0 -> 2249 in the cache */
   /* search1012_1 -> 2250 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 2248, 2250 },
      -1,
   } },

   /* replace1012_0_0_0 -> 760 in the cache */
   /* replace1012_0_0_1 -> 1134 in the cache */
   /* replace1012_0_0 -> 1138 in the cache */
   /* replace1012_0 -> 2679 in the cache */
   /* replace1012 -> 3089 in the cache */

   /* ('umax', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('ior', 'a', 'b'))) */
   /* search1013_0_0_0 -> 760 in the cache */
   /* search1013_0_0 -> 1228 in the cache */
   /* search1013_0 -> 2248 in the cache */
   /* search1013_1_0_0 -> 1134 in the cache */
   /* search1013_1_0 -> 2249 in the cache */
   /* search1013_1 -> 2250 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 2248, 2250 },
      -1,
   } },

   /* replace1013_0_0_0 -> 760 in the cache */
   /* replace1013_0_0_1 -> 1134 in the cache */
   /* replace1013_0_0 -> 1138 in the cache */
   /* replace1013_0 -> 2679 in the cache */
   /* replace1013 -> 3089 in the cache */

   /* ('umin', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('iand', 'a', 'b'))) */
   /* search1014_0_0_0 -> 760 in the cache */
   /* search1014_0_0 -> 1228 in the cache */
   /* search1014_0 -> 2248 in the cache */
   /* search1014_1_0_0 -> 1134 in the cache */
   /* search1014_1_0 -> 2249 in the cache */
   /* search1014_1 -> 2250 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 2248, 2250 },
      -1,
   } },

   /* replace1014_0_0_0 -> 760 in the cache */
   /* replace1014_0_0_1 -> 1134 in the cache */
   /* replace1014_0_0 -> 1146 in the cache */
   /* replace1014_0 -> 2676 in the cache */
   /* replace1014 -> 3087 in the cache */

   /* ('umax', ('b2i', 'a@1'), ('b2i', 'b@1')) => ('b2i', ('ior', 'a', 'b')) */
   /* search1015_0_0 -> 760 in the cache */
   /* search1015_0 -> 1228 in the cache */
   /* search1015_1_0 -> 1134 in the cache */
   /* search1015_1 -> 2249 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 1228, 2249 },
      -1,
   } },

   /* replace1015_0_0 -> 760 in the cache */
   /* replace1015_0_1 -> 1134 in the cache */
   /* replace1015_0 -> 1138 in the cache */
   /* replace1015 -> 2679 in the cache */

   /* ('umin', ('b2i', 'a@1'), ('b2i', 'b@1')) => ('b2i', ('iand', 'a', 'b')) */
   /* search1016_0_0 -> 760 in the cache */
   /* search1016_0 -> 1228 in the cache */
   /* search1016_1_0 -> 1134 in the cache */
   /* search1016_1 -> 2249 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 1228, 2249 },
      -1,
   } },

   /* replace1016_0_0 -> 760 in the cache */
   /* replace1016_0_1 -> 1134 in the cache */
   /* replace1016_0 -> 1146 in the cache */
   /* replace1016 -> 2676 in the cache */

   /* ('ine', ('umin', ('ineg', ('b2i', 'a@1')), 'b'), 0) => ('iand', 'a', ('ine', 'b', 0)) */
   /* search1017_0_0_0_0 -> 760 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 0,
      { 760 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 3105 },
      -1,
   } },
   /* search1017_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 3106, 3 },
      -1,
   } },
   /* search1017_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 3107, 86 },
      -1,
   } },

   /* replace1017_0 -> 760 in the cache */
   /* replace1017_1_0 -> 3 in the cache */
   /* replace1017_1_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 3, 86 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 760, 3109 },
      -1,
   } },

   /* ('ine', ('umax', ('ineg', ('b2i', 'a@1')), 'b'), 0) => ('ior', 'a', ('ine', 'b', 0)) */
   /* search1018_0_0_0_0 -> 760 in the cache */
   /* search1018_0_0_0 -> 3105 in the cache */
   /* search1018_0_0 -> 3106 in the cache */
   /* search1018_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umax,
      1, 1,
      { 3106, 3 },
      -1,
   } },
   /* search1018_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 3111, 86 },
      -1,
   } },

   /* replace1018_0 -> 760 in the cache */
   /* replace1018_1_0 -> 3 in the cache */
   /* replace1018_1_1 -> 86 in the cache */
   /* replace1018_1 -> 3109 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 760, 3109 },
      -1,
   } },

   /* ('f2i', ('ftrunc', 'a')) => ('f2i', 'a') */
   /* search1019_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ftrunc,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2i,
      -1, 0,
      { 3114 },
      -1,
   } },

   /* replace1019_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2i,
      -1, 0,
      { 32 },
      -1,
   } },

   /* ('f2u', ('ftrunc', 'a')) => ('f2u', 'a') */
   /* search1020_0_0 -> 32 in the cache */
   /* search1020_0 -> 3114 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2u,
      -1, 0,
      { 3114 },
      -1,
   } },

   /* replace1020_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2u,
      -1, 0,
      { 32 },
      -1,
   } },

   /* ('inot', ('f2b1', 'a')) => ('feq', 'a', 0.0) */
   /* search1021_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_f2b1,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 3119 },
      -1,
   } },

   /* replace1021_0 -> 32 in the cache */
   /* replace1021_1 -> 111 in the cache */
   /* replace1021 -> 1257 in the cache */

   /* ('f2fmp', ('f2f32', 'a@16')) => a */
   /* search1022_0_0 -> 11 in the cache */
   /* search1022_0 -> 1802 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 1802 },
      -1,
   } },

   /* replace1022 -> 11 in the cache */

   /* ('i2imp', ('i2i32', 'a@16')) => a */
   /* search1023_0_0 -> 11 in the cache */
   /* search1023_0 -> 1804 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 1804 },
      -1,
   } },

   /* replace1023 -> 11 in the cache */

   /* ('i2imp', ('u2u32', 'a@16')) => a */
   /* search1024_0_0 -> 11 in the cache */
   /* search1024_0 -> 1806 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 1806 },
      -1,
   } },

   /* replace1024 -> 11 in the cache */

   /* ('f2imp', ('f2f32', 'a@16')) => ('f2i16', 'a') */
   /* search1025_0_0 -> 11 in the cache */
   /* search1025_0 -> 1802 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2imp,
      -1, 0,
      { 1802 },
      -1,
   } },

   /* replace1025_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2i16,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('f2ump', ('f2f32', 'a@16')) => ('f2u16', 'a') */
   /* search1026_0_0 -> 11 in the cache */
   /* search1026_0 -> 1802 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2ump,
      -1, 0,
      { 1802 },
      -1,
   } },

   /* replace1026_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2u16,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('i2fmp', ('i2i32', 'a@16')) => ('i2f16', 'a') */
   /* search1027_0_0 -> 11 in the cache */
   /* search1027_0 -> 1804 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2fmp,
      -1, 0,
      { 1804 },
      -1,
   } },

   /* replace1027_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2f16,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('u2fmp', ('u2u32', 'a@16')) => ('u2f16', 'a') */
   /* search1028_0_0 -> 11 in the cache */
   /* search1028_0 -> 1806 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2fmp,
      -1, 0,
      { 1806 },
      -1,
   } },

   /* replace1028_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2f16,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('f2fmp', ('b2f32', 'a@1')) => ('b2f16', 'a') */
   /* search1029_0_0 -> 760 in the cache */
   /* search1029_0 -> 1408 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 1408 },
      -1,
   } },

   /* replace1029_0 -> 760 in the cache */
   /* replace1029 -> 1404 in the cache */

   /* ('i2imp', ('b2i32', 'a@1')) => ('b2i16', 'a') */
   /* search1030_0_0 -> 760 in the cache */
   /* search1030_0 -> 1396 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 1396 },
      -1,
   } },

   /* replace1030_0 -> 760 in the cache */
   /* replace1030 -> 1392 in the cache */

   /* ('i2imp', ('b2i32', 'a@1')) => ('b2i16', 'a') */
   /* search1031_0_0 -> 760 in the cache */
   /* search1031_0 -> 1396 in the cache */
   /* search1031 -> 3133 in the cache */

   /* replace1031_0 -> 760 in the cache */
   /* replace1031 -> 1392 in the cache */

   /* ('f2imp', ('b2f32', 'a@1')) => ('b2i16', 'a') */
   /* search1032_0_0 -> 760 in the cache */
   /* search1032_0 -> 1408 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2imp,
      -1, 0,
      { 1408 },
      -1,
   } },

   /* replace1032_0 -> 760 in the cache */
   /* replace1032 -> 1392 in the cache */

   /* ('f2ump', ('b2f32', 'a@1')) => ('b2i16', 'a') */
   /* search1033_0_0 -> 760 in the cache */
   /* search1033_0 -> 1408 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2ump,
      -1, 0,
      { 1408 },
      -1,
   } },

   /* replace1033_0 -> 760 in the cache */
   /* replace1033 -> 1392 in the cache */

   /* ('i2fmp', ('b2i32', 'a@1')) => ('b2f16', 'a') */
   /* search1034_0_0 -> 760 in the cache */
   /* search1034_0 -> 1396 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2fmp,
      -1, 0,
      { 1396 },
      -1,
   } },

   /* replace1034_0 -> 760 in the cache */
   /* replace1034 -> 1404 in the cache */

   /* ('u2fmp', ('b2i32', 'a@1')) => ('b2f16', 'a') */
   /* search1035_0_0 -> 760 in the cache */
   /* search1035_0 -> 1396 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2fmp,
      -1, 0,
      { 1396 },
      -1,
   } },

   /* replace1035_0 -> 760 in the cache */
   /* replace1035 -> 1404 in the cache */

   /* ('f2f32', ('f2fmp', 'a@32')) => a */
   /* search1036_0_0 -> 16 in the cache */
   /* search1036_0 -> 1993 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 1993 },
      -1,
   } },

   /* replace1036 -> 16 in the cache */

   /* ('i2i32', ('i2imp', 'a@32')) => a */
   /* search1037_0_0 -> 16 in the cache */
   /* search1037_0 -> 2004 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 2004 },
      -1,
   } },

   /* replace1037 -> 16 in the cache */

   /* ('u2u32', ('i2imp', 'a@32')) => a */
   /* search1038_0_0 -> 16 in the cache */
   /* search1038_0 -> 2004 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 2004 },
      -1,
   } },

   /* replace1038 -> 16 in the cache */

   /* ('i2i32', ('f2imp', 'a@32')) => ('f2i32', 'a') */
   /* search1039_0_0 -> 16 in the cache */
   /* search1039_0 -> 2003 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 2003 },
      -1,
   } },

   /* replace1039_0 -> 16 in the cache */
   /* replace1039 -> 2001 in the cache */

   /* ('u2u32', ('f2ump', 'a@32')) => ('f2u32', 'a') */
   /* search1040_0_0 -> 16 in the cache */
   /* search1040_0 -> 2000 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 2000 },
      -1,
   } },

   /* replace1040_0 -> 16 in the cache */
   /* replace1040 -> 1998 in the cache */

   /* ('f2f32', ('i2fmp', 'a@32')) => ('i2f32', 'a') */
   /* search1041_0_0 -> 16 in the cache */
   /* search1041_0 -> 1992 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 1992 },
      -1,
   } },

   /* replace1041_0 -> 16 in the cache */
   /* replace1041 -> 1990 in the cache */

   /* ('f2f32', ('u2fmp', 'a@32')) => ('u2f32', 'a') */
   /* search1042_0_0 -> 16 in the cache */
   /* search1042_0 -> 1989 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 1989 },
      -1,
   } },

   /* replace1042_0 -> 16 in the cache */
   /* replace1042 -> 1987 in the cache */

   /* ('~f2f32', ('f2f64', 'a@32')) => a */
   /* search1043_0_0 -> 16 in the cache */
   /* search1043_0 -> 1907 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 1907 },
      -1,
   } },

   /* replace1043 -> 16 in the cache */

   /* ('ffloor', 'a(is_integral)') => a */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      24,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 0,
      { 3146 },
      -1,
   } },

   /* replace1044 -> 32 in the cache */

   /* ('fceil', 'a(is_integral)') => a */
   /* search1045_0 -> 3146 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fceil,
      -1, 0,
      { 3146 },
      -1,
   } },

   /* replace1045 -> 32 in the cache */

   /* ('ftrunc', 'a(is_integral)') => a */
   /* search1046_0 -> 3146 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ftrunc,
      -1, 0,
      { 3146 },
      -1,
   } },

   /* replace1046 -> 32 in the cache */

   /* ('fround_even', 'a(is_integral)') => a */
   /* search1047_0 -> 3146 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fround_even,
      -1, 0,
      { 3146 },
      -1,
   } },

   /* replace1047 -> 32 in the cache */

   /* ('~ffract', 'a(is_integral)') => 0.0 */
   /* search1048_0 -> 3146 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_ffract,
      -1, 0,
      { 3146 },
      -1,
   } },

   /* replace1048 -> 111 in the cache */

   /* ('fabs', 'a(is_not_negative)') => a */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      19,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 3152 },
      -1,
   } },

   /* replace1049 -> 32 in the cache */

   /* ('iabs', 'a(is_not_negative)') => a */
   /* search1050_0 -> 3152 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 3152 },
      -1,
   } },

   /* replace1050 -> 32 in the cache */

   /* ('fsat', 'a(is_not_positive)') => 0.0 */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      25,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 3155 },
      -1,
   } },

   /* replace1051 -> 111 in the cache */

   /* ('~fmin', 'a(is_not_negative)', 1.0) => ('fsat', 'a') */
   /* search1052_0 -> 3152 in the cache */
   /* search1052_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 3152, 241 },
      -1,
   } },

   /* replace1052_0 -> 32 in the cache */
   /* replace1052 -> 217 in the cache */

   /* ('flt', ('fadd', ('fmul', ('fsat', 'a'), ('fneg', ('fsat', 'a'))), 1.0), 0.0) => False */
   /* search1053_0_0_0_0 -> 32 in the cache */
   /* search1053_0_0_0 -> 217 in the cache */
   /* search1053_0_0_1_0_0 -> 32 in the cache */
   /* search1053_0_0_1_0 -> 217 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 217 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 217, 3158 },
      -1,
   } },
   /* search1053_0_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 3159, 241 },
      -1,
   } },
   /* search1053_1 -> 111 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 2,
      { 3160, 111 },
      -1,
   } },

   /* replace1053 -> 1077 in the cache */

   /* ('flt', ('fadd', ('fneg', ('fmul', ('fsat', 'a'), ('fsat', 'a'))), 1.0), 0.0) => False */
   /* search1054_0_0_0_0_0 -> 32 in the cache */
   /* search1054_0_0_0_0 -> 217 in the cache */
   /* search1054_0_0_0_1_0 -> 32 in the cache */
   /* search1054_0_0_0_1 -> 217 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      -1, 0,
      { 217, 217 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 3162 },
      -1,
   } },
   /* search1054_0_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 3163, 241 },
      -1,
   } },
   /* search1054_1 -> 111 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 3164, 111 },
      -1,
   } },

   /* replace1054 -> 1077 in the cache */

   /* ('fmax', ('fadd', ('fmul', ('fsat', 'a'), ('fneg', ('fsat', 'a'))), 1.0), 0.0) => ('fadd', ('fmul', ('fsat', 'a'), ('fneg', ('fsat', 'a'))), 1.0) */
   /* search1055_0_0_0_0 -> 32 in the cache */
   /* search1055_0_0_0 -> 217 in the cache */
   /* search1055_0_0_1_0_0 -> 32 in the cache */
   /* search1055_0_0_1_0 -> 217 in the cache */
   /* search1055_0_0_1 -> 3158 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 217, 3158 },
      -1,
   } },
   /* search1055_0_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 2,
      { 3166, 241 },
      -1,
   } },
   /* search1055_1 -> 111 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 3,
      { 3167, 111 },
      -1,
   } },

   /* replace1055_0_0_0 -> 32 in the cache */
   /* replace1055_0_0 -> 217 in the cache */
   /* replace1055_0_1_0_0 -> 32 in the cache */
   /* replace1055_0_1_0 -> 217 in the cache */
   /* replace1055_0_1 -> 3158 in the cache */
   /* replace1055_0 -> 3159 in the cache */
   /* replace1055_1 -> 241 in the cache */
   /* replace1055 -> 3160 in the cache */

   /* ('fmax', ('fadd', ('fneg', ('fmul', ('fsat', 'a'), ('fsat', 'a'))), 1.0), 0.0) => ('fadd', ('fneg', ('fmul', ('fsat', 'a'), ('fsat', 'a'))), 1.0) */
   /* search1056_0_0_0_0_0 -> 32 in the cache */
   /* search1056_0_0_0_0 -> 217 in the cache */
   /* search1056_0_0_0_1_0 -> 32 in the cache */
   /* search1056_0_0_0_1 -> 217 in the cache */
   /* search1056_0_0_0 -> 3162 in the cache */
   /* search1056_0_0 -> 3163 in the cache */
   /* search1056_0_1 -> 241 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 3163, 241 },
      -1,
   } },
   /* search1056_1 -> 111 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 2,
      { 3169, 111 },
      -1,
   } },

   /* replace1056_0_0_0_0 -> 32 in the cache */
   /* replace1056_0_0_0 -> 217 in the cache */
   /* replace1056_0_0_1_0 -> 32 in the cache */
   /* replace1056_0_0_1 -> 217 in the cache */
   /* replace1056_0_0 -> 3162 in the cache */
   /* replace1056_0 -> 3163 in the cache */
   /* replace1056_1 -> 241 in the cache */
   /* replace1056 -> 3164 in the cache */

   /* ('fneu', 'a(is_not_zero)', 0.0) => True */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      26,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1057_1 -> 111 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 3171, 111 },
      -1,
   } },

   /* replace1057 -> 1235 in the cache */

   /* ('feq', 'a(is_not_zero)', 0.0) => False */
   /* search1058_0 -> 3171 in the cache */
   /* search1058_1 -> 111 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 3171, 111 },
      -1,
   } },

   /* replace1058 -> 1077 in the cache */

   /* ('fge', 'a(is_a_number_not_negative)', 'b(is_a_number_not_positive)') => True */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      27,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      28,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 3174, 3175 },
      -1,
   } },

   /* replace1059 -> 1235 in the cache */

   /* ('fge', 'a(is_not_positive)', 'b(is_gt_zero)') => False */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      25,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      29,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 3177, 3178 },
      -1,
   } },

   /* replace1060 -> 1077 in the cache */

   /* ('fge', 'a(is_lt_zero)', 'b(is_not_negative)') => False */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      30,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      19,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 3180, 3181 },
      -1,
   } },

   /* replace1061 -> 1077 in the cache */

   /* ('flt', 'a(is_not_negative)', 'b(is_not_positive)') => False */
   /* search1062_0 -> 1584 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      25,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1584, 3183 },
      -1,
   } },

   /* replace1062 -> 1077 in the cache */

   /* ('flt', 'a(is_a_number_not_positive)', 'b(is_a_number_gt_zero)') => True */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      28,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      31,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 3185, 3186 },
      -1,
   } },

   /* replace1063 -> 1235 in the cache */

   /* ('flt', 'a(is_a_number_lt_zero)', 'b(is_a_number_not_negative)') => True */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      32,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      27,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 3188, 3189 },
      -1,
   } },

   /* replace1064 -> 1235 in the cache */

   /* ('ine', 'a(is_not_zero)', 0) => True */
   /* search1065_0 -> 3171 in the cache */
   /* search1065_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 3171, 61 },
      -1,
   } },

   /* replace1065 -> 1235 in the cache */

   /* ('ieq', 'a(is_not_zero)', 0) => False */
   /* search1066_0 -> 3171 in the cache */
   /* search1066_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 3171, 61 },
      -1,
   } },

   /* replace1066 -> 1077 in the cache */

   /* ('ige', 'a(is_not_negative)', 'b(is_not_positive)') => True */
   /* search1067_0 -> 1584 in the cache */
   /* search1067_1 -> 3183 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 1584, 3183 },
      -1,
   } },

   /* replace1067 -> 1235 in the cache */

   /* ('ige', 'a(is_not_positive)', 'b(is_gt_zero)') => False */
   /* search1068_0 -> 3177 in the cache */
   /* search1068_1 -> 3178 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3177, 3178 },
      -1,
   } },

   /* replace1068 -> 1077 in the cache */

   /* ('ige', 'a(is_lt_zero)', 'b(is_not_negative)') => False */
   /* search1069_0 -> 3180 in the cache */
   /* search1069_1 -> 3181 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3180, 3181 },
      -1,
   } },

   /* replace1069 -> 1077 in the cache */

   /* ('ilt', 'a(is_not_negative)', 'b(is_not_positive)') => False */
   /* search1070_0 -> 1584 in the cache */
   /* search1070_1 -> 3183 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 1584, 3183 },
      -1,
   } },

   /* replace1070 -> 1077 in the cache */

   /* ('ilt', 'a(is_not_positive)', 'b(is_gt_zero)') => True */
   /* search1071_0 -> 3177 in the cache */
   /* search1071_1 -> 3178 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 3177, 3178 },
      -1,
   } },

   /* replace1071 -> 1235 in the cache */

   /* ('ilt', 'a(is_lt_zero)', 'b(is_not_negative)') => True */
   /* search1072_0 -> 3180 in the cache */
   /* search1072_1 -> 3181 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 3180, 3181 },
      -1,
   } },

   /* replace1072 -> 1235 in the cache */

   /* ('ult', 0, 'a(is_gt_zero)') => True */
   /* search1073_0 -> 61 in the cache */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      29,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 61, 3199 },
      -1,
   } },

   /* replace1073 -> 1235 in the cache */

   /* ('ult', 'a', 0) => False */
   /* search1074_0 -> 32 in the cache */
   /* search1074_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 32, 61 },
      -1,
   } },

   /* replace1074 -> 1077 in the cache */

   /* ('unpack_64_2x32_split_x', ('pack_64_2x32_split', 'a', 'b')) => a */
   /* search1075_0_0 -> 16 in the cache */
   /* search1075_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 0,
      { 16, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_x,
      -1, 0,
      { 3202 },
      -1,
   } },

   /* replace1075 -> 16 in the cache */

   /* ('unpack_64_2x32_split_y', ('pack_64_2x32_split', 'a', 'b')) => b */
   /* search1076_0_0 -> 16 in the cache */
   /* search1076_0_1 -> 36 in the cache */
   /* search1076_0 -> 3202 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_y,
      -1, 0,
      { 3202 },
      -1,
   } },

   /* replace1076 -> 36 in the cache */

   /* ('unpack_64_2x32_split_x', ('pack_64_2x32', 'a')) => a.x */
   /* search1077_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_x,
      -1, 0,
      { 3205 },
      -1,
   } },

   /* replace1077 -> 2427 in the cache */

   /* ('unpack_64_2x32_split_y', ('pack_64_2x32', 'a')) => a.y */
   /* search1078_0_0 -> 16 in the cache */
   /* search1078_0 -> 3205 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_y,
      -1, 0,
      { 3205 },
      -1,
   } },

   /* replace1078 -> 2430 in the cache */

   /* ('unpack_64_2x32_split_x', ('u2u64', 'a@32')) => a */
   /* search1079_0_0 -> 16 in the cache */
   /* search1079_0 -> 1911 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_x,
      -1, 0,
      { 1911 },
      -1,
   } },

   /* replace1079 -> 16 in the cache */

   /* ('unpack_64_2x32_split_y', ('u2u64', 'a')) => 0 */
   /* search1080_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_y,
      -1, 0,
      { 3209 },
      -1,
   } },

   /* replace1080 -> 122 in the cache */

   /* ('unpack_64_2x32_split_x', ('i2i64', 'a@32')) => a */
   /* search1081_0_0 -> 16 in the cache */
   /* search1081_0 -> 1909 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_x,
      -1, 0,
      { 1909 },
      -1,
   } },

   /* replace1081 -> 16 in the cache */

   /* ('unpack_64_2x32_split_y', ('i2i64(is_used_once)', 'a@32')) => ('ishr', 'a', 31) */
   /* search1082_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 16 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_y,
      -1, 0,
      { 3212 },
      -1,
   } },

   /* replace1082_0 -> 16 in the cache */
   /* replace1082_1 -> 19 in the cache */
   /* replace1082 -> 1460 in the cache */

   /* ('unpack_64_2x32', ('pack_64_2x32_split', 'a', 'b')) => ('vec2', 'a', 'b') */
   /* search1083_0_0 -> 16 in the cache */
   /* search1083_0_1 -> 36 in the cache */
   /* search1083_0 -> 3202 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32,
      -1, 0,
      { 3202 },
      -1,
   } },

   /* replace1083_0 -> 16 in the cache */
   /* replace1083_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* ('unpack_64_2x32', ('pack_64_2x32', 'a')) => a */
   /* search1084_0_0 -> 16 in the cache */
   /* search1084_0 -> 3205 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32,
      -1, 0,
      { 3205 },
      -1,
   } },

   /* replace1084 -> 16 in the cache */

   /* ('unpack_double_2x32_dxil', ('pack_double_2x32_dxil', 'a')) => a */
   /* search1085_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_double_2x32_dxil,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_double_2x32_dxil,
      -1, 0,
      { 3217 },
      -1,
   } },

   /* replace1085 -> 16 in the cache */

   /* ('pack_64_2x32_split', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_y', 'a')) => a */
   /* search1086_0_0 -> 21 in the cache */
   /* search1086_0 -> 2732 in the cache */
   /* search1086_1_0 -> 21 in the cache */
   /* search1086_1 -> 2726 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 0,
      { 2732, 2726 },
      -1,
   } },

   /* replace1086 -> 21 in the cache */

   /* ('pack_64_2x32', ('vec2', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_y', 'a'))) => a */
   /* search1087_0_0_0 -> 21 in the cache */
   /* search1087_0_0 -> 2732 in the cache */
   /* search1087_0_1_0 -> 21 in the cache */
   /* search1087_0_1 -> 2726 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 2732, 2726 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32,
      -1, 0,
      { 3220 },
      -1,
   } },

   /* replace1087 -> 21 in the cache */

   /* ('pack_64_2x32', ('unpack_64_2x32', 'a')) => a */
   /* search1088_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32,
      -1, 0,
      { 21 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32,
      -1, 0,
      { 3222 },
      -1,
   } },

   /* replace1088 -> 21 in the cache */

   /* ('pack_double_2x32_dxil', ('unpack_double_2x32_dxil', 'a')) => a */
   /* search1089_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_double_2x32_dxil,
      -1, 0,
      { 21 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_double_2x32_dxil,
      -1, 0,
      { 3224 },
      -1,
   } },

   /* replace1089 -> 21 in the cache */

   /* ('iand', ('ieq', ('unpack_32_2x16_split_x', 'a'), '#b'), ('ieq', ('unpack_32_2x16_split_y', 'a'), '#c')) => ('ieq', 'a', ('pack_32_2x16_split', 'b', 'c')) */
   /* search1090_0_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_unpack_32_2x16_split_x,
      -1, 0,
      { 16 },
      -1,
   } },
   /* search1090_0_1 -> 2011 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 3226, 2011 },
      -1,
   } },
   /* search1090_1_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_unpack_32_2x16_split_y,
      -1, 0,
      { 16 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 16 },
      2, /* c */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 3228, 3229 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 3227, 3230 },
      -1,
   } },

   /* replace1090_0 -> 16 in the cache */
   /* replace1090_1_0 -> 258 in the cache */
   /* replace1090_1_1 -> 469 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_32_2x16_split,
      -1, 0,
      { 258, 469 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 16, 3232 },
      -1,
   } },

   /* ('ushr', 'a@16', 8) => ('extract_u8', 'a', 1) */
   /* search1091_0 -> 11 in the cache */
   /* search1091_1 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 11, 831 },
      -1,
   } },

   /* replace1091_0 -> 11 in the cache */
   /* replace1091_1 -> 1452 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 11, 1452 },
      -1,
   } },

   /* ('ushr', 'a@32', 24) => ('extract_u8', 'a', 3) */
   /* search1092_0 -> 16 in the cache */
   /* search1092_1 -> 1012 in the cache */
   /* search1092 -> 1017 in the cache */

   /* replace1092_0 -> 16 in the cache */
   /* replace1092_1 -> 405 in the cache */
   /* replace1092 -> 422 in the cache */

   /* ('ushr', 'a@64', 56) => ('extract_u8', 'a', 7) */
   /* search1093_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x38 /* 56 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 21, 3236 },
      -1,
   } },

   /* replace1093_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x7 /* 7 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 21, 3238 },
      -1,
   } },

   /* ('ishr', 'a@16', 8) => ('extract_i8', 'a', 1) */
   /* search1094_0 -> 11 in the cache */
   /* search1094_1 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 11, 831 },
      -1,
   } },

   /* replace1094_0 -> 11 in the cache */
   /* replace1094_1 -> 1452 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 11, 1452 },
      -1,
   } },

   /* ('ishr', 'a@32', 24) => ('extract_i8', 'a', 3) */
   /* search1095_0 -> 16 in the cache */
   /* search1095_1 -> 1012 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 16, 1012 },
      -1,
   } },

   /* replace1095_0 -> 16 in the cache */
   /* replace1095_1 -> 405 in the cache */
   /* replace1095 -> 406 in the cache */

   /* ('ishr', 'a@64', 56) => ('extract_i8', 'a', 7) */
   /* search1096_0 -> 21 in the cache */
   /* search1096_1 -> 3236 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 21, 3236 },
      -1,
   } },

   /* replace1096_0 -> 21 in the cache */
   /* replace1096_1 -> 3238 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 21, 3238 },
      -1,
   } },

   /* ('iand', 255, 'a') => ('extract_u8', 'a', 0) */
   /* search1097_0 -> 1590 in the cache */
   /* search1097_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1590, 32 },
      -1,
   } },

   /* replace1097_0 -> 32 in the cache */
   /* replace1097_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 32, 61 },
      -1,
   } },

   /* ('u2u8', ('extract_u16', 'a', 1)) => ('u2u8', ('extract_u8', 'a', 2)) */
   /* search1098_0_0 -> 32 in the cache */
   /* search1098_0_1 -> 35 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 32, 35 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 3247 },
      -1,
   } },

   /* replace1098_0_0 -> 32 in the cache */
   /* replace1098_0_1 -> 2238 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 32, 2238 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 3249 },
      -1,
   } },

   /* ('u2u8', ('ushr', 'a', 8)) => ('u2u8', ('extract_u8', 'a', 1)) */
   /* search1099_0_0 -> 32 in the cache */
   /* search1099_0_1 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 32, 831 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 3251 },
      -1,
   } },

   /* replace1099_0_0 -> 32 in the cache */
   /* replace1099_0_1 -> 35 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 32, 35 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 3253 },
      -1,
   } },

   /* ('i2i16', ('u2u8', ('extract_u8', 'a', 'b'))) => ('i2i16', ('extract_i8', 'a', 'b')) */
   /* search1100_0_0_0 -> 0 in the cache */
   /* search1100_0_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 3255 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 3256 },
      -1,
   } },

   /* replace1100_0_0 -> 0 in the cache */
   /* replace1100_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 3258 },
      -1,
   } },

   /* ('u2u16', ('u2u8', ('extract_u8', 'a', 'b'))) => ('u2u16', ('extract_u8', 'a', 'b')) */
   /* search1101_0_0_0 -> 0 in the cache */
   /* search1101_0_0_1 -> 3 in the cache */
   /* search1101_0_0 -> 3255 in the cache */
   /* search1101_0 -> 3256 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 3256 },
      -1,
   } },

   /* replace1101_0_0 -> 0 in the cache */
   /* replace1101_0_1 -> 3 in the cache */
   /* replace1101_0 -> 3255 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 3255 },
      -1,
   } },

   /* ('ubfe', 'a', 0, 8) => ('extract_u8', 'a', 0) */
   /* search1102_0 -> 16 in the cache */
   /* search1102_1 -> 122 in the cache */
   /* search1102_2 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 16, 122, 831 },
      -1,
   } },

   /* replace1102_0 -> 16 in the cache */
   /* replace1102_1 -> 122 in the cache */
   /* replace1102 -> 412 in the cache */

   /* ('ubfe', 'a', 8, 8) => ('extract_u8', 'a', 1) */
   /* search1103_0 -> 16 in the cache */
   /* search1103_1 -> 831 in the cache */
   /* search1103_2 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 16, 831, 831 },
      -1,
   } },

   /* replace1103_0 -> 16 in the cache */
   /* replace1103_1 -> 396 in the cache */
   /* replace1103 -> 415 in the cache */

   /* ('ubfe', 'a', 16, 8) => ('extract_u8', 'a', 2) */
   /* search1104_0 -> 16 in the cache */
   /* search1104_1 -> 853 in the cache */
   /* search1104_2 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 16, 853, 831 },
      -1,
   } },

   /* replace1104_0 -> 16 in the cache */
   /* replace1104_1 -> 401 in the cache */
   /* replace1104 -> 419 in the cache */

   /* ('ubfe', 'a', 24, 8) => ('extract_u8', 'a', 3) */
   /* search1105_0 -> 16 in the cache */
   /* search1105_1 -> 1012 in the cache */
   /* search1105_2 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 16, 1012, 831 },
      -1,
   } },

   /* replace1105_0 -> 16 in the cache */
   /* replace1105_1 -> 405 in the cache */
   /* replace1105 -> 422 in the cache */

   /* ('ibfe', 'a', 0, 8) => ('extract_i8', 'a', 0) */
   /* search1106_0 -> 16 in the cache */
   /* search1106_1 -> 122 in the cache */
   /* search1106_2 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 16, 122, 831 },
      -1,
   } },

   /* replace1106_0 -> 16 in the cache */
   /* replace1106_1 -> 122 in the cache */
   /* replace1106 -> 393 in the cache */

   /* ('ibfe', 'a', 8, 8) => ('extract_i8', 'a', 1) */
   /* search1107_0 -> 16 in the cache */
   /* search1107_1 -> 831 in the cache */
   /* search1107_2 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 16, 831, 831 },
      -1,
   } },

   /* replace1107_0 -> 16 in the cache */
   /* replace1107_1 -> 396 in the cache */
   /* replace1107 -> 397 in the cache */

   /* ('ibfe', 'a', 16, 8) => ('extract_i8', 'a', 2) */
   /* search1108_0 -> 16 in the cache */
   /* search1108_1 -> 853 in the cache */
   /* search1108_2 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 16, 853, 831 },
      -1,
   } },

   /* replace1108_0 -> 16 in the cache */
   /* replace1108_1 -> 401 in the cache */
   /* replace1108 -> 402 in the cache */

   /* ('ibfe', 'a', 24, 8) => ('extract_i8', 'a', 3) */
   /* search1109_0 -> 16 in the cache */
   /* search1109_1 -> 1012 in the cache */
   /* search1109_2 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 16, 1012, 831 },
      -1,
   } },

   /* replace1109_0 -> 16 in the cache */
   /* replace1109_1 -> 405 in the cache */
   /* replace1109 -> 406 in the cache */

   /* ('extract_u8', ('extract_i8', 'a', 'b'), 0) => ('extract_u8', 'a', 'b') */
   /* search1110_0_0 -> 0 in the cache */
   /* search1110_0_1 -> 3 in the cache */
   /* search1110_0 -> 3258 in the cache */
   /* search1110_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3258, 86 },
      -1,
   } },

   /* replace1110_0 -> 0 in the cache */
   /* replace1110_1 -> 3 in the cache */
   /* replace1110 -> 3255 in the cache */

   /* ('extract_u8', ('extract_u8', 'a', 'b'), 0) => ('extract_u8', 'a', 'b') */
   /* search1111_0_0 -> 0 in the cache */
   /* search1111_0_1 -> 3 in the cache */
   /* search1111_0 -> 3255 in the cache */
   /* search1111_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3255, 86 },
      -1,
   } },

   /* replace1111_0 -> 0 in the cache */
   /* replace1111_1 -> 3 in the cache */
   /* replace1111 -> 3255 in the cache */

   /* ('ushr', ('ishl', 'a@32', 16), 16) => ('extract_u16', 'a', 0) */
   /* search1112_0_0 -> 16 in the cache */
   /* search1112_0_1 -> 853 in the cache */
   /* search1112_0 -> 1005 in the cache */
   /* search1112_1 -> 853 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 1005, 853 },
      -1,
   } },

   /* replace1112_0 -> 16 in the cache */
   /* replace1112_1 -> 122 in the cache */
   /* replace1112 -> 444 in the cache */

   /* ('ushr', 'a@32', 16) => ('extract_u16', 'a', 1) */
   /* search1113_0 -> 16 in the cache */
   /* search1113_1 -> 853 in the cache */
   /* search1113 -> 1008 in the cache */

   /* replace1113_0 -> 16 in the cache */
   /* replace1113_1 -> 396 in the cache */
   /* replace1113 -> 447 in the cache */

   /* ('ishr', ('ishl', 'a@32', 16), 16) => ('extract_i16', 'a', 0) */
   /* search1114_0_0 -> 16 in the cache */
   /* search1114_0_1 -> 853 in the cache */
   /* search1114_0 -> 1005 in the cache */
   /* search1114_1 -> 853 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 1005, 853 },
      -1,
   } },

   /* replace1114_0 -> 16 in the cache */
   /* replace1114_1 -> 122 in the cache */
   /* replace1114 -> 436 in the cache */

   /* ('ishr', 'a@32', 16) => ('extract_i16', 'a', 1) */
   /* search1115_0 -> 16 in the cache */
   /* search1115_1 -> 853 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 16, 853 },
      -1,
   } },

   /* replace1115_0 -> 16 in the cache */
   /* replace1115_1 -> 396 in the cache */
   /* replace1115 -> 439 in the cache */

   /* ('iand', 65535, 'a') => ('extract_u16', 'a', 0) */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_int, { 0xffff /* 65535 */ },
   } },
   /* search1116_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 3275, 32 },
      -1,
   } },

   /* replace1116_0 -> 32 in the cache */
   /* replace1116_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 32, 61 },
      -1,
   } },

   /* ('ubfe', 'a', 0, 16) => ('extract_u16', 'a', 0) */
   /* search1117_0 -> 16 in the cache */
   /* search1117_1 -> 122 in the cache */
   /* search1117_2 -> 853 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 16, 122, 853 },
      -1,
   } },

   /* replace1117_0 -> 16 in the cache */
   /* replace1117_1 -> 122 in the cache */
   /* replace1117 -> 444 in the cache */

   /* ('ubfe', 'a', 16, 16) => ('extract_u16', 'a', 1) */
   /* search1118_0 -> 16 in the cache */
   /* search1118_1 -> 853 in the cache */
   /* search1118_2 -> 853 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 16, 853, 853 },
      -1,
   } },

   /* replace1118_0 -> 16 in the cache */
   /* replace1118_1 -> 396 in the cache */
   /* replace1118 -> 447 in the cache */

   /* ('ibfe', 'a', 0, 16) => ('extract_i16', 'a', 0) */
   /* search1119_0 -> 16 in the cache */
   /* search1119_1 -> 122 in the cache */
   /* search1119_2 -> 853 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 16, 122, 853 },
      -1,
   } },

   /* replace1119_0 -> 16 in the cache */
   /* replace1119_1 -> 122 in the cache */
   /* replace1119 -> 436 in the cache */

   /* ('ibfe', 'a', 16, 16) => ('extract_i16', 'a', 1) */
   /* search1120_0 -> 16 in the cache */
   /* search1120_1 -> 853 in the cache */
   /* search1120_2 -> 853 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 16, 853, 853 },
      -1,
   } },

   /* replace1120_0 -> 16 in the cache */
   /* replace1120_1 -> 396 in the cache */
   /* replace1120 -> 439 in the cache */

   /* ('ior', ('ishl', ('u2u32', 'a@8'), 24), ('ior', ('ishl', ('u2u32', 'b@8'), 16), ('ior', ('ishl', ('u2u32', 'c@8'), 8), ('u2u32', 'd@8')))) => ('pack_32_4x8', ('vec4', 'd', 'c', 'b', 'a')) */
   /* search1121_0_0_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 6 },
      -1,
   } },
   /* search1121_0_1 -> 1012 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3282, 1012 },
      -1,
   } },
   /* search1121_1_0_0_0 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 2112 },
      -1,
   } },
   /* search1121_1_0_1 -> 853 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3284, 853 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 8 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 3286 },
      -1,
   } },
   /* search1121_1_1_0_1 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3287, 831 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 8 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 3289 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      2, 1,
      { 3288, 3290 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      1, 2,
      { 3285, 3291 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 3283, 3292 },
      -1,
   } },

   /* replace1121_0_0 -> 3289 in the cache */
   /* replace1121_0_1 -> 3286 in the cache */
   /* replace1121_0_2 -> 2112 in the cache */
   /* replace1121_0_3 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 3289, 3286, 2112, 6 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_32_4x8,
      -1, 0,
      { 3294 },
      -1,
   } },

   /* ('extract_u16', ('extract_i16', 'a', 'b'), 0) => ('extract_u16', 'a', 'b') */
   /* search1122_0_0 -> 0 in the cache */
   /* search1122_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 0, 3 },
      -1,
   } },
   /* search1122_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 3296, 86 },
      -1,
   } },

   /* replace1122_0 -> 0 in the cache */
   /* replace1122_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* ('extract_u16', ('extract_u16', 'a', 'b'), 0) => ('extract_u16', 'a', 'b') */
   /* search1123_0_0 -> 0 in the cache */
   /* search1123_0_1 -> 3 in the cache */
   /* search1123_0 -> 3298 in the cache */
   /* search1123_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 3298, 86 },
      -1,
   } },

   /* replace1123_0 -> 0 in the cache */
   /* replace1123_1 -> 3 in the cache */
   /* replace1123 -> 3298 in the cache */

   /* ('pack_64_2x32_split', 'a', 'b') => ('ior', ('u2u64', 'a'), ('ishl', ('u2u64', 'b'), 32)) */
   /* search1124_0 -> 16 in the cache */
   /* search1124_1 -> 36 in the cache */
   /* search1124 -> 3202 in the cache */

   /* replace1124_0_0 -> 16 in the cache */
   /* replace1124_0 -> 1911 in the cache */
   /* replace1124_1_0_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 36 },
      -1,
   } },
   /* replace1124_1_1 -> 874 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3300, 874 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1911, 3301 },
      -1,
   } },

   /* ('pack_32_2x16_split', 'a', 'b') => ('ior', ('u2u32', 'a'), ('ishl', ('u2u32', 'b'), 16)) */
   /* search1125_0 -> 11 in the cache */
   /* search1125_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_32_2x16_split,
      -1, 0,
      { 11, 258 },
      -1,
   } },

   /* replace1125_0_0 -> 11 in the cache */
   /* replace1125_0 -> 1806 in the cache */
   /* replace1125_1_0_0 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 258 },
      -1,
   } },
   /* replace1125_1_1 -> 853 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3304, 853 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1806, 3305 },
      -1,
   } },

   /* ('unpack_64_2x32_split_x', 'a') => ('u2u32', 'a') */
   /* search1126_0 -> 21 in the cache */
   /* search1126 -> 2732 in the cache */

   /* replace1126_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 21 },
      -1,
   } },

   /* ('unpack_64_2x32_split_y', 'a') => ('u2u32', ('ushr', 'a', 32)) */
   /* search1127_0 -> 21 in the cache */
   /* search1127 -> 2726 in the cache */

   /* replace1127_0_0 -> 21 in the cache */
   /* replace1127_0_1 -> 874 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 21, 874 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 3308 },
      -1,
   } },

   /* ('unpack_32_2x16_split_x', 'a') => ('u2u16', 'a') */
   /* search1128_0 -> 16 in the cache */
   /* search1128 -> 3226 in the cache */

   /* replace1128_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('unpack_32_2x16_split_y', 'a') => ('u2u16', ('ushr', 'a', 16)) */
   /* search1129_0 -> 16 in the cache */
   /* search1129 -> 3228 in the cache */

   /* replace1129_0_0 -> 16 in the cache */
   /* replace1129_0_1 -> 853 in the cache */
   /* replace1129_0 -> 1008 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 1008 },
      -1,
   } },

   /* ('unpack_half_2x16_split_x', ('iand', 'a', 65535)) => ('unpack_half_2x16_split_x', 'a') */
   /* search1130_0_0 -> 16 in the cache */
   /* search1130_0_1 -> 1002 in the cache */
   /* search1130_0 -> 1003 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_half_2x16_split_x,
      -1, 1,
      { 1003 },
      -1,
   } },

   /* replace1130_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_half_2x16_split_x,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('unpack_32_2x16_split_x', ('iand', 'a', 65535)) => ('unpack_32_2x16_split_x', 'a') */
   /* search1131_0_0 -> 16 in the cache */
   /* search1131_0_1 -> 1002 in the cache */
   /* search1131_0 -> 1003 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_unpack_32_2x16_split_x,
      -1, 1,
      { 1003 },
      -1,
   } },

   /* replace1131_0 -> 16 in the cache */
   /* replace1131 -> 3226 in the cache */

   /* ('unpack_64_2x32_split_x', ('iand', 'a', 4294967295)) => ('unpack_64_2x32_split_x', 'a') */
   /* search1132_0_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0xffffffff /* 4294967295 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 21, 3315 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_x,
      -1, 1,
      { 3316 },
      -1,
   } },

   /* replace1132_0 -> 21 in the cache */
   /* replace1132 -> 2732 in the cache */

   /* ('unpack_half_2x16_split_y', ('iand', 'a', 4294901760)) => ('unpack_half_2x16_split_y', 'a') */
   /* search1133_0_0 -> 16 in the cache */
   /* search1133_0_1 -> 1006 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 16, 1006 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_half_2x16_split_y,
      -1, 1,
      { 3318 },
      -1,
   } },

   /* replace1133_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_half_2x16_split_y,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('unpack_32_2x16_split_y', ('iand', 'a', 4294901760)) => ('unpack_32_2x16_split_y', 'a') */
   /* search1134_0_0 -> 16 in the cache */
   /* search1134_0_1 -> 1006 in the cache */
   /* search1134_0 -> 3318 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_unpack_32_2x16_split_y,
      -1, 1,
      { 3318 },
      -1,
   } },

   /* replace1134_0 -> 16 in the cache */
   /* replace1134 -> 3228 in the cache */

   /* ('unpack_64_2x32_split_y', ('iand', 'a', 18446744069414584320)) => ('unpack_64_2x32_split_y', 'a') */
   /* search1135_0_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0xffffffff00000000 /* 18446744069414584320 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 21, 3322 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_y,
      -1, 1,
      { 3323 },
      -1,
   } },

   /* replace1135_0 -> 21 in the cache */
   /* replace1135 -> 2726 in the cache */

   /* ('unpack_half_2x16_split_x', ('extract_u16', 'a', 0)) => ('unpack_half_2x16_split_x', 'a') */
   /* search1136_0_0 -> 16 in the cache */
   /* search1136_0_1 -> 122 in the cache */
   /* search1136_0 -> 444 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_half_2x16_split_x,
      -1, 0,
      { 444 },
      -1,
   } },

   /* replace1136_0 -> 16 in the cache */
   /* replace1136 -> 3313 in the cache */

   /* ('unpack_half_2x16_split_x', ('extract_u16', 'a', 1)) => ('unpack_half_2x16_split_y', 'a') */
   /* search1137_0_0 -> 16 in the cache */
   /* search1137_0_1 -> 396 in the cache */
   /* search1137_0 -> 447 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_half_2x16_split_x,
      -1, 0,
      { 447 },
      -1,
   } },

   /* replace1137_0 -> 16 in the cache */
   /* replace1137 -> 3320 in the cache */

   /* ('unpack_half_2x16_split_x', ('ushr', 'a', 16)) => ('unpack_half_2x16_split_y', 'a') */
   /* search1138_0_0 -> 16 in the cache */
   /* search1138_0_1 -> 853 in the cache */
   /* search1138_0 -> 1008 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_half_2x16_split_x,
      -1, 0,
      { 1008 },
      -1,
   } },

   /* replace1138_0 -> 16 in the cache */
   /* replace1138 -> 3320 in the cache */

   /* ('unpack_32_2x16_split_x', ('extract_u16', 'a', 0)) => ('unpack_32_2x16_split_x', 'a') */
   /* search1139_0_0 -> 16 in the cache */
   /* search1139_0_1 -> 122 in the cache */
   /* search1139_0 -> 444 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_unpack_32_2x16_split_x,
      -1, 0,
      { 444 },
      -1,
   } },

   /* replace1139_0 -> 16 in the cache */
   /* replace1139 -> 3226 in the cache */

   /* ('unpack_32_2x16_split_x', ('extract_u16', 'a', 1)) => ('unpack_32_2x16_split_y', 'a') */
   /* search1140_0_0 -> 16 in the cache */
   /* search1140_0_1 -> 396 in the cache */
   /* search1140_0 -> 447 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_unpack_32_2x16_split_x,
      -1, 0,
      { 447 },
      -1,
   } },

   /* replace1140_0 -> 16 in the cache */
   /* replace1140 -> 3228 in the cache */

   /* ('ishl', ('pack_half_2x16', ('vec2', 'a', 0)), 16) => ('pack_half_2x16', ('vec2', 0, 'a')) */
   /* search1141_0_0_0 -> 16 in the cache */
   /* search1141_0_0_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 16, 122 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16,
      -1, 0,
      { 3330 },
      -1,
   } },
   /* search1141_1 -> 853 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3331, 853 },
      -1,
   } },

   /* replace1141_0_0 -> 122 in the cache */
   /* replace1141_0_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 122, 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16,
      -1, 0,
      { 3333 },
      -1,
   } },

   /* ('ushr', ('pack_half_2x16', ('vec2', 0, 'a')), 16) => ('pack_half_2x16', ('vec2', 'a', 0)) */
   /* search1142_0_0_0 -> 122 in the cache */
   /* search1142_0_0_1 -> 16 in the cache */
   /* search1142_0_0 -> 3333 in the cache */
   /* search1142_0 -> 3334 in the cache */
   /* search1142_1 -> 853 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 3334, 853 },
      -1,
   } },

   /* replace1142_0_0 -> 16 in the cache */
   /* replace1142_0_1 -> 122 in the cache */
   /* replace1142_0 -> 3330 in the cache */
   /* replace1142 -> 3331 in the cache */

   /* ('iadd', ('pack_half_2x16', ('vec2', 'a', 0)), ('pack_half_2x16', ('vec2', 0, 'b'))) => ('pack_half_2x16', ('vec2', 'a', 'b')) */
   /* search1143_0_0_0 -> 16 in the cache */
   /* search1143_0_0_1 -> 122 in the cache */
   /* search1143_0_0 -> 3330 in the cache */
   /* search1143_0 -> 3331 in the cache */
   /* search1143_1_0_0 -> 122 in the cache */
   /* search1143_1_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 122, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16,
      -1, 0,
      { 3336 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 3331, 3337 },
      -1,
   } },

   /* replace1143_0_0 -> 16 in the cache */
   /* replace1143_0_1 -> 36 in the cache */
   /* replace1143_0 -> 3215 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16,
      -1, 0,
      { 3215 },
      -1,
   } },

   /* ('ior', ('pack_half_2x16', ('vec2', 'a', 0)), ('pack_half_2x16', ('vec2', 0, 'b'))) => ('pack_half_2x16', ('vec2', 'a', 'b')) */
   /* search1144_0_0_0 -> 16 in the cache */
   /* search1144_0_0_1 -> 122 in the cache */
   /* search1144_0_0 -> 3330 in the cache */
   /* search1144_0 -> 3331 in the cache */
   /* search1144_1_0_0 -> 122 in the cache */
   /* search1144_1_0_1 -> 36 in the cache */
   /* search1144_1_0 -> 3336 in the cache */
   /* search1144_1 -> 3337 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 3331, 3337 },
      -1,
   } },

   /* replace1144_0_0 -> 16 in the cache */
   /* replace1144_0_1 -> 36 in the cache */
   /* replace1144_0 -> 3215 in the cache */
   /* replace1144 -> 3339 in the cache */

   /* ('ishl', ('pack_half_2x16_split', 'a', 0), 16) => ('pack_half_2x16_split', 0, 'a') */
   /* search1145_0_0 -> 16 in the cache */
   /* search1145_0_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16_split,
      -1, 0,
      { 16, 122 },
      -1,
   } },
   /* search1145_1 -> 853 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3341, 853 },
      -1,
   } },

   /* replace1145_0 -> 122 in the cache */
   /* replace1145_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16_split,
      -1, 0,
      { 122, 16 },
      -1,
   } },

   /* ('ushr', ('pack_half_2x16_split', 0, 'a'), 16) => ('pack_half_2x16_split', 'a', 0) */
   /* search1146_0_0 -> 122 in the cache */
   /* search1146_0_1 -> 16 in the cache */
   /* search1146_0 -> 3343 in the cache */
   /* search1146_1 -> 853 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 3343, 853 },
      -1,
   } },

   /* replace1146_0 -> 16 in the cache */
   /* replace1146_1 -> 122 in the cache */
   /* replace1146 -> 3341 in the cache */

   /* ('extract_u16', ('pack_half_2x16_split', 0, 'a'), 1) => ('pack_half_2x16_split', 'a', 0) */
   /* search1147_0_0 -> 122 in the cache */
   /* search1147_0_1 -> 16 in the cache */
   /* search1147_0 -> 3343 in the cache */
   /* search1147_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 3343, 396 },
      -1,
   } },

   /* replace1147_0 -> 16 in the cache */
   /* replace1147_1 -> 122 in the cache */
   /* replace1147 -> 3341 in the cache */

   /* ('iadd', ('pack_half_2x16_split', 'a', 0), ('pack_half_2x16_split', 0, 'b')) => ('pack_half_2x16_split', 'a', 'b') */
   /* search1148_0_0 -> 16 in the cache */
   /* search1148_0_1 -> 122 in the cache */
   /* search1148_0 -> 3341 in the cache */
   /* search1148_1_0 -> 122 in the cache */
   /* search1148_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16_split,
      -1, 0,
      { 122, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 3341, 3346 },
      -1,
   } },

   /* replace1148_0 -> 16 in the cache */
   /* replace1148_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16_split,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* ('ior', ('pack_half_2x16_split', 'a', 0), ('pack_half_2x16_split', 0, 'b')) => ('pack_half_2x16_split', 'a', 'b') */
   /* search1149_0_0 -> 16 in the cache */
   /* search1149_0_1 -> 122 in the cache */
   /* search1149_0 -> 3341 in the cache */
   /* search1149_1_0 -> 122 in the cache */
   /* search1149_1_1 -> 36 in the cache */
   /* search1149_1 -> 3346 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 3341, 3346 },
      -1,
   } },

   /* replace1149_0 -> 16 in the cache */
   /* replace1149_1 -> 36 in the cache */
   /* replace1149 -> 3348 in the cache */

   /* ('extract_i8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 0) => ('i2i', 'a') */
   /* search1150_0_0 -> 6 in the cache */
   /* search1150_0_1 -> 2112 in the cache */
   /* search1150_0_2 -> 3286 in the cache */
   /* search1150_0_3 -> 3289 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_32_4x8_split,
      -1, 0,
      { 6, 2112, 3286, 3289 },
      -1,
   } },
   /* search1150_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3350, 122 },
      -1,
   } },

   /* replace1150_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_i2i,
      -1, 0,
      { 6 },
      -1,
   } },

   /* ('extract_i8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 1) => ('i2i', 'b') */
   /* search1151_0_0 -> 6 in the cache */
   /* search1151_0_1 -> 2112 in the cache */
   /* search1151_0_2 -> 3286 in the cache */
   /* search1151_0_3 -> 3289 in the cache */
   /* search1151_0 -> 3350 in the cache */
   /* search1151_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3350, 396 },
      -1,
   } },

   /* replace1151_0 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_i2i,
      -1, 0,
      { 2112 },
      -1,
   } },

   /* ('extract_i8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 2) => ('i2i', 'c') */
   /* search1152_0_0 -> 6 in the cache */
   /* search1152_0_1 -> 2112 in the cache */
   /* search1152_0_2 -> 3286 in the cache */
   /* search1152_0_3 -> 3289 in the cache */
   /* search1152_0 -> 3350 in the cache */
   /* search1152_1 -> 401 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3350, 401 },
      -1,
   } },

   /* replace1152_0 -> 3286 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_i2i,
      -1, 0,
      { 3286 },
      -1,
   } },

   /* ('extract_i8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 3) => ('i2i', 'd') */
   /* search1153_0_0 -> 6 in the cache */
   /* search1153_0_1 -> 2112 in the cache */
   /* search1153_0_2 -> 3286 in the cache */
   /* search1153_0_3 -> 3289 in the cache */
   /* search1153_0 -> 3350 in the cache */
   /* search1153_1 -> 405 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3350, 405 },
      -1,
   } },

   /* replace1153_0 -> 3289 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_i2i,
      -1, 0,
      { 3289 },
      -1,
   } },

   /* ('extract_u8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 0) => ('u2u', 'a') */
   /* search1154_0_0 -> 6 in the cache */
   /* search1154_0_1 -> 2112 in the cache */
   /* search1154_0_2 -> 3286 in the cache */
   /* search1154_0_3 -> 3289 in the cache */
   /* search1154_0 -> 3350 in the cache */
   /* search1154_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3350, 122 },
      -1,
   } },

   /* replace1154_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_u2u,
      -1, 0,
      { 6 },
      -1,
   } },

   /* ('extract_u8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 1) => ('u2u', 'b') */
   /* search1155_0_0 -> 6 in the cache */
   /* search1155_0_1 -> 2112 in the cache */
   /* search1155_0_2 -> 3286 in the cache */
   /* search1155_0_3 -> 3289 in the cache */
   /* search1155_0 -> 3350 in the cache */
   /* search1155_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3350, 396 },
      -1,
   } },

   /* replace1155_0 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_u2u,
      -1, 0,
      { 2112 },
      -1,
   } },

   /* ('extract_u8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 2) => ('u2u', 'c') */
   /* search1156_0_0 -> 6 in the cache */
   /* search1156_0_1 -> 2112 in the cache */
   /* search1156_0_2 -> 3286 in the cache */
   /* search1156_0_3 -> 3289 in the cache */
   /* search1156_0 -> 3350 in the cache */
   /* search1156_1 -> 401 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3350, 401 },
      -1,
   } },

   /* replace1156_0 -> 3286 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_u2u,
      -1, 0,
      { 3286 },
      -1,
   } },

   /* ('extract_u8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 3) => ('u2u', 'd') */
   /* search1157_0_0 -> 6 in the cache */
   /* search1157_0_1 -> 2112 in the cache */
   /* search1157_0_2 -> 3286 in the cache */
   /* search1157_0_3 -> 3289 in the cache */
   /* search1157_0 -> 3350 in the cache */
   /* search1157_1 -> 405 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3350, 405 },
      -1,
   } },

   /* replace1157_0 -> 3289 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_u2u,
      -1, 0,
      { 3289 },
      -1,
   } },

   /* ('extract_u8', ('ushr', 'a@16', 8), 0) => ('extract_u8', 'a', 1) */
   /* search1158_0_0 -> 11 in the cache */
   /* search1158_0_1 -> 831 in the cache */
   /* search1158_0 -> 3234 in the cache */
   /* search1158_1 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3234, 856 },
      -1,
   } },

   /* replace1158_0 -> 11 in the cache */
   /* replace1158_1 -> 1452 in the cache */
   /* replace1158 -> 3235 in the cache */

   /* ('extract_u8', ('ushr', 'a@32', 8), 0) => ('extract_u8', 'a', 1) */
   /* search1159_0_0 -> 16 in the cache */
   /* search1159_0_1 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 16, 831 },
      -1,
   } },
   /* search1159_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3368, 122 },
      -1,
   } },

   /* replace1159_0 -> 16 in the cache */
   /* replace1159_1 -> 396 in the cache */
   /* replace1159 -> 415 in the cache */

   /* ('extract_u8', ('ushr', 'a@32', 16), 0) => ('extract_u8', 'a', 2) */
   /* search1160_0_0 -> 16 in the cache */
   /* search1160_0_1 -> 853 in the cache */
   /* search1160_0 -> 1008 in the cache */
   /* search1160_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 1008, 122 },
      -1,
   } },

   /* replace1160_0 -> 16 in the cache */
   /* replace1160_1 -> 401 in the cache */
   /* replace1160 -> 419 in the cache */

   /* ('extract_u8', ('ushr', 'a@32', 24), 0) => ('extract_u8', 'a', 3) */
   /* search1161_0_0 -> 16 in the cache */
   /* search1161_0_1 -> 1012 in the cache */
   /* search1161_0 -> 1017 in the cache */
   /* search1161_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 1017, 122 },
      -1,
   } },

   /* replace1161_0 -> 16 in the cache */
   /* replace1161_1 -> 405 in the cache */
   /* replace1161 -> 422 in the cache */

   /* ('extract_u8', ('ushr', 'a@64', 8), 0) => ('extract_u8', 'a', 1) */
   /* search1162_0_0 -> 21 in the cache */
   /* search1162_0_1 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 21, 831 },
      -1,
   } },
   /* search1162_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3372, 897 },
      -1,
   } },

   /* replace1162_0 -> 21 in the cache */
   /* replace1162_1 -> 1494 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 21, 1494 },
      -1,
   } },

   /* ('extract_u8', ('ushr', 'a@64', 16), 0) => ('extract_u8', 'a', 2) */
   /* search1163_0_0 -> 21 in the cache */
   /* search1163_0_1 -> 853 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 21, 853 },
      -1,
   } },
   /* search1163_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3375, 897 },
      -1,
   } },

   /* replace1163_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x2 /* 2 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 21, 3377 },
      -1,
   } },

   /* ('extract_u8', ('ushr', 'a@64', 24), 0) => ('extract_u8', 'a', 3) */
   /* search1164_0_0 -> 21 in the cache */
   /* search1164_0_1 -> 1012 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 21, 1012 },
      -1,
   } },
   /* search1164_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3379, 897 },
      -1,
   } },

   /* replace1164_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x3 /* 3 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 21, 3381 },
      -1,
   } },

   /* ('extract_u8', ('ushr', 'a@64', 32), 0) => ('extract_u8', 'a', 4) */
   /* search1165_0_0 -> 21 in the cache */
   /* search1165_0_1 -> 874 in the cache */
   /* search1165_0 -> 3308 in the cache */
   /* search1165_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3308, 897 },
      -1,
   } },

   /* replace1165_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x4 /* 4 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 21, 3384 },
      -1,
   } },

   /* ('extract_u8', ('ushr', 'a@64', 40), 0) => ('extract_u8', 'a', 5) */
   /* search1166_0_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x28 /* 40 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 21, 3386 },
      -1,
   } },
   /* search1166_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3387, 897 },
      -1,
   } },

   /* replace1166_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x5 /* 5 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 21, 3389 },
      -1,
   } },

   /* ('extract_u8', ('ushr', 'a@64', 48), 0) => ('extract_u8', 'a', 6) */
   /* search1167_0_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x30 /* 48 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 21, 3391 },
      -1,
   } },
   /* search1167_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3392, 897 },
      -1,
   } },

   /* replace1167_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x6 /* 6 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 21, 3394 },
      -1,
   } },

   /* ('extract_u8', ('ushr', 'a@64', 56), 0) => ('extract_u8', 'a', 7) */
   /* search1168_0_0 -> 21 in the cache */
   /* search1168_0_1 -> 3236 in the cache */
   /* search1168_0 -> 3237 in the cache */
   /* search1168_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3237, 897 },
      -1,
   } },

   /* replace1168_0 -> 21 in the cache */
   /* replace1168_1 -> 3238 in the cache */
   /* replace1168 -> 3239 in the cache */

   /* ('extract_u8', ('ishr', 'a@16', 8), 0) => ('extract_u8', 'a', 1) */
   /* search1169_0_0 -> 11 in the cache */
   /* search1169_0_1 -> 831 in the cache */
   /* search1169_0 -> 3240 in the cache */
   /* search1169_1 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3240, 856 },
      -1,
   } },

   /* replace1169_0 -> 11 in the cache */
   /* replace1169_1 -> 1452 in the cache */
   /* replace1169 -> 3235 in the cache */

   /* ('extract_u8', ('ishr', 'a@32', 8), 0) => ('extract_u8', 'a', 1) */
   /* search1170_0_0 -> 16 in the cache */
   /* search1170_0_1 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 16, 831 },
      -1,
   } },
   /* search1170_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3398, 122 },
      -1,
   } },

   /* replace1170_0 -> 16 in the cache */
   /* replace1170_1 -> 396 in the cache */
   /* replace1170 -> 415 in the cache */

   /* ('extract_u8', ('ishr', 'a@32', 16), 0) => ('extract_u8', 'a', 2) */
   /* search1171_0_0 -> 16 in the cache */
   /* search1171_0_1 -> 853 in the cache */
   /* search1171_0 -> 3274 in the cache */
   /* search1171_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3274, 122 },
      -1,
   } },

   /* replace1171_0 -> 16 in the cache */
   /* replace1171_1 -> 401 in the cache */
   /* replace1171 -> 419 in the cache */

   /* ('extract_u8', ('ishr', 'a@32', 24), 0) => ('extract_u8', 'a', 3) */
   /* search1172_0_0 -> 16 in the cache */
   /* search1172_0_1 -> 1012 in the cache */
   /* search1172_0 -> 3242 in the cache */
   /* search1172_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3242, 122 },
      -1,
   } },

   /* replace1172_0 -> 16 in the cache */
   /* replace1172_1 -> 405 in the cache */
   /* replace1172 -> 422 in the cache */

   /* ('extract_u8', ('ishr', 'a@64', 8), 0) => ('extract_u8', 'a', 1) */
   /* search1173_0_0 -> 21 in the cache */
   /* search1173_0_1 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 21, 831 },
      -1,
   } },
   /* search1173_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3402, 897 },
      -1,
   } },

   /* replace1173_0 -> 21 in the cache */
   /* replace1173_1 -> 1494 in the cache */
   /* replace1173 -> 3374 in the cache */

   /* ('extract_u8', ('ishr', 'a@64', 16), 0) => ('extract_u8', 'a', 2) */
   /* search1174_0_0 -> 21 in the cache */
   /* search1174_0_1 -> 853 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 21, 853 },
      -1,
   } },
   /* search1174_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3404, 897 },
      -1,
   } },

   /* replace1174_0 -> 21 in the cache */
   /* replace1174_1 -> 3377 in the cache */
   /* replace1174 -> 3378 in the cache */

   /* ('extract_u8', ('ishr', 'a@64', 24), 0) => ('extract_u8', 'a', 3) */
   /* search1175_0_0 -> 21 in the cache */
   /* search1175_0_1 -> 1012 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 21, 1012 },
      -1,
   } },
   /* search1175_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3406, 897 },
      -1,
   } },

   /* replace1175_0 -> 21 in the cache */
   /* replace1175_1 -> 3381 in the cache */
   /* replace1175 -> 3382 in the cache */

   /* ('extract_u8', ('ishr', 'a@64', 32), 0) => ('extract_u8', 'a', 4) */
   /* search1176_0_0 -> 21 in the cache */
   /* search1176_0_1 -> 874 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 21, 874 },
      -1,
   } },
   /* search1176_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3408, 897 },
      -1,
   } },

   /* replace1176_0 -> 21 in the cache */
   /* replace1176_1 -> 3384 in the cache */
   /* replace1176 -> 3385 in the cache */

   /* ('extract_u8', ('ishr', 'a@64', 40), 0) => ('extract_u8', 'a', 5) */
   /* search1177_0_0 -> 21 in the cache */
   /* search1177_0_1 -> 3386 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 21, 3386 },
      -1,
   } },
   /* search1177_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3410, 897 },
      -1,
   } },

   /* replace1177_0 -> 21 in the cache */
   /* replace1177_1 -> 3389 in the cache */
   /* replace1177 -> 3390 in the cache */

   /* ('extract_u8', ('ishr', 'a@64', 48), 0) => ('extract_u8', 'a', 6) */
   /* search1178_0_0 -> 21 in the cache */
   /* search1178_0_1 -> 3391 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 21, 3391 },
      -1,
   } },
   /* search1178_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3412, 897 },
      -1,
   } },

   /* replace1178_0 -> 21 in the cache */
   /* replace1178_1 -> 3394 in the cache */
   /* replace1178 -> 3395 in the cache */

   /* ('extract_u8', ('ishr', 'a@64', 56), 0) => ('extract_u8', 'a', 7) */
   /* search1179_0_0 -> 21 in the cache */
   /* search1179_0_1 -> 3236 in the cache */
   /* search1179_0 -> 3243 in the cache */
   /* search1179_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3243, 897 },
      -1,
   } },

   /* replace1179_0 -> 21 in the cache */
   /* replace1179_1 -> 3238 in the cache */
   /* replace1179 -> 3239 in the cache */

   /* ('extract_u8', ('extract_u16', 'a', 1), 0) => ('extract_u8', 'a', 2) */
   /* search1180_0_0 -> 32 in the cache */
   /* search1180_0_1 -> 35 in the cache */
   /* search1180_0 -> 3247 in the cache */
   /* search1180_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3247, 61 },
      -1,
   } },

   /* replace1180_0 -> 32 in the cache */
   /* replace1180_1 -> 2238 in the cache */
   /* replace1180 -> 3249 in the cache */

   /* ('extract_u8', ('ishl', 'a@16', 8), 1) => ('extract_u8', 'a', 0) */
   /* search1181_0_0 -> 11 in the cache */
   /* search1181_0_1 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 11, 831 },
      -1,
   } },
   /* search1181_1 -> 1452 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3416, 1452 },
      -1,
   } },

   /* replace1181_0 -> 11 in the cache */
   /* replace1181_1 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 11, 856 },
      -1,
   } },

   /* ('extract_u8', ('ishl', 'a@32', 8), 3) => ('extract_u8', 'a', 2) */
   /* search1182_0_0 -> 16 in the cache */
   /* search1182_0_1 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 831 },
      -1,
   } },
   /* search1182_1 -> 405 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3419, 405 },
      -1,
   } },

   /* replace1182_0 -> 16 in the cache */
   /* replace1182_1 -> 401 in the cache */
   /* replace1182 -> 419 in the cache */

   /* ('extract_u8', ('ishl', 'a@32', 16), 3) => ('extract_u8', 'a', 1) */
   /* search1183_0_0 -> 16 in the cache */
   /* search1183_0_1 -> 853 in the cache */
   /* search1183_0 -> 1005 in the cache */
   /* search1183_1 -> 405 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 1005, 405 },
      -1,
   } },

   /* replace1183_0 -> 16 in the cache */
   /* replace1183_1 -> 396 in the cache */
   /* replace1183 -> 415 in the cache */

   /* ('extract_u8', ('ishl', 'a@32', 24), 3) => ('extract_u8', 'a', 0) */
   /* search1184_0_0 -> 16 in the cache */
   /* search1184_0_1 -> 1012 in the cache */
   /* search1184_0 -> 1014 in the cache */
   /* search1184_1 -> 405 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 1014, 405 },
      -1,
   } },

   /* replace1184_0 -> 16 in the cache */
   /* replace1184_1 -> 122 in the cache */
   /* replace1184 -> 412 in the cache */

   /* ('extract_u8', ('ishl', 'a@64', 8), 7) => ('extract_u8', 'a', 6) */
   /* search1185_0_0 -> 21 in the cache */
   /* search1185_0_1 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 831 },
      -1,
   } },
   /* search1185_1 -> 3238 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3423, 3238 },
      -1,
   } },

   /* replace1185_0 -> 21 in the cache */
   /* replace1185_1 -> 3394 in the cache */
   /* replace1185 -> 3395 in the cache */

   /* ('extract_u8', ('ishl', 'a@64', 16), 7) => ('extract_u8', 'a', 5) */
   /* search1186_0_0 -> 21 in the cache */
   /* search1186_0_1 -> 853 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 853 },
      -1,
   } },
   /* search1186_1 -> 3238 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3425, 3238 },
      -1,
   } },

   /* replace1186_0 -> 21 in the cache */
   /* replace1186_1 -> 3389 in the cache */
   /* replace1186 -> 3390 in the cache */

   /* ('extract_u8', ('ishl', 'a@64', 24), 7) => ('extract_u8', 'a', 4) */
   /* search1187_0_0 -> 21 in the cache */
   /* search1187_0_1 -> 1012 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 1012 },
      -1,
   } },
   /* search1187_1 -> 3238 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3427, 3238 },
      -1,
   } },

   /* replace1187_0 -> 21 in the cache */
   /* replace1187_1 -> 3384 in the cache */
   /* replace1187 -> 3385 in the cache */

   /* ('extract_u8', ('ishl', 'a@64', 32), 7) => ('extract_u8', 'a', 3) */
   /* search1188_0_0 -> 21 in the cache */
   /* search1188_0_1 -> 874 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 874 },
      -1,
   } },
   /* search1188_1 -> 3238 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3429, 3238 },
      -1,
   } },

   /* replace1188_0 -> 21 in the cache */
   /* replace1188_1 -> 3381 in the cache */
   /* replace1188 -> 3382 in the cache */

   /* ('extract_u8', ('ishl', 'a@64', 40), 7) => ('extract_u8', 'a', 2) */
   /* search1189_0_0 -> 21 in the cache */
   /* search1189_0_1 -> 3386 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 3386 },
      -1,
   } },
   /* search1189_1 -> 3238 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3431, 3238 },
      -1,
   } },

   /* replace1189_0 -> 21 in the cache */
   /* replace1189_1 -> 3377 in the cache */
   /* replace1189 -> 3378 in the cache */

   /* ('extract_u8', ('ishl', 'a@64', 48), 7) => ('extract_u8', 'a', 1) */
   /* search1190_0_0 -> 21 in the cache */
   /* search1190_0_1 -> 3391 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 3391 },
      -1,
   } },
   /* search1190_1 -> 3238 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3433, 3238 },
      -1,
   } },

   /* replace1190_0 -> 21 in the cache */
   /* replace1190_1 -> 1494 in the cache */
   /* replace1190 -> 3374 in the cache */

   /* ('extract_u8', ('ishl', 'a@64', 56), 7) => ('extract_u8', 'a', 0) */
   /* search1191_0_0 -> 21 in the cache */
   /* search1191_0_1 -> 3236 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 3236 },
      -1,
   } },
   /* search1191_1 -> 3238 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3435, 3238 },
      -1,
   } },

   /* replace1191_0 -> 21 in the cache */
   /* replace1191_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 21, 897 },
      -1,
   } },

   /* ('extract_i8', ('ishl', 'a@16', 8), 1) => ('extract_i8', 'a', 0) */
   /* search1192_0_0 -> 11 in the cache */
   /* search1192_0_1 -> 831 in the cache */
   /* search1192_0 -> 3416 in the cache */
   /* search1192_1 -> 1452 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3416, 1452 },
      -1,
   } },

   /* replace1192_0 -> 11 in the cache */
   /* replace1192_1 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 11, 856 },
      -1,
   } },

   /* ('extract_i8', ('ishl', 'a@32', 8), 3) => ('extract_i8', 'a', 2) */
   /* search1193_0_0 -> 16 in the cache */
   /* search1193_0_1 -> 831 in the cache */
   /* search1193_0 -> 3419 in the cache */
   /* search1193_1 -> 405 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3419, 405 },
      -1,
   } },

   /* replace1193_0 -> 16 in the cache */
   /* replace1193_1 -> 401 in the cache */
   /* replace1193 -> 402 in the cache */

   /* ('extract_i8', ('ishl', 'a@32', 16), 3) => ('extract_i8', 'a', 1) */
   /* search1194_0_0 -> 16 in the cache */
   /* search1194_0_1 -> 853 in the cache */
   /* search1194_0 -> 1005 in the cache */
   /* search1194_1 -> 405 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 1005, 405 },
      -1,
   } },

   /* replace1194_0 -> 16 in the cache */
   /* replace1194_1 -> 396 in the cache */
   /* replace1194 -> 397 in the cache */

   /* ('extract_i8', ('ishl', 'a@32', 24), 3) => ('extract_i8', 'a', 0) */
   /* search1195_0_0 -> 16 in the cache */
   /* search1195_0_1 -> 1012 in the cache */
   /* search1195_0 -> 1014 in the cache */
   /* search1195_1 -> 405 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 1014, 405 },
      -1,
   } },

   /* replace1195_0 -> 16 in the cache */
   /* replace1195_1 -> 122 in the cache */
   /* replace1195 -> 393 in the cache */

   /* ('extract_i8', ('ishl', 'a@64', 8), 7) => ('extract_i8', 'a', 6) */
   /* search1196_0_0 -> 21 in the cache */
   /* search1196_0_1 -> 831 in the cache */
   /* search1196_0 -> 3423 in the cache */
   /* search1196_1 -> 3238 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3423, 3238 },
      -1,
   } },

   /* replace1196_0 -> 21 in the cache */
   /* replace1196_1 -> 3394 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 21, 3394 },
      -1,
   } },

   /* ('extract_i8', ('ishl', 'a@64', 16), 7) => ('extract_i8', 'a', 5) */
   /* search1197_0_0 -> 21 in the cache */
   /* search1197_0_1 -> 853 in the cache */
   /* search1197_0 -> 3425 in the cache */
   /* search1197_1 -> 3238 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3425, 3238 },
      -1,
   } },

   /* replace1197_0 -> 21 in the cache */
   /* replace1197_1 -> 3389 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 21, 3389 },
      -1,
   } },

   /* ('extract_i8', ('ishl', 'a@64', 24), 7) => ('extract_i8', 'a', 4) */
   /* search1198_0_0 -> 21 in the cache */
   /* search1198_0_1 -> 1012 in the cache */
   /* search1198_0 -> 3427 in the cache */
   /* search1198_1 -> 3238 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3427, 3238 },
      -1,
   } },

   /* replace1198_0 -> 21 in the cache */
   /* replace1198_1 -> 3384 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 21, 3384 },
      -1,
   } },

   /* ('extract_i8', ('ishl', 'a@64', 32), 7) => ('extract_i8', 'a', 3) */
   /* search1199_0_0 -> 21 in the cache */
   /* search1199_0_1 -> 874 in the cache */
   /* search1199_0 -> 3429 in the cache */
   /* search1199_1 -> 3238 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3429, 3238 },
      -1,
   } },

   /* replace1199_0 -> 21 in the cache */
   /* replace1199_1 -> 3381 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 21, 3381 },
      -1,
   } },

   /* ('extract_i8', ('ishl', 'a@64', 40), 7) => ('extract_i8', 'a', 2) */
   /* search1200_0_0 -> 21 in the cache */
   /* search1200_0_1 -> 3386 in the cache */
   /* search1200_0 -> 3431 in the cache */
   /* search1200_1 -> 3238 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3431, 3238 },
      -1,
   } },

   /* replace1200_0 -> 21 in the cache */
   /* replace1200_1 -> 3377 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 21, 3377 },
      -1,
   } },

   /* ('extract_i8', ('ishl', 'a@64', 48), 7) => ('extract_i8', 'a', 1) */
   /* search1201_0_0 -> 21 in the cache */
   /* search1201_0_1 -> 3391 in the cache */
   /* search1201_0 -> 3433 in the cache */
   /* search1201_1 -> 3238 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3433, 3238 },
      -1,
   } },

   /* replace1201_0 -> 21 in the cache */
   /* replace1201_1 -> 1494 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 21, 1494 },
      -1,
   } },

   /* ('extract_i8', ('ishl', 'a@64', 56), 7) => ('extract_i8', 'a', 0) */
   /* search1202_0_0 -> 21 in the cache */
   /* search1202_0_1 -> 3236 in the cache */
   /* search1202_0 -> 3435 in the cache */
   /* search1202_1 -> 3238 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3435, 3238 },
      -1,
   } },

   /* replace1202_0 -> 21 in the cache */
   /* replace1202_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 21, 897 },
      -1,
   } },

   /* ('ussub_4x8_vc4', 'a', 0) => a */
   /* search1203_0 -> 16 in the cache */
   /* search1203_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ussub_4x8_vc4,
      -1, 0,
      { 16, 122 },
      -1,
   } },

   /* replace1203 -> 16 in the cache */

   /* ('ussub_4x8_vc4', 'a', -1) => 0 */
   /* search1204_0 -> 16 in the cache */
   /* search1204_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ussub_4x8_vc4,
      -1, 0,
      { 16, 124 },
      -1,
   } },

   /* replace1204 -> 122 in the cache */

   /* ('fsub', 'a', 'b') => ('fadd', 'a', ('fneg', 'b')) */
   /* search1205_0 -> 0 in the cache */
   /* search1205_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* replace1205_0 -> 0 in the cache */
   /* replace1205_1_0 -> 3 in the cache */
   /* replace1205_1 -> 1087 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 0, 1087 },
      -1,
   } },

   /* ('isub', 'a', 'b') => ('iadd', 'a', ('ineg', 'b')) */
   /* search1206_0 -> 0 in the cache */
   /* search1206_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* replace1206_0 -> 0 in the cache */
   /* replace1206_1_0 -> 3 in the cache */
   /* replace1206_1 -> 93 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 0, 93 },
      -1,
   } },

   /* ('uabs_usub', 'a', 'b') => ('bcsel', ('ult', 'a', 'b'), ('ineg', ('isub', 'a', 'b')), ('isub', 'a', 'b')) */
   /* search1207_0 -> 0 in the cache */
   /* search1207_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_uabs_usub,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* replace1207_0_0 -> 0 in the cache */
   /* replace1207_0_1 -> 3 in the cache */
   /* replace1207_0 -> 1067 in the cache */
   /* replace1207_1_0_0 -> 0 in the cache */
   /* replace1207_1_0_1 -> 3 in the cache */
   /* replace1207_1_0 -> 3461 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 3461 },
      -1,
   } },
   /* replace1207_2_0 -> 0 in the cache */
   /* replace1207_2_1 -> 3 in the cache */
   /* replace1207_2 -> 3461 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1067, 3464, 3461 },
      -1,
   } },

   /* ('uabs_isub', 'a', 'b') => ('bcsel', ('ilt', 'a', 'b'), ('ineg', ('isub', 'a', 'b')), ('isub', 'a', 'b')) */
   /* search1208_0 -> 0 in the cache */
   /* search1208_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_uabs_isub,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* replace1208_0_0 -> 0 in the cache */
   /* replace1208_0_1 -> 3 in the cache */
   /* replace1208_0 -> 1064 in the cache */
   /* replace1208_1_0_0 -> 0 in the cache */
   /* replace1208_1_0_1 -> 3 in the cache */
   /* replace1208_1_0 -> 3461 in the cache */
   /* replace1208_1 -> 3464 in the cache */
   /* replace1208_2_0 -> 0 in the cache */
   /* replace1208_2_1 -> 3 in the cache */
   /* replace1208_2 -> 3461 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1064, 3464, 3461 },
      -1,
   } },

   /* ('fmul(is_used_by_non_fsat)', ('fneg', 'a'), 'b') => ('fneg', ('fmul', 'a', 'b')) */
   /* search1209_0_0 -> 0 in the cache */
   /* search1209_0 -> 212 in the cache */
   /* search1209_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 212, 3 },
      3,
   } },

   /* replace1209_0_0 -> 0 in the cache */
   /* replace1209_0_1 -> 3 in the cache */
   /* replace1209_0 -> 264 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 264 },
      -1,
   } },

   /* ('fmulz(is_used_by_non_fsat)', ('fneg', 'a'), 'b') => ('fneg', ('fmulz', 'a', 'b')) */
   /* search1210_0_0 -> 16 in the cache */
   /* search1210_0 -> 280 in the cache */
   /* search1210_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 280, 36 },
      3,
   } },

   /* replace1210_0_0 -> 16 in the cache */
   /* replace1210_0_1 -> 36 in the cache */
   /* replace1210_0 -> 269 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 269 },
      -1,
   } },

   /* ('ffma', ('fneg', 'a'), ('fneg', 'b'), 'c') => ('ffma', 'a', 'b', 'c') */
   /* search1211_0_0 -> 126 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 126 },
      -1,
   } },
   /* search1211_1_0 -> 127 in the cache */
   /* search1211_1 -> 1197 in the cache */
   /* search1211_2 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 3472, 1197, 129 },
      -1,
   } },

   /* replace1211_0 -> 126 in the cache */
   /* replace1211_1 -> 127 in the cache */
   /* replace1211_2 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 126, 127, 129 },
      -1,
   } },

   /* ('ffmaz', ('fneg', 'a'), ('fneg', 'b'), 'c') => ('ffmaz', 'a', 'b', 'c') */
   /* search1212_0_0 -> 16 in the cache */
   /* search1212_0 -> 280 in the cache */
   /* search1212_1_0 -> 36 in the cache */
   /* search1212_1 -> 600 in the cache */
   /* search1212_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 280, 600, 135 },
      -1,
   } },

   /* replace1212_0 -> 16 in the cache */
   /* replace1212_1 -> 36 in the cache */
   /* replace1212_2 -> 135 in the cache */
   /* replace1212 -> 387 in the cache */

   /* ('imul', ('ineg', 'a'), 'b') => ('ineg', ('imul', 'a', 'b')) */
   /* search1213_0_0 -> 0 in the cache */
   /* search1213_0 -> 207 in the cache */
   /* search1213_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 207, 3 },
      -1,
   } },

   /* replace1213_0_0 -> 0 in the cache */
   /* replace1213_0_1 -> 3 in the cache */
   /* replace1213_0 -> 951 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 1,
      { 951 },
      -1,
   } },

   /* ('~fmul(is_used_once)', ('fmul(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c') => ('fmul', ('fmul', 'a', 'c'), 'b') */
   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* a */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* b */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 3478, 3479 },
      0,
   } },
   /* search1214_1 -> 814 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 3480, 814 },
      0,
   } },

   /* replace1214_0_0 -> 126 in the cache */
   /* replace1214_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 126, 129 },
      -1,
   } },
   /* replace1214_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 3482, 127 },
      -1,
   } },

   /* ('~fmulz(is_used_once)', ('fmulz(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c') => ('fmulz', ('fmulz', 'a', 'c'), 'b') */
   /* search1215_0_0 -> 328 in the cache */
   /* search1215_0_1 -> 338 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 328, 338 },
      0,
   } },
   /* search1215_1 -> 309 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fmulz,
      0, 2,
      { 3484, 309 },
      0,
   } },

   /* replace1215_0_0 -> 16 in the cache */
   /* replace1215_0_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 16, 135 },
      -1,
   } },
   /* replace1215_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 2,
      { 3486, 36 },
      -1,
   } },

   /* ('~fmul(is_used_once)', ('fmulz(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c(is_finite_not_zero)') => ('fmulz', ('fmul', 'a', 'c'), 'b') */
   /* search1216_0_0 -> 328 in the cache */
   /* search1216_0_1 -> 338 in the cache */
   /* search1216_0 -> 3484 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* c */
      true,
      nir_type_invalid,
      3,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 3484, 3488 },
      0,
   } },

   /* replace1216_0_0 -> 16 in the cache */
   /* replace1216_0_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 16, 135 },
      -1,
   } },
   /* replace1216_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 2,
      { 3490, 36 },
      -1,
   } },

   /* ('imul(is_used_once)', ('imul(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c') => ('imul', ('imul', 'a', 'c'), 'b') */
   /* search1217_0_0 -> 3478 in the cache */
   /* search1217_0_1 -> 3479 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 3478, 3479 },
      0,
   } },
   /* search1217_1 -> 814 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      0, 2,
      { 3492, 814 },
      0,
   } },

   /* replace1217_0_0 -> 126 in the cache */
   /* replace1217_0_1 -> 129 in the cache */
   /* replace1217_0 -> 1025 in the cache */
   /* replace1217_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      0, 2,
      { 1025, 127 },
      -1,
   } },

   /* ('~ffma', ('fmul(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c', 'd') => ('ffma', ('fmul', 'a', 'c'), 'b', 'd') */
   { .variable = {
      { nir_search_value_variable, -4 },
      0, /* a */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      1, /* b */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 3495, 3496 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      2, /* c */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1218_2 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 3497, 3498, 153 },
      -1,
   } },

   /* replace1218_0_0 -> 149 in the cache */
   /* replace1218_0_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 149, 152 },
      -1,
   } },
   /* replace1218_1 -> 150 in the cache */
   /* replace1218_2 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 3500, 150, 153 },
      -1,
   } },

   /* ('~ffmaz', ('fmulz(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c', 'd') => ('ffmaz', ('fmulz', 'a', 'c'), 'b', 'd') */
   /* search1219_0_0 -> 328 in the cache */
   /* search1219_0_1 -> 338 in the cache */
   /* search1219_0 -> 3484 in the cache */
   /* search1219_1 -> 309 in the cache */
   /* search1219_2 -> 163 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffmaz,
      0, 2,
      { 3484, 309, 163 },
      -1,
   } },

   /* replace1219_0_0 -> 16 in the cache */
   /* replace1219_0_1 -> 135 in the cache */
   /* replace1219_0 -> 3486 in the cache */
   /* replace1219_1 -> 36 in the cache */
   /* replace1219_2 -> 163 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 2,
      { 3486, 36, 163 },
      -1,
   } },

   /* ('~ffma', ('fmulz(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c(is_finite_not_zero)', 'd') => ('ffmaz', ('fmul', 'a', 'c'), 'b', 'd') */
   /* search1220_0_0 -> 328 in the cache */
   /* search1220_0_1 -> 338 in the cache */
   /* search1220_0 -> 3484 in the cache */
   /* search1220_1 -> 3488 in the cache */
   /* search1220_2 -> 163 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 3484, 3488, 163 },
      -1,
   } },

   /* replace1220_0_0 -> 16 in the cache */
   /* replace1220_0_1 -> 135 in the cache */
   /* replace1220_0 -> 3490 in the cache */
   /* replace1220_1 -> 36 in the cache */
   /* replace1220_2 -> 163 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 2,
      { 3490, 36, 163 },
      -1,
   } },

   /* ('~fadd(is_used_once)', ('fadd(is_used_once)', 'a(is_not_const)', 'b(is_fmul)'), '#c') => ('fadd', ('fadd', 'a', 'c'), 'b') */
   /* search1221_0_0 -> 3478 in the cache */
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* b */
      false,
      nir_type_invalid,
      33,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 3478, 3506 },
      0,
   } },
   /* search1221_1 -> 814 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 3507, 814 },
      0,
   } },

   /* replace1221_0_0 -> 126 in the cache */
   /* replace1221_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 126, 129 },
      -1,
   } },
   /* replace1221_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 3509, 127 },
      -1,
   } },

   /* ('~fadd(is_used_once)', ('fadd(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c') => ('fadd', ('fadd', 'a', 'c'), 'b') */
   /* search1222_0_0 -> 3478 in the cache */
   /* search1222_0_1 -> 3479 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 3478, 3479 },
      0,
   } },
   /* search1222_1 -> 814 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 3511, 814 },
      0,
   } },

   /* replace1222_0_0 -> 126 in the cache */
   /* replace1222_0_1 -> 129 in the cache */
   /* replace1222_0 -> 3509 in the cache */
   /* replace1222_1 -> 127 in the cache */
   /* replace1222 -> 3510 in the cache */

   /* ('~fadd(is_used_once)', ('ffma(is_used_once)', 'a(is_not_const)', 'b', 'c(is_not_const)'), '#d') => ('fadd', ('ffma', 'a', 'b', 'd'), 'c') */
   /* search1223_0_0 -> 3495 in the cache */
   /* search1223_0_1 -> 150 in the cache */
   { .variable = {
      { nir_search_value_variable, -4 },
      2, /* c */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 3495, 150, 3513 },
      0,
   } },
   /* search1223_1 -> 1031 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 3514, 1031 },
      0,
   } },

   /* replace1223_0_0 -> 149 in the cache */
   /* replace1223_0_1 -> 150 in the cache */
   /* replace1223_0_2 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 149, 150, 153 },
      -1,
   } },
   /* replace1223_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 3516, 152 },
      -1,
   } },

   /* ('~fadd(is_used_once)', ('ffmaz(is_used_once)', 'a(is_not_const)', 'b', 'c(is_not_const)'), '#d') => ('fadd', ('ffmaz', 'a', 'b', 'd'), 'c') */
   /* search1224_0_0 -> 328 in the cache */
   /* search1224_0_1 -> 36 in the cache */
   /* search1224_0_2 -> 347 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 1,
      { 328, 36, 347 },
      0,
   } },
   /* search1224_1 -> 311 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 3518, 311 },
      0,
   } },

   /* replace1224_0_0 -> 16 in the cache */
   /* replace1224_0_1 -> 36 in the cache */
   /* replace1224_0_2 -> 163 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 1,
      { 16, 36, 163 },
      -1,
   } },
   /* replace1224_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 3520, 135 },
      -1,
   } },

   /* ('iadd(is_used_once)', ('iadd(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c') => ('iadd', ('iadd', 'a', 'c'), 'b') */
   /* search1225_0_0 -> 3478 in the cache */
   /* search1225_0_1 -> 3479 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 3478, 3479 },
      0,
   } },
   /* search1225_1 -> 814 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 3522, 814 },
      0,
   } },

   /* replace1225_0_0 -> 126 in the cache */
   /* replace1225_0_1 -> 129 in the cache */
   /* replace1225_0 -> 1039 in the cache */
   /* replace1225_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 1039, 127 },
      -1,
   } },

   /* ('~fmul', '#a', ('fmul', 'b(is_not_const)', '#c')) => ('fmul', ('fmul', 'a', 'c'), 'b') */
   /* search1226_0 -> 282 in the cache */
   /* search1226_1_0 -> 3479 in the cache */
   /* search1226_1_1 -> 814 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 3479, 814 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 282, 3525 },
      -1,
   } },

   /* replace1226_0_0 -> 126 in the cache */
   /* replace1226_0_1 -> 129 in the cache */
   /* replace1226_0 -> 3482 in the cache */
   /* replace1226_1 -> 127 in the cache */
   /* replace1226 -> 3483 in the cache */

   /* ('~fmulz', '#a', ('fmulz', 'b(is_not_const)', '#c')) => ('fmulz', ('fmulz', 'a', 'c'), 'b') */
   /* search1227_0 -> 286 in the cache */
   /* search1227_1_0 -> 338 in the cache */
   /* search1227_1_1 -> 309 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 338, 309 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fmulz,
      0, 2,
      { 286, 3527 },
      -1,
   } },

   /* replace1227_0_0 -> 16 in the cache */
   /* replace1227_0_1 -> 135 in the cache */
   /* replace1227_0 -> 3486 in the cache */
   /* replace1227_1 -> 36 in the cache */
   /* replace1227 -> 3487 in the cache */

   /* ('~fmul', '#a(is_finite_not_zero)', ('fmulz', 'b(is_not_const)', '#c')) => ('fmulz', ('fmul', 'a', 'c'), 'b') */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      true,
      nir_type_invalid,
      3,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1228_1_0 -> 338 in the cache */
   /* search1228_1_1 -> 309 in the cache */
   /* search1228_1 -> 3527 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 3529, 3527 },
      -1,
   } },

   /* replace1228_0_0 -> 16 in the cache */
   /* replace1228_0_1 -> 135 in the cache */
   /* replace1228_0 -> 3490 in the cache */
   /* replace1228_1 -> 36 in the cache */
   /* replace1228 -> 3491 in the cache */

   /* ('~ffma', '#a', ('fmul', 'b(is_not_const)', '#c'), 'd') => ('ffma', ('fmul', 'a', 'c'), 'b', 'd') */
   { .variable = {
      { nir_search_value_variable, -4 },
      0, /* a */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1229_1_0 -> 3496 in the cache */
   /* search1229_1_1 -> 3498 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 3496, 3498 },
      -1,
   } },
   /* search1229_2 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 3531, 3532, 153 },
      -1,
   } },

   /* replace1229_0_0 -> 149 in the cache */
   /* replace1229_0_1 -> 152 in the cache */
   /* replace1229_0 -> 3500 in the cache */
   /* replace1229_1 -> 150 in the cache */
   /* replace1229_2 -> 153 in the cache */
   /* replace1229 -> 3501 in the cache */

   /* ('~ffmaz', '#a', ('fmulz', 'b(is_not_const)', '#c'), 'd') => ('ffmaz', ('fmulz', 'a', 'c'), 'b', 'd') */
   /* search1230_0 -> 286 in the cache */
   /* search1230_1_0 -> 338 in the cache */
   /* search1230_1_1 -> 309 in the cache */
   /* search1230_1 -> 3527 in the cache */
   /* search1230_2 -> 163 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffmaz,
      0, 2,
      { 286, 3527, 163 },
      -1,
   } },

   /* replace1230_0_0 -> 16 in the cache */
   /* replace1230_0_1 -> 135 in the cache */
   /* replace1230_0 -> 3486 in the cache */
   /* replace1230_1 -> 36 in the cache */
   /* replace1230_2 -> 163 in the cache */
   /* replace1230 -> 3503 in the cache */

   /* ('~ffmaz', '#a(is_finite_not_zero)', ('fmulz', 'b(is_not_const)', '#c'), 'd') => ('ffmaz', ('fmul', 'a', 'c'), 'b', 'd') */
   /* search1231_0 -> 3529 in the cache */
   /* search1231_1_0 -> 338 in the cache */
   /* search1231_1_1 -> 309 in the cache */
   /* search1231_1 -> 3527 in the cache */
   /* search1231_2 -> 163 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffmaz,
      0, 2,
      { 3529, 3527, 163 },
      -1,
   } },

   /* replace1231_0_0 -> 16 in the cache */
   /* replace1231_0_1 -> 135 in the cache */
   /* replace1231_0 -> 3490 in the cache */
   /* replace1231_1 -> 36 in the cache */
   /* replace1231_2 -> 163 in the cache */
   /* replace1231 -> 3505 in the cache */

   /* ('imul', '#a', ('imul', 'b(is_not_const)', '#c')) => ('imul', ('imul', 'a', 'c'), 'b') */
   /* search1232_0 -> 282 in the cache */
   /* search1232_1_0 -> 3479 in the cache */
   /* search1232_1_1 -> 814 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 3479, 814 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      0, 2,
      { 282, 3536 },
      -1,
   } },

   /* replace1232_0_0 -> 126 in the cache */
   /* replace1232_0_1 -> 129 in the cache */
   /* replace1232_0 -> 1025 in the cache */
   /* replace1232_1 -> 127 in the cache */
   /* replace1232 -> 3494 in the cache */

   /* ('~fadd', '#a', ('fadd', 'b(is_not_const)', '#c')) => ('fadd', ('fadd', 'a', 'c'), 'b') */
   /* search1233_0 -> 282 in the cache */
   /* search1233_1_0 -> 3479 in the cache */
   /* search1233_1_1 -> 814 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 3479, 814 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 282, 3538 },
      -1,
   } },

   /* replace1233_0_0 -> 126 in the cache */
   /* replace1233_0_1 -> 129 in the cache */
   /* replace1233_0 -> 3509 in the cache */
   /* replace1233_1 -> 127 in the cache */
   /* replace1233 -> 3510 in the cache */

   /* ('~fadd', '#a', ('fneg', ('fadd', 'b(is_not_const)', '#c'))) => ('fadd', ('fadd', 'a', ('fneg', 'c')), ('fneg', 'b')) */
   /* search1234_0 -> 282 in the cache */
   /* search1234_1_0_0 -> 3479 in the cache */
   /* search1234_1_0_1 -> 814 in the cache */
   /* search1234_1_0 -> 3538 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 3538 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 282, 3540 },
      -1,
   } },

   /* replace1234_0_0 -> 126 in the cache */
   /* replace1234_0_1_0 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 126, 3542 },
      -1,
   } },
   /* replace1234_1_0 -> 127 in the cache */
   /* replace1234_1 -> 1197 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 3543, 1197 },
      -1,
   } },

   /* ('~fadd', '#a', ('ffma', 'b(is_not_const)', 'c(is_not_const)', '#d')) => ('ffma', 'b', 'c', ('fadd', 'a', 'd')) */
   /* search1235_0 -> 3531 in the cache */
   /* search1235_1_0 -> 3496 in the cache */
   /* search1235_1_1 -> 3513 in the cache */
   /* search1235_1_2 -> 1031 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 3496, 3513, 1031 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 3531, 3545 },
      -1,
   } },

   /* replace1235_0 -> 150 in the cache */
   /* replace1235_1 -> 152 in the cache */
   /* replace1235_2_0 -> 149 in the cache */
   /* replace1235_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 149, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 150, 152, 3547 },
      -1,
   } },

   /* ('~fadd', '#a', ('fneg', ('ffma', 'b(is_not_const)', 'c(is_not_const)', '#d'))) => ('ffma', ('fneg', 'b'), 'c', ('fadd', 'a', ('fneg', 'd'))) */
   /* search1236_0 -> 3531 in the cache */
   /* search1236_1_0_0 -> 3496 in the cache */
   /* search1236_1_0_1 -> 3513 in the cache */
   /* search1236_1_0_2 -> 1031 in the cache */
   /* search1236_1_0 -> 3545 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 3545 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 3531, 3549 },
      -1,
   } },

   /* replace1236_0_0 -> 150 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 150 },
      -1,
   } },
   /* replace1236_1 -> 152 in the cache */
   /* replace1236_2_0 -> 149 in the cache */
   /* replace1236_2_1_0 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 149, 3552 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 3551, 152, 3553 },
      -1,
   } },

   /* ('~fadd', '#a', ('ffmaz', 'b(is_not_const)', 'c(is_not_const)', '#d')) => ('ffmaz', 'b', 'c', ('fadd', 'a', 'd')) */
   /* search1237_0 -> 286 in the cache */
   /* search1237_1_0 -> 338 in the cache */
   /* search1237_1_1 -> 347 in the cache */
   /* search1237_1_2 -> 311 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 1,
      { 338, 347, 311 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 286, 3555 },
      -1,
   } },

   /* replace1237_0 -> 36 in the cache */
   /* replace1237_1 -> 135 in the cache */
   /* replace1237_2_0 -> 16 in the cache */
   /* replace1237_2_1 -> 163 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 16, 163 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 2,
      { 36, 135, 3557 },
      -1,
   } },

   /* ('~fadd', '#a', ('fneg', ('ffmaz', 'b(is_not_const)', 'c(is_not_const)', '#d'))) => ('ffmaz', ('fneg', 'b'), 'c', ('fadd', 'a', ('fneg', 'd'))) */
   /* search1238_0 -> 286 in the cache */
   /* search1238_1_0_0 -> 338 in the cache */
   /* search1238_1_0_1 -> 347 in the cache */
   /* search1238_1_0_2 -> 311 in the cache */
   /* search1238_1_0 -> 3555 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 3555 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 286, 3559 },
      -1,
   } },

   /* replace1238_0_0 -> 36 in the cache */
   /* replace1238_0 -> 600 in the cache */
   /* replace1238_1 -> 135 in the cache */
   /* replace1238_2_0 -> 16 in the cache */
   /* replace1238_2_1_0 -> 163 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 163 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 16, 3561 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 2,
      { 600, 135, 3562 },
      -1,
   } },

   /* ('iadd', '#a', ('iadd', 'b(is_not_const)', '#c')) => ('iadd', ('iadd', 'a', 'c'), 'b') */
   /* search1239_0 -> 282 in the cache */
   /* search1239_1_0 -> 3479 in the cache */
   /* search1239_1_1 -> 814 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 3479, 814 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 282, 3564 },
      -1,
   } },

   /* replace1239_0_0 -> 126 in the cache */
   /* replace1239_0_1 -> 129 in the cache */
   /* replace1239_0 -> 1039 in the cache */
   /* replace1239_1 -> 127 in the cache */
   /* replace1239 -> 3524 in the cache */

   /* ('iand', '#a', ('iand', 'b(is_not_const)', '#c')) => ('iand', ('iand', 'a', 'c'), 'b') */
   /* search1240_0 -> 282 in the cache */
   /* search1240_1_0 -> 3479 in the cache */
   /* search1240_1_1 -> 814 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 3479, 814 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 282, 3566 },
      -1,
   } },

   /* replace1240_0_0 -> 126 in the cache */
   /* replace1240_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 126, 129 },
      -1,
   } },
   /* replace1240_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 3568, 127 },
      -1,
   } },

   /* ('ior', '#a', ('ior', 'b(is_not_const)', '#c')) => ('ior', ('ior', 'a', 'c'), 'b') */
   /* search1241_0 -> 282 in the cache */
   /* search1241_1_0 -> 3479 in the cache */
   /* search1241_1_1 -> 814 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 3479, 814 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 282, 3570 },
      -1,
   } },

   /* replace1241_0_0 -> 126 in the cache */
   /* replace1241_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 126, 129 },
      -1,
   } },
   /* replace1241_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 3572, 127 },
      -1,
   } },

   /* ('ixor', '#a', ('ixor', 'b(is_not_const)', '#c')) => ('ixor', ('ixor', 'a', 'c'), 'b') */
   /* search1242_0 -> 282 in the cache */
   /* search1242_1_0 -> 3479 in the cache */
   /* search1242_1_1 -> 814 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ixor,
      1, 1,
      { 3479, 814 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 2,
      { 282, 3574 },
      -1,
   } },

   /* replace1242_0_0 -> 126 in the cache */
   /* replace1242_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ixor,
      1, 1,
      { 126, 129 },
      -1,
   } },
   /* replace1242_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 2,
      { 3576, 127 },
      -1,
   } },

   /* ('~fadd', ('fadd(is_used_once)', 'a(is_fmul)', 'b(is_fmul)'), 'c(is_not_fmul)') => ('fadd', ('fadd', 'a', 'c'), 'b') */
   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* a */
      false,
      nir_type_invalid,
      33,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1243_0_1 -> 3506 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 3578, 3506 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* c */
      false,
      nir_type_invalid,
      8,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 3579, 3580 },
      -1,
   } },

   /* replace1243_0_0 -> 126 in the cache */
   /* replace1243_0_1 -> 129 in the cache */
   /* replace1243_0 -> 3509 in the cache */
   /* replace1243_1 -> 127 in the cache */
   /* replace1243 -> 3510 in the cache */

   /* ('idiv', ('imul(no_signed_wrap)', 'a', 'b'), 'b') => a */
   /* search1244_0_0 -> 0 in the cache */
   /* search1244_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 0, 3 },
      4,
   } },
   /* search1244_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_idiv,
      -1, 1,
      { 3582, 3 },
      -1,
   } },

   /* replace1244 -> 0 in the cache */

   /* ('bcsel', ('ige', ('find_lsb', 'a'), 0), ('find_lsb', 'a'), -1) => ('find_lsb', 'a') */
   /* search1245_0_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_find_lsb,
      -1, 0,
      { 32 },
      -1,
   } },
   /* search1245_0_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3584, 122 },
      -1,
   } },
   /* search1245_1_0 -> 32 in the cache */
   /* search1245_1 -> 3584 in the cache */
   /* search1245_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3585, 3584, 124 },
      -1,
   } },

   /* replace1245_0 -> 32 in the cache */
   /* replace1245 -> 3584 in the cache */

   /* ('bcsel', ('ige', ('ifind_msb', 'a'), 0), ('ifind_msb', 'a'), -1) => ('ifind_msb', 'a') */
   /* search1246_0_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ifind_msb,
      -1, 0,
      { 16 },
      -1,
   } },
   /* search1246_0_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3587, 122 },
      -1,
   } },
   /* search1246_1_0 -> 16 in the cache */
   /* search1246_1 -> 3587 in the cache */
   /* search1246_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3588, 3587, 124 },
      -1,
   } },

   /* replace1246_0 -> 16 in the cache */
   /* replace1246 -> 3587 in the cache */

   /* ('bcsel', ('ige', ('ufind_msb', 'a'), 0), ('ufind_msb', 'a'), -1) => ('ufind_msb', 'a') */
   /* search1247_0_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ufind_msb,
      -1, 0,
      { 32 },
      -1,
   } },
   /* search1247_0_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3590, 122 },
      -1,
   } },
   /* search1247_1_0 -> 32 in the cache */
   /* search1247_1 -> 3590 in the cache */
   /* search1247_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3591, 3590, 124 },
      -1,
   } },

   /* replace1247_0 -> 32 in the cache */
   /* replace1247 -> 3590 in the cache */

   /* ('bcsel', ('ige', ('ifind_msb_rev', 'a'), 0), ('ifind_msb_rev', 'a'), -1) => ('ifind_msb_rev', 'a') */
   /* search1248_0_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ifind_msb_rev,
      -1, 0,
      { 32 },
      -1,
   } },
   /* search1248_0_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3593, 122 },
      -1,
   } },
   /* search1248_1_0 -> 32 in the cache */
   /* search1248_1 -> 3593 in the cache */
   /* search1248_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3594, 3593, 124 },
      -1,
   } },

   /* replace1248_0 -> 32 in the cache */
   /* replace1248 -> 3593 in the cache */

   /* ('bcsel', ('ige', ('ufind_msb_rev', 'a'), 0), ('ufind_msb_rev', 'a'), -1) => ('ufind_msb_rev', 'a') */
   /* search1249_0_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ufind_msb_rev,
      -1, 0,
      { 32 },
      -1,
   } },
   /* search1249_0_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3596, 122 },
      -1,
   } },
   /* search1249_1_0 -> 32 in the cache */
   /* search1249_1 -> 3596 in the cache */
   /* search1249_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3597, 3596, 124 },
      -1,
   } },

   /* replace1249_0 -> 32 in the cache */
   /* replace1249 -> 3596 in the cache */

   /* ('bcsel', ('ine', 'a', 0), ('find_lsb', 'a'), -1) => ('find_lsb', 'a') */
   /* search1250_0_0 -> 32 in the cache */
   /* search1250_0_1 -> 61 in the cache */
   /* search1250_0 -> 1110 in the cache */
   /* search1250_1_0 -> 32 in the cache */
   /* search1250_1 -> 3584 in the cache */
   /* search1250_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 1110, 3584, 124 },
      -1,
   } },

   /* replace1250_0 -> 32 in the cache */
   /* replace1250 -> 3584 in the cache */

   /* ('bcsel', ('ine', 'a', 0), ('ifind_msb', 'a'), -1) => ('ifind_msb', 'a') */
   /* search1251_0_0 -> 16 in the cache */
   /* search1251_0_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 16, 122 },
      -1,
   } },
   /* search1251_1_0 -> 16 in the cache */
   /* search1251_1 -> 3587 in the cache */
   /* search1251_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 3600, 3587, 124 },
      -1,
   } },

   /* replace1251_0 -> 16 in the cache */
   /* replace1251 -> 3587 in the cache */

   /* ('bcsel', ('ine', 'a', 0), ('ufind_msb', 'a'), -1) => ('ufind_msb', 'a') */
   /* search1252_0_0 -> 32 in the cache */
   /* search1252_0_1 -> 61 in the cache */
   /* search1252_0 -> 1110 in the cache */
   /* search1252_1_0 -> 32 in the cache */
   /* search1252_1 -> 3590 in the cache */
   /* search1252_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 1110, 3590, 124 },
      -1,
   } },

   /* replace1252_0 -> 32 in the cache */
   /* replace1252 -> 3590 in the cache */

   /* ('bcsel', ('ine', 'a', 0), ('ifind_msb_rev', 'a'), -1) => ('ifind_msb_rev', 'a') */
   /* search1253_0_0 -> 32 in the cache */
   /* search1253_0_1 -> 61 in the cache */
   /* search1253_0 -> 1110 in the cache */
   /* search1253_1_0 -> 32 in the cache */
   /* search1253_1 -> 3593 in the cache */
   /* search1253_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 1110, 3593, 124 },
      -1,
   } },

   /* replace1253_0 -> 32 in the cache */
   /* replace1253 -> 3593 in the cache */

   /* ('bcsel', ('ine', 'a', 0), ('ufind_msb_rev', 'a'), -1) => ('ufind_msb_rev', 'a') */
   /* search1254_0_0 -> 32 in the cache */
   /* search1254_0_1 -> 61 in the cache */
   /* search1254_0 -> 1110 in the cache */
   /* search1254_1_0 -> 32 in the cache */
   /* search1254_1 -> 3596 in the cache */
   /* search1254_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 1110, 3596, 124 },
      -1,
   } },

   /* replace1254_0 -> 32 in the cache */
   /* replace1254 -> 3596 in the cache */

   /* ('bcsel', ('ine', 'a', -1), ('ifind_msb', 'a'), -1) => ('ifind_msb', 'a') */
   /* search1255_0_0 -> 16 in the cache */
   /* search1255_0_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 16, 124 },
      -1,
   } },
   /* search1255_1_0 -> 16 in the cache */
   /* search1255_1 -> 3587 in the cache */
   /* search1255_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 3605, 3587, 124 },
      -1,
   } },

   /* replace1255_0 -> 16 in the cache */
   /* replace1255 -> 3587 in the cache */

   /* ('bcsel', ('ine', 'a', -1), ('ifind_msb_rev', 'a'), -1) => ('ifind_msb_rev', 'a') */
   /* search1256_0_0 -> 32 in the cache */
   /* search1256_0_1 -> 63 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 32, 63 },
      -1,
   } },
   /* search1256_1_0 -> 32 in the cache */
   /* search1256_1 -> 3593 in the cache */
   /* search1256_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 3607, 3593, 124 },
      -1,
   } },

   /* replace1256_0 -> 32 in the cache */
   /* replace1256 -> 3593 in the cache */

   /* ('bcsel', ('ine', ('ifind_msb', 'a@32'), -1), ('iadd', 31, ('ineg', ('ifind_msb', 'a'))), -1) => ('ifind_msb_rev', 'a') */
   /* search1257_0_0_0 -> 16 in the cache */
   /* search1257_0_0 -> 3587 in the cache */
   /* search1257_0_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 3587, 124 },
      -1,
   } },
   /* search1257_1_0 -> 19 in the cache */
   /* search1257_1_1_0_0 -> 16 in the cache */
   /* search1257_1_1_0 -> 3587 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 3587 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 19, 3610 },
      -1,
   } },
   /* search1257_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3609, 3611, 124 },
      -1,
   } },

   /* replace1257_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ifind_msb_rev,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('bcsel', ('ine', ('ufind_msb', 'a@32'), -1), ('iadd', 31, ('ineg', ('ufind_msb', 'a'))), -1) => ('ufind_msb_rev', 'a') */
   /* search1258_0_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ufind_msb,
      -1, 0,
      { 16 },
      -1,
   } },
   /* search1258_0_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 3614, 124 },
      -1,
   } },
   /* search1258_1_0 -> 19 in the cache */
   /* search1258_1_1_0_0 -> 16 in the cache */
   /* search1258_1_1_0 -> 3614 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 3614 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 19, 3616 },
      -1,
   } },
   /* search1258_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3615, 3617, 124 },
      -1,
   } },

   /* replace1258_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ufind_msb_rev,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('bcsel', ('ieq', ('ifind_msb', 'a@32'), -1), -1, ('iadd', 31, ('ineg', ('ifind_msb', 'a')))) => ('ifind_msb_rev', 'a') */
   /* search1259_0_0_0 -> 16 in the cache */
   /* search1259_0_0 -> 3587 in the cache */
   /* search1259_0_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 3587, 124 },
      -1,
   } },
   /* search1259_1 -> 124 in the cache */
   /* search1259_2_0 -> 19 in the cache */
   /* search1259_2_1_0_0 -> 16 in the cache */
   /* search1259_2_1_0 -> 3587 in the cache */
   /* search1259_2_1 -> 3610 in the cache */
   /* search1259_2 -> 3611 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3620, 124, 3611 },
      -1,
   } },

   /* replace1259_0 -> 16 in the cache */
   /* replace1259 -> 3613 in the cache */

   /* ('bcsel', ('ieq', ('ufind_msb', 'a@32'), -1), -1, ('iadd', 31, ('ineg', ('ufind_msb', 'a')))) => ('ufind_msb_rev', 'a') */
   /* search1260_0_0_0 -> 16 in the cache */
   /* search1260_0_0 -> 3614 in the cache */
   /* search1260_0_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 3614, 124 },
      -1,
   } },
   /* search1260_1 -> 124 in the cache */
   /* search1260_2_0 -> 19 in the cache */
   /* search1260_2_1_0_0 -> 16 in the cache */
   /* search1260_2_1_0 -> 3614 in the cache */
   /* search1260_2_1 -> 3616 in the cache */
   /* search1260_2 -> 3617 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3622, 124, 3617 },
      -1,
   } },

   /* replace1260_0 -> 16 in the cache */
   /* replace1260 -> 3619 in the cache */

   /* ('bcsel', ('ine', ('ifind_msb', 'a@32'), -1), ('iadd', 31, ('ineg', ('ifind_msb', 'a'))), ('ifind_msb', 'a')) => ('ifind_msb_rev', 'a') */
   /* search1261_0_0_0 -> 16 in the cache */
   /* search1261_0_0 -> 3587 in the cache */
   /* search1261_0_1 -> 124 in the cache */
   /* search1261_0 -> 3609 in the cache */
   /* search1261_1_0 -> 19 in the cache */
   /* search1261_1_1_0_0 -> 16 in the cache */
   /* search1261_1_1_0 -> 3587 in the cache */
   /* search1261_1_1 -> 3610 in the cache */
   /* search1261_1 -> 3611 in the cache */
   /* search1261_2_0 -> 16 in the cache */
   /* search1261_2 -> 3587 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3609, 3611, 3587 },
      -1,
   } },

   /* replace1261_0 -> 16 in the cache */
   /* replace1261 -> 3613 in the cache */

   /* ('bcsel', ('ine', ('ufind_msb', 'a@32'), -1), ('iadd', 31, ('ineg', ('ufind_msb', 'a'))), ('ufind_msb', 'a')) => ('ufind_msb_rev', 'a') */
   /* search1262_0_0_0 -> 16 in the cache */
   /* search1262_0_0 -> 3614 in the cache */
   /* search1262_0_1 -> 124 in the cache */
   /* search1262_0 -> 3615 in the cache */
   /* search1262_1_0 -> 19 in the cache */
   /* search1262_1_1_0_0 -> 16 in the cache */
   /* search1262_1_1_0 -> 3614 in the cache */
   /* search1262_1_1 -> 3616 in the cache */
   /* search1262_1 -> 3617 in the cache */
   /* search1262_2_0 -> 16 in the cache */
   /* search1262_2 -> 3614 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3615, 3617, 3614 },
      -1,
   } },

   /* replace1262_0 -> 16 in the cache */
   /* replace1262 -> 3619 in the cache */

   /* ('bcsel', ('ieq', ('ifind_msb', 'a@32'), -1), ('ifind_msb', 'a'), ('iadd', 31, ('ineg', ('ifind_msb', 'a')))) => ('ifind_msb_rev', 'a') */
   /* search1263_0_0_0 -> 16 in the cache */
   /* search1263_0_0 -> 3587 in the cache */
   /* search1263_0_1 -> 124 in the cache */
   /* search1263_0 -> 3620 in the cache */
   /* search1263_1_0 -> 16 in the cache */
   /* search1263_1 -> 3587 in the cache */
   /* search1263_2_0 -> 19 in the cache */
   /* search1263_2_1_0_0 -> 16 in the cache */
   /* search1263_2_1_0 -> 3587 in the cache */
   /* search1263_2_1 -> 3610 in the cache */
   /* search1263_2 -> 3611 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3620, 3587, 3611 },
      -1,
   } },

   /* replace1263_0 -> 16 in the cache */
   /* replace1263 -> 3613 in the cache */

   /* ('bcsel', ('ieq', ('ufind_msb', 'a@32'), -1), ('ufind_msb', 'a'), ('iadd', 31, ('ineg', ('ufind_msb', 'a')))) => ('ufind_msb_rev', 'a') */
   /* search1264_0_0_0 -> 16 in the cache */
   /* search1264_0_0 -> 3614 in the cache */
   /* search1264_0_1 -> 124 in the cache */
   /* search1264_0 -> 3622 in the cache */
   /* search1264_1_0 -> 16 in the cache */
   /* search1264_1 -> 3614 in the cache */
   /* search1264_2_0 -> 19 in the cache */
   /* search1264_2_1_0_0 -> 16 in the cache */
   /* search1264_2_1_0 -> 3614 in the cache */
   /* search1264_2_1 -> 3616 in the cache */
   /* search1264_2 -> 3617 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3622, 3614, 3617 },
      -1,
   } },

   /* replace1264_0 -> 16 in the cache */
   /* replace1264 -> 3619 in the cache */

   /* ('bcsel', ('ine', 'a@32', 0), ('iadd', 31, ('ineg', ('ufind_msb', 'a'))), -1) => ('ufind_msb_rev', 'a') */
   /* search1265_0_0 -> 16 in the cache */
   /* search1265_0_1 -> 122 in the cache */
   /* search1265_0 -> 3600 in the cache */
   /* search1265_1_0 -> 19 in the cache */
   /* search1265_1_1_0_0 -> 16 in the cache */
   /* search1265_1_1_0 -> 3614 in the cache */
   /* search1265_1_1 -> 3616 in the cache */
   /* search1265_1 -> 3617 in the cache */
   /* search1265_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3600, 3617, 124 },
      -1,
   } },

   /* replace1265_0 -> 16 in the cache */
   /* replace1265 -> 3619 in the cache */

   /* ('bcsel', ('ieq', 'a@32', 0), -1, ('iadd', 31, ('ineg', ('ufind_msb', 'a')))) => ('ufind_msb_rev', 'a') */
   /* search1266_0_0 -> 16 in the cache */
   /* search1266_0_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 16, 122 },
      -1,
   } },
   /* search1266_1 -> 124 in the cache */
   /* search1266_2_0 -> 19 in the cache */
   /* search1266_2_1_0_0 -> 16 in the cache */
   /* search1266_2_1_0 -> 3614 in the cache */
   /* search1266_2_1 -> 3616 in the cache */
   /* search1266_2 -> 3617 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3629, 124, 3617 },
      -1,
   } },

   /* replace1266_0 -> 16 in the cache */
   /* replace1266 -> 3619 in the cache */

   /* ('bcsel', ('ine', 'a@32', 0), ('iadd', 31, ('ineg', ('ufind_msb', 'a'))), ('ufind_msb', 'a')) => ('ufind_msb_rev', 'a') */
   /* search1267_0_0 -> 16 in the cache */
   /* search1267_0_1 -> 122 in the cache */
   /* search1267_0 -> 3600 in the cache */
   /* search1267_1_0 -> 19 in the cache */
   /* search1267_1_1_0_0 -> 16 in the cache */
   /* search1267_1_1_0 -> 3614 in the cache */
   /* search1267_1_1 -> 3616 in the cache */
   /* search1267_1 -> 3617 in the cache */
   /* search1267_2_0 -> 16 in the cache */
   /* search1267_2 -> 3614 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3600, 3617, 3614 },
      -1,
   } },

   /* replace1267_0 -> 16 in the cache */
   /* replace1267 -> 3619 in the cache */

   /* ('bcsel', ('ieq', 'a@32', 0), ('ufind_msb', 'a'), ('iadd', 31, ('ineg', ('ufind_msb', 'a')))) => ('ufind_msb_rev', 'a') */
   /* search1268_0_0 -> 16 in the cache */
   /* search1268_0_1 -> 122 in the cache */
   /* search1268_0 -> 3629 in the cache */
   /* search1268_1_0 -> 16 in the cache */
   /* search1268_1 -> 3614 in the cache */
   /* search1268_2_0 -> 19 in the cache */
   /* search1268_2_1_0_0 -> 16 in the cache */
   /* search1268_2_1_0 -> 3614 in the cache */
   /* search1268_2_1 -> 3616 in the cache */
   /* search1268_2 -> 3617 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3629, 3614, 3617 },
      -1,
   } },

   /* replace1268_0 -> 16 in the cache */
   /* replace1268 -> 3619 in the cache */

   /* ('bcsel', ('ine', ('ifind_msb_rev', 'a@32'), -1), ('iadd', 31, ('ineg', ('ifind_msb_rev', 'a'))), -1) => ('ifind_msb', 'a') */
   /* search1269_0_0_0 -> 16 in the cache */
   /* search1269_0_0 -> 3613 in the cache */
   /* search1269_0_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 3613, 124 },
      -1,
   } },
   /* search1269_1_0 -> 19 in the cache */
   /* search1269_1_1_0_0 -> 16 in the cache */
   /* search1269_1_1_0 -> 3613 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 3613 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 19, 3634 },
      -1,
   } },
   /* search1269_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3633, 3635, 124 },
      -1,
   } },

   /* replace1269_0 -> 16 in the cache */
   /* replace1269 -> 3587 in the cache */

   /* ('bcsel', ('ine', ('ufind_msb_rev', 'a@32'), -1), ('iadd', 31, ('ineg', ('ufind_msb_rev', 'a'))), -1) => ('ufind_msb', 'a') */
   /* search1270_0_0_0 -> 16 in the cache */
   /* search1270_0_0 -> 3619 in the cache */
   /* search1270_0_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 3619, 124 },
      -1,
   } },
   /* search1270_1_0 -> 19 in the cache */
   /* search1270_1_1_0_0 -> 16 in the cache */
   /* search1270_1_1_0 -> 3619 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 3619 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 19, 3638 },
      -1,
   } },
   /* search1270_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3637, 3639, 124 },
      -1,
   } },

   /* replace1270_0 -> 16 in the cache */
   /* replace1270 -> 3614 in the cache */

   /* ('bcsel', ('ieq', ('ifind_msb_rev', 'a@32'), -1), -1, ('iadd', 31, ('ineg', ('ifind_msb_rev', 'a')))) => ('ifind_msb', 'a') */
   /* search1271_0_0_0 -> 16 in the cache */
   /* search1271_0_0 -> 3613 in the cache */
   /* search1271_0_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 3613, 124 },
      -1,
   } },
   /* search1271_1 -> 124 in the cache */
   /* search1271_2_0 -> 19 in the cache */
   /* search1271_2_1_0_0 -> 16 in the cache */
   /* search1271_2_1_0 -> 3613 in the cache */
   /* search1271_2_1 -> 3634 in the cache */
   /* search1271_2 -> 3635 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3641, 124, 3635 },
      -1,
   } },

   /* replace1271_0 -> 16 in the cache */
   /* replace1271 -> 3587 in the cache */

   /* ('bcsel', ('ieq', ('ufind_msb_rev', 'a@32'), -1), -1, ('iadd', 31, ('ineg', ('ufind_msb_rev', 'a')))) => ('ufind_msb', 'a') */
   /* search1272_0_0_0 -> 16 in the cache */
   /* search1272_0_0 -> 3619 in the cache */
   /* search1272_0_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 3619, 124 },
      -1,
   } },
   /* search1272_1 -> 124 in the cache */
   /* search1272_2_0 -> 19 in the cache */
   /* search1272_2_1_0_0 -> 16 in the cache */
   /* search1272_2_1_0 -> 3619 in the cache */
   /* search1272_2_1 -> 3638 in the cache */
   /* search1272_2 -> 3639 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3643, 124, 3639 },
      -1,
   } },

   /* replace1272_0 -> 16 in the cache */
   /* replace1272 -> 3614 in the cache */

   /* ('bcsel', ('ine', ('ifind_msb_rev', 'a@32'), -1), ('iadd', 31, ('ineg', ('ifind_msb_rev', 'a'))), ('ifind_msb_rev', 'a')) => ('ifind_msb', 'a') */
   /* search1273_0_0_0 -> 16 in the cache */
   /* search1273_0_0 -> 3613 in the cache */
   /* search1273_0_1 -> 124 in the cache */
   /* search1273_0 -> 3633 in the cache */
   /* search1273_1_0 -> 19 in the cache */
   /* search1273_1_1_0_0 -> 16 in the cache */
   /* search1273_1_1_0 -> 3613 in the cache */
   /* search1273_1_1 -> 3634 in the cache */
   /* search1273_1 -> 3635 in the cache */
   /* search1273_2_0 -> 16 in the cache */
   /* search1273_2 -> 3613 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3633, 3635, 3613 },
      -1,
   } },

   /* replace1273_0 -> 16 in the cache */
   /* replace1273 -> 3587 in the cache */

   /* ('bcsel', ('ine', ('ufind_msb_rev', 'a@32'), -1), ('iadd', 31, ('ineg', ('ufind_msb_rev', 'a'))), ('ufind_msb_rev', 'a')) => ('ufind_msb', 'a') */
   /* search1274_0_0_0 -> 16 in the cache */
   /* search1274_0_0 -> 3619 in the cache */
   /* search1274_0_1 -> 124 in the cache */
   /* search1274_0 -> 3637 in the cache */
   /* search1274_1_0 -> 19 in the cache */
   /* search1274_1_1_0_0 -> 16 in the cache */
   /* search1274_1_1_0 -> 3619 in the cache */
   /* search1274_1_1 -> 3638 in the cache */
   /* search1274_1 -> 3639 in the cache */
   /* search1274_2_0 -> 16 in the cache */
   /* search1274_2 -> 3619 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3637, 3639, 3619 },
      -1,
   } },

   /* replace1274_0 -> 16 in the cache */
   /* replace1274 -> 3614 in the cache */

   /* ('bcsel', ('ieq', ('ifind_msb_rev', 'a@32'), -1), ('ifind_msb_rev', 'a'), ('iadd', 31, ('ineg', ('ifind_msb_rev', 'a')))) => ('ifind_msb', 'a') */
   /* search1275_0_0_0 -> 16 in the cache */
   /* search1275_0_0 -> 3613 in the cache */
   /* search1275_0_1 -> 124 in the cache */
   /* search1275_0 -> 3641 in the cache */
   /* search1275_1_0 -> 16 in the cache */
   /* search1275_1 -> 3613 in the cache */
   /* search1275_2_0 -> 19 in the cache */
   /* search1275_2_1_0_0 -> 16 in the cache */
   /* search1275_2_1_0 -> 3613 in the cache */
   /* search1275_2_1 -> 3634 in the cache */
   /* search1275_2 -> 3635 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3641, 3613, 3635 },
      -1,
   } },

   /* replace1275_0 -> 16 in the cache */
   /* replace1275 -> 3587 in the cache */

   /* ('bcsel', ('ieq', ('ufind_msb_rev', 'a@32'), -1), ('ufind_msb_rev', 'a'), ('iadd', 31, ('ineg', ('ufind_msb_rev', 'a')))) => ('ufind_msb', 'a') */
   /* search1276_0_0_0 -> 16 in the cache */
   /* search1276_0_0 -> 3619 in the cache */
   /* search1276_0_1 -> 124 in the cache */
   /* search1276_0 -> 3643 in the cache */
   /* search1276_1_0 -> 16 in the cache */
   /* search1276_1 -> 3619 in the cache */
   /* search1276_2_0 -> 19 in the cache */
   /* search1276_2_1_0_0 -> 16 in the cache */
   /* search1276_2_1_0 -> 3619 in the cache */
   /* search1276_2_1 -> 3638 in the cache */
   /* search1276_2 -> 3639 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3643, 3619, 3639 },
      -1,
   } },

   /* replace1276_0 -> 16 in the cache */
   /* replace1276 -> 3614 in the cache */

   /* ('bcsel', ('ine', 'a@32', 0), ('iadd', 31, ('ineg', ('ufind_msb_rev', 'a'))), -1) => ('ufind_msb', 'a') */
   /* search1277_0_0 -> 16 in the cache */
   /* search1277_0_1 -> 122 in the cache */
   /* search1277_0 -> 3600 in the cache */
   /* search1277_1_0 -> 19 in the cache */
   /* search1277_1_1_0_0 -> 16 in the cache */
   /* search1277_1_1_0 -> 3619 in the cache */
   /* search1277_1_1 -> 3638 in the cache */
   /* search1277_1 -> 3639 in the cache */
   /* search1277_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3600, 3639, 124 },
      -1,
   } },

   /* replace1277_0 -> 16 in the cache */
   /* replace1277 -> 3614 in the cache */

   /* ('bcsel', ('ieq', 'a@32', 0), -1, ('iadd', 31, ('ineg', ('ufind_msb_rev', 'a')))) => ('ufind_msb', 'a') */
   /* search1278_0_0 -> 16 in the cache */
   /* search1278_0_1 -> 122 in the cache */
   /* search1278_0 -> 3629 in the cache */
   /* search1278_1 -> 124 in the cache */
   /* search1278_2_0 -> 19 in the cache */
   /* search1278_2_1_0_0 -> 16 in the cache */
   /* search1278_2_1_0 -> 3619 in the cache */
   /* search1278_2_1 -> 3638 in the cache */
   /* search1278_2 -> 3639 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3629, 124, 3639 },
      -1,
   } },

   /* replace1278_0 -> 16 in the cache */
   /* replace1278 -> 3614 in the cache */

   /* ('bcsel', ('ine', 'a@32', 0), ('iadd', 31, ('ineg', ('ufind_msb_rev', 'a'))), ('ufind_msb_rev', 'a')) => ('ufind_msb', 'a') */
   /* search1279_0_0 -> 16 in the cache */
   /* search1279_0_1 -> 122 in the cache */
   /* search1279_0 -> 3600 in the cache */
   /* search1279_1_0 -> 19 in the cache */
   /* search1279_1_1_0_0 -> 16 in the cache */
   /* search1279_1_1_0 -> 3619 in the cache */
   /* search1279_1_1 -> 3638 in the cache */
   /* search1279_1 -> 3639 in the cache */
   /* search1279_2_0 -> 16 in the cache */
   /* search1279_2 -> 3619 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3600, 3639, 3619 },
      -1,
   } },

   /* replace1279_0 -> 16 in the cache */
   /* replace1279 -> 3614 in the cache */

   /* ('bcsel', ('ieq', 'a@32', 0), ('ufind_msb_rev', 'a'), ('iadd', 31, ('ineg', ('ufind_msb_rev', 'a')))) => ('ufind_msb', 'a') */
   /* search1280_0_0 -> 16 in the cache */
   /* search1280_0_1 -> 122 in the cache */
   /* search1280_0 -> 3629 in the cache */
   /* search1280_1_0 -> 16 in the cache */
   /* search1280_1 -> 3619 in the cache */
   /* search1280_2_0 -> 19 in the cache */
   /* search1280_2_1_0_0 -> 16 in the cache */
   /* search1280_2_1_0 -> 3619 in the cache */
   /* search1280_2_1 -> 3638 in the cache */
   /* search1280_2 -> 3639 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3629, 3619, 3639 },
      -1,
   } },

   /* replace1280_0 -> 16 in the cache */
   /* replace1280 -> 3614 in the cache */

   /* ('find_lsb', ('bitfield_reverse', 'a')) => ('ufind_msb_rev', 'a') */
   /* search1281_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bitfield_reverse,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_find_lsb,
      -1, 0,
      { 3653 },
      -1,
   } },

   /* replace1281_0 -> 16 in the cache */
   /* replace1281 -> 3619 in the cache */

   /* ('ufind_msb_rev', ('bitfield_reverse', 'a')) => ('find_lsb', 'a') */
   /* search1282_0_0 -> 16 in the cache */
   /* search1282_0 -> 3653 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ufind_msb_rev,
      -1, 0,
      { 3653 },
      -1,
   } },

   /* replace1282_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_find_lsb,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('~fmul', ('bcsel(is_used_once)', 'c', -1.0, 1.0), 'b') => ('bcsel', 'c', ('fneg', 'b'), 'b') */
   /* search1283_0_0 -> 521 in the cache */
   /* search1283_0_1 -> 275 in the cache */
   /* search1283_0_2 -> 270 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 275, 270 },
      0,
   } },
   /* search1283_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 3657, 3 },
      -1,
   } },

   /* replace1283_0 -> 521 in the cache */
   /* replace1283_1_0 -> 3 in the cache */
   /* replace1283_1 -> 1087 in the cache */
   /* replace1283_2 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 1087, 3 },
      -1,
   } },

   /* ('~fmul', ('bcsel(is_used_once)', 'c', 1.0, -1.0), 'b') => ('bcsel', 'c', 'b', ('fneg', 'b')) */
   /* search1284_0_0 -> 521 in the cache */
   /* search1284_0_1 -> 270 in the cache */
   /* search1284_0_2 -> 275 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 270, 275 },
      0,
   } },
   /* search1284_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 3660, 3 },
      -1,
   } },

   /* replace1284_0 -> 521 in the cache */
   /* replace1284_1 -> 3 in the cache */
   /* replace1284_2_0 -> 3 in the cache */
   /* replace1284_2 -> 1087 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 3, 1087 },
      -1,
   } },

   /* ('~fmulz', ('bcsel(is_used_once)', 'c', -1.0, 1.0), 'b') => ('bcsel', 'c', ('fneg', 'b'), 'b') */
   /* search1285_0_0 -> 521 in the cache */
   /* search1285_0_1 -> 278 in the cache */
   /* search1285_0_2 -> 243 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 278, 243 },
      0,
   } },
   /* search1285_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 3663, 36 },
      -1,
   } },

   /* replace1285_0 -> 521 in the cache */
   /* replace1285_1_0 -> 36 in the cache */
   /* replace1285_1 -> 600 in the cache */
   /* replace1285_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 600, 36 },
      -1,
   } },

   /* ('~fmulz', ('bcsel(is_used_once)', 'c', 1.0, -1.0), 'b') => ('bcsel', 'c', 'b', ('fneg', 'b')) */
   /* search1286_0_0 -> 521 in the cache */
   /* search1286_0_1 -> 243 in the cache */
   /* search1286_0_2 -> 278 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 243, 278 },
      0,
   } },
   /* search1286_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 3666, 36 },
      -1,
   } },

   /* replace1286_0 -> 521 in the cache */
   /* replace1286_1 -> 36 in the cache */
   /* replace1286_2_0 -> 36 in the cache */
   /* replace1286_2 -> 600 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 36, 600 },
      -1,
   } },

   /* ('fabs', ('bcsel(is_used_once)', 'b', ('fneg', 'a'), 'a')) => ('fabs', 'a') */
   { .variable = {
      { nir_search_value_variable, 1 },
      0, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1287_0_1_0 -> 1114 in the cache */
   /* search1287_0_1 -> 1577 in the cache */
   /* search1287_0_2 -> 1114 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3669, 1577, 1114 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 3670 },
      -1,
   } },

   /* replace1287_0 -> 1114 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 1114 },
      -1,
   } },

   /* ('fabs', ('bcsel(is_used_once)', 'b', 'a', ('fneg', 'a'))) => ('fabs', 'a') */
   /* search1288_0_0 -> 3669 in the cache */
   /* search1288_0_1 -> 1114 in the cache */
   /* search1288_0_2_0 -> 1114 in the cache */
   /* search1288_0_2 -> 1577 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3669, 1114, 1577 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 3673 },
      -1,
   } },

   /* replace1288_0 -> 1114 in the cache */
   /* replace1288 -> 3672 in the cache */

   /* ('~bcsel', ('flt', 'a', 0.0), ('fneg', 'a'), 'a') => ('fabs', 'a') */
   /* search1289_0_0 -> 32 in the cache */
   /* search1289_0_1 -> 111 in the cache */
   /* search1289_0 -> 1172 in the cache */
   /* search1289_1_0 -> 32 in the cache */
   /* search1289_1 -> 98 in the cache */
   /* search1289_2 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1172, 98, 32 },
      -1,
   } },

   /* replace1289_0 -> 32 in the cache */
   /* replace1289 -> 103 in the cache */

   /* ('bcsel', 'a', ('bcsel', 'b', 'c', 'd'), 'd') => ('bcsel', ('iand', 'a', 'b'), 'c', 'd') */
   /* search1290_0 -> 760 in the cache */
   /* search1290_1_0 -> 1134 in the cache */
   /* search1290_1_1 -> 152 in the cache */
   /* search1290_1_2 -> 153 in the cache */
   /* search1290_1 -> 1322 in the cache */
   /* search1290_2 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 1322, 153 },
      -1,
   } },

   /* replace1290_0_0 -> 760 in the cache */
   /* replace1290_0_1 -> 1134 in the cache */
   /* replace1290_0 -> 1146 in the cache */
   /* replace1290_1 -> 152 in the cache */
   /* replace1290_2 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 1146, 152, 153 },
      -1,
   } },

   /* ('bcsel', 'a', 'b', ('bcsel', 'c', 'b', 'd')) => ('bcsel', ('ior', 'a', 'c'), 'b', 'd') */
   /* search1291_0 -> 760 in the cache */
   /* search1291_1 -> 150 in the cache */
   /* search1291_2_0 -> 464 in the cache */
   /* search1291_2_1 -> 150 in the cache */
   /* search1291_2_2 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 464, 150, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 150, 3678 },
      -1,
   } },

   /* replace1291_0_0 -> 760 in the cache */
   /* replace1291_0_1 -> 464 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 760, 464 },
      -1,
   } },
   /* replace1291_1 -> 150 in the cache */
   /* replace1291_2 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 3680, 150, 153 },
      -1,
   } },

   /* ('fmod', 'a', 'b') => ('fsub', 'a', ('fmul', 'b', ('ffloor', ('fdiv', 'a', 'b')))) */
   /* search1292_0 -> 0 in the cache */
   /* search1292_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmod,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* replace1292_0 -> 0 in the cache */
   /* replace1292_1_0 -> 3 in the cache */
   /* replace1292_1_1_0_0 -> 0 in the cache */
   /* replace1292_1_1_0_1 -> 3 in the cache */
   /* replace1292_1_1_0 -> 2880 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 0,
      { 2880 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 3, 3683 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 1,
      { 0, 3684 },
      -1,
   } },

   /* ('frem', 'a', 'b') => ('fsub', 'a', ('fmul', 'b', ('ftrunc', ('fdiv', 'a', 'b')))) */
   /* search1293_0 -> 0 in the cache */
   /* search1293_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_frem,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* replace1293_0 -> 0 in the cache */
   /* replace1293_1_0 -> 3 in the cache */
   /* replace1293_1_1_0_0 -> 0 in the cache */
   /* replace1293_1_1_0_1 -> 3 in the cache */
   /* replace1293_1_1_0 -> 2880 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ftrunc,
      -1, 0,
      { 2880 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 3, 3687 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 1,
      { 0, 3688 },
      -1,
   } },

   /* ('uadd_carry', 'a', 'b') => ('b2i', ('ult', ('iadd', 'a', 'b'), 'a')) */
   /* search1294_0 -> 0 in the cache */
   /* search1294_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_uadd_carry,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace1294_0_0_0 -> 0 in the cache */
   /* replace1294_0_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 0, 3 },
      -1,
   } },
   /* replace1294_0_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 3691, 0 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 1,
      { 3692 },
      -1,
   } },

   /* ('usub_borrow', 'a', 'b') => ('b2i', ('ult', 'a', 'b')) */
   /* search1295_0 -> 0 in the cache */
   /* search1295_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_usub_borrow,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* replace1295_0_0 -> 0 in the cache */
   /* replace1295_0_1 -> 3 in the cache */
   /* replace1295_0 -> 1067 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 0,
      { 1067 },
      -1,
   } },

   /* ('bitfield_insert', 'base', 'insert', 'offset', 'bits') => ('bcsel', ('ult', 31, 'bits'), 'insert', ('bfi', ('bfm', 'bits', 'offset'), 'insert', 'base')) */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* base */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* insert */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* offset */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      3, /* bits */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bitfield_insert,
      -1, 0,
      { 3696, 3697, 3698, 3699 },
      -1,
   } },

   /* replace1296_0_0 -> 19 in the cache */
   /* replace1296_0_1 -> 3699 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 19, 3699 },
      -1,
   } },
   /* replace1296_1 -> 3697 in the cache */
   /* replace1296_2_0_0 -> 3699 in the cache */
   /* replace1296_2_0_1 -> 3698 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfm,
      -1, 0,
      { 3699, 3698 },
      -1,
   } },
   /* replace1296_2_1 -> 3697 in the cache */
   /* replace1296_2_2 -> 3696 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 3702, 3697, 3696 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3701, 3697, 3703 },
      -1,
   } },

   /* ('ihadd', 'a', 'b') => ('iadd', ('iand', 'a', 'b'), ('ishr', ('ixor', 'a', 'b'), 1)) */
   /* search1297_0 -> 0 in the cache */
   /* search1297_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ihadd,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace1297_0_0 -> 0 in the cache */
   /* replace1297_0_1 -> 3 in the cache */
   /* replace1297_0 -> 201 in the cache */
   /* replace1297_1_0_0 -> 0 in the cache */
   /* replace1297_1_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ixor,
      2, 1,
      { 0, 3 },
      -1,
   } },
   /* replace1297_1_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 3706, 396 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 3,
      { 201, 3707 },
      -1,
   } },

   /* ('uhadd', 'a', 'b') => ('iadd', ('iand', 'a', 'b'), ('ushr', ('ixor', 'a', 'b'), 1)) */
   /* search1298_0 -> 0 in the cache */
   /* search1298_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_uhadd,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace1298_0_0 -> 0 in the cache */
   /* replace1298_0_1 -> 3 in the cache */
   /* replace1298_0 -> 201 in the cache */
   /* replace1298_1_0_0 -> 0 in the cache */
   /* replace1298_1_0_1 -> 3 in the cache */
   /* replace1298_1_0 -> 3706 in the cache */
   /* replace1298_1_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 3706, 396 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 3,
      { 201, 3710 },
      -1,
   } },

   /* ('irhadd', 'a', 'b') => ('isub', ('ior', 'a', 'b'), ('ishr', ('ixor', 'a', 'b'), 1)) */
   /* search1299_0 -> 0 in the cache */
   /* search1299_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_irhadd,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace1299_0_0 -> 0 in the cache */
   /* replace1299_0_1 -> 3 in the cache */
   /* replace1299_0 -> 2338 in the cache */
   /* replace1299_1_0_0 -> 0 in the cache */
   /* replace1299_1_0_1 -> 3 in the cache */
   /* replace1299_1_0 -> 2711 in the cache */
   /* replace1299_1_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 2711, 396 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 2,
      { 2338, 3713 },
      -1,
   } },

   /* ('urhadd', 'a', 'b') => ('isub', ('ior', 'a', 'b'), ('ushr', ('ixor', 'a', 'b'), 1)) */
   /* search1300_0 -> 0 in the cache */
   /* search1300_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_urhadd,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace1300_0_0 -> 0 in the cache */
   /* replace1300_0_1 -> 3 in the cache */
   /* replace1300_0 -> 2338 in the cache */
   /* replace1300_1_0_0 -> 0 in the cache */
   /* replace1300_1_0_1 -> 3 in the cache */
   /* replace1300_1_0 -> 2711 in the cache */
   /* replace1300_1_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 2711, 396 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 2,
      { 2338, 3716 },
      -1,
   } },

   /* ('ihadd@64', 'a', 'b') => ('iadd', ('iand', 'a', 'b'), ('ishr', ('ixor', 'a', 'b'), 1)) */
   /* search1301_0 -> 21 in the cache */
   /* search1301_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ihadd,
      0, 1,
      { 21, 41 },
      -1,
   } },

   /* replace1301_0_0 -> 21 in the cache */
   /* replace1301_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 21, 41 },
      -1,
   } },
   /* replace1301_1_0_0 -> 21 in the cache */
   /* replace1301_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ixor,
      2, 1,
      { 21, 41 },
      -1,
   } },
   /* replace1301_1_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 3720, 396 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 3,
      { 3719, 3721 },
      -1,
   } },

   /* ('uhadd@64', 'a', 'b') => ('iadd', ('iand', 'a', 'b'), ('ushr', ('ixor', 'a', 'b'), 1)) */
   /* search1302_0 -> 21 in the cache */
   /* search1302_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_uhadd,
      0, 1,
      { 21, 41 },
      -1,
   } },

   /* replace1302_0_0 -> 21 in the cache */
   /* replace1302_0_1 -> 41 in the cache */
   /* replace1302_0 -> 3719 in the cache */
   /* replace1302_1_0_0 -> 21 in the cache */
   /* replace1302_1_0_1 -> 41 in the cache */
   /* replace1302_1_0 -> 3720 in the cache */
   /* replace1302_1_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 3720, 396 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 3,
      { 3719, 3724 },
      -1,
   } },

   /* ('irhadd@64', 'a', 'b') => ('isub', ('ior', 'a', 'b'), ('ishr', ('ixor', 'a', 'b'), 1)) */
   /* search1303_0 -> 21 in the cache */
   /* search1303_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_irhadd,
      0, 1,
      { 21, 41 },
      -1,
   } },

   /* replace1303_0_0 -> 21 in the cache */
   /* replace1303_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 21, 41 },
      -1,
   } },
   /* replace1303_1_0_0 -> 21 in the cache */
   /* replace1303_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ixor,
      1, 1,
      { 21, 41 },
      -1,
   } },
   /* replace1303_1_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 3728, 396 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 2,
      { 3727, 3729 },
      -1,
   } },

   /* ('urhadd@64', 'a', 'b') => ('isub', ('ior', 'a', 'b'), ('ushr', ('ixor', 'a', 'b'), 1)) */
   /* search1304_0 -> 21 in the cache */
   /* search1304_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_urhadd,
      0, 1,
      { 21, 41 },
      -1,
   } },

   /* replace1304_0_0 -> 21 in the cache */
   /* replace1304_0_1 -> 41 in the cache */
   /* replace1304_0 -> 3727 in the cache */
   /* replace1304_1_0_0 -> 21 in the cache */
   /* replace1304_1_0_1 -> 41 in the cache */
   /* replace1304_1_0 -> 3728 in the cache */
   /* replace1304_1_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 3728, 396 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 2,
      { 3727, 3732 },
      -1,
   } },

   /* ('imul_32x16', 'a', 'b') => ('imul', 'a', ('extract_i16', 'b', 0)) */
   /* search1305_0 -> 16 in the cache */
   /* search1305_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul_32x16,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* replace1305_0 -> 16 in the cache */
   /* replace1305_1_0 -> 36 in the cache */
   /* replace1305_1_1 -> 122 in the cache */
   /* replace1305_1 -> 437 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 16, 437 },
      -1,
   } },

   /* ('umul_32x16', 'a', 'b') => ('imul', 'a', ('extract_u16', 'b', 0)) */
   /* search1306_0 -> 16 in the cache */
   /* search1306_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umul_32x16,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* replace1306_0 -> 16 in the cache */
   /* replace1306_1_0 -> 36 in the cache */
   /* replace1306_1_1 -> 122 in the cache */
   /* replace1306_1 -> 445 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 16, 445 },
      -1,
   } },

   /* ('uadd_sat@64', 'a', 'b') => ('bcsel', ('ult', ('iadd', 'a', 'b'), 'a'), -1, ('iadd', 'a', 'b')) */
   /* search1307_0 -> 21 in the cache */
   /* search1307_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_uadd_sat,
      0, 1,
      { 21, 41 },
      -1,
   } },

   /* replace1307_0_0_0 -> 21 in the cache */
   /* replace1307_0_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 21, 41 },
      -1,
   } },
   /* replace1307_0_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 3739, 21 },
      -1,
   } },
   /* replace1307_1 -> 1499 in the cache */
   /* replace1307_2_0 -> 21 in the cache */
   /* replace1307_2_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 21, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3740, 1499, 3741 },
      -1,
   } },

   /* ('uadd_sat', 'a', 'b') => ('bcsel', ('ult', ('iadd', 'a', 'b'), 'a'), -1, ('iadd', 'a', 'b')) */
   /* search1308_0 -> 0 in the cache */
   /* search1308_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_uadd_sat,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace1308_0_0_0 -> 0 in the cache */
   /* replace1308_0_0_1 -> 3 in the cache */
   /* replace1308_0_0 -> 3691 in the cache */
   /* replace1308_0_1 -> 0 in the cache */
   /* replace1308_0 -> 3692 in the cache */
   { .constant = {
      { nir_search_value_constant, -2 },
      nir_type_int, { -0x1 /* -1 */ },
   } },
   /* replace1308_2_0 -> 0 in the cache */
   /* replace1308_2_1 -> 3 in the cache */
   /* replace1308_2 -> 208 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3692, 3744, 208 },
      -1,
   } },

   /* ('usub_sat', 'a', 'b') => ('bcsel', ('ult', 'a', 'b'), 0, ('isub', 'a', 'b')) */
   /* search1309_0 -> 0 in the cache */
   /* search1309_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_usub_sat,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* replace1309_0_0 -> 0 in the cache */
   /* replace1309_0_1 -> 3 in the cache */
   /* replace1309_0 -> 1067 in the cache */
   /* replace1309_1 -> 86 in the cache */
   /* replace1309_2_0 -> 0 in the cache */
   /* replace1309_2_1 -> 3 in the cache */
   /* replace1309_2 -> 3461 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1067, 86, 3461 },
      -1,
   } },

   /* ('usub_sat@64', 'a', 'b') => ('bcsel', ('ult', 'a', 'b'), 0, ('isub', 'a', 'b')) */
   /* search1310_0 -> 21 in the cache */
   /* search1310_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_usub_sat,
      -1, 0,
      { 21, 41 },
      -1,
   } },

   /* replace1310_0_0 -> 21 in the cache */
   /* replace1310_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 21, 41 },
      -1,
   } },
   /* replace1310_1 -> 897 in the cache */
   /* replace1310_2_0 -> 21 in the cache */
   /* replace1310_2_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 21, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3749, 897, 3750 },
      -1,
   } },

   /* ('iadd_sat@64', 'a', 'b') => ('bcsel', ('iand', ('iand', ('ilt', 'a', 0), ('ilt', 'b', 0)), ('ige', ('iadd', 'a', 'b'), 0)), 9223372036854775808, ('bcsel', ('ior', ('ior', ('ilt', 'a', 0), ('ilt', 'b', 0)), ('ige', ('iadd', 'a', 'b'), 0)), ('iadd', 'a', 'b'), 9223372036854775807)) */
   /* search1311_0 -> 21 in the cache */
   /* search1311_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd_sat,
      0, 1,
      { 21, 41 },
      -1,
   } },

   /* replace1311_0_0_0_0 -> 21 in the cache */
   /* replace1311_0_0_0_1 -> 897 in the cache */
   /* replace1311_0_0_0 -> 1489 in the cache */
   /* replace1311_0_0_1_0 -> 41 in the cache */
   /* replace1311_0_0_1_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 41, 897 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 1489, 3753 },
      -1,
   } },
   /* replace1311_0_1_0_0 -> 21 in the cache */
   /* replace1311_0_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 1,
      { 21, 41 },
      -1,
   } },
   /* replace1311_0_1_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 3755, 897 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 3754, 3756 },
      -1,
   } },
   /* replace1311_1 -> 22 in the cache */
   /* replace1311_2_0_0_0_0 -> 21 in the cache */
   /* replace1311_2_0_0_0_1 -> 897 in the cache */
   /* replace1311_2_0_0_0 -> 1489 in the cache */
   /* replace1311_2_0_0_1_0 -> 41 in the cache */
   /* replace1311_2_0_0_1_1 -> 897 in the cache */
   /* replace1311_2_0_0_1 -> 3753 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      4, 1,
      { 1489, 3753 },
      -1,
   } },
   /* replace1311_2_0_1_0_0 -> 21 in the cache */
   /* replace1311_2_0_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      5, 1,
      { 21, 41 },
      -1,
   } },
   /* replace1311_2_0_1_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 3759, 897 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      3, 3,
      { 3758, 3760 },
      -1,
   } },
   /* replace1311_2_1_0 -> 21 in the cache */
   /* replace1311_2_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      6, 1,
      { 21, 41 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x7fffffffffffffff /* 9223372036854775807 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 3761, 3762, 3763 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 7,
      { 3757, 22, 3764 },
      -1,
   } },

   /* ('isub_sat@64', 'a', 'b') => ('bcsel', ('iand', ('iand', ('ilt', 'a', 0), ('ige', 'b', 0)), ('ige', ('isub', 'a', 'b'), 0)), 9223372036854775808, ('bcsel', ('ior', ('ior', ('ilt', 'a', 0), ('ige', 'b', 0)), ('ige', ('isub', 'a', 'b'), 0)), ('isub', 'a', 'b'), 9223372036854775807)) */
   /* search1312_0 -> 21 in the cache */
   /* search1312_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_isub_sat,
      -1, 0,
      { 21, 41 },
      -1,
   } },

   /* replace1312_0_0_0_0 -> 21 in the cache */
   /* replace1312_0_0_0_1 -> 897 in the cache */
   /* replace1312_0_0_0 -> 1489 in the cache */
   /* replace1312_0_0_1_0 -> 41 in the cache */
   /* replace1312_0_0_1_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 41, 897 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 1489, 3767 },
      -1,
   } },
   /* replace1312_0_1_0_0 -> 21 in the cache */
   /* replace1312_0_1_0_1 -> 41 in the cache */
   /* replace1312_0_1_0 -> 3750 in the cache */
   /* replace1312_0_1_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3750, 897 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 3768, 3769 },
      -1,
   } },
   /* replace1312_1 -> 22 in the cache */
   /* replace1312_2_0_0_0_0 -> 21 in the cache */
   /* replace1312_2_0_0_0_1 -> 897 in the cache */
   /* replace1312_2_0_0_0 -> 1489 in the cache */
   /* replace1312_2_0_0_1_0 -> 41 in the cache */
   /* replace1312_2_0_0_1_1 -> 897 in the cache */
   /* replace1312_2_0_0_1 -> 3767 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      3, 1,
      { 1489, 3767 },
      -1,
   } },
   /* replace1312_2_0_1_0_0 -> 21 in the cache */
   /* replace1312_2_0_1_0_1 -> 41 in the cache */
   /* replace1312_2_0_1_0 -> 3750 in the cache */
   /* replace1312_2_0_1_1 -> 897 in the cache */
   /* replace1312_2_0_1 -> 3769 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      2, 2,
      { 3771, 3769 },
      -1,
   } },
   /* replace1312_2_1_0 -> 21 in the cache */
   /* replace1312_2_1_1 -> 41 in the cache */
   /* replace1312_2_1 -> 3750 in the cache */
   /* replace1312_2_2 -> 3763 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3772, 3750, 3763 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 3770, 22, 3773 },
      -1,
   } },

   /* ('ilt', ('imax(is_used_once)', 'a@64', 'b@64'), 0) => ('ilt', ('imax', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b')), 0) */
   /* search1313_0_0 -> 21 in the cache */
   /* search1313_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 21, 41 },
      0,
   } },
   /* search1313_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 3775, 897 },
      -1,
   } },

   /* replace1313_0_0_0 -> 21 in the cache */
   /* replace1313_0_0 -> 2726 in the cache */
   /* replace1313_0_1_0 -> 41 in the cache */
   /* replace1313_0_1 -> 2727 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 2726, 2727 },
      -1,
   } },
   /* replace1313_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 3777, 122 },
      -1,
   } },

   /* ('ilt', ('imin(is_used_once)', 'a@64', 'b@64'), 0) => ('ilt', ('imin', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b')), 0) */
   /* search1314_0_0 -> 21 in the cache */
   /* search1314_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 21, 41 },
      0,
   } },
   /* search1314_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 3779, 897 },
      -1,
   } },

   /* replace1314_0_0_0 -> 21 in the cache */
   /* replace1314_0_0 -> 2726 in the cache */
   /* replace1314_0_1_0 -> 41 in the cache */
   /* replace1314_0_1 -> 2727 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 2726, 2727 },
      -1,
   } },
   /* replace1314_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 3781, 122 },
      -1,
   } },

   /* ('ige', ('imax(is_used_once)', 'a@64', 'b@64'), 0) => ('ige', ('imax', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b')), 0) */
   /* search1315_0_0 -> 21 in the cache */
   /* search1315_0_1 -> 41 in the cache */
   /* search1315_0 -> 3775 in the cache */
   /* search1315_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 3775, 897 },
      -1,
   } },

   /* replace1315_0_0_0 -> 21 in the cache */
   /* replace1315_0_0 -> 2726 in the cache */
   /* replace1315_0_1_0 -> 41 in the cache */
   /* replace1315_0_1 -> 2727 in the cache */
   /* replace1315_0 -> 3777 in the cache */
   /* replace1315_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 3777, 122 },
      -1,
   } },

   /* ('ige', ('imin(is_used_once)', 'a@64', 'b@64'), 0) => ('ige', ('imin', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b')), 0) */
   /* search1316_0_0 -> 21 in the cache */
   /* search1316_0_1 -> 41 in the cache */
   /* search1316_0 -> 3779 in the cache */
   /* search1316_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 3779, 897 },
      -1,
   } },

   /* replace1316_0_0_0 -> 21 in the cache */
   /* replace1316_0_0 -> 2726 in the cache */
   /* replace1316_0_1_0 -> 41 in the cache */
   /* replace1316_0_1 -> 2727 in the cache */
   /* replace1316_0 -> 3781 in the cache */
   /* replace1316_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 3781, 122 },
      -1,
   } },

   /* ('ilt', 'a@64', 0) => ('ilt', ('unpack_64_2x32_split_y', 'a'), 0) */
   /* search1317_0 -> 21 in the cache */
   /* search1317_1 -> 897 in the cache */
   /* search1317 -> 1489 in the cache */

   /* replace1317_0_0 -> 21 in the cache */
   /* replace1317_0 -> 2726 in the cache */
   /* replace1317_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 2726, 122 },
      -1,
   } },

   /* ('ige', 'a@64', 0) => ('ige', ('unpack_64_2x32_split_y', 'a'), 0) */
   /* search1318_0 -> 21 in the cache */
   /* search1318_1 -> 897 in the cache */
   /* search1318 -> 1486 in the cache */

   /* replace1318_0_0 -> 21 in the cache */
   /* replace1318_0 -> 2726 in the cache */
   /* replace1318_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 2726, 122 },
      -1,
   } },

   /* ('ine', 'a@64', 0) => ('ine', ('ior', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_y', 'a')), 0) */
   /* search1319_0 -> 21 in the cache */
   /* search1319_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 21, 897 },
      -1,
   } },

   /* replace1319_0_0_0 -> 21 in the cache */
   /* replace1319_0_0 -> 2732 in the cache */
   /* replace1319_0_1_0 -> 21 in the cache */
   /* replace1319_0_1 -> 2726 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 2732, 2726 },
      -1,
   } },
   /* replace1319_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 3790, 122 },
      -1,
   } },

   /* ('ieq', 'a@64', 0) => ('ieq', ('ior', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_y', 'a')), 0) */
   /* search1320_0 -> 21 in the cache */
   /* search1320_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 21, 897 },
      -1,
   } },

   /* replace1320_0_0_0 -> 21 in the cache */
   /* replace1320_0_0 -> 2732 in the cache */
   /* replace1320_0_1_0 -> 21 in the cache */
   /* replace1320_0_1 -> 2726 in the cache */
   /* replace1320_0 -> 3790 in the cache */
   /* replace1320_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 3790, 122 },
      -1,
   } },

   /* ('ult', 0, 'a@64') => ('ine', ('ior', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_y', 'a')), 0) */
   /* search1321_0 -> 897 in the cache */
   /* search1321_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 897, 21 },
      -1,
   } },

   /* replace1321_0_0_0 -> 21 in the cache */
   /* replace1321_0_0 -> 2732 in the cache */
   /* replace1321_0_1_0 -> 21 in the cache */
   /* replace1321_0_1 -> 2726 in the cache */
   /* replace1321_0 -> 3790 in the cache */
   /* replace1321_1 -> 122 in the cache */
   /* replace1321 -> 3791 in the cache */

   /* ('bitfield_insert', 'base', 'insert', 'offset', 'bits') => ('bcsel', ('ult', 31, 'bits'), 'insert', ('ior', ('iand', 'base', ('inot', ('ishl', ('isub', ('ishl', 1, 'bits'), 1), 'offset'))), ('iand', ('ishl', 'insert', 'offset'), ('ishl', ('isub', ('ishl', 1, 'bits'), 1), 'offset')))) */
   /* search1322_0 -> 3696 in the cache */
   /* search1322_1 -> 3697 in the cache */
   /* search1322_2 -> 3698 in the cache */
   /* search1322_3 -> 3699 in the cache */
   /* search1322 -> 3700 in the cache */

   /* replace1322_0_0 -> 19 in the cache */
   /* replace1322_0_1 -> 3699 in the cache */
   /* replace1322_0 -> 3701 in the cache */
   /* replace1322_1 -> 3697 in the cache */
   /* replace1322_2_0_0 -> 3696 in the cache */
   /* replace1322_2_0_1_0_0_0_0 -> 396 in the cache */
   /* replace1322_2_0_1_0_0_0_1 -> 3699 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 396, 3699 },
      -1,
   } },
   /* replace1322_2_0_1_0_0_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 3795, 396 },
      -1,
   } },
   /* replace1322_2_0_1_0_1 -> 3698 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3796, 3698 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 3797 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 3696, 3798 },
      -1,
   } },
   /* replace1322_2_1_0_0 -> 3697 in the cache */
   /* replace1322_2_1_0_1 -> 3698 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3697, 3698 },
      -1,
   } },
   /* replace1322_2_1_1_0_0_0 -> 396 in the cache */
   /* replace1322_2_1_1_0_0_1 -> 3699 in the cache */
   /* replace1322_2_1_1_0_0 -> 3795 in the cache */
   /* replace1322_2_1_1_0_1 -> 396 in the cache */
   /* replace1322_2_1_1_0 -> 3796 in the cache */
   /* replace1322_2_1_1_1 -> 3698 in the cache */
   /* replace1322_2_1_1 -> 3797 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 3800, 3797 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 3799, 3801 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 3,
      { 3701, 3697, 3802 },
      -1,
   } },

   /* ('bitfield_insert', 'base', 'insert', 'offset', 'bits') => ('bcsel', ('ult', 31, 'bits'), 'insert', ('bitfield_select', ('bfm', 'bits', 'offset'), ('ishl', 'insert', 'offset'), 'base')) */
   /* search1323_0 -> 3696 in the cache */
   /* search1323_1 -> 3697 in the cache */
   /* search1323_2 -> 3698 in the cache */
   /* search1323_3 -> 3699 in the cache */
   /* search1323 -> 3700 in the cache */

   /* replace1323_0_0 -> 19 in the cache */
   /* replace1323_0_1 -> 3699 in the cache */
   /* replace1323_0 -> 3701 in the cache */
   /* replace1323_1 -> 3697 in the cache */
   /* replace1323_2_0_0 -> 3699 in the cache */
   /* replace1323_2_0_1 -> 3698 in the cache */
   /* replace1323_2_0 -> 3702 in the cache */
   /* replace1323_2_1_0 -> 3697 in the cache */
   /* replace1323_2_1_1 -> 3698 in the cache */
   /* replace1323_2_1 -> 3800 in the cache */
   /* replace1323_2_2 -> 3696 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bitfield_select,
      -1, 0,
      { 3702, 3800, 3696 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3701, 3697, 3804 },
      -1,
   } },

   /* ('ibitfield_extract', 'value', 'offset', 'bits') => ('bcsel', ('ult', 31, 'bits'), 'value', ('ibfe', 'value', 'offset', 'bits')) */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* value */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* offset */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* bits */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibitfield_extract,
      -1, 0,
      { 3806, 3807, 3808 },
      -1,
   } },

   /* replace1324_0_0 -> 19 in the cache */
   /* replace1324_0_1 -> 3808 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 19, 3808 },
      -1,
   } },
   /* replace1324_1 -> 3806 in the cache */
   /* replace1324_2_0 -> 3806 in the cache */
   /* replace1324_2_1 -> 3807 in the cache */
   /* replace1324_2_2 -> 3808 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 3806, 3807, 3808 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3810, 3806, 3811 },
      -1,
   } },

   /* ('ubitfield_extract', 'value', 'offset', 'bits') => ('bcsel', ('ult', 31, 'bits'), 'value', ('ubfe', 'value', 'offset', 'bits')) */
   /* search1325_0 -> 3806 in the cache */
   /* search1325_1 -> 3807 in the cache */
   /* search1325_2 -> 3808 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubitfield_extract,
      -1, 0,
      { 3806, 3807, 3808 },
      -1,
   } },

   /* replace1325_0_0 -> 19 in the cache */
   /* replace1325_0_1 -> 3808 in the cache */
   /* replace1325_0 -> 3810 in the cache */
   /* replace1325_1 -> 3806 in the cache */
   /* replace1325_2_0 -> 3806 in the cache */
   /* replace1325_2_1 -> 3807 in the cache */
   /* replace1325_2_2 -> 3808 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 3806, 3807, 3808 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3810, 3806, 3814 },
      -1,
   } },

   /* ('bitfield_select', 'a', 'b', 0) => ('iand', 'a', 'b') */
   /* search1326_0 -> 0 in the cache */
   /* search1326_1 -> 3 in the cache */
   /* search1326_2 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bitfield_select,
      -1, 0,
      { 0, 3, 86 },
      -1,
   } },

   /* replace1326_0 -> 0 in the cache */
   /* replace1326_1 -> 3 in the cache */
   /* replace1326 -> 2719 in the cache */

   /* ('bitfield_select', 'a', ('iand', 'a', 'b'), 'c') => ('bitfield_select', 'a', 'b', 'c') */
   /* search1327_0 -> 126 in the cache */
   /* search1327_1_0 -> 126 in the cache */
   /* search1327_1_1 -> 127 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 126, 127 },
      -1,
   } },
   /* search1327_2 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bitfield_select,
      -1, 1,
      { 126, 3817, 129 },
      -1,
   } },

   /* replace1327_0 -> 126 in the cache */
   /* replace1327_1 -> 127 in the cache */
   /* replace1327_2 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bitfield_select,
      -1, 0,
      { 126, 127, 129 },
      -1,
   } },

   /* ('ubfe', 'value', 'offset', ('iand', 31, 'bits')) => ('ubfe', 'value', 'offset', 'bits') */
   /* search1328_0 -> 3806 in the cache */
   /* search1328_1 -> 3807 in the cache */
   /* search1328_2_0 -> 19 in the cache */
   /* search1328_2_1 -> 3808 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 19, 3808 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 1,
      { 3806, 3807, 3820 },
      -1,
   } },

   /* replace1328_0 -> 3806 in the cache */
   /* replace1328_1 -> 3807 in the cache */
   /* replace1328_2 -> 3808 in the cache */
   /* replace1328 -> 3814 in the cache */

   /* ('ubfe', 'value', ('iand', 31, 'offset'), 'bits') => ('ubfe', 'value', 'offset', 'bits') */
   /* search1329_0 -> 3806 in the cache */
   /* search1329_1_0 -> 19 in the cache */
   /* search1329_1_1 -> 3807 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 19, 3807 },
      -1,
   } },
   /* search1329_2 -> 3808 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 1,
      { 3806, 3822, 3808 },
      -1,
   } },

   /* replace1329_0 -> 3806 in the cache */
   /* replace1329_1 -> 3807 in the cache */
   /* replace1329_2 -> 3808 in the cache */
   /* replace1329 -> 3814 in the cache */

   /* ('ibfe', 'value', 'offset', ('iand', 31, 'bits')) => ('ibfe', 'value', 'offset', 'bits') */
   /* search1330_0 -> 3806 in the cache */
   /* search1330_1 -> 3807 in the cache */
   /* search1330_2_0 -> 19 in the cache */
   /* search1330_2_1 -> 3808 in the cache */
   /* search1330_2 -> 3820 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 1,
      { 3806, 3807, 3820 },
      -1,
   } },

   /* replace1330_0 -> 3806 in the cache */
   /* replace1330_1 -> 3807 in the cache */
   /* replace1330_2 -> 3808 in the cache */
   /* replace1330 -> 3811 in the cache */

   /* ('ibfe', 'value', ('iand', 31, 'offset'), 'bits') => ('ibfe', 'value', 'offset', 'bits') */
   /* search1331_0 -> 3806 in the cache */
   /* search1331_1_0 -> 19 in the cache */
   /* search1331_1_1 -> 3807 in the cache */
   /* search1331_1 -> 3822 in the cache */
   /* search1331_2 -> 3808 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 1,
      { 3806, 3822, 3808 },
      -1,
   } },

   /* replace1331_0 -> 3806 in the cache */
   /* replace1331_1 -> 3807 in the cache */
   /* replace1331_2 -> 3808 in the cache */
   /* replace1331 -> 3811 in the cache */

   /* ('bfm', 'bits', ('iand', 31, 'offset')) => ('bfm', 'bits', 'offset') */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* bits */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1332_1_0 -> 19 in the cache */
   /* search1332_1_1 -> 3807 in the cache */
   /* search1332_1 -> 3822 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfm,
      -1, 1,
      { 3826, 3822 },
      -1,
   } },

   /* replace1332_0 -> 3826 in the cache */
   /* replace1332_1 -> 3807 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfm,
      -1, 0,
      { 3826, 3807 },
      -1,
   } },

   /* ('bfm', ('iand', 31, 'bits'), 'offset') => ('bfm', 'bits', 'offset') */
   /* search1333_0_0 -> 19 in the cache */
   /* search1333_0_1 -> 3826 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 19, 3826 },
      -1,
   } },
   /* search1333_1 -> 3807 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfm,
      -1, 1,
      { 3829, 3807 },
      -1,
   } },

   /* replace1333_0 -> 3826 in the cache */
   /* replace1333_1 -> 3807 in the cache */
   /* replace1333 -> 3828 in the cache */

   /* ('ult', 'a', ('umin', ('iand', 'a', 'b'), 'c')) => False */
   /* search1334_0 -> 126 in the cache */
   /* search1334_1_0_0 -> 126 in the cache */
   /* search1334_1_0_1 -> 127 in the cache */
   /* search1334_1_0 -> 194 in the cache */
   /* search1334_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umin,
      0, 2,
      { 194, 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 2,
      { 126, 3831 },
      -1,
   } },

   /* replace1334 -> 1077 in the cache */

   /* ('ult', 31, ('umin', '#bits(is_ult_32)', 'a')) => False */
   { .constant = {
      { nir_search_value_constant, -2 },
      nir_type_int, { 0x1f /* 31 */ },
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* bits */
      true,
      nir_type_invalid,
      34,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1335_1_1 -> 1114 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 3834, 1114 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 3833, 3835 },
      -1,
   } },

   /* replace1335 -> 1077 in the cache */

   /* ('ubfe', 'value', 'offset', ('umin', 'width', ('iadd', 32, ('ineg', ('iand', 31, 'offset'))))) => ('ubfe', 'value', 'offset', 'width') */
   /* search1336_0 -> 3806 in the cache */
   /* search1336_1 -> 3807 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* width */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1336_2_1_0 -> 874 in the cache */
   /* search1336_2_1_1_0_0 -> 19 in the cache */
   /* search1336_2_1_1_0_1 -> 3807 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 19, 3807 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 1,
      { 3838 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 2,
      { 874, 3839 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umin,
      0, 3,
      { 3837, 3840 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 3,
      { 3806, 3807, 3841 },
      -1,
   } },

   /* replace1336_0 -> 3806 in the cache */
   /* replace1336_1 -> 3807 in the cache */
   /* replace1336_2 -> 3837 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 3806, 3807, 3837 },
      -1,
   } },

   /* ('ibfe', 'value', 'offset', ('umin', 'width', ('iadd', 32, ('ineg', ('iand', 31, 'offset'))))) => ('ibfe', 'value', 'offset', 'width') */
   /* search1337_0 -> 3806 in the cache */
   /* search1337_1 -> 3807 in the cache */
   /* search1337_2_0 -> 3837 in the cache */
   /* search1337_2_1_0 -> 874 in the cache */
   /* search1337_2_1_1_0_0 -> 19 in the cache */
   /* search1337_2_1_1_0_1 -> 3807 in the cache */
   /* search1337_2_1_1_0 -> 3838 in the cache */
   /* search1337_2_1_1 -> 3839 in the cache */
   /* search1337_2_1 -> 3840 in the cache */
   /* search1337_2 -> 3841 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 3,
      { 3806, 3807, 3841 },
      -1,
   } },

   /* replace1337_0 -> 3806 in the cache */
   /* replace1337_1 -> 3807 in the cache */
   /* replace1337_2 -> 3837 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 3806, 3807, 3837 },
      -1,
   } },

   /* ('bfm', ('umin', 'width', ('iadd', 32, ('ineg', ('iand', 31, 'offset')))), 'offset') => ('bfm', 'width', 'offset') */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* width */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1338_0_1_0 -> 874 in the cache */
   /* search1338_0_1_1_0_0 -> 19 in the cache */
   /* search1338_0_1_1_0_1 -> 3807 in the cache */
   /* search1338_0_1_1_0 -> 3838 in the cache */
   /* search1338_0_1_1 -> 3839 in the cache */
   /* search1338_0_1 -> 3840 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umin,
      0, 3,
      { 3846, 3840 },
      -1,
   } },
   /* search1338_1 -> 3807 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfm,
      -1, 3,
      { 3847, 3807 },
      -1,
   } },

   /* replace1338_0 -> 3846 in the cache */
   /* replace1338_1 -> 3807 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfm,
      -1, 0,
      { 3846, 3807 },
      -1,
   } },

   /* ('iadd@32', ('ishl', 1, 'a'), -1) => ('bfm', 'a', 0) */
   /* search1339_0_0 -> 396 in the cache */
   /* search1339_0_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 396, 16 },
      -1,
   } },
   /* search1339_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 3850, 124 },
      -1,
   } },

   /* replace1339_0 -> 16 in the cache */
   /* replace1339_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfm,
      -1, 0,
      { 16, 122 },
      -1,
   } },

   /* ('ishl', ('bfm', 'a', 0), 'b') => ('bfm', 'a', 'b') */
   /* search1340_0_0 -> 16 in the cache */
   /* search1340_0_1 -> 122 in the cache */
   /* search1340_0 -> 3852 in the cache */
   /* search1340_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3852, 36 },
      -1,
   } },

   /* replace1340_0 -> 16 in the cache */
   /* replace1340_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfm,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* ('ubfe', 'a', 'b', 0) => 0 */
   /* search1341_0 -> 16 in the cache */
   /* search1341_1 -> 36 in the cache */
   /* search1341_2 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 16, 36, 122 },
      -1,
   } },

   /* replace1341 -> 122 in the cache */

   /* ('ibfe', 'a', 'b', 0) => 0 */
   /* search1342_0 -> 16 in the cache */
   /* search1342_1 -> 36 in the cache */
   /* search1342_2 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 16, 36, 122 },
      -1,
   } },

   /* replace1342 -> 122 in the cache */

   /* ('ubfe', 'a', 0, '#b') => ('iand', 'a', ('ushr', 4294967295, ('ineg', 'b'))) */
   /* search1343_0 -> 16 in the cache */
   /* search1343_1 -> 122 in the cache */
   /* search1343_2 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 16, 122, 33 },
      -1,
   } },

   /* replace1343_0 -> 16 in the cache */
   /* replace1343_1_0 -> 927 in the cache */
   /* replace1343_1_1_0 -> 36 in the cache */
   /* replace1343_1_1 -> 2772 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 927, 2772 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 16, 3858 },
      -1,
   } },

   /* ('b2i32', ('ine', ('ubfe', 'a', 'b', 1), 0)) => ('ubfe', 'a', 'b', 1) */
   /* search1344_0_0_0 -> 16 in the cache */
   /* search1344_0_0_1 -> 36 in the cache */
   /* search1344_0_0_2 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 16, 36, 396 },
      -1,
   } },
   /* search1344_0_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 3860, 122 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 1,
      { 3861 },
      -1,
   } },

   /* replace1344_0 -> 16 in the cache */
   /* replace1344_1 -> 36 in the cache */
   /* replace1344_2 -> 396 in the cache */
   /* replace1344 -> 3860 in the cache */

   /* ('b2i32', ('ine', ('ibfe', 'a', 'b', 1), 0)) => ('ubfe', 'a', 'b', 1) */
   /* search1345_0_0_0 -> 16 in the cache */
   /* search1345_0_0_1 -> 36 in the cache */
   /* search1345_0_0_2 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 16, 36, 396 },
      -1,
   } },
   /* search1345_0_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 3863, 122 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 1,
      { 3864 },
      -1,
   } },

   /* replace1345_0 -> 16 in the cache */
   /* replace1345_1 -> 36 in the cache */
   /* replace1345_2 -> 396 in the cache */
   /* replace1345 -> 3860 in the cache */

   /* ('ine', ('ibfe(is_used_once)', 'a', '#b', '#c'), 0) => ('ine', ('iand', 'a', ('ishl', ('ushr', 4294967295, ('ineg', 'c')), 'b')), 0) */
   /* search1346_0_0 -> 16 in the cache */
   /* search1346_0_1 -> 33 in the cache */
   /* search1346_0_2 -> 309 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 16, 33, 309 },
      0,
   } },
   /* search1346_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 3866, 122 },
      -1,
   } },

   /* replace1346_0_0 -> 16 in the cache */
   /* replace1346_0_1_0_0 -> 927 in the cache */
   /* replace1346_0_1_0_1_0 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 927, 3868 },
      -1,
   } },
   /* replace1346_0_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3869, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 16, 3870 },
      -1,
   } },
   /* replace1346_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 3871, 122 },
      -1,
   } },

   /* ('ieq', ('ibfe(is_used_once)', 'a', '#b', '#c'), 0) => ('ieq', ('iand', 'a', ('ishl', ('ushr', 4294967295, ('ineg', 'c')), 'b')), 0) */
   /* search1347_0_0 -> 16 in the cache */
   /* search1347_0_1 -> 33 in the cache */
   /* search1347_0_2 -> 309 in the cache */
   /* search1347_0 -> 3866 in the cache */
   /* search1347_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 3866, 122 },
      -1,
   } },

   /* replace1347_0_0 -> 16 in the cache */
   /* replace1347_0_1_0_0 -> 927 in the cache */
   /* replace1347_0_1_0_1_0 -> 135 in the cache */
   /* replace1347_0_1_0_1 -> 3868 in the cache */
   /* replace1347_0_1_0 -> 3869 in the cache */
   /* replace1347_0_1_1 -> 36 in the cache */
   /* replace1347_0_1 -> 3870 in the cache */
   /* replace1347_0 -> 3871 in the cache */
   /* replace1347_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 3871, 122 },
      -1,
   } },

   /* ('ine', ('ubfe(is_used_once)', 'a', '#b', '#c'), 0) => ('ine', ('iand', 'a', ('ishl', ('ushr', 4294967295, ('ineg', 'c')), 'b')), 0) */
   /* search1348_0_0 -> 16 in the cache */
   /* search1348_0_1 -> 33 in the cache */
   /* search1348_0_2 -> 309 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 16, 33, 309 },
      0,
   } },
   /* search1348_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 3875, 122 },
      -1,
   } },

   /* replace1348_0_0 -> 16 in the cache */
   /* replace1348_0_1_0_0 -> 927 in the cache */
   /* replace1348_0_1_0_1_0 -> 135 in the cache */
   /* replace1348_0_1_0_1 -> 3868 in the cache */
   /* replace1348_0_1_0 -> 3869 in the cache */
   /* replace1348_0_1_1 -> 36 in the cache */
   /* replace1348_0_1 -> 3870 in the cache */
   /* replace1348_0 -> 3871 in the cache */
   /* replace1348_1 -> 122 in the cache */
   /* replace1348 -> 3872 in the cache */

   /* ('ieq', ('ubfe(is_used_once)', 'a', '#b', '#c'), 0) => ('ieq', ('iand', 'a', ('ishl', ('ushr', 4294967295, ('ineg', 'c')), 'b')), 0) */
   /* search1349_0_0 -> 16 in the cache */
   /* search1349_0_1 -> 33 in the cache */
   /* search1349_0_2 -> 309 in the cache */
   /* search1349_0 -> 3875 in the cache */
   /* search1349_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 3875, 122 },
      -1,
   } },

   /* replace1349_0_0 -> 16 in the cache */
   /* replace1349_0_1_0_0 -> 927 in the cache */
   /* replace1349_0_1_0_1_0 -> 135 in the cache */
   /* replace1349_0_1_0_1 -> 3868 in the cache */
   /* replace1349_0_1_0 -> 3869 in the cache */
   /* replace1349_0_1_1 -> 36 in the cache */
   /* replace1349_0_1 -> 3870 in the cache */
   /* replace1349_0 -> 3871 in the cache */
   /* replace1349_1 -> 122 in the cache */
   /* replace1349 -> 3874 in the cache */

   /* ('ibitfield_extract', 'value', 'offset', 'bits') => ('bcsel', ('ieq', 0, 'bits'), 0, ('ishr', ('ishl', 'value', ('isub', ('isub', 32, 'bits'), 'offset')), ('isub', 32, 'bits'))) */
   /* search1350_0 -> 3806 in the cache */
   /* search1350_1 -> 3807 in the cache */
   /* search1350_2 -> 3808 in the cache */
   /* search1350 -> 3809 in the cache */

   /* replace1350_0_0 -> 122 in the cache */
   /* replace1350_0_1 -> 3808 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 122, 3808 },
      -1,
   } },
   /* replace1350_1 -> 122 in the cache */
   /* replace1350_2_0_0 -> 3806 in the cache */
   /* replace1350_2_0_1_0_0 -> 874 in the cache */
   /* replace1350_2_0_1_0_1 -> 3808 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 874, 3808 },
      -1,
   } },
   /* replace1350_2_0_1_1 -> 3807 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 3879, 3807 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3806, 3880 },
      -1,
   } },
   /* replace1350_2_1_0 -> 874 in the cache */
   /* replace1350_2_1_1 -> 3808 in the cache */
   /* replace1350_2_1 -> 3879 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 3881, 3879 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 3878, 122, 3882 },
      -1,
   } },

   /* ('ubitfield_extract', 'value', 'offset', 'bits') => ('iand', ('ushr', 'value', 'offset'), ('bcsel', ('ieq', 'bits', 32), 4294967295, ('isub', ('ishl', 1, 'bits'), 1))) */
   /* search1351_0 -> 3806 in the cache */
   /* search1351_1 -> 3807 in the cache */
   /* search1351_2 -> 3808 in the cache */
   /* search1351 -> 3813 in the cache */

   /* replace1351_0_0 -> 3806 in the cache */
   /* replace1351_0_1 -> 3807 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 3806, 3807 },
      -1,
   } },
   /* replace1351_1_0_0 -> 3808 in the cache */
   /* replace1351_1_0_1 -> 874 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 3808, 874 },
      -1,
   } },
   /* replace1351_1_1 -> 927 in the cache */
   /* replace1351_1_2_0_0 -> 396 in the cache */
   /* replace1351_1_2_0_1 -> 3808 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 396, 3808 },
      -1,
   } },
   /* replace1351_1_2_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 3886, 396 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 3885, 927, 3887 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 3884, 3888 },
      -1,
   } },

   /* ('ifind_msb', 'value') => ('ufind_msb', ('bcsel', ('ilt', 'value', 0), ('inot', 'value'), 'value')) */
   /* search1352_0 -> 3806 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ifind_msb,
      -1, 0,
      { 3806 },
      -1,
   } },

   /* replace1352_0_0_0 -> 3806 in the cache */
   /* replace1352_0_0_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 3806, 122 },
      -1,
   } },
   /* replace1352_0_1_0 -> 3806 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 3806 },
      -1,
   } },
   /* replace1352_0_2 -> 3806 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3891, 3892, 3806 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ufind_msb,
      -1, 0,
      { 3893 },
      -1,
   } },

   /* ('ifind_msb', 'value') => ('bcsel', ('ige', ('ifind_msb_rev', 'value'), 0), ('isub', 31, ('ifind_msb_rev', 'value')), ('ifind_msb_rev', 'value')) */
   /* search1353_0 -> 3806 in the cache */
   /* search1353 -> 3890 in the cache */

   /* replace1353_0_0_0 -> 3806 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ifind_msb_rev,
      -1, 0,
      { 3806 },
      -1,
   } },
   /* replace1353_0_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3895, 122 },
      -1,
   } },
   /* replace1353_1_0 -> 19 in the cache */
   /* replace1353_1_1_0 -> 3806 in the cache */
   /* replace1353_1_1 -> 3895 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 19, 3895 },
      -1,
   } },
   /* replace1353_2_0 -> 3806 in the cache */
   /* replace1353_2 -> 3895 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3896, 3897, 3895 },
      -1,
   } },

   /* ('ufind_msb', 'value') => ('bcsel', ('ige', ('ufind_msb_rev', 'value'), 0), ('isub', 31, ('ufind_msb_rev', 'value')), ('ufind_msb_rev', 'value')) */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* value */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ufind_msb,
      -1, 0,
      { 3899 },
      -1,
   } },

   /* replace1354_0_0_0 -> 3899 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ufind_msb_rev,
      -1, 0,
      { 3899 },
      -1,
   } },
   /* replace1354_0_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3901, 122 },
      -1,
   } },
   /* replace1354_1_0 -> 19 in the cache */
   /* replace1354_1_1_0 -> 3899 in the cache */
   /* replace1354_1_1 -> 3901 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 19, 3901 },
      -1,
   } },
   /* replace1354_2_0 -> 3899 in the cache */
   /* replace1354_2 -> 3901 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3902, 3903, 3901 },
      -1,
   } },

   /* ('uclz', 'a') => ('umin', 32, ('ufind_msb_rev', 'a')) */
   /* search1355_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_uclz,
      -1, 0,
      { 16 },
      -1,
   } },

   /* replace1355_0 -> 874 in the cache */
   /* replace1355_1_0 -> 16 in the cache */
   /* replace1355_1 -> 3619 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 874, 3619 },
      -1,
   } },

   /* ('find_lsb', 'value') => ('ufind_msb', ('iand', 'value', ('ineg', 'value'))) */
   /* search1356_0 -> 3899 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_find_lsb,
      -1, 0,
      { 3899 },
      -1,
   } },

   /* replace1356_0_0 -> 3899 in the cache */
   /* replace1356_0_1_0 -> 3899 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 3899 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 3899, 3908 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ufind_msb,
      -1, 1,
      { 3909 },
      -1,
   } },

   /* ('extract_i8', 'a', 'b@32') => ('ishr', ('ishl', 'a', ('imul', ('isub', 3, 'b'), 8)), 24) */
   /* search1357_0 -> 16 in the cache */
   /* search1357_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* replace1357_0_0 -> 16 in the cache */
   /* replace1357_0_1_0_0 -> 405 in the cache */
   /* replace1357_0_1_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 405, 36 },
      -1,
   } },
   /* replace1357_0_1_1 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 3912, 831 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 16, 3913 },
      -1,
   } },
   /* replace1357_1 -> 1012 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 3914, 1012 },
      -1,
   } },

   /* ('extract_u8', 'a', 'b@32') => ('iand', ('ushr', 'a', ('imul', 'b', 8)), 255) */
   /* search1358_0 -> 16 in the cache */
   /* search1358_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* replace1358_0_0 -> 16 in the cache */
   /* replace1358_0_1_0 -> 36 in the cache */
   /* replace1358_0_1_1 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 36, 831 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 16, 3917 },
      -1,
   } },
   /* replace1358_1 -> 1010 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 3918, 1010 },
      -1,
   } },

   /* ('extract_i16', 'a', 'b@32') => ('ishr', ('ishl', 'a', ('imul', ('isub', 1, 'b'), 16)), 16) */
   /* search1359_0 -> 16 in the cache */
   /* search1359_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* replace1359_0_0 -> 16 in the cache */
   /* replace1359_0_1_0_0 -> 396 in the cache */
   /* replace1359_0_1_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 396, 36 },
      -1,
   } },
   /* replace1359_0_1_1 -> 853 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 3921, 853 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 16, 3922 },
      -1,
   } },
   /* replace1359_1 -> 853 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 3923, 853 },
      -1,
   } },

   /* ('extract_u16', 'a', 'b@32') => ('iand', ('ushr', 'a', ('imul', 'b', 16)), 65535) */
   /* search1360_0 -> 16 in the cache */
   /* search1360_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* replace1360_0_0 -> 16 in the cache */
   /* replace1360_0_1_0 -> 36 in the cache */
   /* replace1360_0_1_1 -> 853 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 36, 853 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 16, 3926 },
      -1,
   } },
   /* replace1360_1 -> 1002 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 3927, 1002 },
      -1,
   } },

   /* ('pack_unorm_2x16', 'v') => ('pack_uvec2_to_uint', ('f2u32', ('fround_even', ('fmul', ('fsat', 'v'), 65535.0)))) */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* v */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_unorm_2x16,
      -1, 0,
      { 3929 },
      -1,
   } },

   /* replace1361_0_0_0_0_0 -> 3929 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 3929 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0x40efffe000000000 /* 65535.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 3931, 3932 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fround_even,
      -1, 1,
      { 3933 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2u32,
      -1, 1,
      { 3934 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_uvec2_to_uint,
      -1, 1,
      { 3935 },
      -1,
   } },

   /* ('pack_unorm_4x8', 'v') => ('pack_uvec4_to_uint', ('f2u32', ('fround_even', ('fmul', ('fsat', 'v'), 255.0)))) */
   /* search1362_0 -> 3929 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_unorm_4x8,
      -1, 0,
      { 3929 },
      -1,
   } },

   /* replace1362_0_0_0_0_0 -> 3929 in the cache */
   /* replace1362_0_0_0_0 -> 3931 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0x406fe00000000000 /* 255.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 3931, 3938 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fround_even,
      -1, 1,
      { 3939 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2u32,
      -1, 1,
      { 3940 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_uvec4_to_uint,
      -1, 1,
      { 3941 },
      -1,
   } },

   /* ('pack_snorm_2x16', 'v') => ('pack_uvec2_to_uint', ('f2i32', ('fround_even', ('fmul', ('fmin', 1.0, ('fmax', -1.0, 'v')), 32767.0)))) */
   /* search1363_0 -> 3929 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_snorm_2x16,
      -1, 0,
      { 3929 },
      -1,
   } },

   /* replace1363_0_0_0_0_0 -> 243 in the cache */
   /* replace1363_0_0_0_0_1_0 -> 278 in the cache */
   /* replace1363_0_0_0_0_1_1 -> 3929 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmax,
      2, 1,
      { 278, 3929 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 2,
      { 243, 3944 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0x40dfffc000000000 /* 32767.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 3,
      { 3945, 3946 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fround_even,
      -1, 3,
      { 3947 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2i32,
      -1, 3,
      { 3948 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_uvec2_to_uint,
      -1, 3,
      { 3949 },
      -1,
   } },

   /* ('pack_snorm_4x8', 'v') => ('pack_uvec4_to_uint', ('f2i32', ('fround_even', ('fmul', ('fmin', 1.0, ('fmax', -1.0, 'v')), 127.0)))) */
   /* search1364_0 -> 3929 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_snorm_4x8,
      -1, 0,
      { 3929 },
      -1,
   } },

   /* replace1364_0_0_0_0_0 -> 243 in the cache */
   /* replace1364_0_0_0_0_1_0 -> 278 in the cache */
   /* replace1364_0_0_0_0_1_1 -> 3929 in the cache */
   /* replace1364_0_0_0_0_1 -> 3944 in the cache */
   /* replace1364_0_0_0_0 -> 3945 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0x405fc00000000000 /* 127.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 3,
      { 3945, 3952 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fround_even,
      -1, 3,
      { 3953 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2i32,
      -1, 3,
      { 3954 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_uvec4_to_uint,
      -1, 3,
      { 3955 },
      -1,
   } },

   /* ('unpack_unorm_2x16', 'v') => ('fdiv', ('u2f32', ('vec2', ('extract_u16', 'v', 0), ('extract_u16', 'v', 1))), 65535.0) */
   /* search1365_0 -> 3929 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_unorm_2x16,
      -1, 0,
      { 3929 },
      -1,
   } },

   /* replace1365_0_0_0_0 -> 3929 in the cache */
   /* replace1365_0_0_0_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 3929, 122 },
      -1,
   } },
   /* replace1365_0_0_1_0 -> 3929 in the cache */
   /* replace1365_0_0_1_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 3929, 396 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 3958, 3959 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2f32,
      -1, 0,
      { 3960 },
      -1,
   } },
   /* replace1365_1 -> 3932 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fdiv,
      -1, 0,
      { 3961, 3932 },
      -1,
   } },

   /* ('unpack_unorm_4x8', 'v') => ('fdiv', ('u2f32', ('vec4', ('extract_u8', 'v', 0), ('extract_u8', 'v', 1), ('extract_u8', 'v', 2), ('extract_u8', 'v', 3))), 255.0) */
   /* search1366_0 -> 3929 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_unorm_4x8,
      -1, 0,
      { 3929 },
      -1,
   } },

   /* replace1366_0_0_0_0 -> 3929 in the cache */
   /* replace1366_0_0_0_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3929, 122 },
      -1,
   } },
   /* replace1366_0_0_1_0 -> 3929 in the cache */
   /* replace1366_0_0_1_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3929, 396 },
      -1,
   } },
   /* replace1366_0_0_2_0 -> 3929 in the cache */
   /* replace1366_0_0_2_1 -> 401 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3929, 401 },
      -1,
   } },
   /* replace1366_0_0_3_0 -> 3929 in the cache */
   /* replace1366_0_0_3_1 -> 405 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3929, 405 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 3964, 3965, 3966, 3967 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2f32,
      -1, 0,
      { 3968 },
      -1,
   } },
   /* replace1366_1 -> 3938 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fdiv,
      -1, 0,
      { 3969, 3938 },
      -1,
   } },

   /* ('unpack_snorm_2x16', 'v') => ('fmin', 1.0, ('fmax', -1.0, ('fdiv', ('i2f', ('vec2', ('extract_i16', 'v', 0), ('extract_i16', 'v', 1))), 32767.0))) */
   /* search1367_0 -> 3929 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_snorm_2x16,
      -1, 0,
      { 3929 },
      -1,
   } },

   /* replace1367_0 -> 243 in the cache */
   /* replace1367_1_0 -> 278 in the cache */
   /* replace1367_1_1_0_0_0_0 -> 3929 in the cache */
   /* replace1367_1_1_0_0_0_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 3929, 122 },
      -1,
   } },
   /* replace1367_1_1_0_0_1_0 -> 3929 in the cache */
   /* replace1367_1_1_0_0_1_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 3929, 396 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 3972, 3973 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_i2f,
      -1, 0,
      { 3974 },
      -1,
   } },
   /* replace1367_1_1_1 -> 3946 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fdiv,
      -1, 0,
      { 3975, 3946 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 278, 3976 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 2,
      { 243, 3977 },
      -1,
   } },

   /* ('unpack_snorm_4x8', 'v') => ('fmin', 1.0, ('fmax', -1.0, ('fdiv', ('i2f', ('vec4', ('extract_i8', 'v', 0), ('extract_i8', 'v', 1), ('extract_i8', 'v', 2), ('extract_i8', 'v', 3))), 127.0))) */
   /* search1368_0 -> 3929 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_snorm_4x8,
      -1, 0,
      { 3929 },
      -1,
   } },

   /* replace1368_0 -> 243 in the cache */
   /* replace1368_1_0 -> 278 in the cache */
   /* replace1368_1_1_0_0_0_0 -> 3929 in the cache */
   /* replace1368_1_1_0_0_0_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3929, 122 },
      -1,
   } },
   /* replace1368_1_1_0_0_1_0 -> 3929 in the cache */
   /* replace1368_1_1_0_0_1_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3929, 396 },
      -1,
   } },
   /* replace1368_1_1_0_0_2_0 -> 3929 in the cache */
   /* replace1368_1_1_0_0_2_1 -> 401 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3929, 401 },
      -1,
   } },
   /* replace1368_1_1_0_0_3_0 -> 3929 in the cache */
   /* replace1368_1_1_0_0_3_1 -> 405 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3929, 405 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 3980, 3981, 3982, 3983 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_i2f,
      -1, 0,
      { 3984 },
      -1,
   } },
   /* replace1368_1_1_1 -> 3952 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fdiv,
      -1, 0,
      { 3985, 3952 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 278, 3986 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 2,
      { 243, 3987 },
      -1,
   } },

   /* ('pack_half_2x16_split', 'a@32', 'b@32') => ('ior', ('ishl', ('u2u32', ('f2f16', 'b')), 16), ('u2u32', ('f2f16', 'a'))) */
   /* search1369_0 -> 16 in the cache */
   /* search1369_1 -> 36 in the cache */
   /* search1369 -> 3348 in the cache */

   /* replace1369_0_0_0_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 3989 },
      -1,
   } },
   /* replace1369_0_1 -> 853 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3990, 853 },
      -1,
   } },
   /* replace1369_1_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 3992 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 3991, 3993 },
      -1,
   } },

   /* ('unpack_half_2x16_split_x', 'a@32') => ('f2f32', ('u2u16', 'a')) */
   /* search1370_0 -> 16 in the cache */
   /* search1370 -> 3313 in the cache */

   /* replace1370_0_0 -> 16 in the cache */
   /* replace1370_0 -> 3310 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 3310 },
      -1,
   } },

   /* ('unpack_half_2x16_split_y', 'a@32') => ('f2f32', ('u2u16', ('ushr', 'a', 16))) */
   /* search1371_0 -> 16 in the cache */
   /* search1371 -> 3320 in the cache */

   /* replace1371_0_0_0 -> 16 in the cache */
   /* replace1371_0_0_1 -> 853 in the cache */
   /* replace1371_0_0 -> 1008 in the cache */
   /* replace1371_0 -> 3311 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 3311 },
      -1,
   } },

   /* ('pack_32_2x16_split', 'a@16', 'b@16') => ('ior', ('ishl', ('u2u32', 'b'), 16), ('u2u32', 'a')) */
   /* search1372_0 -> 11 in the cache */
   /* search1372_1 -> 258 in the cache */
   /* search1372 -> 3303 in the cache */

   /* replace1372_0_0_0 -> 258 in the cache */
   /* replace1372_0_0 -> 3304 in the cache */
   /* replace1372_0_1 -> 853 in the cache */
   /* replace1372_0 -> 3305 in the cache */
   /* replace1372_1_0 -> 11 in the cache */
   /* replace1372_1 -> 1806 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 3305, 1806 },
      -1,
   } },

   /* ('unpack_32_2x16_split_x', 'a@32') => ('u2u16', 'a') */
   /* search1373_0 -> 16 in the cache */
   /* search1373 -> 3226 in the cache */

   /* replace1373_0 -> 16 in the cache */
   /* replace1373 -> 3310 in the cache */

   /* ('unpack_32_2x16_split_y', 'a@32') => ('u2u16', ('ushr', 'a', 16)) */
   /* search1374_0 -> 16 in the cache */
   /* search1374 -> 3228 in the cache */

   /* replace1374_0_0 -> 16 in the cache */
   /* replace1374_0_1 -> 853 in the cache */
   /* replace1374_0 -> 1008 in the cache */
   /* replace1374 -> 3311 in the cache */

   /* ('isign', 'a') => ('imin', ('imax', 'a', -1), 1) */
   /* search1375_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_isign,
      -1, 0,
      { 32 },
      -1,
   } },

   /* replace1375_0_0 -> 32 in the cache */
   /* replace1375_0_1 -> 63 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imax,
      1, 1,
      { 32, 63 },
      -1,
   } },
   /* replace1375_1 -> 35 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imin,
      0, 2,
      { 3999, 35 },
      -1,
   } },

   /* ('imin', ('imax', 'a', -1), 1) => ('isign', 'a') */
   /* search1376_0_0 -> 32 in the cache */
   /* search1376_0_1 -> 63 in the cache */
   /* search1376_0 -> 3999 in the cache */
   /* search1376_1 -> 35 in the cache */
   /* search1376 -> 4000 in the cache */

   /* replace1376_0 -> 32 in the cache */
   /* replace1376 -> 3998 in the cache */

   /* ('imax', ('imin', 'a', 1), -1) => ('isign', 'a') */
   /* search1377_0_0 -> 32 in the cache */
   /* search1377_0_1 -> 35 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imin,
      1, 1,
      { 32, 35 },
      -1,
   } },
   /* search1377_1 -> 63 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imax,
      0, 2,
      { 4001, 63 },
      -1,
   } },

   /* replace1377_0 -> 32 in the cache */
   /* replace1377 -> 3998 in the cache */

   /* ('fsign', 'a') => ('fsub', ('b2f', ('!flt', 0.0, 'a')), ('b2f', ('!flt', 'a', 0.0))) */
   /* search1378_0 -> 32 in the cache */
   /* search1378 -> 250 in the cache */

   /* replace1378_0_0_0 -> 111 in the cache */
   /* replace1378_0_0_1 -> 32 in the cache */
   /* replace1378_0_0 -> 1526 in the cache */
   /* replace1378_0 -> 1527 in the cache */
   /* replace1378_1_0_0 -> 32 in the cache */
   /* replace1378_1_0_1 -> 111 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_flt,
      -1, 0,
      { 32, 111 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 4003 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 1527, 4004 },
      -1,
   } },

   /* ('amul', 'a', 'b') => ('imul', 'a', 'b') */
   /* search1379_0 -> 0 in the cache */
   /* search1379_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_amul,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace1379_0 -> 0 in the cache */
   /* replace1379_1 -> 3 in the cache */
   /* replace1379 -> 951 in the cache */

   /* ('umul24', 'a', 'b') => ('imul', ('iand', 'a', 16777215), ('iand', 'b', 16777215)) */
   /* search1380_0 -> 16 in the cache */
   /* search1380_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umul24,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace1380_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xffffff /* 16777215 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 16, 4008 },
      -1,
   } },
   /* replace1380_1_0 -> 36 in the cache */
   /* replace1380_1_1 -> 4008 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 36, 4008 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 3,
      { 4009, 4010 },
      -1,
   } },

   /* ('umad24', 'a', 'b', 'c') => ('iadd', ('imul', ('iand', 'a', 16777215), ('iand', 'b', 16777215)), 'c') */
   /* search1381_0 -> 16 in the cache */
   /* search1381_1 -> 36 in the cache */
   /* search1381_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umad24,
      0, 1,
      { 16, 36, 135 },
      -1,
   } },

   /* replace1381_0_0_0 -> 16 in the cache */
   /* replace1381_0_0_1 -> 4008 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 16, 4008 },
      -1,
   } },
   /* replace1381_0_1_0 -> 36 in the cache */
   /* replace1381_0_1_1 -> 4008 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      3, 1,
      { 36, 4008 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      1, 3,
      { 4013, 4014 },
      -1,
   } },
   /* replace1381_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 4,
      { 4015, 135 },
      -1,
   } },

   /* ('imul24_relaxed', 'a', 'b') => ('imul24', 'a', 'b') */
   /* search1382_0 -> 16 in the cache */
   /* search1382_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul24_relaxed,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace1382_0 -> 16 in the cache */
   /* replace1382_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul24,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* ('imul24_relaxed', 'a', 'b') => ('imul', 'a', 'b') */
   /* search1383_0 -> 16 in the cache */
   /* search1383_1 -> 36 in the cache */
   /* search1383 -> 4017 in the cache */

   /* replace1383_0 -> 16 in the cache */
   /* replace1383_1 -> 36 in the cache */
   /* replace1383 -> 49 in the cache */

   /* ('umad24_relaxed', 'a', 'b', 'c') => ('umad24', 'a', 'b', 'c') */
   /* search1384_0 -> 16 in the cache */
   /* search1384_1 -> 36 in the cache */
   /* search1384_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umad24_relaxed,
      0, 1,
      { 16, 36, 135 },
      -1,
   } },

   /* replace1384_0 -> 16 in the cache */
   /* replace1384_1 -> 36 in the cache */
   /* replace1384_2 -> 135 in the cache */
   /* replace1384 -> 4012 in the cache */

   /* ('umad24_relaxed', 'a', 'b', 'c') => ('iadd', ('umul24_relaxed', 'a', 'b'), 'c') */
   /* search1385_0 -> 16 in the cache */
   /* search1385_1 -> 36 in the cache */
   /* search1385_2 -> 135 in the cache */
   /* search1385 -> 4019 in the cache */

   /* replace1385_0_0 -> 16 in the cache */
   /* replace1385_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umul24_relaxed,
      1, 1,
      { 16, 36 },
      -1,
   } },
   /* replace1385_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 4020, 135 },
      -1,
   } },

   /* ('umul24_relaxed', 'a', 'b') => ('umul24', 'a', 'b') */
   /* search1386_0 -> 16 in the cache */
   /* search1386_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umul24_relaxed,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace1386_0 -> 16 in the cache */
   /* replace1386_1 -> 36 in the cache */
   /* replace1386 -> 4007 in the cache */

   /* ('umul24_relaxed', 'a', 'b') => ('imul', 'a', 'b') */
   /* search1387_0 -> 16 in the cache */
   /* search1387_1 -> 36 in the cache */
   /* search1387 -> 4022 in the cache */

   /* replace1387_0 -> 16 in the cache */
   /* replace1387_1 -> 36 in the cache */
   /* replace1387 -> 49 in the cache */

   /* ('imad24_ir3', 'a', 'b', 0) => ('imul24', 'a', 'b') */
   /* search1388_0 -> 16 in the cache */
   /* search1388_1 -> 36 in the cache */
   /* search1388_2 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imad24_ir3,
      0, 1,
      { 16, 36, 122 },
      -1,
   } },

   /* replace1388_0 -> 16 in the cache */
   /* replace1388_1 -> 36 in the cache */
   /* replace1388 -> 4018 in the cache */

   /* ('imad24_ir3', 'a', 0, 'c') => c */
   /* search1389_0 -> 16 in the cache */
   /* search1389_1 -> 122 in the cache */
   /* search1389_2 -> 555 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imad24_ir3,
      0, 1,
      { 16, 122, 555 },
      -1,
   } },

   /* replace1389 -> 555 in the cache */

   /* ('imad24_ir3', 'a', 1, 'c') => ('iadd', 'a', 'c') */
   /* search1390_0 -> 16 in the cache */
   /* search1390_1 -> 396 in the cache */
   /* search1390_2 -> 555 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imad24_ir3,
      0, 1,
      { 16, 396, 555 },
      -1,
   } },

   /* replace1390_0 -> 16 in the cache */
   /* replace1390_1 -> 555 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 16, 555 },
      -1,
   } },

   /* ('imad24_ir3', '#a', '#b', 'c') => ('iadd', ('imul', 'a', 'b'), 'c') */
   /* search1391_0 -> 286 in the cache */
   /* search1391_1 -> 33 in the cache */
   /* search1391_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imad24_ir3,
      0, 1,
      { 286, 33, 135 },
      -1,
   } },

   /* replace1391_0_0 -> 16 in the cache */
   /* replace1391_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 16, 36 },
      -1,
   } },
   /* replace1391_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 4028, 135 },
      -1,
   } },

   /* ('imul24', 'a', '#b@32(is_pos_power_of_two)') => ('ishl', 'a', ('find_lsb', 'b')) */
   /* search1392_0 -> 16 in the cache */
   /* search1392_1 -> 2254 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul24,
      0, 1,
      { 16, 2254 },
      -1,
   } },

   /* replace1392_0 -> 16 in the cache */
   /* replace1392_1_0 -> 36 in the cache */
   /* replace1392_1 -> 2260 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 2260 },
      -1,
   } },

   /* ('imul24', 'a', '#b@32(is_neg_power_of_two)') => ('ineg', ('ishl', 'a', ('find_lsb', ('iabs', 'b')))) */
   /* search1393_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul24,
      0, 1,
      { 16, 4032 },
      -1,
   } },

   /* replace1393_0_0 -> 16 in the cache */
   /* replace1393_0_1_0_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_find_lsb,
      -1, 0,
      { 4034 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 4035 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 4036 },
      -1,
   } },

   /* ('imul24', 'a', 0) => 0 */
   /* search1394_0 -> 16 in the cache */
   /* search1394_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul24,
      0, 1,
      { 16, 122 },
      -1,
   } },

   /* replace1394 -> 122 in the cache */

   /* ('bcsel', ('ult', ('iadd', 'a', 'b'), 'a'), -1, ('iadd@8', 'a', 'b')) => ('uadd_sat', 'a', 'b') */
   /* search1395_0_0_0 -> 6 in the cache */
   /* search1395_0_0_1 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 6, 2112 },
      -1,
   } },
   /* search1395_0_1 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 4039, 6 },
      -1,
   } },
   /* search1395_1 -> 1435 in the cache */
   /* search1395_2_0 -> 6 in the cache */
   /* search1395_2_1 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 6, 2112 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4040, 1435, 4041 },
      -1,
   } },

   /* replace1395_0 -> 6 in the cache */
   /* replace1395_1 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_uadd_sat,
      0, 1,
      { 6, 2112 },
      -1,
   } },

   /* ('bcsel', ('uge', ('iadd', 'a', 'b'), 'a'), ('iadd@8', 'a', 'b'), -1) => ('uadd_sat', 'a', 'b') */
   /* search1396_0_0_0 -> 6 in the cache */
   /* search1396_0_0_1 -> 2112 in the cache */
   /* search1396_0_0 -> 4039 in the cache */
   /* search1396_0_1 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 4039, 6 },
      -1,
   } },
   /* search1396_1_0 -> 6 in the cache */
   /* search1396_1_1 -> 2112 in the cache */
   /* search1396_1 -> 4041 in the cache */
   /* search1396_2 -> 1435 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4044, 4041, 1435 },
      -1,
   } },

   /* replace1396_0 -> 6 in the cache */
   /* replace1396_1 -> 2112 in the cache */
   /* replace1396 -> 4043 in the cache */

   /* ('bcsel', ('ieq', ('uadd_carry', 'a', 'b'), 0), ('iadd@8', 'a', 'b'), -1) => ('uadd_sat', 'a', 'b') */
   /* search1397_0_0_0 -> 6 in the cache */
   /* search1397_0_0_1 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_uadd_carry,
      1, 1,
      { 6, 2112 },
      -1,
   } },
   /* search1397_0_1 -> 835 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 4046, 835 },
      -1,
   } },
   /* search1397_1_0 -> 6 in the cache */
   /* search1397_1_1 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 1,
      { 6, 2112 },
      -1,
   } },
   /* search1397_2 -> 1435 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 3,
      { 4047, 4048, 1435 },
      -1,
   } },

   /* replace1397_0 -> 6 in the cache */
   /* replace1397_1 -> 2112 in the cache */
   /* replace1397 -> 4043 in the cache */

   /* ('bcsel', ('ine', ('uadd_carry', 'a', 'b'), 0), -1, ('iadd@8', 'a', 'b')) => ('uadd_sat', 'a', 'b') */
   /* search1398_0_0_0 -> 6 in the cache */
   /* search1398_0_0_1 -> 2112 in the cache */
   /* search1398_0_0 -> 4046 in the cache */
   /* search1398_0_1 -> 835 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 4046, 835 },
      -1,
   } },
   /* search1398_1 -> 1435 in the cache */
   /* search1398_2_0 -> 6 in the cache */
   /* search1398_2_1 -> 2112 in the cache */
   /* search1398_2 -> 4048 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 3,
      { 4050, 1435, 4048 },
      -1,
   } },

   /* replace1398_0 -> 6 in the cache */
   /* replace1398_1 -> 2112 in the cache */
   /* replace1398 -> 4043 in the cache */

   /* ('bcsel', ('ult', ('iadd', 'a', 'b'), 'a'), -1, ('iadd@16', 'a', 'b')) => ('uadd_sat', 'a', 'b') */
   /* search1399_0_0_0 -> 11 in the cache */
   /* search1399_0_0_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 11, 258 },
      -1,
   } },
   /* search1399_0_1 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 4052, 11 },
      -1,
   } },
   /* search1399_1 -> 1457 in the cache */
   /* search1399_2_0 -> 11 in the cache */
   /* search1399_2_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 11, 258 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4053, 1457, 4054 },
      -1,
   } },

   /* replace1399_0 -> 11 in the cache */
   /* replace1399_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_uadd_sat,
      0, 1,
      { 11, 258 },
      -1,
   } },

   /* ('bcsel', ('uge', ('iadd', 'a', 'b'), 'a'), ('iadd@16', 'a', 'b'), -1) => ('uadd_sat', 'a', 'b') */
   /* search1400_0_0_0 -> 11 in the cache */
   /* search1400_0_0_1 -> 258 in the cache */
   /* search1400_0_0 -> 4052 in the cache */
   /* search1400_0_1 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 4052, 11 },
      -1,
   } },
   /* search1400_1_0 -> 11 in the cache */
   /* search1400_1_1 -> 258 in the cache */
   /* search1400_1 -> 4054 in the cache */
   /* search1400_2 -> 1457 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4057, 4054, 1457 },
      -1,
   } },

   /* replace1400_0 -> 11 in the cache */
   /* replace1400_1 -> 258 in the cache */
   /* replace1400 -> 4056 in the cache */

   /* ('bcsel', ('ieq', ('uadd_carry', 'a', 'b'), 0), ('iadd@16', 'a', 'b'), -1) => ('uadd_sat', 'a', 'b') */
   /* search1401_0_0_0 -> 11 in the cache */
   /* search1401_0_0_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_uadd_carry,
      1, 1,
      { 11, 258 },
      -1,
   } },
   /* search1401_0_1 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 4059, 856 },
      -1,
   } },
   /* search1401_1_0 -> 11 in the cache */
   /* search1401_1_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 1,
      { 11, 258 },
      -1,
   } },
   /* search1401_2 -> 1457 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 3,
      { 4060, 4061, 1457 },
      -1,
   } },

   /* replace1401_0 -> 11 in the cache */
   /* replace1401_1 -> 258 in the cache */
   /* replace1401 -> 4056 in the cache */

   /* ('bcsel', ('ine', ('uadd_carry', 'a', 'b'), 0), -1, ('iadd@16', 'a', 'b')) => ('uadd_sat', 'a', 'b') */
   /* search1402_0_0_0 -> 11 in the cache */
   /* search1402_0_0_1 -> 258 in the cache */
   /* search1402_0_0 -> 4059 in the cache */
   /* search1402_0_1 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 4059, 856 },
      -1,
   } },
   /* search1402_1 -> 1457 in the cache */
   /* search1402_2_0 -> 11 in the cache */
   /* search1402_2_1 -> 258 in the cache */
   /* search1402_2 -> 4061 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 3,
      { 4063, 1457, 4061 },
      -1,
   } },

   /* replace1402_0 -> 11 in the cache */
   /* replace1402_1 -> 258 in the cache */
   /* replace1402 -> 4056 in the cache */

   /* ('bcsel', ('ult', ('iadd', 'a', 'b'), 'a'), -1, ('iadd@32', 'a', 'b')) => ('uadd_sat', 'a', 'b') */
   /* search1403_0_0_0 -> 16 in the cache */
   /* search1403_0_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 16, 36 },
      -1,
   } },
   /* search1403_0_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 4065, 16 },
      -1,
   } },
   /* search1403_1 -> 124 in the cache */
   /* search1403_2_0 -> 16 in the cache */
   /* search1403_2_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 16, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4066, 124, 4067 },
      -1,
   } },

   /* replace1403_0 -> 16 in the cache */
   /* replace1403_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_uadd_sat,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* ('bcsel', ('uge', ('iadd', 'a', 'b'), 'a'), ('iadd@32', 'a', 'b'), -1) => ('uadd_sat', 'a', 'b') */
   /* search1404_0_0_0 -> 16 in the cache */
   /* search1404_0_0_1 -> 36 in the cache */
   /* search1404_0_0 -> 4065 in the cache */
   /* search1404_0_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 4065, 16 },
      -1,
   } },
   /* search1404_1_0 -> 16 in the cache */
   /* search1404_1_1 -> 36 in the cache */
   /* search1404_1 -> 4067 in the cache */
   /* search1404_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4070, 4067, 124 },
      -1,
   } },

   /* replace1404_0 -> 16 in the cache */
   /* replace1404_1 -> 36 in the cache */
   /* replace1404 -> 4069 in the cache */

   /* ('bcsel', ('ieq', ('uadd_carry', 'a', 'b'), 0), ('iadd@32', 'a', 'b'), -1) => ('uadd_sat', 'a', 'b') */
   /* search1405_0_0_0 -> 16 in the cache */
   /* search1405_0_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_uadd_carry,
      1, 1,
      { 16, 36 },
      -1,
   } },
   /* search1405_0_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 4072, 122 },
      -1,
   } },
   /* search1405_1_0 -> 16 in the cache */
   /* search1405_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 1,
      { 16, 36 },
      -1,
   } },
   /* search1405_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 3,
      { 4073, 4074, 124 },
      -1,
   } },

   /* replace1405_0 -> 16 in the cache */
   /* replace1405_1 -> 36 in the cache */
   /* replace1405 -> 4069 in the cache */

   /* ('bcsel', ('ine', ('uadd_carry', 'a', 'b'), 0), -1, ('iadd@32', 'a', 'b')) => ('uadd_sat', 'a', 'b') */
   /* search1406_0_0_0 -> 16 in the cache */
   /* search1406_0_0_1 -> 36 in the cache */
   /* search1406_0_0 -> 4072 in the cache */
   /* search1406_0_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 4072, 122 },
      -1,
   } },
   /* search1406_1 -> 124 in the cache */
   /* search1406_2_0 -> 16 in the cache */
   /* search1406_2_1 -> 36 in the cache */
   /* search1406_2 -> 4074 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 3,
      { 4076, 124, 4074 },
      -1,
   } },

   /* replace1406_0 -> 16 in the cache */
   /* replace1406_1 -> 36 in the cache */
   /* replace1406 -> 4069 in the cache */

   /* ('bcsel', ('ult', ('iadd', 'a', 'b'), 'a'), -1, ('iadd@64', 'a', 'b')) => ('uadd_sat', 'a', 'b') */
   /* search1407_0_0_0 -> 21 in the cache */
   /* search1407_0_0_1 -> 41 in the cache */
   /* search1407_0_0 -> 3739 in the cache */
   /* search1407_0_1 -> 21 in the cache */
   /* search1407_0 -> 3740 in the cache */
   /* search1407_1 -> 1499 in the cache */
   /* search1407_2_0 -> 21 in the cache */
   /* search1407_2_1 -> 41 in the cache */
   /* search1407_2 -> 3741 in the cache */
   /* search1407 -> 3742 in the cache */

   /* replace1407_0 -> 21 in the cache */
   /* replace1407_1 -> 41 in the cache */
   /* replace1407 -> 3738 in the cache */

   /* ('bcsel', ('uge', ('iadd', 'a', 'b'), 'a'), ('iadd@64', 'a', 'b'), -1) => ('uadd_sat', 'a', 'b') */
   /* search1408_0_0_0 -> 21 in the cache */
   /* search1408_0_0_1 -> 41 in the cache */
   /* search1408_0_0 -> 3739 in the cache */
   /* search1408_0_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 3739, 21 },
      -1,
   } },
   /* search1408_1_0 -> 21 in the cache */
   /* search1408_1_1 -> 41 in the cache */
   /* search1408_1 -> 3741 in the cache */
   /* search1408_2 -> 1499 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4078, 3741, 1499 },
      -1,
   } },

   /* replace1408_0 -> 21 in the cache */
   /* replace1408_1 -> 41 in the cache */
   /* replace1408 -> 3738 in the cache */

   /* ('bcsel', ('ieq', ('uadd_carry', 'a', 'b'), 0), ('iadd@64', 'a', 'b'), -1) => ('uadd_sat', 'a', 'b') */
   /* search1409_0_0_0 -> 21 in the cache */
   /* search1409_0_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_uadd_carry,
      1, 1,
      { 21, 41 },
      -1,
   } },
   /* search1409_0_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 4080, 897 },
      -1,
   } },
   /* search1409_1_0 -> 21 in the cache */
   /* search1409_1_1 -> 41 in the cache */
   /* search1409_1 -> 3755 in the cache */
   /* search1409_2 -> 1499 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 3,
      { 4081, 3755, 1499 },
      -1,
   } },

   /* replace1409_0 -> 21 in the cache */
   /* replace1409_1 -> 41 in the cache */
   /* replace1409 -> 3738 in the cache */

   /* ('bcsel', ('ine', ('uadd_carry', 'a', 'b'), 0), -1, ('iadd@64', 'a', 'b')) => ('uadd_sat', 'a', 'b') */
   /* search1410_0_0_0 -> 21 in the cache */
   /* search1410_0_0_1 -> 41 in the cache */
   /* search1410_0_0 -> 4080 in the cache */
   /* search1410_0_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 4080, 897 },
      -1,
   } },
   /* search1410_1 -> 1499 in the cache */
   /* search1410_2_0 -> 21 in the cache */
   /* search1410_2_1 -> 41 in the cache */
   /* search1410_2 -> 3755 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 3,
      { 4083, 1499, 3755 },
      -1,
   } },

   /* replace1410_0 -> 21 in the cache */
   /* replace1410_1 -> 41 in the cache */
   /* replace1410 -> 3738 in the cache */

   /* ('bcsel', ('ult', 'a', 'b'), 0, ('iadd@8', 'a', ('ineg', 'b'))) => ('usub_sat', 'a', 'b') */
   /* search1411_0_0 -> 6 in the cache */
   /* search1411_0_1 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 6, 2112 },
      -1,
   } },
   /* search1411_1 -> 835 in the cache */
   /* search1411_2_0 -> 6 in the cache */
   /* search1411_2_1_0 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 2112 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 6, 4086 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 4085, 835, 4087 },
      -1,
   } },

   /* replace1411_0 -> 6 in the cache */
   /* replace1411_1 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_usub_sat,
      -1, 0,
      { 6, 2112 },
      -1,
   } },

   /* ('bcsel', ('uge', 'a', 'b'), ('iadd@8', 'a', ('ineg', 'b')), 0) => ('usub_sat', 'a', 'b') */
   /* search1412_0_0 -> 6 in the cache */
   /* search1412_0_1 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 6, 2112 },
      -1,
   } },
   /* search1412_1_0 -> 6 in the cache */
   /* search1412_1_1_0 -> 2112 in the cache */
   /* search1412_1_1 -> 4086 in the cache */
   /* search1412_1 -> 4087 in the cache */
   /* search1412_2 -> 835 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 4090, 4087, 835 },
      -1,
   } },

   /* replace1412_0 -> 6 in the cache */
   /* replace1412_1 -> 2112 in the cache */
   /* replace1412 -> 4089 in the cache */

   /* ('bcsel', ('ieq', ('usub_borrow', 'a', 'b'), 0), ('iadd@8', 'a', ('ineg', 'b')), 0) => ('usub_sat', 'a', 'b') */
   /* search1413_0_0_0 -> 6 in the cache */
   /* search1413_0_0_1 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_usub_borrow,
      -1, 0,
      { 6, 2112 },
      -1,
   } },
   /* search1413_0_1 -> 835 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 4092, 835 },
      -1,
   } },
   /* search1413_1_0 -> 6 in the cache */
   /* search1413_1_1_0 -> 2112 in the cache */
   /* search1413_1_1 -> 4086 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 6, 4086 },
      -1,
   } },
   /* search1413_2 -> 835 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4093, 4094, 835 },
      -1,
   } },

   /* replace1413_0 -> 6 in the cache */
   /* replace1413_1 -> 2112 in the cache */
   /* replace1413 -> 4089 in the cache */

   /* ('bcsel', ('ine', ('usub_borrow', 'a', 'b'), 0), 0, ('iadd@8', 'a', ('ineg', 'b'))) => ('usub_sat', 'a', 'b') */
   /* search1414_0_0_0 -> 6 in the cache */
   /* search1414_0_0_1 -> 2112 in the cache */
   /* search1414_0_0 -> 4092 in the cache */
   /* search1414_0_1 -> 835 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 4092, 835 },
      -1,
   } },
   /* search1414_1 -> 835 in the cache */
   /* search1414_2_0 -> 6 in the cache */
   /* search1414_2_1_0 -> 2112 in the cache */
   /* search1414_2_1 -> 4086 in the cache */
   /* search1414_2 -> 4094 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4096, 835, 4094 },
      -1,
   } },

   /* replace1414_0 -> 6 in the cache */
   /* replace1414_1 -> 2112 in the cache */
   /* replace1414 -> 4089 in the cache */

   /* ('bcsel', ('ult', 'a', 'b'), 0, ('iadd@16', 'a', ('ineg', 'b'))) => ('usub_sat', 'a', 'b') */
   /* search1415_0_0 -> 11 in the cache */
   /* search1415_0_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 11, 258 },
      -1,
   } },
   /* search1415_1 -> 856 in the cache */
   /* search1415_2_0 -> 11 in the cache */
   /* search1415_2_1_0 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 258 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 11, 4099 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 4098, 856, 4100 },
      -1,
   } },

   /* replace1415_0 -> 11 in the cache */
   /* replace1415_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_usub_sat,
      -1, 0,
      { 11, 258 },
      -1,
   } },

   /* ('bcsel', ('uge', 'a', 'b'), ('iadd@16', 'a', ('ineg', 'b')), 0) => ('usub_sat', 'a', 'b') */
   /* search1416_0_0 -> 11 in the cache */
   /* search1416_0_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 11, 258 },
      -1,
   } },
   /* search1416_1_0 -> 11 in the cache */
   /* search1416_1_1_0 -> 258 in the cache */
   /* search1416_1_1 -> 4099 in the cache */
   /* search1416_1 -> 4100 in the cache */
   /* search1416_2 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 4103, 4100, 856 },
      -1,
   } },

   /* replace1416_0 -> 11 in the cache */
   /* replace1416_1 -> 258 in the cache */
   /* replace1416 -> 4102 in the cache */

   /* ('bcsel', ('ieq', ('usub_borrow', 'a', 'b'), 0), ('iadd@16', 'a', ('ineg', 'b')), 0) => ('usub_sat', 'a', 'b') */
   /* search1417_0_0_0 -> 11 in the cache */
   /* search1417_0_0_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_usub_borrow,
      -1, 0,
      { 11, 258 },
      -1,
   } },
   /* search1417_0_1 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 4105, 856 },
      -1,
   } },
   /* search1417_1_0 -> 11 in the cache */
   /* search1417_1_1_0 -> 258 in the cache */
   /* search1417_1_1 -> 4099 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 11, 4099 },
      -1,
   } },
   /* search1417_2 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4106, 4107, 856 },
      -1,
   } },

   /* replace1417_0 -> 11 in the cache */
   /* replace1417_1 -> 258 in the cache */
   /* replace1417 -> 4102 in the cache */

   /* ('bcsel', ('ine', ('usub_borrow', 'a', 'b'), 0), 0, ('iadd@16', 'a', ('ineg', 'b'))) => ('usub_sat', 'a', 'b') */
   /* search1418_0_0_0 -> 11 in the cache */
   /* search1418_0_0_1 -> 258 in the cache */
   /* search1418_0_0 -> 4105 in the cache */
   /* search1418_0_1 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 4105, 856 },
      -1,
   } },
   /* search1418_1 -> 856 in the cache */
   /* search1418_2_0 -> 11 in the cache */
   /* search1418_2_1_0 -> 258 in the cache */
   /* search1418_2_1 -> 4099 in the cache */
   /* search1418_2 -> 4107 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4109, 856, 4107 },
      -1,
   } },

   /* replace1418_0 -> 11 in the cache */
   /* replace1418_1 -> 258 in the cache */
   /* replace1418 -> 4102 in the cache */

   /* ('bcsel', ('ult', 'a', 'b'), 0, ('iadd@32', 'a', ('ineg', 'b'))) => ('usub_sat', 'a', 'b') */
   /* search1419_0_0 -> 16 in the cache */
   /* search1419_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 16, 36 },
      -1,
   } },
   /* search1419_1 -> 122 in the cache */
   /* search1419_2_0 -> 16 in the cache */
   /* search1419_2_1_0 -> 36 in the cache */
   /* search1419_2_1 -> 2772 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 16, 2772 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 4111, 122, 4112 },
      -1,
   } },

   /* replace1419_0 -> 16 in the cache */
   /* replace1419_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_usub_sat,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* ('bcsel', ('uge', 'a', 'b'), ('iadd@32', 'a', ('ineg', 'b')), 0) => ('usub_sat', 'a', 'b') */
   /* search1420_0_0 -> 16 in the cache */
   /* search1420_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 16, 36 },
      -1,
   } },
   /* search1420_1_0 -> 16 in the cache */
   /* search1420_1_1_0 -> 36 in the cache */
   /* search1420_1_1 -> 2772 in the cache */
   /* search1420_1 -> 4112 in the cache */
   /* search1420_2 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 4115, 4112, 122 },
      -1,
   } },

   /* replace1420_0 -> 16 in the cache */
   /* replace1420_1 -> 36 in the cache */
   /* replace1420 -> 4114 in the cache */

   /* ('bcsel', ('ieq', ('usub_borrow', 'a', 'b'), 0), ('iadd@32', 'a', ('ineg', 'b')), 0) => ('usub_sat', 'a', 'b') */
   /* search1421_0_0_0 -> 16 in the cache */
   /* search1421_0_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_usub_borrow,
      -1, 0,
      { 16, 36 },
      -1,
   } },
   /* search1421_0_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 4117, 122 },
      -1,
   } },
   /* search1421_1_0 -> 16 in the cache */
   /* search1421_1_1_0 -> 36 in the cache */
   /* search1421_1_1 -> 2772 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 16, 2772 },
      -1,
   } },
   /* search1421_2 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4118, 4119, 122 },
      -1,
   } },

   /* replace1421_0 -> 16 in the cache */
   /* replace1421_1 -> 36 in the cache */
   /* replace1421 -> 4114 in the cache */

   /* ('bcsel', ('ine', ('usub_borrow', 'a', 'b'), 0), 0, ('iadd@32', 'a', ('ineg', 'b'))) => ('usub_sat', 'a', 'b') */
   /* search1422_0_0_0 -> 16 in the cache */
   /* search1422_0_0_1 -> 36 in the cache */
   /* search1422_0_0 -> 4117 in the cache */
   /* search1422_0_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 4117, 122 },
      -1,
   } },
   /* search1422_1 -> 122 in the cache */
   /* search1422_2_0 -> 16 in the cache */
   /* search1422_2_1_0 -> 36 in the cache */
   /* search1422_2_1 -> 2772 in the cache */
   /* search1422_2 -> 4119 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4121, 122, 4119 },
      -1,
   } },

   /* replace1422_0 -> 16 in the cache */
   /* replace1422_1 -> 36 in the cache */
   /* replace1422 -> 4114 in the cache */

   /* ('bcsel', ('ult', 'a', 'b'), 0, ('iadd@64', 'a', ('ineg', 'b'))) => ('usub_sat', 'a', 'b') */
   /* search1423_0_0 -> 21 in the cache */
   /* search1423_0_1 -> 41 in the cache */
   /* search1423_0 -> 3749 in the cache */
   /* search1423_1 -> 897 in the cache */
   /* search1423_2_0 -> 21 in the cache */
   /* search1423_2_1_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 21, 4123 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 3749, 897, 4124 },
      -1,
   } },

   /* replace1423_0 -> 21 in the cache */
   /* replace1423_1 -> 41 in the cache */
   /* replace1423 -> 3748 in the cache */

   /* ('bcsel', ('uge', 'a', 'b'), ('iadd@64', 'a', ('ineg', 'b')), 0) => ('usub_sat', 'a', 'b') */
   /* search1424_0_0 -> 21 in the cache */
   /* search1424_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 21, 41 },
      -1,
   } },
   /* search1424_1_0 -> 21 in the cache */
   /* search1424_1_1_0 -> 41 in the cache */
   /* search1424_1_1 -> 4123 in the cache */
   /* search1424_1 -> 4124 in the cache */
   /* search1424_2 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 4126, 4124, 897 },
      -1,
   } },

   /* replace1424_0 -> 21 in the cache */
   /* replace1424_1 -> 41 in the cache */
   /* replace1424 -> 3748 in the cache */

   /* ('bcsel', ('ieq', ('usub_borrow', 'a', 'b'), 0), ('iadd@64', 'a', ('ineg', 'b')), 0) => ('usub_sat', 'a', 'b') */
   /* search1425_0_0_0 -> 21 in the cache */
   /* search1425_0_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_usub_borrow,
      -1, 0,
      { 21, 41 },
      -1,
   } },
   /* search1425_0_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 4128, 897 },
      -1,
   } },
   /* search1425_1_0 -> 21 in the cache */
   /* search1425_1_1_0 -> 41 in the cache */
   /* search1425_1_1 -> 4123 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 21, 4123 },
      -1,
   } },
   /* search1425_2 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4129, 4130, 897 },
      -1,
   } },

   /* replace1425_0 -> 21 in the cache */
   /* replace1425_1 -> 41 in the cache */
   /* replace1425 -> 3748 in the cache */

   /* ('bcsel', ('ine', ('usub_borrow', 'a', 'b'), 0), 0, ('iadd@64', 'a', ('ineg', 'b'))) => ('usub_sat', 'a', 'b') */
   /* search1426_0_0_0 -> 21 in the cache */
   /* search1426_0_0_1 -> 41 in the cache */
   /* search1426_0_0 -> 4128 in the cache */
   /* search1426_0_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 4128, 897 },
      -1,
   } },
   /* search1426_1 -> 897 in the cache */
   /* search1426_2_0 -> 21 in the cache */
   /* search1426_2_1_0 -> 41 in the cache */
   /* search1426_2_1 -> 4123 in the cache */
   /* search1426_2 -> 4130 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4132, 897, 4130 },
      -1,
   } },

   /* replace1426_0 -> 21 in the cache */
   /* replace1426_1 -> 41 in the cache */
   /* replace1426 -> 3748 in the cache */

   /* ('iadd_sat@8', 'a', 'b') => ('bcsel', ('ige', 'b', 1), ('bcsel', ('ilt', ('iadd', 'a', 'b'), 'a'), 127, ('iadd', 'a', 'b')), ('bcsel', ('ilt', 'a', ('iadd', 'a', 'b')), 128, ('iadd', 'a', 'b'))) */
   /* search1427_0 -> 6 in the cache */
   /* search1427_1 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_iadd_sat,
      0, 1,
      { 6, 2112 },
      -1,
   } },

   /* replace1427_0_0 -> 2112 in the cache */
   /* replace1427_0_1 -> 1430 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 2112, 1430 },
      -1,
   } },
   /* replace1427_1_0_0_0 -> 6 in the cache */
   /* replace1427_1_0_0_1 -> 2112 in the cache */
   /* replace1427_1_0_0 -> 4039 in the cache */
   /* replace1427_1_0_1 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 4039, 6 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 8 },
      nir_type_int, { 0x7f /* 127 */ },
   } },
   /* replace1427_1_2_0 -> 6 in the cache */
   /* replace1427_1_2_1 -> 2112 in the cache */
   /* replace1427_1_2 -> 4041 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4136, 4137, 4041 },
      -1,
   } },
   /* replace1427_2_0_0 -> 6 in the cache */
   /* replace1427_2_0_1_0 -> 6 in the cache */
   /* replace1427_2_0_1_1 -> 2112 in the cache */
   /* replace1427_2_0_1 -> 4048 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 6, 4048 },
      -1,
   } },
   /* replace1427_2_1 -> 7 in the cache */
   /* replace1427_2_2_0 -> 6 in the cache */
   /* replace1427_2_2_1 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_iadd,
      3, 1,
      { 6, 2112 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4139, 7, 4140 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 4135, 4138, 4141 },
      -1,
   } },

   /* ('isub_sat@8', 'a', 'b') => ('bcsel', ('ilt', 'b', 0), ('bcsel', ('ilt', ('isub', 'a', 'b'), 'a'), 127, ('isub', 'a', 'b')), ('bcsel', ('ilt', 'a', ('isub', 'a', 'b')), 128, ('isub', 'a', 'b'))) */
   /* search1428_0 -> 6 in the cache */
   /* search1428_1 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_isub_sat,
      -1, 0,
      { 6, 2112 },
      -1,
   } },

   /* replace1428_0_0 -> 2112 in the cache */
   /* replace1428_0_1 -> 835 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 2112, 835 },
      -1,
   } },
   /* replace1428_1_0_0_0 -> 6 in the cache */
   /* replace1428_1_0_0_1 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 6, 2112 },
      -1,
   } },
   /* replace1428_1_0_1 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4145, 6 },
      -1,
   } },
   /* replace1428_1_1 -> 4137 in the cache */
   /* replace1428_1_2_0 -> 6 in the cache */
   /* replace1428_1_2_1 -> 2112 in the cache */
   /* replace1428_1_2 -> 4145 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 4146, 4137, 4145 },
      -1,
   } },
   /* replace1428_2_0_0 -> 6 in the cache */
   /* replace1428_2_0_1_0 -> 6 in the cache */
   /* replace1428_2_0_1_1 -> 2112 in the cache */
   /* replace1428_2_0_1 -> 4145 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 6, 4145 },
      -1,
   } },
   /* replace1428_2_1 -> 7 in the cache */
   /* replace1428_2_2_0 -> 6 in the cache */
   /* replace1428_2_2_1 -> 2112 in the cache */
   /* replace1428_2_2 -> 4145 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 4148, 7, 4145 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 4144, 4147, 4149 },
      -1,
   } },

   /* ('iadd_sat@16', 'a', 'b') => ('bcsel', ('ige', 'b', 1), ('bcsel', ('ilt', ('iadd', 'a', 'b'), 'a'), 32767, ('iadd', 'a', 'b')), ('bcsel', ('ilt', 'a', ('iadd', 'a', 'b')), 32768, ('iadd', 'a', 'b'))) */
   /* search1429_0 -> 11 in the cache */
   /* search1429_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd_sat,
      0, 1,
      { 11, 258 },
      -1,
   } },

   /* replace1429_0_0 -> 258 in the cache */
   /* replace1429_0_1 -> 1452 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 258, 1452 },
      -1,
   } },
   /* replace1429_1_0_0_0 -> 11 in the cache */
   /* replace1429_1_0_0_1 -> 258 in the cache */
   /* replace1429_1_0_0 -> 4052 in the cache */
   /* replace1429_1_0_1 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 4052, 11 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { 0x7fff /* 32767 */ },
   } },
   /* replace1429_1_2_0 -> 11 in the cache */
   /* replace1429_1_2_1 -> 258 in the cache */
   /* replace1429_1_2 -> 4054 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4153, 4154, 4054 },
      -1,
   } },
   /* replace1429_2_0_0 -> 11 in the cache */
   /* replace1429_2_0_1_0 -> 11 in the cache */
   /* replace1429_2_0_1_1 -> 258 in the cache */
   /* replace1429_2_0_1 -> 4061 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 11, 4061 },
      -1,
   } },
   /* replace1429_2_1 -> 12 in the cache */
   /* replace1429_2_2_0 -> 11 in the cache */
   /* replace1429_2_2_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      3, 1,
      { 11, 258 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4156, 12, 4157 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 4152, 4155, 4158 },
      -1,
   } },

   /* ('isub_sat@16', 'a', 'b') => ('bcsel', ('ilt', 'b', 0), ('bcsel', ('ilt', ('isub', 'a', 'b'), 'a'), 32767, ('isub', 'a', 'b')), ('bcsel', ('ilt', 'a', ('isub', 'a', 'b')), 32768, ('isub', 'a', 'b'))) */
   /* search1430_0 -> 11 in the cache */
   /* search1430_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_isub_sat,
      -1, 0,
      { 11, 258 },
      -1,
   } },

   /* replace1430_0_0 -> 258 in the cache */
   /* replace1430_0_1 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 258, 856 },
      -1,
   } },
   /* replace1430_1_0_0_0 -> 11 in the cache */
   /* replace1430_1_0_0_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 11, 258 },
      -1,
   } },
   /* replace1430_1_0_1 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4162, 11 },
      -1,
   } },
   /* replace1430_1_1 -> 4154 in the cache */
   /* replace1430_1_2_0 -> 11 in the cache */
   /* replace1430_1_2_1 -> 258 in the cache */
   /* replace1430_1_2 -> 4162 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 4163, 4154, 4162 },
      -1,
   } },
   /* replace1430_2_0_0 -> 11 in the cache */
   /* replace1430_2_0_1_0 -> 11 in the cache */
   /* replace1430_2_0_1_1 -> 258 in the cache */
   /* replace1430_2_0_1 -> 4162 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 11, 4162 },
      -1,
   } },
   /* replace1430_2_1 -> 12 in the cache */
   /* replace1430_2_2_0 -> 11 in the cache */
   /* replace1430_2_2_1 -> 258 in the cache */
   /* replace1430_2_2 -> 4162 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 4165, 12, 4162 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 4161, 4164, 4166 },
      -1,
   } },

   /* ('iadd_sat@32', 'a', 'b') => ('bcsel', ('ige', 'b', 1), ('bcsel', ('ilt', ('iadd', 'a', 'b'), 'a'), 2147483647, ('iadd', 'a', 'b')), ('bcsel', ('ilt', 'a', ('iadd', 'a', 'b')), 2147483648, ('iadd', 'a', 'b'))) */
   /* search1431_0 -> 16 in the cache */
   /* search1431_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd_sat,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace1431_0_0 -> 36 in the cache */
   /* replace1431_0_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 36, 396 },
      -1,
   } },
   /* replace1431_1_0_0_0 -> 16 in the cache */
   /* replace1431_1_0_0_1 -> 36 in the cache */
   /* replace1431_1_0_0 -> 4065 in the cache */
   /* replace1431_1_0_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 4065, 16 },
      -1,
   } },
   /* replace1431_1_1 -> 988 in the cache */
   /* replace1431_1_2_0 -> 16 in the cache */
   /* replace1431_1_2_1 -> 36 in the cache */
   /* replace1431_1_2 -> 4067 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4170, 988, 4067 },
      -1,
   } },
   /* replace1431_2_0_0 -> 16 in the cache */
   /* replace1431_2_0_1_0 -> 16 in the cache */
   /* replace1431_2_0_1_1 -> 36 in the cache */
   /* replace1431_2_0_1 -> 4074 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 16, 4074 },
      -1,
   } },
   /* replace1431_2_1 -> 17 in the cache */
   /* replace1431_2_2_0 -> 16 in the cache */
   /* replace1431_2_2_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      3, 1,
      { 16, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4172, 17, 4173 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 4169, 4171, 4174 },
      -1,
   } },

   /* ('isub_sat@32', 'a', 'b') => ('bcsel', ('ilt', 'b', 0), ('bcsel', ('ilt', ('isub', 'a', 'b'), 'a'), 2147483647, ('isub', 'a', 'b')), ('bcsel', ('ilt', 'a', ('isub', 'a', 'b')), 2147483648, ('isub', 'a', 'b'))) */
   /* search1432_0 -> 16 in the cache */
   /* search1432_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub_sat,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* replace1432_0_0 -> 36 in the cache */
   /* replace1432_0_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 36, 122 },
      -1,
   } },
   /* replace1432_1_0_0_0 -> 16 in the cache */
   /* replace1432_1_0_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 16, 36 },
      -1,
   } },
   /* replace1432_1_0_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4178, 16 },
      -1,
   } },
   /* replace1432_1_1 -> 988 in the cache */
   /* replace1432_1_2_0 -> 16 in the cache */
   /* replace1432_1_2_1 -> 36 in the cache */
   /* replace1432_1_2 -> 4178 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 4179, 988, 4178 },
      -1,
   } },
   /* replace1432_2_0_0 -> 16 in the cache */
   /* replace1432_2_0_1_0 -> 16 in the cache */
   /* replace1432_2_0_1_1 -> 36 in the cache */
   /* replace1432_2_0_1 -> 4178 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 16, 4178 },
      -1,
   } },
   /* replace1432_2_1 -> 17 in the cache */
   /* replace1432_2_2_0 -> 16 in the cache */
   /* replace1432_2_2_1 -> 36 in the cache */
   /* replace1432_2_2 -> 4178 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 4181, 17, 4178 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 4177, 4180, 4182 },
      -1,
   } },

   /* ('iadd_sat@64', 'a', 'b') => ('bcsel', ('ige', 'b', 1), ('bcsel', ('ilt', ('iadd', 'a', 'b'), 'a'), 9223372036854775807, ('iadd', 'a', 'b')), ('bcsel', ('ilt', 'a', ('iadd', 'a', 'b')), 9223372036854775808, ('iadd', 'a', 'b'))) */
   /* search1433_0 -> 21 in the cache */
   /* search1433_1 -> 41 in the cache */
   /* search1433 -> 3752 in the cache */

   /* replace1433_0_0 -> 41 in the cache */
   /* replace1433_0_1 -> 1494 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 41, 1494 },
      -1,
   } },
   /* replace1433_1_0_0_0 -> 21 in the cache */
   /* replace1433_1_0_0_1 -> 41 in the cache */
   /* replace1433_1_0_0 -> 3739 in the cache */
   /* replace1433_1_0_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 3739, 21 },
      -1,
   } },
   /* replace1433_1_1 -> 3763 in the cache */
   /* replace1433_1_2_0 -> 21 in the cache */
   /* replace1433_1_2_1 -> 41 in the cache */
   /* replace1433_1_2 -> 3741 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4185, 3763, 3741 },
      -1,
   } },
   /* replace1433_2_0_0 -> 21 in the cache */
   /* replace1433_2_0_1_0 -> 21 in the cache */
   /* replace1433_2_0_1_1 -> 41 in the cache */
   /* replace1433_2_0_1 -> 3755 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 21, 3755 },
      -1,
   } },
   /* replace1433_2_1 -> 22 in the cache */
   /* replace1433_2_2_0 -> 21 in the cache */
   /* replace1433_2_2_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      3, 1,
      { 21, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 4187, 22, 4188 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 4184, 4186, 4189 },
      -1,
   } },

   /* ('isub_sat@64', 'a', 'b') => ('bcsel', ('ilt', 'b', 0), ('bcsel', ('ilt', ('isub', 'a', 'b'), 'a'), 9223372036854775807, ('isub', 'a', 'b')), ('bcsel', ('ilt', 'a', ('isub', 'a', 'b')), 9223372036854775808, ('isub', 'a', 'b'))) */
   /* search1434_0 -> 21 in the cache */
   /* search1434_1 -> 41 in the cache */
   /* search1434 -> 3766 in the cache */

   /* replace1434_0_0 -> 41 in the cache */
   /* replace1434_0_1 -> 897 in the cache */
   /* replace1434_0 -> 3753 in the cache */
   /* replace1434_1_0_0_0 -> 21 in the cache */
   /* replace1434_1_0_0_1 -> 41 in the cache */
   /* replace1434_1_0_0 -> 3750 in the cache */
   /* replace1434_1_0_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 3750, 21 },
      -1,
   } },
   /* replace1434_1_1 -> 3763 in the cache */
   /* replace1434_1_2_0 -> 21 in the cache */
   /* replace1434_1_2_1 -> 41 in the cache */
   /* replace1434_1_2 -> 3750 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 4191, 3763, 3750 },
      -1,
   } },
   /* replace1434_2_0_0 -> 21 in the cache */
   /* replace1434_2_0_1_0 -> 21 in the cache */
   /* replace1434_2_0_1_1 -> 41 in the cache */
   /* replace1434_2_0_1 -> 3750 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 21, 3750 },
      -1,
   } },
   /* replace1434_2_1 -> 22 in the cache */
   /* replace1434_2_2_0 -> 21 in the cache */
   /* replace1434_2_2_1 -> 41 in the cache */
   /* replace1434_2_2 -> 3750 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 4193, 22, 3750 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3753, 4192, 4194 },
      -1,
   } },

   /* ('inot', ('ior(is_used_once)', ('feq', 'a', 'b'), ('feq', 'c', 'd'))) => ('iand', ('fneu', 'a', 'b'), ('fneu', 'c', 'd')) */
   /* search1435_0_0_0 -> 0 in the cache */
   /* search1435_0_0_1 -> 3 in the cache */
   /* search1435_0_0 -> 1074 in the cache */
   /* search1435_0_1_0 -> 152 in the cache */
   /* search1435_0_1_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      2, 1,
      { 152, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 1074, 4196 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 3,
      { 4197 },
      -1,
   } },

   /* replace1435_0_0 -> 0 in the cache */
   /* replace1435_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* replace1435_1_0 -> 152 in the cache */
   /* replace1435_1_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      2, 1,
      { 152, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4199, 4200 },
      -1,
   } },

   /* ('inot', ('iand(is_used_once)', ('feq', 'a', 'b'), ('feq', 'c', 'd'))) => ('ior', ('fneu', 'a', 'b'), ('fneu', 'c', 'd')) */
   /* search1436_0_0_0 -> 0 in the cache */
   /* search1436_0_0_1 -> 3 in the cache */
   /* search1436_0_0 -> 1074 in the cache */
   /* search1436_0_1_0 -> 152 in the cache */
   /* search1436_0_1_1 -> 153 in the cache */
   /* search1436_0_1 -> 4196 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 1074, 4196 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 3,
      { 4202 },
      -1,
   } },

   /* replace1436_0_0 -> 0 in the cache */
   /* replace1436_0_1 -> 3 in the cache */
   /* replace1436_0 -> 4199 in the cache */
   /* replace1436_1_0 -> 152 in the cache */
   /* replace1436_1_1 -> 153 in the cache */
   /* replace1436_1 -> 4200 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4199, 4200 },
      -1,
   } },

   /* ('inot', ('ior(is_used_once)', ('feq', 'a', 'b'), ('fneu', 'c', 'd'))) => ('iand', ('fneu', 'a', 'b'), ('feq', 'c', 'd')) */
   /* search1437_0_0_0 -> 0 in the cache */
   /* search1437_0_0_1 -> 3 in the cache */
   /* search1437_0_0 -> 1074 in the cache */
   /* search1437_0_1_0 -> 152 in the cache */
   /* search1437_0_1_1 -> 153 in the cache */
   /* search1437_0_1 -> 4200 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 1074, 4200 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 3,
      { 4205 },
      -1,
   } },

   /* replace1437_0_0 -> 0 in the cache */
   /* replace1437_0_1 -> 3 in the cache */
   /* replace1437_0 -> 4199 in the cache */
   /* replace1437_1_0 -> 152 in the cache */
   /* replace1437_1_1 -> 153 in the cache */
   /* replace1437_1 -> 4196 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4199, 4196 },
      -1,
   } },

   /* ('inot', ('iand(is_used_once)', ('feq', 'a', 'b'), ('fneu', 'c', 'd'))) => ('ior', ('fneu', 'a', 'b'), ('feq', 'c', 'd')) */
   /* search1438_0_0_0 -> 0 in the cache */
   /* search1438_0_0_1 -> 3 in the cache */
   /* search1438_0_0 -> 1074 in the cache */
   /* search1438_0_1_0 -> 152 in the cache */
   /* search1438_0_1_1 -> 153 in the cache */
   /* search1438_0_1 -> 4200 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 1074, 4200 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 3,
      { 4208 },
      -1,
   } },

   /* replace1438_0_0 -> 0 in the cache */
   /* replace1438_0_1 -> 3 in the cache */
   /* replace1438_0 -> 4199 in the cache */
   /* replace1438_1_0 -> 152 in the cache */
   /* replace1438_1_1 -> 153 in the cache */
   /* replace1438_1 -> 4196 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4199, 4196 },
      -1,
   } },

   /* ('inot', ('ior(is_used_once)', ('fneu', 'a', 'b'), ('fneu', 'c', 'd'))) => ('iand', ('feq', 'a', 'b'), ('feq', 'c', 'd')) */
   /* search1439_0_0_0 -> 0 in the cache */
   /* search1439_0_0_1 -> 3 in the cache */
   /* search1439_0_0 -> 4199 in the cache */
   /* search1439_0_1_0 -> 152 in the cache */
   /* search1439_0_1_1 -> 153 in the cache */
   /* search1439_0_1 -> 4200 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4199, 4200 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 3,
      { 4211 },
      -1,
   } },

   /* replace1439_0_0 -> 0 in the cache */
   /* replace1439_0_1 -> 3 in the cache */
   /* replace1439_0 -> 1074 in the cache */
   /* replace1439_1_0 -> 152 in the cache */
   /* replace1439_1_1 -> 153 in the cache */
   /* replace1439_1 -> 4196 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 1074, 4196 },
      -1,
   } },

   /* ('inot', ('iand(is_used_once)', ('fneu', 'a', 'b'), ('fneu', 'c', 'd'))) => ('ior', ('feq', 'a', 'b'), ('feq', 'c', 'd')) */
   /* search1440_0_0_0 -> 0 in the cache */
   /* search1440_0_0_1 -> 3 in the cache */
   /* search1440_0_0 -> 4199 in the cache */
   /* search1440_0_1_0 -> 152 in the cache */
   /* search1440_0_1_1 -> 153 in the cache */
   /* search1440_0_1 -> 4200 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4199, 4200 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 3,
      { 4214 },
      -1,
   } },

   /* replace1440_0_0 -> 0 in the cache */
   /* replace1440_0_1 -> 3 in the cache */
   /* replace1440_0 -> 1074 in the cache */
   /* replace1440_1_0 -> 152 in the cache */
   /* replace1440_1_1 -> 153 in the cache */
   /* replace1440_1 -> 4196 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 1074, 4196 },
      -1,
   } },

   /* ('f2b1', ('b2f', 'a@1')) => a */
   /* search1441_0_0 -> 760 in the cache */
   /* search1441_0 -> 1129 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_f2b1,
      -1, 0,
      { 1129 },
      -1,
   } },

   /* replace1441 -> 760 in the cache */

   /* ('f2b8', ('b2f', 'a@8')) => a */
   /* search1442_0_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 6 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_f2b8,
      -1, 0,
      { 4218 },
      -1,
   } },

   /* replace1442 -> 6 in the cache */

   /* ('f2b16', ('b2f', 'a@16')) => a */
   /* search1443_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2b16,
      -1, 0,
      { 4220 },
      -1,
   } },

   /* replace1443 -> 11 in the cache */

   /* ('f2b32', ('b2f', 'a@32')) => a */
   /* search1444_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2b32,
      -1, 0,
      { 4222 },
      -1,
   } },

   /* replace1444 -> 16 in the cache */

   /* ('f2f', ('b2f', 'a')) => ('b2f', 'a') */
   /* search1445_0_0 -> 32 in the cache */
   /* search1445_0 -> 1528 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2f,
      -1, 0,
      { 1528 },
      -1,
   } },

   /* replace1445_0 -> 32 in the cache */
   /* replace1445 -> 1528 in the cache */

   /* ('f2u', ('b2f', 'a')) => ('b2i', 'a') */
   /* search1446_0_0 -> 32 in the cache */
   /* search1446_0 -> 1528 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2u,
      -1, 0,
      { 1528 },
      -1,
   } },

   /* replace1446_0 -> 32 in the cache */
   /* replace1446 -> 3095 in the cache */

   /* ('f2i', ('b2f', 'a')) => ('b2i', 'a') */
   /* search1447_0_0 -> 32 in the cache */
   /* search1447_0 -> 1528 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2i,
      -1, 0,
      { 1528 },
      -1,
   } },

   /* replace1447_0 -> 32 in the cache */
   /* replace1447 -> 3095 in the cache */

   /* ('u2f', ('b2i', 'a')) => ('b2f', 'a') */
   /* search1448_0_0 -> 32 in the cache */
   /* search1448_0 -> 3095 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_u2f,
      -1, 0,
      { 3095 },
      -1,
   } },

   /* replace1448_0 -> 32 in the cache */
   /* replace1448 -> 1528 in the cache */

   /* ('u2u', ('b2i', 'a')) => ('b2i', 'a') */
   /* search1449_0_0 -> 32 in the cache */
   /* search1449_0 -> 3095 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_u2u,
      -1, 0,
      { 3095 },
      -1,
   } },

   /* replace1449_0 -> 32 in the cache */
   /* replace1449 -> 3095 in the cache */

   /* ('i2f', ('b2i', 'a')) => ('b2f', 'a') */
   /* search1450_0_0 -> 32 in the cache */
   /* search1450_0 -> 3095 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_i2f,
      -1, 0,
      { 3095 },
      -1,
   } },

   /* replace1450_0 -> 32 in the cache */
   /* replace1450 -> 1528 in the cache */

   /* ('i2i', ('b2i', 'a')) => ('b2i', 'a') */
   /* search1451_0_0 -> 32 in the cache */
   /* search1451_0 -> 3095 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_i2i,
      -1, 0,
      { 3095 },
      -1,
   } },

   /* replace1451_0 -> 32 in the cache */
   /* replace1451 -> 3095 in the cache */

   /* ('i2i1', 'a@1') => a */
   /* search1452_0 -> 760 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2i1,
      -1, 0,
      { 760 },
      -1,
   } },

   /* replace1452 -> 760 in the cache */

   /* ('i2i8', 'a@8') => a */
   /* search1453_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 6 },
      -1,
   } },

   /* replace1453 -> 6 in the cache */

   /* ('i2i16', 'a@16') => a */
   /* search1454_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 11 },
      -1,
   } },

   /* replace1454 -> 11 in the cache */

   /* ('i2i32', 'a@32') => a */
   /* search1455_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 16 },
      -1,
   } },

   /* replace1455 -> 16 in the cache */

   /* ('i2i64', 'a@64') => a */
   /* search1456_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 21 },
      -1,
   } },

   /* replace1456 -> 21 in the cache */

   /* ('u2u1', 'a@1') => a */
   /* search1457_0 -> 760 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_u2u1,
      -1, 0,
      { 760 },
      -1,
   } },

   /* replace1457 -> 760 in the cache */

   /* ('u2u8', 'a@8') => a */
   /* search1458_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 6 },
      -1,
   } },

   /* replace1458 -> 6 in the cache */

   /* ('u2u16', 'a@16') => a */
   /* search1459_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 11 },
      -1,
   } },

   /* replace1459 -> 11 in the cache */

   /* ('u2u32', 'a@32') => a */
   /* search1460_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 16 },
      -1,
   } },

   /* replace1460 -> 16 in the cache */

   /* ('u2u64', 'a@64') => a */
   /* search1461_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 21 },
      -1,
   } },

   /* replace1461 -> 21 in the cache */

   /* ('f2f16', 'a@16') => a */
   /* search1462_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 11 },
      -1,
   } },

   /* replace1462 -> 11 in the cache */

   /* ('f2f32', 'a@32') => a */
   /* search1463_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 16 },
      -1,
   } },

   /* replace1463 -> 16 in the cache */

   /* ('f2f64', 'a@64') => a */
   /* search1464_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2f64,
      -1, 0,
      { 21 },
      -1,
   } },

   /* replace1464 -> 21 in the cache */

   /* ('b2b1', 'a@1') => a */
   /* search1465_0 -> 760 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_b2b1,
      -1, 0,
      { 760 },
      -1,
   } },

   /* replace1465 -> 760 in the cache */

   /* ('b2b8', 'a@8') => a */
   /* search1466_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_b2b8,
      -1, 0,
      { 6 },
      -1,
   } },

   /* replace1466 -> 6 in the cache */

   /* ('b2b16', 'a@16') => a */
   /* search1467_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2b16,
      -1, 0,
      { 11 },
      -1,
   } },

   /* replace1467 -> 11 in the cache */

   /* ('b2b32', 'a@32') => a */
   /* search1468_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2b32,
      -1, 0,
      { 16 },
      -1,
   } },

   /* replace1468 -> 16 in the cache */

   /* ('i2i1', ('i2i8', 'a')) => ('i2i1', 'a') */
   /* search1469_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2i1,
      -1, 0,
      { 4248 },
      -1,
   } },

   /* replace1469_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2i1,
      -1, 0,
      { 32 },
      -1,
   } },

   /* ('i2i1', ('u2u8', 'a')) => ('u2u1', 'a') */
   /* search1470_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2i1,
      -1, 0,
      { 4251 },
      -1,
   } },

   /* replace1470_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_u2u1,
      -1, 0,
      { 32 },
      -1,
   } },

   /* ('u2u1', ('i2i8', 'a')) => ('i2i1', 'a') */
   /* search1471_0_0 -> 32 in the cache */
   /* search1471_0 -> 4248 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_u2u1,
      -1, 0,
      { 4248 },
      -1,
   } },

   /* replace1471_0 -> 32 in the cache */
   /* replace1471 -> 4250 in the cache */

   /* ('u2u1', ('u2u8', 'a')) => ('u2u1', 'a') */
   /* search1472_0_0 -> 32 in the cache */
   /* search1472_0 -> 4251 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_u2u1,
      -1, 0,
      { 4251 },
      -1,
   } },

   /* replace1472_0 -> 32 in the cache */
   /* replace1472 -> 4253 in the cache */

   /* ('i2i1', ('i2i16', 'a')) => ('i2i1', 'a') */
   /* search1473_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2i1,
      -1, 0,
      { 4256 },
      -1,
   } },

   /* replace1473_0 -> 32 in the cache */
   /* replace1473 -> 4250 in the cache */

   /* ('i2i1', ('u2u16', 'a')) => ('u2u1', 'a') */
   /* search1474_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2i1,
      -1, 0,
      { 4258 },
      -1,
   } },

   /* replace1474_0 -> 32 in the cache */
   /* replace1474 -> 4253 in the cache */

   /* ('u2u1', ('i2i16', 'a')) => ('i2i1', 'a') */
   /* search1475_0_0 -> 32 in the cache */
   /* search1475_0 -> 4256 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_u2u1,
      -1, 0,
      { 4256 },
      -1,
   } },

   /* replace1475_0 -> 32 in the cache */
   /* replace1475 -> 4250 in the cache */

   /* ('u2u1', ('u2u16', 'a')) => ('u2u1', 'a') */
   /* search1476_0_0 -> 32 in the cache */
   /* search1476_0 -> 4258 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_u2u1,
      -1, 0,
      { 4258 },
      -1,
   } },

   /* replace1476_0 -> 32 in the cache */
   /* replace1476 -> 4253 in the cache */

   /* ('i2i1', ('i2i32', 'a')) => ('i2i1', 'a') */
   /* search1477_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2i1,
      -1, 0,
      { 4262 },
      -1,
   } },

   /* replace1477_0 -> 32 in the cache */
   /* replace1477 -> 4250 in the cache */

   /* ('i2i1', ('u2u32', 'a')) => ('u2u1', 'a') */
   /* search1478_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2i1,
      -1, 0,
      { 4264 },
      -1,
   } },

   /* replace1478_0 -> 32 in the cache */
   /* replace1478 -> 4253 in the cache */

   /* ('u2u1', ('i2i32', 'a')) => ('i2i1', 'a') */
   /* search1479_0_0 -> 32 in the cache */
   /* search1479_0 -> 4262 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_u2u1,
      -1, 0,
      { 4262 },
      -1,
   } },

   /* replace1479_0 -> 32 in the cache */
   /* replace1479 -> 4250 in the cache */

   /* ('u2u1', ('u2u32', 'a')) => ('u2u1', 'a') */
   /* search1480_0_0 -> 32 in the cache */
   /* search1480_0 -> 4264 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_u2u1,
      -1, 0,
      { 4264 },
      -1,
   } },

   /* replace1480_0 -> 32 in the cache */
   /* replace1480 -> 4253 in the cache */

   /* ('i2i1', ('i2i64', 'a')) => ('i2i1', 'a') */
   /* search1481_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2i1,
      -1, 0,
      { 4268 },
      -1,
   } },

   /* replace1481_0 -> 32 in the cache */
   /* replace1481 -> 4250 in the cache */

   /* ('i2i1', ('u2u64', 'a')) => ('u2u1', 'a') */
   /* search1482_0_0 -> 32 in the cache */
   /* search1482_0 -> 3209 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2i1,
      -1, 0,
      { 3209 },
      -1,
   } },

   /* replace1482_0 -> 32 in the cache */
   /* replace1482 -> 4253 in the cache */

   /* ('u2u1', ('i2i64', 'a')) => ('i2i1', 'a') */
   /* search1483_0_0 -> 32 in the cache */
   /* search1483_0 -> 4268 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_u2u1,
      -1, 0,
      { 4268 },
      -1,
   } },

   /* replace1483_0 -> 32 in the cache */
   /* replace1483 -> 4250 in the cache */

   /* ('u2u1', ('u2u64', 'a')) => ('u2u1', 'a') */
   /* search1484_0_0 -> 32 in the cache */
   /* search1484_0 -> 3209 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_u2u1,
      -1, 0,
      { 3209 },
      -1,
   } },

   /* replace1484_0 -> 32 in the cache */
   /* replace1484 -> 4253 in the cache */

   /* ('i2i8', ('i2i1', 'a@1')) => ('i2i8', 'a') */
   /* search1485_0_0 -> 760 in the cache */
   /* search1485_0 -> 4231 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 4231 },
      -1,
   } },

   /* replace1485_0 -> 760 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 760 },
      -1,
   } },

   /* ('u2u8', ('u2u1', 'a@1')) => ('u2u8', 'a') */
   /* search1486_0_0 -> 760 in the cache */
   /* search1486_0 -> 4236 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 4236 },
      -1,
   } },

   /* replace1486_0 -> 760 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 760 },
      -1,
   } },

   /* ('i2i8', ('i2i16', 'a')) => ('i2i8', 'a') */
   /* search1487_0_0 -> 32 in the cache */
   /* search1487_0 -> 4256 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 4256 },
      -1,
   } },

   /* replace1487_0 -> 32 in the cache */
   /* replace1487 -> 4248 in the cache */

   /* ('i2i8', ('u2u16', 'a')) => ('u2u8', 'a') */
   /* search1488_0_0 -> 32 in the cache */
   /* search1488_0 -> 4258 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 4258 },
      -1,
   } },

   /* replace1488_0 -> 32 in the cache */
   /* replace1488 -> 4251 in the cache */

   /* ('u2u8', ('i2i16', 'a')) => ('i2i8', 'a') */
   /* search1489_0_0 -> 32 in the cache */
   /* search1489_0 -> 4256 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 4256 },
      -1,
   } },

   /* replace1489_0 -> 32 in the cache */
   /* replace1489 -> 4248 in the cache */

   /* ('u2u8', ('u2u16', 'a')) => ('u2u8', 'a') */
   /* search1490_0_0 -> 32 in the cache */
   /* search1490_0 -> 4258 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 4258 },
      -1,
   } },

   /* replace1490_0 -> 32 in the cache */
   /* replace1490 -> 4251 in the cache */

   /* ('i2i8', ('i2i32', 'a')) => ('i2i8', 'a') */
   /* search1491_0_0 -> 32 in the cache */
   /* search1491_0 -> 4262 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 4262 },
      -1,
   } },

   /* replace1491_0 -> 32 in the cache */
   /* replace1491 -> 4248 in the cache */

   /* ('i2i8', ('u2u32', 'a')) => ('u2u8', 'a') */
   /* search1492_0_0 -> 32 in the cache */
   /* search1492_0 -> 4264 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 4264 },
      -1,
   } },

   /* replace1492_0 -> 32 in the cache */
   /* replace1492 -> 4251 in the cache */

   /* ('u2u8', ('i2i32', 'a')) => ('i2i8', 'a') */
   /* search1493_0_0 -> 32 in the cache */
   /* search1493_0 -> 4262 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 4262 },
      -1,
   } },

   /* replace1493_0 -> 32 in the cache */
   /* replace1493 -> 4248 in the cache */

   /* ('u2u8', ('u2u32', 'a')) => ('u2u8', 'a') */
   /* search1494_0_0 -> 32 in the cache */
   /* search1494_0 -> 4264 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 4264 },
      -1,
   } },

   /* replace1494_0 -> 32 in the cache */
   /* replace1494 -> 4251 in the cache */

   /* ('i2i8', ('i2i64', 'a')) => ('i2i8', 'a') */
   /* search1495_0_0 -> 32 in the cache */
   /* search1495_0 -> 4268 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 4268 },
      -1,
   } },

   /* replace1495_0 -> 32 in the cache */
   /* replace1495 -> 4248 in the cache */

   /* ('i2i8', ('u2u64', 'a')) => ('u2u8', 'a') */
   /* search1496_0_0 -> 32 in the cache */
   /* search1496_0 -> 3209 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 3209 },
      -1,
   } },

   /* replace1496_0 -> 32 in the cache */
   /* replace1496 -> 4251 in the cache */

   /* ('u2u8', ('i2i64', 'a')) => ('i2i8', 'a') */
   /* search1497_0_0 -> 32 in the cache */
   /* search1497_0 -> 4268 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 4268 },
      -1,
   } },

   /* replace1497_0 -> 32 in the cache */
   /* replace1497 -> 4248 in the cache */

   /* ('u2u8', ('u2u64', 'a')) => ('u2u8', 'a') */
   /* search1498_0_0 -> 32 in the cache */
   /* search1498_0 -> 3209 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 3209 },
      -1,
   } },

   /* replace1498_0 -> 32 in the cache */
   /* replace1498 -> 4251 in the cache */

   /* ('i2i16', ('i2i1', 'a@1')) => ('i2i16', 'a') */
   /* search1499_0_0 -> 760 in the cache */
   /* search1499_0 -> 4231 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 4231 },
      -1,
   } },

   /* replace1499_0 -> 760 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 760 },
      -1,
   } },

   /* ('u2u16', ('u2u1', 'a@1')) => ('u2u16', 'a') */
   /* search1500_0_0 -> 760 in the cache */
   /* search1500_0 -> 4236 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 4236 },
      -1,
   } },

   /* replace1500_0 -> 760 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 760 },
      -1,
   } },

   /* ('i2i16', ('i2i8', 'a@1')) => ('i2i16', 'a') */
   /* search1501_0_0 -> 760 in the cache */
   /* search1501_0 -> 4274 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 4274 },
      -1,
   } },

   /* replace1501_0 -> 760 in the cache */
   /* replace1501 -> 4290 in the cache */

   /* ('u2u16', ('u2u8', 'a@1')) => ('u2u16', 'a') */
   /* search1502_0_0 -> 760 in the cache */
   /* search1502_0 -> 4276 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 4276 },
      -1,
   } },

   /* replace1502_0 -> 760 in the cache */
   /* replace1502 -> 4292 in the cache */

   /* ('i2i16', ('i2i8', 'a@8')) => ('i2i16', 'a') */
   /* search1503_0_0 -> 6 in the cache */
   /* search1503_0 -> 4232 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 4232 },
      -1,
   } },

   /* replace1503_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 6 },
      -1,
   } },

   /* ('u2u16', ('u2u8', 'a@8')) => ('u2u16', 'a') */
   /* search1504_0_0 -> 6 in the cache */
   /* search1504_0 -> 4237 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 4237 },
      -1,
   } },

   /* replace1504_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 6 },
      -1,
   } },

   /* ('i2i16', ('i2i32', 'a')) => ('i2i16', 'a') */
   /* search1505_0_0 -> 32 in the cache */
   /* search1505_0 -> 4262 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 4262 },
      -1,
   } },

   /* replace1505_0 -> 32 in the cache */
   /* replace1505 -> 4256 in the cache */

   /* ('i2i16', ('u2u32', 'a')) => ('u2u16', 'a') */
   /* search1506_0_0 -> 32 in the cache */
   /* search1506_0 -> 4264 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 4264 },
      -1,
   } },

   /* replace1506_0 -> 32 in the cache */
   /* replace1506 -> 4258 in the cache */

   /* ('u2u16', ('i2i32', 'a')) => ('i2i16', 'a') */
   /* search1507_0_0 -> 32 in the cache */
   /* search1507_0 -> 4262 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 4262 },
      -1,
   } },

   /* replace1507_0 -> 32 in the cache */
   /* replace1507 -> 4256 in the cache */

   /* ('u2u16', ('u2u32', 'a')) => ('u2u16', 'a') */
   /* search1508_0_0 -> 32 in the cache */
   /* search1508_0 -> 4264 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 4264 },
      -1,
   } },

   /* replace1508_0 -> 32 in the cache */
   /* replace1508 -> 4258 in the cache */

   /* ('i2i16', ('i2i64', 'a')) => ('i2i16', 'a') */
   /* search1509_0_0 -> 32 in the cache */
   /* search1509_0 -> 4268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 4268 },
      -1,
   } },

   /* replace1509_0 -> 32 in the cache */
   /* replace1509 -> 4256 in the cache */

   /* ('i2i16', ('u2u64', 'a')) => ('u2u16', 'a') */
   /* search1510_0_0 -> 32 in the cache */
   /* search1510_0 -> 3209 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 3209 },
      -1,
   } },

   /* replace1510_0 -> 32 in the cache */
   /* replace1510 -> 4258 in the cache */

   /* ('u2u16', ('i2i64', 'a')) => ('i2i16', 'a') */
   /* search1511_0_0 -> 32 in the cache */
   /* search1511_0 -> 4268 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 4268 },
      -1,
   } },

   /* replace1511_0 -> 32 in the cache */
   /* replace1511 -> 4256 in the cache */

   /* ('u2u16', ('u2u64', 'a')) => ('u2u16', 'a') */
   /* search1512_0_0 -> 32 in the cache */
   /* search1512_0 -> 3209 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 3209 },
      -1,
   } },

   /* replace1512_0 -> 32 in the cache */
   /* replace1512 -> 4258 in the cache */

   /* ('i2i32', ('i2i1', 'a@1')) => ('i2i32', 'a') */
   /* search1513_0_0 -> 760 in the cache */
   /* search1513_0 -> 4231 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 4231 },
      -1,
   } },

   /* replace1513_0 -> 760 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 760 },
      -1,
   } },

   /* ('u2u32', ('u2u1', 'a@1')) => ('u2u32', 'a') */
   /* search1514_0_0 -> 760 in the cache */
   /* search1514_0 -> 4236 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 4236 },
      -1,
   } },

   /* replace1514_0 -> 760 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 760 },
      -1,
   } },

   /* ('i2i32', ('i2i8', 'a@1')) => ('i2i32', 'a') */
   /* search1515_0_0 -> 760 in the cache */
   /* search1515_0 -> 4274 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 4274 },
      -1,
   } },

   /* replace1515_0 -> 760 in the cache */
   /* replace1515 -> 4308 in the cache */

   /* ('u2u32', ('u2u8', 'a@1')) => ('u2u32', 'a') */
   /* search1516_0_0 -> 760 in the cache */
   /* search1516_0 -> 4276 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 4276 },
      -1,
   } },

   /* replace1516_0 -> 760 in the cache */
   /* replace1516 -> 4310 in the cache */

   /* ('i2i32', ('i2i8', 'a@8')) => ('i2i32', 'a') */
   /* search1517_0_0 -> 6 in the cache */
   /* search1517_0 -> 4232 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 4232 },
      -1,
   } },

   /* replace1517_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 6 },
      -1,
   } },

   /* ('u2u32', ('u2u8', 'a@8')) => ('u2u32', 'a') */
   /* search1518_0_0 -> 6 in the cache */
   /* search1518_0 -> 4237 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 4237 },
      -1,
   } },

   /* replace1518_0 -> 6 in the cache */
   /* replace1518 -> 3282 in the cache */

   /* ('i2i32', ('i2i16', 'a@1')) => ('i2i32', 'a') */
   /* search1519_0_0 -> 760 in the cache */
   /* search1519_0 -> 4290 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 4290 },
      -1,
   } },

   /* replace1519_0 -> 760 in the cache */
   /* replace1519 -> 4308 in the cache */

   /* ('u2u32', ('u2u16', 'a@1')) => ('u2u32', 'a') */
   /* search1520_0_0 -> 760 in the cache */
   /* search1520_0 -> 4292 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 4292 },
      -1,
   } },

   /* replace1520_0 -> 760 in the cache */
   /* replace1520 -> 4310 in the cache */

   /* ('i2i32', ('i2i16', 'a@8')) => ('i2i32', 'a') */
   /* search1521_0_0 -> 6 in the cache */
   /* search1521_0 -> 4296 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 4296 },
      -1,
   } },

   /* replace1521_0 -> 6 in the cache */
   /* replace1521 -> 4314 in the cache */

   /* ('u2u32', ('u2u16', 'a@8')) => ('u2u32', 'a') */
   /* search1522_0_0 -> 6 in the cache */
   /* search1522_0 -> 4298 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 4298 },
      -1,
   } },

   /* replace1522_0 -> 6 in the cache */
   /* replace1522 -> 3282 in the cache */

   /* ('i2i32', ('i2i16', 'a@16')) => ('i2i32', 'a') */
   /* search1523_0_0 -> 11 in the cache */
   /* search1523_0 -> 4233 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 4233 },
      -1,
   } },

   /* replace1523_0 -> 11 in the cache */
   /* replace1523 -> 1804 in the cache */

   /* ('u2u32', ('u2u16', 'a@16')) => ('u2u32', 'a') */
   /* search1524_0_0 -> 11 in the cache */
   /* search1524_0 -> 4238 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 4238 },
      -1,
   } },

   /* replace1524_0 -> 11 in the cache */
   /* replace1524 -> 1806 in the cache */

   /* ('i2i32', ('i2i64', 'a')) => ('i2i32', 'a') */
   /* search1525_0_0 -> 32 in the cache */
   /* search1525_0 -> 4268 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 4268 },
      -1,
   } },

   /* replace1525_0 -> 32 in the cache */
   /* replace1525 -> 4262 in the cache */

   /* ('i2i32', ('u2u64', 'a')) => ('u2u32', 'a') */
   /* search1526_0_0 -> 32 in the cache */
   /* search1526_0 -> 3209 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 3209 },
      -1,
   } },

   /* replace1526_0 -> 32 in the cache */
   /* replace1526 -> 4264 in the cache */

   /* ('u2u32', ('i2i64', 'a')) => ('i2i32', 'a') */
   /* search1527_0_0 -> 32 in the cache */
   /* search1527_0 -> 4268 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 4268 },
      -1,
   } },

   /* replace1527_0 -> 32 in the cache */
   /* replace1527 -> 4262 in the cache */

   /* ('u2u32', ('u2u64', 'a')) => ('u2u32', 'a') */
   /* search1528_0_0 -> 32 in the cache */
   /* search1528_0 -> 3209 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 3209 },
      -1,
   } },

   /* replace1528_0 -> 32 in the cache */
   /* replace1528 -> 4264 in the cache */

   /* ('i2i64', ('i2i1', 'a@1')) => ('i2i64', 'a') */
   /* search1529_0_0 -> 760 in the cache */
   /* search1529_0 -> 4231 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 4231 },
      -1,
   } },

   /* replace1529_0 -> 760 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 760 },
      -1,
   } },

   /* ('u2u64', ('u2u1', 'a@1')) => ('u2u64', 'a') */
   /* search1530_0_0 -> 760 in the cache */
   /* search1530_0 -> 4236 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 4236 },
      -1,
   } },

   /* replace1530_0 -> 760 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 760 },
      -1,
   } },

   /* ('i2i64', ('i2i8', 'a@1')) => ('i2i64', 'a') */
   /* search1531_0_0 -> 760 in the cache */
   /* search1531_0 -> 4274 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 4274 },
      -1,
   } },

   /* replace1531_0 -> 760 in the cache */
   /* replace1531 -> 4327 in the cache */

   /* ('u2u64', ('u2u8', 'a@1')) => ('u2u64', 'a') */
   /* search1532_0_0 -> 760 in the cache */
   /* search1532_0 -> 4276 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 4276 },
      -1,
   } },

   /* replace1532_0 -> 760 in the cache */
   /* replace1532 -> 4329 in the cache */

   /* ('i2i64', ('i2i8', 'a@8')) => ('i2i64', 'a') */
   /* search1533_0_0 -> 6 in the cache */
   /* search1533_0 -> 4232 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 4232 },
      -1,
   } },

   /* replace1533_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 6 },
      -1,
   } },

   /* ('u2u64', ('u2u8', 'a@8')) => ('u2u64', 'a') */
   /* search1534_0_0 -> 6 in the cache */
   /* search1534_0 -> 4237 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 4237 },
      -1,
   } },

   /* replace1534_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 6 },
      -1,
   } },

   /* ('i2i64', ('i2i16', 'a@1')) => ('i2i64', 'a') */
   /* search1535_0_0 -> 760 in the cache */
   /* search1535_0 -> 4290 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 4290 },
      -1,
   } },

   /* replace1535_0 -> 760 in the cache */
   /* replace1535 -> 4327 in the cache */

   /* ('u2u64', ('u2u16', 'a@1')) => ('u2u64', 'a') */
   /* search1536_0_0 -> 760 in the cache */
   /* search1536_0 -> 4292 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 4292 },
      -1,
   } },

   /* replace1536_0 -> 760 in the cache */
   /* replace1536 -> 4329 in the cache */

   /* ('i2i64', ('i2i16', 'a@8')) => ('i2i64', 'a') */
   /* search1537_0_0 -> 6 in the cache */
   /* search1537_0 -> 4296 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 4296 },
      -1,
   } },

   /* replace1537_0 -> 6 in the cache */
   /* replace1537 -> 4333 in the cache */

   /* ('u2u64', ('u2u16', 'a@8')) => ('u2u64', 'a') */
   /* search1538_0_0 -> 6 in the cache */
   /* search1538_0 -> 4298 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 4298 },
      -1,
   } },

   /* replace1538_0 -> 6 in the cache */
   /* replace1538 -> 4335 in the cache */

   /* ('i2i64', ('i2i16', 'a@16')) => ('i2i64', 'a') */
   /* search1539_0_0 -> 11 in the cache */
   /* search1539_0 -> 4233 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 4233 },
      -1,
   } },

   /* replace1539_0 -> 11 in the cache */
   /* replace1539 -> 1827 in the cache */

   /* ('u2u64', ('u2u16', 'a@16')) => ('u2u64', 'a') */
   /* search1540_0_0 -> 11 in the cache */
   /* search1540_0 -> 4238 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 4238 },
      -1,
   } },

   /* replace1540_0 -> 11 in the cache */
   /* replace1540 -> 1829 in the cache */

   /* ('i2i64', ('i2i32', 'a@1')) => ('i2i64', 'a') */
   /* search1541_0_0 -> 760 in the cache */
   /* search1541_0 -> 4308 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 4308 },
      -1,
   } },

   /* replace1541_0 -> 760 in the cache */
   /* replace1541 -> 4327 in the cache */

   /* ('u2u64', ('u2u32', 'a@1')) => ('u2u64', 'a') */
   /* search1542_0_0 -> 760 in the cache */
   /* search1542_0 -> 4310 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 4310 },
      -1,
   } },

   /* replace1542_0 -> 760 in the cache */
   /* replace1542 -> 4329 in the cache */

   /* ('i2i64', ('i2i32', 'a@8')) => ('i2i64', 'a') */
   /* search1543_0_0 -> 6 in the cache */
   /* search1543_0 -> 4314 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 4314 },
      -1,
   } },

   /* replace1543_0 -> 6 in the cache */
   /* replace1543 -> 4333 in the cache */

   /* ('u2u64', ('u2u32', 'a@8')) => ('u2u64', 'a') */
   /* search1544_0_0 -> 6 in the cache */
   /* search1544_0 -> 3282 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 3282 },
      -1,
   } },

   /* replace1544_0 -> 6 in the cache */
   /* replace1544 -> 4335 in the cache */

   /* ('i2i64', ('i2i32', 'a@16')) => ('i2i64', 'a') */
   /* search1545_0_0 -> 11 in the cache */
   /* search1545_0 -> 1804 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 1804 },
      -1,
   } },

   /* replace1545_0 -> 11 in the cache */
   /* replace1545 -> 1827 in the cache */

   /* ('u2u64', ('u2u32', 'a@16')) => ('u2u64', 'a') */
   /* search1546_0_0 -> 11 in the cache */
   /* search1546_0 -> 1806 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 1806 },
      -1,
   } },

   /* replace1546_0 -> 11 in the cache */
   /* replace1546 -> 1829 in the cache */

   /* ('i2i64', ('i2i32', 'a@32')) => ('i2i64', 'a') */
   /* search1547_0_0 -> 16 in the cache */
   /* search1547_0 -> 4234 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 4234 },
      -1,
   } },

   /* replace1547_0 -> 16 in the cache */
   /* replace1547 -> 1909 in the cache */

   /* ('u2u64', ('u2u32', 'a@32')) => ('u2u64', 'a') */
   /* search1548_0_0 -> 16 in the cache */
   /* search1548_0 -> 4239 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 4239 },
      -1,
   } },

   /* replace1548_0 -> 16 in the cache */
   /* replace1548 -> 1911 in the cache */

   /* ('i2i8', ('pack_64_2x32_split', 'a', 'b')) => ('i2i8', 'a') */
   /* search1549_0_0 -> 16 in the cache */
   /* search1549_0_1 -> 36 in the cache */
   /* search1549_0 -> 3202 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 3202 },
      -1,
   } },

   /* replace1549_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('i2i8', ('pack_64_2x32_split', 'a', 'b')) => ('i2i8', 'a') */
   /* search1550_0_0 -> 16 in the cache */
   /* search1550_0_1 -> 36 in the cache */
   /* search1550_0 -> 3202 in the cache */
   /* search1550 -> 4350 in the cache */

   /* replace1550_0 -> 16 in the cache */
   /* replace1550 -> 4351 in the cache */

   /* ('i2i16', ('pack_64_2x32_split', 'a', 'b')) => ('i2i16', 'a') */
   /* search1551_0_0 -> 16 in the cache */
   /* search1551_0_1 -> 36 in the cache */
   /* search1551_0 -> 3202 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 3202 },
      -1,
   } },

   /* replace1551_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('i2i16', ('pack_64_2x32_split', 'a', 'b')) => ('i2i16', 'a') */
   /* search1552_0_0 -> 16 in the cache */
   /* search1552_0_1 -> 36 in the cache */
   /* search1552_0 -> 3202 in the cache */
   /* search1552 -> 4352 in the cache */

   /* replace1552_0 -> 16 in the cache */
   /* replace1552 -> 4353 in the cache */

   /* ('i2i32', ('pack_64_2x32_split', 'a', 'b')) => ('i2i32', 'a') */
   /* search1553_0_0 -> 16 in the cache */
   /* search1553_0_1 -> 36 in the cache */
   /* search1553_0 -> 3202 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 3202 },
      -1,
   } },

   /* replace1553_0 -> 16 in the cache */
   /* replace1553 -> 4234 in the cache */

   /* ('i2i32', ('pack_64_2x32_split', 'a', 'b')) => ('i2i32', 'a') */
   /* search1554_0_0 -> 16 in the cache */
   /* search1554_0_1 -> 36 in the cache */
   /* search1554_0 -> 3202 in the cache */
   /* search1554 -> 4354 in the cache */

   /* replace1554_0 -> 16 in the cache */
   /* replace1554 -> 4234 in the cache */

   /* ('u2u8', ('pack_64_2x32_split', 'a', 'b')) => ('u2u8', 'a') */
   /* search1555_0_0 -> 16 in the cache */
   /* search1555_0_1 -> 36 in the cache */
   /* search1555_0 -> 3202 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 3202 },
      -1,
   } },

   /* replace1555_0 -> 16 in the cache */
   /* replace1555 -> 2025 in the cache */

   /* ('u2u8', ('pack_64_2x32_split', 'a', 'b')) => ('u2u8', 'a') */
   /* search1556_0_0 -> 16 in the cache */
   /* search1556_0_1 -> 36 in the cache */
   /* search1556_0 -> 3202 in the cache */
   /* search1556 -> 4355 in the cache */

   /* replace1556_0 -> 16 in the cache */
   /* replace1556 -> 2025 in the cache */

   /* ('u2u16', ('pack_64_2x32_split', 'a', 'b')) => ('u2u16', 'a') */
   /* search1557_0_0 -> 16 in the cache */
   /* search1557_0_1 -> 36 in the cache */
   /* search1557_0 -> 3202 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 3202 },
      -1,
   } },

   /* replace1557_0 -> 16 in the cache */
   /* replace1557 -> 3310 in the cache */

   /* ('u2u16', ('pack_64_2x32_split', 'a', 'b')) => ('u2u16', 'a') */
   /* search1558_0_0 -> 16 in the cache */
   /* search1558_0_1 -> 36 in the cache */
   /* search1558_0 -> 3202 in the cache */
   /* search1558 -> 4356 in the cache */

   /* replace1558_0 -> 16 in the cache */
   /* replace1558 -> 3310 in the cache */

   /* ('u2u32', ('pack_64_2x32_split', 'a', 'b')) => ('u2u32', 'a') */
   /* search1559_0_0 -> 16 in the cache */
   /* search1559_0_1 -> 36 in the cache */
   /* search1559_0 -> 3202 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 3202 },
      -1,
   } },

   /* replace1559_0 -> 16 in the cache */
   /* replace1559 -> 4239 in the cache */

   /* ('u2u32', ('pack_64_2x32_split', 'a', 'b')) => ('u2u32', 'a') */
   /* search1560_0_0 -> 16 in the cache */
   /* search1560_0_1 -> 36 in the cache */
   /* search1560_0 -> 3202 in the cache */
   /* search1560 -> 4357 in the cache */

   /* replace1560_0 -> 16 in the cache */
   /* replace1560 -> 4239 in the cache */

   /* ('ieq', ('i2i16', 'a@8'), ('i2i16', 'b@8')) => ('ieq', 'a', ('i2i8', 'b')) */
   /* search1561_0_0 -> 6 in the cache */
   /* search1561_0 -> 4296 in the cache */
   /* search1561_1_0 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 2112 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 4296, 4358 },
      -1,
   } },

   /* replace1561_0 -> 6 in the cache */
   /* replace1561_1_0 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 2112 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 6, 4360 },
      -1,
   } },

   /* ('ine', ('i2i16', 'a@8'), ('i2i16', 'b@8')) => ('ine', 'a', ('i2i8', 'b')) */
   /* search1562_0_0 -> 6 in the cache */
   /* search1562_0 -> 4296 in the cache */
   /* search1562_1_0 -> 2112 in the cache */
   /* search1562_1 -> 4358 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 4296, 4358 },
      -1,
   } },

   /* replace1562_0 -> 6 in the cache */
   /* replace1562_1_0 -> 2112 in the cache */
   /* replace1562_1 -> 4360 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 6, 4360 },
      -1,
   } },

   /* ('ige', ('i2i16', 'a@8'), ('i2i16', 'b@8')) => ('ige', 'a', ('i2i8', 'b')) */
   /* search1563_0_0 -> 6 in the cache */
   /* search1563_0 -> 4296 in the cache */
   /* search1563_1_0 -> 2112 in the cache */
   /* search1563_1 -> 4358 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4296, 4358 },
      -1,
   } },

   /* replace1563_0 -> 6 in the cache */
   /* replace1563_1_0 -> 2112 in the cache */
   /* replace1563_1 -> 4360 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 6, 4360 },
      -1,
   } },

   /* ('ilt', ('i2i16', 'a@8'), ('i2i16', 'b@8')) => ('ilt', 'a', ('i2i8', 'b')) */
   /* search1564_0_0 -> 6 in the cache */
   /* search1564_0 -> 4296 in the cache */
   /* search1564_1_0 -> 2112 in the cache */
   /* search1564_1 -> 4358 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4296, 4358 },
      -1,
   } },

   /* replace1564_0 -> 6 in the cache */
   /* replace1564_1_0 -> 2112 in the cache */
   /* replace1564_1 -> 4360 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 6, 4360 },
      -1,
   } },

   /* ('ige', ('i2i16', 'b@8'), ('i2i16', 'a@8')) => ('ige', ('i2i8', 'b'), 'a') */
   { .variable = {
      { nir_search_value_variable, 8 },
      0, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 4368 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 8 },
      1, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 4370 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4369, 4371 },
      -1,
   } },

   /* replace1565_0_0 -> 4368 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 4368 },
      -1,
   } },
   /* replace1565_1 -> 4370 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4373, 4370 },
      -1,
   } },

   /* ('ilt', ('i2i16', 'b@8'), ('i2i16', 'a@8')) => ('ilt', ('i2i8', 'b'), 'a') */
   /* search1566_0_0 -> 4368 in the cache */
   /* search1566_0 -> 4369 in the cache */
   /* search1566_1_0 -> 4370 in the cache */
   /* search1566_1 -> 4371 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4369, 4371 },
      -1,
   } },

   /* replace1566_0_0 -> 4368 in the cache */
   /* replace1566_0 -> 4373 in the cache */
   /* replace1566_1 -> 4370 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4373, 4370 },
      -1,
   } },

   /* ('ieq', ('i2i16', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('i2i8', 'b')), ('ieq', ('i2i16', ('i2i8', 'b')), 'b')) */
   /* search1567_0_0 -> 6 in the cache */
   /* search1567_0 -> 4296 in the cache */
   /* search1567_1 -> 2011 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 4296, 2011 },
      -1,
   } },

   /* replace1567_0_0 -> 6 in the cache */
   /* replace1567_0_1_0 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 258 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 6, 4378 },
      -1,
   } },
   /* replace1567_1_0_0_0 -> 258 in the cache */
   /* replace1567_1_0_0 -> 4378 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 4378 },
      -1,
   } },
   /* replace1567_1_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 4380, 258 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4379, 4381 },
      -1,
   } },

   /* ('ine', ('i2i16', 'a@8'), '#b') => ('ior', ('ine', 'a', ('i2i8', 'b')), ('ine', ('i2i16', ('i2i8', 'b')), 'b')) */
   /* search1568_0_0 -> 6 in the cache */
   /* search1568_0 -> 4296 in the cache */
   /* search1568_1 -> 2011 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 4296, 2011 },
      -1,
   } },

   /* replace1568_0_0 -> 6 in the cache */
   /* replace1568_0_1_0 -> 258 in the cache */
   /* replace1568_0_1 -> 4378 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 6, 4378 },
      -1,
   } },
   /* replace1568_1_0_0_0 -> 258 in the cache */
   /* replace1568_1_0_0 -> 4378 in the cache */
   /* replace1568_1_0 -> 4380 in the cache */
   /* replace1568_1_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 4380, 258 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4384, 4385 },
      -1,
   } },

   /* ('ilt', ('i2i16', 'a@8'), '#b') => ('iand', ('ilt', -128, 'b'), ('ior', ('ilt', 127, 'b'), ('ilt', 'a', ('i2i8', 'b')))) */
   /* search1569_0_0 -> 6 in the cache */
   /* search1569_0 -> 4296 in the cache */
   /* search1569_1 -> 2011 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4296, 2011 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { -0x80 /* -128 */ },
   } },
   /* replace1569_0_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4388, 258 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { 0x7f /* 127 */ },
   } },
   /* replace1569_1_0_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4390, 258 },
      -1,
   } },
   /* replace1569_1_1_0 -> 6 in the cache */
   /* replace1569_1_1_1_0 -> 258 in the cache */
   /* replace1569_1_1_1 -> 4378 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 6, 4378 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4391, 4392 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4389, 4393 },
      -1,
   } },

   /* ('ilt', '#a', ('i2i16', 'b@8')) => ('iand', ('ilt', 'a', 127), ('ior', ('ilt', 'a', -128), ('ilt', ('i2i8', 'a'), 'b'))) */
   { .variable = {
      { nir_search_value_variable, 16 },
      0, /* a */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1570_1_0 -> 2112 in the cache */
   /* search1570_1 -> 4358 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4395, 4358 },
      -1,
   } },

   /* replace1570_0_0 -> 11 in the cache */
   /* replace1570_0_1 -> 4390 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 11, 4390 },
      -1,
   } },
   /* replace1570_1_0_0 -> 11 in the cache */
   /* replace1570_1_0_1 -> 4388 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 11, 4388 },
      -1,
   } },
   /* replace1570_1_1_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 11 },
      -1,
   } },
   /* replace1570_1_1_1 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4399, 2112 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4398, 4400 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4397, 4401 },
      -1,
   } },

   /* ('ige', ('i2i16', 'a@8'), '#b') => ('iand', ('ige', 127, 'b'), ('ior', ('ige', -128, 'b'), ('ige', 'a', ('i2i8', 'b')))) */
   /* search1571_0_0 -> 6 in the cache */
   /* search1571_0 -> 4296 in the cache */
   /* search1571_1 -> 2011 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4296, 2011 },
      -1,
   } },

   /* replace1571_0_0 -> 4390 in the cache */
   /* replace1571_0_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4390, 258 },
      -1,
   } },
   /* replace1571_1_0_0 -> 4388 in the cache */
   /* replace1571_1_0_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4388, 258 },
      -1,
   } },
   /* replace1571_1_1_0 -> 6 in the cache */
   /* replace1571_1_1_1_0 -> 258 in the cache */
   /* replace1571_1_1_1 -> 4378 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 6, 4378 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4405, 4406 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4404, 4407 },
      -1,
   } },

   /* ('ige', '#a', ('i2i16', 'b@8')) => ('iand', ('ige', 'a', -128), ('ior', ('ige', 'a', 127), ('ige', ('i2i8', 'a'), 'b'))) */
   /* search1572_0 -> 4395 in the cache */
   /* search1572_1_0 -> 2112 in the cache */
   /* search1572_1 -> 4358 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4395, 4358 },
      -1,
   } },

   /* replace1572_0_0 -> 11 in the cache */
   /* replace1572_0_1 -> 4388 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 11, 4388 },
      -1,
   } },
   /* replace1572_1_0_0 -> 11 in the cache */
   /* replace1572_1_0_1 -> 4390 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 11, 4390 },
      -1,
   } },
   /* replace1572_1_1_0_0 -> 11 in the cache */
   /* replace1572_1_1_0 -> 4399 in the cache */
   /* replace1572_1_1_1 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4399, 2112 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4411, 4412 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4410, 4413 },
      -1,
   } },

   /* ('ieq', ('i2i32', 'a@8'), ('i2i32', 'b@8')) => ('ieq', 'a', ('i2i8', 'b')) */
   /* search1573_0_0 -> 6 in the cache */
   /* search1573_0 -> 4314 in the cache */
   /* search1573_1_0 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 2112 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 4314, 4415 },
      -1,
   } },

   /* replace1573_0 -> 6 in the cache */
   /* replace1573_1_0 -> 2112 in the cache */
   /* replace1573_1 -> 4360 in the cache */
   /* replace1573 -> 4361 in the cache */

   /* ('ine', ('i2i32', 'a@8'), ('i2i32', 'b@8')) => ('ine', 'a', ('i2i8', 'b')) */
   /* search1574_0_0 -> 6 in the cache */
   /* search1574_0 -> 4314 in the cache */
   /* search1574_1_0 -> 2112 in the cache */
   /* search1574_1 -> 4415 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 4314, 4415 },
      -1,
   } },

   /* replace1574_0 -> 6 in the cache */
   /* replace1574_1_0 -> 2112 in the cache */
   /* replace1574_1 -> 4360 in the cache */
   /* replace1574 -> 4363 in the cache */

   /* ('ige', ('i2i32', 'a@8'), ('i2i32', 'b@8')) => ('ige', 'a', ('i2i8', 'b')) */
   /* search1575_0_0 -> 6 in the cache */
   /* search1575_0 -> 4314 in the cache */
   /* search1575_1_0 -> 2112 in the cache */
   /* search1575_1 -> 4415 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4314, 4415 },
      -1,
   } },

   /* replace1575_0 -> 6 in the cache */
   /* replace1575_1_0 -> 2112 in the cache */
   /* replace1575_1 -> 4360 in the cache */
   /* replace1575 -> 4365 in the cache */

   /* ('ilt', ('i2i32', 'a@8'), ('i2i32', 'b@8')) => ('ilt', 'a', ('i2i8', 'b')) */
   /* search1576_0_0 -> 6 in the cache */
   /* search1576_0 -> 4314 in the cache */
   /* search1576_1_0 -> 2112 in the cache */
   /* search1576_1 -> 4415 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4314, 4415 },
      -1,
   } },

   /* replace1576_0 -> 6 in the cache */
   /* replace1576_1_0 -> 2112 in the cache */
   /* replace1576_1 -> 4360 in the cache */
   /* replace1576 -> 4367 in the cache */

   /* ('ige', ('i2i32', 'b@8'), ('i2i32', 'a@8')) => ('ige', ('i2i8', 'b'), 'a') */
   /* search1577_0_0 -> 4368 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 4368 },
      -1,
   } },
   /* search1577_1_0 -> 4370 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 4370 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4420, 4421 },
      -1,
   } },

   /* replace1577_0_0 -> 4368 in the cache */
   /* replace1577_0 -> 4373 in the cache */
   /* replace1577_1 -> 4370 in the cache */
   /* replace1577 -> 4374 in the cache */

   /* ('ilt', ('i2i32', 'b@8'), ('i2i32', 'a@8')) => ('ilt', ('i2i8', 'b'), 'a') */
   /* search1578_0_0 -> 4368 in the cache */
   /* search1578_0 -> 4420 in the cache */
   /* search1578_1_0 -> 4370 in the cache */
   /* search1578_1 -> 4421 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4420, 4421 },
      -1,
   } },

   /* replace1578_0_0 -> 4368 in the cache */
   /* replace1578_0 -> 4373 in the cache */
   /* replace1578_1 -> 4370 in the cache */
   /* replace1578 -> 4376 in the cache */

   /* ('ieq', ('i2i32', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('i2i8', 'b')), ('ieq', ('i2i32', ('i2i8', 'b')), 'b')) */
   /* search1579_0_0 -> 6 in the cache */
   /* search1579_0 -> 4314 in the cache */
   /* search1579_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 4314, 33 },
      -1,
   } },

   /* replace1579_0_0 -> 6 in the cache */
   /* replace1579_0_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 6, 4425 },
      -1,
   } },
   /* replace1579_1_0_0_0 -> 36 in the cache */
   /* replace1579_1_0_0 -> 4425 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 4425 },
      -1,
   } },
   /* replace1579_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 4427, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4426, 4428 },
      -1,
   } },

   /* ('ine', ('i2i32', 'a@8'), '#b') => ('ior', ('ine', 'a', ('i2i8', 'b')), ('ine', ('i2i32', ('i2i8', 'b')), 'b')) */
   /* search1580_0_0 -> 6 in the cache */
   /* search1580_0 -> 4314 in the cache */
   /* search1580_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 4314, 33 },
      -1,
   } },

   /* replace1580_0_0 -> 6 in the cache */
   /* replace1580_0_1_0 -> 36 in the cache */
   /* replace1580_0_1 -> 4425 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 6, 4425 },
      -1,
   } },
   /* replace1580_1_0_0_0 -> 36 in the cache */
   /* replace1580_1_0_0 -> 4425 in the cache */
   /* replace1580_1_0 -> 4427 in the cache */
   /* replace1580_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 4427, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4431, 4432 },
      -1,
   } },

   /* ('ilt', ('i2i32', 'a@8'), '#b') => ('iand', ('ilt', -128, 'b'), ('ior', ('ilt', 127, 'b'), ('ilt', 'a', ('i2i8', 'b')))) */
   /* search1581_0_0 -> 6 in the cache */
   /* search1581_0 -> 4314 in the cache */
   /* search1581_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4314, 33 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x80 /* -128 */ },
   } },
   /* replace1581_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4435, 36 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x7f /* 127 */ },
   } },
   /* replace1581_1_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4437, 36 },
      -1,
   } },
   /* replace1581_1_1_0 -> 6 in the cache */
   /* replace1581_1_1_1_0 -> 36 in the cache */
   /* replace1581_1_1_1 -> 4425 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 6, 4425 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4438, 4439 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4436, 4440 },
      -1,
   } },

   /* ('ilt', '#a', ('i2i32', 'b@8')) => ('iand', ('ilt', 'a', 127), ('ior', ('ilt', 'a', -128), ('ilt', ('i2i8', 'a'), 'b'))) */
   /* search1582_0 -> 286 in the cache */
   /* search1582_1_0 -> 2112 in the cache */
   /* search1582_1 -> 4415 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 286, 4415 },
      -1,
   } },

   /* replace1582_0_0 -> 16 in the cache */
   /* replace1582_0_1 -> 4437 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 16, 4437 },
      -1,
   } },
   /* replace1582_1_0_0 -> 16 in the cache */
   /* replace1582_1_0_1 -> 4435 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 16, 4435 },
      -1,
   } },
   /* replace1582_1_1_0_0 -> 16 in the cache */
   /* replace1582_1_1_0 -> 4351 in the cache */
   /* replace1582_1_1_1 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4351, 2112 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4444, 4445 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4443, 4446 },
      -1,
   } },

   /* ('ige', ('i2i32', 'a@8'), '#b') => ('iand', ('ige', 127, 'b'), ('ior', ('ige', -128, 'b'), ('ige', 'a', ('i2i8', 'b')))) */
   /* search1583_0_0 -> 6 in the cache */
   /* search1583_0 -> 4314 in the cache */
   /* search1583_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4314, 33 },
      -1,
   } },

   /* replace1583_0_0 -> 4437 in the cache */
   /* replace1583_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4437, 36 },
      -1,
   } },
   /* replace1583_1_0_0 -> 4435 in the cache */
   /* replace1583_1_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4435, 36 },
      -1,
   } },
   /* replace1583_1_1_0 -> 6 in the cache */
   /* replace1583_1_1_1_0 -> 36 in the cache */
   /* replace1583_1_1_1 -> 4425 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 6, 4425 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4450, 4451 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4449, 4452 },
      -1,
   } },

   /* ('ige', '#a', ('i2i32', 'b@8')) => ('iand', ('ige', 'a', -128), ('ior', ('ige', 'a', 127), ('ige', ('i2i8', 'a'), 'b'))) */
   /* search1584_0 -> 286 in the cache */
   /* search1584_1_0 -> 2112 in the cache */
   /* search1584_1 -> 4415 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 286, 4415 },
      -1,
   } },

   /* replace1584_0_0 -> 16 in the cache */
   /* replace1584_0_1 -> 4435 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 16, 4435 },
      -1,
   } },
   /* replace1584_1_0_0 -> 16 in the cache */
   /* replace1584_1_0_1 -> 4437 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 16, 4437 },
      -1,
   } },
   /* replace1584_1_1_0_0 -> 16 in the cache */
   /* replace1584_1_1_0 -> 4351 in the cache */
   /* replace1584_1_1_1 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4351, 2112 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4456, 4457 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4455, 4458 },
      -1,
   } },

   /* ('ieq', ('i2i64', 'a@8'), ('i2i64', 'b@8')) => ('ieq', 'a', ('i2i8', 'b')) */
   /* search1585_0_0 -> 6 in the cache */
   /* search1585_0 -> 4333 in the cache */
   /* search1585_1_0 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 2112 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 4333, 4460 },
      -1,
   } },

   /* replace1585_0 -> 6 in the cache */
   /* replace1585_1_0 -> 2112 in the cache */
   /* replace1585_1 -> 4360 in the cache */
   /* replace1585 -> 4361 in the cache */

   /* ('ine', ('i2i64', 'a@8'), ('i2i64', 'b@8')) => ('ine', 'a', ('i2i8', 'b')) */
   /* search1586_0_0 -> 6 in the cache */
   /* search1586_0 -> 4333 in the cache */
   /* search1586_1_0 -> 2112 in the cache */
   /* search1586_1 -> 4460 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 4333, 4460 },
      -1,
   } },

   /* replace1586_0 -> 6 in the cache */
   /* replace1586_1_0 -> 2112 in the cache */
   /* replace1586_1 -> 4360 in the cache */
   /* replace1586 -> 4363 in the cache */

   /* ('ige', ('i2i64', 'a@8'), ('i2i64', 'b@8')) => ('ige', 'a', ('i2i8', 'b')) */
   /* search1587_0_0 -> 6 in the cache */
   /* search1587_0 -> 4333 in the cache */
   /* search1587_1_0 -> 2112 in the cache */
   /* search1587_1 -> 4460 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4333, 4460 },
      -1,
   } },

   /* replace1587_0 -> 6 in the cache */
   /* replace1587_1_0 -> 2112 in the cache */
   /* replace1587_1 -> 4360 in the cache */
   /* replace1587 -> 4365 in the cache */

   /* ('ilt', ('i2i64', 'a@8'), ('i2i64', 'b@8')) => ('ilt', 'a', ('i2i8', 'b')) */
   /* search1588_0_0 -> 6 in the cache */
   /* search1588_0 -> 4333 in the cache */
   /* search1588_1_0 -> 2112 in the cache */
   /* search1588_1 -> 4460 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4333, 4460 },
      -1,
   } },

   /* replace1588_0 -> 6 in the cache */
   /* replace1588_1_0 -> 2112 in the cache */
   /* replace1588_1 -> 4360 in the cache */
   /* replace1588 -> 4367 in the cache */

   /* ('ige', ('i2i64', 'b@8'), ('i2i64', 'a@8')) => ('ige', ('i2i8', 'b'), 'a') */
   /* search1589_0_0 -> 4368 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 4368 },
      -1,
   } },
   /* search1589_1_0 -> 4370 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 4370 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4465, 4466 },
      -1,
   } },

   /* replace1589_0_0 -> 4368 in the cache */
   /* replace1589_0 -> 4373 in the cache */
   /* replace1589_1 -> 4370 in the cache */
   /* replace1589 -> 4374 in the cache */

   /* ('ilt', ('i2i64', 'b@8'), ('i2i64', 'a@8')) => ('ilt', ('i2i8', 'b'), 'a') */
   /* search1590_0_0 -> 4368 in the cache */
   /* search1590_0 -> 4465 in the cache */
   /* search1590_1_0 -> 4370 in the cache */
   /* search1590_1 -> 4466 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4465, 4466 },
      -1,
   } },

   /* replace1590_0_0 -> 4368 in the cache */
   /* replace1590_0 -> 4373 in the cache */
   /* replace1590_1 -> 4370 in the cache */
   /* replace1590 -> 4376 in the cache */

   /* ('ieq', ('i2i64', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('i2i8', 'b')), ('ieq', ('i2i64', ('i2i8', 'b')), 'b')) */
   /* search1591_0_0 -> 6 in the cache */
   /* search1591_0 -> 4333 in the cache */
   { .variable = {
      { nir_search_value_variable, 64 },
      1, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 4333, 4469 },
      -1,
   } },

   /* replace1591_0_0 -> 6 in the cache */
   /* replace1591_0_1_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 6, 4471 },
      -1,
   } },
   /* replace1591_1_0_0_0 -> 41 in the cache */
   /* replace1591_1_0_0 -> 4471 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 4471 },
      -1,
   } },
   /* replace1591_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 4473, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4472, 4474 },
      -1,
   } },

   /* ('ine', ('i2i64', 'a@8'), '#b') => ('ior', ('ine', 'a', ('i2i8', 'b')), ('ine', ('i2i64', ('i2i8', 'b')), 'b')) */
   /* search1592_0_0 -> 6 in the cache */
   /* search1592_0 -> 4333 in the cache */
   /* search1592_1 -> 4469 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 4333, 4469 },
      -1,
   } },

   /* replace1592_0_0 -> 6 in the cache */
   /* replace1592_0_1_0 -> 41 in the cache */
   /* replace1592_0_1 -> 4471 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 6, 4471 },
      -1,
   } },
   /* replace1592_1_0_0_0 -> 41 in the cache */
   /* replace1592_1_0_0 -> 4471 in the cache */
   /* replace1592_1_0 -> 4473 in the cache */
   /* replace1592_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 4473, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4477, 4478 },
      -1,
   } },

   /* ('ilt', ('i2i64', 'a@8'), '#b') => ('iand', ('ilt', -128, 'b'), ('ior', ('ilt', 127, 'b'), ('ilt', 'a', ('i2i8', 'b')))) */
   /* search1593_0_0 -> 6 in the cache */
   /* search1593_0 -> 4333 in the cache */
   /* search1593_1 -> 4469 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4333, 4469 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { -0x80 /* -128 */ },
   } },
   /* replace1593_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4481, 41 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x7f /* 127 */ },
   } },
   /* replace1593_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4483, 41 },
      -1,
   } },
   /* replace1593_1_1_0 -> 6 in the cache */
   /* replace1593_1_1_1_0 -> 41 in the cache */
   /* replace1593_1_1_1 -> 4471 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 6, 4471 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4484, 4485 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4482, 4486 },
      -1,
   } },

   /* ('ilt', '#a', ('i2i64', 'b@8')) => ('iand', ('ilt', 'a', 127), ('ior', ('ilt', 'a', -128), ('ilt', ('i2i8', 'a'), 'b'))) */
   { .variable = {
      { nir_search_value_variable, 64 },
      0, /* a */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1594_1_0 -> 2112 in the cache */
   /* search1594_1 -> 4460 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4488, 4460 },
      -1,
   } },

   /* replace1594_0_0 -> 21 in the cache */
   /* replace1594_0_1 -> 4483 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 21, 4483 },
      -1,
   } },
   /* replace1594_1_0_0 -> 21 in the cache */
   /* replace1594_1_0_1 -> 4481 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 21, 4481 },
      -1,
   } },
   /* replace1594_1_1_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 21 },
      -1,
   } },
   /* replace1594_1_1_1 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4492, 2112 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4491, 4493 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4490, 4494 },
      -1,
   } },

   /* ('ige', ('i2i64', 'a@8'), '#b') => ('iand', ('ige', 127, 'b'), ('ior', ('ige', -128, 'b'), ('ige', 'a', ('i2i8', 'b')))) */
   /* search1595_0_0 -> 6 in the cache */
   /* search1595_0 -> 4333 in the cache */
   /* search1595_1 -> 4469 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4333, 4469 },
      -1,
   } },

   /* replace1595_0_0 -> 4483 in the cache */
   /* replace1595_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4483, 41 },
      -1,
   } },
   /* replace1595_1_0_0 -> 4481 in the cache */
   /* replace1595_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4481, 41 },
      -1,
   } },
   /* replace1595_1_1_0 -> 6 in the cache */
   /* replace1595_1_1_1_0 -> 41 in the cache */
   /* replace1595_1_1_1 -> 4471 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 6, 4471 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4498, 4499 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4497, 4500 },
      -1,
   } },

   /* ('ige', '#a', ('i2i64', 'b@8')) => ('iand', ('ige', 'a', -128), ('ior', ('ige', 'a', 127), ('ige', ('i2i8', 'a'), 'b'))) */
   /* search1596_0 -> 4488 in the cache */
   /* search1596_1_0 -> 2112 in the cache */
   /* search1596_1 -> 4460 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4488, 4460 },
      -1,
   } },

   /* replace1596_0_0 -> 21 in the cache */
   /* replace1596_0_1 -> 4481 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 21, 4481 },
      -1,
   } },
   /* replace1596_1_0_0 -> 21 in the cache */
   /* replace1596_1_0_1 -> 4483 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 21, 4483 },
      -1,
   } },
   /* replace1596_1_1_0_0 -> 21 in the cache */
   /* replace1596_1_1_0 -> 4492 in the cache */
   /* replace1596_1_1_1 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4492, 2112 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4504, 4505 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4503, 4506 },
      -1,
   } },

   /* ('ieq', ('i2i32', 'a@16'), ('i2i32', 'b@8')) => ('ieq', 'a', ('i2i16', 'b')) */
   /* search1597_0_0 -> 11 in the cache */
   /* search1597_0 -> 1804 in the cache */
   /* search1597_1_0 -> 2112 in the cache */
   /* search1597_1 -> 4415 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1804, 4415 },
      -1,
   } },

   /* replace1597_0 -> 11 in the cache */
   /* replace1597_1_0 -> 2112 in the cache */
   /* replace1597_1 -> 4358 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 11, 4358 },
      -1,
   } },

   /* ('ine', ('i2i32', 'a@16'), ('i2i32', 'b@8')) => ('ine', 'a', ('i2i16', 'b')) */
   /* search1598_0_0 -> 11 in the cache */
   /* search1598_0 -> 1804 in the cache */
   /* search1598_1_0 -> 2112 in the cache */
   /* search1598_1 -> 4415 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1804, 4415 },
      -1,
   } },

   /* replace1598_0 -> 11 in the cache */
   /* replace1598_1_0 -> 2112 in the cache */
   /* replace1598_1 -> 4358 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 11, 4358 },
      -1,
   } },

   /* ('ige', ('i2i32', 'a@16'), ('i2i32', 'b@8')) => ('ige', 'a', ('i2i16', 'b')) */
   /* search1599_0_0 -> 11 in the cache */
   /* search1599_0 -> 1804 in the cache */
   /* search1599_1_0 -> 2112 in the cache */
   /* search1599_1 -> 4415 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 1804, 4415 },
      -1,
   } },

   /* replace1599_0 -> 11 in the cache */
   /* replace1599_1_0 -> 2112 in the cache */
   /* replace1599_1 -> 4358 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 11, 4358 },
      -1,
   } },

   /* ('ilt', ('i2i32', 'a@16'), ('i2i32', 'b@8')) => ('ilt', 'a', ('i2i16', 'b')) */
   /* search1600_0_0 -> 11 in the cache */
   /* search1600_0 -> 1804 in the cache */
   /* search1600_1_0 -> 2112 in the cache */
   /* search1600_1 -> 4415 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 1804, 4415 },
      -1,
   } },

   /* replace1600_0 -> 11 in the cache */
   /* replace1600_1_0 -> 2112 in the cache */
   /* replace1600_1 -> 4358 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 11, 4358 },
      -1,
   } },

   /* ('ige', ('i2i32', 'b@8'), ('i2i32', 'a@16')) => ('ige', ('i2i16', 'b'), 'a') */
   /* search1601_0_0 -> 4368 in the cache */
   /* search1601_0 -> 4420 in the cache */
   /* search1601_1_0 -> 526 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 526 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4420, 4516 },
      -1,
   } },

   /* replace1601_0_0 -> 4368 in the cache */
   /* replace1601_0 -> 4369 in the cache */
   /* replace1601_1 -> 526 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4369, 526 },
      -1,
   } },

   /* ('ilt', ('i2i32', 'b@8'), ('i2i32', 'a@16')) => ('ilt', ('i2i16', 'b'), 'a') */
   /* search1602_0_0 -> 4368 in the cache */
   /* search1602_0 -> 4420 in the cache */
   /* search1602_1_0 -> 526 in the cache */
   /* search1602_1 -> 4516 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4420, 4516 },
      -1,
   } },

   /* replace1602_0_0 -> 4368 in the cache */
   /* replace1602_0 -> 4369 in the cache */
   /* replace1602_1 -> 526 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4369, 526 },
      -1,
   } },

   /* ('ieq', ('i2i32', 'a@16'), ('i2i32', 'b@16')) => ('ieq', 'a', ('i2i16', 'b')) */
   /* search1603_0_0 -> 11 in the cache */
   /* search1603_0 -> 1804 in the cache */
   /* search1603_1_0 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 258 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1804, 4521 },
      -1,
   } },

   /* replace1603_0 -> 11 in the cache */
   /* replace1603_1_0 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 258 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 11, 4523 },
      -1,
   } },

   /* ('ine', ('i2i32', 'a@16'), ('i2i32', 'b@16')) => ('ine', 'a', ('i2i16', 'b')) */
   /* search1604_0_0 -> 11 in the cache */
   /* search1604_0 -> 1804 in the cache */
   /* search1604_1_0 -> 258 in the cache */
   /* search1604_1 -> 4521 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1804, 4521 },
      -1,
   } },

   /* replace1604_0 -> 11 in the cache */
   /* replace1604_1_0 -> 258 in the cache */
   /* replace1604_1 -> 4523 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 11, 4523 },
      -1,
   } },

   /* ('ige', ('i2i32', 'a@16'), ('i2i32', 'b@16')) => ('ige', 'a', ('i2i16', 'b')) */
   /* search1605_0_0 -> 11 in the cache */
   /* search1605_0 -> 1804 in the cache */
   /* search1605_1_0 -> 258 in the cache */
   /* search1605_1 -> 4521 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 1804, 4521 },
      -1,
   } },

   /* replace1605_0 -> 11 in the cache */
   /* replace1605_1_0 -> 258 in the cache */
   /* replace1605_1 -> 4523 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 11, 4523 },
      -1,
   } },

   /* ('ilt', ('i2i32', 'a@16'), ('i2i32', 'b@16')) => ('ilt', 'a', ('i2i16', 'b')) */
   /* search1606_0_0 -> 11 in the cache */
   /* search1606_0 -> 1804 in the cache */
   /* search1606_1_0 -> 258 in the cache */
   /* search1606_1 -> 4521 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 1804, 4521 },
      -1,
   } },

   /* replace1606_0 -> 11 in the cache */
   /* replace1606_1_0 -> 258 in the cache */
   /* replace1606_1 -> 4523 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 11, 4523 },
      -1,
   } },

   /* ('ige', ('i2i32', 'b@16'), ('i2i32', 'a@16')) => ('ige', ('i2i16', 'b'), 'a') */
   /* search1607_0_0 -> 515 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 515 },
      -1,
   } },
   /* search1607_1_0 -> 526 in the cache */
   /* search1607_1 -> 4516 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4531, 4516 },
      -1,
   } },

   /* replace1607_0_0 -> 515 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 515 },
      -1,
   } },
   /* replace1607_1 -> 526 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4533, 526 },
      -1,
   } },

   /* ('ilt', ('i2i32', 'b@16'), ('i2i32', 'a@16')) => ('ilt', ('i2i16', 'b'), 'a') */
   /* search1608_0_0 -> 515 in the cache */
   /* search1608_0 -> 4531 in the cache */
   /* search1608_1_0 -> 526 in the cache */
   /* search1608_1 -> 4516 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4531, 4516 },
      -1,
   } },

   /* replace1608_0_0 -> 515 in the cache */
   /* replace1608_0 -> 4533 in the cache */
   /* replace1608_1 -> 526 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4533, 526 },
      -1,
   } },

   /* ('ieq', ('i2i32', 'a@16'), '#b') => ('iand', ('ieq', 'a', ('i2i16', 'b')), ('ieq', ('i2i32', ('i2i16', 'b')), 'b')) */
   /* search1609_0_0 -> 11 in the cache */
   /* search1609_0 -> 1804 in the cache */
   /* search1609_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1804, 33 },
      -1,
   } },

   /* replace1609_0_0 -> 11 in the cache */
   /* replace1609_0_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 11, 4538 },
      -1,
   } },
   /* replace1609_1_0_0_0 -> 36 in the cache */
   /* replace1609_1_0_0 -> 4538 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 4538 },
      -1,
   } },
   /* replace1609_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 4540, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4539, 4541 },
      -1,
   } },

   /* ('ine', ('i2i32', 'a@16'), '#b') => ('ior', ('ine', 'a', ('i2i16', 'b')), ('ine', ('i2i32', ('i2i16', 'b')), 'b')) */
   /* search1610_0_0 -> 11 in the cache */
   /* search1610_0 -> 1804 in the cache */
   /* search1610_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1804, 33 },
      -1,
   } },

   /* replace1610_0_0 -> 11 in the cache */
   /* replace1610_0_1_0 -> 36 in the cache */
   /* replace1610_0_1 -> 4538 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 11, 4538 },
      -1,
   } },
   /* replace1610_1_0_0_0 -> 36 in the cache */
   /* replace1610_1_0_0 -> 4538 in the cache */
   /* replace1610_1_0 -> 4540 in the cache */
   /* replace1610_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 4540, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4544, 4545 },
      -1,
   } },

   /* ('ilt', ('i2i32', 'a@16'), '#b') => ('iand', ('ilt', -32768, 'b'), ('ior', ('ilt', 32767, 'b'), ('ilt', 'a', ('i2i16', 'b')))) */
   /* search1611_0_0 -> 11 in the cache */
   /* search1611_0 -> 1804 in the cache */
   /* search1611_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 1804, 33 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x8000 /* -32768 */ },
   } },
   /* replace1611_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4548, 36 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x7fff /* 32767 */ },
   } },
   /* replace1611_1_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4550, 36 },
      -1,
   } },
   /* replace1611_1_1_0 -> 11 in the cache */
   /* replace1611_1_1_1_0 -> 36 in the cache */
   /* replace1611_1_1_1 -> 4538 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 11, 4538 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4551, 4552 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4549, 4553 },
      -1,
   } },

   /* ('ilt', '#a', ('i2i32', 'b@16')) => ('iand', ('ilt', 'a', 32767), ('ior', ('ilt', 'a', -32768), ('ilt', ('i2i16', 'a'), 'b'))) */
   /* search1612_0 -> 286 in the cache */
   /* search1612_1_0 -> 258 in the cache */
   /* search1612_1 -> 4521 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 286, 4521 },
      -1,
   } },

   /* replace1612_0_0 -> 16 in the cache */
   /* replace1612_0_1 -> 4550 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 16, 4550 },
      -1,
   } },
   /* replace1612_1_0_0 -> 16 in the cache */
   /* replace1612_1_0_1 -> 4548 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 16, 4548 },
      -1,
   } },
   /* replace1612_1_1_0_0 -> 16 in the cache */
   /* replace1612_1_1_0 -> 4353 in the cache */
   /* replace1612_1_1_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4353, 258 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4557, 4558 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4556, 4559 },
      -1,
   } },

   /* ('ige', ('i2i32', 'a@16'), '#b') => ('iand', ('ige', 32767, 'b'), ('ior', ('ige', -32768, 'b'), ('ige', 'a', ('i2i16', 'b')))) */
   /* search1613_0_0 -> 11 in the cache */
   /* search1613_0 -> 1804 in the cache */
   /* search1613_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 1804, 33 },
      -1,
   } },

   /* replace1613_0_0 -> 4550 in the cache */
   /* replace1613_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4550, 36 },
      -1,
   } },
   /* replace1613_1_0_0 -> 4548 in the cache */
   /* replace1613_1_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4548, 36 },
      -1,
   } },
   /* replace1613_1_1_0 -> 11 in the cache */
   /* replace1613_1_1_1_0 -> 36 in the cache */
   /* replace1613_1_1_1 -> 4538 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 11, 4538 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4563, 4564 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4562, 4565 },
      -1,
   } },

   /* ('ige', '#a', ('i2i32', 'b@16')) => ('iand', ('ige', 'a', -32768), ('ior', ('ige', 'a', 32767), ('ige', ('i2i16', 'a'), 'b'))) */
   /* search1614_0 -> 286 in the cache */
   /* search1614_1_0 -> 258 in the cache */
   /* search1614_1 -> 4521 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 286, 4521 },
      -1,
   } },

   /* replace1614_0_0 -> 16 in the cache */
   /* replace1614_0_1 -> 4548 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 16, 4548 },
      -1,
   } },
   /* replace1614_1_0_0 -> 16 in the cache */
   /* replace1614_1_0_1 -> 4550 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 16, 4550 },
      -1,
   } },
   /* replace1614_1_1_0_0 -> 16 in the cache */
   /* replace1614_1_1_0 -> 4353 in the cache */
   /* replace1614_1_1_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4353, 258 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4569, 4570 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4568, 4571 },
      -1,
   } },

   /* ('ieq', ('i2i64', 'a@16'), ('i2i64', 'b@8')) => ('ieq', 'a', ('i2i16', 'b')) */
   /* search1615_0_0 -> 11 in the cache */
   /* search1615_0 -> 1827 in the cache */
   /* search1615_1_0 -> 2112 in the cache */
   /* search1615_1 -> 4460 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1827, 4460 },
      -1,
   } },

   /* replace1615_0 -> 11 in the cache */
   /* replace1615_1_0 -> 2112 in the cache */
   /* replace1615_1 -> 4358 in the cache */
   /* replace1615 -> 4509 in the cache */

   /* ('ine', ('i2i64', 'a@16'), ('i2i64', 'b@8')) => ('ine', 'a', ('i2i16', 'b')) */
   /* search1616_0_0 -> 11 in the cache */
   /* search1616_0 -> 1827 in the cache */
   /* search1616_1_0 -> 2112 in the cache */
   /* search1616_1 -> 4460 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1827, 4460 },
      -1,
   } },

   /* replace1616_0 -> 11 in the cache */
   /* replace1616_1_0 -> 2112 in the cache */
   /* replace1616_1 -> 4358 in the cache */
   /* replace1616 -> 4511 in the cache */

   /* ('ige', ('i2i64', 'a@16'), ('i2i64', 'b@8')) => ('ige', 'a', ('i2i16', 'b')) */
   /* search1617_0_0 -> 11 in the cache */
   /* search1617_0 -> 1827 in the cache */
   /* search1617_1_0 -> 2112 in the cache */
   /* search1617_1 -> 4460 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 1827, 4460 },
      -1,
   } },

   /* replace1617_0 -> 11 in the cache */
   /* replace1617_1_0 -> 2112 in the cache */
   /* replace1617_1 -> 4358 in the cache */
   /* replace1617 -> 4513 in the cache */

   /* ('ilt', ('i2i64', 'a@16'), ('i2i64', 'b@8')) => ('ilt', 'a', ('i2i16', 'b')) */
   /* search1618_0_0 -> 11 in the cache */
   /* search1618_0 -> 1827 in the cache */
   /* search1618_1_0 -> 2112 in the cache */
   /* search1618_1 -> 4460 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 1827, 4460 },
      -1,
   } },

   /* replace1618_0 -> 11 in the cache */
   /* replace1618_1_0 -> 2112 in the cache */
   /* replace1618_1 -> 4358 in the cache */
   /* replace1618 -> 4515 in the cache */

   /* ('ige', ('i2i64', 'b@8'), ('i2i64', 'a@16')) => ('ige', ('i2i16', 'b'), 'a') */
   /* search1619_0_0 -> 4368 in the cache */
   /* search1619_0 -> 4465 in the cache */
   /* search1619_1_0 -> 526 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 526 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4465, 4577 },
      -1,
   } },

   /* replace1619_0_0 -> 4368 in the cache */
   /* replace1619_0 -> 4369 in the cache */
   /* replace1619_1 -> 526 in the cache */
   /* replace1619 -> 4518 in the cache */

   /* ('ilt', ('i2i64', 'b@8'), ('i2i64', 'a@16')) => ('ilt', ('i2i16', 'b'), 'a') */
   /* search1620_0_0 -> 4368 in the cache */
   /* search1620_0 -> 4465 in the cache */
   /* search1620_1_0 -> 526 in the cache */
   /* search1620_1 -> 4577 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4465, 4577 },
      -1,
   } },

   /* replace1620_0_0 -> 4368 in the cache */
   /* replace1620_0 -> 4369 in the cache */
   /* replace1620_1 -> 526 in the cache */
   /* replace1620 -> 4520 in the cache */

   /* ('ieq', ('i2i64', 'a@16'), ('i2i64', 'b@16')) => ('ieq', 'a', ('i2i16', 'b')) */
   /* search1621_0_0 -> 11 in the cache */
   /* search1621_0 -> 1827 in the cache */
   /* search1621_1_0 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 258 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1827, 4580 },
      -1,
   } },

   /* replace1621_0 -> 11 in the cache */
   /* replace1621_1_0 -> 258 in the cache */
   /* replace1621_1 -> 4523 in the cache */
   /* replace1621 -> 4524 in the cache */

   /* ('ine', ('i2i64', 'a@16'), ('i2i64', 'b@16')) => ('ine', 'a', ('i2i16', 'b')) */
   /* search1622_0_0 -> 11 in the cache */
   /* search1622_0 -> 1827 in the cache */
   /* search1622_1_0 -> 258 in the cache */
   /* search1622_1 -> 4580 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1827, 4580 },
      -1,
   } },

   /* replace1622_0 -> 11 in the cache */
   /* replace1622_1_0 -> 258 in the cache */
   /* replace1622_1 -> 4523 in the cache */
   /* replace1622 -> 4526 in the cache */

   /* ('ige', ('i2i64', 'a@16'), ('i2i64', 'b@16')) => ('ige', 'a', ('i2i16', 'b')) */
   /* search1623_0_0 -> 11 in the cache */
   /* search1623_0 -> 1827 in the cache */
   /* search1623_1_0 -> 258 in the cache */
   /* search1623_1 -> 4580 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 1827, 4580 },
      -1,
   } },

   /* replace1623_0 -> 11 in the cache */
   /* replace1623_1_0 -> 258 in the cache */
   /* replace1623_1 -> 4523 in the cache */
   /* replace1623 -> 4528 in the cache */

   /* ('ilt', ('i2i64', 'a@16'), ('i2i64', 'b@16')) => ('ilt', 'a', ('i2i16', 'b')) */
   /* search1624_0_0 -> 11 in the cache */
   /* search1624_0 -> 1827 in the cache */
   /* search1624_1_0 -> 258 in the cache */
   /* search1624_1 -> 4580 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 1827, 4580 },
      -1,
   } },

   /* replace1624_0 -> 11 in the cache */
   /* replace1624_1_0 -> 258 in the cache */
   /* replace1624_1 -> 4523 in the cache */
   /* replace1624 -> 4530 in the cache */

   /* ('ige', ('i2i64', 'b@16'), ('i2i64', 'a@16')) => ('ige', ('i2i16', 'b'), 'a') */
   /* search1625_0_0 -> 515 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 515 },
      -1,
   } },
   /* search1625_1_0 -> 526 in the cache */
   /* search1625_1 -> 4577 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4585, 4577 },
      -1,
   } },

   /* replace1625_0_0 -> 515 in the cache */
   /* replace1625_0 -> 4533 in the cache */
   /* replace1625_1 -> 526 in the cache */
   /* replace1625 -> 4534 in the cache */

   /* ('ilt', ('i2i64', 'b@16'), ('i2i64', 'a@16')) => ('ilt', ('i2i16', 'b'), 'a') */
   /* search1626_0_0 -> 515 in the cache */
   /* search1626_0 -> 4585 in the cache */
   /* search1626_1_0 -> 526 in the cache */
   /* search1626_1 -> 4577 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4585, 4577 },
      -1,
   } },

   /* replace1626_0_0 -> 515 in the cache */
   /* replace1626_0 -> 4533 in the cache */
   /* replace1626_1 -> 526 in the cache */
   /* replace1626 -> 4536 in the cache */

   /* ('ieq', ('i2i64', 'a@16'), '#b') => ('iand', ('ieq', 'a', ('i2i16', 'b')), ('ieq', ('i2i64', ('i2i16', 'b')), 'b')) */
   /* search1627_0_0 -> 11 in the cache */
   /* search1627_0 -> 1827 in the cache */
   /* search1627_1 -> 4469 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1827, 4469 },
      -1,
   } },

   /* replace1627_0_0 -> 11 in the cache */
   /* replace1627_0_1_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 11, 4589 },
      -1,
   } },
   /* replace1627_1_0_0_0 -> 41 in the cache */
   /* replace1627_1_0_0 -> 4589 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 4589 },
      -1,
   } },
   /* replace1627_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 4591, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4590, 4592 },
      -1,
   } },

   /* ('ine', ('i2i64', 'a@16'), '#b') => ('ior', ('ine', 'a', ('i2i16', 'b')), ('ine', ('i2i64', ('i2i16', 'b')), 'b')) */
   /* search1628_0_0 -> 11 in the cache */
   /* search1628_0 -> 1827 in the cache */
   /* search1628_1 -> 4469 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1827, 4469 },
      -1,
   } },

   /* replace1628_0_0 -> 11 in the cache */
   /* replace1628_0_1_0 -> 41 in the cache */
   /* replace1628_0_1 -> 4589 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 11, 4589 },
      -1,
   } },
   /* replace1628_1_0_0_0 -> 41 in the cache */
   /* replace1628_1_0_0 -> 4589 in the cache */
   /* replace1628_1_0 -> 4591 in the cache */
   /* replace1628_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 4591, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4595, 4596 },
      -1,
   } },

   /* ('ilt', ('i2i64', 'a@16'), '#b') => ('iand', ('ilt', -32768, 'b'), ('ior', ('ilt', 32767, 'b'), ('ilt', 'a', ('i2i16', 'b')))) */
   /* search1629_0_0 -> 11 in the cache */
   /* search1629_0 -> 1827 in the cache */
   /* search1629_1 -> 4469 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 1827, 4469 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { -0x8000 /* -32768 */ },
   } },
   /* replace1629_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4599, 41 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x7fff /* 32767 */ },
   } },
   /* replace1629_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4601, 41 },
      -1,
   } },
   /* replace1629_1_1_0 -> 11 in the cache */
   /* replace1629_1_1_1_0 -> 41 in the cache */
   /* replace1629_1_1_1 -> 4589 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 11, 4589 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4602, 4603 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4600, 4604 },
      -1,
   } },

   /* ('ilt', '#a', ('i2i64', 'b@16')) => ('iand', ('ilt', 'a', 32767), ('ior', ('ilt', 'a', -32768), ('ilt', ('i2i16', 'a'), 'b'))) */
   /* search1630_0 -> 4488 in the cache */
   /* search1630_1_0 -> 258 in the cache */
   /* search1630_1 -> 4580 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4488, 4580 },
      -1,
   } },

   /* replace1630_0_0 -> 21 in the cache */
   /* replace1630_0_1 -> 4601 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 21, 4601 },
      -1,
   } },
   /* replace1630_1_0_0 -> 21 in the cache */
   /* replace1630_1_0_1 -> 4599 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 21, 4599 },
      -1,
   } },
   /* replace1630_1_1_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 21 },
      -1,
   } },
   /* replace1630_1_1_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4609, 258 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4608, 4610 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4607, 4611 },
      -1,
   } },

   /* ('ige', ('i2i64', 'a@16'), '#b') => ('iand', ('ige', 32767, 'b'), ('ior', ('ige', -32768, 'b'), ('ige', 'a', ('i2i16', 'b')))) */
   /* search1631_0_0 -> 11 in the cache */
   /* search1631_0 -> 1827 in the cache */
   /* search1631_1 -> 4469 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 1827, 4469 },
      -1,
   } },

   /* replace1631_0_0 -> 4601 in the cache */
   /* replace1631_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4601, 41 },
      -1,
   } },
   /* replace1631_1_0_0 -> 4599 in the cache */
   /* replace1631_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4599, 41 },
      -1,
   } },
   /* replace1631_1_1_0 -> 11 in the cache */
   /* replace1631_1_1_1_0 -> 41 in the cache */
   /* replace1631_1_1_1 -> 4589 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 11, 4589 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4615, 4616 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4614, 4617 },
      -1,
   } },

   /* ('ige', '#a', ('i2i64', 'b@16')) => ('iand', ('ige', 'a', -32768), ('ior', ('ige', 'a', 32767), ('ige', ('i2i16', 'a'), 'b'))) */
   /* search1632_0 -> 4488 in the cache */
   /* search1632_1_0 -> 258 in the cache */
   /* search1632_1 -> 4580 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4488, 4580 },
      -1,
   } },

   /* replace1632_0_0 -> 21 in the cache */
   /* replace1632_0_1 -> 4599 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 21, 4599 },
      -1,
   } },
   /* replace1632_1_0_0 -> 21 in the cache */
   /* replace1632_1_0_1 -> 4601 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 21, 4601 },
      -1,
   } },
   /* replace1632_1_1_0_0 -> 21 in the cache */
   /* replace1632_1_1_0 -> 4609 in the cache */
   /* replace1632_1_1_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4609, 258 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4621, 4622 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4620, 4623 },
      -1,
   } },

   /* ('ieq', ('i2i64', 'a@32'), ('i2i64', 'b@8')) => ('ieq', 'a', ('i2i32', 'b')) */
   /* search1633_0_0 -> 16 in the cache */
   /* search1633_0 -> 1909 in the cache */
   /* search1633_1_0 -> 2112 in the cache */
   /* search1633_1 -> 4460 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1909, 4460 },
      -1,
   } },

   /* replace1633_0 -> 16 in the cache */
   /* replace1633_1_0 -> 2112 in the cache */
   /* replace1633_1 -> 4415 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 16, 4415 },
      -1,
   } },

   /* ('ine', ('i2i64', 'a@32'), ('i2i64', 'b@8')) => ('ine', 'a', ('i2i32', 'b')) */
   /* search1634_0_0 -> 16 in the cache */
   /* search1634_0 -> 1909 in the cache */
   /* search1634_1_0 -> 2112 in the cache */
   /* search1634_1 -> 4460 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1909, 4460 },
      -1,
   } },

   /* replace1634_0 -> 16 in the cache */
   /* replace1634_1_0 -> 2112 in the cache */
   /* replace1634_1 -> 4415 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 16, 4415 },
      -1,
   } },

   /* ('ige', ('i2i64', 'a@32'), ('i2i64', 'b@8')) => ('ige', 'a', ('i2i32', 'b')) */
   /* search1635_0_0 -> 16 in the cache */
   /* search1635_0 -> 1909 in the cache */
   /* search1635_1_0 -> 2112 in the cache */
   /* search1635_1 -> 4460 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 1909, 4460 },
      -1,
   } },

   /* replace1635_0 -> 16 in the cache */
   /* replace1635_1_0 -> 2112 in the cache */
   /* replace1635_1 -> 4415 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 16, 4415 },
      -1,
   } },

   /* ('ilt', ('i2i64', 'a@32'), ('i2i64', 'b@8')) => ('ilt', 'a', ('i2i32', 'b')) */
   /* search1636_0_0 -> 16 in the cache */
   /* search1636_0 -> 1909 in the cache */
   /* search1636_1_0 -> 2112 in the cache */
   /* search1636_1 -> 4460 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 1909, 4460 },
      -1,
   } },

   /* replace1636_0 -> 16 in the cache */
   /* replace1636_1_0 -> 2112 in the cache */
   /* replace1636_1 -> 4415 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 16, 4415 },
      -1,
   } },

   /* ('ige', ('i2i64', 'b@8'), ('i2i64', 'a@32')) => ('ige', ('i2i32', 'b'), 'a') */
   /* search1637_0_0 -> 4368 in the cache */
   /* search1637_0 -> 4465 in the cache */
   /* search1637_1_0 -> 183 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 183 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4465, 4633 },
      -1,
   } },

   /* replace1637_0_0 -> 4368 in the cache */
   /* replace1637_0 -> 4420 in the cache */
   /* replace1637_1 -> 183 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4420, 183 },
      -1,
   } },

   /* ('ilt', ('i2i64', 'b@8'), ('i2i64', 'a@32')) => ('ilt', ('i2i32', 'b'), 'a') */
   /* search1638_0_0 -> 4368 in the cache */
   /* search1638_0 -> 4465 in the cache */
   /* search1638_1_0 -> 183 in the cache */
   /* search1638_1 -> 4633 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4465, 4633 },
      -1,
   } },

   /* replace1638_0_0 -> 4368 in the cache */
   /* replace1638_0 -> 4420 in the cache */
   /* replace1638_1 -> 183 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4420, 183 },
      -1,
   } },

   /* ('ieq', ('i2i64', 'a@32'), ('i2i64', 'b@16')) => ('ieq', 'a', ('i2i32', 'b')) */
   /* search1639_0_0 -> 16 in the cache */
   /* search1639_0 -> 1909 in the cache */
   /* search1639_1_0 -> 258 in the cache */
   /* search1639_1 -> 4580 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1909, 4580 },
      -1,
   } },

   /* replace1639_0 -> 16 in the cache */
   /* replace1639_1_0 -> 258 in the cache */
   /* replace1639_1 -> 4521 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 16, 4521 },
      -1,
   } },

   /* ('ine', ('i2i64', 'a@32'), ('i2i64', 'b@16')) => ('ine', 'a', ('i2i32', 'b')) */
   /* search1640_0_0 -> 16 in the cache */
   /* search1640_0 -> 1909 in the cache */
   /* search1640_1_0 -> 258 in the cache */
   /* search1640_1 -> 4580 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1909, 4580 },
      -1,
   } },

   /* replace1640_0 -> 16 in the cache */
   /* replace1640_1_0 -> 258 in the cache */
   /* replace1640_1 -> 4521 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 16, 4521 },
      -1,
   } },

   /* ('ige', ('i2i64', 'a@32'), ('i2i64', 'b@16')) => ('ige', 'a', ('i2i32', 'b')) */
   /* search1641_0_0 -> 16 in the cache */
   /* search1641_0 -> 1909 in the cache */
   /* search1641_1_0 -> 258 in the cache */
   /* search1641_1 -> 4580 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 1909, 4580 },
      -1,
   } },

   /* replace1641_0 -> 16 in the cache */
   /* replace1641_1_0 -> 258 in the cache */
   /* replace1641_1 -> 4521 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 16, 4521 },
      -1,
   } },

   /* ('ilt', ('i2i64', 'a@32'), ('i2i64', 'b@16')) => ('ilt', 'a', ('i2i32', 'b')) */
   /* search1642_0_0 -> 16 in the cache */
   /* search1642_0 -> 1909 in the cache */
   /* search1642_1_0 -> 258 in the cache */
   /* search1642_1 -> 4580 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 1909, 4580 },
      -1,
   } },

   /* replace1642_0 -> 16 in the cache */
   /* replace1642_1_0 -> 258 in the cache */
   /* replace1642_1 -> 4521 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 16, 4521 },
      -1,
   } },

   /* ('ige', ('i2i64', 'b@16'), ('i2i64', 'a@32')) => ('ige', ('i2i32', 'b'), 'a') */
   /* search1643_0_0 -> 515 in the cache */
   /* search1643_0 -> 4585 in the cache */
   /* search1643_1_0 -> 183 in the cache */
   /* search1643_1 -> 4633 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4585, 4633 },
      -1,
   } },

   /* replace1643_0_0 -> 515 in the cache */
   /* replace1643_0 -> 4531 in the cache */
   /* replace1643_1 -> 183 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4531, 183 },
      -1,
   } },

   /* ('ilt', ('i2i64', 'b@16'), ('i2i64', 'a@32')) => ('ilt', ('i2i32', 'b'), 'a') */
   /* search1644_0_0 -> 515 in the cache */
   /* search1644_0 -> 4585 in the cache */
   /* search1644_1_0 -> 183 in the cache */
   /* search1644_1 -> 4633 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4585, 4633 },
      -1,
   } },

   /* replace1644_0_0 -> 515 in the cache */
   /* replace1644_0 -> 4531 in the cache */
   /* replace1644_1 -> 183 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4531, 183 },
      -1,
   } },

   /* ('ieq', ('i2i64', 'a@32'), ('i2i64', 'b@32')) => ('ieq', 'a', ('i2i32', 'b')) */
   /* search1645_0_0 -> 16 in the cache */
   /* search1645_0 -> 1909 in the cache */
   /* search1645_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1909, 4650 },
      -1,
   } },

   /* replace1645_0 -> 16 in the cache */
   /* replace1645_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 16, 4652 },
      -1,
   } },

   /* ('ine', ('i2i64', 'a@32'), ('i2i64', 'b@32')) => ('ine', 'a', ('i2i32', 'b')) */
   /* search1646_0_0 -> 16 in the cache */
   /* search1646_0 -> 1909 in the cache */
   /* search1646_1_0 -> 36 in the cache */
   /* search1646_1 -> 4650 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1909, 4650 },
      -1,
   } },

   /* replace1646_0 -> 16 in the cache */
   /* replace1646_1_0 -> 36 in the cache */
   /* replace1646_1 -> 4652 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 16, 4652 },
      -1,
   } },

   /* ('ige', ('i2i64', 'a@32'), ('i2i64', 'b@32')) => ('ige', 'a', ('i2i32', 'b')) */
   /* search1647_0_0 -> 16 in the cache */
   /* search1647_0 -> 1909 in the cache */
   /* search1647_1_0 -> 36 in the cache */
   /* search1647_1 -> 4650 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 1909, 4650 },
      -1,
   } },

   /* replace1647_0 -> 16 in the cache */
   /* replace1647_1_0 -> 36 in the cache */
   /* replace1647_1 -> 4652 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 16, 4652 },
      -1,
   } },

   /* ('ilt', ('i2i64', 'a@32'), ('i2i64', 'b@32')) => ('ilt', 'a', ('i2i32', 'b')) */
   /* search1648_0_0 -> 16 in the cache */
   /* search1648_0 -> 1909 in the cache */
   /* search1648_1_0 -> 36 in the cache */
   /* search1648_1 -> 4650 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 1909, 4650 },
      -1,
   } },

   /* replace1648_0 -> 16 in the cache */
   /* replace1648_1_0 -> 36 in the cache */
   /* replace1648_1 -> 4652 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 16, 4652 },
      -1,
   } },

   /* ('ige', ('i2i64', 'b@32'), ('i2i64', 'a@32')) => ('ige', ('i2i32', 'b'), 'a') */
   /* search1649_0_0 -> 373 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 373 },
      -1,
   } },
   /* search1649_1_0 -> 183 in the cache */
   /* search1649_1 -> 4633 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4660, 4633 },
      -1,
   } },

   /* replace1649_0_0 -> 373 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 373 },
      -1,
   } },
   /* replace1649_1 -> 183 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4662, 183 },
      -1,
   } },

   /* ('ilt', ('i2i64', 'b@32'), ('i2i64', 'a@32')) => ('ilt', ('i2i32', 'b'), 'a') */
   /* search1650_0_0 -> 373 in the cache */
   /* search1650_0 -> 4660 in the cache */
   /* search1650_1_0 -> 183 in the cache */
   /* search1650_1 -> 4633 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4660, 4633 },
      -1,
   } },

   /* replace1650_0_0 -> 373 in the cache */
   /* replace1650_0 -> 4662 in the cache */
   /* replace1650_1 -> 183 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4662, 183 },
      -1,
   } },

   /* ('ieq', ('i2i64', 'a@32'), '#b') => ('iand', ('ieq', 'a', ('i2i32', 'b')), ('ieq', ('i2i64', ('i2i32', 'b')), 'b')) */
   /* search1651_0_0 -> 16 in the cache */
   /* search1651_0 -> 1909 in the cache */
   /* search1651_1 -> 4469 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1909, 4469 },
      -1,
   } },

   /* replace1651_0_0 -> 16 in the cache */
   /* replace1651_0_1_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 16, 4667 },
      -1,
   } },
   /* replace1651_1_0_0_0 -> 41 in the cache */
   /* replace1651_1_0_0 -> 4667 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 4667 },
      -1,
   } },
   /* replace1651_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 4669, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4668, 4670 },
      -1,
   } },

   /* ('ine', ('i2i64', 'a@32'), '#b') => ('ior', ('ine', 'a', ('i2i32', 'b')), ('ine', ('i2i64', ('i2i32', 'b')), 'b')) */
   /* search1652_0_0 -> 16 in the cache */
   /* search1652_0 -> 1909 in the cache */
   /* search1652_1 -> 4469 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1909, 4469 },
      -1,
   } },

   /* replace1652_0_0 -> 16 in the cache */
   /* replace1652_0_1_0 -> 41 in the cache */
   /* replace1652_0_1 -> 4667 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 16, 4667 },
      -1,
   } },
   /* replace1652_1_0_0_0 -> 41 in the cache */
   /* replace1652_1_0_0 -> 4667 in the cache */
   /* replace1652_1_0 -> 4669 in the cache */
   /* replace1652_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 4669, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4673, 4674 },
      -1,
   } },

   /* ('ilt', ('i2i64', 'a@32'), '#b') => ('iand', ('ilt', -2147483648, 'b'), ('ior', ('ilt', 2147483647, 'b'), ('ilt', 'a', ('i2i32', 'b')))) */
   /* search1653_0_0 -> 16 in the cache */
   /* search1653_0 -> 1909 in the cache */
   /* search1653_1 -> 4469 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 1909, 4469 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { -0x80000000 /* -2147483648 */ },
   } },
   /* replace1653_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4677, 41 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x7fffffff /* 2147483647 */ },
   } },
   /* replace1653_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4679, 41 },
      -1,
   } },
   /* replace1653_1_1_0 -> 16 in the cache */
   /* replace1653_1_1_1_0 -> 41 in the cache */
   /* replace1653_1_1_1 -> 4667 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 16, 4667 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4680, 4681 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4678, 4682 },
      -1,
   } },

   /* ('ilt', '#a', ('i2i64', 'b@32')) => ('iand', ('ilt', 'a', 2147483647), ('ior', ('ilt', 'a', -2147483648), ('ilt', ('i2i32', 'a'), 'b'))) */
   /* search1654_0 -> 4488 in the cache */
   /* search1654_1_0 -> 36 in the cache */
   /* search1654_1 -> 4650 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4488, 4650 },
      -1,
   } },

   /* replace1654_0_0 -> 21 in the cache */
   /* replace1654_0_1 -> 4679 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 21, 4679 },
      -1,
   } },
   /* replace1654_1_0_0 -> 21 in the cache */
   /* replace1654_1_0_1 -> 4677 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 21, 4677 },
      -1,
   } },
   /* replace1654_1_1_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 21 },
      -1,
   } },
   /* replace1654_1_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4687, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4686, 4688 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4685, 4689 },
      -1,
   } },

   /* ('ige', ('i2i64', 'a@32'), '#b') => ('iand', ('ige', 2147483647, 'b'), ('ior', ('ige', -2147483648, 'b'), ('ige', 'a', ('i2i32', 'b')))) */
   /* search1655_0_0 -> 16 in the cache */
   /* search1655_0 -> 1909 in the cache */
   /* search1655_1 -> 4469 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 1909, 4469 },
      -1,
   } },

   /* replace1655_0_0 -> 4679 in the cache */
   /* replace1655_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4679, 41 },
      -1,
   } },
   /* replace1655_1_0_0 -> 4677 in the cache */
   /* replace1655_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4677, 41 },
      -1,
   } },
   /* replace1655_1_1_0 -> 16 in the cache */
   /* replace1655_1_1_1_0 -> 41 in the cache */
   /* replace1655_1_1_1 -> 4667 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 16, 4667 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4693, 4694 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4692, 4695 },
      -1,
   } },

   /* ('ige', '#a', ('i2i64', 'b@32')) => ('iand', ('ige', 'a', -2147483648), ('ior', ('ige', 'a', 2147483647), ('ige', ('i2i32', 'a'), 'b'))) */
   /* search1656_0 -> 4488 in the cache */
   /* search1656_1_0 -> 36 in the cache */
   /* search1656_1 -> 4650 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4488, 4650 },
      -1,
   } },

   /* replace1656_0_0 -> 21 in the cache */
   /* replace1656_0_1 -> 4677 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 21, 4677 },
      -1,
   } },
   /* replace1656_1_0_0 -> 21 in the cache */
   /* replace1656_1_0_1 -> 4679 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 21, 4679 },
      -1,
   } },
   /* replace1656_1_1_0_0 -> 21 in the cache */
   /* replace1656_1_1_0 -> 4687 in the cache */
   /* replace1656_1_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4687, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4699, 4700 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4698, 4701 },
      -1,
   } },

   /* ('ieq', ('u2u16', 'a@8'), ('u2u16', 'b@8')) => ('ieq', 'a', ('u2u8', 'b')) */
   /* search1657_0_0 -> 6 in the cache */
   /* search1657_0 -> 4298 in the cache */
   /* search1657_1_0 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 2112 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 4298, 4703 },
      -1,
   } },

   /* replace1657_0 -> 6 in the cache */
   /* replace1657_1_0 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 2112 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 6, 4705 },
      -1,
   } },

   /* ('ine', ('u2u16', 'a@8'), ('u2u16', 'b@8')) => ('ine', 'a', ('u2u8', 'b')) */
   /* search1658_0_0 -> 6 in the cache */
   /* search1658_0 -> 4298 in the cache */
   /* search1658_1_0 -> 2112 in the cache */
   /* search1658_1 -> 4703 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 4298, 4703 },
      -1,
   } },

   /* replace1658_0 -> 6 in the cache */
   /* replace1658_1_0 -> 2112 in the cache */
   /* replace1658_1 -> 4705 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 6, 4705 },
      -1,
   } },

   /* ('uge', ('u2u16', 'a@8'), ('u2u16', 'b@8')) => ('uge', 'a', ('u2u8', 'b')) */
   /* search1659_0_0 -> 6 in the cache */
   /* search1659_0 -> 4298 in the cache */
   /* search1659_1_0 -> 2112 in the cache */
   /* search1659_1 -> 4703 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4298, 4703 },
      -1,
   } },

   /* replace1659_0 -> 6 in the cache */
   /* replace1659_1_0 -> 2112 in the cache */
   /* replace1659_1 -> 4705 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 6, 4705 },
      -1,
   } },

   /* ('ult', ('u2u16', 'a@8'), ('u2u16', 'b@8')) => ('ult', 'a', ('u2u8', 'b')) */
   /* search1660_0_0 -> 6 in the cache */
   /* search1660_0 -> 4298 in the cache */
   /* search1660_1_0 -> 2112 in the cache */
   /* search1660_1 -> 4703 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4298, 4703 },
      -1,
   } },

   /* replace1660_0 -> 6 in the cache */
   /* replace1660_1_0 -> 2112 in the cache */
   /* replace1660_1 -> 4705 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 6, 4705 },
      -1,
   } },

   /* ('uge', ('u2u16', 'b@8'), ('u2u16', 'a@8')) => ('uge', ('u2u8', 'b'), 'a') */
   /* search1661_0_0 -> 4368 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 4368 },
      -1,
   } },
   /* search1661_1_0 -> 4370 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 4370 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4713, 4714 },
      -1,
   } },

   /* replace1661_0_0 -> 4368 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 4368 },
      -1,
   } },
   /* replace1661_1 -> 4370 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4716, 4370 },
      -1,
   } },

   /* ('ult', ('u2u16', 'b@8'), ('u2u16', 'a@8')) => ('ult', ('u2u8', 'b'), 'a') */
   /* search1662_0_0 -> 4368 in the cache */
   /* search1662_0 -> 4713 in the cache */
   /* search1662_1_0 -> 4370 in the cache */
   /* search1662_1 -> 4714 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4713, 4714 },
      -1,
   } },

   /* replace1662_0_0 -> 4368 in the cache */
   /* replace1662_0 -> 4716 in the cache */
   /* replace1662_1 -> 4370 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4716, 4370 },
      -1,
   } },

   /* ('ieq', ('u2u16', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('u2u8', 'b')), ('ieq', ('u2u16', ('u2u8', 'b')), 'b')) */
   /* search1663_0_0 -> 6 in the cache */
   /* search1663_0 -> 4298 in the cache */
   /* search1663_1 -> 2011 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 4298, 2011 },
      -1,
   } },

   /* replace1663_0_0 -> 6 in the cache */
   /* replace1663_0_1_0 -> 258 in the cache */
   /* replace1663_0_1 -> 2019 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 6, 2019 },
      -1,
   } },
   /* replace1663_1_0_0_0 -> 258 in the cache */
   /* replace1663_1_0_0 -> 2019 in the cache */
   /* replace1663_1_0 -> 2020 in the cache */
   /* replace1663_1_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 2020, 258 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4721, 4722 },
      -1,
   } },

   /* ('ine', ('u2u16', 'a@8'), '#b') => ('ior', ('ine', 'a', ('u2u8', 'b')), ('ine', ('u2u16', ('u2u8', 'b')), 'b')) */
   /* search1664_0_0 -> 6 in the cache */
   /* search1664_0 -> 4298 in the cache */
   /* search1664_1 -> 2011 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 4298, 2011 },
      -1,
   } },

   /* replace1664_0_0 -> 6 in the cache */
   /* replace1664_0_1_0 -> 258 in the cache */
   /* replace1664_0_1 -> 2019 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 6, 2019 },
      -1,
   } },
   /* replace1664_1_0_0_0 -> 258 in the cache */
   /* replace1664_1_0_0 -> 2019 in the cache */
   /* replace1664_1_0 -> 2020 in the cache */
   /* replace1664_1_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 2020, 258 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4725, 4726 },
      -1,
   } },

   /* ('ult', ('u2u16', 'a@8'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 255, 'b'), ('ult', 'a', ('u2u8', 'b')))) */
   /* search1665_0_0 -> 6 in the cache */
   /* search1665_0 -> 4298 in the cache */
   /* search1665_1 -> 2011 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4298, 2011 },
      -1,
   } },

   /* replace1665_0_0 -> 856 in the cache */
   /* replace1665_0_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 856, 258 },
      -1,
   } },
   /* replace1665_1_0_0 -> 2013 in the cache */
   /* replace1665_1_0_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 2013, 258 },
      -1,
   } },
   /* replace1665_1_1_0 -> 6 in the cache */
   /* replace1665_1_1_1_0 -> 258 in the cache */
   /* replace1665_1_1_1 -> 2019 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 6, 2019 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4730, 4731 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4729, 4732 },
      -1,
   } },

   /* ('ult', '#a', ('u2u16', 'b@8')) => ('iand', ('ult', 'a', 255), ('ior', ('ult', 'a', 0), ('ult', ('u2u8', 'a'), 'b'))) */
   /* search1666_0 -> 4395 in the cache */
   /* search1666_1_0 -> 2112 in the cache */
   /* search1666_1 -> 4703 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4395, 4703 },
      -1,
   } },

   /* replace1666_0_0 -> 11 in the cache */
   /* replace1666_0_1 -> 2013 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 11, 2013 },
      -1,
   } },
   /* replace1666_1_0_0 -> 11 in the cache */
   /* replace1666_1_0_1 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 11, 856 },
      -1,
   } },
   /* replace1666_1_1_0_0 -> 11 in the cache */
   /* replace1666_1_1_0 -> 2009 in the cache */
   /* replace1666_1_1_1 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 2009, 2112 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4736, 4737 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4735, 4738 },
      -1,
   } },

   /* ('uge', ('u2u16', 'a@8'), '#b') => ('iand', ('uge', 255, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u8', 'b')))) */
   /* search1667_0_0 -> 6 in the cache */
   /* search1667_0 -> 4298 in the cache */
   /* search1667_1 -> 2011 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4298, 2011 },
      -1,
   } },

   /* replace1667_0_0 -> 2013 in the cache */
   /* replace1667_0_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 2013, 258 },
      -1,
   } },
   /* replace1667_1_0_0 -> 856 in the cache */
   /* replace1667_1_0_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 856, 258 },
      -1,
   } },
   /* replace1667_1_1_0 -> 6 in the cache */
   /* replace1667_1_1_1_0 -> 258 in the cache */
   /* replace1667_1_1_1 -> 2019 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 6, 2019 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4742, 4743 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4741, 4744 },
      -1,
   } },

   /* ('uge', '#a', ('u2u16', 'b@8')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 255), ('uge', ('u2u8', 'a'), 'b'))) */
   /* search1668_0 -> 4395 in the cache */
   /* search1668_1_0 -> 2112 in the cache */
   /* search1668_1 -> 4703 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4395, 4703 },
      -1,
   } },

   /* replace1668_0_0 -> 11 in the cache */
   /* replace1668_0_1 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 11, 856 },
      -1,
   } },
   /* replace1668_1_0_0 -> 11 in the cache */
   /* replace1668_1_0_1 -> 2013 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 11, 2013 },
      -1,
   } },
   /* replace1668_1_1_0_0 -> 11 in the cache */
   /* replace1668_1_1_0 -> 2009 in the cache */
   /* replace1668_1_1_1 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 2009, 2112 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4748, 4749 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4747, 4750 },
      -1,
   } },

   /* ('ieq', ('u2u32', 'a@8'), ('u2u32', 'b@8')) => ('ieq', 'a', ('u2u8', 'b')) */
   /* search1669_0_0 -> 6 in the cache */
   /* search1669_0 -> 3282 in the cache */
   /* search1669_1_0 -> 2112 in the cache */
   /* search1669_1 -> 3284 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 3282, 3284 },
      -1,
   } },

   /* replace1669_0 -> 6 in the cache */
   /* replace1669_1_0 -> 2112 in the cache */
   /* replace1669_1 -> 4705 in the cache */
   /* replace1669 -> 4706 in the cache */

   /* ('ine', ('u2u32', 'a@8'), ('u2u32', 'b@8')) => ('ine', 'a', ('u2u8', 'b')) */
   /* search1670_0_0 -> 6 in the cache */
   /* search1670_0 -> 3282 in the cache */
   /* search1670_1_0 -> 2112 in the cache */
   /* search1670_1 -> 3284 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 3282, 3284 },
      -1,
   } },

   /* replace1670_0 -> 6 in the cache */
   /* replace1670_1_0 -> 2112 in the cache */
   /* replace1670_1 -> 4705 in the cache */
   /* replace1670 -> 4708 in the cache */

   /* ('uge', ('u2u32', 'a@8'), ('u2u32', 'b@8')) => ('uge', 'a', ('u2u8', 'b')) */
   /* search1671_0_0 -> 6 in the cache */
   /* search1671_0 -> 3282 in the cache */
   /* search1671_1_0 -> 2112 in the cache */
   /* search1671_1 -> 3284 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 3282, 3284 },
      -1,
   } },

   /* replace1671_0 -> 6 in the cache */
   /* replace1671_1_0 -> 2112 in the cache */
   /* replace1671_1 -> 4705 in the cache */
   /* replace1671 -> 4710 in the cache */

   /* ('ult', ('u2u32', 'a@8'), ('u2u32', 'b@8')) => ('ult', 'a', ('u2u8', 'b')) */
   /* search1672_0_0 -> 6 in the cache */
   /* search1672_0 -> 3282 in the cache */
   /* search1672_1_0 -> 2112 in the cache */
   /* search1672_1 -> 3284 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 3282, 3284 },
      -1,
   } },

   /* replace1672_0 -> 6 in the cache */
   /* replace1672_1_0 -> 2112 in the cache */
   /* replace1672_1 -> 4705 in the cache */
   /* replace1672 -> 4712 in the cache */

   /* ('uge', ('u2u32', 'b@8'), ('u2u32', 'a@8')) => ('uge', ('u2u8', 'b'), 'a') */
   /* search1673_0_0 -> 4368 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 4368 },
      -1,
   } },
   /* search1673_1_0 -> 4370 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 4370 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4756, 4757 },
      -1,
   } },

   /* replace1673_0_0 -> 4368 in the cache */
   /* replace1673_0 -> 4716 in the cache */
   /* replace1673_1 -> 4370 in the cache */
   /* replace1673 -> 4717 in the cache */

   /* ('ult', ('u2u32', 'b@8'), ('u2u32', 'a@8')) => ('ult', ('u2u8', 'b'), 'a') */
   /* search1674_0_0 -> 4368 in the cache */
   /* search1674_0 -> 4756 in the cache */
   /* search1674_1_0 -> 4370 in the cache */
   /* search1674_1 -> 4757 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4756, 4757 },
      -1,
   } },

   /* replace1674_0_0 -> 4368 in the cache */
   /* replace1674_0 -> 4716 in the cache */
   /* replace1674_1 -> 4370 in the cache */
   /* replace1674 -> 4719 in the cache */

   /* ('ieq', ('u2u32', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('u2u8', 'b')), ('ieq', ('u2u32', ('u2u8', 'b')), 'b')) */
   /* search1675_0_0 -> 6 in the cache */
   /* search1675_0 -> 3282 in the cache */
   /* search1675_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 3282, 33 },
      -1,
   } },

   /* replace1675_0_0 -> 6 in the cache */
   /* replace1675_0_1_0 -> 36 in the cache */
   /* replace1675_0_1 -> 2027 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 6, 2027 },
      -1,
   } },
   /* replace1675_1_0_0_0 -> 36 in the cache */
   /* replace1675_1_0_0 -> 2027 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 2027 },
      -1,
   } },
   /* replace1675_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 4762, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4761, 4763 },
      -1,
   } },

   /* ('ine', ('u2u32', 'a@8'), '#b') => ('ior', ('ine', 'a', ('u2u8', 'b')), ('ine', ('u2u32', ('u2u8', 'b')), 'b')) */
   /* search1676_0_0 -> 6 in the cache */
   /* search1676_0 -> 3282 in the cache */
   /* search1676_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 3282, 33 },
      -1,
   } },

   /* replace1676_0_0 -> 6 in the cache */
   /* replace1676_0_1_0 -> 36 in the cache */
   /* replace1676_0_1 -> 2027 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 6, 2027 },
      -1,
   } },
   /* replace1676_1_0_0_0 -> 36 in the cache */
   /* replace1676_1_0_0 -> 2027 in the cache */
   /* replace1676_1_0 -> 4762 in the cache */
   /* replace1676_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 4762, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4766, 4767 },
      -1,
   } },

   /* ('ult', ('u2u32', 'a@8'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 255, 'b'), ('ult', 'a', ('u2u8', 'b')))) */
   /* search1677_0_0 -> 6 in the cache */
   /* search1677_0 -> 3282 in the cache */
   /* search1677_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 3282, 33 },
      -1,
   } },

   /* replace1677_0_0 -> 122 in the cache */
   /* replace1677_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 122, 36 },
      -1,
   } },
   /* replace1677_1_0_0 -> 1010 in the cache */
   /* replace1677_1_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 1010, 36 },
      -1,
   } },
   /* replace1677_1_1_0 -> 6 in the cache */
   /* replace1677_1_1_1_0 -> 36 in the cache */
   /* replace1677_1_1_1 -> 2027 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 6, 2027 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4771, 4772 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4770, 4773 },
      -1,
   } },

   /* ('ult', '#a', ('u2u32', 'b@8')) => ('iand', ('ult', 'a', 255), ('ior', ('ult', 'a', 0), ('ult', ('u2u8', 'a'), 'b'))) */
   /* search1678_0 -> 286 in the cache */
   /* search1678_1_0 -> 2112 in the cache */
   /* search1678_1 -> 3284 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 286, 3284 },
      -1,
   } },

   /* replace1678_0_0 -> 16 in the cache */
   /* replace1678_0_1 -> 1010 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 16, 1010 },
      -1,
   } },
   /* replace1678_1_0_0 -> 16 in the cache */
   /* replace1678_1_0_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 16, 122 },
      -1,
   } },
   /* replace1678_1_1_0_0 -> 16 in the cache */
   /* replace1678_1_1_0 -> 2025 in the cache */
   /* replace1678_1_1_1 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 2025, 2112 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4777, 4778 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4776, 4779 },
      -1,
   } },

   /* ('uge', ('u2u32', 'a@8'), '#b') => ('iand', ('uge', 255, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u8', 'b')))) */
   /* search1679_0_0 -> 6 in the cache */
   /* search1679_0 -> 3282 in the cache */
   /* search1679_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 3282, 33 },
      -1,
   } },

   /* replace1679_0_0 -> 1010 in the cache */
   /* replace1679_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 1010, 36 },
      -1,
   } },
   /* replace1679_1_0_0 -> 122 in the cache */
   /* replace1679_1_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 122, 36 },
      -1,
   } },
   /* replace1679_1_1_0 -> 6 in the cache */
   /* replace1679_1_1_1_0 -> 36 in the cache */
   /* replace1679_1_1_1 -> 2027 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 6, 2027 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4783, 4784 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4782, 4785 },
      -1,
   } },

   /* ('uge', '#a', ('u2u32', 'b@8')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 255), ('uge', ('u2u8', 'a'), 'b'))) */
   /* search1680_0 -> 286 in the cache */
   /* search1680_1_0 -> 2112 in the cache */
   /* search1680_1 -> 3284 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 286, 3284 },
      -1,
   } },

   /* replace1680_0_0 -> 16 in the cache */
   /* replace1680_0_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 16, 122 },
      -1,
   } },
   /* replace1680_1_0_0 -> 16 in the cache */
   /* replace1680_1_0_1 -> 1010 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 16, 1010 },
      -1,
   } },
   /* replace1680_1_1_0_0 -> 16 in the cache */
   /* replace1680_1_1_0 -> 2025 in the cache */
   /* replace1680_1_1_1 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 2025, 2112 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4789, 4790 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4788, 4791 },
      -1,
   } },

   /* ('ieq', ('u2u64', 'a@8'), ('u2u64', 'b@8')) => ('ieq', 'a', ('u2u8', 'b')) */
   /* search1681_0_0 -> 6 in the cache */
   /* search1681_0 -> 4335 in the cache */
   /* search1681_1_0 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 2112 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 4335, 4793 },
      -1,
   } },

   /* replace1681_0 -> 6 in the cache */
   /* replace1681_1_0 -> 2112 in the cache */
   /* replace1681_1 -> 4705 in the cache */
   /* replace1681 -> 4706 in the cache */

   /* ('ine', ('u2u64', 'a@8'), ('u2u64', 'b@8')) => ('ine', 'a', ('u2u8', 'b')) */
   /* search1682_0_0 -> 6 in the cache */
   /* search1682_0 -> 4335 in the cache */
   /* search1682_1_0 -> 2112 in the cache */
   /* search1682_1 -> 4793 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 4335, 4793 },
      -1,
   } },

   /* replace1682_0 -> 6 in the cache */
   /* replace1682_1_0 -> 2112 in the cache */
   /* replace1682_1 -> 4705 in the cache */
   /* replace1682 -> 4708 in the cache */

   /* ('uge', ('u2u64', 'a@8'), ('u2u64', 'b@8')) => ('uge', 'a', ('u2u8', 'b')) */
   /* search1683_0_0 -> 6 in the cache */
   /* search1683_0 -> 4335 in the cache */
   /* search1683_1_0 -> 2112 in the cache */
   /* search1683_1 -> 4793 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4335, 4793 },
      -1,
   } },

   /* replace1683_0 -> 6 in the cache */
   /* replace1683_1_0 -> 2112 in the cache */
   /* replace1683_1 -> 4705 in the cache */
   /* replace1683 -> 4710 in the cache */

   /* ('ult', ('u2u64', 'a@8'), ('u2u64', 'b@8')) => ('ult', 'a', ('u2u8', 'b')) */
   /* search1684_0_0 -> 6 in the cache */
   /* search1684_0 -> 4335 in the cache */
   /* search1684_1_0 -> 2112 in the cache */
   /* search1684_1 -> 4793 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4335, 4793 },
      -1,
   } },

   /* replace1684_0 -> 6 in the cache */
   /* replace1684_1_0 -> 2112 in the cache */
   /* replace1684_1 -> 4705 in the cache */
   /* replace1684 -> 4712 in the cache */

   /* ('uge', ('u2u64', 'b@8'), ('u2u64', 'a@8')) => ('uge', ('u2u8', 'b'), 'a') */
   /* search1685_0_0 -> 4368 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 4368 },
      -1,
   } },
   /* search1685_1_0 -> 4370 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 4370 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4798, 4799 },
      -1,
   } },

   /* replace1685_0_0 -> 4368 in the cache */
   /* replace1685_0 -> 4716 in the cache */
   /* replace1685_1 -> 4370 in the cache */
   /* replace1685 -> 4717 in the cache */

   /* ('ult', ('u2u64', 'b@8'), ('u2u64', 'a@8')) => ('ult', ('u2u8', 'b'), 'a') */
   /* search1686_0_0 -> 4368 in the cache */
   /* search1686_0 -> 4798 in the cache */
   /* search1686_1_0 -> 4370 in the cache */
   /* search1686_1 -> 4799 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4798, 4799 },
      -1,
   } },

   /* replace1686_0_0 -> 4368 in the cache */
   /* replace1686_0 -> 4716 in the cache */
   /* replace1686_1 -> 4370 in the cache */
   /* replace1686 -> 4719 in the cache */

   /* ('ieq', ('u2u64', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('u2u8', 'b')), ('ieq', ('u2u64', ('u2u8', 'b')), 'b')) */
   /* search1687_0_0 -> 6 in the cache */
   /* search1687_0 -> 4335 in the cache */
   /* search1687_1 -> 4469 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 4335, 4469 },
      -1,
   } },

   /* replace1687_0_0 -> 6 in the cache */
   /* replace1687_0_1_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 6, 4803 },
      -1,
   } },
   /* replace1687_1_0_0_0 -> 41 in the cache */
   /* replace1687_1_0_0 -> 4803 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 4803 },
      -1,
   } },
   /* replace1687_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 4805, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4804, 4806 },
      -1,
   } },

   /* ('ine', ('u2u64', 'a@8'), '#b') => ('ior', ('ine', 'a', ('u2u8', 'b')), ('ine', ('u2u64', ('u2u8', 'b')), 'b')) */
   /* search1688_0_0 -> 6 in the cache */
   /* search1688_0 -> 4335 in the cache */
   /* search1688_1 -> 4469 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 4335, 4469 },
      -1,
   } },

   /* replace1688_0_0 -> 6 in the cache */
   /* replace1688_0_1_0 -> 41 in the cache */
   /* replace1688_0_1 -> 4803 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 6, 4803 },
      -1,
   } },
   /* replace1688_1_0_0_0 -> 41 in the cache */
   /* replace1688_1_0_0 -> 4803 in the cache */
   /* replace1688_1_0 -> 4805 in the cache */
   /* replace1688_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 4805, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4809, 4810 },
      -1,
   } },

   /* ('ult', ('u2u64', 'a@8'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 255, 'b'), ('ult', 'a', ('u2u8', 'b')))) */
   /* search1689_0_0 -> 6 in the cache */
   /* search1689_0 -> 4335 in the cache */
   /* search1689_1 -> 4469 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4335, 4469 },
      -1,
   } },

   /* replace1689_0_0 -> 897 in the cache */
   /* replace1689_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 897, 41 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0xff /* 255 */ },
   } },
   /* replace1689_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4814, 41 },
      -1,
   } },
   /* replace1689_1_1_0 -> 6 in the cache */
   /* replace1689_1_1_1_0 -> 41 in the cache */
   /* replace1689_1_1_1 -> 4803 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 6, 4803 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4815, 4816 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4813, 4817 },
      -1,
   } },

   /* ('ult', '#a', ('u2u64', 'b@8')) => ('iand', ('ult', 'a', 255), ('ior', ('ult', 'a', 0), ('ult', ('u2u8', 'a'), 'b'))) */
   /* search1690_0 -> 4488 in the cache */
   /* search1690_1_0 -> 2112 in the cache */
   /* search1690_1 -> 4793 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4488, 4793 },
      -1,
   } },

   /* replace1690_0_0 -> 21 in the cache */
   /* replace1690_0_1 -> 4814 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 21, 4814 },
      -1,
   } },
   /* replace1690_1_0_0 -> 21 in the cache */
   /* replace1690_1_0_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 21, 897 },
      -1,
   } },
   /* replace1690_1_1_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 21 },
      -1,
   } },
   /* replace1690_1_1_1 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4822, 2112 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4821, 4823 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4820, 4824 },
      -1,
   } },

   /* ('uge', ('u2u64', 'a@8'), '#b') => ('iand', ('uge', 255, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u8', 'b')))) */
   /* search1691_0_0 -> 6 in the cache */
   /* search1691_0 -> 4335 in the cache */
   /* search1691_1 -> 4469 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4335, 4469 },
      -1,
   } },

   /* replace1691_0_0 -> 4814 in the cache */
   /* replace1691_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4814, 41 },
      -1,
   } },
   /* replace1691_1_0_0 -> 897 in the cache */
   /* replace1691_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 897, 41 },
      -1,
   } },
   /* replace1691_1_1_0 -> 6 in the cache */
   /* replace1691_1_1_1_0 -> 41 in the cache */
   /* replace1691_1_1_1 -> 4803 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 6, 4803 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4828, 4829 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4827, 4830 },
      -1,
   } },

   /* ('uge', '#a', ('u2u64', 'b@8')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 255), ('uge', ('u2u8', 'a'), 'b'))) */
   /* search1692_0 -> 4488 in the cache */
   /* search1692_1_0 -> 2112 in the cache */
   /* search1692_1 -> 4793 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4488, 4793 },
      -1,
   } },

   /* replace1692_0_0 -> 21 in the cache */
   /* replace1692_0_1 -> 897 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 21, 897 },
      -1,
   } },
   /* replace1692_1_0_0 -> 21 in the cache */
   /* replace1692_1_0_1 -> 4814 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 21, 4814 },
      -1,
   } },
   /* replace1692_1_1_0_0 -> 21 in the cache */
   /* replace1692_1_1_0 -> 4822 in the cache */
   /* replace1692_1_1_1 -> 2112 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4822, 2112 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4834, 4835 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4833, 4836 },
      -1,
   } },

   /* ('ieq', ('u2u32', 'a@16'), ('u2u32', 'b@8')) => ('ieq', 'a', ('u2u16', 'b')) */
   /* search1693_0_0 -> 11 in the cache */
   /* search1693_0 -> 1806 in the cache */
   /* search1693_1_0 -> 2112 in the cache */
   /* search1693_1 -> 3284 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1806, 3284 },
      -1,
   } },

   /* replace1693_0 -> 11 in the cache */
   /* replace1693_1_0 -> 2112 in the cache */
   /* replace1693_1 -> 4703 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 11, 4703 },
      -1,
   } },

   /* ('ine', ('u2u32', 'a@16'), ('u2u32', 'b@8')) => ('ine', 'a', ('u2u16', 'b')) */
   /* search1694_0_0 -> 11 in the cache */
   /* search1694_0 -> 1806 in the cache */
   /* search1694_1_0 -> 2112 in the cache */
   /* search1694_1 -> 3284 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1806, 3284 },
      -1,
   } },

   /* replace1694_0 -> 11 in the cache */
   /* replace1694_1_0 -> 2112 in the cache */
   /* replace1694_1 -> 4703 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 11, 4703 },
      -1,
   } },

   /* ('uge', ('u2u32', 'a@16'), ('u2u32', 'b@8')) => ('uge', 'a', ('u2u16', 'b')) */
   /* search1695_0_0 -> 11 in the cache */
   /* search1695_0 -> 1806 in the cache */
   /* search1695_1_0 -> 2112 in the cache */
   /* search1695_1 -> 3284 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 1806, 3284 },
      -1,
   } },

   /* replace1695_0 -> 11 in the cache */
   /* replace1695_1_0 -> 2112 in the cache */
   /* replace1695_1 -> 4703 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 11, 4703 },
      -1,
   } },

   /* ('ult', ('u2u32', 'a@16'), ('u2u32', 'b@8')) => ('ult', 'a', ('u2u16', 'b')) */
   /* search1696_0_0 -> 11 in the cache */
   /* search1696_0 -> 1806 in the cache */
   /* search1696_1_0 -> 2112 in the cache */
   /* search1696_1 -> 3284 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 1806, 3284 },
      -1,
   } },

   /* replace1696_0 -> 11 in the cache */
   /* replace1696_1_0 -> 2112 in the cache */
   /* replace1696_1 -> 4703 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 11, 4703 },
      -1,
   } },

   /* ('uge', ('u2u32', 'b@8'), ('u2u32', 'a@16')) => ('uge', ('u2u16', 'b'), 'a') */
   /* search1697_0_0 -> 4368 in the cache */
   /* search1697_0 -> 4756 in the cache */
   /* search1697_1_0 -> 526 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 526 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4756, 4846 },
      -1,
   } },

   /* replace1697_0_0 -> 4368 in the cache */
   /* replace1697_0 -> 4713 in the cache */
   /* replace1697_1 -> 526 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4713, 526 },
      -1,
   } },

   /* ('ult', ('u2u32', 'b@8'), ('u2u32', 'a@16')) => ('ult', ('u2u16', 'b'), 'a') */
   /* search1698_0_0 -> 4368 in the cache */
   /* search1698_0 -> 4756 in the cache */
   /* search1698_1_0 -> 526 in the cache */
   /* search1698_1 -> 4846 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4756, 4846 },
      -1,
   } },

   /* replace1698_0_0 -> 4368 in the cache */
   /* replace1698_0 -> 4713 in the cache */
   /* replace1698_1 -> 526 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4713, 526 },
      -1,
   } },

   /* ('ieq', ('u2u32', 'a@16'), ('u2u32', 'b@16')) => ('ieq', 'a', ('u2u16', 'b')) */
   /* search1699_0_0 -> 11 in the cache */
   /* search1699_0 -> 1806 in the cache */
   /* search1699_1_0 -> 258 in the cache */
   /* search1699_1 -> 3304 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1806, 3304 },
      -1,
   } },

   /* replace1699_0 -> 11 in the cache */
   /* replace1699_1_0 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 258 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 11, 4852 },
      -1,
   } },

   /* ('ine', ('u2u32', 'a@16'), ('u2u32', 'b@16')) => ('ine', 'a', ('u2u16', 'b')) */
   /* search1700_0_0 -> 11 in the cache */
   /* search1700_0 -> 1806 in the cache */
   /* search1700_1_0 -> 258 in the cache */
   /* search1700_1 -> 3304 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1806, 3304 },
      -1,
   } },

   /* replace1700_0 -> 11 in the cache */
   /* replace1700_1_0 -> 258 in the cache */
   /* replace1700_1 -> 4852 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 11, 4852 },
      -1,
   } },

   /* ('uge', ('u2u32', 'a@16'), ('u2u32', 'b@16')) => ('uge', 'a', ('u2u16', 'b')) */
   /* search1701_0_0 -> 11 in the cache */
   /* search1701_0 -> 1806 in the cache */
   /* search1701_1_0 -> 258 in the cache */
   /* search1701_1 -> 3304 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 1806, 3304 },
      -1,
   } },

   /* replace1701_0 -> 11 in the cache */
   /* replace1701_1_0 -> 258 in the cache */
   /* replace1701_1 -> 4852 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 11, 4852 },
      -1,
   } },

   /* ('ult', ('u2u32', 'a@16'), ('u2u32', 'b@16')) => ('ult', 'a', ('u2u16', 'b')) */
   /* search1702_0_0 -> 11 in the cache */
   /* search1702_0 -> 1806 in the cache */
   /* search1702_1_0 -> 258 in the cache */
   /* search1702_1 -> 3304 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 1806, 3304 },
      -1,
   } },

   /* replace1702_0 -> 11 in the cache */
   /* replace1702_1_0 -> 258 in the cache */
   /* replace1702_1 -> 4852 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 11, 4852 },
      -1,
   } },

   /* ('uge', ('u2u32', 'b@16'), ('u2u32', 'a@16')) => ('uge', ('u2u16', 'b'), 'a') */
   /* search1703_0_0 -> 515 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 515 },
      -1,
   } },
   /* search1703_1_0 -> 526 in the cache */
   /* search1703_1 -> 4846 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4860, 4846 },
      -1,
   } },

   /* replace1703_0_0 -> 515 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 515 },
      -1,
   } },
   /* replace1703_1 -> 526 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4862, 526 },
      -1,
   } },

   /* ('ult', ('u2u32', 'b@16'), ('u2u32', 'a@16')) => ('ult', ('u2u16', 'b'), 'a') */
   /* search1704_0_0 -> 515 in the cache */
   /* search1704_0 -> 4860 in the cache */
   /* search1704_1_0 -> 526 in the cache */
   /* search1704_1 -> 4846 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4860, 4846 },
      -1,
   } },

   /* replace1704_0_0 -> 515 in the cache */
   /* replace1704_0 -> 4862 in the cache */
   /* replace1704_1 -> 526 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4862, 526 },
      -1,
   } },

   /* ('ieq', ('u2u32', 'a@16'), '#b') => ('iand', ('ieq', 'a', ('u2u16', 'b')), ('ieq', ('u2u32', ('u2u16', 'b')), 'b')) */
   /* search1705_0_0 -> 11 in the cache */
   /* search1705_0 -> 1806 in the cache */
   /* search1705_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1806, 33 },
      -1,
   } },

   /* replace1705_0_0 -> 11 in the cache */
   /* replace1705_0_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 11, 4867 },
      -1,
   } },
   /* replace1705_1_0_0_0 -> 36 in the cache */
   /* replace1705_1_0_0 -> 4867 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 4867 },
      -1,
   } },
   /* replace1705_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 4869, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4868, 4870 },
      -1,
   } },

   /* ('ine', ('u2u32', 'a@16'), '#b') => ('ior', ('ine', 'a', ('u2u16', 'b')), ('ine', ('u2u32', ('u2u16', 'b')), 'b')) */
   /* search1706_0_0 -> 11 in the cache */
   /* search1706_0 -> 1806 in the cache */
   /* search1706_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1806, 33 },
      -1,
   } },

   /* replace1706_0_0 -> 11 in the cache */
   /* replace1706_0_1_0 -> 36 in the cache */
   /* replace1706_0_1 -> 4867 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 11, 4867 },
      -1,
   } },
   /* replace1706_1_0_0_0 -> 36 in the cache */
   /* replace1706_1_0_0 -> 4867 in the cache */
   /* replace1706_1_0 -> 4869 in the cache */
   /* replace1706_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 4869, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4873, 4874 },
      -1,
   } },

   /* ('ult', ('u2u32', 'a@16'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 65535, 'b'), ('ult', 'a', ('u2u16', 'b')))) */
   /* search1707_0_0 -> 11 in the cache */
   /* search1707_0 -> 1806 in the cache */
   /* search1707_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 1806, 33 },
      -1,
   } },

   /* replace1707_0_0 -> 122 in the cache */
   /* replace1707_0_1 -> 36 in the cache */
   /* replace1707_0 -> 4770 in the cache */
   /* replace1707_1_0_0 -> 1002 in the cache */
   /* replace1707_1_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 1002, 36 },
      -1,
   } },
   /* replace1707_1_1_0 -> 11 in the cache */
   /* replace1707_1_1_1_0 -> 36 in the cache */
   /* replace1707_1_1_1 -> 4867 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 11, 4867 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4877, 4878 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4770, 4879 },
      -1,
   } },

   /* ('ult', '#a', ('u2u32', 'b@16')) => ('iand', ('ult', 'a', 65535), ('ior', ('ult', 'a', 0), ('ult', ('u2u16', 'a'), 'b'))) */
   /* search1708_0 -> 286 in the cache */
   /* search1708_1_0 -> 258 in the cache */
   /* search1708_1 -> 3304 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 286, 3304 },
      -1,
   } },

   /* replace1708_0_0 -> 16 in the cache */
   /* replace1708_0_1 -> 1002 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 16, 1002 },
      -1,
   } },
   /* replace1708_1_0_0 -> 16 in the cache */
   /* replace1708_1_0_1 -> 122 in the cache */
   /* replace1708_1_0 -> 4777 in the cache */
   /* replace1708_1_1_0_0 -> 16 in the cache */
   /* replace1708_1_1_0 -> 3310 in the cache */
   /* replace1708_1_1_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 3310, 258 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4777, 4883 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4882, 4884 },
      -1,
   } },

   /* ('uge', ('u2u32', 'a@16'), '#b') => ('iand', ('uge', 65535, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u16', 'b')))) */
   /* search1709_0_0 -> 11 in the cache */
   /* search1709_0 -> 1806 in the cache */
   /* search1709_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 1806, 33 },
      -1,
   } },

   /* replace1709_0_0 -> 1002 in the cache */
   /* replace1709_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 1002, 36 },
      -1,
   } },
   /* replace1709_1_0_0 -> 122 in the cache */
   /* replace1709_1_0_1 -> 36 in the cache */
   /* replace1709_1_0 -> 4783 in the cache */
   /* replace1709_1_1_0 -> 11 in the cache */
   /* replace1709_1_1_1_0 -> 36 in the cache */
   /* replace1709_1_1_1 -> 4867 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 11, 4867 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4783, 4888 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4887, 4889 },
      -1,
   } },

   /* ('uge', '#a', ('u2u32', 'b@16')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 65535), ('uge', ('u2u16', 'a'), 'b'))) */
   /* search1710_0 -> 286 in the cache */
   /* search1710_1_0 -> 258 in the cache */
   /* search1710_1 -> 3304 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 286, 3304 },
      -1,
   } },

   /* replace1710_0_0 -> 16 in the cache */
   /* replace1710_0_1 -> 122 in the cache */
   /* replace1710_0 -> 4788 in the cache */
   /* replace1710_1_0_0 -> 16 in the cache */
   /* replace1710_1_0_1 -> 1002 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 16, 1002 },
      -1,
   } },
   /* replace1710_1_1_0_0 -> 16 in the cache */
   /* replace1710_1_1_0 -> 3310 in the cache */
   /* replace1710_1_1_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 3310, 258 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4892, 4893 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4788, 4894 },
      -1,
   } },

   /* ('ieq', ('u2u64', 'a@16'), ('u2u64', 'b@8')) => ('ieq', 'a', ('u2u16', 'b')) */
   /* search1711_0_0 -> 11 in the cache */
   /* search1711_0 -> 1829 in the cache */
   /* search1711_1_0 -> 2112 in the cache */
   /* search1711_1 -> 4793 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1829, 4793 },
      -1,
   } },

   /* replace1711_0 -> 11 in the cache */
   /* replace1711_1_0 -> 2112 in the cache */
   /* replace1711_1 -> 4703 in the cache */
   /* replace1711 -> 4839 in the cache */

   /* ('ine', ('u2u64', 'a@16'), ('u2u64', 'b@8')) => ('ine', 'a', ('u2u16', 'b')) */
   /* search1712_0_0 -> 11 in the cache */
   /* search1712_0 -> 1829 in the cache */
   /* search1712_1_0 -> 2112 in the cache */
   /* search1712_1 -> 4793 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1829, 4793 },
      -1,
   } },

   /* replace1712_0 -> 11 in the cache */
   /* replace1712_1_0 -> 2112 in the cache */
   /* replace1712_1 -> 4703 in the cache */
   /* replace1712 -> 4841 in the cache */

   /* ('uge', ('u2u64', 'a@16'), ('u2u64', 'b@8')) => ('uge', 'a', ('u2u16', 'b')) */
   /* search1713_0_0 -> 11 in the cache */
   /* search1713_0 -> 1829 in the cache */
   /* search1713_1_0 -> 2112 in the cache */
   /* search1713_1 -> 4793 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 1829, 4793 },
      -1,
   } },

   /* replace1713_0 -> 11 in the cache */
   /* replace1713_1_0 -> 2112 in the cache */
   /* replace1713_1 -> 4703 in the cache */
   /* replace1713 -> 4843 in the cache */

   /* ('ult', ('u2u64', 'a@16'), ('u2u64', 'b@8')) => ('ult', 'a', ('u2u16', 'b')) */
   /* search1714_0_0 -> 11 in the cache */
   /* search1714_0 -> 1829 in the cache */
   /* search1714_1_0 -> 2112 in the cache */
   /* search1714_1 -> 4793 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 1829, 4793 },
      -1,
   } },

   /* replace1714_0 -> 11 in the cache */
   /* replace1714_1_0 -> 2112 in the cache */
   /* replace1714_1 -> 4703 in the cache */
   /* replace1714 -> 4845 in the cache */

   /* ('uge', ('u2u64', 'b@8'), ('u2u64', 'a@16')) => ('uge', ('u2u16', 'b'), 'a') */
   /* search1715_0_0 -> 4368 in the cache */
   /* search1715_0 -> 4798 in the cache */
   /* search1715_1_0 -> 526 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 526 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4798, 4900 },
      -1,
   } },

   /* replace1715_0_0 -> 4368 in the cache */
   /* replace1715_0 -> 4713 in the cache */
   /* replace1715_1 -> 526 in the cache */
   /* replace1715 -> 4848 in the cache */

   /* ('ult', ('u2u64', 'b@8'), ('u2u64', 'a@16')) => ('ult', ('u2u16', 'b'), 'a') */
   /* search1716_0_0 -> 4368 in the cache */
   /* search1716_0 -> 4798 in the cache */
   /* search1716_1_0 -> 526 in the cache */
   /* search1716_1 -> 4900 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4798, 4900 },
      -1,
   } },

   /* replace1716_0_0 -> 4368 in the cache */
   /* replace1716_0 -> 4713 in the cache */
   /* replace1716_1 -> 526 in the cache */
   /* replace1716 -> 4850 in the cache */

   /* ('ieq', ('u2u64', 'a@16'), ('u2u64', 'b@16')) => ('ieq', 'a', ('u2u16', 'b')) */
   /* search1717_0_0 -> 11 in the cache */
   /* search1717_0 -> 1829 in the cache */
   /* search1717_1_0 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 258 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1829, 4903 },
      -1,
   } },

   /* replace1717_0 -> 11 in the cache */
   /* replace1717_1_0 -> 258 in the cache */
   /* replace1717_1 -> 4852 in the cache */
   /* replace1717 -> 4853 in the cache */

   /* ('ine', ('u2u64', 'a@16'), ('u2u64', 'b@16')) => ('ine', 'a', ('u2u16', 'b')) */
   /* search1718_0_0 -> 11 in the cache */
   /* search1718_0 -> 1829 in the cache */
   /* search1718_1_0 -> 258 in the cache */
   /* search1718_1 -> 4903 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1829, 4903 },
      -1,
   } },

   /* replace1718_0 -> 11 in the cache */
   /* replace1718_1_0 -> 258 in the cache */
   /* replace1718_1 -> 4852 in the cache */
   /* replace1718 -> 4855 in the cache */

   /* ('uge', ('u2u64', 'a@16'), ('u2u64', 'b@16')) => ('uge', 'a', ('u2u16', 'b')) */
   /* search1719_0_0 -> 11 in the cache */
   /* search1719_0 -> 1829 in the cache */
   /* search1719_1_0 -> 258 in the cache */
   /* search1719_1 -> 4903 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 1829, 4903 },
      -1,
   } },

   /* replace1719_0 -> 11 in the cache */
   /* replace1719_1_0 -> 258 in the cache */
   /* replace1719_1 -> 4852 in the cache */
   /* replace1719 -> 4857 in the cache */

   /* ('ult', ('u2u64', 'a@16'), ('u2u64', 'b@16')) => ('ult', 'a', ('u2u16', 'b')) */
   /* search1720_0_0 -> 11 in the cache */
   /* search1720_0 -> 1829 in the cache */
   /* search1720_1_0 -> 258 in the cache */
   /* search1720_1 -> 4903 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 1829, 4903 },
      -1,
   } },

   /* replace1720_0 -> 11 in the cache */
   /* replace1720_1_0 -> 258 in the cache */
   /* replace1720_1 -> 4852 in the cache */
   /* replace1720 -> 4859 in the cache */

   /* ('uge', ('u2u64', 'b@16'), ('u2u64', 'a@16')) => ('uge', ('u2u16', 'b'), 'a') */
   /* search1721_0_0 -> 515 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 515 },
      -1,
   } },
   /* search1721_1_0 -> 526 in the cache */
   /* search1721_1 -> 4900 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4908, 4900 },
      -1,
   } },

   /* replace1721_0_0 -> 515 in the cache */
   /* replace1721_0 -> 4862 in the cache */
   /* replace1721_1 -> 526 in the cache */
   /* replace1721 -> 4863 in the cache */

   /* ('ult', ('u2u64', 'b@16'), ('u2u64', 'a@16')) => ('ult', ('u2u16', 'b'), 'a') */
   /* search1722_0_0 -> 515 in the cache */
   /* search1722_0 -> 4908 in the cache */
   /* search1722_1_0 -> 526 in the cache */
   /* search1722_1 -> 4900 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4908, 4900 },
      -1,
   } },

   /* replace1722_0_0 -> 515 in the cache */
   /* replace1722_0 -> 4862 in the cache */
   /* replace1722_1 -> 526 in the cache */
   /* replace1722 -> 4865 in the cache */

   /* ('ieq', ('u2u64', 'a@16'), '#b') => ('iand', ('ieq', 'a', ('u2u16', 'b')), ('ieq', ('u2u64', ('u2u16', 'b')), 'b')) */
   /* search1723_0_0 -> 11 in the cache */
   /* search1723_0 -> 1829 in the cache */
   /* search1723_1 -> 4469 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1829, 4469 },
      -1,
   } },

   /* replace1723_0_0 -> 11 in the cache */
   /* replace1723_0_1_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 11, 4912 },
      -1,
   } },
   /* replace1723_1_0_0_0 -> 41 in the cache */
   /* replace1723_1_0_0 -> 4912 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 4912 },
      -1,
   } },
   /* replace1723_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 4914, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4913, 4915 },
      -1,
   } },

   /* ('ine', ('u2u64', 'a@16'), '#b') => ('ior', ('ine', 'a', ('u2u16', 'b')), ('ine', ('u2u64', ('u2u16', 'b')), 'b')) */
   /* search1724_0_0 -> 11 in the cache */
   /* search1724_0 -> 1829 in the cache */
   /* search1724_1 -> 4469 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1829, 4469 },
      -1,
   } },

   /* replace1724_0_0 -> 11 in the cache */
   /* replace1724_0_1_0 -> 41 in the cache */
   /* replace1724_0_1 -> 4912 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 11, 4912 },
      -1,
   } },
   /* replace1724_1_0_0_0 -> 41 in the cache */
   /* replace1724_1_0_0 -> 4912 in the cache */
   /* replace1724_1_0 -> 4914 in the cache */
   /* replace1724_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 4914, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4918, 4919 },
      -1,
   } },

   /* ('ult', ('u2u64', 'a@16'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 65535, 'b'), ('ult', 'a', ('u2u16', 'b')))) */
   /* search1725_0_0 -> 11 in the cache */
   /* search1725_0 -> 1829 in the cache */
   /* search1725_1 -> 4469 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 1829, 4469 },
      -1,
   } },

   /* replace1725_0_0 -> 897 in the cache */
   /* replace1725_0_1 -> 41 in the cache */
   /* replace1725_0 -> 4813 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0xffff /* 65535 */ },
   } },
   /* replace1725_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4922, 41 },
      -1,
   } },
   /* replace1725_1_1_0 -> 11 in the cache */
   /* replace1725_1_1_1_0 -> 41 in the cache */
   /* replace1725_1_1_1 -> 4912 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 11, 4912 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4923, 4924 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4813, 4925 },
      -1,
   } },

   /* ('ult', '#a', ('u2u64', 'b@16')) => ('iand', ('ult', 'a', 65535), ('ior', ('ult', 'a', 0), ('ult', ('u2u16', 'a'), 'b'))) */
   /* search1726_0 -> 4488 in the cache */
   /* search1726_1_0 -> 258 in the cache */
   /* search1726_1 -> 4903 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4488, 4903 },
      -1,
   } },

   /* replace1726_0_0 -> 21 in the cache */
   /* replace1726_0_1 -> 4922 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 21, 4922 },
      -1,
   } },
   /* replace1726_1_0_0 -> 21 in the cache */
   /* replace1726_1_0_1 -> 897 in the cache */
   /* replace1726_1_0 -> 4821 in the cache */
   /* replace1726_1_1_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 21 },
      -1,
   } },
   /* replace1726_1_1_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4929, 258 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4821, 4930 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4928, 4931 },
      -1,
   } },

   /* ('uge', ('u2u64', 'a@16'), '#b') => ('iand', ('uge', 65535, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u16', 'b')))) */
   /* search1727_0_0 -> 11 in the cache */
   /* search1727_0 -> 1829 in the cache */
   /* search1727_1 -> 4469 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 1829, 4469 },
      -1,
   } },

   /* replace1727_0_0 -> 4922 in the cache */
   /* replace1727_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4922, 41 },
      -1,
   } },
   /* replace1727_1_0_0 -> 897 in the cache */
   /* replace1727_1_0_1 -> 41 in the cache */
   /* replace1727_1_0 -> 4828 in the cache */
   /* replace1727_1_1_0 -> 11 in the cache */
   /* replace1727_1_1_1_0 -> 41 in the cache */
   /* replace1727_1_1_1 -> 4912 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 11, 4912 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4828, 4935 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4934, 4936 },
      -1,
   } },

   /* ('uge', '#a', ('u2u64', 'b@16')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 65535), ('uge', ('u2u16', 'a'), 'b'))) */
   /* search1728_0 -> 4488 in the cache */
   /* search1728_1_0 -> 258 in the cache */
   /* search1728_1 -> 4903 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4488, 4903 },
      -1,
   } },

   /* replace1728_0_0 -> 21 in the cache */
   /* replace1728_0_1 -> 897 in the cache */
   /* replace1728_0 -> 4833 in the cache */
   /* replace1728_1_0_0 -> 21 in the cache */
   /* replace1728_1_0_1 -> 4922 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 21, 4922 },
      -1,
   } },
   /* replace1728_1_1_0_0 -> 21 in the cache */
   /* replace1728_1_1_0 -> 4929 in the cache */
   /* replace1728_1_1_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4929, 258 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4939, 4940 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4833, 4941 },
      -1,
   } },

   /* ('ieq', ('u2u64', 'a@32'), ('u2u64', 'b@8')) => ('ieq', 'a', ('u2u32', 'b')) */
   /* search1729_0_0 -> 16 in the cache */
   /* search1729_0 -> 1911 in the cache */
   /* search1729_1_0 -> 2112 in the cache */
   /* search1729_1 -> 4793 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1911, 4793 },
      -1,
   } },

   /* replace1729_0 -> 16 in the cache */
   /* replace1729_1_0 -> 2112 in the cache */
   /* replace1729_1 -> 3284 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 16, 3284 },
      -1,
   } },

   /* ('ine', ('u2u64', 'a@32'), ('u2u64', 'b@8')) => ('ine', 'a', ('u2u32', 'b')) */
   /* search1730_0_0 -> 16 in the cache */
   /* search1730_0 -> 1911 in the cache */
   /* search1730_1_0 -> 2112 in the cache */
   /* search1730_1 -> 4793 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1911, 4793 },
      -1,
   } },

   /* replace1730_0 -> 16 in the cache */
   /* replace1730_1_0 -> 2112 in the cache */
   /* replace1730_1 -> 3284 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 16, 3284 },
      -1,
   } },

   /* ('uge', ('u2u64', 'a@32'), ('u2u64', 'b@8')) => ('uge', 'a', ('u2u32', 'b')) */
   /* search1731_0_0 -> 16 in the cache */
   /* search1731_0 -> 1911 in the cache */
   /* search1731_1_0 -> 2112 in the cache */
   /* search1731_1 -> 4793 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 1911, 4793 },
      -1,
   } },

   /* replace1731_0 -> 16 in the cache */
   /* replace1731_1_0 -> 2112 in the cache */
   /* replace1731_1 -> 3284 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 16, 3284 },
      -1,
   } },

   /* ('ult', ('u2u64', 'a@32'), ('u2u64', 'b@8')) => ('ult', 'a', ('u2u32', 'b')) */
   /* search1732_0_0 -> 16 in the cache */
   /* search1732_0 -> 1911 in the cache */
   /* search1732_1_0 -> 2112 in the cache */
   /* search1732_1 -> 4793 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 1911, 4793 },
      -1,
   } },

   /* replace1732_0 -> 16 in the cache */
   /* replace1732_1_0 -> 2112 in the cache */
   /* replace1732_1 -> 3284 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 16, 3284 },
      -1,
   } },

   /* ('uge', ('u2u64', 'b@8'), ('u2u64', 'a@32')) => ('uge', ('u2u32', 'b'), 'a') */
   /* search1733_0_0 -> 4368 in the cache */
   /* search1733_0 -> 4798 in the cache */
   /* search1733_1_0 -> 183 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 183 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4798, 4951 },
      -1,
   } },

   /* replace1733_0_0 -> 4368 in the cache */
   /* replace1733_0 -> 4756 in the cache */
   /* replace1733_1 -> 183 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4756, 183 },
      -1,
   } },

   /* ('ult', ('u2u64', 'b@8'), ('u2u64', 'a@32')) => ('ult', ('u2u32', 'b'), 'a') */
   /* search1734_0_0 -> 4368 in the cache */
   /* search1734_0 -> 4798 in the cache */
   /* search1734_1_0 -> 183 in the cache */
   /* search1734_1 -> 4951 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4798, 4951 },
      -1,
   } },

   /* replace1734_0_0 -> 4368 in the cache */
   /* replace1734_0 -> 4756 in the cache */
   /* replace1734_1 -> 183 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4756, 183 },
      -1,
   } },

   /* ('ieq', ('u2u64', 'a@32'), ('u2u64', 'b@16')) => ('ieq', 'a', ('u2u32', 'b')) */
   /* search1735_0_0 -> 16 in the cache */
   /* search1735_0 -> 1911 in the cache */
   /* search1735_1_0 -> 258 in the cache */
   /* search1735_1 -> 4903 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1911, 4903 },
      -1,
   } },

   /* replace1735_0 -> 16 in the cache */
   /* replace1735_1_0 -> 258 in the cache */
   /* replace1735_1 -> 3304 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 16, 3304 },
      -1,
   } },

   /* ('ine', ('u2u64', 'a@32'), ('u2u64', 'b@16')) => ('ine', 'a', ('u2u32', 'b')) */
   /* search1736_0_0 -> 16 in the cache */
   /* search1736_0 -> 1911 in the cache */
   /* search1736_1_0 -> 258 in the cache */
   /* search1736_1 -> 4903 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1911, 4903 },
      -1,
   } },

   /* replace1736_0 -> 16 in the cache */
   /* replace1736_1_0 -> 258 in the cache */
   /* replace1736_1 -> 3304 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 16, 3304 },
      -1,
   } },

   /* ('uge', ('u2u64', 'a@32'), ('u2u64', 'b@16')) => ('uge', 'a', ('u2u32', 'b')) */
   /* search1737_0_0 -> 16 in the cache */
   /* search1737_0 -> 1911 in the cache */
   /* search1737_1_0 -> 258 in the cache */
   /* search1737_1 -> 4903 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 1911, 4903 },
      -1,
   } },

   /* replace1737_0 -> 16 in the cache */
   /* replace1737_1_0 -> 258 in the cache */
   /* replace1737_1 -> 3304 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 16, 3304 },
      -1,
   } },

   /* ('ult', ('u2u64', 'a@32'), ('u2u64', 'b@16')) => ('ult', 'a', ('u2u32', 'b')) */
   /* search1738_0_0 -> 16 in the cache */
   /* search1738_0 -> 1911 in the cache */
   /* search1738_1_0 -> 258 in the cache */
   /* search1738_1 -> 4903 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 1911, 4903 },
      -1,
   } },

   /* replace1738_0 -> 16 in the cache */
   /* replace1738_1_0 -> 258 in the cache */
   /* replace1738_1 -> 3304 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 16, 3304 },
      -1,
   } },

   /* ('uge', ('u2u64', 'b@16'), ('u2u64', 'a@32')) => ('uge', ('u2u32', 'b'), 'a') */
   /* search1739_0_0 -> 515 in the cache */
   /* search1739_0 -> 4908 in the cache */
   /* search1739_1_0 -> 183 in the cache */
   /* search1739_1 -> 4951 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4908, 4951 },
      -1,
   } },

   /* replace1739_0_0 -> 515 in the cache */
   /* replace1739_0 -> 4860 in the cache */
   /* replace1739_1 -> 183 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4860, 183 },
      -1,
   } },

   /* ('ult', ('u2u64', 'b@16'), ('u2u64', 'a@32')) => ('ult', ('u2u32', 'b'), 'a') */
   /* search1740_0_0 -> 515 in the cache */
   /* search1740_0 -> 4908 in the cache */
   /* search1740_1_0 -> 183 in the cache */
   /* search1740_1 -> 4951 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4908, 4951 },
      -1,
   } },

   /* replace1740_0_0 -> 515 in the cache */
   /* replace1740_0 -> 4860 in the cache */
   /* replace1740_1 -> 183 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4860, 183 },
      -1,
   } },

   /* ('ieq', ('u2u64', 'a@32'), ('u2u64', 'b@32')) => ('ieq', 'a', ('u2u32', 'b')) */
   /* search1741_0_0 -> 16 in the cache */
   /* search1741_0 -> 1911 in the cache */
   /* search1741_1_0 -> 36 in the cache */
   /* search1741_1 -> 3300 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1911, 3300 },
      -1,
   } },

   /* replace1741_0 -> 16 in the cache */
   /* replace1741_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 16, 4969 },
      -1,
   } },

   /* ('ine', ('u2u64', 'a@32'), ('u2u64', 'b@32')) => ('ine', 'a', ('u2u32', 'b')) */
   /* search1742_0_0 -> 16 in the cache */
   /* search1742_0 -> 1911 in the cache */
   /* search1742_1_0 -> 36 in the cache */
   /* search1742_1 -> 3300 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1911, 3300 },
      -1,
   } },

   /* replace1742_0 -> 16 in the cache */
   /* replace1742_1_0 -> 36 in the cache */
   /* replace1742_1 -> 4969 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 16, 4969 },
      -1,
   } },

   /* ('uge', ('u2u64', 'a@32'), ('u2u64', 'b@32')) => ('uge', 'a', ('u2u32', 'b')) */
   /* search1743_0_0 -> 16 in the cache */
   /* search1743_0 -> 1911 in the cache */
   /* search1743_1_0 -> 36 in the cache */
   /* search1743_1 -> 3300 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 1911, 3300 },
      -1,
   } },

   /* replace1743_0 -> 16 in the cache */
   /* replace1743_1_0 -> 36 in the cache */
   /* replace1743_1 -> 4969 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 16, 4969 },
      -1,
   } },

   /* ('ult', ('u2u64', 'a@32'), ('u2u64', 'b@32')) => ('ult', 'a', ('u2u32', 'b')) */
   /* search1744_0_0 -> 16 in the cache */
   /* search1744_0 -> 1911 in the cache */
   /* search1744_1_0 -> 36 in the cache */
   /* search1744_1 -> 3300 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 1911, 3300 },
      -1,
   } },

   /* replace1744_0 -> 16 in the cache */
   /* replace1744_1_0 -> 36 in the cache */
   /* replace1744_1 -> 4969 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 16, 4969 },
      -1,
   } },

   /* ('uge', ('u2u64', 'b@32'), ('u2u64', 'a@32')) => ('uge', ('u2u32', 'b'), 'a') */
   /* search1745_0_0 -> 373 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 373 },
      -1,
   } },
   /* search1745_1_0 -> 183 in the cache */
   /* search1745_1 -> 4951 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4977, 4951 },
      -1,
   } },

   /* replace1745_0_0 -> 373 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 373 },
      -1,
   } },
   /* replace1745_1 -> 183 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4979, 183 },
      -1,
   } },

   /* ('ult', ('u2u64', 'b@32'), ('u2u64', 'a@32')) => ('ult', ('u2u32', 'b'), 'a') */
   /* search1746_0_0 -> 373 in the cache */
   /* search1746_0 -> 4977 in the cache */
   /* search1746_1_0 -> 183 in the cache */
   /* search1746_1 -> 4951 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4977, 4951 },
      -1,
   } },

   /* replace1746_0_0 -> 373 in the cache */
   /* replace1746_0 -> 4979 in the cache */
   /* replace1746_1 -> 183 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4979, 183 },
      -1,
   } },

   /* ('ieq', ('u2u64', 'a@32'), '#b') => ('iand', ('ieq', 'a', ('u2u32', 'b')), ('ieq', ('u2u64', ('u2u32', 'b')), 'b')) */
   /* search1747_0_0 -> 16 in the cache */
   /* search1747_0 -> 1911 in the cache */
   /* search1747_1 -> 4469 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1911, 4469 },
      -1,
   } },

   /* replace1747_0_0 -> 16 in the cache */
   /* replace1747_0_1_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 16, 4984 },
      -1,
   } },
   /* replace1747_1_0_0_0 -> 41 in the cache */
   /* replace1747_1_0_0 -> 4984 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 4984 },
      -1,
   } },
   /* replace1747_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 4986, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4985, 4987 },
      -1,
   } },

   /* ('ine', ('u2u64', 'a@32'), '#b') => ('ior', ('ine', 'a', ('u2u32', 'b')), ('ine', ('u2u64', ('u2u32', 'b')), 'b')) */
   /* search1748_0_0 -> 16 in the cache */
   /* search1748_0 -> 1911 in the cache */
   /* search1748_1 -> 4469 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1911, 4469 },
      -1,
   } },

   /* replace1748_0_0 -> 16 in the cache */
   /* replace1748_0_1_0 -> 41 in the cache */
   /* replace1748_0_1 -> 4984 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 16, 4984 },
      -1,
   } },
   /* replace1748_1_0_0_0 -> 41 in the cache */
   /* replace1748_1_0_0 -> 4984 in the cache */
   /* replace1748_1_0 -> 4986 in the cache */
   /* replace1748_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 4986, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4990, 4991 },
      -1,
   } },

   /* ('ult', ('u2u64', 'a@32'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 4294967295, 'b'), ('ult', 'a', ('u2u32', 'b')))) */
   /* search1749_0_0 -> 16 in the cache */
   /* search1749_0 -> 1911 in the cache */
   /* search1749_1 -> 4469 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 1911, 4469 },
      -1,
   } },

   /* replace1749_0_0 -> 897 in the cache */
   /* replace1749_0_1 -> 41 in the cache */
   /* replace1749_0 -> 4813 in the cache */
   /* replace1749_1_0_0 -> 3315 in the cache */
   /* replace1749_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 3315, 41 },
      -1,
   } },
   /* replace1749_1_1_0 -> 16 in the cache */
   /* replace1749_1_1_1_0 -> 41 in the cache */
   /* replace1749_1_1_1 -> 4984 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 16, 4984 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4994, 4995 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4813, 4996 },
      -1,
   } },

   /* ('ult', '#a', ('u2u64', 'b@32')) => ('iand', ('ult', 'a', 4294967295), ('ior', ('ult', 'a', 0), ('ult', ('u2u32', 'a'), 'b'))) */
   /* search1750_0 -> 4488 in the cache */
   /* search1750_1_0 -> 36 in the cache */
   /* search1750_1 -> 3300 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4488, 3300 },
      -1,
   } },

   /* replace1750_0_0 -> 21 in the cache */
   /* replace1750_0_1 -> 3315 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 21, 3315 },
      -1,
   } },
   /* replace1750_1_0_0 -> 21 in the cache */
   /* replace1750_1_0_1 -> 897 in the cache */
   /* replace1750_1_0 -> 4821 in the cache */
   /* replace1750_1_1_0_0 -> 21 in the cache */
   /* replace1750_1_1_0 -> 3307 in the cache */
   /* replace1750_1_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 3307, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4821, 5000 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4999, 5001 },
      -1,
   } },

   /* ('uge', ('u2u64', 'a@32'), '#b') => ('iand', ('uge', 4294967295, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u32', 'b')))) */
   /* search1751_0_0 -> 16 in the cache */
   /* search1751_0 -> 1911 in the cache */
   /* search1751_1 -> 4469 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 1911, 4469 },
      -1,
   } },

   /* replace1751_0_0 -> 3315 in the cache */
   /* replace1751_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 3315, 41 },
      -1,
   } },
   /* replace1751_1_0_0 -> 897 in the cache */
   /* replace1751_1_0_1 -> 41 in the cache */
   /* replace1751_1_0 -> 4828 in the cache */
   /* replace1751_1_1_0 -> 16 in the cache */
   /* replace1751_1_1_1_0 -> 41 in the cache */
   /* replace1751_1_1_1 -> 4984 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 16, 4984 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4828, 5005 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 5004, 5006 },
      -1,
   } },

   /* ('uge', '#a', ('u2u64', 'b@32')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 4294967295), ('uge', ('u2u32', 'a'), 'b'))) */
   /* search1752_0 -> 4488 in the cache */
   /* search1752_1_0 -> 36 in the cache */
   /* search1752_1 -> 3300 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4488, 3300 },
      -1,
   } },

   /* replace1752_0_0 -> 21 in the cache */
   /* replace1752_0_1 -> 897 in the cache */
   /* replace1752_0 -> 4833 in the cache */
   /* replace1752_1_0_0 -> 21 in the cache */
   /* replace1752_1_0_1 -> 3315 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 21, 3315 },
      -1,
   } },
   /* replace1752_1_1_0_0 -> 21 in the cache */
   /* replace1752_1_1_0 -> 3307 in the cache */
   /* replace1752_1_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 3307, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 5009, 5010 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4833, 5011 },
      -1,
   } },

   /* ('feq', ('f2f32', 'a@16'), ('f2f32', 'b@16')) => ('feq', 'a', ('f2f16', 'b')) */
   /* search1753_0_0 -> 11 in the cache */
   /* search1753_0 -> 1802 in the cache */
   /* search1753_1_0 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 258 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1802, 5013 },
      -1,
   } },

   /* replace1753_0 -> 11 in the cache */
   /* replace1753_1_0 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 258 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 11, 5015 },
      -1,
   } },

   /* ('fneu', ('f2f32', 'a@16'), ('f2f32', 'b@16')) => ('fneu', 'a', ('f2f16', 'b')) */
   /* search1754_0_0 -> 11 in the cache */
   /* search1754_0 -> 1802 in the cache */
   /* search1754_1_0 -> 258 in the cache */
   /* search1754_1 -> 5013 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1802, 5013 },
      -1,
   } },

   /* replace1754_0 -> 11 in the cache */
   /* replace1754_1_0 -> 258 in the cache */
   /* replace1754_1 -> 5015 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 11, 5015 },
      -1,
   } },

   /* ('fge', ('f2f32', 'a@16'), ('f2f32', 'b@16')) => ('fge', 'a', ('f2f16', 'b')) */
   /* search1755_0_0 -> 11 in the cache */
   /* search1755_0 -> 1802 in the cache */
   /* search1755_1_0 -> 258 in the cache */
   /* search1755_1 -> 5013 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1802, 5013 },
      -1,
   } },

   /* replace1755_0 -> 11 in the cache */
   /* replace1755_1_0 -> 258 in the cache */
   /* replace1755_1 -> 5015 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 11, 5015 },
      -1,
   } },

   /* ('flt', ('f2f32', 'a@16'), ('f2f32', 'b@16')) => ('flt', 'a', ('f2f16', 'b')) */
   /* search1756_0_0 -> 11 in the cache */
   /* search1756_0 -> 1802 in the cache */
   /* search1756_1_0 -> 258 in the cache */
   /* search1756_1 -> 5013 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1802, 5013 },
      -1,
   } },

   /* replace1756_0 -> 11 in the cache */
   /* replace1756_1_0 -> 258 in the cache */
   /* replace1756_1 -> 5015 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 11, 5015 },
      -1,
   } },

   /* ('fge', ('f2f32', 'b@16'), ('f2f32', 'a@16')) => ('fge', ('f2f16', 'b'), 'a') */
   /* search1757_0_0 -> 515 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 515 },
      -1,
   } },
   /* search1757_1_0 -> 526 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 526 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 5023, 5024 },
      -1,
   } },

   /* replace1757_0_0 -> 515 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 515 },
      -1,
   } },
   /* replace1757_1 -> 526 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 5026, 526 },
      -1,
   } },

   /* ('flt', ('f2f32', 'b@16'), ('f2f32', 'a@16')) => ('flt', ('f2f16', 'b'), 'a') */
   /* search1758_0_0 -> 515 in the cache */
   /* search1758_0 -> 5023 in the cache */
   /* search1758_1_0 -> 526 in the cache */
   /* search1758_1 -> 5024 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 5023, 5024 },
      -1,
   } },

   /* replace1758_0_0 -> 515 in the cache */
   /* replace1758_0 -> 5026 in the cache */
   /* replace1758_1 -> 526 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 5026, 526 },
      -1,
   } },

   /* ('feq', ('f2f64', 'a@16'), ('f2f64', 'b@16')) => ('feq', 'a', ('f2f16', 'b')) */
   /* search1759_0_0 -> 11 in the cache */
   /* search1759_0 -> 1825 in the cache */
   /* search1759_1_0 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2f64,
      -1, 0,
      { 258 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1825, 5030 },
      -1,
   } },

   /* replace1759_0 -> 11 in the cache */
   /* replace1759_1_0 -> 258 in the cache */
   /* replace1759_1 -> 5015 in the cache */
   /* replace1759 -> 5016 in the cache */

   /* ('fneu', ('f2f64', 'a@16'), ('f2f64', 'b@16')) => ('fneu', 'a', ('f2f16', 'b')) */
   /* search1760_0_0 -> 11 in the cache */
   /* search1760_0 -> 1825 in the cache */
   /* search1760_1_0 -> 258 in the cache */
   /* search1760_1 -> 5030 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1825, 5030 },
      -1,
   } },

   /* replace1760_0 -> 11 in the cache */
   /* replace1760_1_0 -> 258 in the cache */
   /* replace1760_1 -> 5015 in the cache */
   /* replace1760 -> 5018 in the cache */

   /* ('fge', ('f2f64', 'a@16'), ('f2f64', 'b@16')) => ('fge', 'a', ('f2f16', 'b')) */
   /* search1761_0_0 -> 11 in the cache */
   /* search1761_0 -> 1825 in the cache */
   /* search1761_1_0 -> 258 in the cache */
   /* search1761_1 -> 5030 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1825, 5030 },
      -1,
   } },

   /* replace1761_0 -> 11 in the cache */
   /* replace1761_1_0 -> 258 in the cache */
   /* replace1761_1 -> 5015 in the cache */
   /* replace1761 -> 5020 in the cache */

   /* ('flt', ('f2f64', 'a@16'), ('f2f64', 'b@16')) => ('flt', 'a', ('f2f16', 'b')) */
   /* search1762_0_0 -> 11 in the cache */
   /* search1762_0 -> 1825 in the cache */
   /* search1762_1_0 -> 258 in the cache */
   /* search1762_1 -> 5030 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1825, 5030 },
      -1,
   } },

   /* replace1762_0 -> 11 in the cache */
   /* replace1762_1_0 -> 258 in the cache */
   /* replace1762_1 -> 5015 in the cache */
   /* replace1762 -> 5022 in the cache */

   /* ('fge', ('f2f64', 'b@16'), ('f2f64', 'a@16')) => ('fge', ('f2f16', 'b'), 'a') */
   /* search1763_0_0 -> 515 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2f64,
      -1, 0,
      { 515 },
      -1,
   } },
   /* search1763_1_0 -> 526 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2f64,
      -1, 0,
      { 526 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 5035, 5036 },
      -1,
   } },

   /* replace1763_0_0 -> 515 in the cache */
   /* replace1763_0 -> 5026 in the cache */
   /* replace1763_1 -> 526 in the cache */
   /* replace1763 -> 5027 in the cache */

   /* ('flt', ('f2f64', 'b@16'), ('f2f64', 'a@16')) => ('flt', ('f2f16', 'b'), 'a') */
   /* search1764_0_0 -> 515 in the cache */
   /* search1764_0 -> 5035 in the cache */
   /* search1764_1_0 -> 526 in the cache */
   /* search1764_1 -> 5036 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 5035, 5036 },
      -1,
   } },

   /* replace1764_0_0 -> 515 in the cache */
   /* replace1764_0 -> 5026 in the cache */
   /* replace1764_1 -> 526 in the cache */
   /* replace1764 -> 5029 in the cache */

   /* ('feq', ('f2f64', 'a@32'), ('f2f64', 'b@16')) => ('feq', 'a', ('f2f32', 'b')) */
   /* search1765_0_0 -> 16 in the cache */
   /* search1765_0 -> 1907 in the cache */
   /* search1765_1_0 -> 258 in the cache */
   /* search1765_1 -> 5030 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1907, 5030 },
      -1,
   } },

   /* replace1765_0 -> 16 in the cache */
   /* replace1765_1_0 -> 258 in the cache */
   /* replace1765_1 -> 5013 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 16, 5013 },
      -1,
   } },

   /* ('fneu', ('f2f64', 'a@32'), ('f2f64', 'b@16')) => ('fneu', 'a', ('f2f32', 'b')) */
   /* search1766_0_0 -> 16 in the cache */
   /* search1766_0 -> 1907 in the cache */
   /* search1766_1_0 -> 258 in the cache */
   /* search1766_1 -> 5030 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1907, 5030 },
      -1,
   } },

   /* replace1766_0 -> 16 in the cache */
   /* replace1766_1_0 -> 258 in the cache */
   /* replace1766_1 -> 5013 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 16, 5013 },
      -1,
   } },

   /* ('fge', ('f2f64', 'a@32'), ('f2f64', 'b@16')) => ('fge', 'a', ('f2f32', 'b')) */
   /* search1767_0_0 -> 16 in the cache */
   /* search1767_0 -> 1907 in the cache */
   /* search1767_1_0 -> 258 in the cache */
   /* search1767_1 -> 5030 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1907, 5030 },
      -1,
   } },

   /* replace1767_0 -> 16 in the cache */
   /* replace1767_1_0 -> 258 in the cache */
   /* replace1767_1 -> 5013 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 16, 5013 },
      -1,
   } },

   /* ('flt', ('f2f64', 'a@32'), ('f2f64', 'b@16')) => ('flt', 'a', ('f2f32', 'b')) */
   /* search1768_0_0 -> 16 in the cache */
   /* search1768_0 -> 1907 in the cache */
   /* search1768_1_0 -> 258 in the cache */
   /* search1768_1 -> 5030 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1907, 5030 },
      -1,
   } },

   /* replace1768_0 -> 16 in the cache */
   /* replace1768_1_0 -> 258 in the cache */
   /* replace1768_1 -> 5013 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 16, 5013 },
      -1,
   } },

   /* ('fge', ('f2f64', 'b@16'), ('f2f64', 'a@32')) => ('fge', ('f2f32', 'b'), 'a') */
   /* search1769_0_0 -> 515 in the cache */
   /* search1769_0 -> 5035 in the cache */
   /* search1769_1_0 -> 183 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2f64,
      -1, 0,
      { 183 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 5035, 5047 },
      -1,
   } },

   /* replace1769_0_0 -> 515 in the cache */
   /* replace1769_0 -> 5023 in the cache */
   /* replace1769_1 -> 183 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 5023, 183 },
      -1,
   } },

   /* ('flt', ('f2f64', 'b@16'), ('f2f64', 'a@32')) => ('flt', ('f2f32', 'b'), 'a') */
   /* search1770_0_0 -> 515 in the cache */
   /* search1770_0 -> 5035 in the cache */
   /* search1770_1_0 -> 183 in the cache */
   /* search1770_1 -> 5047 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 5035, 5047 },
      -1,
   } },

   /* replace1770_0_0 -> 515 in the cache */
   /* replace1770_0 -> 5023 in the cache */
   /* replace1770_1 -> 183 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 5023, 183 },
      -1,
   } },

   /* ('feq', ('f2f64', 'a@32'), ('f2f64', 'b@32')) => ('feq', 'a', ('f2f32', 'b')) */
   /* search1771_0_0 -> 16 in the cache */
   /* search1771_0 -> 1907 in the cache */
   /* search1771_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2f64,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1907, 5052 },
      -1,
   } },

   /* replace1771_0 -> 16 in the cache */
   /* replace1771_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 16, 5054 },
      -1,
   } },

   /* ('fneu', ('f2f64', 'a@32'), ('f2f64', 'b@32')) => ('fneu', 'a', ('f2f32', 'b')) */
   /* search1772_0_0 -> 16 in the cache */
   /* search1772_0 -> 1907 in the cache */
   /* search1772_1_0 -> 36 in the cache */
   /* search1772_1 -> 5052 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1907, 5052 },
      -1,
   } },

   /* replace1772_0 -> 16 in the cache */
   /* replace1772_1_0 -> 36 in the cache */
   /* replace1772_1 -> 5054 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 16, 5054 },
      -1,
   } },

   /* ('fge', ('f2f64', 'a@32'), ('f2f64', 'b@32')) => ('fge', 'a', ('f2f32', 'b')) */
   /* search1773_0_0 -> 16 in the cache */
   /* search1773_0 -> 1907 in the cache */
   /* search1773_1_0 -> 36 in the cache */
   /* search1773_1 -> 5052 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1907, 5052 },
      -1,
   } },

   /* replace1773_0 -> 16 in the cache */
   /* replace1773_1_0 -> 36 in the cache */
   /* replace1773_1 -> 5054 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 16, 5054 },
      -1,
   } },

   /* ('flt', ('f2f64', 'a@32'), ('f2f64', 'b@32')) => ('flt', 'a', ('f2f32', 'b')) */
   /* search1774_0_0 -> 16 in the cache */
   /* search1774_0 -> 1907 in the cache */
   /* search1774_1_0 -> 36 in the cache */
   /* search1774_1 -> 5052 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1907, 5052 },
      -1,
   } },

   /* replace1774_0 -> 16 in the cache */
   /* replace1774_1_0 -> 36 in the cache */
   /* replace1774_1 -> 5054 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 16, 5054 },
      -1,
   } },

   /* ('fge', ('f2f64', 'b@32'), ('f2f64', 'a@32')) => ('fge', ('f2f32', 'b'), 'a') */
   /* search1775_0_0 -> 373 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2f64,
      -1, 0,
      { 373 },
      -1,
   } },
   /* search1775_1_0 -> 183 in the cache */
   /* search1775_1 -> 5047 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 5062, 5047 },
      -1,
   } },

   /* replace1775_0_0 -> 373 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 373 },
      -1,
   } },
   /* replace1775_1 -> 183 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 5064, 183 },
      -1,
   } },

   /* ('flt', ('f2f64', 'b@32'), ('f2f64', 'a@32')) => ('flt', ('f2f32', 'b'), 'a') */
   /* search1776_0_0 -> 373 in the cache */
   /* search1776_0 -> 5062 in the cache */
   /* search1776_1_0 -> 183 in the cache */
   /* search1776_1 -> 5047 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 5062, 5047 },
      -1,
   } },

   /* replace1776_0_0 -> 373 in the cache */
   /* replace1776_0 -> 5064 in the cache */
   /* replace1776_1 -> 183 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 5064, 183 },
      -1,
   } },

   /* ('i2i32', ('iand', 'a@64', 4294967295)) => ('u2u32', 'a') */
   /* search1777_0_0 -> 21 in the cache */
   /* search1777_0_1 -> 3315 in the cache */
   /* search1777_0 -> 3316 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 1,
      { 3316 },
      -1,
   } },

   /* replace1777_0 -> 21 in the cache */
   /* replace1777 -> 3307 in the cache */

   /* ('i2i16', ('iand', 'a@32', 65535)) => ('u2u16', 'a') */
   /* search1778_0_0 -> 16 in the cache */
   /* search1778_0_1 -> 1002 in the cache */
   /* search1778_0 -> 1003 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 1,
      { 1003 },
      -1,
   } },

   /* replace1778_0 -> 16 in the cache */
   /* replace1778 -> 3310 in the cache */

   /* ('i2i16', ('iand', 'a@64', 65535)) => ('u2u16', 'a') */
   /* search1779_0_0 -> 21 in the cache */
   /* search1779_0_1 -> 4922 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 21, 4922 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 1,
      { 5070 },
      -1,
   } },

   /* replace1779_0 -> 21 in the cache */
   /* replace1779 -> 4929 in the cache */

   /* ('i2i8', ('iand', 'a@16', 255)) => ('u2u8', 'a') */
   /* search1780_0_0 -> 11 in the cache */
   /* search1780_0_1 -> 2013 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 11, 2013 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 1,
      { 5072 },
      -1,
   } },

   /* replace1780_0 -> 11 in the cache */
   /* replace1780 -> 2009 in the cache */

   /* ('i2i8', ('iand', 'a@32', 255)) => ('u2u8', 'a') */
   /* search1781_0_0 -> 16 in the cache */
   /* search1781_0_1 -> 1010 in the cache */
   /* search1781_0 -> 1011 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 1,
      { 1011 },
      -1,
   } },

   /* replace1781_0 -> 16 in the cache */
   /* replace1781 -> 2025 in the cache */

   /* ('i2i8', ('iand', 'a@64', 255)) => ('u2u8', 'a') */
   /* search1782_0_0 -> 21 in the cache */
   /* search1782_0_1 -> 4814 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 21, 4814 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 1,
      { 5075 },
      -1,
   } },

   /* replace1782_0 -> 21 in the cache */
   /* replace1782 -> 4822 in the cache */

   /* ('u2u16(only_lower_8_bits_used)', ('u2u8', 'a@16')) => a */
   /* search1783_0_0 -> 11 in the cache */
   /* search1783_0 -> 2009 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 2009 },
      5,
   } },

   /* replace1783 -> 11 in the cache */

   /* ('extract_u8(only_lower_8_bits_used)', 'a@16', 0) => a */
   /* search1784_0 -> 11 in the cache */
   /* search1784_1 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 11, 856 },
      5,
   } },

   /* replace1784 -> 11 in the cache */

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('u2u16', ('u2u8', 'a@16')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1785_0 -> 521 in the cache */
   /* search1785_1_0_0 -> 526 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 526 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 5079 },
      -1,
   } },
   /* search1785_2 -> 488 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5080, 488 },
      5,
   } },

   /* replace1785_0 -> 521 in the cache */
   /* replace1785_1 -> 526 in the cache */
   /* replace1785_2 -> 488 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 526, 488 },
      -1,
   } },

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('u2u16', ('i2i8', 'a@16')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1786_0 -> 521 in the cache */
   /* search1786_1_0_0 -> 526 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 526 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 5083 },
      -1,
   } },
   /* search1786_2 -> 488 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5084, 488 },
      5,
   } },

   /* replace1786_0 -> 521 in the cache */
   /* replace1786_1 -> 526 in the cache */
   /* replace1786_2 -> 488 in the cache */
   /* replace1786 -> 5082 in the cache */

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('extract_u8', 'a@16', 0), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1787_0 -> 521 in the cache */
   /* search1787_1_0 -> 526 in the cache */
   /* search1787_1_1 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 526, 856 },
      -1,
   } },
   /* search1787_2 -> 488 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5086, 488 },
      5,
   } },

   /* replace1787_0 -> 521 in the cache */
   /* replace1787_1 -> 526 in the cache */
   /* replace1787_2 -> 488 in the cache */
   /* replace1787 -> 5082 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('u2u16', ('u2u8', 'a@16')), 'b') => ('iadd', 'a', 'b') */
   /* search1788_0_0_0 -> 11 in the cache */
   /* search1788_0_0 -> 2009 in the cache */
   /* search1788_0 -> 2010 in the cache */
   /* search1788_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 2010, 258 },
      5,
   } },

   /* replace1788_0 -> 11 in the cache */
   /* replace1788_1 -> 258 in the cache */
   /* replace1788 -> 4052 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('u2u16', ('i2i8', 'a@16')), 'b') => ('iadd', 'a', 'b') */
   /* search1789_0_0_0 -> 11 in the cache */
   /* search1789_0_0 -> 4399 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 4399 },
      -1,
   } },
   /* search1789_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 5089, 258 },
      5,
   } },

   /* replace1789_0 -> 11 in the cache */
   /* replace1789_1 -> 258 in the cache */
   /* replace1789 -> 4052 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('extract_u8', 'a@16', 0), 'b') => ('iadd', 'a', 'b') */
   /* search1790_0_0 -> 11 in the cache */
   /* search1790_0_1 -> 856 in the cache */
   /* search1790_0 -> 3418 in the cache */
   /* search1790_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 3418, 258 },
      5,
   } },

   /* replace1790_0 -> 11 in the cache */
   /* replace1790_1 -> 258 in the cache */
   /* replace1790 -> 4052 in the cache */

   /* ('imul(only_lower_8_bits_used)', ('u2u16', ('u2u8', 'a@16')), 'b') => ('imul', 'a', 'b') */
   /* search1791_0_0_0 -> 11 in the cache */
   /* search1791_0_0 -> 2009 in the cache */
   /* search1791_0 -> 2010 in the cache */
   /* search1791_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 2010, 258 },
      5,
   } },

   /* replace1791_0 -> 11 in the cache */
   /* replace1791_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 11, 258 },
      -1,
   } },

   /* ('imul(only_lower_8_bits_used)', ('u2u16', ('i2i8', 'a@16')), 'b') => ('imul', 'a', 'b') */
   /* search1792_0_0_0 -> 11 in the cache */
   /* search1792_0_0 -> 4399 in the cache */
   /* search1792_0 -> 5089 in the cache */
   /* search1792_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 5089, 258 },
      5,
   } },

   /* replace1792_0 -> 11 in the cache */
   /* replace1792_1 -> 258 in the cache */
   /* replace1792 -> 5093 in the cache */

   /* ('imul(only_lower_8_bits_used)', ('extract_u8', 'a@16', 0), 'b') => ('imul', 'a', 'b') */
   /* search1793_0_0 -> 11 in the cache */
   /* search1793_0_1 -> 856 in the cache */
   /* search1793_0 -> 3418 in the cache */
   /* search1793_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 3418, 258 },
      5,
   } },

   /* replace1793_0 -> 11 in the cache */
   /* replace1793_1 -> 258 in the cache */
   /* replace1793 -> 5093 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('u2u16', ('u2u8', 'a@16')), 'b') => ('iand', 'a', 'b') */
   /* search1794_0_0_0 -> 11 in the cache */
   /* search1794_0_0 -> 2009 in the cache */
   /* search1794_0 -> 2010 in the cache */
   /* search1794_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2010, 258 },
      5,
   } },

   /* replace1794_0 -> 11 in the cache */
   /* replace1794_1 -> 258 in the cache */
   /* replace1794 -> 2023 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('u2u16', ('i2i8', 'a@16')), 'b') => ('iand', 'a', 'b') */
   /* search1795_0_0_0 -> 11 in the cache */
   /* search1795_0_0 -> 4399 in the cache */
   /* search1795_0 -> 5089 in the cache */
   /* search1795_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 5089, 258 },
      5,
   } },

   /* replace1795_0 -> 11 in the cache */
   /* replace1795_1 -> 258 in the cache */
   /* replace1795 -> 2023 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('extract_u8', 'a@16', 0), 'b') => ('iand', 'a', 'b') */
   /* search1796_0_0 -> 11 in the cache */
   /* search1796_0_1 -> 856 in the cache */
   /* search1796_0 -> 3418 in the cache */
   /* search1796_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 3418, 258 },
      5,
   } },

   /* replace1796_0 -> 11 in the cache */
   /* replace1796_1 -> 258 in the cache */
   /* replace1796 -> 2023 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('u2u16', ('u2u8', 'a@16')), 'b') => ('ior', 'a', 'b') */
   /* search1797_0_0_0 -> 11 in the cache */
   /* search1797_0_0 -> 2009 in the cache */
   /* search1797_0 -> 2010 in the cache */
   /* search1797_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2010, 258 },
      5,
   } },

   /* replace1797_0 -> 11 in the cache */
   /* replace1797_1 -> 258 in the cache */
   /* replace1797 -> 2065 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('u2u16', ('i2i8', 'a@16')), 'b') => ('ior', 'a', 'b') */
   /* search1798_0_0_0 -> 11 in the cache */
   /* search1798_0_0 -> 4399 in the cache */
   /* search1798_0 -> 5089 in the cache */
   /* search1798_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 5089, 258 },
      5,
   } },

   /* replace1798_0 -> 11 in the cache */
   /* replace1798_1 -> 258 in the cache */
   /* replace1798 -> 2065 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('extract_u8', 'a@16', 0), 'b') => ('ior', 'a', 'b') */
   /* search1799_0_0 -> 11 in the cache */
   /* search1799_0_1 -> 856 in the cache */
   /* search1799_0 -> 3418 in the cache */
   /* search1799_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 3418, 258 },
      5,
   } },

   /* replace1799_0 -> 11 in the cache */
   /* replace1799_1 -> 258 in the cache */
   /* replace1799 -> 2065 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('u2u16', ('u2u8', 'a@16')), 'b') => ('ixor', 'a', 'b') */
   /* search1800_0_0_0 -> 11 in the cache */
   /* search1800_0_0 -> 2009 in the cache */
   /* search1800_0 -> 2010 in the cache */
   /* search1800_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 2010, 258 },
      5,
   } },

   /* replace1800_0 -> 11 in the cache */
   /* replace1800_1 -> 258 in the cache */
   /* replace1800 -> 2089 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('u2u16', ('i2i8', 'a@16')), 'b') => ('ixor', 'a', 'b') */
   /* search1801_0_0_0 -> 11 in the cache */
   /* search1801_0_0 -> 4399 in the cache */
   /* search1801_0 -> 5089 in the cache */
   /* search1801_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 5089, 258 },
      5,
   } },

   /* replace1801_0 -> 11 in the cache */
   /* replace1801_1 -> 258 in the cache */
   /* replace1801 -> 2089 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('extract_u8', 'a@16', 0), 'b') => ('ixor', 'a', 'b') */
   /* search1802_0_0 -> 11 in the cache */
   /* search1802_0_1 -> 856 in the cache */
   /* search1802_0 -> 3418 in the cache */
   /* search1802_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 3418, 258 },
      5,
   } },

   /* replace1802_0 -> 11 in the cache */
   /* replace1802_1 -> 258 in the cache */
   /* replace1802 -> 2089 in the cache */

   /* ('i2i16(only_lower_8_bits_used)', ('u2u8', 'a@16')) => a */
   /* search1803_0_0 -> 11 in the cache */
   /* search1803_0 -> 2009 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 2009 },
      5,
   } },

   /* replace1803 -> 11 in the cache */

   /* ('extract_i8(only_lower_8_bits_used)', 'a@16', 0) => a */
   /* search1804_0 -> 11 in the cache */
   /* search1804_1 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 11, 856 },
      5,
   } },

   /* replace1804 -> 11 in the cache */

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('i2i16', ('u2u8', 'a@16')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1805_0 -> 521 in the cache */
   /* search1805_1_0_0 -> 526 in the cache */
   /* search1805_1_0 -> 5079 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 5079 },
      -1,
   } },
   /* search1805_2 -> 488 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5107, 488 },
      5,
   } },

   /* replace1805_0 -> 521 in the cache */
   /* replace1805_1 -> 526 in the cache */
   /* replace1805_2 -> 488 in the cache */
   /* replace1805 -> 5082 in the cache */

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('i2i16', ('i2i8', 'a@16')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1806_0 -> 521 in the cache */
   /* search1806_1_0_0 -> 526 in the cache */
   /* search1806_1_0 -> 5083 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 5083 },
      -1,
   } },
   /* search1806_2 -> 488 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5109, 488 },
      5,
   } },

   /* replace1806_0 -> 521 in the cache */
   /* replace1806_1 -> 526 in the cache */
   /* replace1806_2 -> 488 in the cache */
   /* replace1806 -> 5082 in the cache */

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('extract_i8', 'a@16', 0), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1807_0 -> 521 in the cache */
   /* search1807_1_0 -> 526 in the cache */
   /* search1807_1_1 -> 856 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 526, 856 },
      -1,
   } },
   /* search1807_2 -> 488 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5111, 488 },
      5,
   } },

   /* replace1807_0 -> 521 in the cache */
   /* replace1807_1 -> 526 in the cache */
   /* replace1807_2 -> 488 in the cache */
   /* replace1807 -> 5082 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('i2i16', ('u2u8', 'a@16')), 'b') => ('iadd', 'a', 'b') */
   /* search1808_0_0_0 -> 11 in the cache */
   /* search1808_0_0 -> 2009 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 2009 },
      -1,
   } },
   /* search1808_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 5113, 258 },
      5,
   } },

   /* replace1808_0 -> 11 in the cache */
   /* replace1808_1 -> 258 in the cache */
   /* replace1808 -> 4052 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('i2i16', ('i2i8', 'a@16')), 'b') => ('iadd', 'a', 'b') */
   /* search1809_0_0_0 -> 11 in the cache */
   /* search1809_0_0 -> 4399 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 4399 },
      -1,
   } },
   /* search1809_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 5115, 258 },
      5,
   } },

   /* replace1809_0 -> 11 in the cache */
   /* replace1809_1 -> 258 in the cache */
   /* replace1809 -> 4052 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('extract_i8', 'a@16', 0), 'b') => ('iadd', 'a', 'b') */
   /* search1810_0_0 -> 11 in the cache */
   /* search1810_0_1 -> 856 in the cache */
   /* search1810_0 -> 3439 in the cache */
   /* search1810_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 3439, 258 },
      5,
   } },

   /* replace1810_0 -> 11 in the cache */
   /* replace1810_1 -> 258 in the cache */
   /* replace1810 -> 4052 in the cache */

   /* ('imul(only_lower_8_bits_used)', ('i2i16', ('u2u8', 'a@16')), 'b') => ('imul', 'a', 'b') */
   /* search1811_0_0_0 -> 11 in the cache */
   /* search1811_0_0 -> 2009 in the cache */
   /* search1811_0 -> 5113 in the cache */
   /* search1811_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 5113, 258 },
      5,
   } },

   /* replace1811_0 -> 11 in the cache */
   /* replace1811_1 -> 258 in the cache */
   /* replace1811 -> 5093 in the cache */

   /* ('imul(only_lower_8_bits_used)', ('i2i16', ('i2i8', 'a@16')), 'b') => ('imul', 'a', 'b') */
   /* search1812_0_0_0 -> 11 in the cache */
   /* search1812_0_0 -> 4399 in the cache */
   /* search1812_0 -> 5115 in the cache */
   /* search1812_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 5115, 258 },
      5,
   } },

   /* replace1812_0 -> 11 in the cache */
   /* replace1812_1 -> 258 in the cache */
   /* replace1812 -> 5093 in the cache */

   /* ('imul(only_lower_8_bits_used)', ('extract_i8', 'a@16', 0), 'b') => ('imul', 'a', 'b') */
   /* search1813_0_0 -> 11 in the cache */
   /* search1813_0_1 -> 856 in the cache */
   /* search1813_0 -> 3439 in the cache */
   /* search1813_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 3439, 258 },
      5,
   } },

   /* replace1813_0 -> 11 in the cache */
   /* replace1813_1 -> 258 in the cache */
   /* replace1813 -> 5093 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('i2i16', ('u2u8', 'a@16')), 'b') => ('iand', 'a', 'b') */
   /* search1814_0_0_0 -> 11 in the cache */
   /* search1814_0_0 -> 2009 in the cache */
   /* search1814_0 -> 5113 in the cache */
   /* search1814_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 5113, 258 },
      5,
   } },

   /* replace1814_0 -> 11 in the cache */
   /* replace1814_1 -> 258 in the cache */
   /* replace1814 -> 2023 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('i2i16', ('i2i8', 'a@16')), 'b') => ('iand', 'a', 'b') */
   /* search1815_0_0_0 -> 11 in the cache */
   /* search1815_0_0 -> 4399 in the cache */
   /* search1815_0 -> 5115 in the cache */
   /* search1815_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 5115, 258 },
      5,
   } },

   /* replace1815_0 -> 11 in the cache */
   /* replace1815_1 -> 258 in the cache */
   /* replace1815 -> 2023 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('extract_i8', 'a@16', 0), 'b') => ('iand', 'a', 'b') */
   /* search1816_0_0 -> 11 in the cache */
   /* search1816_0_1 -> 856 in the cache */
   /* search1816_0 -> 3439 in the cache */
   /* search1816_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 3439, 258 },
      5,
   } },

   /* replace1816_0 -> 11 in the cache */
   /* replace1816_1 -> 258 in the cache */
   /* replace1816 -> 2023 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('i2i16', ('u2u8', 'a@16')), 'b') => ('ior', 'a', 'b') */
   /* search1817_0_0_0 -> 11 in the cache */
   /* search1817_0_0 -> 2009 in the cache */
   /* search1817_0 -> 5113 in the cache */
   /* search1817_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 5113, 258 },
      5,
   } },

   /* replace1817_0 -> 11 in the cache */
   /* replace1817_1 -> 258 in the cache */
   /* replace1817 -> 2065 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('i2i16', ('i2i8', 'a@16')), 'b') => ('ior', 'a', 'b') */
   /* search1818_0_0_0 -> 11 in the cache */
   /* search1818_0_0 -> 4399 in the cache */
   /* search1818_0 -> 5115 in the cache */
   /* search1818_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 5115, 258 },
      5,
   } },

   /* replace1818_0 -> 11 in the cache */
   /* replace1818_1 -> 258 in the cache */
   /* replace1818 -> 2065 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('extract_i8', 'a@16', 0), 'b') => ('ior', 'a', 'b') */
   /* search1819_0_0 -> 11 in the cache */
   /* search1819_0_1 -> 856 in the cache */
   /* search1819_0 -> 3439 in the cache */
   /* search1819_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 3439, 258 },
      5,
   } },

   /* replace1819_0 -> 11 in the cache */
   /* replace1819_1 -> 258 in the cache */
   /* replace1819 -> 2065 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('i2i16', ('u2u8', 'a@16')), 'b') => ('ixor', 'a', 'b') */
   /* search1820_0_0_0 -> 11 in the cache */
   /* search1820_0_0 -> 2009 in the cache */
   /* search1820_0 -> 5113 in the cache */
   /* search1820_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 5113, 258 },
      5,
   } },

   /* replace1820_0 -> 11 in the cache */
   /* replace1820_1 -> 258 in the cache */
   /* replace1820 -> 2089 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('i2i16', ('i2i8', 'a@16')), 'b') => ('ixor', 'a', 'b') */
   /* search1821_0_0_0 -> 11 in the cache */
   /* search1821_0_0 -> 4399 in the cache */
   /* search1821_0 -> 5115 in the cache */
   /* search1821_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 5115, 258 },
      5,
   } },

   /* replace1821_0 -> 11 in the cache */
   /* replace1821_1 -> 258 in the cache */
   /* replace1821 -> 2089 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('extract_i8', 'a@16', 0), 'b') => ('ixor', 'a', 'b') */
   /* search1822_0_0 -> 11 in the cache */
   /* search1822_0_1 -> 856 in the cache */
   /* search1822_0 -> 3439 in the cache */
   /* search1822_1 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 3439, 258 },
      5,
   } },

   /* replace1822_0 -> 11 in the cache */
   /* replace1822_1 -> 258 in the cache */
   /* replace1822 -> 2089 in the cache */

   /* ('u2u32(only_lower_8_bits_used)', ('u2u8', 'a@32')) => a */
   /* search1823_0_0 -> 16 in the cache */
   /* search1823_0 -> 2025 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 2025 },
      5,
   } },

   /* replace1823 -> 16 in the cache */

   /* ('extract_u8(only_lower_8_bits_used)', 'a@32', 0) => a */
   /* search1824_0 -> 16 in the cache */
   /* search1824_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 16, 122 },
      5,
   } },

   /* replace1824 -> 16 in the cache */

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('u2u32', ('u2u8', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1825_0 -> 521 in the cache */
   /* search1825_1_0_0 -> 183 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 183 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 5132 },
      -1,
   } },
   /* search1825_2 -> 559 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5133, 559 },
      5,
   } },

   /* replace1825_0 -> 521 in the cache */
   /* replace1825_1 -> 183 in the cache */
   /* replace1825_2 -> 559 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 183, 559 },
      -1,
   } },

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('u2u32', ('i2i8', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1826_0 -> 521 in the cache */
   /* search1826_1_0_0 -> 183 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 183 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 5136 },
      -1,
   } },
   /* search1826_2 -> 559 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5137, 559 },
      5,
   } },

   /* replace1826_0 -> 521 in the cache */
   /* replace1826_1 -> 183 in the cache */
   /* replace1826_2 -> 559 in the cache */
   /* replace1826 -> 5135 in the cache */

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('extract_u8', 'a@32', 0), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1827_0 -> 521 in the cache */
   /* search1827_1_0 -> 183 in the cache */
   /* search1827_1_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 183, 122 },
      -1,
   } },
   /* search1827_2 -> 559 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5139, 559 },
      5,
   } },

   /* replace1827_0 -> 521 in the cache */
   /* replace1827_1 -> 183 in the cache */
   /* replace1827_2 -> 559 in the cache */
   /* replace1827 -> 5135 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('u2u32', ('u2u8', 'a@32')), 'b') => ('iadd', 'a', 'b') */
   /* search1828_0_0_0 -> 16 in the cache */
   /* search1828_0_0 -> 2025 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 2025 },
      -1,
   } },
   /* search1828_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 5141, 36 },
      5,
   } },

   /* replace1828_0 -> 16 in the cache */
   /* replace1828_1 -> 36 in the cache */
   /* replace1828 -> 4065 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('u2u32', ('i2i8', 'a@32')), 'b') => ('iadd', 'a', 'b') */
   /* search1829_0_0_0 -> 16 in the cache */
   /* search1829_0_0 -> 4351 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 4351 },
      -1,
   } },
   /* search1829_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 5143, 36 },
      5,
   } },

   /* replace1829_0 -> 16 in the cache */
   /* replace1829_1 -> 36 in the cache */
   /* replace1829 -> 4065 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('extract_u8', 'a@32', 0), 'b') => ('iadd', 'a', 'b') */
   /* search1830_0_0 -> 16 in the cache */
   /* search1830_0_1 -> 122 in the cache */
   /* search1830_0 -> 412 in the cache */
   /* search1830_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 412, 36 },
      5,
   } },

   /* replace1830_0 -> 16 in the cache */
   /* replace1830_1 -> 36 in the cache */
   /* replace1830 -> 4065 in the cache */

   /* ('imul(only_lower_8_bits_used)', ('u2u32', ('u2u8', 'a@32')), 'b') => ('imul', 'a', 'b') */
   /* search1831_0_0_0 -> 16 in the cache */
   /* search1831_0_0 -> 2025 in the cache */
   /* search1831_0 -> 5141 in the cache */
   /* search1831_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 5141, 36 },
      5,
   } },

   /* replace1831_0 -> 16 in the cache */
   /* replace1831_1 -> 36 in the cache */
   /* replace1831 -> 49 in the cache */

   /* ('imul(only_lower_8_bits_used)', ('u2u32', ('i2i8', 'a@32')), 'b') => ('imul', 'a', 'b') */
   /* search1832_0_0_0 -> 16 in the cache */
   /* search1832_0_0 -> 4351 in the cache */
   /* search1832_0 -> 5143 in the cache */
   /* search1832_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 5143, 36 },
      5,
   } },

   /* replace1832_0 -> 16 in the cache */
   /* replace1832_1 -> 36 in the cache */
   /* replace1832 -> 49 in the cache */

   /* ('imul(only_lower_8_bits_used)', ('extract_u8', 'a@32', 0), 'b') => ('imul', 'a', 'b') */
   /* search1833_0_0 -> 16 in the cache */
   /* search1833_0_1 -> 122 in the cache */
   /* search1833_0 -> 412 in the cache */
   /* search1833_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 412, 36 },
      5,
   } },

   /* replace1833_0 -> 16 in the cache */
   /* replace1833_1 -> 36 in the cache */
   /* replace1833 -> 49 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('u2u32', ('u2u8', 'a@32')), 'b') => ('iand', 'a', 'b') */
   /* search1834_0_0_0 -> 16 in the cache */
   /* search1834_0_0 -> 2025 in the cache */
   /* search1834_0 -> 5141 in the cache */
   /* search1834_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 5141, 36 },
      5,
   } },

   /* replace1834_0 -> 16 in the cache */
   /* replace1834_1 -> 36 in the cache */
   /* replace1834 -> 2031 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('u2u32', ('i2i8', 'a@32')), 'b') => ('iand', 'a', 'b') */
   /* search1835_0_0_0 -> 16 in the cache */
   /* search1835_0_0 -> 4351 in the cache */
   /* search1835_0 -> 5143 in the cache */
   /* search1835_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 5143, 36 },
      5,
   } },

   /* replace1835_0 -> 16 in the cache */
   /* replace1835_1 -> 36 in the cache */
   /* replace1835 -> 2031 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('extract_u8', 'a@32', 0), 'b') => ('iand', 'a', 'b') */
   /* search1836_0_0 -> 16 in the cache */
   /* search1836_0_1 -> 122 in the cache */
   /* search1836_0 -> 412 in the cache */
   /* search1836_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 412, 36 },
      5,
   } },

   /* replace1836_0 -> 16 in the cache */
   /* replace1836_1 -> 36 in the cache */
   /* replace1836 -> 2031 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('u2u32', ('u2u8', 'a@32')), 'b') => ('ior', 'a', 'b') */
   /* search1837_0_0_0 -> 16 in the cache */
   /* search1837_0_0 -> 2025 in the cache */
   /* search1837_0 -> 5141 in the cache */
   /* search1837_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 5141, 36 },
      5,
   } },

   /* replace1837_0 -> 16 in the cache */
   /* replace1837_1 -> 36 in the cache */
   /* replace1837 -> 2069 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('u2u32', ('i2i8', 'a@32')), 'b') => ('ior', 'a', 'b') */
   /* search1838_0_0_0 -> 16 in the cache */
   /* search1838_0_0 -> 4351 in the cache */
   /* search1838_0 -> 5143 in the cache */
   /* search1838_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 5143, 36 },
      5,
   } },

   /* replace1838_0 -> 16 in the cache */
   /* replace1838_1 -> 36 in the cache */
   /* replace1838 -> 2069 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('extract_u8', 'a@32', 0), 'b') => ('ior', 'a', 'b') */
   /* search1839_0_0 -> 16 in the cache */
   /* search1839_0_1 -> 122 in the cache */
   /* search1839_0 -> 412 in the cache */
   /* search1839_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 412, 36 },
      5,
   } },

   /* replace1839_0 -> 16 in the cache */
   /* replace1839_1 -> 36 in the cache */
   /* replace1839 -> 2069 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('u2u32', ('u2u8', 'a@32')), 'b') => ('ixor', 'a', 'b') */
   /* search1840_0_0_0 -> 16 in the cache */
   /* search1840_0_0 -> 2025 in the cache */
   /* search1840_0 -> 5141 in the cache */
   /* search1840_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 5141, 36 },
      5,
   } },

   /* replace1840_0 -> 16 in the cache */
   /* replace1840_1 -> 36 in the cache */
   /* replace1840 -> 2093 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('u2u32', ('i2i8', 'a@32')), 'b') => ('ixor', 'a', 'b') */
   /* search1841_0_0_0 -> 16 in the cache */
   /* search1841_0_0 -> 4351 in the cache */
   /* search1841_0 -> 5143 in the cache */
   /* search1841_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 5143, 36 },
      5,
   } },

   /* replace1841_0 -> 16 in the cache */
   /* replace1841_1 -> 36 in the cache */
   /* replace1841 -> 2093 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('extract_u8', 'a@32', 0), 'b') => ('ixor', 'a', 'b') */
   /* search1842_0_0 -> 16 in the cache */
   /* search1842_0_1 -> 122 in the cache */
   /* search1842_0 -> 412 in the cache */
   /* search1842_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 412, 36 },
      5,
   } },

   /* replace1842_0 -> 16 in the cache */
   /* replace1842_1 -> 36 in the cache */
   /* replace1842 -> 2093 in the cache */

   /* ('i2i32(only_lower_8_bits_used)', ('u2u8', 'a@32')) => a */
   /* search1843_0_0 -> 16 in the cache */
   /* search1843_0 -> 2025 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 2025 },
      5,
   } },

   /* replace1843 -> 16 in the cache */

   /* ('extract_i8(only_lower_8_bits_used)', 'a@32', 0) => a */
   /* search1844_0 -> 16 in the cache */
   /* search1844_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 16, 122 },
      5,
   } },

   /* replace1844 -> 16 in the cache */

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('i2i32', ('u2u8', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1845_0 -> 521 in the cache */
   /* search1845_1_0_0 -> 183 in the cache */
   /* search1845_1_0 -> 5132 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 5132 },
      -1,
   } },
   /* search1845_2 -> 559 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5160, 559 },
      5,
   } },

   /* replace1845_0 -> 521 in the cache */
   /* replace1845_1 -> 183 in the cache */
   /* replace1845_2 -> 559 in the cache */
   /* replace1845 -> 5135 in the cache */

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('i2i32', ('i2i8', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1846_0 -> 521 in the cache */
   /* search1846_1_0_0 -> 183 in the cache */
   /* search1846_1_0 -> 5136 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 5136 },
      -1,
   } },
   /* search1846_2 -> 559 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5162, 559 },
      5,
   } },

   /* replace1846_0 -> 521 in the cache */
   /* replace1846_1 -> 183 in the cache */
   /* replace1846_2 -> 559 in the cache */
   /* replace1846 -> 5135 in the cache */

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('extract_i8', 'a@32', 0), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1847_0 -> 521 in the cache */
   /* search1847_1_0 -> 183 in the cache */
   /* search1847_1_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 183, 122 },
      -1,
   } },
   /* search1847_2 -> 559 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5164, 559 },
      5,
   } },

   /* replace1847_0 -> 521 in the cache */
   /* replace1847_1 -> 183 in the cache */
   /* replace1847_2 -> 559 in the cache */
   /* replace1847 -> 5135 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('i2i32', ('u2u8', 'a@32')), 'b') => ('iadd', 'a', 'b') */
   /* search1848_0_0_0 -> 16 in the cache */
   /* search1848_0_0 -> 2025 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 2025 },
      -1,
   } },
   /* search1848_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 5166, 36 },
      5,
   } },

   /* replace1848_0 -> 16 in the cache */
   /* replace1848_1 -> 36 in the cache */
   /* replace1848 -> 4065 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('i2i32', ('i2i8', 'a@32')), 'b') => ('iadd', 'a', 'b') */
   /* search1849_0_0_0 -> 16 in the cache */
   /* search1849_0_0 -> 4351 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 4351 },
      -1,
   } },
   /* search1849_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 5168, 36 },
      5,
   } },

   /* replace1849_0 -> 16 in the cache */
   /* replace1849_1 -> 36 in the cache */
   /* replace1849 -> 4065 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('extract_i8', 'a@32', 0), 'b') => ('iadd', 'a', 'b') */
   /* search1850_0_0 -> 16 in the cache */
   /* search1850_0_1 -> 122 in the cache */
   /* search1850_0 -> 393 in the cache */
   /* search1850_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 393, 36 },
      5,
   } },

   /* replace1850_0 -> 16 in the cache */
   /* replace1850_1 -> 36 in the cache */
   /* replace1850 -> 4065 in the cache */

   /* ('imul(only_lower_8_bits_used)', ('i2i32', ('u2u8', 'a@32')), 'b') => ('imul', 'a', 'b') */
   /* search1851_0_0_0 -> 16 in the cache */
   /* search1851_0_0 -> 2025 in the cache */
   /* search1851_0 -> 5166 in the cache */
   /* search1851_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 5166, 36 },
      5,
   } },

   /* replace1851_0 -> 16 in the cache */
   /* replace1851_1 -> 36 in the cache */
   /* replace1851 -> 49 in the cache */

   /* ('imul(only_lower_8_bits_used)', ('i2i32', ('i2i8', 'a@32')), 'b') => ('imul', 'a', 'b') */
   /* search1852_0_0_0 -> 16 in the cache */
   /* search1852_0_0 -> 4351 in the cache */
   /* search1852_0 -> 5168 in the cache */
   /* search1852_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 5168, 36 },
      5,
   } },

   /* replace1852_0 -> 16 in the cache */
   /* replace1852_1 -> 36 in the cache */
   /* replace1852 -> 49 in the cache */

   /* ('imul(only_lower_8_bits_used)', ('extract_i8', 'a@32', 0), 'b') => ('imul', 'a', 'b') */
   /* search1853_0_0 -> 16 in the cache */
   /* search1853_0_1 -> 122 in the cache */
   /* search1853_0 -> 393 in the cache */
   /* search1853_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 393, 36 },
      5,
   } },

   /* replace1853_0 -> 16 in the cache */
   /* replace1853_1 -> 36 in the cache */
   /* replace1853 -> 49 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('i2i32', ('u2u8', 'a@32')), 'b') => ('iand', 'a', 'b') */
   /* search1854_0_0_0 -> 16 in the cache */
   /* search1854_0_0 -> 2025 in the cache */
   /* search1854_0 -> 5166 in the cache */
   /* search1854_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 5166, 36 },
      5,
   } },

   /* replace1854_0 -> 16 in the cache */
   /* replace1854_1 -> 36 in the cache */
   /* replace1854 -> 2031 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('i2i32', ('i2i8', 'a@32')), 'b') => ('iand', 'a', 'b') */
   /* search1855_0_0_0 -> 16 in the cache */
   /* search1855_0_0 -> 4351 in the cache */
   /* search1855_0 -> 5168 in the cache */
   /* search1855_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 5168, 36 },
      5,
   } },

   /* replace1855_0 -> 16 in the cache */
   /* replace1855_1 -> 36 in the cache */
   /* replace1855 -> 2031 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('extract_i8', 'a@32', 0), 'b') => ('iand', 'a', 'b') */
   /* search1856_0_0 -> 16 in the cache */
   /* search1856_0_1 -> 122 in the cache */
   /* search1856_0 -> 393 in the cache */
   /* search1856_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 393, 36 },
      5,
   } },

   /* replace1856_0 -> 16 in the cache */
   /* replace1856_1 -> 36 in the cache */
   /* replace1856 -> 2031 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('i2i32', ('u2u8', 'a@32')), 'b') => ('ior', 'a', 'b') */
   /* search1857_0_0_0 -> 16 in the cache */
   /* search1857_0_0 -> 2025 in the cache */
   /* search1857_0 -> 5166 in the cache */
   /* search1857_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 5166, 36 },
      5,
   } },

   /* replace1857_0 -> 16 in the cache */
   /* replace1857_1 -> 36 in the cache */
   /* replace1857 -> 2069 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('i2i32', ('i2i8', 'a@32')), 'b') => ('ior', 'a', 'b') */
   /* search1858_0_0_0 -> 16 in the cache */
   /* search1858_0_0 -> 4351 in the cache */
   /* search1858_0 -> 5168 in the cache */
   /* search1858_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 5168, 36 },
      5,
   } },

   /* replace1858_0 -> 16 in the cache */
   /* replace1858_1 -> 36 in the cache */
   /* replace1858 -> 2069 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('extract_i8', 'a@32', 0), 'b') => ('ior', 'a', 'b') */
   /* search1859_0_0 -> 16 in the cache */
   /* search1859_0_1 -> 122 in the cache */
   /* search1859_0 -> 393 in the cache */
   /* search1859_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 393, 36 },
      5,
   } },

   /* replace1859_0 -> 16 in the cache */
   /* replace1859_1 -> 36 in the cache */
   /* replace1859 -> 2069 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('i2i32', ('u2u8', 'a@32')), 'b') => ('ixor', 'a', 'b') */
   /* search1860_0_0_0 -> 16 in the cache */
   /* search1860_0_0 -> 2025 in the cache */
   /* search1860_0 -> 5166 in the cache */
   /* search1860_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 5166, 36 },
      5,
   } },

   /* replace1860_0 -> 16 in the cache */
   /* replace1860_1 -> 36 in the cache */
   /* replace1860 -> 2093 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('i2i32', ('i2i8', 'a@32')), 'b') => ('ixor', 'a', 'b') */
   /* search1861_0_0_0 -> 16 in the cache */
   /* search1861_0_0 -> 4351 in the cache */
   /* search1861_0 -> 5168 in the cache */
   /* search1861_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 5168, 36 },
      5,
   } },

   /* replace1861_0 -> 16 in the cache */
   /* replace1861_1 -> 36 in the cache */
   /* replace1861 -> 2093 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('extract_i8', 'a@32', 0), 'b') => ('ixor', 'a', 'b') */
   /* search1862_0_0 -> 16 in the cache */
   /* search1862_0_1 -> 122 in the cache */
   /* search1862_0 -> 393 in the cache */
   /* search1862_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 393, 36 },
      5,
   } },

   /* replace1862_0 -> 16 in the cache */
   /* replace1862_1 -> 36 in the cache */
   /* replace1862 -> 2093 in the cache */

   /* ('u2u32(only_lower_16_bits_used)', ('u2u16', 'a@32')) => a */
   /* search1863_0_0 -> 16 in the cache */
   /* search1863_0 -> 3310 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 3310 },
      6,
   } },

   /* replace1863 -> 16 in the cache */

   /* ('extract_u16(only_lower_16_bits_used)', 'a@32', 0) => a */
   /* search1864_0 -> 16 in the cache */
   /* search1864_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 16, 122 },
      6,
   } },

   /* replace1864 -> 16 in the cache */

   /* ('bcsel(only_lower_16_bits_used)', 'c', ('u2u32', ('u2u16', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1865_0 -> 521 in the cache */
   /* search1865_1_0_0 -> 183 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 183 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 5185 },
      -1,
   } },
   /* search1865_2 -> 559 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5186, 559 },
      6,
   } },

   /* replace1865_0 -> 521 in the cache */
   /* replace1865_1 -> 183 in the cache */
   /* replace1865_2 -> 559 in the cache */
   /* replace1865 -> 5135 in the cache */

   /* ('bcsel(only_lower_16_bits_used)', 'c', ('u2u32', ('i2i16', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1866_0 -> 521 in the cache */
   /* search1866_1_0_0 -> 183 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 183 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 5188 },
      -1,
   } },
   /* search1866_2 -> 559 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5189, 559 },
      6,
   } },

   /* replace1866_0 -> 521 in the cache */
   /* replace1866_1 -> 183 in the cache */
   /* replace1866_2 -> 559 in the cache */
   /* replace1866 -> 5135 in the cache */

   /* ('bcsel(only_lower_16_bits_used)', 'c', ('extract_u16', 'a@32', 0), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1867_0 -> 521 in the cache */
   /* search1867_1_0 -> 183 in the cache */
   /* search1867_1_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 183, 122 },
      -1,
   } },
   /* search1867_2 -> 559 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5191, 559 },
      6,
   } },

   /* replace1867_0 -> 521 in the cache */
   /* replace1867_1 -> 183 in the cache */
   /* replace1867_2 -> 559 in the cache */
   /* replace1867 -> 5135 in the cache */

   /* ('iadd(only_lower_16_bits_used)', ('u2u32', ('u2u16', 'a@32')), 'b') => ('iadd', 'a', 'b') */
   /* search1868_0_0_0 -> 16 in the cache */
   /* search1868_0_0 -> 3310 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 3310 },
      -1,
   } },
   /* search1868_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 5193, 36 },
      6,
   } },

   /* replace1868_0 -> 16 in the cache */
   /* replace1868_1 -> 36 in the cache */
   /* replace1868 -> 4065 in the cache */

   /* ('iadd(only_lower_16_bits_used)', ('u2u32', ('i2i16', 'a@32')), 'b') => ('iadd', 'a', 'b') */
   /* search1869_0_0_0 -> 16 in the cache */
   /* search1869_0_0 -> 4353 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 4353 },
      -1,
   } },
   /* search1869_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 5195, 36 },
      6,
   } },

   /* replace1869_0 -> 16 in the cache */
   /* replace1869_1 -> 36 in the cache */
   /* replace1869 -> 4065 in the cache */

   /* ('iadd(only_lower_16_bits_used)', ('extract_u16', 'a@32', 0), 'b') => ('iadd', 'a', 'b') */
   /* search1870_0_0 -> 16 in the cache */
   /* search1870_0_1 -> 122 in the cache */
   /* search1870_0 -> 444 in the cache */
   /* search1870_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 444, 36 },
      6,
   } },

   /* replace1870_0 -> 16 in the cache */
   /* replace1870_1 -> 36 in the cache */
   /* replace1870 -> 4065 in the cache */

   /* ('imul(only_lower_16_bits_used)', ('u2u32', ('u2u16', 'a@32')), 'b') => ('imul', 'a', 'b') */
   /* search1871_0_0_0 -> 16 in the cache */
   /* search1871_0_0 -> 3310 in the cache */
   /* search1871_0 -> 5193 in the cache */
   /* search1871_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 5193, 36 },
      6,
   } },

   /* replace1871_0 -> 16 in the cache */
   /* replace1871_1 -> 36 in the cache */
   /* replace1871 -> 49 in the cache */

   /* ('imul(only_lower_16_bits_used)', ('u2u32', ('i2i16', 'a@32')), 'b') => ('imul', 'a', 'b') */
   /* search1872_0_0_0 -> 16 in the cache */
   /* search1872_0_0 -> 4353 in the cache */
   /* search1872_0 -> 5195 in the cache */
   /* search1872_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 5195, 36 },
      6,
   } },

   /* replace1872_0 -> 16 in the cache */
   /* replace1872_1 -> 36 in the cache */
   /* replace1872 -> 49 in the cache */

   /* ('imul(only_lower_16_bits_used)', ('extract_u16', 'a@32', 0), 'b') => ('imul', 'a', 'b') */
   /* search1873_0_0 -> 16 in the cache */
   /* search1873_0_1 -> 122 in the cache */
   /* search1873_0 -> 444 in the cache */
   /* search1873_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 444, 36 },
      6,
   } },

   /* replace1873_0 -> 16 in the cache */
   /* replace1873_1 -> 36 in the cache */
   /* replace1873 -> 49 in the cache */

   /* ('iand(only_lower_16_bits_used)', ('u2u32', ('u2u16', 'a@32')), 'b') => ('iand', 'a', 'b') */
   /* search1874_0_0_0 -> 16 in the cache */
   /* search1874_0_0 -> 3310 in the cache */
   /* search1874_0 -> 5193 in the cache */
   /* search1874_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 5193, 36 },
      6,
   } },

   /* replace1874_0 -> 16 in the cache */
   /* replace1874_1 -> 36 in the cache */
   /* replace1874 -> 2031 in the cache */

   /* ('iand(only_lower_16_bits_used)', ('u2u32', ('i2i16', 'a@32')), 'b') => ('iand', 'a', 'b') */
   /* search1875_0_0_0 -> 16 in the cache */
   /* search1875_0_0 -> 4353 in the cache */
   /* search1875_0 -> 5195 in the cache */
   /* search1875_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 5195, 36 },
      6,
   } },

   /* replace1875_0 -> 16 in the cache */
   /* replace1875_1 -> 36 in the cache */
   /* replace1875 -> 2031 in the cache */

   /* ('iand(only_lower_16_bits_used)', ('extract_u16', 'a@32', 0), 'b') => ('iand', 'a', 'b') */
   /* search1876_0_0 -> 16 in the cache */
   /* search1876_0_1 -> 122 in the cache */
   /* search1876_0 -> 444 in the cache */
   /* search1876_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 444, 36 },
      6,
   } },

   /* replace1876_0 -> 16 in the cache */
   /* replace1876_1 -> 36 in the cache */
   /* replace1876 -> 2031 in the cache */

   /* ('ior(only_lower_16_bits_used)', ('u2u32', ('u2u16', 'a@32')), 'b') => ('ior', 'a', 'b') */
   /* search1877_0_0_0 -> 16 in the cache */
   /* search1877_0_0 -> 3310 in the cache */
   /* search1877_0 -> 5193 in the cache */
   /* search1877_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 5193, 36 },
      6,
   } },

   /* replace1877_0 -> 16 in the cache */
   /* replace1877_1 -> 36 in the cache */
   /* replace1877 -> 2069 in the cache */

   /* ('ior(only_lower_16_bits_used)', ('u2u32', ('i2i16', 'a@32')), 'b') => ('ior', 'a', 'b') */
   /* search1878_0_0_0 -> 16 in the cache */
   /* search1878_0_0 -> 4353 in the cache */
   /* search1878_0 -> 5195 in the cache */
   /* search1878_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 5195, 36 },
      6,
   } },

   /* replace1878_0 -> 16 in the cache */
   /* replace1878_1 -> 36 in the cache */
   /* replace1878 -> 2069 in the cache */

   /* ('ior(only_lower_16_bits_used)', ('extract_u16', 'a@32', 0), 'b') => ('ior', 'a', 'b') */
   /* search1879_0_0 -> 16 in the cache */
   /* search1879_0_1 -> 122 in the cache */
   /* search1879_0 -> 444 in the cache */
   /* search1879_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 444, 36 },
      6,
   } },

   /* replace1879_0 -> 16 in the cache */
   /* replace1879_1 -> 36 in the cache */
   /* replace1879 -> 2069 in the cache */

   /* ('ixor(only_lower_16_bits_used)', ('u2u32', ('u2u16', 'a@32')), 'b') => ('ixor', 'a', 'b') */
   /* search1880_0_0_0 -> 16 in the cache */
   /* search1880_0_0 -> 3310 in the cache */
   /* search1880_0 -> 5193 in the cache */
   /* search1880_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 5193, 36 },
      6,
   } },

   /* replace1880_0 -> 16 in the cache */
   /* replace1880_1 -> 36 in the cache */
   /* replace1880 -> 2093 in the cache */

   /* ('ixor(only_lower_16_bits_used)', ('u2u32', ('i2i16', 'a@32')), 'b') => ('ixor', 'a', 'b') */
   /* search1881_0_0_0 -> 16 in the cache */
   /* search1881_0_0 -> 4353 in the cache */
   /* search1881_0 -> 5195 in the cache */
   /* search1881_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 5195, 36 },
      6,
   } },

   /* replace1881_0 -> 16 in the cache */
   /* replace1881_1 -> 36 in the cache */
   /* replace1881 -> 2093 in the cache */

   /* ('ixor(only_lower_16_bits_used)', ('extract_u16', 'a@32', 0), 'b') => ('ixor', 'a', 'b') */
   /* search1882_0_0 -> 16 in the cache */
   /* search1882_0_1 -> 122 in the cache */
   /* search1882_0 -> 444 in the cache */
   /* search1882_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 444, 36 },
      6,
   } },

   /* replace1882_0 -> 16 in the cache */
   /* replace1882_1 -> 36 in the cache */
   /* replace1882 -> 2093 in the cache */

   /* ('i2i32(only_lower_16_bits_used)', ('u2u16', 'a@32')) => a */
   /* search1883_0_0 -> 16 in the cache */
   /* search1883_0 -> 3310 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 3310 },
      6,
   } },

   /* replace1883 -> 16 in the cache */

   /* ('extract_i16(only_lower_16_bits_used)', 'a@32', 0) => a */
   /* search1884_0 -> 16 in the cache */
   /* search1884_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 16, 122 },
      6,
   } },

   /* replace1884 -> 16 in the cache */

   /* ('bcsel(only_lower_16_bits_used)', 'c', ('i2i32', ('u2u16', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1885_0 -> 521 in the cache */
   /* search1885_1_0_0 -> 183 in the cache */
   /* search1885_1_0 -> 5185 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 5185 },
      -1,
   } },
   /* search1885_2 -> 559 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5212, 559 },
      6,
   } },

   /* replace1885_0 -> 521 in the cache */
   /* replace1885_1 -> 183 in the cache */
   /* replace1885_2 -> 559 in the cache */
   /* replace1885 -> 5135 in the cache */

   /* ('bcsel(only_lower_16_bits_used)', 'c', ('i2i32', ('i2i16', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1886_0 -> 521 in the cache */
   /* search1886_1_0_0 -> 183 in the cache */
   /* search1886_1_0 -> 5188 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 5188 },
      -1,
   } },
   /* search1886_2 -> 559 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5214, 559 },
      6,
   } },

   /* replace1886_0 -> 521 in the cache */
   /* replace1886_1 -> 183 in the cache */
   /* replace1886_2 -> 559 in the cache */
   /* replace1886 -> 5135 in the cache */

   /* ('bcsel(only_lower_16_bits_used)', 'c', ('extract_i16', 'a@32', 0), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1887_0 -> 521 in the cache */
   /* search1887_1_0 -> 183 in the cache */
   /* search1887_1_1 -> 122 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 183, 122 },
      -1,
   } },
   /* search1887_2 -> 559 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5216, 559 },
      6,
   } },

   /* replace1887_0 -> 521 in the cache */
   /* replace1887_1 -> 183 in the cache */
   /* replace1887_2 -> 559 in the cache */
   /* replace1887 -> 5135 in the cache */

   /* ('iadd(only_lower_16_bits_used)', ('i2i32', ('u2u16', 'a@32')), 'b') => ('iadd', 'a', 'b') */
   /* search1888_0_0_0 -> 16 in the cache */
   /* search1888_0_0 -> 3310 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 3310 },
      -1,
   } },
   /* search1888_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 5218, 36 },
      6,
   } },

   /* replace1888_0 -> 16 in the cache */
   /* replace1888_1 -> 36 in the cache */
   /* replace1888 -> 4065 in the cache */

   /* ('iadd(only_lower_16_bits_used)', ('i2i32', ('i2i16', 'a@32')), 'b') => ('iadd', 'a', 'b') */
   /* search1889_0_0_0 -> 16 in the cache */
   /* search1889_0_0 -> 4353 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 4353 },
      -1,
   } },
   /* search1889_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 5220, 36 },
      6,
   } },

   /* replace1889_0 -> 16 in the cache */
   /* replace1889_1 -> 36 in the cache */
   /* replace1889 -> 4065 in the cache */

   /* ('iadd(only_lower_16_bits_used)', ('extract_i16', 'a@32', 0), 'b') => ('iadd', 'a', 'b') */
   /* search1890_0_0 -> 16 in the cache */
   /* search1890_0_1 -> 122 in the cache */
   /* search1890_0 -> 436 in the cache */
   /* search1890_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 436, 36 },
      6,
   } },

   /* replace1890_0 -> 16 in the cache */
   /* replace1890_1 -> 36 in the cache */
   /* replace1890 -> 4065 in the cache */

   /* ('imul(only_lower_16_bits_used)', ('i2i32', ('u2u16', 'a@32')), 'b') => ('imul', 'a', 'b') */
   /* search1891_0_0_0 -> 16 in the cache */
   /* search1891_0_0 -> 3310 in the cache */
   /* search1891_0 -> 5218 in the cache */
   /* search1891_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 5218, 36 },
      6,
   } },

   /* replace1891_0 -> 16 in the cache */
   /* replace1891_1 -> 36 in the cache */
   /* replace1891 -> 49 in the cache */

   /* ('imul(only_lower_16_bits_used)', ('i2i32', ('i2i16', 'a@32')), 'b') => ('imul', 'a', 'b') */
   /* search1892_0_0_0 -> 16 in the cache */
   /* search1892_0_0 -> 4353 in the cache */
   /* search1892_0 -> 5220 in the cache */
   /* search1892_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 5220, 36 },
      6,
   } },

   /* replace1892_0 -> 16 in the cache */
   /* replace1892_1 -> 36 in the cache */
   /* replace1892 -> 49 in the cache */

   /* ('imul(only_lower_16_bits_used)', ('extract_i16', 'a@32', 0), 'b') => ('imul', 'a', 'b') */
   /* search1893_0_0 -> 16 in the cache */
   /* search1893_0_1 -> 122 in the cache */
   /* search1893_0 -> 436 in the cache */
   /* search1893_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 436, 36 },
      6,
   } },

   /* replace1893_0 -> 16 in the cache */
   /* replace1893_1 -> 36 in the cache */
   /* replace1893 -> 49 in the cache */

   /* ('iand(only_lower_16_bits_used)', ('i2i32', ('u2u16', 'a@32')), 'b') => ('iand', 'a', 'b') */
   /* search1894_0_0_0 -> 16 in the cache */
   /* search1894_0_0 -> 3310 in the cache */
   /* search1894_0 -> 5218 in the cache */
   /* search1894_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 5218, 36 },
      6,
   } },

   /* replace1894_0 -> 16 in the cache */
   /* replace1894_1 -> 36 in the cache */
   /* replace1894 -> 2031 in the cache */

   /* ('iand(only_lower_16_bits_used)', ('i2i32', ('i2i16', 'a@32')), 'b') => ('iand', 'a', 'b') */
   /* search1895_0_0_0 -> 16 in the cache */
   /* search1895_0_0 -> 4353 in the cache */
   /* search1895_0 -> 5220 in the cache */
   /* search1895_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 5220, 36 },
      6,
   } },

   /* replace1895_0 -> 16 in the cache */
   /* replace1895_1 -> 36 in the cache */
   /* replace1895 -> 2031 in the cache */

   /* ('iand(only_lower_16_bits_used)', ('extract_i16', 'a@32', 0), 'b') => ('iand', 'a', 'b') */
   /* search1896_0_0 -> 16 in the cache */
   /* search1896_0_1 -> 122 in the cache */
   /* search1896_0 -> 436 in the cache */
   /* search1896_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 436, 36 },
      6,
   } },

   /* replace1896_0 -> 16 in the cache */
   /* replace1896_1 -> 36 in the cache */
   /* replace1896 -> 2031 in the cache */

   /* ('ior(only_lower_16_bits_used)', ('i2i32', ('u2u16', 'a@32')), 'b') => ('ior', 'a', 'b') */
   /* search1897_0_0_0 -> 16 in the cache */
   /* search1897_0_0 -> 3310 in the cache */
   /* search1897_0 -> 5218 in the cache */
   /* search1897_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 5218, 36 },
      6,
   } },

   /* replace1897_0 -> 16 in the cache */
   /* replace1897_1 -> 36 in the cache */
   /* replace1897 -> 2069 in the cache */

   /* ('ior(only_lower_16_bits_used)', ('i2i32', ('i2i16', 'a@32')), 'b') => ('ior', 'a', 'b') */
   /* search1898_0_0_0 -> 16 in the cache */
   /* search1898_0_0 -> 4353 in the cache */
   /* search1898_0 -> 5220 in the cache */
   /* search1898_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 5220, 36 },
      6,
   } },

   /* replace1898_0 -> 16 in the cache */
   /* replace1898_1 -> 36 in the cache */
   /* replace1898 -> 2069 in the cache */

   /* ('ior(only_lower_16_bits_used)', ('extract_i16', 'a@32', 0), 'b') => ('ior', 'a', 'b') */
   /* search1899_0_0 -> 16 in the cache */
   /* search1899_0_1 -> 122 in the cache */
   /* search1899_0 -> 436 in the cache */
   /* search1899_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 436, 36 },
      6,
   } },

   /* replace1899_0 -> 16 in the cache */
   /* replace1899_1 -> 36 in the cache */
   /* replace1899 -> 2069 in the cache */

   /* ('ixor(only_lower_16_bits_used)', ('i2i32', ('u2u16', 'a@32')), 'b') => ('ixor', 'a', 'b') */
   /* search1900_0_0_0 -> 16 in the cache */
   /* search1900_0_0 -> 3310 in the cache */
   /* search1900_0 -> 5218 in the cache */
   /* search1900_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 5218, 36 },
      6,
   } },

   /* replace1900_0 -> 16 in the cache */
   /* replace1900_1 -> 36 in the cache */
   /* replace1900 -> 2093 in the cache */

   /* ('ixor(only_lower_16_bits_used)', ('i2i32', ('i2i16', 'a@32')), 'b') => ('ixor', 'a', 'b') */
   /* search1901_0_0_0 -> 16 in the cache */
   /* search1901_0_0 -> 4353 in the cache */
   /* search1901_0 -> 5220 in the cache */
   /* search1901_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 5220, 36 },
      6,
   } },

   /* replace1901_0 -> 16 in the cache */
   /* replace1901_1 -> 36 in the cache */
   /* replace1901 -> 2093 in the cache */

   /* ('ixor(only_lower_16_bits_used)', ('extract_i16', 'a@32', 0), 'b') => ('ixor', 'a', 'b') */
   /* search1902_0_0 -> 16 in the cache */
   /* search1902_0_1 -> 122 in the cache */
   /* search1902_0 -> 436 in the cache */
   /* search1902_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 436, 36 },
      6,
   } },

   /* replace1902_0 -> 16 in the cache */
   /* replace1902_1 -> 36 in the cache */
   /* replace1902 -> 2093 in the cache */

   /* ('ldexp@16', 'x', 'exp') => ('fmul', ('fmul', 'x', ('i2i16', ('ishl', ('iadd', ('ishr', ('imin', ('imax', 'exp', -30), 30), 1), 15), 10))), ('i2i16', ('ishl', ('iadd', ('isub', ('imin', ('imax', 'exp', -30), 30), ('ishr', ('imin', ('imax', 'exp', -30), 30), 1)), 15), 10))) */
   { .variable = {
      { nir_search_value_variable, 16 },
      0, /* x */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* exp */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ldexp,
      -1, 0,
      { 5235, 5236 },
      -1,
   } },

   /* replace1903_0_0 -> 5235 in the cache */
   /* replace1903_0_1_0_0_0_0_0_0 -> 5236 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x1e /* -30 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imax,
      4, 1,
      { 5236, 5238 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x1e /* 30 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      3, 2,
      { 5239, 5240 },
      -1,
   } },
   /* replace1903_0_1_0_0_0_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 2,
      { 5241, 396 },
      -1,
   } },
   /* replace1903_0_1_0_0_1 -> 14 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 3,
      { 5242, 14 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xa /* 10 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 3,
      { 5243, 5244 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 3,
      { 5245 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 4,
      { 5235, 5246 },
      -1,
   } },
   /* replace1903_1_0_0_0_0_0_0 -> 5236 in the cache */
   /* replace1903_1_0_0_0_0_0_1 -> 5238 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imax,
      7, 1,
      { 5236, 5238 },
      -1,
   } },
   /* replace1903_1_0_0_0_0_1 -> 5240 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      6, 2,
      { 5248, 5240 },
      -1,
   } },
   /* replace1903_1_0_0_0_1_0_0_0 -> 5236 in the cache */
   /* replace1903_1_0_0_0_1_0_0_1 -> 5238 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imax,
      9, 1,
      { 5236, 5238 },
      -1,
   } },
   /* replace1903_1_0_0_0_1_0_1 -> 5240 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      8, 2,
      { 5250, 5240 },
      -1,
   } },
   /* replace1903_1_0_0_0_1_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 2,
      { 5251, 396 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 4,
      { 5249, 5252 },
      -1,
   } },
   /* replace1903_1_0_0_1 -> 14 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      5, 5,
      { 5253, 14 },
      -1,
   } },
   /* replace1903_1_0_1 -> 5244 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 5,
      { 5254, 5244 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 5,
      { 5255 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 10,
      { 5247, 5256 },
      -1,
   } },

   /* ('ldexp@32', 'x', 'exp') => ('fmul', ('fmul', 'x', ('ishl', ('iadd', ('ishr', ('imin', ('imax', 'exp', -254), 254), 1), 127), 23)), ('ishl', ('iadd', ('isub', ('imin', ('imax', 'exp', -254), 254), ('ishr', ('imin', ('imax', 'exp', -254), 254), 1)), 127), 23)) */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* x */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1904_1 -> 5236 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ldexp,
      -1, 0,
      { 5258, 5236 },
      -1,
   } },

   /* replace1904_0_0 -> 5258 in the cache */
   /* replace1904_0_1_0_0_0_0_0 -> 5236 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0xfe /* -254 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imax,
      4, 1,
      { 5236, 5260 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xfe /* 254 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      3, 2,
      { 5261, 5262 },
      -1,
   } },
   /* replace1904_0_1_0_0_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 2,
      { 5263, 396 },
      -1,
   } },
   /* replace1904_0_1_0_1 -> 4437 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 3,
      { 5264, 4437 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x17 /* 23 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 3,
      { 5265, 5266 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 4,
      { 5258, 5267 },
      -1,
   } },
   /* replace1904_1_0_0_0_0_0 -> 5236 in the cache */
   /* replace1904_1_0_0_0_0_1 -> 5260 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imax,
      7, 1,
      { 5236, 5260 },
      -1,
   } },
   /* replace1904_1_0_0_0_1 -> 5262 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      6, 2,
      { 5269, 5262 },
      -1,
   } },
   /* replace1904_1_0_0_1_0_0_0 -> 5236 in the cache */
   /* replace1904_1_0_0_1_0_0_1 -> 5260 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imax,
      9, 1,
      { 5236, 5260 },
      -1,
   } },
   /* replace1904_1_0_0_1_0_1 -> 5262 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      8, 2,
      { 5271, 5262 },
      -1,
   } },
   /* replace1904_1_0_0_1_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 2,
      { 5272, 396 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 4,
      { 5270, 5273 },
      -1,
   } },
   /* replace1904_1_0_1 -> 4437 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      5, 5,
      { 5274, 4437 },
      -1,
   } },
   /* replace1904_1_1 -> 5266 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 5,
      { 5275, 5266 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 10,
      { 5268, 5276 },
      -1,
   } },

   /* ('ldexp@64', 'x', 'exp') => ('fmul', ('fmul', 'x', ('pack_64_2x32_split', 0, ('ishl', ('iadd', ('ishr', ('imin', ('imax', 'exp', -2046), 2046), 1), 1023), 20))), ('pack_64_2x32_split', 0, ('ishl', ('iadd', ('isub', ('imin', ('imax', 'exp', -2046), 2046), ('ishr', ('imin', ('imax', 'exp', -2046), 2046), 1)), 1023), 20))) */
   { .variable = {
      { nir_search_value_variable, 64 },
      0, /* x */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1905_1 -> 5236 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ldexp,
      -1, 0,
      { 5278, 5236 },
      -1,
   } },

   /* replace1905_0_0 -> 5278 in the cache */
   /* replace1905_0_1_0 -> 122 in the cache */
   /* replace1905_0_1_1_0_0_0_0_0 -> 5236 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x7fe /* -2046 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imax,
      4, 1,
      { 5236, 5280 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x7fe /* 2046 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      3, 2,
      { 5281, 5282 },
      -1,
   } },
   /* replace1905_0_1_1_0_0_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 2,
      { 5283, 396 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x3ff /* 1023 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 3,
      { 5284, 5285 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x14 /* 20 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 3,
      { 5286, 5287 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 3,
      { 122, 5288 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 4,
      { 5278, 5289 },
      -1,
   } },
   /* replace1905_1_0 -> 122 in the cache */
   /* replace1905_1_1_0_0_0_0_0 -> 5236 in the cache */
   /* replace1905_1_1_0_0_0_0_1 -> 5280 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imax,
      7, 1,
      { 5236, 5280 },
      -1,
   } },
   /* replace1905_1_1_0_0_0_1 -> 5282 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      6, 2,
      { 5291, 5282 },
      -1,
   } },
   /* replace1905_1_1_0_0_1_0_0_0 -> 5236 in the cache */
   /* replace1905_1_1_0_0_1_0_0_1 -> 5280 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imax,
      9, 1,
      { 5236, 5280 },
      -1,
   } },
   /* replace1905_1_1_0_0_1_0_1 -> 5282 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      8, 2,
      { 5293, 5282 },
      -1,
   } },
   /* replace1905_1_1_0_0_1_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 2,
      { 5294, 396 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 4,
      { 5292, 5295 },
      -1,
   } },
   /* replace1905_1_1_0_1 -> 5285 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      5, 5,
      { 5296, 5285 },
      -1,
   } },
   /* replace1905_1_1_1 -> 5287 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 5,
      { 5297, 5287 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 5,
      { 122, 5298 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 10,
      { 5290, 5299 },
      -1,
   } },

   /* ('ior(many-comm-expr)', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 252645135), 4), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 4042322160), 4)), 858993459), 2), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 252645135), 4), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 4042322160), 4)), 3435973836), 2)), 1431655765), 1), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 252645135), 4), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 4042322160), 4)), 858993459), 2), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 252645135), 4), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 4042322160), 4)), 3435973836), 2)), 2863311530), 1)) => ('bitfield_reverse', 'x') */
   /* search1906_0_0_0_0_0_0_0_0_0_0_0_0_0_0 -> 5258 in the cache */
   /* search1906_0_0_0_0_0_0_0_0_0_0_0_0_0_1 -> 853 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 5258, 853 },
      -1,
   } },
   /* search1906_0_0_0_0_0_0_0_0_0_0_0_0_1_0 -> 5258 in the cache */
   /* search1906_0_0_0_0_0_0_0_0_0_0_0_0_1_1 -> 853 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 5258, 853 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      8, 1,
      { 5301, 5302 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xff00ff /* 16711935 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      7, 2,
      { 5303, 5304 },
      -1,
   } },
   /* search1906_0_0_0_0_0_0_0_0_0_0_1 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 2,
      { 5305, 831 },
      -1,
   } },
   /* search1906_0_0_0_0_0_0_0_0_0_1_0_0_0_0 -> 5258 in the cache */
   /* search1906_0_0_0_0_0_0_0_0_0_1_0_0_0_1 -> 853 in the cache */
   /* search1906_0_0_0_0_0_0_0_0_0_1_0_0_0 -> 5301 in the cache */
   /* search1906_0_0_0_0_0_0_0_0_0_1_0_0_1_0 -> 5258 in the cache */
   /* search1906_0_0_0_0_0_0_0_0_0_1_0_0_1_1 -> 853 in the cache */
   /* search1906_0_0_0_0_0_0_0_0_0_1_0_0_1 -> 5302 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      10, 1,
      { 5301, 5302 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xff00ff00 /* 4278255360 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      9, 2,
      { 5307, 5308 },
      -1,
   } },
   /* search1906_0_0_0_0_0_0_0_0_0_1_1 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 2,
      { 5309, 831 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      6, 5,
      { 5306, 5310 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xf0f0f0f /* 252645135 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      5, 6,
      { 5311, 5312 },
      -1,
   } },
   /* search1906_0_0_0_0_0_0_0_1 -> 2198 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 6,
      { 5313, 2198 },
      -1,
   } },
   /* search1906_0_0_0_0_0_0_1_0_0_0_0_0_0_0 -> 5258 in the cache */
   /* search1906_0_0_0_0_0_0_1_0_0_0_0_0_0_1 -> 853 in the cache */
   /* search1906_0_0_0_0_0_0_1_0_0_0_0_0_0 -> 5301 in the cache */
   /* search1906_0_0_0_0_0_0_1_0_0_0_0_0_1_0 -> 5258 in the cache */
   /* search1906_0_0_0_0_0_0_1_0_0_0_0_0_1_1 -> 853 in the cache */
   /* search1906_0_0_0_0_0_0_1_0_0_0_0_0_1 -> 5302 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      14, 1,
      { 5301, 5302 },
      -1,
   } },
   /* search1906_0_0_0_0_0_0_1_0_0_0_0_1 -> 5304 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      13, 2,
      { 5315, 5304 },
      -1,
   } },
   /* search1906_0_0_0_0_0_0_1_0_0_0_1 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 2,
      { 5316, 831 },
      -1,
   } },
   /* search1906_0_0_0_0_0_0_1_0_0_1_0_0_0_0 -> 5258 in the cache */
   /* search1906_0_0_0_0_0_0_1_0_0_1_0_0_0_1 -> 853 in the cache */
   /* search1906_0_0_0_0_0_0_1_0_0_1_0_0_0 -> 5301 in the cache */
   /* search1906_0_0_0_0_0_0_1_0_0_1_0_0_1_0 -> 5258 in the cache */
   /* search1906_0_0_0_0_0_0_1_0_0_1_0_0_1_1 -> 853 in the cache */
   /* search1906_0_0_0_0_0_0_1_0_0_1_0_0_1 -> 5302 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      16, 1,
      { 5301, 5302 },
      -1,
   } },
   /* search1906_0_0_0_0_0_0_1_0_0_1_0_1 -> 5308 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      15, 2,
      { 5318, 5308 },
      -1,
   } },
   /* search1906_0_0_0_0_0_0_1_0_0_1_1 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 2,
      { 5319, 831 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      12, 5,
      { 5317, 5320 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xf0f0f0f0 /* 4042322160 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      11, 6,
      { 5321, 5322 },
      -1,
   } },
   /* search1906_0_0_0_0_0_0_1_1 -> 2198 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 6,
      { 5323, 2198 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      4, 13,
      { 5314, 5324 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x33333333 /* 858993459 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      3, 14,
      { 5325, 5326 },
      -1,
   } },
   /* search1906_0_0_0_0_1 -> 401 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 14,
      { 5327, 401 },
      -1,
   } },
   /* search1906_0_0_0_1_0_0_0_0_0_0_0_0_0_0 -> 5258 in the cache */
   /* search1906_0_0_0_1_0_0_0_0_0_0_0_0_0_1 -> 853 in the cache */
   /* search1906_0_0_0_1_0_0_0_0_0_0_0_0_0 -> 5301 in the cache */
   /* search1906_0_0_0_1_0_0_0_0_0_0_0_0_1_0 -> 5258 in the cache */
   /* search1906_0_0_0_1_0_0_0_0_0_0_0_0_1_1 -> 853 in the cache */
   /* search1906_0_0_0_1_0_0_0_0_0_0_0_0_1 -> 5302 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      22, 1,
      { 5301, 5302 },
      -1,
   } },
   /* search1906_0_0_0_1_0_0_0_0_0_0_0_1 -> 5304 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      21, 2,
      { 5329, 5304 },
      -1,
   } },
   /* search1906_0_0_0_1_0_0_0_0_0_0_1 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 2,
      { 5330, 831 },
      -1,
   } },
   /* search1906_0_0_0_1_0_0_0_0_0_1_0_0_0_0 -> 5258 in the cache */
   /* search1906_0_0_0_1_0_0_0_0_0_1_0_0_0_1 -> 853 in the cache */
   /* search1906_0_0_0_1_0_0_0_0_0_1_0_0_0 -> 5301 in the cache */
   /* search1906_0_0_0_1_0_0_0_0_0_1_0_0_1_0 -> 5258 in the cache */
   /* search1906_0_0_0_1_0_0_0_0_0_1_0_0_1_1 -> 853 in the cache */
   /* search1906_0_0_0_1_0_0_0_0_0_1_0_0_1 -> 5302 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      24, 1,
      { 5301, 5302 },
      -1,
   } },
   /* search1906_0_0_0_1_0_0_0_0_0_1_0_1 -> 5308 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      23, 2,
      { 5332, 5308 },
      -1,
   } },
   /* search1906_0_0_0_1_0_0_0_0_0_1_1 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 2,
      { 5333, 831 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      20, 5,
      { 5331, 5334 },
      -1,
   } },
   /* search1906_0_0_0_1_0_0_0_0_1 -> 5312 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      19, 6,
      { 5335, 5312 },
      -1,
   } },
   /* search1906_0_0_0_1_0_0_0_1 -> 2198 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 6,
      { 5336, 2198 },
      -1,
   } },
   /* search1906_0_0_0_1_0_0_1_0_0_0_0_0_0_0 -> 5258 in the cache */
   /* search1906_0_0_0_1_0_0_1_0_0_0_0_0_0_1 -> 853 in the cache */
   /* search1906_0_0_0_1_0_0_1_0_0_0_0_0_0 -> 5301 in the cache */
   /* search1906_0_0_0_1_0_0_1_0_0_0_0_0_1_0 -> 5258 in the cache */
   /* search1906_0_0_0_1_0_0_1_0_0_0_0_0_1_1 -> 853 in the cache */
   /* search1906_0_0_0_1_0_0_1_0_0_0_0_0_1 -> 5302 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      28, 1,
      { 5301, 5302 },
      -1,
   } },
   /* search1906_0_0_0_1_0_0_1_0_0_0_0_1 -> 5304 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      27, 2,
      { 5338, 5304 },
      -1,
   } },
   /* search1906_0_0_0_1_0_0_1_0_0_0_1 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 2,
      { 5339, 831 },
      -1,
   } },
   /* search1906_0_0_0_1_0_0_1_0_0_1_0_0_0_0 -> 5258 in the cache */
   /* search1906_0_0_0_1_0_0_1_0_0_1_0_0_0_1 -> 853 in the cache */
   /* search1906_0_0_0_1_0_0_1_0_0_1_0_0_0 -> 5301 in the cache */
   /* search1906_0_0_0_1_0_0_1_0_0_1_0_0_1_0 -> 5258 in the cache */
   /* search1906_0_0_0_1_0_0_1_0_0_1_0_0_1_1 -> 853 in the cache */
   /* search1906_0_0_0_1_0_0_1_0_0_1_0_0_1 -> 5302 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      30, 1,
      { 5301, 5302 },
      -1,
   } },
   /* search1906_0_0_0_1_0_0_1_0_0_1_0_1 -> 5308 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      29, 2,
      { 5341, 5308 },
      -1,
   } },
   /* search1906_0_0_0_1_0_0_1_0_0_1_1 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 2,
      { 5342, 831 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      26, 5,
      { 5340, 5343 },
      -1,
   } },
   /* search1906_0_0_0_1_0_0_1_0_1 -> 5322 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      25, 6,
      { 5344, 5322 },
      -1,
   } },
   /* search1906_0_0_0_1_0_0_1_1 -> 2198 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 6,
      { 5345, 2198 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      18, 13,
      { 5337, 5346 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xcccccccc /* 3435973836 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      17, 14,
      { 5347, 5348 },
      -1,
   } },
   /* search1906_0_0_0_1_1 -> 401 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 14,
      { 5349, 401 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      2, 29,
      { 5328, 5350 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x55555555 /* 1431655765 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 30,
      { 5351, 5352 },
      -1,
   } },
   /* search1906_0_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 30,
      { 5353, 396 },
      -1,
   } },
   /* search1906_1_0_0_0_0_0_0_0_0_0_0_0_0_0 -> 5258 in the cache */
   /* search1906_1_0_0_0_0_0_0_0_0_0_0_0_0_1 -> 853 in the cache */
   /* search1906_1_0_0_0_0_0_0_0_0_0_0_0_0 -> 5301 in the cache */
   /* search1906_1_0_0_0_0_0_0_0_0_0_0_0_1_0 -> 5258 in the cache */
   /* search1906_1_0_0_0_0_0_0_0_0_0_0_0_1_1 -> 853 in the cache */
   /* search1906_1_0_0_0_0_0_0_0_0_0_0_0_1 -> 5302 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      38, 1,
      { 5301, 5302 },
      -1,
   } },
   /* search1906_1_0_0_0_0_0_0_0_0_0_0_1 -> 5304 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      37, 2,
      { 5355, 5304 },
      -1,
   } },
   /* search1906_1_0_0_0_0_0_0_0_0_0_1 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 2,
      { 5356, 831 },
      -1,
   } },
   /* search1906_1_0_0_0_0_0_0_0_0_1_0_0_0_0 -> 5258 in the cache */
   /* search1906_1_0_0_0_0_0_0_0_0_1_0_0_0_1 -> 853 in the cache */
   /* search1906_1_0_0_0_0_0_0_0_0_1_0_0_0 -> 5301 in the cache */
   /* search1906_1_0_0_0_0_0_0_0_0_1_0_0_1_0 -> 5258 in the cache */
   /* search1906_1_0_0_0_0_0_0_0_0_1_0_0_1_1 -> 853 in the cache */
   /* search1906_1_0_0_0_0_0_0_0_0_1_0_0_1 -> 5302 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      40, 1,
      { 5301, 5302 },
      -1,
   } },
   /* search1906_1_0_0_0_0_0_0_0_0_1_0_1 -> 5308 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      39, 2,
      { 5358, 5308 },
      -1,
   } },
   /* search1906_1_0_0_0_0_0_0_0_0_1_1 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 2,
      { 5359, 831 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      36, 5,
      { 5357, 5360 },
      -1,
   } },
   /* search1906_1_0_0_0_0_0_0_0_1 -> 5312 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      35, 6,
      { 5361, 5312 },
      -1,
   } },
   /* search1906_1_0_0_0_0_0_0_1 -> 2198 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 6,
      { 5362, 2198 },
      -1,
   } },
   /* search1906_1_0_0_0_0_0_1_0_0_0_0_0_0_0 -> 5258 in the cache */
   /* search1906_1_0_0_0_0_0_1_0_0_0_0_0_0_1 -> 853 in the cache */
   /* search1906_1_0_0_0_0_0_1_0_0_0_0_0_0 -> 5301 in the cache */
   /* search1906_1_0_0_0_0_0_1_0_0_0_0_0_1_0 -> 5258 in the cache */
   /* search1906_1_0_0_0_0_0_1_0_0_0_0_0_1_1 -> 853 in the cache */
   /* search1906_1_0_0_0_0_0_1_0_0_0_0_0_1 -> 5302 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      44, 1,
      { 5301, 5302 },
      -1,
   } },
   /* search1906_1_0_0_0_0_0_1_0_0_0_0_1 -> 5304 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      43, 2,
      { 5364, 5304 },
      -1,
   } },
   /* search1906_1_0_0_0_0_0_1_0_0_0_1 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 2,
      { 5365, 831 },
      -1,
   } },
   /* search1906_1_0_0_0_0_0_1_0_0_1_0_0_0_0 -> 5258 in the cache */
   /* search1906_1_0_0_0_0_0_1_0_0_1_0_0_0_1 -> 853 in the cache */
   /* search1906_1_0_0_0_0_0_1_0_0_1_0_0_0 -> 5301 in the cache */
   /* search1906_1_0_0_0_0_0_1_0_0_1_0_0_1_0 -> 5258 in the cache */
   /* search1906_1_0_0_0_0_0_1_0_0_1_0_0_1_1 -> 853 in the cache */
   /* search1906_1_0_0_0_0_0_1_0_0_1_0_0_1 -> 5302 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      46, 1,
      { 5301, 5302 },
      -1,
   } },
   /* search1906_1_0_0_0_0_0_1_0_0_1_0_1 -> 5308 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      45, 2,
      { 5367, 5308 },
      -1,
   } },
   /* search1906_1_0_0_0_0_0_1_0_0_1_1 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 2,
      { 5368, 831 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      42, 5,
      { 5366, 5369 },
      -1,
   } },
   /* search1906_1_0_0_0_0_0_1_0_1 -> 5322 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      41, 6,
      { 5370, 5322 },
      -1,
   } },
   /* search1906_1_0_0_0_0_0_1_1 -> 2198 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 6,
      { 5371, 2198 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      34, 13,
      { 5363, 5372 },
      -1,
   } },
   /* search1906_1_0_0_0_0_1 -> 5326 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      33, 14,
      { 5373, 5326 },
      -1,
   } },
   /* search1906_1_0_0_0_1 -> 401 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 14,
      { 5374, 401 },
      -1,
   } },
   /* search1906_1_0_0_1_0_0_0_0_0_0_0_0_0_0 -> 5258 in the cache */
   /* search1906_1_0_0_1_0_0_0_0_0_0_0_0_0_1 -> 853 in the cache */
   /* search1906_1_0_0_1_0_0_0_0_0_0_0_0_0 -> 5301 in the cache */
   /* search1906_1_0_0_1_0_0_0_0_0_0_0_0_1_0 -> 5258 in the cache */
   /* search1906_1_0_0_1_0_0_0_0_0_0_0_0_1_1 -> 853 in the cache */
   /* search1906_1_0_0_1_0_0_0_0_0_0_0_0_1 -> 5302 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      52, 1,
      { 5301, 5302 },
      -1,
   } },
   /* search1906_1_0_0_1_0_0_0_0_0_0_0_1 -> 5304 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      51, 2,
      { 5376, 5304 },
      -1,
   } },
   /* search1906_1_0_0_1_0_0_0_0_0_0_1 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 2,
      { 5377, 831 },
      -1,
   } },
   /* search1906_1_0_0_1_0_0_0_0_0_1_0_0_0_0 -> 5258 in the cache */
   /* search1906_1_0_0_1_0_0_0_0_0_1_0_0_0_1 -> 853 in the cache */
   /* search1906_1_0_0_1_0_0_0_0_0_1_0_0_0 -> 5301 in the cache */
   /* search1906_1_0_0_1_0_0_0_0_0_1_0_0_1_0 -> 5258 in the cache */
   /* search1906_1_0_0_1_0_0_0_0_0_1_0_0_1_1 -> 853 in the cache */
   /* search1906_1_0_0_1_0_0_0_0_0_1_0_0_1 -> 5302 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      54, 1,
      { 5301, 5302 },
      -1,
   } },
   /* search1906_1_0_0_1_0_0_0_0_0_1_0_1 -> 5308 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      53, 2,
      { 5379, 5308 },
      -1,
   } },
   /* search1906_1_0_0_1_0_0_0_0_0_1_1 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 2,
      { 5380, 831 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      50, 5,
      { 5378, 5381 },
      -1,
   } },
   /* search1906_1_0_0_1_0_0_0_0_1 -> 5312 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      49, 6,
      { 5382, 5312 },
      -1,
   } },
   /* search1906_1_0_0_1_0_0_0_1 -> 2198 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 6,
      { 5383, 2198 },
      -1,
   } },
   /* search1906_1_0_0_1_0_0_1_0_0_0_0_0_0_0 -> 5258 in the cache */
   /* search1906_1_0_0_1_0_0_1_0_0_0_0_0_0_1 -> 853 in the cache */
   /* search1906_1_0_0_1_0_0_1_0_0_0_0_0_0 -> 5301 in the cache */
   /* search1906_1_0_0_1_0_0_1_0_0_0_0_0_1_0 -> 5258 in the cache */
   /* search1906_1_0_0_1_0_0_1_0_0_0_0_0_1_1 -> 853 in the cache */
   /* search1906_1_0_0_1_0_0_1_0_0_0_0_0_1 -> 5302 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      58, 1,
      { 5301, 5302 },
      -1,
   } },
   /* search1906_1_0_0_1_0_0_1_0_0_0_0_1 -> 5304 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      57, 2,
      { 5385, 5304 },
      -1,
   } },
   /* search1906_1_0_0_1_0_0_1_0_0_0_1 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 2,
      { 5386, 831 },
      -1,
   } },
   /* search1906_1_0_0_1_0_0_1_0_0_1_0_0_0_0 -> 5258 in the cache */
   /* search1906_1_0_0_1_0_0_1_0_0_1_0_0_0_1 -> 853 in the cache */
   /* search1906_1_0_0_1_0_0_1_0_0_1_0_0_0 -> 5301 in the cache */
   /* search1906_1_0_0_1_0_0_1_0_0_1_0_0_1_0 -> 5258 in the cache */
   /* search1906_1_0_0_1_0_0_1_0_0_1_0_0_1_1 -> 853 in the cache */
   /* search1906_1_0_0_1_0_0_1_0_0_1_0_0_1 -> 5302 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      60, 1,
      { 5301, 5302 },
      -1,
   } },
   /* search1906_1_0_0_1_0_0_1_0_0_1_0_1 -> 5308 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      59, 2,
      { 5388, 5308 },
      -1,
   } },
   /* search1906_1_0_0_1_0_0_1_0_0_1_1 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 2,
      { 5389, 831 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      56, 5,
      { 5387, 5390 },
      -1,
   } },
   /* search1906_1_0_0_1_0_0_1_0_1 -> 5322 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      55, 6,
      { 5391, 5322 },
      -1,
   } },
   /* search1906_1_0_0_1_0_0_1_1 -> 2198 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 6,
      { 5392, 2198 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      48, 13,
      { 5384, 5393 },
      -1,
   } },
   /* search1906_1_0_0_1_0_1 -> 5348 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      47, 14,
      { 5394, 5348 },
      -1,
   } },
   /* search1906_1_0_0_1_1 -> 401 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 14,
      { 5395, 401 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      32, 29,
      { 5375, 5396 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xaaaaaaaa /* 2863311530 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      31, 30,
      { 5397, 5398 },
      -1,
   } },
   /* search1906_1_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 30,
      { 5399, 396 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 61,
      { 5354, 5400 },
      -1,
   } },

   /* replace1906_0 -> 5258 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bitfield_reverse,
      -1, 0,
      { 5258 },
      -1,
   } },

   /* ('ior(many-comm-expr)', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 3435973836), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 858993459)), 4), 4042322160), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 3435973836), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 858993459)), 4), 252645135)), 8), 4278255360), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 3435973836), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 858993459)), 4), 4042322160), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 3435973836), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 858993459)), 4), 252645135)), 8), 16711935)) => ('bitfield_reverse', 'x') */
   /* search1907_0_0_0_0_0_0_0_0_0_0_0_0_0_0 -> 5258 in the cache */
   /* search1907_0_0_0_0_0_0_0_0_0_0_0_0_0_1 -> 853 in the cache */
   /* search1907_0_0_0_0_0_0_0_0_0_0_0_0_0 -> 5301 in the cache */
   /* search1907_0_0_0_0_0_0_0_0_0_0_0_0_1_0 -> 5258 in the cache */
   /* search1907_0_0_0_0_0_0_0_0_0_0_0_0_1_1 -> 853 in the cache */
   /* search1907_0_0_0_0_0_0_0_0_0_0_0_0_1 -> 5302 in the cache */
   /* search1907_0_0_0_0_0_0_0_0_0_0_0_0 -> 5303 in the cache */
   /* search1907_0_0_0_0_0_0_0_0_0_0_0_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 5303, 396 },
      -1,
   } },
   /* search1907_0_0_0_0_0_0_0_0_0_0_1 -> 5398 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      7, 2,
      { 5403, 5398 },
      -1,
   } },
   /* search1907_0_0_0_0_0_0_0_0_0_1_0_0_0_0 -> 5258 in the cache */
   /* search1907_0_0_0_0_0_0_0_0_0_1_0_0_0_1 -> 853 in the cache */
   /* search1907_0_0_0_0_0_0_0_0_0_1_0_0_0 -> 5301 in the cache */
   /* search1907_0_0_0_0_0_0_0_0_0_1_0_0_1_0 -> 5258 in the cache */
   /* search1907_0_0_0_0_0_0_0_0_0_1_0_0_1_1 -> 853 in the cache */
   /* search1907_0_0_0_0_0_0_0_0_0_1_0_0_1 -> 5302 in the cache */
   /* search1907_0_0_0_0_0_0_0_0_0_1_0_0 -> 5307 in the cache */
   /* search1907_0_0_0_0_0_0_0_0_0_1_0_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 5307, 396 },
      -1,
   } },
   /* search1907_0_0_0_0_0_0_0_0_0_1_1 -> 5352 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      9, 2,
      { 5405, 5352 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      6, 5,
      { 5404, 5406 },
      -1,
   } },
   /* search1907_0_0_0_0_0_0_0_0_1 -> 401 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 5,
      { 5407, 401 },
      -1,
   } },
   /* search1907_0_0_0_0_0_0_0_1 -> 5348 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      5, 6,
      { 5408, 5348 },
      -1,
   } },
   /* search1907_0_0_0_0_0_0_1_0_0_0_0_0_0_0 -> 5258 in the cache */
   /* search1907_0_0_0_0_0_0_1_0_0_0_0_0_0_1 -> 853 in the cache */
   /* search1907_0_0_0_0_0_0_1_0_0_0_0_0_0 -> 5301 in the cache */
   /* search1907_0_0_0_0_0_0_1_0_0_0_0_0_1_0 -> 5258 in the cache */
   /* search1907_0_0_0_0_0_0_1_0_0_0_0_0_1_1 -> 853 in the cache */
   /* search1907_0_0_0_0_0_0_1_0_0_0_0_0_1 -> 5302 in the cache */
   /* search1907_0_0_0_0_0_0_1_0_0_0_0_0 -> 5315 in the cache */
   /* search1907_0_0_0_0_0_0_1_0_0_0_0_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 5315, 396 },
      -1,
   } },
   /* search1907_0_0_0_0_0_0_1_0_0_0_1 -> 5398 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      13, 2,
      { 5410, 5398 },
      -1,
   } },
   /* search1907_0_0_0_0_0_0_1_0_0_1_0_0_0_0 -> 5258 in the cache */
   /* search1907_0_0_0_0_0_0_1_0_0_1_0_0_0_1 -> 853 in the cache */
   /* search1907_0_0_0_0_0_0_1_0_0_1_0_0_0 -> 5301 in the cache */
   /* search1907_0_0_0_0_0_0_1_0_0_1_0_0_1_0 -> 5258 in the cache */
   /* search1907_0_0_0_0_0_0_1_0_0_1_0_0_1_1 -> 853 in the cache */
   /* search1907_0_0_0_0_0_0_1_0_0_1_0_0_1 -> 5302 in the cache */
   /* search1907_0_0_0_0_0_0_1_0_0_1_0_0 -> 5318 in the cache */
   /* search1907_0_0_0_0_0_0_1_0_0_1_0_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 5318, 396 },
      -1,
   } },
   /* search1907_0_0_0_0_0_0_1_0_0_1_1 -> 5352 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      15, 2,
      { 5412, 5352 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      12, 5,
      { 5411, 5413 },
      -1,
   } },
   /* search1907_0_0_0_0_0_0_1_0_1 -> 401 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 5,
      { 5414, 401 },
      -1,
   } },
   /* search1907_0_0_0_0_0_0_1_1 -> 5326 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      11, 6,
      { 5415, 5326 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      4, 13,
      { 5409, 5416 },
      -1,
   } },
   /* search1907_0_0_0_0_0_1 -> 2198 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 13,
      { 5417, 2198 },
      -1,
   } },
   /* search1907_0_0_0_0_1 -> 5322 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      3, 14,
      { 5418, 5322 },
      -1,
   } },
   /* search1907_0_0_0_1_0_0_0_0_0_0_0_0_0_0 -> 5258 in the cache */
   /* search1907_0_0_0_1_0_0_0_0_0_0_0_0_0_1 -> 853 in the cache */
   /* search1907_0_0_0_1_0_0_0_0_0_0_0_0_0 -> 5301 in the cache */
   /* search1907_0_0_0_1_0_0_0_0_0_0_0_0_1_0 -> 5258 in the cache */
   /* search1907_0_0_0_1_0_0_0_0_0_0_0_0_1_1 -> 853 in the cache */
   /* search1907_0_0_0_1_0_0_0_0_0_0_0_0_1 -> 5302 in the cache */
   /* search1907_0_0_0_1_0_0_0_0_0_0_0_0 -> 5329 in the cache */
   /* search1907_0_0_0_1_0_0_0_0_0_0_0_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 5329, 396 },
      -1,
   } },
   /* search1907_0_0_0_1_0_0_0_0_0_0_1 -> 5398 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      21, 2,
      { 5420, 5398 },
      -1,
   } },
   /* search1907_0_0_0_1_0_0_0_0_0_1_0_0_0_0 -> 5258 in the cache */
   /* search1907_0_0_0_1_0_0_0_0_0_1_0_0_0_1 -> 853 in the cache */
   /* search1907_0_0_0_1_0_0_0_0_0_1_0_0_0 -> 5301 in the cache */
   /* search1907_0_0_0_1_0_0_0_0_0_1_0_0_1_0 -> 5258 in the cache */
   /* search1907_0_0_0_1_0_0_0_0_0_1_0_0_1_1 -> 853 in the cache */
   /* search1907_0_0_0_1_0_0_0_0_0_1_0_0_1 -> 5302 in the cache */
   /* search1907_0_0_0_1_0_0_0_0_0_1_0_0 -> 5332 in the cache */
   /* search1907_0_0_0_1_0_0_0_0_0_1_0_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 5332, 396 },
      -1,
   } },
   /* search1907_0_0_0_1_0_0_0_0_0_1_1 -> 5352 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      23, 2,
      { 5422, 5352 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      20, 5,
      { 5421, 5423 },
      -1,
   } },
   /* search1907_0_0_0_1_0_0_0_0_1 -> 401 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 5,
      { 5424, 401 },
      -1,
   } },
   /* search1907_0_0_0_1_0_0_0_1 -> 5348 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      19, 6,
      { 5425, 5348 },
      -1,
   } },
   /* search1907_0_0_0_1_0_0_1_0_0_0_0_0_0_0 -> 5258 in the cache */
   /* search1907_0_0_0_1_0_0_1_0_0_0_0_0_0_1 -> 853 in the cache */
   /* search1907_0_0_0_1_0_0_1_0_0_0_0_0_0 -> 5301 in the cache */
   /* search1907_0_0_0_1_0_0_1_0_0_0_0_0_1_0 -> 5258 in the cache */
   /* search1907_0_0_0_1_0_0_1_0_0_0_0_0_1_1 -> 853 in the cache */
   /* search1907_0_0_0_1_0_0_1_0_0_0_0_0_1 -> 5302 in the cache */
   /* search1907_0_0_0_1_0_0_1_0_0_0_0_0 -> 5338 in the cache */
   /* search1907_0_0_0_1_0_0_1_0_0_0_0_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 5338, 396 },
      -1,
   } },
   /* search1907_0_0_0_1_0_0_1_0_0_0_1 -> 5398 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      27, 2,
      { 5427, 5398 },
      -1,
   } },
   /* search1907_0_0_0_1_0_0_1_0_0_1_0_0_0_0 -> 5258 in the cache */
   /* search1907_0_0_0_1_0_0_1_0_0_1_0_0_0_1 -> 853 in the cache */
   /* search1907_0_0_0_1_0_0_1_0_0_1_0_0_0 -> 5301 in the cache */
   /* search1907_0_0_0_1_0_0_1_0_0_1_0_0_1_0 -> 5258 in the cache */
   /* search1907_0_0_0_1_0_0_1_0_0_1_0_0_1_1 -> 853 in the cache */
   /* search1907_0_0_0_1_0_0_1_0_0_1_0_0_1 -> 5302 in the cache */
   /* search1907_0_0_0_1_0_0_1_0_0_1_0_0 -> 5341 in the cache */
   /* search1907_0_0_0_1_0_0_1_0_0_1_0_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 5341, 396 },
      -1,
   } },
   /* search1907_0_0_0_1_0_0_1_0_0_1_1 -> 5352 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      29, 2,
      { 5429, 5352 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      26, 5,
      { 5428, 5430 },
      -1,
   } },
   /* search1907_0_0_0_1_0_0_1_0_1 -> 401 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 5,
      { 5431, 401 },
      -1,
   } },
   /* search1907_0_0_0_1_0_0_1_1 -> 5326 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      25, 6,
      { 5432, 5326 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      18, 13,
      { 5426, 5433 },
      -1,
   } },
   /* search1907_0_0_0_1_0_1 -> 2198 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 13,
      { 5434, 2198 },
      -1,
   } },
   /* search1907_0_0_0_1_1 -> 5312 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      17, 14,
      { 5435, 5312 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      2, 29,
      { 5419, 5436 },
      -1,
   } },
   /* search1907_0_0_1 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 29,
      { 5437, 831 },
      -1,
   } },
   /* search1907_0_1 -> 5308 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 30,
      { 5438, 5308 },
      -1,
   } },
   /* search1907_1_0_0_0_0_0_0_0_0_0_0_0_0_0 -> 5258 in the cache */
   /* search1907_1_0_0_0_0_0_0_0_0_0_0_0_0_1 -> 853 in the cache */
   /* search1907_1_0_0_0_0_0_0_0_0_0_0_0_0 -> 5301 in the cache */
   /* search1907_1_0_0_0_0_0_0_0_0_0_0_0_1_0 -> 5258 in the cache */
   /* search1907_1_0_0_0_0_0_0_0_0_0_0_0_1_1 -> 853 in the cache */
   /* search1907_1_0_0_0_0_0_0_0_0_0_0_0_1 -> 5302 in the cache */
   /* search1907_1_0_0_0_0_0_0_0_0_0_0_0 -> 5355 in the cache */
   /* search1907_1_0_0_0_0_0_0_0_0_0_0_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 5355, 396 },
      -1,
   } },
   /* search1907_1_0_0_0_0_0_0_0_0_0_1 -> 5398 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      37, 2,
      { 5440, 5398 },
      -1,
   } },
   /* search1907_1_0_0_0_0_0_0_0_0_1_0_0_0_0 -> 5258 in the cache */
   /* search1907_1_0_0_0_0_0_0_0_0_1_0_0_0_1 -> 853 in the cache */
   /* search1907_1_0_0_0_0_0_0_0_0_1_0_0_0 -> 5301 in the cache */
   /* search1907_1_0_0_0_0_0_0_0_0_1_0_0_1_0 -> 5258 in the cache */
   /* search1907_1_0_0_0_0_0_0_0_0_1_0_0_1_1 -> 853 in the cache */
   /* search1907_1_0_0_0_0_0_0_0_0_1_0_0_1 -> 5302 in the cache */
   /* search1907_1_0_0_0_0_0_0_0_0_1_0_0 -> 5358 in the cache */
   /* search1907_1_0_0_0_0_0_0_0_0_1_0_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 5358, 396 },
      -1,
   } },
   /* search1907_1_0_0_0_0_0_0_0_0_1_1 -> 5352 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      39, 2,
      { 5442, 5352 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      36, 5,
      { 5441, 5443 },
      -1,
   } },
   /* search1907_1_0_0_0_0_0_0_0_1 -> 401 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 5,
      { 5444, 401 },
      -1,
   } },
   /* search1907_1_0_0_0_0_0_0_1 -> 5348 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      35, 6,
      { 5445, 5348 },
      -1,
   } },
   /* search1907_1_0_0_0_0_0_1_0_0_0_0_0_0_0 -> 5258 in the cache */
   /* search1907_1_0_0_0_0_0_1_0_0_0_0_0_0_1 -> 853 in the cache */
   /* search1907_1_0_0_0_0_0_1_0_0_0_0_0_0 -> 5301 in the cache */
   /* search1907_1_0_0_0_0_0_1_0_0_0_0_0_1_0 -> 5258 in the cache */
   /* search1907_1_0_0_0_0_0_1_0_0_0_0_0_1_1 -> 853 in the cache */
   /* search1907_1_0_0_0_0_0_1_0_0_0_0_0_1 -> 5302 in the cache */
   /* search1907_1_0_0_0_0_0_1_0_0_0_0_0 -> 5364 in the cache */
   /* search1907_1_0_0_0_0_0_1_0_0_0_0_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 5364, 396 },
      -1,
   } },
   /* search1907_1_0_0_0_0_0_1_0_0_0_1 -> 5398 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      43, 2,
      { 5447, 5398 },
      -1,
   } },
   /* search1907_1_0_0_0_0_0_1_0_0_1_0_0_0_0 -> 5258 in the cache */
   /* search1907_1_0_0_0_0_0_1_0_0_1_0_0_0_1 -> 853 in the cache */
   /* search1907_1_0_0_0_0_0_1_0_0_1_0_0_0 -> 5301 in the cache */
   /* search1907_1_0_0_0_0_0_1_0_0_1_0_0_1_0 -> 5258 in the cache */
   /* search1907_1_0_0_0_0_0_1_0_0_1_0_0_1_1 -> 853 in the cache */
   /* search1907_1_0_0_0_0_0_1_0_0_1_0_0_1 -> 5302 in the cache */
   /* search1907_1_0_0_0_0_0_1_0_0_1_0_0 -> 5367 in the cache */
   /* search1907_1_0_0_0_0_0_1_0_0_1_0_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 5367, 396 },
      -1,
   } },
   /* search1907_1_0_0_0_0_0_1_0_0_1_1 -> 5352 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      45, 2,
      { 5449, 5352 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      42, 5,
      { 5448, 5450 },
      -1,
   } },
   /* search1907_1_0_0_0_0_0_1_0_1 -> 401 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 5,
      { 5451, 401 },
      -1,
   } },
   /* search1907_1_0_0_0_0_0_1_1 -> 5326 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      41, 6,
      { 5452, 5326 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      34, 13,
      { 5446, 5453 },
      -1,
   } },
   /* search1907_1_0_0_0_0_1 -> 2198 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 13,
      { 5454, 2198 },
      -1,
   } },
   /* search1907_1_0_0_0_1 -> 5322 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      33, 14,
      { 5455, 5322 },
      -1,
   } },
   /* search1907_1_0_0_1_0_0_0_0_0_0_0_0_0_0 -> 5258 in the cache */
   /* search1907_1_0_0_1_0_0_0_0_0_0_0_0_0_1 -> 853 in the cache */
   /* search1907_1_0_0_1_0_0_0_0_0_0_0_0_0 -> 5301 in the cache */
   /* search1907_1_0_0_1_0_0_0_0_0_0_0_0_1_0 -> 5258 in the cache */
   /* search1907_1_0_0_1_0_0_0_0_0_0_0_0_1_1 -> 853 in the cache */
   /* search1907_1_0_0_1_0_0_0_0_0_0_0_0_1 -> 5302 in the cache */
   /* search1907_1_0_0_1_0_0_0_0_0_0_0_0 -> 5376 in the cache */
   /* search1907_1_0_0_1_0_0_0_0_0_0_0_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 5376, 396 },
      -1,
   } },
   /* search1907_1_0_0_1_0_0_0_0_0_0_1 -> 5398 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      51, 2,
      { 5457, 5398 },
      -1,
   } },
   /* search1907_1_0_0_1_0_0_0_0_0_1_0_0_0_0 -> 5258 in the cache */
   /* search1907_1_0_0_1_0_0_0_0_0_1_0_0_0_1 -> 853 in the cache */
   /* search1907_1_0_0_1_0_0_0_0_0_1_0_0_0 -> 5301 in the cache */
   /* search1907_1_0_0_1_0_0_0_0_0_1_0_0_1_0 -> 5258 in the cache */
   /* search1907_1_0_0_1_0_0_0_0_0_1_0_0_1_1 -> 853 in the cache */
   /* search1907_1_0_0_1_0_0_0_0_0_1_0_0_1 -> 5302 in the cache */
   /* search1907_1_0_0_1_0_0_0_0_0_1_0_0 -> 5379 in the cache */
   /* search1907_1_0_0_1_0_0_0_0_0_1_0_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 5379, 396 },
      -1,
   } },
   /* search1907_1_0_0_1_0_0_0_0_0_1_1 -> 5352 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      53, 2,
      { 5459, 5352 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      50, 5,
      { 5458, 5460 },
      -1,
   } },
   /* search1907_1_0_0_1_0_0_0_0_1 -> 401 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 5,
      { 5461, 401 },
      -1,
   } },
   /* search1907_1_0_0_1_0_0_0_1 -> 5348 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      49, 6,
      { 5462, 5348 },
      -1,
   } },
   /* search1907_1_0_0_1_0_0_1_0_0_0_0_0_0_0 -> 5258 in the cache */
   /* search1907_1_0_0_1_0_0_1_0_0_0_0_0_0_1 -> 853 in the cache */
   /* search1907_1_0_0_1_0_0_1_0_0_0_0_0_0 -> 5301 in the cache */
   /* search1907_1_0_0_1_0_0_1_0_0_0_0_0_1_0 -> 5258 in the cache */
   /* search1907_1_0_0_1_0_0_1_0_0_0_0_0_1_1 -> 853 in the cache */
   /* search1907_1_0_0_1_0_0_1_0_0_0_0_0_1 -> 5302 in the cache */
   /* search1907_1_0_0_1_0_0_1_0_0_0_0_0 -> 5385 in the cache */
   /* search1907_1_0_0_1_0_0_1_0_0_0_0_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 5385, 396 },
      -1,
   } },
   /* search1907_1_0_0_1_0_0_1_0_0_0_1 -> 5398 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      57, 2,
      { 5464, 5398 },
      -1,
   } },
   /* search1907_1_0_0_1_0_0_1_0_0_1_0_0_0_0 -> 5258 in the cache */
   /* search1907_1_0_0_1_0_0_1_0_0_1_0_0_0_1 -> 853 in the cache */
   /* search1907_1_0_0_1_0_0_1_0_0_1_0_0_0 -> 5301 in the cache */
   /* search1907_1_0_0_1_0_0_1_0_0_1_0_0_1_0 -> 5258 in the cache */
   /* search1907_1_0_0_1_0_0_1_0_0_1_0_0_1_1 -> 853 in the cache */
   /* search1907_1_0_0_1_0_0_1_0_0_1_0_0_1 -> 5302 in the cache */
   /* search1907_1_0_0_1_0_0_1_0_0_1_0_0 -> 5388 in the cache */
   /* search1907_1_0_0_1_0_0_1_0_0_1_0_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 5388, 396 },
      -1,
   } },
   /* search1907_1_0_0_1_0_0_1_0_0_1_1 -> 5352 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      59, 2,
      { 5466, 5352 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      56, 5,
      { 5465, 5467 },
      -1,
   } },
   /* search1907_1_0_0_1_0_0_1_0_1 -> 401 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 5,
      { 5468, 401 },
      -1,
   } },
   /* search1907_1_0_0_1_0_0_1_1 -> 5326 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      55, 6,
      { 5469, 5326 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      48, 13,
      { 5463, 5470 },
      -1,
   } },
   /* search1907_1_0_0_1_0_1 -> 2198 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 13,
      { 5471, 2198 },
      -1,
   } },
   /* search1907_1_0_0_1_1 -> 5312 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      47, 14,
      { 5472, 5312 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      32, 29,
      { 5456, 5473 },
      -1,
   } },
   /* search1907_1_0_1 -> 831 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 29,
      { 5474, 831 },
      -1,
   } },
   /* search1907_1_1 -> 5304 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      31, 30,
      { 5475, 5304 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 61,
      { 5439, 5476 },
      -1,
   } },

   /* replace1907_0 -> 5258 in the cache */
   /* replace1907 -> 5402 in the cache */

   /* ('ball_iequal2', ('ieq', 'a', 'b'), -1) => ('ball_iequal2', 'a', 'b') */
   /* search1908_0_0 -> 0 in the cache */
   /* search1908_0_1 -> 3 in the cache */
   /* search1908_0 -> 1080 in the cache */
   { .constant = {
      { nir_search_value_constant, 1 },
      nir_type_int, { -0x1 /* -1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal2,
      0, 2,
      { 1080, 5478 },
      -1,
   } },

   /* replace1908_0 -> 0 in the cache */
   /* replace1908_1 -> 3 in the cache */
   /* replace1908 -> 2464 in the cache */

   /* ('ball_iequal2', ('feq', 'a', 'b'), -1) => ('ball_fequal2', 'a', 'b') */
   /* search1909_0_0 -> 0 in the cache */
   /* search1909_0_1 -> 3 in the cache */
   /* search1909_0 -> 1074 in the cache */
   /* search1909_1 -> 5478 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal2,
      0, 2,
      { 1074, 5478 },
      -1,
   } },

   /* replace1909_0 -> 0 in the cache */
   /* replace1909_1 -> 3 in the cache */
   /* replace1909 -> 2534 in the cache */

   /* ('bany_inequal2', ('ine', 'a', 'b'), 0) => ('bany_inequal2', 'a', 'b') */
   /* search1910_0_0 -> 0 in the cache */
   /* search1910_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 0, 3 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 1 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal2,
      0, 2,
      { 5481, 5482 },
      -1,
   } },

   /* replace1910_0 -> 0 in the cache */
   /* replace1910_1 -> 3 in the cache */
   /* replace1910 -> 2575 in the cache */

   /* ('bany_inequal2', ('fneu', 'a', 'b'), 0) => ('bany_fnequal2', 'a', 'b') */
   /* search1911_0_0 -> 0 in the cache */
   /* search1911_0_1 -> 3 in the cache */
   /* search1911_0 -> 4199 in the cache */
   /* search1911_1 -> 5482 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal2,
      0, 2,
      { 4199, 5482 },
      -1,
   } },

   /* replace1911_0 -> 0 in the cache */
   /* replace1911_1 -> 3 in the cache */
   /* replace1911 -> 2616 in the cache */

   /* ('ball_iequal3', ('ieq', 'a', 'b'), -1) => ('ball_iequal3', 'a', 'b') */
   /* search1912_0_0 -> 0 in the cache */
   /* search1912_0_1 -> 3 in the cache */
   /* search1912_0 -> 1080 in the cache */
   /* search1912_1 -> 5478 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal3,
      0, 2,
      { 1080, 5478 },
      -1,
   } },

   /* replace1912_0 -> 0 in the cache */
   /* replace1912_1 -> 3 in the cache */
   /* replace1912 -> 2470 in the cache */

   /* ('ball_iequal3', ('feq', 'a', 'b'), -1) => ('ball_fequal3', 'a', 'b') */
   /* search1913_0_0 -> 0 in the cache */
   /* search1913_0_1 -> 3 in the cache */
   /* search1913_0 -> 1074 in the cache */
   /* search1913_1 -> 5478 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal3,
      0, 2,
      { 1074, 5478 },
      -1,
   } },

   /* replace1913_0 -> 0 in the cache */
   /* replace1913_1 -> 3 in the cache */
   /* replace1913 -> 2538 in the cache */

   /* ('bany_inequal3', ('ine', 'a', 'b'), 0) => ('bany_inequal3', 'a', 'b') */
   /* search1914_0_0 -> 0 in the cache */
   /* search1914_0_1 -> 3 in the cache */
   /* search1914_0 -> 5481 in the cache */
   /* search1914_1 -> 5482 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal3,
      0, 2,
      { 5481, 5482 },
      -1,
   } },

   /* replace1914_0 -> 0 in the cache */
   /* replace1914_1 -> 3 in the cache */
   /* replace1914 -> 2579 in the cache */

   /* ('bany_inequal3', ('fneu', 'a', 'b'), 0) => ('bany_fnequal3', 'a', 'b') */
   /* search1915_0_0 -> 0 in the cache */
   /* search1915_0_1 -> 3 in the cache */
   /* search1915_0 -> 4199 in the cache */
   /* search1915_1 -> 5482 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal3,
      0, 2,
      { 4199, 5482 },
      -1,
   } },

   /* replace1915_0 -> 0 in the cache */
   /* replace1915_1 -> 3 in the cache */
   /* replace1915 -> 2620 in the cache */

   /* ('ball_iequal4', ('ieq', 'a', 'b'), -1) => ('ball_iequal4', 'a', 'b') */
   /* search1916_0_0 -> 0 in the cache */
   /* search1916_0_1 -> 3 in the cache */
   /* search1916_0 -> 1080 in the cache */
   /* search1916_1 -> 5478 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal4,
      0, 2,
      { 1080, 5478 },
      -1,
   } },

   /* replace1916_0 -> 0 in the cache */
   /* replace1916_1 -> 3 in the cache */
   /* replace1916 -> 2477 in the cache */

   /* ('ball_iequal4', ('feq', 'a', 'b'), -1) => ('ball_fequal4', 'a', 'b') */
   /* search1917_0_0 -> 0 in the cache */
   /* search1917_0_1 -> 3 in the cache */
   /* search1917_0 -> 1074 in the cache */
   /* search1917_1 -> 5478 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal4,
      0, 2,
      { 1074, 5478 },
      -1,
   } },

   /* replace1917_0 -> 0 in the cache */
   /* replace1917_1 -> 3 in the cache */
   /* replace1917 -> 2544 in the cache */

   /* ('bany_inequal4', ('ine', 'a', 'b'), 0) => ('bany_inequal4', 'a', 'b') */
   /* search1918_0_0 -> 0 in the cache */
   /* search1918_0_1 -> 3 in the cache */
   /* search1918_0 -> 5481 in the cache */
   /* search1918_1 -> 5482 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal4,
      0, 2,
      { 5481, 5482 },
      -1,
   } },

   /* replace1918_0 -> 0 in the cache */
   /* replace1918_1 -> 3 in the cache */
   /* replace1918 -> 2585 in the cache */

   /* ('bany_inequal4', ('fneu', 'a', 'b'), 0) => ('bany_fnequal4', 'a', 'b') */
   /* search1919_0_0 -> 0 in the cache */
   /* search1919_0_1 -> 3 in the cache */
   /* search1919_0 -> 4199 in the cache */
   /* search1919_1 -> 5482 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal4,
      0, 2,
      { 4199, 5482 },
      -1,
   } },

   /* replace1919_0 -> 0 in the cache */
   /* replace1919_1 -> 3 in the cache */
   /* replace1919 -> 2626 in the cache */

   /* ('ball_iequal8', ('ieq', 'a', 'b'), -1) => ('ball_iequal8', 'a', 'b') */
   /* search1920_0_0 -> 0 in the cache */
   /* search1920_0_1 -> 3 in the cache */
   /* search1920_0 -> 1080 in the cache */
   /* search1920_1 -> 5478 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal8,
      0, 2,
      { 1080, 5478 },
      -1,
   } },

   /* replace1920_0 -> 0 in the cache */
   /* replace1920_1 -> 3 in the cache */
   /* replace1920 -> 2484 in the cache */

   /* ('ball_iequal8', ('feq', 'a', 'b'), -1) => ('ball_fequal8', 'a', 'b') */
   /* search1921_0_0 -> 0 in the cache */
   /* search1921_0_1 -> 3 in the cache */
   /* search1921_0 -> 1074 in the cache */
   /* search1921_1 -> 5478 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal8,
      0, 2,
      { 1074, 5478 },
      -1,
   } },

   /* replace1921_0 -> 0 in the cache */
   /* replace1921_1 -> 3 in the cache */
   /* replace1921 -> 2549 in the cache */

   /* ('bany_inequal8', ('ine', 'a', 'b'), 0) => ('bany_inequal8', 'a', 'b') */
   /* search1922_0_0 -> 0 in the cache */
   /* search1922_0_1 -> 3 in the cache */
   /* search1922_0 -> 5481 in the cache */
   /* search1922_1 -> 5482 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal8,
      0, 2,
      { 5481, 5482 },
      -1,
   } },

   /* replace1922_0 -> 0 in the cache */
   /* replace1922_1 -> 3 in the cache */
   /* replace1922 -> 2590 in the cache */

   /* ('bany_inequal8', ('fneu', 'a', 'b'), 0) => ('bany_fnequal8', 'a', 'b') */
   /* search1923_0_0 -> 0 in the cache */
   /* search1923_0_1 -> 3 in the cache */
   /* search1923_0 -> 4199 in the cache */
   /* search1923_1 -> 5482 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal8,
      0, 2,
      { 4199, 5482 },
      -1,
   } },

   /* replace1923_0 -> 0 in the cache */
   /* replace1923_1 -> 3 in the cache */
   /* replace1923 -> 2631 in the cache */

   /* ('ball_iequal16', ('ieq', 'a', 'b'), -1) => ('ball_iequal16', 'a', 'b') */
   /* search1924_0_0 -> 0 in the cache */
   /* search1924_0_1 -> 3 in the cache */
   /* search1924_0 -> 1080 in the cache */
   /* search1924_1 -> 5478 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal16,
      0, 2,
      { 1080, 5478 },
      -1,
   } },

   /* replace1924_0 -> 0 in the cache */
   /* replace1924_1 -> 3 in the cache */
   /* replace1924 -> 2501 in the cache */

   /* ('ball_iequal16', ('feq', 'a', 'b'), -1) => ('ball_fequal16', 'a', 'b') */
   /* search1925_0_0 -> 0 in the cache */
   /* search1925_0_1 -> 3 in the cache */
   /* search1925_0 -> 1074 in the cache */
   /* search1925_1 -> 5478 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal16,
      0, 2,
      { 1074, 5478 },
      -1,
   } },

   /* replace1925_0 -> 0 in the cache */
   /* replace1925_1 -> 3 in the cache */
   /* replace1925 -> 2558 in the cache */

   /* ('bany_inequal16', ('ine', 'a', 'b'), 0) => ('bany_inequal16', 'a', 'b') */
   /* search1926_0_0 -> 0 in the cache */
   /* search1926_0_1 -> 3 in the cache */
   /* search1926_0 -> 5481 in the cache */
   /* search1926_1 -> 5482 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal16,
      0, 2,
      { 5481, 5482 },
      -1,
   } },

   /* replace1926_0 -> 0 in the cache */
   /* replace1926_1 -> 3 in the cache */
   /* replace1926 -> 2599 in the cache */

   /* ('bany_inequal16', ('fneu', 'a', 'b'), 0) => ('bany_fnequal16', 'a', 'b') */
   /* search1927_0_0 -> 0 in the cache */
   /* search1927_0_1 -> 3 in the cache */
   /* search1927_0 -> 4199 in the cache */
   /* search1927_1 -> 5482 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal16,
      0, 2,
      { 4199, 5482 },
      -1,
   } },

   /* replace1927_0 -> 0 in the cache */
   /* replace1927_1 -> 3 in the cache */
   /* replace1927 -> 2640 in the cache */

   /* ('iand', ('feq', 'a', 'a'), ('flt', 'a', 'b')) => ('!flt', 'a', 'b') */
   /* search1928_0_0 -> 0 in the cache */
   /* search1928_0_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      -1, 0,
      { 0, 0 },
      -1,
   } },
   /* search1928_1_0 -> 0 in the cache */
   /* search1928_1_1 -> 3 in the cache */
   /* search1928_1 -> 1049 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 5501, 1049 },
      -1,
   } },

   /* replace1928_0 -> 0 in the cache */
   /* replace1928_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_flt,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* ('iand', ('feq', 'a', 'a'), ('flt', 'b', 'a')) => ('!flt', 'b', 'a') */
   /* search1929_0_0 -> 32 in the cache */
   /* search1929_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      -1, 0,
      { 32, 32 },
      -1,
   } },
   /* search1929_1_0 -> 2297 in the cache */
   /* search1929_1_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 2297, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 5504, 5505 },
      -1,
   } },

   /* replace1929_0 -> 2297 in the cache */
   /* replace1929_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_flt,
      -1, 0,
      { 2297, 32 },
      -1,
   } },

   /* ('iand', ('feq', 'a', 'a'), ('fge', 'a', 'b')) => ('!fge', 'a', 'b') */
   /* search1930_0_0 -> 0 in the cache */
   /* search1930_0_1 -> 0 in the cache */
   /* search1930_0 -> 5501 in the cache */
   /* search1930_1_0 -> 0 in the cache */
   /* search1930_1_1 -> 3 in the cache */
   /* search1930_1 -> 1046 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 5501, 1046 },
      -1,
   } },

   /* replace1930_0 -> 0 in the cache */
   /* replace1930_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_fge,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* ('iand', ('feq', 'a', 'a'), ('fge', 'b', 'a')) => ('!fge', 'b', 'a') */
   /* search1931_0_0 -> 32 in the cache */
   /* search1931_0_1 -> 32 in the cache */
   /* search1931_0 -> 5504 in the cache */
   /* search1931_1_0 -> 2297 in the cache */
   /* search1931_1_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 2297, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 5504, 5510 },
      -1,
   } },

   /* replace1931_0 -> 2297 in the cache */
   /* replace1931_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_fge,
      -1, 0,
      { 2297, 32 },
      -1,
   } },

   /* ('iand', ('feq', 'a', 'a'), ('feq', 'a', 'b')) => ('!feq', 'a', 'b') */
   /* search1932_0_0 -> 0 in the cache */
   /* search1932_0_1 -> 0 in the cache */
   /* search1932_0 -> 5501 in the cache */
   /* search1932_1_0 -> 0 in the cache */
   /* search1932_1_1 -> 3 in the cache */
   /* search1932_1 -> 1074 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 5501, 1074 },
      -1,
   } },

   /* replace1932_0 -> 0 in the cache */
   /* replace1932_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_feq,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('iand', ('feq', 'a', 'a'), ('feq', 'b', 'a')) => ('!feq', 'b', 'a') */
   /* search1933_0_0 -> 32 in the cache */
   /* search1933_0_1 -> 32 in the cache */
   /* search1933_0 -> 5504 in the cache */
   /* search1933_1_0 -> 2297 in the cache */
   /* search1933_1_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      1, 1,
      { 2297, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 5504, 5515 },
      -1,
   } },

   /* replace1933_0 -> 2297 in the cache */
   /* replace1933_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_feq,
      0, 1,
      { 2297, 32 },
      -1,
   } },

   /* ('feq', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('feq', 'b', 'd'), ('feq', 'c', 'd')) */
   /* search1934_0_0 -> 760 in the cache */
   /* search1934_0_1 -> 1028 in the cache */
   /* search1934_0_2 -> 3498 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 1028, 3498 },
      -1,
   } },
   /* search1934_1 -> 1031 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 5518, 1031 },
      -1,
   } },

   /* replace1934_0 -> 760 in the cache */
   /* replace1934_1_0 -> 150 in the cache */
   /* replace1934_1_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 150, 153 },
      -1,
   } },
   /* replace1934_2_0 -> 152 in the cache */
   /* replace1934_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      1, 1,
      { 152, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5520, 5521 },
      -1,
   } },

   /* ('fneu', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('fneu', 'b', 'd'), ('fneu', 'c', 'd')) */
   /* search1935_0_0 -> 760 in the cache */
   /* search1935_0_1 -> 1028 in the cache */
   /* search1935_0_2 -> 3498 in the cache */
   /* search1935_0 -> 5518 in the cache */
   /* search1935_1 -> 1031 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 5518, 1031 },
      -1,
   } },

   /* replace1935_0 -> 760 in the cache */
   /* replace1935_1_0 -> 150 in the cache */
   /* replace1935_1_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 150, 153 },
      -1,
   } },
   /* replace1935_2_0 -> 152 in the cache */
   /* replace1935_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      1, 1,
      { 152, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5524, 5525 },
      -1,
   } },

   /* ('ieq', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('ieq', 'b', 'd'), ('ieq', 'c', 'd')) */
   /* search1936_0_0 -> 760 in the cache */
   /* search1936_0_1 -> 1028 in the cache */
   /* search1936_0_2 -> 3498 in the cache */
   /* search1936_0 -> 5518 in the cache */
   /* search1936_1 -> 1031 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 5518, 1031 },
      -1,
   } },

   /* replace1936_0 -> 760 in the cache */
   /* replace1936_1_0 -> 150 in the cache */
   /* replace1936_1_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 150, 153 },
      -1,
   } },
   /* replace1936_2_0 -> 152 in the cache */
   /* replace1936_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 152, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5528, 5529 },
      -1,
   } },

   /* ('ine', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('ine', 'b', 'd'), ('ine', 'c', 'd')) */
   /* search1937_0_0 -> 760 in the cache */
   /* search1937_0_1 -> 1028 in the cache */
   /* search1937_0_2 -> 3498 in the cache */
   /* search1937_0 -> 5518 in the cache */
   /* search1937_1 -> 1031 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 5518, 1031 },
      -1,
   } },

   /* replace1937_0 -> 760 in the cache */
   /* replace1937_1_0 -> 150 in the cache */
   /* replace1937_1_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 150, 153 },
      -1,
   } },
   /* replace1937_2_0 -> 152 in the cache */
   /* replace1937_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 152, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5532, 5533 },
      -1,
   } },

   /* ('flt', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('flt', 'b', 'd'), ('flt', 'c', 'd')) */
   /* search1938_0_0 -> 760 in the cache */
   /* search1938_0_1 -> 1028 in the cache */
   /* search1938_0_2 -> 3498 in the cache */
   /* search1938_0 -> 5518 in the cache */
   /* search1938_1 -> 1031 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 5518, 1031 },
      -1,
   } },

   /* replace1938_0 -> 760 in the cache */
   /* replace1938_1_0 -> 150 in the cache */
   /* replace1938_1_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 150, 153 },
      -1,
   } },
   /* replace1938_2_0 -> 152 in the cache */
   /* replace1938_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 152, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 5536, 5537 },
      -1,
   } },

   /* ('flt', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('flt', 'd', 'b'), ('flt', 'd', 'c')) */
   { .variable = {
      { nir_search_value_variable, -4 },
      0, /* d */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1939_1_0 -> 1176 in the cache */
   { .variable = {
      { nir_search_value_variable, -4 },
      2, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      3, /* c */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1176, 5540, 5541 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 5539, 5542 },
      -1,
   } },

   /* replace1939_0 -> 1176 in the cache */
   { .variable = {
      { nir_search_value_variable, -4 },
      0, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      2, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 5544, 5545 },
      -1,
   } },
   /* replace1939_2_0 -> 5544 in the cache */
   /* replace1939_2_1 -> 1726 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 5544, 1726 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1176, 5546, 5547 },
      -1,
   } },

   /* ('fge', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('fge', 'b', 'd'), ('fge', 'c', 'd')) */
   /* search1940_0_0 -> 760 in the cache */
   /* search1940_0_1 -> 1028 in the cache */
   /* search1940_0_2 -> 3498 in the cache */
   /* search1940_0 -> 5518 in the cache */
   /* search1940_1 -> 1031 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 5518, 1031 },
      -1,
   } },

   /* replace1940_0 -> 760 in the cache */
   /* replace1940_1_0 -> 150 in the cache */
   /* replace1940_1_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 150, 153 },
      -1,
   } },
   /* replace1940_2_0 -> 152 in the cache */
   /* replace1940_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 152, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 5550, 5551 },
      -1,
   } },

   /* ('fge', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('fge', 'd', 'b'), ('fge', 'd', 'c')) */
   /* search1941_0 -> 5539 in the cache */
   /* search1941_1_0 -> 1176 in the cache */
   /* search1941_1_1 -> 5540 in the cache */
   /* search1941_1_2 -> 5541 in the cache */
   /* search1941_1 -> 5542 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 5539, 5542 },
      -1,
   } },

   /* replace1941_0 -> 1176 in the cache */
   /* replace1941_1_0 -> 5544 in the cache */
   /* replace1941_1_1 -> 5545 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 5544, 5545 },
      -1,
   } },
   /* replace1941_2_0 -> 5544 in the cache */
   /* replace1941_2_1 -> 1726 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 5544, 1726 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1176, 5554, 5555 },
      -1,
   } },

   /* ('ilt', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('ilt', 'b', 'd'), ('ilt', 'c', 'd')) */
   /* search1942_0_0 -> 760 in the cache */
   /* search1942_0_1 -> 1028 in the cache */
   /* search1942_0_2 -> 3498 in the cache */
   /* search1942_0 -> 5518 in the cache */
   /* search1942_1 -> 1031 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 5518, 1031 },
      -1,
   } },

   /* replace1942_0 -> 760 in the cache */
   /* replace1942_1_0 -> 150 in the cache */
   /* replace1942_1_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 150, 153 },
      -1,
   } },
   /* replace1942_2_0 -> 152 in the cache */
   /* replace1942_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 152, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 5558, 5559 },
      -1,
   } },

   /* ('ilt', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('ilt', 'd', 'b'), ('ilt', 'd', 'c')) */
   /* search1943_0 -> 5539 in the cache */
   /* search1943_1_0 -> 1176 in the cache */
   /* search1943_1_1 -> 5540 in the cache */
   /* search1943_1_2 -> 5541 in the cache */
   /* search1943_1 -> 5542 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 5539, 5542 },
      -1,
   } },

   /* replace1943_0 -> 1176 in the cache */
   /* replace1943_1_0 -> 5544 in the cache */
   /* replace1943_1_1 -> 5545 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 5544, 5545 },
      -1,
   } },
   /* replace1943_2_0 -> 5544 in the cache */
   /* replace1943_2_1 -> 1726 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 5544, 1726 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1176, 5562, 5563 },
      -1,
   } },

   /* ('ige', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('ige', 'b', 'd'), ('ige', 'c', 'd')) */
   /* search1944_0_0 -> 760 in the cache */
   /* search1944_0_1 -> 1028 in the cache */
   /* search1944_0_2 -> 3498 in the cache */
   /* search1944_0 -> 5518 in the cache */
   /* search1944_1 -> 1031 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 5518, 1031 },
      -1,
   } },

   /* replace1944_0 -> 760 in the cache */
   /* replace1944_1_0 -> 150 in the cache */
   /* replace1944_1_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 150, 153 },
      -1,
   } },
   /* replace1944_2_0 -> 152 in the cache */
   /* replace1944_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 152, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 5566, 5567 },
      -1,
   } },

   /* ('ige', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('ige', 'd', 'b'), ('ige', 'd', 'c')) */
   /* search1945_0 -> 5539 in the cache */
   /* search1945_1_0 -> 1176 in the cache */
   /* search1945_1_1 -> 5540 in the cache */
   /* search1945_1_2 -> 5541 in the cache */
   /* search1945_1 -> 5542 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 5539, 5542 },
      -1,
   } },

   /* replace1945_0 -> 1176 in the cache */
   /* replace1945_1_0 -> 5544 in the cache */
   /* replace1945_1_1 -> 5545 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 5544, 5545 },
      -1,
   } },
   /* replace1945_2_0 -> 5544 in the cache */
   /* replace1945_2_1 -> 1726 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 5544, 1726 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1176, 5570, 5571 },
      -1,
   } },

   /* ('ult', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('ult', 'b', 'd'), ('ult', 'c', 'd')) */
   /* search1946_0_0 -> 760 in the cache */
   /* search1946_0_1 -> 1028 in the cache */
   /* search1946_0_2 -> 3498 in the cache */
   /* search1946_0 -> 5518 in the cache */
   /* search1946_1 -> 1031 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 5518, 1031 },
      -1,
   } },

   /* replace1946_0 -> 760 in the cache */
   /* replace1946_1_0 -> 150 in the cache */
   /* replace1946_1_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 150, 153 },
      -1,
   } },
   /* replace1946_2_0 -> 152 in the cache */
   /* replace1946_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 152, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 5574, 5575 },
      -1,
   } },

   /* ('ult', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('ult', 'd', 'b'), ('ult', 'd', 'c')) */
   /* search1947_0 -> 5539 in the cache */
   /* search1947_1_0 -> 1176 in the cache */
   /* search1947_1_1 -> 5540 in the cache */
   /* search1947_1_2 -> 5541 in the cache */
   /* search1947_1 -> 5542 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 5539, 5542 },
      -1,
   } },

   /* replace1947_0 -> 1176 in the cache */
   /* replace1947_1_0 -> 5544 in the cache */
   /* replace1947_1_1 -> 5545 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 5544, 5545 },
      -1,
   } },
   /* replace1947_2_0 -> 5544 in the cache */
   /* replace1947_2_1 -> 1726 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 5544, 1726 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1176, 5578, 5579 },
      -1,
   } },

   /* ('uge', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('uge', 'b', 'd'), ('uge', 'c', 'd')) */
   /* search1948_0_0 -> 760 in the cache */
   /* search1948_0_1 -> 1028 in the cache */
   /* search1948_0_2 -> 3498 in the cache */
   /* search1948_0 -> 5518 in the cache */
   /* search1948_1 -> 1031 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 5518, 1031 },
      -1,
   } },

   /* replace1948_0 -> 760 in the cache */
   /* replace1948_1_0 -> 150 in the cache */
   /* replace1948_1_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 150, 153 },
      -1,
   } },
   /* replace1948_2_0 -> 152 in the cache */
   /* replace1948_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 152, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 5582, 5583 },
      -1,
   } },

   /* ('uge', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('uge', 'd', 'b'), ('uge', 'd', 'c')) */
   /* search1949_0 -> 5539 in the cache */
   /* search1949_1_0 -> 1176 in the cache */
   /* search1949_1_1 -> 5540 in the cache */
   /* search1949_1_2 -> 5541 in the cache */
   /* search1949_1 -> 5542 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 5539, 5542 },
      -1,
   } },

   /* replace1949_0 -> 1176 in the cache */
   /* replace1949_1_0 -> 5544 in the cache */
   /* replace1949_1_1 -> 5545 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 5544, 5545 },
      -1,
   } },
   /* replace1949_2_0 -> 5544 in the cache */
   /* replace1949_2_1 -> 1726 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 5544, 1726 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1176, 5586, 5587 },
      -1,
   } },

   /* ('fadd', ('bcsel(is_used_once)', 'a', '#b', 'c'), '#d') => ('bcsel', 'a', ('fadd', 'b', 'd'), ('fadd', 'c', 'd')) */
   /* search1950_0_0 -> 760 in the cache */
   /* search1950_0_1 -> 1028 in the cache */
   /* search1950_0_2 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 1028, 152 },
      0,
   } },
   /* search1950_1 -> 1031 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 5589, 1031 },
      -1,
   } },

   /* replace1950_0 -> 760 in the cache */
   /* replace1950_1_0 -> 150 in the cache */
   /* replace1950_1_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 150, 153 },
      -1,
   } },
   /* replace1950_2_0 -> 152 in the cache */
   /* replace1950_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 152, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5591, 5592 },
      -1,
   } },

   /* ('fmul', ('bcsel(is_used_once)', 'a', '#b', 'c'), '#d') => ('bcsel', 'a', ('fmul', 'b', 'd'), ('fmul', 'c', 'd')) */
   /* search1951_0_0 -> 760 in the cache */
   /* search1951_0_1 -> 1028 in the cache */
   /* search1951_0_2 -> 152 in the cache */
   /* search1951_0 -> 5589 in the cache */
   /* search1951_1 -> 1031 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 5589, 1031 },
      -1,
   } },

   /* replace1951_0 -> 760 in the cache */
   /* replace1951_1_0 -> 150 in the cache */
   /* replace1951_1_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 150, 153 },
      -1,
   } },
   /* replace1951_2_0 -> 152 in the cache */
   /* replace1951_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 152, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5595, 5596 },
      -1,
   } },

   /* ('fmulz', ('bcsel(is_used_once)', 'a', '#b', 'c'), '#d') => ('bcsel', 'a', ('fmulz', 'b', 'd'), ('fmulz', 'c', 'd')) */
   /* search1952_0_0 -> 760 in the cache */
   /* search1952_0_1 -> 33 in the cache */
   /* search1952_0_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 33, 135 },
      0,
   } },
   /* search1952_1 -> 311 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 5598, 311 },
      -1,
   } },

   /* replace1952_0 -> 760 in the cache */
   /* replace1952_1_0 -> 36 in the cache */
   /* replace1952_1_1 -> 163 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 36, 163 },
      -1,
   } },
   /* replace1952_2_0 -> 135 in the cache */
   /* replace1952_2_1 -> 163 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 135, 163 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5600, 5601 },
      -1,
   } },

   /* ('iadd', ('bcsel(is_used_once)', 'a', '#b', 'c'), '#d') => ('bcsel', 'a', ('iadd', 'b', 'd'), ('iadd', 'c', 'd')) */
   /* search1953_0_0 -> 760 in the cache */
   /* search1953_0_1 -> 1028 in the cache */
   /* search1953_0_2 -> 152 in the cache */
   /* search1953_0 -> 5589 in the cache */
   /* search1953_1 -> 1031 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 5589, 1031 },
      -1,
   } },

   /* replace1953_0 -> 760 in the cache */
   /* replace1953_1_0 -> 150 in the cache */
   /* replace1953_1_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 150, 153 },
      -1,
   } },
   /* replace1953_2_0 -> 152 in the cache */
   /* replace1953_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 152, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5604, 5605 },
      -1,
   } },

   /* ('imul', ('bcsel(is_used_once)', 'a', '#b', 'c'), '#d') => ('bcsel', 'a', ('imul', 'b', 'd'), ('imul', 'c', 'd')) */
   /* search1954_0_0 -> 760 in the cache */
   /* search1954_0_1 -> 1028 in the cache */
   /* search1954_0_2 -> 152 in the cache */
   /* search1954_0 -> 5589 in the cache */
   /* search1954_1 -> 1031 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 5589, 1031 },
      -1,
   } },

   /* replace1954_0 -> 760 in the cache */
   /* replace1954_1_0 -> 150 in the cache */
   /* replace1954_1_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 150, 153 },
      -1,
   } },
   /* replace1954_2_0 -> 152 in the cache */
   /* replace1954_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 152, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5608, 5609 },
      -1,
   } },

   /* ('fddx', 'a') => 0.0 */
   /* search1955_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fddx,
      -1, 0,
      { 32 },
      -1,
   } },

   /* replace1955 -> 111 in the cache */

   /* ('fddx_fine', 'a') => 0.0 */
   /* search1956_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fddx_fine,
      -1, 0,
      { 32 },
      -1,
   } },

   /* replace1956 -> 111 in the cache */

   /* ('fddx_coarse', 'a') => 0.0 */
   /* search1957_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fddx_coarse,
      -1, 0,
      { 32 },
      -1,
   } },

   /* replace1957 -> 111 in the cache */

   /* ('fddy', 'a') => 0.0 */
   /* search1958_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fddy,
      -1, 0,
      { 32 },
      -1,
   } },

   /* replace1958 -> 111 in the cache */

   /* ('fddy_fine', 'a') => 0.0 */
   /* search1959_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fddy_fine,
      -1, 0,
      { 32 },
      -1,
   } },

   /* replace1959 -> 111 in the cache */

   /* ('fddy_coarse', 'a') => 0.0 */
   /* search1960_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fddy_coarse,
      -1, 0,
      { 32 },
      -1,
   } },

   /* replace1960 -> 111 in the cache */

   /* ('umul_low', '#a(is_lower_half_zero)', 'b') => 0 */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      true,
      nir_type_invalid,
      20,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1961_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umul_low,
      0, 1,
      { 5617, 36 },
      -1,
   } },

   /* replace1961 -> 122 in the cache */

   /* ('imadsh_mix16', '#a@32(is_lower_half_zero)', 'b@32', 'c@32') => c */
   /* search1962_0 -> 5617 in the cache */
   /* search1962_1 -> 36 in the cache */
   /* search1962_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imadsh_mix16,
      -1, 0,
      { 5617, 36, 135 },
      -1,
   } },

   /* replace1962 -> 135 in the cache */

   /* ('imadsh_mix16', 'a@32', '#b@32(is_upper_half_zero)', 'c@32') => c */
   /* search1963_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      21,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1963_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imadsh_mix16,
      -1, 0,
      { 16, 5620, 135 },
      -1,
   } },

   /* replace1963 -> 135 in the cache */

   /* ('bcsel', 'a', ('flrp(is_used_once)', 'b', 'c', 'd'), ('flrp', 'b', 'c', 'e')) => ('flrp', 'b', 'c', ('bcsel', 'a', 'd', 'e')) */
   /* search1964_0 -> 760 in the cache */
   /* search1964_1_0 -> 752 in the cache */
   /* search1964_1_1 -> 753 in the cache */
   /* search1964_1_2 -> 762 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 752, 753, 762 },
      0,
   } },
   /* search1964_2_0 -> 752 in the cache */
   /* search1964_2_1 -> 753 in the cache */
   /* search1964_2_2 -> 764 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 752, 753, 764 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 5622, 5623 },
      -1,
   } },

   /* replace1964_0 -> 752 in the cache */
   /* replace1964_1 -> 753 in the cache */
   /* replace1964_2_0 -> 760 in the cache */
   /* replace1964_2_1 -> 762 in the cache */
   /* replace1964_2_2 -> 764 in the cache */
   /* replace1964_2 -> 1317 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 752, 753, 1317 },
      -1,
   } },

   /* ('bcsel', 'a', ('flrp', 'b', 'c', 'd'), ('flrp(is_used_once)', 'b', 'c', 'e')) => ('flrp', 'b', 'c', ('bcsel', 'a', 'd', 'e')) */
   /* search1965_0 -> 760 in the cache */
   /* search1965_1_0 -> 752 in the cache */
   /* search1965_1_1 -> 753 in the cache */
   /* search1965_1_2 -> 762 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 752, 753, 762 },
      -1,
   } },
   /* search1965_2_0 -> 752 in the cache */
   /* search1965_2_1 -> 753 in the cache */
   /* search1965_2_2 -> 764 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 752, 753, 764 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 5626, 5627 },
      -1,
   } },

   /* replace1965_0 -> 752 in the cache */
   /* replace1965_1 -> 753 in the cache */
   /* replace1965_2_0 -> 760 in the cache */
   /* replace1965_2_1 -> 762 in the cache */
   /* replace1965_2_2 -> 764 in the cache */
   /* replace1965_2 -> 1317 in the cache */
   /* replace1965 -> 5625 in the cache */

   /* ('bcsel', 'a', ('flrp(is_used_once)', 'b', 'c', 'd'), ('flrp', 'b', 'e', 'd')) => ('flrp', 'b', ('bcsel', 'a', 'c', 'e'), 'd') */
   /* search1966_0 -> 760 in the cache */
   /* search1966_1_0 -> 752 in the cache */
   /* search1966_1_1 -> 753 in the cache */
   /* search1966_1_2 -> 762 in the cache */
   /* search1966_1 -> 5622 in the cache */
   /* search1966_2_0 -> 752 in the cache */
   /* search1966_2_1 -> 764 in the cache */
   /* search1966_2_2 -> 762 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 752, 764, 762 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 5622, 5629 },
      -1,
   } },

   /* replace1966_0 -> 752 in the cache */
   /* replace1966_1_0 -> 760 in the cache */
   /* replace1966_1_1 -> 753 in the cache */
   /* replace1966_1_2 -> 764 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 753, 764 },
      -1,
   } },
   /* replace1966_2 -> 762 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 752, 5631, 762 },
      -1,
   } },

   /* ('bcsel', 'a', ('flrp', 'b', 'c', 'd'), ('flrp(is_used_once)', 'b', 'e', 'd')) => ('flrp', 'b', ('bcsel', 'a', 'c', 'e'), 'd') */
   /* search1967_0 -> 760 in the cache */
   /* search1967_1_0 -> 752 in the cache */
   /* search1967_1_1 -> 753 in the cache */
   /* search1967_1_2 -> 762 in the cache */
   /* search1967_1 -> 5626 in the cache */
   /* search1967_2_0 -> 752 in the cache */
   /* search1967_2_1 -> 764 in the cache */
   /* search1967_2_2 -> 762 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 752, 764, 762 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 5626, 5633 },
      -1,
   } },

   /* replace1967_0 -> 752 in the cache */
   /* replace1967_1_0 -> 760 in the cache */
   /* replace1967_1_1 -> 753 in the cache */
   /* replace1967_1_2 -> 764 in the cache */
   /* replace1967_1 -> 5631 in the cache */
   /* replace1967_2 -> 762 in the cache */
   /* replace1967 -> 5632 in the cache */

   /* ('bcsel', 'a', ('flrp(is_used_once)', 'b', 'c', 'd'), ('flrp', 'e', 'c', 'd')) => ('flrp', ('bcsel', 'a', 'b', 'e'), 'c', 'd') */
   /* search1968_0 -> 760 in the cache */
   /* search1968_1_0 -> 150 in the cache */
   /* search1968_1_1 -> 152 in the cache */
   /* search1968_1_2 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 150, 152, 153 },
      0,
   } },
   /* search1968_2_0 -> 1323 in the cache */
   /* search1968_2_1 -> 152 in the cache */
   /* search1968_2_2 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 1323, 152, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 5635, 5636 },
      -1,
   } },

   /* replace1968_0_0 -> 760 in the cache */
   /* replace1968_0_1 -> 150 in the cache */
   /* replace1968_0_2 -> 1323 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 150, 1323 },
      -1,
   } },
   /* replace1968_1 -> 152 in the cache */
   /* replace1968_2 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 5638, 152, 153 },
      -1,
   } },

   /* ('bcsel', 'a', ('flrp', 'b', 'c', 'd'), ('flrp(is_used_once)', 'e', 'c', 'd')) => ('flrp', ('bcsel', 'a', 'b', 'e'), 'c', 'd') */
   /* search1969_0 -> 760 in the cache */
   /* search1969_1_0 -> 150 in the cache */
   /* search1969_1_1 -> 152 in the cache */
   /* search1969_1_2 -> 153 in the cache */
   /* search1969_1 -> 682 in the cache */
   /* search1969_2_0 -> 1323 in the cache */
   /* search1969_2_1 -> 152 in the cache */
   /* search1969_2_2 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 1323, 152, 153 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 682, 5640 },
      -1,
   } },

   /* replace1969_0_0 -> 760 in the cache */
   /* replace1969_0_1 -> 150 in the cache */
   /* replace1969_0_2 -> 1323 in the cache */
   /* replace1969_0 -> 5638 in the cache */
   /* replace1969_1 -> 152 in the cache */
   /* replace1969_2 -> 153 in the cache */
   /* replace1969 -> 5639 in the cache */

   /* ('bcsel', 'a', ('fmulz(is_used_once)', 'b', 'c'), ('fmulz', 'b', 'd(is_not_const)')) => ('fmulz', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1970_0 -> 760 in the cache */
   /* search1970_1_0 -> 36 in the cache */
   /* search1970_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 36, 135 },
      0,
   } },
   /* search1970_2_0 -> 36 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      3, /* d */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 36, 5643 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5642, 5644 },
      -1,
   } },

   /* replace1970_0 -> 36 in the cache */
   /* replace1970_1_0 -> 760 in the cache */
   /* replace1970_1_1 -> 135 in the cache */
   /* replace1970_1_2 -> 163 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 135, 163 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 36, 5646 },
      -1,
   } },

   /* ('bcsel', 'a', ('fmulz(is_used_once)', 'b', 'c(is_not_const)'), ('fmulz', 'b', 'd')) => ('fmulz', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1971_0 -> 760 in the cache */
   /* search1971_1_0 -> 36 in the cache */
   /* search1971_1_1 -> 347 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 36, 347 },
      0,
   } },
   /* search1971_2_0 -> 36 in the cache */
   /* search1971_2_1 -> 163 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 36, 163 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5648, 5649 },
      -1,
   } },

   /* replace1971_0 -> 36 in the cache */
   /* replace1971_1_0 -> 760 in the cache */
   /* replace1971_1_1 -> 135 in the cache */
   /* replace1971_1_2 -> 163 in the cache */
   /* replace1971_1 -> 5646 in the cache */
   /* replace1971 -> 5647 in the cache */

   /* ('bcsel', 'a', ('fmulz', 'b', 'c(is_not_const)'), ('fmulz(is_used_once)', 'b', 'd')) => ('fmulz', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1972_0 -> 760 in the cache */
   /* search1972_1_0 -> 36 in the cache */
   /* search1972_1_1 -> 347 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 36, 347 },
      -1,
   } },
   /* search1972_2_0 -> 36 in the cache */
   /* search1972_2_1 -> 163 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 36, 163 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5651, 5652 },
      -1,
   } },

   /* replace1972_0 -> 36 in the cache */
   /* replace1972_1_0 -> 760 in the cache */
   /* replace1972_1_1 -> 135 in the cache */
   /* replace1972_1_2 -> 163 in the cache */
   /* replace1972_1 -> 5646 in the cache */
   /* replace1972 -> 5647 in the cache */

   /* ('bcsel', 'a', ('fmulz', 'b', 'c'), ('fmulz(is_used_once)', 'b', 'd(is_not_const)')) => ('fmulz', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1973_0 -> 760 in the cache */
   /* search1973_1_0 -> 36 in the cache */
   /* search1973_1_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 36, 135 },
      -1,
   } },
   /* search1973_2_0 -> 36 in the cache */
   /* search1973_2_1 -> 5643 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 36, 5643 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5654, 5655 },
      -1,
   } },

   /* replace1973_0 -> 36 in the cache */
   /* replace1973_1_0 -> 760 in the cache */
   /* replace1973_1_1 -> 135 in the cache */
   /* replace1973_1_2 -> 163 in the cache */
   /* replace1973_1 -> 5646 in the cache */
   /* replace1973 -> 5647 in the cache */

   /* ('bcsel', 'a', ('fmul(is_used_once)', 'b', 'c'), ('fmul', 'b', 'd(is_not_const)')) => ('fmul', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1974_0 -> 760 in the cache */
   /* search1974_1_0 -> 150 in the cache */
   /* search1974_1_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 150, 152 },
      0,
   } },
   /* search1974_2_0 -> 150 in the cache */
   { .variable = {
      { nir_search_value_variable, -4 },
      3, /* d */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 150, 5658 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5657, 5659 },
      -1,
   } },

   /* replace1974_0 -> 150 in the cache */
   /* replace1974_1_0 -> 760 in the cache */
   /* replace1974_1_1 -> 152 in the cache */
   /* replace1974_1_2 -> 153 in the cache */
   /* replace1974_1 -> 1312 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 150, 1312 },
      -1,
   } },

   /* ('bcsel', 'a', ('fmul(is_used_once)', 'b', 'c(is_not_const)'), ('fmul', 'b', 'd')) => ('fmul', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1975_0 -> 760 in the cache */
   /* search1975_1_0 -> 150 in the cache */
   /* search1975_1_1 -> 3513 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 150, 3513 },
      0,
   } },
   /* search1975_2_0 -> 150 in the cache */
   /* search1975_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 150, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5662, 5663 },
      -1,
   } },

   /* replace1975_0 -> 150 in the cache */
   /* replace1975_1_0 -> 760 in the cache */
   /* replace1975_1_1 -> 152 in the cache */
   /* replace1975_1_2 -> 153 in the cache */
   /* replace1975_1 -> 1312 in the cache */
   /* replace1975 -> 5661 in the cache */

   /* ('bcsel', 'a', ('fmul', 'b', 'c(is_not_const)'), ('fmul(is_used_once)', 'b', 'd')) => ('fmul', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1976_0 -> 760 in the cache */
   /* search1976_1_0 -> 150 in the cache */
   /* search1976_1_1 -> 3513 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 150, 3513 },
      -1,
   } },
   /* search1976_2_0 -> 150 in the cache */
   /* search1976_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 150, 153 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5665, 5666 },
      -1,
   } },

   /* replace1976_0 -> 150 in the cache */
   /* replace1976_1_0 -> 760 in the cache */
   /* replace1976_1_1 -> 152 in the cache */
   /* replace1976_1_2 -> 153 in the cache */
   /* replace1976_1 -> 1312 in the cache */
   /* replace1976 -> 5661 in the cache */

   /* ('bcsel', 'a', ('fmul', 'b', 'c'), ('fmul(is_used_once)', 'b', 'd(is_not_const)')) => ('fmul', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1977_0 -> 760 in the cache */
   /* search1977_1_0 -> 150 in the cache */
   /* search1977_1_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 150, 152 },
      -1,
   } },
   /* search1977_2_0 -> 150 in the cache */
   /* search1977_2_1 -> 5658 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 150, 5658 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5668, 5669 },
      -1,
   } },

   /* replace1977_0 -> 150 in the cache */
   /* replace1977_1_0 -> 760 in the cache */
   /* replace1977_1_1 -> 152 in the cache */
   /* replace1977_1_2 -> 153 in the cache */
   /* replace1977_1 -> 1312 in the cache */
   /* replace1977 -> 5661 in the cache */

   /* ('bcsel', 'a', ('iadd(is_used_once)', 'b', 'c'), ('iadd', 'b', 'd(is_not_const)')) => ('iadd', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1978_0 -> 760 in the cache */
   /* search1978_1_0 -> 150 in the cache */
   /* search1978_1_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 150, 152 },
      0,
   } },
   /* search1978_2_0 -> 150 in the cache */
   /* search1978_2_1 -> 5658 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 150, 5658 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5671, 5672 },
      -1,
   } },

   /* replace1978_0 -> 150 in the cache */
   /* replace1978_1_0 -> 760 in the cache */
   /* replace1978_1_1 -> 152 in the cache */
   /* replace1978_1_2 -> 153 in the cache */
   /* replace1978_1 -> 1312 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 150, 1312 },
      -1,
   } },

   /* ('bcsel', 'a', ('iadd(is_used_once)', 'b', 'c(is_not_const)'), ('iadd', 'b', 'd')) => ('iadd', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1979_0 -> 760 in the cache */
   /* search1979_1_0 -> 150 in the cache */
   /* search1979_1_1 -> 3513 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 150, 3513 },
      0,
   } },
   /* search1979_2_0 -> 150 in the cache */
   /* search1979_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 150, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5675, 5676 },
      -1,
   } },

   /* replace1979_0 -> 150 in the cache */
   /* replace1979_1_0 -> 760 in the cache */
   /* replace1979_1_1 -> 152 in the cache */
   /* replace1979_1_2 -> 153 in the cache */
   /* replace1979_1 -> 1312 in the cache */
   /* replace1979 -> 5674 in the cache */

   /* ('bcsel', 'a', ('iadd', 'b', 'c(is_not_const)'), ('iadd(is_used_once)', 'b', 'd')) => ('iadd', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1980_0 -> 760 in the cache */
   /* search1980_1_0 -> 150 in the cache */
   /* search1980_1_1 -> 3513 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 150, 3513 },
      -1,
   } },
   /* search1980_2_0 -> 150 in the cache */
   /* search1980_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 150, 153 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5678, 5679 },
      -1,
   } },

   /* replace1980_0 -> 150 in the cache */
   /* replace1980_1_0 -> 760 in the cache */
   /* replace1980_1_1 -> 152 in the cache */
   /* replace1980_1_2 -> 153 in the cache */
   /* replace1980_1 -> 1312 in the cache */
   /* replace1980 -> 5674 in the cache */

   /* ('bcsel', 'a', ('iadd', 'b', 'c'), ('iadd(is_used_once)', 'b', 'd(is_not_const)')) => ('iadd', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1981_0 -> 760 in the cache */
   /* search1981_1_0 -> 150 in the cache */
   /* search1981_1_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 150, 152 },
      -1,
   } },
   /* search1981_2_0 -> 150 in the cache */
   /* search1981_2_1 -> 5658 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 150, 5658 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5681, 5682 },
      -1,
   } },

   /* replace1981_0 -> 150 in the cache */
   /* replace1981_1_0 -> 760 in the cache */
   /* replace1981_1_1 -> 152 in the cache */
   /* replace1981_1_2 -> 153 in the cache */
   /* replace1981_1 -> 1312 in the cache */
   /* replace1981 -> 5674 in the cache */

   /* ('bcsel', 'a', ('imul(is_used_once)', 'b', 'c'), ('imul', 'b', 'd(is_not_const)')) => ('imul', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1982_0 -> 760 in the cache */
   /* search1982_1_0 -> 150 in the cache */
   /* search1982_1_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 150, 152 },
      0,
   } },
   /* search1982_2_0 -> 150 in the cache */
   /* search1982_2_1 -> 5658 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 150, 5658 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5684, 5685 },
      -1,
   } },

   /* replace1982_0 -> 150 in the cache */
   /* replace1982_1_0 -> 760 in the cache */
   /* replace1982_1_1 -> 152 in the cache */
   /* replace1982_1_2 -> 153 in the cache */
   /* replace1982_1 -> 1312 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 150, 1312 },
      -1,
   } },

   /* ('bcsel', 'a', ('imul(is_used_once)', 'b', 'c(is_not_const)'), ('imul', 'b', 'd')) => ('imul', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1983_0 -> 760 in the cache */
   /* search1983_1_0 -> 150 in the cache */
   /* search1983_1_1 -> 3513 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 150, 3513 },
      0,
   } },
   /* search1983_2_0 -> 150 in the cache */
   /* search1983_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 150, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5688, 5689 },
      -1,
   } },

   /* replace1983_0 -> 150 in the cache */
   /* replace1983_1_0 -> 760 in the cache */
   /* replace1983_1_1 -> 152 in the cache */
   /* replace1983_1_2 -> 153 in the cache */
   /* replace1983_1 -> 1312 in the cache */
   /* replace1983 -> 5687 in the cache */

   /* ('bcsel', 'a', ('imul', 'b', 'c(is_not_const)'), ('imul(is_used_once)', 'b', 'd')) => ('imul', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1984_0 -> 760 in the cache */
   /* search1984_1_0 -> 150 in the cache */
   /* search1984_1_1 -> 3513 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 150, 3513 },
      -1,
   } },
   /* search1984_2_0 -> 150 in the cache */
   /* search1984_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 150, 153 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5691, 5692 },
      -1,
   } },

   /* replace1984_0 -> 150 in the cache */
   /* replace1984_1_0 -> 760 in the cache */
   /* replace1984_1_1 -> 152 in the cache */
   /* replace1984_1_2 -> 153 in the cache */
   /* replace1984_1 -> 1312 in the cache */
   /* replace1984 -> 5687 in the cache */

   /* ('bcsel', 'a', ('imul', 'b', 'c'), ('imul(is_used_once)', 'b', 'd(is_not_const)')) => ('imul', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1985_0 -> 760 in the cache */
   /* search1985_1_0 -> 150 in the cache */
   /* search1985_1_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 150, 152 },
      -1,
   } },
   /* search1985_2_0 -> 150 in the cache */
   /* search1985_2_1 -> 5658 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 150, 5658 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5694, 5695 },
      -1,
   } },

   /* replace1985_0 -> 150 in the cache */
   /* replace1985_1_0 -> 760 in the cache */
   /* replace1985_1_1 -> 152 in the cache */
   /* replace1985_1_2 -> 153 in the cache */
   /* replace1985_1 -> 1312 in the cache */
   /* replace1985 -> 5687 in the cache */

   /* ('bcsel', 'a', ('iand(is_used_once)', 'b', 'c'), ('iand', 'b', 'd(is_not_const)')) => ('iand', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1986_0 -> 760 in the cache */
   /* search1986_1_0 -> 150 in the cache */
   /* search1986_1_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 150, 152 },
      0,
   } },
   /* search1986_2_0 -> 150 in the cache */
   /* search1986_2_1 -> 5658 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 150, 5658 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5697, 5698 },
      -1,
   } },

   /* replace1986_0 -> 150 in the cache */
   /* replace1986_1_0 -> 760 in the cache */
   /* replace1986_1_1 -> 152 in the cache */
   /* replace1986_1_2 -> 153 in the cache */
   /* replace1986_1 -> 1312 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 150, 1312 },
      -1,
   } },

   /* ('bcsel', 'a', ('iand(is_used_once)', 'b', 'c(is_not_const)'), ('iand', 'b', 'd')) => ('iand', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1987_0 -> 760 in the cache */
   /* search1987_1_0 -> 150 in the cache */
   /* search1987_1_1 -> 3513 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 150, 3513 },
      0,
   } },
   /* search1987_2_0 -> 150 in the cache */
   /* search1987_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 150, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5701, 5702 },
      -1,
   } },

   /* replace1987_0 -> 150 in the cache */
   /* replace1987_1_0 -> 760 in the cache */
   /* replace1987_1_1 -> 152 in the cache */
   /* replace1987_1_2 -> 153 in the cache */
   /* replace1987_1 -> 1312 in the cache */
   /* replace1987 -> 5700 in the cache */

   /* ('bcsel', 'a', ('iand', 'b', 'c(is_not_const)'), ('iand(is_used_once)', 'b', 'd')) => ('iand', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1988_0 -> 760 in the cache */
   /* search1988_1_0 -> 150 in the cache */
   /* search1988_1_1 -> 3513 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 150, 3513 },
      -1,
   } },
   /* search1988_2_0 -> 150 in the cache */
   /* search1988_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 150, 153 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5704, 5705 },
      -1,
   } },

   /* replace1988_0 -> 150 in the cache */
   /* replace1988_1_0 -> 760 in the cache */
   /* replace1988_1_1 -> 152 in the cache */
   /* replace1988_1_2 -> 153 in the cache */
   /* replace1988_1 -> 1312 in the cache */
   /* replace1988 -> 5700 in the cache */

   /* ('bcsel', 'a', ('iand', 'b', 'c'), ('iand(is_used_once)', 'b', 'd(is_not_const)')) => ('iand', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1989_0 -> 760 in the cache */
   /* search1989_1_0 -> 150 in the cache */
   /* search1989_1_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 150, 152 },
      -1,
   } },
   /* search1989_2_0 -> 150 in the cache */
   /* search1989_2_1 -> 5658 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 150, 5658 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5707, 5708 },
      -1,
   } },

   /* replace1989_0 -> 150 in the cache */
   /* replace1989_1_0 -> 760 in the cache */
   /* replace1989_1_1 -> 152 in the cache */
   /* replace1989_1_2 -> 153 in the cache */
   /* replace1989_1 -> 1312 in the cache */
   /* replace1989 -> 5700 in the cache */

   /* ('bcsel', 'a', ('ior(is_used_once)', 'b', 'c'), ('ior', 'b', 'd(is_not_const)')) => ('ior', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1990_0 -> 760 in the cache */
   /* search1990_1_0 -> 150 in the cache */
   /* search1990_1_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 150, 152 },
      0,
   } },
   /* search1990_2_0 -> 150 in the cache */
   /* search1990_2_1 -> 5658 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 150, 5658 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5710, 5711 },
      -1,
   } },

   /* replace1990_0 -> 150 in the cache */
   /* replace1990_1_0 -> 760 in the cache */
   /* replace1990_1_1 -> 152 in the cache */
   /* replace1990_1_2 -> 153 in the cache */
   /* replace1990_1 -> 1312 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 150, 1312 },
      -1,
   } },

   /* ('bcsel', 'a', ('ior(is_used_once)', 'b', 'c(is_not_const)'), ('ior', 'b', 'd')) => ('ior', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1991_0 -> 760 in the cache */
   /* search1991_1_0 -> 150 in the cache */
   /* search1991_1_1 -> 3513 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 150, 3513 },
      0,
   } },
   /* search1991_2_0 -> 150 in the cache */
   /* search1991_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 150, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5714, 5715 },
      -1,
   } },

   /* replace1991_0 -> 150 in the cache */
   /* replace1991_1_0 -> 760 in the cache */
   /* replace1991_1_1 -> 152 in the cache */
   /* replace1991_1_2 -> 153 in the cache */
   /* replace1991_1 -> 1312 in the cache */
   /* replace1991 -> 5713 in the cache */

   /* ('bcsel', 'a', ('ior', 'b', 'c(is_not_const)'), ('ior(is_used_once)', 'b', 'd')) => ('ior', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1992_0 -> 760 in the cache */
   /* search1992_1_0 -> 150 in the cache */
   /* search1992_1_1 -> 3513 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 150, 3513 },
      -1,
   } },
   /* search1992_2_0 -> 150 in the cache */
   /* search1992_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 150, 153 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5717, 5718 },
      -1,
   } },

   /* replace1992_0 -> 150 in the cache */
   /* replace1992_1_0 -> 760 in the cache */
   /* replace1992_1_1 -> 152 in the cache */
   /* replace1992_1_2 -> 153 in the cache */
   /* replace1992_1 -> 1312 in the cache */
   /* replace1992 -> 5713 in the cache */

   /* ('bcsel', 'a', ('ior', 'b', 'c'), ('ior(is_used_once)', 'b', 'd(is_not_const)')) => ('ior', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1993_0 -> 760 in the cache */
   /* search1993_1_0 -> 150 in the cache */
   /* search1993_1_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 150, 152 },
      -1,
   } },
   /* search1993_2_0 -> 150 in the cache */
   /* search1993_2_1 -> 5658 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 150, 5658 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5720, 5721 },
      -1,
   } },

   /* replace1993_0 -> 150 in the cache */
   /* replace1993_1_0 -> 760 in the cache */
   /* replace1993_1_1 -> 152 in the cache */
   /* replace1993_1_2 -> 153 in the cache */
   /* replace1993_1 -> 1312 in the cache */
   /* replace1993 -> 5713 in the cache */

   /* ('bcsel', 'a', ('ixor(is_used_once)', 'b', 'c'), ('ixor', 'b', 'd(is_not_const)')) => ('ixor', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1994_0 -> 760 in the cache */
   /* search1994_1_0 -> 150 in the cache */
   /* search1994_1_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 150, 152 },
      0,
   } },
   /* search1994_2_0 -> 150 in the cache */
   /* search1994_2_1 -> 5658 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ixor,
      1, 1,
      { 150, 5658 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5723, 5724 },
      -1,
   } },

   /* replace1994_0 -> 150 in the cache */
   /* replace1994_1_0 -> 760 in the cache */
   /* replace1994_1_1 -> 152 in the cache */
   /* replace1994_1_2 -> 153 in the cache */
   /* replace1994_1 -> 1312 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 150, 1312 },
      -1,
   } },

   /* ('bcsel', 'a', ('ixor(is_used_once)', 'b', 'c(is_not_const)'), ('ixor', 'b', 'd')) => ('ixor', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1995_0 -> 760 in the cache */
   /* search1995_1_0 -> 150 in the cache */
   /* search1995_1_1 -> 3513 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 150, 3513 },
      0,
   } },
   /* search1995_2_0 -> 150 in the cache */
   /* search1995_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ixor,
      1, 1,
      { 150, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5727, 5728 },
      -1,
   } },

   /* replace1995_0 -> 150 in the cache */
   /* replace1995_1_0 -> 760 in the cache */
   /* replace1995_1_1 -> 152 in the cache */
   /* replace1995_1_2 -> 153 in the cache */
   /* replace1995_1 -> 1312 in the cache */
   /* replace1995 -> 5726 in the cache */

   /* ('bcsel', 'a', ('ixor', 'b', 'c(is_not_const)'), ('ixor(is_used_once)', 'b', 'd')) => ('ixor', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1996_0 -> 760 in the cache */
   /* search1996_1_0 -> 150 in the cache */
   /* search1996_1_1 -> 3513 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 150, 3513 },
      -1,
   } },
   /* search1996_2_0 -> 150 in the cache */
   /* search1996_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ixor,
      1, 1,
      { 150, 153 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5730, 5731 },
      -1,
   } },

   /* replace1996_0 -> 150 in the cache */
   /* replace1996_1_0 -> 760 in the cache */
   /* replace1996_1_1 -> 152 in the cache */
   /* replace1996_1_2 -> 153 in the cache */
   /* replace1996_1 -> 1312 in the cache */
   /* replace1996 -> 5726 in the cache */

   /* ('bcsel', 'a', ('ixor', 'b', 'c'), ('ixor(is_used_once)', 'b', 'd(is_not_const)')) => ('ixor', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1997_0 -> 760 in the cache */
   /* search1997_1_0 -> 150 in the cache */
   /* search1997_1_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 150, 152 },
      -1,
   } },
   /* search1997_2_0 -> 150 in the cache */
   /* search1997_2_1 -> 5658 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ixor,
      1, 1,
      { 150, 5658 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5733, 5734 },
      -1,
   } },

   /* replace1997_0 -> 150 in the cache */
   /* replace1997_1_0 -> 760 in the cache */
   /* replace1997_1_1 -> 152 in the cache */
   /* replace1997_1_2 -> 153 in the cache */
   /* replace1997_1 -> 1312 in the cache */
   /* replace1997 -> 5726 in the cache */

   /* ('bcsel', 'a', ('fmin(is_used_once)', 'b', 'c'), ('fmin', 'b', 'd(is_not_const)')) => ('fmin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1998_0 -> 760 in the cache */
   /* search1998_1_0 -> 150 in the cache */
   /* search1998_1_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 150, 152 },
      0,
   } },
   /* search1998_2_0 -> 150 in the cache */
   /* search1998_2_1 -> 5658 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 150, 5658 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5736, 5737 },
      -1,
   } },

   /* replace1998_0 -> 150 in the cache */
   /* replace1998_1_0 -> 760 in the cache */
   /* replace1998_1_1 -> 152 in the cache */
   /* replace1998_1_2 -> 153 in the cache */
   /* replace1998_1 -> 1312 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 150, 1312 },
      -1,
   } },

   /* ('bcsel', 'a', ('fmin(is_used_once)', 'b', 'c(is_not_const)'), ('fmin', 'b', 'd')) => ('fmin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1999_0 -> 760 in the cache */
   /* search1999_1_0 -> 150 in the cache */
   /* search1999_1_1 -> 3513 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 150, 3513 },
      0,
   } },
   /* search1999_2_0 -> 150 in the cache */
   /* search1999_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 150, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5740, 5741 },
      -1,
   } },

   /* replace1999_0 -> 150 in the cache */
   /* replace1999_1_0 -> 760 in the cache */
   /* replace1999_1_1 -> 152 in the cache */
   /* replace1999_1_2 -> 153 in the cache */
   /* replace1999_1 -> 1312 in the cache */
   /* replace1999 -> 5739 in the cache */

   /* ('bcsel', 'a', ('fmin', 'b', 'c(is_not_const)'), ('fmin(is_used_once)', 'b', 'd')) => ('fmin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2000_0 -> 760 in the cache */
   /* search2000_1_0 -> 150 in the cache */
   /* search2000_1_1 -> 3513 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 150, 3513 },
      -1,
   } },
   /* search2000_2_0 -> 150 in the cache */
   /* search2000_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 150, 153 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5743, 5744 },
      -1,
   } },

   /* replace2000_0 -> 150 in the cache */
   /* replace2000_1_0 -> 760 in the cache */
   /* replace2000_1_1 -> 152 in the cache */
   /* replace2000_1_2 -> 153 in the cache */
   /* replace2000_1 -> 1312 in the cache */
   /* replace2000 -> 5739 in the cache */

   /* ('bcsel', 'a', ('fmin', 'b', 'c'), ('fmin(is_used_once)', 'b', 'd(is_not_const)')) => ('fmin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2001_0 -> 760 in the cache */
   /* search2001_1_0 -> 150 in the cache */
   /* search2001_1_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 150, 152 },
      -1,
   } },
   /* search2001_2_0 -> 150 in the cache */
   /* search2001_2_1 -> 5658 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 150, 5658 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5746, 5747 },
      -1,
   } },

   /* replace2001_0 -> 150 in the cache */
   /* replace2001_1_0 -> 760 in the cache */
   /* replace2001_1_1 -> 152 in the cache */
   /* replace2001_1_2 -> 153 in the cache */
   /* replace2001_1 -> 1312 in the cache */
   /* replace2001 -> 5739 in the cache */

   /* ('bcsel', 'a', ('fmax(is_used_once)', 'b', 'c'), ('fmax', 'b', 'd(is_not_const)')) => ('fmax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2002_0 -> 760 in the cache */
   /* search2002_1_0 -> 150 in the cache */
   /* search2002_1_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 150, 152 },
      0,
   } },
   /* search2002_2_0 -> 150 in the cache */
   /* search2002_2_1 -> 5658 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 150, 5658 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5749, 5750 },
      -1,
   } },

   /* replace2002_0 -> 150 in the cache */
   /* replace2002_1_0 -> 760 in the cache */
   /* replace2002_1_1 -> 152 in the cache */
   /* replace2002_1_2 -> 153 in the cache */
   /* replace2002_1 -> 1312 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 150, 1312 },
      -1,
   } },

   /* ('bcsel', 'a', ('fmax(is_used_once)', 'b', 'c(is_not_const)'), ('fmax', 'b', 'd')) => ('fmax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2003_0 -> 760 in the cache */
   /* search2003_1_0 -> 150 in the cache */
   /* search2003_1_1 -> 3513 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 150, 3513 },
      0,
   } },
   /* search2003_2_0 -> 150 in the cache */
   /* search2003_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 150, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5753, 5754 },
      -1,
   } },

   /* replace2003_0 -> 150 in the cache */
   /* replace2003_1_0 -> 760 in the cache */
   /* replace2003_1_1 -> 152 in the cache */
   /* replace2003_1_2 -> 153 in the cache */
   /* replace2003_1 -> 1312 in the cache */
   /* replace2003 -> 5752 in the cache */

   /* ('bcsel', 'a', ('fmax', 'b', 'c(is_not_const)'), ('fmax(is_used_once)', 'b', 'd')) => ('fmax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2004_0 -> 760 in the cache */
   /* search2004_1_0 -> 150 in the cache */
   /* search2004_1_1 -> 3513 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 150, 3513 },
      -1,
   } },
   /* search2004_2_0 -> 150 in the cache */
   /* search2004_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 150, 153 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5756, 5757 },
      -1,
   } },

   /* replace2004_0 -> 150 in the cache */
   /* replace2004_1_0 -> 760 in the cache */
   /* replace2004_1_1 -> 152 in the cache */
   /* replace2004_1_2 -> 153 in the cache */
   /* replace2004_1 -> 1312 in the cache */
   /* replace2004 -> 5752 in the cache */

   /* ('bcsel', 'a', ('fmax', 'b', 'c'), ('fmax(is_used_once)', 'b', 'd(is_not_const)')) => ('fmax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2005_0 -> 760 in the cache */
   /* search2005_1_0 -> 150 in the cache */
   /* search2005_1_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 150, 152 },
      -1,
   } },
   /* search2005_2_0 -> 150 in the cache */
   /* search2005_2_1 -> 5658 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 150, 5658 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5759, 5760 },
      -1,
   } },

   /* replace2005_0 -> 150 in the cache */
   /* replace2005_1_0 -> 760 in the cache */
   /* replace2005_1_1 -> 152 in the cache */
   /* replace2005_1_2 -> 153 in the cache */
   /* replace2005_1 -> 1312 in the cache */
   /* replace2005 -> 5752 in the cache */

   /* ('bcsel', 'a', ('imin(is_used_once)', 'b', 'c'), ('imin', 'b', 'd(is_not_const)')) => ('imin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2006_0 -> 760 in the cache */
   /* search2006_1_0 -> 150 in the cache */
   /* search2006_1_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 150, 152 },
      0,
   } },
   /* search2006_2_0 -> 150 in the cache */
   /* search2006_2_1 -> 5658 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imin,
      1, 1,
      { 150, 5658 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5762, 5763 },
      -1,
   } },

   /* replace2006_0 -> 150 in the cache */
   /* replace2006_1_0 -> 760 in the cache */
   /* replace2006_1_1 -> 152 in the cache */
   /* replace2006_1_2 -> 153 in the cache */
   /* replace2006_1 -> 1312 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 150, 1312 },
      -1,
   } },

   /* ('bcsel', 'a', ('imin(is_used_once)', 'b', 'c(is_not_const)'), ('imin', 'b', 'd')) => ('imin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2007_0 -> 760 in the cache */
   /* search2007_1_0 -> 150 in the cache */
   /* search2007_1_1 -> 3513 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 150, 3513 },
      0,
   } },
   /* search2007_2_0 -> 150 in the cache */
   /* search2007_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imin,
      1, 1,
      { 150, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5766, 5767 },
      -1,
   } },

   /* replace2007_0 -> 150 in the cache */
   /* replace2007_1_0 -> 760 in the cache */
   /* replace2007_1_1 -> 152 in the cache */
   /* replace2007_1_2 -> 153 in the cache */
   /* replace2007_1 -> 1312 in the cache */
   /* replace2007 -> 5765 in the cache */

   /* ('bcsel', 'a', ('imin', 'b', 'c(is_not_const)'), ('imin(is_used_once)', 'b', 'd')) => ('imin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2008_0 -> 760 in the cache */
   /* search2008_1_0 -> 150 in the cache */
   /* search2008_1_1 -> 3513 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 150, 3513 },
      -1,
   } },
   /* search2008_2_0 -> 150 in the cache */
   /* search2008_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imin,
      1, 1,
      { 150, 153 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5769, 5770 },
      -1,
   } },

   /* replace2008_0 -> 150 in the cache */
   /* replace2008_1_0 -> 760 in the cache */
   /* replace2008_1_1 -> 152 in the cache */
   /* replace2008_1_2 -> 153 in the cache */
   /* replace2008_1 -> 1312 in the cache */
   /* replace2008 -> 5765 in the cache */

   /* ('bcsel', 'a', ('imin', 'b', 'c'), ('imin(is_used_once)', 'b', 'd(is_not_const)')) => ('imin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2009_0 -> 760 in the cache */
   /* search2009_1_0 -> 150 in the cache */
   /* search2009_1_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 150, 152 },
      -1,
   } },
   /* search2009_2_0 -> 150 in the cache */
   /* search2009_2_1 -> 5658 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imin,
      1, 1,
      { 150, 5658 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5772, 5773 },
      -1,
   } },

   /* replace2009_0 -> 150 in the cache */
   /* replace2009_1_0 -> 760 in the cache */
   /* replace2009_1_1 -> 152 in the cache */
   /* replace2009_1_2 -> 153 in the cache */
   /* replace2009_1 -> 1312 in the cache */
   /* replace2009 -> 5765 in the cache */

   /* ('bcsel', 'a', ('imax(is_used_once)', 'b', 'c'), ('imax', 'b', 'd(is_not_const)')) => ('imax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2010_0 -> 760 in the cache */
   /* search2010_1_0 -> 150 in the cache */
   /* search2010_1_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 150, 152 },
      0,
   } },
   /* search2010_2_0 -> 150 in the cache */
   /* search2010_2_1 -> 5658 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imax,
      1, 1,
      { 150, 5658 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5775, 5776 },
      -1,
   } },

   /* replace2010_0 -> 150 in the cache */
   /* replace2010_1_0 -> 760 in the cache */
   /* replace2010_1_1 -> 152 in the cache */
   /* replace2010_1_2 -> 153 in the cache */
   /* replace2010_1 -> 1312 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 150, 1312 },
      -1,
   } },

   /* ('bcsel', 'a', ('imax(is_used_once)', 'b', 'c(is_not_const)'), ('imax', 'b', 'd')) => ('imax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2011_0 -> 760 in the cache */
   /* search2011_1_0 -> 150 in the cache */
   /* search2011_1_1 -> 3513 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 150, 3513 },
      0,
   } },
   /* search2011_2_0 -> 150 in the cache */
   /* search2011_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imax,
      1, 1,
      { 150, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5779, 5780 },
      -1,
   } },

   /* replace2011_0 -> 150 in the cache */
   /* replace2011_1_0 -> 760 in the cache */
   /* replace2011_1_1 -> 152 in the cache */
   /* replace2011_1_2 -> 153 in the cache */
   /* replace2011_1 -> 1312 in the cache */
   /* replace2011 -> 5778 in the cache */

   /* ('bcsel', 'a', ('imax', 'b', 'c(is_not_const)'), ('imax(is_used_once)', 'b', 'd')) => ('imax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2012_0 -> 760 in the cache */
   /* search2012_1_0 -> 150 in the cache */
   /* search2012_1_1 -> 3513 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 150, 3513 },
      -1,
   } },
   /* search2012_2_0 -> 150 in the cache */
   /* search2012_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imax,
      1, 1,
      { 150, 153 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5782, 5783 },
      -1,
   } },

   /* replace2012_0 -> 150 in the cache */
   /* replace2012_1_0 -> 760 in the cache */
   /* replace2012_1_1 -> 152 in the cache */
   /* replace2012_1_2 -> 153 in the cache */
   /* replace2012_1 -> 1312 in the cache */
   /* replace2012 -> 5778 in the cache */

   /* ('bcsel', 'a', ('imax', 'b', 'c'), ('imax(is_used_once)', 'b', 'd(is_not_const)')) => ('imax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2013_0 -> 760 in the cache */
   /* search2013_1_0 -> 150 in the cache */
   /* search2013_1_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 150, 152 },
      -1,
   } },
   /* search2013_2_0 -> 150 in the cache */
   /* search2013_2_1 -> 5658 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imax,
      1, 1,
      { 150, 5658 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5785, 5786 },
      -1,
   } },

   /* replace2013_0 -> 150 in the cache */
   /* replace2013_1_0 -> 760 in the cache */
   /* replace2013_1_1 -> 152 in the cache */
   /* replace2013_1_2 -> 153 in the cache */
   /* replace2013_1 -> 1312 in the cache */
   /* replace2013 -> 5778 in the cache */

   /* ('bcsel', 'a', ('umin(is_used_once)', 'b', 'c'), ('umin', 'b', 'd(is_not_const)')) => ('umin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2014_0 -> 760 in the cache */
   /* search2014_1_0 -> 150 in the cache */
   /* search2014_1_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 150, 152 },
      0,
   } },
   /* search2014_2_0 -> 150 in the cache */
   /* search2014_2_1 -> 5658 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 150, 5658 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5788, 5789 },
      -1,
   } },

   /* replace2014_0 -> 150 in the cache */
   /* replace2014_1_0 -> 760 in the cache */
   /* replace2014_1_1 -> 152 in the cache */
   /* replace2014_1_2 -> 153 in the cache */
   /* replace2014_1 -> 1312 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 150, 1312 },
      -1,
   } },

   /* ('bcsel', 'a', ('umin(is_used_once)', 'b', 'c(is_not_const)'), ('umin', 'b', 'd')) => ('umin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2015_0 -> 760 in the cache */
   /* search2015_1_0 -> 150 in the cache */
   /* search2015_1_1 -> 3513 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 150, 3513 },
      0,
   } },
   /* search2015_2_0 -> 150 in the cache */
   /* search2015_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 150, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5792, 5793 },
      -1,
   } },

   /* replace2015_0 -> 150 in the cache */
   /* replace2015_1_0 -> 760 in the cache */
   /* replace2015_1_1 -> 152 in the cache */
   /* replace2015_1_2 -> 153 in the cache */
   /* replace2015_1 -> 1312 in the cache */
   /* replace2015 -> 5791 in the cache */

   /* ('bcsel', 'a', ('umin', 'b', 'c(is_not_const)'), ('umin(is_used_once)', 'b', 'd')) => ('umin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2016_0 -> 760 in the cache */
   /* search2016_1_0 -> 150 in the cache */
   /* search2016_1_1 -> 3513 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 150, 3513 },
      -1,
   } },
   /* search2016_2_0 -> 150 in the cache */
   /* search2016_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 150, 153 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5795, 5796 },
      -1,
   } },

   /* replace2016_0 -> 150 in the cache */
   /* replace2016_1_0 -> 760 in the cache */
   /* replace2016_1_1 -> 152 in the cache */
   /* replace2016_1_2 -> 153 in the cache */
   /* replace2016_1 -> 1312 in the cache */
   /* replace2016 -> 5791 in the cache */

   /* ('bcsel', 'a', ('umin', 'b', 'c'), ('umin(is_used_once)', 'b', 'd(is_not_const)')) => ('umin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2017_0 -> 760 in the cache */
   /* search2017_1_0 -> 150 in the cache */
   /* search2017_1_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 150, 152 },
      -1,
   } },
   /* search2017_2_0 -> 150 in the cache */
   /* search2017_2_1 -> 5658 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 150, 5658 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5798, 5799 },
      -1,
   } },

   /* replace2017_0 -> 150 in the cache */
   /* replace2017_1_0 -> 760 in the cache */
   /* replace2017_1_1 -> 152 in the cache */
   /* replace2017_1_2 -> 153 in the cache */
   /* replace2017_1 -> 1312 in the cache */
   /* replace2017 -> 5791 in the cache */

   /* ('bcsel', 'a', ('umax(is_used_once)', 'b', 'c'), ('umax', 'b', 'd(is_not_const)')) => ('umax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2018_0 -> 760 in the cache */
   /* search2018_1_0 -> 150 in the cache */
   /* search2018_1_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 150, 152 },
      0,
   } },
   /* search2018_2_0 -> 150 in the cache */
   /* search2018_2_1 -> 5658 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umax,
      1, 1,
      { 150, 5658 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5801, 5802 },
      -1,
   } },

   /* replace2018_0 -> 150 in the cache */
   /* replace2018_1_0 -> 760 in the cache */
   /* replace2018_1_1 -> 152 in the cache */
   /* replace2018_1_2 -> 153 in the cache */
   /* replace2018_1 -> 1312 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 150, 1312 },
      -1,
   } },

   /* ('bcsel', 'a', ('umax(is_used_once)', 'b', 'c(is_not_const)'), ('umax', 'b', 'd')) => ('umax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2019_0 -> 760 in the cache */
   /* search2019_1_0 -> 150 in the cache */
   /* search2019_1_1 -> 3513 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 150, 3513 },
      0,
   } },
   /* search2019_2_0 -> 150 in the cache */
   /* search2019_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umax,
      1, 1,
      { 150, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5805, 5806 },
      -1,
   } },

   /* replace2019_0 -> 150 in the cache */
   /* replace2019_1_0 -> 760 in the cache */
   /* replace2019_1_1 -> 152 in the cache */
   /* replace2019_1_2 -> 153 in the cache */
   /* replace2019_1 -> 1312 in the cache */
   /* replace2019 -> 5804 in the cache */

   /* ('bcsel', 'a', ('umax', 'b', 'c(is_not_const)'), ('umax(is_used_once)', 'b', 'd')) => ('umax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2020_0 -> 760 in the cache */
   /* search2020_1_0 -> 150 in the cache */
   /* search2020_1_1 -> 3513 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 150, 3513 },
      -1,
   } },
   /* search2020_2_0 -> 150 in the cache */
   /* search2020_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umax,
      1, 1,
      { 150, 153 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5808, 5809 },
      -1,
   } },

   /* replace2020_0 -> 150 in the cache */
   /* replace2020_1_0 -> 760 in the cache */
   /* replace2020_1_1 -> 152 in the cache */
   /* replace2020_1_2 -> 153 in the cache */
   /* replace2020_1 -> 1312 in the cache */
   /* replace2020 -> 5804 in the cache */

   /* ('bcsel', 'a', ('umax', 'b', 'c'), ('umax(is_used_once)', 'b', 'd(is_not_const)')) => ('umax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2021_0 -> 760 in the cache */
   /* search2021_1_0 -> 150 in the cache */
   /* search2021_1_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 150, 152 },
      -1,
   } },
   /* search2021_2_0 -> 150 in the cache */
   /* search2021_2_1 -> 5658 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umax,
      1, 1,
      { 150, 5658 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 760, 5811, 5812 },
      -1,
   } },

   /* replace2021_0 -> 150 in the cache */
   /* replace2021_1_0 -> 760 in the cache */
   /* replace2021_1_1 -> 152 in the cache */
   /* replace2021_1_2 -> 153 in the cache */
   /* replace2021_1 -> 1312 in the cache */
   /* replace2021 -> 5804 in the cache */

   /* ('bcsel', 'a', ('fpow(is_used_once)', 'b', 'c'), ('fpow', 'b', 'd')) => ('fpow', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2022_0 -> 760 in the cache */
   /* search2022_1_0 -> 150 in the cache */
   /* search2022_1_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 150, 152 },
      0,
   } },
   /* search2022_2_0 -> 150 in the cache */
   /* search2022_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 150, 153 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 5814, 5815 },
      -1,
   } },

   /* replace2022_0 -> 150 in the cache */
   /* replace2022_1_0 -> 760 in the cache */
   /* replace2022_1_1 -> 152 in the cache */
   /* replace2022_1_2 -> 153 in the cache */
   /* replace2022_1 -> 1312 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 150, 1312 },
      -1,
   } },

   /* ('bcsel', 'a', ('fpow', 'b', 'c'), ('fpow(is_used_once)', 'b', 'd')) => ('fpow', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2023_0 -> 760 in the cache */
   /* search2023_1_0 -> 150 in the cache */
   /* search2023_1_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 150, 152 },
      -1,
   } },
   /* search2023_2_0 -> 150 in the cache */
   /* search2023_2_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 150, 153 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 5818, 5819 },
      -1,
   } },

   /* replace2023_0 -> 150 in the cache */
   /* replace2023_1_0 -> 760 in the cache */
   /* replace2023_1_1 -> 152 in the cache */
   /* replace2023_1_2 -> 153 in the cache */
   /* replace2023_1 -> 1312 in the cache */
   /* replace2023 -> 5817 in the cache */

   /* ('bcsel', 'a', ('fpow(is_used_once)', 'b', 'c'), ('fpow', 'd', 'c')) => ('fpow', ('bcsel', 'a', 'b', 'd'), 'c') */
   /* search2024_0 -> 760 in the cache */
   /* search2024_1_0 -> 127 in the cache */
   /* search2024_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 127, 129 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search2024_2_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 5822, 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 5821, 5823 },
      -1,
   } },

   /* replace2024_0_0 -> 760 in the cache */
   /* replace2024_0_1 -> 127 in the cache */
   /* replace2024_0_2 -> 5822 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 127, 5822 },
      -1,
   } },
   /* replace2024_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 5825, 129 },
      -1,
   } },

   /* ('bcsel', 'a', ('fpow', 'b', 'c'), ('fpow(is_used_once)', 'd', 'c')) => ('fpow', ('bcsel', 'a', 'b', 'd'), 'c') */
   /* search2025_0 -> 760 in the cache */
   /* search2025_1_0 -> 127 in the cache */
   /* search2025_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 127, 129 },
      -1,
   } },
   /* search2025_2_0 -> 5822 in the cache */
   /* search2025_2_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 5822, 129 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 760, 5827, 5828 },
      -1,
   } },

   /* replace2025_0_0 -> 760 in the cache */
   /* replace2025_0_1 -> 127 in the cache */
   /* replace2025_0_2 -> 5822 in the cache */
   /* replace2025_0 -> 5825 in the cache */
   /* replace2025_1 -> 129 in the cache */
   /* replace2025 -> 5826 in the cache */

   /* ('bcsel', 'c', ('frcp(is_used_once)', 'a'), ('frcp(is_used_once)', 'b')) => ('frcp', ('bcsel', 'c', 'a', 'b')) */
   /* search2026_0 -> 521 in the cache */
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 5830 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 5832 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5831, 5833 },
      -1,
   } },

   /* replace2026_0_0 -> 521 in the cache */
   /* replace2026_0_1 -> 5830 in the cache */
   /* replace2026_0_2 -> 5832 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5830, 5832 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 5835 },
      -1,
   } },

   /* ('bcsel', 'c', ('frsq(is_used_once)', 'a'), ('frsq(is_used_once)', 'b')) => ('frsq', ('bcsel', 'c', 'a', 'b')) */
   /* search2027_0 -> 521 in the cache */
   /* search2027_1_0 -> 5830 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_frsq,
      -1, 0,
      { 5830 },
      0,
   } },
   /* search2027_2_0 -> 5832 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_frsq,
      -1, 0,
      { 5832 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5837, 5838 },
      -1,
   } },

   /* replace2027_0_0 -> 521 in the cache */
   /* replace2027_0_1 -> 5830 in the cache */
   /* replace2027_0_2 -> 5832 in the cache */
   /* replace2027_0 -> 5835 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_frsq,
      -1, 0,
      { 5835 },
      -1,
   } },

   /* ('bcsel', 'c', ('fsqrt(is_used_once)', 'a'), ('fsqrt(is_used_once)', 'b')) => ('fsqrt', ('bcsel', 'c', 'a', 'b')) */
   /* search2028_0 -> 521 in the cache */
   /* search2028_1_0 -> 5830 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsqrt,
      -1, 0,
      { 5830 },
      0,
   } },
   /* search2028_2_0 -> 5832 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsqrt,
      -1, 0,
      { 5832 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5841, 5842 },
      -1,
   } },

   /* replace2028_0_0 -> 521 in the cache */
   /* replace2028_0_1 -> 5830 in the cache */
   /* replace2028_0_2 -> 5832 in the cache */
   /* replace2028_0 -> 5835 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsqrt,
      -1, 0,
      { 5835 },
      -1,
   } },

   /* ('bcsel', 'c', ('fexp2(is_used_once)', 'a'), ('fexp2(is_used_once)', 'b')) => ('fexp2', ('bcsel', 'c', 'a', 'b')) */
   /* search2029_0 -> 521 in the cache */
   /* search2029_1_0 -> 5830 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 0,
      { 5830 },
      0,
   } },
   /* search2029_2_0 -> 5832 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 0,
      { 5832 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5845, 5846 },
      -1,
   } },

   /* replace2029_0_0 -> 521 in the cache */
   /* replace2029_0_1 -> 5830 in the cache */
   /* replace2029_0_2 -> 5832 in the cache */
   /* replace2029_0 -> 5835 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 0,
      { 5835 },
      -1,
   } },

   /* ('bcsel', 'c', ('flog2(is_used_once)', 'a'), ('flog2(is_used_once)', 'b')) => ('flog2', ('bcsel', 'c', 'a', 'b')) */
   /* search2030_0 -> 521 in the cache */
   /* search2030_1_0 -> 5830 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 5830 },
      0,
   } },
   /* search2030_2_0 -> 5832 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 5832 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5849, 5850 },
      -1,
   } },

   /* replace2030_0_0 -> 521 in the cache */
   /* replace2030_0_1 -> 5830 in the cache */
   /* replace2030_0_2 -> 5832 in the cache */
   /* replace2030_0 -> 5835 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 5835 },
      -1,
   } },

   /* ('bcsel', 'c', ('fsign(is_used_once)', 'a'), ('fsign(is_used_once)', 'b')) => ('fsign', ('bcsel', 'c', 'a', 'b')) */
   /* search2031_0 -> 521 in the cache */
   /* search2031_1_0 -> 5830 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 5830 },
      0,
   } },
   /* search2031_2_0 -> 5832 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 5832 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5853, 5854 },
      -1,
   } },

   /* replace2031_0_0 -> 521 in the cache */
   /* replace2031_0_1 -> 5830 in the cache */
   /* replace2031_0_2 -> 5832 in the cache */
   /* replace2031_0 -> 5835 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 5835 },
      -1,
   } },

   /* ('bcsel', 'c', ('fsin(is_used_once)', 'a'), ('fsin(is_used_once)', 'b')) => ('fsin', ('bcsel', 'c', 'a', 'b')) */
   /* search2032_0 -> 521 in the cache */
   /* search2032_1_0 -> 5830 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsin,
      -1, 0,
      { 5830 },
      0,
   } },
   /* search2032_2_0 -> 5832 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsin,
      -1, 0,
      { 5832 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5857, 5858 },
      -1,
   } },

   /* replace2032_0_0 -> 521 in the cache */
   /* replace2032_0_1 -> 5830 in the cache */
   /* replace2032_0_2 -> 5832 in the cache */
   /* replace2032_0 -> 5835 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsin,
      -1, 0,
      { 5835 },
      -1,
   } },

   /* ('bcsel', 'c', ('fcos(is_used_once)', 'a'), ('fcos(is_used_once)', 'b')) => ('fcos', ('bcsel', 'c', 'a', 'b')) */
   /* search2033_0 -> 521 in the cache */
   /* search2033_1_0 -> 5830 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fcos,
      -1, 0,
      { 5830 },
      0,
   } },
   /* search2033_2_0 -> 5832 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fcos,
      -1, 0,
      { 5832 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5861, 5862 },
      -1,
   } },

   /* replace2033_0_0 -> 521 in the cache */
   /* replace2033_0_1 -> 5830 in the cache */
   /* replace2033_0_2 -> 5832 in the cache */
   /* replace2033_0 -> 5835 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fcos,
      -1, 0,
      { 5835 },
      -1,
   } },

   /* ('bcsel', 'c', ('fsin_amd(is_used_once)', 'a'), ('fsin_amd(is_used_once)', 'b')) => ('fsin_amd', ('bcsel', 'c', 'a', 'b')) */
   /* search2034_0 -> 521 in the cache */
   /* search2034_1_0 -> 5830 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsin_amd,
      -1, 0,
      { 5830 },
      0,
   } },
   /* search2034_2_0 -> 5832 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsin_amd,
      -1, 0,
      { 5832 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5865, 5866 },
      -1,
   } },

   /* replace2034_0_0 -> 521 in the cache */
   /* replace2034_0_1 -> 5830 in the cache */
   /* replace2034_0_2 -> 5832 in the cache */
   /* replace2034_0 -> 5835 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsin_amd,
      -1, 0,
      { 5835 },
      -1,
   } },

   /* ('bcsel', 'c', ('fcos_amd(is_used_once)', 'a'), ('fcos_amd(is_used_once)', 'b')) => ('fcos_amd', ('bcsel', 'c', 'a', 'b')) */
   /* search2035_0 -> 521 in the cache */
   /* search2035_1_0 -> 5830 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fcos_amd,
      -1, 0,
      { 5830 },
      0,
   } },
   /* search2035_2_0 -> 5832 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fcos_amd,
      -1, 0,
      { 5832 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5869, 5870 },
      -1,
   } },

   /* replace2035_0_0 -> 521 in the cache */
   /* replace2035_0_1 -> 5830 in the cache */
   /* replace2035_0_2 -> 5832 in the cache */
   /* replace2035_0 -> 5835 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fcos_amd,
      -1, 0,
      { 5835 },
      -1,
   } },

   /* ('bcsel', 'c', ('fneg(is_used_once)', 'a'), ('fneg(is_used_once)', 'b')) => ('fneg', ('bcsel', 'c', 'a', 'b')) */
   /* search2036_0 -> 521 in the cache */
   /* search2036_1_0 -> 5830 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 5830 },
      0,
   } },
   /* search2036_2_0 -> 5832 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 5832 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5873, 5874 },
      -1,
   } },

   /* replace2036_0_0 -> 521 in the cache */
   /* replace2036_0_1 -> 5830 in the cache */
   /* replace2036_0_2 -> 5832 in the cache */
   /* replace2036_0 -> 5835 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 5835 },
      -1,
   } },

   /* ('bcsel', 'c', ('fabs(is_used_once)', 'a'), ('fabs(is_used_once)', 'b')) => ('fabs', ('bcsel', 'c', 'a', 'b')) */
   /* search2037_0 -> 521 in the cache */
   /* search2037_1_0 -> 5830 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 5830 },
      0,
   } },
   /* search2037_2_0 -> 5832 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 5832 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5877, 5878 },
      -1,
   } },

   /* replace2037_0_0 -> 521 in the cache */
   /* replace2037_0_1 -> 5830 in the cache */
   /* replace2037_0_2 -> 5832 in the cache */
   /* replace2037_0 -> 5835 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 5835 },
      -1,
   } },

   /* ('bcsel', 'c', ('fsign(is_used_once)', 'a'), ('fsign(is_used_once)', 'b')) => ('fsign', ('bcsel', 'c', 'a', 'b')) */
   /* search2038_0 -> 521 in the cache */
   /* search2038_1_0 -> 5830 in the cache */
   /* search2038_1 -> 5853 in the cache */
   /* search2038_2_0 -> 5832 in the cache */
   /* search2038_2 -> 5854 in the cache */
   /* search2038 -> 5855 in the cache */

   /* replace2038_0_0 -> 521 in the cache */
   /* replace2038_0_1 -> 5830 in the cache */
   /* replace2038_0_2 -> 5832 in the cache */
   /* replace2038_0 -> 5835 in the cache */
   /* replace2038 -> 5856 in the cache */

   /* ('ineg', ('bcsel', 'c', '#a', '#b')) => ('bcsel', 'c', ('ineg', 'a'), ('ineg', 'b')) */
   /* search2039_0_0 -> 521 in the cache */
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* a */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5881, 5882 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 5883 },
      -1,
   } },

   /* replace2039_0 -> 521 in the cache */
   /* replace2039_1_0 -> 5830 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 5830 },
      -1,
   } },
   /* replace2039_2_0 -> 5832 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 5832 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5885, 5886 },
      -1,
   } },

   /* ('iabs', ('bcsel', 'c', '#a', '#b')) => ('bcsel', 'c', ('iabs', 'a'), ('iabs', 'b')) */
   /* search2040_0_0 -> 521 in the cache */
   /* search2040_0_1 -> 5881 in the cache */
   /* search2040_0_2 -> 5882 in the cache */
   /* search2040_0 -> 5883 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 5883 },
      -1,
   } },

   /* replace2040_0 -> 521 in the cache */
   /* replace2040_1_0 -> 5830 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 5830 },
      -1,
   } },
   /* replace2040_2_0 -> 5832 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 5832 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5889, 5890 },
      -1,
   } },

   /* ('inot', ('bcsel', 'c', '#a', '#b')) => ('bcsel', 'c', ('inot', 'a'), ('inot', 'b')) */
   /* search2041_0_0 -> 521 in the cache */
   /* search2041_0_1 -> 5881 in the cache */
   /* search2041_0_2 -> 5882 in the cache */
   /* search2041_0 -> 5883 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 5883 },
      -1,
   } },

   /* replace2041_0 -> 521 in the cache */
   /* replace2041_1_0 -> 5830 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 5830 },
      -1,
   } },
   /* replace2041_2_0 -> 5832 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 5832 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5893, 5894 },
      -1,
   } },

   /* ('isign', ('bcsel', 'c', '#a', '#b')) => ('bcsel', 'c', ('isign', 'a'), ('isign', 'b')) */
   /* search2042_0_0 -> 521 in the cache */
   /* search2042_0_1 -> 5881 in the cache */
   /* search2042_0_2 -> 5882 in the cache */
   /* search2042_0 -> 5883 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_isign,
      -1, 0,
      { 5883 },
      -1,
   } },

   /* replace2042_0 -> 521 in the cache */
   /* replace2042_1_0 -> 5830 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_isign,
      -1, 0,
      { 5830 },
      -1,
   } },
   /* replace2042_2_0 -> 5832 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_isign,
      -1, 0,
      { 5832 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 521, 5897, 5898 },
      -1,
   } },

   /* ('fisnormal', 'a@16') => ('ult', 4095, ('iadd', ('ishl', 'a', 1), 2048)) */
   /* search2043_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fisnormal,
      -1, 0,
      { 11 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { 0xfff /* 4095 */ },
   } },
   /* replace2043_1_0_0 -> 11 in the cache */
   /* replace2043_1_0_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 11, 396 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { 0x800 /* 2048 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 5902, 5903 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 5901, 5904 },
      -1,
   } },

   /* ('fisnormal', 'a@32') => ('ult', 33554431, ('iadd', ('ishl', 'a', 1), 16777216)) */
   /* search2044_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fisnormal,
      -1, 0,
      { 16 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x1ffffff /* 33554431 */ },
   } },
   /* replace2044_1_0_0 -> 16 in the cache */
   /* replace2044_1_0_1 -> 396 in the cache */
   /* replace2044_1_0 -> 991 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x1000000 /* 16777216 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 991, 5908 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 5907, 5909 },
      -1,
   } },

   /* ('fisnormal', 'a@64') => ('ult', 18014398509481983, ('iadd', ('ishl', 'a', 1), 9007199254740992)) */
   /* search2045_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fisnormal,
      -1, 0,
      { 21 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x3fffffffffffff /* 18014398509481983 */ },
   } },
   /* replace2045_1_0_0 -> 21 in the cache */
   /* replace2045_1_0_1 -> 396 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 396 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x20000000000000 /* 9007199254740992 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 5913, 5914 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 5912, 5915 },
      -1,
   } },

   /* ('f2f16', ('vec2', 'a@32', 'b@32')) => ('vec2', ('f2f16', 'a'), ('f2f16', 'b')) */
   /* search2046_0_0 -> 16 in the cache */
   /* search2046_0_1 -> 36 in the cache */
   /* search2046_0 -> 3215 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 3215 },
      -1,
   } },

   /* replace2046_0_0 -> 16 in the cache */
   /* replace2046_0 -> 3992 in the cache */
   /* replace2046_1_0 -> 36 in the cache */
   /* replace2046_1 -> 3989 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 3992, 3989 },
      -1,
   } },

   /* ('f2fmp', ('vec2', 'a@32', 'b@32')) => ('vec2', ('f2fmp', 'a'), ('f2fmp', 'b')) */
   /* search2047_0_0 -> 16 in the cache */
   /* search2047_0_1 -> 36 in the cache */
   /* search2047_0 -> 3215 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 3215 },
      -1,
   } },

   /* replace2047_0_0 -> 16 in the cache */
   /* replace2047_0 -> 1993 in the cache */
   /* replace2047_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 1993, 5920 },
      -1,
   } },

   /* ('u2u16', ('vec2', 'a@32', 'b@32')) => ('vec2', ('u2u16', 'a'), ('u2u16', 'b')) */
   /* search2048_0_0 -> 16 in the cache */
   /* search2048_0_1 -> 36 in the cache */
   /* search2048_0 -> 3215 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 3215 },
      -1,
   } },

   /* replace2048_0_0 -> 16 in the cache */
   /* replace2048_0 -> 3310 in the cache */
   /* replace2048_1_0 -> 36 in the cache */
   /* replace2048_1 -> 4867 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 3310, 4867 },
      -1,
   } },

   /* ('i2i16', ('vec2', 'a@32', 'b@32')) => ('vec2', ('i2i16', 'a'), ('i2i16', 'b')) */
   /* search2049_0_0 -> 16 in the cache */
   /* search2049_0_1 -> 36 in the cache */
   /* search2049_0 -> 3215 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 3215 },
      -1,
   } },

   /* replace2049_0_0 -> 16 in the cache */
   /* replace2049_0 -> 4353 in the cache */
   /* replace2049_1_0 -> 36 in the cache */
   /* replace2049_1 -> 4538 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 4353, 4538 },
      -1,
   } },

   /* ('i2imp', ('vec2', 'a@32', 'b@32')) => ('vec2', ('i2imp', 'a'), ('i2imp', 'b')) */
   /* search2050_0_0 -> 16 in the cache */
   /* search2050_0_1 -> 36 in the cache */
   /* search2050_0 -> 3215 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 3215 },
      -1,
   } },

   /* replace2050_0_0 -> 16 in the cache */
   /* replace2050_0 -> 2004 in the cache */
   /* replace2050_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 2004, 5927 },
      -1,
   } },

   /* ('f2f16', ('vec3', 'a@32', 'b@32', 'c@32')) => ('vec3', ('f2f16', 'a'), ('f2f16', 'b'), ('f2f16', 'c')) */
   /* search2051_0_0 -> 16 in the cache */
   /* search2051_0_1 -> 36 in the cache */
   /* search2051_0_2 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec3,
      -1, 0,
      { 16, 36, 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 5929 },
      -1,
   } },

   /* replace2051_0_0 -> 16 in the cache */
   /* replace2051_0 -> 3992 in the cache */
   /* replace2051_1_0 -> 36 in the cache */
   /* replace2051_1 -> 3989 in the cache */
   /* replace2051_2_0 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec3,
      -1, 0,
      { 3992, 3989, 5931 },
      -1,
   } },

   /* ('f2fmp', ('vec3', 'a@32', 'b@32', 'c@32')) => ('vec3', ('f2fmp', 'a'), ('f2fmp', 'b'), ('f2fmp', 'c')) */
   /* search2052_0_0 -> 16 in the cache */
   /* search2052_0_1 -> 36 in the cache */
   /* search2052_0_2 -> 135 in the cache */
   /* search2052_0 -> 5929 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 5929 },
      -1,
   } },

   /* replace2052_0_0 -> 16 in the cache */
   /* replace2052_0 -> 1993 in the cache */
   /* replace2052_1_0 -> 36 in the cache */
   /* replace2052_1 -> 5920 in the cache */
   /* replace2052_2_0 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec3,
      -1, 0,
      { 1993, 5920, 5934 },
      -1,
   } },

   /* ('u2u16', ('vec3', 'a@32', 'b@32', 'c@32')) => ('vec3', ('u2u16', 'a'), ('u2u16', 'b'), ('u2u16', 'c')) */
   /* search2053_0_0 -> 16 in the cache */
   /* search2053_0_1 -> 36 in the cache */
   /* search2053_0_2 -> 135 in the cache */
   /* search2053_0 -> 5929 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 5929 },
      -1,
   } },

   /* replace2053_0_0 -> 16 in the cache */
   /* replace2053_0 -> 3310 in the cache */
   /* replace2053_1_0 -> 36 in the cache */
   /* replace2053_1 -> 4867 in the cache */
   /* replace2053_2_0 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec3,
      -1, 0,
      { 3310, 4867, 5937 },
      -1,
   } },

   /* ('i2i16', ('vec3', 'a@32', 'b@32', 'c@32')) => ('vec3', ('i2i16', 'a'), ('i2i16', 'b'), ('i2i16', 'c')) */
   /* search2054_0_0 -> 16 in the cache */
   /* search2054_0_1 -> 36 in the cache */
   /* search2054_0_2 -> 135 in the cache */
   /* search2054_0 -> 5929 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 5929 },
      -1,
   } },

   /* replace2054_0_0 -> 16 in the cache */
   /* replace2054_0 -> 4353 in the cache */
   /* replace2054_1_0 -> 36 in the cache */
   /* replace2054_1 -> 4538 in the cache */
   /* replace2054_2_0 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec3,
      -1, 0,
      { 4353, 4538, 5940 },
      -1,
   } },

   /* ('i2imp', ('vec3', 'a@32', 'b@32', 'c@32')) => ('vec3', ('i2imp', 'a'), ('i2imp', 'b'), ('i2imp', 'c')) */
   /* search2055_0_0 -> 16 in the cache */
   /* search2055_0_1 -> 36 in the cache */
   /* search2055_0_2 -> 135 in the cache */
   /* search2055_0 -> 5929 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 5929 },
      -1,
   } },

   /* replace2055_0_0 -> 16 in the cache */
   /* replace2055_0 -> 2004 in the cache */
   /* replace2055_1_0 -> 36 in the cache */
   /* replace2055_1 -> 5927 in the cache */
   /* replace2055_2_0 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec3,
      -1, 0,
      { 2004, 5927, 5943 },
      -1,
   } },

   /* ('f2f16', ('vec4', 'a@32', 'b@32', 'c@32', 'd@32')) => ('vec4', ('f2f16', 'a'), ('f2f16', 'b'), ('f2f16', 'c'), ('f2f16', 'd')) */
   /* search2056_0_0 -> 16 in the cache */
   /* search2056_0_1 -> 36 in the cache */
   /* search2056_0_2 -> 135 in the cache */
   /* search2056_0_3 -> 163 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 16, 36, 135, 163 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 5945 },
      -1,
   } },

   /* replace2056_0_0 -> 16 in the cache */
   /* replace2056_0 -> 3992 in the cache */
   /* replace2056_1_0 -> 36 in the cache */
   /* replace2056_1 -> 3989 in the cache */
   /* replace2056_2_0 -> 135 in the cache */
   /* replace2056_2 -> 5931 in the cache */
   /* replace2056_3_0 -> 163 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 163 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 3992, 3989, 5931, 5947 },
      -1,
   } },

   /* ('f2fmp', ('vec4', 'a@32', 'b@32', 'c@32', 'd@32')) => ('vec4', ('f2fmp', 'a'), ('f2fmp', 'b'), ('f2fmp', 'c'), ('f2fmp', 'd')) */
   /* search2057_0_0 -> 16 in the cache */
   /* search2057_0_1 -> 36 in the cache */
   /* search2057_0_2 -> 135 in the cache */
   /* search2057_0_3 -> 163 in the cache */
   /* search2057_0 -> 5945 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 5945 },
      -1,
   } },

   /* replace2057_0_0 -> 16 in the cache */
   /* replace2057_0 -> 1993 in the cache */
   /* replace2057_1_0 -> 36 in the cache */
   /* replace2057_1 -> 5920 in the cache */
   /* replace2057_2_0 -> 135 in the cache */
   /* replace2057_2 -> 5934 in the cache */
   /* replace2057_3_0 -> 163 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 163 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 1993, 5920, 5934, 5950 },
      -1,
   } },

   /* ('u2u16', ('vec4', 'a@32', 'b@32', 'c@32', 'd@32')) => ('vec4', ('u2u16', 'a'), ('u2u16', 'b'), ('u2u16', 'c'), ('u2u16', 'd')) */
   /* search2058_0_0 -> 16 in the cache */
   /* search2058_0_1 -> 36 in the cache */
   /* search2058_0_2 -> 135 in the cache */
   /* search2058_0_3 -> 163 in the cache */
   /* search2058_0 -> 5945 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 5945 },
      -1,
   } },

   /* replace2058_0_0 -> 16 in the cache */
   /* replace2058_0 -> 3310 in the cache */
   /* replace2058_1_0 -> 36 in the cache */
   /* replace2058_1 -> 4867 in the cache */
   /* replace2058_2_0 -> 135 in the cache */
   /* replace2058_2 -> 5937 in the cache */
   /* replace2058_3_0 -> 163 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 163 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 3310, 4867, 5937, 5953 },
      -1,
   } },

   /* ('i2i16', ('vec4', 'a@32', 'b@32', 'c@32', 'd@32')) => ('vec4', ('i2i16', 'a'), ('i2i16', 'b'), ('i2i16', 'c'), ('i2i16', 'd')) */
   /* search2059_0_0 -> 16 in the cache */
   /* search2059_0_1 -> 36 in the cache */
   /* search2059_0_2 -> 135 in the cache */
   /* search2059_0_3 -> 163 in the cache */
   /* search2059_0 -> 5945 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 5945 },
      -1,
   } },

   /* replace2059_0_0 -> 16 in the cache */
   /* replace2059_0 -> 4353 in the cache */
   /* replace2059_1_0 -> 36 in the cache */
   /* replace2059_1 -> 4538 in the cache */
   /* replace2059_2_0 -> 135 in the cache */
   /* replace2059_2 -> 5940 in the cache */
   /* replace2059_3_0 -> 163 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 163 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 4353, 4538, 5940, 5956 },
      -1,
   } },

   /* ('i2imp', ('vec4', 'a@32', 'b@32', 'c@32', 'd@32')) => ('vec4', ('i2imp', 'a'), ('i2imp', 'b'), ('i2imp', 'c'), ('i2imp', 'd')) */
   /* search2060_0_0 -> 16 in the cache */
   /* search2060_0_1 -> 36 in the cache */
   /* search2060_0_2 -> 135 in the cache */
   /* search2060_0_3 -> 163 in the cache */
   /* search2060_0 -> 5945 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 5945 },
      -1,
   } },

   /* replace2060_0_0 -> 16 in the cache */
   /* replace2060_0 -> 2004 in the cache */
   /* replace2060_1_0 -> 36 in the cache */
   /* replace2060_1 -> 5927 in the cache */
   /* replace2060_2_0 -> 135 in the cache */
   /* replace2060_2 -> 5943 in the cache */
   /* replace2060_3_0 -> 163 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 163 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 2004, 5927, 5943, 5959 },
      -1,
   } },

};

static const nir_search_expression_cond nir_opt_algebraic_expression_cond[] = {
   (is_used_once),
   (is_only_used_as_float),
   (is_not_used_by_if),
   (is_used_by_non_fsat),
   (no_signed_wrap),
   (only_lower_8_bits_used),
   (only_lower_16_bits_used),
};

static const nir_search_variable_cond nir_opt_algebraic_variable_cond[] = {
   (is_pos_power_of_two),
   (is_neg_power_of_two),
   (is_bitcount2),
   (is_finite_not_zero),
   (is_finite),
   (is_not_const),
   (is_not_const_zero),
   (is_ult_0xfffc07fc),
   (is_not_fmul),
   (is_first_5_bits_uge_2),
   (is_unsigned_multiple_of_4),
   (is_unsigned_multiple_of_2),
   (is_unsigned_multiple_of_8),
   (is_unsigned_multiple_of_16),
   (is_unsigned_multiple_of_32),
   (is_unsigned_multiple_of_64),
   (is_a_number),
   (is_gt_0_and_lt_1),
   (is_zero_to_one),
   (is_not_negative),
   (is_lower_half_zero),
   (is_upper_half_zero),
   (is_lower_half_negative_one),
   (is_upper_half_negative_one),
   (is_integral),
   (is_not_positive),
   (is_not_zero),
   (is_a_number_not_negative),
   (is_a_number_not_positive),
   (is_gt_zero),
   (is_lt_zero),
   (is_a_number_gt_zero),
   (is_a_number_lt_zero),
   (is_fmul),
   (is_ult_32),
};

static const struct transform nir_opt_algebraic_transforms[] = {
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 238, 61, 0 },
   { 246, 32, 0 },
   { 249, 100, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 2765, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2764, 1229, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 2764, 1229, 0 },
   { 2765, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2732, 3307, 63 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 52, 54, 4 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 55, 57, 4 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 58, 32, 0 },
   { 67, 68, 1 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 59, 32, 0 },
   { 69, 73, 1 },
   { 74, 77, 1 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 60, 61, 0 },
   { 78, 81, 1 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 62, 61, 0 },
   { 64, 61, 0 },
   { 82, 81, 1 },
   { 83, 88, 1 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 61, 0 },
   { 66, 61, 0 },
   { 89, 95, 1 },
   { 96, 97, 1 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3153, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3154, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 32, 0 },
   { 114, 11, 5 },
   { 116, 16, 6 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 117, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3752, 3765, 79 },
   { 4134, 4142, 114 },
   { 4151, 4159, 114 },
   { 4168, 4175, 114 },
   { 3752, 4190, 114 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 118, 32, 0 },
   { 3752, 3765, 79 },
   { 4134, 4142, 114 },
   { 4151, 4159, 114 },
   { 4168, 4175, 114 },
   { 3752, 4190, 114 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3766, 3774, 79 },
   { 4143, 4150, 114 },
   { 4160, 4167, 114 },
   { 4176, 4183, 114 },
   { 3766, 4195, 114 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 119, 32, 0 },
   { 3766, 3774, 79 },
   { 4143, 4150, 114 },
   { 4160, 4167, 114 },
   { 4176, 4183, 114 },
   { 3766, 4195, 114 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3738, 3742, 75 },
   { 3743, 3745, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 120, 32, 0 },
   { 3738, 3742, 75 },
   { 3743, 3745, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3746, 3747, 77 },
   { 3748, 3751, 78 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 121, 32, 0 },
   { 3746, 3747, 77 },
   { 3748, 3751, 78 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 123, 16, 0 },
   { 125, 124, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 221, 111, 0 },
   { 222, 113, 5 },
   { 223, 115, 6 },
   { 242, 32, 0 },
   { 245, 32, 0 },
   { 248, 98, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 226, 227, 6 },
   { 230, 227, 0 },
   { 231, 232, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 224, 115, 0 },
   { 226, 227, 6 },
   { 230, 227, 0 },
   { 231, 232, 0 },
   { 244, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 263, 264, 0 },
   { 265, 266, 5 },
   { 267, 227, 6 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 257, 3, 0 },
   { 259, 258, 5 },
   { 260, 36, 6 },
   { 271, 272, 0 },
   { 276, 277, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 257, 3, 0 },
   { 259, 258, 5 },
   { 260, 36, 6 },
   { 263, 264, 0 },
   { 265, 266, 5 },
   { 267, 227, 6 },
   { 271, 272, 0 },
   { 276, 277, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 257, 3, 0 },
   { 259, 258, 5 },
   { 260, 36, 6 },
   { 271, 272, 0 },
   { 276, 277, 0 },
   { 284, 285, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 257, 3, 0 },
   { 259, 258, 5 },
   { 260, 36, 6 },
   { 263, 264, 0 },
   { 265, 266, 5 },
   { 267, 227, 6 },
   { 271, 272, 0 },
   { 276, 277, 0 },
   { 284, 285, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 233, 234, 6 },
   { 235, 234, 0 },
   { 236, 237, 0 },
   { 387, 288, 30 },
   { 748, 288, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 233, 234, 6 },
   { 235, 234, 0 },
   { 236, 237, 0 },
   { 268, 269, 6 },
   { 387, 288, 30 },
   { 748, 288, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 233, 234, 6 },
   { 235, 234, 0 },
   { 236, 237, 0 },
   { 261, 262, 0 },
   { 273, 274, 6 },
   { 279, 281, 6 },
   { 387, 288, 30 },
   { 748, 288, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 233, 234, 6 },
   { 235, 234, 0 },
   { 236, 237, 0 },
   { 261, 262, 0 },
   { 268, 269, 6 },
   { 273, 274, 6 },
   { 279, 281, 6 },
   { 387, 288, 30 },
   { 748, 288, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 233, 234, 6 },
   { 235, 234, 0 },
   { 236, 237, 0 },
   { 261, 262, 0 },
   { 273, 274, 6 },
   { 279, 281, 6 },
   { 287, 288, 0 },
   { 387, 288, 30 },
   { 748, 288, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 233, 234, 6 },
   { 235, 234, 0 },
   { 236, 237, 0 },
   { 261, 262, 0 },
   { 268, 269, 6 },
   { 273, 274, 6 },
   { 279, 281, 6 },
   { 287, 288, 0 },
   { 387, 288, 30 },
   { 748, 288, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2683, 2684, 0 },
   { 2686, 2687, 0 },
   { 2700, 32, 0 },
   { 2701, 61, 0 },
   { 2702, 32, 0 },
   { 2725, 2729, 0 },
   { 2731, 2735, 0 },
   { 2737, 2738, 0 },
   { 2740, 2741, 0 },
   { 3245, 3246, 58 },
   { 3276, 3277, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2706, 32, 0 },
   { 2707, 63, 0 },
   { 2742, 2744, 0 },
   { 2745, 2747, 0 },
   { 2748, 2749, 0 },
   { 2750, 2751, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2691, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2691, 1235, 0 },
   { 3040, 760, 0 },
   { 3043, 1131, 0 },
   { 3192, 1077, 0 },
   { 3792, 3793, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2692, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2692, 1077, 0 },
   { 3041, 1131, 0 },
   { 3042, 760, 0 },
   { 3191, 1235, 0 },
   { 3789, 3791, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 217, 1517, 40 },
   { 3156, 111, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 239, 122, 0 },
   { 240, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 250, 4005, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 291, 0, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 289, 0, 0 },
   { 290, 3, 0 },
   { 291, 0, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 291, 0, 0 },
   { 685, 687, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 289, 0, 0 },
   { 290, 3, 0 },
   { 291, 0, 0 },
   { 685, 687, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 291, 0, 0 },
   { 292, 264, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 289, 0, 0 },
   { 290, 3, 0 },
   { 291, 0, 0 },
   { 292, 264, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 291, 0, 0 },
   { 292, 264, 0 },
   { 685, 687, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 289, 0, 0 },
   { 290, 3, 0 },
   { 291, 0, 0 },
   { 292, 264, 0 },
   { 685, 687, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 331, 411, 10 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 331, 411, 10 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 297, 36, 0 },
   { 331, 411, 10 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 297, 36, 0 },
   { 331, 411, 10 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 297, 36, 0 },
   { 348, 350, 0 },
   { 331, 411, 10 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 297, 36, 0 },
   { 348, 350, 0 },
   { 331, 411, 10 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 334, 427, 11 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 334, 427, 11 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 298, 36, 0 },
   { 334, 427, 11 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 298, 36, 0 },
   { 334, 427, 11 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 298, 36, 0 },
   { 351, 353, 0 },
   { 334, 427, 11 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 298, 36, 0 },
   { 351, 353, 0 },
   { 334, 427, 11 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 456, 457, 10 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 299, 36, 0 },
   { 456, 457, 10 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 299, 36, 0 },
   { 363, 364, 7 },
   { 456, 457, 10 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 454, 455, 11 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 453, 334, 0 },
   { 454, 455, 11 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 300, 36, 0 },
   { 454, 455, 11 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 300, 36, 0 },
   { 453, 334, 0 },
   { 454, 455, 11 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 300, 36, 0 },
   { 365, 366, 8 },
   { 454, 455, 11 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 300, 36, 0 },
   { 365, 366, 8 },
   { 453, 334, 0 },
   { 454, 455, 11 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 343, 443, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 343, 443, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 301, 36, 0 },
   { 343, 443, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 301, 36, 0 },
   { 343, 443, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 301, 36, 0 },
   { 357, 359, 0 },
   { 343, 443, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 301, 36, 0 },
   { 357, 359, 0 },
   { 343, 443, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 346, 451, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 346, 451, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 302, 36, 0 },
   { 346, 451, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 302, 36, 0 },
   { 346, 451, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 302, 36, 0 },
   { 360, 362, 0 },
   { 346, 451, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 302, 36, 0 },
   { 360, 362, 0 },
   { 346, 451, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 462, 463, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 303, 36, 0 },
   { 462, 463, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 303, 36, 0 },
   { 369, 370, 7 },
   { 462, 463, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 460, 461, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 304, 36, 0 },
   { 460, 461, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 304, 36, 0 },
   { 371, 372, 8 },
   { 460, 461, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 337, 435, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 337, 435, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 305, 36, 0 },
   { 337, 435, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 305, 36, 0 },
   { 337, 435, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 306, 36, 0 },
   { 337, 435, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 306, 36, 0 },
   { 337, 435, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 305, 36, 0 },
   { 306, 36, 0 },
   { 354, 356, 0 },
   { 337, 435, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 305, 36, 0 },
   { 306, 36, 0 },
   { 354, 356, 0 },
   { 337, 435, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 458, 459, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 307, 36, 0 },
   { 458, 459, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 308, 36, 0 },
   { 458, 459, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 307, 36, 0 },
   { 308, 36, 0 },
   { 367, 368, 7 },
   { 458, 459, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3044, 760, 0 },
   { 3045, 1131, 0 },
   { 3048, 3049, 0 },
   { 3050, 3051, 0 },
   { 3054, 3055, 0 },
   { 3056, 3057, 0 },
   { 3058, 3059, 0 },
   { 3060, 3061, 0 },
   { 3063, 3064, 0 },
   { 3065, 3066, 0 },
   { 3067, 3068, 57 },
   { 3069, 3070, 57 },
   { 3073, 3074, 57 },
   { 3075, 3076, 57 },
   { 3077, 3, 0 },
   { 3079, 2248, 0 },
   { 3080, 3082, 0 },
   { 3084, 1228, 0 },
   { 3085, 3081, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3044, 760, 0 },
   { 3045, 1131, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3048, 3049, 0 },
   { 3050, 3051, 0 },
   { 3054, 3055, 0 },
   { 3056, 3057, 0 },
   { 3058, 3059, 0 },
   { 3060, 3061, 0 },
   { 3063, 3064, 0 },
   { 3065, 3066, 0 },
   { 3067, 3068, 57 },
   { 3069, 3070, 57 },
   { 3073, 3074, 57 },
   { 3075, 3076, 57 },
   { 3077, 3, 0 },
   { 3079, 2248, 0 },
   { 3080, 3082, 0 },
   { 3084, 1228, 0 },
   { 3085, 3081, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2698, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2698, 1235, 0 },
   { 3173, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 688, 693, 20 },
   { 694, 699, 20 },
   { 700, 706, 21 },
   { 3149, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 707, 709, 22 },
   { 710, 712, 22 },
   { 713, 715, 23 },
   { 3147, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 708, 731, 26 },
   { 711, 732, 26 },
   { 714, 733, 27 },
   { 3151, 111, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 734, 736, 28 },
   { 3148, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 768, 773, 35 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 804, 805, 37 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 802, 803, 37 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 800, 801, 37 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 796, 799, 0 },
   { 800, 801, 37 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 806, 799, 37 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2771, 32, 0 },
   { 3234, 3235, 58 },
   { 1017, 422, 58 },
   { 3237, 3239, 58 },
   { 1008, 447, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2770, 1229, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2770, 1229, 0 },
   { 2771, 32, 0 },
   { 3234, 3235, 58 },
   { 1017, 422, 58 },
   { 3237, 3239, 58 },
   { 1008, 447, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2769, 32, 0 },
   { 3240, 3241, 58 },
   { 3242, 406, 58 },
   { 3243, 3244, 58 },
   { 3274, 439, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2766, 1229, 0 },
   { 2768, 2767, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2766, 1229, 0 },
   { 2768, 2767, 0 },
   { 2769, 32, 0 },
   { 3240, 3241, 58 },
   { 3242, 406, 58 },
   { 3243, 3244, 58 },
   { 3274, 439, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2699, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2699, 1077, 0 },
   { 3172, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2689, 1077, 0 },
   { 3196, 1077, 0 },
   { 3197, 1235, 0 },
   { 3198, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2689, 1077, 0 },
   { 3196, 1077, 0 },
   { 3197, 1235, 0 },
   { 3198, 1235, 0 },
   { 1489, 3787, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2693, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2693, 1077, 0 },
   { 3201, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2693, 1077, 0 },
   { 3200, 1235, 0 },
   { 3794, 3791, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2693, 1077, 0 },
   { 3200, 1235, 0 },
   { 3201, 1077, 0 },
   { 3794, 3791, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2690, 1235, 0 },
   { 3193, 1235, 0 },
   { 3194, 1077, 0 },
   { 3195, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2690, 1235, 0 },
   { 3193, 1235, 0 },
   { 3194, 1077, 0 },
   { 3195, 1077, 0 },
   { 1486, 3788, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2694, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2694, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 3157, 217, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1338, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1338, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { 1340, 61, 0 },
   { 1341, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1342, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1342, 32, 0 },
   { 1343, 32, 0 },
   { 1344, 63, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3916, 3919, 91 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3916, 3919, 91 },
   { 5078, 11, 0 },
   { 5131, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4241, 11, 0 },
   { 4242, 16, 0 },
   { 4243, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4231, 760, 0 },
   { 4232, 6, 0 },
   { 4233, 11, 0 },
   { 4234, 16, 0 },
   { 4235, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4236, 760, 0 },
   { 4237, 6, 0 },
   { 4238, 11, 0 },
   { 4239, 16, 0 },
   { 4240, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3911, 3915, 91 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3911, 3915, 91 },
   { 5106, 11, 0 },
   { 5159, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3920, 3924, 92 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3920, 3924, 92 },
   { 5211, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3925, 3928, 92 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3925, 3928, 92 },
   { 5184, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 2710, 32, 0 },
   { 2713, 2703, 0 },
   { 2752, 2754, 0 },
   { 2755, 2757, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2396, 2404, 47 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2398, 2405, 47 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2400, 2406, 47 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2402, 2407, 47 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2426, 2433, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2434, 2436, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2437, 2439, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2440, 2442, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2443, 2445, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2446, 2449, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2450, 2454, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2455, 2457, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2458, 2460, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2461, 2463, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2464, 2469, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2470, 2476, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2477, 2483, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 2500, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2501, 2533, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2534, 2537, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2538, 2543, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2544, 2548, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2549, 2557, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2558, 2574, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2575, 2578, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2579, 2584, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2585, 2589, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2590, 2598, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2599, 2615, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2616, 2619, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2620, 2625, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2626, 2630, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2631, 2639, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2640, 2656, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3461, 3462, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3461, 3462, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2797, 2800, 50 },
   { 2776, 2779, 50 },
   { 2783, 2786, 50 },
   { 2801, 2804, 50 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2805, 2807, 50 },
   { 2789, 2791, 50 },
   { 2794, 2796, 50 },
   { 2808, 2810, 50 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2815, 2818, 51 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2815, 2818, 51 },
   { 2840, 32, 0 },
   { 2841, 251, 0 },
   { 2842, 2839, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2815, 2818, 51 },
   { 2843, 2813, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2815, 2818, 51 },
   { 2840, 32, 0 },
   { 2841, 251, 0 },
   { 2842, 2839, 0 },
   { 2843, 2813, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2832, 2885, 54 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2880, 2882, 53 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2879, 2861, 0 },
   { 2880, 2882, 53 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2887, 2969, 56 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2970, 3039, 56 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3078, 36, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3150, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3202, 3302, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2726, 3309, 63 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3226, 3310, 64 },
   { 3226, 3310, 101 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3228, 3311, 64 },
   { 3228, 3311, 101 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3855, 122, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3262, 412, 58 },
   { 3263, 415, 58 },
   { 3264, 419, 58 },
   { 3265, 422, 58 },
   { 3278, 444, 59 },
   { 3279, 447, 59 },
   { 3855, 122, 0 },
   { 3857, 3859, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3856, 122, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3266, 393, 58 },
   { 3267, 397, 58 },
   { 3268, 402, 58 },
   { 3269, 406, 58 },
   { 3280, 436, 59 },
   { 3281, 439, 59 },
   { 3856, 122, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3303, 3306, 62 },
   { 3303, 3997, 101 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3313, 3995, 101 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3320, 3996, 101 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3348, 3994, 101 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3348, 3994, 101 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3348, 3994, 101 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3348, 3994, 101 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3457, 16, 0 },
   { 3458, 122, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3459, 3460, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3463, 3465, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3466, 3467, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3907, 3910, 90 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3890, 3894, 87 },
   { 3890, 3898, 88 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3900, 3904, 88 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3682, 3685, 68 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3686, 3689, 68 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3690, 3693, 69 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3694, 3695, 70 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3700, 3704, 71 },
   { 3700, 3803, 82 },
   { 3700, 3805, 83 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3705, 3708, 72 },
   { 3718, 3722, 73 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3709, 3711, 72 },
   { 3723, 3725, 73 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3712, 3714, 72 },
   { 3726, 3730, 73 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3715, 3717, 72 },
   { 3731, 3733, 73 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3734, 3735, 74 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3736, 3737, 74 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3809, 3812, 84 },
   { 3809, 3883, 86 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3813, 3815, 84 },
   { 3813, 3889, 86 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3816, 2719, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3905, 3906, 89 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3930, 3936, 93 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3937, 3942, 94 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3943, 3950, 95 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3951, 3956, 96 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3957, 3962, 97 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3963, 3970, 98 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3971, 3978, 99 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3979, 3988, 100 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3998, 4000, 102 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4006, 951, 105 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4007, 4011, 106 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4012, 4016, 107 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4017, 4018, 108 },
   { 4017, 49, 105 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4019, 4012, 109 },
   { 4019, 4021, 107 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4022, 4007, 110 },
   { 4022, 49, 106 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4023, 4018, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4024, 555, 0 },
   { 4025, 4026, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4023, 4018, 0 },
   { 4024, 555, 0 },
   { 4025, 4026, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4024, 555, 0 },
   { 4025, 4026, 0 },
   { 4027, 4029, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4023, 4018, 0 },
   { 4024, 555, 0 },
   { 4025, 4026, 0 },
   { 4027, 4029, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4030, 4031, 1 },
   { 4033, 4037, 1 },
   { 4038, 122, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4244, 760, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4245, 6, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4246, 11, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4247, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5237, 5257, 117 },
   { 5259, 5277, 117 },
   { 5279, 5300, 117 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5611, 111, 119 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5612, 111, 119 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5613, 111, 119 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5614, 111, 119 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5615, 111, 119 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5616, 111, 119 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5618, 122, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5621, 135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5619, 135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5619, 135, 0 },
   { 5621, 135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5900, 5905, 120 },
   { 5906, 5910, 120 },
   { 5911, 5916, 120 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3476, 3477, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5095, 5093, 0 },
   { 5148, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5120, 5093, 0 },
   { 5173, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5225, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5200, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 238, 61, 0 },
   { 246, 32, 0 },
   { 249, 100, 0 },
   { 3493, 3494, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 238, 61, 0 },
   { 246, 32, 0 },
   { 249, 100, 0 },
   { 3493, 3494, 0 },
   { 3537, 3494, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 238, 61, 0 },
   { 246, 32, 0 },
   { 249, 100, 0 },
   { 3476, 3477, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 238, 61, 0 },
   { 246, 32, 0 },
   { 249, 100, 0 },
   { 1024, 1027, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 238, 61, 0 },
   { 246, 32, 0 },
   { 249, 100, 0 },
   { 5607, 5610, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 238, 61, 0 },
   { 246, 32, 0 },
   { 249, 100, 0 },
   { 5095, 5093, 0 },
   { 5148, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 238, 61, 0 },
   { 246, 32, 0 },
   { 249, 100, 0 },
   { 5120, 5093, 0 },
   { 5173, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 238, 61, 0 },
   { 246, 32, 0 },
   { 249, 100, 0 },
   { 5225, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 238, 61, 0 },
   { 246, 32, 0 },
   { 249, 100, 0 },
   { 5200, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3476, 3477, 0 },
   { 5095, 5093, 0 },
   { 5148, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3476, 3477, 0 },
   { 5120, 5093, 0 },
   { 5173, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3476, 3477, 0 },
   { 5225, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3476, 3477, 0 },
   { 5200, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2675, 2676, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5095, 5093, 0 },
   { 5120, 5093, 0 },
   { 5148, 49, 0 },
   { 5173, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5095, 5093, 0 },
   { 5148, 49, 0 },
   { 5225, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5095, 5093, 0 },
   { 5148, 49, 0 },
   { 5200, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5120, 5093, 0 },
   { 5173, 49, 0 },
   { 5225, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5120, 5093, 0 },
   { 5173, 49, 0 },
   { 5200, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5200, 49, 0 },
   { 5225, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3583, 0, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 59, 32, 0 },
   { 69, 73, 1 },
   { 74, 77, 1 },
   { 3583, 0, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 206, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 330, 331, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 333, 334, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 342, 343, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 345, 346, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 336, 337, 0 },
   { 340, 337, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5091, 4052, 0 },
   { 5145, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5117, 4052, 0 },
   { 5170, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5222, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5197, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 117, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 117, 32, 0 },
   { 3851, 3852, 85 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 117, 32, 0 },
   { 206, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 117, 32, 0 },
   { 3523, 3524, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 117, 32, 0 },
   { 3523, 3524, 0 },
   { 3565, 3524, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 117, 32, 0 },
   { 312, 314, 0 },
   { 330, 331, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 117, 32, 0 },
   { 316, 317, 0 },
   { 333, 334, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 117, 32, 0 },
   { 323, 324, 0 },
   { 342, 343, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 117, 32, 0 },
   { 326, 327, 0 },
   { 345, 346, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 117, 32, 0 },
   { 319, 321, 0 },
   { 336, 337, 0 },
   { 340, 337, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 117, 32, 0 },
   { 5603, 5606, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 117, 32, 0 },
   { 5091, 4052, 0 },
   { 5145, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 117, 32, 0 },
   { 5117, 4052, 0 },
   { 5170, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 117, 32, 0 },
   { 5222, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 117, 32, 0 },
   { 5197, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 179, 181, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 186, 188, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 206, 61, 0 },
   { 209, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 206, 61, 0 },
   { 209, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 206, 61, 0 },
   { 330, 331, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 206, 61, 0 },
   { 333, 334, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 206, 61, 0 },
   { 342, 343, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 206, 61, 0 },
   { 345, 346, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 206, 61, 0 },
   { 336, 337, 0 },
   { 340, 337, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 206, 61, 0 },
   { 5091, 4052, 0 },
   { 5145, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 206, 61, 0 },
   { 5117, 4052, 0 },
   { 5170, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 206, 61, 0 },
   { 5222, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 206, 61, 0 },
   { 5197, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 330, 331, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 333, 334, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 342, 343, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 345, 346, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 336, 337, 0 },
   { 340, 337, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5091, 4052, 0 },
   { 5145, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5117, 4052, 0 },
   { 5170, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5222, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5197, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 330, 331, 0 },
   { 333, 334, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 330, 331, 0 },
   { 342, 343, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 330, 331, 0 },
   { 345, 346, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 330, 331, 0 },
   { 336, 337, 0 },
   { 340, 337, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 330, 331, 0 },
   { 5091, 4052, 0 },
   { 5145, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 330, 331, 0 },
   { 5117, 4052, 0 },
   { 5170, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 330, 331, 0 },
   { 5222, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 330, 331, 0 },
   { 5197, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 333, 334, 0 },
   { 342, 343, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 333, 334, 0 },
   { 345, 346, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 333, 334, 0 },
   { 336, 337, 0 },
   { 340, 337, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 333, 334, 0 },
   { 5091, 4052, 0 },
   { 5145, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 333, 334, 0 },
   { 5117, 4052, 0 },
   { 5170, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 333, 334, 0 },
   { 5222, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 333, 334, 0 },
   { 5197, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 342, 343, 0 },
   { 345, 346, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 336, 337, 0 },
   { 340, 337, 0 },
   { 342, 343, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 342, 343, 0 },
   { 5091, 4052, 0 },
   { 5145, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 342, 343, 0 },
   { 5117, 4052, 0 },
   { 5170, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 342, 343, 0 },
   { 5222, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 342, 343, 0 },
   { 5197, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 336, 337, 0 },
   { 340, 337, 0 },
   { 345, 346, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 345, 346, 0 },
   { 5091, 4052, 0 },
   { 5145, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 345, 346, 0 },
   { 5117, 4052, 0 },
   { 5170, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 345, 346, 0 },
   { 5222, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 345, 346, 0 },
   { 5197, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 336, 337, 0 },
   { 340, 337, 0 },
   { 5091, 4052, 0 },
   { 5145, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 336, 337, 0 },
   { 340, 337, 0 },
   { 5117, 4052, 0 },
   { 5170, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 336, 337, 0 },
   { 340, 337, 0 },
   { 5222, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 336, 337, 0 },
   { 340, 337, 0 },
   { 5197, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5091, 4052, 0 },
   { 5117, 4052, 0 },
   { 5145, 4065, 0 },
   { 5170, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5091, 4052, 0 },
   { 5145, 4065, 0 },
   { 5222, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5091, 4052, 0 },
   { 5145, 4065, 0 },
   { 5197, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5117, 4052, 0 },
   { 5170, 4065, 0 },
   { 5222, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5117, 4052, 0 },
   { 5170, 4065, 0 },
   { 5197, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5197, 4065, 0 },
   { 5222, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3347, 3348, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1313, 1311, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5686, 5687, 0 },
   { 5690, 5687, 0 },
   { 5693, 5687, 0 },
   { 5696, 5687, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5875, 5876, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5879, 5880, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5673, 5674, 0 },
   { 5677, 5674, 0 },
   { 5680, 5674, 0 },
   { 5683, 5674, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5660, 5661, 0 },
   { 5664, 5661, 0 },
   { 5667, 5661, 0 },
   { 5670, 5661, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5645, 5647, 0 },
   { 5650, 5647, 0 },
   { 5653, 5647, 0 },
   { 5656, 5647, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5699, 5700, 0 },
   { 5703, 5700, 0 },
   { 5706, 5700, 0 },
   { 5709, 5700, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5712, 5713, 0 },
   { 5716, 5713, 0 },
   { 5719, 5713, 0 },
   { 5722, 5713, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5855, 5856, 0 },
   { 5855, 5856, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5624, 5625, 0 },
   { 5628, 5625, 0 },
   { 5630, 5632, 0 },
   { 5634, 5632, 0 },
   { 5637, 5639, 0 },
   { 5641, 5639, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1310, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1310, 1311, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1310, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1310, 1311, 0 },
   { 1313, 1311, 0 },
   { 1316, 1318, 0 },
   { 1321, 1318, 0 },
   { 1325, 1327, 0 },
   { 1330, 1327, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5847, 5848, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 1779, 1781, 0 },
   { 1884, 1886, 0 },
   { 1964, 1966, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5738, 5739, 0 },
   { 5742, 5739, 0 },
   { 5745, 5739, 0 },
   { 5748, 5739, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5751, 5752, 0 },
   { 5755, 5752, 0 },
   { 5758, 5752, 0 },
   { 5761, 5752, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5764, 5765, 0 },
   { 5768, 5765, 0 },
   { 5771, 5765, 0 },
   { 5774, 5765, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5777, 5778, 0 },
   { 5781, 5778, 0 },
   { 5784, 5778, 0 },
   { 5787, 5778, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5790, 5791, 0 },
   { 5794, 5791, 0 },
   { 5797, 5791, 0 },
   { 5800, 5791, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5803, 5804, 0 },
   { 5807, 5804, 0 },
   { 5810, 5804, 0 },
   { 5813, 5804, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5725, 5726, 0 },
   { 5729, 5726, 0 },
   { 5732, 5726, 0 },
   { 5735, 5726, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5851, 5852, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5816, 5817, 0 },
   { 5820, 5817, 0 },
   { 5824, 5826, 0 },
   { 5829, 5826, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5843, 5844, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5834, 5836, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5839, 5840, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5859, 5860, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5863, 5864, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5867, 5868, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5871, 5872, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1313, 1311, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5686, 5687, 0 },
   { 5690, 5687, 0 },
   { 5693, 5687, 0 },
   { 5696, 5687, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5875, 5876, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5879, 5880, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5673, 5674, 0 },
   { 5677, 5674, 0 },
   { 5680, 5674, 0 },
   { 5683, 5674, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5660, 5661, 0 },
   { 5664, 5661, 0 },
   { 5667, 5661, 0 },
   { 5670, 5661, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5645, 5647, 0 },
   { 5650, 5647, 0 },
   { 5653, 5647, 0 },
   { 5656, 5647, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5699, 5700, 0 },
   { 5703, 5700, 0 },
   { 5706, 5700, 0 },
   { 5709, 5700, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5712, 5713, 0 },
   { 5716, 5713, 0 },
   { 5719, 5713, 0 },
   { 5722, 5713, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5855, 5856, 0 },
   { 5855, 5856, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5624, 5625, 0 },
   { 5628, 5625, 0 },
   { 5630, 5632, 0 },
   { 5634, 5632, 0 },
   { 5637, 5639, 0 },
   { 5641, 5639, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1310, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1310, 1311, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1310, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1310, 1311, 0 },
   { 1313, 1311, 0 },
   { 1316, 1318, 0 },
   { 1321, 1318, 0 },
   { 1325, 1327, 0 },
   { 1330, 1327, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5847, 5848, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 1779, 1781, 0 },
   { 1884, 1886, 0 },
   { 1964, 1966, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5738, 5739, 0 },
   { 5742, 5739, 0 },
   { 5745, 5739, 0 },
   { 5748, 5739, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5751, 5752, 0 },
   { 5755, 5752, 0 },
   { 5758, 5752, 0 },
   { 5761, 5752, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5764, 5765, 0 },
   { 5768, 5765, 0 },
   { 5771, 5765, 0 },
   { 5774, 5765, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5777, 5778, 0 },
   { 5781, 5778, 0 },
   { 5784, 5778, 0 },
   { 5787, 5778, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5790, 5791, 0 },
   { 5794, 5791, 0 },
   { 5797, 5791, 0 },
   { 5800, 5791, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5803, 5804, 0 },
   { 5807, 5804, 0 },
   { 5810, 5804, 0 },
   { 5813, 5804, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5725, 5726, 0 },
   { 5729, 5726, 0 },
   { 5732, 5726, 0 },
   { 5735, 5726, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5851, 5852, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5816, 5817, 0 },
   { 5820, 5817, 0 },
   { 5824, 5826, 0 },
   { 5829, 5826, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5843, 5844, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5834, 5836, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5839, 5840, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5859, 5860, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5863, 5864, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5867, 5868, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5871, 5872, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3599, 3584, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3601, 3587, 0 },
   { 3606, 3587, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3602, 3590, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3603, 3593, 0 },
   { 3608, 3593, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3604, 3596, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3044, 760, 0 },
   { 3045, 1131, 0 },
   { 3048, 3049, 0 },
   { 3050, 3051, 0 },
   { 3054, 3055, 0 },
   { 3056, 3057, 0 },
   { 3058, 3059, 0 },
   { 3060, 3061, 0 },
   { 3063, 3064, 0 },
   { 3065, 3066, 0 },
   { 3067, 3068, 57 },
   { 3069, 3070, 57 },
   { 3073, 3074, 57 },
   { 3075, 3076, 57 },
   { 3077, 3, 0 },
   { 3079, 2248, 0 },
   { 3080, 3082, 0 },
   { 3084, 1228, 0 },
   { 3085, 3081, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1313, 1311, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3044, 760, 0 },
   { 3045, 1131, 0 },
   { 3048, 3049, 0 },
   { 3050, 3051, 0 },
   { 3054, 3055, 0 },
   { 3056, 3057, 0 },
   { 3058, 3059, 0 },
   { 3060, 3061, 0 },
   { 3063, 3064, 0 },
   { 3065, 3066, 0 },
   { 3067, 3068, 57 },
   { 3069, 3070, 57 },
   { 3073, 3074, 57 },
   { 3075, 3076, 57 },
   { 3077, 3, 0 },
   { 3079, 2248, 0 },
   { 3080, 3082, 0 },
   { 3084, 1228, 0 },
   { 3085, 3081, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1313, 1311, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5686, 5687, 0 },
   { 5690, 5687, 0 },
   { 5693, 5687, 0 },
   { 5696, 5687, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5875, 5876, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5879, 5880, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5673, 5674, 0 },
   { 5677, 5674, 0 },
   { 5680, 5674, 0 },
   { 5683, 5674, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5660, 5661, 0 },
   { 5664, 5661, 0 },
   { 5667, 5661, 0 },
   { 5670, 5661, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5645, 5647, 0 },
   { 5650, 5647, 0 },
   { 5653, 5647, 0 },
   { 5656, 5647, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5699, 5700, 0 },
   { 5703, 5700, 0 },
   { 5706, 5700, 0 },
   { 5709, 5700, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5712, 5713, 0 },
   { 5716, 5713, 0 },
   { 5719, 5713, 0 },
   { 5722, 5713, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5855, 5856, 0 },
   { 5855, 5856, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5624, 5625, 0 },
   { 5628, 5625, 0 },
   { 5630, 5632, 0 },
   { 5634, 5632, 0 },
   { 5637, 5639, 0 },
   { 5641, 5639, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1310, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1310, 1311, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1310, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1310, 1311, 0 },
   { 1313, 1311, 0 },
   { 1316, 1318, 0 },
   { 1321, 1318, 0 },
   { 1325, 1327, 0 },
   { 1330, 1327, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5847, 5848, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 1779, 1781, 0 },
   { 1884, 1886, 0 },
   { 1964, 1966, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5738, 5739, 0 },
   { 5742, 5739, 0 },
   { 5745, 5739, 0 },
   { 5748, 5739, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5751, 5752, 0 },
   { 5755, 5752, 0 },
   { 5758, 5752, 0 },
   { 5761, 5752, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5764, 5765, 0 },
   { 5768, 5765, 0 },
   { 5771, 5765, 0 },
   { 5774, 5765, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5777, 5778, 0 },
   { 5781, 5778, 0 },
   { 5784, 5778, 0 },
   { 5787, 5778, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5790, 5791, 0 },
   { 5794, 5791, 0 },
   { 5797, 5791, 0 },
   { 5800, 5791, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5803, 5804, 0 },
   { 5807, 5804, 0 },
   { 5810, 5804, 0 },
   { 5813, 5804, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5725, 5726, 0 },
   { 5729, 5726, 0 },
   { 5732, 5726, 0 },
   { 5735, 5726, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5851, 5852, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5816, 5817, 0 },
   { 5820, 5817, 0 },
   { 5824, 5826, 0 },
   { 5829, 5826, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5843, 5844, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5834, 5836, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5839, 5840, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5859, 5860, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5863, 5864, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5867, 5868, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5871, 5872, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3044, 760, 0 },
   { 3045, 1131, 0 },
   { 3048, 3049, 0 },
   { 3050, 3051, 0 },
   { 3054, 3055, 0 },
   { 3056, 3057, 0 },
   { 3058, 3059, 0 },
   { 3060, 3061, 0 },
   { 3063, 3064, 0 },
   { 3065, 3066, 0 },
   { 3067, 3068, 57 },
   { 3069, 3070, 57 },
   { 3073, 3074, 57 },
   { 3075, 3076, 57 },
   { 3077, 3, 0 },
   { 3079, 2248, 0 },
   { 3080, 3082, 0 },
   { 3084, 1228, 0 },
   { 3085, 3081, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1313, 1311, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5686, 5687, 0 },
   { 5690, 5687, 0 },
   { 5693, 5687, 0 },
   { 5696, 5687, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5875, 5876, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5879, 5880, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5673, 5674, 0 },
   { 5677, 5674, 0 },
   { 5680, 5674, 0 },
   { 5683, 5674, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5660, 5661, 0 },
   { 5664, 5661, 0 },
   { 5667, 5661, 0 },
   { 5670, 5661, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5645, 5647, 0 },
   { 5650, 5647, 0 },
   { 5653, 5647, 0 },
   { 5656, 5647, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5699, 5700, 0 },
   { 5703, 5700, 0 },
   { 5706, 5700, 0 },
   { 5709, 5700, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5712, 5713, 0 },
   { 5716, 5713, 0 },
   { 5719, 5713, 0 },
   { 5722, 5713, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5855, 5856, 0 },
   { 5855, 5856, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5624, 5625, 0 },
   { 5628, 5625, 0 },
   { 5630, 5632, 0 },
   { 5634, 5632, 0 },
   { 5637, 5639, 0 },
   { 5641, 5639, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1310, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1310, 1311, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1310, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1310, 1311, 0 },
   { 1313, 1311, 0 },
   { 1316, 1318, 0 },
   { 1321, 1318, 0 },
   { 1325, 1327, 0 },
   { 1330, 1327, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5847, 5848, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 1779, 1781, 0 },
   { 1884, 1886, 0 },
   { 1964, 1966, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5738, 5739, 0 },
   { 5742, 5739, 0 },
   { 5745, 5739, 0 },
   { 5748, 5739, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5751, 5752, 0 },
   { 5755, 5752, 0 },
   { 5758, 5752, 0 },
   { 5761, 5752, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5764, 5765, 0 },
   { 5768, 5765, 0 },
   { 5771, 5765, 0 },
   { 5774, 5765, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5777, 5778, 0 },
   { 5781, 5778, 0 },
   { 5784, 5778, 0 },
   { 5787, 5778, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5790, 5791, 0 },
   { 5794, 5791, 0 },
   { 5797, 5791, 0 },
   { 5800, 5791, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5803, 5804, 0 },
   { 5807, 5804, 0 },
   { 5810, 5804, 0 },
   { 5813, 5804, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5725, 5726, 0 },
   { 5729, 5726, 0 },
   { 5732, 5726, 0 },
   { 5735, 5726, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5851, 5852, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5816, 5817, 0 },
   { 5820, 5817, 0 },
   { 5824, 5826, 0 },
   { 5829, 5826, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5843, 5844, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5834, 5836, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5839, 5840, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5859, 5860, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5863, 5864, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5867, 5868, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5871, 5872, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 2873, 2875, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3675, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3675, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3675, 103, 0 },
   { 5875, 5876, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3675, 103, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3044, 760, 0 },
   { 3045, 1131, 0 },
   { 3048, 3049, 0 },
   { 3050, 3051, 0 },
   { 3054, 3055, 0 },
   { 3056, 3057, 0 },
   { 3058, 3059, 0 },
   { 3060, 3061, 0 },
   { 3063, 3064, 0 },
   { 3065, 3066, 0 },
   { 3067, 3068, 57 },
   { 3069, 3070, 57 },
   { 3073, 3074, 57 },
   { 3075, 3076, 57 },
   { 3077, 3, 0 },
   { 3079, 2248, 0 },
   { 3080, 3082, 0 },
   { 3084, 1228, 0 },
   { 3085, 3081, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1313, 1311, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5686, 5687, 0 },
   { 5690, 5687, 0 },
   { 5693, 5687, 0 },
   { 5696, 5687, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5875, 5876, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5879, 5880, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5673, 5674, 0 },
   { 5677, 5674, 0 },
   { 5680, 5674, 0 },
   { 5683, 5674, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5660, 5661, 0 },
   { 5664, 5661, 0 },
   { 5667, 5661, 0 },
   { 5670, 5661, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5645, 5647, 0 },
   { 5650, 5647, 0 },
   { 5653, 5647, 0 },
   { 5656, 5647, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5699, 5700, 0 },
   { 5703, 5700, 0 },
   { 5706, 5700, 0 },
   { 5709, 5700, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5712, 5713, 0 },
   { 5716, 5713, 0 },
   { 5719, 5713, 0 },
   { 5722, 5713, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5855, 5856, 0 },
   { 5855, 5856, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5624, 5625, 0 },
   { 5628, 5625, 0 },
   { 5630, 5632, 0 },
   { 5634, 5632, 0 },
   { 5637, 5639, 0 },
   { 5641, 5639, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1310, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1310, 1311, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1310, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1310, 1311, 0 },
   { 1313, 1311, 0 },
   { 1316, 1318, 0 },
   { 1321, 1318, 0 },
   { 1325, 1327, 0 },
   { 1330, 1327, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5847, 5848, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 1779, 1781, 0 },
   { 1884, 1886, 0 },
   { 1964, 1966, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5738, 5739, 0 },
   { 5742, 5739, 0 },
   { 5745, 5739, 0 },
   { 5748, 5739, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5751, 5752, 0 },
   { 5755, 5752, 0 },
   { 5758, 5752, 0 },
   { 5761, 5752, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5764, 5765, 0 },
   { 5768, 5765, 0 },
   { 5771, 5765, 0 },
   { 5774, 5765, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5777, 5778, 0 },
   { 5781, 5778, 0 },
   { 5784, 5778, 0 },
   { 5787, 5778, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5790, 5791, 0 },
   { 5794, 5791, 0 },
   { 5797, 5791, 0 },
   { 5800, 5791, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5803, 5804, 0 },
   { 5807, 5804, 0 },
   { 5810, 5804, 0 },
   { 5813, 5804, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5725, 5726, 0 },
   { 5729, 5726, 0 },
   { 5732, 5726, 0 },
   { 5735, 5726, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5851, 5852, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5816, 5817, 0 },
   { 5820, 5817, 0 },
   { 5824, 5826, 0 },
   { 5829, 5826, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5843, 5844, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5834, 5836, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5839, 5840, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5859, 5860, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5863, 5864, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5867, 5868, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5871, 5872, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3044, 760, 0 },
   { 3045, 1131, 0 },
   { 3048, 3049, 0 },
   { 3050, 3051, 0 },
   { 3054, 3055, 0 },
   { 3056, 3057, 0 },
   { 3058, 3059, 0 },
   { 3060, 3061, 0 },
   { 3063, 3064, 0 },
   { 3065, 3066, 0 },
   { 3067, 3068, 57 },
   { 3069, 3070, 57 },
   { 3073, 3074, 57 },
   { 3075, 3076, 57 },
   { 3077, 3, 0 },
   { 3079, 2248, 0 },
   { 3080, 3082, 0 },
   { 3084, 1228, 0 },
   { 3085, 3081, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1313, 1311, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5686, 5687, 0 },
   { 5690, 5687, 0 },
   { 5693, 5687, 0 },
   { 5696, 5687, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5875, 5876, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5879, 5880, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5673, 5674, 0 },
   { 5677, 5674, 0 },
   { 5680, 5674, 0 },
   { 5683, 5674, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5660, 5661, 0 },
   { 5664, 5661, 0 },
   { 5667, 5661, 0 },
   { 5670, 5661, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5645, 5647, 0 },
   { 5650, 5647, 0 },
   { 5653, 5647, 0 },
   { 5656, 5647, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5699, 5700, 0 },
   { 5703, 5700, 0 },
   { 5706, 5700, 0 },
   { 5709, 5700, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5712, 5713, 0 },
   { 5716, 5713, 0 },
   { 5719, 5713, 0 },
   { 5722, 5713, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5855, 5856, 0 },
   { 5855, 5856, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5624, 5625, 0 },
   { 5628, 5625, 0 },
   { 5630, 5632, 0 },
   { 5634, 5632, 0 },
   { 5637, 5639, 0 },
   { 5641, 5639, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1310, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1310, 1311, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1310, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1310, 1311, 0 },
   { 1313, 1311, 0 },
   { 1316, 1318, 0 },
   { 1321, 1318, 0 },
   { 1325, 1327, 0 },
   { 1330, 1327, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5847, 5848, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 1779, 1781, 0 },
   { 1884, 1886, 0 },
   { 1964, 1966, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5738, 5739, 0 },
   { 5742, 5739, 0 },
   { 5745, 5739, 0 },
   { 5748, 5739, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5751, 5752, 0 },
   { 5755, 5752, 0 },
   { 5758, 5752, 0 },
   { 5761, 5752, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5764, 5765, 0 },
   { 5768, 5765, 0 },
   { 5771, 5765, 0 },
   { 5774, 5765, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5777, 5778, 0 },
   { 5781, 5778, 0 },
   { 5784, 5778, 0 },
   { 5787, 5778, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5790, 5791, 0 },
   { 5794, 5791, 0 },
   { 5797, 5791, 0 },
   { 5800, 5791, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5803, 5804, 0 },
   { 5807, 5804, 0 },
   { 5810, 5804, 0 },
   { 5813, 5804, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5725, 5726, 0 },
   { 5729, 5726, 0 },
   { 5732, 5726, 0 },
   { 5735, 5726, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5851, 5852, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5816, 5817, 0 },
   { 5820, 5817, 0 },
   { 5824, 5826, 0 },
   { 5829, 5826, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5843, 5844, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5834, 5836, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5839, 5840, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5859, 5860, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5863, 5864, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5867, 5868, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5871, 5872, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3044, 760, 0 },
   { 3045, 1131, 0 },
   { 3048, 3049, 0 },
   { 3050, 3051, 0 },
   { 3054, 3055, 0 },
   { 3056, 3057, 0 },
   { 3058, 3059, 0 },
   { 3060, 3061, 0 },
   { 3063, 3064, 0 },
   { 3065, 3066, 0 },
   { 3067, 3068, 57 },
   { 3069, 3070, 57 },
   { 3073, 3074, 57 },
   { 3075, 3076, 57 },
   { 3077, 3, 0 },
   { 3079, 2248, 0 },
   { 3080, 3082, 0 },
   { 3084, 1228, 0 },
   { 3085, 3081, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1313, 1311, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5686, 5687, 0 },
   { 5690, 5687, 0 },
   { 5693, 5687, 0 },
   { 5696, 5687, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5875, 5876, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5879, 5880, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5673, 5674, 0 },
   { 5677, 5674, 0 },
   { 5680, 5674, 0 },
   { 5683, 5674, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5660, 5661, 0 },
   { 5664, 5661, 0 },
   { 5667, 5661, 0 },
   { 5670, 5661, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5645, 5647, 0 },
   { 5650, 5647, 0 },
   { 5653, 5647, 0 },
   { 5656, 5647, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5699, 5700, 0 },
   { 5703, 5700, 0 },
   { 5706, 5700, 0 },
   { 5709, 5700, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5712, 5713, 0 },
   { 5716, 5713, 0 },
   { 5719, 5713, 0 },
   { 5722, 5713, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5855, 5856, 0 },
   { 5855, 5856, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5624, 5625, 0 },
   { 5628, 5625, 0 },
   { 5630, 5632, 0 },
   { 5634, 5632, 0 },
   { 5637, 5639, 0 },
   { 5641, 5639, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1310, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1310, 1311, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1310, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1310, 1311, 0 },
   { 1313, 1311, 0 },
   { 1316, 1318, 0 },
   { 1321, 1318, 0 },
   { 1325, 1327, 0 },
   { 1330, 1327, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5847, 5848, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 1779, 1781, 0 },
   { 1884, 1886, 0 },
   { 1964, 1966, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5738, 5739, 0 },
   { 5742, 5739, 0 },
   { 5745, 5739, 0 },
   { 5748, 5739, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5751, 5752, 0 },
   { 5755, 5752, 0 },
   { 5758, 5752, 0 },
   { 5761, 5752, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5764, 5765, 0 },
   { 5768, 5765, 0 },
   { 5771, 5765, 0 },
   { 5774, 5765, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5777, 5778, 0 },
   { 5781, 5778, 0 },
   { 5784, 5778, 0 },
   { 5787, 5778, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5790, 5791, 0 },
   { 5794, 5791, 0 },
   { 5797, 5791, 0 },
   { 5800, 5791, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5803, 5804, 0 },
   { 5807, 5804, 0 },
   { 5810, 5804, 0 },
   { 5813, 5804, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5725, 5726, 0 },
   { 5729, 5726, 0 },
   { 5732, 5726, 0 },
   { 5735, 5726, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5851, 5852, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5816, 5817, 0 },
   { 5820, 5817, 0 },
   { 5824, 5826, 0 },
   { 5829, 5826, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5843, 5844, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5834, 5836, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5839, 5840, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5859, 5860, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5863, 5864, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5867, 5868, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5871, 5872, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3044, 760, 0 },
   { 3045, 1131, 0 },
   { 3048, 3049, 0 },
   { 3050, 3051, 0 },
   { 3054, 3055, 0 },
   { 3056, 3057, 0 },
   { 3058, 3059, 0 },
   { 3060, 3061, 0 },
   { 3063, 3064, 0 },
   { 3065, 3066, 0 },
   { 3067, 3068, 57 },
   { 3069, 3070, 57 },
   { 3073, 3074, 57 },
   { 3075, 3076, 57 },
   { 3077, 3, 0 },
   { 3079, 2248, 0 },
   { 3080, 3082, 0 },
   { 3084, 1228, 0 },
   { 3085, 3081, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1313, 1311, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5686, 5687, 0 },
   { 5690, 5687, 0 },
   { 5693, 5687, 0 },
   { 5696, 5687, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5875, 5876, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5879, 5880, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5673, 5674, 0 },
   { 5677, 5674, 0 },
   { 5680, 5674, 0 },
   { 5683, 5674, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5660, 5661, 0 },
   { 5664, 5661, 0 },
   { 5667, 5661, 0 },
   { 5670, 5661, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5645, 5647, 0 },
   { 5650, 5647, 0 },
   { 5653, 5647, 0 },
   { 5656, 5647, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5699, 5700, 0 },
   { 5703, 5700, 0 },
   { 5706, 5700, 0 },
   { 5709, 5700, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5712, 5713, 0 },
   { 5716, 5713, 0 },
   { 5719, 5713, 0 },
   { 5722, 5713, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5855, 5856, 0 },
   { 5855, 5856, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5624, 5625, 0 },
   { 5628, 5625, 0 },
   { 5630, 5632, 0 },
   { 5634, 5632, 0 },
   { 5637, 5639, 0 },
   { 5641, 5639, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1310, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1310, 1311, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1310, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1310, 1311, 0 },
   { 1313, 1311, 0 },
   { 1316, 1318, 0 },
   { 1321, 1318, 0 },
   { 1325, 1327, 0 },
   { 1330, 1327, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5847, 5848, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 1779, 1781, 0 },
   { 1884, 1886, 0 },
   { 1964, 1966, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5738, 5739, 0 },
   { 5742, 5739, 0 },
   { 5745, 5739, 0 },
   { 5748, 5739, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5751, 5752, 0 },
   { 5755, 5752, 0 },
   { 5758, 5752, 0 },
   { 5761, 5752, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5764, 5765, 0 },
   { 5768, 5765, 0 },
   { 5771, 5765, 0 },
   { 5774, 5765, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5777, 5778, 0 },
   { 5781, 5778, 0 },
   { 5784, 5778, 0 },
   { 5787, 5778, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5790, 5791, 0 },
   { 5794, 5791, 0 },
   { 5797, 5791, 0 },
   { 5800, 5791, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5803, 5804, 0 },
   { 5807, 5804, 0 },
   { 5810, 5804, 0 },
   { 5813, 5804, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5725, 5726, 0 },
   { 5729, 5726, 0 },
   { 5732, 5726, 0 },
   { 5735, 5726, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5851, 5852, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5816, 5817, 0 },
   { 5820, 5817, 0 },
   { 5824, 5826, 0 },
   { 5829, 5826, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5843, 5844, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5834, 5836, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5839, 5840, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5859, 5860, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5863, 5864, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5867, 5868, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5871, 5872, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3044, 760, 0 },
   { 3045, 1131, 0 },
   { 3048, 3049, 0 },
   { 3050, 3051, 0 },
   { 3054, 3055, 0 },
   { 3056, 3057, 0 },
   { 3058, 3059, 0 },
   { 3060, 3061, 0 },
   { 3063, 3064, 0 },
   { 3065, 3066, 0 },
   { 3067, 3068, 57 },
   { 3069, 3070, 57 },
   { 3073, 3074, 57 },
   { 3075, 3076, 57 },
   { 3077, 3, 0 },
   { 3079, 2248, 0 },
   { 3080, 3082, 0 },
   { 3084, 1228, 0 },
   { 3085, 3081, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1313, 1311, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5686, 5687, 0 },
   { 5690, 5687, 0 },
   { 5693, 5687, 0 },
   { 5696, 5687, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5875, 5876, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5879, 5880, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5673, 5674, 0 },
   { 5677, 5674, 0 },
   { 5680, 5674, 0 },
   { 5683, 5674, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5660, 5661, 0 },
   { 5664, 5661, 0 },
   { 5667, 5661, 0 },
   { 5670, 5661, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5645, 5647, 0 },
   { 5650, 5647, 0 },
   { 5653, 5647, 0 },
   { 5656, 5647, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5699, 5700, 0 },
   { 5703, 5700, 0 },
   { 5706, 5700, 0 },
   { 5709, 5700, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5712, 5713, 0 },
   { 5716, 5713, 0 },
   { 5719, 5713, 0 },
   { 5722, 5713, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5855, 5856, 0 },
   { 5855, 5856, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5624, 5625, 0 },
   { 5628, 5625, 0 },
   { 5630, 5632, 0 },
   { 5634, 5632, 0 },
   { 5637, 5639, 0 },
   { 5641, 5639, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1310, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1310, 1311, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1310, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1310, 1311, 0 },
   { 1313, 1311, 0 },
   { 1316, 1318, 0 },
   { 1321, 1318, 0 },
   { 1325, 1327, 0 },
   { 1330, 1327, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3676, 3677, 0 },
   { 3679, 3681, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5847, 5848, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 1779, 1781, 0 },
   { 1884, 1886, 0 },
   { 1964, 1966, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5738, 5739, 0 },
   { 5742, 5739, 0 },
   { 5745, 5739, 0 },
   { 5748, 5739, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5751, 5752, 0 },
   { 5755, 5752, 0 },
   { 5758, 5752, 0 },
   { 5761, 5752, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5764, 5765, 0 },
   { 5768, 5765, 0 },
   { 5771, 5765, 0 },
   { 5774, 5765, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5777, 5778, 0 },
   { 5781, 5778, 0 },
   { 5784, 5778, 0 },
   { 5787, 5778, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5790, 5791, 0 },
   { 5794, 5791, 0 },
   { 5797, 5791, 0 },
   { 5800, 5791, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5803, 5804, 0 },
   { 5807, 5804, 0 },
   { 5810, 5804, 0 },
   { 5813, 5804, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5087, 5082, 0 },
   { 5140, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5112, 5082, 0 },
   { 5165, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5217, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5192, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5725, 5726, 0 },
   { 5729, 5726, 0 },
   { 5732, 5726, 0 },
   { 5735, 5726, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5851, 5852, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5816, 5817, 0 },
   { 5820, 5817, 0 },
   { 5824, 5826, 0 },
   { 5829, 5826, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5843, 5844, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5834, 5836, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5839, 5840, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5859, 5860, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5863, 5864, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5867, 5868, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5871, 5872, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2132, 2133, 0 },
   { 2163, 2164, 0 },
   { 2193, 946, 0 },
   { 2223, 2224, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2132, 2133, 0 },
   { 2163, 2164, 0 },
   { 2193, 946, 0 },
   { 2223, 2224, 0 },
   { 2764, 1229, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2764, 1229, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 823, 825, 0 },
   { 2765, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 827, 836, 0 },
   { 849, 857, 0 },
   { 870, 877, 0 },
   { 890, 898, 0 },
   { 2765, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 1020, 1022, 0 },
   { 2765, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 990, 991, 0 },
   { 997, 998, 0 },
   { 1004, 1005, 0 },
   { 1013, 1014, 0 },
   { 2765, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 910, 913, 0 },
   { 918, 921, 0 },
   { 926, 929, 0 },
   { 934, 937, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 910, 913, 0 },
   { 918, 921, 0 },
   { 926, 929, 0 },
   { 934, 937, 0 },
   { 2765, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 910, 913, 0 },
   { 918, 921, 0 },
   { 926, 929, 0 },
   { 934, 937, 0 },
   { 2132, 2133, 0 },
   { 2163, 2164, 0 },
   { 2193, 946, 0 },
   { 2223, 2224, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 910, 913, 0 },
   { 918, 921, 0 },
   { 926, 929, 0 },
   { 934, 937, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 914, 913, 0 },
   { 922, 921, 0 },
   { 930, 929, 0 },
   { 938, 937, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 914, 913, 0 },
   { 922, 921, 0 },
   { 930, 929, 0 },
   { 938, 937, 0 },
   { 2765, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 914, 913, 0 },
   { 922, 921, 0 },
   { 930, 929, 0 },
   { 938, 937, 0 },
   { 2132, 2133, 0 },
   { 2163, 2164, 0 },
   { 2193, 946, 0 },
   { 2223, 2224, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 914, 913, 0 },
   { 922, 921, 0 },
   { 930, 929, 0 },
   { 938, 937, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 2765, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 2765, 32, 0 },
   { 3342, 3343, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3853, 3854, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 2765, 32, 0 },
   { 3853, 3854, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2132, 2133, 0 },
   { 2163, 2164, 0 },
   { 2193, 946, 0 },
   { 2223, 2224, 0 },
   { 3853, 3854, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3853, 3854, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2718, 2719, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2717, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2704, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5098, 2023, 0 },
   { 5151, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5123, 2023, 0 },
   { 5176, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5228, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5203, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 950, 951, 0 },
   { 2683, 2684, 0 },
   { 2686, 2687, 0 },
   { 2700, 32, 0 },
   { 2701, 61, 0 },
   { 2702, 32, 0 },
   { 2725, 2729, 0 },
   { 2731, 2735, 0 },
   { 2737, 2738, 0 },
   { 2740, 2741, 0 },
   { 3245, 3246, 58 },
   { 3276, 3277, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 945, 946, 0 },
   { 992, 991, 0 },
   { 999, 998, 0 },
   { 1007, 1005, 0 },
   { 1016, 1014, 0 },
   { 2683, 2684, 0 },
   { 2686, 2687, 0 },
   { 2700, 32, 0 },
   { 2701, 61, 0 },
   { 2702, 32, 0 },
   { 2725, 2729, 0 },
   { 2731, 2735, 0 },
   { 2737, 2738, 0 },
   { 2740, 2741, 0 },
   { 3245, 3246, 58 },
   { 3276, 3277, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 955, 957, 0 },
   { 961, 963, 0 },
   { 967, 969, 0 },
   { 973, 975, 0 },
   { 979, 981, 0 },
   { 985, 987, 0 },
   { 2683, 2684, 0 },
   { 2686, 2687, 0 },
   { 2700, 32, 0 },
   { 2701, 61, 0 },
   { 2702, 32, 0 },
   { 2725, 2729, 0 },
   { 2731, 2735, 0 },
   { 2737, 2738, 0 },
   { 2740, 2741, 0 },
   { 3245, 3246, 58 },
   { 3276, 3277, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2683, 2684, 0 },
   { 2686, 2687, 0 },
   { 2700, 32, 0 },
   { 2701, 61, 0 },
   { 2702, 32, 0 },
   { 2718, 2719, 0 },
   { 2725, 2729, 0 },
   { 2731, 2735, 0 },
   { 2737, 2738, 0 },
   { 2740, 2741, 0 },
   { 3245, 3246, 58 },
   { 3276, 3277, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2683, 2684, 0 },
   { 2686, 2687, 0 },
   { 2700, 32, 0 },
   { 2701, 61, 0 },
   { 2702, 32, 0 },
   { 2718, 2719, 0 },
   { 2725, 2729, 0 },
   { 2731, 2735, 0 },
   { 2737, 2738, 0 },
   { 2740, 2741, 0 },
   { 3245, 3246, 58 },
   { 3276, 3277, 59 },
   { 3567, 3569, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2683, 2684, 0 },
   { 2686, 2687, 0 },
   { 2700, 32, 0 },
   { 2701, 61, 0 },
   { 2702, 32, 0 },
   { 2717, 3, 0 },
   { 2725, 2729, 0 },
   { 2731, 2735, 0 },
   { 2737, 2738, 0 },
   { 2740, 2741, 0 },
   { 3245, 3246, 58 },
   { 3276, 3277, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 994, 993, 0 },
   { 1001, 1000, 0 },
   { 1009, 1008, 0 },
   { 1018, 1017, 0 },
   { 2683, 2684, 0 },
   { 2686, 2687, 0 },
   { 2700, 32, 0 },
   { 2701, 61, 0 },
   { 2702, 32, 0 },
   { 2725, 2729, 0 },
   { 2731, 2735, 0 },
   { 2737, 2738, 0 },
   { 2740, 2741, 0 },
   { 3245, 3246, 58 },
   { 3276, 3277, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2683, 2684, 0 },
   { 2686, 2687, 0 },
   { 2700, 32, 0 },
   { 2701, 61, 0 },
   { 2702, 32, 0 },
   { 2704, 61, 0 },
   { 2725, 2729, 0 },
   { 2731, 2735, 0 },
   { 2737, 2738, 0 },
   { 2740, 2741, 0 },
   { 3245, 3246, 58 },
   { 3276, 3277, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2683, 2684, 0 },
   { 2686, 2687, 0 },
   { 2700, 32, 0 },
   { 2701, 61, 0 },
   { 2702, 32, 0 },
   { 2725, 2729, 0 },
   { 2731, 2735, 0 },
   { 2737, 2738, 0 },
   { 2740, 2741, 0 },
   { 3245, 3246, 58 },
   { 3276, 3277, 59 },
   { 5098, 2023, 0 },
   { 5151, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2683, 2684, 0 },
   { 2686, 2687, 0 },
   { 2700, 32, 0 },
   { 2701, 61, 0 },
   { 2702, 32, 0 },
   { 2725, 2729, 0 },
   { 2731, 2735, 0 },
   { 2737, 2738, 0 },
   { 2740, 2741, 0 },
   { 3245, 3246, 58 },
   { 3276, 3277, 59 },
   { 5123, 2023, 0 },
   { 5176, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2683, 2684, 0 },
   { 2686, 2687, 0 },
   { 2700, 32, 0 },
   { 2701, 61, 0 },
   { 2702, 32, 0 },
   { 2725, 2729, 0 },
   { 2731, 2735, 0 },
   { 2737, 2738, 0 },
   { 2740, 2741, 0 },
   { 3245, 3246, 58 },
   { 3276, 3277, 59 },
   { 5228, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2683, 2684, 0 },
   { 2686, 2687, 0 },
   { 2700, 32, 0 },
   { 2701, 61, 0 },
   { 2702, 32, 0 },
   { 2725, 2729, 0 },
   { 2731, 2735, 0 },
   { 2737, 2738, 0 },
   { 2740, 2741, 0 },
   { 3245, 3246, 58 },
   { 3276, 3277, 59 },
   { 5203, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2061, 2062, 0 },
   { 2700, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2718, 2719, 0 },
   { 2722, 2723, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2717, 3, 0 },
   { 2718, 2719, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2704, 61, 0 },
   { 2718, 2719, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2718, 2719, 0 },
   { 5098, 2023, 0 },
   { 5151, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2718, 2719, 0 },
   { 5123, 2023, 0 },
   { 5176, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2718, 2719, 0 },
   { 5228, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2718, 2719, 0 },
   { 5203, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2278, 2279, 0 },
   { 2700, 32, 0 },
   { 2718, 2719, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 191, 193, 0 },
   { 2700, 32, 0 },
   { 2717, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2704, 61, 0 },
   { 2717, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2717, 3, 0 },
   { 5098, 2023, 0 },
   { 5151, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2717, 3, 0 },
   { 5123, 2023, 0 },
   { 5176, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2717, 3, 0 },
   { 5228, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2717, 3, 0 },
   { 5203, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1082, 1077, 0 },
   { 2700, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2114, 2116, 43 },
   { 2114, 2122, 45 },
   { 2145, 2147, 43 },
   { 2145, 2153, 45 },
   { 2175, 2177, 43 },
   { 2175, 2183, 45 },
   { 2205, 2207, 43 },
   { 2205, 2213, 45 },
   { 2700, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2126, 2127, 46 },
   { 2157, 2158, 46 },
   { 2187, 2188, 46 },
   { 2217, 2218, 46 },
   { 2700, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5513, 5514, 0 },
   { 5516, 5517, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5502, 5503, 0 },
   { 5506, 5507, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5508, 5509, 0 },
   { 5511, 5512, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1076, 1077, 0 },
   { 2700, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2677, 2676, 0 },
   { 2700, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2052, 2054, 0 },
   { 2700, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2057, 2058, 0 },
   { 2700, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2704, 61, 0 },
   { 2760, 2761, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2704, 61, 0 },
   { 5098, 2023, 0 },
   { 5151, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2704, 61, 0 },
   { 5123, 2023, 0 },
   { 5176, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2704, 61, 0 },
   { 5228, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2704, 61, 0 },
   { 5203, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1079, 1077, 0 },
   { 1628, 1630, 0 },
   { 1631, 1633, 0 },
   { 2700, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1079, 1077, 0 },
   { 1628, 1630, 0 },
   { 1631, 1633, 0 },
   { 1644, 1630, 0 },
   { 2700, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1079, 1077, 0 },
   { 1628, 1630, 0 },
   { 1631, 1633, 0 },
   { 1757, 1758, 0 },
   { 1759, 1758, 0 },
   { 1862, 1863, 0 },
   { 1864, 1863, 0 },
   { 1942, 1943, 0 },
   { 1944, 1943, 0 },
   { 2700, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1079, 1077, 0 },
   { 1628, 1630, 0 },
   { 1631, 1633, 0 },
   { 1644, 1630, 0 },
   { 1757, 1758, 0 },
   { 1759, 1758, 0 },
   { 1862, 1863, 0 },
   { 1864, 1863, 0 },
   { 1942, 1943, 0 },
   { 1944, 1943, 0 },
   { 2700, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1079, 1077, 0 },
   { 1628, 1630, 0 },
   { 1631, 1633, 0 },
   { 1649, 1633, 0 },
   { 2700, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1079, 1077, 0 },
   { 1628, 1630, 0 },
   { 1631, 1633, 0 },
   { 1649, 1633, 0 },
   { 1757, 1758, 0 },
   { 1759, 1758, 0 },
   { 1862, 1863, 0 },
   { 1864, 1863, 0 },
   { 1942, 1943, 0 },
   { 1944, 1943, 0 },
   { 2700, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1079, 1077, 0 },
   { 1628, 1630, 0 },
   { 1631, 1633, 0 },
   { 1644, 1630, 0 },
   { 1649, 1633, 0 },
   { 1757, 1758, 0 },
   { 1759, 1758, 0 },
   { 1862, 1863, 0 },
   { 1864, 1863, 0 },
   { 1942, 1943, 0 },
   { 1944, 1943, 0 },
   { 2700, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1634, 1636, 0 },
   { 1637, 1639, 0 },
   { 2700, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1634, 1636, 0 },
   { 1637, 1639, 0 },
   { 1652, 1636, 0 },
   { 2700, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1259, 1260, 38 },
   { 1634, 1636, 0 },
   { 1637, 1639, 0 },
   { 1760, 1761, 0 },
   { 1762, 1761, 0 },
   { 1865, 1866, 0 },
   { 1867, 1866, 0 },
   { 1945, 1946, 0 },
   { 1947, 1946, 0 },
   { 2700, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1259, 1260, 38 },
   { 1634, 1636, 0 },
   { 1637, 1639, 0 },
   { 1652, 1636, 0 },
   { 1760, 1761, 0 },
   { 1762, 1761, 0 },
   { 1865, 1866, 0 },
   { 1867, 1866, 0 },
   { 1945, 1946, 0 },
   { 1947, 1946, 0 },
   { 2700, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1634, 1636, 0 },
   { 1637, 1639, 0 },
   { 1655, 1639, 0 },
   { 2700, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1259, 1260, 38 },
   { 1634, 1636, 0 },
   { 1637, 1639, 0 },
   { 1655, 1639, 0 },
   { 1760, 1761, 0 },
   { 1762, 1761, 0 },
   { 1865, 1866, 0 },
   { 1867, 1866, 0 },
   { 1945, 1946, 0 },
   { 1947, 1946, 0 },
   { 2700, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1259, 1260, 38 },
   { 1634, 1636, 0 },
   { 1637, 1639, 0 },
   { 1652, 1636, 0 },
   { 1655, 1639, 0 },
   { 1760, 1761, 0 },
   { 1762, 1761, 0 },
   { 1865, 1866, 0 },
   { 1867, 1866, 0 },
   { 1945, 1946, 0 },
   { 1947, 1946, 0 },
   { 2700, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1084, 1077, 0 },
   { 1696, 1697, 0 },
   { 1698, 1699, 0 },
   { 2700, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1086, 1077, 0 },
   { 1704, 1705, 0 },
   { 1706, 1707, 0 },
   { 2700, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1700, 1701, 0 },
   { 1702, 1703, 0 },
   { 2700, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1708, 1709, 0 },
   { 1710, 1711, 0 },
   { 2700, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2040, 2041, 0 },
   { 2700, 32, 0 },
   { 5098, 2023, 0 },
   { 5151, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5098, 2023, 0 },
   { 5123, 2023, 0 },
   { 5151, 2031, 0 },
   { 5176, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5098, 2023, 0 },
   { 5151, 2031, 0 },
   { 5228, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5098, 2023, 0 },
   { 5151, 2031, 0 },
   { 5203, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2035, 2037, 0 },
   { 2700, 32, 0 },
   { 5123, 2023, 0 },
   { 5176, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5123, 2023, 0 },
   { 5176, 2031, 0 },
   { 5228, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5123, 2023, 0 },
   { 5176, 2031, 0 },
   { 5203, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2044, 2045, 0 },
   { 2700, 32, 0 },
   { 5228, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5203, 2031, 0 },
   { 5228, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2048, 2049, 0 },
   { 2700, 32, 0 },
   { 5203, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2715, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2716, 2338, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2247, 1235, 0 },
   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2708, 63, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5101, 2065, 0 },
   { 5154, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5126, 2065, 0 },
   { 5179, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5231, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5206, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2706, 32, 0 },
   { 2707, 63, 0 },
   { 2715, 3, 0 },
   { 2742, 2744, 0 },
   { 2745, 2747, 0 },
   { 2748, 2749, 0 },
   { 2750, 2751, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2706, 32, 0 },
   { 2707, 63, 0 },
   { 2716, 2338, 0 },
   { 2742, 2744, 0 },
   { 2745, 2747, 0 },
   { 2748, 2749, 0 },
   { 2750, 2751, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2706, 32, 0 },
   { 2707, 63, 0 },
   { 2716, 2338, 0 },
   { 2742, 2744, 0 },
   { 2745, 2747, 0 },
   { 2748, 2749, 0 },
   { 2750, 2751, 0 },
   { 3571, 3573, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2247, 1235, 0 },
   { 2705, 32, 0 },
   { 2706, 32, 0 },
   { 2707, 63, 0 },
   { 2742, 2744, 0 },
   { 2745, 2747, 0 },
   { 2748, 2749, 0 },
   { 2750, 2751, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2706, 32, 0 },
   { 2707, 63, 0 },
   { 2708, 63, 0 },
   { 2742, 2744, 0 },
   { 2745, 2747, 0 },
   { 2748, 2749, 0 },
   { 2750, 2751, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2706, 32, 0 },
   { 2707, 63, 0 },
   { 2742, 2744, 0 },
   { 2745, 2747, 0 },
   { 2748, 2749, 0 },
   { 2750, 2751, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2706, 32, 0 },
   { 2707, 63, 0 },
   { 2742, 2744, 0 },
   { 2745, 2747, 0 },
   { 2748, 2749, 0 },
   { 2750, 2751, 0 },
   { 5101, 2065, 0 },
   { 5154, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2706, 32, 0 },
   { 2707, 63, 0 },
   { 2742, 2744, 0 },
   { 2745, 2747, 0 },
   { 2748, 2749, 0 },
   { 2750, 2751, 0 },
   { 5126, 2065, 0 },
   { 5179, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2706, 32, 0 },
   { 2707, 63, 0 },
   { 2742, 2744, 0 },
   { 2745, 2747, 0 },
   { 2748, 2749, 0 },
   { 2750, 2751, 0 },
   { 5231, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2706, 32, 0 },
   { 2707, 63, 0 },
   { 2742, 2744, 0 },
   { 2745, 2747, 0 },
   { 2748, 2749, 0 },
   { 2750, 2751, 0 },
   { 5206, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2085, 2086, 0 },
   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 196, 198, 0 },
   { 2705, 32, 0 },
   { 2715, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2715, 3, 0 },
   { 2716, 2338, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2247, 1235, 0 },
   { 2705, 32, 0 },
   { 2715, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2708, 63, 0 },
   { 2715, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2715, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2715, 3, 0 },
   { 5101, 2065, 0 },
   { 5154, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2715, 3, 0 },
   { 5126, 2065, 0 },
   { 5179, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2715, 3, 0 },
   { 5231, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2715, 3, 0 },
   { 5206, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2275, 2277, 0 },
   { 2705, 32, 0 },
   { 2715, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2247, 1235, 0 },
   { 2705, 32, 0 },
   { 2716, 2338, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2708, 63, 0 },
   { 2716, 2338, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2716, 2338, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2716, 2338, 0 },
   { 5101, 2065, 0 },
   { 5154, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2716, 2338, 0 },
   { 5126, 2065, 0 },
   { 5179, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2716, 2338, 0 },
   { 5231, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2716, 2338, 0 },
   { 5206, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2123, 2125, 46 },
   { 2154, 2156, 46 },
   { 2184, 2186, 46 },
   { 2214, 2216, 46 },
   { 2236, 2237, 0 },
   { 2247, 1235, 0 },
   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2247, 1235, 0 },
   { 2705, 32, 0 },
   { 2708, 63, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2247, 1235, 0 },
   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2240, 2241, 0 },
   { 2244, 2245, 0 },
   { 2247, 1235, 0 },
   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2247, 1235, 0 },
   { 2705, 32, 0 },
   { 5101, 2065, 0 },
   { 5154, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2247, 1235, 0 },
   { 2705, 32, 0 },
   { 5126, 2065, 0 },
   { 5179, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2247, 1235, 0 },
   { 2705, 32, 0 },
   { 5231, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2247, 1235, 0 },
   { 2705, 32, 0 },
   { 5206, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2119, 2120, 44 },
   { 2119, 2128, 45 },
   { 2150, 2151, 44 },
   { 2150, 2159, 45 },
   { 2180, 2181, 44 },
   { 2180, 2189, 45 },
   { 2210, 2211, 44 },
   { 2210, 2219, 45 },
   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2678, 2679, 0 },
   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2080, 2082, 0 },
   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2083, 2084, 0 },
   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2708, 63, 0 },
   { 2762, 2763, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2708, 63, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2708, 63, 0 },
   { 5101, 2065, 0 },
   { 5154, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2708, 63, 0 },
   { 5126, 2065, 0 },
   { 5179, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2708, 63, 0 },
   { 5231, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2708, 63, 0 },
   { 5206, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1595, 1597, 0 },
   { 1601, 1603, 0 },
   { 1737, 1052, 0 },
   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5101, 2065, 0 },
   { 5154, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5126, 2065, 0 },
   { 5179, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5231, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5206, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1595, 1597, 0 },
   { 1601, 1603, 0 },
   { 1616, 1597, 0 },
   { 1737, 1052, 0 },
   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1595, 1597, 0 },
   { 1601, 1603, 0 },
   { 1737, 1052, 0 },
   { 1743, 1745, 0 },
   { 1748, 1745, 0 },
   { 1848, 1850, 0 },
   { 1853, 1850, 0 },
   { 1928, 1930, 0 },
   { 1933, 1930, 0 },
   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1595, 1597, 0 },
   { 1601, 1603, 0 },
   { 1616, 1597, 0 },
   { 1737, 1052, 0 },
   { 1743, 1745, 0 },
   { 1748, 1745, 0 },
   { 1848, 1850, 0 },
   { 1853, 1850, 0 },
   { 1928, 1930, 0 },
   { 1933, 1930, 0 },
   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1595, 1597, 0 },
   { 1601, 1603, 0 },
   { 1621, 1603, 0 },
   { 1737, 1052, 0 },
   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1595, 1597, 0 },
   { 1601, 1603, 0 },
   { 1621, 1603, 0 },
   { 1737, 1052, 0 },
   { 1743, 1745, 0 },
   { 1748, 1745, 0 },
   { 1848, 1850, 0 },
   { 1853, 1850, 0 },
   { 1928, 1930, 0 },
   { 1933, 1930, 0 },
   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1595, 1597, 0 },
   { 1601, 1603, 0 },
   { 1616, 1597, 0 },
   { 1621, 1603, 0 },
   { 1737, 1052, 0 },
   { 1743, 1745, 0 },
   { 1748, 1745, 0 },
   { 1848, 1850, 0 },
   { 1853, 1850, 0 },
   { 1928, 1930, 0 },
   { 1933, 1930, 0 },
   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1606, 1608, 0 },
   { 1611, 1613, 0 },
   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1606, 1608, 0 },
   { 1611, 1613, 0 },
   { 1624, 1608, 0 },
   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1606, 1608, 0 },
   { 1611, 1613, 0 },
   { 1751, 1753, 0 },
   { 1756, 1753, 0 },
   { 1856, 1858, 0 },
   { 1861, 1858, 0 },
   { 1936, 1938, 0 },
   { 1941, 1938, 0 },
   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1606, 1608, 0 },
   { 1611, 1613, 0 },
   { 1624, 1608, 0 },
   { 1751, 1753, 0 },
   { 1756, 1753, 0 },
   { 1856, 1858, 0 },
   { 1861, 1858, 0 },
   { 1936, 1938, 0 },
   { 1941, 1938, 0 },
   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1606, 1608, 0 },
   { 1611, 1613, 0 },
   { 1627, 1613, 0 },
   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1606, 1608, 0 },
   { 1611, 1613, 0 },
   { 1627, 1613, 0 },
   { 1751, 1753, 0 },
   { 1756, 1753, 0 },
   { 1856, 1858, 0 },
   { 1861, 1858, 0 },
   { 1936, 1938, 0 },
   { 1941, 1938, 0 },
   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1606, 1608, 0 },
   { 1611, 1613, 0 },
   { 1624, 1608, 0 },
   { 1627, 1613, 0 },
   { 1751, 1753, 0 },
   { 1756, 1753, 0 },
   { 1856, 1858, 0 },
   { 1861, 1858, 0 },
   { 1936, 1938, 0 },
   { 1941, 1938, 0 },
   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1658, 1660, 0 },
   { 1663, 1665, 0 },
   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1678, 1680, 0 },
   { 1683, 1685, 0 },
   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1668, 1670, 0 },
   { 1673, 1675, 0 },
   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1688, 1690, 0 },
   { 1693, 1695, 0 },
   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2074, 2075, 0 },
   { 2705, 32, 0 },
   { 5101, 2065, 0 },
   { 5154, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5101, 2065, 0 },
   { 5126, 2065, 0 },
   { 5154, 2069, 0 },
   { 5179, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5101, 2065, 0 },
   { 5154, 2069, 0 },
   { 5231, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5101, 2065, 0 },
   { 5154, 2069, 0 },
   { 5206, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2071, 2073, 0 },
   { 2705, 32, 0 },
   { 5126, 2065, 0 },
   { 5179, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5126, 2065, 0 },
   { 5179, 2069, 0 },
   { 5231, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5126, 2065, 0 },
   { 5179, 2069, 0 },
   { 5206, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2076, 2077, 0 },
   { 2705, 32, 0 },
   { 5231, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5206, 2069, 0 },
   { 5231, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2078, 2079, 0 },
   { 2705, 32, 0 },
   { 5206, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 3349, 3348, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5104, 2089, 0 },
   { 5157, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5129, 2089, 0 },
   { 5182, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5234, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5209, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 2712, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 2710, 32, 0 },
   { 2713, 2703, 0 },
   { 2752, 2754, 0 },
   { 2755, 2757, 0 },
   { 5104, 2089, 0 },
   { 5157, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 2710, 32, 0 },
   { 2713, 2703, 0 },
   { 2752, 2754, 0 },
   { 2755, 2757, 0 },
   { 5129, 2089, 0 },
   { 5182, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 2710, 32, 0 },
   { 2713, 2703, 0 },
   { 2752, 2754, 0 },
   { 2755, 2757, 0 },
   { 5234, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 2710, 32, 0 },
   { 2713, 2703, 0 },
   { 2752, 2754, 0 },
   { 2755, 2757, 0 },
   { 5209, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 2710, 32, 0 },
   { 2712, 3, 0 },
   { 2713, 2703, 0 },
   { 2752, 2754, 0 },
   { 2755, 2757, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 2710, 32, 0 },
   { 2712, 3, 0 },
   { 2713, 2703, 0 },
   { 2752, 2754, 0 },
   { 2755, 2757, 0 },
   { 3575, 3577, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2109, 2110, 0 },
   { 2709, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2104, 2106, 0 },
   { 2709, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2107, 2108, 0 },
   { 2709, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2098, 2099, 0 },
   { 2709, 61, 0 },
   { 5104, 2089, 0 },
   { 5157, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5104, 2089, 0 },
   { 5129, 2089, 0 },
   { 5157, 2093, 0 },
   { 5182, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5104, 2089, 0 },
   { 5157, 2093, 0 },
   { 5234, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5104, 2089, 0 },
   { 5157, 2093, 0 },
   { 5209, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 2712, 3, 0 },
   { 5104, 2089, 0 },
   { 5157, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2095, 2097, 0 },
   { 2709, 61, 0 },
   { 5129, 2089, 0 },
   { 5182, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5129, 2089, 0 },
   { 5182, 2093, 0 },
   { 5234, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5129, 2089, 0 },
   { 5182, 2093, 0 },
   { 5209, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 2712, 3, 0 },
   { 5129, 2089, 0 },
   { 5182, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2100, 2101, 0 },
   { 2709, 61, 0 },
   { 5234, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5209, 2093, 0 },
   { 5234, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 2712, 3, 0 },
   { 5234, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2102, 2103, 0 },
   { 2709, 61, 0 },
   { 5209, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 2712, 3, 0 },
   { 5209, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2136, 2137, 0 },
   { 2167, 2168, 0 },
   { 2196, 2197, 0 },
   { 2227, 2228, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2136, 2137, 0 },
   { 2167, 2168, 0 },
   { 2196, 2197, 0 },
   { 2227, 2228, 0 },
   { 2770, 1229, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2770, 1229, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 915, 917, 0 },
   { 923, 925, 0 },
   { 931, 933, 0 },
   { 939, 941, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 915, 917, 0 },
   { 923, 925, 0 },
   { 931, 933, 0 },
   { 939, 941, 0 },
   { 2771, 32, 0 },
   { 3234, 3235, 58 },
   { 1017, 422, 58 },
   { 3237, 3239, 58 },
   { 3272, 444, 59 },
   { 1008, 447, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 915, 917, 0 },
   { 923, 925, 0 },
   { 931, 933, 0 },
   { 939, 941, 0 },
   { 2136, 2137, 0 },
   { 2167, 2168, 0 },
   { 2196, 2197, 0 },
   { 2227, 2228, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 915, 917, 0 },
   { 923, 925, 0 },
   { 931, 933, 0 },
   { 939, 941, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 838, 840, 0 },
   { 859, 861, 0 },
   { 879, 881, 0 },
   { 900, 902, 0 },
   { 2771, 32, 0 },
   { 3234, 3235, 58 },
   { 1017, 422, 58 },
   { 3237, 3239, 58 },
   { 1008, 447, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2771, 32, 0 },
   { 3234, 3235, 58 },
   { 1017, 422, 58 },
   { 3237, 3239, 58 },
   { 1008, 447, 59 },
   { 3344, 3341, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2134, 2135, 0 },
   { 2165, 2166, 0 },
   { 2194, 2195, 0 },
   { 2225, 2226, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2134, 2135, 0 },
   { 2165, 2166, 0 },
   { 2194, 2195, 0 },
   { 2225, 2226, 0 },
   { 2766, 1229, 0 },
   { 2768, 2767, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2769, 32, 0 },
   { 3240, 3241, 58 },
   { 3242, 406, 58 },
   { 3243, 3244, 58 },
   { 3273, 436, 59 },
   { 3274, 439, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2769, 32, 0 },
   { 3240, 3241, 58 },
   { 3242, 406, 58 },
   { 3243, 3244, 58 },
   { 3274, 439, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2134, 2135, 0 },
   { 2165, 2166, 0 },
   { 2194, 2195, 0 },
   { 2225, 2226, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 842, 847, 0 },
   { 863, 868, 0 },
   { 883, 888, 0 },
   { 904, 909, 0 },
   { 2769, 32, 0 },
   { 3240, 3241, 58 },
   { 3242, 406, 58 },
   { 3243, 3244, 58 },
   { 3274, 439, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3417, 3418, 0 },
   { 3420, 419, 0 },
   { 3421, 415, 0 },
   { 3422, 412, 0 },
   { 3424, 3395, 0 },
   { 3426, 3390, 0 },
   { 3428, 3385, 0 },
   { 3430, 3382, 0 },
   { 3432, 3378, 0 },
   { 3434, 3374, 0 },
   { 3436, 3437, 0 },
   { 3916, 3919, 91 },
   { 5078, 11, 0 },
   { 5131, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3367, 3235, 0 },
   { 3369, 415, 0 },
   { 3370, 419, 0 },
   { 3371, 422, 0 },
   { 3373, 3374, 0 },
   { 3376, 3378, 0 },
   { 3380, 3382, 0 },
   { 3383, 3385, 0 },
   { 3388, 3390, 0 },
   { 3393, 3395, 0 },
   { 3396, 3239, 0 },
   { 3916, 3919, 91 },
   { 5078, 11, 0 },
   { 5131, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3397, 3235, 0 },
   { 3399, 415, 0 },
   { 3400, 419, 0 },
   { 3401, 422, 0 },
   { 3403, 3374, 0 },
   { 3405, 3378, 0 },
   { 3407, 3382, 0 },
   { 3409, 3385, 0 },
   { 3411, 3390, 0 },
   { 3413, 3395, 0 },
   { 3414, 3239, 0 },
   { 3916, 3919, 91 },
   { 5078, 11, 0 },
   { 5131, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3271, 3255, 0 },
   { 3916, 3919, 91 },
   { 5078, 11, 0 },
   { 5131, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3270, 3255, 0 },
   { 3916, 3919, 91 },
   { 5078, 11, 0 },
   { 5131, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3415, 3249, 0 },
   { 3916, 3919, 91 },
   { 5078, 11, 0 },
   { 5131, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3359, 3360, 0 },
   { 3361, 3362, 0 },
   { 3363, 3364, 0 },
   { 3365, 3366, 0 },
   { 3916, 3919, 91 },
   { 5078, 11, 0 },
   { 5131, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3438, 3439, 0 },
   { 3440, 402, 0 },
   { 3441, 397, 0 },
   { 3442, 393, 0 },
   { 3443, 3444, 0 },
   { 3445, 3446, 0 },
   { 3447, 3448, 0 },
   { 3449, 3450, 0 },
   { 3451, 3452, 0 },
   { 3453, 3454, 0 },
   { 3455, 3456, 0 },
   { 3911, 3915, 91 },
   { 5106, 11, 0 },
   { 5159, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3351, 3352, 0 },
   { 3353, 3354, 0 },
   { 3355, 3356, 0 },
   { 3357, 3358, 0 },
   { 3911, 3915, 91 },
   { 5106, 11, 0 },
   { 5159, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3219, 21, 0 },
   { 3202, 3302, 61 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 48, 49, 0 },
   { 2732, 3307, 63 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 51, 49, 0 },
   { 2732, 3307, 63 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2732, 3307, 63 },
   { 3317, 2732, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3211, 16, 0 },
   { 2732, 3307, 63 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3208, 16, 0 },
   { 2732, 3307, 63 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3203, 16, 0 },
   { 2732, 3307, 63 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3206, 2427, 0 },
   { 2732, 3307, 63 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 99, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 102, 103, 0 },
   { 3153, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 105, 104, 0 },
   { 3153, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3153, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2397, 2396, 0 },
   { 3153, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2399, 2398, 0 },
   { 3153, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2401, 2400, 0 },
   { 3153, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2403, 2402, 0 },
   { 3153, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3153, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 109, 110, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4217, 760, 0 },
   { 4219, 6, 0 },
   { 4221, 11, 0 },
   { 4223, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 205, 111, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3581, 3510, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 32, 0 },
   { 114, 11, 5 },
   { 116, 16, 6 },
   { 205, 111, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 32, 0 },
   { 114, 11, 5 },
   { 116, 16, 6 },
   { 3508, 3510, 0 },
   { 3512, 3510, 0 },
   { 3581, 3510, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 32, 0 },
   { 114, 11, 5 },
   { 116, 16, 6 },
   { 3508, 3510, 0 },
   { 3512, 3510, 0 },
   { 3539, 3510, 0 },
   { 3581, 3510, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 32, 0 },
   { 114, 11, 5 },
   { 116, 16, 6 },
   { 3515, 3517, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 32, 0 },
   { 114, 11, 5 },
   { 116, 16, 6 },
   { 3515, 3517, 0 },
   { 3546, 3548, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 32, 0 },
   { 114, 11, 5 },
   { 116, 16, 6 },
   { 3519, 3521, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 32, 0 },
   { 114, 11, 5 },
   { 116, 16, 6 },
   { 3519, 3521, 0 },
   { 3556, 3558, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 32, 0 },
   { 114, 11, 5 },
   { 116, 16, 6 },
   { 5590, 5593, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 205, 111, 0 },
   { 214, 3, 0 },
   { 3581, 3510, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 131, 133, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 155, 157, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 139, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 171, 166, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3468, 3469, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3658, 3659, 0 },
   { 3661, 3662, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 221, 111, 0 },
   { 222, 113, 5 },
   { 223, 115, 6 },
   { 242, 32, 0 },
   { 245, 32, 0 },
   { 248, 98, 0 },
   { 3468, 3469, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 221, 111, 0 },
   { 222, 113, 5 },
   { 223, 115, 6 },
   { 242, 32, 0 },
   { 245, 32, 0 },
   { 248, 98, 0 },
   { 3481, 3483, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 221, 111, 0 },
   { 222, 113, 5 },
   { 223, 115, 6 },
   { 242, 32, 0 },
   { 245, 32, 0 },
   { 248, 98, 0 },
   { 3481, 3483, 0 },
   { 3526, 3483, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 221, 111, 0 },
   { 222, 113, 5 },
   { 223, 115, 6 },
   { 242, 32, 0 },
   { 245, 32, 0 },
   { 248, 98, 0 },
   { 3489, 3491, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 221, 111, 0 },
   { 222, 113, 5 },
   { 223, 115, 6 },
   { 242, 32, 0 },
   { 245, 32, 0 },
   { 248, 98, 0 },
   { 3489, 3491, 0 },
   { 3530, 3491, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 221, 111, 0 },
   { 222, 113, 5 },
   { 223, 115, 6 },
   { 242, 32, 0 },
   { 245, 32, 0 },
   { 248, 98, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 221, 111, 0 },
   { 222, 113, 5 },
   { 223, 115, 6 },
   { 242, 32, 0 },
   { 245, 32, 0 },
   { 248, 98, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 221, 111, 0 },
   { 222, 113, 5 },
   { 223, 115, 6 },
   { 242, 32, 0 },
   { 245, 32, 0 },
   { 248, 98, 0 },
   { 5594, 5597, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 221, 111, 0 },
   { 222, 113, 5 },
   { 223, 115, 6 },
   { 242, 32, 0 },
   { 245, 32, 0 },
   { 248, 98, 0 },
   { 3658, 3659, 0 },
   { 3661, 3662, 0 },
   { 5594, 5597, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 221, 111, 0 },
   { 222, 113, 5 },
   { 223, 115, 6 },
   { 242, 32, 0 },
   { 245, 32, 0 },
   { 248, 98, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 221, 111, 0 },
   { 222, 113, 5 },
   { 223, 115, 6 },
   { 242, 32, 0 },
   { 245, 32, 0 },
   { 248, 98, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3468, 3469, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3468, 3469, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3468, 3469, 0 },
   { 3658, 3659, 0 },
   { 3661, 3662, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3468, 3469, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3468, 3469, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 252, 253, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3658, 3659, 0 },
   { 3661, 3662, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3658, 3659, 0 },
   { 3661, 3662, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3658, 3659, 0 },
   { 3661, 3662, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3658, 3659, 0 },
   { 3661, 3662, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2871, 2872, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2680, 1266, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2876, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 226, 227, 6 },
   { 230, 227, 0 },
   { 231, 232, 0 },
   { 3470, 3471, 6 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 226, 227, 6 },
   { 230, 227, 0 },
   { 231, 232, 0 },
   { 3664, 3665, 0 },
   { 3667, 3668, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 224, 115, 0 },
   { 226, 227, 6 },
   { 230, 227, 0 },
   { 231, 232, 0 },
   { 244, 16, 0 },
   { 3470, 3471, 6 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 224, 115, 0 },
   { 226, 227, 6 },
   { 230, 227, 0 },
   { 231, 232, 0 },
   { 244, 16, 0 },
   { 3485, 3487, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 224, 115, 0 },
   { 226, 227, 6 },
   { 230, 227, 0 },
   { 231, 232, 0 },
   { 244, 16, 0 },
   { 3485, 3487, 0 },
   { 3528, 3487, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 224, 115, 0 },
   { 226, 227, 6 },
   { 230, 227, 0 },
   { 231, 232, 0 },
   { 244, 16, 0 },
   { 5599, 5602, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 224, 115, 0 },
   { 226, 227, 6 },
   { 230, 227, 0 },
   { 231, 232, 0 },
   { 244, 16, 0 },
   { 3664, 3665, 0 },
   { 3667, 3668, 0 },
   { 5599, 5602, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 226, 227, 6 },
   { 230, 227, 0 },
   { 231, 232, 0 },
   { 3470, 3471, 6 },
   { 3664, 3665, 0 },
   { 3667, 3668, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 226, 227, 6 },
   { 230, 227, 0 },
   { 231, 232, 0 },
   { 2878, 696, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 147, 148, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 143, 145, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 147, 148, 0 },
   { 257, 3, 0 },
   { 259, 258, 5 },
   { 260, 36, 6 },
   { 271, 272, 0 },
   { 276, 277, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 143, 145, 0 },
   { 257, 3, 0 },
   { 259, 258, 5 },
   { 260, 36, 6 },
   { 271, 272, 0 },
   { 276, 277, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 263, 264, 0 },
   { 265, 266, 5 },
   { 267, 227, 6 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 147, 148, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 143, 145, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 147, 148, 0 },
   { 160, 162, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 147, 148, 0 },
   { 257, 3, 0 },
   { 259, 258, 5 },
   { 260, 36, 6 },
   { 271, 272, 0 },
   { 276, 277, 0 },
   { 284, 285, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 143, 145, 0 },
   { 257, 3, 0 },
   { 259, 258, 5 },
   { 260, 36, 6 },
   { 271, 272, 0 },
   { 276, 277, 0 },
   { 284, 285, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 257, 3, 0 },
   { 259, 258, 5 },
   { 260, 36, 6 },
   { 271, 272, 0 },
   { 276, 277, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 257, 3, 0 },
   { 259, 258, 5 },
   { 260, 36, 6 },
   { 263, 264, 0 },
   { 265, 266, 5 },
   { 267, 227, 6 },
   { 271, 272, 0 },
   { 276, 277, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 147, 148, 0 },
   { 257, 3, 0 },
   { 259, 258, 5 },
   { 260, 36, 6 },
   { 271, 272, 0 },
   { 276, 277, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 143, 145, 0 },
   { 257, 3, 0 },
   { 259, 258, 5 },
   { 260, 36, 6 },
   { 271, 272, 0 },
   { 276, 277, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 257, 3, 0 },
   { 259, 258, 5 },
   { 260, 36, 6 },
   { 271, 272, 0 },
   { 276, 277, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { 3499, 3501, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 257, 3, 0 },
   { 259, 258, 5 },
   { 260, 36, 6 },
   { 263, 264, 0 },
   { 265, 266, 5 },
   { 267, 227, 6 },
   { 271, 272, 0 },
   { 276, 277, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { 3499, 3501, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 147, 148, 0 },
   { 160, 162, 0 },
   { 257, 3, 0 },
   { 259, 258, 5 },
   { 260, 36, 6 },
   { 271, 272, 0 },
   { 276, 277, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { 3499, 3501, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 143, 145, 0 },
   { 257, 3, 0 },
   { 259, 258, 5 },
   { 260, 36, 6 },
   { 271, 272, 0 },
   { 276, 277, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { 3499, 3501, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 257, 3, 0 },
   { 259, 258, 5 },
   { 260, 36, 6 },
   { 271, 272, 0 },
   { 276, 277, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { 3499, 3501, 0 },
   { 3533, 3501, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 257, 3, 0 },
   { 259, 258, 5 },
   { 260, 36, 6 },
   { 263, 264, 0 },
   { 265, 266, 5 },
   { 267, 227, 6 },
   { 271, 272, 0 },
   { 276, 277, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { 3499, 3501, 0 },
   { 3533, 3501, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 147, 148, 0 },
   { 160, 162, 0 },
   { 257, 3, 0 },
   { 259, 258, 5 },
   { 260, 36, 6 },
   { 271, 272, 0 },
   { 276, 277, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { 3499, 3501, 0 },
   { 3533, 3501, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 143, 145, 0 },
   { 257, 3, 0 },
   { 259, 258, 5 },
   { 260, 36, 6 },
   { 271, 272, 0 },
   { 276, 277, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { 3499, 3501, 0 },
   { 3533, 3501, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 257, 3, 0 },
   { 259, 258, 5 },
   { 260, 36, 6 },
   { 271, 272, 0 },
   { 276, 277, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { 3504, 3505, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 257, 3, 0 },
   { 259, 258, 5 },
   { 260, 36, 6 },
   { 263, 264, 0 },
   { 265, 266, 5 },
   { 267, 227, 6 },
   { 271, 272, 0 },
   { 276, 277, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { 3504, 3505, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 147, 148, 0 },
   { 257, 3, 0 },
   { 259, 258, 5 },
   { 260, 36, 6 },
   { 271, 272, 0 },
   { 276, 277, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { 3504, 3505, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 143, 145, 0 },
   { 257, 3, 0 },
   { 259, 258, 5 },
   { 260, 36, 6 },
   { 271, 272, 0 },
   { 276, 277, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { 3504, 3505, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { 3473, 3474, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 263, 264, 0 },
   { 265, 266, 5 },
   { 267, 227, 6 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { 3473, 3474, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 147, 148, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { 3473, 3474, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 143, 145, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { 3473, 3474, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 147, 148, 0 },
   { 160, 162, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 263, 264, 0 },
   { 265, 266, 5 },
   { 267, 227, 6 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 147, 148, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 143, 145, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 168, 139, 0 },
   { 233, 234, 6 },
   { 235, 234, 0 },
   { 236, 237, 0 },
   { 387, 288, 30 },
   { 748, 288, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 165, 166, 0 },
   { 233, 234, 6 },
   { 235, 234, 0 },
   { 236, 237, 0 },
   { 387, 288, 30 },
   { 748, 288, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 168, 139, 0 },
   { 233, 234, 6 },
   { 235, 234, 0 },
   { 236, 237, 0 },
   { 261, 262, 0 },
   { 273, 274, 6 },
   { 279, 281, 6 },
   { 387, 288, 30 },
   { 748, 288, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 165, 166, 0 },
   { 233, 234, 6 },
   { 235, 234, 0 },
   { 236, 237, 0 },
   { 261, 262, 0 },
   { 273, 274, 6 },
   { 279, 281, 6 },
   { 387, 288, 30 },
   { 748, 288, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 168, 139, 0 },
   { 174, 176, 0 },
   { 233, 234, 6 },
   { 235, 234, 0 },
   { 236, 237, 0 },
   { 387, 288, 30 },
   { 748, 288, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 168, 139, 0 },
   { 233, 234, 6 },
   { 235, 234, 0 },
   { 236, 237, 0 },
   { 261, 262, 0 },
   { 273, 274, 6 },
   { 279, 281, 6 },
   { 287, 288, 0 },
   { 387, 288, 30 },
   { 748, 288, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 165, 166, 0 },
   { 233, 234, 6 },
   { 235, 234, 0 },
   { 236, 237, 0 },
   { 261, 262, 0 },
   { 273, 274, 6 },
   { 279, 281, 6 },
   { 287, 288, 0 },
   { 387, 288, 30 },
   { 748, 288, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 233, 234, 6 },
   { 235, 234, 0 },
   { 236, 237, 0 },
   { 261, 262, 0 },
   { 273, 274, 6 },
   { 279, 281, 6 },
   { 387, 288, 30 },
   { 748, 288, 33 },
   { 3502, 3503, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 233, 234, 6 },
   { 235, 234, 0 },
   { 236, 237, 0 },
   { 261, 262, 0 },
   { 268, 269, 6 },
   { 273, 274, 6 },
   { 279, 281, 6 },
   { 387, 288, 30 },
   { 748, 288, 33 },
   { 3502, 3503, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 168, 139, 0 },
   { 174, 176, 0 },
   { 233, 234, 6 },
   { 235, 234, 0 },
   { 236, 237, 0 },
   { 261, 262, 0 },
   { 273, 274, 6 },
   { 279, 281, 6 },
   { 387, 288, 30 },
   { 748, 288, 33 },
   { 3502, 3503, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 165, 166, 0 },
   { 233, 234, 6 },
   { 235, 234, 0 },
   { 236, 237, 0 },
   { 261, 262, 0 },
   { 273, 274, 6 },
   { 279, 281, 6 },
   { 387, 288, 30 },
   { 748, 288, 33 },
   { 3502, 3503, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 233, 234, 6 },
   { 235, 234, 0 },
   { 236, 237, 0 },
   { 261, 262, 0 },
   { 273, 274, 6 },
   { 279, 281, 6 },
   { 387, 288, 30 },
   { 748, 288, 33 },
   { 3502, 3503, 0 },
   { 3534, 3503, 0 },
   { 3535, 3505, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 233, 234, 6 },
   { 235, 234, 0 },
   { 236, 237, 0 },
   { 261, 262, 0 },
   { 268, 269, 6 },
   { 273, 274, 6 },
   { 279, 281, 6 },
   { 387, 288, 30 },
   { 748, 288, 33 },
   { 3502, 3503, 0 },
   { 3534, 3503, 0 },
   { 3535, 3505, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 168, 139, 0 },
   { 174, 176, 0 },
   { 233, 234, 6 },
   { 235, 234, 0 },
   { 236, 237, 0 },
   { 261, 262, 0 },
   { 273, 274, 6 },
   { 279, 281, 6 },
   { 387, 288, 30 },
   { 748, 288, 33 },
   { 3502, 3503, 0 },
   { 3534, 3503, 0 },
   { 3535, 3505, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 165, 166, 0 },
   { 233, 234, 6 },
   { 235, 234, 0 },
   { 236, 237, 0 },
   { 261, 262, 0 },
   { 273, 274, 6 },
   { 279, 281, 6 },
   { 387, 288, 30 },
   { 748, 288, 33 },
   { 3502, 3503, 0 },
   { 3534, 3503, 0 },
   { 3535, 3505, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 233, 234, 6 },
   { 235, 234, 0 },
   { 236, 237, 0 },
   { 387, 288, 30 },
   { 748, 288, 33 },
   { 3475, 387, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 233, 234, 6 },
   { 235, 234, 0 },
   { 236, 237, 0 },
   { 268, 269, 6 },
   { 387, 288, 30 },
   { 748, 288, 33 },
   { 3475, 387, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 168, 139, 0 },
   { 233, 234, 6 },
   { 235, 234, 0 },
   { 236, 237, 0 },
   { 387, 288, 30 },
   { 748, 288, 33 },
   { 3475, 387, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 165, 166, 0 },
   { 233, 234, 6 },
   { 235, 234, 0 },
   { 236, 237, 0 },
   { 387, 288, 30 },
   { 748, 288, 33 },
   { 3475, 387, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 217, 1517, 40 },
   { 3156, 111, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 217, 1517, 40 },
   { 1530, 217, 0 },
   { 3156, 111, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1525, 1527, 0 },
   { 217, 1517, 40 },
   { 3156, 111, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1529, 1528, 0 },
   { 217, 1517, 40 },
   { 3156, 111, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2674, 1257, 0 },
   { 2698, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1191, 1192, 0 },
   { 2698, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1105, 1106, 0 },
   { 2674, 1257, 0 },
   { 2698, 1235, 0 },
   { 3173, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1233, 1108, 0 },
   { 2698, 1235, 0 },
   { 3173, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1191, 1192, 0 },
   { 2698, 1235, 0 },
   { 3173, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1191, 1192, 0 },
   { 1210, 1212, 0 },
   { 2698, 1235, 0 },
   { 3173, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1122, 1055, 0 },
   { 2698, 1235, 0 },
   { 3173, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2698, 1235, 0 },
   { 3173, 1077, 0 },
   { 5519, 5522, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1226, 1131, 0 },
   { 2698, 1235, 0 },
   { 3173, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1243, 1108, 0 },
   { 2698, 1235, 0 },
   { 3173, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2659, 1049, 0 },
   { 2667, 1046, 0 },
   { 2698, 1235, 0 },
   { 3173, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2660, 1046, 0 },
   { 2668, 1049, 0 },
   { 2698, 1235, 0 },
   { 3173, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2657, 1055, 0 },
   { 2665, 1052, 0 },
   { 2698, 1235, 0 },
   { 3173, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2658, 1052, 0 },
   { 2666, 1055, 0 },
   { 2698, 1235, 0 },
   { 3173, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1091, 1092, 0 },
   { 2674, 1257, 0 },
   { 2698, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1191, 1192, 0 },
   { 2674, 1257, 0 },
   { 2698, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1163, 1162, 0 },
   { 1406, 1162, 0 },
   { 1410, 1162, 0 },
   { 1414, 1162, 0 },
   { 2698, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2698, 1235, 0 },
   { 5014, 5016, 116 },
   { 5031, 5016, 116 },
   { 5039, 5040, 0 },
   { 5053, 5055, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1171, 1172, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2305, 1077, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2299, 2300, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1097, 1098, 0 },
   { 1171, 1172, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1252, 1253, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1116, 1118, 0 },
   { 1126, 1127, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { 5543, 5548, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2305, 1077, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1250, 1251, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2299, 2300, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1095, 1096, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1088, 1078, 0 },
   { 1171, 1172, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1238, 1077, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1187, 1188, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1187, 1188, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1171, 1172, 0 },
   { 1187, 1188, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1187, 1188, 0 },
   { 2305, 1077, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1187, 1188, 0 },
   { 2299, 2300, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1187, 1188, 0 },
   { 1196, 1199, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { 5535, 5538, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2301, 1078, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2301, 1078, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1171, 1172, 0 },
   { 2301, 1078, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2301, 1078, 0 },
   { 2305, 1077, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2299, 2300, 0 },
   { 2301, 1078, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1248, 1249, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2306, 1077, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2306, 1077, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1171, 1172, 0 },
   { 2306, 1077, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2305, 1077, 0 },
   { 2306, 1077, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2299, 2300, 0 },
   { 2306, 1077, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { 5021, 5022, 116 },
   { 5028, 5029, 116 },
   { 5034, 5022, 116 },
   { 5038, 5029, 116 },
   { 5045, 5046, 0 },
   { 5050, 5051, 0 },
   { 5060, 5061, 0 },
   { 5066, 5067, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1173, 1174, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2303, 1235, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2307, 2308, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1101, 1102, 0 },
   { 1173, 1174, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1256, 1257, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1236, 1237, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { 5553, 5556, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1130, 1131, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2303, 1235, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1246, 1247, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2307, 2308, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1099, 1100, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1089, 1090, 0 },
   { 1173, 1174, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1234, 1235, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1189, 1190, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1189, 1190, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1173, 1174, 0 },
   { 1189, 1190, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1189, 1190, 0 },
   { 2303, 1235, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1189, 1190, 0 },
   { 2307, 2308, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1189, 1190, 0 },
   { 1205, 1206, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1121, 1046, 0 },
   { 1125, 1102, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { 5549, 5552, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2309, 1090, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2309, 1090, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1173, 1174, 0 },
   { 2309, 1090, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2303, 1235, 0 },
   { 2309, 1090, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2307, 2308, 0 },
   { 2309, 1090, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1244, 1245, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2304, 1235, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2304, 1235, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1173, 1174, 0 },
   { 2304, 1235, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2303, 1235, 0 },
   { 2304, 1235, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2304, 1235, 0 },
   { 2307, 2308, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { 5019, 5020, 116 },
   { 5025, 5027, 116 },
   { 5033, 5020, 116 },
   { 5037, 5027, 116 },
   { 5043, 5044, 0 },
   { 5048, 5049, 0 },
   { 5058, 5059, 0 },
   { 5063, 5065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2673, 1253, 0 },
   { 2699, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1193, 1194, 0 },
   { 2699, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1103, 1104, 0 },
   { 2673, 1253, 0 },
   { 2699, 1077, 0 },
   { 3172, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1232, 1110, 0 },
   { 2699, 1077, 0 },
   { 3172, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1193, 1194, 0 },
   { 2699, 1077, 0 },
   { 3172, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1193, 1194, 0 },
   { 1218, 1219, 0 },
   { 2699, 1077, 0 },
   { 3172, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1123, 1052, 0 },
   { 2699, 1077, 0 },
   { 3172, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2699, 1077, 0 },
   { 3172, 1235, 0 },
   { 5523, 5526, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1227, 760, 0 },
   { 2699, 1077, 0 },
   { 3172, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1242, 1110, 0 },
   { 2699, 1077, 0 },
   { 3172, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2663, 1049, 0 },
   { 2671, 1046, 0 },
   { 2699, 1077, 0 },
   { 3172, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2664, 1046, 0 },
   { 2672, 1049, 0 },
   { 2699, 1077, 0 },
   { 3172, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2661, 1055, 0 },
   { 2669, 1052, 0 },
   { 2699, 1077, 0 },
   { 3172, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2662, 1052, 0 },
   { 2670, 1055, 0 },
   { 2699, 1077, 0 },
   { 3172, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1093, 1094, 0 },
   { 2673, 1253, 0 },
   { 2699, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1193, 1194, 0 },
   { 2673, 1253, 0 },
   { 2699, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1152, 1151, 0 },
   { 1407, 1391, 0 },
   { 1411, 1391, 0 },
   { 1415, 1391, 0 },
   { 2699, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2699, 1077, 0 },
   { 5017, 5018, 116 },
   { 5032, 5018, 116 },
   { 5041, 5042, 0 },
   { 5056, 5057, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1502, 1254, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1507, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1352, 1283, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1502, 1254, 0 },
   { 3157, 217, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1507, 32, 0 },
   { 3157, 217, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1571, 1572, 0 },
   { 3157, 217, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1275, 0 },
   { 1335, 32, 0 },
   { 3157, 217, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1352, 1283, 0 },
   { 3157, 217, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 3157, 217, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1517, 217, 38 },
   { 1521, 1522, 38 },
   { 3157, 217, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1502, 1254, 0 },
   { 1507, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1352, 1283, 0 },
   { 1502, 1254, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1502, 1254, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1352, 1283, 0 },
   { 1507, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1507, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1268, 1266, 0 },
   { 1335, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1352, 1283, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 1513, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 1511, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 1346, 1281, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 1513, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 1511, 103, 0 },
   { 1515, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 1569, 1570, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 1519, 217, 38 },
   { 1524, 1522, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 1346, 1281, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 1511, 103, 0 },
   { 1513, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 1346, 1281, 0 },
   { 1513, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 1346, 1281, 0 },
   { 1511, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1262, 1263, 0 },
   { 1336, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 101, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5884, 5887, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2130, 6, 0 },
   { 2161, 11, 0 },
   { 2191, 16, 0 },
   { 2221, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 108, 106, 0 },
   { 3154, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 107, 106, 0 },
   { 3154, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3154, 32, 0 },
   { 5888, 5891, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1222, 1223, 0 },
   { 2691, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 200, 202, 0 },
   { 2691, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1107, 1108, 0 },
   { 2691, 1235, 0 },
   { 3040, 760, 0 },
   { 3043, 1131, 0 },
   { 3192, 1077, 0 },
   { 3792, 3793, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1111, 1108, 0 },
   { 2691, 1235, 0 },
   { 3040, 760, 0 },
   { 3043, 1131, 0 },
   { 3192, 1077, 0 },
   { 3792, 3793, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1222, 1223, 0 },
   { 2691, 1235, 0 },
   { 3040, 760, 0 },
   { 3043, 1131, 0 },
   { 3192, 1077, 0 },
   { 3792, 3793, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 200, 202, 0 },
   { 1421, 1422, 0 },
   { 1426, 1425, 0 },
   { 1443, 1444, 0 },
   { 1448, 1447, 0 },
   { 1465, 1466, 0 },
   { 1470, 1469, 0 },
   { 1485, 1486, 0 },
   { 1490, 1489, 0 },
   { 2691, 1235, 0 },
   { 3040, 760, 0 },
   { 3043, 1131, 0 },
   { 3192, 1077, 0 },
   { 3792, 3793, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2691, 1235, 0 },
   { 3040, 760, 0 },
   { 3043, 1131, 0 },
   { 3192, 1077, 0 },
   { 3792, 3793, 81 },
   { 5527, 5530, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1230, 1131, 0 },
   { 2691, 1235, 0 },
   { 3040, 760, 0 },
   { 3043, 1131, 0 },
   { 3192, 1077, 0 },
   { 3792, 3793, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1429, 1422, 0 },
   { 1431, 1425, 0 },
   { 1451, 1444, 0 },
   { 1453, 1447, 0 },
   { 1473, 1466, 0 },
   { 1474, 1469, 0 },
   { 1493, 1486, 0 },
   { 1495, 1489, 0 },
   { 2691, 1235, 0 },
   { 3040, 760, 0 },
   { 3043, 1131, 0 },
   { 3192, 1077, 0 },
   { 3792, 3793, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1434, 1422, 0 },
   { 1436, 1425, 0 },
   { 1456, 1444, 0 },
   { 1458, 1447, 0 },
   { 1477, 1466, 0 },
   { 1478, 1469, 0 },
   { 1498, 1486, 0 },
   { 1500, 1489, 0 },
   { 2691, 1235, 0 },
   { 3040, 760, 0 },
   { 3043, 1131, 0 },
   { 3192, 1077, 0 },
   { 3792, 3793, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2691, 1235, 0 },
   { 3040, 760, 0 },
   { 3043, 1131, 0 },
   { 3192, 1077, 0 },
   { 3792, 3793, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2691, 1235, 0 },
   { 3040, 760, 0 },
   { 3043, 1131, 0 },
   { 3192, 1077, 0 },
   { 3792, 3793, 81 },
   { 4377, 4382, 115 },
   { 4424, 4429, 115 },
   { 4470, 4475, 115 },
   { 4537, 4542, 116 },
   { 4588, 4593, 116 },
   { 4666, 4671, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2691, 1235, 0 },
   { 3040, 760, 0 },
   { 3043, 1131, 0 },
   { 3192, 1077, 0 },
   { 3792, 3793, 81 },
   { 4720, 4723, 115 },
   { 4760, 4764, 115 },
   { 4802, 4807, 115 },
   { 4866, 4871, 116 },
   { 4911, 4916, 116 },
   { 4983, 4988, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2691, 1235, 0 },
   { 3040, 760, 0 },
   { 3043, 1131, 0 },
   { 3192, 1077, 0 },
   { 3792, 3793, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2691, 1235, 0 },
   { 3040, 760, 0 },
   { 3043, 1131, 0 },
   { 3192, 1077, 0 },
   { 3792, 3793, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2691, 1235, 0 },
   { 3040, 760, 0 },
   { 3043, 1131, 0 },
   { 3192, 1077, 0 },
   { 3792, 3793, 81 },
   { 3877, 3874, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2691, 1235, 0 },
   { 3040, 760, 0 },
   { 3043, 1131, 0 },
   { 3192, 1077, 0 },
   { 3792, 3793, 81 },
   { 3873, 3874, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2691, 1235, 0 },
   { 3040, 760, 0 },
   { 3043, 1131, 0 },
   { 3192, 1077, 0 },
   { 3792, 3793, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2691, 1235, 0 },
   { 3040, 760, 0 },
   { 3043, 1131, 0 },
   { 3192, 1077, 0 },
   { 3792, 3793, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2691, 1235, 0 },
   { 3040, 760, 0 },
   { 3043, 1131, 0 },
   { 3192, 1077, 0 },
   { 3792, 3793, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2691, 1235, 0 },
   { 3040, 760, 0 },
   { 3043, 1131, 0 },
   { 3192, 1077, 0 },
   { 3792, 3793, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2691, 1235, 0 },
   { 3040, 760, 0 },
   { 3043, 1131, 0 },
   { 3192, 1077, 0 },
   { 3792, 3793, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2691, 1235, 0 },
   { 3040, 760, 0 },
   { 3043, 1131, 0 },
   { 3192, 1077, 0 },
   { 3792, 3793, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 200, 202, 0 },
   { 1222, 1223, 0 },
   { 2691, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1389, 1162, 0 },
   { 1394, 1162, 0 },
   { 1398, 1162, 0 },
   { 1402, 1162, 0 },
   { 2691, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2691, 1235, 0 },
   { 4359, 4361, 115 },
   { 4416, 4361, 115 },
   { 4461, 4361, 115 },
   { 4508, 4509, 116 },
   { 4522, 4524, 116 },
   { 4573, 4509, 116 },
   { 4581, 4524, 116 },
   { 4625, 4626, 0 },
   { 4638, 4639, 0 },
   { 4651, 4653, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2691, 1235, 0 },
   { 4704, 4706, 115 },
   { 4752, 4706, 115 },
   { 4794, 4706, 115 },
   { 4838, 4839, 116 },
   { 4851, 4853, 116 },
   { 4896, 4839, 116 },
   { 4904, 4853, 116 },
   { 4943, 4944, 0 },
   { 4956, 4957, 0 },
   { 4968, 4970, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1224, 1225, 0 },
   { 2692, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 203, 204, 0 },
   { 2692, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1109, 1110, 0 },
   { 2692, 1077, 0 },
   { 3041, 1131, 0 },
   { 3042, 760, 0 },
   { 3191, 1235, 0 },
   { 3789, 3791, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1112, 1110, 0 },
   { 2692, 1077, 0 },
   { 3041, 1131, 0 },
   { 3042, 760, 0 },
   { 3191, 1235, 0 },
   { 3789, 3791, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1224, 1225, 0 },
   { 2692, 1077, 0 },
   { 3041, 1131, 0 },
   { 3042, 760, 0 },
   { 3191, 1235, 0 },
   { 3789, 3791, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 203, 204, 0 },
   { 1423, 1422, 0 },
   { 1424, 1425, 0 },
   { 1445, 1444, 0 },
   { 1446, 1447, 0 },
   { 1467, 1466, 0 },
   { 1468, 1469, 0 },
   { 1487, 1486, 0 },
   { 1488, 1489, 0 },
   { 2692, 1077, 0 },
   { 3041, 1131, 0 },
   { 3042, 760, 0 },
   { 3191, 1235, 0 },
   { 3789, 3791, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2692, 1077, 0 },
   { 3041, 1131, 0 },
   { 3042, 760, 0 },
   { 3191, 1235, 0 },
   { 3789, 3791, 81 },
   { 5531, 5534, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1231, 760, 0 },
   { 2692, 1077, 0 },
   { 3041, 1131, 0 },
   { 3042, 760, 0 },
   { 3191, 1235, 0 },
   { 3789, 3791, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1428, 1425, 0 },
   { 1432, 1422, 0 },
   { 1450, 1447, 0 },
   { 1454, 1444, 0 },
   { 1472, 1469, 0 },
   { 1475, 1466, 0 },
   { 1492, 1489, 0 },
   { 1496, 1486, 0 },
   { 2692, 1077, 0 },
   { 3041, 1131, 0 },
   { 3042, 760, 0 },
   { 3191, 1235, 0 },
   { 3789, 3791, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1433, 1425, 0 },
   { 1437, 1422, 0 },
   { 1455, 1447, 0 },
   { 1459, 1444, 0 },
   { 1476, 1469, 0 },
   { 1479, 1466, 0 },
   { 1497, 1489, 0 },
   { 1501, 1486, 0 },
   { 2692, 1077, 0 },
   { 3041, 1131, 0 },
   { 3042, 760, 0 },
   { 3191, 1235, 0 },
   { 3789, 3791, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2692, 1077, 0 },
   { 3041, 1131, 0 },
   { 3042, 760, 0 },
   { 3191, 1235, 0 },
   { 3789, 3791, 81 },
   { 4383, 4386, 115 },
   { 4430, 4433, 115 },
   { 4476, 4479, 115 },
   { 4543, 4546, 116 },
   { 4594, 4597, 116 },
   { 4672, 4675, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2692, 1077, 0 },
   { 3041, 1131, 0 },
   { 3042, 760, 0 },
   { 3191, 1235, 0 },
   { 3789, 3791, 81 },
   { 4724, 4727, 115 },
   { 4765, 4768, 115 },
   { 4808, 4811, 115 },
   { 4872, 4875, 116 },
   { 4917, 4920, 116 },
   { 4989, 4992, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2692, 1077, 0 },
   { 3041, 1131, 0 },
   { 3042, 760, 0 },
   { 3191, 1235, 0 },
   { 3789, 3791, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2692, 1077, 0 },
   { 3041, 1131, 0 },
   { 3042, 760, 0 },
   { 3191, 1235, 0 },
   { 3789, 3791, 81 },
   { 3876, 3872, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2692, 1077, 0 },
   { 3041, 1131, 0 },
   { 3042, 760, 0 },
   { 3191, 1235, 0 },
   { 3789, 3791, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2692, 1077, 0 },
   { 3041, 1131, 0 },
   { 3042, 760, 0 },
   { 3191, 1235, 0 },
   { 3789, 3791, 81 },
   { 3867, 3872, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2692, 1077, 0 },
   { 3041, 1131, 0 },
   { 3042, 760, 0 },
   { 3191, 1235, 0 },
   { 3789, 3791, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2692, 1077, 0 },
   { 3041, 1131, 0 },
   { 3042, 760, 0 },
   { 3191, 1235, 0 },
   { 3789, 3791, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2692, 1077, 0 },
   { 3041, 1131, 0 },
   { 3042, 760, 0 },
   { 3191, 1235, 0 },
   { 3789, 3791, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2692, 1077, 0 },
   { 3041, 1131, 0 },
   { 3042, 760, 0 },
   { 3191, 1235, 0 },
   { 3789, 3791, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2692, 1077, 0 },
   { 3041, 1131, 0 },
   { 3042, 760, 0 },
   { 3191, 1235, 0 },
   { 3789, 3791, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2692, 1077, 0 },
   { 3041, 1131, 0 },
   { 3042, 760, 0 },
   { 3191, 1235, 0 },
   { 3789, 3791, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 203, 204, 0 },
   { 1224, 1225, 0 },
   { 2692, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1390, 1391, 0 },
   { 1395, 1391, 0 },
   { 1399, 1391, 0 },
   { 1403, 1391, 0 },
   { 2692, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2692, 1077, 0 },
   { 4362, 4363, 115 },
   { 4417, 4363, 115 },
   { 4462, 4363, 115 },
   { 4510, 4511, 116 },
   { 4525, 4526, 116 },
   { 4574, 4511, 116 },
   { 4582, 4526, 116 },
   { 4627, 4628, 0 },
   { 4640, 4641, 0 },
   { 4654, 4655, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2692, 1077, 0 },
   { 4707, 4708, 115 },
   { 4753, 4708, 115 },
   { 4795, 4708, 115 },
   { 4840, 4841, 116 },
   { 4854, 4855, 116 },
   { 4897, 4841, 116 },
   { 4905, 4855, 116 },
   { 4945, 4946, 0 },
   { 4958, 4959, 0 },
   { 4971, 4972, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { 1503, 1504, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { 1508, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { 1356, 1303, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { 1503, 1504, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { 1508, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { 1356, 1303, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { 4000, 3998, 103 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { 1503, 1504, 0 },
   { 1508, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { 1356, 1303, 0 },
   { 1503, 1504, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { 1503, 1504, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { 1356, 1303, 0 },
   { 1508, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { 1508, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { 1356, 1303, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1338, 32, 0 },
   { 1514, 106, 39 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1338, 32, 0 },
   { 1512, 106, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1338, 32, 0 },
   { 1350, 1301, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1338, 32, 0 },
   { 1514, 106, 39 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1338, 32, 0 },
   { 1512, 106, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1338, 32, 0 },
   { 4002, 3998, 103 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1338, 32, 0 },
   { 1350, 1301, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1338, 32, 0 },
   { 1512, 106, 0 },
   { 1514, 106, 39 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1338, 32, 0 },
   { 1350, 1301, 0 },
   { 1514, 106, 39 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1338, 32, 0 },
   { 1350, 1301, 0 },
   { 1512, 106, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1787, 11, 0 },
   { 1892, 16, 0 },
   { 1972, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4226, 3095, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3115, 3116, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1785, 11, 0 },
   { 1890, 16, 0 },
   { 1970, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1813, 1814, 0 },
   { 1836, 1837, 0 },
   { 1918, 1919, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1996, 1997, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1789, 11, 0 },
   { 1894, 16, 0 },
   { 1974, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4225, 3095, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3117, 3118, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1788, 11, 0 },
   { 1893, 16, 0 },
   { 1973, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1811, 1812, 0 },
   { 1834, 1835, 0 },
   { 1916, 1917, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1994, 1995, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1816, 1817, 0 },
   { 1839, 1817, 0 },
   { 1920, 1815, 0 },
   { 4241, 11, 0 },
   { 4242, 16, 0 },
   { 4243, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1808, 1404, 0 },
   { 1831, 1404, 0 },
   { 1913, 1408, 0 },
   { 4224, 1528, 0 },
   { 4241, 11, 0 },
   { 4242, 16, 0 },
   { 4243, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4241, 11, 0 },
   { 4242, 16, 0 },
   { 4243, 21, 0 },
   { 5946, 5948, 121 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4241, 11, 0 },
   { 4242, 16, 0 },
   { 4243, 21, 0 },
   { 5930, 5932, 121 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4241, 11, 0 },
   { 4242, 16, 0 },
   { 4243, 21, 0 },
   { 5917, 5918, 121 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1819, 1820, 0 },
   { 1841, 1820, 0 },
   { 1921, 1818, 0 },
   { 4241, 11, 0 },
   { 4242, 16, 0 },
   { 4243, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1803, 11, 0 },
   { 1826, 11, 0 },
   { 1908, 16, 0 },
   { 3145, 16, 0 },
   { 4241, 11, 0 },
   { 4242, 16, 0 },
   { 4243, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3138, 16, 0 },
   { 4241, 11, 0 },
   { 4242, 16, 0 },
   { 4243, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3143, 1990, 0 },
   { 4241, 11, 0 },
   { 4242, 16, 0 },
   { 4243, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3144, 1987, 0 },
   { 4241, 11, 0 },
   { 4242, 16, 0 },
   { 4243, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1988, 1989, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3132, 1404, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5949, 5951, 121 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5933, 5935, 121 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5919, 5921, 121 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1991, 1992, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3121, 11, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3120, 1257, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1069, 1070, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1072, 1073, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5892, 5895, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1051, 1052, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2714, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1045, 1046, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1048, 1049, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1054, 1055, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1057, 1058, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1060, 1061, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1063, 1064, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1066, 1067, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 291, 0, 0 },
   { 466, 467, 14 },
   { 541, 542, 16 },
   { 607, 608, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 291, 0, 0 },
   { 466, 467, 14 },
   { 541, 542, 16 },
   { 607, 608, 18 },
   { 685, 687, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 291, 0, 0 },
   { 294, 296, 0 },
   { 470, 472, 14 },
   { 543, 545, 16 },
   { 611, 613, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 289, 0, 0 },
   { 290, 3, 0 },
   { 291, 0, 0 },
   { 294, 296, 0 },
   { 470, 472, 14 },
   { 543, 545, 16 },
   { 611, 613, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 291, 0, 0 },
   { 294, 296, 0 },
   { 466, 467, 14 },
   { 470, 472, 14 },
   { 541, 542, 16 },
   { 543, 545, 16 },
   { 607, 608, 18 },
   { 611, 613, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 291, 0, 0 },
   { 480, 483, 14 },
   { 552, 554, 16 },
   { 621, 624, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 289, 0, 0 },
   { 290, 3, 0 },
   { 291, 0, 0 },
   { 480, 483, 14 },
   { 552, 554, 16 },
   { 621, 624, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 291, 0, 0 },
   { 466, 467, 14 },
   { 480, 483, 14 },
   { 541, 542, 16 },
   { 552, 554, 16 },
   { 607, 608, 18 },
   { 621, 624, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 291, 0, 0 },
   { 292, 264, 0 },
   { 466, 467, 14 },
   { 541, 542, 16 },
   { 607, 608, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 291, 0, 0 },
   { 292, 264, 0 },
   { 466, 467, 14 },
   { 541, 542, 16 },
   { 607, 608, 18 },
   { 685, 687, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 291, 0, 0 },
   { 292, 264, 0 },
   { 294, 296, 0 },
   { 470, 472, 14 },
   { 543, 545, 16 },
   { 611, 613, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 289, 0, 0 },
   { 290, 3, 0 },
   { 291, 0, 0 },
   { 292, 264, 0 },
   { 294, 296, 0 },
   { 470, 472, 14 },
   { 543, 545, 16 },
   { 611, 613, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 291, 0, 0 },
   { 292, 264, 0 },
   { 294, 296, 0 },
   { 466, 467, 14 },
   { 470, 472, 14 },
   { 541, 542, 16 },
   { 543, 545, 16 },
   { 607, 608, 18 },
   { 611, 613, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 291, 0, 0 },
   { 292, 264, 0 },
   { 480, 483, 14 },
   { 552, 554, 16 },
   { 621, 624, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 289, 0, 0 },
   { 290, 3, 0 },
   { 291, 0, 0 },
   { 292, 264, 0 },
   { 480, 483, 14 },
   { 552, 554, 16 },
   { 621, 624, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 291, 0, 0 },
   { 292, 264, 0 },
   { 466, 467, 14 },
   { 480, 483, 14 },
   { 541, 542, 16 },
   { 552, 554, 16 },
   { 607, 608, 18 },
   { 621, 624, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 291, 0, 0 },
   { 294, 296, 0 },
   { 470, 472, 14 },
   { 476, 478, 14 },
   { 543, 545, 16 },
   { 548, 550, 16 },
   { 611, 613, 18 },
   { 617, 619, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 289, 0, 0 },
   { 290, 3, 0 },
   { 291, 0, 0 },
   { 294, 296, 0 },
   { 470, 472, 14 },
   { 476, 478, 14 },
   { 543, 545, 16 },
   { 548, 550, 16 },
   { 611, 613, 18 },
   { 617, 619, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 291, 0, 0 },
   { 294, 296, 0 },
   { 466, 467, 14 },
   { 470, 472, 14 },
   { 476, 478, 14 },
   { 541, 542, 16 },
   { 543, 545, 16 },
   { 548, 550, 16 },
   { 607, 608, 18 },
   { 611, 613, 18 },
   { 617, 619, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 291, 0, 0 },
   { 480, 483, 14 },
   { 552, 554, 16 },
   { 621, 624, 18 },
   { 681, 683, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 289, 0, 0 },
   { 290, 3, 0 },
   { 291, 0, 0 },
   { 480, 483, 14 },
   { 552, 554, 16 },
   { 621, 624, 18 },
   { 681, 683, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 291, 0, 0 },
   { 466, 467, 14 },
   { 480, 483, 14 },
   { 541, 542, 16 },
   { 552, 554, 16 },
   { 607, 608, 18 },
   { 621, 624, 18 },
   { 681, 683, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2337, 1077, 0 },
   { 2693, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2330, 1077, 0 },
   { 2693, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2318, 2319, 0 },
   { 2693, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2337, 1077, 0 },
   { 2693, 1077, 0 },
   { 3200, 1235, 0 },
   { 3794, 3791, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2693, 1077, 0 },
   { 3200, 1235, 0 },
   { 3794, 3791, 81 },
   { 5577, 5580, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2330, 1077, 0 },
   { 2693, 1077, 0 },
   { 3200, 1235, 0 },
   { 3794, 3791, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2330, 1077, 0 },
   { 2382, 2383, 0 },
   { 2693, 1077, 0 },
   { 3200, 1235, 0 },
   { 3794, 3791, 81 },
   { 3836, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2318, 2319, 0 },
   { 2693, 1077, 0 },
   { 3200, 1235, 0 },
   { 3794, 3791, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2318, 2319, 0 },
   { 2359, 2361, 0 },
   { 2693, 1077, 0 },
   { 3200, 1235, 0 },
   { 3794, 3791, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2693, 1077, 0 },
   { 3200, 1235, 0 },
   { 3794, 3791, 81 },
   { 4734, 4739, 115 },
   { 4775, 4780, 115 },
   { 4819, 4825, 115 },
   { 4881, 4885, 116 },
   { 4927, 4932, 116 },
   { 4998, 5002, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2693, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2693, 1077, 0 },
   { 3201, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2337, 1077, 0 },
   { 2693, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2330, 1077, 0 },
   { 2693, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2318, 2319, 0 },
   { 2693, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2339, 1077, 0 },
   { 2693, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2339, 1077, 0 },
   { 2693, 1077, 0 },
   { 3201, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2337, 1077, 0 },
   { 2339, 1077, 0 },
   { 2693, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2330, 1077, 0 },
   { 2339, 1077, 0 },
   { 2693, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2318, 2319, 0 },
   { 2339, 1077, 0 },
   { 2693, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2693, 1077, 0 },
   { 3201, 1077, 0 },
   { 5573, 5576, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2320, 1085, 0 },
   { 2693, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2320, 1085, 0 },
   { 2693, 1077, 0 },
   { 3201, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2320, 1085, 0 },
   { 2337, 1077, 0 },
   { 2693, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2320, 1085, 0 },
   { 2330, 1077, 0 },
   { 2693, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2318, 2319, 0 },
   { 2320, 1085, 0 },
   { 2693, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2320, 1085, 0 },
   { 2363, 2365, 0 },
   { 2693, 1077, 0 },
   { 3201, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2331, 1077, 0 },
   { 2693, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2331, 1077, 0 },
   { 2693, 1077, 0 },
   { 3201, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2331, 1077, 0 },
   { 2337, 1077, 0 },
   { 2693, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2330, 1077, 0 },
   { 2331, 1077, 0 },
   { 2693, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2318, 2319, 0 },
   { 2331, 1077, 0 },
   { 2693, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2331, 1077, 0 },
   { 2384, 2385, 0 },
   { 2693, 1077, 0 },
   { 3201, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2693, 1077, 0 },
   { 3201, 1077, 0 },
   { 4728, 4733, 115 },
   { 4769, 4774, 115 },
   { 4812, 4818, 115 },
   { 4876, 4880, 116 },
   { 4921, 4926, 116 },
   { 4993, 4997, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2693, 1077, 0 },
   { 4711, 4712, 115 },
   { 4718, 4719, 115 },
   { 4755, 4712, 115 },
   { 4759, 4719, 115 },
   { 4797, 4712, 115 },
   { 4801, 4719, 115 },
   { 4844, 4845, 116 },
   { 4849, 4850, 116 },
   { 4858, 4859, 116 },
   { 4864, 4865, 116 },
   { 4899, 4845, 116 },
   { 4902, 4850, 116 },
   { 4907, 4859, 116 },
   { 4910, 4865, 116 },
   { 4949, 4950, 0 },
   { 4954, 4955, 0 },
   { 4962, 4963, 0 },
   { 4966, 4967, 0 },
   { 4975, 4976, 0 },
   { 4981, 4982, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2340, 1235, 0 },
   { 2694, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2322, 1235, 0 },
   { 2694, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2332, 2333, 0 },
   { 2694, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2340, 1235, 0 },
   { 2694, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2694, 1235, 0 },
   { 5585, 5588, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2322, 1235, 0 },
   { 2694, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2322, 1235, 0 },
   { 2367, 2369, 0 },
   { 2694, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2332, 2333, 0 },
   { 2694, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2332, 2333, 0 },
   { 2386, 2387, 0 },
   { 2694, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2694, 1235, 0 },
   { 4746, 4751, 115 },
   { 4787, 4792, 115 },
   { 4832, 4837, 115 },
   { 4891, 4895, 116 },
   { 4938, 4942, 116 },
   { 5008, 5012, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2694, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2340, 1235, 0 },
   { 2694, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2322, 1235, 0 },
   { 2694, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2332, 2333, 0 },
   { 2694, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2341, 1235, 0 },
   { 2694, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2340, 1235, 0 },
   { 2341, 1235, 0 },
   { 2694, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2322, 1235, 0 },
   { 2341, 1235, 0 },
   { 2694, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2332, 2333, 0 },
   { 2341, 1235, 0 },
   { 2694, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2694, 1235, 0 },
   { 5581, 5584, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2334, 2335, 0 },
   { 2694, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2334, 2335, 0 },
   { 2340, 1235, 0 },
   { 2694, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2322, 1235, 0 },
   { 2334, 2335, 0 },
   { 2694, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2332, 2333, 0 },
   { 2334, 2335, 0 },
   { 2694, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2334, 2335, 0 },
   { 2388, 2389, 0 },
   { 2694, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2323, 1235, 0 },
   { 2694, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2323, 1235, 0 },
   { 2340, 1235, 0 },
   { 2694, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2322, 1235, 0 },
   { 2323, 1235, 0 },
   { 2694, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2323, 1235, 0 },
   { 2332, 2333, 0 },
   { 2694, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2323, 1235, 0 },
   { 2371, 2373, 0 },
   { 2694, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2694, 1235, 0 },
   { 4740, 4745, 115 },
   { 4781, 4786, 115 },
   { 4826, 4831, 115 },
   { 4886, 4890, 116 },
   { 4933, 4937, 116 },
   { 5003, 5007, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2694, 1235, 0 },
   { 4709, 4710, 115 },
   { 4715, 4717, 115 },
   { 4754, 4710, 115 },
   { 4758, 4717, 115 },
   { 4796, 4710, 115 },
   { 4800, 4717, 115 },
   { 4842, 4843, 116 },
   { 4847, 4848, 116 },
   { 4856, 4857, 116 },
   { 4861, 4863, 116 },
   { 4898, 4843, 116 },
   { 4901, 4848, 116 },
   { 4906, 4857, 116 },
   { 4909, 4863, 116 },
   { 4947, 4948, 0 },
   { 4952, 4953, 0 },
   { 4960, 4961, 0 },
   { 4964, 4965, 0 },
   { 4973, 4974, 0 },
   { 4978, 4980, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2812, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { 1354, 1293, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { 1340, 61, 0 },
   { 1341, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { 1340, 61, 0 },
   { 1341, 32, 0 },
   { 1354, 1293, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { 1340, 61, 0 },
   { 1341, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { 1739, 1740, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { 1354, 1293, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { 3104, 2676, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { 1354, 1293, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3818, 3819, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3816, 2719, 0 },
   { 3818, 3819, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4231, 760, 0 },
   { 4232, 6, 0 },
   { 4233, 11, 0 },
   { 4234, 16, 0 },
   { 4235, 21, 0 },
   { 5068, 3307, 0 },
   { 5069, 3310, 0 },
   { 5071, 4929, 0 },
   { 5073, 2009, 0 },
   { 5074, 2025, 0 },
   { 5076, 4822, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1809, 1392, 0 },
   { 1832, 1392, 0 },
   { 1914, 1396, 0 },
   { 4230, 3095, 0 },
   { 4231, 760, 0 },
   { 4232, 6, 0 },
   { 4233, 11, 0 },
   { 4234, 16, 0 },
   { 4235, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4231, 760, 0 },
   { 4232, 6, 0 },
   { 4233, 11, 0 },
   { 4234, 16, 0 },
   { 4235, 21, 0 },
   { 5955, 5957, 121 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4231, 760, 0 },
   { 4232, 6, 0 },
   { 4233, 11, 0 },
   { 4234, 16, 0 },
   { 4235, 21, 0 },
   { 5939, 5941, 121 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4231, 760, 0 },
   { 4232, 6, 0 },
   { 4233, 11, 0 },
   { 4234, 16, 0 },
   { 4235, 21, 0 },
   { 5924, 5925, 121 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1805, 11, 0 },
   { 1828, 11, 0 },
   { 1910, 16, 0 },
   { 4231, 760, 0 },
   { 4232, 6, 0 },
   { 4233, 11, 0 },
   { 4234, 16, 0 },
   { 4235, 21, 0 },
   { 4249, 4250, 0 },
   { 4257, 4250, 0 },
   { 4263, 4250, 0 },
   { 4269, 4250, 0 },
   { 4273, 4274, 0 },
   { 4277, 4248, 0 },
   { 4281, 4248, 0 },
   { 4285, 4248, 0 },
   { 4289, 4290, 0 },
   { 4293, 4290, 0 },
   { 4295, 4296, 0 },
   { 4299, 4256, 0 },
   { 4303, 4256, 0 },
   { 4307, 4308, 0 },
   { 4311, 4308, 0 },
   { 4313, 4314, 0 },
   { 4316, 4308, 0 },
   { 4318, 4314, 0 },
   { 4320, 1804, 0 },
   { 4322, 4262, 0 },
   { 4326, 4327, 0 },
   { 4330, 4327, 0 },
   { 4332, 4333, 0 },
   { 4336, 4327, 0 },
   { 4338, 4333, 0 },
   { 4340, 1827, 0 },
   { 4342, 4327, 0 },
   { 4344, 4333, 0 },
   { 4346, 1827, 0 },
   { 4348, 1909, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4231, 760, 0 },
   { 4232, 6, 0 },
   { 4233, 11, 0 },
   { 4234, 16, 0 },
   { 4235, 21, 0 },
   { 4252, 4253, 0 },
   { 4259, 4253, 0 },
   { 4265, 4253, 0 },
   { 4270, 4253, 0 },
   { 4278, 4251, 0 },
   { 4282, 4251, 0 },
   { 4286, 4251, 0 },
   { 4300, 4258, 0 },
   { 4304, 4258, 0 },
   { 4323, 4264, 0 },
   { 5105, 11, 0 },
   { 5158, 16, 0 },
   { 5210, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3139, 16, 0 },
   { 4231, 760, 0 },
   { 4232, 6, 0 },
   { 4233, 11, 0 },
   { 4234, 16, 0 },
   { 4235, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3141, 2001, 0 },
   { 4231, 760, 0 },
   { 4232, 6, 0 },
   { 4233, 11, 0 },
   { 4234, 16, 0 },
   { 4235, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4231, 760, 0 },
   { 4232, 6, 0 },
   { 4233, 11, 0 },
   { 4234, 16, 0 },
   { 4235, 21, 0 },
   { 4350, 4351, 0 },
   { 4350, 4351, 0 },
   { 4352, 4353, 0 },
   { 4352, 4353, 0 },
   { 4354, 4234, 0 },
   { 4354, 4234, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4236, 760, 0 },
   { 4237, 6, 0 },
   { 4238, 11, 0 },
   { 4239, 16, 0 },
   { 4240, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1810, 1392, 0 },
   { 1833, 1392, 0 },
   { 1915, 1396, 0 },
   { 4228, 3095, 0 },
   { 4236, 760, 0 },
   { 4237, 6, 0 },
   { 4238, 11, 0 },
   { 4239, 16, 0 },
   { 4240, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4236, 760, 0 },
   { 4237, 6, 0 },
   { 4238, 11, 0 },
   { 4239, 16, 0 },
   { 4240, 21, 0 },
   { 5952, 5954, 121 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4236, 760, 0 },
   { 4237, 6, 0 },
   { 4238, 11, 0 },
   { 4239, 16, 0 },
   { 4240, 21, 0 },
   { 5936, 5938, 121 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4236, 760, 0 },
   { 4237, 6, 0 },
   { 4238, 11, 0 },
   { 4239, 16, 0 },
   { 4240, 21, 0 },
   { 5922, 5923, 121 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3252, 3254, 58 },
   { 4236, 760, 0 },
   { 4237, 6, 0 },
   { 4238, 11, 0 },
   { 4239, 16, 0 },
   { 4240, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4236, 760, 0 },
   { 4237, 6, 0 },
   { 4238, 11, 0 },
   { 4239, 16, 0 },
   { 4240, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4236, 760, 0 },
   { 4237, 6, 0 },
   { 4238, 11, 0 },
   { 4239, 16, 0 },
   { 4240, 21, 0 },
   { 4254, 4250, 0 },
   { 4260, 4250, 0 },
   { 4266, 4250, 0 },
   { 4271, 4250, 0 },
   { 4279, 4248, 0 },
   { 4283, 4248, 0 },
   { 4287, 4248, 0 },
   { 4301, 4256, 0 },
   { 4305, 4256, 0 },
   { 4324, 4262, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1807, 11, 0 },
   { 1830, 11, 0 },
   { 1912, 16, 0 },
   { 4236, 760, 0 },
   { 4237, 6, 0 },
   { 4238, 11, 0 },
   { 4239, 16, 0 },
   { 4240, 21, 0 },
   { 4255, 4253, 0 },
   { 4261, 4253, 0 },
   { 4267, 4253, 0 },
   { 4272, 4253, 0 },
   { 4275, 4276, 0 },
   { 4280, 4251, 0 },
   { 4284, 4251, 0 },
   { 4288, 4251, 0 },
   { 4291, 4292, 0 },
   { 4294, 4292, 0 },
   { 4297, 4298, 0 },
   { 4302, 4258, 0 },
   { 4306, 4258, 0 },
   { 4309, 4310, 0 },
   { 4312, 4310, 0 },
   { 4315, 3282, 0 },
   { 4317, 4310, 0 },
   { 4319, 3282, 0 },
   { 4321, 1806, 0 },
   { 4325, 4264, 0 },
   { 4328, 4329, 0 },
   { 4331, 4329, 0 },
   { 4334, 4335, 0 },
   { 4337, 4329, 0 },
   { 4339, 4335, 0 },
   { 4341, 1829, 0 },
   { 4343, 4329, 0 },
   { 4345, 4335, 0 },
   { 4347, 1829, 0 },
   { 4349, 1911, 0 },
   { 5077, 11, 0 },
   { 5130, 16, 0 },
   { 5183, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3140, 16, 0 },
   { 4236, 760, 0 },
   { 4237, 6, 0 },
   { 4238, 11, 0 },
   { 4239, 16, 0 },
   { 4240, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3248, 3250, 58 },
   { 4236, 760, 0 },
   { 4237, 6, 0 },
   { 4238, 11, 0 },
   { 4239, 16, 0 },
   { 4240, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3142, 1998, 0 },
   { 4236, 760, 0 },
   { 4237, 6, 0 },
   { 4238, 11, 0 },
   { 4239, 16, 0 },
   { 4240, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4236, 760, 0 },
   { 4237, 6, 0 },
   { 4238, 11, 0 },
   { 4239, 16, 0 },
   { 4240, 21, 0 },
   { 4355, 2025, 0 },
   { 4355, 2025, 0 },
   { 4356, 3310, 0 },
   { 4356, 3310, 0 },
   { 4357, 4239, 0 },
   { 4357, 4239, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2726, 3309, 63 },
   { 3324, 2726, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3213, 1460, 0 },
   { 2726, 3309, 63 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3210, 122, 0 },
   { 2726, 3309, 63 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3204, 36, 0 },
   { 2726, 3309, 63 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3207, 2430, 0 },
   { 2726, 3309, 63 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3226, 3310, 64 },
   { 3314, 3226, 0 },
   { 3226, 3310, 101 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3226, 3310, 64 },
   { 3328, 3226, 0 },
   { 3329, 3228, 0 },
   { 3226, 3310, 101 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3228, 3311, 64 },
   { 3321, 3228, 0 },
   { 3228, 3311, 101 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3821, 3814, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3823, 3814, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3823, 3814, 0 },
   { 3855, 122, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3821, 3814, 0 },
   { 3823, 3814, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3824, 3811, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3825, 3811, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3825, 3811, 0 },
   { 3856, 122, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3824, 3811, 0 },
   { 3825, 3811, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3312, 3313, 0 },
   { 3313, 3995, 101 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3327, 3320, 0 },
   { 3313, 3995, 101 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3325, 3313, 0 },
   { 3326, 3320, 0 },
   { 3313, 3995, 101 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3319, 3320, 0 },
   { 3320, 3996, 101 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3827, 3828, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3830, 3828, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3830, 3828, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3827, 3828, 0 },
   { 3830, 3828, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2464, 2469, 48 },
   { 5479, 2464, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2464, 2469, 48 },
   { 5480, 2534, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2470, 2476, 48 },
   { 5485, 2470, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2470, 2476, 48 },
   { 5486, 2538, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2477, 2483, 48 },
   { 5489, 2477, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2477, 2483, 48 },
   { 5490, 2544, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 2500, 48 },
   { 5493, 2484, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 2500, 48 },
   { 5494, 2549, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2501, 2533, 48 },
   { 5497, 2501, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2501, 2533, 48 },
   { 5498, 2558, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2252, 2253, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2575, 2578, 48 },
   { 5483, 2575, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2575, 2578, 48 },
   { 5484, 2616, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2579, 2584, 48 },
   { 5487, 2579, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2579, 2584, 48 },
   { 5488, 2620, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2585, 2589, 48 },
   { 5491, 2585, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2585, 2589, 48 },
   { 5492, 2626, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2590, 2598, 48 },
   { 5495, 2590, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2590, 2598, 48 },
   { 5496, 2631, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2599, 2615, 48 },
   { 5499, 2599, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2599, 2615, 48 },
   { 5500, 2640, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2324, 1077, 0 },
   { 2689, 1077, 0 },
   { 3196, 1077, 0 },
   { 3197, 1235, 0 },
   { 3198, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2311, 2312, 0 },
   { 2689, 1077, 0 },
   { 3196, 1077, 0 },
   { 3197, 1235, 0 },
   { 3198, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2689, 1077, 0 },
   { 3196, 1077, 0 },
   { 3197, 1235, 0 },
   { 3198, 1235, 0 },
   { 5561, 5564, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2324, 1077, 0 },
   { 2374, 2375, 0 },
   { 2689, 1077, 0 },
   { 3196, 1077, 0 },
   { 3197, 1235, 0 },
   { 3198, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2311, 2312, 0 },
   { 2343, 2345, 0 },
   { 2689, 1077, 0 },
   { 3196, 1077, 0 },
   { 3197, 1235, 0 },
   { 3198, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2689, 1077, 0 },
   { 3196, 1077, 0 },
   { 3197, 1235, 0 },
   { 3198, 1235, 0 },
   { 4396, 4402, 115 },
   { 4442, 4447, 115 },
   { 4489, 4495, 115 },
   { 4555, 4560, 116 },
   { 4606, 4612, 116 },
   { 4684, 4690, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2689, 1077, 0 },
   { 3196, 1077, 0 },
   { 3197, 1235, 0 },
   { 3198, 1235, 0 },
   { 1489, 3787, 81 },
   { 5557, 5560, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2313, 1083, 0 },
   { 2689, 1077, 0 },
   { 3196, 1077, 0 },
   { 3197, 1235, 0 },
   { 3198, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2313, 1083, 0 },
   { 2689, 1077, 0 },
   { 3196, 1077, 0 },
   { 3197, 1235, 0 },
   { 3198, 1235, 0 },
   { 3780, 3782, 80 },
   { 1489, 3787, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2313, 1083, 0 },
   { 2324, 1077, 0 },
   { 2689, 1077, 0 },
   { 3196, 1077, 0 },
   { 3197, 1235, 0 },
   { 3198, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2311, 2312, 0 },
   { 2313, 1083, 0 },
   { 2689, 1077, 0 },
   { 3196, 1077, 0 },
   { 3197, 1235, 0 },
   { 3198, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2313, 1083, 0 },
   { 2347, 2349, 0 },
   { 2689, 1077, 0 },
   { 3196, 1077, 0 },
   { 3197, 1235, 0 },
   { 3198, 1235, 0 },
   { 3780, 3782, 80 },
   { 1489, 3787, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2325, 1077, 0 },
   { 2689, 1077, 0 },
   { 3196, 1077, 0 },
   { 3197, 1235, 0 },
   { 3198, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2325, 1077, 0 },
   { 2689, 1077, 0 },
   { 3196, 1077, 0 },
   { 3197, 1235, 0 },
   { 3198, 1235, 0 },
   { 3776, 3778, 80 },
   { 1489, 3787, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2324, 1077, 0 },
   { 2325, 1077, 0 },
   { 2689, 1077, 0 },
   { 3196, 1077, 0 },
   { 3197, 1235, 0 },
   { 3198, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2311, 2312, 0 },
   { 2325, 1077, 0 },
   { 2689, 1077, 0 },
   { 3196, 1077, 0 },
   { 3197, 1235, 0 },
   { 3198, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2325, 1077, 0 },
   { 2376, 2377, 0 },
   { 2689, 1077, 0 },
   { 3196, 1077, 0 },
   { 3197, 1235, 0 },
   { 3198, 1235, 0 },
   { 3776, 3778, 80 },
   { 1489, 3787, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2689, 1077, 0 },
   { 3196, 1077, 0 },
   { 3197, 1235, 0 },
   { 3198, 1235, 0 },
   { 1489, 3787, 81 },
   { 4387, 4394, 115 },
   { 4434, 4441, 115 },
   { 4480, 4487, 115 },
   { 4547, 4554, 116 },
   { 4598, 4605, 116 },
   { 4676, 4683, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2689, 1077, 0 },
   { 3196, 1077, 0 },
   { 3197, 1235, 0 },
   { 3198, 1235, 0 },
   { 4366, 4367, 115 },
   { 4375, 4376, 115 },
   { 4419, 4367, 115 },
   { 4423, 4376, 115 },
   { 4464, 4367, 115 },
   { 4468, 4376, 115 },
   { 4514, 4515, 116 },
   { 4519, 4520, 116 },
   { 4529, 4530, 116 },
   { 4535, 4536, 116 },
   { 4576, 4515, 116 },
   { 4579, 4520, 116 },
   { 4584, 4530, 116 },
   { 4587, 4536, 116 },
   { 4631, 4632, 0 },
   { 4636, 4637, 0 },
   { 4644, 4645, 0 },
   { 4648, 4649, 0 },
   { 4658, 4659, 0 },
   { 4664, 4665, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2315, 1235, 0 },
   { 2690, 1235, 0 },
   { 3193, 1235, 0 },
   { 3194, 1077, 0 },
   { 3195, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2326, 2327, 0 },
   { 2690, 1235, 0 },
   { 3193, 1235, 0 },
   { 3194, 1077, 0 },
   { 3195, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2690, 1235, 0 },
   { 3193, 1235, 0 },
   { 3194, 1077, 0 },
   { 3195, 1077, 0 },
   { 5569, 5572, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2315, 1235, 0 },
   { 2351, 2353, 0 },
   { 2690, 1235, 0 },
   { 3193, 1235, 0 },
   { 3194, 1077, 0 },
   { 3195, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2326, 2327, 0 },
   { 2378, 2379, 0 },
   { 2690, 1235, 0 },
   { 3193, 1235, 0 },
   { 3194, 1077, 0 },
   { 3195, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2690, 1235, 0 },
   { 3193, 1235, 0 },
   { 3194, 1077, 0 },
   { 3195, 1077, 0 },
   { 4409, 4414, 115 },
   { 4454, 4459, 115 },
   { 4502, 4507, 115 },
   { 4567, 4572, 116 },
   { 4619, 4624, 116 },
   { 4697, 4702, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2690, 1235, 0 },
   { 3193, 1235, 0 },
   { 3194, 1077, 0 },
   { 3195, 1077, 0 },
   { 1486, 3788, 81 },
   { 5565, 5568, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2328, 2329, 0 },
   { 2690, 1235, 0 },
   { 3193, 1235, 0 },
   { 3194, 1077, 0 },
   { 3195, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2328, 2329, 0 },
   { 2690, 1235, 0 },
   { 3193, 1235, 0 },
   { 3194, 1077, 0 },
   { 3195, 1077, 0 },
   { 3785, 3786, 80 },
   { 1486, 3788, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2315, 1235, 0 },
   { 2328, 2329, 0 },
   { 2690, 1235, 0 },
   { 3193, 1235, 0 },
   { 3194, 1077, 0 },
   { 3195, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2326, 2327, 0 },
   { 2328, 2329, 0 },
   { 2690, 1235, 0 },
   { 3193, 1235, 0 },
   { 3194, 1077, 0 },
   { 3195, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2328, 2329, 0 },
   { 2380, 2381, 0 },
   { 2690, 1235, 0 },
   { 3193, 1235, 0 },
   { 3194, 1077, 0 },
   { 3195, 1077, 0 },
   { 3785, 3786, 80 },
   { 1486, 3788, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2316, 1235, 0 },
   { 2690, 1235, 0 },
   { 3193, 1235, 0 },
   { 3194, 1077, 0 },
   { 3195, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2316, 1235, 0 },
   { 2690, 1235, 0 },
   { 3193, 1235, 0 },
   { 3194, 1077, 0 },
   { 3195, 1077, 0 },
   { 3783, 3784, 80 },
   { 1486, 3788, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2315, 1235, 0 },
   { 2316, 1235, 0 },
   { 2690, 1235, 0 },
   { 3193, 1235, 0 },
   { 3194, 1077, 0 },
   { 3195, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2316, 1235, 0 },
   { 2326, 2327, 0 },
   { 2690, 1235, 0 },
   { 3193, 1235, 0 },
   { 3194, 1077, 0 },
   { 3195, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2316, 1235, 0 },
   { 2355, 2357, 0 },
   { 2690, 1235, 0 },
   { 3193, 1235, 0 },
   { 3194, 1077, 0 },
   { 3195, 1077, 0 },
   { 3783, 3784, 80 },
   { 1486, 3788, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2690, 1235, 0 },
   { 3193, 1235, 0 },
   { 3194, 1077, 0 },
   { 3195, 1077, 0 },
   { 1486, 3788, 81 },
   { 4403, 4408, 115 },
   { 4448, 4453, 115 },
   { 4496, 4501, 115 },
   { 4561, 4566, 116 },
   { 4613, 4618, 116 },
   { 4691, 4696, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2690, 1235, 0 },
   { 3193, 1235, 0 },
   { 3194, 1077, 0 },
   { 3195, 1077, 0 },
   { 4364, 4365, 115 },
   { 4372, 4374, 115 },
   { 4418, 4365, 115 },
   { 4422, 4374, 115 },
   { 4463, 4365, 115 },
   { 4467, 4374, 115 },
   { 4512, 4513, 116 },
   { 4517, 4518, 116 },
   { 4527, 4528, 116 },
   { 4532, 4534, 116 },
   { 4575, 4513, 116 },
   { 4578, 4518, 116 },
   { 4583, 4528, 116 },
   { 4586, 4534, 116 },
   { 4629, 4630, 0 },
   { 4634, 4635, 0 },
   { 4642, 4643, 0 },
   { 4646, 4647, 0 },
   { 4656, 4657, 0 },
   { 4661, 4663, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2690, 1235, 0 },
   { 3193, 1235, 0 },
   { 3194, 1077, 0 },
   { 3195, 1077, 0 },
   { 1486, 3788, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2690, 1235, 0 },
   { 3193, 1235, 0 },
   { 3194, 1077, 0 },
   { 3195, 1077, 0 },
   { 1486, 3788, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2690, 1235, 0 },
   { 3193, 1235, 0 },
   { 3194, 1077, 0 },
   { 3195, 1077, 0 },
   { 1486, 3788, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2690, 1235, 0 },
   { 3193, 1235, 0 },
   { 3194, 1077, 0 },
   { 3195, 1077, 0 },
   { 1486, 3788, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2690, 1235, 0 },
   { 3193, 1235, 0 },
   { 3194, 1077, 0 },
   { 3195, 1077, 0 },
   { 1486, 3788, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3998, 4000, 102 },
   { 5896, 5899, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2814, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2867, 2868, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2859, 2860, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2862, 2863, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2865, 2866, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2850, 2852, 0 },
   { 2832, 2885, 54 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2853, 2854, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2884, 2864, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2883, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2886, 2832, 55 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2855, 2858, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3134, 1392, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3124, 3125, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3135, 1392, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3126, 3127, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4227, 1528, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1821, 1822, 0 },
   { 1842, 1843, 0 },
   { 1922, 1923, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2005, 2006, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4229, 1528, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1783, 688, 0 },
   { 1888, 694, 0 },
   { 1968, 700, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1823, 1824, 0 },
   { 1844, 1845, 0 },
   { 1924, 1925, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2007, 2008, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1342, 32, 0 },
   { 1348, 1291, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1342, 32, 0 },
   { 1343, 32, 0 },
   { 1344, 63, 0 },
   { 1348, 1291, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1342, 32, 0 },
   { 3103, 2679, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3133, 1392, 0 },
   { 3133, 1392, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5958, 5960, 121 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5942, 5944, 121 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5926, 5928, 121 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2002, 2003, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1999, 2000, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3122, 11, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3123, 11, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3136, 1404, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3128, 3129, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3137, 1404, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3130, 3131, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 776, 778, 36 },
   { 788, 789, 0 },
   { 804, 805, 37 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 776, 778, 36 },
   { 783, 785, 0 },
   { 788, 789, 0 },
   { 804, 805, 37 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 776, 778, 36 },
   { 780, 264, 0 },
   { 783, 785, 0 },
   { 788, 789, 0 },
   { 804, 805, 37 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 793, 785, 0 },
   { 802, 803, 37 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 791, 264, 0 },
   { 793, 785, 0 },
   { 802, 803, 37 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 795, 264, 0 },
   { 800, 801, 37 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 795, 264, 0 },
   { 796, 799, 0 },
   { 800, 801, 37 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3297, 3298, 0 },
   { 3925, 3928, 92 },
   { 5184, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3299, 3298, 0 },
   { 3925, 3928, 92 },
   { 5184, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3345, 3341, 0 },
   { 3925, 3928, 92 },
   { 5184, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2400, 2406, 47 },
   { 2412, 2396, 0 },
   { 2420, 2398, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2400, 2406, 47 },
   { 2413, 2398, 0 },
   { 2421, 2396, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2400, 2406, 47 },
   { 2409, 2400, 0 },
   { 2418, 2402, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2400, 2406, 47 },
   { 2411, 2402, 0 },
   { 2419, 2400, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2402, 2407, 47 },
   { 2416, 2396, 0 },
   { 2424, 2398, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2402, 2407, 47 },
   { 2417, 2398, 0 },
   { 2425, 2396, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2402, 2407, 47 },
   { 2414, 2400, 0 },
   { 2422, 2402, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2402, 2407, 47 },
   { 2415, 2402, 0 },
   { 2423, 2400, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2815, 2818, 51 },
   { 2840, 32, 0 },
   { 2841, 251, 0 },
   { 2842, 2839, 0 },
   { 2847, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3214, 3215, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3216, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3223, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3225, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3218, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3654, 3619, 65 },
   { 3907, 3910, 90 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3655, 3656, 67 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 211, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 211, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5116, 4052, 0 },
   { 5169, 4065, 0 },
   { 5221, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5114, 4052, 0 },
   { 5167, 4065, 0 },
   { 5219, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5090, 4052, 0 },
   { 5144, 4065, 0 },
   { 5196, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5088, 4052, 0 },
   { 5142, 4065, 0 },
   { 5194, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 117, 32, 0 },
   { 211, 3, 0 },
   { 3523, 3524, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 117, 32, 0 },
   { 211, 3, 0 },
   { 3523, 3524, 0 },
   { 3565, 3524, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 117, 32, 0 },
   { 206, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 117, 32, 0 },
   { 206, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 117, 32, 0 },
   { 206, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 117, 32, 0 },
   { 206, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 117, 32, 0 },
   { 206, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 117, 32, 0 },
   { 5116, 4052, 0 },
   { 5169, 4065, 0 },
   { 5221, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 117, 32, 0 },
   { 5114, 4052, 0 },
   { 5167, 4065, 0 },
   { 5219, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 117, 32, 0 },
   { 5090, 4052, 0 },
   { 5144, 4065, 0 },
   { 5196, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 117, 32, 0 },
   { 5088, 4052, 0 },
   { 5142, 4065, 0 },
   { 5194, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 206, 61, 0 },
   { 209, 3, 0 },
   { 211, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 206, 61, 0 },
   { 209, 3, 0 },
   { 211, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 206, 61, 0 },
   { 5116, 4052, 0 },
   { 5169, 4065, 0 },
   { 5221, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 206, 61, 0 },
   { 5114, 4052, 0 },
   { 5167, 4065, 0 },
   { 5219, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 206, 61, 0 },
   { 5090, 4052, 0 },
   { 5144, 4065, 0 },
   { 5196, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 206, 61, 0 },
   { 5088, 4052, 0 },
   { 5142, 4065, 0 },
   { 5194, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5116, 4052, 0 },
   { 5169, 4065, 0 },
   { 5221, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5114, 4052, 0 },
   { 5167, 4065, 0 },
   { 5219, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5090, 4052, 0 },
   { 5144, 4065, 0 },
   { 5196, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5088, 4052, 0 },
   { 5142, 4065, 0 },
   { 5194, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 211, 3, 0 },
   { 330, 331, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 211, 3, 0 },
   { 330, 331, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 330, 331, 0 },
   { 5116, 4052, 0 },
   { 5169, 4065, 0 },
   { 5221, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 330, 331, 0 },
   { 5114, 4052, 0 },
   { 5167, 4065, 0 },
   { 5219, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 330, 331, 0 },
   { 5090, 4052, 0 },
   { 5144, 4065, 0 },
   { 5196, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 330, 331, 0 },
   { 5088, 4052, 0 },
   { 5142, 4065, 0 },
   { 5194, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 211, 3, 0 },
   { 333, 334, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 211, 3, 0 },
   { 333, 334, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 333, 334, 0 },
   { 5116, 4052, 0 },
   { 5169, 4065, 0 },
   { 5221, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 333, 334, 0 },
   { 5114, 4052, 0 },
   { 5167, 4065, 0 },
   { 5219, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 333, 334, 0 },
   { 5090, 4052, 0 },
   { 5144, 4065, 0 },
   { 5196, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 333, 334, 0 },
   { 5088, 4052, 0 },
   { 5142, 4065, 0 },
   { 5194, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 211, 3, 0 },
   { 342, 343, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 211, 3, 0 },
   { 342, 343, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 342, 343, 0 },
   { 5116, 4052, 0 },
   { 5169, 4065, 0 },
   { 5221, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 342, 343, 0 },
   { 5114, 4052, 0 },
   { 5167, 4065, 0 },
   { 5219, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 342, 343, 0 },
   { 5090, 4052, 0 },
   { 5144, 4065, 0 },
   { 5196, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 342, 343, 0 },
   { 5088, 4052, 0 },
   { 5142, 4065, 0 },
   { 5194, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 211, 3, 0 },
   { 345, 346, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 211, 3, 0 },
   { 345, 346, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 345, 346, 0 },
   { 5116, 4052, 0 },
   { 5169, 4065, 0 },
   { 5221, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 345, 346, 0 },
   { 5114, 4052, 0 },
   { 5167, 4065, 0 },
   { 5219, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 345, 346, 0 },
   { 5090, 4052, 0 },
   { 5144, 4065, 0 },
   { 5196, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 345, 346, 0 },
   { 5088, 4052, 0 },
   { 5142, 4065, 0 },
   { 5194, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 211, 3, 0 },
   { 336, 337, 0 },
   { 340, 337, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 211, 3, 0 },
   { 336, 337, 0 },
   { 340, 337, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 336, 337, 0 },
   { 340, 337, 0 },
   { 5116, 4052, 0 },
   { 5169, 4065, 0 },
   { 5221, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 336, 337, 0 },
   { 340, 337, 0 },
   { 5114, 4052, 0 },
   { 5167, 4065, 0 },
   { 5219, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 336, 337, 0 },
   { 340, 337, 0 },
   { 5090, 4052, 0 },
   { 5144, 4065, 0 },
   { 5196, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 336, 337, 0 },
   { 340, 337, 0 },
   { 5088, 4052, 0 },
   { 5142, 4065, 0 },
   { 5194, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 211, 3, 0 },
   { 5091, 4052, 0 },
   { 5145, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 211, 3, 0 },
   { 5091, 4052, 0 },
   { 5145, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5091, 4052, 0 },
   { 5116, 4052, 0 },
   { 5145, 4065, 0 },
   { 5169, 4065, 0 },
   { 5221, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5091, 4052, 0 },
   { 5114, 4052, 0 },
   { 5145, 4065, 0 },
   { 5167, 4065, 0 },
   { 5219, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5090, 4052, 0 },
   { 5091, 4052, 0 },
   { 5144, 4065, 0 },
   { 5145, 4065, 0 },
   { 5196, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5088, 4052, 0 },
   { 5091, 4052, 0 },
   { 5142, 4065, 0 },
   { 5145, 4065, 0 },
   { 5194, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 211, 3, 0 },
   { 5117, 4052, 0 },
   { 5170, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 211, 3, 0 },
   { 5117, 4052, 0 },
   { 5170, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5116, 4052, 0 },
   { 5117, 4052, 0 },
   { 5169, 4065, 0 },
   { 5170, 4065, 0 },
   { 5221, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5114, 4052, 0 },
   { 5117, 4052, 0 },
   { 5167, 4065, 0 },
   { 5170, 4065, 0 },
   { 5219, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5090, 4052, 0 },
   { 5117, 4052, 0 },
   { 5144, 4065, 0 },
   { 5170, 4065, 0 },
   { 5196, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5088, 4052, 0 },
   { 5117, 4052, 0 },
   { 5142, 4065, 0 },
   { 5170, 4065, 0 },
   { 5194, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 211, 3, 0 },
   { 5222, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 211, 3, 0 },
   { 5222, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5116, 4052, 0 },
   { 5169, 4065, 0 },
   { 5221, 4065, 0 },
   { 5222, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5114, 4052, 0 },
   { 5167, 4065, 0 },
   { 5219, 4065, 0 },
   { 5222, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5090, 4052, 0 },
   { 5144, 4065, 0 },
   { 5196, 4065, 0 },
   { 5222, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5088, 4052, 0 },
   { 5142, 4065, 0 },
   { 5194, 4065, 0 },
   { 5222, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 211, 3, 0 },
   { 5197, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 211, 3, 0 },
   { 5197, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5116, 4052, 0 },
   { 5169, 4065, 0 },
   { 5197, 4065, 0 },
   { 5221, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5114, 4052, 0 },
   { 5167, 4065, 0 },
   { 5197, 4065, 0 },
   { 5219, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5090, 4052, 0 },
   { 5144, 4065, 0 },
   { 5196, 4065, 0 },
   { 5197, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5088, 4052, 0 },
   { 5142, 4065, 0 },
   { 5194, 4065, 0 },
   { 5197, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 211, 3, 0 },
   { 5116, 4052, 0 },
   { 5169, 4065, 0 },
   { 5221, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 211, 3, 0 },
   { 5114, 4052, 0 },
   { 5167, 4065, 0 },
   { 5219, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 211, 3, 0 },
   { 5090, 4052, 0 },
   { 5144, 4065, 0 },
   { 5196, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 211, 3, 0 },
   { 5088, 4052, 0 },
   { 5142, 4065, 0 },
   { 5194, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 211, 3, 0 },
   { 5116, 4052, 0 },
   { 5169, 4065, 0 },
   { 5221, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 211, 3, 0 },
   { 5114, 4052, 0 },
   { 5167, 4065, 0 },
   { 5219, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 211, 3, 0 },
   { 5090, 4052, 0 },
   { 5144, 4065, 0 },
   { 5196, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 211, 3, 0 },
   { 5088, 4052, 0 },
   { 5142, 4065, 0 },
   { 5194, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5114, 4052, 0 },
   { 5116, 4052, 0 },
   { 5167, 4065, 0 },
   { 5169, 4065, 0 },
   { 5219, 4065, 0 },
   { 5221, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5090, 4052, 0 },
   { 5116, 4052, 0 },
   { 5144, 4065, 0 },
   { 5169, 4065, 0 },
   { 5196, 4065, 0 },
   { 5221, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5088, 4052, 0 },
   { 5116, 4052, 0 },
   { 5142, 4065, 0 },
   { 5169, 4065, 0 },
   { 5194, 4065, 0 },
   { 5221, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5090, 4052, 0 },
   { 5114, 4052, 0 },
   { 5144, 4065, 0 },
   { 5167, 4065, 0 },
   { 5196, 4065, 0 },
   { 5219, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5088, 4052, 0 },
   { 5114, 4052, 0 },
   { 5142, 4065, 0 },
   { 5167, 4065, 0 },
   { 5194, 4065, 0 },
   { 5219, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5088, 4052, 0 },
   { 5090, 4052, 0 },
   { 5142, 4065, 0 },
   { 5144, 4065, 0 },
   { 5194, 4065, 0 },
   { 5196, 4065, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3338, 3339, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3046, 1117, 0 },
   { 3047, 684, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 391, 392, 9 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1307, 1308, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 1279, 0 },
   { 1280, 1281, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 1283, 0 },
   { 1285, 1286, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1298, 1299, 0 },
   { 1300, 1301, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 2392, 2393, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 4088, 4089, 112 },
   { 4101, 4102, 112 },
   { 4113, 4114, 112 },
   { 4125, 3748, 113 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 4091, 4089, 112 },
   { 4104, 4102, 112 },
   { 4116, 4114, 112 },
   { 4127, 3748, 113 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5110, 5082, 0 },
   { 5163, 5135, 0 },
   { 5215, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5108, 5082, 0 },
   { 5161, 5135, 0 },
   { 5213, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5085, 5082, 0 },
   { 5138, 5135, 0 },
   { 5190, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1313, 1311, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3679, 3681, 0 },
   { 5081, 5082, 0 },
   { 5134, 5135, 0 },
   { 5187, 5135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 4049, 4043, 8 },
   { 4062, 4056, 8 },
   { 4075, 4069, 8 },
   { 4082, 3738, 111 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 4095, 4089, 112 },
   { 4108, 4102, 112 },
   { 4120, 4114, 112 },
   { 4131, 3748, 113 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 4051, 4043, 8 },
   { 4064, 4056, 8 },
   { 4077, 4069, 8 },
   { 4084, 3738, 111 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 4097, 4089, 112 },
   { 4110, 4102, 112 },
   { 4122, 4114, 112 },
   { 4133, 3748, 113 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 4042, 4043, 8 },
   { 4055, 4056, 8 },
   { 4068, 4069, 8 },
   { 3742, 3738, 111 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1288, 1289, 0 },
   { 1290, 1291, 0 },
   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 4042, 4043, 8 },
   { 4055, 4056, 8 },
   { 4068, 4069, 8 },
   { 3742, 3738, 111 },
   { 4088, 4089, 112 },
   { 4101, 4102, 112 },
   { 4113, 4114, 112 },
   { 4125, 3748, 113 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 4045, 4043, 8 },
   { 4058, 4056, 8 },
   { 4071, 4069, 8 },
   { 4079, 3738, 111 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1292, 1293, 0 },
   { 1295, 1296, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 4045, 4043, 8 },
   { 4058, 4056, 8 },
   { 4071, 4069, 8 },
   { 4079, 3738, 111 },
   { 4091, 4089, 112 },
   { 4104, 4102, 112 },
   { 4116, 4114, 112 },
   { 4127, 3748, 113 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3586, 3584, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3589, 3587, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3592, 3590, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3595, 3593, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1302, 1303, 0 },
   { 1305, 1306, 0 },
   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3598, 3596, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5119, 5093, 0 },
   { 5172, 49, 0 },
   { 5224, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5118, 5093, 0 },
   { 5171, 49, 0 },
   { 5223, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5094, 5093, 0 },
   { 5147, 49, 0 },
   { 5199, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5092, 5093, 0 },
   { 5146, 49, 0 },
   { 5198, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 238, 61, 0 },
   { 246, 32, 0 },
   { 249, 100, 0 },
   { 1032, 1036, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 238, 61, 0 },
   { 246, 32, 0 },
   { 249, 100, 0 },
   { 1024, 1027, 0 },
   { 1032, 1036, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 238, 61, 0 },
   { 246, 32, 0 },
   { 249, 100, 0 },
   { 5119, 5093, 0 },
   { 5172, 49, 0 },
   { 5224, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 238, 61, 0 },
   { 246, 32, 0 },
   { 249, 100, 0 },
   { 5118, 5093, 0 },
   { 5171, 49, 0 },
   { 5223, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 238, 61, 0 },
   { 246, 32, 0 },
   { 249, 100, 0 },
   { 5094, 5093, 0 },
   { 5147, 49, 0 },
   { 5199, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 238, 61, 0 },
   { 246, 32, 0 },
   { 249, 100, 0 },
   { 5092, 5093, 0 },
   { 5146, 49, 0 },
   { 5198, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3476, 3477, 0 },
   { 5119, 5093, 0 },
   { 5172, 49, 0 },
   { 5224, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3476, 3477, 0 },
   { 5118, 5093, 0 },
   { 5171, 49, 0 },
   { 5223, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3476, 3477, 0 },
   { 5094, 5093, 0 },
   { 5147, 49, 0 },
   { 5199, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3476, 3477, 0 },
   { 5092, 5093, 0 },
   { 5146, 49, 0 },
   { 5198, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5095, 5093, 0 },
   { 5119, 5093, 0 },
   { 5148, 49, 0 },
   { 5172, 49, 0 },
   { 5224, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5095, 5093, 0 },
   { 5118, 5093, 0 },
   { 5148, 49, 0 },
   { 5171, 49, 0 },
   { 5223, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5094, 5093, 0 },
   { 5095, 5093, 0 },
   { 5147, 49, 0 },
   { 5148, 49, 0 },
   { 5199, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5092, 5093, 0 },
   { 5095, 5093, 0 },
   { 5146, 49, 0 },
   { 5148, 49, 0 },
   { 5198, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5119, 5093, 0 },
   { 5120, 5093, 0 },
   { 5172, 49, 0 },
   { 5173, 49, 0 },
   { 5224, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5118, 5093, 0 },
   { 5120, 5093, 0 },
   { 5171, 49, 0 },
   { 5173, 49, 0 },
   { 5223, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5094, 5093, 0 },
   { 5120, 5093, 0 },
   { 5147, 49, 0 },
   { 5173, 49, 0 },
   { 5199, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5092, 5093, 0 },
   { 5120, 5093, 0 },
   { 5146, 49, 0 },
   { 5173, 49, 0 },
   { 5198, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5119, 5093, 0 },
   { 5172, 49, 0 },
   { 5224, 49, 0 },
   { 5225, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5118, 5093, 0 },
   { 5171, 49, 0 },
   { 5223, 49, 0 },
   { 5225, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5094, 5093, 0 },
   { 5147, 49, 0 },
   { 5199, 49, 0 },
   { 5225, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5092, 5093, 0 },
   { 5146, 49, 0 },
   { 5198, 49, 0 },
   { 5225, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5119, 5093, 0 },
   { 5172, 49, 0 },
   { 5200, 49, 0 },
   { 5224, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5118, 5093, 0 },
   { 5171, 49, 0 },
   { 5200, 49, 0 },
   { 5223, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5094, 5093, 0 },
   { 5147, 49, 0 },
   { 5199, 49, 0 },
   { 5200, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5092, 5093, 0 },
   { 5146, 49, 0 },
   { 5198, 49, 0 },
   { 5200, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5118, 5093, 0 },
   { 5119, 5093, 0 },
   { 5171, 49, 0 },
   { 5172, 49, 0 },
   { 5223, 49, 0 },
   { 5224, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5094, 5093, 0 },
   { 5119, 5093, 0 },
   { 5147, 49, 0 },
   { 5172, 49, 0 },
   { 5199, 49, 0 },
   { 5224, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5092, 5093, 0 },
   { 5119, 5093, 0 },
   { 5146, 49, 0 },
   { 5172, 49, 0 },
   { 5198, 49, 0 },
   { 5224, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5094, 5093, 0 },
   { 5118, 5093, 0 },
   { 5147, 49, 0 },
   { 5171, 49, 0 },
   { 5199, 49, 0 },
   { 5223, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5092, 5093, 0 },
   { 5118, 5093, 0 },
   { 5146, 49, 0 },
   { 5171, 49, 0 },
   { 5198, 49, 0 },
   { 5223, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5092, 5093, 0 },
   { 5094, 5093, 0 },
   { 5146, 49, 0 },
   { 5147, 49, 0 },
   { 5198, 49, 0 },
   { 5199, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2764, 1229, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 910, 913, 0 },
   { 918, 921, 0 },
   { 926, 929, 0 },
   { 934, 937, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 914, 913, 0 },
   { 922, 921, 0 },
   { 930, 929, 0 },
   { 938, 937, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3853, 3854, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 1038, 1041, 0 },
   { 2765, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 816, 820, 0 },
   { 1020, 1022, 0 },
   { 2765, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 1020, 1022, 0 },
   { 1038, 1041, 0 },
   { 2765, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 2765, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 2765, 32, 0 },
   { 3332, 3334, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2140, 2142, 0 },
   { 2170, 2172, 0 },
   { 2200, 2202, 0 },
   { 2231, 2233, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2770, 1229, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2140, 2142, 0 },
   { 2170, 2172, 0 },
   { 2200, 2202, 0 },
   { 2231, 2233, 0 },
   { 2770, 1229, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 915, 917, 0 },
   { 923, 925, 0 },
   { 931, 933, 0 },
   { 939, 941, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 915, 917, 0 },
   { 923, 925, 0 },
   { 931, 933, 0 },
   { 939, 941, 0 },
   { 2140, 2142, 0 },
   { 2170, 2172, 0 },
   { 2200, 2202, 0 },
   { 2231, 2233, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2771, 32, 0 },
   { 3234, 3235, 58 },
   { 1017, 422, 58 },
   { 3237, 3239, 58 },
   { 1008, 447, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2771, 32, 0 },
   { 3234, 3235, 58 },
   { 1017, 422, 58 },
   { 3237, 3239, 58 },
   { 1008, 447, 59 },
   { 3335, 3331, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3153, 32, 0 },
   { 3674, 3672, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3153, 32, 0 },
   { 3671, 3672, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3153, 32, 0 },
   { 3671, 3672, 0 },
   { 3674, 3672, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1156, 1155, 0 },
   { 2698, 1235, 0 },
   { 3173, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1160, 1159, 0 },
   { 2698, 1235, 0 },
   { 3173, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1105, 1106, 0 },
   { 1214, 1217, 0 },
   { 2674, 1257, 0 },
   { 2698, 1235, 0 },
   { 3173, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1154, 1155, 0 },
   { 1191, 1192, 0 },
   { 2698, 1235, 0 },
   { 3173, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1158, 1159, 0 },
   { 2698, 1235, 0 },
   { 3173, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1157, 1155, 0 },
   { 2674, 1257, 0 },
   { 2698, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1091, 1092, 0 },
   { 1164, 1162, 0 },
   { 2674, 1257, 0 },
   { 2698, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1141, 1138, 0 },
   { 2699, 1077, 0 },
   { 3172, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1148, 1146, 0 },
   { 2699, 1077, 0 },
   { 3172, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1103, 1104, 0 },
   { 1220, 1221, 0 },
   { 2673, 1253, 0 },
   { 2699, 1077, 0 },
   { 3172, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1137, 1138, 0 },
   { 1193, 1194, 0 },
   { 2699, 1077, 0 },
   { 3172, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1145, 1146, 0 },
   { 2699, 1077, 0 },
   { 3172, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1143, 1138, 0 },
   { 2673, 1253, 0 },
   { 2699, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1093, 1094, 0 },
   { 1153, 1151, 0 },
   { 2673, 1253, 0 },
   { 2699, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 383, 269, 9 },
   { 3658, 3659, 0 },
   { 3661, 3662, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 255, 253, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 221, 111, 0 },
   { 222, 113, 5 },
   { 223, 115, 6 },
   { 242, 32, 0 },
   { 245, 32, 0 },
   { 248, 98, 0 },
   { 383, 269, 9 },
   { 3658, 3659, 0 },
   { 3661, 3662, 0 },
   { 5594, 5597, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 221, 111, 0 },
   { 222, 113, 5 },
   { 223, 115, 6 },
   { 242, 32, 0 },
   { 245, 32, 0 },
   { 248, 98, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 221, 111, 0 },
   { 222, 113, 5 },
   { 223, 115, 6 },
   { 242, 32, 0 },
   { 245, 32, 0 },
   { 248, 98, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 221, 111, 0 },
   { 222, 113, 5 },
   { 223, 115, 6 },
   { 242, 32, 0 },
   { 245, 32, 0 },
   { 248, 98, 0 },
   { 255, 253, 0 },
   { 3481, 3483, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 221, 111, 0 },
   { 222, 113, 5 },
   { 223, 115, 6 },
   { 242, 32, 0 },
   { 245, 32, 0 },
   { 248, 98, 0 },
   { 255, 253, 0 },
   { 3481, 3483, 0 },
   { 3526, 3483, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 383, 269, 9 },
   { 3468, 3469, 0 },
   { 3658, 3659, 0 },
   { 3661, 3662, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3468, 3469, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3468, 3469, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 255, 253, 0 },
   { 3468, 3469, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 383, 269, 9 },
   { 3658, 3659, 0 },
   { 3661, 3662, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3468, 3469, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 255, 253, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 383, 269, 9 },
   { 3658, 3659, 0 },
   { 3661, 3662, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 252, 253, 0 },
   { 255, 253, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3658, 3659, 0 },
   { 3661, 3662, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3658, 3659, 0 },
   { 3661, 3662, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 255, 253, 0 },
   { 3658, 3659, 0 },
   { 3661, 3662, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 383, 269, 9 },
   { 3658, 3659, 0 },
   { 3661, 3662, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 255, 253, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 383, 269, 9 },
   { 3658, 3659, 0 },
   { 3661, 3662, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 255, 253, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 378, 379, 9 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 378, 379, 9 },
   { 383, 269, 9 },
   { 3658, 3659, 0 },
   { 3661, 3662, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 383, 269, 9 },
   { 3658, 3659, 0 },
   { 3661, 3662, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 383, 269, 9 },
   { 3658, 3659, 0 },
   { 3661, 3662, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 255, 253, 0 },
   { 383, 269, 9 },
   { 3658, 3659, 0 },
   { 3661, 3662, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 255, 253, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 255, 253, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 143, 145, 0 },
   { 519, 520, 0 },
   { 587, 588, 0 },
   { 659, 660, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 386, 387, 9 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 263, 264, 0 },
   { 265, 266, 5 },
   { 267, 227, 6 },
   { 386, 387, 9 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 147, 148, 0 },
   { 386, 387, 9 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 143, 145, 0 },
   { 386, 387, 9 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 143, 145, 0 },
   { 257, 3, 0 },
   { 259, 258, 5 },
   { 260, 36, 6 },
   { 271, 272, 0 },
   { 276, 277, 0 },
   { 519, 520, 0 },
   { 587, 588, 0 },
   { 659, 660, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 257, 3, 0 },
   { 259, 258, 5 },
   { 260, 36, 6 },
   { 271, 272, 0 },
   { 276, 277, 0 },
   { 386, 387, 9 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 257, 3, 0 },
   { 259, 258, 5 },
   { 260, 36, 6 },
   { 263, 264, 0 },
   { 265, 266, 5 },
   { 267, 227, 6 },
   { 271, 272, 0 },
   { 276, 277, 0 },
   { 386, 387, 9 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 147, 148, 0 },
   { 257, 3, 0 },
   { 259, 258, 5 },
   { 260, 36, 6 },
   { 271, 272, 0 },
   { 276, 277, 0 },
   { 386, 387, 9 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 143, 145, 0 },
   { 257, 3, 0 },
   { 259, 258, 5 },
   { 260, 36, 6 },
   { 271, 272, 0 },
   { 276, 277, 0 },
   { 386, 387, 9 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 143, 145, 0 },
   { 519, 520, 0 },
   { 587, 588, 0 },
   { 659, 660, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 386, 387, 9 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 263, 264, 0 },
   { 265, 266, 5 },
   { 267, 227, 6 },
   { 386, 387, 9 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 147, 148, 0 },
   { 386, 387, 9 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 143, 145, 0 },
   { 386, 387, 9 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 147, 148, 0 },
   { 160, 162, 0 },
   { 386, 387, 9 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 143, 145, 0 },
   { 386, 387, 9 },
   { 519, 520, 0 },
   { 587, 588, 0 },
   { 659, 660, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 524, 525, 0 },
   { 591, 592, 0 },
   { 663, 664, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 263, 264, 0 },
   { 265, 266, 5 },
   { 267, 227, 6 },
   { 524, 525, 0 },
   { 591, 592, 0 },
   { 663, 664, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 147, 148, 0 },
   { 524, 525, 0 },
   { 591, 592, 0 },
   { 663, 664, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 143, 145, 0 },
   { 524, 525, 0 },
   { 591, 592, 0 },
   { 663, 664, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 143, 145, 0 },
   { 519, 520, 0 },
   { 524, 525, 0 },
   { 587, 588, 0 },
   { 591, 592, 0 },
   { 659, 660, 0 },
   { 663, 664, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 147, 148, 0 },
   { 530, 532, 0 },
   { 596, 598, 0 },
   { 669, 671, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 384, 385, 9 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 263, 264, 0 },
   { 265, 266, 5 },
   { 267, 227, 6 },
   { 384, 385, 9 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 147, 148, 0 },
   { 384, 385, 9 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 143, 145, 0 },
   { 384, 385, 9 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 384, 385, 9 },
   { 386, 387, 9 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 263, 264, 0 },
   { 265, 266, 5 },
   { 267, 227, 6 },
   { 384, 385, 9 },
   { 386, 387, 9 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 147, 148, 0 },
   { 384, 385, 9 },
   { 386, 387, 9 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 143, 145, 0 },
   { 384, 385, 9 },
   { 386, 387, 9 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5125, 2065, 0 },
   { 5178, 2069, 0 },
   { 5230, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5124, 2065, 0 },
   { 5177, 2069, 0 },
   { 5229, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5100, 2065, 0 },
   { 5153, 2069, 0 },
   { 5205, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5099, 2065, 0 },
   { 5152, 2069, 0 },
   { 5204, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2706, 32, 0 },
   { 2707, 63, 0 },
   { 2742, 2744, 0 },
   { 2745, 2747, 0 },
   { 2748, 2749, 0 },
   { 2750, 2751, 0 },
   { 5125, 2065, 0 },
   { 5178, 2069, 0 },
   { 5230, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2706, 32, 0 },
   { 2707, 63, 0 },
   { 2742, 2744, 0 },
   { 2745, 2747, 0 },
   { 2748, 2749, 0 },
   { 2750, 2751, 0 },
   { 5124, 2065, 0 },
   { 5177, 2069, 0 },
   { 5229, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2706, 32, 0 },
   { 2707, 63, 0 },
   { 2742, 2744, 0 },
   { 2745, 2747, 0 },
   { 2748, 2749, 0 },
   { 2750, 2751, 0 },
   { 5100, 2065, 0 },
   { 5153, 2069, 0 },
   { 5205, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2706, 32, 0 },
   { 2707, 63, 0 },
   { 2742, 2744, 0 },
   { 2745, 2747, 0 },
   { 2748, 2749, 0 },
   { 2750, 2751, 0 },
   { 5099, 2065, 0 },
   { 5152, 2069, 0 },
   { 5204, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2779, 2776, 49 },
   { 2786, 2783, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2715, 3, 0 },
   { 5125, 2065, 0 },
   { 5178, 2069, 0 },
   { 5230, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2715, 3, 0 },
   { 5124, 2065, 0 },
   { 5177, 2069, 0 },
   { 5229, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2715, 3, 0 },
   { 5100, 2065, 0 },
   { 5153, 2069, 0 },
   { 5205, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2715, 3, 0 },
   { 5099, 2065, 0 },
   { 5152, 2069, 0 },
   { 5204, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2716, 2338, 0 },
   { 5125, 2065, 0 },
   { 5178, 2069, 0 },
   { 5230, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2716, 2338, 0 },
   { 5124, 2065, 0 },
   { 5177, 2069, 0 },
   { 5229, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2716, 2338, 0 },
   { 5100, 2065, 0 },
   { 5153, 2069, 0 },
   { 5205, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2716, 2338, 0 },
   { 5099, 2065, 0 },
   { 5152, 2069, 0 },
   { 5204, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2247, 1235, 0 },
   { 2705, 32, 0 },
   { 5125, 2065, 0 },
   { 5178, 2069, 0 },
   { 5230, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2247, 1235, 0 },
   { 2705, 32, 0 },
   { 5124, 2065, 0 },
   { 5177, 2069, 0 },
   { 5229, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2247, 1235, 0 },
   { 2705, 32, 0 },
   { 5100, 2065, 0 },
   { 5153, 2069, 0 },
   { 5205, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2247, 1235, 0 },
   { 2705, 32, 0 },
   { 5099, 2065, 0 },
   { 5152, 2069, 0 },
   { 5204, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2791, 2789, 49 },
   { 2796, 2794, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2708, 63, 0 },
   { 5125, 2065, 0 },
   { 5178, 2069, 0 },
   { 5230, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2708, 63, 0 },
   { 5124, 2065, 0 },
   { 5177, 2069, 0 },
   { 5229, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2708, 63, 0 },
   { 5100, 2065, 0 },
   { 5153, 2069, 0 },
   { 5205, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2708, 63, 0 },
   { 5099, 2065, 0 },
   { 5152, 2069, 0 },
   { 5204, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1716, 1719, 0 },
   { 1723, 1719, 0 },
   { 1728, 1731, 0 },
   { 1735, 1731, 0 },
   { 2705, 32, 0 },
   { 2716, 2338, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5125, 2065, 0 },
   { 5178, 2069, 0 },
   { 5230, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5124, 2065, 0 },
   { 5177, 2069, 0 },
   { 5229, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5100, 2065, 0 },
   { 5153, 2069, 0 },
   { 5205, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5099, 2065, 0 },
   { 5152, 2069, 0 },
   { 5204, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5101, 2065, 0 },
   { 5125, 2065, 0 },
   { 5154, 2069, 0 },
   { 5178, 2069, 0 },
   { 5230, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5101, 2065, 0 },
   { 5124, 2065, 0 },
   { 5154, 2069, 0 },
   { 5177, 2069, 0 },
   { 5229, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5100, 2065, 0 },
   { 5101, 2065, 0 },
   { 5153, 2069, 0 },
   { 5154, 2069, 0 },
   { 5205, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5099, 2065, 0 },
   { 5101, 2065, 0 },
   { 5152, 2069, 0 },
   { 5154, 2069, 0 },
   { 5204, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5125, 2065, 0 },
   { 5126, 2065, 0 },
   { 5178, 2069, 0 },
   { 5179, 2069, 0 },
   { 5230, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5124, 2065, 0 },
   { 5126, 2065, 0 },
   { 5177, 2069, 0 },
   { 5179, 2069, 0 },
   { 5229, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5100, 2065, 0 },
   { 5126, 2065, 0 },
   { 5153, 2069, 0 },
   { 5179, 2069, 0 },
   { 5205, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5099, 2065, 0 },
   { 5126, 2065, 0 },
   { 5152, 2069, 0 },
   { 5179, 2069, 0 },
   { 5204, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5125, 2065, 0 },
   { 5178, 2069, 0 },
   { 5230, 2069, 0 },
   { 5231, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5124, 2065, 0 },
   { 5177, 2069, 0 },
   { 5229, 2069, 0 },
   { 5231, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5100, 2065, 0 },
   { 5153, 2069, 0 },
   { 5205, 2069, 0 },
   { 5231, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5099, 2065, 0 },
   { 5152, 2069, 0 },
   { 5204, 2069, 0 },
   { 5231, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5125, 2065, 0 },
   { 5178, 2069, 0 },
   { 5206, 2069, 0 },
   { 5230, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5124, 2065, 0 },
   { 5177, 2069, 0 },
   { 5206, 2069, 0 },
   { 5229, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5100, 2065, 0 },
   { 5153, 2069, 0 },
   { 5205, 2069, 0 },
   { 5206, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5099, 2065, 0 },
   { 5152, 2069, 0 },
   { 5204, 2069, 0 },
   { 5206, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2779, 2776, 49 },
   { 2786, 2783, 49 },
   { 2791, 2789, 49 },
   { 2796, 2794, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5100, 2065, 0 },
   { 5153, 2069, 0 },
   { 5205, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5099, 2065, 0 },
   { 5152, 2069, 0 },
   { 5204, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 3088, 3089, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5124, 2065, 0 },
   { 5125, 2065, 0 },
   { 5177, 2069, 0 },
   { 5178, 2069, 0 },
   { 5229, 2069, 0 },
   { 5230, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5100, 2065, 0 },
   { 5125, 2065, 0 },
   { 5153, 2069, 0 },
   { 5178, 2069, 0 },
   { 5205, 2069, 0 },
   { 5230, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5099, 2065, 0 },
   { 5125, 2065, 0 },
   { 5152, 2069, 0 },
   { 5178, 2069, 0 },
   { 5204, 2069, 0 },
   { 5230, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5100, 2065, 0 },
   { 5124, 2065, 0 },
   { 5153, 2069, 0 },
   { 5177, 2069, 0 },
   { 5205, 2069, 0 },
   { 5229, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5099, 2065, 0 },
   { 5124, 2065, 0 },
   { 5152, 2069, 0 },
   { 5177, 2069, 0 },
   { 5204, 2069, 0 },
   { 5229, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5099, 2065, 0 },
   { 5100, 2065, 0 },
   { 5152, 2069, 0 },
   { 5153, 2069, 0 },
   { 5204, 2069, 0 },
   { 5205, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5099, 2065, 0 },
   { 5152, 2069, 0 },
   { 5204, 2069, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 3340, 3339, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4203, 4204, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4209, 4210, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4215, 4216, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4198, 4201, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4206, 4207, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4212, 4213, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2717, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1418, 1419, 0 },
   { 1440, 1441, 0 },
   { 1462, 1463, 0 },
   { 1482, 1483, 0 },
   { 2700, 32, 0 },
   { 2704, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5122, 2023, 0 },
   { 5175, 2031, 0 },
   { 5227, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5121, 2023, 0 },
   { 5174, 2031, 0 },
   { 5226, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5097, 2023, 0 },
   { 5150, 2031, 0 },
   { 5202, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5096, 2023, 0 },
   { 5149, 2031, 0 },
   { 5201, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2683, 2684, 0 },
   { 2686, 2687, 0 },
   { 2700, 32, 0 },
   { 2701, 61, 0 },
   { 2702, 32, 0 },
   { 2717, 3, 0 },
   { 2725, 2729, 0 },
   { 2731, 2735, 0 },
   { 2737, 2738, 0 },
   { 2740, 2741, 0 },
   { 3245, 3246, 58 },
   { 3276, 3277, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2683, 2684, 0 },
   { 2686, 2687, 0 },
   { 2700, 32, 0 },
   { 2701, 61, 0 },
   { 2702, 32, 0 },
   { 2717, 3, 0 },
   { 2725, 2729, 0 },
   { 2731, 2735, 0 },
   { 2737, 2738, 0 },
   { 2740, 2741, 0 },
   { 3245, 3246, 58 },
   { 3276, 3277, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2683, 2684, 0 },
   { 2686, 2687, 0 },
   { 2700, 32, 0 },
   { 2701, 61, 0 },
   { 2702, 32, 0 },
   { 2725, 2729, 0 },
   { 2731, 2735, 0 },
   { 2737, 2738, 0 },
   { 2740, 2741, 0 },
   { 3097, 1528, 0 },
   { 3098, 3095, 0 },
   { 3245, 3246, 58 },
   { 3276, 3277, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2683, 2684, 0 },
   { 2686, 2687, 0 },
   { 2700, 32, 0 },
   { 2701, 61, 0 },
   { 2702, 32, 0 },
   { 2725, 2729, 0 },
   { 2731, 2735, 0 },
   { 2737, 2738, 0 },
   { 2740, 2741, 0 },
   { 3245, 3246, 58 },
   { 3276, 3277, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1418, 1419, 0 },
   { 1440, 1441, 0 },
   { 1462, 1463, 0 },
   { 1482, 1483, 0 },
   { 2683, 2684, 0 },
   { 2686, 2687, 0 },
   { 2700, 32, 0 },
   { 2701, 61, 0 },
   { 2702, 32, 0 },
   { 2704, 61, 0 },
   { 2725, 2729, 0 },
   { 2731, 2735, 0 },
   { 2737, 2738, 0 },
   { 2740, 2741, 0 },
   { 3245, 3246, 58 },
   { 3276, 3277, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2683, 2684, 0 },
   { 2686, 2687, 0 },
   { 2700, 32, 0 },
   { 2701, 61, 0 },
   { 2702, 32, 0 },
   { 2725, 2729, 0 },
   { 2731, 2735, 0 },
   { 2737, 2738, 0 },
   { 2740, 2741, 0 },
   { 3245, 3246, 58 },
   { 3276, 3277, 59 },
   { 5122, 2023, 0 },
   { 5175, 2031, 0 },
   { 5227, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2683, 2684, 0 },
   { 2686, 2687, 0 },
   { 2700, 32, 0 },
   { 2701, 61, 0 },
   { 2702, 32, 0 },
   { 2725, 2729, 0 },
   { 2731, 2735, 0 },
   { 2737, 2738, 0 },
   { 2740, 2741, 0 },
   { 3245, 3246, 58 },
   { 3276, 3277, 59 },
   { 5121, 2023, 0 },
   { 5174, 2031, 0 },
   { 5226, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2683, 2684, 0 },
   { 2686, 2687, 0 },
   { 2700, 32, 0 },
   { 2701, 61, 0 },
   { 2702, 32, 0 },
   { 2725, 2729, 0 },
   { 2731, 2735, 0 },
   { 2737, 2738, 0 },
   { 2740, 2741, 0 },
   { 3245, 3246, 58 },
   { 3276, 3277, 59 },
   { 5097, 2023, 0 },
   { 5150, 2031, 0 },
   { 5202, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2012, 2015, 0 },
   { 2683, 2684, 0 },
   { 2686, 2687, 0 },
   { 2700, 32, 0 },
   { 2701, 61, 0 },
   { 2702, 32, 0 },
   { 2725, 2729, 0 },
   { 2731, 2735, 0 },
   { 2737, 2738, 0 },
   { 2740, 2741, 0 },
   { 3245, 3246, 58 },
   { 3276, 3277, 59 },
   { 5096, 2023, 0 },
   { 5149, 2031, 0 },
   { 5201, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2717, 3, 0 },
   { 2718, 2719, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2718, 2719, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1418, 1419, 0 },
   { 1440, 1441, 0 },
   { 1462, 1463, 0 },
   { 1482, 1483, 0 },
   { 2700, 32, 0 },
   { 2704, 61, 0 },
   { 2718, 2719, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2718, 2719, 0 },
   { 5122, 2023, 0 },
   { 5175, 2031, 0 },
   { 5227, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2718, 2719, 0 },
   { 5121, 2023, 0 },
   { 5174, 2031, 0 },
   { 5226, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2718, 2719, 0 },
   { 5097, 2023, 0 },
   { 5150, 2031, 0 },
   { 5202, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2718, 2719, 0 },
   { 5096, 2023, 0 },
   { 5149, 2031, 0 },
   { 5201, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 191, 193, 0 },
   { 2700, 32, 0 },
   { 2717, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2717, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1418, 1419, 0 },
   { 1440, 1441, 0 },
   { 1462, 1463, 0 },
   { 1482, 1483, 0 },
   { 2700, 32, 0 },
   { 2704, 61, 0 },
   { 2717, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2717, 3, 0 },
   { 5122, 2023, 0 },
   { 5175, 2031, 0 },
   { 5227, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2717, 3, 0 },
   { 5121, 2023, 0 },
   { 5174, 2031, 0 },
   { 5226, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2717, 3, 0 },
   { 5097, 2023, 0 },
   { 5150, 2031, 0 },
   { 5202, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2717, 3, 0 },
   { 5096, 2023, 0 },
   { 5149, 2031, 0 },
   { 5201, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2114, 2116, 43 },
   { 2114, 2122, 45 },
   { 2145, 2147, 43 },
   { 2145, 2153, 45 },
   { 2175, 2177, 43 },
   { 2175, 2183, 45 },
   { 2205, 2207, 43 },
   { 2205, 2213, 45 },
   { 2700, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1082, 1077, 0 },
   { 2700, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2704, 61, 0 },
   { 2717, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2704, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1418, 1419, 0 },
   { 1440, 1441, 0 },
   { 1462, 1463, 0 },
   { 1482, 1483, 0 },
   { 2700, 32, 0 },
   { 2704, 61, 0 },
   { 2760, 2761, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2704, 61, 0 },
   { 5122, 2023, 0 },
   { 5175, 2031, 0 },
   { 5227, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2704, 61, 0 },
   { 5121, 2023, 0 },
   { 5174, 2031, 0 },
   { 5226, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2704, 61, 0 },
   { 5097, 2023, 0 },
   { 5150, 2031, 0 },
   { 5202, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2704, 61, 0 },
   { 5096, 2023, 0 },
   { 5149, 2031, 0 },
   { 5201, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2717, 3, 0 },
   { 5098, 2023, 0 },
   { 5151, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5098, 2023, 0 },
   { 5151, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1418, 1419, 0 },
   { 1440, 1441, 0 },
   { 1462, 1463, 0 },
   { 1482, 1483, 0 },
   { 2700, 32, 0 },
   { 2704, 61, 0 },
   { 5098, 2023, 0 },
   { 5151, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5098, 2023, 0 },
   { 5122, 2023, 0 },
   { 5151, 2031, 0 },
   { 5175, 2031, 0 },
   { 5227, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5098, 2023, 0 },
   { 5121, 2023, 0 },
   { 5151, 2031, 0 },
   { 5174, 2031, 0 },
   { 5226, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5097, 2023, 0 },
   { 5098, 2023, 0 },
   { 5150, 2031, 0 },
   { 5151, 2031, 0 },
   { 5202, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5096, 2023, 0 },
   { 5098, 2023, 0 },
   { 5149, 2031, 0 },
   { 5151, 2031, 0 },
   { 5201, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2717, 3, 0 },
   { 5123, 2023, 0 },
   { 5176, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5123, 2023, 0 },
   { 5176, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1418, 1419, 0 },
   { 1440, 1441, 0 },
   { 1462, 1463, 0 },
   { 1482, 1483, 0 },
   { 2700, 32, 0 },
   { 2704, 61, 0 },
   { 5123, 2023, 0 },
   { 5176, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5122, 2023, 0 },
   { 5123, 2023, 0 },
   { 5175, 2031, 0 },
   { 5176, 2031, 0 },
   { 5227, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5121, 2023, 0 },
   { 5123, 2023, 0 },
   { 5174, 2031, 0 },
   { 5176, 2031, 0 },
   { 5226, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5097, 2023, 0 },
   { 5123, 2023, 0 },
   { 5150, 2031, 0 },
   { 5176, 2031, 0 },
   { 5202, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5096, 2023, 0 },
   { 5123, 2023, 0 },
   { 5149, 2031, 0 },
   { 5176, 2031, 0 },
   { 5201, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2717, 3, 0 },
   { 5228, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5228, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1418, 1419, 0 },
   { 1440, 1441, 0 },
   { 1462, 1463, 0 },
   { 1482, 1483, 0 },
   { 2700, 32, 0 },
   { 2704, 61, 0 },
   { 5228, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5122, 2023, 0 },
   { 5175, 2031, 0 },
   { 5227, 2031, 0 },
   { 5228, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5121, 2023, 0 },
   { 5174, 2031, 0 },
   { 5226, 2031, 0 },
   { 5228, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5097, 2023, 0 },
   { 5150, 2031, 0 },
   { 5202, 2031, 0 },
   { 5228, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5096, 2023, 0 },
   { 5149, 2031, 0 },
   { 5201, 2031, 0 },
   { 5228, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2717, 3, 0 },
   { 5203, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5203, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1418, 1419, 0 },
   { 1440, 1441, 0 },
   { 1462, 1463, 0 },
   { 1482, 1483, 0 },
   { 2700, 32, 0 },
   { 2704, 61, 0 },
   { 5203, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5122, 2023, 0 },
   { 5175, 2031, 0 },
   { 5203, 2031, 0 },
   { 5227, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5121, 2023, 0 },
   { 5174, 2031, 0 },
   { 5203, 2031, 0 },
   { 5226, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5097, 2023, 0 },
   { 5150, 2031, 0 },
   { 5202, 2031, 0 },
   { 5203, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5096, 2023, 0 },
   { 5149, 2031, 0 },
   { 5201, 2031, 0 },
   { 5203, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2717, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1418, 1419, 0 },
   { 1440, 1441, 0 },
   { 1462, 1463, 0 },
   { 1482, 1483, 0 },
   { 2700, 32, 0 },
   { 2704, 61, 0 },
   { 2717, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2717, 3, 0 },
   { 5122, 2023, 0 },
   { 5175, 2031, 0 },
   { 5227, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2717, 3, 0 },
   { 5121, 2023, 0 },
   { 5174, 2031, 0 },
   { 5226, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2717, 3, 0 },
   { 5097, 2023, 0 },
   { 5150, 2031, 0 },
   { 5202, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 2717, 3, 0 },
   { 5096, 2023, 0 },
   { 5149, 2031, 0 },
   { 5201, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 3086, 3087, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1418, 1419, 0 },
   { 1440, 1441, 0 },
   { 1462, 1463, 0 },
   { 1482, 1483, 0 },
   { 2700, 32, 0 },
   { 2704, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5122, 2023, 0 },
   { 5175, 2031, 0 },
   { 5227, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5121, 2023, 0 },
   { 5174, 2031, 0 },
   { 5226, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5097, 2023, 0 },
   { 5150, 2031, 0 },
   { 5202, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5096, 2023, 0 },
   { 5149, 2031, 0 },
   { 5201, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2114, 2116, 43 },
   { 2114, 2122, 45 },
   { 2145, 2147, 43 },
   { 2145, 2153, 45 },
   { 2175, 2177, 43 },
   { 2175, 2183, 45 },
   { 2205, 2207, 43 },
   { 2205, 2213, 45 },
   { 2700, 32, 0 },
   { 3231, 3233, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1418, 1419, 0 },
   { 1440, 1441, 0 },
   { 1462, 1463, 0 },
   { 1482, 1483, 0 },
   { 2700, 32, 0 },
   { 2704, 61, 0 },
   { 5122, 2023, 0 },
   { 5175, 2031, 0 },
   { 5227, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1418, 1419, 0 },
   { 1440, 1441, 0 },
   { 1462, 1463, 0 },
   { 1482, 1483, 0 },
   { 2700, 32, 0 },
   { 2704, 61, 0 },
   { 5121, 2023, 0 },
   { 5174, 2031, 0 },
   { 5226, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1418, 1419, 0 },
   { 1440, 1441, 0 },
   { 1462, 1463, 0 },
   { 1482, 1483, 0 },
   { 2700, 32, 0 },
   { 2704, 61, 0 },
   { 5097, 2023, 0 },
   { 5150, 2031, 0 },
   { 5202, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1418, 1419, 0 },
   { 1440, 1441, 0 },
   { 1462, 1463, 0 },
   { 1482, 1483, 0 },
   { 2700, 32, 0 },
   { 2704, 61, 0 },
   { 5096, 2023, 0 },
   { 5149, 2031, 0 },
   { 5201, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5121, 2023, 0 },
   { 5122, 2023, 0 },
   { 5174, 2031, 0 },
   { 5175, 2031, 0 },
   { 5226, 2031, 0 },
   { 5227, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5097, 2023, 0 },
   { 5122, 2023, 0 },
   { 5150, 2031, 0 },
   { 5175, 2031, 0 },
   { 5202, 2031, 0 },
   { 5227, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5096, 2023, 0 },
   { 5122, 2023, 0 },
   { 5149, 2031, 0 },
   { 5175, 2031, 0 },
   { 5201, 2031, 0 },
   { 5227, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5097, 2023, 0 },
   { 5121, 2023, 0 },
   { 5150, 2031, 0 },
   { 5174, 2031, 0 },
   { 5202, 2031, 0 },
   { 5226, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5096, 2023, 0 },
   { 5121, 2023, 0 },
   { 5149, 2031, 0 },
   { 5174, 2031, 0 },
   { 5201, 2031, 0 },
   { 5226, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5096, 2023, 0 },
   { 5097, 2023, 0 },
   { 5149, 2031, 0 },
   { 5150, 2031, 0 },
   { 5201, 2031, 0 },
   { 5202, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2700, 32, 0 },
   { 5096, 2023, 0 },
   { 5149, 2031, 0 },
   { 5201, 2031, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2395, 2394, 0 },
   { 3154, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3221, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1168, 1138, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1095, 1096, 0 },
   { 1255, 1253, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1095, 1096, 0 },
   { 1201, 1204, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1095, 1096, 0 },
   { 2688, 760, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1187, 1188, 0 },
   { 1277, 1049, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1187, 1188, 0 },
   { 1196, 1199, 0 },
   { 1277, 1049, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1170, 1155, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1099, 1100, 0 },
   { 1258, 1257, 0 },
   { 1258, 1257, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1099, 1100, 0 },
   { 1207, 1208, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1099, 1100, 0 },
   { 1133, 1131, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1502, 1254, 0 },
   { 1505, 1254, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1502, 1254, 0 },
   { 1581, 1583, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1352, 1283, 0 },
   { 1374, 1376, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1502, 1254, 0 },
   { 1505, 1254, 0 },
   { 3157, 217, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1502, 1254, 0 },
   { 1581, 1583, 0 },
   { 3157, 217, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1352, 1283, 0 },
   { 1374, 1376, 0 },
   { 3157, 217, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1352, 1283, 0 },
   { 1374, 1376, 0 },
   { 1502, 1254, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1502, 1254, 0 },
   { 1505, 1254, 0 },
   { 1507, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1502, 1254, 0 },
   { 1507, 32, 0 },
   { 1581, 1583, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1352, 1283, 0 },
   { 1374, 1376, 0 },
   { 1507, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1352, 1283, 0 },
   { 1502, 1254, 0 },
   { 1505, 1254, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1352, 1283, 0 },
   { 1502, 1254, 0 },
   { 1581, 1583, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1502, 1254, 0 },
   { 1505, 1254, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1502, 1254, 0 },
   { 1581, 1583, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1352, 1283, 0 },
   { 1374, 1376, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1502, 1254, 0 },
   { 1505, 1254, 0 },
   { 1581, 1583, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1352, 1283, 0 },
   { 1374, 1376, 0 },
   { 1502, 1254, 0 },
   { 1505, 1254, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1269, 1270, 0 },
   { 1335, 32, 0 },
   { 1502, 1254, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1352, 1283, 0 },
   { 1374, 1376, 0 },
   { 1502, 1254, 0 },
   { 1581, 1583, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 1509, 32, 0 },
   { 1513, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 1513, 103, 0 },
   { 1575, 1578, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 1346, 1281, 0 },
   { 1359, 1361, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 1509, 32, 0 },
   { 1513, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 1513, 103, 0 },
   { 1575, 1578, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 1587, 1588, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 1519, 217, 38 },
   { 1524, 1522, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 1346, 1281, 0 },
   { 1359, 1361, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 1513, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 1346, 1281, 0 },
   { 1359, 1361, 0 },
   { 1513, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 1509, 32, 0 },
   { 1511, 103, 0 },
   { 1513, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 1511, 103, 0 },
   { 1513, 103, 0 },
   { 1575, 1578, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 1511, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 1346, 1281, 0 },
   { 1359, 1361, 0 },
   { 1511, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 1346, 1281, 0 },
   { 1509, 32, 0 },
   { 1513, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 1346, 1281, 0 },
   { 1513, 103, 0 },
   { 1575, 1578, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 1346, 1281, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 1509, 32, 0 },
   { 1513, 103, 0 },
   { 1575, 1578, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 1509, 32, 0 },
   { 1513, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 1346, 1281, 0 },
   { 1359, 1361, 0 },
   { 1509, 32, 0 },
   { 1513, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1265, 1267, 0 },
   { 1336, 32, 0 },
   { 1513, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 1513, 103, 0 },
   { 1575, 1578, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 1346, 1281, 0 },
   { 1359, 1361, 0 },
   { 1513, 103, 0 },
   { 1575, 1578, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 1346, 1281, 0 },
   { 1359, 1361, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 217, 1517, 40 },
   { 1533, 1535, 38 },
   { 3156, 111, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 217, 1517, 40 },
   { 1538, 1540, 38 },
   { 3156, 111, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 217, 1517, 40 },
   { 1543, 1545, 41 },
   { 3156, 111, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 217, 1517, 40 },
   { 1547, 1551, 38 },
   { 3156, 111, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 217, 1517, 40 },
   { 3156, 111, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 217, 1517, 40 },
   { 2681, 1263, 0 },
   { 3156, 111, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 205, 111, 0 },
   { 728, 707, 24 },
   { 729, 710, 24 },
   { 730, 713, 25 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 216, 3, 0 },
   { 3581, 3510, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 32, 0 },
   { 114, 11, 5 },
   { 116, 16, 6 },
   { 205, 111, 0 },
   { 3541, 3544, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 32, 0 },
   { 114, 11, 5 },
   { 116, 16, 6 },
   { 205, 111, 0 },
   { 3550, 3554, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 32, 0 },
   { 114, 11, 5 },
   { 116, 16, 6 },
   { 205, 111, 0 },
   { 3560, 3563, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 32, 0 },
   { 114, 11, 5 },
   { 116, 16, 6 },
   { 205, 111, 0 },
   { 810, 812, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 32, 0 },
   { 114, 11, 5 },
   { 116, 16, 6 },
   { 205, 111, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 32, 0 },
   { 114, 11, 5 },
   { 116, 16, 6 },
   { 205, 111, 0 },
   { 728, 707, 24 },
   { 729, 710, 24 },
   { 730, 713, 25 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 32, 0 },
   { 114, 11, 5 },
   { 116, 16, 6 },
   { 216, 3, 0 },
   { 3508, 3510, 0 },
   { 3512, 3510, 0 },
   { 3581, 3510, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 32, 0 },
   { 114, 11, 5 },
   { 116, 16, 6 },
   { 216, 3, 0 },
   { 3508, 3510, 0 },
   { 3512, 3510, 0 },
   { 3539, 3510, 0 },
   { 3581, 3510, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 205, 111, 0 },
   { 214, 3, 0 },
   { 216, 3, 0 },
   { 3581, 3510, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 205, 111, 0 },
   { 214, 3, 0 },
   { 728, 707, 24 },
   { 729, 710, 24 },
   { 730, 713, 25 },
   { 3581, 3510, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 219, 220, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 205, 111, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 205, 111, 0 },
   { 214, 3, 0 },
   { 216, 3, 0 },
   { 728, 707, 24 },
   { 729, 710, 24 },
   { 730, 713, 25 },
   { 3581, 3510, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 131, 133, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2815, 2818, 51 },
   { 2840, 32, 0 },
   { 2841, 251, 0 },
   { 2842, 2839, 0 },
   { 2849, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2819, 2815, 52 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2819, 2815, 52 },
   { 2831, 2832, 0 },
   { 2835, 251, 0 },
   { 2838, 2839, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { 1503, 1504, 0 },
   { 1506, 1504, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { 1356, 1303, 0 },
   { 1384, 1386, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { 1503, 1504, 0 },
   { 1506, 1504, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { 1356, 1303, 0 },
   { 1384, 1386, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { 1356, 1303, 0 },
   { 1384, 1386, 0 },
   { 1503, 1504, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { 1503, 1504, 0 },
   { 1506, 1504, 0 },
   { 1508, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { 1356, 1303, 0 },
   { 1384, 1386, 0 },
   { 1508, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { 1356, 1303, 0 },
   { 1503, 1504, 0 },
   { 1506, 1504, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { 1503, 1504, 0 },
   { 1506, 1504, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { 1356, 1303, 0 },
   { 1384, 1386, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { 1356, 1303, 0 },
   { 1384, 1386, 0 },
   { 1503, 1504, 0 },
   { 1506, 1504, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { 1503, 1504, 0 },
   { 3100, 3089, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1338, 32, 0 },
   { 1510, 32, 0 },
   { 1514, 106, 39 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1338, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1338, 32, 0 },
   { 1350, 1301, 0 },
   { 1369, 1371, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1338, 32, 0 },
   { 1510, 32, 0 },
   { 1514, 106, 39 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1338, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1338, 32, 0 },
   { 4002, 3998, 103 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1338, 32, 0 },
   { 1350, 1301, 0 },
   { 1369, 1371, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1338, 32, 0 },
   { 1514, 106, 39 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1338, 32, 0 },
   { 1350, 1301, 0 },
   { 1369, 1371, 0 },
   { 1514, 106, 39 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1338, 32, 0 },
   { 1510, 32, 0 },
   { 1512, 106, 0 },
   { 1514, 106, 39 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1338, 32, 0 },
   { 1512, 106, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1338, 32, 0 },
   { 1350, 1301, 0 },
   { 1369, 1371, 0 },
   { 1512, 106, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1338, 32, 0 },
   { 1350, 1301, 0 },
   { 1510, 32, 0 },
   { 1514, 106, 39 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1338, 32, 0 },
   { 1350, 1301, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1338, 32, 0 },
   { 1510, 32, 0 },
   { 1514, 106, 39 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1338, 32, 0 },
   { 1350, 1301, 0 },
   { 1369, 1371, 0 },
   { 1510, 32, 0 },
   { 1514, 106, 39 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1338, 32, 0 },
   { 1514, 106, 39 },
   { 3099, 3087, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1338, 32, 0 },
   { 1350, 1301, 0 },
   { 1369, 1371, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1107, 1108, 0 },
   { 2691, 1235, 0 },
   { 3040, 760, 0 },
   { 3043, 1131, 0 },
   { 3090, 760, 0 },
   { 3192, 1077, 0 },
   { 3792, 3793, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1109, 1110, 0 },
   { 2692, 1077, 0 },
   { 3041, 1131, 0 },
   { 3042, 760, 0 },
   { 3091, 1131, 0 },
   { 3191, 1235, 0 },
   { 3789, 3791, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2251, 1391, 0 },
   { 2692, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2689, 1077, 0 },
   { 3093, 760, 0 },
   { 3196, 1077, 0 },
   { 3197, 1235, 0 },
   { 3198, 1235, 0 },
   { 1489, 3787, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2330, 1077, 0 },
   { 2693, 1077, 0 },
   { 3832, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2693, 1077, 0 },
   { 3094, 760, 0 },
   { 3200, 1235, 0 },
   { 3794, 3791, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2330, 1077, 0 },
   { 2693, 1077, 0 },
   { 3200, 1235, 0 },
   { 3794, 3791, 81 },
   { 3832, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2330, 1077, 0 },
   { 2382, 2383, 0 },
   { 2693, 1077, 0 },
   { 3200, 1235, 0 },
   { 3794, 3791, 81 },
   { 3832, 1077, 0 },
   { 3836, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2330, 1077, 0 },
   { 2693, 1077, 0 },
   { 3832, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2330, 1077, 0 },
   { 2339, 1077, 0 },
   { 2693, 1077, 0 },
   { 3832, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2320, 1085, 0 },
   { 2330, 1077, 0 },
   { 2693, 1077, 0 },
   { 3832, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2330, 1077, 0 },
   { 2331, 1077, 0 },
   { 2693, 1077, 0 },
   { 3832, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2690, 1235, 0 },
   { 3092, 1131, 0 },
   { 3193, 1235, 0 },
   { 3194, 1077, 0 },
   { 3195, 1077, 0 },
   { 1486, 3788, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { 1354, 1293, 0 },
   { 1379, 1381, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { 1340, 61, 0 },
   { 1341, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { 1340, 61, 0 },
   { 1341, 32, 0 },
   { 1354, 1293, 0 },
   { 1379, 1381, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { 1354, 1293, 0 },
   { 1379, 1381, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { 1354, 1293, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { 1354, 1293, 0 },
   { 1379, 1381, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { 3102, 3087, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { 1354, 1293, 0 },
   { 1379, 1381, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1342, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1342, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1342, 32, 0 },
   { 1348, 1291, 0 },
   { 1364, 1366, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1342, 32, 0 },
   { 1343, 32, 0 },
   { 1344, 63, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1342, 32, 0 },
   { 1343, 32, 0 },
   { 1344, 63, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1342, 32, 0 },
   { 1343, 32, 0 },
   { 1344, 63, 0 },
   { 1348, 1291, 0 },
   { 1364, 1366, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1342, 32, 0 },
   { 1348, 1291, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1342, 32, 0 },
   { 1348, 1291, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1342, 32, 0 },
   { 3101, 3089, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1342, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1342, 32, 0 },
   { 1348, 1291, 0 },
   { 1364, 1366, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1342, 32, 0 },
   { 1348, 1291, 0 },
   { 1364, 1366, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2252, 2253, 0 },
   { 2257, 2272, 1 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2252, 2253, 0 },
   { 3862, 3860, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2252, 2253, 0 },
   { 3865, 3860, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1592, 1591, 0 },
   { 3916, 3919, 91 },
   { 5078, 11, 0 },
   { 5131, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5128, 2089, 0 },
   { 5181, 2093, 0 },
   { 5233, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5127, 2089, 0 },
   { 5180, 2093, 0 },
   { 5232, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5103, 2089, 0 },
   { 5156, 2093, 0 },
   { 5208, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5102, 2089, 0 },
   { 5155, 2093, 0 },
   { 5207, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 2710, 32, 0 },
   { 2713, 2703, 0 },
   { 2752, 2754, 0 },
   { 2755, 2757, 0 },
   { 5128, 2089, 0 },
   { 5181, 2093, 0 },
   { 5233, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 2710, 32, 0 },
   { 2713, 2703, 0 },
   { 2752, 2754, 0 },
   { 2755, 2757, 0 },
   { 5127, 2089, 0 },
   { 5180, 2093, 0 },
   { 5232, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 2710, 32, 0 },
   { 2713, 2703, 0 },
   { 2752, 2754, 0 },
   { 2755, 2757, 0 },
   { 5103, 2089, 0 },
   { 5156, 2093, 0 },
   { 5208, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 2710, 32, 0 },
   { 2713, 2703, 0 },
   { 2752, 2754, 0 },
   { 2755, 2757, 0 },
   { 5102, 2089, 0 },
   { 5155, 2093, 0 },
   { 5207, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5104, 2089, 0 },
   { 5128, 2089, 0 },
   { 5157, 2093, 0 },
   { 5181, 2093, 0 },
   { 5233, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5104, 2089, 0 },
   { 5127, 2089, 0 },
   { 5157, 2093, 0 },
   { 5180, 2093, 0 },
   { 5232, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5103, 2089, 0 },
   { 5104, 2089, 0 },
   { 5156, 2093, 0 },
   { 5157, 2093, 0 },
   { 5208, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5102, 2089, 0 },
   { 5104, 2089, 0 },
   { 5155, 2093, 0 },
   { 5157, 2093, 0 },
   { 5207, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5128, 2089, 0 },
   { 5129, 2089, 0 },
   { 5181, 2093, 0 },
   { 5182, 2093, 0 },
   { 5233, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5127, 2089, 0 },
   { 5129, 2089, 0 },
   { 5180, 2093, 0 },
   { 5182, 2093, 0 },
   { 5232, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5103, 2089, 0 },
   { 5129, 2089, 0 },
   { 5156, 2093, 0 },
   { 5182, 2093, 0 },
   { 5208, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5102, 2089, 0 },
   { 5129, 2089, 0 },
   { 5155, 2093, 0 },
   { 5182, 2093, 0 },
   { 5207, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5128, 2089, 0 },
   { 5181, 2093, 0 },
   { 5233, 2093, 0 },
   { 5234, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5127, 2089, 0 },
   { 5180, 2093, 0 },
   { 5232, 2093, 0 },
   { 5234, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5103, 2089, 0 },
   { 5156, 2093, 0 },
   { 5208, 2093, 0 },
   { 5234, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5102, 2089, 0 },
   { 5155, 2093, 0 },
   { 5207, 2093, 0 },
   { 5234, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5128, 2089, 0 },
   { 5181, 2093, 0 },
   { 5209, 2093, 0 },
   { 5233, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5127, 2089, 0 },
   { 5180, 2093, 0 },
   { 5209, 2093, 0 },
   { 5232, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5103, 2089, 0 },
   { 5156, 2093, 0 },
   { 5208, 2093, 0 },
   { 5209, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5102, 2089, 0 },
   { 5155, 2093, 0 },
   { 5207, 2093, 0 },
   { 5209, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 2712, 3, 0 },
   { 5128, 2089, 0 },
   { 5181, 2093, 0 },
   { 5233, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 2712, 3, 0 },
   { 5127, 2089, 0 },
   { 5180, 2093, 0 },
   { 5232, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 2712, 3, 0 },
   { 5103, 2089, 0 },
   { 5156, 2093, 0 },
   { 5208, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 2712, 3, 0 },
   { 5102, 2089, 0 },
   { 5155, 2093, 0 },
   { 5207, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5127, 2089, 0 },
   { 5128, 2089, 0 },
   { 5180, 2093, 0 },
   { 5181, 2093, 0 },
   { 5232, 2093, 0 },
   { 5233, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5103, 2089, 0 },
   { 5128, 2089, 0 },
   { 5156, 2093, 0 },
   { 5181, 2093, 0 },
   { 5208, 2093, 0 },
   { 5233, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5102, 2089, 0 },
   { 5128, 2089, 0 },
   { 5155, 2093, 0 },
   { 5181, 2093, 0 },
   { 5207, 2093, 0 },
   { 5233, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5103, 2089, 0 },
   { 5127, 2089, 0 },
   { 5156, 2093, 0 },
   { 5180, 2093, 0 },
   { 5208, 2093, 0 },
   { 5232, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5102, 2089, 0 },
   { 5127, 2089, 0 },
   { 5155, 2093, 0 },
   { 5180, 2093, 0 },
   { 5207, 2093, 0 },
   { 5232, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5102, 2089, 0 },
   { 5103, 2089, 0 },
   { 5155, 2093, 0 },
   { 5156, 2093, 0 },
   { 5207, 2093, 0 },
   { 5208, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2709, 61, 0 },
   { 5102, 2089, 0 },
   { 5155, 2093, 0 },
   { 5207, 2093, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1807, 11, 0 },
   { 1830, 11, 0 },
   { 1912, 16, 0 },
   { 2018, 2015, 0 },
   { 4236, 760, 0 },
   { 4237, 6, 0 },
   { 4238, 11, 0 },
   { 4239, 16, 0 },
   { 4240, 21, 0 },
   { 4255, 4253, 0 },
   { 4261, 4253, 0 },
   { 4267, 4253, 0 },
   { 4272, 4253, 0 },
   { 4275, 4276, 0 },
   { 4280, 4251, 0 },
   { 4284, 4251, 0 },
   { 4288, 4251, 0 },
   { 4291, 4292, 0 },
   { 4294, 4292, 0 },
   { 4297, 4298, 0 },
   { 4302, 4258, 0 },
   { 4306, 4258, 0 },
   { 4309, 4310, 0 },
   { 4312, 4310, 0 },
   { 4315, 3282, 0 },
   { 4317, 4310, 0 },
   { 4319, 3282, 0 },
   { 4321, 1806, 0 },
   { 4325, 4264, 0 },
   { 4328, 4329, 0 },
   { 4331, 4329, 0 },
   { 4334, 4335, 0 },
   { 4337, 4329, 0 },
   { 4339, 4335, 0 },
   { 4341, 1829, 0 },
   { 4343, 4329, 0 },
   { 4345, 4335, 0 },
   { 4347, 1829, 0 },
   { 4349, 1911, 0 },
   { 5077, 11, 0 },
   { 5130, 16, 0 },
   { 5183, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1807, 11, 0 },
   { 1830, 11, 0 },
   { 1912, 16, 0 },
   { 3260, 3261, 0 },
   { 4236, 760, 0 },
   { 4237, 6, 0 },
   { 4238, 11, 0 },
   { 4239, 16, 0 },
   { 4240, 21, 0 },
   { 4255, 4253, 0 },
   { 4261, 4253, 0 },
   { 4267, 4253, 0 },
   { 4272, 4253, 0 },
   { 4275, 4276, 0 },
   { 4280, 4251, 0 },
   { 4284, 4251, 0 },
   { 4288, 4251, 0 },
   { 4291, 4292, 0 },
   { 4294, 4292, 0 },
   { 4297, 4298, 0 },
   { 4302, 4258, 0 },
   { 4306, 4258, 0 },
   { 4309, 4310, 0 },
   { 4312, 4310, 0 },
   { 4315, 3282, 0 },
   { 4317, 4310, 0 },
   { 4319, 3282, 0 },
   { 4321, 1806, 0 },
   { 4325, 4264, 0 },
   { 4328, 4329, 0 },
   { 4331, 4329, 0 },
   { 4334, 4335, 0 },
   { 4337, 4329, 0 },
   { 4339, 4335, 0 },
   { 4341, 1829, 0 },
   { 4343, 4329, 0 },
   { 4345, 4335, 0 },
   { 4347, 1829, 0 },
   { 4349, 1911, 0 },
   { 5077, 11, 0 },
   { 5130, 16, 0 },
   { 5183, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3257, 3259, 0 },
   { 4231, 760, 0 },
   { 4232, 6, 0 },
   { 4233, 11, 0 },
   { 4234, 16, 0 },
   { 4235, 21, 0 },
   { 4252, 4253, 0 },
   { 4259, 4253, 0 },
   { 4265, 4253, 0 },
   { 4270, 4253, 0 },
   { 4278, 4251, 0 },
   { 4282, 4251, 0 },
   { 4286, 4251, 0 },
   { 4300, 4258, 0 },
   { 4304, 4258, 0 },
   { 4323, 4264, 0 },
   { 5105, 11, 0 },
   { 5158, 16, 0 },
   { 5210, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2130, 6, 0 },
   { 2161, 11, 0 },
   { 2191, 16, 0 },
   { 2221, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { 1565, 1566, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { 1340, 61, 0 },
   { 1341, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { 1340, 61, 0 },
   { 1341, 32, 0 },
   { 1565, 1566, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { 1565, 1566, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { 1354, 1293, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { 1354, 1293, 0 },
   { 1565, 1566, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { 1565, 1566, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { 1354, 1293, 0 },
   { 1379, 1381, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { 1354, 1293, 0 },
   { 1379, 1381, 0 },
   { 1565, 1566, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1339, 32, 0 },
   { 1565, 1566, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3630, 3619, 65 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3650, 3614, 66 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3632, 3619, 65 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3652, 3614, 66 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3628, 3619, 65 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3631, 3619, 65 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3649, 3614, 66 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3651, 3614, 66 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3621, 3613, 65 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3626, 3613, 65 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3623, 3619, 65 },
   { 3630, 3619, 65 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3627, 3619, 65 },
   { 3632, 3619, 65 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3642, 3587, 66 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3647, 3587, 66 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3644, 3614, 66 },
   { 3650, 3614, 66 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3648, 3614, 66 },
   { 3652, 3614, 66 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3612, 3613, 65 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3624, 3613, 65 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3618, 3619, 65 },
   { 3628, 3619, 65 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3625, 3619, 65 },
   { 3631, 3619, 65 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3636, 3587, 66 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3645, 3587, 66 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1333, 1146, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3640, 3614, 66 },
   { 3649, 3614, 66 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 3077, 3, 0 },
   { 3646, 3614, 66 },
   { 3651, 3614, 66 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2775, 2776, 49 },
   { 2782, 2783, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2788, 2789, 49 },
   { 2793, 2794, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2775, 2776, 49 },
   { 2782, 2783, 49 },
   { 2791, 2789, 49 },
   { 2796, 2794, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2716, 2338, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2779, 2776, 49 },
   { 2786, 2783, 49 },
   { 2788, 2789, 49 },
   { 2793, 2794, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2775, 2776, 49 },
   { 2782, 2783, 49 },
   { 2788, 2789, 49 },
   { 2793, 2794, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 945, 946, 0 },
   { 992, 991, 0 },
   { 999, 998, 0 },
   { 1007, 1005, 0 },
   { 1016, 1014, 0 },
   { 2683, 2684, 0 },
   { 2686, 2687, 0 },
   { 2700, 32, 0 },
   { 2701, 61, 0 },
   { 2702, 32, 0 },
   { 2725, 2729, 0 },
   { 2731, 2735, 0 },
   { 2737, 2738, 0 },
   { 2740, 2741, 0 },
   { 3245, 3246, 58 },
   { 3276, 3277, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 994, 993, 0 },
   { 1001, 1000, 0 },
   { 1009, 1008, 0 },
   { 1018, 1017, 0 },
   { 2683, 2684, 0 },
   { 2686, 2687, 0 },
   { 2700, 32, 0 },
   { 2701, 61, 0 },
   { 2702, 32, 0 },
   { 2725, 2729, 0 },
   { 2731, 2735, 0 },
   { 2737, 2738, 0 },
   { 2740, 2741, 0 },
   { 3245, 3246, 58 },
   { 3276, 3277, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 501, 491, 15 },
   { 571, 562, 17 },
   { 642, 632, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 501, 491, 15 },
   { 511, 509, 14 },
   { 571, 562, 17 },
   { 580, 578, 16 },
   { 642, 632, 19 },
   { 651, 649, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 216, 3, 0 },
   { 718, 719, 24 },
   { 722, 723, 24 },
   { 726, 727, 25 },
   { 3581, 3510, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 32, 0 },
   { 114, 11, 5 },
   { 116, 16, 6 },
   { 501, 491, 15 },
   { 571, 562, 17 },
   { 642, 632, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 32, 0 },
   { 114, 11, 5 },
   { 116, 16, 6 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 32, 0 },
   { 114, 11, 5 },
   { 116, 16, 6 },
   { 501, 491, 15 },
   { 511, 509, 14 },
   { 571, 562, 17 },
   { 580, 578, 16 },
   { 642, 632, 19 },
   { 651, 649, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 32, 0 },
   { 114, 11, 5 },
   { 116, 16, 6 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 32, 0 },
   { 114, 11, 5 },
   { 116, 16, 6 },
   { 216, 3, 0 },
   { 718, 719, 24 },
   { 722, 723, 24 },
   { 726, 727, 25 },
   { 3508, 3510, 0 },
   { 3512, 3510, 0 },
   { 3581, 3510, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 32, 0 },
   { 114, 11, 5 },
   { 116, 16, 6 },
   { 216, 3, 0 },
   { 718, 719, 24 },
   { 722, 723, 24 },
   { 726, 727, 25 },
   { 3508, 3510, 0 },
   { 3512, 3510, 0 },
   { 3539, 3510, 0 },
   { 3581, 3510, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 32, 0 },
   { 114, 11, 5 },
   { 116, 16, 6 },
   { 205, 111, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 205, 111, 0 },
   { 501, 491, 15 },
   { 571, 562, 17 },
   { 642, 632, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 205, 111, 0 },
   { 501, 491, 15 },
   { 511, 509, 14 },
   { 571, 562, 17 },
   { 580, 578, 16 },
   { 642, 632, 19 },
   { 651, 649, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 205, 111, 0 },
   { 214, 3, 0 },
   { 216, 3, 0 },
   { 718, 719, 24 },
   { 722, 723, 24 },
   { 726, 727, 25 },
   { 3581, 3510, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 501, 491, 15 },
   { 571, 562, 17 },
   { 642, 632, 19 },
   { 3581, 3510, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 501, 491, 15 },
   { 511, 509, 14 },
   { 571, 562, 17 },
   { 580, 578, 16 },
   { 642, 632, 19 },
   { 651, 649, 18 },
   { 3581, 3510, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 131, 133, 0 },
   { 501, 491, 15 },
   { 571, 562, 17 },
   { 642, 632, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 131, 133, 0 },
   { 490, 491, 15 },
   { 501, 491, 15 },
   { 561, 562, 17 },
   { 571, 562, 17 },
   { 631, 632, 19 },
   { 642, 632, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 131, 133, 0 },
   { 501, 491, 15 },
   { 511, 509, 14 },
   { 571, 562, 17 },
   { 580, 578, 16 },
   { 642, 632, 19 },
   { 651, 649, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 131, 133, 0 },
   { 490, 491, 15 },
   { 501, 491, 15 },
   { 511, 509, 14 },
   { 561, 562, 17 },
   { 571, 562, 17 },
   { 580, 578, 16 },
   { 631, 632, 19 },
   { 642, 632, 19 },
   { 651, 649, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 155, 157, 0 },
   { 501, 491, 15 },
   { 571, 562, 17 },
   { 642, 632, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 155, 157, 0 },
   { 501, 491, 15 },
   { 511, 509, 14 },
   { 571, 562, 17 },
   { 580, 578, 16 },
   { 642, 632, 19 },
   { 651, 649, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 205, 111, 0 },
   { 501, 491, 15 },
   { 571, 562, 17 },
   { 642, 632, 19 },
   { 728, 707, 24 },
   { 729, 710, 24 },
   { 730, 713, 25 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 205, 111, 0 },
   { 501, 491, 15 },
   { 511, 509, 14 },
   { 571, 562, 17 },
   { 580, 578, 16 },
   { 642, 632, 19 },
   { 651, 649, 18 },
   { 728, 707, 24 },
   { 729, 710, 24 },
   { 730, 713, 25 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 205, 111, 0 },
   { 214, 3, 0 },
   { 216, 3, 0 },
   { 718, 719, 24 },
   { 722, 723, 24 },
   { 726, 727, 25 },
   { 728, 707, 24 },
   { 729, 710, 24 },
   { 730, 713, 25 },
   { 3581, 3510, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 216, 3, 0 },
   { 501, 491, 15 },
   { 571, 562, 17 },
   { 642, 632, 19 },
   { 3581, 3510, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 216, 3, 0 },
   { 501, 491, 15 },
   { 511, 509, 14 },
   { 571, 562, 17 },
   { 580, 578, 16 },
   { 642, 632, 19 },
   { 651, 649, 18 },
   { 3581, 3510, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 131, 133, 0 },
   { 501, 491, 15 },
   { 571, 562, 17 },
   { 642, 632, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 131, 133, 0 },
   { 490, 491, 15 },
   { 501, 491, 15 },
   { 561, 562, 17 },
   { 571, 562, 17 },
   { 631, 632, 19 },
   { 642, 632, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 205, 111, 0 },
   { 1793, 1794, 42 },
   { 1898, 1899, 42 },
   { 1978, 1979, 42 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 216, 3, 0 },
   { 501, 491, 15 },
   { 571, 562, 17 },
   { 642, 632, 19 },
   { 718, 719, 24 },
   { 722, 723, 24 },
   { 726, 727, 25 },
   { 3581, 3510, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 216, 3, 0 },
   { 501, 491, 15 },
   { 511, 509, 14 },
   { 571, 562, 17 },
   { 580, 578, 16 },
   { 642, 632, 19 },
   { 651, 649, 18 },
   { 718, 719, 24 },
   { 722, 723, 24 },
   { 726, 727, 25 },
   { 3581, 3510, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2064, 2066, 0 },
   { 2068, 2070, 0 },
   { 4236, 760, 0 },
   { 4237, 6, 0 },
   { 4238, 11, 0 },
   { 4239, 16, 0 },
   { 4240, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2022, 2024, 0 },
   { 2030, 2032, 0 },
   { 4236, 760, 0 },
   { 4237, 6, 0 },
   { 4238, 11, 0 },
   { 4239, 16, 0 },
   { 4240, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2088, 2090, 0 },
   { 2092, 2094, 0 },
   { 4236, 760, 0 },
   { 4237, 6, 0 },
   { 4238, 11, 0 },
   { 4239, 16, 0 },
   { 4240, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 990, 991, 0 },
   { 997, 998, 0 },
   { 1004, 1005, 0 },
   { 1013, 1014, 0 },
   { 2765, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 2258, 2271, 0 },
   { 2765, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2771, 32, 0 },
   { 3234, 3235, 58 },
   { 1017, 422, 58 },
   { 3237, 3239, 58 },
   { 1008, 447, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1555, 1556, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1502, 1254, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1555, 1556, 0 },
   { 3157, 217, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1517, 217, 38 },
   { 1521, 1522, 38 },
   { 1555, 1556, 0 },
   { 3157, 217, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1502, 1254, 0 },
   { 3157, 217, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1502, 1254, 0 },
   { 1555, 1556, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1507, 32, 0 },
   { 1555, 1556, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1502, 1254, 0 },
   { 1507, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1352, 1283, 0 },
   { 1555, 1556, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1352, 1283, 0 },
   { 1502, 1254, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1502, 1254, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1502, 1254, 0 },
   { 1505, 1254, 0 },
   { 1555, 1556, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1502, 1254, 0 },
   { 1505, 1254, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1502, 1254, 0 },
   { 1555, 1556, 0 },
   { 1581, 1583, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1502, 1254, 0 },
   { 1581, 1583, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1352, 1283, 0 },
   { 1374, 1376, 0 },
   { 1555, 1556, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1352, 1283, 0 },
   { 1374, 1376, 0 },
   { 1502, 1254, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1335, 32, 0 },
   { 1502, 1254, 0 },
   { 1555, 1556, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 221, 111, 0 },
   { 222, 113, 5 },
   { 223, 115, 6 },
   { 242, 32, 0 },
   { 245, 32, 0 },
   { 248, 98, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 221, 111, 0 },
   { 222, 113, 5 },
   { 223, 115, 6 },
   { 242, 32, 0 },
   { 245, 32, 0 },
   { 248, 98, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3468, 3469, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3468, 3469, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3658, 3659, 0 },
   { 3661, 3662, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3658, 3659, 0 },
   { 3661, 3662, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 383, 269, 9 },
   { 3658, 3659, 0 },
   { 3661, 3662, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 383, 269, 9 },
   { 3658, 3659, 0 },
   { 3661, 3662, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3468, 3469, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 255, 253, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 255, 253, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 147, 148, 0 },
   { 514, 509, 0 },
   { 583, 578, 0 },
   { 654, 649, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 147, 148, 0 },
   { 257, 3, 0 },
   { 259, 258, 5 },
   { 260, 36, 6 },
   { 271, 272, 0 },
   { 276, 277, 0 },
   { 514, 509, 0 },
   { 583, 578, 0 },
   { 654, 649, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 147, 148, 0 },
   { 514, 509, 0 },
   { 583, 578, 0 },
   { 654, 649, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 147, 148, 0 },
   { 160, 162, 0 },
   { 514, 509, 0 },
   { 583, 578, 0 },
   { 654, 649, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 147, 148, 0 },
   { 514, 509, 0 },
   { 524, 525, 0 },
   { 583, 578, 0 },
   { 591, 592, 0 },
   { 654, 649, 0 },
   { 663, 664, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 147, 148, 0 },
   { 386, 387, 9 },
   { 514, 509, 0 },
   { 583, 578, 0 },
   { 654, 649, 0 },
   { 737, 739, 29 },
   { 234, 741, 30 },
   { 742, 744, 31 },
   { 745, 739, 32 },
   { 746, 741, 33 },
   { 747, 744, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1161, 1162, 0 },
   { 1191, 1192, 0 },
   { 2698, 1235, 0 },
   { 3173, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1150, 1151, 0 },
   { 1193, 1194, 0 },
   { 2699, 1077, 0 },
   { 3172, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2825, 2828, 52 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1095, 1096, 0 },
   { 1167, 1138, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1099, 1100, 0 },
   { 1169, 1155, 0 },
   { 2697, 1235, 0 },
   { 3176, 1235, 0 },
   { 3179, 1077, 0 },
   { 3182, 1077, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { 1560, 1561, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { 1560, 1561, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { 1560, 1561, 0 },
   { 4000, 3998, 103 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { 1503, 1504, 0 },
   { 1560, 1561, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { 1508, 32, 0 },
   { 1560, 1561, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { 1356, 1303, 0 },
   { 1560, 1561, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { 1503, 1504, 0 },
   { 1506, 1504, 0 },
   { 1560, 1561, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1337, 32, 0 },
   { 1356, 1303, 0 },
   { 1384, 1386, 0 },
   { 1560, 1561, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2692, 1077, 0 },
   { 3041, 1131, 0 },
   { 3042, 760, 0 },
   { 3108, 3110, 0 },
   { 3191, 1235, 0 },
   { 3789, 3791, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2692, 1077, 0 },
   { 3041, 1131, 0 },
   { 3042, 760, 0 },
   { 3112, 3113, 0 },
   { 3191, 1235, 0 },
   { 3789, 3791, 81 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 206, 61, 0 },
   { 1800, 1801, 42 },
   { 1905, 1906, 42 },
   { 1985, 1986, 42 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3842, 3843, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3823, 3814, 0 },
   { 3842, 3843, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3844, 3845, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3825, 3811, 0 },
   { 3844, 3845, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3848, 3849, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3848, 3849, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3827, 3828, 0 },
   { 3848, 3849, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 2716, 2338, 0 },
   { 3293, 3295, 60 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 196, 198, 0 },
   { 2705, 32, 0 },
   { 2715, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1187, 1188, 0 },
   { 1196, 1199, 0 },
   { 2695, 1077, 0 },
   { 3161, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1187, 1188, 0 },
   { 1196, 1199, 0 },
   { 1277, 1049, 0 },
   { 2695, 1077, 0 },
   { 3165, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1183, 1186, 0 },
   { 2301, 1078, 0 },
   { 2695, 1077, 0 },
   { 3184, 1077, 0 },
   { 3187, 1235, 0 },
   { 3190, 1235, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 3168, 3160, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1336, 32, 0 },
   { 1587, 1588, 38 },
   { 3170, 3164, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 221, 111, 0 },
   { 222, 113, 5 },
   { 223, 115, 6 },
   { 242, 32, 0 },
   { 245, 32, 0 },
   { 248, 98, 0 },
   { 759, 767, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 32, 0 },
   { 114, 11, 5 },
   { 116, 16, 6 },
   { 205, 111, 0 },
   { 538, 539, 15 },
   { 604, 605, 17 },
   { 677, 678, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 131, 133, 0 },
   { 496, 497, 15 },
   { 567, 568, 17 },
   { 637, 638, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 131, 133, 0 },
   { 496, 497, 15 },
   { 501, 491, 15 },
   { 567, 568, 17 },
   { 571, 562, 17 },
   { 637, 638, 19 },
   { 642, 632, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 131, 133, 0 },
   { 490, 491, 15 },
   { 496, 497, 15 },
   { 501, 491, 15 },
   { 561, 562, 17 },
   { 567, 568, 17 },
   { 571, 562, 17 },
   { 631, 632, 19 },
   { 637, 638, 19 },
   { 642, 632, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 131, 133, 0 },
   { 490, 491, 15 },
   { 501, 491, 15 },
   { 508, 509, 14 },
   { 561, 562, 17 },
   { 571, 562, 17 },
   { 577, 578, 16 },
   { 631, 632, 19 },
   { 642, 632, 19 },
   { 648, 649, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 131, 133, 0 },
   { 490, 491, 15 },
   { 501, 491, 15 },
   { 508, 509, 14 },
   { 511, 509, 14 },
   { 561, 562, 17 },
   { 571, 562, 17 },
   { 577, 578, 16 },
   { 580, 578, 16 },
   { 631, 632, 19 },
   { 642, 632, 19 },
   { 648, 649, 18 },
   { 651, 649, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 131, 133, 0 },
   { 490, 491, 15 },
   { 496, 497, 15 },
   { 501, 491, 15 },
   { 508, 509, 14 },
   { 561, 562, 17 },
   { 567, 568, 17 },
   { 571, 562, 17 },
   { 577, 578, 16 },
   { 631, 632, 19 },
   { 637, 638, 19 },
   { 642, 632, 19 },
   { 648, 649, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 131, 133, 0 },
   { 496, 497, 15 },
   { 567, 568, 17 },
   { 637, 638, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 131, 133, 0 },
   { 490, 491, 15 },
   { 501, 491, 15 },
   { 508, 509, 14 },
   { 561, 562, 17 },
   { 571, 562, 17 },
   { 577, 578, 16 },
   { 631, 632, 19 },
   { 642, 632, 19 },
   { 648, 649, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2284, 2287, 0 },
   { 2700, 32, 0 },
   { 2704, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2292, 2296, 0 },
   { 2700, 32, 0 },
   { 2704, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2284, 2287, 0 },
   { 2292, 2296, 0 },
   { 2700, 32, 0 },
   { 2704, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 2765, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2771, 32, 0 },
   { 3234, 3235, 58 },
   { 1017, 422, 58 },
   { 3237, 3239, 58 },
   { 1008, 447, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2683, 2684, 0 },
   { 2686, 2687, 0 },
   { 2700, 32, 0 },
   { 2701, 61, 0 },
   { 2702, 32, 0 },
   { 2717, 3, 0 },
   { 2725, 2729, 0 },
   { 2731, 2735, 0 },
   { 2737, 2738, 0 },
   { 2740, 2741, 0 },
   { 3245, 3246, 58 },
   { 3276, 3277, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1331, 1138, 0 },
   { 1332, 1138, 0 },
   { 1334, 1146, 0 },
   { 1769, 1776, 0 },
   { 1874, 1881, 0 },
   { 1954, 1961, 0 },
   { 3077, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 945, 946, 0 },
   { 992, 991, 0 },
   { 999, 998, 0 },
   { 1007, 1005, 0 },
   { 1016, 1014, 0 },
   { 2683, 2684, 0 },
   { 2686, 2687, 0 },
   { 2700, 32, 0 },
   { 2701, 61, 0 },
   { 2702, 32, 0 },
   { 2725, 2729, 0 },
   { 2731, 2735, 0 },
   { 2737, 2738, 0 },
   { 2740, 2741, 0 },
   { 3245, 3246, 58 },
   { 3276, 3277, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 994, 993, 0 },
   { 1001, 1000, 0 },
   { 1009, 1008, 0 },
   { 1018, 1017, 0 },
   { 2683, 2684, 0 },
   { 2686, 2687, 0 },
   { 2700, 32, 0 },
   { 2701, 61, 0 },
   { 2702, 32, 0 },
   { 2725, 2729, 0 },
   { 2731, 2735, 0 },
   { 2737, 2738, 0 },
   { 2740, 2741, 0 },
   { 3245, 3246, 58 },
   { 3276, 3277, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 990, 991, 0 },
   { 997, 998, 0 },
   { 1004, 1005, 0 },
   { 1013, 1014, 0 },
   { 2765, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2771, 32, 0 },
   { 3234, 3235, 58 },
   { 1017, 422, 58 },
   { 3237, 3239, 58 },
   { 1008, 447, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 196, 198, 0 },
   { 2705, 32, 0 },
   { 2715, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 2765, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2771, 32, 0 },
   { 3234, 3235, 58 },
   { 1017, 422, 58 },
   { 3237, 3239, 58 },
   { 1008, 447, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2683, 2684, 0 },
   { 2686, 2687, 0 },
   { 2700, 32, 0 },
   { 2701, 61, 0 },
   { 2702, 32, 0 },
   { 2717, 3, 0 },
   { 2725, 2729, 0 },
   { 2731, 2735, 0 },
   { 2737, 2738, 0 },
   { 2740, 2741, 0 },
   { 3245, 3246, 58 },
   { 3276, 3277, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 945, 946, 0 },
   { 992, 991, 0 },
   { 999, 998, 0 },
   { 1007, 1005, 0 },
   { 1016, 1014, 0 },
   { 2683, 2684, 0 },
   { 2686, 2687, 0 },
   { 2700, 32, 0 },
   { 2701, 61, 0 },
   { 2702, 32, 0 },
   { 2725, 2729, 0 },
   { 2731, 2735, 0 },
   { 2737, 2738, 0 },
   { 2740, 2741, 0 },
   { 3245, 3246, 58 },
   { 3276, 3277, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 994, 993, 0 },
   { 1001, 1000, 0 },
   { 1009, 1008, 0 },
   { 1018, 1017, 0 },
   { 2683, 2684, 0 },
   { 2686, 2687, 0 },
   { 2700, 32, 0 },
   { 2701, 61, 0 },
   { 2702, 32, 0 },
   { 2725, 2729, 0 },
   { 2731, 2735, 0 },
   { 2737, 2738, 0 },
   { 2740, 2741, 0 },
   { 3245, 3246, 58 },
   { 3276, 3277, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 990, 991, 0 },
   { 997, 998, 0 },
   { 1004, 1005, 0 },
   { 1013, 1014, 0 },
   { 2765, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2771, 32, 0 },
   { 3234, 3235, 58 },
   { 1017, 422, 58 },
   { 3237, 3239, 58 },
   { 1008, 447, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 196, 198, 0 },
   { 2705, 32, 0 },
   { 2715, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 2765, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2771, 32, 0 },
   { 3234, 3235, 58 },
   { 1017, 422, 58 },
   { 3237, 3239, 58 },
   { 1008, 447, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2683, 2684, 0 },
   { 2686, 2687, 0 },
   { 2700, 32, 0 },
   { 2701, 61, 0 },
   { 2702, 32, 0 },
   { 2717, 3, 0 },
   { 2725, 2729, 0 },
   { 2731, 2735, 0 },
   { 2737, 2738, 0 },
   { 2740, 2741, 0 },
   { 3245, 3246, 58 },
   { 3276, 3277, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 945, 946, 0 },
   { 992, 991, 0 },
   { 999, 998, 0 },
   { 1007, 1005, 0 },
   { 1016, 1014, 0 },
   { 2683, 2684, 0 },
   { 2686, 2687, 0 },
   { 2700, 32, 0 },
   { 2701, 61, 0 },
   { 2702, 32, 0 },
   { 2725, 2729, 0 },
   { 2731, 2735, 0 },
   { 2737, 2738, 0 },
   { 2740, 2741, 0 },
   { 3245, 3246, 58 },
   { 3276, 3277, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 994, 993, 0 },
   { 1001, 1000, 0 },
   { 1009, 1008, 0 },
   { 1018, 1017, 0 },
   { 2683, 2684, 0 },
   { 2686, 2687, 0 },
   { 2700, 32, 0 },
   { 2701, 61, 0 },
   { 2702, 32, 0 },
   { 2725, 2729, 0 },
   { 2731, 2735, 0 },
   { 2737, 2738, 0 },
   { 2740, 2741, 0 },
   { 3245, 3246, 58 },
   { 3276, 3277, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 990, 991, 0 },
   { 997, 998, 0 },
   { 1004, 1005, 0 },
   { 1013, 1014, 0 },
   { 2765, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2771, 32, 0 },
   { 3234, 3235, 58 },
   { 1017, 422, 58 },
   { 3237, 3239, 58 },
   { 1008, 447, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 196, 198, 0 },
   { 2705, 32, 0 },
   { 2715, 3, 0 },
   { 5477, 5402, 118 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2705, 32, 0 },
   { 5401, 5402, 118 },
   { ~0, ~0, ~0 }, /* Sentinel */

};

static const struct per_op_table nir_opt_algebraic_pass_op_table[nir_num_search_ops] = {
   [nir_op_imul] = {
      .filter = (const uint16_t []) {
         0,
         1,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         0,
         0,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         10,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         13,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         17,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         11,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         5,
         13,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         12,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         17,
         17,
         15,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         6,
         0,
         6,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 18,
      .table = (const uint16_t []) {
      
         2,
         3,
         2,
         2,
         355,
         2,
         2,
         2,
         356,
         357,
         358,
         359,
         2,
         2,
         2233,
         2234,
         2235,
         2236,
         3,
         3,
         360,
         361,
         362,
         363,
         364,
         3,
         365,
         366,
         367,
         368,
         2237,
         2238,
         2239,
         2240,
         2241,
         2242,
         2,
         360,
         2,
         2,
         355,
         2,
         2,
         2,
         356,
         357,
         358,
         359,
         2,
         2,
         2233,
         2234,
         2235,
         2236,
         2,
         361,
         2,
         2,
         355,
         2,
         2,
         2,
         356,
         357,
         358,
         359,
         2,
         2,
         2233,
         2234,
         2235,
         2236,
         355,
         362,
         355,
         355,
         355,
         355,
         355,
         355,
         369,
         370,
         371,
         372,
         355,
         355,
         2243,
         2244,
         2245,
         2246,
         2,
         363,
         2,
         2,
         355,
         2,
         2,
         2,
         356,
         357,
         358,
         359,
         2,
         2,
         2233,
         2234,
         2235,
         2236,
         2,
         364,
         2,
         2,
         355,
         2,
         2,
         2,
         356,
         357,
         358,
         359,
         2,
         2,
         2233,
         2234,
         2235,
         2236,
         2,
         3,
         2,
         2,
         355,
         2,
         2,
         373,
         356,
         357,
         358,
         359,
         2,
         2,
         2233,
         2234,
         2235,
         2236,
         356,
         365,
         356,
         356,
         369,
         356,
         356,
         356,
         356,
         374,
         375,
         376,
         356,
         356,
         2247,
         2248,
         2249,
         2250,
         357,
         366,
         357,
         357,
         370,
         357,
         357,
         357,
         374,
         357,
         377,
         378,
         357,
         357,
         2251,
         2252,
         2253,
         2254,
         358,
         367,
         358,
         358,
         371,
         358,
         358,
         358,
         375,
         377,
         358,
         379,
         358,
         358,
         2255,
         2256,
         2257,
         2258,
         359,
         368,
         359,
         359,
         372,
         359,
         359,
         359,
         376,
         378,
         379,
         359,
         359,
         359,
         2259,
         2260,
         2261,
         2262,
         2,
         2237,
         2,
         2,
         355,
         2,
         2,
         2,
         356,
         357,
         358,
         359,
         2,
         2,
         2233,
         2234,
         2235,
         2236,
         2,
         2238,
         2,
         2,
         355,
         2,
         2,
         2,
         356,
         357,
         358,
         359,
         2,
         2,
         2233,
         2234,
         2235,
         2236,
         2233,
         2239,
         2233,
         2233,
         2243,
         2233,
         2233,
         2233,
         2247,
         2251,
         2255,
         2259,
         2233,
         2233,
         2233,
         2263,
         2264,
         2265,
         2234,
         2240,
         2234,
         2234,
         2244,
         2234,
         2234,
         2234,
         2248,
         2252,
         2256,
         2260,
         2234,
         2234,
         2263,
         2234,
         2266,
         2267,
         2235,
         2241,
         2235,
         2235,
         2245,
         2235,
         2235,
         2235,
         2249,
         2253,
         2257,
         2261,
         2235,
         2235,
         2264,
         2266,
         2235,
         2268,
         2236,
         2242,
         2236,
         2236,
         2246,
         2236,
         2236,
         2236,
         2250,
         2254,
         2258,
         2262,
         2236,
         2236,
         2265,
         2267,
         2268,
         2236,
      },
   },
   [nir_op_ishl] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         2,
         0,
         3,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         6,
         0,
         7,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         4,
         14,
         4,
         15,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         0,
         3,
         0,
         0,
         0,
         3,
         0,
         0,
         3,
         3,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         6,
         6,
         0,
         0,
         7,
         0,
         7,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         17,
         0,
         17,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         4,
         15,
         14,
         14,
         14,
         14,
         14,
         4,
         4,
         4,
         4,
         0,
         12,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         18,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         19,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         0,
         0,
         0,
         3,
         3,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         0,
         0,
         20,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         6,
         21,
         0,
         5,
         5,
         3,
         6,
         22,
         16,
         3,
         6,
         23,
         5,
         5,
         3,
         6,
         24,
         16,
         3,
         6,
         25,
         5,
         5,
         3,
         6,
         0,
         16,
      },
      
      .num_filtered_states = 26,
      .table = (const uint16_t []) {
      
         4,
         5,
         4,
         4,
         4,
         1023,
         4,
         4,
         4,
         1024,
         4,
         4,
         4,
         4,
         2269,
         4,
         4,
         4,
         1023,
         4,
         4,
         1023,
         4,
         1023,
         4,
         1023,
         6,
         7,
         6,
         6,
         6,
         1025,
         6,
         6,
         6,
         1026,
         6,
         6,
         6,
         6,
         2270,
         6,
         6,
         6,
         1025,
         6,
         6,
         1025,
         6,
         1025,
         6,
         1025,
         4,
         1027,
         4,
         4,
         4,
         1023,
         4,
         4,
         4,
         1024,
         4,
         4,
         4,
         4,
         2269,
         4,
         4,
         4,
         1023,
         4,
         4,
         1023,
         4,
         1023,
         4,
         1023,
         4,
         1028,
         4,
         4,
         4,
         1023,
         4,
         4,
         4,
         1024,
         4,
         4,
         4,
         4,
         2269,
         4,
         4,
         4,
         1023,
         4,
         4,
         1023,
         4,
         1023,
         4,
         1023,
         4,
         1029,
         4,
         4,
         4,
         1023,
         4,
         4,
         4,
         1024,
         4,
         4,
         4,
         4,
         2269,
         4,
         4,
         4,
         1023,
         4,
         4,
         1023,
         4,
         1023,
         4,
         1023,
         4,
         1030,
         4,
         4,
         4,
         1023,
         4,
         4,
         4,
         1024,
         4,
         4,
         4,
         4,
         2269,
         4,
         4,
         4,
         1023,
         4,
         4,
         1023,
         4,
         1023,
         4,
         1023,
         1031,
         1032,
         1031,
         1031,
         1031,
         1033,
         1031,
         1031,
         1031,
         1034,
         1031,
         1031,
         1031,
         1031,
         2271,
         1031,
         1031,
         1031,
         1033,
         1031,
         1031,
         1033,
         1031,
         1033,
         1031,
         1033,
         1035,
         1036,
         1035,
         1035,
         1035,
         1037,
         1035,
         1035,
         1035,
         1038,
         1035,
         1035,
         1035,
         1035,
         2272,
         1035,
         1035,
         1035,
         1037,
         1035,
         1035,
         1037,
         1035,
         1037,
         1035,
         1037,
         4,
         1039,
         4,
         4,
         4,
         1023,
         4,
         4,
         4,
         1024,
         4,
         4,
         4,
         4,
         2269,
         4,
         4,
         4,
         1023,
         4,
         4,
         1023,
         4,
         1023,
         4,
         1023,
         4,
         5,
         4,
         4,
         4,
         1023,
         4,
         4,
         4,
         1024,
         4,
         4,
         4,
         4,
         2269,
         4,
         4,
         4,
         1023,
         4,
         4,
         1023,
         4,
         1023,
         4,
         1023,
         4,
         1040,
         4,
         4,
         4,
         1023,
         4,
         4,
         4,
         1024,
         4,
         4,
         4,
         4,
         2269,
         4,
         4,
         4,
         1023,
         4,
         4,
         1023,
         4,
         1023,
         4,
         1023,
         1041,
         1042,
         1041,
         1041,
         1041,
         1043,
         1041,
         1041,
         1041,
         1044,
         1041,
         1041,
         1041,
         1041,
         2273,
         1041,
         1041,
         1041,
         1043,
         1041,
         1041,
         1043,
         1041,
         1043,
         1041,
         1043,
         4,
         2274,
         4,
         4,
         4,
         1023,
         4,
         4,
         4,
         1024,
         4,
         4,
         4,
         4,
         2269,
         4,
         4,
         4,
         1023,
         4,
         4,
         1023,
         4,
         1023,
         4,
         1023,
         4,
         2275,
         4,
         4,
         4,
         1023,
         4,
         4,
         4,
         1024,
         4,
         4,
         4,
         4,
         2269,
         4,
         4,
         4,
         1023,
         4,
         4,
         1023,
         4,
         1023,
         4,
         1023,
         4,
         1029,
         4,
         4,
         4,
         1023,
         4,
         4,
         4,
         1024,
         4,
         4,
         4,
         4,
         2269,
         4,
         4,
         4,
         1023,
         4,
         4,
         1023,
         4,
         1023,
         4,
         1023,
         4,
         2276,
         4,
         4,
         4,
         1023,
         4,
         4,
         4,
         1024,
         4,
         4,
         4,
         4,
         2269,
         4,
         4,
         4,
         1023,
         4,
         4,
         1023,
         4,
         1023,
         4,
         1023,
         4,
         2277,
         4,
         4,
         4,
         1023,
         4,
         4,
         4,
         1024,
         4,
         4,
         4,
         4,
         2269,
         4,
         4,
         4,
         1023,
         4,
         4,
         1023,
         4,
         1023,
         4,
         1023,
         4,
         2278,
         4,
         4,
         4,
         1023,
         4,
         4,
         4,
         1024,
         4,
         4,
         4,
         4,
         2269,
         4,
         4,
         4,
         1023,
         4,
         4,
         1023,
         4,
         1023,
         4,
         1023,
         4,
         2812,
         4,
         4,
         4,
         1023,
         4,
         4,
         4,
         1024,
         4,
         4,
         4,
         4,
         2269,
         4,
         4,
         4,
         1023,
         4,
         4,
         1023,
         4,
         1023,
         4,
         1023,
         4,
         2813,
         4,
         4,
         4,
         1023,
         4,
         4,
         4,
         1024,
         4,
         4,
         4,
         4,
         2269,
         4,
         4,
         4,
         1023,
         4,
         4,
         1023,
         4,
         1023,
         4,
         1023,
         4,
         2909,
         4,
         4,
         4,
         1023,
         4,
         4,
         4,
         1024,
         4,
         4,
         4,
         4,
         2269,
         4,
         4,
         4,
         1023,
         4,
         4,
         1023,
         4,
         1023,
         4,
         1023,
         4,
         2915,
         4,
         4,
         4,
         1023,
         4,
         4,
         4,
         1024,
         4,
         4,
         4,
         4,
         2269,
         4,
         4,
         4,
         1023,
         4,
         4,
         1023,
         4,
         1023,
         4,
         1023,
         4,
         2919,
         4,
         4,
         4,
         1023,
         4,
         4,
         4,
         1024,
         4,
         4,
         4,
         4,
         2269,
         4,
         4,
         4,
         1023,
         4,
         4,
         1023,
         4,
         1023,
         4,
         1023,
         4,
         2924,
         4,
         4,
         4,
         1023,
         4,
         4,
         4,
         1024,
         4,
         4,
         4,
         4,
         2269,
         4,
         4,
         4,
         1023,
         4,
         4,
         1023,
         4,
         1023,
         4,
         1023,
         4,
         2928,
         4,
         4,
         4,
         1023,
         4,
         4,
         4,
         1024,
         4,
         4,
         4,
         4,
         2269,
         4,
         4,
         4,
         1023,
         4,
         4,
         1023,
         4,
         1023,
         4,
         1023,
         4,
         2933,
         4,
         4,
         4,
         1023,
         4,
         4,
         4,
         1024,
         4,
         4,
         4,
         4,
         2269,
         4,
         4,
         4,
         1023,
         4,
         4,
         1023,
         4,
         1023,
         4,
         1023,
      },
   },
   [nir_op_unpack_64_2x32_split_x] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 8,
      .table = (const uint16_t []) {
      
         8,
         1271,
         1272,
         1273,
         1274,
         1275,
         1276,
         1277,
      },
   },
   [nir_op_imul_2x32_64] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         9,
      },
   },
   [nir_op_umul_2x32_64] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         10,
      },
   },
   [nir_op_udiv] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         11,
         0,
         11,
      },
   },
   [nir_op_idiv] = {
      .filter = (const uint16_t []) {
         0,
         1,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         0,
         12,
         0,
         0,
         12,
         0,
         380,
         381,
         380,
      },
   },
   [nir_op_umod] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         13,
         0,
         13,
      },
   },
   [nir_op_imod] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         14,
         0,
         14,
      },
   },
   [nir_op_irem] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         15,
         0,
         15,
      },
   },
   [nir_op_fneg] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         2,
         0,
         0,
         0,
         3,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         6,
         6,
         0,
         7,
         0,
         7,
         0,
         7,
         0,
         8,
         0,
         8,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         0,
         13,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         3,
         3,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         3,
         3,
         3,
         3,
         3,
         14,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         15,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         8,
         0,
         0,
         9,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         10,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         17,
         5,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         9,
         9,
         9,
         9,
         9,
         9,
         3,
         3,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         5,
         5,
         5,
         5,
         5,
         5,
         15,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         17,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         5,
         4,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 18,
      .table = (const uint16_t []) {
      
         16,
         1278,
         1279,
         1280,
         1281,
         1282,
         1283,
         1284,
         1285,
         1286,
         1287,
         1288,
         1289,
         1290,
         2614,
         2615,
         2616,
         2805,
      },
   },
   [nir_op_ineg] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         8,
         9,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         11,
         5,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 12,
      .table = (const uint16_t []) {
      
         17,
         1555,
         1556,
         1557,
         1558,
         1559,
         1560,
         1561,
         1562,
         1563,
         1564,
         2727,
      },
   },
   [nir_op_fabs] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         5,
         6,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         9,
         0,
         0,
         0,
         10,
         10,
         11,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         9,
         0,
         0,
         0,
         10,
         10,
         11,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         10,
         10,
         11,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         10,
         10,
         11,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         11,
         10,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         10,
         10,
         11,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         10,
         10,
         11,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         10,
         10,
         11,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         10,
         10,
         11,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         10,
         10,
         11,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         7,
         7,
         7,
         7,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 12,
      .table = (const uint16_t []) {
      
         18,
         1291,
         1292,
         1293,
         1294,
         1295,
         1296,
         1297,
         1298,
         2287,
         2288,
         2289,
      },
   },
   [nir_search_op_u2f] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         19,
         1922,
         1923,
         1924,
      },
   },
   [nir_op_iabs] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 5,
      .table = (const uint16_t []) {
      
         20,
         1565,
         1566,
         1567,
         2534,
      },
   },
   [nir_search_op_f2b] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         21,
         1299,
         1300,
      },
   },
   [nir_op_fadd] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         6,
         6,
         7,
         8,
         7,
         8,
         7,
         8,
         9,
         10,
         9,
         10,
         9,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         0,
         0,
         12,
         12,
         0,
         0,
         0,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         14,
         2,
         2,
         15,
         16,
         17,
         18,
         19,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         20,
         3,
         21,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         20,
         3,
         3,
         3,
         3,
         3,
         5,
         22,
         5,
         5,
         23,
         24,
         5,
         5,
         5,
         5,
         5,
         22,
         5,
         5,
         5,
         23,
         24,
         22,
         5,
         5,
         23,
         24,
         5,
         22,
         22,
         22,
         25,
         26,
         5,
         23,
         24,
         23,
         24,
         5,
         23,
         27,
         5,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         7,
         7,
         7,
         7,
         7,
         8,
         7,
         7,
         7,
         7,
         7,
         7,
         8,
         7,
         7,
         7,
         8,
         7,
         7,
         7,
         8,
         7,
         7,
         7,
         8,
         7,
         7,
         7,
         8,
         7,
         7,
         7,
         7,
         8,
         7,
         7,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         10,
         9,
         9,
         9,
         10,
         9,
         9,
         9,
         10,
         9,
         9,
         28,
         11,
         11,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         29,
         29,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         30,
         31,
         5,
         5,
         30,
         31,
         5,
         5,
         5,
         30,
         31,
         5,
         22,
         32,
         33,
         34,
         22,
         5,
         30,
         31,
         5,
         30,
         31,
         5,
         23,
         35,
         36,
         23,
         24,
         37,
         38,
         24,
         5,
         5,
         30,
         31,
         5,
         30,
         31,
         31,
         7,
         7,
         8,
         7,
         7,
         7,
         7,
         8,
         7,
         7,
         7,
         7,
         8,
         7,
         7,
         7,
         7,
         7,
         8,
         7,
         7,
         7,
         7,
         7,
         8,
         7,
         7,
         7,
         8,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         39,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         28,
         28,
         28,
         11,
         11,
         11,
         40,
         3,
         21,
         21,
         21,
         21,
         21,
         41,
         4,
         4,
         20,
         40,
         3,
         20,
         40,
         3,
         0,
         2,
         42,
         43,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         12,
         0,
         12,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         4,
         4,
         4,
         4,
         4,
         4,
         21,
         20,
         20,
         20,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         40,
         40,
         40,
         3,
         3,
         3,
         3,
         20,
         3,
         3,
         44,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         45,
         46,
         45,
         46,
         45,
         46,
         45,
         47,
         45,
         46,
         45,
         46,
         48,
         49,
         50,
         51,
         45,
         46,
         52,
         53,
         54,
         46,
         45,
         46,
         55,
         7,
         7,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         21,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 56,
      .table = (const uint16_t []) {
      
         22,
         23,
         1301,
         1302,
         1302,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         1301,
         1301,
         1301,
         1301,
         1301,
         2597,
         2598,
         2598,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         2774,
         2774,
         22,
         2774,
         2774,
         2775,
         2775,
         2774,
         2774,
         22,
         2776,
         2776,
         1301,
         1301,
         1301,
         22,
         2774,
         2774,
         22,
         2775,
         22,
         2774,
         22,
         2774,
         2774,
         2774,
         23,
         23,
         1303,
         1304,
         1305,
         23,
         23,
         1306,
         1307,
         1308,
         1309,
         23,
         1310,
         23,
         2599,
         2600,
         2601,
         2602,
         2603,
         2604,
         2605,
         2606,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         2777,
         2777,
         2778,
         2777,
         2777,
         2779,
         2779,
         2777,
         2777,
         2780,
         2781,
         2782,
         2783,
         2603,
         2896,
         23,
         2777,
         2777,
         23,
         2779,
         23,
         2777,
         2778,
         2777,
         2777,
         2777,
         1301,
         1303,
         1301,
         1311,
         1311,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         2597,
         2607,
         2607,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         2784,
         2784,
         1301,
         2784,
         2784,
         2785,
         2785,
         2784,
         2784,
         1301,
         2786,
         2786,
         1301,
         1301,
         1301,
         1301,
         2784,
         2784,
         1301,
         2785,
         1301,
         2784,
         1301,
         2784,
         2784,
         2784,
         1302,
         1304,
         1311,
         1302,
         1302,
         1302,
         1302,
         1302,
         1302,
         1302,
         1302,
         1302,
         1302,
         1302,
         1311,
         1311,
         1311,
         1311,
         1311,
         2608,
         2598,
         2598,
         1302,
         1302,
         1302,
         1302,
         1302,
         1302,
         1302,
         1302,
         2787,
         2787,
         1302,
         2787,
         2787,
         2788,
         2788,
         2787,
         2787,
         1302,
         2776,
         2776,
         1311,
         1311,
         1311,
         1302,
         2787,
         2787,
         1302,
         2788,
         1302,
         2787,
         1302,
         2787,
         2787,
         2787,
         1302,
         1305,
         1311,
         1302,
         1302,
         1302,
         1302,
         1302,
         1302,
         1302,
         1302,
         1302,
         1302,
         1302,
         1311,
         1311,
         1311,
         1311,
         1311,
         2608,
         2598,
         2598,
         1302,
         1302,
         1302,
         1302,
         1302,
         1302,
         1302,
         1302,
         2787,
         2787,
         1302,
         2787,
         2787,
         2788,
         2788,
         2787,
         2787,
         1302,
         2776,
         2776,
         1311,
         1311,
         1311,
         1302,
         2787,
         2787,
         1302,
         2788,
         1302,
         2787,
         1302,
         2787,
         2787,
         2787,
         22,
         23,
         1301,
         1302,
         1302,
         1312,
         22,
         1313,
         1313,
         22,
         22,
         22,
         22,
         22,
         1301,
         1301,
         1301,
         1301,
         1301,
         2597,
         2598,
         2598,
         1312,
         1312,
         1312,
         1312,
         1312,
         1312,
         22,
         22,
         2789,
         2790,
         1312,
         2789,
         2790,
         2791,
         2792,
         2789,
         2790,
         22,
         2776,
         2776,
         1301,
         1301,
         1301,
         1312,
         2790,
         2790,
         1312,
         2792,
         1312,
         2790,
         1312,
         2789,
         2790,
         2790,
         22,
         23,
         1301,
         1302,
         1302,
         22,
         1314,
         22,
         22,
         1315,
         1315,
         22,
         22,
         22,
         1301,
         1301,
         1301,
         1301,
         1301,
         2597,
         2598,
         2598,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         2774,
         2774,
         22,
         2774,
         2774,
         2775,
         2775,
         2774,
         2774,
         22,
         2776,
         2776,
         1301,
         1301,
         1301,
         22,
         2774,
         2774,
         22,
         2775,
         22,
         2774,
         22,
         2774,
         2774,
         2774,
         22,
         1306,
         1301,
         1302,
         1302,
         1313,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         1301,
         1301,
         1301,
         1301,
         1301,
         2597,
         2598,
         2598,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         22,
         22,
         2793,
         2793,
         1313,
         2793,
         2793,
         2794,
         2794,
         2793,
         2793,
         22,
         2776,
         2776,
         1301,
         1301,
         1301,
         1313,
         2793,
         2793,
         1313,
         2794,
         1313,
         2793,
         1313,
         2793,
         2793,
         2793,
         22,
         1307,
         1301,
         1302,
         1302,
         1313,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         1301,
         1301,
         1301,
         1301,
         1301,
         2597,
         2598,
         2598,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         22,
         22,
         2793,
         2793,
         1313,
         2793,
         2793,
         2794,
         2794,
         2793,
         2793,
         22,
         2776,
         2776,
         1301,
         1301,
         1301,
         1313,
         2793,
         2793,
         1313,
         2794,
         1313,
         2793,
         1313,
         2793,
         2793,
         2793,
         22,
         1308,
         1301,
         1302,
         1302,
         22,
         1315,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         1301,
         1301,
         1301,
         1301,
         1301,
         2597,
         2598,
         2598,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         2774,
         2774,
         22,
         2774,
         2774,
         2775,
         2775,
         2774,
         2774,
         22,
         2776,
         2776,
         1301,
         1301,
         1301,
         22,
         2774,
         2774,
         22,
         2775,
         22,
         2774,
         22,
         2774,
         2774,
         2774,
         22,
         1309,
         1301,
         1302,
         1302,
         22,
         1315,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         1301,
         1301,
         1301,
         1301,
         1301,
         2597,
         2598,
         2598,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         2774,
         2774,
         22,
         2774,
         2774,
         2775,
         2775,
         2774,
         2774,
         22,
         2776,
         2776,
         1301,
         1301,
         1301,
         22,
         2774,
         2774,
         22,
         2775,
         22,
         2774,
         22,
         2774,
         2774,
         2774,
         22,
         23,
         1301,
         1302,
         1302,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         1301,
         1301,
         1301,
         1301,
         1301,
         2597,
         2598,
         2598,
         22,
         22,
         22,
         22,
         22,
         22,
         2609,
         22,
         2774,
         2774,
         22,
         2774,
         2774,
         2775,
         2775,
         2774,
         2774,
         22,
         2776,
         2776,
         1301,
         1301,
         1301,
         22,
         2774,
         2774,
         22,
         2775,
         22,
         2774,
         22,
         2774,
         2774,
         2774,
         22,
         1310,
         1301,
         1302,
         1302,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         1301,
         1301,
         1301,
         1301,
         1301,
         2597,
         2598,
         2598,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         2774,
         2774,
         22,
         2774,
         2774,
         2775,
         2775,
         2774,
         2774,
         22,
         2776,
         2776,
         1301,
         1301,
         1301,
         22,
         2774,
         2774,
         22,
         2775,
         22,
         2774,
         22,
         2774,
         2774,
         2774,
         22,
         23,
         1301,
         1302,
         1302,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         1316,
         1301,
         1301,
         1301,
         1301,
         2610,
         2597,
         2598,
         2598,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         1316,
         2774,
         2774,
         22,
         2774,
         2774,
         2775,
         2775,
         2774,
         2774,
         22,
         2776,
         2776,
         1301,
         2610,
         1301,
         22,
         2774,
         2774,
         22,
         2775,
         22,
         2774,
         22,
         2774,
         2774,
         2774,
         1301,
         2599,
         1301,
         1311,
         1311,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         2597,
         2607,
         2607,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         2784,
         2784,
         1301,
         2784,
         2784,
         2785,
         2785,
         2784,
         2784,
         1301,
         2786,
         2786,
         1301,
         1301,
         1301,
         1301,
         2784,
         2784,
         1301,
         2785,
         1301,
         2784,
         1301,
         2784,
         2784,
         2784,
         1301,
         2600,
         1301,
         1311,
         1311,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         2597,
         2607,
         2607,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         2784,
         2784,
         1301,
         2784,
         2784,
         2785,
         2785,
         2784,
         2784,
         1301,
         2786,
         2786,
         1301,
         1301,
         1301,
         1301,
         2784,
         2784,
         1301,
         2785,
         1301,
         2784,
         1301,
         2784,
         2784,
         2784,
         1301,
         2601,
         1301,
         1311,
         1311,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         2597,
         2607,
         2607,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         2784,
         2784,
         1301,
         2784,
         2784,
         2785,
         2785,
         2784,
         2784,
         1301,
         2786,
         2786,
         1301,
         1301,
         1301,
         1301,
         2784,
         2784,
         1301,
         2785,
         1301,
         2784,
         1301,
         2784,
         2784,
         2784,
         1301,
         2602,
         1301,
         1311,
         1311,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         2597,
         2607,
         2607,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         2784,
         2784,
         1301,
         2784,
         2784,
         2785,
         2785,
         2784,
         2784,
         1301,
         2786,
         2786,
         1301,
         1301,
         1301,
         1301,
         2784,
         2784,
         1301,
         2785,
         1301,
         2784,
         1301,
         2784,
         2784,
         2784,
         1301,
         2603,
         1301,
         1311,
         1311,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         2610,
         1301,
         1301,
         1301,
         1301,
         1301,
         2597,
         2607,
         2607,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         2610,
         2784,
         2784,
         1301,
         2784,
         2784,
         2785,
         2785,
         2784,
         2784,
         1301,
         2786,
         2786,
         1301,
         1301,
         1301,
         1301,
         2784,
         2784,
         1301,
         2785,
         1301,
         2784,
         1301,
         2784,
         2784,
         2784,
         2597,
         2604,
         2597,
         2608,
         2608,
         2597,
         2597,
         2597,
         2597,
         2597,
         2597,
         2597,
         2597,
         2597,
         2597,
         2597,
         2597,
         2597,
         2597,
         2597,
         2611,
         2611,
         2597,
         2597,
         2597,
         2597,
         2597,
         2597,
         2597,
         2597,
         2795,
         2795,
         2597,
         2795,
         2795,
         2796,
         2796,
         2795,
         2795,
         2597,
         2797,
         2797,
         2597,
         2597,
         2597,
         2597,
         2795,
         2795,
         2597,
         2796,
         2597,
         2795,
         2597,
         2795,
         2795,
         2795,
         2598,
         2605,
         2607,
         2598,
         2598,
         2598,
         2598,
         2598,
         2598,
         2598,
         2598,
         2598,
         2598,
         2598,
         2607,
         2607,
         2607,
         2607,
         2607,
         2611,
         2598,
         2598,
         2598,
         2598,
         2598,
         2598,
         2598,
         2598,
         2598,
         2598,
         2798,
         2798,
         2598,
         2798,
         2798,
         2799,
         2799,
         2798,
         2798,
         2598,
         2776,
         2776,
         2607,
         2607,
         2607,
         2598,
         2798,
         2798,
         2598,
         2799,
         2598,
         2798,
         2598,
         2798,
         2798,
         2798,
         2598,
         2606,
         2607,
         2598,
         2598,
         2598,
         2598,
         2598,
         2598,
         2598,
         2598,
         2598,
         2598,
         2598,
         2607,
         2607,
         2607,
         2607,
         2607,
         2611,
         2598,
         2598,
         2598,
         2598,
         2598,
         2598,
         2598,
         2598,
         2598,
         2598,
         2798,
         2798,
         2598,
         2798,
         2798,
         2799,
         2799,
         2798,
         2798,
         2598,
         2776,
         2776,
         2607,
         2607,
         2607,
         2598,
         2798,
         2798,
         2598,
         2799,
         2598,
         2798,
         2598,
         2798,
         2798,
         2798,
         22,
         23,
         1301,
         1302,
         1302,
         1312,
         22,
         1313,
         1313,
         22,
         22,
         22,
         22,
         22,
         1301,
         1301,
         1301,
         1301,
         1301,
         2597,
         2598,
         2598,
         1312,
         1312,
         1312,
         1312,
         1312,
         1312,
         22,
         22,
         2789,
         2790,
         1312,
         2789,
         2790,
         2791,
         2792,
         2789,
         2790,
         22,
         2776,
         2776,
         1301,
         1301,
         1301,
         2897,
         2790,
         2790,
         2897,
         2792,
         2897,
         2790,
         2897,
         2898,
         2899,
         2899,
         22,
         23,
         1301,
         1302,
         1302,
         1312,
         22,
         1313,
         1313,
         22,
         22,
         22,
         22,
         22,
         1301,
         1301,
         1301,
         1301,
         1301,
         2597,
         2598,
         2598,
         1312,
         1312,
         1312,
         1312,
         1312,
         1312,
         22,
         22,
         2789,
         2790,
         1312,
         2789,
         2790,
         2791,
         2792,
         2789,
         2790,
         22,
         2776,
         2776,
         1301,
         1301,
         1301,
         1312,
         2900,
         2900,
         1312,
         2901,
         1312,
         2900,
         1312,
         2789,
         2790,
         2900,
         22,
         23,
         1301,
         1302,
         1302,
         1312,
         22,
         1313,
         1313,
         22,
         22,
         22,
         22,
         22,
         1301,
         1301,
         1301,
         1301,
         1301,
         2597,
         2598,
         2598,
         1312,
         1312,
         2612,
         1312,
         2612,
         2612,
         22,
         22,
         2789,
         2790,
         1312,
         2789,
         2790,
         2791,
         2792,
         2800,
         2801,
         22,
         2776,
         2776,
         1301,
         1301,
         1301,
         1312,
         2790,
         2790,
         1312,
         2792,
         2612,
         2801,
         1312,
         2789,
         2790,
         2790,
         22,
         23,
         1301,
         1302,
         1302,
         1312,
         22,
         1313,
         1313,
         22,
         22,
         22,
         22,
         22,
         1301,
         1301,
         1301,
         1301,
         1301,
         2597,
         2598,
         2598,
         1312,
         1312,
         1312,
         1312,
         1312,
         1312,
         22,
         22,
         2789,
         2790,
         1312,
         2789,
         2790,
         2791,
         2792,
         2789,
         2790,
         22,
         2776,
         2776,
         1301,
         1301,
         1301,
         2897,
         2900,
         2900,
         2897,
         2901,
         2897,
         2900,
         2897,
         2898,
         2899,
         2902,
         22,
         23,
         1301,
         1302,
         1302,
         1312,
         22,
         1313,
         1313,
         22,
         22,
         22,
         22,
         22,
         1301,
         1301,
         1301,
         1301,
         1301,
         2597,
         2598,
         2598,
         1312,
         1312,
         2612,
         1312,
         2612,
         2612,
         22,
         22,
         2789,
         2790,
         1312,
         2789,
         2790,
         2791,
         2792,
         2800,
         2801,
         22,
         2776,
         2776,
         1301,
         1301,
         1301,
         2897,
         2790,
         2790,
         2897,
         2792,
         2903,
         2801,
         2897,
         2898,
         2899,
         2899,
         22,
         23,
         1301,
         1302,
         1302,
         1312,
         22,
         1313,
         1313,
         22,
         22,
         22,
         22,
         22,
         1301,
         1301,
         1301,
         1301,
         1301,
         2597,
         2598,
         2598,
         1312,
         1312,
         2612,
         1312,
         2612,
         2612,
         22,
         22,
         2789,
         2790,
         1312,
         2789,
         2790,
         2791,
         2792,
         2800,
         2801,
         22,
         2776,
         2776,
         1301,
         1301,
         1301,
         1312,
         2900,
         2900,
         1312,
         2901,
         2612,
         2904,
         1312,
         2789,
         2790,
         2900,
         22,
         23,
         1301,
         1302,
         1302,
         22,
         22,
         22,
         22,
         22,
         22,
         2609,
         22,
         22,
         1301,
         1301,
         1301,
         1301,
         1301,
         2597,
         2598,
         2598,
         22,
         22,
         22,
         22,
         22,
         22,
         2609,
         22,
         2774,
         2774,
         22,
         2774,
         2774,
         2775,
         2775,
         2774,
         2774,
         22,
         2776,
         2776,
         1301,
         1301,
         1301,
         22,
         2774,
         2774,
         22,
         2775,
         22,
         2774,
         22,
         2774,
         2774,
         2774,
         22,
         23,
         1301,
         1302,
         1302,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         1316,
         1301,
         1301,
         1301,
         1301,
         2610,
         2597,
         2598,
         2598,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         1316,
         2774,
         2774,
         22,
         2774,
         2774,
         2775,
         2775,
         2774,
         2774,
         22,
         2776,
         2776,
         1301,
         2802,
         1301,
         22,
         2774,
         2774,
         22,
         2775,
         22,
         2774,
         22,
         2774,
         2774,
         2774,
         2774,
         2777,
         2784,
         2787,
         2787,
         2789,
         2774,
         2793,
         2793,
         2774,
         2774,
         2774,
         2774,
         2774,
         2784,
         2784,
         2784,
         2784,
         2784,
         2795,
         2798,
         2798,
         2789,
         2789,
         2789,
         2789,
         2789,
         2789,
         2774,
         2774,
         2789,
         2790,
         2789,
         2789,
         2790,
         2791,
         2792,
         2789,
         2790,
         2774,
         2803,
         2803,
         2784,
         2784,
         2784,
         2789,
         2790,
         2790,
         2789,
         2792,
         2789,
         2790,
         2789,
         2789,
         2790,
         2790,
         2774,
         2777,
         2784,
         2787,
         2787,
         2790,
         2774,
         2793,
         2793,
         2774,
         2774,
         2774,
         2774,
         2774,
         2784,
         2784,
         2784,
         2784,
         2784,
         2795,
         2798,
         2798,
         2790,
         2790,
         2790,
         2790,
         2790,
         2790,
         2774,
         2774,
         2790,
         2790,
         2790,
         2790,
         2790,
         2792,
         2792,
         2790,
         2790,
         2774,
         2803,
         2803,
         2784,
         2784,
         2784,
         2790,
         2790,
         2790,
         2790,
         2792,
         2790,
         2790,
         2790,
         2790,
         2790,
         2790,
         22,
         2778,
         1301,
         1302,
         1302,
         1312,
         22,
         1313,
         1313,
         22,
         22,
         22,
         22,
         22,
         1301,
         1301,
         1301,
         1301,
         1301,
         2597,
         2598,
         2598,
         1312,
         1312,
         1312,
         1312,
         1312,
         1312,
         22,
         22,
         2789,
         2790,
         1312,
         2789,
         2790,
         2791,
         2792,
         2789,
         2790,
         22,
         2776,
         2776,
         1301,
         1301,
         1301,
         2897,
         2790,
         2790,
         2897,
         2792,
         2897,
         2790,
         2897,
         2898,
         2899,
         2899,
         2774,
         2777,
         2784,
         2787,
         2787,
         2789,
         2774,
         2793,
         2793,
         2774,
         2774,
         2774,
         2774,
         2774,
         2784,
         2784,
         2784,
         2784,
         2784,
         2795,
         2798,
         2798,
         2789,
         2789,
         2789,
         2789,
         2789,
         2789,
         2774,
         2774,
         2789,
         2790,
         2789,
         2789,
         2790,
         2791,
         2792,
         2789,
         2790,
         2774,
         2803,
         2803,
         2784,
         2784,
         2784,
         2898,
         2790,
         2790,
         2898,
         2792,
         2898,
         2790,
         2898,
         2898,
         2899,
         2899,
         2774,
         2777,
         2784,
         2787,
         2787,
         2790,
         2774,
         2793,
         2793,
         2774,
         2774,
         2774,
         2774,
         2774,
         2784,
         2784,
         2784,
         2784,
         2784,
         2795,
         2798,
         2798,
         2790,
         2790,
         2790,
         2790,
         2790,
         2790,
         2774,
         2774,
         2790,
         2790,
         2790,
         2790,
         2790,
         2792,
         2792,
         2790,
         2790,
         2774,
         2803,
         2803,
         2784,
         2784,
         2784,
         2899,
         2790,
         2790,
         2899,
         2792,
         2899,
         2790,
         2899,
         2899,
         2899,
         2899,
         2775,
         2779,
         2785,
         2788,
         2788,
         2791,
         2775,
         2794,
         2794,
         2775,
         2775,
         2775,
         2775,
         2775,
         2785,
         2785,
         2785,
         2785,
         2785,
         2796,
         2799,
         2799,
         2791,
         2791,
         2791,
         2791,
         2791,
         2791,
         2775,
         2775,
         2791,
         2792,
         2791,
         2791,
         2792,
         2791,
         2792,
         2791,
         2792,
         2775,
         2804,
         2804,
         2785,
         2785,
         2785,
         2791,
         2901,
         2901,
         2791,
         2901,
         2791,
         2901,
         2791,
         2791,
         2792,
         2901,
         2775,
         2779,
         2785,
         2788,
         2788,
         2792,
         2775,
         2794,
         2794,
         2775,
         2775,
         2775,
         2775,
         2775,
         2785,
         2785,
         2785,
         2785,
         2785,
         2796,
         2799,
         2799,
         2792,
         2792,
         2792,
         2792,
         2792,
         2792,
         2775,
         2775,
         2792,
         2792,
         2792,
         2792,
         2792,
         2792,
         2792,
         2792,
         2792,
         2775,
         2804,
         2804,
         2785,
         2785,
         2785,
         2792,
         2901,
         2901,
         2792,
         2901,
         2792,
         2901,
         2792,
         2792,
         2792,
         2901,
         2774,
         2777,
         2784,
         2787,
         2787,
         2789,
         2774,
         2793,
         2793,
         2774,
         2774,
         2774,
         2774,
         2774,
         2784,
         2784,
         2784,
         2784,
         2784,
         2795,
         2798,
         2798,
         2789,
         2789,
         2800,
         2789,
         2800,
         2800,
         2774,
         2774,
         2789,
         2790,
         2789,
         2789,
         2790,
         2791,
         2792,
         2800,
         2801,
         2774,
         2803,
         2803,
         2784,
         2784,
         2784,
         2789,
         2790,
         2790,
         2789,
         2792,
         2800,
         2801,
         2789,
         2789,
         2790,
         2790,
         2774,
         2777,
         2784,
         2787,
         2787,
         2790,
         2774,
         2793,
         2793,
         2774,
         2774,
         2774,
         2774,
         2774,
         2784,
         2784,
         2784,
         2784,
         2784,
         2795,
         2798,
         2798,
         2790,
         2790,
         2801,
         2790,
         2801,
         2801,
         2774,
         2774,
         2790,
         2790,
         2790,
         2790,
         2790,
         2792,
         2792,
         2801,
         2801,
         2774,
         2803,
         2803,
         2784,
         2784,
         2784,
         2790,
         2790,
         2790,
         2790,
         2792,
         2801,
         2801,
         2790,
         2790,
         2790,
         2790,
         22,
         2780,
         1301,
         1302,
         1302,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         1301,
         1301,
         1301,
         1301,
         1301,
         2597,
         2598,
         2598,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         2774,
         2774,
         22,
         2774,
         2774,
         2775,
         2775,
         2774,
         2774,
         22,
         2776,
         2776,
         1301,
         1301,
         1301,
         22,
         2774,
         2774,
         22,
         2775,
         22,
         2774,
         22,
         2774,
         2774,
         2774,
         2776,
         2781,
         2786,
         2776,
         2776,
         2776,
         2776,
         2776,
         2776,
         2776,
         2776,
         2776,
         2776,
         2776,
         2786,
         2786,
         2786,
         2786,
         2786,
         2797,
         2776,
         2776,
         2776,
         2776,
         2776,
         2776,
         2776,
         2776,
         2776,
         2776,
         2803,
         2803,
         2776,
         2803,
         2803,
         2804,
         2804,
         2803,
         2803,
         2776,
         2776,
         2776,
         2786,
         2786,
         2786,
         2776,
         2803,
         2803,
         2776,
         2804,
         2776,
         2803,
         2776,
         2803,
         2803,
         2803,
         2776,
         2782,
         2786,
         2776,
         2776,
         2776,
         2776,
         2776,
         2776,
         2776,
         2776,
         2776,
         2776,
         2776,
         2786,
         2786,
         2786,
         2786,
         2786,
         2797,
         2776,
         2776,
         2776,
         2776,
         2776,
         2776,
         2776,
         2776,
         2776,
         2776,
         2803,
         2803,
         2776,
         2803,
         2803,
         2804,
         2804,
         2803,
         2803,
         2776,
         2776,
         2776,
         2786,
         2786,
         2786,
         2776,
         2803,
         2803,
         2776,
         2804,
         2776,
         2803,
         2776,
         2803,
         2803,
         2803,
         1301,
         2783,
         1301,
         1311,
         1311,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         2597,
         2607,
         2607,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         2784,
         2784,
         1301,
         2784,
         2784,
         2785,
         2785,
         2784,
         2784,
         1301,
         2786,
         2786,
         1301,
         1301,
         1301,
         1301,
         2784,
         2784,
         1301,
         2785,
         1301,
         2784,
         1301,
         2784,
         2784,
         2784,
         1301,
         2603,
         1301,
         1311,
         1311,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         2610,
         1301,
         1301,
         1301,
         1301,
         1301,
         2597,
         2607,
         2607,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         2802,
         2784,
         2784,
         1301,
         2784,
         2784,
         2785,
         2785,
         2784,
         2784,
         1301,
         2786,
         2786,
         1301,
         1301,
         1301,
         1301,
         2784,
         2784,
         1301,
         2785,
         1301,
         2784,
         1301,
         2784,
         2784,
         2784,
         1301,
         2896,
         1301,
         1311,
         1311,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         2597,
         2607,
         2607,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         1301,
         2784,
         2784,
         1301,
         2784,
         2784,
         2785,
         2785,
         2784,
         2784,
         1301,
         2786,
         2786,
         1301,
         1301,
         1301,
         1301,
         2784,
         2784,
         1301,
         2785,
         1301,
         2784,
         1301,
         2784,
         2784,
         2784,
         22,
         23,
         1301,
         1302,
         1302,
         1312,
         22,
         1313,
         1313,
         22,
         22,
         22,
         22,
         22,
         1301,
         1301,
         1301,
         1301,
         1301,
         2597,
         2598,
         2598,
         2897,
         1312,
         1312,
         2897,
         2897,
         1312,
         22,
         22,
         2789,
         2790,
         2897,
         2898,
         2899,
         2791,
         2792,
         2789,
         2790,
         22,
         2776,
         2776,
         1301,
         1301,
         1301,
         2897,
         2790,
         2899,
         2897,
         2792,
         2897,
         2790,
         2897,
         2898,
         2899,
         2899,
         2774,
         2777,
         2784,
         2787,
         2787,
         2790,
         2774,
         2793,
         2793,
         2774,
         2774,
         2774,
         2774,
         2774,
         2784,
         2784,
         2784,
         2784,
         2784,
         2795,
         2798,
         2798,
         2790,
         2900,
         2790,
         2900,
         2790,
         2900,
         2774,
         2774,
         2790,
         2790,
         2790,
         2790,
         2790,
         2901,
         2901,
         2790,
         2790,
         2774,
         2803,
         2803,
         2784,
         2784,
         2784,
         2790,
         2790,
         2790,
         2900,
         2901,
         2790,
         2790,
         2790,
         2790,
         2790,
         2790,
         2774,
         2777,
         2784,
         2787,
         2787,
         2790,
         2774,
         2793,
         2793,
         2774,
         2774,
         2774,
         2774,
         2774,
         2784,
         2784,
         2784,
         2784,
         2784,
         2795,
         2798,
         2798,
         2790,
         2900,
         2790,
         2900,
         2790,
         2900,
         2774,
         2774,
         2790,
         2790,
         2790,
         2790,
         2790,
         2901,
         2901,
         2790,
         2790,
         2774,
         2803,
         2803,
         2784,
         2784,
         2784,
         2899,
         2790,
         2790,
         2902,
         2901,
         2899,
         2790,
         2899,
         2899,
         2899,
         2899,
         22,
         23,
         1301,
         1302,
         1302,
         1312,
         22,
         1313,
         1313,
         22,
         22,
         22,
         22,
         22,
         1301,
         1301,
         1301,
         1301,
         1301,
         2597,
         2598,
         2598,
         2897,
         1312,
         1312,
         2897,
         2897,
         1312,
         22,
         22,
         2789,
         2790,
         2897,
         2898,
         2899,
         2791,
         2792,
         2789,
         2790,
         22,
         2776,
         2776,
         1301,
         1301,
         1301,
         2897,
         2900,
         2902,
         2897,
         2901,
         2897,
         2900,
         2897,
         2898,
         2899,
         2902,
         2775,
         2779,
         2785,
         2788,
         2788,
         2792,
         2775,
         2794,
         2794,
         2775,
         2775,
         2775,
         2775,
         2775,
         2785,
         2785,
         2785,
         2785,
         2785,
         2796,
         2799,
         2799,
         2792,
         2901,
         2792,
         2901,
         2792,
         2901,
         2775,
         2775,
         2792,
         2792,
         2792,
         2792,
         2792,
         2901,
         2901,
         2792,
         2792,
         2775,
         2804,
         2804,
         2785,
         2785,
         2785,
         2792,
         2901,
         2901,
         2901,
         2901,
         2792,
         2901,
         2792,
         2792,
         2792,
         2901,
         22,
         23,
         1301,
         1302,
         1302,
         1312,
         22,
         1313,
         1313,
         22,
         22,
         22,
         22,
         22,
         1301,
         1301,
         1301,
         1301,
         1301,
         2597,
         2598,
         2598,
         2897,
         1312,
         2612,
         2897,
         2903,
         2612,
         22,
         22,
         2789,
         2790,
         2897,
         2898,
         2899,
         2791,
         2792,
         2800,
         2801,
         22,
         2776,
         2776,
         1301,
         1301,
         1301,
         2897,
         2790,
         2899,
         2897,
         2792,
         2903,
         2801,
         2897,
         2898,
         2899,
         2899,
         2774,
         2777,
         2784,
         2787,
         2787,
         2790,
         2774,
         2793,
         2793,
         2774,
         2774,
         2774,
         2774,
         2774,
         2784,
         2784,
         2784,
         2784,
         2784,
         2795,
         2798,
         2798,
         2790,
         2900,
         2801,
         2900,
         2801,
         2904,
         2774,
         2774,
         2790,
         2790,
         2790,
         2790,
         2790,
         2901,
         2901,
         2801,
         2801,
         2774,
         2803,
         2803,
         2784,
         2784,
         2784,
         2790,
         2790,
         2790,
         2900,
         2901,
         2801,
         2801,
         2790,
         2790,
         2790,
         2790,
         22,
         2778,
         1301,
         1302,
         1302,
         1312,
         22,
         1313,
         1313,
         22,
         22,
         22,
         22,
         22,
         1301,
         1301,
         1301,
         1301,
         1301,
         2597,
         2598,
         2598,
         2897,
         1312,
         1312,
         2897,
         2897,
         1312,
         22,
         22,
         2789,
         2790,
         2897,
         2898,
         2899,
         2791,
         2792,
         2789,
         2790,
         22,
         2776,
         2776,
         1301,
         1301,
         1301,
         2897,
         2790,
         2899,
         2897,
         2792,
         2897,
         2790,
         2897,
         2898,
         2899,
         2899,
         2774,
         2777,
         2784,
         2787,
         2787,
         2789,
         2774,
         2793,
         2793,
         2774,
         2774,
         2774,
         2774,
         2774,
         2784,
         2784,
         2784,
         2784,
         2784,
         2795,
         2798,
         2798,
         2898,
         2789,
         2789,
         2898,
         2898,
         2789,
         2774,
         2774,
         2789,
         2790,
         2898,
         2898,
         2899,
         2791,
         2792,
         2789,
         2790,
         2774,
         2803,
         2803,
         2784,
         2784,
         2784,
         2898,
         2790,
         2899,
         2898,
         2792,
         2898,
         2790,
         2898,
         2898,
         2899,
         2899,
         2774,
         2777,
         2784,
         2787,
         2787,
         2790,
         2774,
         2793,
         2793,
         2774,
         2774,
         2774,
         2774,
         2774,
         2784,
         2784,
         2784,
         2784,
         2784,
         2795,
         2798,
         2798,
         2899,
         2790,
         2790,
         2899,
         2899,
         2790,
         2774,
         2774,
         2790,
         2790,
         2899,
         2899,
         2899,
         2792,
         2792,
         2790,
         2790,
         2774,
         2803,
         2803,
         2784,
         2784,
         2784,
         2899,
         2790,
         2899,
         2899,
         2792,
         2899,
         2790,
         2899,
         2899,
         2899,
         2899,
         2774,
         2777,
         2784,
         2787,
         2787,
         2790,
         2774,
         2793,
         2793,
         2774,
         2774,
         2774,
         2774,
         2774,
         2784,
         2784,
         2784,
         2784,
         2784,
         2795,
         2798,
         2798,
         2899,
         2900,
         2790,
         2902,
         2899,
         2900,
         2774,
         2774,
         2790,
         2790,
         2899,
         2899,
         2899,
         2901,
         2901,
         2790,
         2790,
         2774,
         2803,
         2803,
         2784,
         2784,
         2784,
         2899,
         2790,
         2899,
         2902,
         2901,
         2899,
         2790,
         2899,
         2899,
         2899,
         2899,
      },
   },
   [nir_op_iadd] = {
      .filter = (const uint16_t []) {
         0,
         1,
         2,
         3,
         4,
         4,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         9,
         0,
         9,
         0,
         10,
         0,
         10,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         0,
         11,
         0,
         11,
         0,
         12,
         0,
         12,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         0,
         13,
         0,
         13,
         0,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         14,
         0,
         0,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         17,
         0,
         18,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         19,
         20,
         21,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         22,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         8,
         8,
         23,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         7,
         7,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         0,
         0,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         14,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         14,
         14,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         14,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         14,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         14,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         14,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         14,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         14,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         5,
         5,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         15,
         15,
         15,
         15,
         15,
         15,
         16,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         24,
         6,
         6,
         6,
         25,
         26,
         27,
         28,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         29,
         30,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         31,
         32,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         33,
         33,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         34,
         35,
         36,
         0,
         0,
         0,
         18,
         18,
         18,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         7,
         7,
         8,
         8,
         23,
         23,
         23,
         23,
         23,
         8,
         8,
         8,
         8,
         22,
         22,
         22,
         22,
         22,
         22,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         4,
         5,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         32,
         32,
         30,
         37,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         0,
         14,
         0,
         14,
         0,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         22,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         14,
         0,
         0,
         4,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 38,
      .table = (const uint16_t []) {
      
         24,
         25,
         24,
         24,
         24,
         24,
         382,
         24,
         383,
         384,
         385,
         386,
         387,
         388,
         24,
         389,
         390,
         391,
         392,
         24,
         24,
         24,
         1976,
         1977,
         382,
         382,
         382,
         382,
         382,
         1978,
         1979,
         1980,
         1981,
         24,
         24,
         24,
         24,
         382,
         25,
         25,
         25,
         393,
         25,
         394,
         395,
         396,
         397,
         398,
         399,
         400,
         401,
         402,
         403,
         404,
         405,
         406,
         407,
         25,
         25,
         25,
         1982,
         1983,
         1984,
         1985,
         1986,
         1987,
         1988,
         1989,
         1990,
         1991,
         1992,
         25,
         25,
         25,
         25,
         395,
         24,
         25,
         408,
         408,
         24,
         24,
         382,
         24,
         383,
         384,
         385,
         386,
         387,
         388,
         24,
         389,
         390,
         391,
         392,
         24,
         24,
         24,
         1976,
         1977,
         382,
         382,
         382,
         382,
         382,
         1978,
         1979,
         1980,
         1981,
         24,
         24,
         24,
         24,
         382,
         24,
         393,
         408,
         408,
         24,
         24,
         382,
         24,
         383,
         384,
         385,
         386,
         387,
         388,
         24,
         389,
         390,
         391,
         392,
         24,
         24,
         24,
         1976,
         1977,
         382,
         382,
         382,
         382,
         382,
         1978,
         1979,
         1980,
         1981,
         24,
         24,
         24,
         24,
         382,
         24,
         25,
         24,
         24,
         409,
         409,
         382,
         24,
         383,
         384,
         385,
         386,
         387,
         388,
         24,
         389,
         390,
         391,
         392,
         24,
         24,
         24,
         1976,
         1977,
         382,
         382,
         382,
         382,
         382,
         1978,
         1979,
         1980,
         1981,
         24,
         24,
         24,
         24,
         382,
         24,
         394,
         24,
         24,
         409,
         409,
         382,
         24,
         383,
         384,
         385,
         386,
         387,
         388,
         24,
         389,
         390,
         391,
         392,
         24,
         24,
         24,
         1976,
         1977,
         382,
         382,
         382,
         382,
         382,
         1978,
         1979,
         1980,
         1981,
         24,
         24,
         24,
         24,
         382,
         382,
         395,
         382,
         382,
         382,
         382,
         382,
         410,
         411,
         412,
         413,
         414,
         415,
         416,
         382,
         417,
         418,
         419,
         420,
         382,
         382,
         382,
         1993,
         1994,
         382,
         382,
         382,
         382,
         382,
         1995,
         1996,
         1997,
         1998,
         382,
         382,
         382,
         382,
         382,
         24,
         396,
         24,
         24,
         24,
         24,
         410,
         24,
         383,
         384,
         385,
         386,
         387,
         388,
         24,
         389,
         390,
         391,
         392,
         24,
         24,
         24,
         1976,
         1977,
         410,
         410,
         410,
         410,
         410,
         1978,
         1979,
         1980,
         1981,
         24,
         24,
         24,
         24,
         410,
         383,
         397,
         383,
         383,
         383,
         383,
         411,
         383,
         383,
         421,
         422,
         423,
         424,
         425,
         383,
         426,
         427,
         428,
         429,
         383,
         383,
         383,
         1977,
         1977,
         411,
         411,
         411,
         411,
         411,
         1999,
         2000,
         2001,
         2002,
         383,
         383,
         383,
         383,
         411,
         384,
         398,
         384,
         384,
         384,
         384,
         412,
         384,
         421,
         384,
         430,
         431,
         432,
         433,
         384,
         434,
         435,
         436,
         437,
         384,
         384,
         384,
         2003,
         2004,
         412,
         412,
         412,
         412,
         412,
         2005,
         2006,
         2007,
         2008,
         384,
         384,
         384,
         384,
         412,
         385,
         399,
         385,
         385,
         385,
         385,
         413,
         385,
         422,
         430,
         385,
         438,
         439,
         440,
         385,
         441,
         442,
         443,
         444,
         385,
         385,
         385,
         2009,
         2010,
         413,
         413,
         413,
         413,
         413,
         2011,
         2012,
         2013,
         2014,
         385,
         385,
         385,
         385,
         413,
         386,
         400,
         386,
         386,
         386,
         386,
         414,
         386,
         423,
         431,
         438,
         386,
         445,
         446,
         386,
         447,
         448,
         449,
         450,
         386,
         386,
         386,
         2015,
         2016,
         414,
         414,
         414,
         414,
         414,
         2017,
         2018,
         2019,
         2020,
         386,
         386,
         386,
         386,
         414,
         387,
         401,
         387,
         387,
         387,
         387,
         415,
         387,
         424,
         432,
         439,
         445,
         387,
         451,
         387,
         452,
         453,
         454,
         455,
         387,
         387,
         387,
         2021,
         2022,
         415,
         415,
         415,
         415,
         415,
         2023,
         2024,
         2025,
         2026,
         387,
         387,
         387,
         387,
         415,
         388,
         402,
         388,
         388,
         388,
         388,
         416,
         388,
         425,
         433,
         440,
         446,
         451,
         388,
         388,
         456,
         457,
         458,
         459,
         388,
         388,
         388,
         2027,
         2028,
         416,
         416,
         416,
         416,
         416,
         2029,
         2030,
         2031,
         2032,
         388,
         388,
         388,
         388,
         416,
         24,
         403,
         24,
         24,
         24,
         24,
         382,
         24,
         383,
         384,
         385,
         386,
         387,
         388,
         24,
         389,
         390,
         391,
         392,
         24,
         24,
         24,
         1976,
         1977,
         382,
         382,
         382,
         382,
         382,
         1978,
         1979,
         1980,
         1981,
         24,
         24,
         24,
         24,
         382,
         389,
         404,
         389,
         389,
         389,
         389,
         417,
         389,
         426,
         434,
         441,
         447,
         452,
         456,
         389,
         389,
         460,
         461,
         462,
         389,
         389,
         389,
         2033,
         2034,
         417,
         417,
         417,
         417,
         417,
         2035,
         2036,
         2037,
         2038,
         389,
         389,
         389,
         389,
         417,
         390,
         405,
         390,
         390,
         390,
         390,
         418,
         390,
         427,
         435,
         442,
         448,
         453,
         457,
         390,
         460,
         390,
         463,
         464,
         390,
         390,
         390,
         2039,
         2040,
         418,
         418,
         418,
         418,
         418,
         2041,
         2042,
         2043,
         2044,
         390,
         390,
         390,
         390,
         418,
         391,
         406,
         391,
         391,
         391,
         391,
         419,
         391,
         428,
         436,
         443,
         449,
         454,
         458,
         391,
         461,
         463,
         391,
         465,
         391,
         391,
         391,
         2045,
         2046,
         419,
         419,
         419,
         419,
         419,
         2047,
         2048,
         2049,
         2050,
         391,
         391,
         391,
         391,
         419,
         392,
         407,
         392,
         392,
         392,
         392,
         420,
         392,
         429,
         437,
         444,
         450,
         455,
         459,
         392,
         462,
         464,
         465,
         392,
         392,
         392,
         392,
         2051,
         2052,
         420,
         420,
         420,
         420,
         420,
         2053,
         2054,
         2055,
         2056,
         392,
         392,
         392,
         392,
         420,
         24,
         25,
         24,
         24,
         24,
         24,
         382,
         24,
         383,
         384,
         385,
         386,
         387,
         388,
         24,
         389,
         390,
         391,
         392,
         24,
         466,
         466,
         1976,
         1977,
         382,
         382,
         382,
         382,
         382,
         1978,
         1979,
         1980,
         1981,
         24,
         24,
         24,
         24,
         382,
         24,
         25,
         24,
         24,
         24,
         24,
         382,
         24,
         383,
         384,
         385,
         386,
         387,
         388,
         24,
         389,
         390,
         391,
         392,
         466,
         24,
         466,
         1976,
         1977,
         382,
         382,
         382,
         382,
         382,
         1978,
         1979,
         1980,
         1981,
         24,
         24,
         24,
         24,
         382,
         24,
         25,
         24,
         24,
         24,
         24,
         382,
         24,
         383,
         384,
         385,
         386,
         387,
         388,
         24,
         389,
         390,
         391,
         392,
         466,
         466,
         466,
         1976,
         1977,
         382,
         382,
         382,
         382,
         382,
         1978,
         1979,
         1980,
         1981,
         24,
         24,
         24,
         24,
         382,
         1976,
         1982,
         1976,
         1976,
         1976,
         1976,
         1993,
         1976,
         1977,
         2003,
         2009,
         2015,
         2021,
         2027,
         1976,
         2033,
         2039,
         2045,
         2051,
         1976,
         1976,
         1976,
         1976,
         1977,
         1993,
         1993,
         1993,
         1993,
         1993,
         2057,
         2058,
         2059,
         2060,
         1976,
         1976,
         1976,
         1976,
         1993,
         1977,
         1983,
         1977,
         1977,
         1977,
         1977,
         1994,
         1977,
         1977,
         2004,
         2010,
         2016,
         2022,
         2028,
         1977,
         2034,
         2040,
         2046,
         2052,
         1977,
         1977,
         1977,
         1977,
         1977,
         1994,
         1994,
         1994,
         1994,
         1994,
         2061,
         2062,
         2063,
         2064,
         1977,
         1977,
         1977,
         1977,
         1994,
         382,
         1984,
         382,
         382,
         382,
         382,
         382,
         410,
         411,
         412,
         413,
         414,
         415,
         416,
         382,
         417,
         418,
         419,
         420,
         382,
         382,
         382,
         1993,
         1994,
         382,
         382,
         382,
         382,
         382,
         1995,
         1996,
         1997,
         1998,
         382,
         382,
         382,
         382,
         382,
         382,
         1985,
         382,
         382,
         382,
         382,
         382,
         410,
         411,
         412,
         413,
         414,
         415,
         416,
         382,
         417,
         418,
         419,
         420,
         382,
         382,
         382,
         1993,
         1994,
         382,
         382,
         382,
         382,
         382,
         1995,
         1996,
         1997,
         1998,
         382,
         382,
         382,
         382,
         382,
         382,
         1986,
         382,
         382,
         382,
         382,
         382,
         410,
         411,
         412,
         413,
         414,
         415,
         416,
         382,
         417,
         418,
         419,
         420,
         382,
         382,
         382,
         1993,
         1994,
         382,
         382,
         382,
         382,
         382,
         1995,
         1996,
         1997,
         1998,
         382,
         382,
         382,
         382,
         382,
         382,
         1987,
         382,
         382,
         382,
         382,
         382,
         410,
         411,
         412,
         413,
         414,
         415,
         416,
         382,
         417,
         418,
         419,
         420,
         382,
         382,
         382,
         1993,
         1994,
         382,
         382,
         382,
         382,
         382,
         1995,
         1996,
         1997,
         1998,
         382,
         382,
         382,
         382,
         382,
         382,
         1988,
         382,
         382,
         382,
         382,
         382,
         410,
         411,
         412,
         413,
         414,
         415,
         416,
         382,
         417,
         418,
         419,
         420,
         382,
         382,
         382,
         1993,
         1994,
         382,
         382,
         382,
         382,
         382,
         1995,
         1996,
         1997,
         1998,
         382,
         382,
         382,
         382,
         382,
         1978,
         1989,
         1978,
         1978,
         1978,
         1978,
         1995,
         1978,
         1999,
         2005,
         2011,
         2017,
         2023,
         2029,
         1978,
         2035,
         2041,
         2047,
         2053,
         1978,
         1978,
         1978,
         2057,
         2061,
         1995,
         1995,
         1995,
         1995,
         1995,
         1978,
         2065,
         2066,
         2067,
         1978,
         1978,
         1978,
         1978,
         1995,
         1979,
         1990,
         1979,
         1979,
         1979,
         1979,
         1996,
         1979,
         2000,
         2006,
         2012,
         2018,
         2024,
         2030,
         1979,
         2036,
         2042,
         2048,
         2054,
         1979,
         1979,
         1979,
         2058,
         2062,
         1996,
         1996,
         1996,
         1996,
         1996,
         2065,
         1979,
         2068,
         2069,
         1979,
         1979,
         1979,
         1979,
         1996,
         1980,
         1991,
         1980,
         1980,
         1980,
         1980,
         1997,
         1980,
         2001,
         2007,
         2013,
         2019,
         2025,
         2031,
         1980,
         2037,
         2043,
         2049,
         2055,
         1980,
         1980,
         1980,
         2059,
         2063,
         1997,
         1997,
         1997,
         1997,
         1997,
         2066,
         2068,
         1980,
         2070,
         1980,
         1980,
         1980,
         1980,
         1997,
         1981,
         1992,
         1981,
         1981,
         1981,
         1981,
         1998,
         1981,
         2002,
         2008,
         2014,
         2020,
         2026,
         2032,
         1981,
         2038,
         2044,
         2050,
         2056,
         1981,
         1981,
         1981,
         2060,
         2064,
         1998,
         1998,
         1998,
         1998,
         1998,
         2067,
         2069,
         2070,
         1981,
         1981,
         1981,
         1981,
         1981,
         1998,
         24,
         25,
         24,
         24,
         24,
         24,
         382,
         24,
         383,
         384,
         385,
         386,
         387,
         388,
         24,
         389,
         390,
         391,
         392,
         24,
         24,
         24,
         1976,
         1977,
         382,
         382,
         382,
         382,
         382,
         1978,
         1979,
         1980,
         1981,
         24,
         24,
         24,
         24,
         2879,
         24,
         25,
         24,
         24,
         24,
         24,
         382,
         24,
         383,
         384,
         385,
         386,
         387,
         388,
         24,
         389,
         390,
         391,
         392,
         24,
         24,
         24,
         1976,
         1977,
         382,
         382,
         382,
         382,
         382,
         1978,
         1979,
         1980,
         1981,
         24,
         24,
         2071,
         2071,
         382,
         24,
         25,
         24,
         24,
         24,
         24,
         382,
         24,
         383,
         384,
         385,
         386,
         387,
         388,
         24,
         389,
         390,
         391,
         392,
         24,
         24,
         24,
         1976,
         1977,
         382,
         382,
         382,
         382,
         382,
         1978,
         1979,
         1980,
         1981,
         24,
         2071,
         24,
         2071,
         382,
         24,
         25,
         24,
         24,
         24,
         24,
         382,
         24,
         383,
         384,
         385,
         386,
         387,
         388,
         24,
         389,
         390,
         391,
         392,
         24,
         24,
         24,
         1976,
         1977,
         382,
         382,
         382,
         382,
         382,
         1978,
         1979,
         1980,
         1981,
         24,
         2071,
         2071,
         2071,
         382,
         382,
         395,
         382,
         382,
         382,
         382,
         382,
         410,
         411,
         412,
         413,
         414,
         415,
         416,
         382,
         417,
         418,
         419,
         420,
         382,
         382,
         382,
         1993,
         1994,
         382,
         382,
         382,
         382,
         382,
         1995,
         1996,
         1997,
         1998,
         2879,
         382,
         382,
         382,
         382,
      },
   },
   [nir_op_iadd_sat] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         26,
         27,
         27,
         27,
      },
   },
   [nir_op_isub_sat] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         28,
         29,
         28,
         29,
      },
   },
   [nir_op_uadd_sat] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         30,
         31,
         31,
         31,
      },
   },
   [nir_op_usub_sat] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         32,
         33,
         32,
         33,
      },
   },
   [nir_op_usadd_4x8_vc4] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         34,
         34,
         34,
      },
   },
   [nir_op_fmul] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         5,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         0,
         0,
         9,
         10,
         0,
         0,
         11,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         0,
         0,
         0,
         0,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         16,
         15,
         15,
         15,
         0,
         0,
         0,
         0,
         9,
         10,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         17,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         18,
         0,
         19,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         18,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         20,
         3,
         3,
         3,
         4,
         4,
         4,
         4,
         4,
         4,
         21,
         4,
         4,
         4,
         4,
         3,
         20,
         3,
         3,
         3,
         20,
         3,
         20,
         3,
         3,
         3,
         20,
         20,
         20,
         3,
         3,
         3,
         3,
         3,
         3,
         5,
         5,
         6,
         6,
         6,
         6,
         6,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         13,
         13,
         13,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         4,
         4,
         4,
         4,
         4,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         20,
         20,
         20,
         20,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         22,
         7,
         18,
         0,
         19,
         19,
         19,
         19,
         23,
         19,
         0,
         0,
         18,
         18,
         0,
         18,
         18,
         0,
         0,
         2,
         2,
         2,
         11,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         9,
         0,
         9,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         24,
         0,
         0,
         19,
         18,
         18,
         18,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         18,
         18,
         18,
         0,
         0,
         0,
         0,
         18,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         4,
         4,
         3,
         3,
         3,
         3,
         20,
         20,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         19,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 25,
      .table = (const uint16_t []) {
      
         35,
         36,
         1317,
         35,
         35,
         35,
         35,
         1318,
         1319,
         35,
         1320,
         35,
         1321,
         1322,
         35,
         35,
         2304,
         1317,
         2305,
         2306,
         2307,
         2307,
         2833,
         2834,
         35,
         36,
         36,
         1323,
         1324,
         1325,
         1326,
         1327,
         1328,
         1329,
         1330,
         1331,
         36,
         1332,
         1333,
         36,
         1330,
         2308,
         1323,
         2309,
         2310,
         2311,
         2312,
         2835,
         2836,
         2895,
         1317,
         1323,
         1317,
         1317,
         1317,
         1317,
         1317,
         1334,
         1335,
         1317,
         1336,
         1317,
         1337,
         1338,
         1317,
         1317,
         2313,
         1317,
         2314,
         2315,
         2316,
         2316,
         2837,
         2838,
         1317,
         35,
         1324,
         1317,
         35,
         35,
         35,
         35,
         1318,
         1339,
         35,
         1320,
         35,
         1321,
         1322,
         35,
         35,
         2304,
         1317,
         2305,
         2306,
         2307,
         2307,
         2833,
         2834,
         35,
         35,
         1325,
         1317,
         35,
         35,
         35,
         35,
         1318,
         1339,
         35,
         1320,
         35,
         1321,
         1322,
         35,
         35,
         2304,
         1317,
         2305,
         2306,
         2307,
         2307,
         2833,
         2834,
         35,
         35,
         1326,
         1317,
         35,
         35,
         35,
         35,
         1318,
         1319,
         35,
         1320,
         35,
         1321,
         1322,
         35,
         35,
         2304,
         1317,
         2305,
         2306,
         2307,
         2307,
         2833,
         2834,
         35,
         35,
         1327,
         1317,
         35,
         35,
         35,
         35,
         1318,
         1319,
         35,
         1320,
         35,
         1321,
         1322,
         35,
         35,
         2304,
         1317,
         2305,
         2306,
         2307,
         2307,
         2833,
         2834,
         35,
         1318,
         1328,
         1334,
         1318,
         1318,
         1318,
         1318,
         1340,
         1341,
         1318,
         1342,
         1318,
         1343,
         1344,
         1318,
         1318,
         2317,
         2318,
         2319,
         2320,
         2321,
         2321,
         2839,
         2840,
         1318,
         1319,
         1329,
         1335,
         1339,
         1339,
         1319,
         1319,
         1341,
         1319,
         1319,
         1345,
         1319,
         1346,
         1347,
         1319,
         1319,
         2322,
         1335,
         2323,
         2324,
         2325,
         2325,
         2841,
         2842,
         1319,
         35,
         1330,
         1317,
         35,
         35,
         35,
         35,
         1318,
         1319,
         35,
         1320,
         35,
         1321,
         1322,
         35,
         35,
         2304,
         1317,
         2305,
         2306,
         2307,
         2307,
         2833,
         2834,
         35,
         1320,
         1331,
         1336,
         1320,
         1320,
         1320,
         1320,
         1342,
         1345,
         1320,
         1320,
         1320,
         1348,
         1349,
         1320,
         1320,
         2304,
         1336,
         2326,
         2327,
         2328,
         2328,
         2843,
         2844,
         1320,
         35,
         36,
         1317,
         35,
         35,
         35,
         35,
         1318,
         1319,
         35,
         1320,
         1350,
         1321,
         1322,
         35,
         35,
         2304,
         1317,
         2305,
         2306,
         2307,
         2307,
         2833,
         2834,
         35,
         1321,
         1332,
         1337,
         1321,
         1321,
         1321,
         1321,
         1343,
         1346,
         1321,
         1348,
         1321,
         1351,
         1352,
         1321,
         1321,
         2329,
         1337,
         2330,
         2331,
         2332,
         2332,
         2845,
         2846,
         1321,
         1322,
         1333,
         1338,
         1322,
         1322,
         1322,
         1322,
         1344,
         1347,
         1322,
         1349,
         1322,
         1352,
         1322,
         1322,
         1322,
         2333,
         1338,
         2334,
         2335,
         2336,
         2336,
         2847,
         2848,
         1322,
         35,
         36,
         1317,
         35,
         35,
         35,
         35,
         1318,
         1319,
         35,
         1320,
         35,
         1321,
         1322,
         1353,
         35,
         2304,
         1317,
         2305,
         2306,
         2307,
         2307,
         2833,
         2834,
         35,
         35,
         1330,
         1317,
         35,
         35,
         35,
         35,
         1318,
         1319,
         35,
         1320,
         35,
         1321,
         1322,
         35,
         2337,
         2338,
         1317,
         2305,
         2306,
         2307,
         2307,
         2833,
         2834,
         35,
         2304,
         2308,
         2313,
         2304,
         2304,
         2304,
         2304,
         2317,
         2322,
         2304,
         2304,
         2304,
         2329,
         2333,
         2304,
         2338,
         2338,
         2313,
         2339,
         2340,
         2341,
         2341,
         2849,
         2850,
         2304,
         1317,
         1323,
         1317,
         1317,
         1317,
         1317,
         1317,
         2318,
         1335,
         1317,
         1336,
         1317,
         1337,
         1338,
         1317,
         1317,
         2313,
         1317,
         2314,
         2315,
         2316,
         2316,
         2851,
         2838,
         1317,
         2305,
         2309,
         2314,
         2305,
         2305,
         2305,
         2305,
         2319,
         2323,
         2305,
         2326,
         2305,
         2330,
         2334,
         2305,
         2305,
         2339,
         2314,
         2305,
         2306,
         2342,
         2342,
         2852,
         2834,
         2305,
         2306,
         2310,
         2315,
         2306,
         2306,
         2306,
         2306,
         2320,
         2324,
         2306,
         2327,
         2306,
         2331,
         2335,
         2306,
         2306,
         2340,
         2315,
         2306,
         2343,
         2344,
         2344,
         2853,
         2854,
         2306,
         2307,
         2311,
         2316,
         2307,
         2307,
         2307,
         2307,
         2321,
         2325,
         2307,
         2328,
         2307,
         2332,
         2336,
         2307,
         2307,
         2341,
         2316,
         2342,
         2344,
         2307,
         2307,
         2855,
         2856,
         2307,
         2307,
         2312,
         2316,
         2307,
         2307,
         2307,
         2307,
         2321,
         2325,
         2307,
         2328,
         2307,
         2332,
         2336,
         2307,
         2307,
         2341,
         2316,
         2342,
         2344,
         2307,
         2307,
         2855,
         2856,
         2307,
         2833,
         2835,
         2837,
         2833,
         2833,
         2833,
         2833,
         2839,
         2841,
         2833,
         2843,
         2833,
         2845,
         2847,
         2833,
         2833,
         2849,
         2851,
         2852,
         2853,
         2855,
         2855,
         2839,
         2857,
         2833,
         2834,
         2836,
         2838,
         2834,
         2834,
         2834,
         2834,
         2840,
         2842,
         2834,
         2844,
         2834,
         2846,
         2848,
         2834,
         2834,
         2850,
         2838,
         2834,
         2854,
         2856,
         2856,
         2857,
         2854,
         2834,
         35,
         2895,
         1317,
         35,
         35,
         35,
         35,
         1318,
         1319,
         35,
         1320,
         35,
         1321,
         1322,
         35,
         35,
         2304,
         1317,
         2305,
         2306,
         2307,
         2307,
         2833,
         2834,
         35,
      },
   },
   [nir_op_fmulz] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         0,
         0,
         5,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         4,
         4,
         4,
         4,
         4,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         5,
         0,
         5,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 8,
      .table = (const uint16_t []) {
      
         37,
         38,
         1354,
         37,
         37,
         37,
         1355,
         37,
         38,
         38,
         1356,
         1357,
         1358,
         1359,
         1360,
         38,
         1354,
         1356,
         1354,
         1354,
         1354,
         1354,
         1361,
         1354,
         37,
         1357,
         1354,
         37,
         37,
         37,
         1355,
         37,
         37,
         1358,
         1354,
         37,
         37,
         37,
         1355,
         37,
         37,
         1359,
         1354,
         37,
         37,
         37,
         1355,
         37,
         1355,
         1360,
         1361,
         1355,
         1355,
         1355,
         1355,
         1355,
         37,
         38,
         1354,
         37,
         37,
         37,
         1355,
         1362,
      },
   },
   [nir_op_ffma] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         5,
         5,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         9,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         11,
         3,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         12,
         4,
         3,
         3,
         3,
         11,
         3,
         3,
         3,
         3,
         3,
         11,
         3,
         3,
         11,
         3,
         11,
         3,
         3,
         11,
         11,
         3,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         6,
         6,
         6,
         6,
         13,
         13,
         13,
         13,
         6,
         6,
         6,
         13,
         13,
         13,
         13,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         13,
         13,
         13,
         13,
         13,
         14,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         4,
         4,
         4,
         4,
         4,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         11,
         11,
         11,
         11,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         14,
         13,
         13,
         13,
         13,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         0,
         10,
         10,
         10,
         10,
         15,
         10,
         0,
         0,
         10,
         10,
         0,
         10,
         10,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         0,
         0,
         0,
         0,
         10,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         4,
         4,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         11,
         11,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         6,
         6,
         13,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 16,
      .table = (const uint16_t []) {
      
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         41,
         42,
         41,
         1365,
         1365,
         41,
         1366,
         41,
         41,
         41,
         41,
         1365,
         1365,
         1366,
         1366,
         41,
         1367,
         1368,
         1367,
         1369,
         1369,
         1367,
         1370,
         1367,
         1367,
         1367,
         1367,
         1369,
         1369,
         1370,
         1370,
         1367,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         2345,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         2346,
         2347,
         2346,
         2348,
         2348,
         2346,
         2349,
         2346,
         2346,
         2346,
         2346,
         2348,
         2348,
         2349,
         2349,
         2346,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         2858,
         2858,
         1364,
         1364,
         39,
         41,
         42,
         41,
         1365,
         1365,
         41,
         1366,
         41,
         41,
         41,
         41,
         1365,
         1365,
         1366,
         1366,
         41,
         43,
         44,
         43,
         1372,
         1372,
         43,
         1373,
         43,
         43,
         43,
         43,
         1372,
         1372,
         1373,
         1373,
         43,
         1374,
         1375,
         1374,
         1376,
         1376,
         1374,
         1377,
         1374,
         1374,
         1374,
         1374,
         1376,
         1376,
         1377,
         1377,
         1374,
         1378,
         1379,
         1378,
         1380,
         1380,
         1378,
         1381,
         1378,
         1378,
         1378,
         1378,
         1380,
         1380,
         1381,
         1381,
         1378,
         1382,
         1383,
         1382,
         1384,
         1384,
         1382,
         1385,
         1382,
         1382,
         1382,
         1382,
         1384,
         1384,
         1385,
         1385,
         1382,
         1386,
         1387,
         1386,
         1388,
         1388,
         1386,
         1389,
         1386,
         1386,
         1386,
         1386,
         1388,
         1388,
         1389,
         1389,
         1386,
         41,
         42,
         41,
         1365,
         1365,
         41,
         1366,
         41,
         41,
         41,
         41,
         1365,
         1365,
         1366,
         1366,
         41,
         41,
         42,
         41,
         1365,
         1365,
         41,
         1366,
         41,
         41,
         41,
         41,
         1365,
         1365,
         1366,
         2350,
         41,
         41,
         42,
         41,
         1365,
         1365,
         41,
         1366,
         41,
         41,
         41,
         41,
         1365,
         1365,
         1366,
         1366,
         41,
         2351,
         2352,
         2351,
         2353,
         2353,
         2351,
         2354,
         2351,
         2351,
         2351,
         2351,
         2353,
         2353,
         2354,
         2354,
         2351,
         41,
         42,
         41,
         1365,
         1365,
         41,
         1366,
         41,
         41,
         41,
         41,
         1365,
         1365,
         1366,
         1366,
         41,
         1378,
         1379,
         1378,
         1380,
         1380,
         1378,
         1381,
         1378,
         1378,
         1378,
         1378,
         1380,
         1380,
         1381,
         1381,
         1378,
         1382,
         1383,
         1382,
         1384,
         1384,
         1382,
         1385,
         1382,
         1382,
         1382,
         1382,
         1384,
         1384,
         1385,
         1385,
         1382,
         41,
         42,
         41,
         1365,
         1365,
         41,
         1366,
         41,
         41,
         41,
         41,
         1365,
         1365,
         1366,
         1366,
         41,
         41,
         42,
         41,
         1365,
         1365,
         41,
         1366,
         41,
         41,
         41,
         41,
         1365,
         1365,
         1366,
         1366,
         41,
         41,
         42,
         41,
         1365,
         1365,
         41,
         1366,
         41,
         41,
         41,
         41,
         2859,
         2859,
         1366,
         1366,
         41,
         1367,
         1368,
         1367,
         1369,
         1369,
         1367,
         1370,
         1367,
         1367,
         1367,
         1367,
         1369,
         1369,
         1370,
         1370,
         1367,
         1374,
         1375,
         1374,
         1376,
         1376,
         1374,
         1377,
         1374,
         1374,
         1374,
         1374,
         1376,
         1376,
         1377,
         1377,
         1374,
         1390,
         1391,
         1390,
         1392,
         1392,
         1390,
         1393,
         1390,
         1390,
         1390,
         1390,
         1392,
         1392,
         1393,
         1393,
         1390,
         1367,
         1368,
         1367,
         1394,
         1394,
         1367,
         1370,
         1367,
         1367,
         1367,
         1367,
         1394,
         1394,
         1370,
         1370,
         1367,
         1367,
         1368,
         1367,
         1394,
         1394,
         1367,
         1370,
         1367,
         1367,
         1367,
         1367,
         1394,
         1394,
         1370,
         1370,
         1367,
         1367,
         1368,
         1367,
         1369,
         1369,
         1367,
         1370,
         1367,
         1367,
         1367,
         1367,
         1369,
         1369,
         1370,
         1370,
         1367,
         1367,
         1368,
         1367,
         1369,
         1369,
         1367,
         1370,
         1367,
         1367,
         1367,
         1367,
         1369,
         1369,
         1370,
         1370,
         1367,
         1395,
         1396,
         1395,
         1397,
         1397,
         1395,
         1398,
         1395,
         1395,
         1395,
         1395,
         1397,
         1397,
         1398,
         2355,
         1395,
         1367,
         1368,
         1367,
         1369,
         1369,
         1367,
         1370,
         1367,
         1367,
         1367,
         1367,
         1369,
         1369,
         1370,
         1370,
         1367,
         2356,
         2357,
         2356,
         2358,
         2358,
         2356,
         2359,
         2356,
         2356,
         2356,
         2356,
         2358,
         2358,
         2359,
         2359,
         2356,
         1367,
         1368,
         1367,
         1369,
         1369,
         1367,
         1370,
         1367,
         1367,
         1367,
         1367,
         1369,
         1369,
         1370,
         1370,
         1367,
         1367,
         1368,
         1367,
         1394,
         1394,
         1367,
         1370,
         1367,
         1367,
         1367,
         1367,
         1394,
         1394,
         1370,
         1370,
         1367,
         1367,
         1368,
         1367,
         1394,
         1394,
         1367,
         1370,
         1367,
         1367,
         1367,
         1367,
         1394,
         1394,
         1370,
         1370,
         1367,
         1367,
         1368,
         1367,
         1369,
         1369,
         1367,
         1370,
         1367,
         1367,
         1367,
         1367,
         1369,
         1369,
         1370,
         1370,
         1367,
         1367,
         1368,
         1367,
         1369,
         1369,
         1367,
         1370,
         1367,
         1367,
         1367,
         1367,
         1369,
         1369,
         1370,
         1370,
         1367,
         1367,
         1368,
         1367,
         1369,
         1369,
         1367,
         1370,
         1367,
         1367,
         1367,
         1367,
         2860,
         2860,
         1370,
         1370,
         1367,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         1378,
         1379,
         1378,
         1380,
         1380,
         1378,
         1381,
         1378,
         1378,
         1378,
         1378,
         1380,
         1380,
         1381,
         1381,
         1378,
         1367,
         1368,
         1367,
         1394,
         1394,
         1367,
         1370,
         1367,
         1367,
         1367,
         1367,
         1394,
         1394,
         1370,
         1370,
         1367,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         2345,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         2346,
         2347,
         2346,
         2360,
         2360,
         2346,
         2349,
         2346,
         2346,
         2346,
         2346,
         2360,
         2360,
         2349,
         2349,
         2346,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         2861,
         2861,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         1382,
         1383,
         1382,
         1384,
         1384,
         1382,
         1385,
         1382,
         1382,
         1382,
         1382,
         1384,
         1384,
         1385,
         1385,
         1382,
         1367,
         1368,
         1367,
         1394,
         1394,
         1367,
         1370,
         1367,
         1367,
         1367,
         1367,
         1394,
         1394,
         1370,
         1370,
         1367,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         2345,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         2346,
         2347,
         2346,
         2360,
         2360,
         2346,
         2349,
         2346,
         2346,
         2346,
         2346,
         2360,
         2360,
         2349,
         2349,
         2346,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         2861,
         2861,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         1386,
         1387,
         1386,
         1388,
         1388,
         1386,
         1389,
         1386,
         1386,
         1386,
         1386,
         1388,
         1388,
         1389,
         1389,
         1386,
         1367,
         1368,
         1367,
         1369,
         1369,
         1367,
         1370,
         1367,
         1367,
         1367,
         1367,
         1369,
         1369,
         1370,
         1370,
         1367,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         2345,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         2346,
         2347,
         2346,
         2348,
         2348,
         2346,
         2349,
         2346,
         2346,
         2346,
         2346,
         2348,
         2348,
         2349,
         2349,
         2346,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         2858,
         2858,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         41,
         42,
         41,
         1365,
         1365,
         41,
         1366,
         41,
         41,
         41,
         41,
         1365,
         1365,
         1366,
         1366,
         41,
         1367,
         1368,
         1367,
         1369,
         1369,
         1367,
         1370,
         1367,
         1367,
         1367,
         1367,
         1369,
         1369,
         1370,
         1370,
         1367,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         2345,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         2346,
         2347,
         2346,
         2348,
         2348,
         2346,
         2349,
         2346,
         2346,
         2346,
         2346,
         2348,
         2348,
         2349,
         2349,
         2346,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         2858,
         2858,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         2345,
         39,
         41,
         42,
         41,
         1365,
         1365,
         41,
         1366,
         41,
         41,
         41,
         41,
         1365,
         1365,
         1366,
         2350,
         41,
         1395,
         1396,
         1395,
         1397,
         1397,
         1395,
         1398,
         1395,
         1395,
         1395,
         1395,
         1397,
         1397,
         1398,
         2355,
         1395,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         2345,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         2345,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         2345,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         2345,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         2345,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         2345,
         39,
         2346,
         2347,
         2346,
         2348,
         2348,
         2346,
         2349,
         2346,
         2346,
         2346,
         2346,
         2348,
         2348,
         2349,
         2361,
         2346,
         2362,
         2363,
         2362,
         2364,
         2364,
         2362,
         2365,
         2362,
         2362,
         2362,
         2362,
         2364,
         2364,
         2365,
         2366,
         2362,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         2345,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         2345,
         39,
         39,
         40,
         39,
         2367,
         2367,
         39,
         1364,
         39,
         39,
         39,
         39,
         2367,
         2367,
         1364,
         2345,
         39,
         39,
         40,
         39,
         2367,
         2367,
         39,
         1364,
         39,
         39,
         39,
         39,
         2367,
         2367,
         1364,
         2345,
         39,
         2362,
         2363,
         2362,
         2364,
         2364,
         2362,
         2365,
         2362,
         2362,
         2362,
         2362,
         2862,
         2862,
         2365,
         2366,
         2362,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         41,
         42,
         41,
         1365,
         1365,
         41,
         1366,
         41,
         41,
         41,
         41,
         1365,
         1365,
         1366,
         1366,
         41,
         1367,
         1368,
         1367,
         1369,
         1369,
         1367,
         1370,
         1367,
         1367,
         1367,
         1367,
         1369,
         1369,
         1370,
         1370,
         1367,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         2345,
         39,
         2368,
         2369,
         2368,
         2370,
         2370,
         2368,
         2371,
         2368,
         2368,
         2368,
         2368,
         2370,
         2370,
         2371,
         2371,
         2368,
         2372,
         2373,
         2372,
         2374,
         2374,
         2372,
         2375,
         2372,
         2372,
         2372,
         2372,
         2374,
         2374,
         2375,
         2375,
         2372,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         2858,
         2858,
         1364,
         1364,
         39,
         2346,
         2347,
         2346,
         2348,
         2348,
         2346,
         2349,
         2346,
         2346,
         2346,
         2346,
         2348,
         2348,
         2349,
         2349,
         2346,
         2351,
         2352,
         2351,
         2353,
         2353,
         2351,
         2354,
         2351,
         2351,
         2351,
         2351,
         2353,
         2353,
         2354,
         2354,
         2351,
         2356,
         2357,
         2356,
         2358,
         2358,
         2356,
         2359,
         2356,
         2356,
         2356,
         2356,
         2358,
         2358,
         2359,
         2359,
         2356,
         2346,
         2347,
         2346,
         2360,
         2360,
         2346,
         2349,
         2346,
         2346,
         2346,
         2346,
         2360,
         2360,
         2349,
         2349,
         2346,
         2346,
         2347,
         2346,
         2360,
         2360,
         2346,
         2349,
         2346,
         2346,
         2346,
         2346,
         2360,
         2360,
         2349,
         2349,
         2346,
         2346,
         2347,
         2346,
         2348,
         2348,
         2346,
         2349,
         2346,
         2346,
         2346,
         2346,
         2348,
         2348,
         2349,
         2349,
         2346,
         2346,
         2347,
         2346,
         2348,
         2348,
         2346,
         2349,
         2346,
         2346,
         2346,
         2346,
         2348,
         2348,
         2349,
         2349,
         2346,
         2346,
         2347,
         2346,
         2348,
         2348,
         2346,
         2349,
         2346,
         2346,
         2346,
         2346,
         2348,
         2348,
         2349,
         2361,
         2346,
         2372,
         2373,
         2372,
         2374,
         2374,
         2372,
         2375,
         2372,
         2372,
         2372,
         2372,
         2374,
         2374,
         2375,
         2375,
         2372,
         2372,
         2373,
         2372,
         2374,
         2374,
         2372,
         2375,
         2372,
         2372,
         2372,
         2372,
         2374,
         2374,
         2375,
         2375,
         2372,
         2346,
         2347,
         2346,
         2348,
         2348,
         2346,
         2349,
         2346,
         2346,
         2346,
         2346,
         2348,
         2348,
         2349,
         2349,
         2346,
         2346,
         2347,
         2346,
         2360,
         2360,
         2346,
         2349,
         2346,
         2346,
         2346,
         2346,
         2360,
         2360,
         2349,
         2349,
         2346,
         2346,
         2347,
         2346,
         2360,
         2360,
         2346,
         2349,
         2346,
         2346,
         2346,
         2346,
         2360,
         2360,
         2349,
         2349,
         2346,
         2346,
         2347,
         2346,
         2348,
         2348,
         2346,
         2349,
         2346,
         2346,
         2346,
         2346,
         2348,
         2348,
         2349,
         2349,
         2346,
         2346,
         2347,
         2346,
         2348,
         2348,
         2346,
         2349,
         2346,
         2346,
         2346,
         2346,
         2348,
         2348,
         2349,
         2349,
         2346,
         2346,
         2347,
         2346,
         2348,
         2348,
         2346,
         2349,
         2346,
         2346,
         2346,
         2346,
         2863,
         2863,
         2349,
         2349,
         2346,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         41,
         42,
         41,
         1365,
         1365,
         41,
         1366,
         41,
         41,
         41,
         41,
         1365,
         1365,
         1366,
         1366,
         41,
         1367,
         1368,
         1367,
         1369,
         1369,
         1367,
         1370,
         1367,
         1367,
         1367,
         1367,
         1369,
         1369,
         1370,
         1370,
         1367,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         2362,
         2363,
         2362,
         2364,
         2364,
         2362,
         2365,
         2362,
         2362,
         2362,
         2362,
         2364,
         2364,
         2365,
         2366,
         2362,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         2346,
         2347,
         2346,
         2348,
         2348,
         2346,
         2349,
         2346,
         2346,
         2346,
         2346,
         2348,
         2348,
         2349,
         2349,
         2346,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         2858,
         2858,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         1378,
         1379,
         1378,
         1380,
         1380,
         1378,
         1381,
         1378,
         1378,
         1378,
         1378,
         1380,
         1380,
         1381,
         1381,
         1378,
         1367,
         1368,
         1367,
         1394,
         1394,
         1367,
         1370,
         1367,
         1367,
         1367,
         1367,
         1394,
         1394,
         1370,
         1370,
         1367,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         2345,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         2346,
         2347,
         2346,
         2360,
         2360,
         2346,
         2349,
         2346,
         2346,
         2346,
         2346,
         2360,
         2360,
         2349,
         2349,
         2346,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         2861,
         2861,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         1382,
         1383,
         1382,
         1384,
         1384,
         1382,
         1385,
         1382,
         1382,
         1382,
         1382,
         1384,
         1384,
         1385,
         1385,
         1382,
         1367,
         1368,
         1367,
         1394,
         1394,
         1367,
         1370,
         1367,
         1367,
         1367,
         1367,
         1394,
         1394,
         1370,
         1370,
         1367,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         2345,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         2346,
         2347,
         2346,
         2360,
         2360,
         2346,
         2349,
         2346,
         2346,
         2346,
         2346,
         2360,
         2360,
         2349,
         2349,
         2346,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         2861,
         2861,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         41,
         42,
         41,
         1365,
         1365,
         41,
         1366,
         41,
         41,
         41,
         41,
         1365,
         1365,
         1366,
         1366,
         41,
         1367,
         1368,
         1367,
         1369,
         1369,
         1367,
         1370,
         1367,
         1367,
         1367,
         1367,
         1369,
         1369,
         1370,
         1370,
         1367,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         39,
         40,
         39,
         2367,
         2367,
         39,
         1364,
         39,
         39,
         39,
         39,
         2367,
         2367,
         1364,
         2345,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         2346,
         2347,
         2346,
         2348,
         2348,
         2346,
         2349,
         2346,
         2346,
         2346,
         2346,
         2348,
         2348,
         2349,
         2349,
         2346,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         2858,
         2858,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         41,
         42,
         41,
         1365,
         1365,
         41,
         1366,
         41,
         41,
         41,
         41,
         1365,
         1365,
         1366,
         1366,
         41,
         1367,
         1368,
         1367,
         1369,
         1369,
         1367,
         1370,
         1367,
         1367,
         1367,
         1367,
         1369,
         1369,
         1370,
         1370,
         1367,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         39,
         40,
         39,
         2367,
         2367,
         39,
         1364,
         39,
         39,
         39,
         39,
         2367,
         2367,
         1364,
         2345,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         2346,
         2347,
         2346,
         2348,
         2348,
         2346,
         2349,
         2346,
         2346,
         2346,
         2346,
         2348,
         2348,
         2349,
         2349,
         2346,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         1371,
         1371,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         1363,
         1363,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         2858,
         2858,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         2858,
         2858,
         1364,
         1364,
         39,
         41,
         42,
         41,
         1365,
         1365,
         41,
         1366,
         41,
         41,
         41,
         41,
         2859,
         2859,
         1366,
         1366,
         41,
         1367,
         1368,
         1367,
         1369,
         1369,
         1367,
         1370,
         1367,
         1367,
         1367,
         1367,
         2860,
         2860,
         1370,
         1370,
         1367,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         2861,
         2861,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         2861,
         2861,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         2858,
         2858,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         2858,
         2858,
         1364,
         1364,
         39,
         2362,
         2363,
         2362,
         2364,
         2364,
         2362,
         2365,
         2362,
         2362,
         2362,
         2362,
         2862,
         2862,
         2365,
         2366,
         2362,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         2858,
         2858,
         1364,
         1364,
         39,
         2346,
         2347,
         2346,
         2348,
         2348,
         2346,
         2349,
         2346,
         2346,
         2346,
         2346,
         2863,
         2863,
         2349,
         2349,
         2346,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         2858,
         2858,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         2861,
         2861,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1371,
         1371,
         39,
         1364,
         39,
         39,
         39,
         39,
         2861,
         2861,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         2858,
         2858,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         2858,
         2858,
         1364,
         1364,
         39,
         39,
         40,
         39,
         1363,
         1363,
         39,
         1364,
         39,
         39,
         39,
         39,
         2858,
         2858,
         1364,
         1364,
         39,
      },
   },
   [nir_op_ffmaz] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         4,
         4,
         4,
         4,
         4,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 6,
      .table = (const uint16_t []) {
      
         45,
         46,
         45,
         1399,
         1399,
         1400,
         47,
         48,
         47,
         1401,
         1401,
         1402,
         45,
         46,
         45,
         1399,
         1399,
         1400,
         45,
         46,
         45,
         1403,
         1403,
         1400,
         45,
         46,
         45,
         1403,
         1403,
         1400,
         45,
         46,
         45,
         1399,
         1399,
         1400,
         47,
         48,
         47,
         1401,
         1401,
         1402,
         49,
         50,
         49,
         1404,
         1404,
         1405,
         47,
         48,
         47,
         1401,
         1401,
         1402,
         1406,
         1407,
         1406,
         1408,
         1408,
         1409,
         1410,
         1411,
         1410,
         1412,
         1412,
         1413,
         47,
         48,
         47,
         1401,
         1401,
         1402,
         45,
         46,
         45,
         1399,
         1399,
         1400,
         47,
         48,
         47,
         1401,
         1401,
         1402,
         1414,
         1415,
         1414,
         1416,
         1416,
         1417,
         45,
         46,
         45,
         1403,
         1403,
         1400,
         45,
         46,
         45,
         1403,
         1403,
         1400,
         45,
         46,
         45,
         1399,
         1399,
         1400,
         45,
         46,
         45,
         1403,
         1403,
         1400,
         1406,
         1407,
         1406,
         1408,
         1408,
         1409,
         45,
         46,
         45,
         1403,
         1403,
         1400,
         45,
         46,
         45,
         1403,
         1403,
         1400,
         45,
         46,
         45,
         1403,
         1403,
         1400,
         45,
         46,
         45,
         1403,
         1403,
         1400,
         45,
         46,
         45,
         1403,
         1403,
         1400,
         1410,
         1411,
         1410,
         1412,
         1412,
         1413,
         45,
         46,
         45,
         1403,
         1403,
         1400,
         45,
         46,
         45,
         1403,
         1403,
         1400,
         45,
         46,
         45,
         1403,
         1403,
         1400,
         45,
         46,
         45,
         1403,
         1403,
         1400,
         45,
         46,
         45,
         1399,
         1399,
         1400,
         47,
         48,
         47,
         1401,
         1401,
         1402,
         45,
         46,
         45,
         1399,
         1399,
         1400,
         45,
         46,
         45,
         1403,
         1403,
         1400,
         45,
         46,
         45,
         1403,
         1403,
         1400,
         45,
         46,
         45,
         1399,
         1399,
         1400,
      },
   },
   [nir_op_iand] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         2,
         3,
         4,
         3,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         8,
         9,
         9,
         10,
         11,
         12,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         17,
         16,
         17,
         18,
         19,
         18,
         19,
         20,
         21,
         22,
         23,
         24,
         25,
         26,
         27,
         28,
         29,
         29,
         30,
         31,
         32,
         32,
         32,
         32,
         33,
         33,
         34,
         34,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         35,
         36,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         37,
         38,
         39,
         40,
         41,
         42,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         4,
         4,
         4,
         4,
         3,
         4,
         3,
         3,
         3,
         4,
         3,
         3,
         4,
         4,
         3,
         4,
         3,
         3,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         9,
         9,
         43,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         43,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         44,
         9,
         9,
         43,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         43,
         9,
         9,
         9,
         9,
         9,
         9,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         16,
         16,
         16,
         16,
         17,
         16,
         16,
         17,
         17,
         18,
         18,
         19,
         18,
         19,
         18,
         19,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         38,
         38,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         21,
         21,
         21,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         22,
         21,
         22,
         21,
         22,
         21,
         21,
         21,
         22,
         22,
         21,
         22,
         21,
         21,
         21,
         22,
         21,
         22,
         21,
         21,
         21,
         21,
         25,
         25,
         25,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         26,
         25,
         26,
         25,
         26,
         25,
         25,
         25,
         26,
         26,
         26,
         25,
         26,
         25,
         25,
         25,
         26,
         25,
         26,
         25,
         25,
         25,
         25,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         45,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         46,
         11,
         11,
         47,
         48,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         10,
         10,
         10,
         10,
         12,
         12,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         20,
         20,
         20,
         20,
         20,
         49,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         50,
         51,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         52,
         53,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         15,
         15,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         30,
         30,
         30,
         30,
         30,
         30,
         31,
         30,
         31,
         30,
         30,
         31,
         30,
         31,
         30,
         30,
         31,
         31,
         30,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         42,
         42,
         42,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         4,
         4,
         4,
         54,
         4,
         16,
         16,
         16,
         16,
         16,
         16,
         55,
         17,
         0,
         0,
         0,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         43,
         43,
         43,
         43,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         20,
         20,
         20,
         20,
         20,
         20,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         0,
         0,
         23,
         22,
         22,
         22,
         22,
         22,
         27,
         26,
         26,
         26,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         13,
         12,
         31,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         33,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         15,
         15,
         36,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         53,
         53,
         51,
         45,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         9,
         9,
         9,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         56,
         57,
         58,
         4,
         4,
         17,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         29,
         0,
         22,
         26,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         59,
         22,
         22,
         22,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         0,
         60,
         61,
         8,
         0,
         8,
         8,
         4,
         17,
         9,
         62,
         63,
         64,
         8,
         8,
         8,
         4,
         17,
         9,
         65,
         66,
         67,
         8,
         8,
         8,
         4,
         17,
         9,
         65,
      },
      
      .num_filtered_states = 68,
      .table = (const uint16_t []) {
      
         51,
         52,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         51,
         1047,
         1047,
         1047,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1046,
         51,
         51,
         52,
         52,
         1052,
         52,
         1053,
         1054,
         52,
         1055,
         1056,
         1057,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1058,
         52,
         52,
         1059,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1060,
         52,
         1061,
         52,
         1062,
         52,
         1063,
         2449,
         2450,
         2451,
         2452,
         52,
         52,
         2453,
         2454,
         2455,
         2456,
         2457,
         2772,
         2773,
         1059,
         1059,
         1059,
         2911,
         2913,
         2914,
         2921,
         2922,
         2923,
         2930,
         2931,
         2932,
         51,
         1052,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         51,
         1047,
         1047,
         1047,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1046,
         51,
         51,
         51,
         52,
         51,
         51,
         1064,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         1064,
         51,
         1047,
         1047,
         1047,
         1046,
         1064,
         51,
         1046,
         1064,
         51,
         1046,
         1064,
         51,
         51,
         1053,
         51,
         1064,
         1064,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         1064,
         51,
         1047,
         1047,
         1047,
         1046,
         1064,
         51,
         1046,
         1064,
         51,
         1046,
         1064,
         51,
         51,
         1054,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         51,
         1047,
         1047,
         1047,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1046,
         51,
         51,
         51,
         52,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         2458,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         51,
         1047,
         1047,
         1047,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1045,
         1055,
         1045,
         1045,
         1045,
         1045,
         1045,
         1065,
         1065,
         1066,
         1045,
         1045,
         1045,
         1045,
         1045,
         1045,
         1045,
         1045,
         1045,
         1045,
         1067,
         1045,
         1045,
         1045,
         1045,
         1045,
         1045,
         1045,
         1045,
         1045,
         1045,
         1045,
         1045,
         1045,
         1045,
         1045,
         1068,
         1045,
         1069,
         1045,
         1070,
         1045,
         1071,
         2459,
         1066,
         1045,
         2460,
         1045,
         1045,
         2461,
         2462,
         2463,
         2464,
         2465,
         1045,
         1045,
         1067,
         1067,
         1067,
         1066,
         1045,
         1045,
         1066,
         1045,
         1045,
         1066,
         1045,
         1045,
         1045,
         1056,
         1045,
         1045,
         1045,
         1045,
         1045,
         1065,
         1065,
         1066,
         1045,
         1045,
         1045,
         1045,
         1045,
         1072,
         1045,
         1045,
         1045,
         1045,
         1067,
         1045,
         1045,
         1045,
         1045,
         1045,
         1045,
         1045,
         1045,
         1045,
         1045,
         1045,
         1045,
         1045,
         1045,
         1045,
         1068,
         1045,
         1069,
         1045,
         1070,
         1045,
         1071,
         2459,
         1066,
         1045,
         2460,
         1045,
         1045,
         2461,
         2462,
         2463,
         2464,
         2465,
         1045,
         1045,
         1067,
         1067,
         1067,
         1066,
         1045,
         1045,
         1066,
         1045,
         1045,
         1066,
         1045,
         1045,
         1046,
         1057,
         1046,
         1046,
         1046,
         1046,
         1046,
         1066,
         1066,
         1073,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1074,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1075,
         1046,
         1076,
         1046,
         1077,
         1046,
         1078,
         2466,
         1073,
         1046,
         2467,
         1046,
         1046,
         2468,
         2469,
         2470,
         2471,
         2472,
         1046,
         1046,
         1074,
         1074,
         1074,
         1073,
         1046,
         1046,
         1073,
         1046,
         1046,
         1073,
         1046,
         1046,
         51,
         52,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         1079,
         1079,
         51,
         51,
         51,
         51,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         51,
         1047,
         1047,
         1047,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1046,
         51,
         51,
         51,
         52,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         1080,
         1079,
         1079,
         51,
         51,
         51,
         51,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2473,
         1080,
         1080,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         51,
         1047,
         1047,
         1047,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1046,
         51,
         51,
         51,
         52,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         1079,
         1079,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2474,
         1079,
         1079,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         51,
         1047,
         1047,
         1047,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1046,
         51,
         51,
         51,
         52,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         1079,
         1079,
         51,
         1081,
         51,
         51,
         51,
         51,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2474,
         1079,
         1079,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         51,
         1047,
         1047,
         1047,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1046,
         51,
         51,
         51,
         52,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         1082,
         51,
         51,
         51,
         51,
         51,
         1047,
         1083,
         1083,
         1083,
         1083,
         1084,
         1084,
         1084,
         1084,
         1085,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         51,
         1047,
         1047,
         1047,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1046,
         51,
         51,
         51,
         52,
         51,
         51,
         51,
         51,
         51,
         1045,
         1072,
         1046,
         51,
         51,
         51,
         51,
         51,
         1086,
         51,
         51,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         51,
         1047,
         1047,
         1047,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1046,
         51,
         51,
         51,
         52,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1087,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         1087,
         1047,
         1047,
         1047,
         1046,
         51,
         1087,
         1046,
         51,
         1087,
         1046,
         51,
         1087,
         51,
         1058,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         51,
         51,
         1087,
         1087,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         1087,
         1047,
         1047,
         1047,
         1046,
         51,
         1087,
         1046,
         51,
         1087,
         1046,
         51,
         1087,
         51,
         52,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1088,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         51,
         1047,
         1047,
         1047,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1046,
         51,
         51,
         51,
         52,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1088,
         1088,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         51,
         1047,
         1047,
         1047,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1047,
         1059,
         1047,
         1047,
         1047,
         1047,
         1047,
         1067,
         1067,
         1074,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1089,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1090,
         1047,
         1091,
         1047,
         1092,
         1047,
         1093,
         2475,
         1074,
         1047,
         2476,
         1047,
         1047,
         2477,
         2478,
         2479,
         2480,
         2481,
         1047,
         1047,
         1089,
         1089,
         1089,
         1074,
         1047,
         1047,
         1074,
         1047,
         1047,
         1074,
         1047,
         1047,
         51,
         52,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         1083,
         51,
         51,
         51,
         51,
         51,
         1047,
         1094,
         1094,
         1094,
         1094,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         51,
         1047,
         1047,
         1047,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1046,
         51,
         51,
         51,
         52,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         1083,
         51,
         51,
         51,
         51,
         51,
         1047,
         1094,
         1095,
         1096,
         1097,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         51,
         1047,
         1047,
         1047,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1046,
         51,
         51,
         51,
         52,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         1083,
         51,
         51,
         51,
         51,
         51,
         1047,
         1094,
         1096,
         1098,
         1099,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         51,
         1047,
         1047,
         1047,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1046,
         51,
         51,
         51,
         52,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         1083,
         51,
         51,
         51,
         51,
         51,
         1047,
         1094,
         1097,
         1099,
         1100,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         51,
         1047,
         1047,
         1047,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1046,
         51,
         51,
         51,
         52,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         1084,
         51,
         51,
         51,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         1101,
         1101,
         1101,
         1101,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         51,
         1047,
         1047,
         1047,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1046,
         51,
         51,
         51,
         52,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         1084,
         51,
         51,
         51,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         1101,
         1102,
         1103,
         1104,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         51,
         1047,
         1047,
         1047,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1046,
         51,
         51,
         51,
         52,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         1084,
         51,
         51,
         51,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         1101,
         1103,
         1105,
         1106,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         51,
         1047,
         1047,
         1047,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1046,
         51,
         51,
         51,
         52,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         1084,
         51,
         51,
         51,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         1101,
         1104,
         1106,
         1107,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         51,
         1047,
         1047,
         1047,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1046,
         51,
         51,
         51,
         52,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         1085,
         51,
         51,
         51,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1108,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         51,
         1047,
         1047,
         1047,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1046,
         51,
         51,
         51,
         52,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1109,
         1109,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         51,
         1047,
         1047,
         1047,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1046,
         51,
         51,
         51,
         52,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1109,
         1109,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         51,
         2905,
         2906,
         2907,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1046,
         51,
         51,
         51,
         52,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1110,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         51,
         1047,
         1047,
         1047,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1046,
         51,
         51,
         51,
         52,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1111,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         51,
         1047,
         1047,
         1047,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1046,
         51,
         51,
         51,
         52,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1112,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         51,
         1047,
         1047,
         1047,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1046,
         51,
         51,
         51,
         52,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1113,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         51,
         1047,
         1047,
         1047,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1048,
         1060,
         1048,
         1048,
         1048,
         1048,
         1048,
         1068,
         1068,
         1075,
         1048,
         1048,
         1048,
         1048,
         1048,
         1048,
         1048,
         1048,
         1048,
         1048,
         1090,
         1048,
         1048,
         1048,
         1048,
         1048,
         1048,
         1048,
         1048,
         1048,
         1048,
         1048,
         1048,
         1048,
         1048,
         1113,
         1113,
         1048,
         1114,
         1048,
         1115,
         1048,
         1116,
         2482,
         1075,
         1048,
         2483,
         1048,
         1048,
         2484,
         2485,
         2486,
         2487,
         2488,
         1048,
         1048,
         1090,
         1090,
         1090,
         1075,
         1048,
         1048,
         1075,
         1048,
         1048,
         1075,
         1048,
         1048,
         51,
         52,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1117,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         51,
         1047,
         1047,
         1047,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1049,
         1061,
         1049,
         1049,
         1049,
         1049,
         1049,
         1069,
         1069,
         1076,
         1049,
         1049,
         1049,
         1049,
         1049,
         1049,
         1049,
         1049,
         1049,
         1049,
         1091,
         1049,
         1049,
         1049,
         1049,
         1049,
         1049,
         1049,
         1049,
         1049,
         1049,
         1049,
         1049,
         1049,
         1049,
         1049,
         1114,
         1117,
         1117,
         1049,
         1118,
         1049,
         1119,
         2489,
         1076,
         1049,
         2490,
         1049,
         1049,
         2491,
         2492,
         2493,
         2494,
         2495,
         1049,
         1049,
         1091,
         1091,
         1091,
         1076,
         1049,
         1049,
         1076,
         1049,
         1049,
         1076,
         1049,
         1049,
         51,
         52,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1120,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         51,
         1047,
         1047,
         1047,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1050,
         1062,
         1050,
         1050,
         1050,
         1050,
         1050,
         1070,
         1070,
         1077,
         1050,
         1050,
         1050,
         1050,
         1050,
         1050,
         1050,
         1050,
         1050,
         1050,
         1092,
         1050,
         1050,
         1050,
         1050,
         1050,
         1050,
         1050,
         1050,
         1050,
         1050,
         1050,
         1050,
         1050,
         1050,
         1050,
         1115,
         1050,
         1118,
         1120,
         1120,
         1050,
         1121,
         2496,
         1077,
         1050,
         2497,
         1050,
         1050,
         2498,
         2499,
         2500,
         2501,
         2502,
         1050,
         1050,
         1092,
         1092,
         1092,
         1077,
         1050,
         1050,
         1077,
         1050,
         1050,
         1077,
         1050,
         1050,
         51,
         52,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1122,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         51,
         1047,
         1047,
         1047,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1051,
         1063,
         1051,
         1051,
         1051,
         1051,
         1051,
         1071,
         1071,
         1078,
         1051,
         1051,
         1051,
         1051,
         1051,
         1051,
         1051,
         1051,
         1051,
         1051,
         1093,
         1051,
         1051,
         1051,
         1051,
         1051,
         1051,
         1051,
         1051,
         1051,
         1051,
         1051,
         1051,
         1051,
         1051,
         1051,
         1116,
         1051,
         1119,
         1051,
         1121,
         1122,
         1122,
         2503,
         1078,
         1051,
         2504,
         1051,
         1051,
         2505,
         2506,
         2507,
         2508,
         2509,
         1051,
         1051,
         1093,
         1093,
         1093,
         1078,
         1051,
         1051,
         1078,
         1051,
         1051,
         1078,
         1051,
         1051,
         2442,
         2449,
         2442,
         2442,
         2442,
         2442,
         2442,
         2459,
         2459,
         2466,
         2442,
         2442,
         2442,
         2442,
         2442,
         2442,
         2442,
         2442,
         2442,
         2442,
         2475,
         2442,
         2442,
         2442,
         2442,
         2442,
         2442,
         2442,
         2442,
         2442,
         2442,
         2442,
         2442,
         2442,
         2442,
         2442,
         2482,
         2442,
         2489,
         2442,
         2496,
         2442,
         2503,
         2466,
         2466,
         2442,
         2510,
         2442,
         2442,
         2511,
         2512,
         2513,
         2514,
         2515,
         2442,
         2442,
         2475,
         2475,
         2475,
         2466,
         2442,
         2442,
         2466,
         2442,
         2442,
         2466,
         2442,
         2442,
         1046,
         2450,
         1046,
         1046,
         1046,
         1046,
         1046,
         1066,
         1066,
         1073,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1074,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1075,
         1046,
         1076,
         1046,
         1077,
         1046,
         1078,
         2466,
         1073,
         1046,
         2467,
         1046,
         1046,
         2468,
         2469,
         2470,
         2471,
         2472,
         1046,
         1046,
         1074,
         1074,
         1074,
         1073,
         1046,
         1046,
         1073,
         1046,
         1046,
         1073,
         1046,
         1046,
         51,
         2451,
         51,
         51,
         51,
         51,
         2458,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         2516,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         51,
         1047,
         1047,
         1047,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1046,
         51,
         51,
         2443,
         2452,
         2443,
         2443,
         2443,
         2443,
         2443,
         2460,
         2460,
         2467,
         2443,
         2473,
         2474,
         2474,
         2443,
         2443,
         2443,
         2443,
         2443,
         2443,
         2476,
         2443,
         2443,
         2443,
         2443,
         2443,
         2443,
         2443,
         2443,
         2443,
         2443,
         2443,
         2443,
         2443,
         2443,
         2443,
         2483,
         2443,
         2490,
         2443,
         2497,
         2443,
         2504,
         2510,
         2467,
         2443,
         2473,
         2473,
         2473,
         2517,
         2518,
         2519,
         2520,
         2521,
         2443,
         2443,
         2476,
         2476,
         2476,
         2467,
         2443,
         2443,
         2467,
         2443,
         2443,
         2467,
         2443,
         2443,
         51,
         52,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         1080,
         1079,
         1079,
         51,
         51,
         51,
         51,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2473,
         1080,
         2522,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         51,
         1047,
         1047,
         1047,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1046,
         51,
         51,
         51,
         52,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         1080,
         1079,
         1079,
         51,
         51,
         51,
         51,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2473,
         2522,
         1080,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         51,
         1047,
         1047,
         1047,
         1046,
         51,
         51,
         1046,
         51,
         51,
         1046,
         51,
         51,
         2444,
         2453,
         2444,
         2444,
         2444,
         2444,
         2444,
         2461,
         2461,
         2468,
         2444,
         2444,
         2444,
         2444,
         2444,
         2444,
         2444,
         2444,
         2444,
         2444,
         2477,
         2444,
         2444,
         2444,
         2444,
         2444,
         2444,
         2444,
         2444,
         2444,
         2444,
         2444,
         2444,
         2444,
         2444,
         2444,
         2484,
         2444,
         2491,
         2444,
         2498,
         2444,
         2505,
         2511,
         2468,
         2444,
         2517,
         2444,
         2444,
         2477,
         2523,
         2524,
         2525,
         2526,
         2444,
         2444,
         2477,
         2477,
         2477,
         2468,
         2444,
         2444,
         2468,
         2444,
         2444,
         2468,
         2444,
         2444,
         2445,
         2454,
         2445,
         2445,
         2445,
         2445,
         2445,
         2462,
         2462,
         2469,
         2445,
         2445,
         2445,
         2445,
         2445,
         2445,
         2445,
         2445,
         2445,
         2445,
         2478,
         2445,
         2445,
         2445,
         2445,
         2445,
         2445,
         2445,
         2445,
         2445,
         2445,
         2445,
         2445,
         2445,
         2445,
         2445,
         2485,
         2445,
         2492,
         2445,
         2499,
         2445,
         2506,
         2512,
         2469,
         2445,
         2518,
         2445,
         2445,
         2523,
         2445,
         2527,
         2528,
         2529,
         2445,
         2445,
         2478,
         2478,
         2478,
         2469,
         2445,
         2445,
         2469,
         2445,
         2445,
         2469,
         2445,
         2445,
         2446,
         2455,
         2446,
         2446,
         2446,
         2446,
         2446,
         2463,
         2463,
         2470,
         2446,
         2446,
         2446,
         2446,
         2446,
         2446,
         2446,
         2446,
         2446,
         2446,
         2479,
         2446,
         2446,
         2446,
         2446,
         2446,
         2446,
         2446,
         2446,
         2446,
         2446,
         2446,
         2446,
         2446,
         2446,
         2446,
         2486,
         2446,
         2493,
         2446,
         2500,
         2446,
         2507,
         2513,
         2470,
         2446,
         2519,
         2446,
         2446,
         2524,
         2527,
         2446,
         2530,
         2531,
         2446,
         2446,
         2479,
         2479,
         2479,
         2470,
         2446,
         2446,
         2470,
         2446,
         2446,
         2470,
         2446,
         2446,
         2447,
         2456,
         2447,
         2447,
         2447,
         2447,
         2447,
         2464,
         2464,
         2471,
         2447,
         2447,
         2447,
         2447,
         2447,
         2447,
         2447,
         2447,
         2447,
         2447,
         2480,
         2447,
         2447,
         2447,
         2447,
         2447,
         2447,
         2447,
         2447,
         2447,
         2447,
         2447,
         2447,
         2447,
         2447,
         2447,
         2487,
         2447,
         2494,
         2447,
         2501,
         2447,
         2508,
         2514,
         2471,
         2447,
         2520,
         2447,
         2447,
         2525,
         2528,
         2530,
         2447,
         2532,
         2447,
         2447,
         2480,
         2480,
         2480,
         2471,
         2447,
         2447,
         2471,
         2447,
         2447,
         2471,
         2447,
         2447,
         2448,
         2457,
         2448,
         2448,
         2448,
         2448,
         2448,
         2465,
         2465,
         2472,
         2448,
         2448,
         2448,
         2448,
         2448,
         2448,
         2448,
         2448,
         2448,
         2448,
         2481,
         2448,
         2448,
         2448,
         2448,
         2448,
         2448,
         2448,
         2448,
         2448,
         2448,
         2448,
         2448,
         2448,
         2448,
         2448,
         2488,
         2448,
         2495,
         2448,
         2502,
         2448,
         2509,
         2515,
         2472,
         2448,
         2521,
         2448,
         2448,
         2526,
         2529,
         2531,
         2532,
         2533,
         2448,
         2448,
         2481,
         2481,
         2481,
         2472,
         2448,
         2448,
         2472,
         2448,
         2448,
         2472,
         2448,
         2448,
         51,
         2772,
         51,
         1064,
         1064,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         1064,
         51,
         1047,
         1047,
         1047,
         1046,
         1064,
         51,
         1046,
         1064,
         51,
         1046,
         1064,
         51,
         51,
         2773,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         51,
         51,
         1087,
         1087,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         1087,
         1047,
         1047,
         1047,
         1046,
         51,
         1087,
         1046,
         51,
         1087,
         1046,
         51,
         1087,
         1047,
         1059,
         1047,
         1047,
         1047,
         1047,
         1047,
         1067,
         1067,
         1074,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1089,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         2905,
         1047,
         1047,
         1047,
         1047,
         1090,
         1047,
         1091,
         1047,
         1092,
         1047,
         1093,
         2475,
         1074,
         1047,
         2476,
         1047,
         1047,
         2477,
         2478,
         2479,
         2480,
         2481,
         1047,
         1047,
         1089,
         1089,
         1089,
         1074,
         1047,
         1047,
         1074,
         1047,
         1047,
         1074,
         1047,
         1047,
         1047,
         1059,
         1047,
         1047,
         1047,
         1047,
         1047,
         1067,
         1067,
         1074,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1089,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         2906,
         1047,
         1047,
         1047,
         1047,
         1090,
         1047,
         1091,
         1047,
         1092,
         1047,
         1093,
         2475,
         1074,
         1047,
         2476,
         1047,
         1047,
         2477,
         2478,
         2479,
         2480,
         2481,
         1047,
         1047,
         1089,
         1089,
         1089,
         1074,
         1047,
         1047,
         1074,
         1047,
         1047,
         1074,
         1047,
         1047,
         1047,
         1059,
         1047,
         1047,
         1047,
         1047,
         1047,
         1067,
         1067,
         1074,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1089,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         1047,
         2907,
         1047,
         1047,
         1047,
         1047,
         1090,
         1047,
         1091,
         1047,
         1092,
         1047,
         1093,
         2475,
         1074,
         1047,
         2476,
         1047,
         1047,
         2477,
         2478,
         2479,
         2480,
         2481,
         1047,
         1047,
         1089,
         1089,
         1089,
         1074,
         1047,
         1047,
         1074,
         1047,
         1047,
         1074,
         1047,
         1047,
         1046,
         2911,
         1046,
         1046,
         1046,
         1046,
         1046,
         1066,
         1066,
         1073,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1074,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1075,
         1046,
         1076,
         1046,
         1077,
         1046,
         1078,
         2466,
         1073,
         1046,
         2467,
         1046,
         1046,
         2468,
         2469,
         2470,
         2471,
         2472,
         1046,
         1046,
         1074,
         1074,
         1074,
         1073,
         1046,
         1046,
         1073,
         1046,
         1046,
         1073,
         1046,
         1046,
         51,
         2913,
         51,
         1064,
         1064,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         1064,
         51,
         1047,
         1047,
         1047,
         1046,
         1064,
         51,
         1046,
         1064,
         51,
         1046,
         1064,
         51,
         51,
         2914,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         51,
         51,
         1087,
         1087,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         1087,
         1047,
         1047,
         1047,
         1046,
         51,
         1087,
         1046,
         51,
         1087,
         1046,
         51,
         1087,
         1046,
         2921,
         1046,
         1046,
         1046,
         1046,
         1046,
         1066,
         1066,
         1073,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1074,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1075,
         1046,
         1076,
         1046,
         1077,
         1046,
         1078,
         2466,
         1073,
         1046,
         2467,
         1046,
         1046,
         2468,
         2469,
         2470,
         2471,
         2472,
         1046,
         1046,
         1074,
         1074,
         1074,
         1073,
         1046,
         1046,
         1073,
         1046,
         1046,
         1073,
         1046,
         1046,
         51,
         2922,
         51,
         1064,
         1064,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         1064,
         51,
         1047,
         1047,
         1047,
         1046,
         1064,
         51,
         1046,
         1064,
         51,
         1046,
         1064,
         51,
         51,
         2923,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         51,
         51,
         1087,
         1087,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         1087,
         1047,
         1047,
         1047,
         1046,
         51,
         1087,
         1046,
         51,
         1087,
         1046,
         51,
         1087,
         1046,
         2930,
         1046,
         1046,
         1046,
         1046,
         1046,
         1066,
         1066,
         1073,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1074,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1046,
         1075,
         1046,
         1076,
         1046,
         1077,
         1046,
         1078,
         2466,
         1073,
         1046,
         2467,
         1046,
         1046,
         2468,
         2469,
         2470,
         2471,
         2472,
         1046,
         1046,
         1074,
         1074,
         1074,
         1073,
         1046,
         1046,
         1073,
         1046,
         1046,
         1073,
         1046,
         1046,
         51,
         2931,
         51,
         1064,
         1064,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         1064,
         51,
         1047,
         1047,
         1047,
         1046,
         1064,
         51,
         1046,
         1064,
         51,
         1046,
         1064,
         51,
         51,
         2932,
         51,
         51,
         51,
         51,
         51,
         1045,
         1045,
         1046,
         51,
         51,
         51,
         51,
         51,
         51,
         1087,
         1087,
         51,
         51,
         1047,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         51,
         1048,
         51,
         1049,
         51,
         1050,
         51,
         1051,
         2442,
         1046,
         51,
         2443,
         51,
         51,
         2444,
         2445,
         2446,
         2447,
         2448,
         51,
         1087,
         1047,
         1047,
         1047,
         1046,
         51,
         1087,
         1046,
         51,
         1087,
         1046,
         51,
         1087,
      },
   },
   [nir_op_ior] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         2,
         3,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         5,
         6,
         7,
         0,
         8,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         13,
         12,
         13,
         14,
         15,
         14,
         15,
         16,
         17,
         18,
         19,
         20,
         21,
         22,
         23,
         24,
         25,
         25,
         26,
         26,
         27,
         27,
         27,
         27,
         28,
         28,
         29,
         30,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         31,
         32,
         0,
         0,
         0,
         0,
         33,
         0,
         0,
         34,
         35,
         36,
         37,
         38,
         39,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         40,
         41,
         42,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         43,
         2,
         43,
         3,
         3,
         3,
         3,
         2,
         3,
         2,
         43,
         2,
         3,
         2,
         43,
         44,
         3,
         2,
         3,
         2,
         43,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         6,
         6,
         6,
         6,
         45,
         6,
         6,
         6,
         6,
         7,
         7,
         7,
         7,
         7,
         46,
         7,
         7,
         7,
         7,
         6,
         6,
         6,
         6,
         6,
         6,
         45,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         45,
         6,
         6,
         6,
         6,
         6,
         6,
         45,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         45,
         6,
         6,
         6,
         6,
         45,
         45,
         45,
         45,
         45,
         45,
         45,
         45,
         45,
         45,
         45,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         47,
         12,
         47,
         12,
         13,
         12,
         47,
         13,
         13,
         14,
         14,
         15,
         14,
         15,
         14,
         15,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         35,
         35,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         17,
         17,
         17,
         19,
         19,
         19,
         19,
         19,
         19,
         19,
         19,
         18,
         17,
         18,
         17,
         18,
         17,
         17,
         17,
         18,
         18,
         17,
         18,
         17,
         17,
         17,
         18,
         17,
         18,
         17,
         17,
         17,
         17,
         21,
         21,
         21,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         22,
         21,
         22,
         21,
         22,
         21,
         21,
         21,
         22,
         22,
         22,
         21,
         22,
         21,
         21,
         21,
         22,
         21,
         22,
         21,
         21,
         21,
         21,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         48,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         29,
         29,
         29,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         49,
         50,
         0,
         0,
         0,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         51,
         52,
         33,
         33,
         33,
         33,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         11,
         11,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         53,
         54,
         55,
         0,
         0,
         0,
         39,
         39,
         39,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         56,
         56,
         56,
         56,
         56,
         3,
         3,
         3,
         3,
         3,
         57,
         12,
         57,
         12,
         57,
         12,
         13,
         13,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         7,
         7,
         7,
         7,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         45,
         45,
         45,
         45,
         45,
         6,
         6,
         6,
         6,
         58,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         58,
         58,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         16,
         16,
         16,
         16,
         16,
         16,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         19,
         18,
         18,
         18,
         18,
         18,
         23,
         22,
         22,
         22,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         9,
         0,
         26,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         28,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         11,
         11,
         32,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         52,
         52,
         50,
         48,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         59,
         6,
         6,
         60,
         61,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         33,
         33,
         33,
         16,
         16,
         16,
         62,
         3,
         63,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         25,
         0,
         18,
         22,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         18,
         18,
         18,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         0,
         3,
         13,
         5,
         0,
         64,
         65,
         66,
         67,
         6,
         6,
         3,
         13,
         5,
         68,
         69,
         70,
         71,
         6,
         6,
         3,
         13,
         5,
         72,
         73,
         74,
         75,
         6,
         6,
      },
      
      .num_filtered_states = 76,
      .table = (const uint16_t []) {
      
         53,
         54,
         53,
         53,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         53,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         54,
         54,
         54,
         54,
         1132,
         1132,
         1133,
         1134,
         1135,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1136,
         1137,
         1137,
         1137,
         1137,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1138,
         54,
         54,
         1139,
         54,
         1140,
         54,
         1141,
         54,
         54,
         54,
         54,
         54,
         1133,
         1134,
         54,
         54,
         2380,
         2381,
         2382,
         2383,
         54,
         54,
         54,
         54,
         54,
         1133,
         1133,
         1132,
         1132,
         54,
         54,
         1132,
         1132,
         54,
         54,
         1132,
         1132,
         54,
         54,
         1132,
         1132,
         54,
         54,
         53,
         54,
         53,
         1142,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         53,
         1142,
         1124,
         1124,
         2384,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         53,
         2766,
         1124,
         1124,
         1123,
         1123,
         1142,
         53,
         1123,
         1123,
         1142,
         53,
         1123,
         1123,
         1142,
         53,
         1123,
         1123,
         1142,
         53,
         53,
         54,
         1142,
         1142,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         53,
         1143,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         1142,
         1142,
         1124,
         1124,
         2384,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         1142,
         2766,
         1124,
         1124,
         1123,
         1123,
         1142,
         1143,
         1123,
         1123,
         1142,
         1143,
         1123,
         1123,
         1142,
         1143,
         1123,
         1123,
         1142,
         1143,
         1123,
         1132,
         1123,
         1123,
         1144,
         1144,
         1145,
         1145,
         1146,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1147,
         1148,
         1148,
         1148,
         1148,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1149,
         1123,
         1123,
         1150,
         1123,
         1151,
         1123,
         1152,
         1123,
         1123,
         1123,
         1123,
         1123,
         1145,
         1145,
         1123,
         1123,
         2385,
         2386,
         2387,
         2388,
         1123,
         1123,
         1123,
         1123,
         1123,
         1145,
         1145,
         1144,
         1144,
         1123,
         1123,
         1144,
         1144,
         1123,
         1123,
         1144,
         1144,
         1123,
         1123,
         1144,
         1144,
         1123,
         1123,
         1123,
         1132,
         1123,
         1123,
         1144,
         1144,
         1145,
         1145,
         1146,
         1123,
         1123,
         1153,
         1123,
         1123,
         1123,
         1123,
         1147,
         1148,
         1148,
         1148,
         1148,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1149,
         1123,
         1123,
         1150,
         1123,
         1151,
         1123,
         1152,
         1123,
         1123,
         1123,
         1123,
         1123,
         1145,
         1145,
         1123,
         1123,
         2385,
         2386,
         2387,
         2388,
         1123,
         1123,
         1123,
         1123,
         1123,
         1145,
         1145,
         1144,
         1144,
         1123,
         1123,
         1144,
         1144,
         1123,
         1123,
         1144,
         1144,
         1123,
         1123,
         1144,
         1144,
         1123,
         1123,
         1124,
         1133,
         1124,
         1124,
         1145,
         1145,
         1124,
         1124,
         1154,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1155,
         1156,
         1156,
         1156,
         1156,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1157,
         1124,
         1124,
         1158,
         1124,
         1159,
         1124,
         1160,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         2389,
         2390,
         2391,
         2392,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1145,
         1145,
         1124,
         1124,
         1145,
         1145,
         1124,
         1124,
         1145,
         1145,
         1124,
         1124,
         1145,
         1145,
         1124,
         1124,
         1124,
         1134,
         1124,
         1124,
         1145,
         1145,
         1124,
         1124,
         1154,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1155,
         1156,
         1156,
         1156,
         1156,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1157,
         1124,
         1124,
         1158,
         1124,
         1159,
         1124,
         1160,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         2389,
         2390,
         2391,
         2392,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1145,
         1145,
         1124,
         1124,
         1145,
         1145,
         1124,
         1124,
         1145,
         1145,
         1124,
         1124,
         1145,
         1145,
         1124,
         1124,
         1125,
         1135,
         1125,
         1125,
         1146,
         1146,
         1154,
         1154,
         1161,
         1125,
         1125,
         1125,
         1125,
         1125,
         1125,
         1125,
         1162,
         1163,
         1163,
         1163,
         1163,
         1125,
         1125,
         1125,
         1125,
         1125,
         1125,
         1125,
         1125,
         1125,
         1164,
         1125,
         1165,
         1125,
         1125,
         1166,
         1125,
         1167,
         1125,
         1168,
         1125,
         1125,
         1125,
         1125,
         1125,
         1154,
         1154,
         1125,
         1125,
         2393,
         2394,
         2395,
         2396,
         1125,
         1125,
         1125,
         1125,
         1125,
         1154,
         1154,
         1146,
         1146,
         1125,
         1125,
         1146,
         1146,
         1125,
         1125,
         1146,
         1146,
         1125,
         1125,
         1146,
         1146,
         1125,
         1125,
         53,
         54,
         53,
         53,
         1123,
         1123,
         1124,
         1124,
         1125,
         1169,
         53,
         53,
         53,
         53,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         53,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         53,
         54,
         53,
         53,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         1170,
         53,
         53,
         53,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         1171,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         53,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         53,
         54,
         53,
         53,
         1123,
         1153,
         1124,
         1124,
         1125,
         53,
         53,
         1172,
         53,
         53,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         53,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         1153,
         1153,
         53,
         53,
         1153,
         1153,
         53,
         53,
         1153,
         1153,
         53,
         53,
         1153,
         1153,
         53,
         53,
         53,
         54,
         53,
         53,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         53,
         1173,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         2397,
         53,
         1124,
         1124,
         53,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         2767,
         53,
         1124,
         1124,
         1123,
         1123,
         53,
         1173,
         1123,
         1123,
         53,
         1173,
         1123,
         1123,
         53,
         1173,
         1123,
         1123,
         53,
         1173,
         53,
         54,
         53,
         1143,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         1173,
         1173,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         2397,
         1143,
         1124,
         1124,
         1173,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         2767,
         1173,
         1124,
         1124,
         1123,
         1123,
         1143,
         1173,
         1123,
         1123,
         1143,
         1173,
         1123,
         1123,
         1143,
         1173,
         1123,
         1123,
         1143,
         1173,
         53,
         54,
         53,
         53,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         53,
         53,
         53,
         1174,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         53,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         53,
         54,
         53,
         53,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         53,
         53,
         1174,
         1174,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         53,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1126,
         1136,
         1126,
         1126,
         1147,
         1147,
         1155,
         1155,
         1162,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1175,
         1176,
         1176,
         1176,
         1176,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1177,
         1126,
         1126,
         1178,
         1126,
         1179,
         1126,
         1180,
         1126,
         1126,
         1126,
         1126,
         1126,
         1155,
         1155,
         1126,
         1126,
         2398,
         2399,
         2400,
         2401,
         1126,
         1126,
         1126,
         1126,
         1126,
         1155,
         1155,
         1147,
         1147,
         1126,
         1126,
         1147,
         1147,
         1126,
         1126,
         1147,
         1147,
         1126,
         1126,
         1147,
         1147,
         1126,
         1126,
         1127,
         1137,
         1127,
         1127,
         1148,
         1148,
         1156,
         1156,
         1163,
         1127,
         1127,
         1127,
         1127,
         1127,
         1127,
         1127,
         1176,
         1181,
         1181,
         1181,
         1181,
         1127,
         1127,
         1127,
         1127,
         1127,
         1127,
         1127,
         1127,
         1127,
         1127,
         1127,
         1182,
         1127,
         1127,
         1183,
         1127,
         1184,
         1127,
         1185,
         1127,
         1127,
         1127,
         1127,
         1127,
         2402,
         2402,
         1127,
         1127,
         2403,
         2404,
         2405,
         2406,
         1127,
         1127,
         1127,
         1127,
         1127,
         1156,
         1156,
         1148,
         1148,
         1127,
         1127,
         1148,
         1148,
         1127,
         1127,
         1148,
         1148,
         1127,
         1127,
         1148,
         1148,
         1127,
         1127,
         1127,
         1137,
         1127,
         1127,
         1148,
         1148,
         1156,
         1156,
         1163,
         1127,
         1127,
         1127,
         1127,
         1127,
         1127,
         1127,
         1176,
         1181,
         1186,
         1187,
         1188,
         1127,
         1127,
         1127,
         1127,
         1127,
         1127,
         1127,
         1127,
         1127,
         1127,
         1127,
         1182,
         1127,
         1127,
         1183,
         1127,
         1184,
         1127,
         1185,
         1127,
         1127,
         1127,
         1127,
         1127,
         2402,
         2402,
         1127,
         1127,
         2403,
         2404,
         2405,
         2406,
         1127,
         1127,
         1127,
         1127,
         1127,
         1156,
         1156,
         1148,
         1148,
         1127,
         1127,
         1148,
         1148,
         1127,
         1127,
         1148,
         1148,
         1127,
         1127,
         1148,
         1148,
         1127,
         1127,
         1127,
         1137,
         1127,
         1127,
         1148,
         1148,
         1156,
         1156,
         1163,
         1127,
         1127,
         1127,
         1127,
         1127,
         1127,
         1127,
         1176,
         1181,
         1187,
         1189,
         1190,
         1127,
         1127,
         1127,
         1127,
         1127,
         1127,
         1127,
         1127,
         1127,
         1127,
         1127,
         1182,
         1127,
         1127,
         1183,
         1127,
         1184,
         1127,
         1185,
         1127,
         1127,
         1127,
         1127,
         1127,
         2402,
         2402,
         1127,
         1127,
         2403,
         2404,
         2405,
         2406,
         1127,
         1127,
         1127,
         1127,
         1127,
         1156,
         1156,
         1148,
         1148,
         1127,
         1127,
         1148,
         1148,
         1127,
         1127,
         1148,
         1148,
         1127,
         1127,
         1148,
         1148,
         1127,
         1127,
         1127,
         1137,
         1127,
         1127,
         1148,
         1148,
         1156,
         1156,
         1163,
         1127,
         1127,
         1127,
         1127,
         1127,
         1127,
         1127,
         1176,
         1181,
         1188,
         1190,
         1191,
         1127,
         1127,
         1127,
         1127,
         1127,
         1127,
         1127,
         1127,
         1127,
         1127,
         1127,
         1182,
         1127,
         1127,
         1183,
         1127,
         1184,
         1127,
         1185,
         1127,
         1127,
         1127,
         1127,
         1127,
         2402,
         2402,
         1127,
         1127,
         2403,
         2404,
         2405,
         2406,
         1127,
         1127,
         1127,
         1127,
         1127,
         1156,
         1156,
         1148,
         1148,
         1127,
         1127,
         1148,
         1148,
         1127,
         1127,
         1148,
         1148,
         1127,
         1127,
         1148,
         1148,
         1127,
         1127,
         53,
         54,
         53,
         53,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         1192,
         1192,
         1192,
         1192,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         53,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         53,
         54,
         53,
         53,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         1192,
         1193,
         1194,
         1195,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         53,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         53,
         54,
         53,
         53,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         1192,
         1194,
         1196,
         1197,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         53,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         53,
         54,
         53,
         53,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         1192,
         1195,
         1197,
         1198,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         53,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         53,
         54,
         53,
         53,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         1171,
         53,
         53,
         53,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         1199,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         53,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         53,
         54,
         53,
         53,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         1200,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         53,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         53,
         54,
         53,
         53,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         1201,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         53,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         53,
         54,
         53,
         53,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1202,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         53,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         53,
         54,
         53,
         53,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1203,
         1203,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         53,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         53,
         54,
         53,
         53,
         1123,
         1123,
         1124,
         1124,
         1164,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1203,
         1203,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         53,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         53,
         54,
         53,
         53,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1204,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         53,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1128,
         1138,
         1128,
         1128,
         1149,
         1149,
         1157,
         1157,
         1165,
         1128,
         1128,
         1128,
         1128,
         1128,
         1128,
         1128,
         1177,
         1182,
         1182,
         1182,
         1182,
         1128,
         1128,
         1128,
         1128,
         1128,
         1128,
         1128,
         1128,
         1128,
         1128,
         1204,
         1204,
         1128,
         1128,
         1205,
         1128,
         1206,
         1128,
         1207,
         1128,
         1128,
         1128,
         1128,
         1128,
         1157,
         1157,
         1128,
         1128,
         2407,
         2408,
         2409,
         2410,
         1128,
         1128,
         1128,
         1128,
         1128,
         1157,
         1157,
         1149,
         1149,
         1128,
         1128,
         1149,
         1149,
         1128,
         1128,
         1149,
         1149,
         1128,
         1128,
         1149,
         1149,
         1128,
         1128,
         53,
         54,
         53,
         53,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         53,
         2411,
         1124,
         1124,
         53,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         53,
         54,
         53,
         53,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1208,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         53,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1129,
         1139,
         1129,
         1129,
         1150,
         1150,
         1158,
         1158,
         1166,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1178,
         1183,
         1183,
         1183,
         1183,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1205,
         1129,
         1208,
         1208,
         1129,
         1209,
         1129,
         1210,
         1129,
         1129,
         1129,
         1129,
         1129,
         1158,
         1158,
         1129,
         1129,
         2412,
         2413,
         2414,
         2415,
         1129,
         1129,
         1129,
         1129,
         1129,
         1158,
         1158,
         1150,
         1150,
         1129,
         1129,
         1150,
         1150,
         1129,
         1129,
         1150,
         1150,
         1129,
         1129,
         1150,
         1150,
         1129,
         1129,
         53,
         54,
         53,
         53,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1211,
         53,
         1131,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         53,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1130,
         1140,
         1130,
         1130,
         1151,
         1151,
         1159,
         1159,
         1167,
         1130,
         1130,
         1130,
         1130,
         1130,
         1130,
         1130,
         1179,
         1184,
         1184,
         1184,
         1184,
         1130,
         1130,
         1130,
         1130,
         1130,
         1130,
         1130,
         1130,
         1130,
         1130,
         1130,
         1206,
         1130,
         1130,
         1209,
         1211,
         1211,
         1130,
         1212,
         1130,
         1130,
         1130,
         1130,
         1130,
         1159,
         1159,
         1130,
         1130,
         2416,
         2417,
         2418,
         2419,
         1130,
         1130,
         1130,
         1130,
         1130,
         1159,
         1159,
         1151,
         1151,
         1130,
         1130,
         1151,
         1151,
         1130,
         1130,
         1151,
         1151,
         1130,
         1130,
         1151,
         1151,
         1130,
         1130,
         53,
         54,
         53,
         53,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1213,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         53,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1131,
         1141,
         1131,
         1131,
         1152,
         1152,
         1160,
         1160,
         1168,
         1131,
         1131,
         1131,
         1131,
         1131,
         1131,
         1131,
         1180,
         1185,
         1185,
         1185,
         1185,
         1131,
         1131,
         1131,
         1131,
         1131,
         1131,
         1131,
         1131,
         1131,
         1131,
         1131,
         1207,
         1131,
         1131,
         1210,
         1131,
         1212,
         1213,
         1213,
         1131,
         1131,
         1131,
         1131,
         1131,
         1160,
         1160,
         1131,
         1131,
         2420,
         2421,
         2422,
         2423,
         1131,
         1131,
         1131,
         1131,
         1131,
         1160,
         1160,
         1152,
         1152,
         1131,
         1131,
         1152,
         1152,
         1131,
         1131,
         1152,
         1152,
         1131,
         1131,
         1152,
         1152,
         1131,
         1131,
         53,
         54,
         53,
         53,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         1214,
         1214,
         53,
         53,
         1124,
         1124,
         53,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         53,
         54,
         53,
         53,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         1214,
         53,
         1214,
         53,
         53,
         1124,
         1124,
         53,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         53,
         54,
         53,
         53,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         1214,
         1214,
         1214,
         53,
         53,
         1124,
         1124,
         53,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         53,
         54,
         53,
         1142,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         2397,
         2397,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         53,
         1142,
         1124,
         1124,
         2424,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         53,
         2768,
         1124,
         1124,
         1123,
         1123,
         1142,
         2397,
         1123,
         1123,
         1142,
         2397,
         1123,
         1123,
         1142,
         2397,
         1123,
         1123,
         1142,
         2397,
         53,
         54,
         1142,
         1142,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         53,
         1143,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         2411,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         1142,
         1142,
         1124,
         1124,
         2384,
         53,
         2376,
         2377,
         2425,
         2426,
         53,
         53,
         53,
         1142,
         2766,
         2769,
         2887,
         1123,
         1123,
         1142,
         1143,
         1123,
         1123,
         1142,
         1143,
         1123,
         1123,
         1142,
         1143,
         1123,
         1123,
         1142,
         1143,
         1124,
         1133,
         1124,
         1124,
         1145,
         1145,
         1124,
         1124,
         1154,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1155,
         2402,
         2402,
         2402,
         2402,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1157,
         1124,
         1124,
         1158,
         1124,
         1159,
         1124,
         1160,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         2389,
         2390,
         2391,
         2392,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1145,
         1145,
         1124,
         1124,
         1145,
         1145,
         1124,
         1124,
         1145,
         1145,
         1124,
         1124,
         1145,
         1145,
         1124,
         1124,
         1124,
         1134,
         1124,
         1124,
         1145,
         1145,
         1124,
         1124,
         1154,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1155,
         2402,
         2402,
         2402,
         2402,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1157,
         1124,
         1124,
         1158,
         1124,
         1159,
         1124,
         1160,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         2389,
         2390,
         2391,
         2392,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1145,
         1145,
         1124,
         1124,
         1145,
         1145,
         1124,
         1124,
         1145,
         1145,
         1124,
         1124,
         1145,
         1145,
         1124,
         1124,
         53,
         54,
         2384,
         2384,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         53,
         1173,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         2424,
         2384,
         1124,
         1124,
         53,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         2770,
         53,
         1124,
         1124,
         1123,
         1123,
         2384,
         1173,
         1123,
         1123,
         2384,
         1173,
         1123,
         1123,
         2384,
         1173,
         1123,
         1123,
         2384,
         1173,
         53,
         54,
         53,
         53,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         53,
         2427,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         2376,
         2380,
         2376,
         2376,
         2385,
         2385,
         2389,
         2389,
         2393,
         2376,
         2376,
         2376,
         2376,
         2376,
         2376,
         2376,
         2398,
         2403,
         2403,
         2403,
         2403,
         2376,
         2376,
         2376,
         2376,
         2376,
         2376,
         2376,
         2376,
         2376,
         2376,
         2376,
         2407,
         2376,
         2376,
         2412,
         2376,
         2416,
         2376,
         2420,
         2376,
         2376,
         2376,
         2376,
         2376,
         2389,
         2389,
         2376,
         2376,
         2376,
         2428,
         2429,
         2430,
         2376,
         2376,
         2376,
         2376,
         2376,
         2389,
         2389,
         2385,
         2385,
         2376,
         2376,
         2385,
         2385,
         2376,
         2376,
         2385,
         2385,
         2376,
         2376,
         2385,
         2385,
         2376,
         2376,
         2377,
         2381,
         2377,
         2377,
         2386,
         2386,
         2390,
         2390,
         2394,
         2377,
         2377,
         2377,
         2377,
         2377,
         2377,
         2377,
         2399,
         2404,
         2404,
         2404,
         2404,
         2377,
         2377,
         2377,
         2377,
         2377,
         2377,
         2377,
         2377,
         2377,
         2377,
         2377,
         2408,
         2377,
         2377,
         2413,
         2377,
         2417,
         2377,
         2421,
         2377,
         2377,
         2377,
         2377,
         2377,
         2390,
         2390,
         2377,
         2377,
         2428,
         2377,
         2431,
         2432,
         2377,
         2377,
         2377,
         2377,
         2377,
         2390,
         2390,
         2386,
         2386,
         2377,
         2377,
         2386,
         2386,
         2377,
         2377,
         2386,
         2386,
         2377,
         2377,
         2386,
         2386,
         2377,
         2377,
         2378,
         2382,
         2378,
         2378,
         2387,
         2387,
         2391,
         2391,
         2395,
         2378,
         2378,
         2378,
         2378,
         2378,
         2378,
         2378,
         2400,
         2405,
         2405,
         2405,
         2405,
         2378,
         2378,
         2378,
         2378,
         2378,
         2378,
         2378,
         2378,
         2378,
         2378,
         2378,
         2409,
         2378,
         2378,
         2414,
         2378,
         2418,
         2378,
         2422,
         2378,
         2378,
         2378,
         2378,
         2425,
         2391,
         2391,
         2378,
         2378,
         2429,
         2431,
         2378,
         2433,
         2378,
         2378,
         2378,
         2378,
         2378,
         2391,
         2391,
         2387,
         2387,
         2378,
         2378,
         2387,
         2387,
         2378,
         2378,
         2387,
         2387,
         2378,
         2378,
         2387,
         2387,
         2378,
         2378,
         2379,
         2383,
         2379,
         2379,
         2388,
         2388,
         2392,
         2392,
         2396,
         2379,
         2379,
         2379,
         2379,
         2379,
         2379,
         2379,
         2401,
         2406,
         2406,
         2406,
         2406,
         2379,
         2379,
         2379,
         2379,
         2379,
         2379,
         2379,
         2379,
         2379,
         2379,
         2379,
         2410,
         2379,
         2379,
         2415,
         2379,
         2419,
         2379,
         2423,
         2379,
         2379,
         2379,
         2379,
         2426,
         2392,
         2392,
         2379,
         2379,
         2430,
         2432,
         2433,
         2434,
         2379,
         2379,
         2379,
         2379,
         2379,
         2392,
         2392,
         2388,
         2388,
         2379,
         2379,
         2388,
         2388,
         2379,
         2379,
         2388,
         2388,
         2379,
         2379,
         2388,
         2388,
         2379,
         2379,
         53,
         54,
         53,
         53,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         53,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         2435,
         2435,
         53,
         53,
         1124,
         1124,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         53,
         54,
         53,
         53,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         53,
         53,
         2376,
         2377,
         2378,
         2379,
         2435,
         53,
         2435,
         53,
         53,
         1124,
         1124,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         53,
         54,
         53,
         53,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         53,
         53,
         1124,
         1124,
         53,
         53,
         2376,
         2377,
         2378,
         2379,
         2435,
         2435,
         2435,
         53,
         53,
         1124,
         1124,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         1123,
         1123,
         53,
         53,
         53,
         54,
         53,
         1142,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         2767,
         2767,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         53,
         1142,
         1124,
         1124,
         2770,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         53,
         2771,
         1124,
         1124,
         1123,
         1123,
         1142,
         2767,
         1123,
         1123,
         1142,
         2767,
         1123,
         1123,
         1142,
         2767,
         1123,
         1123,
         1142,
         2767,
         53,
         54,
         2766,
         2766,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         53,
         1173,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         2768,
         2766,
         1124,
         1124,
         53,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         2771,
         53,
         1124,
         1124,
         1123,
         1123,
         2766,
         1173,
         1123,
         1123,
         2766,
         1173,
         1123,
         1123,
         2766,
         1173,
         1123,
         1123,
         2766,
         1173,
         1124,
         1133,
         1124,
         1124,
         1145,
         1145,
         1124,
         1124,
         1154,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1155,
         1156,
         1156,
         1156,
         1156,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1157,
         1124,
         1124,
         1158,
         1124,
         1159,
         1124,
         1160,
         1124,
         1124,
         1124,
         1124,
         2769,
         1124,
         1124,
         1124,
         1124,
         2389,
         2390,
         2391,
         2392,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1145,
         1145,
         1124,
         1124,
         1145,
         1145,
         1124,
         1124,
         1145,
         1145,
         1124,
         1124,
         1145,
         1145,
         1124,
         1124,
         1124,
         1133,
         1124,
         1124,
         1145,
         1145,
         1124,
         1124,
         1154,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1155,
         1156,
         1156,
         1156,
         1156,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1157,
         1124,
         1124,
         1158,
         1124,
         1159,
         1124,
         1160,
         1124,
         1124,
         1124,
         1124,
         2887,
         1124,
         1124,
         1124,
         1124,
         2389,
         2390,
         2391,
         2392,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1124,
         1145,
         1145,
         1124,
         1124,
         1145,
         1145,
         1124,
         1124,
         1145,
         1145,
         1124,
         1124,
         1145,
         1145,
         1124,
         1124,
         1123,
         1132,
         1123,
         1123,
         1144,
         1144,
         1145,
         1145,
         1146,
         1123,
         1123,
         1153,
         1123,
         1123,
         1123,
         1123,
         1147,
         1148,
         1148,
         1148,
         1148,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1149,
         1123,
         1123,
         1150,
         1123,
         1151,
         1123,
         1152,
         1123,
         1123,
         1123,
         1123,
         1123,
         1145,
         1145,
         1123,
         1123,
         2385,
         2386,
         2387,
         2388,
         1123,
         1123,
         1123,
         1123,
         1123,
         1145,
         1145,
         1144,
         2888,
         1123,
         1123,
         1144,
         1144,
         1123,
         1123,
         1144,
         1144,
         1123,
         1123,
         1144,
         1144,
         1123,
         1123,
         1123,
         1132,
         1123,
         1123,
         1144,
         1144,
         1145,
         1145,
         1146,
         1123,
         1123,
         1153,
         1123,
         1123,
         1123,
         1123,
         1147,
         1148,
         1148,
         1148,
         1148,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1149,
         1123,
         1123,
         1150,
         1123,
         1151,
         1123,
         1152,
         1123,
         1123,
         1123,
         1123,
         1123,
         1145,
         1145,
         1123,
         1123,
         2385,
         2386,
         2387,
         2388,
         1123,
         1123,
         1123,
         1123,
         1123,
         1145,
         1145,
         2888,
         1144,
         1123,
         1123,
         1144,
         1144,
         1123,
         1123,
         1144,
         1144,
         1123,
         1123,
         1144,
         1144,
         1123,
         1123,
         53,
         54,
         1142,
         1142,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         53,
         1143,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         1142,
         1142,
         1124,
         1124,
         2384,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         1142,
         2766,
         1124,
         1124,
         1123,
         1123,
         1142,
         2889,
         1123,
         1123,
         1142,
         2889,
         1123,
         1123,
         1142,
         2889,
         1123,
         1123,
         1142,
         2889,
         53,
         54,
         53,
         1143,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         1173,
         1173,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         2397,
         1143,
         1124,
         1124,
         1173,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         2767,
         1173,
         1124,
         1124,
         1123,
         1123,
         2889,
         1173,
         1123,
         1123,
         2889,
         1173,
         1123,
         1123,
         2889,
         1173,
         1123,
         1123,
         2889,
         1173,
         1123,
         1132,
         1123,
         1123,
         1144,
         1144,
         1145,
         1145,
         1146,
         1123,
         1123,
         1153,
         1123,
         1123,
         1123,
         1123,
         1147,
         1148,
         1148,
         1148,
         1148,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1149,
         1123,
         1123,
         1150,
         1123,
         1151,
         1123,
         1152,
         1123,
         1123,
         1123,
         1123,
         1123,
         1145,
         1145,
         1123,
         1123,
         2385,
         2386,
         2387,
         2388,
         1123,
         1123,
         1123,
         1123,
         1123,
         1145,
         1145,
         1144,
         1144,
         1123,
         1123,
         1144,
         2917,
         1123,
         1123,
         1144,
         1144,
         1123,
         1123,
         1144,
         1144,
         1123,
         1123,
         1123,
         1132,
         1123,
         1123,
         1144,
         1144,
         1145,
         1145,
         1146,
         1123,
         1123,
         1153,
         1123,
         1123,
         1123,
         1123,
         1147,
         1148,
         1148,
         1148,
         1148,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1149,
         1123,
         1123,
         1150,
         1123,
         1151,
         1123,
         1152,
         1123,
         1123,
         1123,
         1123,
         1123,
         1145,
         1145,
         1123,
         1123,
         2385,
         2386,
         2387,
         2388,
         1123,
         1123,
         1123,
         1123,
         1123,
         1145,
         1145,
         1144,
         1144,
         1123,
         1123,
         2917,
         1144,
         1123,
         1123,
         1144,
         1144,
         1123,
         1123,
         1144,
         1144,
         1123,
         1123,
         53,
         54,
         1142,
         1142,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         53,
         1143,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         1142,
         1142,
         1124,
         1124,
         2384,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         1142,
         2766,
         1124,
         1124,
         1123,
         1123,
         1142,
         2889,
         1123,
         1123,
         1142,
         2918,
         1123,
         1123,
         1142,
         2918,
         1123,
         1123,
         1142,
         2918,
         53,
         54,
         53,
         1143,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         1173,
         1173,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         2397,
         1143,
         1124,
         1124,
         1173,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         2767,
         1173,
         1124,
         1124,
         1123,
         1123,
         2889,
         1173,
         1123,
         1123,
         2918,
         1173,
         1123,
         1123,
         2918,
         1173,
         1123,
         1123,
         2918,
         1173,
         1123,
         1132,
         1123,
         1123,
         1144,
         1144,
         1145,
         1145,
         1146,
         1123,
         1123,
         1153,
         1123,
         1123,
         1123,
         1123,
         1147,
         1148,
         1148,
         1148,
         1148,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1149,
         1123,
         1123,
         1150,
         1123,
         1151,
         1123,
         1152,
         1123,
         1123,
         1123,
         1123,
         1123,
         1145,
         1145,
         1123,
         1123,
         2385,
         2386,
         2387,
         2388,
         1123,
         1123,
         1123,
         1123,
         1123,
         1145,
         1145,
         1144,
         1144,
         1123,
         1123,
         1144,
         1144,
         1123,
         1123,
         1144,
         2926,
         1123,
         1123,
         1144,
         1144,
         1123,
         1123,
         1123,
         1132,
         1123,
         1123,
         1144,
         1144,
         1145,
         1145,
         1146,
         1123,
         1123,
         1153,
         1123,
         1123,
         1123,
         1123,
         1147,
         1148,
         1148,
         1148,
         1148,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1149,
         1123,
         1123,
         1150,
         1123,
         1151,
         1123,
         1152,
         1123,
         1123,
         1123,
         1123,
         1123,
         1145,
         1145,
         1123,
         1123,
         2385,
         2386,
         2387,
         2388,
         1123,
         1123,
         1123,
         1123,
         1123,
         1145,
         1145,
         1144,
         1144,
         1123,
         1123,
         1144,
         1144,
         1123,
         1123,
         2926,
         1144,
         1123,
         1123,
         1144,
         1144,
         1123,
         1123,
         53,
         54,
         1142,
         1142,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         53,
         1143,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         1142,
         1142,
         1124,
         1124,
         2384,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         1142,
         2766,
         1124,
         1124,
         1123,
         1123,
         1142,
         2889,
         1123,
         1123,
         1142,
         2918,
         1123,
         1123,
         1142,
         2927,
         1123,
         1123,
         1142,
         2927,
         53,
         54,
         53,
         1143,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         1173,
         1173,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         2397,
         1143,
         1124,
         1124,
         1173,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         2767,
         1173,
         1124,
         1124,
         1123,
         1123,
         2889,
         1173,
         1123,
         1123,
         2918,
         1173,
         1123,
         1123,
         2927,
         1173,
         1123,
         1123,
         2927,
         1173,
         1123,
         1132,
         1123,
         1123,
         1144,
         1144,
         1145,
         1145,
         1146,
         1123,
         1123,
         1153,
         1123,
         1123,
         1123,
         1123,
         1147,
         1148,
         1148,
         1148,
         1148,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1149,
         1123,
         1123,
         1150,
         1123,
         1151,
         1123,
         1152,
         1123,
         1123,
         1123,
         1123,
         1123,
         1145,
         1145,
         1123,
         1123,
         2385,
         2386,
         2387,
         2388,
         1123,
         1123,
         1123,
         1123,
         1123,
         1145,
         1145,
         1144,
         1144,
         1123,
         1123,
         1144,
         1144,
         1123,
         1123,
         1144,
         1144,
         1123,
         1123,
         1144,
         2935,
         1123,
         1123,
         1123,
         1132,
         1123,
         1123,
         1144,
         1144,
         1145,
         1145,
         1146,
         1123,
         1123,
         1153,
         1123,
         1123,
         1123,
         1123,
         1147,
         1148,
         1148,
         1148,
         1148,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1123,
         1149,
         1123,
         1123,
         1150,
         1123,
         1151,
         1123,
         1152,
         1123,
         1123,
         1123,
         1123,
         1123,
         1145,
         1145,
         1123,
         1123,
         2385,
         2386,
         2387,
         2388,
         1123,
         1123,
         1123,
         1123,
         1123,
         1145,
         1145,
         1144,
         1144,
         1123,
         1123,
         1144,
         1144,
         1123,
         1123,
         1144,
         1144,
         1123,
         1123,
         2935,
         1144,
         1123,
         1123,
         53,
         54,
         1142,
         1142,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         53,
         1143,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         1142,
         1142,
         1124,
         1124,
         2384,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         1142,
         2766,
         1124,
         1124,
         1123,
         1123,
         1142,
         2889,
         1123,
         1123,
         1142,
         2918,
         1123,
         1123,
         1142,
         2927,
         1123,
         1123,
         1142,
         2936,
         53,
         54,
         53,
         1143,
         1123,
         1123,
         1124,
         1124,
         1125,
         53,
         53,
         53,
         1173,
         1173,
         53,
         53,
         1126,
         1127,
         1127,
         1127,
         1127,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1128,
         53,
         53,
         1129,
         53,
         1130,
         53,
         1131,
         53,
         53,
         53,
         2397,
         1143,
         1124,
         1124,
         1173,
         53,
         2376,
         2377,
         2378,
         2379,
         53,
         53,
         53,
         2767,
         1173,
         1124,
         1124,
         1123,
         1123,
         2889,
         1173,
         1123,
         1123,
         2918,
         1173,
         1123,
         1123,
         2927,
         1173,
         1123,
         1123,
         2936,
         1173,
      },
   },
   [nir_op_ieq] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         3,
         0,
         0,
         0,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         8,
         0,
         9,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         14,
         0,
         15,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         17,
         18,
         19,
         20,
         0,
         0,
         0,
         21,
         22,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         8,
         8,
         0,
         0,
         9,
         0,
         9,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         23,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         0,
         0,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         20,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         11,
         23,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         5,
         0,
         5,
         5,
         0,
         8,
         0,
         0,
         0,
         8,
         5,
         5,
         5,
         0,
         8,
         0,
         0,
         0,
         8,
         5,
         5,
         5,
         0,
         8,
         0,
         0,
      },
      
      .num_filtered_states = 24,
      .table = (const uint16_t []) {
      
         55,
         56,
         55,
         55,
         1568,
         1569,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         56,
         56,
         1570,
         1571,
         1572,
         1573,
         1574,
         1575,
         1576,
         1577,
         1578,
         1579,
         1580,
         1581,
         1582,
         1583,
         1584,
         1585,
         1586,
         1587,
         1588,
         1589,
         1590,
         2649,
         55,
         1570,
         55,
         55,
         1568,
         1569,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         1571,
         55,
         55,
         1568,
         1569,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         1568,
         1572,
         1568,
         1568,
         1568,
         1591,
         1568,
         1568,
         1568,
         1568,
         1568,
         1568,
         1568,
         1568,
         1568,
         1568,
         1568,
         1568,
         1568,
         1568,
         1568,
         1568,
         1568,
         1568,
         1569,
         1573,
         1569,
         1569,
         1591,
         1569,
         1569,
         1569,
         1569,
         1569,
         1569,
         1569,
         1569,
         1569,
         1569,
         1569,
         1569,
         1569,
         1569,
         1569,
         1569,
         1569,
         1569,
         1569,
         55,
         1574,
         55,
         55,
         1568,
         1569,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         1575,
         55,
         55,
         1568,
         1569,
         55,
         1592,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         1576,
         55,
         55,
         1568,
         1569,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         1577,
         55,
         55,
         1568,
         1569,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         1578,
         55,
         55,
         1568,
         1569,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         1579,
         55,
         55,
         1568,
         1569,
         55,
         55,
         55,
         55,
         55,
         1593,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         1580,
         55,
         55,
         1568,
         1569,
         55,
         55,
         55,
         55,
         55,
         55,
         1594,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         1581,
         55,
         55,
         1568,
         1569,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         1582,
         55,
         55,
         1568,
         1569,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         1583,
         55,
         55,
         1568,
         1569,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         1584,
         55,
         55,
         1568,
         1569,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         1585,
         55,
         55,
         1568,
         1569,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         1586,
         55,
         55,
         1568,
         1569,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         1587,
         55,
         55,
         1568,
         1569,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         1588,
         55,
         55,
         1568,
         1569,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         1589,
         55,
         55,
         1568,
         1569,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         1590,
         55,
         55,
         1568,
         1569,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         2649,
         55,
         55,
         1568,
         1569,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
      },
   },
   [nir_op_ine] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         3,
         0,
         0,
         0,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         8,
         0,
         9,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         13,
         14,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         17,
         18,
         19,
         0,
         0,
         0,
         20,
         21,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         8,
         8,
         0,
         0,
         9,
         0,
         9,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         22,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         19,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         23,
         0,
         23,
         0,
         23,
         0,
         23,
         23,
         0,
         23,
         23,
         24,
         0,
         0,
         24,
         0,
         0,
         24,
         0,
         24,
         24,
         24,
         0,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         11,
         10,
         22,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         23,
         23,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         11,
         11,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         5,
         0,
         5,
         5,
         0,
         8,
         0,
         0,
         0,
         8,
         5,
         5,
         5,
         0,
         8,
         0,
         0,
         0,
         8,
         5,
         5,
         5,
         0,
         8,
         0,
         0,
      },
      
      .num_filtered_states = 25,
      .table = (const uint16_t []) {
      
         57,
         58,
         57,
         57,
         1595,
         1596,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         58,
         58,
         1597,
         1598,
         1599,
         1600,
         1601,
         1602,
         1603,
         1604,
         1605,
         1606,
         1607,
         1608,
         1609,
         1610,
         1611,
         1612,
         1613,
         1614,
         1615,
         1616,
         2650,
         2877,
         2878,
         57,
         1597,
         57,
         57,
         1595,
         1596,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         1598,
         57,
         57,
         1595,
         1596,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         1595,
         1599,
         1595,
         1595,
         1595,
         1617,
         1595,
         1595,
         1595,
         1595,
         1595,
         1595,
         1595,
         1595,
         1595,
         1595,
         1595,
         1595,
         1595,
         1595,
         1595,
         1595,
         1595,
         1595,
         1595,
         1596,
         1600,
         1596,
         1596,
         1617,
         1596,
         1596,
         1596,
         1596,
         1596,
         1596,
         1596,
         1596,
         1596,
         1596,
         1596,
         1596,
         1596,
         1596,
         1596,
         1596,
         1596,
         1596,
         1596,
         1596,
         57,
         1601,
         57,
         57,
         1595,
         1596,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         1602,
         57,
         57,
         1595,
         1596,
         57,
         1618,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         1603,
         57,
         57,
         1595,
         1596,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         1604,
         57,
         57,
         1595,
         1596,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         1605,
         57,
         57,
         1595,
         1596,
         57,
         57,
         57,
         57,
         1619,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         1606,
         57,
         57,
         1595,
         1596,
         57,
         57,
         57,
         57,
         57,
         1620,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         1607,
         57,
         57,
         1595,
         1596,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         1608,
         57,
         57,
         1595,
         1596,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         1609,
         57,
         57,
         1595,
         1596,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         1610,
         57,
         57,
         1595,
         1596,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         1611,
         57,
         57,
         1595,
         1596,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         1612,
         57,
         57,
         1595,
         1596,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         1613,
         57,
         57,
         1595,
         1596,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         1614,
         57,
         57,
         1595,
         1596,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         1615,
         57,
         57,
         1595,
         1596,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         1616,
         57,
         57,
         1595,
         1596,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         2650,
         57,
         57,
         1595,
         1596,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         2651,
         57,
         57,
         57,
         2877,
         57,
         57,
         1595,
         1596,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         2878,
         57,
         57,
         1595,
         1596,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
         57,
      },
   },
   [nir_op_fsat] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         5,
         5,
         6,
         7,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         9,
         9,
         9,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 11,
      .table = (const uint16_t []) {
      
         59,
         1418,
         1419,
         1420,
         1421,
         2591,
         2592,
         2593,
         2594,
         2595,
         2596,
      },
   },
   [nir_op_umul_unorm_4x8_vc4] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         60,
         60,
         60,
      },
   },
   [nir_op_fsign] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         61,
      },
   },
   [nir_op_flrp] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 5,
      .table = (const uint16_t []) {
      
         62,
         63,
         62,
         62,
         1689,
         64,
         65,
         64,
         64,
         1690,
         1691,
         1692,
         1691,
         1691,
         1693,
         1694,
         1695,
         1694,
         1694,
         1696,
         62,
         63,
         62,
         62,
         1689,
         66,
         67,
         66,
         66,
         1697,
         68,
         69,
         68,
         68,
         1698,
         1699,
         1700,
         1699,
         1699,
         1701,
         1702,
         1703,
         1702,
         1702,
         1704,
         66,
         67,
         66,
         66,
         1697,
         62,
         63,
         62,
         62,
         1689,
         64,
         65,
         64,
         64,
         1690,
         1705,
         1706,
         1705,
         1705,
         1707,
         1694,
         1695,
         1694,
         1694,
         1696,
         62,
         63,
         62,
         62,
         1689,
         62,
         63,
         62,
         62,
         1689,
         64,
         65,
         64,
         64,
         1690,
         1691,
         1692,
         1691,
         1691,
         1693,
         1708,
         1709,
         1708,
         1708,
         1710,
         62,
         63,
         62,
         62,
         1689,
         62,
         63,
         62,
         62,
         1689,
         64,
         65,
         64,
         64,
         1690,
         1691,
         1692,
         1691,
         1691,
         1693,
         1694,
         1695,
         1694,
         1694,
         1696,
         62,
         63,
         62,
         62,
         1689,
      },
   },
   [nir_op_sdot_4x8_iadd] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         70,
         71,
         72,
         73,
         72,
         73,
         74,
         75,
      },
   },
   [nir_op_udot_4x8_uadd] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         76,
         77,
         78,
         79,
         78,
         79,
         80,
         81,
      },
   },
   [nir_op_sdot_4x8_iadd_sat] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         82,
         82,
         83,
         83,
         83,
         83,
         84,
         84,
      },
   },
   [nir_op_udot_4x8_uadd_sat] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         85,
         86,
         87,
         88,
         87,
         88,
         89,
         90,
      },
   },
   [nir_op_sdot_2x16_iadd] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         91,
         92,
         93,
         94,
         93,
         94,
         95,
         96,
      },
   },
   [nir_op_udot_2x16_uadd] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         97,
         98,
         99,
         100,
         99,
         100,
         101,
         102,
      },
   },
   [nir_op_sdot_2x16_iadd_sat] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         103,
         103,
         104,
         104,
         104,
         104,
         105,
         105,
      },
   },
   [nir_op_udot_2x16_uadd_sat] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         106,
         106,
         107,
         107,
         107,
         107,
         108,
         108,
      },
   },
   [nir_op_sudot_4x8_iadd] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         109,
         110,
         111,
         112,
         113,
         114,
         115,
         116,
      },
   },
   [nir_op_sudot_4x8_iadd_sat] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         117,
         117,
         118,
         118,
         119,
         119,
         120,
         120,
      },
   },
   [nir_op_bcsel] = {
      .filter = (const uint16_t []) {
         0,
         1,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         9,
         9,
         0,
         10,
         0,
         11,
         0,
         0,
         12,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         0,
         15,
         16,
         17,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         18,
         18,
         18,
         18,
         19,
         20,
         21,
         20,
         21,
         22,
         22,
         22,
         22,
         0,
         0,
         23,
         24,
         25,
         25,
         25,
         25,
         26,
         26,
         27,
         27,
         28,
         28,
         0,
         29,
         29,
         30,
         30,
         31,
         31,
         32,
         32,
         33,
         33,
         0,
         34,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         35,
         0,
         36,
         0,
         37,
         38,
         38,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         39,
         40,
         40,
         40,
         40,
         41,
         42,
         43,
         0,
         0,
         44,
         45,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         46,
         47,
         48,
         49,
         50,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         51,
         52,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         53,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         53,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         18,
         18,
         18,
         18,
         18,
         18,
         18,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         35,
         35,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         15,
         15,
         15,
         15,
         15,
         15,
         15,
         15,
         15,
         15,
         15,
         0,
         0,
         0,
         0,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         21,
         20,
         21,
         20,
         21,
         20,
         20,
         20,
         21,
         21,
         20,
         21,
         20,
         20,
         20,
         21,
         20,
         21,
         20,
         20,
         20,
         20,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         0,
         0,
         0,
         0,
         0,
         54,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         55,
         56,
         57,
         58,
         59,
         60,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         61,
         62,
         63,
         64,
         65,
         66,
         0,
         0,
         0,
         0,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         19,
         19,
         19,
         19,
         19,
         19,
         19,
         19,
         19,
         19,
         19,
         19,
         19,
         19,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         67,
         67,
         67,
         67,
         67,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         68,
         68,
         68,
         68,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         69,
         16,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         70,
         71,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         72,
         73,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         23,
         23,
         23,
         23,
         23,
         23,
         24,
         23,
         24,
         23,
         23,
         24,
         23,
         24,
         23,
         23,
         24,
         24,
         23,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         74,
         75,
         76,
         77,
         78,
         0,
         39,
         39,
         39,
         39,
         39,
         41,
         42,
         42,
         42,
         42,
         43,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         33,
         33,
         33,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         17,
         17,
         17,
         37,
         37,
         37,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         40,
         0,
         0,
         0,
         0,
         0,
         46,
         50,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         53,
         79,
         80,
         81,
         82,
         5,
         5,
         5,
         5,
         53,
         53,
         53,
         53,
         53,
         53,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         15,
         15,
         15,
         15,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         19,
         19,
         19,
         19,
         19,
         19,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         0,
         0,
         20,
         21,
         21,
         21,
         21,
         21,
         22,
         22,
         22,
         22,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         40,
         3,
         3,
         3,
         69,
         69,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         10,
         11,
         0,
         24,
         25,
         25,
         25,
         25,
         67,
         25,
         25,
         25,
         26,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         0,
         0,
         0,
         34,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         73,
         73,
         71,
         0,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         9,
         9,
         9,
         9,
         9,
         9,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         19,
         19,
         19,
         0,
         0,
         0,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         0,
         16,
         21,
         22,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         11,
         11,
         53,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         21,
         21,
         21,
         29,
         29,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         83,
         0,
         0,
         8,
         14,
         8,
         8,
         0,
         0,
         9,
         9,
         0,
         0,
         8,
         8,
         8,
         0,
         0,
         9,
         9,
         0,
         0,
         8,
         8,
         8,
         0,
         0,
         9,
         9,
      },
      
      .num_filtered_states = 84,
      .table = (const uint16_t []) {
      
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         127,
         128,
         127,
         526,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         527,
         127,
         127,
         528,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         127,
         125,
         126,
         529,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         530,
         531,
         530,
         532,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         533,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         125,
         126,
         125,
         524,
         534,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         535,
         535,
         535,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         536,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         537,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         538,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         539,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         540,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         541,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         542,
         543,
         542,
         544,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         545,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         542,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         546,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         546,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         547,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         548,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         549,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         550,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         551,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         552,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         553,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         554,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         555,
         556,
         555,
         557,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         558,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         555,
         559,
         560,
         559,
         561,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         562,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         559,
         563,
         564,
         563,
         565,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         566,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         563,
         567,
         568,
         567,
         569,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         570,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         567,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         571,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         572,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         573,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         574,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         575,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         576,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         577,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         578,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         579,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         580,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         535,
         535,
         535,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         546,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         546,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         2088,
         2089,
         2088,
         2090,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2091,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2088,
         2092,
         2093,
         2092,
         2094,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2095,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2092,
         2096,
         2097,
         2096,
         2098,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2099,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2096,
         2100,
         2101,
         2100,
         2102,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2103,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         2100,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         126,
         125,
         524,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         535,
         535,
         535,
         125,
         125,
         126,
         125,
         524,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         535,
         535,
         535,
         125,
         125,
         126,
         125,
         524,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         535,
         535,
         535,
         125,
         125,
         126,
         125,
         524,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         535,
         535,
         535,
         535,
         125,
         125,
         126,
         125,
         524,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         525,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         125,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         2742,
         123,
         2743,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2744,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2745,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         581,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         582,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         583,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         584,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         585,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         2746,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2747,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         2748,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2749,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         586,
         587,
         586,
         588,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         589,
         586,
         586,
         590,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         2104,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         586,
         2912,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         595,
         596,
         595,
         597,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         598,
         595,
         595,
         599,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         595,
         591,
         592,
         600,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         601,
         602,
         601,
         603,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         604,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         601,
         591,
         592,
         591,
         593,
         605,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         606,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         606,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         606,
         606,
         606,
         606,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         607,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         608,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         609,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         610,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         611,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         612,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         613,
         614,
         613,
         615,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         616,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         613,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         617,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         617,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         618,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         619,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         620,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         621,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         622,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         623,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         624,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         625,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         626,
         627,
         626,
         628,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         629,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         626,
         630,
         631,
         630,
         632,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         633,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         630,
         634,
         635,
         634,
         636,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         637,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         634,
         638,
         639,
         638,
         640,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         641,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         638,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         642,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         643,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         644,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         645,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         646,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         647,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         648,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         649,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         650,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         651,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         606,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         606,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         606,
         606,
         606,
         606,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         617,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         617,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         2105,
         2106,
         2105,
         2107,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2108,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2105,
         2109,
         2110,
         2109,
         2111,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2112,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2109,
         2113,
         2114,
         2113,
         2115,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2116,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2113,
         2117,
         2118,
         2117,
         2119,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2120,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         2117,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         592,
         591,
         593,
         591,
         606,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         606,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         606,
         606,
         606,
         606,
         591,
         591,
         592,
         591,
         593,
         591,
         606,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         606,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         606,
         606,
         606,
         606,
         591,
         591,
         592,
         591,
         593,
         591,
         606,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         606,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         606,
         606,
         606,
         606,
         591,
         591,
         592,
         591,
         593,
         591,
         606,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         606,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         606,
         606,
         606,
         606,
         591,
         591,
         592,
         591,
         593,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         594,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         591,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         656,
         657,
         656,
         658,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         659,
         656,
         656,
         660,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         652,
         653,
         661,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         662,
         663,
         662,
         664,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         665,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         652,
         653,
         652,
         654,
         666,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         667,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         667,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         667,
         667,
         667,
         667,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         668,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         669,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         670,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         671,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         672,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         673,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         674,
         675,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         678,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         678,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         679,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         680,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         681,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         682,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         683,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         684,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         685,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         686,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         687,
         688,
         687,
         689,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         690,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         691,
         692,
         691,
         693,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         694,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         695,
         696,
         695,
         697,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         698,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         699,
         700,
         699,
         701,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         702,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         703,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         704,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         705,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         706,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         707,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         708,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         709,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         710,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         711,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         712,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         667,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         667,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         667,
         667,
         667,
         667,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         678,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         678,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         2121,
         2122,
         2121,
         2123,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2124,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2125,
         2126,
         2125,
         2127,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2128,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2129,
         2130,
         2129,
         2131,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2132,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2133,
         2134,
         2133,
         2135,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2136,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         667,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         667,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         667,
         667,
         667,
         667,
         652,
         652,
         653,
         652,
         654,
         652,
         667,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         667,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         667,
         667,
         667,
         667,
         652,
         652,
         653,
         652,
         654,
         652,
         667,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         667,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         667,
         667,
         667,
         667,
         652,
         652,
         653,
         652,
         654,
         652,
         667,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         667,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         667,
         667,
         667,
         667,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         656,
         657,
         656,
         658,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         659,
         656,
         656,
         660,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         713,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         656,
         652,
         653,
         661,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         714,
         715,
         714,
         716,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         717,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         714,
         652,
         653,
         652,
         654,
         666,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         667,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         667,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         667,
         667,
         667,
         667,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         668,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         669,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         670,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         671,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         672,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         673,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         674,
         675,
         674,
         676,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         677,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         674,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         678,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         678,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         679,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         680,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         681,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         682,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         683,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         684,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         685,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         686,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         687,
         688,
         687,
         689,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         690,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         691,
         692,
         691,
         693,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         694,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         691,
         695,
         696,
         695,
         697,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         698,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         695,
         699,
         700,
         699,
         701,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         702,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         699,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         703,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         704,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         705,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         706,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         707,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         708,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         709,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         710,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         711,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         712,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         667,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         667,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         667,
         667,
         667,
         667,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         678,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         678,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         2121,
         2122,
         2121,
         2123,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2124,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2121,
         2125,
         2126,
         2125,
         2127,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2128,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2125,
         2129,
         2130,
         2129,
         2131,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2132,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2129,
         2133,
         2134,
         2133,
         2135,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2136,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         2133,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         653,
         652,
         654,
         652,
         667,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         667,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         667,
         667,
         667,
         667,
         652,
         652,
         653,
         652,
         654,
         652,
         667,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         667,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         667,
         667,
         667,
         667,
         652,
         652,
         653,
         652,
         654,
         652,
         667,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         667,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         667,
         667,
         667,
         667,
         652,
         652,
         653,
         652,
         654,
         652,
         667,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         667,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         667,
         667,
         667,
         667,
         652,
         652,
         653,
         652,
         654,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         655,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         652,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         722,
         723,
         722,
         724,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         725,
         722,
         722,
         726,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         722,
         718,
         719,
         727,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         728,
         729,
         728,
         730,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         731,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         728,
         718,
         719,
         718,
         720,
         732,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         733,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         733,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         733,
         733,
         733,
         733,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         734,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         735,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         736,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         737,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         738,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         739,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         740,
         741,
         740,
         742,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         743,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         740,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         744,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         744,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         745,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         746,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         747,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         748,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         749,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         750,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         751,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         752,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         753,
         754,
         753,
         755,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         756,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         753,
         757,
         758,
         757,
         759,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         760,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         757,
         761,
         762,
         761,
         763,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         764,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         761,
         765,
         766,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         768,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         769,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         770,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         771,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         772,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         773,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         774,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         775,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         776,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         777,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         778,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         733,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         733,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         733,
         733,
         733,
         733,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         744,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         744,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         2137,
         2138,
         2137,
         2139,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2140,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2137,
         2141,
         2142,
         2141,
         2143,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2144,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2141,
         2145,
         2146,
         2145,
         2147,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2148,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2145,
         2149,
         2150,
         2149,
         2151,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2152,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         2149,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         719,
         718,
         720,
         718,
         733,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         733,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         733,
         733,
         733,
         733,
         718,
         718,
         719,
         718,
         720,
         718,
         733,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         733,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         733,
         733,
         733,
         733,
         718,
         718,
         719,
         718,
         720,
         718,
         733,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         733,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         733,
         733,
         733,
         733,
         718,
         718,
         719,
         718,
         720,
         718,
         733,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         733,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         733,
         733,
         733,
         733,
         718,
         718,
         719,
         718,
         720,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         721,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         718,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         783,
         784,
         783,
         785,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         786,
         783,
         783,
         787,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         779,
         780,
         788,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         789,
         790,
         789,
         791,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         792,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         779,
         780,
         779,
         781,
         793,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         794,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         794,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         794,
         794,
         794,
         794,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         795,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         796,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         797,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         798,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         799,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         800,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         801,
         802,
         801,
         803,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         804,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         805,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         805,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         806,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         807,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         808,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         809,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         810,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         811,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         812,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         813,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         814,
         815,
         814,
         816,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         817,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         818,
         819,
         818,
         820,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         821,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         822,
         823,
         822,
         824,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         825,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         826,
         827,
         826,
         828,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         829,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         830,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         831,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         832,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         833,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         834,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         835,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         836,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         837,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         838,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         839,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         794,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         794,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         794,
         794,
         794,
         794,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         805,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         805,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         2153,
         2154,
         2153,
         2155,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2156,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2157,
         2158,
         2157,
         2159,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2160,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2161,
         2162,
         2161,
         2163,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2164,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2165,
         2166,
         2165,
         2167,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2168,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         794,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         794,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         794,
         794,
         794,
         794,
         779,
         779,
         780,
         779,
         781,
         779,
         794,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         794,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         794,
         794,
         794,
         794,
         779,
         779,
         780,
         779,
         781,
         779,
         794,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         794,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         794,
         794,
         794,
         794,
         779,
         779,
         780,
         779,
         781,
         779,
         794,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         794,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         794,
         794,
         794,
         794,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         783,
         784,
         783,
         785,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         786,
         783,
         783,
         787,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         783,
         779,
         780,
         788,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         789,
         790,
         789,
         791,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         792,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         789,
         779,
         780,
         779,
         781,
         793,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         794,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         794,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         794,
         794,
         794,
         794,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         795,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         796,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         797,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         798,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         799,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         800,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         801,
         802,
         801,
         803,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         804,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         801,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         805,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         805,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         806,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         807,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         808,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         809,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         810,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         811,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         812,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         813,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         814,
         815,
         814,
         816,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         817,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         814,
         818,
         819,
         818,
         820,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         821,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         818,
         822,
         823,
         822,
         824,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         825,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         822,
         826,
         827,
         826,
         828,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         829,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         826,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         830,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         831,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         832,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         833,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         834,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         835,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         836,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         837,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         838,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         839,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         794,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         794,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         794,
         794,
         794,
         794,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         2169,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         805,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         805,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         2153,
         2154,
         2153,
         2155,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2156,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2153,
         2157,
         2158,
         2157,
         2159,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2160,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2157,
         2161,
         2162,
         2161,
         2163,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2164,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2161,
         2165,
         2166,
         2165,
         2167,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2168,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         2165,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         780,
         779,
         781,
         779,
         794,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         794,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         794,
         794,
         794,
         794,
         779,
         779,
         780,
         779,
         781,
         779,
         794,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         794,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         794,
         794,
         794,
         794,
         779,
         779,
         780,
         779,
         781,
         779,
         794,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         794,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         794,
         794,
         794,
         794,
         779,
         779,
         780,
         779,
         781,
         779,
         794,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         794,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         794,
         794,
         794,
         794,
         779,
         779,
         780,
         779,
         781,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         782,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         779,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         844,
         845,
         844,
         846,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         847,
         844,
         844,
         848,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         2170,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         2170,
         2170,
         2170,
         2170,
         844,
         840,
         841,
         849,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         850,
         851,
         850,
         852,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         853,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         840,
         841,
         840,
         842,
         854,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         855,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         855,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         855,
         855,
         855,
         855,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         856,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         857,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         858,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         859,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         860,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         861,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         862,
         863,
         862,
         864,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         865,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         866,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         866,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         867,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         868,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         869,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         870,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         871,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         872,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         873,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         874,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         875,
         876,
         875,
         877,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         878,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         879,
         880,
         879,
         881,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         882,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         883,
         884,
         883,
         885,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         886,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         887,
         888,
         887,
         889,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         890,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         891,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         892,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         893,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         894,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         895,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         896,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         897,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         898,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         899,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         900,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         855,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         855,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         855,
         855,
         855,
         855,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         866,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         866,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         2171,
         2172,
         2171,
         2173,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2174,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2175,
         2176,
         2175,
         2177,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2178,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2179,
         2180,
         2179,
         2181,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2182,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2183,
         2184,
         2183,
         2185,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2186,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         855,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         855,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         855,
         855,
         855,
         855,
         840,
         840,
         841,
         840,
         842,
         840,
         855,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         855,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         855,
         855,
         855,
         855,
         840,
         840,
         841,
         840,
         842,
         840,
         855,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         855,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         855,
         855,
         855,
         855,
         840,
         840,
         841,
         840,
         842,
         840,
         855,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         855,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         855,
         855,
         855,
         855,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         905,
         906,
         905,
         907,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         908,
         905,
         905,
         909,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         901,
         902,
         910,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         911,
         912,
         911,
         913,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         914,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         901,
         902,
         901,
         903,
         915,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         917,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         918,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         919,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         920,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         921,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         922,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         923,
         924,
         923,
         925,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         926,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         927,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         927,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         928,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         929,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         930,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         931,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         932,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         933,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         934,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         935,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         936,
         937,
         936,
         938,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         939,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         940,
         941,
         940,
         942,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         943,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         944,
         945,
         944,
         946,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         947,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         948,
         949,
         948,
         950,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         951,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         952,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         953,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         954,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         955,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         956,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         957,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         958,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         959,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         960,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         961,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         927,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         927,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         2187,
         2188,
         2187,
         2189,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2190,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2191,
         2192,
         2191,
         2193,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2194,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2195,
         2196,
         2195,
         2197,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2198,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2199,
         2200,
         2199,
         2201,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2202,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         966,
         967,
         966,
         968,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         969,
         966,
         966,
         970,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         962,
         963,
         971,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         972,
         973,
         972,
         974,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         975,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         962,
         963,
         962,
         964,
         976,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         977,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         977,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         977,
         977,
         977,
         977,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         978,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         979,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         980,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         981,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         982,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         983,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         984,
         985,
         984,
         986,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         987,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         988,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         988,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         989,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         990,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         991,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         992,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         993,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         994,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         995,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         996,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         997,
         998,
         997,
         999,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         1000,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         1001,
         1002,
         1001,
         1003,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1004,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1005,
         1006,
         1005,
         1007,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1008,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1009,
         1010,
         1009,
         1011,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1012,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         1013,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         1014,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         1015,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         1016,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         1017,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         1018,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         1019,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         1020,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         1021,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         1022,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         2203,
         962,
         964,
         962,
         977,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         977,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         977,
         977,
         977,
         977,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         988,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         988,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         2204,
         2205,
         2204,
         2206,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2207,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2208,
         2209,
         2208,
         2210,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2211,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2212,
         2213,
         2212,
         2214,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2215,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2216,
         2217,
         2216,
         2218,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2219,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         2203,
         962,
         964,
         962,
         977,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         977,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         977,
         977,
         977,
         977,
         962,
         962,
         2203,
         962,
         964,
         962,
         977,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         977,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         977,
         977,
         977,
         977,
         962,
         962,
         2203,
         962,
         964,
         962,
         977,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         977,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         977,
         977,
         977,
         977,
         962,
         962,
         2203,
         962,
         964,
         962,
         977,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         977,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         977,
         977,
         977,
         977,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         2750,
         2742,
         123,
         2743,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2751,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2744,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2745,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         2752,
         123,
         2743,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2753,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2745,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         2742,
         2754,
         2743,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2744,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2755,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2745,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         2742,
         123,
         2756,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2744,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2757,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         2742,
         123,
         2743,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2220,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2744,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2745,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2220,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2220,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         2220,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         2220,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         2220,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         2742,
         123,
         2743,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2744,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2745,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2221,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2221,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         2221,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         2221,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         2221,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         581,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         582,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         583,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         584,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         585,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2758,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2759,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         2746,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2747,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         2748,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2749,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         581,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         582,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         583,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         584,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         585,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         2760,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2761,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         2748,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2749,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         581,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         582,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         583,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         584,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         585,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         2746,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2747,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         2762,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2763,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         2748,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2749,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         581,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         582,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         583,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         584,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         585,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         2746,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2747,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         2764,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2765,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         2222,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         2222,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         2222,
         2222,
         2222,
         2222,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         581,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         582,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         583,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         584,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         585,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         2746,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2747,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         2748,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2749,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         2223,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         2223,
         2223,
         2223,
         2223,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         581,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         582,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         583,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         584,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         585,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         2746,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2747,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         2748,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2749,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         844,
         845,
         844,
         846,
         844,
         2224,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         847,
         844,
         844,
         848,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         2225,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         844,
         2225,
         2225,
         2225,
         2225,
         844,
         840,
         841,
         849,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         850,
         851,
         850,
         852,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         853,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         850,
         840,
         841,
         840,
         842,
         854,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         855,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         855,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         855,
         855,
         855,
         855,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         856,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         857,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         858,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         859,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         860,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         861,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         862,
         863,
         862,
         864,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         865,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         862,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         866,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         866,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         867,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         868,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         869,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         870,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         871,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         872,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         873,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         874,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         875,
         876,
         875,
         877,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         878,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         875,
         879,
         880,
         879,
         881,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         882,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         879,
         883,
         884,
         883,
         885,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         886,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         883,
         887,
         888,
         887,
         889,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         890,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         887,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         891,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         892,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         893,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         894,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         895,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         896,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         897,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         898,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         899,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         900,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         855,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         855,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         855,
         855,
         855,
         855,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         866,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         866,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         2171,
         2172,
         2171,
         2173,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2174,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2175,
         2176,
         2175,
         2177,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2178,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2175,
         2179,
         2180,
         2179,
         2181,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2182,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2179,
         2183,
         2184,
         2183,
         2185,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2186,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         2183,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         841,
         840,
         842,
         840,
         855,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         855,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         855,
         855,
         855,
         855,
         840,
         840,
         841,
         840,
         842,
         840,
         855,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         855,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         855,
         855,
         855,
         855,
         840,
         840,
         841,
         840,
         842,
         840,
         855,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         855,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         855,
         855,
         855,
         855,
         840,
         840,
         841,
         840,
         842,
         840,
         855,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         855,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         855,
         855,
         855,
         855,
         840,
         840,
         841,
         840,
         842,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         843,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         840,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         966,
         967,
         966,
         968,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         969,
         966,
         966,
         970,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         966,
         962,
         963,
         971,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         972,
         973,
         972,
         974,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         975,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         972,
         962,
         963,
         962,
         964,
         976,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         2226,
         962,
         964,
         962,
         977,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         977,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         977,
         977,
         977,
         977,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         978,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         979,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         980,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         981,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         982,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         983,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         984,
         985,
         984,
         986,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         987,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         984,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         988,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         988,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         989,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         990,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         991,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         992,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         993,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         994,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         995,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         996,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         997,
         998,
         997,
         999,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         1000,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         997,
         1001,
         1002,
         1001,
         1003,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1004,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1005,
         1006,
         1005,
         1007,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1008,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1005,
         1009,
         1010,
         1009,
         1011,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1012,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         1009,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         1013,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         1014,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         1015,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         1016,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         1017,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         1018,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         1019,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         1020,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         1021,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         1022,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         2227,
         962,
         964,
         962,
         977,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         977,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         977,
         977,
         977,
         977,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         988,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         988,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         2204,
         2205,
         2204,
         2206,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2207,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2204,
         2208,
         2209,
         2208,
         2210,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2211,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2208,
         2212,
         2213,
         2212,
         2214,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2215,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2212,
         2216,
         2217,
         2216,
         2218,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2219,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         2216,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         2227,
         962,
         964,
         962,
         977,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         977,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         977,
         977,
         977,
         977,
         962,
         962,
         2227,
         962,
         964,
         962,
         977,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         977,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         977,
         977,
         977,
         977,
         962,
         962,
         2227,
         962,
         964,
         962,
         977,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         977,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         977,
         977,
         977,
         977,
         962,
         962,
         2227,
         962,
         964,
         962,
         977,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         977,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         977,
         977,
         977,
         977,
         962,
         962,
         963,
         962,
         964,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         965,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         962,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         905,
         906,
         905,
         907,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         908,
         905,
         905,
         909,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         901,
         902,
         910,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         911,
         912,
         911,
         913,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         914,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         901,
         902,
         901,
         903,
         915,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         917,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         918,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         919,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         920,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         921,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         922,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         923,
         924,
         923,
         925,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         926,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         927,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         927,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         928,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         929,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         930,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         931,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         932,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         933,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         934,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         935,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         936,
         937,
         936,
         938,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         939,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         940,
         941,
         940,
         942,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         943,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         944,
         945,
         944,
         946,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         947,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         948,
         949,
         948,
         950,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         951,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         952,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         953,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         954,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         955,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         956,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         957,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         958,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         959,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         2228,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         960,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         961,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         927,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         927,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         2187,
         2188,
         2187,
         2189,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2190,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2191,
         2192,
         2191,
         2193,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2194,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2195,
         2196,
         2195,
         2197,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2198,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2199,
         2200,
         2199,
         2201,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2202,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         905,
         906,
         905,
         907,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         908,
         905,
         905,
         909,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         901,
         902,
         910,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         911,
         912,
         911,
         913,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         914,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         901,
         902,
         901,
         903,
         915,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         917,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         918,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         919,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         920,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         921,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         922,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         923,
         924,
         923,
         925,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         926,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         927,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         927,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         928,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         929,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         930,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         931,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         932,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         933,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         934,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         935,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         936,
         937,
         936,
         938,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         939,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         940,
         941,
         940,
         942,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         943,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         944,
         945,
         944,
         946,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         947,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         948,
         949,
         948,
         950,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         951,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         952,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         953,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         954,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         955,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         956,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         957,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         958,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         959,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         2229,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         960,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         961,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         927,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         927,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         2187,
         2188,
         2187,
         2189,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2190,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2191,
         2192,
         2191,
         2193,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2194,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2195,
         2196,
         2195,
         2197,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2198,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2199,
         2200,
         2199,
         2201,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2202,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         905,
         906,
         905,
         907,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         908,
         905,
         905,
         909,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         901,
         902,
         910,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         911,
         912,
         911,
         913,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         914,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         901,
         902,
         901,
         903,
         915,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         917,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         918,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         919,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         920,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         921,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         922,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         923,
         924,
         923,
         925,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         926,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         927,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         927,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         928,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         929,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         930,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         931,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         932,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         933,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         934,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         935,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         936,
         937,
         936,
         938,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         939,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         940,
         941,
         940,
         942,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         943,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         944,
         945,
         944,
         946,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         947,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         948,
         949,
         948,
         950,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         951,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         952,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         953,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         954,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         955,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         956,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         957,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         958,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         959,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         2230,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         960,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         961,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         927,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         927,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         2187,
         2188,
         2187,
         2189,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2190,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2191,
         2192,
         2191,
         2193,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2194,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2195,
         2196,
         2195,
         2197,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2198,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2199,
         2200,
         2199,
         2201,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2202,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         905,
         906,
         905,
         907,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         908,
         905,
         905,
         909,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         901,
         902,
         910,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         911,
         912,
         911,
         913,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         914,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         901,
         902,
         901,
         903,
         915,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         917,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         918,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         919,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         920,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         921,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         922,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         923,
         924,
         923,
         925,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         926,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         927,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         927,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         928,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         929,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         930,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         931,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         932,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         933,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         934,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         935,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         936,
         937,
         936,
         938,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         939,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         940,
         941,
         940,
         942,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         943,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         944,
         945,
         944,
         946,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         947,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         948,
         949,
         948,
         950,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         951,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         952,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         953,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         954,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         955,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         956,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         957,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         958,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         959,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         2231,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         960,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         961,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         927,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         927,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         2187,
         2188,
         2187,
         2189,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2190,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2191,
         2192,
         2191,
         2193,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2194,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2195,
         2196,
         2195,
         2197,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2198,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2199,
         2200,
         2199,
         2201,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2202,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         905,
         906,
         905,
         907,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         908,
         905,
         905,
         909,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         901,
         902,
         910,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         911,
         912,
         911,
         913,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         914,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         911,
         901,
         902,
         901,
         903,
         915,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         917,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         918,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         919,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         920,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         921,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         922,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         923,
         924,
         923,
         925,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         926,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         923,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         927,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         927,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         928,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         929,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         930,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         931,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         932,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         933,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         934,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         935,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         936,
         937,
         936,
         938,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         939,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         936,
         940,
         941,
         940,
         942,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         943,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         940,
         944,
         945,
         944,
         946,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         947,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         944,
         948,
         949,
         948,
         950,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         951,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         948,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         952,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         953,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         954,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         955,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         956,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         957,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         958,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         959,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         2232,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         960,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         961,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         927,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         927,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         2187,
         2188,
         2187,
         2189,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2190,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2187,
         2191,
         2192,
         2191,
         2193,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2194,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2191,
         2195,
         2196,
         2195,
         2197,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2198,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2195,
         2199,
         2200,
         2199,
         2201,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2202,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         2199,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         916,
         916,
         916,
         916,
         901,
         901,
         902,
         901,
         903,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         904,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         901,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         123,
         124,
         123,
         469,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         470,
         123,
         123,
         471,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         123,
         121,
         122,
         472,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         473,
         474,
         473,
         475,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         476,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         473,
         121,
         122,
         121,
         467,
         477,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         479,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         480,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         481,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         482,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         483,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         484,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         485,
         486,
         485,
         487,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         488,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         485,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         490,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         491,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         492,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         493,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         494,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         495,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         496,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         497,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         498,
         499,
         498,
         500,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         501,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         498,
         502,
         503,
         502,
         504,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         505,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         502,
         506,
         507,
         506,
         508,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         509,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         506,
         510,
         511,
         510,
         512,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         513,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         510,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         514,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         515,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         516,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         517,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         518,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         519,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         520,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         521,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         522,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         523,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         489,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         2072,
         2073,
         2072,
         2074,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2075,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2072,
         2076,
         2077,
         2076,
         2078,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2079,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2076,
         2080,
         2081,
         2080,
         2082,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2083,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2080,
         2084,
         2085,
         2084,
         2086,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2087,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         2084,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         478,
         478,
         478,
         478,
         121,
         121,
         122,
         121,
         467,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         468,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
         121,
      },
   },
   [nir_op_feq] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         3,
         0,
         0,
         4,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         12,
         13,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         17,
         2,
         2,
         2,
         2,
         2,
         18,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         4,
         4,
         4,
         4,
         4,
         19,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         20,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         9,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         0,
         0,
         0,
         13,
         13,
         13,
         13,
         14,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         6,
         4,
         4,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         4,
         4,
         4,
         21,
         4,
         4,
         0,
         2,
         2,
         18,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         21,
         4,
         4,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 22,
      .table = (const uint16_t []) {
      
         129,
         130,
         1422,
         129,
         1423,
         1423,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         1422,
         1422,
         1423,
         129,
         1423,
         130,
         130,
         1424,
         1425,
         1426,
         1427,
         1428,
         1429,
         1430,
         1431,
         130,
         1432,
         1433,
         1434,
         1435,
         2290,
         2291,
         2292,
         1424,
         2293,
         2294,
         2864,
         1422,
         1424,
         1436,
         1422,
         1437,
         1437,
         1422,
         1422,
         1422,
         1422,
         1422,
         1422,
         1422,
         1422,
         1422,
         1422,
         1422,
         1436,
         1436,
         1437,
         1422,
         1437,
         129,
         1425,
         1422,
         129,
         1423,
         1423,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         1422,
         1422,
         1423,
         129,
         1423,
         1423,
         1426,
         1437,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1437,
         1437,
         1423,
         1423,
         1423,
         1423,
         1427,
         1437,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1437,
         1437,
         1423,
         1423,
         1423,
         129,
         1428,
         1422,
         129,
         1423,
         1423,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         1422,
         1422,
         1423,
         129,
         1423,
         129,
         1429,
         1422,
         129,
         1423,
         1423,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         1422,
         1422,
         1423,
         129,
         1423,
         129,
         1430,
         1422,
         129,
         1423,
         1423,
         129,
         129,
         1438,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         1422,
         2295,
         1423,
         129,
         1423,
         129,
         1431,
         1422,
         129,
         1423,
         1423,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         1422,
         1422,
         1423,
         129,
         1423,
         129,
         130,
         1422,
         129,
         1423,
         1423,
         129,
         129,
         129,
         129,
         1439,
         129,
         129,
         129,
         129,
         129,
         129,
         1422,
         1422,
         1423,
         129,
         1423,
         129,
         1432,
         1422,
         129,
         1423,
         1423,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         1422,
         1422,
         1423,
         129,
         1423,
         129,
         1433,
         1422,
         129,
         1423,
         1423,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         1422,
         1422,
         1423,
         129,
         1423,
         129,
         1434,
         1422,
         129,
         1423,
         1423,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         1422,
         1422,
         1423,
         129,
         1423,
         129,
         1435,
         1422,
         129,
         1423,
         1423,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         1422,
         1422,
         1423,
         129,
         1423,
         129,
         2290,
         1422,
         129,
         1423,
         1423,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         1422,
         1422,
         1423,
         129,
         1423,
         129,
         2291,
         1422,
         129,
         1423,
         1423,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         1422,
         1422,
         1423,
         129,
         1423,
         1422,
         2292,
         1436,
         1422,
         1437,
         1437,
         1422,
         1422,
         1422,
         1422,
         1422,
         1422,
         1422,
         1422,
         1422,
         1422,
         1422,
         1436,
         1436,
         1437,
         1422,
         1437,
         1422,
         1424,
         1436,
         1422,
         1437,
         1437,
         1422,
         1422,
         2295,
         1422,
         1422,
         1422,
         1422,
         1422,
         1422,
         1422,
         1422,
         1436,
         2296,
         1437,
         1422,
         1437,
         1423,
         2293,
         1437,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1437,
         1437,
         1423,
         1423,
         1423,
         129,
         2294,
         1422,
         129,
         1423,
         1423,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         129,
         1422,
         1422,
         1423,
         129,
         1423,
         1423,
         2864,
         1437,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1423,
         1437,
         1437,
         1423,
         1423,
         1423,
      },
   },
   [nir_op_fexp2] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         3,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         4,
         1,
         1,
         1,
         1,
         3,
         1,
         1,
         1,
         1,
         1,
         3,
         1,
         1,
         3,
         1,
         3,
         1,
         1,
         3,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         3,
         3,
         3,
         3,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         3,
         3,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 6,
      .table = (const uint16_t []) {
      
         131,
         1777,
         1778,
         2617,
         2618,
         2866,
      },
   },
   [nir_search_op_b2f] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         1,
         0,
         1,
         0,
         1,
         0,
         0,
         0,
         1,
         1,
         0,
         1,
         0,
         0,
         0,
         1,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         132,
         1954,
         1955,
         1956,
      },
   },
   [nir_op_ftrunc] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         133,
      },
   },
   [nir_op_ffloor] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         134,
      },
   },
   [nir_op_ffract] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         135,
      },
   },
   [nir_op_fceil] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         136,
      },
   },
   [nir_search_op_b2i] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         0,
         2,
         0,
         2,
         0,
         0,
         0,
         2,
         2,
         0,
         2,
         0,
         0,
         0,
         2,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         5,
         1,
         1,
         1,
         1,
         1,
         1,
         6,
         6,
         7,
         7,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 8,
      .table = (const uint16_t []) {
      
         137,
         1849,
         1850,
         1851,
         1852,
         2685,
         2686,
         2687,
      },
   },
   [nir_op_fdph] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         138,
      },
   },
   [nir_op_fdot4] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         139,
         1944,
         1945,
         1946,
         1944,
         1944,
         1945,
         1946,
         1945,
         1945,
         1945,
         1946,
         1946,
         1946,
         1946,
         1946,
      },
   },
   [nir_op_vec4] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         140,
         141,
         140,
         142,
         140,
         141,
         140,
         143,
         140,
         141,
         140,
         142,
         140,
         141,
         140,
         143,
      },
   },
   [nir_op_fdot3] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         144,
         1947,
         1948,
         1947,
         1947,
         1948,
         1948,
         1948,
         1948,
      },
   },
   [nir_op_vec3] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         145,
         146,
         145,
         147,
         145,
         146,
         145,
         147,
      },
   },
   [nir_op_fdot2] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         148,
         149,
         1949,
         149,
         149,
         1950,
         1949,
         1950,
         1949,
      },
   },
   [nir_op_vec2] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         150,
         151,
         150,
         150,
         152,
         153,
         152,
         152,
         150,
         151,
         150,
         1269,
         150,
         151,
         150,
         150,
      },
   },
   [nir_op_fsum2] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         154,
      },
   },
   [nir_op_ushr] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         2,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         8,
         0,
         2,
         0,
         0,
         0,
         2,
         0,
         0,
         2,
         2,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         11,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         2,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         4,
         13,
         0,
         3,
         3,
         8,
         4,
         14,
         9,
         2,
         4,
         15,
         3,
         3,
         8,
         4,
         16,
         9,
         2,
         4,
         17,
         3,
         3,
         8,
         4,
         0,
         9,
      },
      
      .num_filtered_states = 18,
      .table = (const uint16_t []) {
      
         155,
         156,
         155,
         1243,
         155,
         1244,
         155,
         2279,
         2280,
         155,
         155,
         1243,
         155,
         1243,
         155,
         1243,
         155,
         1243,
         157,
         158,
         157,
         1245,
         157,
         1246,
         157,
         2281,
         2282,
         157,
         157,
         1245,
         157,
         1245,
         157,
         1245,
         157,
         1245,
         1247,
         1248,
         1247,
         1249,
         1247,
         1250,
         1247,
         2283,
         2284,
         1247,
         1247,
         1249,
         1247,
         1249,
         1247,
         1249,
         1247,
         1249,
         155,
         156,
         155,
         1243,
         155,
         1244,
         155,
         2279,
         2280,
         155,
         155,
         1243,
         155,
         1243,
         155,
         1243,
         155,
         1243,
         155,
         1251,
         155,
         1243,
         155,
         1244,
         155,
         2279,
         2280,
         155,
         155,
         1243,
         155,
         1243,
         155,
         1243,
         155,
         1243,
         155,
         156,
         155,
         1243,
         155,
         1244,
         155,
         2279,
         2280,
         155,
         155,
         1243,
         155,
         1243,
         155,
         1243,
         155,
         1243,
         155,
         1252,
         155,
         1243,
         155,
         1244,
         155,
         2279,
         2280,
         155,
         155,
         1243,
         155,
         1243,
         155,
         1243,
         155,
         1243,
         155,
         156,
         155,
         1243,
         155,
         1244,
         155,
         2279,
         2280,
         155,
         155,
         1243,
         155,
         1243,
         155,
         1243,
         155,
         1243,
         1247,
         1248,
         1247,
         1249,
         1247,
         1250,
         1247,
         2283,
         2284,
         1247,
         1247,
         1249,
         1247,
         1249,
         1247,
         1249,
         1247,
         1249,
         155,
         2285,
         155,
         1243,
         155,
         1244,
         155,
         2279,
         2280,
         155,
         155,
         1243,
         155,
         1243,
         155,
         1243,
         155,
         1243,
         155,
         2286,
         155,
         1243,
         155,
         1244,
         155,
         2279,
         2280,
         155,
         155,
         1243,
         155,
         1243,
         155,
         1243,
         155,
         1243,
         155,
         2814,
         155,
         1243,
         155,
         1244,
         155,
         2279,
         2280,
         155,
         155,
         1243,
         155,
         1243,
         155,
         1243,
         155,
         1243,
         155,
         2910,
         155,
         1243,
         155,
         1244,
         155,
         2279,
         2280,
         155,
         155,
         1243,
         155,
         1243,
         155,
         1243,
         155,
         1243,
         155,
         2916,
         155,
         1243,
         155,
         1244,
         155,
         2279,
         2280,
         155,
         155,
         1243,
         155,
         1243,
         155,
         1243,
         155,
         1243,
         155,
         2920,
         155,
         1243,
         155,
         1244,
         155,
         2279,
         2280,
         155,
         155,
         1243,
         155,
         1243,
         155,
         1243,
         155,
         1243,
         155,
         2925,
         155,
         1243,
         155,
         1244,
         155,
         2279,
         2280,
         155,
         155,
         1243,
         155,
         1243,
         155,
         1243,
         155,
         1243,
         155,
         2929,
         155,
         1243,
         155,
         1244,
         155,
         2279,
         2280,
         155,
         155,
         1243,
         155,
         1243,
         155,
         1243,
         155,
         1243,
         155,
         2934,
         155,
         1243,
         155,
         1244,
         155,
         2279,
         2280,
         155,
         155,
         1243,
         155,
         1243,
         155,
         1243,
         155,
         1243,
      },
   },
   [nir_op_ishr] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         2,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         0,
         2,
         0,
         0,
         0,
         2,
         0,
         0,
         2,
         2,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         5,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         4,
         0,
         4,
         4,
         2,
         0,
         0,
         0,
         2,
         0,
         4,
         4,
         4,
         2,
         0,
         0,
         0,
         2,
         0,
         4,
         4,
         4,
         2,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 6,
      .table = (const uint16_t []) {
      
         159,
         160,
         159,
         159,
         1253,
         159,
         161,
         162,
         161,
         161,
         1254,
         161,
         159,
         1255,
         159,
         159,
         1253,
         159,
         1256,
         1257,
         1256,
         1256,
         1258,
         1256,
         159,
         160,
         159,
         159,
         1253,
         159,
         159,
         1259,
         159,
         159,
         1253,
         159,
      },
   },
   [nir_op_inot] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         4,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         6,
         7,
         8,
         8,
         8,
         8,
         9,
         9,
         9,
         9,
         10,
         10,
         11,
         11,
         12,
         12,
         12,
         12,
         13,
         13,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         17,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         18,
         19,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         20,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         6,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         7,
         7,
         21,
         22,
         0,
         0,
         0,
         0,
         0,
         21,
         0,
         0,
         22,
         0,
         0,
         0,
         0,
         0,
         0,
         21,
         22,
         0,
         0,
         0,
         0,
         0,
         21,
         22,
         0,
         0,
         0,
         0,
         0,
         22,
         22,
         21,
         22,
         0,
         0,
         0,
         0,
         0,
         21,
         22,
         0,
         0,
         0,
         0,
         0,
         21,
         22,
         0,
         0,
         0,
         0,
         0,
         21,
         22,
         0,
         0,
         0,
         0,
         0,
         21,
         22,
         0,
         0,
         0,
         0,
         0,
         23,
         21,
         21,
         21,
         21,
         21,
         0,
         22,
         22,
         22,
         22,
         22,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         8,
         8,
         8,
         9,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         3,
         11,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         10,
         0,
         8,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 24,
      .table = (const uint16_t []) {
      
         163,
         1675,
         1676,
         1677,
         1678,
         1679,
         1680,
         1681,
         1682,
         1683,
         1684,
         1685,
         1686,
         1687,
         1688,
         2436,
         2437,
         2438,
         2439,
         2440,
         2441,
         2809,
         2810,
         2811,
      },
   },
   [nir_op_flt] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         3,
         4,
         0,
         0,
         5,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         10,
         11,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         13,
         2,
         14,
         2,
         2,
         2,
         2,
         2,
         15,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         16,
         5,
         17,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         16,
         5,
         5,
         5,
         5,
         18,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         10,
         10,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         7,
         7,
         7,
         7,
         7,
         7,
         16,
         5,
         17,
         17,
         17,
         17,
         17,
         17,
         6,
         6,
         16,
         16,
         5,
         16,
         16,
         5,
         0,
         19,
         2,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         6,
         20,
         6,
         6,
         6,
         6,
         21,
         16,
         16,
         16,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         16,
         16,
         16,
         5,
         5,
         5,
         5,
         16,
         5,
         5,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         22,
         9,
         9,
         22,
         9,
         9,
         22,
         9,
         22,
         22,
         9,
         22,
         9,
         22,
         9,
         22,
         22,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         11,
         0,
         17,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 23,
      .table = (const uint16_t []) {
      
         164,
         165,
         1440,
         164,
         164,
         164,
         164,
         164,
         164,
         1441,
         164,
         1442,
         164,
         1440,
         1440,
         1440,
         164,
         164,
         164,
         1440,
         164,
         164,
         1441,
         166,
         167,
         1443,
         1444,
         1445,
         166,
         166,
         1446,
         1447,
         1448,
         1449,
         1450,
         166,
         1443,
         1443,
         1443,
         166,
         166,
         2536,
         1443,
         166,
         166,
         1448,
         164,
         1451,
         1452,
         164,
         164,
         164,
         164,
         164,
         164,
         1441,
         164,
         1442,
         164,
         1452,
         1452,
         1452,
         164,
         164,
         164,
         1452,
         164,
         164,
         1441,
         164,
         165,
         1440,
         164,
         164,
         164,
         164,
         164,
         164,
         1441,
         164,
         1442,
         164,
         1440,
         1440,
         1440,
         164,
         164,
         164,
         1440,
         164,
         164,
         1441,
         164,
         1453,
         1440,
         164,
         164,
         164,
         164,
         164,
         164,
         1441,
         164,
         1442,
         164,
         1440,
         1440,
         1440,
         164,
         164,
         164,
         1440,
         164,
         164,
         1441,
         1454,
         1455,
         1456,
         1454,
         1454,
         1454,
         1454,
         1454,
         1454,
         1457,
         1454,
         1458,
         1454,
         1456,
         1456,
         1456,
         1454,
         1454,
         1454,
         1456,
         1454,
         1454,
         1457,
         1454,
         1459,
         1456,
         1454,
         1454,
         1454,
         1454,
         1454,
         1454,
         1457,
         1454,
         1458,
         1454,
         1456,
         1456,
         1456,
         1454,
         1454,
         1454,
         1456,
         1454,
         1454,
         1457,
         164,
         165,
         1440,
         164,
         164,
         164,
         164,
         164,
         164,
         1441,
         164,
         1442,
         164,
         1440,
         1440,
         1440,
         164,
         164,
         164,
         1440,
         164,
         164,
         1441,
         164,
         1460,
         1440,
         164,
         164,
         164,
         164,
         164,
         164,
         1441,
         164,
         1442,
         164,
         1440,
         1440,
         1440,
         164,
         164,
         164,
         1440,
         164,
         164,
         1441,
         1461,
         1462,
         1463,
         1461,
         1461,
         1461,
         1461,
         1461,
         1461,
         1464,
         1461,
         1465,
         1461,
         1463,
         1463,
         1463,
         1461,
         1461,
         1461,
         1463,
         1461,
         1461,
         1464,
         164,
         1466,
         1440,
         164,
         164,
         164,
         164,
         164,
         164,
         1441,
         164,
         1442,
         164,
         1440,
         1440,
         1440,
         164,
         164,
         164,
         1440,
         164,
         164,
         1441,
         1467,
         1468,
         1469,
         1467,
         1467,
         1467,
         1467,
         1467,
         1467,
         1470,
         1467,
         1471,
         1467,
         1469,
         1469,
         1469,
         1467,
         1467,
         1467,
         1469,
         1467,
         1467,
         1470,
         164,
         165,
         1440,
         164,
         164,
         164,
         164,
         164,
         164,
         1441,
         164,
         1442,
         1472,
         1440,
         1440,
         1440,
         164,
         164,
         164,
         1440,
         164,
         164,
         1441,
         164,
         2537,
         1452,
         164,
         164,
         164,
         164,
         164,
         164,
         1441,
         164,
         1442,
         164,
         1452,
         1452,
         1452,
         164,
         164,
         164,
         1452,
         164,
         164,
         1441,
         164,
         2538,
         1452,
         164,
         164,
         164,
         164,
         164,
         164,
         1441,
         164,
         1442,
         164,
         1452,
         1452,
         1452,
         164,
         164,
         164,
         1452,
         164,
         164,
         1441,
         164,
         2539,
         1452,
         164,
         164,
         164,
         164,
         164,
         164,
         1441,
         164,
         1442,
         164,
         1452,
         1452,
         1452,
         164,
         164,
         164,
         1452,
         164,
         164,
         1441,
         1454,
         2540,
         1456,
         1454,
         1454,
         1454,
         1454,
         1454,
         1454,
         1457,
         1454,
         1458,
         1454,
         1456,
         1456,
         1456,
         1454,
         1454,
         1454,
         1456,
         1454,
         1454,
         1457,
         1454,
         2541,
         1456,
         1454,
         1454,
         1454,
         1454,
         1454,
         1454,
         1457,
         1454,
         1458,
         1454,
         1456,
         1456,
         1456,
         1454,
         1454,
         1454,
         1456,
         1454,
         1454,
         1457,
         1454,
         1455,
         1456,
         1454,
         1454,
         1454,
         1454,
         1454,
         1454,
         1457,
         1454,
         1458,
         1454,
         1456,
         1456,
         1456,
         1454,
         1454,
         1454,
         1456,
         1454,
         1454,
         1457,
         164,
         2867,
         1452,
         164,
         164,
         164,
         164,
         164,
         164,
         1441,
         164,
         1442,
         164,
         1452,
         1452,
         1452,
         164,
         164,
         164,
         1452,
         164,
         164,
         1441,
         1454,
         2890,
         1456,
         1454,
         1454,
         1454,
         1454,
         1454,
         1454,
         1457,
         1454,
         1458,
         1454,
         1456,
         1456,
         1456,
         1454,
         1454,
         1454,
         1456,
         1454,
         1454,
         1457,
         1454,
         2891,
         1456,
         1454,
         1454,
         1454,
         1454,
         1454,
         1454,
         1457,
         1454,
         1458,
         1454,
         1456,
         1456,
         1456,
         1454,
         1454,
         1454,
         1456,
         1454,
         1454,
         1457,
         1461,
         2892,
         1463,
         1461,
         1461,
         1461,
         1461,
         1461,
         1461,
         1464,
         1461,
         1465,
         1461,
         1463,
         1463,
         1463,
         1461,
         1461,
         1461,
         1463,
         1461,
         1461,
         1464,
      },
   },
   [nir_op_fge] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         3,
         4,
         0,
         0,
         5,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         11,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         14,
         2,
         15,
         2,
         2,
         2,
         2,
         2,
         16,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         5,
         5,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         5,
         5,
         5,
         5,
         5,
         17,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         11,
         11,
         11,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         7,
         7,
         7,
         7,
         7,
         7,
         5,
         5,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         18,
         2,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         0,
         6,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 19,
      .table = (const uint16_t []) {
      
         168,
         169,
         1473,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         1474,
         168,
         1475,
         168,
         1473,
         1473,
         1473,
         168,
         1473,
         170,
         171,
         1476,
         1477,
         1478,
         170,
         170,
         170,
         1479,
         1480,
         1481,
         1482,
         1483,
         170,
         1476,
         1476,
         1476,
         2542,
         1476,
         168,
         1484,
         1485,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         1474,
         168,
         1475,
         168,
         1485,
         1485,
         1485,
         168,
         1485,
         168,
         169,
         1473,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         1474,
         168,
         1475,
         168,
         1473,
         1473,
         1473,
         168,
         1473,
         168,
         1486,
         1473,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         1474,
         168,
         1475,
         168,
         1473,
         1473,
         1473,
         168,
         1473,
         1487,
         1488,
         1489,
         1487,
         1487,
         1487,
         1487,
         1487,
         1487,
         1487,
         1490,
         1487,
         1491,
         1487,
         1489,
         1489,
         1489,
         1487,
         1489,
         1487,
         1492,
         1489,
         1487,
         1487,
         1487,
         1487,
         1487,
         1487,
         1487,
         1490,
         1487,
         1491,
         1487,
         1489,
         1489,
         1489,
         1487,
         1489,
         168,
         1493,
         1473,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         1474,
         168,
         1475,
         168,
         1473,
         1473,
         1473,
         168,
         1473,
         168,
         1494,
         1473,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         1474,
         168,
         1475,
         168,
         1473,
         1473,
         1473,
         168,
         1473,
         168,
         169,
         1473,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         1474,
         168,
         1475,
         168,
         1473,
         1473,
         1473,
         168,
         1473,
         1495,
         1496,
         1497,
         1495,
         1495,
         1495,
         1495,
         1495,
         1495,
         1495,
         1498,
         1495,
         1499,
         1495,
         1497,
         1497,
         1497,
         1495,
         1497,
         168,
         1500,
         1473,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         1474,
         168,
         1475,
         168,
         1473,
         1473,
         1473,
         168,
         1473,
         1501,
         1502,
         1503,
         1501,
         1501,
         1501,
         1501,
         1501,
         1501,
         1501,
         1504,
         1501,
         1505,
         1501,
         1503,
         1503,
         1503,
         1501,
         1503,
         168,
         169,
         1473,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         1474,
         168,
         1475,
         1506,
         1473,
         1473,
         1473,
         168,
         1473,
         168,
         2543,
         1485,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         1474,
         168,
         1475,
         168,
         1485,
         1485,
         1485,
         168,
         1485,
         168,
         2544,
         1485,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         1474,
         168,
         1475,
         168,
         1485,
         1485,
         1485,
         168,
         1485,
         168,
         2545,
         1485,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         1474,
         168,
         1475,
         168,
         1485,
         1485,
         1485,
         168,
         1485,
         1487,
         1488,
         1489,
         1487,
         1487,
         1487,
         1487,
         1487,
         1487,
         1487,
         1490,
         1487,
         1491,
         1487,
         1489,
         1489,
         1489,
         1487,
         1489,
         168,
         2868,
         1485,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         1474,
         168,
         1475,
         168,
         1485,
         1485,
         1485,
         168,
         1485,
      },
   },
   [nir_op_fneu] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         3,
         0,
         0,
         4,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         12,
         13,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         17,
         2,
         2,
         2,
         2,
         2,
         18,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         4,
         4,
         4,
         4,
         4,
         19,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         20,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         9,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         0,
         0,
         0,
         13,
         13,
         13,
         13,
         14,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         6,
         4,
         4,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         4,
         4,
         4,
         21,
         4,
         4,
         0,
         2,
         2,
         18,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         21,
         4,
         4,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 22,
      .table = (const uint16_t []) {
      
         172,
         173,
         1507,
         172,
         1508,
         1508,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         1507,
         1507,
         1508,
         172,
         1508,
         173,
         173,
         1509,
         1510,
         1511,
         1512,
         1513,
         1514,
         1515,
         1516,
         173,
         1517,
         1518,
         1519,
         1520,
         2297,
         2298,
         2299,
         1509,
         2300,
         2301,
         2865,
         1507,
         1509,
         1521,
         1507,
         1522,
         1522,
         1507,
         1507,
         1507,
         1507,
         1507,
         1507,
         1507,
         1507,
         1507,
         1507,
         1507,
         1521,
         1521,
         1522,
         1507,
         1522,
         172,
         1510,
         1507,
         172,
         1508,
         1508,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         1507,
         1507,
         1508,
         172,
         1508,
         1508,
         1511,
         1522,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1522,
         1522,
         1508,
         1508,
         1508,
         1508,
         1512,
         1522,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1522,
         1522,
         1508,
         1508,
         1508,
         172,
         1513,
         1507,
         172,
         1508,
         1508,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         1507,
         1507,
         1508,
         172,
         1508,
         172,
         1514,
         1507,
         172,
         1508,
         1508,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         1507,
         1507,
         1508,
         172,
         1508,
         172,
         1515,
         1507,
         172,
         1508,
         1508,
         172,
         172,
         1523,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         1507,
         2302,
         1508,
         172,
         1508,
         172,
         1516,
         1507,
         172,
         1508,
         1508,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         1507,
         1507,
         1508,
         172,
         1508,
         172,
         173,
         1507,
         172,
         1508,
         1508,
         172,
         172,
         172,
         172,
         1524,
         172,
         172,
         172,
         172,
         172,
         172,
         1507,
         1507,
         1508,
         172,
         1508,
         172,
         1517,
         1507,
         172,
         1508,
         1508,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         1507,
         1507,
         1508,
         172,
         1508,
         172,
         1518,
         1507,
         172,
         1508,
         1508,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         1507,
         1507,
         1508,
         172,
         1508,
         172,
         1519,
         1507,
         172,
         1508,
         1508,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         1507,
         1507,
         1508,
         172,
         1508,
         172,
         1520,
         1507,
         172,
         1508,
         1508,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         1507,
         1507,
         1508,
         172,
         1508,
         172,
         2297,
         1507,
         172,
         1508,
         1508,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         1507,
         1507,
         1508,
         172,
         1508,
         172,
         2298,
         1507,
         172,
         1508,
         1508,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         1507,
         1507,
         1508,
         172,
         1508,
         1507,
         2299,
         1521,
         1507,
         1522,
         1522,
         1507,
         1507,
         1507,
         1507,
         1507,
         1507,
         1507,
         1507,
         1507,
         1507,
         1507,
         1521,
         1521,
         1522,
         1507,
         1522,
         1507,
         1509,
         1521,
         1507,
         1522,
         1522,
         1507,
         1507,
         2302,
         1507,
         1507,
         1507,
         1507,
         1507,
         1507,
         1507,
         1507,
         1521,
         2303,
         1522,
         1507,
         1522,
         1508,
         2300,
         1522,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1522,
         1522,
         1508,
         1508,
         1508,
         172,
         2301,
         1507,
         172,
         1508,
         1508,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         172,
         1507,
         1507,
         1508,
         172,
         1508,
         1508,
         2865,
         1522,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1508,
         1522,
         1522,
         1508,
         1508,
         1508,
      },
   },
   [nir_op_ilt] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         5,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         4,
         4,
         4,
         4,
         4,
         3,
         3,
         3,
         3,
         3,
         3,
         5,
         5,
         5,
         6,
         6,
         6,
         6,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         4,
         4,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         5,
         5,
         5,
         6,
         6,
         6,
         6,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         4,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 9,
      .table = (const uint16_t []) {
      
         174,
         175,
         174,
         1863,
         1863,
         1864,
         1864,
         174,
         174,
         174,
         175,
         1865,
         1863,
         1866,
         1864,
         1867,
         1868,
         174,
         174,
         1869,
         174,
         1863,
         1863,
         1864,
         1864,
         174,
         174,
         1870,
         1871,
         1870,
         1872,
         1872,
         1873,
         1873,
         1870,
         1870,
         1870,
         1874,
         1870,
         1872,
         1872,
         1873,
         1873,
         1870,
         1870,
         1875,
         1876,
         1875,
         1877,
         1877,
         1878,
         1878,
         1875,
         1875,
         1875,
         1879,
         1875,
         1877,
         1877,
         1878,
         1878,
         1875,
         1875,
         174,
         1880,
         174,
         1863,
         1863,
         1864,
         1864,
         1881,
         174,
         174,
         2652,
         174,
         1863,
         1863,
         1864,
         1864,
         174,
         174,
      },
   },
   [nir_op_ult] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         7,
         8,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         6,
         6,
         13,
         7,
         7,
         12,
         12,
         12,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         9,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         7,
         7,
         12,
         12,
         6,
         6,
         6,
         6,
         6,
         8,
         8,
         8,
         9,
         9,
         9,
         8,
         8,
         8,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         0,
         11,
         6,
         6,
         7,
         7,
         12,
         12,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         4,
         4,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         3,
         0,
         3,
         3,
         0,
         0,
         4,
         4,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         4,
         4,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         4,
         4,
      },
      
      .num_filtered_states = 14,
      .table = (const uint16_t []) {
      
         176,
         177,
         176,
         1711,
         176,
         176,
         1712,
         1712,
         1713,
         1713,
         176,
         176,
         2653,
         2653,
         178,
         179,
         178,
         1714,
         178,
         1715,
         1716,
         1717,
         1718,
         1719,
         1720,
         2654,
         2655,
         2656,
         1721,
         1722,
         1721,
         1723,
         1721,
         1721,
         1724,
         1724,
         1725,
         1725,
         1721,
         1721,
         2657,
         2657,
         176,
         177,
         176,
         1711,
         176,
         176,
         1712,
         1712,
         1713,
         1713,
         176,
         176,
         2653,
         2653,
         1726,
         1727,
         1726,
         1728,
         1726,
         1726,
         1729,
         1729,
         1730,
         1730,
         1726,
         1726,
         2658,
         2658,
         176,
         1731,
         176,
         1711,
         176,
         176,
         1712,
         1712,
         1713,
         1713,
         176,
         176,
         2653,
         2653,
         1732,
         1733,
         1732,
         1734,
         1732,
         1732,
         1735,
         1735,
         1736,
         1736,
         1732,
         1732,
         2659,
         2659,
         1732,
         1737,
         1732,
         1734,
         1732,
         1732,
         1735,
         1735,
         1736,
         1736,
         1732,
         1732,
         2659,
         2659,
         1738,
         1739,
         1738,
         1740,
         1738,
         1738,
         1741,
         1741,
         1742,
         1742,
         1738,
         1738,
         2660,
         2660,
         1738,
         1743,
         1738,
         1740,
         1738,
         1738,
         1741,
         1741,
         1742,
         1742,
         1738,
         1738,
         2660,
         2660,
         176,
         1744,
         176,
         1711,
         176,
         176,
         1712,
         1712,
         1713,
         1713,
         1745,
         176,
         2653,
         2653,
         176,
         177,
         176,
         1711,
         176,
         176,
         1712,
         1712,
         1713,
         1713,
         176,
         176,
         2653,
         2653,
         1732,
         1733,
         1732,
         1734,
         1732,
         1732,
         1735,
         1735,
         1736,
         1736,
         1732,
         1732,
         2659,
         2659,
         1732,
         1737,
         1732,
         1734,
         1732,
         1732,
         1735,
         1735,
         1736,
         1736,
         1732,
         1732,
         2659,
         2659,
      },
   },
   [nir_op_ige] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         5,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         9,
         10,
         11,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         4,
         4,
         4,
         4,
         4,
         3,
         3,
         3,
         3,
         3,
         3,
         5,
         5,
         5,
         6,
         6,
         6,
         6,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         4,
         4,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         5,
         5,
         5,
         6,
         6,
         6,
         6,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         4,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 14,
      .table = (const uint16_t []) {
      
         180,
         181,
         180,
         1882,
         1882,
         1883,
         1883,
         180,
         180,
         180,
         180,
         180,
         180,
         180,
         180,
         181,
         1884,
         1882,
         1885,
         1883,
         1886,
         1887,
         180,
         180,
         180,
         180,
         180,
         180,
         180,
         1888,
         180,
         1882,
         1882,
         1883,
         1883,
         180,
         180,
         180,
         180,
         180,
         180,
         180,
         1889,
         1890,
         1889,
         1891,
         1891,
         1892,
         1892,
         1889,
         1889,
         1889,
         1889,
         1889,
         1889,
         1889,
         1889,
         1893,
         1889,
         1891,
         1891,
         1892,
         1892,
         1889,
         1889,
         1889,
         1889,
         1889,
         1889,
         1889,
         1894,
         1895,
         1894,
         1896,
         1896,
         1897,
         1897,
         1894,
         1894,
         1894,
         1894,
         1894,
         1894,
         1894,
         1894,
         1898,
         1894,
         1896,
         1896,
         1897,
         1897,
         1894,
         1894,
         1894,
         1894,
         1894,
         1894,
         1894,
         180,
         1899,
         180,
         1882,
         1882,
         1883,
         1883,
         1900,
         180,
         180,
         180,
         180,
         180,
         180,
         180,
         1901,
         180,
         1882,
         1882,
         1883,
         1883,
         180,
         180,
         180,
         180,
         180,
         180,
         180,
         180,
         1902,
         180,
         1882,
         1882,
         1883,
         1883,
         180,
         180,
         180,
         180,
         180,
         180,
         180,
         180,
         1903,
         180,
         1882,
         1882,
         1883,
         1883,
         180,
         180,
         180,
         180,
         180,
         180,
         180,
         180,
         1904,
         180,
         1882,
         1882,
         1883,
         1883,
         180,
         180,
         180,
         180,
         180,
         180,
         180,
         180,
         1905,
         180,
         1882,
         1882,
         1883,
         1883,
         180,
         180,
         180,
         180,
         180,
         180,
         180,
         180,
         2661,
         180,
         1882,
         1882,
         1883,
         1883,
         180,
         180,
         180,
         180,
         180,
         180,
         180,
      },
   },
   [nir_op_uge] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         7,
         8,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         7,
         7,
         7,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         9,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         7,
         7,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         8,
         8,
         8,
         9,
         9,
         9,
         8,
         8,
         8,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         0,
         0,
         6,
         6,
         7,
         7,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         4,
         4,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         3,
         0,
         3,
         3,
         0,
         0,
         4,
         4,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         4,
         4,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         4,
         4,
      },
      
      .num_filtered_states = 11,
      .table = (const uint16_t []) {
      
         182,
         182,
         182,
         1746,
         182,
         182,
         1747,
         1747,
         1748,
         1748,
         182,
         183,
         183,
         183,
         1749,
         183,
         1750,
         1751,
         1752,
         1753,
         1754,
         1755,
         1756,
         1756,
         1756,
         1757,
         1756,
         1756,
         1758,
         1758,
         1759,
         1759,
         1756,
         182,
         182,
         182,
         1746,
         182,
         182,
         1747,
         1747,
         1748,
         1748,
         182,
         1760,
         1760,
         1760,
         1761,
         1760,
         1760,
         1762,
         1762,
         1763,
         1763,
         1760,
         182,
         1764,
         182,
         1746,
         182,
         182,
         1747,
         1747,
         1748,
         1748,
         182,
         1765,
         1765,
         1765,
         1766,
         1765,
         1765,
         1767,
         1767,
         1768,
         1768,
         1765,
         1765,
         1769,
         1765,
         1766,
         1765,
         1765,
         1767,
         1767,
         1768,
         1768,
         1765,
         1770,
         1770,
         1770,
         1771,
         1770,
         1770,
         1772,
         1772,
         1773,
         1773,
         1770,
         1770,
         1774,
         1770,
         1771,
         1770,
         1770,
         1772,
         1772,
         1773,
         1773,
         1770,
         182,
         1775,
         182,
         1746,
         182,
         182,
         1747,
         1747,
         1748,
         1748,
         1776,
      },
   },
   [nir_op_fmin] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         0,
         7,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         9,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         10,
         2,
         2,
         11,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         12,
         6,
         6,
         6,
         6,
         6,
         12,
         6,
         6,
         6,
         12,
         6,
         12,
         6,
         6,
         12,
         7,
         7,
         7,
         8,
         8,
         8,
         8,
         8,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         12,
         6,
         6,
         12,
         12,
         6,
         6,
         12,
         12,
         12,
         6,
         6,
         12,
         6,
         12,
         6,
         12,
         7,
         7,
         13,
         7,
         8,
         8,
         8,
         14,
         14,
         8,
         13,
         7,
         7,
         7,
         13,
         7,
         7,
         7,
         13,
         7,
         13,
         7,
         7,
         13,
         7,
         13,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         2,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         12,
         12,
         6,
         6,
         6,
         6,
         6,
         12,
         12,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 16,
      .table = (const uint16_t []) {
      
         184,
         185,
         1525,
         1526,
         184,
         184,
         1527,
         1528,
         1528,
         2546,
         1525,
         2547,
         2548,
         2815,
         2815,
         2816,
         185,
         185,
         1529,
         1530,
         1531,
         1532,
         1533,
         1534,
         1535,
         2549,
         1529,
         2550,
         2551,
         2817,
         2818,
         2819,
         1525,
         1529,
         1525,
         1536,
         1525,
         1525,
         1537,
         1538,
         1538,
         2546,
         1525,
         2547,
         2552,
         2820,
         2820,
         2816,
         1526,
         1530,
         1536,
         1526,
         1526,
         1526,
         1539,
         1540,
         1540,
         2553,
         1536,
         2554,
         2555,
         2821,
         2821,
         2822,
         184,
         1531,
         1525,
         1526,
         184,
         184,
         1527,
         1528,
         1528,
         2546,
         1525,
         2547,
         2548,
         2815,
         2815,
         2816,
         184,
         1532,
         1525,
         1526,
         184,
         1541,
         1527,
         1528,
         1528,
         2546,
         1525,
         2547,
         2548,
         2815,
         2815,
         2816,
         1527,
         1533,
         1537,
         1539,
         1527,
         1527,
         1527,
         1542,
         1542,
         2556,
         1537,
         2557,
         2548,
         2823,
         2823,
         2824,
         1528,
         1534,
         1538,
         1540,
         1528,
         1528,
         1542,
         1528,
         1528,
         2558,
         1538,
         2559,
         2560,
         2815,
         2815,
         2825,
         1528,
         1535,
         1538,
         1540,
         1528,
         1528,
         1542,
         1528,
         1528,
         2558,
         1538,
         2559,
         2560,
         2815,
         2815,
         2825,
         2546,
         2549,
         2546,
         2553,
         2546,
         2546,
         2556,
         2558,
         2558,
         2546,
         2546,
         2561,
         2562,
         2826,
         2826,
         2827,
         1525,
         1529,
         1525,
         1536,
         1525,
         1525,
         1537,
         1538,
         1538,
         2546,
         2563,
         2547,
         2552,
         2820,
         2820,
         2816,
         2547,
         2550,
         2547,
         2554,
         2547,
         2547,
         2557,
         2559,
         2559,
         2561,
         2547,
         2547,
         2564,
         2828,
         2828,
         2829,
         2548,
         2551,
         2552,
         2555,
         2548,
         2548,
         2548,
         2560,
         2560,
         2562,
         2552,
         2564,
         2548,
         2830,
         2830,
         2831,
         2815,
         2817,
         2820,
         2821,
         2815,
         2815,
         2823,
         2815,
         2815,
         2826,
         2820,
         2828,
         2830,
         2815,
         2815,
         2832,
         2815,
         2818,
         2820,
         2821,
         2815,
         2815,
         2823,
         2815,
         2815,
         2826,
         2820,
         2828,
         2830,
         2815,
         2815,
         2832,
         2816,
         2819,
         2816,
         2822,
         2816,
         2816,
         2824,
         2825,
         2825,
         2827,
         2816,
         2829,
         2831,
         2832,
         2832,
         2816,
      },
   },
   [nir_search_op_i2f] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         3,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 6,
      .table = (const uint16_t []) {
      
         186,
         1925,
         1926,
         1927,
         1928,
         2908,
      },
   },
   [nir_op_fmax] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         8,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         9,
         2,
         10,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         6,
         6,
         6,
         6,
         6,
         13,
         13,
         0,
         0,
         12,
         0,
         12,
         0,
         12,
         7,
         7,
         14,
         7,
         7,
         7,
         7,
         14,
         7,
         14,
         14,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         14,
         7,
         7,
         7,
         7,
         7,
         14,
         7,
         14,
         7,
         7,
         7,
         14,
         14,
         14,
         14,
         7,
         7,
         14,
         7,
         7,
         14,
         14,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         11,
         11,
         11,
         11,
         11,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         13,
         13,
         6,
         12,
         12,
         0,
         12,
         0,
         12,
         12,
         0,
         12,
         0,
         12,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 17,
      .table = (const uint16_t []) {
      
         187,
         188,
         1543,
         1544,
         187,
         187,
         187,
         1545,
         2565,
         1543,
         2566,
         187,
         2567,
         2567,
         2568,
         187,
         187,
         188,
         188,
         1546,
         1547,
         1548,
         188,
         1549,
         1550,
         2569,
         1546,
         2570,
         2571,
         2572,
         2573,
         2574,
         2893,
         2894,
         1543,
         1546,
         1543,
         1551,
         1543,
         1543,
         1543,
         1552,
         2565,
         1543,
         2566,
         1543,
         2575,
         2575,
         2576,
         1543,
         1543,
         1544,
         1547,
         1551,
         1544,
         1544,
         1544,
         1544,
         1553,
         2577,
         1551,
         2578,
         1544,
         2579,
         2579,
         2580,
         1544,
         1544,
         187,
         1548,
         1543,
         1544,
         187,
         187,
         187,
         1545,
         2565,
         1543,
         2566,
         187,
         2567,
         2567,
         2568,
         187,
         187,
         187,
         188,
         1543,
         1544,
         187,
         1554,
         187,
         1545,
         2565,
         1543,
         2566,
         187,
         2567,
         2567,
         2568,
         187,
         187,
         187,
         1549,
         1543,
         1544,
         187,
         187,
         187,
         1545,
         2565,
         1543,
         2566,
         187,
         2567,
         2567,
         2568,
         187,
         187,
         1545,
         1550,
         1552,
         1553,
         1545,
         1545,
         1545,
         1545,
         2581,
         1552,
         2582,
         1545,
         2583,
         2583,
         2568,
         1545,
         1545,
         2565,
         2569,
         2565,
         2577,
         2565,
         2565,
         2565,
         2581,
         2565,
         2565,
         2584,
         2565,
         2585,
         2585,
         2586,
         2565,
         2565,
         1543,
         1546,
         1543,
         1551,
         1543,
         1543,
         1543,
         1552,
         2565,
         2587,
         2566,
         1543,
         2575,
         2575,
         2576,
         1543,
         1543,
         2566,
         2570,
         2566,
         2578,
         2566,
         2566,
         2566,
         2582,
         2584,
         2566,
         2566,
         2566,
         2588,
         2588,
         2589,
         2566,
         2566,
         187,
         2571,
         1543,
         1544,
         187,
         187,
         187,
         1545,
         2565,
         1543,
         2566,
         187,
         2567,
         2567,
         2568,
         187,
         187,
         2567,
         2572,
         2575,
         2579,
         2567,
         2567,
         2567,
         2583,
         2585,
         2575,
         2588,
         2567,
         2567,
         2567,
         2590,
         2567,
         2567,
         2567,
         2573,
         2575,
         2579,
         2567,
         2567,
         2567,
         2583,
         2585,
         2575,
         2588,
         2567,
         2567,
         2567,
         2590,
         2567,
         2567,
         2568,
         2574,
         2576,
         2580,
         2568,
         2568,
         2568,
         2568,
         2586,
         2576,
         2589,
         2568,
         2590,
         2590,
         2568,
         2568,
         2568,
         187,
         2893,
         1543,
         1544,
         187,
         187,
         187,
         1545,
         2565,
         1543,
         2566,
         187,
         2567,
         2567,
         2568,
         187,
         187,
         187,
         2894,
         1543,
         1544,
         187,
         187,
         187,
         1545,
         2565,
         1543,
         2566,
         187,
         2567,
         2567,
         2568,
         187,
         187,
      },
   },
   [nir_op_imin] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         5,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         7,
         2,
         2,
         8,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         9,
         4,
         4,
         4,
         9,
         4,
         4,
         4,
         9,
         4,
         9,
         4,
         9,
         5,
         5,
         5,
         6,
         6,
         6,
         6,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         9,
         4,
         9,
         9,
         4,
         9,
         9,
         4,
         9,
         9,
         4,
         5,
         10,
         5,
         6,
         11,
         11,
         6,
         10,
         5,
         5,
         10,
         5,
         5,
         10,
         10,
         5,
         5,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         4,
         9,
         4,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 12,
      .table = (const uint16_t []) {
      
         189,
         190,
         1621,
         1622,
         1623,
         1624,
         1624,
         2619,
         1621,
         2620,
         2869,
         2869,
         190,
         190,
         1625,
         1626,
         1627,
         1628,
         1629,
         2621,
         1625,
         2622,
         2870,
         2871,
         1621,
         1625,
         1621,
         1630,
         1631,
         1632,
         1632,
         2619,
         1621,
         2623,
         2872,
         2872,
         1622,
         1626,
         1630,
         1622,
         1633,
         1634,
         1634,
         2624,
         1630,
         2625,
         2873,
         2873,
         1623,
         1627,
         1631,
         1633,
         1623,
         1635,
         1635,
         2626,
         1631,
         2620,
         2874,
         2874,
         1624,
         1628,
         1632,
         1634,
         1635,
         1624,
         1624,
         2627,
         1632,
         2628,
         2869,
         2869,
         1624,
         1629,
         1632,
         1634,
         1635,
         1624,
         1624,
         2627,
         1632,
         2628,
         2869,
         2869,
         2619,
         2621,
         2619,
         2624,
         2626,
         2627,
         2627,
         2619,
         2619,
         2629,
         2875,
         2875,
         1621,
         1625,
         1621,
         1630,
         1631,
         1632,
         1632,
         2619,
         2630,
         2623,
         2872,
         2872,
         2620,
         2622,
         2623,
         2625,
         2620,
         2628,
         2628,
         2629,
         2623,
         2620,
         2876,
         2876,
         2869,
         2870,
         2872,
         2873,
         2874,
         2869,
         2869,
         2875,
         2872,
         2876,
         2869,
         2869,
         2869,
         2871,
         2872,
         2873,
         2874,
         2869,
         2869,
         2875,
         2872,
         2876,
         2869,
         2869,
      },
   },
   [nir_op_imax] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         6,
         2,
         2,
         7,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         4,
         4,
         4,
         9,
         9,
         0,
         0,
         8,
         0,
         8,
         8,
         5,
         5,
         10,
         5,
         5,
         5,
         10,
         5,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         0,
         0,
         0,
         0,
         8,
         8,
         0,
         0,
         5,
         5,
         10,
         5,
         5,
         5,
         10,
         5,
         10,
         5,
         5,
         10,
         10,
         10,
         5,
         10,
         5,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         9,
         9,
         8,
         8,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 11,
      .table = (const uint16_t []) {
      
         191,
         192,
         1636,
         1637,
         191,
         1638,
         2631,
         1636,
         2632,
         2632,
         2633,
         192,
         192,
         1639,
         1640,
         1641,
         1642,
         2634,
         1639,
         2635,
         2636,
         2637,
         1636,
         1639,
         1636,
         1643,
         1636,
         1644,
         2631,
         1636,
         2638,
         2638,
         2639,
         1637,
         1640,
         1643,
         1637,
         1637,
         1645,
         2640,
         1643,
         2641,
         2641,
         2642,
         191,
         1641,
         1636,
         1637,
         191,
         1638,
         2631,
         1636,
         2632,
         2632,
         2633,
         1638,
         1642,
         1644,
         1645,
         1638,
         1638,
         2643,
         1644,
         2644,
         2644,
         2633,
         2631,
         2634,
         2631,
         2640,
         2631,
         2643,
         2631,
         2631,
         2645,
         2645,
         2646,
         1636,
         1639,
         1636,
         1643,
         1636,
         1644,
         2631,
         2647,
         2638,
         2638,
         2639,
         2632,
         2635,
         2638,
         2641,
         2632,
         2644,
         2645,
         2638,
         2632,
         2632,
         2648,
         2632,
         2636,
         2638,
         2641,
         2632,
         2644,
         2645,
         2638,
         2632,
         2632,
         2648,
         2633,
         2637,
         2639,
         2642,
         2633,
         2633,
         2646,
         2639,
         2648,
         2648,
         2633,
      },
   },
   [nir_op_umin] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         8,
         5,
         5,
         8,
         5,
         5,
         8,
         5,
         5,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         8,
         5,
         8,
         5,
         8,
         8,
         5,
         8,
         5,
         8,
         6,
         10,
         6,
         6,
         10,
         6,
         6,
         10,
         6,
         10,
         6,
         10,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         5,
         5,
         5,
         5,
         5,
         5,
         8,
         8,
         5,
         5,
         5,
         8,
         8,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         3,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 11,
      .table = (const uint16_t []) {
      
         193,
         194,
         1779,
         1779,
         193,
         1780,
         1781,
         2662,
         2663,
         2728,
         2729,
         194,
         194,
         1782,
         1782,
         194,
         1783,
         1784,
         2664,
         2665,
         2730,
         2731,
         1779,
         1782,
         1779,
         1785,
         1779,
         1786,
         1787,
         2666,
         2667,
         2732,
         2733,
         1779,
         1782,
         1785,
         1785,
         1779,
         1786,
         1787,
         2666,
         2667,
         2732,
         2733,
         193,
         194,
         1779,
         1779,
         1788,
         1780,
         1781,
         2662,
         2663,
         2728,
         2729,
         1780,
         1783,
         1786,
         1786,
         1780,
         1780,
         1789,
         2668,
         2663,
         2734,
         2735,
         1781,
         1784,
         1787,
         1787,
         1781,
         1789,
         1781,
         2669,
         2670,
         2736,
         2729,
         2662,
         2664,
         2666,
         2666,
         2662,
         2668,
         2669,
         2671,
         2672,
         2737,
         2738,
         2663,
         2665,
         2667,
         2667,
         2663,
         2663,
         2670,
         2672,
         2663,
         2739,
         2740,
         2728,
         2730,
         2732,
         2732,
         2728,
         2734,
         2736,
         2737,
         2739,
         2728,
         2741,
         2729,
         2731,
         2733,
         2733,
         2729,
         2735,
         2729,
         2738,
         2740,
         2741,
         2729,
      },
   },
   [nir_op_umax] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         5,
         0,
         0,
         5,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         0,
         0,
         3,
         3,
         6,
         3,
         3,
         6,
         6,
         6,
         3,
         3,
         6,
         6,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         5,
         0,
         5,
         0,
         5,
         0,
         5,
         5,
         0,
         5,
         0,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 7,
      .table = (const uint16_t []) {
      
         195,
         196,
         195,
         1929,
         2673,
         2674,
         2675,
         196,
         196,
         196,
         1930,
         2676,
         2677,
         2678,
         195,
         196,
         1931,
         1929,
         2673,
         2674,
         2675,
         1929,
         1930,
         1929,
         1929,
         2679,
         2680,
         2675,
         2673,
         2676,
         2673,
         2679,
         2681,
         2682,
         2683,
         2674,
         2677,
         2674,
         2680,
         2682,
         2674,
         2684,
         2675,
         2678,
         2675,
         2675,
         2683,
         2684,
         2675,
      },
   },
   [nir_op_extract_u8] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         2,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         3,
         0,
         4,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         0,
         2,
         0,
         0,
         0,
         2,
         0,
         0,
         2,
         2,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         3,
         3,
         0,
         0,
         4,
         0,
         4,
         0,
         4,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         0,
         0,
         0,
         0,
         2,
         3,
         0,
         0,
         2,
         3,
         0,
         0,
         0,
         2,
         3,
         0,
         0,
         2,
         3,
         0,
         0,
         0,
         2,
         3,
         0,
         0,
      },
      
      .num_filtered_states = 10,
      .table = (const uint16_t []) {
      
         197,
         198,
         197,
         197,
         197,
         197,
         197,
         197,
         197,
         197,
         197,
         198,
         197,
         197,
         197,
         197,
         197,
         197,
         197,
         197,
         197,
         1260,
         197,
         197,
         197,
         197,
         197,
         197,
         197,
         197,
         197,
         1261,
         197,
         197,
         197,
         197,
         197,
         197,
         197,
         197,
         197,
         1262,
         197,
         197,
         197,
         197,
         197,
         197,
         197,
         197,
         197,
         1263,
         197,
         197,
         197,
         197,
         197,
         197,
         197,
         197,
         197,
         1264,
         197,
         197,
         197,
         197,
         197,
         197,
         197,
         197,
         197,
         1265,
         197,
         197,
         197,
         197,
         197,
         197,
         197,
         197,
         197,
         1266,
         197,
         197,
         197,
         197,
         197,
         197,
         197,
         197,
         197,
         2688,
         197,
         197,
         197,
         197,
         197,
         197,
         197,
         197,
      },
   },
   [nir_search_op_f2i] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 7,
      .table = (const uint16_t []) {
      
         199,
         1646,
         1647,
         1648,
         1649,
         1650,
         1651,
      },
   },
   [nir_search_op_f2u] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 7,
      .table = (const uint16_t []) {
      
         200,
         1652,
         1653,
         1654,
         1655,
         1656,
         1657,
      },
   },
   [nir_search_op_f2f] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         0,
         4,
         4,
         4,
         0,
         0,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 11,
      .table = (const uint16_t []) {
      
         201,
         1658,
         1659,
         1660,
         1661,
         1662,
         1663,
         1664,
         1665,
         1666,
         1667,
      },
   },
   [nir_search_op_i2i] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         3,
         3,
         3,
         3,
         0,
         4,
         4,
         4,
         0,
         0,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         7,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         7,
         7,
         7,
         7,
         7,
         7,
         11,
         7,
         7,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 12,
      .table = (const uint16_t []) {
      
         202,
         1792,
         1793,
         1794,
         1795,
         1796,
         1797,
         1798,
         1799,
         1800,
         1801,
         2726,
      },
   },
   [nir_search_op_u2u] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         3,
         3,
         3,
         3,
         0,
         4,
         4,
         4,
         0,
         0,
         5,
         5,
         5,
         5,
         0,
         0,
         6,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         0,
         0,
         0,
         8,
         9,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         0,
         0,
         5,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         14,
         9,
         9,
         9,
         9,
         9,
         15,
         9,
         9,
         9,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         11,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         17,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         18,
         9,
         9,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         1,
         0,
         1,
         1,
         0,
         6,
         0,
         0,
         0,
         6,
         1,
         1,
         1,
         0,
         6,
         0,
         0,
         0,
         6,
         1,
         1,
         1,
         0,
         6,
         0,
         0,
      },
      
      .num_filtered_states = 19,
      .table = (const uint16_t []) {
      
         203,
         1802,
         1803,
         1804,
         1805,
         1806,
         1807,
         1808,
         1809,
         1810,
         1811,
         1812,
         1813,
         1814,
         2724,
         2725,
         2806,
         2807,
         2808,
      },
   },
   [nir_op_f2fmp] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         0,
         4,
         4,
         4,
         0,
         0,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 8,
      .table = (const uint16_t []) {
      
         204,
         1668,
         1669,
         1670,
         1671,
         1672,
         1673,
         1674,
      },
   },
   [nir_op_i2imp] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         2,
         2,
         2,
         2,
         0,
         3,
         3,
         3,
         0,
         0,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         0,
         7,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 9,
      .table = (const uint16_t []) {
      
         205,
         1932,
         1933,
         1934,
         1935,
         1936,
         1937,
         1938,
         1939,
      },
   },
   [nir_op_extract_i8] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         2,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         0,
         2,
         0,
         0,
         0,
         2,
         0,
         0,
         2,
         2,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         206,
         207,
         206,
         206,
         206,
         207,
         206,
         206,
         206,
         1267,
         206,
         206,
         206,
         1268,
         206,
         206,
      },
   },
   [nir_op_extract_i16] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         208,
         209,
         208,
         209,
      },
   },
   [nir_op_extract_u16] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 5,
      .table = (const uint16_t []) {
      
         210,
         211,
         210,
         210,
         210,
         210,
         211,
         210,
         210,
         210,
         210,
         1957,
         210,
         210,
         210,
         210,
         1958,
         210,
         210,
         210,
         210,
         1959,
         210,
         210,
         210,
      },
   },
   [nir_op_ixor] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         2,
         3,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         5,
         4,
         5,
         6,
         7,
         6,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         11,
         12,
         13,
         14,
         15,
         16,
         17,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         2,
         3,
         2,
         2,
         2,
         3,
         2,
         2,
         3,
         3,
         2,
         3,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         16,
         16,
         16,
         16,
         17,
         17,
         17,
         17,
         17,
         17,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         4,
         4,
         4,
         4,
         4,
         5,
         4,
         4,
         5,
         5,
         6,
         6,
         7,
         6,
         7,
         6,
         7,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         11,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         18,
         19,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         20,
         21,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         15,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         4,
         4,
         4,
         4,
         4,
         4,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         16,
         16,
         16,
         16,
         17,
         17,
         17,
         17,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         21,
         21,
         19,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         5,
         0,
         0,
         0,
         0,
         3,
         5,
         0,
         0,
         3,
         5,
         0,
         0,
         0,
         3,
         5,
         0,
         0,
         3,
         5,
         0,
         0,
         0,
         3,
         5,
         0,
         0,
      },
      
      .num_filtered_states = 22,
      .table = (const uint16_t []) {
      
         212,
         213,
         212,
         212,
         212,
         212,
         212,
         212,
         212,
         1215,
         212,
         1216,
         212,
         1217,
         212,
         1218,
         1219,
         1219,
         2689,
         2690,
         2691,
         2692,
         213,
         213,
         213,
         213,
         213,
         213,
         213,
         213,
         213,
         1220,
         213,
         1221,
         213,
         1222,
         213,
         1223,
         1224,
         1225,
         2693,
         2694,
         2695,
         2696,
         212,
         213,
         212,
         1226,
         212,
         212,
         212,
         212,
         212,
         1215,
         212,
         1216,
         212,
         1217,
         212,
         1218,
         1219,
         1219,
         2689,
         2690,
         2691,
         2692,
         212,
         213,
         1226,
         1226,
         212,
         212,
         212,
         212,
         212,
         1215,
         212,
         1216,
         212,
         1217,
         212,
         1218,
         1219,
         1219,
         2689,
         2690,
         2691,
         2692,
         212,
         213,
         212,
         212,
         212,
         1227,
         212,
         212,
         212,
         1215,
         212,
         1216,
         212,
         1217,
         212,
         1218,
         1219,
         1219,
         2689,
         2690,
         2691,
         2692,
         212,
         213,
         212,
         212,
         1227,
         1227,
         212,
         212,
         212,
         1215,
         212,
         1216,
         212,
         1217,
         212,
         1218,
         1219,
         1219,
         2689,
         2690,
         2691,
         2692,
         212,
         213,
         212,
         212,
         212,
         212,
         212,
         1228,
         212,
         1215,
         212,
         1216,
         212,
         1217,
         212,
         1218,
         1219,
         1219,
         2689,
         2690,
         2691,
         2692,
         212,
         213,
         212,
         212,
         212,
         212,
         1228,
         1228,
         212,
         1215,
         212,
         1216,
         212,
         1217,
         212,
         1218,
         1219,
         1219,
         2689,
         2690,
         2691,
         2692,
         212,
         213,
         212,
         212,
         212,
         212,
         212,
         212,
         212,
         1229,
         212,
         1216,
         212,
         1217,
         212,
         1218,
         1219,
         1219,
         2689,
         2690,
         2691,
         2692,
         1215,
         1220,
         1215,
         1215,
         1215,
         1215,
         1215,
         1215,
         1229,
         1229,
         1215,
         1230,
         1215,
         1231,
         1215,
         1232,
         1233,
         1233,
         2697,
         2698,
         2699,
         2700,
         212,
         213,
         212,
         212,
         212,
         212,
         212,
         212,
         212,
         1215,
         212,
         1234,
         212,
         1217,
         212,
         1218,
         1219,
         1219,
         2689,
         2690,
         2691,
         2692,
         1216,
         1221,
         1216,
         1216,
         1216,
         1216,
         1216,
         1216,
         1216,
         1230,
         1234,
         1234,
         1216,
         1235,
         1216,
         1236,
         1237,
         1237,
         2701,
         2702,
         2703,
         2704,
         212,
         213,
         212,
         212,
         212,
         212,
         212,
         212,
         212,
         1215,
         212,
         1216,
         212,
         1238,
         212,
         1218,
         1219,
         1219,
         2689,
         2690,
         2691,
         2692,
         1217,
         1222,
         1217,
         1217,
         1217,
         1217,
         1217,
         1217,
         1217,
         1231,
         1217,
         1235,
         1238,
         1238,
         1217,
         1239,
         1240,
         1240,
         2705,
         2706,
         2707,
         2708,
         212,
         213,
         212,
         212,
         212,
         212,
         212,
         212,
         212,
         1215,
         212,
         1216,
         212,
         1217,
         212,
         1241,
         1219,
         1219,
         2689,
         2690,
         2691,
         2692,
         1218,
         1223,
         1218,
         1218,
         1218,
         1218,
         1218,
         1218,
         1218,
         1232,
         1218,
         1236,
         1218,
         1239,
         1241,
         1241,
         1242,
         1242,
         2709,
         2710,
         2711,
         2712,
         1219,
         1224,
         1219,
         1219,
         1219,
         1219,
         1219,
         1219,
         1219,
         1233,
         1219,
         1237,
         1219,
         1240,
         1219,
         1242,
         1219,
         1219,
         2713,
         2714,
         2715,
         2716,
         1219,
         1225,
         1219,
         1219,
         1219,
         1219,
         1219,
         1219,
         1219,
         1233,
         1219,
         1237,
         1219,
         1240,
         1219,
         1242,
         1219,
         1219,
         2713,
         2714,
         2715,
         2716,
         2689,
         2693,
         2689,
         2689,
         2689,
         2689,
         2689,
         2689,
         2689,
         2697,
         2689,
         2701,
         2689,
         2705,
         2689,
         2709,
         2713,
         2713,
         2689,
         2717,
         2718,
         2719,
         2690,
         2694,
         2690,
         2690,
         2690,
         2690,
         2690,
         2690,
         2690,
         2698,
         2690,
         2702,
         2690,
         2706,
         2690,
         2710,
         2714,
         2714,
         2717,
         2690,
         2720,
         2721,
         2691,
         2695,
         2691,
         2691,
         2691,
         2691,
         2691,
         2691,
         2691,
         2699,
         2691,
         2703,
         2691,
         2707,
         2691,
         2711,
         2715,
         2715,
         2718,
         2720,
         2691,
         2722,
         2692,
         2696,
         2692,
         2692,
         2692,
         2692,
         2692,
         2692,
         2692,
         2700,
         2692,
         2704,
         2692,
         2708,
         2692,
         2712,
         2716,
         2716,
         2719,
         2721,
         2722,
         2723,
      },
   },
   [nir_op_slt] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         214,
      },
   },
   [nir_op_sge] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         215,
      },
   },
   [nir_op_seq] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         4,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 6,
      .table = (const uint16_t []) {
      
         216,
         216,
         216,
         216,
         216,
         216,
         216,
         216,
         1960,
         1961,
         1962,
         1963,
         216,
         1960,
         216,
         216,
         216,
         216,
         216,
         1961,
         216,
         216,
         216,
         216,
         216,
         1962,
         216,
         216,
         216,
         216,
         216,
         1963,
         216,
         216,
         216,
         216,
      },
   },
   [nir_op_sne] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         4,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 6,
      .table = (const uint16_t []) {
      
         217,
         217,
         217,
         217,
         217,
         217,
         217,
         217,
         1964,
         1965,
         1966,
         1967,
         217,
         1964,
         217,
         217,
         217,
         217,
         217,
         1965,
         217,
         217,
         217,
         217,
         217,
         1966,
         217,
         217,
         217,
         217,
         217,
         1967,
         217,
         217,
         217,
         217,
      },
   },
   [nir_op_fall_equal2] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         218,
      },
   },
   [nir_op_fall_equal3] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         219,
      },
   },
   [nir_op_fall_equal4] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         220,
      },
   },
   [nir_op_fall_equal8] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         221,
      },
   },
   [nir_op_fall_equal16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         222,
      },
   },
   [nir_op_fany_nequal2] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         223,
      },
   },
   [nir_op_fany_nequal3] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         224,
      },
   },
   [nir_op_fany_nequal4] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         225,
      },
   },
   [nir_op_fany_nequal8] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         226,
      },
   },
   [nir_op_fany_nequal16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         227,
      },
   },
   [nir_op_ball_iequal2] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         228,
         228,
         228,
         228,
         228,
         228,
         1839,
         1840,
         228,
         1839,
         228,
         228,
         228,
         1840,
         228,
         228,
      },
   },
   [nir_op_ball_iequal3] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         229,
         229,
         229,
         229,
         229,
         229,
         1841,
         1842,
         229,
         1841,
         229,
         229,
         229,
         1842,
         229,
         229,
      },
   },
   [nir_op_ball_iequal4] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         230,
         230,
         230,
         230,
         230,
         230,
         1843,
         1844,
         230,
         1843,
         230,
         230,
         230,
         1844,
         230,
         230,
      },
   },
   [nir_op_ball_iequal8] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         231,
         231,
         231,
         231,
         231,
         231,
         1845,
         1846,
         231,
         1845,
         231,
         231,
         231,
         1846,
         231,
         231,
      },
   },
   [nir_op_ball_iequal16] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         232,
         232,
         232,
         232,
         232,
         232,
         1847,
         1848,
         232,
         1847,
         232,
         232,
         232,
         1848,
         232,
         232,
      },
   },
   [nir_op_ball_fequal2] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         233,
      },
   },
   [nir_op_ball_fequal3] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         234,
      },
   },
   [nir_op_ball_fequal4] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         235,
      },
   },
   [nir_op_ball_fequal8] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         236,
      },
   },
   [nir_op_ball_fequal16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         237,
      },
   },
   [nir_op_bany_inequal2] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         238,
         238,
         238,
         238,
         238,
         238,
         1853,
         1854,
         238,
         1853,
         238,
         238,
         238,
         1854,
         238,
         238,
      },
   },
   [nir_op_bany_inequal3] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         239,
         239,
         239,
         239,
         239,
         239,
         1855,
         1856,
         239,
         1855,
         239,
         239,
         239,
         1856,
         239,
         239,
      },
   },
   [nir_op_bany_inequal4] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         240,
         240,
         240,
         240,
         240,
         240,
         1857,
         1858,
         240,
         1857,
         240,
         240,
         240,
         1858,
         240,
         240,
      },
   },
   [nir_op_bany_inequal8] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         241,
         241,
         241,
         241,
         241,
         241,
         1859,
         1860,
         241,
         1859,
         241,
         241,
         241,
         1860,
         241,
         241,
      },
   },
   [nir_op_bany_inequal16] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         242,
         242,
         242,
         242,
         242,
         242,
         1861,
         1862,
         242,
         1861,
         242,
         242,
         242,
         1862,
         242,
         242,
      },
   },
   [nir_op_bany_fnequal2] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         243,
      },
   },
   [nir_op_bany_fnequal3] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         244,
      },
   },
   [nir_op_bany_fnequal4] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         245,
      },
   },
   [nir_op_bany_fnequal8] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         246,
      },
   },
   [nir_op_bany_fnequal16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         247,
      },
   },
   [nir_op_isub] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         248,
         248,
         249,
         249,
      },
   },
   [nir_op_urol] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         250,
      },
   },
   [nir_op_uror] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         251,
      },
   },
   [nir_op_flog2] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         3,
         4,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         4,
         4,
         4,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 6,
      .table = (const uint16_t []) {
      
         252,
         1907,
         1908,
         1909,
         1910,
         1911,
      },
   },
   [nir_op_fpow] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         253,
         254,
         253,
         253,
         255,
         256,
         255,
         255,
         253,
         1968,
         253,
         253,
         253,
         2613,
         253,
         253,
      },
   },
   [nir_op_fsqrt] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         257,
         1912,
      },
   },
   [nir_op_frcp] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         3,
         3,
         3,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 5,
      .table = (const uint16_t []) {
      
         258,
         1913,
         1914,
         1915,
         1916,
      },
   },
   [nir_op_frsq] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         259,
         1917,
      },
   },
   [nir_op_fdiv] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         260,
         260,
         261,
         261,
      },
   },
   [nir_op_fsin] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         262,
      },
   },
   [nir_op_fcos] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         263,
      },
   },
   [nir_op_fcsel] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         264,
      },
   },
   [nir_op_f2imp] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         265,
         1918,
         1919,
      },
   },
   [nir_op_f2ump] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         266,
         1920,
         1921,
      },
   },
   [nir_op_i2fmp] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         267,
         1940,
         1941,
      },
   },
   [nir_op_u2fmp] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         268,
         1942,
         1943,
      },
   },
   [nir_op_fround_even] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         269,
      },
   },
   [nir_op_pack_64_2x32_split] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         270,
         270,
         270,
         270,
         270,
         1270,
         270,
         270,
         270,
      },
   },
   [nir_op_unpack_64_2x32_split_y] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 6,
      .table = (const uint16_t []) {
      
         271,
         1815,
         1816,
         1817,
         1818,
         1819,
      },
   },
   [nir_op_pack_64_2x32] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         272,
         1971,
         2535,
      },
   },
   [nir_op_unpack_64_2x32] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         273,
         1969,
         1970,
      },
   },
   [nir_op_unpack_double_2x32_dxil] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         274,
         1973,
      },
   },
   [nir_op_pack_double_2x32_dxil] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         275,
         1972,
      },
   },
   [nir_op_unpack_32_2x16_split_x] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         276,
         1820,
         1821,
      },
   },
   [nir_op_unpack_32_2x16_split_y] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         277,
         1822,
      },
   },
   [nir_op_ubfe] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         3,
         0,
         3,
         0,
         3,
         0,
         3,
         3,
         0,
         3,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         0,
         278,
         1823,
         2880,
         0,
         279,
         1823,
         2880,
         1824,
         1825,
         1826,
         2881,
         0,
         278,
         1823,
         2880,
         0,
         278,
         1823,
         2880,
         0,
         279,
         1823,
         2880,
         1824,
         1825,
         1826,
         2881,
         0,
         278,
         1823,
         2880,
         0,
         278,
         1823,
         2880,
         0,
         279,
         1823,
         2880,
         1824,
         1825,
         1826,
         2881,
         0,
         278,
         1823,
         2880,
         0,
         278,
         1823,
         2880,
         0,
         279,
         1823,
         2880,
         1824,
         1825,
         1826,
         2881,
         0,
         278,
         1823,
         2880,
      },
   },
   [nir_op_ibfe] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         3,
         0,
         3,
         0,
         3,
         0,
         3,
         3,
         0,
         3,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         0,
         280,
         1827,
         2882,
         0,
         281,
         1827,
         2882,
         1828,
         1829,
         1830,
         2883,
         0,
         280,
         1827,
         2882,
         0,
         280,
         1827,
         2882,
         0,
         281,
         1827,
         2882,
         1828,
         1829,
         1830,
         2883,
         0,
         280,
         1827,
         2882,
         0,
         280,
         1827,
         2882,
         0,
         281,
         1827,
         2882,
         1828,
         1829,
         1830,
         2883,
         0,
         280,
         1827,
         2882,
         0,
         280,
         1827,
         2882,
         0,
         281,
         1827,
         2882,
         1828,
         1829,
         1830,
         2883,
         0,
         280,
         1827,
         2882,
      },
   },
   [nir_op_pack_32_2x16_split] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         282,
      },
   },
   [nir_op_unpack_half_2x16_split_x] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         1,
         0,
         1,
         1,
         0,
         2,
         0,
         0,
         0,
         2,
         1,
         1,
         1,
         0,
         2,
         0,
         0,
         0,
         2,
         1,
         1,
         1,
         0,
         2,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         283,
         1831,
         1832,
         1833,
      },
   },
   [nir_op_unpack_half_2x16_split_y] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         284,
         1834,
      },
   },
   [nir_op_pack_half_2x16] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         0,
         1951,
         1952,
         1953,
      },
   },
   [nir_op_pack_half_2x16_split] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         285,
         286,
         287,
         288,
      },
   },
   [nir_op_pack_32_4x8_split] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         289,
      },
   },
   [nir_op_ussub_4x8_vc4] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         290,
         0,
         290,
      },
   },
   [nir_op_fsub] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         291,
      },
   },
   [nir_op_uabs_usub] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         292,
      },
   },
   [nir_op_uabs_isub] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         293,
      },
   },
   [nir_op_find_lsb] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         294,
         1974,
      },
   },
   [nir_op_ifind_msb] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         295,
      },
   },
   [nir_op_ufind_msb] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         296,
      },
   },
   [nir_op_ifind_msb_rev] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         297,
      },
   },
   [nir_op_ufind_msb_rev] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         298,
         1975,
      },
   },
   [nir_op_bitfield_reverse] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         299,
      },
   },
   [nir_op_fmod] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         300,
      },
   },
   [nir_op_frem] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         301,
      },
   },
   [nir_op_uadd_carry] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         302,
      },
   },
   [nir_op_usub_borrow] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         303,
      },
   },
   [nir_op_bitfield_insert] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         304,
      },
   },
   [nir_op_ihadd] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         305,
      },
   },
   [nir_op_uhadd] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         306,
      },
   },
   [nir_op_irhadd] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         307,
      },
   },
   [nir_op_urhadd] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         308,
      },
   },
   [nir_op_imul_32x16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         309,
      },
   },
   [nir_op_umul_32x16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         310,
      },
   },
   [nir_op_ibitfield_extract] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         311,
      },
   },
   [nir_op_ubitfield_extract] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         312,
      },
   },
   [nir_op_bitfield_select] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         2,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         0,
         313,
         0,
         0,
         313,
         0,
         1790,
         1791,
         1790,
         0,
         313,
         0,
         0,
         313,
         0,
         1790,
         1791,
         1790,
         0,
         313,
         0,
         0,
         313,
         0,
         1790,
         1791,
         1790,
      },
   },
   [nir_op_bfm] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         3,
         0,
         3,
         0,
         3,
         0,
         3,
         3,
         0,
         3,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         0,
         314,
         1835,
         0,
         0,
         314,
         1835,
         0,
         1836,
         1837,
         1838,
         1836,
         2884,
         2885,
         2886,
         2884,
      },
   },
   [nir_op_uclz] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         315,
      },
   },
   [nir_op_pack_unorm_2x16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         316,
      },
   },
   [nir_op_pack_unorm_4x8] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         317,
      },
   },
   [nir_op_pack_snorm_2x16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         318,
      },
   },
   [nir_op_pack_snorm_4x8] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         319,
      },
   },
   [nir_op_unpack_unorm_2x16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         320,
      },
   },
   [nir_op_unpack_unorm_4x8] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         321,
      },
   },
   [nir_op_unpack_snorm_2x16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         322,
      },
   },
   [nir_op_unpack_snorm_4x8] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         323,
      },
   },
   [nir_op_isign] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         324,
         1906,
      },
   },
   [nir_op_amul] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         325,
      },
   },
   [nir_op_umul24] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         326,
      },
   },
   [nir_op_umad24] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         327,
      },
   },
   [nir_op_imul24_relaxed] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         328,
      },
   },
   [nir_op_umad24_relaxed] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         329,
      },
   },
   [nir_op_umul24_relaxed] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         330,
      },
   },
   [nir_op_imad24_ir3] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         331,
         332,
         333,
         332,
         333,
         334,
         335,
      },
   },
   [nir_op_imul24] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         336,
         336,
         336,
      },
   },
   [nir_op_b2b1] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         337,
      },
   },
   [nir_op_b2b8] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         338,
      },
   },
   [nir_op_b2b16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         339,
      },
   },
   [nir_op_b2b32] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         340,
      },
   },
   [nir_op_ldexp] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         341,
      },
   },
   [nir_op_fddx] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         342,
      },
   },
   [nir_op_fddx_fine] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         343,
      },
   },
   [nir_op_fddx_coarse] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         344,
      },
   },
   [nir_op_fddy] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         345,
      },
   },
   [nir_op_fddy_fine] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         346,
      },
   },
   [nir_op_fddy_coarse] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         347,
      },
   },
   [nir_op_umul_low] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         348,
         348,
         348,
      },
   },
   [nir_op_imadsh_mix16] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         0,
         349,
         349,
         350,
         350,
         351,
         351,
      },
   },
   [nir_op_fsin_amd] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         352,
      },
   },
   [nir_op_fcos_amd] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         353,
      },
   },
   [nir_op_fisnormal] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         354,
      },
   },
};

/* Mapping from state index to offset in transforms (0 being no transforms) */
static const uint16_t nir_opt_algebraic_transform_offsets[] = {
   0,
   0,
   0,
   1,
   0,
   12,
   15,
   17,
   21,
   23,
   25,
   27,
   30,
   34,
   37,
   42,
   0,
   0,
   47,
   0,
   49,
   0,
   0,
   51,
   0,
   55,
   57,
   63,
   70,
   76,
   83,
   86,
   90,
   93,
   97,
   0,
   100,
   107,
   111,
   117,
   124,
   134,
   146,
   161,
   174,
   190,
   196,
   203,
   212,
   222,
   232,
   243,
   245,
   257,
   259,
   267,
   269,
   275,
   277,
   283,
   286,
   289,
   291,
   293,
   297,
   300,
   305,
   308,
   313,
   317,
   323,
   325,
   327,
   330,
   333,
   337,
   341,
   343,
   345,
   348,
   351,
   355,
   359,
   361,
   364,
   368,
   370,
   373,
   376,
   380,
   384,
   389,
   391,
   393,
   396,
   399,
   403,
   407,
   409,
   411,
   414,
   417,
   421,
   425,
   427,
   430,
   434,
   436,
   439,
   443,
   445,
   447,
   450,
   453,
   456,
   459,
   464,
   469,
   471,
   474,
   477,
   482,
   486,
   491,
   496,
   520,
   526,
   533,
   540,
   566,
   568,
   0,
   0,
   571,
   576,
   581,
   586,
   0,
   589,
   591,
   0,
   0,
   0,
   0,
   593,
   0,
   0,
   0,
   595,
   597,
   0,
   0,
   0,
   0,
   600,
   0,
   602,
   608,
   610,
   0,
   617,
   623,
   626,
   0,
   634,
   639,
   644,
   649,
   654,
   659,
   664,
   669,
   674,
   676,
   679,
   684,
   690,
   692,
   695,
   699,
   704,
   709,
   715,
   717,
   719,
   721,
   0,
   724,
   726,
   728,
   730,
   732,
   734,
   736,
   738,
   742,
   744,
   748,
   750,
   0,
   0,
   754,
   758,
   764,
   0,
   0,
   770,
   772,
   776,
   778,
   781,
   783,
   786,
   788,
   794,
   796,
   798,
   800,
   802,
   804,
   806,
   808,
   810,
   812,
   814,
   816,
   818,
   820,
   822,
   824,
   826,
   828,
   830,
   832,
   834,
   836,
   838,
   840,
   842,
   844,
   846,
   848,
   850,
   852,
   854,
   856,
   858,
   860,
   862,
   864,
   866,
   871,
   0,
   876,
   878,
   883,
   886,
   892,
   0,
   0,
   894,
   896,
   899,
   901,
   903,
   0,
   0,
   0,
   0,
   905,
   907,
   909,
   0,
   0,
   0,
   0,
   911,
   914,
   917,
   919,
   928,
   930,
   938,
   941,
   943,
   945,
   947,
   949,
   951,
   0,
   953,
   956,
   958,
   960,
   962,
   964,
   967,
   0,
   0,
   0,
   969,
   971,
   973,
   975,
   977,
   981,
   984,
   987,
   990,
   993,
   995,
   997,
   1000,
   1003,
   0,
   1005,
   1007,
   1009,
   1011,
   1013,
   1015,
   1017,
   1019,
   1021,
   1023,
   1025,
   1027,
   1029,
   1031,
   1034,
   1037,
   1040,
   1042,
   1045,
   1049,
   1053,
   1058,
   1062,
   1064,
   1066,
   1068,
   1070,
   1074,
   1076,
   1078,
   1080,
   1082,
   1084,
   1086,
   1088,
   1090,
   1092,
   0,
   0,
   1095,
   1099,
   1101,
   1104,
   1107,
   1109,
   1111,
   1123,
   1136,
   1148,
   1160,
   1172,
   1185,
   1198,
   1210,
   1222,
   1226,
   1230,
   1233,
   1236,
   1238,
   1243,
   1247,
   1251,
   1255,
   1259,
   1262,
   1264,
   1269,
   0,
   1271,
   1273,
   1275,
   1277,
   1279,
   1282,
   1285,
   1288,
   1290,
   1292,
   1294,
   1297,
   1300,
   1303,
   1307,
   1311,
   1315,
   1319,
   1323,
   1328,
   1331,
   1335,
   1339,
   1342,
   1345,
   1347,
   1349,
   1352,
   1355,
   1358,
   1361,
   1364,
   1367,
   1371,
   1375,
   1379,
   1382,
   1385,
   1387,
   1389,
   1391,
   1393,
   1396,
   1399,
   1402,
   1404,
   1406,
   1409,
   1412,
   1415,
   1419,
   1423,
   1427,
   1430,
   1433,
   1436,
   1439,
   1443,
   1447,
   1451,
   1454,
   1457,
   1460,
   1464,
   1468,
   1472,
   1475,
   1478,
   1482,
   1486,
   1490,
   1493,
   1496,
   1501,
   1506,
   1510,
   1514,
   1519,
   1523,
   1527,
   1531,
   1535,
   1538,
   1540,
   1544,
   1550,
   1555,
   1562,
   1567,
   1575,
   1579,
   1584,
   1589,
   1595,
   1600,
   1608,
   1616,
   1624,
   1632,
   1640,
   1646,
   1656,
   1662,
   1669,
   1675,
   1687,
   1692,
   1697,
   1704,
   1712,
   1720,
   1728,
   1736,
   1744,
   1752,
   1758,
   1765,
   1771,
   1779,
   1785,
   1792,
   1798,
   1806,
   1811,
   1817,
   1822,
   1829,
   1834,
   1840,
   1845,
   1852,
   1860,
   1865,
   1873,
   1878,
   1883,
   1888,
   1893,
   1898,
   1903,
   1908,
   1914,
   1922,
   1929,
   1938,
   1945,
   1955,
   1961,
   1968,
   1975,
   1983,
   1990,
   2000,
   2010,
   2020,
   2030,
   2040,
   2048,
   2060,
   2068,
   2077,
   2085,
   2099,
   2106,
   2113,
   2122,
   2132,
   2142,
   2152,
   2162,
   2172,
   2182,
   2190,
   2199,
   2207,
   2217,
   2225,
   2234,
   2242,
   2252,
   2259,
   2267,
   2274,
   2283,
   2290,
   2298,
   2305,
   2314,
   2324,
   2331,
   2341,
   2348,
   2355,
   2362,
   2369,
   2376,
   2383,
   2390,
   2396,
   2403,
   2409,
   2416,
   2422,
   2427,
   2451,
   2456,
   2463,
   2468,
   2473,
   2479,
   2484,
   2491,
   2497,
   2522,
   2528,
   2536,
   2542,
   2551,
   2556,
   2562,
   2568,
   2575,
   2581,
   2590,
   2599,
   2608,
   2617,
   2626,
   2633,
   2644,
   2651,
   2659,
   2666,
   2679,
   2685,
   2691,
   2699,
   2708,
   2717,
   2726,
   2735,
   2744,
   2753,
   2760,
   2768,
   2775,
   2784,
   2791,
   2799,
   2806,
   2815,
   2821,
   2828,
   2834,
   2842,
   2848,
   2855,
   2861,
   2869,
   2878,
   2884,
   2893,
   2899,
   2905,
   2911,
   2917,
   2923,
   2929,
   2935,
   2941,
   2948,
   2954,
   2962,
   2969,
   2995,
   3002,
   3011,
   3018,
   3028,
   3034,
   3041,
   3048,
   3056,
   3063,
   3073,
   3083,
   3093,
   3103,
   3113,
   3121,
   3133,
   3141,
   3150,
   3158,
   3172,
   3179,
   3186,
   3195,
   3205,
   3215,
   3225,
   3235,
   3245,
   3255,
   3263,
   3272,
   3280,
   3290,
   3298,
   3307,
   3315,
   3325,
   3332,
   3340,
   3347,
   3356,
   3363,
   3371,
   3378,
   3387,
   3397,
   3404,
   3414,
   3421,
   3428,
   3435,
   3442,
   3449,
   3456,
   3463,
   3471,
   3478,
   3486,
   3494,
   3503,
   3509,
   3516,
   3522,
   3530,
   3537,
   3563,
   3570,
   3579,
   3586,
   3596,
   3602,
   3609,
   3616,
   3624,
   3631,
   3641,
   3651,
   3661,
   3671,
   3681,
   3689,
   3701,
   3709,
   3718,
   3726,
   3740,
   3747,
   3754,
   3763,
   3773,
   3783,
   3793,
   3803,
   3813,
   3823,
   3831,
   3840,
   3848,
   3858,
   3866,
   3875,
   3883,
   3893,
   3900,
   3908,
   3915,
   3924,
   3931,
   3939,
   3946,
   3955,
   3965,
   3972,
   3982,
   3989,
   3996,
   4003,
   4010,
   4017,
   4024,
   4031,
   4037,
   4044,
   4050,
   4058,
   4065,
   4091,
   4098,
   4107,
   4114,
   4124,
   4130,
   4137,
   4144,
   4152,
   4159,
   4169,
   4179,
   4189,
   4199,
   4209,
   4217,
   4229,
   4237,
   4246,
   4254,
   4268,
   4275,
   4282,
   4291,
   4301,
   4311,
   4321,
   4331,
   4341,
   4351,
   4359,
   4368,
   4376,
   4386,
   4394,
   4403,
   4411,
   4421,
   4428,
   4436,
   4443,
   4452,
   4459,
   4467,
   4474,
   4483,
   4493,
   4500,
   4510,
   4517,
   4524,
   4531,
   4538,
   4545,
   4552,
   4559,
   4565,
   4572,
   4578,
   4586,
   4593,
   4619,
   4626,
   4635,
   4642,
   4652,
   4658,
   4665,
   4672,
   4680,
   4687,
   4697,
   4707,
   4717,
   4727,
   4737,
   4745,
   4757,
   4765,
   4774,
   4782,
   4796,
   4803,
   4810,
   4819,
   4829,
   4839,
   4849,
   4859,
   4869,
   4879,
   4887,
   4896,
   4904,
   4914,
   4922,
   4931,
   4939,
   4949,
   4956,
   4964,
   4971,
   4980,
   4987,
   4995,
   5002,
   5011,
   5021,
   5028,
   5038,
   5045,
   5052,
   5059,
   5066,
   5073,
   5080,
   5087,
   5093,
   5100,
   5106,
   5114,
   5121,
   5147,
   5154,
   5163,
   5170,
   5180,
   5186,
   5193,
   5200,
   5208,
   5215,
   5225,
   5235,
   5245,
   5255,
   5265,
   5273,
   5285,
   5293,
   5302,
   5310,
   5324,
   5331,
   5338,
   5347,
   5357,
   5367,
   5377,
   5387,
   5397,
   5407,
   5415,
   5424,
   5432,
   5442,
   5450,
   5459,
   5467,
   5477,
   5484,
   5492,
   5499,
   5508,
   5515,
   5523,
   5530,
   5539,
   5549,
   5556,
   5566,
   5573,
   5580,
   5587,
   5594,
   5601,
   5608,
   5615,
   5621,
   5628,
   5634,
   5642,
   5649,
   5675,
   5682,
   5691,
   5698,
   5708,
   5714,
   5721,
   5728,
   5736,
   5743,
   5753,
   5763,
   5773,
   5783,
   5793,
   5801,
   5813,
   5821,
   5830,
   5838,
   5852,
   5859,
   5866,
   5875,
   5885,
   5895,
   5905,
   5915,
   5925,
   5935,
   5943,
   5952,
   5960,
   5970,
   5978,
   5987,
   5995,
   6005,
   6012,
   6020,
   6027,
   6036,
   6043,
   6051,
   6058,
   6067,
   6077,
   6084,
   6094,
   6101,
   6108,
   6115,
   6122,
   6129,
   6136,
   6143,
   0,
   6148,
   6154,
   6156,
   6160,
   6167,
   6171,
   6178,
   6183,
   6190,
   6199,
   6204,
   6209,
   6216,
   6225,
   6230,
   6233,
   6237,
   6239,
   6243,
   6249,
   6251,
   6254,
   6257,
   6260,
   6264,
   6268,
   6271,
   6274,
   6287,
   6304,
   6322,
   6335,
   6349,
   6362,
   6378,
   6391,
   6405,
   6419,
   6432,
   6445,
   6448,
   6452,
   6456,
   6460,
   6465,
   6470,
   6474,
   6478,
   6482,
   6486,
   6490,
   6495,
   6500,
   6504,
   6508,
   6511,
   6521,
   6527,
   6531,
   6535,
   6539,
   6542,
   6545,
   6548,
   6551,
   6555,
   6560,
   6565,
   6569,
   6573,
   6578,
   6584,
   6595,
   6607,
   6613,
   6625,
   6638,
   6642,
   6647,
   6658,
   6670,
   6675,
   6687,
   6700,
   6702,
   6707,
   6712,
   6716,
   6720,
   6725,
   6731,
   6736,
   6741,
   6746,
   6751,
   6756,
   6760,
   6764,
   6768,
   6771,
   6774,
   6777,
   6780,
   6782,
   6786,
   6790,
   6793,
   6796,
   6805,
   6814,
   6824,
   6833,
   6842,
   6850,
   6860,
   6870,
   6879,
   6888,
   6891,
   6893,
   6897,
   6901,
   6905,
   6909,
   6912,
   6917,
   6922,
   6926,
   6930,
   6934,
   6938,
   6942,
   6945,
   6950,
   6955,
   6959,
   6963,
   6971,
   6975,
   6978,
   6983,
   6988,
   6993,
   6997,
   7001,
   7011,
   7013,
   7015,
   7018,
   7021,
   7024,
   7028,
   7031,
   7036,
   7041,
   7045,
   7049,
   7054,
   7058,
   7062,
   7065,
   7068,
   7074,
   7085,
   7097,
   7103,
   7115,
   7128,
   7132,
   7137,
   7147,
   7158,
   7163,
   7174,
   7186,
   7188,
   7192,
   7196,
   7200,
   7204,
   7209,
   7215,
   7220,
   7225,
   7230,
   7235,
   7240,
   7244,
   7248,
   7252,
   7255,
   7259,
   7263,
   7266,
   7269,
   7272,
   7280,
   7288,
   7295,
   7302,
   7309,
   7317,
   7320,
   7323,
   7326,
   7331,
   7337,
   7342,
   7347,
   7352,
   7357,
   7362,
   7367,
   7372,
   7376,
   7380,
   7384,
   7388,
   7392,
   0,
   7397,
   7403,
   7405,
   7410,
   7421,
   7430,
   7435,
   7445,
   7452,
   7457,
   7464,
   0,
   7471,
   7477,
   7482,
   7492,
   7507,
   7522,
   7537,
   7542,
   7547,
   7552,
   7560,
   7575,
   0,
   7583,
   7586,
   7589,
   7592,
   7595,
   7598,
   7601,
   7604,
   7607,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   7609,
   7612,
   7615,
   7617,
   7620,
   7623,
   7626,
   7629,
   7631,
   7633,
   7638,
   7640,
   7642,
   7647,
   7654,
   7662,
   7667,
   7673,
   7678,
   7684,
   7689,
   7693,
   7695,
   7697,
   7699,
   0,
   7701,
   0,
   0,
   7703,
   0,
   0,
   7706,
   7714,
   7722,
   7731,
   7739,
   7748,
   7755,
   7762,
   7770,
   7780,
   7787,
   7794,
   7796,
   7798,
   7802,
   7804,
   7806,
   0,
   0,
   7808,
   0,
   0,
   7811,
   0,
   0,
   7814,
   7817,
   7820,
   7822,
   0,
   7824,
   7826,
   7831,
   7837,
   7844,
   7851,
   7859,
   7866,
   7875,
   7882,
   7887,
   7895,
   7903,
   7916,
   7929,
   7936,
   7946,
   7954,
   7962,
   7971,
   7985,
   7999,
   8011,
   8026,
   8039,
   8052,
   8065,
   8081,
   8096,
   8110,
   8124,
   8141,
   8157,
   8172,
   8185,
   8201,
   8215,
   8229,
   8237,
   8248,
   8257,
   8266,
   8275,
   8282,
   8292,
   8300,
   8308,
   8315,
   8322,
   8332,
   8342,
   8350,
   8361,
   8372,
   8382,
   8393,
   8405,
   8416,
   8428,
   8441,
   8455,
   8468,
   8475,
   8483,
   8491,
   8499,
   8502,
   8506,
   8510,
   8514,
   8517,
   8520,
   8525,
   8529,
   8533,
   8538,
   8542,
   8546,
   8550,
   8554,
   8559,
   8564,
   8569,
   8574,
   8578,
   8582,
   8588,
   8594,
   8600,
   8606,
   8612,
   8619,
   8625,
   8631,
   8638,
   8644,
   8650,
   8656,
   8662,
   8668,
   8675,
   8681,
   8687,
   8693,
   8700,
   8707,
   8714,
   8721,
   8727,
   8733,
   8739,
   8746,
   8753,
   8760,
   8766,
   8772,
   8778,
   8785,
   8792,
   8799,
   8812,
   8818,
   8824,
   8830,
   8837,
   8843,
   8849,
   8855,
   8861,
   8867,
   8873,
   8879,
   8885,
   8892,
   8898,
   8904,
   8910,
   8917,
   8924,
   8931,
   8938,
   8945,
   8951,
   8957,
   8963,
   8970,
   8977,
   8984,
   8990,
   8996,
   9002,
   9009,
   9016,
   9023,
   9036,
   9039,
   9042,
   9047,
   9051,
   9055,
   9060,
   9064,
   9068,
   9072,
   9076,
   9081,
   9086,
   9091,
   9096,
   9100,
   9104,
   9110,
   9116,
   9119,
   9122,
   9125,
   9127,
   9131,
   9135,
   9139,
   9143,
   9147,
   9150,
   9155,
   9159,
   9163,
   9166,
   9170,
   9173,
   9176,
   9179,
   9182,
   9185,
   9188,
   9191,
   9195,
   9198,
   9202,
   9205,
   9209,
   9213,
   9217,
   9220,
   0,
   0,
   9222,
   9224,
   0,
   0,
   0,
   0,
   0,
   9229,
   9232,
   9235,
   9238,
   9241,
   9244,
   9251,
   9258,
   9265,
   9280,
   9287,
   9294,
   9308,
   9322,
   9328,
   9340,
   9352,
   9358,
   9364,
   9371,
   9378,
   9384,
   9390,
   9396,
   9402,
   9408,
   9414,
   9418,
   9424,
   9436,
   9448,
   9451,
   9454,
   9461,
   9468,
   9475,
   9490,
   9497,
   9504,
   9518,
   9532,
   9544,
   9556,
   9562,
   9569,
   9575,
   9582,
   9588,
   9594,
   9600,
   9606,
   9612,
   9618,
   9622,
   9628,
   9640,
   9652,
   9655,
   9658,
   9661,
   9663,
   9666,
   9669,
   9672,
   9674,
   9677,
   9681,
   9685,
   9688,
   9692,
   9695,
   9698,
   9701,
   9704,
   9707,
   9710,
   9713,
   9716,
   9719,
   9723,
   9727,
   9731,
   9735,
   9737,
   9739,
   9743,
   9747,
   9749,
   9753,
   9755,
   9757,
   9761,
   9765,
   9767,
   9774,
   9782,
   9787,
   9792,
   9797,
   9804,
   9812,
   9817,
   9822,
   9827,
   9829,
   9831,
   9833,
   9835,
   9837,
   9839,
   9841,
   9843,
   9845,
   9847,
   9849,
   0,
   9851,
   9853,
   9855,
   9857,
   9859,
   9861,
   9863,
   9865,
   9867,
   9872,
   9878,
   9884,
   9892,
   9901,
   9906,
   9913,
   9921,
   9927,
   9934,
   9941,
   9950,
   9960,
   9966,
   9974,
   9983,
   9992,
   10003,
   10015,
   10021,
   10029,
   10038,
   10041,
   10044,
   10047,
   10052,
   10057,
   10062,
   10069,
   10074,
   10080,
   10090,
   10092,
   10095,
   10098,
   10101,
   10104,
   10107,
   10111,
   10115,
   10119,
   10123,
   10127,
   10130,
   10134,
   10138,
   10142,
   10146,
   10151,
   10154,
   10158,
   10162,
   10166,
   10170,
   10175,
   10184,
   10206,
   10209,
   10212,
   10215,
   10218,
   10221,
   10224,
   10228,
   10231,
   10235,
   10243,
   10245,
   10248,
   10251,
   10254,
   10257,
   10261,
   10265,
   10269,
   10272,
   10275,
   10279,
   10283,
   10287,
   10291,
   10294,
   10298,
   10302,
   10306,
   10310,
   10318,
   0,
   10340,
   10342,
   10344,
   10347,
   10349,
   10353,
   10358,
   10362,
   10365,
   10368,
   10370,
   10373,
   10376,
   10378,
   10381,
   10393,
   10403,
   10410,
   10417,
   10424,
   10463,
   10482,
   10489,
   10496,
   10508,
   10514,
   10524,
   10531,
   10538,
   10545,
   10552,
   10558,
   10574,
   10616,
   10623,
   10630,
   10637,
   10649,
   10652,
   10655,
   10658,
   10661,
   10664,
   10668,
   10673,
   10677,
   10679,
   10681,
   10684,
   10687,
   10689,
   10691,
   10694,
   10697,
   10700,
   10703,
   10707,
   10710,
   10712,
   10714,
   10716,
   10719,
   10722,
   10725,
   10728,
   10731,
   10734,
   10737,
   10740,
   10743,
   10746,
   10749,
   0,
   0,
   0,
   10751,
   10754,
   10757,
   10760,
   10763,
   10766,
   10769,
   10772,
   10775,
   10778,
   10781,
   10787,
   10793,
   10799,
   10806,
   10813,
   10824,
   10831,
   10837,
   10845,
   10852,
   10859,
   10868,
   10874,
   10882,
   10889,
   10896,
   10905,
   10917,
   10942,
   10948,
   10954,
   10960,
   10967,
   10974,
   10985,
   10992,
   10998,
   11006,
   11013,
   11020,
   11029,
   11035,
   11043,
   11050,
   11057,
   11066,
   11078,
   11103,
   11109,
   11115,
   11121,
   11127,
   11133,
   11136,
   11138,
   11140,
   11142,
   11144,
   11146,
   11149,
   11151,
   11153,
   11155,
   11157,
   11159,
   11161,
   11163,
   11165,
   11167,
   11169,
   11173,
   11175,
   11177,
   11181,
   11185,
   11187,
   11190,
   11195,
   11198,
   11201,
   11203,
   11205,
   11207,
   11209,
   11211,
   11213,
   11215,
   11217,
   11219,
   11221,
   11223,
   11227,
   11232,
   11238,
   11241,
   11245,
   11248,
   0,
   0,
   0,
   0,
   0,
   0,
   11252,
   11256,
   11260,
   11264,
   11268,
   11272,
   11276,
   11280,
   11284,
   11288,
   11292,
   11296,
   11302,
   11304,
   11306,
   11308,
   11310,
   11312,
   11315,
   11317,
   11319,
   11321,
   11325,
   11329,
   11333,
   11337,
   11341,
   11346,
   11349,
   11352,
   11355,
   11358,
   11361,
   11366,
   11371,
   11376,
   11381,
   11385,
   11389,
   11394,
   11399,
   11404,
   11409,
   11413,
   11417,
   11421,
   11425,
   11428,
   11431,
   11436,
   11441,
   11446,
   11451,
   11454,
   11457,
   11462,
   11467,
   11472,
   11477,
   11480,
   11483,
   11488,
   11493,
   11498,
   11503,
   11506,
   11509,
   11514,
   11519,
   11524,
   11529,
   11533,
   11537,
   11543,
   11549,
   11555,
   11561,
   11565,
   11569,
   11575,
   11581,
   11587,
   11593,
   11597,
   11601,
   11607,
   11613,
   11619,
   11625,
   11628,
   11631,
   11636,
   11641,
   11646,
   11651,
   11654,
   11657,
   11662,
   11667,
   11672,
   11677,
   11682,
   11687,
   11692,
   11697,
   11702,
   11707,
   11712,
   11717,
   11724,
   11731,
   11738,
   11745,
   11752,
   11759,
   11761,
   11768,
   11776,
   11783,
   11792,
   11799,
   11807,
   11814,
   11823,
   11830,
   11838,
   11845,
   11854,
   11861,
   11869,
   11876,
   11885,
   11894,
   11904,
   11913,
   11924,
   11933,
   11943,
   11952,
   11963,
   11972,
   11982,
   11991,
   12002,
   12011,
   12021,
   12030,
   12041,
   12047,
   12055,
   12064,
   12072,
   12082,
   12090,
   12099,
   12107,
   12117,
   12125,
   12134,
   12142,
   12152,
   12160,
   12169,
   12177,
   12187,
   12196,
   12206,
   12215,
   12226,
   12235,
   12245,
   12254,
   12265,
   12274,
   12284,
   12293,
   12304,
   12313,
   12323,
   12332,
   12343,
   12352,
   12362,
   12371,
   12382,
   12391,
   12401,
   12410,
   12421,
   12430,
   12440,
   12449,
   12460,
   12469,
   12479,
   12488,
   12499,
   12508,
   12518,
   12527,
   12538,
   12547,
   12557,
   12566,
   12577,
   12586,
   12596,
   12605,
   12616,
   12625,
   12635,
   12644,
   12655,
   12662,
   12673,
   12682,
   12692,
   12701,
   12712,
   12721,
   12731,
   12740,
   12751,
   12760,
   12770,
   12779,
   12790,
   12799,
   12809,
   12818,
   12829,
   12838,
   12848,
   12857,
   12868,
   12877,
   12887,
   12896,
   12907,
   12916,
   12926,
   12935,
   12946,
   12955,
   12965,
   12974,
   12985,
   12996,
   13005,
   13015,
   13024,
   13035,
   13044,
   13054,
   13063,
   13074,
   13083,
   13093,
   13102,
   13113,
   13122,
   13132,
   13141,
   13152,
   13161,
   13170,
   13179,
   13188,
   13199,
   13214,
   13225,
   13240,
   13248,
   13256,
   13264,
   13272,
   13280,
   13284,
   13288,
   13292,
   13296,
   13308,
   13321,
   13335,
   13349,
   13363,
   13377,
   13382,
   13387,
   13392,
   13397,
   13403,
   13409,
   13415,
   13421,
   13427,
   13433,
   13439,
   13445,
   13450,
   13455,
   13460,
   13465,
   13470,
   13475,
   13480,
   13485,
   13492,
   13499,
   13506,
   13513,
   13520,
   0,
   13527,
   13529,
   13534,
   13539,
   13541,
   13545,
   13550,
   13555,
   13558,
   0,
   13562,
   13567,
   13569,
   13575,
   13580,
   13589,
   13595,
   13602,
   13605,
   13608,
   13612,
   13616,
   13620,
   13626,
   13631,
   13635,
   13639,
   13644,
   13648,
   13652,
   13658,
   13663,
   13667,
   13671,
   13676,
   0,
   0,
   13680,
   13682,
   13693,
   13700,
   13707,
   13716,
   13726,
   13731,
   13733,
   13735,
   13738,
   13742,
   0,
   0,
   13744,
   13746,
   0,
   0,
   13750,
   13753,
   13756,
   13759,
   13763,
   0,
   0,
   13767,
   13769,
   0,
   0,
   13773,
   13775,
   13777,
   13782,
   13786,
   13790,
   13795,
   0,
   13797,
   13799,
   13810,
   13818,
   13829,
   13838,
   13847,
   13863,
   13876,
   13892,
   13906,
   13920,
   13931,
   13939,
   13950,
   13959,
   13968,
   13978,
   13990,
   14000,
   14013,
   14024,
   14035,
   14049,
   14060,
   14068,
   14079,
   14088,
   14097,
   14106,
   14118,
   14128,
   14138,
   14143,
   14148,
   14153,
   14158,
   14169,
   14180,
   14191,
   14202,
   14206,
   14212,
   14218,
   14224,
   14230,
   14236,
   14242,
   14248,
   14254,
   14260,
   14266,
   14272,
   14278,
   14282,
   14288,
   14294,
   14300,
   14306,
   14313,
   14318,
   14323,
   14328,
   14333,
   14340,
   14347,
   14354,
   14361,
   14363,
   14370,
   14377,
   14384,
   14391,
   14397,
   14403,
   14409,
   14415,
   14421,
   14427,
   14433,
   14439,
   14445,
   14450,
   14455,
   14458,
   14466,
   14474,
   14482,
   14490,
   14498,
   14506,
   14511,
   14514,
   14516,
   14518,
   14520,
   14522,
   14524,
   14526,
   14529,
   14531,
   14538,
   14543,
   14548,
   14553,
   14558,
   14571,
   14584,
   14598,
   14610,
   14627,
   14642,
   14657,
   14672,
   14688,
   14690,
   14694,
   14697,
   14705,
   14711,
   14717,
   14723,
   14729,
   14733,
   14736,
   14744,
   14750,
   14756,
   14762,
   14768,
   14778,
   14781,
   14785,
   14788,
   14796,
   14802,
   14808,
   14814,
   14820,
   14825,
   14829,
   14838,
   14845,
   14852,
   14859,
   14866,
   14871,
   14875,
   14884,
   14891,
   14898,
   14905,
   14912,
   14916,
   14919,
   14927,
   14933,
   14939,
   14945,
   14951,
   14955,
   14958,
   14966,
   14972,
   14978,
   14984,
   14990,
   14993,
   15001,
   15007,
   15013,
   15019,
   15025,
   15028,
   15035,
   15040,
   15045,
   15050,
   15055,
   15066,
   15076,
   15086,
   15096,
   15106,
   15114,
   15122,
   15130,
   15138,
   15146,
   15154,
   15159,
   15162,
   15164,
   15170,
   15177,
   15184,
   15191,
   15198,
   15206,
   15212,
   15220,
   15227,
   15234,
   15238,
   15242,
   15246,
   15251,
   15256,
   15261,
   15266,
   15271,
   15276,
   15281,
   15286,
   15291,
   15295,
   15299,
   15303,
   15308,
   15314,
   15318,
   15324,
   15328,
   15332,
   15334,
   15338,
   15342,
   15346,
   15349,
   15351,
   15355,
   15359,
   15362,
   15367,
   15372,
   15377,
   15380,
   15385,
   15390,
   15395,
   15398,
   15403,
   15407,
   15413,
   15417,
   15421,
   15427,
   15431,
   15435,
   15439,
   15443,
   15447,
   15450,
   15454,
   15459,
   15462,
   15468,
   15474,
   15480,
   15486,
   15491,
   15499,
   15507,
   15516,
   15521,
   15528,
   15530,
   15532,
   15540,
   15542,
   0,
   0,
   0,
   15548,
   15550,
   15555,
   15559,
   15563,
   15567,
   15571,
   15576,
   15581,
   15586,
   15591,
   15595,
   15599,
   15605,
   15609,
   15613,
   15615,
   15619,
   15623,
   15625,
   15628,
   15632,
   15635,
   15640,
   15645,
   15648,
   15653,
   15658,
   15661,
   15665,
   15671,
   15675,
   15679,
   15687,
   15695,
   15698,
   15705,
   15709,
   15714,
   15720,
   15728,
   15732,
   15737,
   15742,
   15747,
   15754,
   15756,
   15760,
   15764,
   15770,
   15772,
   15776,
   15779,
   15781,
   15785,
   15788,
   15792,
   15794,
   15796,
   15800,
   15804,
   15808,
   15814,
   15817,
   15820,
   15823,
   15825,
   15829,
   15833,
   15836,
   15839,
   15842,
   15847,
   15852,
   15857,
   15862,
   15867,
   15876,
   15885,
   15894,
   15903,
   15910,
   15917,
   15924,
   15931,
   15938,
   15945,
   15952,
   15959,
   15965,
   15971,
   15977,
   15983,
   15989,
   15995,
   16001,
   16007,
   16013,
   16019,
   16025,
   16031,
   16039,
   16047,
   16055,
   16063,
   16071,
   16079,
   16084,
   16127,
   16170,
   16190,
   16195,
   16197,
   16200,
   16204,
   16209,
   16211,
   16214,
   16217,
   16221,
   16223,
   16225,
   16228,
   16232,
   16237,
   16240,
   16246,
   16252,
   16257,
   16262,
   16268,
   16273,
   16279,
   16284,
   16290,
   16295,
   16302,
   16308,
   16314,
   16319,
   16326,
   16332,
   16338,
   16343,
   16350,
   16356,
   16362,
   16367,
   16374,
   16380,
   16384,
   16388,
   16394,
   16397,
   16403,
   16409,
   16426,
   16442,
   16446,
   16453,
   16459,
   16466,
   16470,
   16480,
   16484,
   16495,
   16507,
   16512,
   16517,
   16525,
   16533,
   16538,
   16546,
   16551,
   16559,
   16567,
   16578,
   16583,
   16591,
   16599,
   16610,
   16621,
   16627,
   16636,
   16641,
   16649,
   16654,
   16663,
   0,
   16675,
   16683,
   16691,
   0,
   0,
   0,
   16699,
   16706,
   16710,
   16716,
   16719,
   16722,
   16726,
   16732,
   16736,
   16740,
   16744,
   16748,
   16752,
   16756,
   16759,
   16764,
   16768,
   16773,
   16777,
   16782,
   16787,
   0,
   0,
   16791,
   16798,
   16805,
   16807,
   0,
   0,
   0,
   0,
   16809,
   16812,
   0,
   0,
   0,
   0,
   16815,
   16819,
   16823,
   0,
   0,
   0,
   16825,
   16827,
   0,
   16829,
   16840,
   16856,
   16867,
   16879,
   16893,
   16905,
   16910,
   16915,
   16917,
   16924,
   16931,
   16934,
   16937,
   16941,
   16945,
   16949,
   16953,
   16958,
   16963,
   16970,
   16977,
   16982,
   16984,
   16987,
   16989,
   16992,
   16994,
   16996,
   16999,
   17003,
   17007,
   17009,
   17017,
   17026,
   17033,
   17036,
   17040,
   17048,
   17056,
   17061,
   17069,
   17080,
   17091,
   17105,
   17119,
   17124,
   17135,
   17139,
   17143,
   0,
   17148,
   17151,
   17157,
   17170,
   17178,
   17195,
   17211,
   17218,
   17224,
   17228,
   17230,
   17233,
   17239,
   17252,
   17269,
   17285,
   17292,
   17298,
   17302,
   17304,
   17307,
   17313,
   17326,
   17343,
   17359,
   17366,
   17372,
   17377,
};

static const nir_algebraic_table nir_opt_algebraic_table = {
   .transforms = nir_opt_algebraic_transforms,
   .transform_offsets = nir_opt_algebraic_transform_offsets,
   .pass_op_table = nir_opt_algebraic_pass_op_table,
   .values = nir_opt_algebraic_values,
   .expression_cond = nir_opt_algebraic_expression_cond,
   .variable_cond = nir_opt_algebraic_variable_cond,
};

bool
nir_opt_algebraic(nir_shader *shader)
{
   bool progress = false;
   bool condition_flags[122];
   const nir_shader_compiler_options *options = shader->options;
   const shader_info *info = &shader->info;
   (void) options;
   (void) info;

   /* This is not a great place for this, but it seems to be the best place
    * for it. Check that at most one kind of lowering is requested for
    * bitfield extract and bitfield insert. Otherwise the lowering can fight
    * with each other and optimizations.
    */
   assert((int)options->lower_bitfield_extract +
          (int)options->lower_bitfield_extract_to_shifts <= 1);
   assert((int)options->lower_bitfield_insert +
          (int)options->lower_bitfield_insert_to_shifts +
          (int)options->lower_bitfield_insert_to_bitfield_select <= 1);


   STATIC_ASSERT(5961 == ARRAY_SIZE(nir_opt_algebraic_values));
   condition_flags[0] = true;
   condition_flags[1] = !options->lower_bitops;
   condition_flags[2] = options->lower_bitops;
   condition_flags[3] = !options->lower_bitops && (options->lower_int64_options & (nir_lower_imul64 | nir_lower_shift64)) == nir_lower_imul64;
   condition_flags[4] = options->lower_mul_2x32_64;
   condition_flags[5] = !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 16);
   condition_flags[6] = !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[7] = !options->lower_iadd_sat;
   condition_flags[8] = !options->lower_uadd_sat;
   condition_flags[9] = options->has_fmulz && !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[10] = !options->has_sdot_4x8;
   condition_flags[11] = !options->has_udot_4x8;
   condition_flags[12] = !options->has_sudot_4x8;
   condition_flags[13] = !options->has_dot_2x16;
   condition_flags[14] = options->lower_flrp16;
   condition_flags[15] = !options->lower_flrp16;
   condition_flags[16] = options->lower_flrp32;
   condition_flags[17] = !options->lower_flrp32;
   condition_flags[18] = options->lower_flrp64;
   condition_flags[19] = !options->lower_flrp64;
   condition_flags[20] = options->lower_ftrunc;
   condition_flags[21] = options->lower_ftrunc || (options->lower_doubles_options & nir_lower_dtrunc);
   condition_flags[22] = options->lower_ffloor;
   condition_flags[23] = (options->lower_ffloor || (options->lower_doubles_options & nir_lower_dfloor)) && !(options->lower_doubles_options & nir_lower_dfract);
   condition_flags[24] = !options->lower_ffloor;
   condition_flags[25] = !options->lower_ffloor && !(options->lower_doubles_options & nir_lower_dfloor);
   condition_flags[26] = options->lower_ffract;
   condition_flags[27] = options->lower_ffract || (options->lower_doubles_options & nir_lower_dfract);
   condition_flags[28] = options->lower_fceil;
   condition_flags[29] = options->lower_ffma16;
   condition_flags[30] = options->lower_ffma32;
   condition_flags[31] = options->lower_ffma64;
   condition_flags[32] = options->fuse_ffma16;
   condition_flags[33] = options->fuse_ffma32;
   condition_flags[34] = options->fuse_ffma64;
   condition_flags[35] = options->lower_fdph;
   condition_flags[36] = !options->lower_fdph;
   condition_flags[37] = options->lower_fdot;
   condition_flags[38] = !options->lower_fsat;
   condition_flags[39] = !options->lower_iabs;
   condition_flags[40] = options->lower_fsat;
   condition_flags[41] = !options->lower_fsat && !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[42] = !options->lower_fsign;
   condition_flags[43] = options->lower_umax;
   condition_flags[44] = options->lower_umin;
   condition_flags[45] = !options->lower_umax;
   condition_flags[46] = !options->lower_umin;
   condition_flags[47] = options->lower_scmp;
   condition_flags[48] = options->lower_vector_cmp;
   condition_flags[49] = !options->lower_rotate;
   condition_flags[50] = options->lower_rotate;
   condition_flags[51] = options->lower_fpow;
   condition_flags[52] = !options->lower_fpow;
   condition_flags[53] = options->lower_fdiv;
   condition_flags[54] = options->lower_fsqrt;
   condition_flags[55] = !options->lower_fsqrt;
   condition_flags[56] = options->lower_sincos;
   condition_flags[57] = !(options->lower_doubles_options & nir_lower_fp64_full_software);
   condition_flags[58] = !options->lower_extract_byte;
   condition_flags[59] = !options->lower_extract_word;
   condition_flags[60] = options->has_pack_32_4x8;
   condition_flags[61] = options->lower_pack_64_2x32_split;
   condition_flags[62] = options->lower_pack_32_2x16_split;
   condition_flags[63] = options->lower_unpack_64_2x32_split;
   condition_flags[64] = options->lower_unpack_32_2x16_split;
   condition_flags[65] = options->has_find_msb_rev;
   condition_flags[66] = !options->lower_find_msb_to_reverse;
   condition_flags[67] = !options->lower_find_lsb;
   condition_flags[68] = options->lower_fmod;
   condition_flags[69] = options->lower_uadd_carry;
   condition_flags[70] = options->lower_usub_borrow;
   condition_flags[71] = options->lower_bitfield_insert;
   condition_flags[72] = options->lower_hadd;
   condition_flags[73] = options->lower_hadd64 || (options->lower_int64_options & nir_lower_iadd64) != 0;
   condition_flags[74] = options->lower_mul_32x16;
   condition_flags[75] = options->lower_uadd_sat || (options->lower_int64_options & nir_lower_iadd64) != 0;
   condition_flags[76] = options->lower_uadd_sat;
   condition_flags[77] = options->lower_usub_sat;
   condition_flags[78] = (options->lower_int64_options & nir_lower_usub_sat64) != 0;
   condition_flags[79] = (options->lower_int64_options & nir_lower_iadd_sat64) != 0;
   condition_flags[80] = (options->lower_int64_options & nir_lower_minmax64) != 0;
   condition_flags[81] = (options->lower_int64_options & nir_lower_icmp64) != 0;
   condition_flags[82] = options->lower_bitfield_insert_to_shifts;
   condition_flags[83] = options->lower_bitfield_insert_to_bitfield_select;
   condition_flags[84] = options->lower_bitfield_extract;
   condition_flags[85] = options->lower_bitfield_insert_to_bitfield_select || options->lower_bitfield_insert;
   condition_flags[86] = options->lower_bitfield_extract_to_shifts;
   condition_flags[87] = options->lower_ifind_msb;
   condition_flags[88] = options->lower_find_msb_to_reverse;
   condition_flags[89] = options->lower_uclz;
   condition_flags[90] = options->lower_find_lsb;
   condition_flags[91] = options->lower_extract_byte;
   condition_flags[92] = options->lower_extract_word;
   condition_flags[93] = options->lower_pack_unorm_2x16;
   condition_flags[94] = options->lower_pack_unorm_4x8;
   condition_flags[95] = options->lower_pack_snorm_2x16;
   condition_flags[96] = options->lower_pack_snorm_4x8;
   condition_flags[97] = options->lower_unpack_unorm_2x16;
   condition_flags[98] = options->lower_unpack_unorm_4x8;
   condition_flags[99] = options->lower_unpack_snorm_2x16;
   condition_flags[100] = options->lower_unpack_snorm_4x8;
   condition_flags[101] = options->lower_pack_split;
   condition_flags[102] = options->lower_isign;
   condition_flags[103] = !options->lower_isign;
   condition_flags[104] = options->lower_fsign;
   condition_flags[105] = !options->has_imul24;
   condition_flags[106] = !options->has_umul24;
   condition_flags[107] = !options->has_umad24;
   condition_flags[108] = options->has_imul24;
   condition_flags[109] = options->has_umad24;
   condition_flags[110] = options->has_umul24;
   condition_flags[111] = !options->lower_uadd_sat && !(options->lower_int64_options & nir_lower_iadd64);
   condition_flags[112] = !options->lower_usub_sat;
   condition_flags[113] = !options->lower_usub_sat && !(options->lower_int64_options & nir_lower_usub_sat64);
   condition_flags[114] = options->lower_iadd_sat;
   condition_flags[115] = options->support_8bit_alu;
   condition_flags[116] = options->support_16bit_alu;
   condition_flags[117] = options->lower_ldexp;
   condition_flags[118] = !options->lower_bitfield_reverse;
   condition_flags[119] = info->stage == MESA_SHADER_COMPUTE && info->cs.derivative_group == DERIVATIVE_GROUP_NONE;
   condition_flags[120] = options->lower_fisnormal;
   condition_flags[121] = !options->vectorize_vec2_16bit;

   nir_foreach_function(function, shader) {
      if (function->impl) {
         progress |= nir_algebraic_impl(function->impl, condition_flags,
                                        &nir_opt_algebraic_table);
      }
   }

   return progress;
}


#include "nir.h"
#include "nir_builder.h"
#include "nir_search.h"
#include "nir_search_helpers.h"

/* What follows is NIR algebraic transform code for the following 15
 * transforms:
 *    ('~fmul(is_used_once)', ('fmul(is_used_once)', 'a(is_not_const)', '#b'), 'c(is_not_const)') => ('fmul', ('fmul', 'a', 'c'), 'b')
 *    ('imul(is_used_once)', ('imul(is_used_once)', 'a(is_not_const)', '#b'), 'c(is_not_const)') => ('imul', ('imul', 'a', 'c'), 'b')
 *    ('~fadd(is_used_once)', ('fadd(is_used_once)', 'a(is_not_const)', '#b'), 'c(is_not_const)') => ('fadd', ('fadd', 'a', 'c'), 'b')
 *    ('iadd(is_used_once)', ('iadd(is_used_once)', 'a(is_not_const)', '#b'), 'c(is_not_const)') => ('iadd', ('iadd', 'a', 'c'), 'b')
 *    ('~fadd', ('fmul', 'a', 'b'), ('fmul', 'a', 'c')) => ('fmul', 'a', ('fadd', 'b', 'c'))
 *    ('iadd', ('imul', 'a', 'b'), ('imul', 'a', 'c')) => ('imul', 'a', ('iadd', 'b', 'c'))
 *    ('~fadd', ('fneg', 'a'), 'a') => 0.0
 *    ('iadd', ('ineg', 'a'), 'a') => 0
 *    ('iadd', ('ineg', 'a'), ('iadd', 'a', 'b')) => b
 *    ('iadd', 'a', ('iadd', ('ineg', 'a'), 'b')) => b
 *    ('~fadd', ('fneg', 'a'), ('fadd', 'a', 'b')) => b
 *    ('~fadd', 'a', ('fadd', ('fneg', 'a'), 'b')) => b
 *    ('~flrp', ('fadd(is_used_once)', 'a', -1.0), ('fadd(is_used_once)', 'a', 1.0), 'd') => ('fadd', ('flrp', -1.0, 1.0, 'd'), 'a')
 *    ('~flrp', ('fadd(is_used_once)', 'a', 1.0), ('fadd(is_used_once)', 'a', -1.0), 'd') => ('fadd', ('flrp', 1.0, -1.0, 'd'), 'a')
 *    ('~flrp', ('fadd(is_used_once)', 'a', '#b'), ('fadd(is_used_once)', 'a', '#c'), 'd') => ('fadd', ('fmul', 'd', ('fadd', 'c', ('fneg', 'b'))), ('fadd', 'a', 'b'))
 */


static const nir_search_value_union nir_opt_algebraic_before_ffma_values[] = {
   /* ('~fmul(is_used_once)', ('fmul(is_used_once)', 'a(is_not_const)', '#b'), 'c(is_not_const)') => ('fmul', ('fmul', 'a', 'c'), 'b') */
   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* a */
      false,
      nir_type_invalid,
      0,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 0, 1 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* c */
      false,
      nir_type_invalid,
      0,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 2, 3 },
      0,
   } },

   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 5, 6 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 7, 8 },
      -1,
   } },

   /* ('imul(is_used_once)', ('imul(is_used_once)', 'a(is_not_const)', '#b'), 'c(is_not_const)') => ('imul', ('imul', 'a', 'c'), 'b') */
   /* search2062_0_0 -> 0 in the cache */
   /* search2062_0_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 0, 1 },
      0,
   } },
   /* search2062_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      0, 2,
      { 10, 3 },
      0,
   } },

   /* replace2062_0_0 -> 5 in the cache */
   /* replace2062_0_1 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 5, 6 },
      -1,
   } },
   /* replace2062_1 -> 8 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      0, 2,
      { 12, 8 },
      -1,
   } },

   /* ('~fadd(is_used_once)', ('fadd(is_used_once)', 'a(is_not_const)', '#b'), 'c(is_not_const)') => ('fadd', ('fadd', 'a', 'c'), 'b') */
   /* search2063_0_0 -> 0 in the cache */
   /* search2063_0_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 0, 1 },
      0,
   } },
   /* search2063_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 14, 3 },
      0,
   } },

   /* replace2063_0_0 -> 5 in the cache */
   /* replace2063_0_1 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 5, 6 },
      -1,
   } },
   /* replace2063_1 -> 8 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 16, 8 },
      -1,
   } },

   /* ('iadd(is_used_once)', ('iadd(is_used_once)', 'a(is_not_const)', '#b'), 'c(is_not_const)') => ('iadd', ('iadd', 'a', 'c'), 'b') */
   /* search2064_0_0 -> 0 in the cache */
   /* search2064_0_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 0, 1 },
      0,
   } },
   /* search2064_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 18, 3 },
      0,
   } },

   /* replace2064_0_0 -> 5 in the cache */
   /* replace2064_0_1 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 5, 6 },
      -1,
   } },
   /* replace2064_1 -> 8 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 20, 8 },
      -1,
   } },

   /* ('~fadd', ('fmul', 'a', 'b'), ('fmul', 'a', 'c')) => ('fmul', 'a', ('fadd', 'b', 'c')) */
   /* search2065_0_0 -> 5 in the cache */
   /* search2065_0_1 -> 8 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 5, 8 },
      -1,
   } },
   /* search2065_1_0 -> 5 in the cache */
   /* search2065_1_1 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 5, 6 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 22, 23 },
      -1,
   } },

   /* replace2065_0 -> 5 in the cache */
   /* replace2065_1_0 -> 8 in the cache */
   /* replace2065_1_1 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 8, 6 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 5, 25 },
      -1,
   } },

   /* ('iadd', ('imul', 'a', 'b'), ('imul', 'a', 'c')) => ('imul', 'a', ('iadd', 'b', 'c')) */
   /* search2066_0_0 -> 5 in the cache */
   /* search2066_0_1 -> 8 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 5, 8 },
      -1,
   } },
   /* search2066_1_0 -> 5 in the cache */
   /* search2066_1_1 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      2, 1,
      { 5, 6 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 3,
      { 27, 28 },
      -1,
   } },

   /* replace2066_0 -> 5 in the cache */
   /* replace2066_1_0 -> 8 in the cache */
   /* replace2066_1_1 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 8, 6 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      0, 2,
      { 5, 30 },
      -1,
   } },

   /* ('~fadd', ('fneg', 'a'), 'a') => 0.0 */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 32 },
      -1,
   } },
   /* search2067_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 33, 32 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },

   /* ('iadd', ('ineg', 'a'), 'a') => 0 */
   /* search2068_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 32 },
      -1,
   } },
   /* search2068_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 36, 32 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_int, { 0x0 /* 0 */ },
   } },

   /* ('iadd', ('ineg', 'a'), ('iadd', 'a', 'b')) => b */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 39 },
      -1,
   } },
   /* search2069_1_0 -> 39 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 39, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 40, 42 },
      -1,
   } },

   /* replace2069 -> 41 in the cache */

   /* ('iadd', 'a', ('iadd', ('ineg', 'a'), 'b')) => b */
   /* search2070_0 -> 39 in the cache */
   /* search2070_1_0_0 -> 39 in the cache */
   /* search2070_1_0 -> 40 in the cache */
   /* search2070_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 40, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 39, 44 },
      -1,
   } },

   /* replace2070 -> 41 in the cache */

   /* ('~fadd', ('fneg', 'a'), ('fadd', 'a', 'b')) => b */
   /* search2071_0_0 -> 39 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 39 },
      -1,
   } },
   /* search2071_1_0 -> 39 in the cache */
   /* search2071_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 39, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 46, 47 },
      -1,
   } },

   /* replace2071 -> 41 in the cache */

   /* ('~fadd', 'a', ('fadd', ('fneg', 'a'), 'b')) => b */
   /* search2072_0 -> 39 in the cache */
   /* search2072_1_0_0 -> 39 in the cache */
   /* search2072_1_0 -> 46 in the cache */
   /* search2072_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 46, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 39, 49 },
      -1,
   } },

   /* replace2072 -> 41 in the cache */

   /* ('~flrp', ('fadd(is_used_once)', 'a', -1.0), ('fadd(is_used_once)', 'a', 1.0), 'd') => ('fadd', ('flrp', -1.0, 1.0, 'd'), 'a') */
   /* search2073_0_0 -> 39 in the cache */
   { .constant = {
      { nir_search_value_constant, -2 },
      nir_type_float, { 0xbff0000000000000 /* -1.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 39, 51 },
      0,
   } },
   /* search2073_1_0 -> 39 in the cache */
   { .constant = {
      { nir_search_value_constant, -2 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 39, 53 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 2,
      { 52, 54, 55 },
      -1,
   } },

   /* replace2073_0_0 -> 51 in the cache */
   /* replace2073_0_1 -> 53 in the cache */
   /* replace2073_0_2 -> 55 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 51, 53, 55 },
      -1,
   } },
   /* replace2073_1 -> 39 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 57, 39 },
      -1,
   } },

   /* ('~flrp', ('fadd(is_used_once)', 'a', 1.0), ('fadd(is_used_once)', 'a', -1.0), 'd') => ('fadd', ('flrp', 1.0, -1.0, 'd'), 'a') */
   /* search2074_0_0 -> 39 in the cache */
   /* search2074_0_1 -> 53 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 39, 53 },
      0,
   } },
   /* search2074_1_0 -> 39 in the cache */
   /* search2074_1_1 -> 51 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 39, 51 },
      0,
   } },
   /* search2074_2 -> 55 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 2,
      { 59, 60, 55 },
      -1,
   } },

   /* replace2074_0_0 -> 53 in the cache */
   /* replace2074_0_1 -> 51 in the cache */
   /* replace2074_0_2 -> 55 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 53, 51, 55 },
      -1,
   } },
   /* replace2074_1 -> 39 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 62, 39 },
      -1,
   } },

   /* ('~flrp', ('fadd(is_used_once)', 'a', '#b'), ('fadd(is_used_once)', 'a', '#c'), 'd') => ('fadd', ('fmul', 'd', ('fadd', 'c', ('fneg', 'b'))), ('fadd', 'a', 'b')) */
   { .variable = {
      { nir_search_value_variable, -4 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      1, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 64, 65 },
      0,
   } },
   /* search2075_1_0 -> 64 in the cache */
   { .variable = {
      { nir_search_value_variable, -4 },
      2, /* c */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 64, 67 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 2,
      { 66, 68, 69 },
      -1,
   } },

   /* replace2075_0_0 -> 69 in the cache */
   { .variable = {
      { nir_search_value_variable, -4 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 72 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 71, 73 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 69, 74 },
      -1,
   } },
   /* replace2075_1_0 -> 64 in the cache */
   /* replace2075_1_1 -> 72 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      3, 1,
      { 64, 72 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 75, 76 },
      -1,
   } },

};

static const nir_search_expression_cond nir_opt_algebraic_before_ffma_expression_cond[] = {
   (is_used_once),
};

static const nir_search_variable_cond nir_opt_algebraic_before_ffma_variable_cond[] = {
   (is_not_const),
};

static const struct transform nir_opt_algebraic_before_ffma_transforms[] = {
   { ~0, ~0, ~0 }, /* Sentinel */

   { 15, 17, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 35, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 15, 17, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 35, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 24, 26, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 35, 0 },
   { 48, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 15, 17, 0 },
   { 34, 35, 0 },
   { 48, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4, 9, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4, 9, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 19, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 37, 38, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 19, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 37, 38, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 29, 31, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 37, 38, 0 },
   { 43, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 19, 21, 0 },
   { 37, 38, 0 },
   { 43, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 11, 13, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 11, 13, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 56, 58, 0 },
   { 61, 63, 0 },
   { 70, 77, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 50, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 15, 17, 0 },
   { 50, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 50, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 15, 17, 0 },
   { 50, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 35, 0 },
   { 48, 41, 0 },
   { 50, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 15, 17, 0 },
   { 34, 35, 0 },
   { 48, 41, 0 },
   { 50, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 45, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 19, 21, 0 },
   { 45, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 45, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 19, 21, 0 },
   { 45, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 37, 38, 0 },
   { 43, 41, 0 },
   { 45, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 19, 21, 0 },
   { 37, 38, 0 },
   { 43, 41, 0 },
   { 45, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

};

static const struct per_op_table nir_opt_algebraic_before_ffma_pass_op_table[nir_num_search_ops] = {
   [nir_op_fmul] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         2,
         3,
         19,
         3,
         3,
         20,
         19,
         20,
         19,
      },
   },
   [nir_op_imul] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         4,
         5,
         28,
         5,
         5,
         29,
         28,
         29,
         28,
      },
   },
   [nir_op_fadd] = {
      .filter = (const uint16_t []) {
         0,
         1,
         2,
         2,
         0,
         0,
         3,
         4,
         0,
         0,
         5,
         0,
         3,
         6,
         4,
         7,
         3,
         6,
         6,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         4,
         4,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 8,
      .table = (const uint16_t []) {
      
         6,
         7,
         6,
         6,
         12,
         13,
         31,
         32,
         7,
         7,
         7,
         7,
         14,
         15,
         33,
         34,
         6,
         7,
         16,
         6,
         12,
         13,
         31,
         32,
         6,
         7,
         6,
         6,
         12,
         17,
         31,
         32,
         12,
         14,
         12,
         12,
         12,
         18,
         32,
         32,
         13,
         15,
         13,
         17,
         18,
         13,
         35,
         36,
         31,
         33,
         31,
         31,
         32,
         35,
         31,
         32,
         32,
         34,
         32,
         32,
         32,
         36,
         32,
         32,
      },
   },
   [nir_op_iadd] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         2,
         2,
         0,
         0,
         3,
         4,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         6,
         4,
         7,
         3,
         6,
         6,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         4,
         4,
         6,
         6,
      },
      
      .num_filtered_states = 8,
      .table = (const uint16_t []) {
      
         8,
         9,
         8,
         8,
         21,
         22,
         37,
         38,
         9,
         9,
         9,
         9,
         23,
         24,
         39,
         40,
         8,
         9,
         25,
         8,
         21,
         22,
         37,
         38,
         8,
         9,
         8,
         8,
         21,
         26,
         37,
         38,
         21,
         23,
         21,
         21,
         21,
         27,
         38,
         38,
         22,
         24,
         22,
         26,
         27,
         22,
         41,
         42,
         37,
         39,
         37,
         37,
         38,
         41,
         37,
         38,
         38,
         40,
         38,
         38,
         38,
         42,
         38,
         38,
      },
   },
   [nir_op_fneg] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         10,
      },
   },
   [nir_op_ineg] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         11,
      },
   },
   [nir_op_flrp] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         0,
         0,
         0,
         30,
         30,
      },
   },
};

/* Mapping from state index to offset in transforms (0 being no transforms) */
static const uint16_t nir_opt_algebraic_before_ffma_transform_offsets[] = {
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   1,
   3,
   5,
   7,
   9,
   11,
   14,
   18,
   20,
   22,
   24,
   26,
   28,
   30,
   32,
   35,
   39,
   41,
   43,
   47,
   49,
   52,
   54,
   57,
   61,
   66,
   68,
   71,
   73,
   76,
   80,
};

static const nir_algebraic_table nir_opt_algebraic_before_ffma_table = {
   .transforms = nir_opt_algebraic_before_ffma_transforms,
   .transform_offsets = nir_opt_algebraic_before_ffma_transform_offsets,
   .pass_op_table = nir_opt_algebraic_before_ffma_pass_op_table,
   .values = nir_opt_algebraic_before_ffma_values,
   .expression_cond = nir_opt_algebraic_before_ffma_expression_cond,
   .variable_cond = nir_opt_algebraic_before_ffma_variable_cond,
};

bool
nir_opt_algebraic_before_ffma(nir_shader *shader)
{
   bool progress = false;
   bool condition_flags[122];
   const nir_shader_compiler_options *options = shader->options;
   const shader_info *info = &shader->info;
   (void) options;
   (void) info;

   /* This is not a great place for this, but it seems to be the best place
    * for it. Check that at most one kind of lowering is requested for
    * bitfield extract and bitfield insert. Otherwise the lowering can fight
    * with each other and optimizations.
    */
   assert((int)options->lower_bitfield_extract +
          (int)options->lower_bitfield_extract_to_shifts <= 1);
   assert((int)options->lower_bitfield_insert +
          (int)options->lower_bitfield_insert_to_shifts +
          (int)options->lower_bitfield_insert_to_bitfield_select <= 1);


   STATIC_ASSERT(78 == ARRAY_SIZE(nir_opt_algebraic_before_ffma_values));
   condition_flags[0] = true;
   condition_flags[1] = !options->lower_bitops;
   condition_flags[2] = options->lower_bitops;
   condition_flags[3] = !options->lower_bitops && (options->lower_int64_options & (nir_lower_imul64 | nir_lower_shift64)) == nir_lower_imul64;
   condition_flags[4] = options->lower_mul_2x32_64;
   condition_flags[5] = !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 16);
   condition_flags[6] = !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[7] = !options->lower_iadd_sat;
   condition_flags[8] = !options->lower_uadd_sat;
   condition_flags[9] = options->has_fmulz && !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[10] = !options->has_sdot_4x8;
   condition_flags[11] = !options->has_udot_4x8;
   condition_flags[12] = !options->has_sudot_4x8;
   condition_flags[13] = !options->has_dot_2x16;
   condition_flags[14] = options->lower_flrp16;
   condition_flags[15] = !options->lower_flrp16;
   condition_flags[16] = options->lower_flrp32;
   condition_flags[17] = !options->lower_flrp32;
   condition_flags[18] = options->lower_flrp64;
   condition_flags[19] = !options->lower_flrp64;
   condition_flags[20] = options->lower_ftrunc;
   condition_flags[21] = options->lower_ftrunc || (options->lower_doubles_options & nir_lower_dtrunc);
   condition_flags[22] = options->lower_ffloor;
   condition_flags[23] = (options->lower_ffloor || (options->lower_doubles_options & nir_lower_dfloor)) && !(options->lower_doubles_options & nir_lower_dfract);
   condition_flags[24] = !options->lower_ffloor;
   condition_flags[25] = !options->lower_ffloor && !(options->lower_doubles_options & nir_lower_dfloor);
   condition_flags[26] = options->lower_ffract;
   condition_flags[27] = options->lower_ffract || (options->lower_doubles_options & nir_lower_dfract);
   condition_flags[28] = options->lower_fceil;
   condition_flags[29] = options->lower_ffma16;
   condition_flags[30] = options->lower_ffma32;
   condition_flags[31] = options->lower_ffma64;
   condition_flags[32] = options->fuse_ffma16;
   condition_flags[33] = options->fuse_ffma32;
   condition_flags[34] = options->fuse_ffma64;
   condition_flags[35] = options->lower_fdph;
   condition_flags[36] = !options->lower_fdph;
   condition_flags[37] = options->lower_fdot;
   condition_flags[38] = !options->lower_fsat;
   condition_flags[39] = !options->lower_iabs;
   condition_flags[40] = options->lower_fsat;
   condition_flags[41] = !options->lower_fsat && !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[42] = !options->lower_fsign;
   condition_flags[43] = options->lower_umax;
   condition_flags[44] = options->lower_umin;
   condition_flags[45] = !options->lower_umax;
   condition_flags[46] = !options->lower_umin;
   condition_flags[47] = options->lower_scmp;
   condition_flags[48] = options->lower_vector_cmp;
   condition_flags[49] = !options->lower_rotate;
   condition_flags[50] = options->lower_rotate;
   condition_flags[51] = options->lower_fpow;
   condition_flags[52] = !options->lower_fpow;
   condition_flags[53] = options->lower_fdiv;
   condition_flags[54] = options->lower_fsqrt;
   condition_flags[55] = !options->lower_fsqrt;
   condition_flags[56] = options->lower_sincos;
   condition_flags[57] = !(options->lower_doubles_options & nir_lower_fp64_full_software);
   condition_flags[58] = !options->lower_extract_byte;
   condition_flags[59] = !options->lower_extract_word;
   condition_flags[60] = options->has_pack_32_4x8;
   condition_flags[61] = options->lower_pack_64_2x32_split;
   condition_flags[62] = options->lower_pack_32_2x16_split;
   condition_flags[63] = options->lower_unpack_64_2x32_split;
   condition_flags[64] = options->lower_unpack_32_2x16_split;
   condition_flags[65] = options->has_find_msb_rev;
   condition_flags[66] = !options->lower_find_msb_to_reverse;
   condition_flags[67] = !options->lower_find_lsb;
   condition_flags[68] = options->lower_fmod;
   condition_flags[69] = options->lower_uadd_carry;
   condition_flags[70] = options->lower_usub_borrow;
   condition_flags[71] = options->lower_bitfield_insert;
   condition_flags[72] = options->lower_hadd;
   condition_flags[73] = options->lower_hadd64 || (options->lower_int64_options & nir_lower_iadd64) != 0;
   condition_flags[74] = options->lower_mul_32x16;
   condition_flags[75] = options->lower_uadd_sat || (options->lower_int64_options & nir_lower_iadd64) != 0;
   condition_flags[76] = options->lower_uadd_sat;
   condition_flags[77] = options->lower_usub_sat;
   condition_flags[78] = (options->lower_int64_options & nir_lower_usub_sat64) != 0;
   condition_flags[79] = (options->lower_int64_options & nir_lower_iadd_sat64) != 0;
   condition_flags[80] = (options->lower_int64_options & nir_lower_minmax64) != 0;
   condition_flags[81] = (options->lower_int64_options & nir_lower_icmp64) != 0;
   condition_flags[82] = options->lower_bitfield_insert_to_shifts;
   condition_flags[83] = options->lower_bitfield_insert_to_bitfield_select;
   condition_flags[84] = options->lower_bitfield_extract;
   condition_flags[85] = options->lower_bitfield_insert_to_bitfield_select || options->lower_bitfield_insert;
   condition_flags[86] = options->lower_bitfield_extract_to_shifts;
   condition_flags[87] = options->lower_ifind_msb;
   condition_flags[88] = options->lower_find_msb_to_reverse;
   condition_flags[89] = options->lower_uclz;
   condition_flags[90] = options->lower_find_lsb;
   condition_flags[91] = options->lower_extract_byte;
   condition_flags[92] = options->lower_extract_word;
   condition_flags[93] = options->lower_pack_unorm_2x16;
   condition_flags[94] = options->lower_pack_unorm_4x8;
   condition_flags[95] = options->lower_pack_snorm_2x16;
   condition_flags[96] = options->lower_pack_snorm_4x8;
   condition_flags[97] = options->lower_unpack_unorm_2x16;
   condition_flags[98] = options->lower_unpack_unorm_4x8;
   condition_flags[99] = options->lower_unpack_snorm_2x16;
   condition_flags[100] = options->lower_unpack_snorm_4x8;
   condition_flags[101] = options->lower_pack_split;
   condition_flags[102] = options->lower_isign;
   condition_flags[103] = !options->lower_isign;
   condition_flags[104] = options->lower_fsign;
   condition_flags[105] = !options->has_imul24;
   condition_flags[106] = !options->has_umul24;
   condition_flags[107] = !options->has_umad24;
   condition_flags[108] = options->has_imul24;
   condition_flags[109] = options->has_umad24;
   condition_flags[110] = options->has_umul24;
   condition_flags[111] = !options->lower_uadd_sat && !(options->lower_int64_options & nir_lower_iadd64);
   condition_flags[112] = !options->lower_usub_sat;
   condition_flags[113] = !options->lower_usub_sat && !(options->lower_int64_options & nir_lower_usub_sat64);
   condition_flags[114] = options->lower_iadd_sat;
   condition_flags[115] = options->support_8bit_alu;
   condition_flags[116] = options->support_16bit_alu;
   condition_flags[117] = options->lower_ldexp;
   condition_flags[118] = !options->lower_bitfield_reverse;
   condition_flags[119] = info->stage == MESA_SHADER_COMPUTE && info->cs.derivative_group == DERIVATIVE_GROUP_NONE;
   condition_flags[120] = options->lower_fisnormal;
   condition_flags[121] = !options->vectorize_vec2_16bit;

   nir_foreach_function(function, shader) {
      if (function->impl) {
         progress |= nir_algebraic_impl(function->impl, condition_flags,
                                        &nir_opt_algebraic_before_ffma_table);
      }
   }

   return progress;
}


#include "nir.h"
#include "nir_builder.h"
#include "nir_search.h"
#include "nir_search_helpers.h"

/* What follows is NIR algebraic transform code for the following 211
 * transforms:
 *    ('flt', ('fadd(is_used_once)', 'a', 'b'), 0.0) => ('flt', 'a', ('fneg', 'b'))
 *    ('flt', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', 'b')), 0.0) => ('flt', ('fneg', 'a'), 'b')
 *    ('flt', 0.0, ('fadd(is_used_once)', 'a', 'b')) => ('flt', ('fneg', 'a'), 'b')
 *    ('flt', 0.0, ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', 'b'))) => ('flt', 'a', ('fneg', 'b'))
 *    ('~fge', ('fadd(is_used_once)', 'a', 'b'), 0.0) => ('fge', 'a', ('fneg', 'b'))
 *    ('~fge', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', 'b')), 0.0) => ('fge', ('fneg', 'a'), 'b')
 *    ('~fge', 0.0, ('fadd(is_used_once)', 'a', 'b')) => ('fge', ('fneg', 'a'), 'b')
 *    ('~fge', 0.0, ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', 'b'))) => ('fge', 'a', ('fneg', 'b'))
 *    ('~feq', ('fadd(is_used_once)', 'a', 'b'), 0.0) => ('feq', 'a', ('fneg', 'b'))
 *    ('~fneu', ('fadd(is_used_once)', 'a', 'b'), 0.0) => ('fneu', 'a', ('fneg', 'b'))
 *    ('fge', ('fadd(is_used_once)', 'a(is_finite)', 'b'), 0.0) => ('fge', 'a', ('fneg', 'b'))
 *    ('fge', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a(is_finite)', 'b')), 0.0) => ('fge', ('fneg', 'a'), 'b')
 *    ('fge', 0.0, ('fadd(is_used_once)', 'a(is_finite)', 'b')) => ('fge', ('fneg', 'a'), 'b')
 *    ('fge', 0.0, ('fneg(is_used_once)', ('fadd(is_used_once)', 'a(is_finite)', 'b'))) => ('fge', 'a', ('fneg', 'b'))
 *    ('feq', ('fadd(is_used_once)', 'a(is_finite)', 'b'), 0.0) => ('feq', 'a', ('fneg', 'b'))
 *    ('fneu', ('fadd(is_used_once)', 'a(is_finite)', 'b'), 0.0) => ('fneu', 'a', ('fneg', 'b'))
 *    ('iand', ('fneu', 'a', 'b'), ('iand', ('feq', 'a', 'a'), ('feq', 'b', 'b'))) => ('ior', ('!flt', 'a', 'b'), ('!flt', 'b', 'a'))
 *    ('iand', ('fneu', 'a', 0.0), ('feq', 'a', 'a')) => ('!flt', 0.0, ('fabs', 'a'))
 *    ('ior', ('feq', 'a', 'b'), ('ior', ('fneu', 'a', 'a'), ('fneu', 'b', 'b'))) => ('inot', ('ior', ('!flt', 'a', 'b'), ('!flt', 'b', 'a')))
 *    ('ior', ('feq', 'a', 0.0), ('fneu', 'a', 'a')) => ('inot', ('!flt', 0.0, ('fabs', 'a')))
 *    ('fneg', ('fneg', 'a')) => a
 *    ('~fadd@16', ('fmul(is_only_used_by_fadd)', 'a', 'b'), 'c') => ('ffma', 'a', 'b', 'c')
 *    ('~fadd@16', ('fneg(is_only_used_by_fadd)', ('fmul(is_only_used_by_fadd)', 'a', 'b')), 'c') => ('ffma', ('fneg', 'a'), 'b', 'c')
 *    ('~fadd@16', ('fabs(is_only_used_by_fadd)', ('fmul(is_only_used_by_fadd)', 'a', 'b')), 'c') => ('ffma', ('fabs', 'a'), ('fabs', 'b'), 'c')
 *    ('~fadd@16', ('fneg(is_only_used_by_fadd)', ('fabs', ('fmul(is_only_used_by_fadd)', 'a', 'b'))), 'c') => ('ffma', ('fneg', ('fabs', 'a')), ('fabs', 'b'), 'c')
 *    ('~fadd@32', ('fmul(is_only_used_by_fadd)', 'a', 'b'), 'c') => ('ffma', 'a', 'b', 'c')
 *    ('~fadd@32', ('fneg(is_only_used_by_fadd)', ('fmul(is_only_used_by_fadd)', 'a', 'b')), 'c') => ('ffma', ('fneg', 'a'), 'b', 'c')
 *    ('~fadd@32', ('fabs(is_only_used_by_fadd)', ('fmul(is_only_used_by_fadd)', 'a', 'b')), 'c') => ('ffma', ('fabs', 'a'), ('fabs', 'b'), 'c')
 *    ('~fadd@32', ('fneg(is_only_used_by_fadd)', ('fabs', ('fmul(is_only_used_by_fadd)', 'a', 'b'))), 'c') => ('ffma', ('fneg', ('fabs', 'a')), ('fabs', 'b'), 'c')
 *    ('~fadd@32', ('fmulz(is_only_used_by_fadd)', 'a', 'b'), 'c') => ('ffmaz', 'a', 'b', 'c')
 *    ('~fadd@32', ('fneg(is_only_used_by_fadd)', ('fmulz(is_only_used_by_fadd)', 'a', 'b')), 'c') => ('ffmaz', ('fneg', 'a'), 'b', 'c')
 *    ('~fadd@32', ('fabs(is_only_used_by_fadd)', ('fmulz(is_only_used_by_fadd)', 'a', 'b')), 'c') => ('ffmaz', ('fabs', 'a'), ('fabs', 'b'), 'c')
 *    ('~fadd@32', ('fneg(is_only_used_by_fadd)', ('fabs', ('fmulz(is_only_used_by_fadd)', 'a', 'b'))), 'c') => ('ffmaz', ('fneg', ('fabs', 'a')), ('fabs', 'b'), 'c')
 *    ('~fadd@64', ('fmul(is_only_used_by_fadd)', 'a', 'b'), 'c') => ('ffma', 'a', 'b', 'c')
 *    ('~fadd@64', ('fneg(is_only_used_by_fadd)', ('fmul(is_only_used_by_fadd)', 'a', 'b')), 'c') => ('ffma', ('fneg', 'a'), 'b', 'c')
 *    ('~fadd@64', ('fabs(is_only_used_by_fadd)', ('fmul(is_only_used_by_fadd)', 'a', 'b')), 'c') => ('ffma', ('fabs', 'a'), ('fabs', 'b'), 'c')
 *    ('~fadd@64', ('fneg(is_only_used_by_fadd)', ('fabs', ('fmul(is_only_used_by_fadd)', 'a', 'b'))), 'c') => ('ffma', ('fneg', ('fabs', 'a')), ('fabs', 'b'), 'c')
 *    ('fadd@8', 'a', ('fneg', 'b')) => ('fsub', 'a', 'b')
 *    ('fadd@16', 'a', ('fneg', 'b')) => ('fsub', 'a', 'b')
 *    ('fadd@32', 'a', ('fneg', 'b')) => ('fsub', 'a', 'b')
 *    ('fadd@64', 'a', ('fneg', 'b')) => ('fsub', 'a', 'b')
 *    ('fneg', 'a') => ('fmul', 'a', -1.0)
 *    ('iadd', 'a', ('ineg', 'b')) => ('isub', 'a', 'b')
 *    ('ineg', 'a') => ('isub', 0, 'a')
 *    ('iabs', 'a') => ('imax', 'a', ('ineg', 'a'))
 *    ('iadd', ('iadd(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), 'c(is_not_const)') => ('iadd3', 'a', 'b', 'c')
 *    ('iadd', ('isub(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), 'c(is_not_const)') => ('iadd3', 'a', ('ineg', 'b'), 'c')
 *    ('isub', ('isub(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), 'c(is_not_const)') => ('iadd3', 'a', ('ineg', 'b'), ('ineg', 'c'))
 *    ('vec2(is_only_used_as_float)', ('fneg@16', 'a'), 'b') => ('fmul', ('vec2', 'a', 'b'), ('vec2', -1.0, 1.0))
 *    ('vec2(is_only_used_as_float)', 'a', ('fneg@16', 'b')) => ('fmul', ('vec2', 'a', 'b'), ('vec2', 1.0, -1.0))
 *    ('flt', '#b(is_gt_0_and_lt_1)', ('fsat(is_used_once)', 'a')) => ('flt', 'b', 'a')
 *    ('fge', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('fge', 'a', 'b')
 *    ('feq', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('feq', 'a', 'b')
 *    ('fneu', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('fneu', 'a', 'b')
 *    ('fge', ('fsat(is_used_once)', 'a'), 1.0) => ('fge', 'a', 1.0)
 *    ('~fge', ('fmin(is_used_once)', ('fadd(is_used_once)', 'a', 'b'), ('fadd', 'c', 'd')), 0.0) => ('iand', ('fge', 'a', ('fneg', 'b')), ('fge', 'c', ('fneg', 'd')))
 *    ('flt', ('fneg', 'a'), ('fneg', 'b')) => ('flt', 'b', 'a')
 *    ('fge', ('fneg', 'a'), ('fneg', 'b')) => ('fge', 'b', 'a')
 *    ('feq', ('fneg', 'a'), ('fneg', 'b')) => ('feq', 'b', 'a')
 *    ('fneu', ('fneg', 'a'), ('fneg', 'b')) => ('fneu', 'b', 'a')
 *    ('flt', ('fneg', 'a'), -1.0) => ('flt', 1.0, 'a')
 *    ('flt', -1.0, ('fneg', 'a')) => ('flt', 'a', 1.0)
 *    ('fge', ('fneg', 'a'), -1.0) => ('fge', 1.0, 'a')
 *    ('fge', -1.0, ('fneg', 'a')) => ('fge', 'a', 1.0)
 *    ('fneu', ('fneg', 'a'), -1.0) => ('fneu', 1.0, 'a')
 *    ('feq', -1.0, ('fneg', 'a')) => ('feq', 'a', 1.0)
 *    ('ior', 'a', 'a') => a
 *    ('iand', 'a', 'a') => a
 *    ('~fadd', ('fneg(is_used_once)', ('fsat(is_used_once)', 'a(is_not_fmul)')), 1.0) => ('fsat', ('fadd', 1.0, ('fneg', 'a')))
 *    ('fdot2', 'a', 'b') => ('fdot2_replicated', 'a', 'b')
 *    ('fdot3', 'a', 'b') => ('fdot3_replicated', 'a', 'b')
 *    ('fdot4', 'a', 'b') => ('fdot4_replicated', 'a', 'b')
 *    ('fdph', 'a', 'b') => ('fdph_replicated', 'a', 'b')
 *    ('~flrp', ('fadd(is_used_once)', 'a', 'b'), ('fadd(is_used_once)', 'a', 'c'), 'd') => ('fadd', ('flrp', 'b', 'c', 'd'), 'a')
 *    ('fround_even', 'a') => ('bcsel', ('feq', ('ffract', 'a'), 0.5), ('fadd', ('ffloor', ('fadd', 'a', 0.5)), 1.0), ('ffloor', ('fadd', 'a', 0.5)))
 *    ('~ffma@32', 'a', 2.0, -1.0) => ('flrp', -1.0, 1.0, 'a')
 *    ('~ffma@32', 'a', -2.0, -1.0) => ('flrp', -1.0, 1.0, ('fneg', 'a'))
 *    ('~ffma@32', 'a', -2.0, 1.0) => ('flrp', 1.0, -1.0, 'a')
 *    ('~ffma@32', 'a', 2.0, 1.0) => ('flrp', 1.0, -1.0, ('fneg', 'a'))
 *    ('~fadd@32', ('fmul(is_used_once)', 2.0, 'a'), -1.0) => ('flrp', -1.0, 1.0, 'a')
 *    ('~fadd@32', ('fmul(is_used_once)', -2.0, 'a'), -1.0) => ('flrp', -1.0, 1.0, ('fneg', 'a'))
 *    ('~fadd@32', ('fmul(is_used_once)', -2.0, 'a'), 1.0) => ('flrp', 1.0, -1.0, 'a')
 *    ('~fadd@32', ('fmul(is_used_once)', 2.0, 'a'), 1.0) => ('flrp', 1.0, -1.0, ('fneg', 'a'))
 *    ('~ffma@32', ('fadd', 'b', ('fneg', 'a')), 'a', 'a') => ('flrp', 'a', 'b', 'a')
 *    ('~ffma@32', 'a', 2.0, ('fneg', ('fmul', 'a', 'a'))) => ('flrp', 'a', 1.0, 'a')
 *    ('~ffma@32', 'a', 2.0, ('fmul', ('fneg', 'a'), 'a')) => ('flrp', 'a', 1.0, 'a')
 *    ('~ffma@32', 'a', ('fneg', 'a'), ('fmul', 2.0, 'a')) => ('flrp', 'a', 1.0, 'a')
 *    ('~fmul@32', 'a', ('fadd', 2.0, ('fneg', 'a'))) => ('flrp', 'a', 1.0, 'a')
 *    ('fmin', ('fadd(is_used_once)', '#c', 'a'), ('fadd(is_used_once)', '#c', 'b')) => ('fadd', 'c', ('fmin', 'a', 'b'))
 *    ('fmax', ('fadd(is_used_once)', '#c', 'a'), ('fadd(is_used_once)', '#c', 'b')) => ('fadd', 'c', ('fmax', 'a', 'b'))
 *    ('bcsel', ('feq', ('fsqrt', 'a(is_not_negative)'), 0.0), 3.4028234663852886e+38, ('frsq', 'a')) => ('fmin', ('frsq', 'a'), 3.4028234663852886e+38)
 *    ('~fadd', ('ffma(is_used_once)', 'a', 'b', ('ffma', 'c', 'd', ('fmul(is_used_once)', 'e(is_not_const_and_not_fsign)', 'f(is_not_const_and_not_fsign)'))), 'g(is_not_const)') => ('ffma', 'a', 'b', ('ffma', 'c', 'd', ('ffma', 'e', 'f', 'g')))
 *    ('~fadd', ('ffma(is_used_once)', 'a', 'b', ('fmul(is_used_once)', 'c(is_not_const_and_not_fsign)', 'd(is_not_const_and_not_fsign)')), 'e(is_not_const)') => ('ffma', 'a', 'b', ('ffma', 'c', 'd', 'e'))
 *    ('~fadd', ('fneg', ('ffma(is_used_once)', 'a', 'b', ('ffma', 'c', 'd', ('fmul(is_used_once)', 'e(is_not_const_and_not_fsign)', 'f(is_not_const_and_not_fsign)')))), 'g(is_not_const)') => ('ffma', ('fneg', 'a'), 'b', ('ffma', ('fneg', 'c'), 'd', ('ffma', ('fneg', 'e'), 'f', 'g')))
 *    ('~fadd', ('ffmaz(is_used_once)', 'a', 'b', ('ffmaz', 'c', 'd', ('fmulz(is_used_once)', 'e(is_not_const_and_not_fsign)', 'f(is_not_const_and_not_fsign)'))), 'g(is_not_const)') => ('ffmaz', 'a', 'b', ('ffmaz', 'c', 'd', ('ffmaz', 'e', 'f', 'g')))
 *    ('~fadd', ('ffmaz(is_used_once)', 'a', 'b', ('fmulz(is_used_once)', 'c(is_not_const_and_not_fsign)', 'd(is_not_const_and_not_fsign)')), 'e(is_not_const)') => ('ffmaz', 'a', 'b', ('ffmaz', 'c', 'd', 'e'))
 *    ('~fadd', ('fneg', ('ffmaz(is_used_once)', 'a', 'b', ('ffmaz', 'c', 'd', ('fmulz(is_used_once)', 'e(is_not_const_and_not_fsign)', 'f(is_not_const_and_not_fsign)')))), 'g(is_not_const)') => ('ffmaz', ('fneg', 'a'), 'b', ('ffmaz', ('fneg', 'c'), 'd', ('ffmaz', ('fneg', 'e'), 'f', 'g')))
 *    ('ubfe', 'a', 'b', 0) => 0
 *    ('ibfe', 'a', 'b', 0) => 0
 *    ('ubfe', 'a', '#b', '#c') => ('iand', ('ushr', 4294967295, ('ineg', 'c')), ('ushr', 'a', 'b'))
 *    ('ibfe', 'a', '#b', '#c') => ('ishr', ('ishl', 'a', ('ineg', ('iadd', 'b', 'c'))), ('ineg', 'c'))
 *    ('ishl', 'a', 0) => a
 *    ('ishl', 'a', -32) => a
 *    ('ishr', 'a', 0) => a
 *    ('ishr', 'a', -32) => a
 *    ('ushr', 'a', 0) => a
 *    ('extract_i8', ('extract_i8', 'a', 'b'), 0) => ('extract_i8', 'a', 'b')
 *    ('extract_i8', ('extract_u8', 'a', 'b'), 0) => ('extract_i8', 'a', 'b')
 *    ('extract_u8', ('extract_i8', 'a', 'b'), 0) => ('extract_u8', 'a', 'b')
 *    ('extract_u8', ('extract_u8', 'a', 'b'), 0) => ('extract_u8', 'a', 'b')
 *    ('u2u16', ('u2u8', 'a@16')) => ('extract_u8', 'a', 0)
 *    ('u2u16', ('i2i8', 'a@16')) => ('extract_u8', 'a', 0)
 *    ('i2i16', ('u2u8', 'a@16')) => ('extract_i8', 'a', 0)
 *    ('i2i16', ('i2i8', 'a@16')) => ('extract_i8', 'a', 0)
 *    ('u2u32', ('u2u8', 'a@32')) => ('extract_u8', 'a', 0)
 *    ('u2u32', ('i2i8', 'a@32')) => ('extract_u8', 'a', 0)
 *    ('u2u32', ('u2u16', 'a@32')) => ('extract_u16', 'a', 0)
 *    ('u2u32', ('i2i16', 'a@32')) => ('extract_u16', 'a', 0)
 *    ('i2i32', ('u2u8', 'a@32')) => ('extract_i8', 'a', 0)
 *    ('i2i32', ('i2i8', 'a@32')) => ('extract_i8', 'a', 0)
 *    ('i2i32', ('u2u16', 'a@32')) => ('extract_i16', 'a', 0)
 *    ('i2i32', ('i2i16', 'a@32')) => ('extract_i16', 'a', 0)
 *    ('ishl', ('extract_u8', 'a@32', 0), 8) => ('insert_u8', 'a', 1)
 *    ('ishl', ('extract_u8', 'a@32', 0), 16) => ('insert_u8', 'a', 2)
 *    ('ishl', ('extract_u8', 'a@32', 0), 24) => ('insert_u8', 'a', 3)
 *    ('iand', ('ishl', 'a@32', 8), 65280) => ('insert_u8', 'a', 1)
 *    ('iand', ('ishl', 'a@32', 16), 16711680) => ('insert_u8', 'a', 2)
 *    ('iand', ('ishl', 'a@32', 24), 4278190080) => ('insert_u8', 'a', 3)
 *    ('ishl', 'a@32', 24) => ('insert_u8', 'a', 3)
 *    ('ishl', 'a@32', 16) => ('insert_u16', 'a', 1)
 *    ('insert_u8', ('extract_u8', 'a', 0), 'b') => ('insert_u8', 'a', 'b')
 *    ('insert_u16', ('extract_u16', 'a', 0), 'b') => ('insert_u16', 'a', 'b')
 *    ('iand', ('ine(is_used_once)', 'a@8', 0), ('ine', 'b@8', 0)) => ('ine', ('umin', 'a', 'b'), 0)
 *    ('ior', ('ieq(is_used_once)', 'a@8', 0), ('ieq', 'b@8', 0)) => ('ieq', ('umin', 'a', 'b'), 0)
 *    ('iand', ('ine(is_used_once)', 'a@16', 0), ('ine', 'b@16', 0)) => ('ine', ('umin', 'a', 'b'), 0)
 *    ('ior', ('ieq(is_used_once)', 'a@16', 0), ('ieq', 'b@16', 0)) => ('ieq', ('umin', 'a', 'b'), 0)
 *    ('iand', ('ine(is_used_once)', 'a@32', 0), ('ine', 'b@32', 0)) => ('ine', ('umin', 'a', 'b'), 0)
 *    ('ior', ('ieq(is_used_once)', 'a@32', 0), ('ieq', 'b@32', 0)) => ('ieq', ('umin', 'a', 'b'), 0)
 *    ('iand', ('ine(is_used_once)', 'a@64', 0), ('ine', 'b@64', 0)) => ('ine', ('umin', 'a', 'b'), 0)
 *    ('ior', ('ieq(is_used_once)', 'a@64', 0), ('ieq', 'b@64', 0)) => ('ieq', ('umin', 'a', 'b'), 0)
 *    ('~fadd@16', 1.0, ('fmul(is_used_once)', 'c', ('fadd', 'b', -1.0))) => ('fadd', ('fadd', 1.0, ('fneg', 'c')), ('fmul', 'b', 'c'))
 *    ('bcsel', 'a', 0, ('b2f16', ('inot', 'b@bool'))) => ('b2f16', ('inot', ('ior', 'a', 'b')))
 *    ('~fadd@32', 1.0, ('fmul(is_used_once)', 'c', ('fadd', 'b', -1.0))) => ('fadd', ('fadd', 1.0, ('fneg', 'c')), ('fmul', 'b', 'c'))
 *    ('bcsel', 'a', 0, ('b2f32', ('inot', 'b@bool'))) => ('b2f32', ('inot', ('ior', 'a', 'b')))
 *    ('~fadd@64', 1.0, ('fmul(is_used_once)', 'c', ('fadd', 'b', -1.0))) => ('fadd', ('fadd', 1.0, ('fneg', 'c')), ('fmul', 'b', 'c'))
 *    ('bcsel', 'a', 0, ('b2f64', ('inot', 'b@bool'))) => ('b2f64', ('inot', ('ior', 'a', 'b')))
 *    ('bcsel', 'a', ('fadd(is_used_once)', 'b', 'c'), ('fadd', 'b', 'd')) => ('fadd', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fadd', 'b', 'c'), ('fadd(is_used_once)', 'b', 'd')) => ('fadd', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('ffma(is_used_once)', 'b', 'c', 'd'), ('ffma', 'b', 'c', 'e')) => ('ffma', 'b', 'c', ('bcsel', 'a', 'd', 'e'))
 *    ('bcsel', 'a', ('ffma', 'b', 'c', 'd'), ('ffma(is_used_once)', 'b', 'c', 'e')) => ('ffma', 'b', 'c', ('bcsel', 'a', 'd', 'e'))
 *    ('bcsel', 'a', ('ffma(is_used_once)', 'b', 'c', 'd'), ('ffma', 'b', 'e', 'd')) => ('ffma', 'b', ('bcsel', 'a', 'c', 'e'), 'd')
 *    ('bcsel', 'a', ('ffma', 'b', 'c', 'd'), ('ffma(is_used_once)', 'b', 'e', 'd')) => ('ffma', 'b', ('bcsel', 'a', 'c', 'e'), 'd')
 *    ('bcsel', 'a', ('ffmaz(is_used_once)', 'b', 'c', 'd'), ('ffmaz', 'b', 'c', 'e')) => ('ffmaz', 'b', 'c', ('bcsel', 'a', 'd', 'e'))
 *    ('bcsel', 'a', ('ffmaz', 'b', 'c', 'd'), ('ffmaz(is_used_once)', 'b', 'c', 'e')) => ('ffmaz', 'b', 'c', ('bcsel', 'a', 'd', 'e'))
 *    ('bcsel', 'a', ('ffmaz(is_used_once)', 'b', 'c', 'd'), ('ffmaz', 'b', 'e', 'd')) => ('ffmaz', 'b', ('bcsel', 'a', 'c', 'e'), 'd')
 *    ('bcsel', 'a', ('ffmaz', 'b', 'c', 'd'), ('ffmaz(is_used_once)', 'b', 'e', 'd')) => ('ffmaz', 'b', ('bcsel', 'a', 'c', 'e'), 'd')
 *    ('~f2f32', ('fabs', ('f2fmp', 'a'))) => ('fabs', 'a')
 *    ('~f2f32', ('fceil', ('f2fmp', 'a'))) => ('fceil', 'a')
 *    ('~f2f32', ('fcos', ('f2fmp', 'a'))) => ('fcos', 'a')
 *    ('~f2f32', ('fddx', ('f2fmp', 'a'))) => ('fddx', 'a')
 *    ('~f2f32', ('fddx_coarse', ('f2fmp', 'a'))) => ('fddx_coarse', 'a')
 *    ('~f2f32', ('fddx_fine', ('f2fmp', 'a'))) => ('fddx_fine', 'a')
 *    ('~f2f32', ('fddy', ('f2fmp', 'a'))) => ('fddy', 'a')
 *    ('~f2f32', ('fddy_coarse', ('f2fmp', 'a'))) => ('fddy_coarse', 'a')
 *    ('~f2f32', ('fddy_fine', ('f2fmp', 'a'))) => ('fddy_fine', 'a')
 *    ('~f2f32', ('fexp2', ('f2fmp', 'a'))) => ('fexp2', 'a')
 *    ('~f2f32', ('ffloor', ('f2fmp', 'a'))) => ('ffloor', 'a')
 *    ('~f2f32', ('ffract', ('f2fmp', 'a'))) => ('ffract', 'a')
 *    ('~f2f32', ('flog2', ('f2fmp', 'a'))) => ('flog2', 'a')
 *    ('~f2f32', ('fneg', ('f2fmp', 'a'))) => ('fneg', 'a')
 *    ('~f2f32', ('frcp', ('f2fmp', 'a'))) => ('frcp', 'a')
 *    ('~f2f32', ('fround_even', ('f2fmp', 'a'))) => ('fround_even', 'a')
 *    ('~f2f32', ('frsq', ('f2fmp', 'a'))) => ('frsq', 'a')
 *    ('~f2f32', ('fsat', ('f2fmp', 'a'))) => ('fsat', 'a')
 *    ('~f2f32', ('fsign', ('f2fmp', 'a'))) => ('fsign', 'a')
 *    ('~f2f32', ('fsin', ('f2fmp', 'a'))) => ('fsin', 'a')
 *    ('~f2f32', ('fsqrt', ('f2fmp', 'a'))) => ('fsqrt', 'a')
 *    ('~f2f32', ('fadd', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fadd', 'a', 'b')
 *    ('~f2f32', ('fdiv', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fdiv', 'a', 'b')
 *    ('~f2f32', ('fmax', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fmax', 'a', 'b')
 *    ('~f2f32', ('fmin', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fmin', 'a', 'b')
 *    ('~f2f32', ('fmod', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fmod', 'a', 'b')
 *    ('~f2f32', ('fmul', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fmul', 'a', 'b')
 *    ('~f2f32', ('fpow', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fpow', 'a', 'b')
 *    ('~f2f32', ('frem', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('frem', 'a', 'b')
 *    ('~f2f32', ('ffma', ('f2fmp', 'a'), ('f2fmp', 'b'), ('f2fmp', 'c'))) => ('ffma', 'a', 'b', 'c')
 *    ('~f2f32', ('flrp', ('f2fmp', 'a'), ('f2fmp', 'b'), ('f2fmp', 'c'))) => ('flrp', 'a', 'b', 'c')
 *    ('~feq', ('f2fmp', 'a'), ('f2fmp', 'b')) => ('feq', 'a', 'b')
 *    ('~fge', ('f2fmp', 'a'), ('f2fmp', 'b')) => ('fge', 'a', 'b')
 *    ('~flt', ('f2fmp', 'a'), ('f2fmp', 'b')) => ('flt', 'a', 'b')
 *    ('~fneu', ('f2fmp', 'a'), ('f2fmp', 'b')) => ('fneu', 'a', 'b')
 *    ('f2fmp', 'a') => ('f2f16', 'a')
 *    ('f2imp', 'a') => ('f2i16', 'a')
 *    ('f2ump', 'a') => ('f2u16', 'a')
 *    ('i2imp', 'a') => ('i2i16', 'a')
 *    ('i2fmp', 'a') => ('i2f16', 'a')
 *    ('i2imp', 'a') => ('u2u16', 'a')
 *    ('u2fmp', 'a') => ('u2f16', 'a')
 *    ('fisfinite', 'a') => ('flt', ('fabs', 'a'), inf)
 *    ('fcsel', ('slt', 0, 'a'), 'b', 'c') => ('fcsel_gt', 'a', 'b', 'c')
 *    ('fcsel', ('slt', 'a', 0), 'b', 'c') => ('fcsel_gt', ('fneg', 'a'), 'b', 'c')
 *    ('fcsel', ('sge', 'a', 0), 'b', 'c') => ('fcsel_ge', 'a', 'b', 'c')
 *    ('fcsel', ('sge', 0, 'a'), 'b', 'c') => ('fcsel_ge', ('fneg', 'a'), 'b', 'c')
 *    ('bcsel', ('ilt', 0, 'a@32'), 'b@32', 'c@32') => ('i32csel_gt', 'a', 'b', 'c')
 *    ('bcsel', ('ilt', 'a@32', 0), 'b@32', 'c@32') => ('i32csel_ge', 'a', 'c', 'b')
 *    ('bcsel', ('ige', 'a@32', 0), 'b@32', 'c@32') => ('i32csel_ge', 'a', 'b', 'c')
 *    ('bcsel', ('ige', 0, 'a@32'), 'b@32', 'c@32') => ('i32csel_gt', 'a', 'c', 'b')
 *    ('bcsel', ('flt', 0, 'a@32'), 'b@32', 'c@32') => ('fcsel_gt', 'a', 'b', 'c')
 *    ('bcsel', ('flt', 'a@32', 0), 'b@32', 'c@32') => ('fcsel_gt', ('fneg', 'a'), 'b', 'c')
 *    ('bcsel', ('fge', 'a@32', 0), 'b@32', 'c@32') => ('fcsel_ge', 'a', 'b', 'c')
 *    ('bcsel', ('fge', 0, 'a@32'), 'b@32', 'c@32') => ('fcsel_ge', ('fneg', 'a'), 'b', 'c')
 */


static const nir_search_value_union nir_opt_algebraic_late_values[] = {
   /* ('flt', ('fadd(is_used_once)', 'a', 'b'), 0.0) => ('flt', 'a', ('fneg', 'b')) */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 0, 1 },
      0,
   } },
   { .constant = {
      { nir_search_value_constant, -2 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 2, 3 },
      -1,
   } },

   /* replace2076_0 -> 0 in the cache */
   /* replace2076_1_0 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 1 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 0, 5 },
      -1,
   } },

   /* ('flt', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', 'b')), 0.0) => ('flt', ('fneg', 'a'), 'b') */
   /* search2077_0_0_0 -> 0 in the cache */
   /* search2077_0_0_1 -> 1 in the cache */
   /* search2077_0_0 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 2 },
      0,
   } },
   /* search2077_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 7, 3 },
      -1,
   } },

   /* replace2077_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace2077_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 9, 1 },
      -1,
   } },

   /* ('flt', 0.0, ('fadd(is_used_once)', 'a', 'b')) => ('flt', ('fneg', 'a'), 'b') */
   /* search2078_0 -> 3 in the cache */
   /* search2078_1_0 -> 0 in the cache */
   /* search2078_1_1 -> 1 in the cache */
   /* search2078_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 3, 2 },
      -1,
   } },

   /* replace2078_0_0 -> 0 in the cache */
   /* replace2078_0 -> 9 in the cache */
   /* replace2078_1 -> 1 in the cache */
   /* replace2078 -> 10 in the cache */

   /* ('flt', 0.0, ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', 'b'))) => ('flt', 'a', ('fneg', 'b')) */
   /* search2079_0 -> 3 in the cache */
   /* search2079_1_0_0 -> 0 in the cache */
   /* search2079_1_0_1 -> 1 in the cache */
   /* search2079_1_0 -> 2 in the cache */
   /* search2079_1 -> 7 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 3, 7 },
      -1,
   } },

   /* replace2079_0 -> 0 in the cache */
   /* replace2079_1_0 -> 1 in the cache */
   /* replace2079_1 -> 5 in the cache */
   /* replace2079 -> 6 in the cache */

   /* ('~fge', ('fadd(is_used_once)', 'a', 'b'), 0.0) => ('fge', 'a', ('fneg', 'b')) */
   /* search2080_0_0 -> 0 in the cache */
   /* search2080_0_1 -> 1 in the cache */
   /* search2080_0 -> 2 in the cache */
   /* search2080_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 2, 3 },
      -1,
   } },

   /* replace2080_0 -> 0 in the cache */
   /* replace2080_1_0 -> 1 in the cache */
   /* replace2080_1 -> 5 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 0, 5 },
      -1,
   } },

   /* ('~fge', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', 'b')), 0.0) => ('fge', ('fneg', 'a'), 'b') */
   /* search2081_0_0_0 -> 0 in the cache */
   /* search2081_0_0_1 -> 1 in the cache */
   /* search2081_0_0 -> 2 in the cache */
   /* search2081_0 -> 7 in the cache */
   /* search2081_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 7, 3 },
      -1,
   } },

   /* replace2081_0_0 -> 0 in the cache */
   /* replace2081_0 -> 9 in the cache */
   /* replace2081_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 9, 1 },
      -1,
   } },

   /* ('~fge', 0.0, ('fadd(is_used_once)', 'a', 'b')) => ('fge', ('fneg', 'a'), 'b') */
   /* search2082_0 -> 3 in the cache */
   /* search2082_1_0 -> 0 in the cache */
   /* search2082_1_1 -> 1 in the cache */
   /* search2082_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 3, 2 },
      -1,
   } },

   /* replace2082_0_0 -> 0 in the cache */
   /* replace2082_0 -> 9 in the cache */
   /* replace2082_1 -> 1 in the cache */
   /* replace2082 -> 16 in the cache */

   /* ('~fge', 0.0, ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', 'b'))) => ('fge', 'a', ('fneg', 'b')) */
   /* search2083_0 -> 3 in the cache */
   /* search2083_1_0_0 -> 0 in the cache */
   /* search2083_1_0_1 -> 1 in the cache */
   /* search2083_1_0 -> 2 in the cache */
   /* search2083_1 -> 7 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 3, 7 },
      -1,
   } },

   /* replace2083_0 -> 0 in the cache */
   /* replace2083_1_0 -> 1 in the cache */
   /* replace2083_1 -> 5 in the cache */
   /* replace2083 -> 14 in the cache */

   /* ('~feq', ('fadd(is_used_once)', 'a', 'b'), 0.0) => ('feq', 'a', ('fneg', 'b')) */
   /* search2084_0_0 -> 0 in the cache */
   /* search2084_0_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 0, 1 },
      0,
   } },
   /* search2084_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 19, 3 },
      -1,
   } },

   /* replace2084_0 -> 0 in the cache */
   /* replace2084_1_0 -> 1 in the cache */
   /* replace2084_1 -> 5 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 0, 5 },
      -1,
   } },

   /* ('~fneu', ('fadd(is_used_once)', 'a', 'b'), 0.0) => ('fneu', 'a', ('fneg', 'b')) */
   /* search2085_0_0 -> 0 in the cache */
   /* search2085_0_1 -> 1 in the cache */
   /* search2085_0 -> 19 in the cache */
   /* search2085_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 19, 3 },
      -1,
   } },

   /* replace2085_0 -> 0 in the cache */
   /* replace2085_1_0 -> 1 in the cache */
   /* replace2085_1 -> 5 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 0, 5 },
      -1,
   } },

   /* ('fge', ('fadd(is_used_once)', 'a(is_finite)', 'b'), 0.0) => ('fge', 'a', ('fneg', 'b')) */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      0,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search2086_0_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 24, 1 },
      0,
   } },
   /* search2086_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 25, 3 },
      -1,
   } },

   /* replace2086_0 -> 0 in the cache */
   /* replace2086_1_0 -> 1 in the cache */
   /* replace2086_1 -> 5 in the cache */
   /* replace2086 -> 14 in the cache */

   /* ('fge', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a(is_finite)', 'b')), 0.0) => ('fge', ('fneg', 'a'), 'b') */
   /* search2087_0_0_0 -> 24 in the cache */
   /* search2087_0_0_1 -> 1 in the cache */
   /* search2087_0_0 -> 25 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 25 },
      0,
   } },
   /* search2087_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 27, 3 },
      -1,
   } },

   /* replace2087_0_0 -> 0 in the cache */
   /* replace2087_0 -> 9 in the cache */
   /* replace2087_1 -> 1 in the cache */
   /* replace2087 -> 16 in the cache */

   /* ('fge', 0.0, ('fadd(is_used_once)', 'a(is_finite)', 'b')) => ('fge', ('fneg', 'a'), 'b') */
   /* search2088_0 -> 3 in the cache */
   /* search2088_1_0 -> 24 in the cache */
   /* search2088_1_1 -> 1 in the cache */
   /* search2088_1 -> 25 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 3, 25 },
      -1,
   } },

   /* replace2088_0_0 -> 0 in the cache */
   /* replace2088_0 -> 9 in the cache */
   /* replace2088_1 -> 1 in the cache */
   /* replace2088 -> 16 in the cache */

   /* ('fge', 0.0, ('fneg(is_used_once)', ('fadd(is_used_once)', 'a(is_finite)', 'b'))) => ('fge', 'a', ('fneg', 'b')) */
   /* search2089_0 -> 3 in the cache */
   /* search2089_1_0_0 -> 24 in the cache */
   /* search2089_1_0_1 -> 1 in the cache */
   /* search2089_1_0 -> 25 in the cache */
   /* search2089_1 -> 27 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 3, 27 },
      -1,
   } },

   /* replace2089_0 -> 0 in the cache */
   /* replace2089_1_0 -> 1 in the cache */
   /* replace2089_1 -> 5 in the cache */
   /* replace2089 -> 14 in the cache */

   /* ('feq', ('fadd(is_used_once)', 'a(is_finite)', 'b'), 0.0) => ('feq', 'a', ('fneg', 'b')) */
   /* search2090_0_0 -> 24 in the cache */
   /* search2090_0_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 24, 1 },
      0,
   } },
   /* search2090_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 31, 3 },
      -1,
   } },

   /* replace2090_0 -> 0 in the cache */
   /* replace2090_1_0 -> 1 in the cache */
   /* replace2090_1 -> 5 in the cache */
   /* replace2090 -> 21 in the cache */

   /* ('fneu', ('fadd(is_used_once)', 'a(is_finite)', 'b'), 0.0) => ('fneu', 'a', ('fneg', 'b')) */
   /* search2091_0_0 -> 24 in the cache */
   /* search2091_0_1 -> 1 in the cache */
   /* search2091_0 -> 31 in the cache */
   /* search2091_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 31, 3 },
      -1,
   } },

   /* replace2091_0 -> 0 in the cache */
   /* replace2091_1_0 -> 1 in the cache */
   /* replace2091_1 -> 5 in the cache */
   /* replace2091 -> 23 in the cache */

   /* ('iand', ('fneu', 'a', 'b'), ('iand', ('feq', 'a', 'a'), ('feq', 'b', 'b'))) => ('ior', ('!flt', 'a', 'b'), ('!flt', 'b', 'a')) */
   /* search2092_0_0 -> 0 in the cache */
   /* search2092_0_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      1, 1,
      { 0, 1 },
      -1,
   } },
   /* search2092_1_0_0 -> 0 in the cache */
   /* search2092_1_0_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      -1, 0,
      { 0, 0 },
      -1,
   } },
   /* search2092_1_1_0 -> 1 in the cache */
   /* search2092_1_1_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      -1, 0,
      { 1, 1 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 35, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 34, 37 },
      -1,
   } },

   /* replace2092_0_0 -> 0 in the cache */
   /* replace2092_0_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_flt,
      -1, 0,
      { 0, 1 },
      -1,
   } },
   /* replace2092_1_0 -> 1 in the cache */
   /* replace2092_1_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_flt,
      -1, 0,
      { 1, 0 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 39, 40 },
      -1,
   } },

   /* ('iand', ('fneu', 'a', 0.0), ('feq', 'a', 'a')) => ('!flt', 0.0, ('fabs', 'a')) */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      1, 1,
      { 42, 43 },
      -1,
   } },
   /* search2093_1_0 -> 42 in the cache */
   /* search2093_1_1 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      -1, 0,
      { 42, 42 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 44, 45 },
      -1,
   } },

   /* replace2093_0 -> 43 in the cache */
   /* replace2093_1_0 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 42 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_flt,
      -1, 0,
      { 43, 47 },
      -1,
   } },

   /* ('ior', ('feq', 'a', 'b'), ('ior', ('fneu', 'a', 'a'), ('fneu', 'b', 'b'))) => ('inot', ('ior', ('!flt', 'a', 'b'), ('!flt', 'b', 'a'))) */
   /* search2094_0_0 -> 0 in the cache */
   /* search2094_0_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      1, 1,
      { 0, 1 },
      -1,
   } },
   /* search2094_1_0_0 -> 0 in the cache */
   /* search2094_1_0_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      -1, 0,
      { 0, 0 },
      -1,
   } },
   /* search2094_1_1_0 -> 1 in the cache */
   /* search2094_1_1_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      -1, 0,
      { 1, 1 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      2, 1,
      { 50, 51 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 49, 52 },
      -1,
   } },

   /* replace2094_0_0_0 -> 0 in the cache */
   /* replace2094_0_0_1 -> 1 in the cache */
   /* replace2094_0_0 -> 39 in the cache */
   /* replace2094_0_1_0 -> 1 in the cache */
   /* replace2094_0_1_1 -> 0 in the cache */
   /* replace2094_0_1 -> 40 in the cache */
   /* replace2094_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 1,
      { 41 },
      -1,
   } },

   /* ('ior', ('feq', 'a', 0.0), ('fneu', 'a', 'a')) => ('inot', ('!flt', 0.0, ('fabs', 'a'))) */
   /* search2095_0_0 -> 42 in the cache */
   /* search2095_0_1 -> 43 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      1, 1,
      { 42, 43 },
      -1,
   } },
   /* search2095_1_0 -> 42 in the cache */
   /* search2095_1_1 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      -1, 0,
      { 42, 42 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 55, 56 },
      -1,
   } },

   /* replace2095_0_0 -> 43 in the cache */
   /* replace2095_0_1_0 -> 42 in the cache */
   /* replace2095_0_1 -> 47 in the cache */
   /* replace2095_0 -> 48 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 48 },
      -1,
   } },

   /* ('fneg', ('fneg', 'a')) => a */
   /* search2096_0_0 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 42 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 59 },
      -1,
   } },

   /* replace2096 -> 42 in the cache */

   /* ('~fadd@16', ('fmul(is_only_used_by_fadd)', 'a', 'b'), 'c') => ('ffma', 'a', 'b', 'c') */
   { .variable = {
      { nir_search_value_variable, 16 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 16 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 61, 62 },
      1,
   } },
   { .variable = {
      { nir_search_value_variable, 16 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 63, 64 },
      -1,
   } },

   /* replace2097_0 -> 61 in the cache */
   /* replace2097_1 -> 62 in the cache */
   /* replace2097_2 -> 64 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 61, 62, 64 },
      -1,
   } },

   /* ('~fadd@16', ('fneg(is_only_used_by_fadd)', ('fmul(is_only_used_by_fadd)', 'a', 'b')), 'c') => ('ffma', ('fneg', 'a'), 'b', 'c') */
   /* search2098_0_0_0 -> 61 in the cache */
   /* search2098_0_0_1 -> 62 in the cache */
   /* search2098_0_0 -> 63 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 63 },
      1,
   } },
   /* search2098_1 -> 64 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 67, 64 },
      -1,
   } },

   /* replace2098_0_0 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 61 },
      -1,
   } },
   /* replace2098_1 -> 62 in the cache */
   /* replace2098_2 -> 64 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 69, 62, 64 },
      -1,
   } },

   /* ('~fadd@16', ('fabs(is_only_used_by_fadd)', ('fmul(is_only_used_by_fadd)', 'a', 'b')), 'c') => ('ffma', ('fabs', 'a'), ('fabs', 'b'), 'c') */
   /* search2099_0_0_0 -> 61 in the cache */
   /* search2099_0_0_1 -> 62 in the cache */
   /* search2099_0_0 -> 63 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 1,
      { 63 },
      1,
   } },
   /* search2099_1 -> 64 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 71, 64 },
      -1,
   } },

   /* replace2099_0_0 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 61 },
      -1,
   } },
   /* replace2099_1_0 -> 62 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 62 },
      -1,
   } },
   /* replace2099_2 -> 64 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 73, 74, 64 },
      -1,
   } },

   /* ('~fadd@16', ('fneg(is_only_used_by_fadd)', ('fabs', ('fmul(is_only_used_by_fadd)', 'a', 'b'))), 'c') => ('ffma', ('fneg', ('fabs', 'a')), ('fabs', 'b'), 'c') */
   /* search2100_0_0_0_0 -> 61 in the cache */
   /* search2100_0_0_0_1 -> 62 in the cache */
   /* search2100_0_0_0 -> 63 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 1,
      { 63 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 76 },
      1,
   } },
   /* search2100_1 -> 64 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 77, 64 },
      -1,
   } },

   /* replace2100_0_0_0 -> 61 in the cache */
   /* replace2100_0_0 -> 73 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 73 },
      -1,
   } },
   /* replace2100_1_0 -> 62 in the cache */
   /* replace2100_1 -> 74 in the cache */
   /* replace2100_2 -> 64 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 79, 74, 64 },
      -1,
   } },

   /* ('~fadd@32', ('fmul(is_only_used_by_fadd)', 'a', 'b'), 'c') => ('ffma', 'a', 'b', 'c') */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 81, 82 },
      1,
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 83, 84 },
      -1,
   } },

   /* replace2101_0 -> 81 in the cache */
   /* replace2101_1 -> 82 in the cache */
   /* replace2101_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 81, 82, 84 },
      -1,
   } },

   /* ('~fadd@32', ('fneg(is_only_used_by_fadd)', ('fmul(is_only_used_by_fadd)', 'a', 'b')), 'c') => ('ffma', ('fneg', 'a'), 'b', 'c') */
   /* search2102_0_0_0 -> 81 in the cache */
   /* search2102_0_0_1 -> 82 in the cache */
   /* search2102_0_0 -> 83 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 83 },
      1,
   } },
   /* search2102_1 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 87, 84 },
      -1,
   } },

   /* replace2102_0_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 81 },
      -1,
   } },
   /* replace2102_1 -> 82 in the cache */
   /* replace2102_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 89, 82, 84 },
      -1,
   } },

   /* ('~fadd@32', ('fabs(is_only_used_by_fadd)', ('fmul(is_only_used_by_fadd)', 'a', 'b')), 'c') => ('ffma', ('fabs', 'a'), ('fabs', 'b'), 'c') */
   /* search2103_0_0_0 -> 81 in the cache */
   /* search2103_0_0_1 -> 82 in the cache */
   /* search2103_0_0 -> 83 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 1,
      { 83 },
      1,
   } },
   /* search2103_1 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 91, 84 },
      -1,
   } },

   /* replace2103_0_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 81 },
      -1,
   } },
   /* replace2103_1_0 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 82 },
      -1,
   } },
   /* replace2103_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 93, 94, 84 },
      -1,
   } },

   /* ('~fadd@32', ('fneg(is_only_used_by_fadd)', ('fabs', ('fmul(is_only_used_by_fadd)', 'a', 'b'))), 'c') => ('ffma', ('fneg', ('fabs', 'a')), ('fabs', 'b'), 'c') */
   /* search2104_0_0_0_0 -> 81 in the cache */
   /* search2104_0_0_0_1 -> 82 in the cache */
   /* search2104_0_0_0 -> 83 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 1,
      { 83 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 96 },
      1,
   } },
   /* search2104_1 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 97, 84 },
      -1,
   } },

   /* replace2104_0_0_0 -> 81 in the cache */
   /* replace2104_0_0 -> 93 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 93 },
      -1,
   } },
   /* replace2104_1_0 -> 82 in the cache */
   /* replace2104_1 -> 94 in the cache */
   /* replace2104_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 99, 94, 84 },
      -1,
   } },

   /* ('~fadd@32', ('fmulz(is_only_used_by_fadd)', 'a', 'b'), 'c') => ('ffmaz', 'a', 'b', 'c') */
   /* search2105_0_0 -> 81 in the cache */
   /* search2105_0_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 81, 82 },
      1,
   } },
   /* search2105_1 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 101, 84 },
      -1,
   } },

   /* replace2105_0 -> 81 in the cache */
   /* replace2105_1 -> 82 in the cache */
   /* replace2105_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 81, 82, 84 },
      -1,
   } },

   /* ('~fadd@32', ('fneg(is_only_used_by_fadd)', ('fmulz(is_only_used_by_fadd)', 'a', 'b')), 'c') => ('ffmaz', ('fneg', 'a'), 'b', 'c') */
   /* search2106_0_0_0 -> 81 in the cache */
   /* search2106_0_0_1 -> 82 in the cache */
   /* search2106_0_0 -> 101 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 101 },
      1,
   } },
   /* search2106_1 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 104, 84 },
      -1,
   } },

   /* replace2106_0_0 -> 81 in the cache */
   /* replace2106_0 -> 89 in the cache */
   /* replace2106_1 -> 82 in the cache */
   /* replace2106_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 89, 82, 84 },
      -1,
   } },

   /* ('~fadd@32', ('fabs(is_only_used_by_fadd)', ('fmulz(is_only_used_by_fadd)', 'a', 'b')), 'c') => ('ffmaz', ('fabs', 'a'), ('fabs', 'b'), 'c') */
   /* search2107_0_0_0 -> 81 in the cache */
   /* search2107_0_0_1 -> 82 in the cache */
   /* search2107_0_0 -> 101 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 1,
      { 101 },
      1,
   } },
   /* search2107_1 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 107, 84 },
      -1,
   } },

   /* replace2107_0_0 -> 81 in the cache */
   /* replace2107_0 -> 93 in the cache */
   /* replace2107_1_0 -> 82 in the cache */
   /* replace2107_1 -> 94 in the cache */
   /* replace2107_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 93, 94, 84 },
      -1,
   } },

   /* ('~fadd@32', ('fneg(is_only_used_by_fadd)', ('fabs', ('fmulz(is_only_used_by_fadd)', 'a', 'b'))), 'c') => ('ffmaz', ('fneg', ('fabs', 'a')), ('fabs', 'b'), 'c') */
   /* search2108_0_0_0_0 -> 81 in the cache */
   /* search2108_0_0_0_1 -> 82 in the cache */
   /* search2108_0_0_0 -> 101 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 1,
      { 101 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 110 },
      1,
   } },
   /* search2108_1 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 111, 84 },
      -1,
   } },

   /* replace2108_0_0_0 -> 81 in the cache */
   /* replace2108_0_0 -> 93 in the cache */
   /* replace2108_0 -> 99 in the cache */
   /* replace2108_1_0 -> 82 in the cache */
   /* replace2108_1 -> 94 in the cache */
   /* replace2108_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 99, 94, 84 },
      -1,
   } },

   /* ('~fadd@64', ('fmul(is_only_used_by_fadd)', 'a', 'b'), 'c') => ('ffma', 'a', 'b', 'c') */
   { .variable = {
      { nir_search_value_variable, 64 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 64 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 114, 115 },
      1,
   } },
   { .variable = {
      { nir_search_value_variable, 64 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 116, 117 },
      -1,
   } },

   /* replace2109_0 -> 114 in the cache */
   /* replace2109_1 -> 115 in the cache */
   /* replace2109_2 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 114, 115, 117 },
      -1,
   } },

   /* ('~fadd@64', ('fneg(is_only_used_by_fadd)', ('fmul(is_only_used_by_fadd)', 'a', 'b')), 'c') => ('ffma', ('fneg', 'a'), 'b', 'c') */
   /* search2110_0_0_0 -> 114 in the cache */
   /* search2110_0_0_1 -> 115 in the cache */
   /* search2110_0_0 -> 116 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 116 },
      1,
   } },
   /* search2110_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 120, 117 },
      -1,
   } },

   /* replace2110_0_0 -> 114 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 114 },
      -1,
   } },
   /* replace2110_1 -> 115 in the cache */
   /* replace2110_2 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 122, 115, 117 },
      -1,
   } },

   /* ('~fadd@64', ('fabs(is_only_used_by_fadd)', ('fmul(is_only_used_by_fadd)', 'a', 'b')), 'c') => ('ffma', ('fabs', 'a'), ('fabs', 'b'), 'c') */
   /* search2111_0_0_0 -> 114 in the cache */
   /* search2111_0_0_1 -> 115 in the cache */
   /* search2111_0_0 -> 116 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 1,
      { 116 },
      1,
   } },
   /* search2111_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 124, 117 },
      -1,
   } },

   /* replace2111_0_0 -> 114 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 114 },
      -1,
   } },
   /* replace2111_1_0 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 115 },
      -1,
   } },
   /* replace2111_2 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 126, 127, 117 },
      -1,
   } },

   /* ('~fadd@64', ('fneg(is_only_used_by_fadd)', ('fabs', ('fmul(is_only_used_by_fadd)', 'a', 'b'))), 'c') => ('ffma', ('fneg', ('fabs', 'a')), ('fabs', 'b'), 'c') */
   /* search2112_0_0_0_0 -> 114 in the cache */
   /* search2112_0_0_0_1 -> 115 in the cache */
   /* search2112_0_0_0 -> 116 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 1,
      { 116 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 129 },
      1,
   } },
   /* search2112_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 130, 117 },
      -1,
   } },

   /* replace2112_0_0_0 -> 114 in the cache */
   /* replace2112_0_0 -> 126 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 126 },
      -1,
   } },
   /* replace2112_1_0 -> 115 in the cache */
   /* replace2112_1 -> 127 in the cache */
   /* replace2112_2 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 132, 127, 117 },
      -1,
   } },

   /* ('fadd@8', 'a', ('fneg', 'b')) => ('fsub', 'a', 'b') */
   { .variable = {
      { nir_search_value_variable, 8 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 8 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 134, 136 },
      -1,
   } },

   /* replace2113_0 -> 134 in the cache */
   /* replace2113_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 134, 135 },
      -1,
   } },

   /* ('fadd@16', 'a', ('fneg', 'b')) => ('fsub', 'a', 'b') */
   /* search2114_0 -> 61 in the cache */
   /* search2114_1_0 -> 62 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 62 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 61, 139 },
      -1,
   } },

   /* replace2114_0 -> 61 in the cache */
   /* replace2114_1 -> 62 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 61, 62 },
      -1,
   } },

   /* ('fadd@32', 'a', ('fneg', 'b')) => ('fsub', 'a', 'b') */
   /* search2115_0 -> 81 in the cache */
   /* search2115_1_0 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 82 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 81, 142 },
      -1,
   } },

   /* replace2115_0 -> 81 in the cache */
   /* replace2115_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 81, 82 },
      -1,
   } },

   /* ('fadd@64', 'a', ('fneg', 'b')) => ('fsub', 'a', 'b') */
   /* search2116_0 -> 114 in the cache */
   /* search2116_1_0 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 115 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 114, 145 },
      -1,
   } },

   /* replace2116_0 -> 114 in the cache */
   /* replace2116_1 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 114, 115 },
      -1,
   } },

   /* ('fneg', 'a') => ('fmul', 'a', -1.0) */
   /* search2117_0 -> 42 in the cache */
   /* search2117 -> 59 in the cache */

   /* replace2117_0 -> 42 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0xbff0000000000000 /* -1.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 42, 148 },
      -1,
   } },

   /* ('iadd', 'a', ('ineg', 'b')) => ('isub', 'a', 'b') */
   /* search2118_0 -> 0 in the cache */
   /* search2118_1_0 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 1 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 0, 150 },
      -1,
   } },

   /* replace2118_0 -> 0 in the cache */
   /* replace2118_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* ('ineg', 'a') => ('isub', 0, 'a') */
   /* search2119_0 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 42 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   /* replace2119_1 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 154, 42 },
      -1,
   } },

   /* ('iabs', 'a') => ('imax', 'a', ('ineg', 'a')) */
   /* search2120_0 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 42 },
      -1,
   } },

   /* replace2120_0 -> 42 in the cache */
   /* replace2120_1_0 -> 42 in the cache */
   /* replace2120_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 42, 153 },
      -1,
   } },

   /* ('iadd', ('iadd(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), 'c(is_not_const)') => ('iadd3', 'a', 'b', 'c') */
   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* a */
      false,
      nir_type_invalid,
      1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* b */
      false,
      nir_type_invalid,
      1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 158, 159 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* c */
      false,
      nir_type_invalid,
      1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 160, 161 },
      -1,
   } },

   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd3,
      0, 1,
      { 163, 164, 165 },
      -1,
   } },

   /* ('iadd', ('isub(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), 'c(is_not_const)') => ('iadd3', 'a', ('ineg', 'b'), 'c') */
   /* search2122_0_0 -> 158 in the cache */
   /* search2122_0_1 -> 159 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 158, 159 },
      0,
   } },
   /* search2122_1 -> 161 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 167, 161 },
      -1,
   } },

   /* replace2122_0 -> 163 in the cache */
   /* replace2122_1_0 -> 164 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 164 },
      -1,
   } },
   /* replace2122_2 -> 165 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd3,
      0, 1,
      { 163, 169, 165 },
      -1,
   } },

   /* ('isub', ('isub(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), 'c(is_not_const)') => ('iadd3', 'a', ('ineg', 'b'), ('ineg', 'c')) */
   /* search2123_0_0 -> 158 in the cache */
   /* search2123_0_1 -> 159 in the cache */
   /* search2123_0 -> 167 in the cache */
   /* search2123_1 -> 161 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 167, 161 },
      -1,
   } },

   /* replace2123_0 -> 163 in the cache */
   /* replace2123_1_0 -> 164 in the cache */
   /* replace2123_1 -> 169 in the cache */
   /* replace2123_2_0 -> 165 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 165 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd3,
      0, 1,
      { 163, 169, 172 },
      -1,
   } },

   /* ('vec2(is_only_used_as_float)', ('fneg@16', 'a'), 'b') => ('fmul', ('vec2', 'a', 'b'), ('vec2', -1.0, 1.0)) */
   /* search2124_0_0 -> 61 in the cache */
   /* search2124_0 -> 69 in the cache */
   /* search2124_1 -> 62 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 69, 62 },
      2,
   } },

   /* replace2124_0_0 -> 61 in the cache */
   /* replace2124_0_1 -> 62 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 61, 62 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_float, { 0xbff0000000000000 /* -1.0 */ },
   } },
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 176, 177 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 175, 178 },
      -1,
   } },

   /* ('vec2(is_only_used_as_float)', 'a', ('fneg@16', 'b')) => ('fmul', ('vec2', 'a', 'b'), ('vec2', 1.0, -1.0)) */
   /* search2125_0 -> 61 in the cache */
   /* search2125_1_0 -> 62 in the cache */
   /* search2125_1 -> 139 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 61, 139 },
      2,
   } },

   /* replace2125_0_0 -> 61 in the cache */
   /* replace2125_0_1 -> 62 in the cache */
   /* replace2125_0 -> 175 in the cache */
   /* replace2125_1_0 -> 177 in the cache */
   /* replace2125_1_1 -> 176 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 177, 176 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 175, 181 },
      -1,
   } },

   /* ('flt', '#b(is_gt_0_and_lt_1)', ('fsat(is_used_once)', 'a')) => ('flt', 'b', 'a') */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* b */
      true,
      nir_type_invalid,
      2,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 184 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 183, 185 },
      -1,
   } },

   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* replace2126_1 -> 184 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 187, 184 },
      -1,
   } },

   /* ('fge', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('fge', 'a', 'b') */
   /* search2127_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 0 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      true,
      nir_type_invalid,
      2,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 189, 190 },
      -1,
   } },

   /* replace2127_0 -> 0 in the cache */
   /* replace2127_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* ('feq', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('feq', 'a', 'b') */
   /* search2128_0_0 -> 0 in the cache */
   /* search2128_0 -> 189 in the cache */
   /* search2128_1 -> 190 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 189, 190 },
      -1,
   } },

   /* replace2128_0 -> 0 in the cache */
   /* replace2128_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 0, 1 },
      -1,
   } },

   /* ('fneu', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('fneu', 'a', 'b') */
   /* search2129_0_0 -> 0 in the cache */
   /* search2129_0 -> 189 in the cache */
   /* search2129_1 -> 190 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 189, 190 },
      -1,
   } },

   /* replace2129_0 -> 0 in the cache */
   /* replace2129_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 0, 1 },
      -1,
   } },

   /* ('fge', ('fsat(is_used_once)', 'a'), 1.0) => ('fge', 'a', 1.0) */
   /* search2130_0_0 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 42 },
      0,
   } },
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 197, 198 },
      -1,
   } },

   /* replace2130_0 -> 42 in the cache */
   /* replace2130_1 -> 198 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 42, 198 },
      -1,
   } },

   /* ('~fge', ('fmin(is_used_once)', ('fadd(is_used_once)', 'a', 'b'), ('fadd', 'c', 'd')), 0.0) => ('iand', ('fge', 'a', ('fneg', 'b')), ('fge', 'c', ('fneg', 'd'))) */
   { .variable = {
      { nir_search_value_variable, -4 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 201, 202 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 204, 205 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 3,
      { 203, 206 },
      0,
   } },
   { .constant = {
      { nir_search_value_constant, -4 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 3,
      { 207, 208 },
      -1,
   } },

   /* replace2131_0_0 -> 201 in the cache */
   /* replace2131_0_1_0 -> 202 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 202 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 201, 210 },
      -1,
   } },
   /* replace2131_1_0 -> 204 in the cache */
   /* replace2131_1_1_0 -> 205 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 205 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 204, 212 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 211, 213 },
      -1,
   } },

   /* ('flt', ('fneg', 'a'), ('fneg', 'b')) => ('flt', 'b', 'a') */
   /* search2132_0_0 -> 0 in the cache */
   /* search2132_0 -> 9 in the cache */
   /* search2132_1_0 -> 1 in the cache */
   /* search2132_1 -> 5 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 9, 5 },
      -1,
   } },

   /* replace2132_0 -> 1 in the cache */
   /* replace2132_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1, 0 },
      -1,
   } },

   /* ('fge', ('fneg', 'a'), ('fneg', 'b')) => ('fge', 'b', 'a') */
   /* search2133_0_0 -> 0 in the cache */
   /* search2133_0 -> 9 in the cache */
   /* search2133_1_0 -> 1 in the cache */
   /* search2133_1 -> 5 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 9, 5 },
      -1,
   } },

   /* replace2133_0 -> 1 in the cache */
   /* replace2133_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1, 0 },
      -1,
   } },

   /* ('feq', ('fneg', 'a'), ('fneg', 'b')) => ('feq', 'b', 'a') */
   /* search2134_0_0 -> 0 in the cache */
   /* search2134_0 -> 9 in the cache */
   /* search2134_1_0 -> 1 in the cache */
   /* search2134_1 -> 5 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 9, 5 },
      -1,
   } },

   /* replace2134_0 -> 1 in the cache */
   /* replace2134_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1, 0 },
      -1,
   } },

   /* ('fneu', ('fneg', 'a'), ('fneg', 'b')) => ('fneu', 'b', 'a') */
   /* search2135_0_0 -> 0 in the cache */
   /* search2135_0 -> 9 in the cache */
   /* search2135_1_0 -> 1 in the cache */
   /* search2135_1 -> 5 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 9, 5 },
      -1,
   } },

   /* replace2135_0 -> 1 in the cache */
   /* replace2135_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1, 0 },
      -1,
   } },

   /* ('flt', ('fneg', 'a'), -1.0) => ('flt', 1.0, 'a') */
   /* search2136_0_0 -> 42 in the cache */
   /* search2136_0 -> 59 in the cache */
   /* search2136_1 -> 148 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 59, 148 },
      -1,
   } },

   /* replace2136_0 -> 198 in the cache */
   /* replace2136_1 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 198, 42 },
      -1,
   } },

   /* ('flt', -1.0, ('fneg', 'a')) => ('flt', 'a', 1.0) */
   /* search2137_0 -> 148 in the cache */
   /* search2137_1_0 -> 42 in the cache */
   /* search2137_1 -> 59 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 148, 59 },
      -1,
   } },

   /* replace2137_0 -> 42 in the cache */
   /* replace2137_1 -> 198 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 42, 198 },
      -1,
   } },

   /* ('fge', ('fneg', 'a'), -1.0) => ('fge', 1.0, 'a') */
   /* search2138_0_0 -> 42 in the cache */
   /* search2138_0 -> 59 in the cache */
   /* search2138_1 -> 148 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 59, 148 },
      -1,
   } },

   /* replace2138_0 -> 198 in the cache */
   /* replace2138_1 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 198, 42 },
      -1,
   } },

   /* ('fge', -1.0, ('fneg', 'a')) => ('fge', 'a', 1.0) */
   /* search2139_0 -> 148 in the cache */
   /* search2139_1_0 -> 42 in the cache */
   /* search2139_1 -> 59 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 148, 59 },
      -1,
   } },

   /* replace2139_0 -> 42 in the cache */
   /* replace2139_1 -> 198 in the cache */
   /* replace2139 -> 200 in the cache */

   /* ('fneu', ('fneg', 'a'), -1.0) => ('fneu', 1.0, 'a') */
   /* search2140_0_0 -> 42 in the cache */
   /* search2140_0 -> 59 in the cache */
   /* search2140_1 -> 148 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 59, 148 },
      -1,
   } },

   /* replace2140_0 -> 198 in the cache */
   /* replace2140_1 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 198, 42 },
      -1,
   } },

   /* ('feq', -1.0, ('fneg', 'a')) => ('feq', 'a', 1.0) */
   /* search2141_0 -> 148 in the cache */
   /* search2141_1_0 -> 42 in the cache */
   /* search2141_1 -> 59 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 148, 59 },
      -1,
   } },

   /* replace2141_0 -> 42 in the cache */
   /* replace2141_1 -> 198 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 42, 198 },
      -1,
   } },

   /* ('ior', 'a', 'a') => a */
   /* search2142_0 -> 42 in the cache */
   /* search2142_1 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 0,
      { 42, 42 },
      -1,
   } },

   /* replace2142 -> 42 in the cache */

   /* ('iand', 'a', 'a') => a */
   /* search2143_0 -> 42 in the cache */
   /* search2143_1 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 0,
      { 42, 42 },
      -1,
   } },

   /* replace2143 -> 42 in the cache */

   /* ('~fadd', ('fneg(is_used_once)', ('fsat(is_used_once)', 'a(is_not_fmul)')), 1.0) => ('fsat', ('fadd', 1.0, ('fneg', 'a'))) */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      3,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 236 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 237 },
      0,
   } },
   /* search2144_1 -> 198 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 238, 198 },
      -1,
   } },

   /* replace2144_0_0 -> 198 in the cache */
   /* replace2144_0_1_0 -> 42 in the cache */
   /* replace2144_0_1 -> 59 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 198, 59 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 240 },
      -1,
   } },

   /* ('fdot2', 'a', 'b') => ('fdot2_replicated', 'a', 'b') */
   /* search2145_0 -> 0 in the cache */
   /* search2145_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot2,
      0, 1,
      { 0, 1 },
      -1,
   } },

   /* replace2145_0 -> 0 in the cache */
   /* replace2145_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot2_replicated,
      0, 1,
      { 0, 1 },
      -1,
   } },

   /* ('fdot3', 'a', 'b') => ('fdot3_replicated', 'a', 'b') */
   /* search2146_0 -> 0 in the cache */
   /* search2146_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot3,
      0, 1,
      { 0, 1 },
      -1,
   } },

   /* replace2146_0 -> 0 in the cache */
   /* replace2146_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot3_replicated,
      0, 1,
      { 0, 1 },
      -1,
   } },

   /* ('fdot4', 'a', 'b') => ('fdot4_replicated', 'a', 'b') */
   /* search2147_0 -> 0 in the cache */
   /* search2147_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot4,
      0, 1,
      { 0, 1 },
      -1,
   } },

   /* replace2147_0 -> 0 in the cache */
   /* replace2147_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot4_replicated,
      0, 1,
      { 0, 1 },
      -1,
   } },

   /* ('fdph', 'a', 'b') => ('fdph_replicated', 'a', 'b') */
   /* search2148_0 -> 0 in the cache */
   /* search2148_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdph,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* replace2148_0 -> 0 in the cache */
   /* replace2148_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdph_replicated,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* ('~flrp', ('fadd(is_used_once)', 'a', 'b'), ('fadd(is_used_once)', 'a', 'c'), 'd') => ('fadd', ('flrp', 'b', 'c', 'd'), 'a') */
   /* search2149_0_0 -> 201 in the cache */
   /* search2149_0_1 -> 202 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 201, 202 },
      0,
   } },
   /* search2149_1_0 -> 201 in the cache */
   /* search2149_1_1 -> 204 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 201, 204 },
      0,
   } },
   /* search2149_2 -> 205 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 2,
      { 250, 251, 205 },
      -1,
   } },

   /* replace2149_0_0 -> 202 in the cache */
   /* replace2149_0_1 -> 204 in the cache */
   /* replace2149_0_2 -> 205 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 202, 204, 205 },
      -1,
   } },
   /* replace2149_1 -> 201 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 253, 201 },
      -1,
   } },

   /* ('fround_even', 'a') => ('bcsel', ('feq', ('ffract', 'a'), 0.5), ('fadd', ('ffloor', ('fadd', 'a', 0.5)), 1.0), ('ffloor', ('fadd', 'a', 0.5))) */
   /* search2150_0 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fround_even,
      -1, 0,
      { 42 },
      -1,
   } },

   /* replace2150_0_0_0 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 0,
      { 42 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x3fe0000000000000 /* 0.5 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 256, 257 },
      -1,
   } },
   /* replace2150_1_0_0_0 -> 42 in the cache */
   /* replace2150_1_0_0_1 -> 257 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 42, 257 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 1,
      { 259 },
      -1,
   } },
   /* replace2150_1_1 -> 198 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 2,
      { 260, 198 },
      -1,
   } },
   /* replace2150_2_0_0 -> 42 in the cache */
   /* replace2150_2_0_1 -> 257 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      3, 1,
      { 42, 257 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 1,
      { 262 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 258, 261, 263 },
      -1,
   } },

   /* ('~ffma@32', 'a', 2.0, -1.0) => ('flrp', -1.0, 1.0, 'a') */
   /* search2151_0 -> 81 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0x4000000000000000 /* 2.0 */ },
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0xbff0000000000000 /* -1.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 81, 265, 266 },
      -1,
   } },

   /* replace2151_0 -> 266 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   /* replace2151_2 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 266, 268, 81 },
      -1,
   } },

   /* ('~ffma@32', 'a', -2.0, -1.0) => ('flrp', -1.0, 1.0, ('fneg', 'a')) */
   /* search2152_0 -> 81 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0xc000000000000000 /* -2.0 */ },
   } },
   /* search2152_2 -> 266 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 81, 270, 266 },
      -1,
   } },

   /* replace2152_0 -> 266 in the cache */
   /* replace2152_1 -> 268 in the cache */
   /* replace2152_2_0 -> 81 in the cache */
   /* replace2152_2 -> 89 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 266, 268, 89 },
      -1,
   } },

   /* ('~ffma@32', 'a', -2.0, 1.0) => ('flrp', 1.0, -1.0, 'a') */
   /* search2153_0 -> 81 in the cache */
   /* search2153_1 -> 270 in the cache */
   /* search2153_2 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 81, 270, 268 },
      -1,
   } },

   /* replace2153_0 -> 268 in the cache */
   /* replace2153_1 -> 266 in the cache */
   /* replace2153_2 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 268, 266, 81 },
      -1,
   } },

   /* ('~ffma@32', 'a', 2.0, 1.0) => ('flrp', 1.0, -1.0, ('fneg', 'a')) */
   /* search2154_0 -> 81 in the cache */
   /* search2154_1 -> 265 in the cache */
   /* search2154_2 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 81, 265, 268 },
      -1,
   } },

   /* replace2154_0 -> 268 in the cache */
   /* replace2154_1 -> 266 in the cache */
   /* replace2154_2_0 -> 81 in the cache */
   /* replace2154_2 -> 89 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 268, 266, 89 },
      -1,
   } },

   /* ('~fadd@32', ('fmul(is_used_once)', 2.0, 'a'), -1.0) => ('flrp', -1.0, 1.0, 'a') */
   /* search2155_0_0 -> 265 in the cache */
   /* search2155_0_1 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 265, 81 },
      0,
   } },
   /* search2155_1 -> 266 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 277, 266 },
      -1,
   } },

   /* replace2155_0 -> 266 in the cache */
   /* replace2155_1 -> 268 in the cache */
   /* replace2155_2 -> 81 in the cache */
   /* replace2155 -> 269 in the cache */

   /* ('~fadd@32', ('fmul(is_used_once)', -2.0, 'a'), -1.0) => ('flrp', -1.0, 1.0, ('fneg', 'a')) */
   /* search2156_0_0 -> 270 in the cache */
   /* search2156_0_1 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 270, 81 },
      0,
   } },
   /* search2156_1 -> 266 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 279, 266 },
      -1,
   } },

   /* replace2156_0 -> 266 in the cache */
   /* replace2156_1 -> 268 in the cache */
   /* replace2156_2_0 -> 81 in the cache */
   /* replace2156_2 -> 89 in the cache */
   /* replace2156 -> 272 in the cache */

   /* ('~fadd@32', ('fmul(is_used_once)', -2.0, 'a'), 1.0) => ('flrp', 1.0, -1.0, 'a') */
   /* search2157_0_0 -> 270 in the cache */
   /* search2157_0_1 -> 81 in the cache */
   /* search2157_0 -> 279 in the cache */
   /* search2157_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 279, 268 },
      -1,
   } },

   /* replace2157_0 -> 268 in the cache */
   /* replace2157_1 -> 266 in the cache */
   /* replace2157_2 -> 81 in the cache */
   /* replace2157 -> 274 in the cache */

   /* ('~fadd@32', ('fmul(is_used_once)', 2.0, 'a'), 1.0) => ('flrp', 1.0, -1.0, ('fneg', 'a')) */
   /* search2158_0_0 -> 265 in the cache */
   /* search2158_0_1 -> 81 in the cache */
   /* search2158_0 -> 277 in the cache */
   /* search2158_1 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 277, 268 },
      -1,
   } },

   /* replace2158_0 -> 268 in the cache */
   /* replace2158_1 -> 266 in the cache */
   /* replace2158_2_0 -> 81 in the cache */
   /* replace2158_2 -> 89 in the cache */
   /* replace2158 -> 276 in the cache */

   /* ('~ffma@32', ('fadd', 'b', ('fneg', 'a')), 'a', 'a') => ('flrp', 'a', 'b', 'a') */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 284 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 283, 285 },
      -1,
   } },
   /* search2159_1 -> 284 in the cache */
   /* search2159_2 -> 284 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 286, 284, 284 },
      -1,
   } },

   /* replace2159_0 -> 284 in the cache */
   /* replace2159_1 -> 283 in the cache */
   /* replace2159_2 -> 284 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 284, 283, 284 },
      -1,
   } },

   /* ('~ffma@32', 'a', 2.0, ('fneg', ('fmul', 'a', 'a'))) => ('flrp', 'a', 1.0, 'a') */
   /* search2160_0 -> 81 in the cache */
   /* search2160_1 -> 265 in the cache */
   /* search2160_2_0_0 -> 81 in the cache */
   /* search2160_2_0_1 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      -1, 0,
      { 81, 81 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 289 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 81, 265, 290 },
      -1,
   } },

   /* replace2160_0 -> 81 in the cache */
   /* replace2160_1 -> 268 in the cache */
   /* replace2160_2 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 81, 268, 81 },
      -1,
   } },

   /* ('~ffma@32', 'a', 2.0, ('fmul', ('fneg', 'a'), 'a')) => ('flrp', 'a', 1.0, 'a') */
   /* search2161_0 -> 81 in the cache */
   /* search2161_1 -> 265 in the cache */
   /* search2161_2_0_0 -> 81 in the cache */
   /* search2161_2_0 -> 89 in the cache */
   /* search2161_2_1 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 89, 81 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 81, 265, 293 },
      -1,
   } },

   /* replace2161_0 -> 81 in the cache */
   /* replace2161_1 -> 268 in the cache */
   /* replace2161_2 -> 81 in the cache */
   /* replace2161 -> 292 in the cache */

   /* ('~ffma@32', 'a', ('fneg', 'a'), ('fmul', 2.0, 'a')) => ('flrp', 'a', 1.0, 'a') */
   /* search2162_0 -> 81 in the cache */
   /* search2162_1_0 -> 81 in the cache */
   /* search2162_1 -> 89 in the cache */
   /* search2162_2_0 -> 265 in the cache */
   /* search2162_2_1 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 265, 81 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 81, 89, 295 },
      -1,
   } },

   /* replace2162_0 -> 81 in the cache */
   /* replace2162_1 -> 268 in the cache */
   /* replace2162_2 -> 81 in the cache */
   /* replace2162 -> 292 in the cache */

   /* ('~fmul@32', 'a', ('fadd', 2.0, ('fneg', 'a'))) => ('flrp', 'a', 1.0, 'a') */
   /* search2163_0 -> 81 in the cache */
   /* search2163_1_0 -> 265 in the cache */
   /* search2163_1_1_0 -> 81 in the cache */
   /* search2163_1_1 -> 89 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 265, 89 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 81, 297 },
      -1,
   } },

   /* replace2163_0 -> 81 in the cache */
   /* replace2163_1 -> 268 in the cache */
   /* replace2163_2 -> 81 in the cache */
   /* replace2163 -> 292 in the cache */

   /* ('fmin', ('fadd(is_used_once)', '#c', 'a'), ('fadd(is_used_once)', '#c', 'b')) => ('fadd', 'c', ('fmin', 'a', 'b')) */
   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* c */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 299, 300 },
      0,
   } },
   /* search2164_1_0 -> 299 in the cache */
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 299, 302 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 3,
      { 301, 303 },
      -1,
   } },

   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* replace2164_1_0 -> 300 in the cache */
   /* replace2164_1_1 -> 302 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 300, 302 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 305, 306 },
      -1,
   } },

   /* ('fmax', ('fadd(is_used_once)', '#c', 'a'), ('fadd(is_used_once)', '#c', 'b')) => ('fadd', 'c', ('fmax', 'a', 'b')) */
   /* search2165_0_0 -> 299 in the cache */
   /* search2165_0_1 -> 300 in the cache */
   /* search2165_0 -> 301 in the cache */
   /* search2165_1_0 -> 299 in the cache */
   /* search2165_1_1 -> 302 in the cache */
   /* search2165_1 -> 303 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 3,
      { 301, 303 },
      -1,
   } },

   /* replace2165_0 -> 305 in the cache */
   /* replace2165_1_0 -> 300 in the cache */
   /* replace2165_1_1 -> 302 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 300, 302 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 305, 309 },
      -1,
   } },

   /* ('bcsel', ('feq', ('fsqrt', 'a(is_not_negative)'), 0.0), 3.4028234663852886e+38, ('frsq', 'a')) => ('fmin', ('frsq', 'a'), 3.4028234663852886e+38) */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      4,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsqrt,
      -1, 0,
      { 311 },
      -1,
   } },
   /* search2166_0_1 -> 43 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 312, 43 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x47efffffe0000000 /* 3.4028234663852886e+38 */ },
   } },
   /* search2166_2_0 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_frsq,
      -1, 0,
      { 42 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 313, 314, 315 },
      -1,
   } },

   /* replace2166_0_0 -> 42 in the cache */
   /* replace2166_0 -> 315 in the cache */
   /* replace2166_1 -> 314 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 315, 314 },
      -1,
   } },

   /* ('~fadd', ('ffma(is_used_once)', 'a', 'b', ('ffma', 'c', 'd', ('fmul(is_used_once)', 'e(is_not_const_and_not_fsign)', 'f(is_not_const_and_not_fsign)'))), 'g(is_not_const)') => ('ffma', 'a', 'b', ('ffma', 'c', 'd', ('ffma', 'e', 'f', 'g'))) */
   { .variable = {
      { nir_search_value_variable, -7 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -7 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -7 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -7 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -7 },
      4, /* e */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -7 },
      5, /* f */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 322, 323 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_ffma,
      2, 2,
      { 320, 321, 324 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 3,
      { 318, 319, 325 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -7 },
      6, /* g */
      false,
      nir_type_invalid,
      1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -7 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 326, 327 },
      -1,
   } },

   /* replace2167_0 -> 318 in the cache */
   /* replace2167_1 -> 319 in the cache */
   /* replace2167_2_0 -> 320 in the cache */
   /* replace2167_2_1 -> 321 in the cache */
   { .variable = {
      { nir_search_value_variable, -7 },
      4, /* e */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -7 },
      5, /* f */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -7 },
      6, /* g */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_ffma,
      2, 1,
      { 329, 330, 331 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 2,
      { 320, 321, 332 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 3,
      { 318, 319, 333 },
      -1,
   } },

   /* ('~fadd', ('ffma(is_used_once)', 'a', 'b', ('fmul(is_used_once)', 'c(is_not_const_and_not_fsign)', 'd(is_not_const_and_not_fsign)')), 'e(is_not_const)') => ('ffma', 'a', 'b', ('ffma', 'c', 'd', 'e')) */
   { .variable = {
      { nir_search_value_variable, -5 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      2, /* c */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      3, /* d */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 337, 338 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 2,
      { 335, 336, 339 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      4, /* e */
      false,
      nir_type_invalid,
      1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 340, 341 },
      -1,
   } },

   /* replace2168_0 -> 335 in the cache */
   /* replace2168_1 -> 336 in the cache */
   { .variable = {
      { nir_search_value_variable, -5 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      4, /* e */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 343, 344, 345 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 335, 336, 346 },
      -1,
   } },

   /* ('~fadd', ('fneg', ('ffma(is_used_once)', 'a', 'b', ('ffma', 'c', 'd', ('fmul(is_used_once)', 'e(is_not_const_and_not_fsign)', 'f(is_not_const_and_not_fsign)')))), 'g(is_not_const)') => ('ffma', ('fneg', 'a'), 'b', ('ffma', ('fneg', 'c'), 'd', ('ffma', ('fneg', 'e'), 'f', 'g'))) */
   /* search2169_0_0_0 -> 318 in the cache */
   /* search2169_0_0_1 -> 319 in the cache */
   /* search2169_0_0_2_0 -> 320 in the cache */
   /* search2169_0_0_2_1 -> 321 in the cache */
   /* search2169_0_0_2_2_0 -> 322 in the cache */
   /* search2169_0_0_2_2_1 -> 323 in the cache */
   /* search2169_0_0_2_2 -> 324 in the cache */
   /* search2169_0_0_2 -> 325 in the cache */
   /* search2169_0_0 -> 326 in the cache */
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 3,
      { 326 },
      -1,
   } },
   /* search2169_1 -> 327 in the cache */
   { .expression = {
      { nir_search_value_expression, -7 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 348, 327 },
      -1,
   } },

   /* replace2169_0_0 -> 318 in the cache */
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 318 },
      -1,
   } },
   /* replace2169_1 -> 319 in the cache */
   /* replace2169_2_0_0 -> 320 in the cache */
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 320 },
      -1,
   } },
   /* replace2169_2_1 -> 321 in the cache */
   /* replace2169_2_2_0_0 -> 329 in the cache */
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 329 },
      -1,
   } },
   /* replace2169_2_2_1 -> 330 in the cache */
   /* replace2169_2_2_2 -> 331 in the cache */
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_ffma,
      2, 1,
      { 352, 330, 331 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 2,
      { 351, 321, 353 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 3,
      { 350, 319, 354 },
      -1,
   } },

   /* ('~fadd', ('ffmaz(is_used_once)', 'a', 'b', ('ffmaz', 'c', 'd', ('fmulz(is_used_once)', 'e(is_not_const_and_not_fsign)', 'f(is_not_const_and_not_fsign)'))), 'g(is_not_const)') => ('ffmaz', 'a', 'b', ('ffmaz', 'c', 'd', ('ffmaz', 'e', 'f', 'g'))) */
   /* search2170_0_0 -> 81 in the cache */
   /* search2170_0_1 -> 82 in the cache */
   /* search2170_0_2_0 -> 84 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      4, /* e */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      5, /* f */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      3, 1,
      { 357, 358 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      2, 2,
      { 84, 356, 359 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 3,
      { 81, 82, 360 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      6, /* g */
      false,
      nir_type_invalid,
      1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 361, 362 },
      -1,
   } },

   /* replace2170_0 -> 81 in the cache */
   /* replace2170_1 -> 82 in the cache */
   /* replace2170_2_0 -> 84 in the cache */
   /* replace2170_2_1 -> 356 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      4, /* e */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      5, /* f */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      6, /* g */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      2, 1,
      { 364, 365, 366 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 2,
      { 84, 356, 367 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 3,
      { 81, 82, 368 },
      -1,
   } },

   /* ('~fadd', ('ffmaz(is_used_once)', 'a', 'b', ('fmulz(is_used_once)', 'c(is_not_const_and_not_fsign)', 'd(is_not_const_and_not_fsign)')), 'e(is_not_const)') => ('ffmaz', 'a', 'b', ('ffmaz', 'c', 'd', 'e')) */
   /* search2171_0_0 -> 81 in the cache */
   /* search2171_0_1 -> 82 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* c */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      3, /* d */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      2, 1,
      { 370, 371 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 2,
      { 81, 82, 372 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      4, /* e */
      false,
      nir_type_invalid,
      1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 373, 374 },
      -1,
   } },

   /* replace2171_0 -> 81 in the cache */
   /* replace2171_1 -> 82 in the cache */
   /* replace2171_2_0 -> 84 in the cache */
   /* replace2171_2_1 -> 356 in the cache */
   /* replace2171_2_2 -> 364 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 1,
      { 84, 356, 364 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 2,
      { 81, 82, 376 },
      -1,
   } },

   /* ('~fadd', ('fneg', ('ffmaz(is_used_once)', 'a', 'b', ('ffmaz', 'c', 'd', ('fmulz(is_used_once)', 'e(is_not_const_and_not_fsign)', 'f(is_not_const_and_not_fsign)')))), 'g(is_not_const)') => ('ffmaz', ('fneg', 'a'), 'b', ('ffmaz', ('fneg', 'c'), 'd', ('ffmaz', ('fneg', 'e'), 'f', 'g'))) */
   /* search2172_0_0_0 -> 81 in the cache */
   /* search2172_0_0_1 -> 82 in the cache */
   /* search2172_0_0_2_0 -> 84 in the cache */
   /* search2172_0_0_2_1 -> 356 in the cache */
   /* search2172_0_0_2_2_0 -> 357 in the cache */
   /* search2172_0_0_2_2_1 -> 358 in the cache */
   /* search2172_0_0_2_2 -> 359 in the cache */
   /* search2172_0_0_2 -> 360 in the cache */
   /* search2172_0_0 -> 361 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 3,
      { 361 },
      -1,
   } },
   /* search2172_1 -> 362 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 378, 362 },
      -1,
   } },

   /* replace2172_0_0 -> 81 in the cache */
   /* replace2172_0 -> 89 in the cache */
   /* replace2172_1 -> 82 in the cache */
   /* replace2172_2_0_0 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 84 },
      -1,
   } },
   /* replace2172_2_1 -> 356 in the cache */
   /* replace2172_2_2_0_0 -> 364 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 364 },
      -1,
   } },
   /* replace2172_2_2_1 -> 365 in the cache */
   /* replace2172_2_2_2 -> 366 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      2, 1,
      { 381, 365, 366 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 2,
      { 380, 356, 382 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 3,
      { 89, 82, 383 },
      -1,
   } },

   /* ('ubfe', 'a', 'b', 0) => 0 */
   /* search2173_0 -> 81 in the cache */
   /* search2173_1 -> 82 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 81, 82, 385 },
      -1,
   } },

   /* replace2173 -> 385 in the cache */

   /* ('ibfe', 'a', 'b', 0) => 0 */
   /* search2174_0 -> 81 in the cache */
   /* search2174_1 -> 82 in the cache */
   /* search2174_2 -> 385 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 81, 82, 385 },
      -1,
   } },

   /* replace2174 -> 385 in the cache */

   /* ('ubfe', 'a', '#b', '#c') => ('iand', ('ushr', 4294967295, ('ineg', 'c')), ('ushr', 'a', 'b')) */
   /* search2175_0 -> 81 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* c */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 81, 388, 389 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xffffffff /* 4294967295 */ },
   } },
   /* replace2175_0_1_0 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 84 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 391, 392 },
      -1,
   } },
   /* replace2175_1_0 -> 81 in the cache */
   /* replace2175_1_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 81, 82 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 393, 394 },
      -1,
   } },

   /* ('ibfe', 'a', '#b', '#c') => ('ishr', ('ishl', 'a', ('ineg', ('iadd', 'b', 'c'))), ('ineg', 'c')) */
   /* search2176_0 -> 81 in the cache */
   /* search2176_1 -> 388 in the cache */
   /* search2176_2 -> 389 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 81, 388, 389 },
      -1,
   } },

   /* replace2176_0_0 -> 81 in the cache */
   /* replace2176_0_1_0_0 -> 82 in the cache */
   /* replace2176_0_1_0_1 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 82, 84 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 1,
      { 397 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 81, 398 },
      -1,
   } },
   /* replace2176_1_0 -> 84 in the cache */
   /* replace2176_1 -> 392 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 399, 392 },
      -1,
   } },

   /* ('ishl', 'a', 0) => a */
   /* search2177_0 -> 42 in the cache */
   /* search2177_1 -> 385 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 42, 385 },
      -1,
   } },

   /* replace2177 -> 42 in the cache */

   /* ('ishl', 'a', -32) => a */
   /* search2178_0 -> 42 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x20 /* -32 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 42, 402 },
      -1,
   } },

   /* replace2178 -> 42 in the cache */

   /* ('ishr', 'a', 0) => a */
   /* search2179_0 -> 42 in the cache */
   /* search2179_1 -> 385 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 42, 385 },
      -1,
   } },

   /* replace2179 -> 42 in the cache */

   /* ('ishr', 'a', -32) => a */
   /* search2180_0 -> 42 in the cache */
   /* search2180_1 -> 402 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 42, 402 },
      -1,
   } },

   /* replace2180 -> 42 in the cache */

   /* ('ushr', 'a', 0) => a */
   /* search2181_0 -> 42 in the cache */
   /* search2181_1 -> 385 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 42, 385 },
      -1,
   } },

   /* replace2181 -> 42 in the cache */

   /* ('extract_i8', ('extract_i8', 'a', 'b'), 0) => ('extract_i8', 'a', 'b') */
   /* search2182_0_0 -> 0 in the cache */
   /* search2182_0_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 0, 1 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, -2 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 407, 408 },
      -1,
   } },

   /* replace2182_0 -> 0 in the cache */
   /* replace2182_1 -> 1 in the cache */
   /* replace2182 -> 407 in the cache */

   /* ('extract_i8', ('extract_u8', 'a', 'b'), 0) => ('extract_i8', 'a', 'b') */
   /* search2183_0_0 -> 0 in the cache */
   /* search2183_0_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 0, 1 },
      -1,
   } },
   /* search2183_1 -> 408 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 410, 408 },
      -1,
   } },

   /* replace2183_0 -> 0 in the cache */
   /* replace2183_1 -> 1 in the cache */
   /* replace2183 -> 407 in the cache */

   /* ('extract_u8', ('extract_i8', 'a', 'b'), 0) => ('extract_u8', 'a', 'b') */
   /* search2184_0_0 -> 0 in the cache */
   /* search2184_0_1 -> 1 in the cache */
   /* search2184_0 -> 407 in the cache */
   /* search2184_1 -> 408 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 407, 408 },
      -1,
   } },

   /* replace2184_0 -> 0 in the cache */
   /* replace2184_1 -> 1 in the cache */
   /* replace2184 -> 410 in the cache */

   /* ('extract_u8', ('extract_u8', 'a', 'b'), 0) => ('extract_u8', 'a', 'b') */
   /* search2185_0_0 -> 0 in the cache */
   /* search2185_0_1 -> 1 in the cache */
   /* search2185_0 -> 410 in the cache */
   /* search2185_1 -> 408 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 410, 408 },
      -1,
   } },

   /* replace2185_0 -> 0 in the cache */
   /* replace2185_1 -> 1 in the cache */
   /* replace2185 -> 410 in the cache */

   /* ('u2u16', ('u2u8', 'a@16')) => ('extract_u8', 'a', 0) */
   /* search2186_0_0 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 61 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 414 },
      -1,
   } },

   /* replace2186_0 -> 61 in the cache */
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 61, 416 },
      -1,
   } },

   /* ('u2u16', ('i2i8', 'a@16')) => ('extract_u8', 'a', 0) */
   /* search2187_0_0 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 61 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 418 },
      -1,
   } },

   /* replace2187_0 -> 61 in the cache */
   /* replace2187_1 -> 416 in the cache */
   /* replace2187 -> 417 in the cache */

   /* ('i2i16', ('u2u8', 'a@16')) => ('extract_i8', 'a', 0) */
   /* search2188_0_0 -> 61 in the cache */
   /* search2188_0 -> 414 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 414 },
      -1,
   } },

   /* replace2188_0 -> 61 in the cache */
   /* replace2188_1 -> 416 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 61, 416 },
      -1,
   } },

   /* ('i2i16', ('i2i8', 'a@16')) => ('extract_i8', 'a', 0) */
   /* search2189_0_0 -> 61 in the cache */
   /* search2189_0 -> 418 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 418 },
      -1,
   } },

   /* replace2189_0 -> 61 in the cache */
   /* replace2189_1 -> 416 in the cache */
   /* replace2189 -> 421 in the cache */

   /* ('u2u32', ('u2u8', 'a@32')) => ('extract_u8', 'a', 0) */
   /* search2190_0_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 81 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 423 },
      -1,
   } },

   /* replace2190_0 -> 81 in the cache */
   /* replace2190_1 -> 385 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 81, 385 },
      -1,
   } },

   /* ('u2u32', ('i2i8', 'a@32')) => ('extract_u8', 'a', 0) */
   /* search2191_0_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 81 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 426 },
      -1,
   } },

   /* replace2191_0 -> 81 in the cache */
   /* replace2191_1 -> 385 in the cache */
   /* replace2191 -> 425 in the cache */

   /* ('u2u32', ('u2u16', 'a@32')) => ('extract_u16', 'a', 0) */
   /* search2192_0_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 81 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 428 },
      -1,
   } },

   /* replace2192_0 -> 81 in the cache */
   /* replace2192_1 -> 385 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 81, 385 },
      -1,
   } },

   /* ('u2u32', ('i2i16', 'a@32')) => ('extract_u16', 'a', 0) */
   /* search2193_0_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 81 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 431 },
      -1,
   } },

   /* replace2193_0 -> 81 in the cache */
   /* replace2193_1 -> 385 in the cache */
   /* replace2193 -> 430 in the cache */

   /* ('i2i32', ('u2u8', 'a@32')) => ('extract_i8', 'a', 0) */
   /* search2194_0_0 -> 81 in the cache */
   /* search2194_0 -> 423 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 423 },
      -1,
   } },

   /* replace2194_0 -> 81 in the cache */
   /* replace2194_1 -> 385 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 81, 385 },
      -1,
   } },

   /* ('i2i32', ('i2i8', 'a@32')) => ('extract_i8', 'a', 0) */
   /* search2195_0_0 -> 81 in the cache */
   /* search2195_0 -> 426 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 426 },
      -1,
   } },

   /* replace2195_0 -> 81 in the cache */
   /* replace2195_1 -> 385 in the cache */
   /* replace2195 -> 434 in the cache */

   /* ('i2i32', ('u2u16', 'a@32')) => ('extract_i16', 'a', 0) */
   /* search2196_0_0 -> 81 in the cache */
   /* search2196_0 -> 428 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 428 },
      -1,
   } },

   /* replace2196_0 -> 81 in the cache */
   /* replace2196_1 -> 385 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 81, 385 },
      -1,
   } },

   /* ('i2i32', ('i2i16', 'a@32')) => ('extract_i16', 'a', 0) */
   /* search2197_0_0 -> 81 in the cache */
   /* search2197_0 -> 431 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 431 },
      -1,
   } },

   /* replace2197_0 -> 81 in the cache */
   /* replace2197_1 -> 385 in the cache */
   /* replace2197 -> 437 in the cache */

   /* ('ishl', ('extract_u8', 'a@32', 0), 8) => ('insert_u8', 'a', 1) */
   /* search2198_0_0 -> 81 in the cache */
   /* search2198_0_1 -> 385 in the cache */
   /* search2198_0 -> 425 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x8 /* 8 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 425, 439 },
      -1,
   } },

   /* replace2198_0 -> 81 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x1 /* 1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_insert_u8,
      -1, 0,
      { 81, 441 },
      -1,
   } },

   /* ('ishl', ('extract_u8', 'a@32', 0), 16) => ('insert_u8', 'a', 2) */
   /* search2199_0_0 -> 81 in the cache */
   /* search2199_0_1 -> 385 in the cache */
   /* search2199_0 -> 425 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x10 /* 16 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 425, 443 },
      -1,
   } },

   /* replace2199_0 -> 81 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x2 /* 2 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_insert_u8,
      -1, 0,
      { 81, 445 },
      -1,
   } },

   /* ('ishl', ('extract_u8', 'a@32', 0), 24) => ('insert_u8', 'a', 3) */
   /* search2200_0_0 -> 81 in the cache */
   /* search2200_0_1 -> 385 in the cache */
   /* search2200_0 -> 425 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x18 /* 24 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 425, 447 },
      -1,
   } },

   /* replace2200_0 -> 81 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x3 /* 3 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_insert_u8,
      -1, 0,
      { 81, 449 },
      -1,
   } },

   /* ('iand', ('ishl', 'a@32', 8), 65280) => ('insert_u8', 'a', 1) */
   /* search2201_0_0 -> 81 in the cache */
   /* search2201_0_1 -> 439 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 81, 439 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xff00 /* 65280 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 451, 452 },
      -1,
   } },

   /* replace2201_0 -> 81 in the cache */
   /* replace2201_1 -> 441 in the cache */
   /* replace2201 -> 442 in the cache */

   /* ('iand', ('ishl', 'a@32', 16), 16711680) => ('insert_u8', 'a', 2) */
   /* search2202_0_0 -> 81 in the cache */
   /* search2202_0_1 -> 443 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 81, 443 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xff0000 /* 16711680 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 454, 455 },
      -1,
   } },

   /* replace2202_0 -> 81 in the cache */
   /* replace2202_1 -> 445 in the cache */
   /* replace2202 -> 446 in the cache */

   /* ('iand', ('ishl', 'a@32', 24), 4278190080) => ('insert_u8', 'a', 3) */
   /* search2203_0_0 -> 81 in the cache */
   /* search2203_0_1 -> 447 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 81, 447 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xff000000 /* 4278190080 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 457, 458 },
      -1,
   } },

   /* replace2203_0 -> 81 in the cache */
   /* replace2203_1 -> 449 in the cache */
   /* replace2203 -> 450 in the cache */

   /* ('ishl', 'a@32', 24) => ('insert_u8', 'a', 3) */
   /* search2204_0 -> 81 in the cache */
   /* search2204_1 -> 447 in the cache */
   /* search2204 -> 457 in the cache */

   /* replace2204_0 -> 81 in the cache */
   /* replace2204_1 -> 449 in the cache */
   /* replace2204 -> 450 in the cache */

   /* ('ishl', 'a@32', 16) => ('insert_u16', 'a', 1) */
   /* search2205_0 -> 81 in the cache */
   /* search2205_1 -> 443 in the cache */
   /* search2205 -> 454 in the cache */

   /* replace2205_0 -> 81 in the cache */
   /* replace2205_1 -> 441 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_insert_u16,
      -1, 0,
      { 81, 441 },
      -1,
   } },

   /* ('insert_u8', ('extract_u8', 'a', 0), 'b') => ('insert_u8', 'a', 'b') */
   /* search2206_0_0 -> 0 in the cache */
   /* search2206_0_1 -> 408 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 0, 408 },
      -1,
   } },
   /* search2206_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_insert_u8,
      -1, 0,
      { 461, 1 },
      -1,
   } },

   /* replace2206_0 -> 0 in the cache */
   /* replace2206_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_insert_u8,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* ('insert_u16', ('extract_u16', 'a', 0), 'b') => ('insert_u16', 'a', 'b') */
   /* search2207_0_0 -> 0 in the cache */
   /* search2207_0_1 -> 408 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 0, 408 },
      -1,
   } },
   /* search2207_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_insert_u16,
      -1, 0,
      { 464, 1 },
      -1,
   } },

   /* replace2207_0 -> 0 in the cache */
   /* replace2207_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_insert_u16,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* ('iand', ('ine(is_used_once)', 'a@8', 0), ('ine', 'b@8', 0)) => ('ine', ('umin', 'a', 'b'), 0) */
   /* search2208_0_0 -> 134 in the cache */
   { .constant = {
      { nir_search_value_constant, 8 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 134, 467 },
      0,
   } },
   /* search2208_1_0 -> 135 in the cache */
   /* search2208_1_1 -> 467 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 135, 467 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 468, 469 },
      -1,
   } },

   /* replace2208_0_0 -> 134 in the cache */
   /* replace2208_0_1 -> 135 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 134, 135 },
      -1,
   } },
   /* replace2208_1 -> 467 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 471, 467 },
      -1,
   } },

   /* ('ior', ('ieq(is_used_once)', 'a@8', 0), ('ieq', 'b@8', 0)) => ('ieq', ('umin', 'a', 'b'), 0) */
   /* search2209_0_0 -> 134 in the cache */
   /* search2209_0_1 -> 467 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 134, 467 },
      0,
   } },
   /* search2209_1_0 -> 135 in the cache */
   /* search2209_1_1 -> 467 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 135, 467 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 473, 474 },
      -1,
   } },

   /* replace2209_0_0 -> 134 in the cache */
   /* replace2209_0_1 -> 135 in the cache */
   /* replace2209_0 -> 471 in the cache */
   /* replace2209_1 -> 467 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 471, 467 },
      -1,
   } },

   /* ('iand', ('ine(is_used_once)', 'a@16', 0), ('ine', 'b@16', 0)) => ('ine', ('umin', 'a', 'b'), 0) */
   /* search2210_0_0 -> 61 in the cache */
   /* search2210_0_1 -> 416 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 61, 416 },
      0,
   } },
   /* search2210_1_0 -> 62 in the cache */
   /* search2210_1_1 -> 416 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 62, 416 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 477, 478 },
      -1,
   } },

   /* replace2210_0_0 -> 61 in the cache */
   /* replace2210_0_1 -> 62 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 61, 62 },
      -1,
   } },
   /* replace2210_1 -> 416 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 480, 416 },
      -1,
   } },

   /* ('ior', ('ieq(is_used_once)', 'a@16', 0), ('ieq', 'b@16', 0)) => ('ieq', ('umin', 'a', 'b'), 0) */
   /* search2211_0_0 -> 61 in the cache */
   /* search2211_0_1 -> 416 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 61, 416 },
      0,
   } },
   /* search2211_1_0 -> 62 in the cache */
   /* search2211_1_1 -> 416 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 62, 416 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 482, 483 },
      -1,
   } },

   /* replace2211_0_0 -> 61 in the cache */
   /* replace2211_0_1 -> 62 in the cache */
   /* replace2211_0 -> 480 in the cache */
   /* replace2211_1 -> 416 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 480, 416 },
      -1,
   } },

   /* ('iand', ('ine(is_used_once)', 'a@32', 0), ('ine', 'b@32', 0)) => ('ine', ('umin', 'a', 'b'), 0) */
   /* search2212_0_0 -> 81 in the cache */
   /* search2212_0_1 -> 385 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 81, 385 },
      0,
   } },
   /* search2212_1_0 -> 82 in the cache */
   /* search2212_1_1 -> 385 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 82, 385 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 486, 487 },
      -1,
   } },

   /* replace2212_0_0 -> 81 in the cache */
   /* replace2212_0_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 81, 82 },
      -1,
   } },
   /* replace2212_1 -> 385 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 489, 385 },
      -1,
   } },

   /* ('ior', ('ieq(is_used_once)', 'a@32', 0), ('ieq', 'b@32', 0)) => ('ieq', ('umin', 'a', 'b'), 0) */
   /* search2213_0_0 -> 81 in the cache */
   /* search2213_0_1 -> 385 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 81, 385 },
      0,
   } },
   /* search2213_1_0 -> 82 in the cache */
   /* search2213_1_1 -> 385 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 82, 385 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 491, 492 },
      -1,
   } },

   /* replace2213_0_0 -> 81 in the cache */
   /* replace2213_0_1 -> 82 in the cache */
   /* replace2213_0 -> 489 in the cache */
   /* replace2213_1 -> 385 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 489, 385 },
      -1,
   } },

   /* ('iand', ('ine(is_used_once)', 'a@64', 0), ('ine', 'b@64', 0)) => ('ine', ('umin', 'a', 'b'), 0) */
   /* search2214_0_0 -> 114 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 114, 495 },
      0,
   } },
   /* search2214_1_0 -> 115 in the cache */
   /* search2214_1_1 -> 495 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 115, 495 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 496, 497 },
      -1,
   } },

   /* replace2214_0_0 -> 114 in the cache */
   /* replace2214_0_1 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 114, 115 },
      -1,
   } },
   /* replace2214_1 -> 495 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 499, 495 },
      -1,
   } },

   /* ('ior', ('ieq(is_used_once)', 'a@64', 0), ('ieq', 'b@64', 0)) => ('ieq', ('umin', 'a', 'b'), 0) */
   /* search2215_0_0 -> 114 in the cache */
   /* search2215_0_1 -> 495 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 114, 495 },
      0,
   } },
   /* search2215_1_0 -> 115 in the cache */
   /* search2215_1_1 -> 495 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 115, 495 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 501, 502 },
      -1,
   } },

   /* replace2215_0_0 -> 114 in the cache */
   /* replace2215_0_1 -> 115 in the cache */
   /* replace2215_0 -> 499 in the cache */
   /* replace2215_1 -> 495 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 499, 495 },
      -1,
   } },

   /* ('~fadd@16', 1.0, ('fmul(is_used_once)', 'c', ('fadd', 'b', -1.0))) => ('fadd', ('fadd', 1.0, ('fneg', 'c')), ('fmul', 'b', 'c')) */
   /* search2216_0 -> 177 in the cache */
   { .variable = {
      { nir_search_value_variable, 16 },
      0, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search2216_1_1_0 -> 62 in the cache */
   /* search2216_1_1_1 -> 176 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 62, 176 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 505, 506 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 177, 507 },
      -1,
   } },

   /* replace2216_0_0 -> 177 in the cache */
   /* replace2216_0_1_0 -> 505 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 505 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 177, 509 },
      -1,
   } },
   /* replace2216_1_0 -> 62 in the cache */
   /* replace2216_1_1 -> 505 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 62, 505 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 510, 511 },
      -1,
   } },

   /* ('bcsel', 'a', 0, ('b2f16', ('inot', 'b@bool'))) => ('b2f16', ('inot', ('ior', 'a', 'b'))) */
   { .variable = {
      { nir_search_value_variable, 1 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search2217_1 -> 416 in the cache */
   { .variable = {
      { nir_search_value_variable, 1 },
      1, /* b */
      false,
      nir_type_bool,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 514 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2f16,
      -1, 0,
      { 515 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 513, 416, 516 },
      -1,
   } },

   /* replace2217_0_0_0 -> 513 in the cache */
   { .variable = {
      { nir_search_value_variable, 1 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 513, 518 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 1,
      { 519 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2f16,
      -1, 1,
      { 520 },
      -1,
   } },

   /* ('~fadd@32', 1.0, ('fmul(is_used_once)', 'c', ('fadd', 'b', -1.0))) => ('fadd', ('fadd', 1.0, ('fneg', 'c')), ('fmul', 'b', 'c')) */
   /* search2218_0 -> 268 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search2218_1_1_0 -> 82 in the cache */
   /* search2218_1_1_1 -> 266 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 82, 266 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 522, 523 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 268, 524 },
      -1,
   } },

   /* replace2218_0_0 -> 268 in the cache */
   /* replace2218_0_1_0 -> 522 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 522 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 268, 526 },
      -1,
   } },
   /* replace2218_1_0 -> 82 in the cache */
   /* replace2218_1_1 -> 522 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 82, 522 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 527, 528 },
      -1,
   } },

   /* ('bcsel', 'a', 0, ('b2f32', ('inot', 'b@bool'))) => ('b2f32', ('inot', ('ior', 'a', 'b'))) */
   /* search2219_0 -> 513 in the cache */
   /* search2219_1 -> 385 in the cache */
   /* search2219_2_0_0 -> 514 in the cache */
   /* search2219_2_0 -> 515 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2f32,
      -1, 0,
      { 515 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 513, 385, 530 },
      -1,
   } },

   /* replace2219_0_0_0 -> 513 in the cache */
   /* replace2219_0_0_1 -> 518 in the cache */
   /* replace2219_0_0 -> 519 in the cache */
   /* replace2219_0 -> 520 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2f32,
      -1, 1,
      { 520 },
      -1,
   } },

   /* ('~fadd@64', 1.0, ('fmul(is_used_once)', 'c', ('fadd', 'b', -1.0))) => ('fadd', ('fadd', 1.0, ('fneg', 'c')), ('fmul', 'b', 'c')) */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   { .variable = {
      { nir_search_value_variable, 64 },
      0, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search2220_1_1_0 -> 115 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_float, { 0xbff0000000000000 /* -1.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 115, 535 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 534, 536 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 533, 537 },
      -1,
   } },

   /* replace2220_0_0 -> 533 in the cache */
   /* replace2220_0_1_0 -> 534 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 534 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 533, 539 },
      -1,
   } },
   /* replace2220_1_0 -> 115 in the cache */
   /* replace2220_1_1 -> 534 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 115, 534 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 540, 541 },
      -1,
   } },

   /* ('bcsel', 'a', 0, ('b2f64', ('inot', 'b@bool'))) => ('b2f64', ('inot', ('ior', 'a', 'b'))) */
   /* search2221_0 -> 513 in the cache */
   /* search2221_1 -> 495 in the cache */
   /* search2221_2_0_0 -> 514 in the cache */
   /* search2221_2_0 -> 515 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2f64,
      -1, 0,
      { 515 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 513, 495, 543 },
      -1,
   } },

   /* replace2221_0_0_0 -> 513 in the cache */
   /* replace2221_0_0_1 -> 518 in the cache */
   /* replace2221_0_0 -> 519 in the cache */
   /* replace2221_0 -> 520 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2f64,
      -1, 1,
      { 520 },
      -1,
   } },

   /* ('bcsel', 'a', ('fadd(is_used_once)', 'b', 'c'), ('fadd', 'b', 'd')) => ('fadd', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2222_0 -> 513 in the cache */
   /* search2222_1_0 -> 202 in the cache */
   /* search2222_1_1 -> 204 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 202, 204 },
      0,
   } },
   /* search2222_2_0 -> 202 in the cache */
   /* search2222_2_1 -> 205 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 202, 205 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 513, 546, 547 },
      -1,
   } },

   /* replace2222_0 -> 202 in the cache */
   /* replace2222_1_0 -> 513 in the cache */
   /* replace2222_1_1 -> 204 in the cache */
   /* replace2222_1_2 -> 205 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 513, 204, 205 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 202, 549 },
      -1,
   } },

   /* ('bcsel', 'a', ('fadd', 'b', 'c'), ('fadd(is_used_once)', 'b', 'd')) => ('fadd', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2223_0 -> 513 in the cache */
   /* search2223_1_0 -> 202 in the cache */
   /* search2223_1_1 -> 204 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 202, 204 },
      -1,
   } },
   /* search2223_2_0 -> 202 in the cache */
   /* search2223_2_1 -> 205 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 202, 205 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 513, 551, 552 },
      -1,
   } },

   /* replace2223_0 -> 202 in the cache */
   /* replace2223_1_0 -> 513 in the cache */
   /* replace2223_1_1 -> 204 in the cache */
   /* replace2223_1_2 -> 205 in the cache */
   /* replace2223_1 -> 549 in the cache */
   /* replace2223 -> 550 in the cache */

   /* ('bcsel', 'a', ('ffma(is_used_once)', 'b', 'c', 'd'), ('ffma', 'b', 'c', 'e')) => ('ffma', 'b', 'c', ('bcsel', 'a', 'd', 'e')) */
   /* search2224_0 -> 513 in the cache */
   /* search2224_1_0 -> 336 in the cache */
   /* search2224_1_1 -> 343 in the cache */
   /* search2224_1_2 -> 344 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 336, 343, 344 },
      0,
   } },
   /* search2224_2_0 -> 336 in the cache */
   /* search2224_2_1 -> 343 in the cache */
   /* search2224_2_2 -> 345 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 336, 343, 345 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 513, 554, 555 },
      -1,
   } },

   /* replace2224_0 -> 336 in the cache */
   /* replace2224_1 -> 343 in the cache */
   /* replace2224_2_0 -> 513 in the cache */
   /* replace2224_2_1 -> 344 in the cache */
   /* replace2224_2_2 -> 345 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 513, 344, 345 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 336, 343, 557 },
      -1,
   } },

   /* ('bcsel', 'a', ('ffma', 'b', 'c', 'd'), ('ffma(is_used_once)', 'b', 'c', 'e')) => ('ffma', 'b', 'c', ('bcsel', 'a', 'd', 'e')) */
   /* search2225_0 -> 513 in the cache */
   /* search2225_1_0 -> 336 in the cache */
   /* search2225_1_1 -> 343 in the cache */
   /* search2225_1_2 -> 344 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 336, 343, 344 },
      -1,
   } },
   /* search2225_2_0 -> 336 in the cache */
   /* search2225_2_1 -> 343 in the cache */
   /* search2225_2_2 -> 345 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 336, 343, 345 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 513, 559, 560 },
      -1,
   } },

   /* replace2225_0 -> 336 in the cache */
   /* replace2225_1 -> 343 in the cache */
   /* replace2225_2_0 -> 513 in the cache */
   /* replace2225_2_1 -> 344 in the cache */
   /* replace2225_2_2 -> 345 in the cache */
   /* replace2225_2 -> 557 in the cache */
   /* replace2225 -> 558 in the cache */

   /* ('bcsel', 'a', ('ffma(is_used_once)', 'b', 'c', 'd'), ('ffma', 'b', 'e', 'd')) => ('ffma', 'b', ('bcsel', 'a', 'c', 'e'), 'd') */
   /* search2226_0 -> 513 in the cache */
   /* search2226_1_0 -> 336 in the cache */
   /* search2226_1_1 -> 343 in the cache */
   /* search2226_1_2 -> 344 in the cache */
   /* search2226_1 -> 554 in the cache */
   /* search2226_2_0 -> 336 in the cache */
   /* search2226_2_1 -> 345 in the cache */
   /* search2226_2_2 -> 344 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 336, 345, 344 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 513, 554, 562 },
      -1,
   } },

   /* replace2226_0 -> 336 in the cache */
   /* replace2226_1_0 -> 513 in the cache */
   /* replace2226_1_1 -> 343 in the cache */
   /* replace2226_1_2 -> 345 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 513, 343, 345 },
      -1,
   } },
   /* replace2226_2 -> 344 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 336, 564, 344 },
      -1,
   } },

   /* ('bcsel', 'a', ('ffma', 'b', 'c', 'd'), ('ffma(is_used_once)', 'b', 'e', 'd')) => ('ffma', 'b', ('bcsel', 'a', 'c', 'e'), 'd') */
   /* search2227_0 -> 513 in the cache */
   /* search2227_1_0 -> 336 in the cache */
   /* search2227_1_1 -> 343 in the cache */
   /* search2227_1_2 -> 344 in the cache */
   /* search2227_1 -> 559 in the cache */
   /* search2227_2_0 -> 336 in the cache */
   /* search2227_2_1 -> 345 in the cache */
   /* search2227_2_2 -> 344 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 336, 345, 344 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 513, 559, 566 },
      -1,
   } },

   /* replace2227_0 -> 336 in the cache */
   /* replace2227_1_0 -> 513 in the cache */
   /* replace2227_1_1 -> 343 in the cache */
   /* replace2227_1_2 -> 345 in the cache */
   /* replace2227_1 -> 564 in the cache */
   /* replace2227_2 -> 344 in the cache */
   /* replace2227 -> 565 in the cache */

   /* ('bcsel', 'a', ('ffmaz(is_used_once)', 'b', 'c', 'd'), ('ffmaz', 'b', 'c', 'e')) => ('ffmaz', 'b', 'c', ('bcsel', 'a', 'd', 'e')) */
   /* search2228_0 -> 513 in the cache */
   /* search2228_1_0 -> 82 in the cache */
   /* search2228_1_1 -> 84 in the cache */
   /* search2228_1_2 -> 356 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 82, 84, 356 },
      0,
   } },
   /* search2228_2_0 -> 82 in the cache */
   /* search2228_2_1 -> 84 in the cache */
   /* search2228_2_2 -> 364 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 1,
      { 82, 84, 364 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 513, 568, 569 },
      -1,
   } },

   /* replace2228_0 -> 82 in the cache */
   /* replace2228_1 -> 84 in the cache */
   /* replace2228_2_0 -> 513 in the cache */
   /* replace2228_2_1 -> 356 in the cache */
   /* replace2228_2_2 -> 364 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 513, 356, 364 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 82, 84, 571 },
      -1,
   } },

   /* ('bcsel', 'a', ('ffmaz', 'b', 'c', 'd'), ('ffmaz(is_used_once)', 'b', 'c', 'e')) => ('ffmaz', 'b', 'c', ('bcsel', 'a', 'd', 'e')) */
   /* search2229_0 -> 513 in the cache */
   /* search2229_1_0 -> 82 in the cache */
   /* search2229_1_1 -> 84 in the cache */
   /* search2229_1_2 -> 356 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 82, 84, 356 },
      -1,
   } },
   /* search2229_2_0 -> 82 in the cache */
   /* search2229_2_1 -> 84 in the cache */
   /* search2229_2_2 -> 364 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 1,
      { 82, 84, 364 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 513, 573, 574 },
      -1,
   } },

   /* replace2229_0 -> 82 in the cache */
   /* replace2229_1 -> 84 in the cache */
   /* replace2229_2_0 -> 513 in the cache */
   /* replace2229_2_1 -> 356 in the cache */
   /* replace2229_2_2 -> 364 in the cache */
   /* replace2229_2 -> 571 in the cache */
   /* replace2229 -> 572 in the cache */

   /* ('bcsel', 'a', ('ffmaz(is_used_once)', 'b', 'c', 'd'), ('ffmaz', 'b', 'e', 'd')) => ('ffmaz', 'b', ('bcsel', 'a', 'c', 'e'), 'd') */
   /* search2230_0 -> 513 in the cache */
   /* search2230_1_0 -> 82 in the cache */
   /* search2230_1_1 -> 84 in the cache */
   /* search2230_1_2 -> 356 in the cache */
   /* search2230_1 -> 568 in the cache */
   /* search2230_2_0 -> 82 in the cache */
   /* search2230_2_1 -> 364 in the cache */
   /* search2230_2_2 -> 356 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 1,
      { 82, 364, 356 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 513, 568, 576 },
      -1,
   } },

   /* replace2230_0 -> 82 in the cache */
   /* replace2230_1_0 -> 513 in the cache */
   /* replace2230_1_1 -> 84 in the cache */
   /* replace2230_1_2 -> 364 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 513, 84, 364 },
      -1,
   } },
   /* replace2230_2 -> 356 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 82, 578, 356 },
      -1,
   } },

   /* ('bcsel', 'a', ('ffmaz', 'b', 'c', 'd'), ('ffmaz(is_used_once)', 'b', 'e', 'd')) => ('ffmaz', 'b', ('bcsel', 'a', 'c', 'e'), 'd') */
   /* search2231_0 -> 513 in the cache */
   /* search2231_1_0 -> 82 in the cache */
   /* search2231_1_1 -> 84 in the cache */
   /* search2231_1_2 -> 356 in the cache */
   /* search2231_1 -> 573 in the cache */
   /* search2231_2_0 -> 82 in the cache */
   /* search2231_2_1 -> 364 in the cache */
   /* search2231_2_2 -> 356 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 1,
      { 82, 364, 356 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 513, 573, 580 },
      -1,
   } },

   /* replace2231_0 -> 82 in the cache */
   /* replace2231_1_0 -> 513 in the cache */
   /* replace2231_1_1 -> 84 in the cache */
   /* replace2231_1_2 -> 364 in the cache */
   /* replace2231_1 -> 578 in the cache */
   /* replace2231_2 -> 356 in the cache */
   /* replace2231 -> 579 in the cache */

   /* ('~f2f32', ('fabs', ('f2fmp', 'a'))) => ('fabs', 'a') */
   /* search2232_0_0_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 81 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 582 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 583 },
      -1,
   } },

   /* replace2232_0 -> 81 in the cache */
   /* replace2232 -> 93 in the cache */

   /* ('~f2f32', ('fceil', ('f2fmp', 'a'))) => ('fceil', 'a') */
   /* search2233_0_0_0 -> 81 in the cache */
   /* search2233_0_0 -> 582 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fceil,
      -1, 0,
      { 582 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 585 },
      -1,
   } },

   /* replace2233_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fceil,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('~f2f32', ('fcos', ('f2fmp', 'a'))) => ('fcos', 'a') */
   /* search2234_0_0_0 -> 81 in the cache */
   /* search2234_0_0 -> 582 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fcos,
      -1, 0,
      { 582 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 588 },
      -1,
   } },

   /* replace2234_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fcos,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('~f2f32', ('fddx', ('f2fmp', 'a'))) => ('fddx', 'a') */
   /* search2235_0_0_0 -> 81 in the cache */
   /* search2235_0_0 -> 582 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fddx,
      -1, 0,
      { 582 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 591 },
      -1,
   } },

   /* replace2235_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fddx,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('~f2f32', ('fddx_coarse', ('f2fmp', 'a'))) => ('fddx_coarse', 'a') */
   /* search2236_0_0_0 -> 81 in the cache */
   /* search2236_0_0 -> 582 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fddx_coarse,
      -1, 0,
      { 582 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 594 },
      -1,
   } },

   /* replace2236_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fddx_coarse,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('~f2f32', ('fddx_fine', ('f2fmp', 'a'))) => ('fddx_fine', 'a') */
   /* search2237_0_0_0 -> 81 in the cache */
   /* search2237_0_0 -> 582 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fddx_fine,
      -1, 0,
      { 582 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 597 },
      -1,
   } },

   /* replace2237_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fddx_fine,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('~f2f32', ('fddy', ('f2fmp', 'a'))) => ('fddy', 'a') */
   /* search2238_0_0_0 -> 81 in the cache */
   /* search2238_0_0 -> 582 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fddy,
      -1, 0,
      { 582 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 600 },
      -1,
   } },

   /* replace2238_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fddy,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('~f2f32', ('fddy_coarse', ('f2fmp', 'a'))) => ('fddy_coarse', 'a') */
   /* search2239_0_0_0 -> 81 in the cache */
   /* search2239_0_0 -> 582 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fddy_coarse,
      -1, 0,
      { 582 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 603 },
      -1,
   } },

   /* replace2239_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fddy_coarse,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('~f2f32', ('fddy_fine', ('f2fmp', 'a'))) => ('fddy_fine', 'a') */
   /* search2240_0_0_0 -> 81 in the cache */
   /* search2240_0_0 -> 582 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fddy_fine,
      -1, 0,
      { 582 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 606 },
      -1,
   } },

   /* replace2240_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fddy_fine,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('~f2f32', ('fexp2', ('f2fmp', 'a'))) => ('fexp2', 'a') */
   /* search2241_0_0_0 -> 81 in the cache */
   /* search2241_0_0 -> 582 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 0,
      { 582 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 609 },
      -1,
   } },

   /* replace2241_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('~f2f32', ('ffloor', ('f2fmp', 'a'))) => ('ffloor', 'a') */
   /* search2242_0_0_0 -> 81 in the cache */
   /* search2242_0_0 -> 582 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 0,
      { 582 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 612 },
      -1,
   } },

   /* replace2242_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('~f2f32', ('ffract', ('f2fmp', 'a'))) => ('ffract', 'a') */
   /* search2243_0_0_0 -> 81 in the cache */
   /* search2243_0_0 -> 582 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 0,
      { 582 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 615 },
      -1,
   } },

   /* replace2243_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('~f2f32', ('flog2', ('f2fmp', 'a'))) => ('flog2', 'a') */
   /* search2244_0_0_0 -> 81 in the cache */
   /* search2244_0_0 -> 582 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 582 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 618 },
      -1,
   } },

   /* replace2244_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('~f2f32', ('fneg', ('f2fmp', 'a'))) => ('fneg', 'a') */
   /* search2245_0_0_0 -> 81 in the cache */
   /* search2245_0_0 -> 582 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 582 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 621 },
      -1,
   } },

   /* replace2245_0 -> 81 in the cache */
   /* replace2245 -> 89 in the cache */

   /* ('~f2f32', ('frcp', ('f2fmp', 'a'))) => ('frcp', 'a') */
   /* search2246_0_0_0 -> 81 in the cache */
   /* search2246_0_0 -> 582 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 582 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 623 },
      -1,
   } },

   /* replace2246_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('~f2f32', ('fround_even', ('f2fmp', 'a'))) => ('fround_even', 'a') */
   /* search2247_0_0_0 -> 81 in the cache */
   /* search2247_0_0 -> 582 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fround_even,
      -1, 0,
      { 582 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 626 },
      -1,
   } },

   /* replace2247_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fround_even,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('~f2f32', ('frsq', ('f2fmp', 'a'))) => ('frsq', 'a') */
   /* search2248_0_0_0 -> 81 in the cache */
   /* search2248_0_0 -> 582 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_frsq,
      -1, 0,
      { 582 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 629 },
      -1,
   } },

   /* replace2248_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_frsq,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('~f2f32', ('fsat', ('f2fmp', 'a'))) => ('fsat', 'a') */
   /* search2249_0_0_0 -> 81 in the cache */
   /* search2249_0_0 -> 582 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 582 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 632 },
      -1,
   } },

   /* replace2249_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('~f2f32', ('fsign', ('f2fmp', 'a'))) => ('fsign', 'a') */
   /* search2250_0_0_0 -> 81 in the cache */
   /* search2250_0_0 -> 582 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 582 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 635 },
      -1,
   } },

   /* replace2250_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('~f2f32', ('fsin', ('f2fmp', 'a'))) => ('fsin', 'a') */
   /* search2251_0_0_0 -> 81 in the cache */
   /* search2251_0_0 -> 582 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fsin,
      -1, 0,
      { 582 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 638 },
      -1,
   } },

   /* replace2251_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsin,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('~f2f32', ('fsqrt', ('f2fmp', 'a'))) => ('fsqrt', 'a') */
   /* search2252_0_0_0 -> 81 in the cache */
   /* search2252_0_0 -> 582 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fsqrt,
      -1, 0,
      { 582 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 641 },
      -1,
   } },

   /* replace2252_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsqrt,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('~f2f32', ('fadd', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fadd', 'a', 'b') */
   /* search2253_0_0_0 -> 81 in the cache */
   /* search2253_0_0 -> 582 in the cache */
   /* search2253_0_1_0 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 82 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 582, 644 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 1,
      { 645 },
      -1,
   } },

   /* replace2253_0 -> 81 in the cache */
   /* replace2253_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 81, 82 },
      -1,
   } },

   /* ('~f2f32', ('fdiv', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fdiv', 'a', 'b') */
   /* search2254_0_0_0 -> 81 in the cache */
   /* search2254_0_0 -> 582 in the cache */
   /* search2254_0_1_0 -> 82 in the cache */
   /* search2254_0_1 -> 644 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fdiv,
      -1, 0,
      { 582, 644 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 648 },
      -1,
   } },

   /* replace2254_0 -> 81 in the cache */
   /* replace2254_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fdiv,
      -1, 0,
      { 81, 82 },
      -1,
   } },

   /* ('~f2f32', ('fmax', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fmax', 'a', 'b') */
   /* search2255_0_0_0 -> 81 in the cache */
   /* search2255_0_0 -> 582 in the cache */
   /* search2255_0_1_0 -> 82 in the cache */
   /* search2255_0_1 -> 644 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 582, 644 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 1,
      { 651 },
      -1,
   } },

   /* replace2255_0 -> 81 in the cache */
   /* replace2255_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 81, 82 },
      -1,
   } },

   /* ('~f2f32', ('fmin', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fmin', 'a', 'b') */
   /* search2256_0_0_0 -> 81 in the cache */
   /* search2256_0_0 -> 582 in the cache */
   /* search2256_0_1_0 -> 82 in the cache */
   /* search2256_0_1 -> 644 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 582, 644 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 1,
      { 654 },
      -1,
   } },

   /* replace2256_0 -> 81 in the cache */
   /* replace2256_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 81, 82 },
      -1,
   } },

   /* ('~f2f32', ('fmod', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fmod', 'a', 'b') */
   /* search2257_0_0_0 -> 81 in the cache */
   /* search2257_0_0 -> 582 in the cache */
   /* search2257_0_1_0 -> 82 in the cache */
   /* search2257_0_1 -> 644 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmod,
      -1, 0,
      { 582, 644 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 657 },
      -1,
   } },

   /* replace2257_0 -> 81 in the cache */
   /* replace2257_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmod,
      -1, 0,
      { 81, 82 },
      -1,
   } },

   /* ('~f2f32', ('fmul', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fmul', 'a', 'b') */
   /* search2258_0_0_0 -> 81 in the cache */
   /* search2258_0_0 -> 582 in the cache */
   /* search2258_0_1_0 -> 82 in the cache */
   /* search2258_0_1 -> 644 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 582, 644 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 1,
      { 660 },
      -1,
   } },

   /* replace2258_0 -> 81 in the cache */
   /* replace2258_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 81, 82 },
      -1,
   } },

   /* ('~f2f32', ('fpow', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fpow', 'a', 'b') */
   /* search2259_0_0_0 -> 81 in the cache */
   /* search2259_0_0 -> 582 in the cache */
   /* search2259_0_1_0 -> 82 in the cache */
   /* search2259_0_1 -> 644 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 582, 644 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 663 },
      -1,
   } },

   /* replace2259_0 -> 81 in the cache */
   /* replace2259_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 81, 82 },
      -1,
   } },

   /* ('~f2f32', ('frem', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('frem', 'a', 'b') */
   /* search2260_0_0_0 -> 81 in the cache */
   /* search2260_0_0 -> 582 in the cache */
   /* search2260_0_1_0 -> 82 in the cache */
   /* search2260_0_1 -> 644 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_frem,
      -1, 0,
      { 582, 644 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 666 },
      -1,
   } },

   /* replace2260_0 -> 81 in the cache */
   /* replace2260_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_frem,
      -1, 0,
      { 81, 82 },
      -1,
   } },

   /* ('~f2f32', ('ffma', ('f2fmp', 'a'), ('f2fmp', 'b'), ('f2fmp', 'c'))) => ('ffma', 'a', 'b', 'c') */
   /* search2261_0_0_0 -> 81 in the cache */
   /* search2261_0_0 -> 582 in the cache */
   /* search2261_0_1_0 -> 82 in the cache */
   /* search2261_0_1 -> 644 in the cache */
   /* search2261_0_2_0 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 84 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 582, 644, 669 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 1,
      { 670 },
      -1,
   } },

   /* replace2261_0 -> 81 in the cache */
   /* replace2261_1 -> 82 in the cache */
   /* replace2261_2 -> 84 in the cache */
   /* replace2261 -> 86 in the cache */

   /* ('~f2f32', ('flrp', ('f2fmp', 'a'), ('f2fmp', 'b'), ('f2fmp', 'c'))) => ('flrp', 'a', 'b', 'c') */
   /* search2262_0_0_0 -> 81 in the cache */
   /* search2262_0_0 -> 582 in the cache */
   /* search2262_0_1_0 -> 82 in the cache */
   /* search2262_0_1 -> 644 in the cache */
   /* search2262_0_2_0 -> 84 in the cache */
   /* search2262_0_2 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 582, 644, 669 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 672 },
      -1,
   } },

   /* replace2262_0 -> 81 in the cache */
   /* replace2262_1 -> 82 in the cache */
   /* replace2262_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 81, 82, 84 },
      -1,
   } },

   /* ('~feq', ('f2fmp', 'a'), ('f2fmp', 'b')) => ('feq', 'a', 'b') */
   /* search2263_0_0 -> 81 in the cache */
   /* search2263_0 -> 582 in the cache */
   /* search2263_1_0 -> 82 in the cache */
   /* search2263_1 -> 644 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 582, 644 },
      -1,
   } },

   /* replace2263_0 -> 81 in the cache */
   /* replace2263_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 81, 82 },
      -1,
   } },

   /* ('~fge', ('f2fmp', 'a'), ('f2fmp', 'b')) => ('fge', 'a', 'b') */
   /* search2264_0_0 -> 81 in the cache */
   /* search2264_0 -> 582 in the cache */
   /* search2264_1_0 -> 82 in the cache */
   /* search2264_1 -> 644 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 582, 644 },
      -1,
   } },

   /* replace2264_0 -> 81 in the cache */
   /* replace2264_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 81, 82 },
      -1,
   } },

   /* ('~flt', ('f2fmp', 'a'), ('f2fmp', 'b')) => ('flt', 'a', 'b') */
   /* search2265_0_0 -> 81 in the cache */
   /* search2265_0 -> 582 in the cache */
   /* search2265_1_0 -> 82 in the cache */
   /* search2265_1 -> 644 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 582, 644 },
      -1,
   } },

   /* replace2265_0 -> 81 in the cache */
   /* replace2265_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 81, 82 },
      -1,
   } },

   /* ('~fneu', ('f2fmp', 'a'), ('f2fmp', 'b')) => ('fneu', 'a', 'b') */
   /* search2266_0_0 -> 81 in the cache */
   /* search2266_0 -> 582 in the cache */
   /* search2266_1_0 -> 82 in the cache */
   /* search2266_1 -> 644 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 582, 644 },
      -1,
   } },

   /* replace2266_0 -> 81 in the cache */
   /* replace2266_1 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 81, 82 },
      -1,
   } },

   /* ('f2fmp', 'a') => ('f2f16', 'a') */
   /* search2267_0 -> 81 in the cache */
   /* search2267 -> 582 in the cache */

   /* replace2267_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('f2imp', 'a') => ('f2i16', 'a') */
   /* search2268_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2imp,
      -1, 0,
      { 81 },
      -1,
   } },

   /* replace2268_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2i16,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('f2ump', 'a') => ('f2u16', 'a') */
   /* search2269_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2ump,
      -1, 0,
      { 81 },
      -1,
   } },

   /* replace2269_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2u16,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('i2imp', 'a') => ('i2i16', 'a') */
   /* search2270_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 81 },
      -1,
   } },

   /* replace2270_0 -> 81 in the cache */
   /* replace2270 -> 431 in the cache */

   /* ('i2fmp', 'a') => ('i2f16', 'a') */
   /* search2271_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2fmp,
      -1, 0,
      { 81 },
      -1,
   } },

   /* replace2271_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2f16,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('i2imp', 'a') => ('u2u16', 'a') */
   /* search2272_0 -> 81 in the cache */
   /* search2272 -> 688 in the cache */

   /* replace2272_0 -> 81 in the cache */
   /* replace2272 -> 428 in the cache */

   /* ('u2fmp', 'a') => ('u2f16', 'a') */
   /* search2273_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2fmp,
      -1, 0,
      { 81 },
      -1,
   } },

   /* replace2273_0 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2f16,
      -1, 0,
      { 81 },
      -1,
   } },

   /* ('fisfinite', 'a') => ('flt', ('fabs', 'a'), inf) */
   /* search2274_0 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fisfinite,
      -1, 0,
      { 42 },
      -1,
   } },

   /* replace2274_0_0 -> 42 in the cache */
   /* replace2274_0 -> 47 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x7ff0000000000000 /* inf */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 47, 694 },
      -1,
   } },

   /* ('fcsel', ('slt', 0, 'a'), 'b', 'c') => ('fcsel_gt', 'a', 'b', 'c') */
   /* search2275_0_0 -> 385 in the cache */
   /* search2275_0_1 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_slt,
      -1, 0,
      { 385, 81 },
      -1,
   } },
   /* search2275_1 -> 82 in the cache */
   /* search2275_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fcsel,
      -1, 0,
      { 696, 82, 84 },
      -1,
   } },

   /* replace2275_0 -> 81 in the cache */
   /* replace2275_1 -> 82 in the cache */
   /* replace2275_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fcsel_gt,
      -1, 0,
      { 81, 82, 84 },
      -1,
   } },

   /* ('fcsel', ('slt', 'a', 0), 'b', 'c') => ('fcsel_gt', ('fneg', 'a'), 'b', 'c') */
   /* search2276_0_0 -> 81 in the cache */
   /* search2276_0_1 -> 385 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_slt,
      -1, 0,
      { 81, 385 },
      -1,
   } },
   /* search2276_1 -> 82 in the cache */
   /* search2276_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fcsel,
      -1, 0,
      { 699, 82, 84 },
      -1,
   } },

   /* replace2276_0_0 -> 81 in the cache */
   /* replace2276_0 -> 89 in the cache */
   /* replace2276_1 -> 82 in the cache */
   /* replace2276_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fcsel_gt,
      -1, 0,
      { 89, 82, 84 },
      -1,
   } },

   /* ('fcsel', ('sge', 'a', 0), 'b', 'c') => ('fcsel_ge', 'a', 'b', 'c') */
   /* search2277_0_0 -> 81 in the cache */
   /* search2277_0_1 -> 385 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sge,
      -1, 0,
      { 81, 385 },
      -1,
   } },
   /* search2277_1 -> 82 in the cache */
   /* search2277_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fcsel,
      -1, 0,
      { 702, 82, 84 },
      -1,
   } },

   /* replace2277_0 -> 81 in the cache */
   /* replace2277_1 -> 82 in the cache */
   /* replace2277_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fcsel_ge,
      -1, 0,
      { 81, 82, 84 },
      -1,
   } },

   /* ('fcsel', ('sge', 0, 'a'), 'b', 'c') => ('fcsel_ge', ('fneg', 'a'), 'b', 'c') */
   /* search2278_0_0 -> 385 in the cache */
   /* search2278_0_1 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sge,
      -1, 0,
      { 385, 81 },
      -1,
   } },
   /* search2278_1 -> 82 in the cache */
   /* search2278_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fcsel,
      -1, 0,
      { 705, 82, 84 },
      -1,
   } },

   /* replace2278_0_0 -> 81 in the cache */
   /* replace2278_0 -> 89 in the cache */
   /* replace2278_1 -> 82 in the cache */
   /* replace2278_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fcsel_ge,
      -1, 0,
      { 89, 82, 84 },
      -1,
   } },

   /* ('bcsel', ('ilt', 0, 'a@32'), 'b@32', 'c@32') => ('i32csel_gt', 'a', 'b', 'c') */
   /* search2279_0_0 -> 385 in the cache */
   /* search2279_0_1 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 385, 81 },
      -1,
   } },
   /* search2279_1 -> 82 in the cache */
   /* search2279_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 708, 82, 84 },
      -1,
   } },

   /* replace2279_0 -> 81 in the cache */
   /* replace2279_1 -> 82 in the cache */
   /* replace2279_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i32csel_gt,
      -1, 0,
      { 81, 82, 84 },
      -1,
   } },

   /* ('bcsel', ('ilt', 'a@32', 0), 'b@32', 'c@32') => ('i32csel_ge', 'a', 'c', 'b') */
   /* search2280_0_0 -> 81 in the cache */
   /* search2280_0_1 -> 385 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 81, 385 },
      -1,
   } },
   /* search2280_1 -> 82 in the cache */
   /* search2280_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 711, 82, 84 },
      -1,
   } },

   /* replace2280_0 -> 81 in the cache */
   /* replace2280_1 -> 84 in the cache */
   /* replace2280_2 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i32csel_ge,
      -1, 0,
      { 81, 84, 82 },
      -1,
   } },

   /* ('bcsel', ('ige', 'a@32', 0), 'b@32', 'c@32') => ('i32csel_ge', 'a', 'b', 'c') */
   /* search2281_0_0 -> 81 in the cache */
   /* search2281_0_1 -> 385 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 81, 385 },
      -1,
   } },
   /* search2281_1 -> 82 in the cache */
   /* search2281_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 714, 82, 84 },
      -1,
   } },

   /* replace2281_0 -> 81 in the cache */
   /* replace2281_1 -> 82 in the cache */
   /* replace2281_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i32csel_ge,
      -1, 0,
      { 81, 82, 84 },
      -1,
   } },

   /* ('bcsel', ('ige', 0, 'a@32'), 'b@32', 'c@32') => ('i32csel_gt', 'a', 'c', 'b') */
   /* search2282_0_0 -> 385 in the cache */
   /* search2282_0_1 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 385, 81 },
      -1,
   } },
   /* search2282_1 -> 82 in the cache */
   /* search2282_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 717, 82, 84 },
      -1,
   } },

   /* replace2282_0 -> 81 in the cache */
   /* replace2282_1 -> 84 in the cache */
   /* replace2282_2 -> 82 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i32csel_gt,
      -1, 0,
      { 81, 84, 82 },
      -1,
   } },

   /* ('bcsel', ('flt', 0, 'a@32'), 'b@32', 'c@32') => ('fcsel_gt', 'a', 'b', 'c') */
   /* search2283_0_0 -> 385 in the cache */
   /* search2283_0_1 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 385, 81 },
      -1,
   } },
   /* search2283_1 -> 82 in the cache */
   /* search2283_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 720, 82, 84 },
      -1,
   } },

   /* replace2283_0 -> 81 in the cache */
   /* replace2283_1 -> 82 in the cache */
   /* replace2283_2 -> 84 in the cache */
   /* replace2283 -> 698 in the cache */

   /* ('bcsel', ('flt', 'a@32', 0), 'b@32', 'c@32') => ('fcsel_gt', ('fneg', 'a'), 'b', 'c') */
   /* search2284_0_0 -> 81 in the cache */
   /* search2284_0_1 -> 385 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 81, 385 },
      -1,
   } },
   /* search2284_1 -> 82 in the cache */
   /* search2284_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 82, 84 },
      -1,
   } },

   /* replace2284_0_0 -> 81 in the cache */
   /* replace2284_0 -> 89 in the cache */
   /* replace2284_1 -> 82 in the cache */
   /* replace2284_2 -> 84 in the cache */
   /* replace2284 -> 701 in the cache */

   /* ('bcsel', ('fge', 'a@32', 0), 'b@32', 'c@32') => ('fcsel_ge', 'a', 'b', 'c') */
   /* search2285_0_0 -> 81 in the cache */
   /* search2285_0_1 -> 385 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 81, 385 },
      -1,
   } },
   /* search2285_1 -> 82 in the cache */
   /* search2285_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 724, 82, 84 },
      -1,
   } },

   /* replace2285_0 -> 81 in the cache */
   /* replace2285_1 -> 82 in the cache */
   /* replace2285_2 -> 84 in the cache */
   /* replace2285 -> 704 in the cache */

   /* ('bcsel', ('fge', 0, 'a@32'), 'b@32', 'c@32') => ('fcsel_ge', ('fneg', 'a'), 'b', 'c') */
   /* search2286_0_0 -> 385 in the cache */
   /* search2286_0_1 -> 81 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 385, 81 },
      -1,
   } },
   /* search2286_1 -> 82 in the cache */
   /* search2286_2 -> 84 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 726, 82, 84 },
      -1,
   } },

   /* replace2286_0_0 -> 81 in the cache */
   /* replace2286_0 -> 89 in the cache */
   /* replace2286_1 -> 82 in the cache */
   /* replace2286_2 -> 84 in the cache */
   /* replace2286 -> 707 in the cache */

};

static const nir_search_expression_cond nir_opt_algebraic_late_expression_cond[] = {
   (is_used_once),
   (is_only_used_by_fadd),
   (is_only_used_as_float),
};

static const nir_search_variable_cond nir_opt_algebraic_late_variable_cond[] = {
   (is_finite),
   (is_not_const),
   (is_gt_0_and_lt_1),
   (is_not_fmul),
   (is_not_negative),
   (is_not_const_and_not_fsign),
};

static const struct transform nir_opt_algebraic_late_transforms[] = {
   { ~0, ~0, ~0 }, /* Sentinel */

   { 59, 149, 124 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 235, 42, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 234, 42, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 153, 155, 126 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 156, 157, 127 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 242, 243, 130 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 244, 245, 130 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 246, 247, 130 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 248, 249, 130 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 255, 264, 131 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 267, 269, 17 },
   { 271, 272, 17 },
   { 273, 274, 17 },
   { 275, 276, 17 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 386, 385, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 386, 385, 0 },
   { 390, 395, 133 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 387, 385, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 387, 385, 0 },
   { 396, 400, 133 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 401, 42, 0 },
   { 403, 42, 0 },
   { 457, 450, 134 },
   { 454, 460, 135 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 404, 42, 0 },
   { 405, 42, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 406, 42, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 582, 683, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 684, 685, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 686, 687, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 688, 431, 0 },
   { 688, 428, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 689, 690, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 691, 692, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 693, 695, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 548, 550, 0 },
   { 553, 550, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 556, 558, 0 },
   { 561, 558, 0 },
   { 563, 565, 0 },
   { 567, 565, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 570, 572, 0 },
   { 575, 572, 0 },
   { 577, 579, 0 },
   { 581, 579, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 723, 701, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 548, 550, 0 },
   { 553, 550, 0 },
   { 723, 701, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 556, 558, 0 },
   { 561, 558, 0 },
   { 563, 565, 0 },
   { 567, 565, 0 },
   { 723, 701, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 570, 572, 0 },
   { 575, 572, 0 },
   { 577, 579, 0 },
   { 581, 579, 0 },
   { 723, 701, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 721, 698, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 548, 550, 0 },
   { 553, 550, 0 },
   { 721, 698, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 556, 558, 0 },
   { 561, 558, 0 },
   { 563, 565, 0 },
   { 567, 565, 0 },
   { 721, 698, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 570, 572, 0 },
   { 575, 572, 0 },
   { 577, 579, 0 },
   { 581, 579, 0 },
   { 721, 698, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 721, 698, 136 },
   { 723, 701, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 548, 550, 0 },
   { 553, 550, 0 },
   { 721, 698, 136 },
   { 723, 701, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 556, 558, 0 },
   { 561, 558, 0 },
   { 563, 565, 0 },
   { 567, 565, 0 },
   { 721, 698, 136 },
   { 723, 701, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 570, 572, 0 },
   { 575, 572, 0 },
   { 577, 579, 0 },
   { 581, 579, 0 },
   { 721, 698, 136 },
   { 723, 701, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 725, 704, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 548, 550, 0 },
   { 553, 550, 0 },
   { 725, 704, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 556, 558, 0 },
   { 561, 558, 0 },
   { 563, 565, 0 },
   { 567, 565, 0 },
   { 725, 704, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 570, 572, 0 },
   { 575, 572, 0 },
   { 577, 579, 0 },
   { 581, 579, 0 },
   { 725, 704, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 727, 707, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 548, 550, 0 },
   { 553, 550, 0 },
   { 727, 707, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 556, 558, 0 },
   { 561, 558, 0 },
   { 563, 565, 0 },
   { 567, 565, 0 },
   { 727, 707, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 570, 572, 0 },
   { 575, 572, 0 },
   { 577, 579, 0 },
   { 581, 579, 0 },
   { 727, 707, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 725, 704, 136 },
   { 727, 707, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 548, 550, 0 },
   { 553, 550, 0 },
   { 725, 704, 136 },
   { 727, 707, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 556, 558, 0 },
   { 561, 558, 0 },
   { 563, 565, 0 },
   { 567, 565, 0 },
   { 725, 704, 136 },
   { 727, 707, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 570, 572, 0 },
   { 575, 572, 0 },
   { 577, 579, 0 },
   { 581, 579, 0 },
   { 725, 704, 136 },
   { 727, 707, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 712, 713, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 548, 550, 0 },
   { 553, 550, 0 },
   { 712, 713, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 556, 558, 0 },
   { 561, 558, 0 },
   { 563, 565, 0 },
   { 567, 565, 0 },
   { 712, 713, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 570, 572, 0 },
   { 575, 572, 0 },
   { 577, 579, 0 },
   { 581, 579, 0 },
   { 712, 713, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 709, 710, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 548, 550, 0 },
   { 553, 550, 0 },
   { 709, 710, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 556, 558, 0 },
   { 561, 558, 0 },
   { 563, 565, 0 },
   { 567, 565, 0 },
   { 709, 710, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 570, 572, 0 },
   { 575, 572, 0 },
   { 577, 579, 0 },
   { 581, 579, 0 },
   { 709, 710, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 709, 710, 136 },
   { 712, 713, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 548, 550, 0 },
   { 553, 550, 0 },
   { 709, 710, 136 },
   { 712, 713, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 556, 558, 0 },
   { 561, 558, 0 },
   { 563, 565, 0 },
   { 567, 565, 0 },
   { 709, 710, 136 },
   { 712, 713, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 570, 572, 0 },
   { 575, 572, 0 },
   { 577, 579, 0 },
   { 581, 579, 0 },
   { 709, 710, 136 },
   { 712, 713, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 715, 716, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 548, 550, 0 },
   { 553, 550, 0 },
   { 715, 716, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 556, 558, 0 },
   { 561, 558, 0 },
   { 563, 565, 0 },
   { 567, 565, 0 },
   { 715, 716, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 570, 572, 0 },
   { 575, 572, 0 },
   { 577, 579, 0 },
   { 581, 579, 0 },
   { 715, 716, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 718, 719, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 548, 550, 0 },
   { 553, 550, 0 },
   { 718, 719, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 556, 558, 0 },
   { 561, 558, 0 },
   { 563, 565, 0 },
   { 567, 565, 0 },
   { 718, 719, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 570, 572, 0 },
   { 575, 572, 0 },
   { 577, 579, 0 },
   { 581, 579, 0 },
   { 718, 719, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 715, 716, 136 },
   { 718, 719, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 548, 550, 0 },
   { 553, 550, 0 },
   { 715, 716, 136 },
   { 718, 719, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 556, 558, 0 },
   { 561, 558, 0 },
   { 563, 565, 0 },
   { 567, 565, 0 },
   { 715, 716, 136 },
   { 718, 719, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 570, 572, 0 },
   { 575, 572, 0 },
   { 577, 579, 0 },
   { 581, 579, 0 },
   { 715, 716, 136 },
   { 718, 719, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 11, 10, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 225, 226, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 186, 188, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4, 6, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 223, 224, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 215, 216, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 679, 680, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 59, 149, 124 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 60, 42, 0 },
   { 59, 149, 124 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 59, 149, 124 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 59, 149, 124 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 59, 149, 124 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 59, 149, 124 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 17, 16, 0 },
   { 29, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 229, 200, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 13, 14, 0 },
   { 26, 14, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 227, 228, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 217, 218, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 191, 192, 0 },
   { 199, 200, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 677, 678, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 20, 21, 0 },
   { 32, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 232, 233, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 193, 194, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 219, 220, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 675, 676, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 22, 23, 0 },
   { 33, 23, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 230, 231, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 195, 196, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 221, 222, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 681, 682, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 304, 307, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 252, 254, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 308, 310, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 32 },
   { 85, 86, 33 },
   { 118, 119, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 102, 103, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 32 },
   { 85, 86, 33 },
   { 118, 119, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 32 },
   { 85, 86, 33 },
   { 118, 119, 34 },
   { 278, 269, 17 },
   { 280, 272, 17 },
   { 281, 274, 17 },
   { 282, 276, 17 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 102, 103, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 32 },
   { 85, 86, 33 },
   { 118, 119, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 102, 103, 33 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 32 },
   { 85, 86, 33 },
   { 102, 103, 33 },
   { 118, 119, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 180, 182, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 174, 179, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 174, 179, 129 },
   { 180, 182, 129 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 296, 292, 17 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 235, 42, 0 },
   { 453, 442, 134 },
   { 456, 446, 134 },
   { 459, 450, 134 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 235, 42, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 46, 48, 0 },
   { 235, 42, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 235, 42, 0 },
   { 470, 472, 0 },
   { 479, 481, 0 },
   { 488, 490, 0 },
   { 498, 500, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 57, 58, 0 },
   { 234, 42, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 234, 42, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 234, 42, 0 },
   { 475, 476, 0 },
   { 484, 485, 0 },
   { 493, 494, 0 },
   { 503, 504, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 162, 166, 128 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 151, 152, 125 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 168, 170, 128 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 151, 152, 125 },
   { 162, 166, 128 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 162, 166, 128 },
   { 168, 170, 128 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 151, 152, 125 },
   { 168, 170, 128 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 171, 173, 128 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 409, 407, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 411, 407, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 412, 410, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 413, 410, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 401, 42, 0 },
   { 403, 42, 0 },
   { 440, 442, 134 },
   { 444, 446, 134 },
   { 448, 450, 134 },
   { 457, 450, 134 },
   { 454, 460, 135 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 462, 463, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 415, 417, 58 },
   { 424, 425, 58 },
   { 429, 430, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 419, 417, 58 },
   { 427, 425, 58 },
   { 432, 430, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 420, 421, 58 },
   { 433, 434, 58 },
   { 436, 437, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 422, 421, 58 },
   { 435, 434, 58 },
   { 438, 437, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 465, 466, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 255, 264, 131 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 700, 701, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 697, 698, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 697, 698, 136 },
   { 700, 701, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 703, 704, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 706, 707, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 703, 704, 136 },
   { 706, 707, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 12, 6, 0 },
   { 225, 226, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 8, 10, 0 },
   { 223, 224, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 18, 14, 0 },
   { 30, 14, 0 },
   { 229, 200, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 15, 16, 0 },
   { 28, 16, 0 },
   { 227, 228, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 209, 214, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 68, 70, 32 },
   { 88, 90, 33 },
   { 121, 123, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 105, 106, 33 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 342, 347, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 72, 75, 32 },
   { 92, 95, 33 },
   { 125, 128, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 108, 109, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 375, 377, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 68, 70, 32 },
   { 88, 90, 33 },
   { 121, 123, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 105, 106, 33 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 239, 241, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 32 },
   { 85, 86, 33 },
   { 118, 119, 34 },
   { 508, 512, 14 },
   { 525, 529, 16 },
   { 538, 542, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 32 },
   { 85, 86, 33 },
   { 118, 119, 34 },
   { 278, 269, 17 },
   { 280, 272, 17 },
   { 281, 274, 17 },
   { 282, 276, 17 },
   { 508, 512, 14 },
   { 525, 529, 16 },
   { 538, 542, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 342, 347, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 72, 75, 32 },
   { 92, 95, 33 },
   { 125, 128, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 108, 109, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 375, 377, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 342, 347, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 72, 75, 32 },
   { 92, 95, 33 },
   { 125, 128, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 108, 109, 33 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 375, 377, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 32 },
   { 68, 70, 32 },
   { 85, 86, 33 },
   { 88, 90, 33 },
   { 118, 119, 34 },
   { 121, 123, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 32 },
   { 85, 86, 33 },
   { 105, 106, 33 },
   { 118, 119, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 32 },
   { 85, 86, 33 },
   { 118, 119, 34 },
   { 342, 347, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 32 },
   { 72, 75, 32 },
   { 85, 86, 33 },
   { 92, 95, 33 },
   { 118, 119, 34 },
   { 125, 128, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 32 },
   { 85, 86, 33 },
   { 108, 109, 33 },
   { 118, 119, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 32 },
   { 85, 86, 33 },
   { 118, 119, 34 },
   { 375, 377, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 68, 70, 32 },
   { 88, 90, 33 },
   { 102, 103, 33 },
   { 121, 123, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 102, 103, 33 },
   { 105, 106, 33 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 102, 103, 33 },
   { 342, 347, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 72, 75, 32 },
   { 92, 95, 33 },
   { 102, 103, 33 },
   { 125, 128, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 102, 103, 33 },
   { 108, 109, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 102, 103, 33 },
   { 375, 377, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 68, 70, 32 },
   { 88, 90, 33 },
   { 105, 106, 33 },
   { 121, 123, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 68, 70, 32 },
   { 88, 90, 33 },
   { 121, 123, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 342, 347, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 68, 70, 32 },
   { 72, 75, 32 },
   { 88, 90, 33 },
   { 92, 95, 33 },
   { 121, 123, 34 },
   { 125, 128, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 68, 70, 32 },
   { 88, 90, 33 },
   { 108, 109, 33 },
   { 121, 123, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 68, 70, 32 },
   { 88, 90, 33 },
   { 121, 123, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 375, 377, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 105, 106, 33 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 342, 347, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 72, 75, 32 },
   { 92, 95, 33 },
   { 105, 106, 33 },
   { 125, 128, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 105, 106, 33 },
   { 108, 109, 33 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 105, 106, 33 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 375, 377, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 72, 75, 32 },
   { 92, 95, 33 },
   { 125, 128, 34 },
   { 342, 347, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 108, 109, 33 },
   { 342, 347, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 342, 347, 132 },
   { 375, 377, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 72, 75, 32 },
   { 92, 95, 33 },
   { 108, 109, 33 },
   { 125, 128, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 72, 75, 32 },
   { 92, 95, 33 },
   { 125, 128, 34 },
   { 375, 377, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 108, 109, 33 },
   { 375, 377, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 291, 292, 17 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 294, 292, 17 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 287, 288, 17 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 287, 288, 17 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 287, 288, 17 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 294, 292, 17 },
   { 296, 292, 17 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 267, 269, 17 },
   { 271, 272, 17 },
   { 273, 274, 17 },
   { 275, 276, 17 },
   { 287, 288, 17 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 287, 288, 17 },
   { 291, 292, 17 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 287, 288, 17 },
   { 294, 292, 17 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 287, 288, 17 },
   { 296, 292, 17 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 622, 89, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 655, 656, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 673, 674, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 661, 662, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 652, 653, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 646, 647, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 671, 86, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 584, 93, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 633, 634, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 627, 628, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 642, 643, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 630, 631, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 586, 587, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 589, 590, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 592, 593, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 595, 596, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 598, 599, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 601, 602, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 604, 605, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 607, 608, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 610, 611, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 613, 614, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 616, 617, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 619, 620, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 624, 625, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 636, 637, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 639, 640, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 649, 650, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 658, 659, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 664, 665, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 667, 668, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 517, 521, 0 },
   { 531, 532, 0 },
   { 544, 545, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 517, 521, 0 },
   { 531, 532, 0 },
   { 544, 545, 0 },
   { 723, 701, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 517, 521, 0 },
   { 531, 532, 0 },
   { 544, 545, 0 },
   { 721, 698, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 517, 521, 0 },
   { 531, 532, 0 },
   { 544, 545, 0 },
   { 721, 698, 136 },
   { 723, 701, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 517, 521, 0 },
   { 531, 532, 0 },
   { 544, 545, 0 },
   { 725, 704, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 517, 521, 0 },
   { 531, 532, 0 },
   { 544, 545, 0 },
   { 727, 707, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 517, 521, 0 },
   { 531, 532, 0 },
   { 544, 545, 0 },
   { 725, 704, 136 },
   { 727, 707, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 517, 521, 0 },
   { 531, 532, 0 },
   { 544, 545, 0 },
   { 712, 713, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 517, 521, 0 },
   { 531, 532, 0 },
   { 544, 545, 0 },
   { 709, 710, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 517, 521, 0 },
   { 531, 532, 0 },
   { 544, 545, 0 },
   { 709, 710, 136 },
   { 712, 713, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 517, 521, 0 },
   { 531, 532, 0 },
   { 544, 545, 0 },
   { 715, 716, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 517, 521, 0 },
   { 531, 532, 0 },
   { 544, 545, 0 },
   { 718, 719, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 517, 521, 0 },
   { 531, 532, 0 },
   { 544, 545, 0 },
   { 715, 716, 136 },
   { 718, 719, 136 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 316, 317, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 298, 292, 17 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 298, 292, 17 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 298, 292, 17 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 38, 41, 0 },
   { 235, 42, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 53, 54, 0 },
   { 234, 42, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 59, 149, 124 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 59, 149, 124 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 328, 334, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 78, 80, 32 },
   { 98, 100, 33 },
   { 131, 133, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 113, 33 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 363, 369, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 328, 334, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 78, 80, 32 },
   { 98, 100, 33 },
   { 131, 133, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 113, 33 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 363, 369, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 328, 334, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 363, 369, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 32 },
   { 85, 86, 33 },
   { 118, 119, 34 },
   { 328, 334, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 32 },
   { 78, 80, 32 },
   { 85, 86, 33 },
   { 98, 100, 33 },
   { 118, 119, 34 },
   { 131, 133, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 32 },
   { 85, 86, 33 },
   { 112, 113, 33 },
   { 118, 119, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 32 },
   { 85, 86, 33 },
   { 118, 119, 34 },
   { 363, 369, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 102, 103, 33 },
   { 328, 334, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 78, 80, 32 },
   { 98, 100, 33 },
   { 102, 103, 33 },
   { 131, 133, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 102, 103, 33 },
   { 112, 113, 33 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 102, 103, 33 },
   { 363, 369, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 68, 70, 32 },
   { 88, 90, 33 },
   { 121, 123, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 328, 334, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 68, 70, 32 },
   { 78, 80, 32 },
   { 88, 90, 33 },
   { 98, 100, 33 },
   { 121, 123, 34 },
   { 131, 133, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 68, 70, 32 },
   { 88, 90, 33 },
   { 112, 113, 33 },
   { 121, 123, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 68, 70, 32 },
   { 88, 90, 33 },
   { 121, 123, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 363, 369, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 105, 106, 33 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 328, 334, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 78, 80, 32 },
   { 98, 100, 33 },
   { 105, 106, 33 },
   { 131, 133, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 105, 106, 33 },
   { 112, 113, 33 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 105, 106, 33 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 363, 369, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 328, 334, 132 },
   { 342, 347, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 78, 80, 32 },
   { 98, 100, 33 },
   { 131, 133, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 342, 347, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 113, 33 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 342, 347, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 342, 347, 132 },
   { 363, 369, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 72, 75, 32 },
   { 92, 95, 33 },
   { 125, 128, 34 },
   { 328, 334, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 72, 75, 32 },
   { 78, 80, 32 },
   { 92, 95, 33 },
   { 98, 100, 33 },
   { 125, 128, 34 },
   { 131, 133, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 72, 75, 32 },
   { 92, 95, 33 },
   { 112, 113, 33 },
   { 125, 128, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 72, 75, 32 },
   { 92, 95, 33 },
   { 125, 128, 34 },
   { 363, 369, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 108, 109, 33 },
   { 328, 334, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 78, 80, 32 },
   { 98, 100, 33 },
   { 108, 109, 33 },
   { 131, 133, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 108, 109, 33 },
   { 112, 113, 33 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 108, 109, 33 },
   { 363, 369, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 328, 334, 132 },
   { 375, 377, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 78, 80, 32 },
   { 98, 100, 33 },
   { 131, 133, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 375, 377, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 113, 33 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 375, 377, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 363, 369, 132 },
   { 375, 377, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 78, 80, 32 },
   { 98, 100, 33 },
   { 131, 133, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 328, 334, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 113, 33 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 328, 334, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 328, 334, 132 },
   { 363, 369, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 78, 80, 32 },
   { 98, 100, 33 },
   { 112, 113, 33 },
   { 131, 133, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 78, 80, 32 },
   { 98, 100, 33 },
   { 131, 133, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 363, 369, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 113, 33 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 363, 369, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 59, 149, 124 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 59, 149, 124 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 349, 355, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 379, 384, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 349, 355, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 379, 384, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 32 },
   { 85, 86, 33 },
   { 118, 119, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 349, 355, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 32 },
   { 85, 86, 33 },
   { 118, 119, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 379, 384, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 102, 103, 33 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 349, 355, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 102, 103, 33 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 379, 384, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 68, 70, 32 },
   { 88, 90, 33 },
   { 121, 123, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 349, 355, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 68, 70, 32 },
   { 88, 90, 33 },
   { 121, 123, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 379, 384, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 105, 106, 33 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 349, 355, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 105, 106, 33 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 379, 384, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 342, 347, 132 },
   { 349, 355, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 342, 347, 132 },
   { 379, 384, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 72, 75, 32 },
   { 92, 95, 33 },
   { 125, 128, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 349, 355, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 72, 75, 32 },
   { 92, 95, 33 },
   { 125, 128, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 379, 384, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 108, 109, 33 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 349, 355, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 108, 109, 33 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 379, 384, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 349, 355, 132 },
   { 375, 377, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 375, 377, 132 },
   { 379, 384, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 328, 334, 132 },
   { 349, 355, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 328, 334, 132 },
   { 379, 384, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 78, 80, 32 },
   { 98, 100, 33 },
   { 131, 133, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 349, 355, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 78, 80, 32 },
   { 98, 100, 33 },
   { 131, 133, 34 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 379, 384, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 113, 33 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 349, 355, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 112, 113, 33 },
   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 379, 384, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 349, 355, 132 },
   { 363, 369, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 363, 369, 132 },
   { 379, 384, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 137, 138, 122 },
   { 140, 141, 122 },
   { 143, 144, 122 },
   { 146, 147, 123 },
   { 349, 355, 132 },
   { 379, 384, 132 },
   { ~0, ~0, ~0 }, /* Sentinel */

};

static const struct per_op_table nir_opt_algebraic_late_pass_op_table[nir_num_search_ops] = {
   [nir_op_flt] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         2,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
      },
      
      .num_filtered_states = 7,
      .table = (const uint16_t []) {
      
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         121,
         122,
         123,
         3,
         241,
         0,
         124,
         0,
         0,
         0,
         0,
         0,
         0,
         125,
         0,
         126,
         0,
         0,
         126,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         127,
         0,
         0,
         242,
         0,
         126,
         0,
         0,
         126,
      },
   },
   [nir_op_fadd] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         7,
         8,
         9,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         3,
         11,
         4,
         10,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         14,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         12,
         0,
         12,
         16,
         12,
         0,
         0,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         11,
         10,
         0,
         0,
         17,
         18,
         19,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         20,
         21,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 22,
      .table = (const uint16_t []) {
      
         5,
         6,
         165,
         166,
         166,
         167,
         5,
         246,
         247,
         165,
         166,
         166,
         248,
         249,
         250,
         251,
         356,
         357,
         358,
         359,
         406,
         407,
         6,
         6,
         168,
         169,
         170,
         171,
         6,
         252,
         253,
         254,
         255,
         256,
         257,
         258,
         259,
         260,
         360,
         361,
         362,
         363,
         408,
         409,
         165,
         168,
         165,
         172,
         172,
         173,
         165,
         246,
         247,
         165,
         172,
         172,
         261,
         262,
         263,
         264,
         364,
         357,
         358,
         365,
         406,
         407,
         166,
         169,
         172,
         166,
         166,
         174,
         166,
         265,
         266,
         172,
         166,
         166,
         267,
         268,
         269,
         270,
         366,
         367,
         368,
         369,
         410,
         411,
         166,
         170,
         172,
         166,
         166,
         174,
         166,
         265,
         266,
         172,
         166,
         166,
         267,
         268,
         269,
         270,
         366,
         367,
         368,
         369,
         410,
         411,
         167,
         171,
         173,
         174,
         174,
         167,
         167,
         271,
         272,
         173,
         174,
         174,
         273,
         274,
         275,
         276,
         370,
         371,
         372,
         373,
         412,
         413,
         5,
         6,
         165,
         166,
         166,
         167,
         175,
         246,
         247,
         165,
         166,
         166,
         248,
         249,
         250,
         251,
         356,
         357,
         358,
         359,
         406,
         407,
         246,
         252,
         246,
         265,
         265,
         271,
         246,
         246,
         277,
         246,
         265,
         265,
         278,
         279,
         280,
         281,
         374,
         375,
         376,
         377,
         414,
         415,
         247,
         253,
         247,
         266,
         266,
         272,
         247,
         277,
         247,
         247,
         266,
         266,
         282,
         283,
         284,
         285,
         378,
         379,
         380,
         381,
         416,
         417,
         165,
         254,
         165,
         172,
         172,
         173,
         165,
         246,
         247,
         165,
         172,
         172,
         261,
         262,
         263,
         264,
         364,
         357,
         358,
         365,
         406,
         407,
         166,
         255,
         172,
         166,
         166,
         174,
         166,
         265,
         266,
         172,
         166,
         166,
         267,
         268,
         269,
         270,
         366,
         367,
         368,
         369,
         410,
         411,
         166,
         256,
         172,
         166,
         166,
         174,
         166,
         265,
         266,
         172,
         166,
         166,
         267,
         268,
         269,
         270,
         366,
         367,
         368,
         369,
         410,
         411,
         248,
         257,
         261,
         267,
         267,
         273,
         248,
         278,
         282,
         261,
         267,
         267,
         248,
         286,
         287,
         288,
         382,
         383,
         384,
         385,
         418,
         419,
         249,
         258,
         262,
         268,
         268,
         274,
         249,
         279,
         283,
         262,
         268,
         268,
         286,
         249,
         289,
         290,
         386,
         387,
         388,
         389,
         420,
         421,
         250,
         259,
         263,
         269,
         269,
         275,
         250,
         280,
         284,
         263,
         269,
         269,
         287,
         289,
         250,
         291,
         390,
         391,
         392,
         393,
         422,
         423,
         251,
         260,
         264,
         270,
         270,
         276,
         251,
         281,
         285,
         264,
         270,
         270,
         288,
         290,
         291,
         251,
         394,
         395,
         396,
         397,
         424,
         425,
         356,
         360,
         364,
         366,
         366,
         370,
         356,
         374,
         378,
         364,
         366,
         366,
         382,
         386,
         390,
         394,
         356,
         398,
         399,
         400,
         426,
         427,
         357,
         361,
         357,
         367,
         367,
         371,
         357,
         375,
         379,
         357,
         367,
         367,
         383,
         387,
         391,
         395,
         398,
         357,
         401,
         402,
         428,
         429,
         358,
         362,
         358,
         368,
         368,
         372,
         358,
         376,
         380,
         358,
         368,
         368,
         384,
         388,
         392,
         396,
         399,
         401,
         358,
         403,
         430,
         431,
         359,
         363,
         365,
         369,
         369,
         373,
         359,
         377,
         381,
         365,
         369,
         369,
         385,
         389,
         393,
         397,
         400,
         402,
         403,
         359,
         432,
         433,
         406,
         408,
         406,
         410,
         410,
         412,
         406,
         414,
         416,
         406,
         410,
         410,
         418,
         420,
         422,
         424,
         426,
         428,
         430,
         432,
         406,
         434,
         407,
         409,
         407,
         411,
         411,
         413,
         407,
         415,
         417,
         407,
         411,
         411,
         419,
         421,
         423,
         425,
         427,
         429,
         431,
         433,
         434,
         407,
      },
   },
   [nir_op_fneg] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         4,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         9,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         2,
         2,
         10,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         2,
         2,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
      },
      
      .num_filtered_states = 11,
      .table = (const uint16_t []) {
      
         7,
         128,
         129,
         130,
         131,
         132,
         133,
         353,
         354,
         404,
         405,
      },
   },
   [nir_op_fge] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         2,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
      },
      
      .num_filtered_states = 8,
      .table = (const uint16_t []) {
      
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         134,
         135,
         9,
         9,
         243,
         9,
         0,
         136,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         137,
         0,
         138,
         0,
         0,
         138,
         0,
         0,
         139,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         140,
         0,
         0,
         0,
         244,
         0,
         138,
         0,
         0,
         138,
         0,
         0,
         245,
         0,
         0,
         0,
         0,
         0,
         0,
      },
   },
   [nir_op_feq] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         2,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
      },
      
      .num_filtered_states = 7,
      .table = (const uint16_t []) {
      
         11,
         12,
         11,
         11,
         11,
         11,
         11,
         12,
         12,
         141,
         142,
         143,
         144,
         12,
         11,
         141,
         11,
         11,
         11,
         11,
         11,
         11,
         142,
         11,
         145,
         11,
         11,
         11,
         11,
         143,
         11,
         11,
         11,
         11,
         11,
         11,
         144,
         11,
         11,
         11,
         11,
         11,
         11,
         12,
         11,
         11,
         11,
         11,
         146,
      },
   },
   [nir_op_fneu] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         2,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
      },
      
      .num_filtered_states = 6,
      .table = (const uint16_t []) {
      
         13,
         14,
         13,
         13,
         13,
         13,
         14,
         14,
         147,
         148,
         149,
         14,
         13,
         147,
         13,
         13,
         13,
         13,
         13,
         148,
         13,
         150,
         13,
         13,
         13,
         149,
         13,
         13,
         13,
         13,
         13,
         14,
         13,
         13,
         13,
         151,
      },
   },
   [nir_op_iand] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         3,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         4,
         4,
         4,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 8,
      .table = (const uint16_t []) {
      
         15,
         15,
         15,
         15,
         15,
         15,
         15,
         15,
         15,
         15,
         15,
         15,
         15,
         182,
         15,
         15,
         15,
         15,
         183,
         15,
         184,
         15,
         15,
         15,
         15,
         15,
         15,
         15,
         15,
         15,
         15,
         351,
         15,
         15,
         184,
         15,
         15,
         15,
         15,
         351,
         15,
         182,
         15,
         15,
         15,
         15,
         15,
         15,
         15,
         15,
         15,
         15,
         15,
         15,
         185,
         15,
         15,
         15,
         15,
         351,
         351,
         15,
         15,
         15,
      },
   },
   [nir_op_ior] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         2,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         1,
         1,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 6,
      .table = (const uint16_t []) {
      
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         352,
         16,
         16,
         16,
         186,
         16,
         352,
         16,
         16,
         186,
         187,
         16,
         16,
         16,
         16,
         16,
         16,
         188,
         16,
         16,
         352,
         352,
         16,
         16,
         16,
      },
   },
   [nir_op_fmul] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         2,
         5,
         5,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 6,
      .table = (const uint16_t []) {
      
         17,
         18,
         157,
         158,
         17,
         348,
         18,
         18,
         159,
         160,
         18,
         349,
         157,
         159,
         157,
         161,
         157,
         348,
         158,
         160,
         161,
         158,
         158,
         350,
         17,
         18,
         157,
         158,
         162,
         348,
         348,
         349,
         348,
         350,
         348,
         348,
      },
   },
   [nir_op_fabs] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         0,
         189,
         190,
         191,
      },
   },
   [nir_op_fmulz] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         19,
      },
   },
   [nir_op_iadd] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         2,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         20,
         193,
         194,
         195,
         193,
         193,
         196,
         197,
         194,
         196,
         194,
         198,
         195,
         197,
         198,
         195,
      },
   },
   [nir_op_ineg] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         21,
      },
   },
   [nir_op_iabs] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         22,
      },
   },
   [nir_op_isub] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         23,
         23,
         199,
         199,
      },
   },
   [nir_op_vec2] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         176,
         177,
         178,
      },
   },
   [nir_op_fsat] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         24,
         212,
      },
   },
   [nir_op_fmin] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         1,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         2,
         2,
         2,
         2,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         2,
         2,
         2,
         2,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         1,
         1,
         2,
         2,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         0,
         0,
         152,
         152,
         0,
         0,
         152,
         153,
         0,
         0,
         0,
         0,
         154,
      },
   },
   [nir_op_fdot2] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         25,
      },
   },
   [nir_op_fdot3] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         26,
      },
   },
   [nir_op_fdot4] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         27,
      },
   },
   [nir_op_fdph] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         28,
      },
   },
   [nir_op_flrp] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         155,
         155,
         155,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         156,
      },
   },
   [nir_op_fround_even] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         29,
         213,
      },
   },
   [nir_op_ffma] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         6,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         7,
         4,
         8,
         7,
         3,
         0,
         0,
         9,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         9,
         9,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         0,
         10,
         0,
         10,
         0,
         0,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         7,
         0,
         0,
         2,
         2,
         0,
         0,
         9,
         9,
         0,
         0,
         9,
         9,
         0,
         9,
         9,
         0,
         9,
         9,
         0,
         0,
         9,
         9,
         0,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         0,
         9,
         9,
         0,
         0,
         9,
         9,
         0,
         0,
         9,
         9,
         0,
         0,
         9,
         9,
         0,
         9,
         9,
         0,
         9,
         9,
         9,
         2,
         2,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
      },
      
      .num_filtered_states = 11,
      .table = (const uint16_t []) {
      
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         31,
         30,
         179,
         179,
         30,
         293,
         294,
         294,
         30,
         292,
         30,
         30,
         30,
         179,
         180,
         30,
         30,
         179,
         180,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         30,
         30,
         179,
         180,
         30,
         30,
         179,
         180,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         295,
         295,
         295,
         296,
         296,
         295,
         295,
         296,
         296,
         295,
         297,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         31,
         30,
         179,
         179,
         30,
         293,
         294,
         294,
         30,
         292,
         30,
         31,
         30,
         179,
         179,
         30,
         293,
         294,
         294,
         30,
         292,
         30,
         31,
         30,
         179,
         180,
         30,
         293,
         294,
         298,
         30,
         292,
         30,
         31,
         30,
         179,
         179,
         30,
         293,
         294,
         294,
         30,
         292,
         30,
         31,
         30,
         179,
         179,
         30,
         293,
         294,
         294,
         30,
         292,
         30,
         31,
         30,
         179,
         179,
         30,
         293,
         294,
         294,
         30,
         292,
         30,
         31,
         30,
         179,
         180,
         30,
         293,
         294,
         298,
         30,
         292,
         30,
         31,
         30,
         179,
         179,
         30,
         293,
         294,
         294,
         30,
         292,
         30,
         31,
         30,
         179,
         179,
         30,
         293,
         294,
         294,
         30,
         292,
         295,
         299,
         295,
         296,
         296,
         295,
         300,
         301,
         301,
         295,
         297,
         30,
         31,
         30,
         179,
         179,
         30,
         293,
         294,
         294,
         30,
         292,
         30,
         30,
         30,
         179,
         180,
         30,
         30,
         179,
         180,
         30,
         292,
         30,
         31,
         30,
         179,
         180,
         30,
         293,
         294,
         298,
         30,
         292,
         30,
         30,
         30,
         179,
         180,
         30,
         30,
         179,
         180,
         30,
         292,
         30,
         30,
         30,
         179,
         180,
         30,
         30,
         179,
         180,
         30,
         292,
         30,
         30,
         30,
         179,
         180,
         30,
         30,
         179,
         180,
         30,
         292,
         30,
         30,
         30,
         179,
         180,
         30,
         30,
         179,
         180,
         30,
         292,
         30,
         30,
         30,
         179,
         180,
         30,
         30,
         179,
         180,
         30,
         292,
         30,
         30,
         30,
         179,
         180,
         30,
         30,
         179,
         180,
         30,
         292,
         30,
         30,
         30,
         179,
         180,
         30,
         30,
         179,
         180,
         30,
         292,
         295,
         295,
         295,
         296,
         302,
         295,
         295,
         296,
         302,
         295,
         297,
         30,
         30,
         30,
         179,
         180,
         30,
         30,
         179,
         180,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         31,
         30,
         179,
         179,
         30,
         293,
         294,
         294,
         30,
         292,
         30,
         30,
         30,
         179,
         180,
         30,
         30,
         179,
         180,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         30,
         30,
         179,
         180,
         30,
         30,
         179,
         180,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         295,
         295,
         295,
         296,
         296,
         295,
         295,
         296,
         296,
         295,
         297,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         31,
         30,
         179,
         179,
         30,
         293,
         294,
         294,
         30,
         292,
         30,
         30,
         30,
         179,
         180,
         30,
         30,
         179,
         180,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         30,
         30,
         179,
         180,
         30,
         30,
         179,
         180,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         295,
         295,
         295,
         296,
         296,
         295,
         295,
         296,
         296,
         295,
         297,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         31,
         30,
         179,
         179,
         30,
         293,
         294,
         294,
         30,
         292,
         30,
         30,
         30,
         179,
         180,
         30,
         30,
         179,
         180,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         181,
         30,
         179,
         179,
         30,
         292,
         30,
         30,
         30,
         179,
         180,
         30,
         30,
         179,
         180,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         295,
         295,
         295,
         296,
         296,
         295,
         295,
         296,
         296,
         295,
         297,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         30,
         30,
         179,
         180,
         30,
         30,
         179,
         180,
         30,
         292,
         30,
         31,
         30,
         179,
         180,
         30,
         293,
         294,
         298,
         30,
         292,
         30,
         30,
         30,
         179,
         180,
         30,
         30,
         179,
         180,
         30,
         292,
         30,
         30,
         30,
         179,
         180,
         30,
         30,
         179,
         180,
         30,
         292,
         30,
         30,
         30,
         179,
         180,
         30,
         30,
         179,
         180,
         30,
         292,
         30,
         30,
         30,
         179,
         180,
         30,
         30,
         179,
         180,
         30,
         292,
         30,
         30,
         30,
         179,
         180,
         30,
         30,
         179,
         180,
         30,
         292,
         30,
         30,
         30,
         179,
         180,
         30,
         30,
         179,
         180,
         30,
         292,
         30,
         30,
         30,
         179,
         180,
         30,
         30,
         179,
         180,
         30,
         292,
         295,
         295,
         295,
         296,
         302,
         295,
         295,
         296,
         302,
         295,
         297,
         30,
         30,
         30,
         179,
         180,
         30,
         30,
         179,
         180,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         31,
         30,
         179,
         179,
         30,
         293,
         294,
         294,
         30,
         292,
         30,
         30,
         30,
         179,
         180,
         30,
         30,
         179,
         180,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         30,
         30,
         179,
         180,
         30,
         30,
         179,
         180,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         295,
         295,
         295,
         296,
         296,
         295,
         295,
         296,
         296,
         295,
         297,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         31,
         30,
         179,
         179,
         30,
         293,
         294,
         294,
         30,
         292,
         30,
         30,
         30,
         179,
         180,
         30,
         30,
         179,
         180,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         30,
         30,
         179,
         180,
         30,
         30,
         179,
         180,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         295,
         295,
         295,
         296,
         296,
         295,
         295,
         296,
         296,
         295,
         297,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         295,
         295,
         295,
         296,
         296,
         295,
         295,
         296,
         296,
         295,
         297,
         295,
         299,
         295,
         296,
         296,
         295,
         300,
         301,
         301,
         295,
         297,
         295,
         295,
         295,
         296,
         302,
         295,
         295,
         296,
         302,
         295,
         297,
         295,
         295,
         295,
         296,
         296,
         295,
         295,
         296,
         296,
         295,
         297,
         295,
         295,
         295,
         296,
         296,
         295,
         295,
         296,
         296,
         295,
         297,
         295,
         295,
         295,
         296,
         296,
         295,
         295,
         296,
         296,
         295,
         297,
         295,
         295,
         295,
         296,
         302,
         295,
         295,
         296,
         302,
         295,
         297,
         295,
         295,
         295,
         296,
         296,
         295,
         295,
         296,
         296,
         295,
         297,
         295,
         295,
         295,
         296,
         296,
         295,
         295,
         296,
         296,
         295,
         297,
         295,
         295,
         295,
         296,
         296,
         295,
         295,
         296,
         296,
         295,
         297,
         295,
         295,
         295,
         296,
         296,
         295,
         295,
         296,
         296,
         295,
         297,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         31,
         30,
         179,
         179,
         30,
         293,
         294,
         294,
         30,
         292,
         30,
         30,
         30,
         179,
         180,
         30,
         30,
         179,
         180,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         30,
         30,
         179,
         180,
         30,
         30,
         179,
         180,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
         295,
         295,
         295,
         296,
         296,
         295,
         295,
         296,
         296,
         295,
         297,
         30,
         30,
         30,
         179,
         179,
         30,
         30,
         179,
         179,
         30,
         292,
      },
   },
   [nir_op_fmax] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         0,
         163,
         0,
         0,
         0,
         164,
      },
   },
   [nir_op_bcsel] = {
      .filter = (const uint16_t []) {
         0,
         1,
         2,
         3,
         4,
         5,
         5,
         0,
         6,
         7,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         0,
         10,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         13,
         14,
         15,
         16,
         17,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         6,
         6,
         0,
         6,
         0,
         0,
         0,
         0,
         18,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         19,
         0,
         0,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         2,
         7,
         6,
         6,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
      },
      
      .num_filtered_states = 20,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         334,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         70,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         71,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         72,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         334,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         70,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         71,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         72,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         335,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         74,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         75,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         76,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         73,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         336,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         78,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         79,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         80,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         77,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         337,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         82,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         83,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         84,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         81,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         334,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         70,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         71,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         72,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         338,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         86,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         87,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         88,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         85,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         339,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         90,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         91,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         92,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         89,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         340,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         94,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         95,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         96,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         93,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         334,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         70,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         71,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         72,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         334,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         70,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         71,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         72,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         334,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         70,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         71,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         72,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         341,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         98,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         99,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         100,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         97,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         342,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         102,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         103,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         104,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         101,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         343,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         106,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         107,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         108,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         105,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         344,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         110,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         111,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         112,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         109,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         345,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         114,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         115,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         116,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         113,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         346,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         118,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         119,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         120,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         117,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         347,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         334,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         70,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         71,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         72,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         334,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         70,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         71,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         72,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
   },
   [nir_op_fsqrt] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         32,
         214,
      },
   },
   [nir_op_frsq] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         33,
         215,
      },
   },
   [nir_op_ffmaz] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         34,
         192,
         355,
         34,
         192,
         355,
         34,
         192,
         355,
         34,
         192,
         355,
         34,
         192,
         355,
         34,
         192,
         355,
         34,
         192,
         355,
         34,
         192,
         355,
         34,
         192,
         355,
      },
   },
   [nir_op_ubfe] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         35,
         0,
         36,
         0,
         35,
         0,
         36,
      },
   },
   [nir_op_ibfe] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         37,
         0,
         38,
         0,
         37,
         0,
         38,
      },
   },
   [nir_op_ishl] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         0,
         39,
         0,
         0,
         39,
         0,
         0,
         204,
         0,
      },
   },
   [nir_op_ishr] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         40,
         0,
         40,
      },
   },
   [nir_op_ushr] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         41,
         0,
         41,
      },
   },
   [nir_op_extract_i8] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         42,
         42,
         42,
         42,
         42,
         42,
         42,
         42,
         42,
         200,
         42,
         42,
         42,
         201,
         42,
         42,
      },
   },
   [nir_op_extract_u8] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         43,
         44,
         43,
         43,
         43,
         44,
         43,
         43,
         43,
         202,
         43,
         43,
         43,
         203,
         43,
         43,
      },
   },
   [nir_search_op_u2u] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         45,
         206,
         207,
      },
   },
   [nir_search_op_i2i] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         46,
         208,
         209,
      },
   },
   [nir_op_insert_u8] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         0,
         205,
         205,
      },
   },
   [nir_op_insert_u16] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         0,
         210,
         210,
      },
   },
   [nir_op_extract_u16] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         47,
         0,
         47,
      },
   },
   [nir_op_ine] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         48,
         48,
         48,
      },
   },
   [nir_op_ieq] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         49,
         49,
         49,
      },
   },
   [nir_search_op_b2f] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         211,
      },
   },
   [nir_op_inot] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         50,
      },
   },
   [nir_search_op_f2f] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         11,
         12,
         13,
         14,
         15,
         16,
         17,
         18,
         19,
         20,
         21,
         22,
         23,
         24,
         25,
         26,
         27,
         28,
         29,
         30,
         31,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 32,
      .table = (const uint16_t []) {
      
         0,
         303,
         304,
         305,
         306,
         307,
         308,
         309,
         310,
         311,
         312,
         313,
         314,
         315,
         316,
         317,
         318,
         319,
         320,
         321,
         322,
         323,
         324,
         325,
         326,
         327,
         328,
         329,
         330,
         331,
         332,
         333,
      },
   },
   [nir_op_f2fmp] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         51,
      },
   },
   [nir_op_fceil] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         216,
      },
   },
   [nir_op_fcos] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         217,
      },
   },
   [nir_op_fddx] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         218,
      },
   },
   [nir_op_fddx_coarse] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         219,
      },
   },
   [nir_op_fddx_fine] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         220,
      },
   },
   [nir_op_fddy] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         221,
      },
   },
   [nir_op_fddy_coarse] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         222,
      },
   },
   [nir_op_fddy_fine] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         223,
      },
   },
   [nir_op_fexp2] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         224,
      },
   },
   [nir_op_ffloor] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         225,
      },
   },
   [nir_op_ffract] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         226,
      },
   },
   [nir_op_flog2] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         227,
      },
   },
   [nir_op_frcp] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         228,
      },
   },
   [nir_op_fsign] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         229,
      },
   },
   [nir_op_fsin] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         230,
      },
   },
   [nir_op_fdiv] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         231,
      },
   },
   [nir_op_fmod] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         232,
      },
   },
   [nir_op_fpow] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         233,
      },
   },
   [nir_op_frem] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         234,
      },
   },
   [nir_op_f2imp] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         52,
      },
   },
   [nir_op_f2ump] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         53,
      },
   },
   [nir_op_i2imp] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         54,
      },
   },
   [nir_op_i2fmp] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         55,
      },
   },
   [nir_op_u2fmp] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         56,
      },
   },
   [nir_op_fisfinite] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         57,
      },
   },
   [nir_op_fcsel] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         2,
         3,
         4,
         5,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 7,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         235,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         236,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         237,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         238,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         239,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
         240,
      },
   },
   [nir_op_slt] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         58,
         59,
         60,
      },
   },
   [nir_op_sge] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         61,
         62,
         63,
      },
   },
   [nir_op_ilt] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         64,
         65,
         66,
      },
   },
   [nir_op_ige] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         67,
         68,
         69,
      },
   },
};

/* Mapping from state index to offset in transforms (0 being no transforms) */
static const uint16_t nir_opt_algebraic_late_transform_offsets[] = {
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   1,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   3,
   5,
   0,
   0,
   0,
   0,
   7,
   9,
   0,
   0,
   11,
   13,
   15,
   17,
   19,
   0,
   21,
   0,
   0,
   0,
   26,
   28,
   31,
   33,
   36,
   41,
   44,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   46,
   48,
   50,
   52,
   55,
   57,
   59,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   61,
   64,
   69,
   74,
   76,
   80,
   86,
   92,
   94,
   98,
   104,
   110,
   113,
   118,
   125,
   132,
   134,
   138,
   144,
   150,
   152,
   156,
   162,
   168,
   171,
   176,
   183,
   190,
   192,
   196,
   202,
   208,
   210,
   214,
   220,
   226,
   229,
   234,
   241,
   248,
   250,
   254,
   260,
   266,
   268,
   272,
   278,
   284,
   287,
   292,
   299,
   306,
   308,
   310,
   312,
   314,
   316,
   318,
   320,
   322,
   325,
   327,
   329,
   331,
   333,
   336,
   338,
   341,
   343,
   345,
   348,
   350,
   353,
   355,
   0,
   357,
   359,
   361,
   364,
   366,
   368,
   370,
   0,
   372,
   0,
   374,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   376,
   0,
   378,
   383,
   387,
   389,
   394,
   398,
   406,
   408,
   416,
   422,
   0,
   427,
   429,
   431,
   0,
   434,
   0,
   436,
   441,
   443,
   446,
   452,
   455,
   457,
   0,
   0,
   0,
   0,
   463,
   465,
   467,
   469,
   472,
   475,
   478,
   480,
   482,
   484,
   486,
   488,
   496,
   498,
   502,
   506,
   510,
   514,
   0,
   0,
   516,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   518,
   520,
   522,
   525,
   527,
   529,
   532,
   535,
   538,
   542,
   546,
   548,
   556,
   562,
   564,
   568,
   570,
   572,
   580,
   586,
   592,
   599,
   610,
   612,
   616,
   618,
   620,
   626,
   634,
   640,
   646,
   657,
   666,
   671,
   678,
   683,
   688,
   697,
   704,
   707,
   712,
   715,
   718,
   727,
   736,
   747,
   756,
   765,
   772,
   781,
   788,
   795,
   800,
   803,
   806,
   811,
   816,
   0,
   819,
   821,
   823,
   825,
   827,
   829,
   832,
   838,
   841,
   844,
   847,
   849,
   851,
   853,
   855,
   857,
   859,
   861,
   863,
   865,
   867,
   869,
   871,
   873,
   875,
   877,
   879,
   881,
   883,
   885,
   887,
   889,
   891,
   893,
   895,
   897,
   899,
   901,
   903,
   905,
   907,
   909,
   913,
   918,
   923,
   929,
   934,
   939,
   945,
   950,
   955,
   961,
   966,
   971,
   977,
   979,
   981,
   983,
   985,
   988,
   991,
   993,
   0,
   995,
   997,
   1005,
   1011,
   1013,
   1015,
   1023,
   1029,
   1031,
   1037,
   1043,
   1048,
   1059,
   1068,
   1073,
   1076,
   1085,
   1092,
   1095,
   1104,
   1115,
   1124,
   1133,
   1140,
   1149,
   1156,
   1163,
   1166,
   1175,
   1182,
   1185,
   1190,
   1201,
   1210,
   1215,
   1218,
   1227,
   1234,
   1237,
   1240,
   1249,
   1256,
   1259,
   1268,
   1275,
   1278,
   1287,
   1296,
   1303,
   1305,
   1307,
   1313,
   1319,
   1325,
   1331,
   1340,
   1349,
   1356,
   1363,
   1372,
   1381,
   1388,
   1395,
   1402,
   1409,
   1418,
   1427,
   1434,
   1441,
   1448,
   1455,
   1462,
   1469,
   1478,
   1487,
   1494,
   1501,
   1508,
   1515,
};

static const nir_algebraic_table nir_opt_algebraic_late_table = {
   .transforms = nir_opt_algebraic_late_transforms,
   .transform_offsets = nir_opt_algebraic_late_transform_offsets,
   .pass_op_table = nir_opt_algebraic_late_pass_op_table,
   .values = nir_opt_algebraic_late_values,
   .expression_cond = nir_opt_algebraic_late_expression_cond,
   .variable_cond = nir_opt_algebraic_late_variable_cond,
};

bool
nir_opt_algebraic_late(nir_shader *shader)
{
   bool progress = false;
   bool condition_flags[137];
   const nir_shader_compiler_options *options = shader->options;
   const shader_info *info = &shader->info;
   (void) options;
   (void) info;

   /* This is not a great place for this, but it seems to be the best place
    * for it. Check that at most one kind of lowering is requested for
    * bitfield extract and bitfield insert. Otherwise the lowering can fight
    * with each other and optimizations.
    */
   assert((int)options->lower_bitfield_extract +
          (int)options->lower_bitfield_extract_to_shifts <= 1);
   assert((int)options->lower_bitfield_insert +
          (int)options->lower_bitfield_insert_to_shifts +
          (int)options->lower_bitfield_insert_to_bitfield_select <= 1);


   STATIC_ASSERT(728 == ARRAY_SIZE(nir_opt_algebraic_late_values));
   condition_flags[0] = true;
   condition_flags[1] = !options->lower_bitops;
   condition_flags[2] = options->lower_bitops;
   condition_flags[3] = !options->lower_bitops && (options->lower_int64_options & (nir_lower_imul64 | nir_lower_shift64)) == nir_lower_imul64;
   condition_flags[4] = options->lower_mul_2x32_64;
   condition_flags[5] = !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 16);
   condition_flags[6] = !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[7] = !options->lower_iadd_sat;
   condition_flags[8] = !options->lower_uadd_sat;
   condition_flags[9] = options->has_fmulz && !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[10] = !options->has_sdot_4x8;
   condition_flags[11] = !options->has_udot_4x8;
   condition_flags[12] = !options->has_sudot_4x8;
   condition_flags[13] = !options->has_dot_2x16;
   condition_flags[14] = options->lower_flrp16;
   condition_flags[15] = !options->lower_flrp16;
   condition_flags[16] = options->lower_flrp32;
   condition_flags[17] = !options->lower_flrp32;
   condition_flags[18] = options->lower_flrp64;
   condition_flags[19] = !options->lower_flrp64;
   condition_flags[20] = options->lower_ftrunc;
   condition_flags[21] = options->lower_ftrunc || (options->lower_doubles_options & nir_lower_dtrunc);
   condition_flags[22] = options->lower_ffloor;
   condition_flags[23] = (options->lower_ffloor || (options->lower_doubles_options & nir_lower_dfloor)) && !(options->lower_doubles_options & nir_lower_dfract);
   condition_flags[24] = !options->lower_ffloor;
   condition_flags[25] = !options->lower_ffloor && !(options->lower_doubles_options & nir_lower_dfloor);
   condition_flags[26] = options->lower_ffract;
   condition_flags[27] = options->lower_ffract || (options->lower_doubles_options & nir_lower_dfract);
   condition_flags[28] = options->lower_fceil;
   condition_flags[29] = options->lower_ffma16;
   condition_flags[30] = options->lower_ffma32;
   condition_flags[31] = options->lower_ffma64;
   condition_flags[32] = options->fuse_ffma16;
   condition_flags[33] = options->fuse_ffma32;
   condition_flags[34] = options->fuse_ffma64;
   condition_flags[35] = options->lower_fdph;
   condition_flags[36] = !options->lower_fdph;
   condition_flags[37] = options->lower_fdot;
   condition_flags[38] = !options->lower_fsat;
   condition_flags[39] = !options->lower_iabs;
   condition_flags[40] = options->lower_fsat;
   condition_flags[41] = !options->lower_fsat && !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[42] = !options->lower_fsign;
   condition_flags[43] = options->lower_umax;
   condition_flags[44] = options->lower_umin;
   condition_flags[45] = !options->lower_umax;
   condition_flags[46] = !options->lower_umin;
   condition_flags[47] = options->lower_scmp;
   condition_flags[48] = options->lower_vector_cmp;
   condition_flags[49] = !options->lower_rotate;
   condition_flags[50] = options->lower_rotate;
   condition_flags[51] = options->lower_fpow;
   condition_flags[52] = !options->lower_fpow;
   condition_flags[53] = options->lower_fdiv;
   condition_flags[54] = options->lower_fsqrt;
   condition_flags[55] = !options->lower_fsqrt;
   condition_flags[56] = options->lower_sincos;
   condition_flags[57] = !(options->lower_doubles_options & nir_lower_fp64_full_software);
   condition_flags[58] = !options->lower_extract_byte;
   condition_flags[59] = !options->lower_extract_word;
   condition_flags[60] = options->has_pack_32_4x8;
   condition_flags[61] = options->lower_pack_64_2x32_split;
   condition_flags[62] = options->lower_pack_32_2x16_split;
   condition_flags[63] = options->lower_unpack_64_2x32_split;
   condition_flags[64] = options->lower_unpack_32_2x16_split;
   condition_flags[65] = options->has_find_msb_rev;
   condition_flags[66] = !options->lower_find_msb_to_reverse;
   condition_flags[67] = !options->lower_find_lsb;
   condition_flags[68] = options->lower_fmod;
   condition_flags[69] = options->lower_uadd_carry;
   condition_flags[70] = options->lower_usub_borrow;
   condition_flags[71] = options->lower_bitfield_insert;
   condition_flags[72] = options->lower_hadd;
   condition_flags[73] = options->lower_hadd64 || (options->lower_int64_options & nir_lower_iadd64) != 0;
   condition_flags[74] = options->lower_mul_32x16;
   condition_flags[75] = options->lower_uadd_sat || (options->lower_int64_options & nir_lower_iadd64) != 0;
   condition_flags[76] = options->lower_uadd_sat;
   condition_flags[77] = options->lower_usub_sat;
   condition_flags[78] = (options->lower_int64_options & nir_lower_usub_sat64) != 0;
   condition_flags[79] = (options->lower_int64_options & nir_lower_iadd_sat64) != 0;
   condition_flags[80] = (options->lower_int64_options & nir_lower_minmax64) != 0;
   condition_flags[81] = (options->lower_int64_options & nir_lower_icmp64) != 0;
   condition_flags[82] = options->lower_bitfield_insert_to_shifts;
   condition_flags[83] = options->lower_bitfield_insert_to_bitfield_select;
   condition_flags[84] = options->lower_bitfield_extract;
   condition_flags[85] = options->lower_bitfield_insert_to_bitfield_select || options->lower_bitfield_insert;
   condition_flags[86] = options->lower_bitfield_extract_to_shifts;
   condition_flags[87] = options->lower_ifind_msb;
   condition_flags[88] = options->lower_find_msb_to_reverse;
   condition_flags[89] = options->lower_uclz;
   condition_flags[90] = options->lower_find_lsb;
   condition_flags[91] = options->lower_extract_byte;
   condition_flags[92] = options->lower_extract_word;
   condition_flags[93] = options->lower_pack_unorm_2x16;
   condition_flags[94] = options->lower_pack_unorm_4x8;
   condition_flags[95] = options->lower_pack_snorm_2x16;
   condition_flags[96] = options->lower_pack_snorm_4x8;
   condition_flags[97] = options->lower_unpack_unorm_2x16;
   condition_flags[98] = options->lower_unpack_unorm_4x8;
   condition_flags[99] = options->lower_unpack_snorm_2x16;
   condition_flags[100] = options->lower_unpack_snorm_4x8;
   condition_flags[101] = options->lower_pack_split;
   condition_flags[102] = options->lower_isign;
   condition_flags[103] = !options->lower_isign;
   condition_flags[104] = options->lower_fsign;
   condition_flags[105] = !options->has_imul24;
   condition_flags[106] = !options->has_umul24;
   condition_flags[107] = !options->has_umad24;
   condition_flags[108] = options->has_imul24;
   condition_flags[109] = options->has_umad24;
   condition_flags[110] = options->has_umul24;
   condition_flags[111] = !options->lower_uadd_sat && !(options->lower_int64_options & nir_lower_iadd64);
   condition_flags[112] = !options->lower_usub_sat;
   condition_flags[113] = !options->lower_usub_sat && !(options->lower_int64_options & nir_lower_usub_sat64);
   condition_flags[114] = options->lower_iadd_sat;
   condition_flags[115] = options->support_8bit_alu;
   condition_flags[116] = options->support_16bit_alu;
   condition_flags[117] = options->lower_ldexp;
   condition_flags[118] = !options->lower_bitfield_reverse;
   condition_flags[119] = info->stage == MESA_SHADER_COMPUTE && info->cs.derivative_group == DERIVATIVE_GROUP_NONE;
   condition_flags[120] = options->lower_fisnormal;
   condition_flags[121] = !options->vectorize_vec2_16bit;
   condition_flags[122] = options->has_fsub;
   condition_flags[123] = options->has_fsub && !(options->lower_doubles_options & nir_lower_dsub);
   condition_flags[124] = options->lower_fneg;
   condition_flags[125] = options->has_isub || options->lower_ineg;
   condition_flags[126] = options->lower_ineg;
   condition_flags[127] = options->lower_iabs;
   condition_flags[128] = options->has_iadd3;
   condition_flags[129] = options->vectorize_vec2_16bit;
   condition_flags[130] = options->fdot_replicates;
   condition_flags[131] = options->lower_fround_even;
   condition_flags[132] = (info->stage != MESA_SHADER_VERTEX && info->stage != MESA_SHADER_GEOMETRY) && !options->intel_vec4;
   condition_flags[133] = options->avoid_ternary_with_two_constants;
   condition_flags[134] = !options->lower_insert_byte;
   condition_flags[135] = !options->lower_insert_word;
   condition_flags[136] = options->has_fused_comp_and_csel;

   nir_foreach_function(function, shader) {
      if (function->impl) {
         progress |= nir_algebraic_impl(function->impl, condition_flags,
                                        &nir_opt_algebraic_late_table);
      }
   }

   return progress;
}


#include "nir.h"
#include "nir_builder.h"
#include "nir_search.h"
#include "nir_search_helpers.h"

/* What follows is NIR algebraic transform code for the following 19
 * transforms:
 *    ('fmul', ('fneg', 'a'), ('fneg', 'b')) => ('fmul', 'a', 'b')
 *    ('ffma', ('fneg', 'a'), ('fneg', 'b'), 'c') => ('ffma', 'a', 'b', 'c')
 *    ('fdot2_replicated', ('fneg', 'a'), ('fneg', 'b')) => ('fdot2_replicated', 'a', 'b')
 *    ('fdot3_replicated', ('fneg', 'a'), ('fneg', 'b')) => ('fdot3_replicated', 'a', 'b')
 *    ('fdot4_replicated', ('fneg', 'a'), ('fneg', 'b')) => ('fdot4_replicated', 'a', 'b')
 *    ('fneg', ('fneg', 'a')) => a
 *    ('fneg', ('fmul(is_used_once)', 'a', 'b')) => ('fmul', ('fneg', 'a'), 'b')
 *    ('fabs', ('fmul(is_used_once)', 'a', 'b')) => ('fmul', ('fabs', 'a'), ('fabs', 'b'))
 *    ('fneg', ('ffma(is_used_once)', 'a', 'b', 'c')) => ('ffma', ('fneg', 'a'), 'b', ('fneg', 'c'))
 *    ('fneg', ('flrp(is_used_once)', 'a', 'b', 'c')) => ('flrp', ('fneg', 'a'), ('fneg', 'b'), 'c')
 *    ('fneg', ('~fadd(is_used_once)', 'a', 'b')) => ('fadd', ('fneg', 'a'), ('fneg', 'b'))
 *    ('fneg', ('fmin(is_used_once)', 'a', 'b')) => ('fmax', ('fneg', 'a'), ('fneg', 'b'))
 *    ('fneg', ('fmax(is_used_once)', 'a', 'b')) => ('fmin', ('fneg', 'a'), ('fneg', 'b'))
 *    ('fneg', ('fdot2_replicated(is_used_once)', 'a', 'b')) => ('fdot2_replicated', ('fneg', 'a'), 'b')
 *    ('fneg', ('fdot3_replicated(is_used_once)', 'a', 'b')) => ('fdot3_replicated', ('fneg', 'a'), 'b')
 *    ('fneg', ('fdot4_replicated(is_used_once)', 'a', 'b')) => ('fdot4_replicated', ('fneg', 'a'), 'b')
 *    ('fneg', ('fdph_replicated(is_used_once)', 'a', 'b')) => ('fdph_replicated', 'a', ('fneg', 'b'))
 *    ('fneg', ('fsign(is_used_once)', 'a')) => ('fsign', ('fneg', 'a'))
 *    ('fabs', ('fsign(is_used_once)', 'a')) => ('fsign', ('fabs', 'a'))
 */


static const nir_search_value_union nir_opt_algebraic_distribute_src_mods_values[] = {
   /* ('fmul', ('fneg', 'a'), ('fneg', 'b')) => ('fmul', 'a', 'b') */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 0 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 2 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 1, 3 },
      -1,
   } },

   /* replace2287_0 -> 0 in the cache */
   /* replace2287_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 0, 2 },
      -1,
   } },

   /* ('ffma', ('fneg', 'a'), ('fneg', 'b'), 'c') => ('ffma', 'a', 'b', 'c') */
   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 6 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 8 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 7, 9, 10 },
      -1,
   } },

   /* replace2288_0 -> 6 in the cache */
   /* replace2288_1 -> 8 in the cache */
   /* replace2288_2 -> 10 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 6, 8, 10 },
      -1,
   } },

   /* ('fdot2_replicated', ('fneg', 'a'), ('fneg', 'b')) => ('fdot2_replicated', 'a', 'b') */
   /* search2289_0_0 -> 0 in the cache */
   /* search2289_0 -> 1 in the cache */
   /* search2289_1_0 -> 2 in the cache */
   /* search2289_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot2_replicated,
      0, 1,
      { 1, 3 },
      -1,
   } },

   /* replace2289_0 -> 0 in the cache */
   /* replace2289_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot2_replicated,
      0, 1,
      { 0, 2 },
      -1,
   } },

   /* ('fdot3_replicated', ('fneg', 'a'), ('fneg', 'b')) => ('fdot3_replicated', 'a', 'b') */
   /* search2290_0_0 -> 0 in the cache */
   /* search2290_0 -> 1 in the cache */
   /* search2290_1_0 -> 2 in the cache */
   /* search2290_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot3_replicated,
      0, 1,
      { 1, 3 },
      -1,
   } },

   /* replace2290_0 -> 0 in the cache */
   /* replace2290_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot3_replicated,
      0, 1,
      { 0, 2 },
      -1,
   } },

   /* ('fdot4_replicated', ('fneg', 'a'), ('fneg', 'b')) => ('fdot4_replicated', 'a', 'b') */
   /* search2291_0_0 -> 0 in the cache */
   /* search2291_0 -> 1 in the cache */
   /* search2291_1_0 -> 2 in the cache */
   /* search2291_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot4_replicated,
      0, 1,
      { 1, 3 },
      -1,
   } },

   /* replace2291_0 -> 0 in the cache */
   /* replace2291_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot4_replicated,
      0, 1,
      { 0, 2 },
      -1,
   } },

   /* ('fneg', ('fneg', 'a')) => a */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 19 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 20 },
      -1,
   } },

   /* replace2292 -> 19 in the cache */

   /* ('fneg', ('fmul(is_used_once)', 'a', 'b')) => ('fmul', ('fneg', 'a'), 'b') */
   /* search2293_0_0 -> 0 in the cache */
   /* search2293_0_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 0, 2 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 22 },
      -1,
   } },

   /* replace2293_0_0 -> 0 in the cache */
   /* replace2293_0 -> 1 in the cache */
   /* replace2293_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 1, 2 },
      -1,
   } },

   /* ('fabs', ('fmul(is_used_once)', 'a', 'b')) => ('fmul', ('fabs', 'a'), ('fabs', 'b')) */
   /* search2294_0_0 -> 0 in the cache */
   /* search2294_0_1 -> 2 in the cache */
   /* search2294_0 -> 22 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 1,
      { 22 },
      -1,
   } },

   /* replace2294_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace2294_1_0 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 2 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 26, 27 },
      -1,
   } },

   /* ('fneg', ('ffma(is_used_once)', 'a', 'b', 'c')) => ('ffma', ('fneg', 'a'), 'b', ('fneg', 'c')) */
   /* search2295_0_0 -> 6 in the cache */
   /* search2295_0_1 -> 8 in the cache */
   /* search2295_0_2 -> 10 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 6, 8, 10 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 29 },
      -1,
   } },

   /* replace2295_0_0 -> 6 in the cache */
   /* replace2295_0 -> 7 in the cache */
   /* replace2295_1 -> 8 in the cache */
   /* replace2295_2_0 -> 10 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 10 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 7, 8, 31 },
      -1,
   } },

   /* ('fneg', ('flrp(is_used_once)', 'a', 'b', 'c')) => ('flrp', ('fneg', 'a'), ('fneg', 'b'), 'c') */
   /* search2296_0_0 -> 6 in the cache */
   /* search2296_0_1 -> 8 in the cache */
   /* search2296_0_2 -> 10 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 6, 8, 10 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 33 },
      -1,
   } },

   /* replace2296_0_0 -> 6 in the cache */
   /* replace2296_0 -> 7 in the cache */
   /* replace2296_1_0 -> 8 in the cache */
   /* replace2296_1 -> 9 in the cache */
   /* replace2296_2 -> 10 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 7, 9, 10 },
      -1,
   } },

   /* ('fneg', ('~fadd(is_used_once)', 'a', 'b')) => ('fadd', ('fneg', 'a'), ('fneg', 'b')) */
   /* search2297_0_0 -> 0 in the cache */
   /* search2297_0_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 0, 2 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 36 },
      -1,
   } },

   /* replace2297_0_0 -> 0 in the cache */
   /* replace2297_0 -> 1 in the cache */
   /* replace2297_1_0 -> 2 in the cache */
   /* replace2297_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 1, 3 },
      -1,
   } },

   /* ('fneg', ('fmin(is_used_once)', 'a', 'b')) => ('fmax', ('fneg', 'a'), ('fneg', 'b')) */
   /* search2298_0_0 -> 0 in the cache */
   /* search2298_0_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 0, 2 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 39 },
      -1,
   } },

   /* replace2298_0_0 -> 0 in the cache */
   /* replace2298_0 -> 1 in the cache */
   /* replace2298_1_0 -> 2 in the cache */
   /* replace2298_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 1, 3 },
      -1,
   } },

   /* ('fneg', ('fmax(is_used_once)', 'a', 'b')) => ('fmin', ('fneg', 'a'), ('fneg', 'b')) */
   /* search2299_0_0 -> 0 in the cache */
   /* search2299_0_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 0, 2 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 42 },
      -1,
   } },

   /* replace2299_0_0 -> 0 in the cache */
   /* replace2299_0 -> 1 in the cache */
   /* replace2299_1_0 -> 2 in the cache */
   /* replace2299_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 1, 3 },
      -1,
   } },

   /* ('fneg', ('fdot2_replicated(is_used_once)', 'a', 'b')) => ('fdot2_replicated', ('fneg', 'a'), 'b') */
   /* search2300_0_0 -> 0 in the cache */
   /* search2300_0_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot2_replicated,
      0, 1,
      { 0, 2 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 45 },
      -1,
   } },

   /* replace2300_0_0 -> 0 in the cache */
   /* replace2300_0 -> 1 in the cache */
   /* replace2300_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot2_replicated,
      0, 1,
      { 1, 2 },
      -1,
   } },

   /* ('fneg', ('fdot3_replicated(is_used_once)', 'a', 'b')) => ('fdot3_replicated', ('fneg', 'a'), 'b') */
   /* search2301_0_0 -> 0 in the cache */
   /* search2301_0_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot3_replicated,
      0, 1,
      { 0, 2 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 48 },
      -1,
   } },

   /* replace2301_0_0 -> 0 in the cache */
   /* replace2301_0 -> 1 in the cache */
   /* replace2301_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot3_replicated,
      0, 1,
      { 1, 2 },
      -1,
   } },

   /* ('fneg', ('fdot4_replicated(is_used_once)', 'a', 'b')) => ('fdot4_replicated', ('fneg', 'a'), 'b') */
   /* search2302_0_0 -> 0 in the cache */
   /* search2302_0_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot4_replicated,
      0, 1,
      { 0, 2 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 51 },
      -1,
   } },

   /* replace2302_0_0 -> 0 in the cache */
   /* replace2302_0 -> 1 in the cache */
   /* replace2302_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot4_replicated,
      0, 1,
      { 1, 2 },
      -1,
   } },

   /* ('fneg', ('fdph_replicated(is_used_once)', 'a', 'b')) => ('fdph_replicated', 'a', ('fneg', 'b')) */
   /* search2303_0_0 -> 0 in the cache */
   /* search2303_0_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdph_replicated,
      -1, 0,
      { 0, 2 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 54 },
      -1,
   } },

   /* replace2303_0 -> 0 in the cache */
   /* replace2303_1_0 -> 2 in the cache */
   /* replace2303_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdph_replicated,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* ('fneg', ('fsign(is_used_once)', 'a')) => ('fsign', ('fneg', 'a')) */
   /* search2304_0_0 -> 19 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 19 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 57 },
      -1,
   } },

   /* replace2304_0_0 -> 19 in the cache */
   /* replace2304_0 -> 20 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 20 },
      -1,
   } },

   /* ('fabs', ('fsign(is_used_once)', 'a')) => ('fsign', ('fabs', 'a')) */
   /* search2305_0_0 -> 19 in the cache */
   /* search2305_0 -> 57 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 57 },
      -1,
   } },

   /* replace2305_0_0 -> 19 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 19 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 61 },
      -1,
   } },

};

static const nir_search_expression_cond nir_opt_algebraic_distribute_src_mods_expression_cond[] = {
   (is_used_once),
};


static const struct transform nir_opt_algebraic_distribute_src_mods_transforms[] = {
   { ~0, ~0, ~0 }, /* Sentinel */

   { 23, 24, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 21, 19, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 30, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 46, 47, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 49, 50, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 52, 53, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 35, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 37, 38, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 40, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 43, 44, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 55, 56, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 58, 59, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 25, 28, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 60, 62, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4, 5, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 11, 12, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 13, 14, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 15, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 17, 18, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

};

static const struct per_op_table nir_opt_algebraic_distribute_src_mods_pass_op_table[nir_num_search_ops] = {
   [nir_op_fmul] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         2,
         2,
         2,
         28,
      },
   },
   [nir_op_fneg] = {
      .filter = (const uint16_t []) {
         0,
         0,
         1,
         2,
         3,
         4,
         5,
         6,
         7,
         8,
         9,
         10,
         11,
         12,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         1,
         3,
         4,
         5,
         6,
      },
      
      .num_filtered_states = 13,
      .table = (const uint16_t []) {
      
         3,
         14,
         15,
         16,
         17,
         18,
         19,
         20,
         21,
         22,
         23,
         24,
         25,
      },
   },
   [nir_op_ffma] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         4,
         4,
         4,
         4,
         4,
         4,
         29,
         29,
      },
   },
   [nir_op_fdot2_replicated] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         5,
         5,
         5,
         30,
      },
   },
   [nir_op_fdot3_replicated] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         6,
         6,
         6,
         31,
      },
   },
   [nir_op_fdot4_replicated] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         7,
         7,
         7,
         32,
      },
   },
   [nir_op_fabs] = {
      .filter = (const uint16_t []) {
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         0,
         26,
         27,
      },
   },
   [nir_op_flrp] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         8,
      },
   },
   [nir_op_fadd] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         9,
      },
   },
   [nir_op_fmin] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         10,
      },
   },
   [nir_op_fmax] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         11,
      },
   },
   [nir_op_fdph_replicated] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         12,
      },
   },
   [nir_op_fsign] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         13,
      },
   },
};

/* Mapping from state index to offset in transforms (0 being no transforms) */
static const uint16_t nir_opt_algebraic_distribute_src_mods_transform_offsets[] = {
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   1,
   3,
   5,
   7,
   9,
   11,
   13,
   15,
   17,
   19,
   21,
   23,
   25,
   27,
   29,
   31,
   33,
   35,
   37,
};

static const nir_algebraic_table nir_opt_algebraic_distribute_src_mods_table = {
   .transforms = nir_opt_algebraic_distribute_src_mods_transforms,
   .transform_offsets = nir_opt_algebraic_distribute_src_mods_transform_offsets,
   .pass_op_table = nir_opt_algebraic_distribute_src_mods_pass_op_table,
   .values = nir_opt_algebraic_distribute_src_mods_values,
   .expression_cond = nir_opt_algebraic_distribute_src_mods_expression_cond,
   .variable_cond = NULL,
};

bool
nir_opt_algebraic_distribute_src_mods(nir_shader *shader)
{
   bool progress = false;
   bool condition_flags[137];
   const nir_shader_compiler_options *options = shader->options;
   const shader_info *info = &shader->info;
   (void) options;
   (void) info;

   /* This is not a great place for this, but it seems to be the best place
    * for it. Check that at most one kind of lowering is requested for
    * bitfield extract and bitfield insert. Otherwise the lowering can fight
    * with each other and optimizations.
    */
   assert((int)options->lower_bitfield_extract +
          (int)options->lower_bitfield_extract_to_shifts <= 1);
   assert((int)options->lower_bitfield_insert +
          (int)options->lower_bitfield_insert_to_shifts +
          (int)options->lower_bitfield_insert_to_bitfield_select <= 1);


   STATIC_ASSERT(63 == ARRAY_SIZE(nir_opt_algebraic_distribute_src_mods_values));
   condition_flags[0] = true;
   condition_flags[1] = !options->lower_bitops;
   condition_flags[2] = options->lower_bitops;
   condition_flags[3] = !options->lower_bitops && (options->lower_int64_options & (nir_lower_imul64 | nir_lower_shift64)) == nir_lower_imul64;
   condition_flags[4] = options->lower_mul_2x32_64;
   condition_flags[5] = !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 16);
   condition_flags[6] = !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[7] = !options->lower_iadd_sat;
   condition_flags[8] = !options->lower_uadd_sat;
   condition_flags[9] = options->has_fmulz && !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[10] = !options->has_sdot_4x8;
   condition_flags[11] = !options->has_udot_4x8;
   condition_flags[12] = !options->has_sudot_4x8;
   condition_flags[13] = !options->has_dot_2x16;
   condition_flags[14] = options->lower_flrp16;
   condition_flags[15] = !options->lower_flrp16;
   condition_flags[16] = options->lower_flrp32;
   condition_flags[17] = !options->lower_flrp32;
   condition_flags[18] = options->lower_flrp64;
   condition_flags[19] = !options->lower_flrp64;
   condition_flags[20] = options->lower_ftrunc;
   condition_flags[21] = options->lower_ftrunc || (options->lower_doubles_options & nir_lower_dtrunc);
   condition_flags[22] = options->lower_ffloor;
   condition_flags[23] = (options->lower_ffloor || (options->lower_doubles_options & nir_lower_dfloor)) && !(options->lower_doubles_options & nir_lower_dfract);
   condition_flags[24] = !options->lower_ffloor;
   condition_flags[25] = !options->lower_ffloor && !(options->lower_doubles_options & nir_lower_dfloor);
   condition_flags[26] = options->lower_ffract;
   condition_flags[27] = options->lower_ffract || (options->lower_doubles_options & nir_lower_dfract);
   condition_flags[28] = options->lower_fceil;
   condition_flags[29] = options->lower_ffma16;
   condition_flags[30] = options->lower_ffma32;
   condition_flags[31] = options->lower_ffma64;
   condition_flags[32] = options->fuse_ffma16;
   condition_flags[33] = options->fuse_ffma32;
   condition_flags[34] = options->fuse_ffma64;
   condition_flags[35] = options->lower_fdph;
   condition_flags[36] = !options->lower_fdph;
   condition_flags[37] = options->lower_fdot;
   condition_flags[38] = !options->lower_fsat;
   condition_flags[39] = !options->lower_iabs;
   condition_flags[40] = options->lower_fsat;
   condition_flags[41] = !options->lower_fsat && !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[42] = !options->lower_fsign;
   condition_flags[43] = options->lower_umax;
   condition_flags[44] = options->lower_umin;
   condition_flags[45] = !options->lower_umax;
   condition_flags[46] = !options->lower_umin;
   condition_flags[47] = options->lower_scmp;
   condition_flags[48] = options->lower_vector_cmp;
   condition_flags[49] = !options->lower_rotate;
   condition_flags[50] = options->lower_rotate;
   condition_flags[51] = options->lower_fpow;
   condition_flags[52] = !options->lower_fpow;
   condition_flags[53] = options->lower_fdiv;
   condition_flags[54] = options->lower_fsqrt;
   condition_flags[55] = !options->lower_fsqrt;
   condition_flags[56] = options->lower_sincos;
   condition_flags[57] = !(options->lower_doubles_options & nir_lower_fp64_full_software);
   condition_flags[58] = !options->lower_extract_byte;
   condition_flags[59] = !options->lower_extract_word;
   condition_flags[60] = options->has_pack_32_4x8;
   condition_flags[61] = options->lower_pack_64_2x32_split;
   condition_flags[62] = options->lower_pack_32_2x16_split;
   condition_flags[63] = options->lower_unpack_64_2x32_split;
   condition_flags[64] = options->lower_unpack_32_2x16_split;
   condition_flags[65] = options->has_find_msb_rev;
   condition_flags[66] = !options->lower_find_msb_to_reverse;
   condition_flags[67] = !options->lower_find_lsb;
   condition_flags[68] = options->lower_fmod;
   condition_flags[69] = options->lower_uadd_carry;
   condition_flags[70] = options->lower_usub_borrow;
   condition_flags[71] = options->lower_bitfield_insert;
   condition_flags[72] = options->lower_hadd;
   condition_flags[73] = options->lower_hadd64 || (options->lower_int64_options & nir_lower_iadd64) != 0;
   condition_flags[74] = options->lower_mul_32x16;
   condition_flags[75] = options->lower_uadd_sat || (options->lower_int64_options & nir_lower_iadd64) != 0;
   condition_flags[76] = options->lower_uadd_sat;
   condition_flags[77] = options->lower_usub_sat;
   condition_flags[78] = (options->lower_int64_options & nir_lower_usub_sat64) != 0;
   condition_flags[79] = (options->lower_int64_options & nir_lower_iadd_sat64) != 0;
   condition_flags[80] = (options->lower_int64_options & nir_lower_minmax64) != 0;
   condition_flags[81] = (options->lower_int64_options & nir_lower_icmp64) != 0;
   condition_flags[82] = options->lower_bitfield_insert_to_shifts;
   condition_flags[83] = options->lower_bitfield_insert_to_bitfield_select;
   condition_flags[84] = options->lower_bitfield_extract;
   condition_flags[85] = options->lower_bitfield_insert_to_bitfield_select || options->lower_bitfield_insert;
   condition_flags[86] = options->lower_bitfield_extract_to_shifts;
   condition_flags[87] = options->lower_ifind_msb;
   condition_flags[88] = options->lower_find_msb_to_reverse;
   condition_flags[89] = options->lower_uclz;
   condition_flags[90] = options->lower_find_lsb;
   condition_flags[91] = options->lower_extract_byte;
   condition_flags[92] = options->lower_extract_word;
   condition_flags[93] = options->lower_pack_unorm_2x16;
   condition_flags[94] = options->lower_pack_unorm_4x8;
   condition_flags[95] = options->lower_pack_snorm_2x16;
   condition_flags[96] = options->lower_pack_snorm_4x8;
   condition_flags[97] = options->lower_unpack_unorm_2x16;
   condition_flags[98] = options->lower_unpack_unorm_4x8;
   condition_flags[99] = options->lower_unpack_snorm_2x16;
   condition_flags[100] = options->lower_unpack_snorm_4x8;
   condition_flags[101] = options->lower_pack_split;
   condition_flags[102] = options->lower_isign;
   condition_flags[103] = !options->lower_isign;
   condition_flags[104] = options->lower_fsign;
   condition_flags[105] = !options->has_imul24;
   condition_flags[106] = !options->has_umul24;
   condition_flags[107] = !options->has_umad24;
   condition_flags[108] = options->has_imul24;
   condition_flags[109] = options->has_umad24;
   condition_flags[110] = options->has_umul24;
   condition_flags[111] = !options->lower_uadd_sat && !(options->lower_int64_options & nir_lower_iadd64);
   condition_flags[112] = !options->lower_usub_sat;
   condition_flags[113] = !options->lower_usub_sat && !(options->lower_int64_options & nir_lower_usub_sat64);
   condition_flags[114] = options->lower_iadd_sat;
   condition_flags[115] = options->support_8bit_alu;
   condition_flags[116] = options->support_16bit_alu;
   condition_flags[117] = options->lower_ldexp;
   condition_flags[118] = !options->lower_bitfield_reverse;
   condition_flags[119] = info->stage == MESA_SHADER_COMPUTE && info->cs.derivative_group == DERIVATIVE_GROUP_NONE;
   condition_flags[120] = options->lower_fisnormal;
   condition_flags[121] = !options->vectorize_vec2_16bit;
   condition_flags[122] = options->has_fsub;
   condition_flags[123] = options->has_fsub && !(options->lower_doubles_options & nir_lower_dsub);
   condition_flags[124] = options->lower_fneg;
   condition_flags[125] = options->has_isub || options->lower_ineg;
   condition_flags[126] = options->lower_ineg;
   condition_flags[127] = options->lower_iabs;
   condition_flags[128] = options->has_iadd3;
   condition_flags[129] = options->vectorize_vec2_16bit;
   condition_flags[130] = options->fdot_replicates;
   condition_flags[131] = options->lower_fround_even;
   condition_flags[132] = (info->stage != MESA_SHADER_VERTEX && info->stage != MESA_SHADER_GEOMETRY) && !options->intel_vec4;
   condition_flags[133] = options->avoid_ternary_with_two_constants;
   condition_flags[134] = !options->lower_insert_byte;
   condition_flags[135] = !options->lower_insert_word;
   condition_flags[136] = options->has_fused_comp_and_csel;

   nir_foreach_function(function, shader) {
      if (function->impl) {
         progress |= nir_algebraic_impl(function->impl, condition_flags,
                                        &nir_opt_algebraic_distribute_src_mods_table);
      }
   }

   return progress;
}

